# "@(#) $Id: ACMM:2dFutil/2dFsounds.tcl,v 5.2 10-Sep-2013 09:12:12+10 tjf $"
#
#
# A simple script which provides commands which uses the
# NetworkSound sub-system Tcl commands to play appropiate 2dF
# sounds.  Presumes the NetworkSound system is running.
#
# Define the sound for each event.  Each entry value is a list containing the
# sound name, the bell alternative (1/2/3/4/3s/4s) and the volume (0-100)
#
namespace eval 2dFsound {
    variable sounds
    set sounds(FinishedConfigure)  "\$NETWORKSOUND_DIR/sounds/gong.wav 3 50"
    set sounds(FinishedInitialise) "\$NETWORKSOUND_DIR/sounds/buzzer1.wav 4 50"
    set sounds(PositionerError)    "\$NETWORKSOUND_DIR/sounds/firealrm.wav 2 50"
    set sounds(ExposureComplete)   "\$NETWORKSOUND_DIR/sounds/siren.wav 2 50"
    set sounds(PositionFibre)      "\$NETWORKSOUND_DIR/sounds/chirp1.wav 1 80"
    set sounds(FinishedTumble)     "\$NETWORKSOUND_DIR/sounds/zapperoo.wav 2 50"
    set sounds(Shutdown)           "\$NETWORKSOUND_DIR/sounds/Snicker.wav 1 50"
    set sounds(GeneralError)       "\$NETWORKSOUND_DIR/sounds/danger.wav 1 50"
    set sounds(ReadoutStart)       "\$NETWORKSOUND_DIR/sounds/cuckoo.wav 1 50"

    set reason(FinishedConfigure)  "Sound played when fibre configuration completes. "
    set reason(FinishedInitialise) "Sound played when system initialise completes. "
    set reason(PositionerError)    "Sound played when a fibre positioner error occurs."
    set reason(ExposureComplete)   "Sound played when the CCD run has completed."
    set reason(PositionFibre)      "Sound played each time a fibre is positioned."
    set reason(FinishedTumble)     "Sound played when a tumble has completed. (CT Only)"
    set reason(Shutdown)           "Sound played at the end of system shutdown."
    set reason(GeneralError)       "Sound played for errors (other then positioner errors)."
    set reason(ReadoutStart)       "Sound played at start of last readout."
    #
    # A command which plays one of the specified sounds.
    #
    proc Play { which } {
        variable sounds
        set sound   [lindex $sounds($which) 0]
        set bellAlt [lindex $sounds($which) 1]
        set volume  [lindex $sounds($which) 2]
        NetworkSound::Play $volume $sound $bellAlt
    }

    #
    # Add a cacasde menu to the specified menu window which can be
    # used to control the 2dF sound.
    #
    proc AddCtrlMenu { w } {
        $w add cascade -label "Sound control" -menu $w.soundCtrl
        menu $w.soundCtrl
        $w.soundCtrl add checkbutton -label "Sound On" \
            -variable NetworkSound::SoundOn
        
        # The command spec below is because if it may not be possible to
        # change this value, but the checkbutton won't pick that fact up
        # by itself - the after command ensured the checkbutton is right.
        $w.soundCtrl add checkbutton -label "Bell Only" \
            -variable NetworkSound::BellOnly \
            -command {after 50 set NetworkSound::BellOnly $NetworkSound::BellOnly}
    
        $w.soundCtrl add command -label "Volume" -command 2dFsound::SetVolume

        $w.soundCtrl add command -label "Demo Sounds" -command 2dFsound::DemoSounds
    
    }

    proc SetVolume { } {

        set w .2dFsoundSetVolume
        set NotCreated [catch "wm state $w" state]
        if { $NotCreated } {
            toplevel $w
            wm title $w "2dF Volume Control"
            wm iconname $w 2dFvolume
            wm minsize  $w 5 5
            wm maxsize  $w 1000 800
            
            label $w.title -text "Please set the sound volume"
            scale $w.scale -label "Sound Volume %" -from 0 -to 100 -length 3i \
                -tickinterval 20 -orient horizontal \
                -variable NetworkSound::BaseVolume \
                -command {NetworkSound::Play 50 \$NETWORKSOUND_DIR/sounds/dougheee.wav 2 ;#}
            button $w.close -text "Close" -command "destroy $w"
            
            pack $w.title $w.scale $w.close -side top -expand 1 -fill x
            
            
        }
        
    }

    proc DemoSounds { } {
        variable sounds 
        variable reason
 
        set w .2dFsoundDemo
        set NotCreated [catch "wm state $w" state]
        if { $NotCreated } {
            toplevel $w
            wm title $w "2dF Sound Demo"
            wm iconname $w 2dFvolume
            wm minsize  $w 5 5
            wm maxsize  $w 1000 800

            label $w.lab1 -text "Press buttons to play sounds" -anchor w
            label $w.lab2 -text "Engineering interface is still inconsistent" -anchor w
            pack $w.lab1 $w.lab2 -side top -fill x -expand 1
       
            set soundNames [array names sounds]
            foreach i $soundNames {
                #puts "Sounds $i"
                button $w.but_$i -text $reason($i) -command "2dFsound::Play $i" -width 50
                pack $w.but_$i -side top -fill x -expand 1
            }
 
            if { [info command DtcluDialogPosSet] != "" } {
                DtcluDialogPosSet $w
            }
        } elseif { [info command DtcluDialogPosReset] != "" } {
            DtcluDialogPosReset $w
        } else {
            #       Deiconify and raise the window.
            wm deiconify $w
            raise $w
            
        }
        
        
    }

    #
    # A simple test of the 2dF sounds.  Can be run with
    #
    #   $NETWORKSOUND_DEV/dtknas -stdin -t $TDF_UTIL_DIR/2dFsounds.tcl
    #   % 2dFsound::Test
    #
    #   Which will create a window with a set of buttons which can be used
    #   to test the 2dF sounds.
    #
    #
    proc Test { } {
        variable sounds
        menubutton .menubut -text "Menu Test" -menu .menubut.menu
        menu .menubut.menu
        pack .menubut -side top -fill x -expand 1
        set soundNames [array names sounds]
        foreach i $soundNames {
            #puts "Sounds $i"
            button .but_$i -text $i -command "2dFsound::Play $i"
            pack .but_$i -side top -fill x -expand 1
        }
        2dFsound::AddCtrlMenu .menubut.menu
    }



} ;# namespace


