/*+                T D F U T I L

 *  Module name:
      tdFcollision

 *  Function:
      Contains functions used to detect for fibre/fibre, fibre/button+jaw, 
      and button/button collisions.

 *  Description:
      Functions are those written by JOS, but optimised for our needs (ie, 
      some code is removed from each code as we do not need to check for 
      things such as "could a collision ever occur regardless of the 
      orientation of the button").

 *  Language:
      C

 *  Support: James Wilcox, AAO

 *-

 *  History:
      01-Jul-1994  JW   Original version
      14-Dec-1999  TJF  Simpilfy and document the MAXMIN macro.
      03-Nov-2006  KS   Added LineGeometry call to button-fibre calculation.
      19-Aug-2009  KS   Turned MAXMIN into a function to get rid of a host of
                        compiler warnings.
      {@change entry@}

 *  @(#) $Id: ACMM:2dFutil/tdFcollision.c,v 5.2 10-Sep-2013 09:12:11+10 tjf $ (mm/dd/yy)
 */

/*
 *  Include files
 */


static char *rcsId="@(#) $Id: ACMM:2dFutil/tdFcollision.c,v 5.2 10-Sep-2013 09:12:11+10 tjf $";
static void *use_rcsId = (0 ? (void *)(&use_rcsId) : (void *) &rcsId);


#include <math.h>
#include "tdFutil.h"
#include "tdFbuttonDims.h"
#include "tdFfieldDims.h"

/*
 *  Definitions
 */
#define  PI   3.1415926535
#define  NO              0
#define  YES             1

/*
 *  Macros.
 */
#define  SQRD(x)  ((x)*(x))

/*
 *  MAXMIN used to be defined as follows: 
 *  MACRO to put the max of a and b in c and the min in d.
 *  (and return the max value).
 *
 *  #define  MAXMIN(a,b,c,d) ((c=a)>(d=b)?(c):(d=a,c=b))
 * 
 *  but it generated loads of warnings, because it would end up
 *  expanding into code like xmax = xmax = some_value
 *  which gcc doesn't think highly of.
 *
 *  So it's been redefined here as a function that returns the
 *  maximum of two values and sets the min and max into passed
 *  variables. The references to MAXMIN in the code have been
 *  modified accordingly - it just needed '&'s in front of the
 *  last two arguments in each case. It may be slower, but it
 *  doesn't generate all those worrying warnings. KS.
 */

double MAXMIN (double a, double b, double* maxab, double* minab) {
   if (a > b) { *maxab = a; *minab = b; return a; }
   else { *maxab = b; *minab = a; return b; }
}

/*
 *  Macro used to tag parameters as known to be unused and hence
 * prevent warnings when compiled with strict warning under GNU C.
 */
#ifndef DUNUSED
#if defined(__GNUC__) || defined(GNUC)
#if __GNUC_MINOR__ >= 7 || __GNUC__ > 2
#       define DUNUSED __attribute__ ((unused))
#else
#       define DUNUSED
#endif
#endif
#endif

/*  Internal routine declarations */

double DistPointSegSq ( /* Returns square of shortest line-segment distance */
   double Px,           /* X-coordinate of point */
   double Py,           /* Y-coordinate of point */
   double S1x,          /* X-coordinate of one end of line segment */
   double S1y,          /* Y-coordinate of one end of line segment */
   double S2x,          /* X-coordinate of other end of line segment */
   double S2y);         /* Y-coordinate of other end of line segment */


/*+        T D F C O L L I S I O N

 *  Function name:
      tdFcollisionFibFib

 *  Function:
      Checks for a collision in the xy plane between two fibres - fibre 
      A and fibre B.

      If a collision is detected the function will return YES.
      If no crossover is detected the function will return NO.

 *  Description:
      From JOS collision detection code used during 2dF Configure code, 
      but optimised for our needs.

 *  Language:
      C

 *  Call:
      (int) = tdFcollisionFibFib (pivAX,pivAY,fvpAX,fvpAY,pivBX,pivBY,fvpBX,fvpBY)

 *  Parameters:  (">" input, "!" modified, "W" workspace, "<" output)
      (>) pivAX, pivAY    (double)    Fibre A pivot coordinates.
      (>) fvpAX, fvpAY    (double)    Fibre A virtual pivot coordinates.
      (>) pivBX, pivBY    (double)    Fibre B pivot coordinates.
      (>) fvpBX, fvpBY    (double)    Fibre B virtual pivot coordinates.

 *  Proir Requirements:

 *  Support: James Wilcox, AAO

 *-

 *  History:
      01-Jul-1994  JW   Original version
      {@change entry@}
 */
TDFUTIL_INTERNAL int  tdFcollisionFibFib (
        double  xapiv,   /* fibre A - pivot x coord  */
        double  yapiv,   /*         - pivot y coord  */
        double  xa,      /*         - fvp x coord    */
        double  ya,      /*         - fvp y coord    */
        double  xbpiv,   /* fibre B - pivot x coord  */
        double  ybpiv,   /*         - pivot y coord  */
        double  xb,      /*         - fvp x coord    */
        double  yb)      /*         - fvp y coord    */
{
    double  dxa, dya;
    double  ssa, sa;
    double  sina, cosa;
    double  xtb, ytb;
    double  xpb, ypb;
    double  xBr, yBr;
    double  xBPr, yBPr;

    /*
     *  Probe A fibre direction.
     */
    dxa  = xapiv - xa;
    dya  = yapiv - ya;
    ssa  = dxa*dxa + dya*dya;
    sa   = sqrt(ssa);
    cosa = dxa/sa;
    sina = dya/sa;

    /*
     *  Probe B fibre coordinates.
     */
    xtb = xb - xa;
    ytb = yb - ya;
    xpb = xbpiv - xa;
    ypb = ybpiv - ya;

    /*
     *  Rotate fibre B coordinates.
     */
    xBr  = xtb*cosa + ytb*sina;
    yBr  = ytb*cosa - xtb*sina;
    xBPr = xpb*cosa + ypb*sina;
    yBPr = ypb*cosa - xpb*sina;

    /*
     *  Check if fibres cross.
     */
    if (yBr*yBPr > 0.0)                      return (NO);
    if (xBr > 0.0 && xBPr > 0.0)             return (YES);
    if (xBr < 0.0 && xBPr < 0.0)             return (NO);
    if (xBr > yBr*(xBPr - xBr)/(yBPr - yBr)) return (YES);
    return(NO);
}


/*+        T D F C O L L I S I O N

 *  Function name:
      tdFcollisionButFib

 *  Function:
      Checks for a collision between a fibre and a button and gripper jaws.

 *  Description:
      From JOS collision detection code used during 2dF Configure code, but optimised
      for our needs.

 *  Language:
      C

 *  Call:
      (int) = tdFcollisionButFib (butX,butY,theta,fvpX,fvpY,pivX,pivY,fibreClear)

 *  Parameters:  (">" input, "!" modified, "W" workspace, "<" output)
      (>) butX, butY , theta  (double)    Button coordinates.
      (>) fvpX, fvpY          (double)    Coordinates of fibre virtual pivot point.
      (>) pivX, pivY          (double)    Coordinates of fibre at pivot.
      (>) fibreClear          (long int)  Clearance allowed around a fibre.

 *  Proir Requirements:

 *  Support: James Wilcox, AAO

 *-

 *  History:
      01-Jul-1994  JW   Original version
      03-Nov-2006  KS   Added DistPointSegSq() call. Reorganised multiple
                        returns so there is one exit point where diagnostics
                        (such as did DistPointSegSq() ever give the wrong
                        answer) can be inserted. 
      {@change entry@}
 */
TDFUTIL_INTERNAL int  tdFcollisionButFib (
        double    butX,                  /* Fibre-end - x ordinate            */
        double    butY,                  /*           - y ordinate            */
        double    theta,                 /*           - theta                 */
        double    fvpX,                  /* Fibre - x at button end           */
        double    fvpY,                  /*       - y at button end           */
        double    pivX,                  /*       - x at pivot                */
        double    pivY,                  /*       - y at pivot                */
        long int  fibreClear)            /* Clearance around fibre            */
{
    double  sinta, costa, sinfb, cosfb,   /* Button orientations               */
            sina,  cosa,  asina, acosa;
    double  dxb, dyb, dsb;                /* Direction of fibre                */
    double  xas, yas, xsr, ysr;           /* Shifted and rotated coordinates   */
    double  dxple, dyple,                 /* Contributions from button coords. */
            dxfvp, dyfvp,
            dxble, dyble,
            dxbte, dybte,
            dxphw, dyphw,
            dxbhw, dybhw,
            dxthw, dythw;
    double  clearance;
    double  xmin, xmax, ymin, ymax;       /* Used to detect overlapping        */
    double  safeRadSq;
    int     result = NO;                  /* Result returned */

    /*
     *  Minimum allowed clearance to centre of fibre.
     */
    clearance = FIBRE_DIA*0.5 + fibreClear;
    safeRadSq = (BUTTON_FVP + BUTTON_BHW + clearance) * 
                               (BUTTON_FVP + BUTTON_BHW + clearance);

    /*  One very quick test first. Calculate the distance from the
     *  button to the fibre. If this is more than the radius of a
     *  given safe circle (a very conservative estimate, here) then
     *  there cannot be a collision.
     */
      
    if (DistPointSegSq(butX,butY,pivX,pivY,fvpX,fvpY) > safeRadSq) {
       result = NO;
       goto Exit;
    }

    /*
     *  Orientation of button (convert field theta to local theta).
     */
    sinta = sin(theta+(PI/2));
    costa = cos(theta+(PI/2));

    /*
     *  Direction of fibre.
     */
    dxb = pivX - fvpX;
    dyb = pivY - fvpY;
    dsb = sqrt(dxb*dxb + dyb*dyb);
    sinfb = dyb/dsb;
    cosfb = dxb/dsb;

    /*
     *  Coordinates of button in fibre coordinate system.
     */
    xas = butX - fvpX;
    yas = butY - fvpY;
    xsr = xas*cosfb + yas*sinfb;
    ysr = yas*cosfb - xas*sinfb;

    /*
     *  Orientation of button in fibre coordinatee system.
     */
    sina = sinta*cosfb - costa*sinfb;
    cosa = costa*cosfb + sinta*sinfb;
    asina = fabs(sina);
    acosa = fabs(cosa);

    /*
     *  Contributions from extreme parts of button.
     */
    dxple = BUTTON_PLE*cosa;     dyple = BUTTON_PLE*sina;
    dxfvp = BUTTON_FVP*cosa;     dyfvp = BUTTON_FVP*sina;
    dxbhw = BUTTON_BHW*asina;    dybhw = BUTTON_BHW*acosa;

    /*
     *  Check to see if collision is likely.
     */
    xmax = xsr + MAXMIN(dxple,dxfvp,&xmax,&xmin) + dxbhw + clearance;
    ymax = ysr + MAXMIN(dyple,dyfvp,&ymax,&ymin) + dybhw + clearance;
    ymin += ysr - dybhw - clearance;

    /*
     *  Using Straede's conjecture, a collision is likely if xmax>0 and 
     *  ymax>0 and ymin<0. Check this, and if no collision is possible 
     *  return "NO".
     */
    if (!(xmax>0.0 && ymax>0.0 && ymin<0.0)) {
       result = NO;
       goto Exit;
    }

    /*
     *  Check to see if there is a collision with the button footprint.
     */
    dxble = BUTTON_BLE*cosa;     dyble = BUTTON_BLE*sina;
    dxbte = BUTTON_BTE*cosa;     dybte = BUTTON_BTE*sina;
    xmax = xsr + MAXMIN(dxble,dxbte,&xmax,&xmin) + clearance;
    ymax = ysr + MAXMIN(dyble,dybte,&ymax,&ymin) + clearance;
    ymin += ysr - clearance;
    if (xmax>-dxbhw && ymax>-dybhw && ymin<dybhw) {
       result = YES;
       goto Exit;
    }

    /*
     *  Check if fibre collides with button tail.
     */
    dxfvp = BUTTON_FVP*cosa;      dyfvp = BUTTON_FVP*sina;
    dxthw = BUTTON_THW*asina;     dythw = BUTTON_THW*acosa;
    xmax = xsr + MAXMIN(dxbte,dxfvp,&xmax,&xmin) + clearance;
    ymax = ysr + MAXMIN(dybte,dyfvp,&ymax,&ymin) + clearance;
    ymin += ysr - clearance;
    if (xmax>-dxthw && ymax>-dythw && ymin<dythw) {
       result = YES;
       goto Exit;
    }

    /*
     *  Check if fibre collides with prism.
     */
    dxphw = BUTTON_PHW*asina;     dyphw = BUTTON_PHW*acosa;
    xmax = xsr + MAXMIN(dxple,dxble,&xmax,&xmin) + clearance;
    ymax = ysr + MAXMIN(dyple,dyble,&ymax,&ymin) + clearance;
    ymin += ysr - clearance;
    if (xmax>-dxphw && ymax>-dyphw && ymin<dyphw) {
       result = YES;
       goto Exit;
    }

    /*
     *  If reached this far there is no button/fibre collision possible.
     */
    result = NO;
    
Exit:;
    return (result);

}


/*+        T D F C O L L I S I O N

 *  Function name:
      tdFcollisionButBut

 *  Function:
      Checks for a collision between two buttons.

      If a collision is detected the function will return YES.
      If no crossover is detected the function will return NO.

 *  Description:
      From JOS collision detection code used during 2dF Configure code, 
      but optimised for our needs.

      Straede's conjecture - Two rectangles overlap if

          1.  Rotate Rectangles such that  A is aligned with X axis.  
              There must be overlap in either X or Y directions
          2.  Rotate Rectangles such that B is aligned with the X axis.
              There must be overlap in either X or Y directions


 *  Language:
      C

 *  Call:
      (int) = tdFcollisionButBut (butXa,butYa,thetaa,butXb,butYb,thetab,buttClear)

 *  Parameters:  (">" input, "!" modified, "W" workspace, "<" output)
      (>) butXa, butYa , thetaa  (double)    Button A coordinates.
      (>) butXb, butYb , thetab  (double)    Button B coordinates.
      (>) buttClear              (long int)  Clearance around button.

 *  Proir Requirements:
      Angles supplied must be in the 2dF coordinate system (ie, theta = 0 
      at 12o'clock, increasing anti-clockwise and in radians).  Also note 
      that the function parameters are doubles and not integers.

 *  Support: James Wilcox, AAO

 *-

 *  History:
      01-Jul-1994  JW   Original version
      18-Jun-1999  TJF  Was using BUTTON_CLEAR macro instead of buttClear
			argument in all cases but the first.  Fixed this.
      14-Dec-1999  TJF  Improve documentation.
      11-Jun-2010  TJF  Add some comments on rectangle calcuation, obtained
                          from a trying to understand the code.
      {@change entry@}
 */
TDFUTIL_INTERNAL int  tdFcollisionButBut (
        double    butXa,                /* Button A - x ordinate            */
        double    butYa,                /*          - y ordinate            */
        double    thetaa,               /*          - theta                 */
        double    butXb,                /* Button B - x ordinate            */
        double    butYb,                /*          - y ordinate            */
        double    thetab,               /*          - theta                 */
        long int  buttClear)            /* Clearance around buttons         */
{
    double  sinta, costa, sintb, costb;  /* Button orientations                         */
    double  xah, yah, xas, yas,          /* A coordinates rotated/shifted to A orient'n */
            xbh, ybh, xbs, ybs;          /* B coordinates rotated/shifted to A orient'n */
    double  sinbma, cosbma, asin,acos;   /* Orientation and magnitude                   */
    double                    /* X values for button B rectangles in sys A   */
            axl,                         /* Leading edge of button B in 
                                            coorindate system of button A */ 
            axt,                         /* Trailing edge of button B in
                                            coorindate systme of button B */
            axbl, 
            axbt;       

    double                  /* Max and min values for button B rectangles..*/
                            /* ..in system A                               */
            axp,            /* Right most X point of B in system A         */
            axm,            /* Left most  X point of B in system A         */
            abxp, 
            abxm,
            apxp, 
            apxm, 
            atxp, 
            atxm;
    double  ayl, ayt, aybl, aybt;        /* Button B rectangle in system A              */
    double  ayp, aym, abyp, abym,        /* Max and min y values for butt B in sys A    */
            apyp, apym, atyp, atym;
    double  bxl, bxt, bxbl, bxbt;        /* X values for button A rectangle in system B */
    double  bxp, bxm, bbxp, bbxm,        /* Max and min x values for butt A rectangles..*/
            bpxp, bpxm, btxp, btxm;      /* ..in system B                               */
    double  byl, byt, bybl, bybt;        /* Y values for button A rectangle in system B */
    double  byp, bym, bbyp, bbym,        /* Max and min y values for butt A in system B */
            bpyp, bpym, btyp, btym;
    double  xpl, ypl, xtt, ytt,          /* Coordinates of centre-line of button        */
            xbl, ybl, xbt, ybt;
    double  dxp, dyp, dxt, dyt;          /* Prism and tail halfwidths                   */
    double  dxb, dyb;

    /*
     *  Align plate theta axis (theta = 0 at 12 o'clock) and theta axis used in
     *  these calculations (theta = 0 at 3 o'clock).
     */
    thetaa += PI/2;
    thetab += PI/2;

    /*
     *  Rotate the coordinate system to make button A "horizontal" and shift 
     *  the coordinate system to make the prism of A the origin.
     */
    sinta = sin(thetaa);                costa = cos(thetaa);
    /* Rotate A to align with theta = 0 */
    xah = butXa*costa + butYa*sinta;    yah = butYa*costa - butXa*sinta;
    /* Rotate B to mmatch */
    xbh = butXb*costa + butYb*sinta;    ybh = butYb*costa - butXb*sinta;
    /* xbs/ybs becomes the location of rotated B, such that A is at origin */
    xbs = xbh - xah;                    ybs = ybh - yah ;

    /*
     *  Orientation of B in coordinate system of A.
     */
    sintb  = sin(thetab);               costb  = cos(thetab) ;
    sinbma = sintb*costa - costb*sinta; cosbma = costa*costb + sinta*sintb ;
    asin   = fabs(sinbma);              acos   = fabs(cosbma) ;

    /*
     *  Use Straede's conjecture to determine if a collision between the two 
     *  buttons is likely by treating each button as a rectangle of full 
     *  length and width.
     *
     * BUTTON_PLE = left most part of button,
     * BUTTON_FVP = right most part of button.
     * BUTTON_BWH = button half width.
     */

    /* xpl is the left most part of button B rotated */
    xpl = BUTTON_PLE*cosbma;            
    /* xtt is the right most part of button B, rotated */
    xtt = BUTTON_FVP*cosbma;
    /* get the x component of the rotated button width, add in clearance */
    dxb = BUTTON_BHW*asin + buttClear;
    /* Add the origin of the button */
    axl = xbs + xpl;
    axt = xbs + xtt;
    /* determine the max/min of axl and axt.  Add origin to b. */
    axp = MAXMIN(axl, axt, &axp, &axm) + dxb;
    
    axm -= dxb;
    /*
     *  Check if button A is totally to the left or right of button B. 
     *  IF yes, then no collision is possible.  If not, we must do more checks.
     */
    if (BUTTON_PLE<axp && axm<BUTTON_FVP) {
        /* Similar calculations to the above, but now in y */
        ypl = BUTTON_PLE*sinbma;
        ytt = BUTTON_FVP*sinbma;
        dyb = BUTTON_BHW*acos + buttClear;
        ayl = ybs + ypl;
        ayt = ybs + ytt;
        ayp = MAXMIN(ayl, ayt, &ayp, &aym) + dyb;
        aym -= dyb;
    }
    else
        return (NO);

    /*
     *  Check if button A is totally above or belowbutton B. 
     *  If yes, then no collision is possible.  If not, we must do more checks.
     *
     *  The next set of checks are based on rotating the buttons such
     *  that button B is alligned on the X axis rather then button A.
     */
    if (-BUTTON_BHW<ayp && aym<BUTTON_BHW) {
        /*
         *  The next set of checks are based on rotating the buttons such
         *  that button B is alligned on the X axis rather then button A.
         */
        xas = -xbs*cosbma - ybs*sinbma;
        bxl = xas + xpl;
        bxt = xas + xtt;
        bxp = MAXMIN(bxl, bxt, &bxp, &bxm) + dxb;
        bxm -= dxb;
    }
    else
        return (NO);
    /*
     *  X overlap check.
     */
    if (BUTTON_PLE<bxp && bxm<BUTTON_FVP) {
        yas = -ybs*cosbma + xbs*sinbma;
        byl = yas - ypl;
        byt = yas - ytt;
        byp = MAXMIN(byl, byt, &byp, &bym) + dyb;
        bym -= dyb;
    }
    else
        return (NO);

    /*
     * Y overlap check.
     */
    if (!(-BUTTON_BHW<byp && bym<BUTTON_BHW))
        return (NO);

    /*
     *  If reached here, it is likely that there is a button/button collision.
     *  We now check each component of button A against each component of 
     *  button B to determine if there is a collision. (There are nine sets 
     *  of possible collisions) using the same algrothim as used above.
     */
    xbl = BUTTON_BLE*cosbma;                         ybl = BUTTON_BLE*sinbma;
    xbt = BUTTON_BTE*cosbma;                         ybt = BUTTON_BTE*sinbma;
    axbl = xbs + xbl;                                aybl = ybs + ybl;
    axbt = xbs + xbt;                                aybt = ybs + ybt;
    abxp = MAXMIN(axbl, axbt, &abxp, &abxm) + dxb;   abxm -= dxb;
    abyp = MAXMIN(aybl, aybt, &abyp, &abym) + dyb;   abym -= dyb;
    bxbl = xas + xbl;                                bxbt = xas + xbt;
    bybl = yas - ybl;                                bybt = yas - ybt;
    bbxp = MAXMIN(bxbl, bxbt, &bbxp, &bbxm) + dxb;   bbxm -= dxb;
    bbyp = MAXMIN(bybl, bybt, &bbyp, &bbym) + dyb;   bbym -= dyb;

    /*
     *  Check for footprint/footprint collision.
     */
    if (BUTTON_BLE < abxp  &&  abxm < BUTTON_BTE   &&
        BUTTON_BLE < bbxp  &&  bbxm < BUTTON_BTE   &&
       -BUTTON_BHW < abyp  &&  abym < BUTTON_BHW   &&
       -BUTTON_BHW < bbyp  &&  bbym < BUTTON_BHW)      return (YES);

    /*
     *  Check for footprintA/prismB collision.
     */
    dxp = BUTTON_PHW*asin + buttClear;
    dyp = BUTTON_PHW*acos + buttClear;
    apxp = MAXMIN(axl, axbl, &apxp, &apxm) + dxp;    apxm -= dxp;
    apyp = MAXMIN(ayl, aybl, &apyp, &apym) + dyp;    apym -= dyp;

    if (BUTTON_BLE < apxp  &&  apxm < BUTTON_BTE   &&
        BUTTON_PLE < bbxp  &&  bbxm < BUTTON_BLE   &&
       -BUTTON_BHW < apyp  &&  apym < BUTTON_BHW   &&
       -BUTTON_PHW < bbyp  &&  bbym < BUTTON_PHW)      return (YES);

    /*
     *  Check for prismA/footprintB collision.
     */
    bpxp = MAXMIN(bxl, bxbl, &bpxp, &bpxm) + dxp;    bpxm -= dxp;
    bpyp = MAXMIN(byl, bybl, &bpyp, &bpym) + dyp;    bpym -= dyp;

    if (BUTTON_PLE < abxp  &&  abxm < BUTTON_BLE   &&
        BUTTON_BLE < bpxp  &&  bpxm < BUTTON_BTE   &&
       -BUTTON_PHW < abyp  &&  abym < BUTTON_PHW   &&
       -BUTTON_BHW < bpyp  &&  bpym < BUTTON_BHW)       return (YES);

    /*
     *  Check for prism/prism collision.
     */
    if (BUTTON_PLE < apxp  &&  apxm < BUTTON_BLE   &&
        BUTTON_PLE < bpxp  &&  bpxm < BUTTON_BLE   &&
       -BUTTON_PHW < apyp  &&  apym < BUTTON_PHW   &&
       -BUTTON_PHW < bpyp  &&  bpym < BUTTON_PHW)      return (YES);

    /*
     *  Check for footprintA/prismB collision.
     */
    dxt = BUTTON_THW*asin + buttClear;
    dyt = BUTTON_THW*acos + buttClear;
    atxp = MAXMIN(axbt, axt, &atxp, &atxm) + dxt;    atxm -= dxt;
    atyp = MAXMIN(aybt, ayt, &atyp, &atym) + dyt;    atym -= dyt;

    if (BUTTON_BLE < atxp  &&  atxm < BUTTON_BTE   &&
        BUTTON_BTE < bbxp  &&  bbxm < BUTTON_FVP   &&
       -BUTTON_BHW < atyp  &&  atym < BUTTON_BHW   &&
       -BUTTON_THW < bbyp  &&  bbym < BUTTON_THW)      return (YES);

    /*
     *  Check for tailA/footprintB collision.
     */
    btxp = MAXMIN(bxbt, bxt, &btxp, &btxm) + dxt;    btxm -= dxt;
    btyp = MAXMIN(bybt, byt, &btyp, &btym) + dyt;    btym -= dyt;

    if (BUTTON_BTE < abxp  &&  abxm < BUTTON_FVP   &&
        BUTTON_BLE < btxp  &&  btxm < BUTTON_BTE   &&
       -BUTTON_THW < abyp  &&  abym < BUTTON_THW   &&
       -BUTTON_BHW < btyp  &&  btym < BUTTON_BHW)      return (YES);

    /*
     *  Check for tail/tail collision.
     */
    if (BUTTON_BTE < atxp  &&  atxm < BUTTON_FVP   &&
        BUTTON_BTE < btxp  &&  btxm < BUTTON_FVP   &&
       -BUTTON_THW < atyp  &&  atym < BUTTON_THW   &&
       -BUTTON_THW < btyp  &&  btym < BUTTON_THW)      return (YES);

    /*
     *  Check for prismA/tailB collision.
     */
    if (BUTTON_PLE < atxp  &&  atxm < BUTTON_BLE   &&
        BUTTON_BTE < bpxp  &&  bpxm < BUTTON_FVP   &&
       -BUTTON_PHW < atyp  &&  atym < BUTTON_PHW   &&
       -BUTTON_THW < bpyp  &&  bpym < BUTTON_THW)      return (YES);

    /*
     *  Check for tailA/prismB collision.
     */
    if (BUTTON_BTE < apxp  &&  apxm < BUTTON_FVP   &&
        BUTTON_PLE < btxp  &&  btxm < BUTTON_BLE   &&
       -BUTTON_THW < apyp  &&  apym < BUTTON_THW   &&
       -BUTTON_PHW < btyp  &&  btym < BUTTON_PHW)      return (YES);

    /*
     *  No collision was detected.
     */
    return (NO);
}


/*+        T D F C O L L I S I O N

 *  Function name:
      tdFcollisionMarkFib

 *  Function:
      Checks if a fiducial mark is being obstructed by a fibre.

      If an obstruction is detected the function will return YES.
      If no obstruction is detected the function will return NO.

 *  Description:
      Given the fibre end points, and the fiducial location, determine if 
      the fiducial mark is obstructed by the fibre.

      Given m is the gradient of the line between the fibre ends and
      b is the intercept, then, d the distance between the poine x1,y1 and
      the line is given by
      
      d =    m.x1 - y + b
          -----------------
          +/- sqrt(m^2 + 1)
          

      An only the positive solution is considered.

 *  Language:
      C

 *  Call:
      (int) = tdFcollisionMarkFib (end1x,end1y,end2x,end2y,markx,marky,tolSqrd)

 *  Parameters:  (">" input, "!" modified, "W" workspace, "<" output)
      (>) end1x, end1y, end2x, end2y  (long int)  Fibre end points.
      (>) markx, marky                (long int)  Fiducial mark location.
      (>) tolSqrd                     (double)    Square of tolerance indicating that
                                                  fiducial is obstrcuted

 *  Proir Requirements:

 *  Support: James Wilcox, AAO

 *-

 *  History:
      01-Jul-1994  JW   Original version
      29-Oct-1996  TJF  Total rewrite.  Original algrothim did not appear right.
      {@change entry@}
 */
TDFUTIL_INTERNAL int  tdFcollisionMarkFib (
        long int  end1x,
        long int  end1y,
        long int  end2x,
        long int  end2y,
        long int  markx,
        long int  marky,
        double    tolSqrd)
{
/*
 *  a is the length of the fibre
 *  b is the distance from the fiducial to the pivot
 *  c is the distance from the fiduical to the fibre end.
 */

    double a = SQRD((double)(end1x - end2x)) + SQRD((double)(end1y - end2y));
    double b = SQRD((double)(end1x - markx)) + SQRD((double)(end1y - marky));
    double c = SQRD((double)(end2x - markx)) + SQRD((double)(end2y - marky));
/*
 *  That if the distance from the pivot to the fiducial is longer than the
 *  fibre, then there can be no collision.  Likewise for the fibre end
 *  to the fiducial.
 */
    if (((a+tolSqrd) < b) || ((a+tolSqrd) < c))
        return NO;
    
    else if (end1x == end2x)
    {
/*
 *      Fibre lays along the X axis and y of point is within the lenght
 *      of the fibre (ensure by the above check) 
 */
        long int d = markx - end2x;
        return (d < (sqrt(tolSqrd))? YES: NO);        
    }
    else
    {        
/*
 *      Get slope and intercept of line between fibre ends.
 */
        double m = (double)(end2y - end1y)/(double)(end2x - end1x);
        double b = end2y - m*end2x;
        
        double dd = sqrt(m*m + 1);
        double dn = m * markx - marky + b;
        
        double d;
        
        if (dn < 0)
            d = dn * -1 / dd;
        else
            d = dn / dd;         
        
        return (d < (sqrt(tolSqrd))? YES: NO);
    }
    
}


/*+        T D F C O L L I S I O N

 *  Function name:
      tdFcollisionMarkBut

 *  Function:
      Checks if a fiducial mark is being obstructed by a button.

      If an obstruction is detected the function will return YES.
      If no obstruction is detected the function will return NO.

 *  Description:

 *  Language:
      C

 *  Call:
      (int) = tdFcollisionMarkBut (butx,buty,theta,markx,marky,tolSqrd)

 *  Parameters:  (">" input, "!" modified, "W" workspace, "<" output)
      (>) butx, buty    (long int)   Button coordinates.
      (>) theta         (double)     Button orientation.
      (>) markx, marky  (long int)   Fiducial location.
      (>) tolSqrd       (double)     Square of tolerance indicating that
                                     fiducial is obstrcuted

 *  Proir Requirements:

 *  Support: James Wilcox, AAO

 *-

 *  History:
      01-Jul-1994  JW   Original version
      31-Jan-2000  TJF  We had not been using tolSqrd at all.  So, take
                        the square root of it and add that to the fiducial
                        diameter to get the presumed fiducial diameter.
      09-Jun-2004  TJF  The check against the prisim position was wrong - it
                        was using (xMark <= ((double)(BUTTON_PHW)  - fidClearDia))
                        instead of (xMark <= ((double)(BUTTON_PHW)  + fidClearDia)).
                        This is probably what was causing some problems with
                        obsured fiducials.
      {@change entry@}
 */
TDFUTIL_INTERNAL int  tdFcollisionMarkBut (
        long int  butx,
        long int  buty,
        double    theta,
        long int  markx,
        long int  marky,
        double    tolSqrd)
{
    double  seperationSqrd,
            limitSqrd,
            xShift, yShift,
            xMark, yMark,
            cosTheta, sinTheta;
    double  fidClearDia;  /* Clearance plus fiducial diameter */

    fidClearDia = sqrt(tolSqrd) + (double)(FID_DIA);

    /*
     *  Perform a rough check to see if a collision is likely.
     */
    seperationSqrd = SQRD((double)(butx-markx)) + SQRD((double)(buty-marky));
    limitSqrd = SQRD(fidClearDia+(double)BUTTON_FVP);
    if (seperationSqrd > limitSqrd)
        return NO;

    /*
     *  Likelyhood of a collision is high - need to perform an accurate check.
     *
     *  The check involves determining if the mark (consider it a dot) is inside any
     *  of the 3 rectangles that make the button (ie, the prism, the footprint, and
     *  the tail).
     *
     *    ie,    xRectMin <= xMark <= xRectMax and
     *           yRectMin <= yMark <= yRectMax
     *
     *  To perform these checks, we must first shift and rotate the mark such that
     *  the local coordinate system has it's origin at butx,buty with the y axis runing
     *  along the button handle.
     */
    cosTheta = cos(-theta);
    sinTheta = sin(-theta);
    xShift = (double)(markx - butx);
    yShift = (double)(marky - buty);
    xMark = xShift*cosTheta - yShift*sinTheta;
    yMark = yShift*cosTheta + xShift*sinTheta;

    /*
     *  Check against button footprint.
     */
    if ((xMark >  ((double)(-BUTTON_BHW) - fidClearDia)) &&
        (xMark <= ((double)( BUTTON_BHW) + fidClearDia)) &&
        (yMark >  ((double)( BUTTON_BLE) - fidClearDia)) &&
        (yMark <= ((double)( BUTTON_BTE) + fidClearDia)))
        return YES;

    /*
     *  Check against button tail.
     */
    if ((xMark >  ((double)(-BUTTON_THW) - fidClearDia)) &&
        (xMark <= ((double)( BUTTON_THW) + fidClearDia)) &&
        (yMark >  ((double)( BUTTON_BTE) - fidClearDia)) &&
        (yMark <= ((double)( BUTTON_FVP) + fidClearDia)))
        return YES;

    /*
     *  Check against button prism.
     */
    if ((xMark >  ((double)(-BUTTON_PHW) - fidClearDia)) &&
        (xMark <= ((double)( BUTTON_PHW) + fidClearDia)) &&
        (yMark >  ((double)( BUTTON_PLE) - fidClearDia)) &&
        (yMark <= ((double)( BUTTON_BLE) + fidClearDia)))
        return YES;

    /*
     *  If we reached here there must be no collision.
     */
    return NO;
}


/*+        T D F C O L L I S I O N

 *  Function name:
      tdFcollisionScrew

 *  Function:
      Checks if a button is overlapping a screw hole (used to secure the magnetic
      field plate surface to the plate structure).

      If an obstruction is detected the function will return YES.
      If no obstruction is detected the function will return NO.

 *  Description:
      This check is different to the check between the button and the fiducial as
      we are only concerned with the button outline (and not the tail or prism)
      coliding with the screw hole (also a much larger diameter than the fiducial).

 *  Language:
      C

 *  Call:
      (int) = tdFcollisionScrew (butx,buty,theta,tolSqrd)

 *  Parameters:  (">" input, "!" modified, "W" workspace, "<" output)
      (>) butx, buty      (long int)   Button coordinates.
      (>) theta           (double)     Button orientation.
      (>) tolSqrd         (double)     Square of tolerance.

 *  Proir Requirements:

 *  Support: James Wilcox, AAO

 *-

 *  History:
      15-Mar-1996  JW   Original version
      31-Jan-2000  TJF  We had not been using tolSqrd at all.  So, take
                        the square root of it and add that to the plate hole
                        diameter to get the presumed fiducal diameter.      
      11-Aug-2005  TJF  No longer needed - there are no longer any screws
                        on the 2dF plate.
      {@change entry@}
 */
TDFUTIL_INTERNAL int  tdFcollisionScrew (
        long int  butx DUNUSED,
        long int  buty DUNUSED,
        double    theta DUNUSED ,
        double    tolSqrd DUNUSED)
{
    return NO;
}

/*
 *                   D i s t  P o i n t  S e g  S q
 *
 *  Function:
 *     Returns shortest distance (squared) from a point to a line segment.
 *
 *  Description:
 *     Given the X,Y coordinates of a point, and the X,Y coordinates
 *     of each end of a line segment, this returns the square of the
 *     shortest distance from that point to a point on the line
 *     segment. (In many cases, that point will be one of the ends
 *     of the line segment.)
 *
 *  Author:
 *     Keith Shortridge, AAO.
 *
 *  Credit:
 *     There are loads of examples of code to do this on the web, but
 *     nowadays they are all very abstract, often in C++ and depending on
 *     pre-existing Point, Vector and Segment classes, and work in 
 *     arbitrary (well, usually at least 3) dimensions. This code is 
 *     quick, in C, and 2-dimensional. (The extension to 3 is trivial,
 *     but not wanted here.) The basic expression used follows that
 *     given by Paul Bourke, www.wasp.uwa.edu.au/~pbourke/geometry/pointline.
 *
 *  History:
 *      3rd Nov 2006. Original version, KS.
 */
 
double DistPointSegSq ( /* Returns square of shortest line-segment distance */
   double Px,           /* X-coordinate of point */
   double Py,           /* Y-coordinate of point */
   double S1x,          /* X-coordinate of one end of line segment */
   double S1y,          /* Y-coordinate of one end of line segment */
   double S2x,          /* X-coordinate of other end of line segment */
   double S2y)          /* Y-coordinate of other end of line segment */
{
   /*  The trick here it to work out the point on the segment that is
    *  closest to the point. We call this point I. What we work out is the
    *  fraction of the way along the line (between S1 and S2) at which I
    *  falls. If that fraction is outside the range 0-1 then the closest
    *  point is outside the segment and I must be S1 or S2. To work out
    *  the fraction, we have an expression whose denominator is the
    *  length of the segment squared. Working that out separately also 
    *  checks for the case where the segment is in fact a point, in which 
    *  case I is that point.
    */
     
   double Ix;            /* X-coordinate of I (closest point on line to P) */
   double Iy;            /* Y-coordinate of I (closest point on line to P) */
   double SegLenSq;      /* Square of length of segment */
   double Frac;          /* Fraction of way along segment to point I */
   
   SegLenSq = ((S2x - S1x) * (S2x - S1x)) + ((S2y - S1y) * (S2y - S1y));
   if (SegLenSq <= 0.0) {
   
      /* Segment is a point, intersection point could be either end. */
      
      Ix = S1x;
      Iy = S1y;
      
   } else {
   
      Frac = (((Px - S1x) * (S2x - S1x)) + ((Py - S1y) * (S2y - S1y))) /
                                                                 SegLenSq;
      if (Frac < 0.0) {
      
         /*  Intersection is not in the segment, and closest segment
          *  end to the point is end 1.
          */
          
         Ix = S1x;
         Iy = S1y;
         
      } else if (Frac > 1.0) {
      
         /*  Intersection is not in the segment, and closest segment
          *  end to the point is end 2.
          */
          
         Ix = S2x;
         Iy = S2y;
         
      } else {
      
         /*  Intersection is in the segment, at point I, which is
          *  LineFrac along the line.
          */
         
         Ix = S1x + (Frac * (S2x - S1x));
         Iy = S1y + (Frac * (S2y - S1y));
      }
   }
      
   /*  Now work out distance from point P to point I. */
      
   return ((Px - Ix) * (Px - Ix)) + ((Py - Iy) * (Py - Iy));
   
}      
