/*
 *+                         t d F f o r b i d d e n . c
  
 * Function name:
      tdFforbidden.c
  
 * Function:
   	Determine if we are allowed to lower the Jaw       
  
 * Description:
       This routine determines whether a position is forbidden for the
       purposes of lowering the positioner. Z axis.
       
       We can have the jaw lower either inside retractor ring or outside
       it.  We must now lower it over the retractor ring or cross the
       retractor ring with the jaw too low.  This routine checks for
       all such suitations.

       Based on 2dF Drawing E3956, dated 4 Jul 1996.  Note that the 2dF
       field plate is NOT round.
       
 * Language:
      C
  
 * Call:
     (int) = tdFforbidden.c (x,y,theta,rq,ri,ro,hw,jhwp,jhwm,jl,clearance,
     				inside,outside)
  
 * Parameters:   (">" input, "!" modified, "W" workspace, "<" output)
 	(>) x		(double) X position to test.
 	(>) y		(double) Y position to test.
 	(>) theta	(double) Theta of gripper
 	(>) rq		(double) Quadrant radius, from drawing.
 	(>) ri		(double) Inside plate radius, inside this circle
 				is clearly safe
 	(>) ro		(double) Output plate radius, outside this circle
 				is clearly outside the retractors.
 	(>) hw		(double) Separation of quadrants to account for
 				 guide fibres
	(>) jhwp        (double) Jaw halfwidth increasing theta side
        (>) jhwm      	(double) Jaw halfwidth decreasing theta side
 	(>) jl		(double) Jaw length
 	(>) clearance	(double) Minimum allowed clearance.  Will not allow
 				this to be below 500
 	(>) inside	(int *) Are we inside the retrator ring
 	(>) output	(int *) Are we outside the retractor ring
 	
 * Include files: 
  
 * Return value:
 	Returns true if the position is forbidden.
        
  
 * Support: John Straede, AAO
  
 *-
  
 * History:
      02-Oct-1996 - JOS - Original version.
  
 */


static char *rcsId="@(#) $Id: ACMM:2dFutil/tdFforbidden.c,v 5.2 10-Sep-2013 09:12:12+10 tjf $";
static void *use_rcsId = (0 ? (void *)(&use_rcsId) : (void *) &rcsId);


#include <math.h>

#define TAN4P5 0.07870171
#define SIN9   0.156434465
#define COS9   0.987688341
#define SIN4P5 0.0784590957
#define COS4P5  0.9969173337
#define CLEARANCE 500.0
#define PI     3.1415926535
#define TWOPI  6.283185308
#define PION2  1.570796327
#define R4P5   0.07853981634
#define R9     0.1570796327


#include "tdFfieldDims.h"

int tdFforbidden ( double x, double y,	/* coordinates			*/
		double theta,		/* probe position angle		*/
		double rq,		/* quadrant radius		*/
		double ri, 		/* inside "plate radius"	*/
		double ro,		/* outside "plate radius"	*/
		double hw,		/* half guide fibre "expansion"	*/
		double jhwp,            /* jaws halfwidth +ve theta	*/
		double jhwm,            /* jaws halfwidth -ve theta     */
		double jl,      	/* jaws length from optical axis*/
		double clearance,	/* minimum allowed clearance	*/
		int *inside,		/* flag: 1 = inside o.k.	*/
		int *outside)		/* flag: 1 = outside o.k.	*/
{
/*  Strongly recommended values for "fixed" arguments (microns)
 *
 *		rq = 277000 
 *		ri = 270430
 *		ro = 347300
 *		hw = 22500
 *		jwp = 12220
 *              jwm = 8300
 *		jl = 5250
 *		clearance = 500
 */

   double t ;		/* temporary					*/
   double rr ;		/* positioner radius squared			*/
   double xb, yb ;	/* coordinates after rotation			*/
   double jhw ;         /* greater jaw halfwidth                        */
   double jo ;		/* jaw offset from reference position		*/
   double jjo ;		/* jaw offset squared				*/
   int sector ;         /* 0 = guide sector, otherwise > 0              */


   if (clearance < CLEARANCE)   clearance = CLEARANCE ;

   jhw = jhwp ;
   if (jhwm > jhw)   jhw = jhwm ;  
   rr = x*x + y*y ;
   jjo = jhw*jhw + jl*jl ;
   jo = sqrt(jjo) + clearance ;
   
/* quick check for inside						*/

   if (rr  < (ri - jo)*(ri - jo))
   {
      *inside = 1 ;  *outside = 0 ;  return (0) ;
   }

/* quick check for outside						*/

   if (rr > (ro + jo)*(ro + jo))
   {
      *inside = 0 ;  *outside = 1 ;   return (0) ;
   }

   *outside = 0 ;

/* Change theta to mathematical coordinates !                           */

   theta += PION2 ;

/* Flip to get mirror image.  *** MAY NEED TO REMOVE THIS. ***		*/

   x = -x ;
   theta = TWOPI - theta ;
   
/* Collapse to single quadrant.						*/

   if (x < hw  &&  y >= hw)
     {
       t = -x ;  x = y ;  y = t ;  theta -= PION2 ;
     }
   else if (x < -hw  &&  y < hw)
     {
       x = -x ;  y = -y ;  theta += PI ;
     }
   else if (x >= -hw  &&  y < -hw)
     {
       t = -y ;  y = x ;  x = t ;  theta += PION2 ;
     }
 
   while (theta > PI)   theta -= TWOPI ;
   while (theta < -PI)  theta += TWOPI ;

/* Shift coordinates to single quadrant.				*/

   x -= hw ;
   y -= hw ;

/* Rotate into first sector.						*/

   xb = x*COS4P5 + y*SIN4P5 ;
   yb = y*COS4P5 - x*SIN4P5 ;
   x = xb ;
   y = yb ;
   theta -= R4P5 ;
   sector = 0 ;

/* Rotate to relevant sector.						*/

   while (y/x > TAN4P5)
   {
      xb = x*COS9 + y*SIN9 ;
      yb = y*COS9 - x*SIN9 ;
      x = xb ;
      y = yb ;
      theta -= R9 ;
      sector++ ;
   }

/*   Final tests						       	*/
/*   ===========                                                        */

/*   First check for inside current sector.                             */

   jo = jl*cos(theta) + clearance ;
   if (theta > 0.0) jo += jhwp*sin(theta) ; /* REMINDER: I have fliped   */
      else          jo -= jhwm*sin(theta) ; /* the coordinate system !   */

   if (x < rq - jo)
     {

/*  So far, so good.  What about the adjacent sector ?                  */

       if (y > 0.0)
	 {
/* Next sector                                                          */
           if (sector == 9)   x -= 2.*hw ;   /* Next sector may be the  */
	   xb = x*COS9 + y*SIN9 ;            /* next quadrant !         */
	   yb = y*COS9 - x*SIN9 ;
	   theta -= R9 ;
	   jo = jl*cos(theta) + clearance - jhwm*sin(theta) ;
	   
	   if (xb < rq - jo)
	     {
	       *inside = 1 ; return (0) ;  
	     }

	 }
       else
         {
/* Previous sector                                                      */
	   if (sector == 0)   y += 2.*hw ;  /* Previous sector can be   */
	   xb = x*COS9 - y*SIN9 ;           /* the previous quadrant !  */
	   yb = y*COS9 + x*SIN9 ;
	   theta += R9 ;
	   jo = jl*cos(theta) + clearance + jhwp*sin(theta) ;

	   if (xb < rq - jo)
	     {
	       *inside = 1 ; return (0) ;
	     }

	 }

     }

/*  Not inside - must be forbidden.                                     */

   *inside = 0 ;  return(1) ;

}


