/*                               t d F p o s N S M a s k . c
  
 * Module name:
        tdFposNSMask
  
 * Function:
        Functions to check the Nod and Shuffle masks status.
  
 * Description:
        Provides functions to check the Nod and Shuffle mask status.

 * Language:
      C
  
 * Support: Tony Farrell, AAO
  
 * Copyright (c) Anglo-Australian Telescope Board, 2006.
   Not to be used for commercial purposes without AATB permission.
  
 * @(#) $Id: ACMM:2dFutil/tdFposNSMask.c,v 5.2 10-Sep-2013 09:12:12+10 tjf $
  
 * History:
      31-Jul-2006 - TJF - Original version
  
 */
#ifdef VxWorks
#    include <vxWorks.h>
#    include "vmeIOSys.h"
#    include "vmeIOPos.h"
#endif


static char *rcsId="@(#) $Id: ACMM:2dFutil/tdFposNSMask.c,v 5.2 10-Sep-2013 09:12:12+10 tjf $";
static void *use_rcsId = (0 ? (void *)(&use_rcsId) : (void *) &rcsId);


#include "tdFutil.h"
#include "Sdp.h"
#include "DitsMsgOut.h"
#include "Ers.h"

/*
 *+                         t d F p o s N S M a s k C h e c k
  
 * Function name:
      tdFposNSMaskCheck
  
 * Function:
      Check the Nod and Shuffle mask status.
  
 * Description:
      The status of the Nod and Shuffle mask is checked.  If the mask
      is on for the specified plate, then *safe is set to false,  Otherwise
      it is set to true.

      When simulating, if the parameter NS_MASK_SIM exists, then
      it gives us the value we presume has been read from the
      hardware.
    
 * Language:
      C
  
 * Call:
     (void) = tdFposNSMaskCheck (debug, simFlag, plate, parameter, safe, status)

 * Parameters:
         (>) debug    (int)  If true, then output debuging.
         (>) simFlag  (int)  If true, then we are in simulation
         (>) plate    (int)  The plate ot check, must be 0 or 1.  If not, 
                              then we presume that no plate is in place
                              return *safe = 1.
         (>) paramameters (char *[]) If non-zero, points to the name of three task
                              parameters which will get the value of the
                              N&S mask (two bits indicating plates 0 and 1),
                              the value for Plate 0 and the value for plate 1.
         (<) safe      (int *) If the N&S mask on the specified plate is NOT
                              in place, return true.  OTherwise return false.
         (!) status    (StatusType *) Modified status.
                              
 * Include files: tdFutil.h
  
 * Support: Tony Farrell, AAO
  
 *-
  
 * History:
     31-Jul-2006 - TJF - Original version
     17-Aug-2006 - TJF - VxWorks code was not correctly enabled - fixed.
     07-Mar-2010 - TJF - Now have parameters for each plate. "parameter" argument
                          becomes an array.
  
 */
#define NSMASK_INTERLOCK_PLATE_0 1
#define NSMASK_INTERLOCK_PLATE_1 2
#define NSMASK_INTERLOCK_BOTH    3
extern void tdFposNSMaskCheck(
    int debug,
    int simFlag,
    int plate,
    const char *parameters[],
    int * safe,
    StatusType *status)
{
    unsigned char data = 0;
    int vmeStat = 0;
    int flags;
    if (*status != STATUS__OK) return;

#ifndef VxWorks  /* Something else should pick up this error before now */
    if (simFlag == 0)
        simFlag = 1;
#endif
    if (simFlag)
    {
        short simVal;
        ErsPush();
        SdpGets("NS_MASK_SIM", &simVal, status);
        if (*status != STATUS__OK)
        {
            ErsAnnul(status);
            /*
             * Presume both flags clear - which is to set the bits
             * true at this point - they are inverted later.
             */
            data =  NSMASK_INTERLOCK_BOTH; 
        }
        else
        {
            /* Invert the parameter to get the value of interest */
            data = ~simVal;
        }
    }
#ifdef VxWorks
    else
    {
#warning "VxWorks code compiled"
        /*
         * The N&S interlocks are on VME Interlock board 2, register 3.
         */
        vmeStat = vmeIORead(VME_IO_ILK_BRD_2|VME_IO_ILK_R3, &data);
    }
#endif    
    /*
     * Invert and grab the required bits.
     */
    flags = ~data & NSMASK_INTERLOCK_BOTH ;
    /*
     * The raw data goes into the first parameter, and plate specific values
     * in the two other parameters.  Only update parameter if value has changed
     * (to avoid monitors being triggered all the time)
     */
    if (parameters)
    {
        short curVal;
        short newVal;
        SdpGets(parameters[0], &curVal, status);
        if (curVal != flags)
            SdpPuti(parameters[0], flags, status);

        SdpGets(parameters[1], &curVal, status);
        newVal = ((flags& NSMASK_INTERLOCK_PLATE_0) ? 1 : 0);
        if (curVal != newVal)
            SdpPuti(parameters[1], newVal , status);


        SdpGets(parameters[2], &curVal, status);
        newVal = ((flags& NSMASK_INTERLOCK_PLATE_1) ? 1 : 0);
        if (curVal != newVal)
            SdpPuti(parameters[2], newVal , status);

    }

    /*
     * Debugging message.
     */
    if (debug)
        MsgOut(status, "tdFposNSMaskCheck: VmeStatus %d Raw data %d, Masked %d, plate %d, Simulation %d, Parameter %s",
               vmeStat, (int)data, flags, plate, simFlag,
               parameters ? parameters[0] : "<none>");

    if (*status != STATUS__OK)
        *safe = 0;  /* Something when wrong - presume we are not safe */
    else if ((plate == 0)&&(flags & NSMASK_INTERLOCK_PLATE_0))
        *safe = 0;  
    else if ((plate == 1)&&(flags & NSMASK_INTERLOCK_PLATE_1))
        *safe = 0;
    else
        *safe = 1; /* The relevant plate is clear - we are safe */
}
