/*                               t d F p o s V m e I O
  
 * Module name:
        tdFposVmeIO
  
 * Function:
        Interface to the 2dF positioner VME IO bus.
  
 * Description:
        The VME IO bus is used by the 2dF positioner to control various
        relays.  it provides a read and write calls and interrupt support.


	Just pass control to vmeDIO routine specfying the 
	positioner mask.



        The routines provided ared

            tdFposVmeIOInit => Initialise the package.  Should be invoked using
                               VxWorks start up to ensure the appropriate
                               bits are clear.

            tdFposVmeIOWrite => Write to the IO system, clearing or setting
                                appropiate bits in the mask.
            tdFposVmeIORead  => Read the current value of the mask.
            tdFposVmeIOMaskAddr  => To interface with packages which do the
                                writing themselves, this routine returns the
                                address of the internally maintained mask which
                                should indicates the currently enabled set
                                of relays.  The value pointed to should be
                                modified as appropiate. SHOULD NOT BE USED
				NOW, SINCE WAS ONLY USED IN CALLS TO APS
				ROUTINES, WHICH NOW IGNORE THE VALUE ON INPUT,
				JUST RETUNS A POINTER TO MODULE DATA.
  

   ORIGINAL COMMENTS:  They no longer apply but help to explain the
	structure of the code.

        Unfortunately, there are two problems

             1.  The read call is not a realiable indicator of what bits 
                 are on and off.  This is because they are dependent on 
                 the current through the relays.

             2.  In order write to a bit, we need to know the current status
                 so that we can only change the bits we are interested in.

        This means that the state of the VME IO system must be maintained on
        a global basis.  This package provides an interface to doing this.

 * Language:
      C
  
 * Support: Tony Farrell, AAO
  
 * Copyright (c) Anglo-Australian Telescope Board, 1997.
   Not to be used for commercial purposes without AATB permission.
  
 *     @(#) $Id: ACMM:2dFutil/tdFposVmeIO.c,v 5.2 10-Sep-2013 09:12:11+10 tjf $
  
 * History:
      15-Sep-1997 - TJF - Original version
      13-Nov-1997 - TJF - Relay mask is now kept in driver.  We ignore our
		own mask and use the new routines to activate relays.
  
 */
#ifdef VxWorks
#    include <vxWorks.h>
#    include "vmeIO.h"
#    include <logLib.h>
#else
     typedef int STATUS;
#endif


static char *rcsId="@(#) $Id: ACMM:2dFutil/tdFposVmeIO.c,v 5.2 10-Sep-2013 09:12:11+10 tjf $";
static void *use_rcsId = (0 ? (void *)(&use_rcsId) : (void *) &rcsId);


#include "tdFutil.h"
#include "vmeIOPosDIO.h"

static unsigned char IOMask=0;   /* Maintains the current mask */

/*
 *+                         t d F p o s V m e I O I n i t
  
 * Function name:
      tdFposVmeIOInit
  
 * Function:
      Initialise the Positioner VME IO Module
  
 * Description:
      Clears the internal mask used to keep track of the state of the
      Positioner VME IO system.  Writes the cleared mask to the VME
      IO system to ensure we are in a known state.
    
 * Language:
      C
  
 * Call:
     (void) = tdFposVmeIOInit ()
  
  
 * Include files: tdFutil.h
  
 * Support: Tony Farrell, AAO
  
 *-
  
 * History:
      15-Sep-1997 - TJF - Original version
      13-Nov-1997 - TJF - Replace call to vmeIOWrite with one to
				new routine vmeIODIOInitRly
  
 */
extern void tdFposVmeIOInit (void)
{
    IOMask = 0;
#ifdef VxWorks
    if (vmeIODIOInitRly(VME_IO_POS_DIO_RW_RLY) == -1)
    {
        logMsg("tdFposVmeIOInit:Failed to initialise vmeIO system\n",
                  0,0,0,0,0,0);
    }
    else
        logMsg("tdFposVmeIOInit:Successfully started vme IO system\n",
                  0,0,0,0,0,0);
#endif
}
/*
 *+                         t d F p o s V m e I O W r i t e
  
 * Function name:
      tdFposVmeIOWrite
  
 * Function:
      Write the Positioner VME IO Module
  
 * Description:
      Write to the Positioner VME IO module using the vmeIOWrite call.

      disable the relays in the CLEAR mask.   Enable the relays in the
      SET mask.

 * Language:
      C
  
 * Call:
     (void) = tdFposVmeIOWrite (clear,set)

 * Parameters:   (">" input, "!" modified, "W" workspace, "<" output)
  	(>) clear (unsigned char) These bits in the mask are cleared.
        (>) set   (unsigned char) These bits in the mask are set.
   
 * Returned value: -1 (VxWorks ERROR) on error, 0 (VxWorks OK) on success.
 
 * Include files: tdFutil.h
  
 * Support: Tony Farrell, AAO
  
 *-
  
 * History:
      15-Sep-1997 - TJF - Original version
      13-Sep-1997 - TJF - replace call to vmeIOWrite with new routine
			  vmeIODIOCtrlRly
  
 */
extern int tdFposVmeIOWrite (unsigned char clear, 
                              unsigned char set)
{
    int valueMask  = set;		/* Value of bits	*/
    int enableMask  = clear | set;	/* Bits to change	*/

#ifdef VxWorks
        return (vmeIODIOCtrlRly(VME_IO_POS_DIO_RW_RLY, valueMask, enableMask));
#else
        valueMask = 0;  /* Just to shut up compiler */
        enableMask = 0;
        return -1;
#endif
 
}


/*
 *+                         t d F p o s V m e I O R e a d
  
 * Function name:
      tdFposVmeIORead
  
 * Function:
      Read the Positioner VME IO Module mask
  
 * Description:
      Read  the Positioner VME IO Module current mask.

 * Language:
      C
  
 * Call:
     (unsigned char) = tdFposVmeIORead ()

  
 * Returned value: The current value of the mask.
 
 * Include files: tdFutil.h
  
 * Support: Tony Farrell, AAO
  
 *-
  
 * History:
      15-Sep-1997 - TJF - Original version
      13-Nov-1997 - TJF - Use new vmeIODIOGetRlyMask routine.
  
 */
extern unsigned char tdFposVmeIORead ()
{
    unsigned char mask = 0;
#ifdef VxWorks
    if (vmeIODIOGetRlyMask(VME_IO_POS_DIO_RW_RLY, &mask) == -1)
        logMsg("tdFposVmeIOInit:Failed to read vmeIO relay mask",
                  0,0,0,0,0,0);
#endif
    return (mask);
}


/*
 *+                         t d F p o s V m e I O M a s k A d d r
  
 * Function name:
      tdFposVmeIOMaskAddr
  
 * Function:
      Return the address of the the Positioner VME IO Module mask
  
 * Description:

      REDUNDANT ROUTINE, MAINTAINED BECAUSE THE ONLY THINGS WHICH USE
      IT NOW IGNORE THE VALUE RETURNED.

 * ORIGINAL COMMENTS:

      Return the  the Positioner VME IO Module current mask used by
      tdFposVmeIOWrite.  This address can be passed to routines which
      call vmeIOWrite directly, allowing them to update the mask themselves.

      Higher level code is responsible for managing asscess to this variable.
    

      Note that this is done using the internal mask of the tdFposVmeIO
      module.  No hardware read is attempted.

 * Language:
      C
  
 * Call:
     (unsigned char *) = tdFposVmeIOMaskAddr ()

  
 * Returned value: The address of the mask
 
 * Include files: tdFutil.h
  
 * Support: Tony Farrell, AAO
  
 *-
  
 * History:
      15-Sep-1997 - TJF - Original version
      13-Nov-1997 - TJF - Now redundant, update comments.
  
 */
extern unsigned char * tdFposVmeIOMaskAddr ()
{
    return (&IOMask);
}




