/*
 *+                                t d F s a v e
  
 * Function name:
      tdFsave
  
 * Function:
	Saves 2dF configuration file after optionally saving a backup.
  
 * Description:
  	In many places, the 2dF positioner software writes Sds files
  	out to disk to save configuration files.  This routine does
  	this and if compiled withe CREATE_BACKUP macro defined to be 1,
        saves a copy of the file in a sub-directory the name 
  	of which is based on the current time.      
  
 * Language:
      C
  
 * Call:
     (void) = TdfSave (file,usePath,pathName,sdsId,status)
  
 * Parameters:   (">" input, "!" modified, "W" workspace, "<" output)
 	(>) file (const char *) The base name of the file involved.  
 	(>) usePath (int) If true, then save in the directory refered to
 				by pathName, otherwise, save in the current
 				directory
 	(>) pathName (const char *) Only used if usePath is true, in which
 				case it indicates the directory in which
 				to save the files.
 	(>) sdsId	(SdsIdType) The Sds data to save 
	(>) show	(int) If true, log writing of file
	(>) message	(const char *) If show is true, a message to be output
			if the write successed and show it true.  It can 
			contain one %s character to indicate where the
			filename should be formated into the string.
	(!) status	(StatusType *) Modified status
        
  
 * Include files: tdFutil.h
  
 * External functions used:
        
  
 * External values used:
        
  
 * Prior requirements:
        
  
 * Support: Tony Farrell, AAO
  
 *-
  
 * History:
      10-Sep-1996 - TJF - Original version
      27-Oct-1999 - TJF - Don't need the backup anymore.  Disable it by
                          wrapping the code up in a macro.
      08-Oct-2004 - TJF - Fix MacOsX compile problem.
  
 */



static char *rcsId="@(#) $Id: ACMM:2dFutil/tdFsave.c,v 5.2 10-Sep-2013 09:12:12+10 tjf $";
static void *use_rcsId = (0 ? (void *)(&use_rcsId) : (void *) &rcsId);


#include <stdio.h>
#include <errno.h>
#ifdef unix
#    include <sys/types.h>
#    include <sys/stat.h>
#    include <sys/time.h>
/* Most unix versions don't declare this in include files, but linux and
 * MacOsX do.
 * Need to find out what the standards say about it
 */
#    if !defined(__linux__) && !defined(__APPLE__)
     extern char *sys_errlist[];
#    endif
#elif defined(VxWorks)
#   include <fcntl.h>
#   include <nfsLib.h>
    extern STATUS copy (char *in,char *out);
#endif
#include <time.h>

#include "DitsTypes.h"
#include "sds.h"
#include "DitsMsgOut.h"
#include "tdFutil.h"
#include "tdfxy_err.h"
#include "Ers.h"

#ifdef DFUDGE_PROTOTYPES_UNIX
    extern int strftime(char *buf, int bufsize, const char *fmt, 
    		const struct tm *tm);
	/*
    extern int mkdir(const char *path, mode_t mode);
	*/
    extern int rename (const char *oldname, const char *newname);

#endif
#ifdef DFUDGE_PROTOTYPES_C
   extern time_t time(time_t *);
#endif

/*
 *  Define this to be 1 if you want to create the backup in the
 *  date and time based directory.
 */
#ifndef CREATE_BACKUP
#define CREATE_BACKUP 0
#endif

void TdfSave (const char * file, int usePath, const char *pathName,
	      SdsIdType sdsId, int show DUNUSED,
	      const char *message DUNUSED, StatusType *status)
{
    char fname[200];	   /* Name for actual file	*/
#   if CREATE_BACKUP
        char sname[200];   /* Name for saved version of file */
        char datedir[200]; /* Component of saved file name with just date */
        char timedir[200]; /* Component of savedfile name with data and time*/
    
        time_t t;
        struct tm timeBuffer;
    
        char day[10];
        char curtime[10];
#   endif    

    if (*status != STATUS__OK) return;
    


#   if CREATE_BACKUP
/*
 *  Get the current time.
 */
        t = time(0);
#       ifdef VxWorks
            gmtime_r(&t,&timeBuffer);
#       else
            timeBuffer = *gmtime(&t);
#       endif /* ! VxWorks */

/*
 *      Format a date string of the form
 *  		yymmdd
 */
        strftime(day,sizeof(day),"%y%m%d",&timeBuffer);
    
/*
 *      Format the time in the form.  Note that we assume the seconds
 *      compoent is 0 so that all files written within the same minute
 *      are written in the same sub-directory.
 *  
 *  		hh:mm:ss
 */
        strftime(curtime,sizeof(curtime),"%H:%M:00",&timeBuffer);

#   endif /* CREATE_BACKUP */    
    if (usePath)
    {
        if (ErsSPrintf(sizeof(fname),fname,"%s/%s.sds",pathName,file) == EOF)
                *status = TDFXY__SPRINTF;
     
#     if CREATE_BACKUP           
        else if (ErsSPrintf(sizeof(datedir),datedir,"%s/%s",pathName,day)
        					== EOF)
                *status = TDFXY__SPRINTF;
                
        else if (ErsSPrintf(sizeof(timedir),timedir,"%s/%s/%s",pathName,day,
        		curtime) == EOF)
                *status = TDFXY__SPRINTF;
                
        else if (ErsSPrintf(sizeof(sname),sname,"%s/%s/%s/%s.sds",pathName,day,
        		curtime,file) == EOF)
                *status = TDFXY__SPRINTF;
#     endif
    }
    else
    {
        if (ErsSPrintf(sizeof(fname),fname,"%s.sds",file) == EOF)
                *status = TDFXY__SPRINTF;
#     if CREATE_BACKUP                
        else if (ErsSPrintf(sizeof(datedir),datedir,"%s",day)
        					== EOF)
                *status = TDFXY__SPRINTF;
                
        else if (ErsSPrintf(sizeof(timedir),timedir,"%s/%s",day,
        		curtime) == EOF)
                *status = TDFXY__SPRINTF;
                
        else if (ErsSPrintf(sizeof(sname),sname,"%s/%s/%s.sds",day,
        		curtime,file) == EOF)
                *status = TDFXY__SPRINTF;
#     endif
    }
    
    if (*status != STATUS__OK) return;
    
/*
 *  Make the directory corresponding to today's date.
 */
#   if CREATE_BACKUP
#      ifdef VxWorks
           if ((mkdir(datedir) == ERROR)&&(errno != S_nfsLib_NFSERR_EXIST))
           {
               *status = TDFXY__DIRCREERR;
               ErsRep(0,status,"Error creating directory \"%s\", errno = %x",
                      datedir,errno);
    	
               return;
           }
#      else
           if ((mkdir(datedir,S_IRWXG|S_IRWXU) == -1)&&(errno != EEXIST))
           {
               *status = TDFXY__DIRCREERR;
               ErsRep(0,status,"Error creating directory \"%s\", errno = %x",
                      datedir,errno);
               ErsRep(0,status,"Error text is \"%s\"",sys_errlist[errno]);
               return;
           }
#      endif    
       errno = 0;
    
/*
 *  Make the directory corresponding to the current time.
 */
#      ifdef VxWorks
           if ((mkdir(timedir) == ERROR)&&(errno != S_nfsLib_NFSERR_EXIST))
           {
               *status = TDFXY__DIRCREERR;
               ErsRep(0,status,"Error creating directory \"%s\", errno = %x",
                      timedir,errno);
               return;
           }
#      else
           if ((mkdir(timedir,S_IRWXG|S_IRWXU) == -1)&&(errno != EEXIST))
           {
               *status = TDFXY__DIRCREERR;
               ErsRep(0,status,"Error creating directory \"%s\", errno = %x",
                      timedir,errno);
               ErsRep(0,status,"Error text is \"%s\"",sys_errlist[errno]);
               return;
           }
#      endif    
       errno = 0;
#   endif /* CREATE_BACKUP */    

/*
 *  Write file twice, once in the spot where it will be picked up next 
 *  time and once in the saved location.
 */
    SdsWrite(sdsId,fname,status);
    if (*status != STATUS__OK)
	ErsRep(0,status,"Error writing file \"%s\", errno = %d",fname,errno);
#   if CREATE_BACKUP
        else
        {
            SdsWrite(sdsId,sname,status);
            
            if ((*status == STATUS__OK)&&(show))
                MsgOut(status,message,fname);
            else if (*status != STATUS__OK)
                ErsRep(0,status,"Error writing file \"%s\", errno = %d",
                       sname,errno);   
        }
#   endif
    
}
