/*                               t d F s t a t e
  
 * Module name:
        tdFstate
  
 * Function:
       Functions for handling the TASK_STATE parameter of a 2dF positioner task.
  
 * Description:
	The four lower level 2dF positioner tasks each contain a parameter
	name TASK_STATE.  This parameter is an unsigned short in which the
	last eight bits are used to maintain details about the state of
	each task (bits defined in tdFiomaps.h).  
	
	This module provides access to setting and clearing bits in this
	parameter.
  
 * Language:
      C
  
 * Support: Tony Farrell, AAO
  
 * Copyright (c) Anglo-Australian Telescope Board, 1997.
   Not to be used for commercial purposes without AATB permission.
  
 *     @(#) $Id: ACMM:2dFutil/tdFstate.c,v 5.2 10-Sep-2013 09:12:11+10 tjf $
  
 * History:
      26-Mar-1997 - TJF - Original version
  
 */


static char *rcsId="@(#) $Id: ACMM:2dFutil/tdFstate.c,v 5.2 10-Sep-2013 09:12:11+10 tjf $";
static void *use_rcsId = (0 ? (void *)(&use_rcsId) : (void *) &rcsId);


#include "DitsTypes.h"
#include "Sdp.h"
#include "tdFutil.h"

/*
 *+                         t d F s t a t e B i t S e t
  
 * Function name:
      tdFstateBitSet
  
 * Function:
        Sets a bit in the TASK_STATE parameter.
  
 * Description:
  	Sets the specified bit in the TASK_STATE parameter.      
  
 * Language:
      C
  
 * Call:
     (void) = tdFstateBitSet (bit,status)
  
 * Parameters:   (">" input, "!" modified, "W" workspace, "<" output)
  	(>) bit (unsigned char) The bit/s to set
  	(!) status (StatusType *) Modified status.
  
 * Include files: tdFutil.h
  
 * Support: Tony Farrell, AAO
  
 *-
  
 * History:
      26-Mar-1997 - TJF - Original version
  
 */
extern void tdFstateBitSet (unsigned char bit, StatusType *status)
{
    unsigned short oldval,newval=0;
    SdpGetus("TASK_STATE",&oldval,status);
    
    newval = oldval | bit;
  
    /* Only update the parameter (triggering monitors) if value has changed*/ 
    if (newval != oldval) 
        SdpPutus("TASK_STATE",newval,status);
}

/*
 *+                         t d F s t a t e B i t C l e a r
  
 * Function name:
      tdFstateBitClear
  
 * Function:
        Clear a bit in the TASK_STATE parameter.
  
 * Description:
  	Clear the specified bit in the TASK_STATE parameter.      
  
 * Language:
      C
  
 * Call:
     (void) = tdFstateBitClear (bit,status)
  
 * Parameters:   (">" input, "!" modified, "W" workspace, "<" output)
  	(>) bit (unsigned char) The bit/s to clear
  	(!) status (StatusType *) Modified status.
  
 * Include files: tdFutil.h
  
 * Support: Tony Farrell, AAO
  
 *-
  
 * History:
      26-Mar-1997 - TJF - Original version
  
 */
extern void tdFstateBitClear (unsigned char bit, StatusType *status)
{
    unsigned short oldval,newval=0;
    SdpGetus("TASK_STATE",&oldval,status);
    
    newval = oldval & (~bit);
    
    /* Only update the parameter (triggering monitors) if value has changed*/ 
    if (newval != oldval) 
        SdpPutus("TASK_STATE",newval,status);
}

