

static char *rcsId="@(#) $Id: ACMM:2dFutil/tdfflex.c,v 5.2 10-Sep-2013 09:12:11+10 tjf $";
static void *use_rcsId = (0 ? (void *)(&use_rcsId) : (void *) &rcsId);


#include <math.h>
#include <stdio.h>
#include "status.h"
#include "slalib.h"
#include "slamac.h" 
#include "sds.h"


#include "tdfflex.h"


/*+				T d f G e t F l e x

 *  Function name:
      TdfGetFlex

 *  Function:
      Get Flexure Parameters

 *  Description:
      Get the 2dF flexure parameters from an SDS file.

      
 *  Language:
      C

 *  Declaration:
       TdfGetFlex(char *file, TdfFlexType *pars, StatusType *status);

      
 *  Parameters:   (">" input, "!" modified, "W" workspace, "<" output)
      (>) file    (char *)        Name of sds file to read parameters from.
      (<) pars    (TdfFlexType*)  Flexure parameters.
      (!) status  (StatusType *)  Modified status.


 *  Support: Jeremy Bailey, {AAO}

 *  Version date: 1-Jun-94
 *-
 */

void TdfGetFlex(char *file, TdfFlexType *pars, StatusType *status)

{
   SdsIdType id,id2;
   unsigned long actlen;
   
   SdsRead(file,&id,status);
   
   SdsFind(id,"offset",&id2,status);
   SdsGet(id2,sizeof(double),0,&(pars->offset),&actlen,status);
   SdsFreeId(id2,status);
   SdsFind(id,"length",&id2,status);
   SdsGet(id2,sizeof(double),0,&(pars->length),&actlen,status);
   SdsFreeId(id2,status);
   SdsFind(id,"k",&id2,status);
   SdsGet(id2,sizeof(double),0,&(pars->k),&actlen,status);
   SdsFreeId(id2,status);

   SdsReadFree(id, status);
   SdsFreeId(id,status);
}

/*+				T d f F l e x u r e

 *  Function name:
      TdfFlexure

 *  Function:
      Apply Correction for flexure 

 *  Description:
      For a given x,y and telescope position determine the corrections
      to x and y due to flexure in the positioner.

      
 *  Language:
      C

 *  Declaration:
       TdfFlexure(long int x, long int y, double ha, double dec,
                  TdfFlexType *pars, long int *dx, long int *dy,
                  StatusType *status);

      
 *  Parameters:   (">" input, "!" modified, "W" workspace, "<" output)
      (>) x       (long int)      Gantry position - x coordinate (microns)
      (>) y       (long int)      Gantry position - y coordinate (microns)
      (>) ha      (double)        Hour angle of telescope (radians)
      (>) dec     (double)        Declination of telescope (radians)
      (>) pars    (TdfFlexType*)  Flexure parameters - can be obtained
                                    by calling TdfGetFlex.
      (<) dx      (long int *)    Correction to x (microns).
      (<) dy      (long int *)    Correction to y (microns).
      (!) status  (StatusType *)  Modified status.


 *  Support: Jeremy Bailey, {AAO}

 *  Version date: 14-Jun-94
 *-
 */


void TdfFlexure(long int x, long int y, double ha, double dec,
                TdfFlexType *pars, long int *dx, long int *dy, StatusType *status)

{
    double pa,zd;
    double phi = -0.54589;
    double yd;
    double dyd;

    if (*status != STATUS__OK) return;

    *dy = 0;

/*  parallactic angle */    
    pa = slaPa(ha,dec,phi);
    
/*  Zenith distance */
    zd = slaZd(ha,dec,phi);
    
    yd = (double)y - pars->offset;
    
    dyd = pars->k * yd * (pars->length - yd)*cos(pa)*sin(zd);
    
    *dx = (long)(dyd * 4.0/((pars->length)*(pars->length)));
       
}

/*+				T d f S e t F l e x

 *  Function name:
      TdfSetFlex

 *  Function:
      Set Flexure Parameters

 *  Description:
      Write the 2dF flexure parameters to an SDS file.

      
 *  Language:
      C

 *  Declaration:
       TdfSetFlex(TdfFlexType pars, char *file, StatusType *status);

      
 *  Parameters:   (">" input, "!" modified, "W" workspace, "<" output)

      (>) pars    (TdfFlexType)   Flexure parameters.
      (>) file    (char *)        Name of sds file to write parameters to.
      (!) status  (StatusType *)  Modified status.


 *  Support: Jeremy Bailey, {AAO}

 *  Version date: 15-Jun-94
 *-
 */

void TdfSetFlex(TdfFlexType pars, char *file, StatusType *status)

{
    SdsIdType id,id2;
	          
	SdsNew(0,"flexure",0,NULL,SDS_STRUCT,0,NULL,&id,status);
	SdsNew(id,"offset",0,NULL,SDS_DOUBLE,0,NULL,&id2,status);
	SdsPut(id2,sizeof(double),0,&(pars.offset),status);
	SdsFreeId(id2,status);
	SdsNew(id,"length",0,NULL,SDS_DOUBLE,0,NULL,&id2,status);
	SdsPut(id2,sizeof(double),0,&(pars.length),status);
	SdsFreeId(id2,status);
	SdsNew(id,"k",0,NULL,SDS_DOUBLE,0,NULL,&id2,status);
	SdsPut(id2,sizeof(double),0,&(pars.k),status);
	SdsFreeId(id2,status);
    SdsWrite(id,file,status);
    SdsDelete(id,status);
    SdsFreeId(id,status);
}
