/*+		T D F F P I L F U L L
 * Module name:
     TDFFPILFULL

 * Function:
      Wrap up 2dF constants file access and telescope model access.

 * Description:
      This module implements TdfFpilFullInit(3), a routine which invokes
      FpilInit(3) with appropiate values for 2dF in FPIL's full mode.  
      (collision detection, constants access and coorindate transformation) 
      Since the original equivalent 2dF routines don't quite meet the required
      function prototypes, we provide wrappers (which can be removed
      when the original 2dF routines are modified, but since that
      requires changes to a lot of code, it won't be done for a while).

      We make use of TdfFpil___SetupMinStructs to fetch the FPIL Mininal
      mode setup (collision detection, basic instrument details etc.)
      
 * Language:
      C

 * Support:
      Tony Farrell, AAO
      Keith Shortridge, AAO

 *-

 *     @(#) $Id: ACMM:2dFutil/tdffpilfull.c,v 5.2 10-Sep-2013 09:12:11+10 tjf $

 * History:
     07-Dec-1999 - TJF - Original version
     13-Mar-2000 - TJF - Added QuickAppToObs.  Support new approach tel
                         telescope parameters.
     16-Mar-2000 - TJF - Add DrawButFib routine.
     17-Mar-2000 - TJF - Add SpecNames.
      5-Dec-2001 - KS  - Now treats guide fibres as just another fibre
                         type, so 2dF has three fibre types, not just 2
                         types and guide fibres.
      5-Jun-2002 - KS  - PATH_MAX defined if include files miss it out.
     14-Dec-2004 - TJF - Changed to support 400 rather then 404 fibres.
                         constants file renamed to  "tdFconstants400.sds.  
                         TdfFpilFullInit() has been renamed to TdfFpilFullInit400(), and
                         a new version named TdfFpilFullInit404() created, such that
                         configure can access old configurations correctly.
      7-Mar-2005 - KS  - Added support for AAOmega. Introduced the new 
                         routine TdfFpilFullInitAAOmega() and its associated
                         routines.
      1-Apr-2005 - KS  - Modified code in AccessConsAAOmega() to avoid the
                         use of SdsInsert(). Doing so seems to have fixed some
                         intermittent segmentation errors and other odd,
                         erratic errors.
     19-Oct-2005 - KS  - No code changes, but comments changed to reflect the
                         fact that the tentative values for AAOmega guide 
                         fibre numbers turned out to be correct.
     17-Nov-2005 - KS  - It was eventually decided that field2 for AAOmega
                         should be the combined field plate, rather than
                         field 0. Code reworked accordingly.
     10-May-2006 - TJF - Change use_rcsId to standards compatible form.
     29-Jun-2006 - KS  - Corrected code in AccessConsAAOmega() that freed 
                         external sds Id too soon. WhichSpecAAOmega() now
                         uses proper set of AAOmega names.
     23-Mar-2007 - KS  - Added support for 'virtual' plate 2.
     04-Apr-2012 - MVV - Added support for HERMES. Introduced the new
                         routine TdfFpilFullInitHERMES() and its associated
                         routines.
     25-Feb-2013 - TJF - AAOmega specific TdfFpilFullInitAAOmega(), 
                          WhichSpecAAOmega(),  AccessConsAAOmega() and related
                          constants extracted into tdffpilfull_aaomega.c, as
                          part of completing HERMES support.
 */
/*
 *  Include files
 */
#define __EXTENSIONS__   /* To get PATH_MAX on solaris */


static char *rcsId="@(#) $Id: ACMM:2dFutil/tdffpilfull.c,v 5.2 10-Sep-2013 09:12:11+10 tjf $";
static void *use_rcsId = (0 ? (void *)(&use_rcsId) : (void *) &rcsId);

#include <limits.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <stdio.h>
#include "drama.h"
#include "sds.h"
#define TDFFPIL_BUILD  /* Used by the include file */
#include "tdffpil.h"

#include "tdfxy.h"
#include "slalib.h"
#include "tdFbuttonDims.h"
#include "Ers.h"
#include "tdfxy_err.h"

#ifndef PATH_MAX
#define PATH_MAX 1024
#endif

#ifdef DEBUG
#undef DEBUG
#endif
#define DEBUG printf

/* 
 * Constant definitions
 */
#define TDFCONSFILE400 "tdFconstants400.sds"      /* 2dF constants file - 400 fibre ver */
#define TDFCONSFILE404 "tdFconstants.sds"         /* 2dF constants file - 404 fibre ver */
#define TDF_P0_LINFILE "tdFlinear0.sds"        /* 2dF linear model files */
#define TDF_P1_LINFILE "tdFlinear1.sds"

#define TDF_P0_DISTFILE "tdFdistortion0.sds"   /* 2dF distortion model files */
#define TDF_P1_DISTFILE "tdFdistortion1.sds"

#define TDFMODELDIR "/instsoft/2dF/config"     /* Directory at the AAT/AAO 
                                                  for linear and distortion
                                                  files. */
/*
 * Macro function definitions
 */

/*
 * type definitions
 */
typedef struct {
    FpilModelType model;
    TdfDistType   dist;    /* Set by AccessModel using TdfGetDist() */
    TdfLinType    linpars; /* Set by AccessModel using TdfGetLin()  */
    TdfXyType     xypars;  /* Set by CvtInit using TdfXyInit()      */
} TdfModel, *TdfModelPnt;

/*
 * Global variables and functions imported (but not declared in include files)
 */

/*
 * Global variables exported (definitions and initialization of globals
 * declared in the associated include file)
 */

/*
 * Variables private to this module (static variables)
 */

/*
 * Spectrograh names, used by WhichSpec() and TdfFpilFullInit()
 */

static const char * const SpecNames[] = { "Guide", "Spect 1", "Spect 2" };
static const char * const HERMESSpecNames[] = { "Guide", "HERMES"};


/* Originally, it was not clear what the pivot numbers for the HERMES plates
 * would be, so the code was written to use a look-up table that could be
 * changed later. The number of pivots is deduced from the size of this
 * array.
 */
static const int HERMESGuidePivs[] = { 50, 100, 150, 200, 250, 300, 350, 400 };

/*
 * Telescope model parameter names;
 */
static const char * const ParamNames[] = {
    "MA",
    "ME",
    "NP",
    "CH",
    "HF",
    "ZGT"
};

/*
 * Telescope model parameter descriptions.
 */
static const char * const ParamDescrs[] = {
    "Azimuth misalignment - MA",
    "Elevation  misalignment - ME",
    "Non-perpendicularity - NP",
    "Collimation error in HA - CH",
    "Horseshoe flexure - HF",
    "Z-gear transducer reading"
};
    
#define MA 24.71
#define ME 37.9
#define NP -1.65
#define CH -32.0
#define HF -7.24
#define ZGT 0

/*
 * Telescope parameter default values.
 */
static const double ParamDefaults[] = {
    MA, ME, NP, CH, HF, ZGT 
};
    

/*
 * Internal Function name:
     TdfFpilFull___AccessCons
  
 * Description:
      Accesses the 2dF positioner constants file.
  
 * History:
     10-Dec-1999 - TJF - Original version.
     10-Mar-2000 - TJF - Add extra arguments to support search paths
                         and reporting of which file was opended and
                         implemented that functionality.
     25-Feb-2013 - TJF - Renamed from AccessCons() to  TdfFpilFull___AccessCons()
                          so it can be used by extracted AAOmega and
                          HERMES specific code.  Made extern.
 */
extern void TdfFpilFull___AccessCons (
    const char *baseFile,
    const char *dirOverride,
    const char *dirFallback,
    FpilRptRoutineType RptRoutine,
    void * RptArg,
    SdsIdType *id,
    StatusType *status)
{
    char filename[PATH_MAX];

    if (*status != STATUS__OK) return;

    ErsPush();
/*
 *  First attempt, using the override directory if we have one.
 */
    if (dirOverride)
    {
        ErsSPrintf(sizeof(filename), filename, "%s/%s", 
                   dirOverride, baseFile);
        SdsRead(filename, id, status);
    }
    if ((!dirOverride)||(*status == SDS__FOPEN)||(*status == SDS__FREAD))
    {
/*
 *      It is not in the override directory, try the /instsoft/2dF/positioner
 *      directory.
 */

        ErsAnnul(status);
        ErsSPrintf(sizeof(filename), filename, "%s/%s", 
                   TDFCONSDIR, baseFile);
        SdsRead(filename, id, status);
        if (((*status == SDS__FOPEN)||(*status == SDS__FREAD))&&dirFallback)
        {
/*
 *           It is not in the default directory, the fallback
 *           directory if these is one.
 */
            ErsAnnul(status);
            ErsSPrintf(sizeof(filename), filename, "%s/%s", 
                       dirFallback, baseFile);
            SdsRead(filename, id, status);
            
        }
    }
    ErsPop();
    if ((*status == SDS__FOPEN)||(*status == SDS__FREAD))
    {
        ErsRep(0,status,"Failed to open 2dF constants file \"%s\"",
               baseFile);

        if (dirOverride)
            ErsRep(0,status,"   Tried looking in \"%s\"",dirOverride);

        ErsRep(0,status,"   Tried looking in \"%s\"",TDFCONSDIR);

        if (dirFallback)
            ErsRep(0,status,"   Tried looking in \"%s\"",dirFallback);
    }
    else if (*status != STATUS__OK)
    {
        ErsRep(0,status,"Failed to open 2dF constants file \"%s\"",
               baseFile);
        ErsRep(0, status, "General failure, status code %lx",
               (long int )*status);
    }
    else if (RptRoutine)
        (*RptRoutine)(RptArg,filename,status);
        
}


/*
 * Internal Function name:
      AccessCons400
  
 * Description:
      Accesses the 2dF positioner constants file (400 fibre version)
  
 * History:
     15-Dec-2004 - TJF - Original version.
 */
static void AccessCons400 (
    void * clientData DUNUSED,
    const char *dirOverride,
    const char *dirFallback,
    FpilRptRoutineType RptRoutine,
    void * RptArg,
    SdsIdType *id,
    StatusType *status)
{
    TdfFpilFull___AccessCons(TDFCONSFILE400, dirOverride, dirFallback, RptRoutine, RptArg, id, status);
}


/*
 * Internal Function name:
      AccessCons404
  
 * Description:
      Accesses the 2dF positioner constants file (404 fibre version)
  
 * History:
     15-Dec-2004 - TJF - Original version.
 */
static void AccessCons404 (
    void * clientData DUNUSED,
    const char *dirOverride,
    const char *dirFallback,
    FpilRptRoutineType RptRoutine,
    void * RptArg,
    SdsIdType *id,
    StatusType *status)
{
    TdfFpilFull___AccessCons(TDFCONSFILE404, dirOverride, dirFallback, RptRoutine, RptArg, id, status);
}



/*
 * Internal Function name:
      WhichSpec404
  
 * Description:
      Determine the spectrograph associated with a pivot.  (Old 2dF field plate with
      404 fibres).

      For 2dF, we assume fibre one goes to spectrograph two, and that
      both plates have the same configuration.
  
 * History:
     16-Mar-2000 - TJF - Original version, based on Gavin's Dalton's code
                         which had been in the configure.c ConfWhichSpec
                         function.
      5-Dec-2001 - KS  - SpecNames now contains three names - guides are
                         handled like just another fibre type.
     14-Dec-2004 - TJF - 2dF now has 400 fibres rather then 404.  This version renamed
                          so that configure can support both.
 */
static void WhichSpec404 (
    void * clientData DUNUSED,
    unsigned field    DUNUSED,
    unsigned pivot,
    SdsIdType Constants DUNUSED,
    unsigned *spec,
    const char ** SpecName,
    StatusType *status)
{
    div_t result;

    if (*status != STATUS__OK) return;

/*
 *  There are 404 fibres, numbered 1 to 404, on 40 retractors.  Alternative
 *  retractors goto different spectrographs.  Most retractors have 10
 *  fibers, but 10, 20, 30 and 40 have an extra fibre, being the guide 
 *  fibres, such that fibres 101, 202, 303 and 404 are guide fibres.
 *
 *  Divide by 101
 */    
    result = div(pivot,101);
    if (result.rem == 0) 
    {
/*
 *      Remainder is 0, this must be a guide fibre.
 */
        *spec = 0;
    }
    else
    {
        int val3;
/*
 *      result.rem is now in the range 1 to 100.
 *      val2 will be in the range 0 to 99.  
 */
        int val2 =  result.rem - 1;
/*
 *      Divide by 10, the number of fibres on each retractor.
 */
        result = div(val2,10);

/*
 *      The quotient indicates the retractor number within the quadrant and
 *      will be in teh range 0 to 9.
 */
        val3 = result.quot;
/*
 *      Divide this by two.  The reminder (0/1) indicates the spectrograph set.
 */
        result = div(val3,2);
        if (result.rem == 0)
        {
            /* Spectrograph is number 2 */
            *spec = 2;
        }
        else
        {
            *spec = 1;
        }
    }
    *SpecName = SpecNames[*spec];
}

/*
 * Internal Function name:
      WhichSpec400
  
 * Description:
      Determine the spectrograph associated with a pivot.

      For 2dF, we assume fibre one goes to spectrograph two, and that
      both plates have the same configuration.
  
 * History:
     16-Mar-2000 - TJF - Original version, based on Gavin's Dalton's code
                         which had been in the configure.c ConfWhichSpec
                         function.
      5-Dec-2001 - KS  - SpecNames now contains three names - guides are
                         handled like just another fibre type.
     14-Dec-2004 - TJF - 2dF now has 400 fibres rather then 404.  Create 
                         this version and rename old version to WhichSpec404.
      7-Jan-2004 - TJF - Get the last fibre on the retactor correct.
 */
static void WhichSpec400 (
    void * clientData DUNUSED,
    unsigned field    DUNUSED,
    unsigned pivot,
    SdsIdType Constants DUNUSED,
    unsigned *spec,
    const char ** SpecName,
    StatusType *status)
{
    div_t result;

    if (*status != STATUS__OK) return;

/*
 *  There are 400 (NUM_PIVOTS_2DF) fibres, numbered 1 to 400, on 40 retractors.  
 *  Alternative retractors goto different spectrographs.  Fibres 100, 200, 300
 *  and 400 are guide fibres.
 *
 */    
    result = div(pivot,100);
    if (result.rem == 0) 
    {
/*
 *      Remainder is 0, this must be a guide fibre.
 */
        *spec = 0;
    }
    else
    {
        int val3;
/*
 *      result.rem is now in the range 1 to 99, the fibre number
 *      within the quadrant, with fibre 100 on the quadrant handled above.
 *
 *      Divide by 10, the number of fibres on each retractor
 */
        result = div(result.rem,10);

/*
 *      If the remainder is non-zero, quotient indicates the retractor 
 *      number within the quadrant and will be in the range 0 to 9.
 *
 *      If the remainder is 0, then we are the last fibre in the previous
 *      retractor.  Decrement the quotant to ensure this value is
 *      consistent.
 */        
        if  (result.rem == 0)
            val3 = result.quot - 1;
        else
            val3 = result.quot;
/*
 *      Divide this by two.  The reminder (0/1) indicates the spectrograph 
 *      set.
 */
        result = div(val3,2);
        if (result.rem == 0)
        {
            /* Spectrograph is number 2 */
            *spec = 2;
        }
        else
        {
            *spec = 1;
        }
    }
    *SpecName = SpecNames[*spec];
}




/*
 * Internal Function name:
      AccessModel
  
 * Description:
      Mallocs spec for the 2dF model data structure.

      Reads the 2dF distortion and linear models from files.
  
 * History:
     10-Dec-1999 - TJF - Original version.
     10-Mar-2000 - TJF - Add extra arguments to support search paths
                         and reporting of which file was opended and
                         implemented that functionallity.
     23-Mar-2007 - KS  - Added (rather late!) support for field 2, which
                         uses a combination of the data from the two real
                         plates, and so needs to read both sets of input files.
 */
static void AccessModel(
    void * clientData DUNUSED,
    const char *dirOverride,
    const char *dirFallback,
    FpilRptRoutineType RptRoutine,
    void * RptArg,
    unsigned int field,
    FpilModelType ** model,
    StatusType *status)
{
    char DistFilename[PATH_MAX];
    char LinFilename[PATH_MAX];
    char DistFilename1[PATH_MAX];
    char LinFilename1[PATH_MAX];
    char *DistNamePart;
    char *LinNamePart;
    char *DistNamePart1 = "";
    char *LinNamePart1 = "";
    TdfModelPnt MyModel;
    if (*status != STATUS__OK) return;

    *model = 0;
    if ((MyModel = malloc(sizeof(TdfModel))) == 0)
    {
        *status = TDFFPIL__MALLOCERR;
        return;
    }

/*
 *  Work out which files we are looking for, the names depend on the field.
 */
    if (field == 0)
    {
        DistNamePart = TDF_P0_DISTFILE;
        LinNamePart   = TDF_P0_LINFILE;
    }
    else if (field == 1) 
    {
        DistNamePart = TDF_P1_DISTFILE;
        LinNamePart   = TDF_P1_LINFILE;
    }
    else if (field == 2) 
    {
        DistNamePart = TDF_P0_DISTFILE;
        LinNamePart   = TDF_P0_LINFILE;
        DistNamePart1 = TDF_P1_DISTFILE;
        LinNamePart1   = TDF_P1_LINFILE;
    }
    else
    {
        ErsRep(0,status,
               "Invalid field (%d) specified for distortion data",field);
        *status = TDFXY__ILLFIELD;
        return;
    }
        
/*
 *  We need to open both the 2dF distortion model and linear model
 *  files.  We want to pick up both files from the same place.
 */
    ErsPush();
/*
 *  First attempt, using the override directory if we have one. Remember
 *  that plate 2 needs both sets of input files.
 */
    if (dirOverride)
    {
        ErsSPrintf(sizeof(DistFilename), DistFilename, "%s/%s", 
                   dirOverride, DistNamePart);

        ErsSPrintf(sizeof(LinFilename), LinFilename, "%s/%s", 
                   dirOverride, LinNamePart);
        if (field != 2) 
        {
            TdfGetDist(DistFilename, &MyModel->dist, status);
            TdfGetLin(LinFilename, &MyModel->linpars, status);
        }
        else
        {
            ErsSPrintf(sizeof(DistFilename1), DistFilename1, "%s/%s", 
                   dirOverride, DistNamePart1);

            ErsSPrintf(sizeof(LinFilename1), LinFilename1, "%s/%s", 
                   dirOverride, LinNamePart1);

            TdfGetDist2(DistFilename, DistFilename1, &MyModel->dist, status);
            TdfGetLin2(LinFilename, LinFilename1, &MyModel->linpars, status);
        }
    }
    if ((!dirOverride)||(*status == SDS__FOPEN)||(*status == SDS__FREAD))
    {
/*
 *      It is not in the override directory, try the /instsoft/2dF/positioner
 *      directory.
 */

        ErsAnnul(status);
        ErsSPrintf(sizeof(DistFilename), DistFilename, "%s/%s", 
                   TDFMODELDIR, DistNamePart);

        ErsSPrintf(sizeof(LinFilename), LinFilename, "%s/%s", 
                   TDFMODELDIR, LinNamePart);

        if (field != 2) 
        {
            TdfGetDist(DistFilename, &MyModel->dist, status);
            TdfGetLin(LinFilename, &MyModel->linpars, status);
        }
        else
        {
            ErsSPrintf(sizeof(DistFilename1), DistFilename1, "%s/%s", 
                   TDFMODELDIR, DistNamePart1);

            ErsSPrintf(sizeof(LinFilename1), LinFilename1, "%s/%s", 
                   TDFMODELDIR, LinNamePart1);

            TdfGetDist2(DistFilename, DistFilename1, &MyModel->dist, status);
            TdfGetLin2(LinFilename, LinFilename1, &MyModel->linpars, status);
        }

        if (((*status == SDS__FOPEN)||(*status == SDS__FREAD))&&dirFallback)
        {
/*
 *           It is not in the default directory, the fallback
 *           directory if there is one.
 */
            ErsAnnul(status);
            ErsSPrintf(sizeof(DistFilename), DistFilename, "%s/%s", 
                       dirFallback, DistNamePart);

            ErsSPrintf(sizeof(LinFilename), LinFilename, "%s/%s", 
                   dirFallback, LinNamePart);

            if (field != 2) 
            {
                TdfGetDist(DistFilename, &MyModel->dist, status);
                TdfGetLin(LinFilename, &MyModel->linpars, status);
            }
            else
            {
                ErsSPrintf(sizeof(DistFilename1), DistFilename1, "%s/%s", 
                   dirFallback, DistNamePart1);

                ErsSPrintf(sizeof(LinFilename1), LinFilename1, "%s/%s", 
                   dirFallback, LinNamePart1);

                TdfGetDist2(DistFilename, DistFilename1, &MyModel->dist, status);
                TdfGetLin2(LinFilename, LinFilename1, &MyModel->linpars, status);
            }            
        }
    }
    ErsPop();
    if (*status != STATUS__OK)
    {
        if (field != 2) 
        {
            ErsRep(0,status,
               "Failed to open 2dF telescope model files \"%s\" and \"%s\"",
               DistNamePart, LinNamePart);
        }
        else
        {
            ErsRep(0,status,
               "Failed to open 2dF model files \"%s\",\"%s\",\"%s\",\"%s\"",
               DistNamePart, LinNamePart, DistNamePart1, LinNamePart1);
        }
    }
    if ((*status == SDS__FOPEN)||(*status == SDS__FREAD))
    {
        if (dirOverride)
            ErsRep(0,status,"   Tried looking in \"%s\"",dirOverride);

        ErsRep(0,status,"   Tried looking in \"%s\"",TDFMODELDIR);

        if (dirFallback)
            ErsRep(0,status,"   Tried looking in \"%s\"",dirFallback);
    }
    else if (*status != STATUS__OK)
    {
        ErsRep(0, status, "General failure, status code %lx",
               (long int )*status);
    }
    else if (RptRoutine)
    {
        (*RptRoutine)(RptArg,DistFilename,status);
        (*RptRoutine)(RptArg,LinFilename,status);
    }
            


    if (*status == STATUS__OK)
        *model = (FpilModelType *)MyModel;
    else
        free(MyModel);
    
}


/*
 * Internal Function name:
      CloneModel
  
 * Description:
      Mallocs spec for a 2dF model data structure and copies in data from a source model.

  
 * History:
     22-Aug-2008 - TJF - Original version.
 */
static void CloneModel(
    const FpilModelType * modelSource,
    FpilModelType ** modelClone,
    StatusType *status)
{
    TdfModelPnt MyModelSource = (TdfModelPnt)(modelSource);
    TdfModelPnt MyModelClone;

    

    if (*status != STATUS__OK) return;

    *modelClone = 0;
    if ((MyModelClone = malloc(sizeof(TdfModel))) == 0)
    {
        *status = TDFFPIL__MALLOCERR;
        return;
    }
    *MyModelClone = *MyModelSource;
    *modelClone = (FpilModelType *)MyModelSource;
    
}



/*
 * Internal Function name:
       CvtInit
  
 * Description:
       Initialise for transformation of RA,Dec to 2dF x,y.
  
 * History:
     10-Dec-1999 - TJF - Original version.
     30-Sep-2002 - TJF - Add cenWave argument.  Rename wave to obsWave.
  
 */
static void CvtInit(
    FpilModelType *model,
    double mjd,
    double dut,
    double temp,
    double pres,
    double humid,
    double cenWave,
    double obsWave,
    const double *params,
    FpilStatusType *status)
{
    TdfModelPnt MyModel = (TdfModelPnt)model;
    if (*status != STATUS__OK) return;

    TdfXyInit(mjd, dut, temp, pres, humid, cenWave, obsWave,
              params[0],
              params[1],
              params[2],
              params[3],
              params[4],
              params[5],
              0,  /* ADC Position angle - element A, assume 0 */
              0,  /* ADC Position angle - element B, assume 0 */
              MyModel->dist,
              &MyModel->xypars,
              status);

}
/*
 * Internal Function name:
       Rd2Xy
  
 * Description:
       Determine 2dF X/Y from RA/Dec
  
 * History:
     10-Dec-1999 - TJF - Original version.
  
 */
static void Rd2Xy(
    const FpilModelType *model,
    double cra,
    double cdec,
    double ra,
    double dec,
    double mjd,
    double *x,
    double *y,
    FpilStatusType *status)
{
    TdfModelPnt MyModel = (TdfModelPnt)model;
    if (*status != STATUS__OK) return;

    TdfRd2xy(&MyModel->xypars, 
             cra, cdec, ra, dec, mjd, x, y, status);

}
/*
 * Internal Function name:
      Xy2Pos
 * Description:
      Convert from field plate coordinates to positioner coordinates.
  
 * History:
     10-Dec-1999 - TJF - Original version.
  
 */
static void Xy2Pos (
    const FpilModelType *model,
    double x,
    double y,
    double *xp,
    double *yp,
    FpilStatusType *status)
{
    TdfModelPnt MyModel = (TdfModelPnt)model;
    if (*status != STATUS__OK) return;

    TdfXy2pos(&MyModel->linpars, x, y, xp, yp);
}
/*
 * Internal Function name:
       Xy2Rd

 * Description:
       Determine apparent RA and DEC from 2dF x and y.
  
 * History:
     10-Dec-1999 - TJF - Original version.
  
 */
static void Xy2Rd  (
    const FpilModelType *model,
    double cra,
    double cdec,
    double x,
    double y,
    double mjd,
    double *ra,
    double *dec,
    FpilStatusType *status)

{
    TdfModelPnt MyModel = (TdfModelPnt)model;
    if (*status != STATUS__OK) return;

    TdfXy2rd(&MyModel->xypars, 
             cra, cdec, x, y, mjd, ra,dec, status);

}
/*
 * Internal Function name:
      Pos2Xy

 * Description:
      Convert positioner coordinates to field plate coordinates
  
 * History:
     10-Dec-1999 - TJF - Original version.
  
 */
static void Pos2Xy  (
    const FpilModelType *model,
    double xp,
    double yp,
    double *x,
    double *y,
    FpilStatusType *status)
{
    TdfModelPnt MyModel = (TdfModelPnt)model;
    if (*status != STATUS__OK) return;

    TdfPos2xy(&MyModel->linpars, xp, yp, x, y);
}
/*
 * Internal Function name:
      QuickAppToObs

 * Description:
     Quick apparent to observed position conversion.

     Just a straight conversion to the routine slaAopqk using the
     aoparms compoents of xypars (we don't have a proper way of
     accessing this).

  
 * History:
     13-Mar-2000 - TJF - Original version.
  
 */
static void QuickAppToObs  (
    const FpilModelType *model,
    double ra,
    double dec,
    double *aob,
    double *zob,
    double *hob,
    double *dob,
    double *rob,
    FpilStatusType *status)
{
    TdfModelPnt MyModel = (TdfModelPnt)model;
    if (*status != STATUS__OK) return;

    slaAopqk(ra,dec,MyModel->xypars.obsAoprms,aob,zob,hob,dob,rob);

}
/*
 * Internal Function name:
      DrawButFib

 * Description:
      Return the appropiate information for Drawing a button and fibre.

      The Tcl/Tk cavans widget graphic we return is for a polygon
      which represents a 2dF button.

      Note, will return the bad status TDFXY__SPRINTF if the
      available string length is insufficent for the button
      graphic string.

  
 * History:
     16-Mar-2000 - TJF - Original version, based on the original 2dF
                         configure button and fibre drawing 
                         code (ConfGetButton()).
  
 */
static void DrawButFib  (
        void * clientData DUNUSED,
        double scaling DUNUSED,
        double zoom,            
        FpilCvtRoutineType CvtX,  /* For converting Microns in X to pixels */
        FpilCvtRoutineType CvtY,  /* For converting Microns in Y to pixels */
        void * CvtData,           /* To be passed to CvtX and CvtY         */
        double X,                 /* Fibre position in X, microns on plate */
        double Y,                 /* Fibre position in Y, microns on plate */
        double Theta,             /* Button theta in radians               */
        int Type DUNUSED,         /* Fibre type                            */
        unsigned ButGrapLen,      /* Length of ButGrapic string            */
        double *FibX,             /* Fibre X position in GUI returned here */
        double *FibY,             /* Fibre Y position in GUI returned here */
        char *ButGraphic1,        /* Tcl/Tk cavnas create command arugment 
                                     for drawing the button is returned 
                                     here                                  */
        char *ButGraphic2 DUNUSED,/* As per ButGraphic2, but not required 
                                     for 2dF and can be ignored            */
    
        FpilStatusType *status)
{
    register unsigned i;
    float xb[12],yb[12];
    char buffer[40];
    char tempString[500];
    float x0,y0;
    float c,s;
    if (*status != STATUS__OK) return;
    
/*
 *  See tdFbuttonDims.h for a description of a 2dF button, but rotate
 *  the diagram by 90 degrees clockwise to get the coorindate system
 *  used here.
 *
 *  The orgin of a button is the fibre location.
 * 
 *  We need to create a Tk canvas widget polygon, so we need
 *  a list of points for that command. The xb array represents
 *  the 12 points needs 
 *
 *  First create one side of the button
 */
    xb[0] = -BUTTON_PHW;   /* Top left corner of microprism */
    yb[0] = BUTTON_PLE;

    xb[1] = -BUTTON_PHW;   /* Bottom left corner of microprism */
    yb[1] = BUTTON_BLE;

    xb[2] = -BUTTON_BHW;   /* Top left corner of button itself */
    yb[2] = BUTTON_BLE;

    xb[3] = -BUTTON_BHW;   /* Bottom left corner of button itself */
    yb[3] = BUTTON_BTE;

    xb[4] = -BUTTON_THW;   /* Top left corner of tail (where it exits */
    yb[4] = BUTTON_BTE;    /* the button) */

    xb[5] = -BUTTON_THW;   /* Bottom left cornder of tail */
    yb[5] = BUTTON_FVP;

/*
 *  Now we just need the mirror reflection about the y axis to
 *  get the other side of the button.
 */    
    for (i=0 ; i<6 ; i++) 
    { 
        xb[11-i] = -xb[i];
        yb[11-i] = yb[i];
    }
/*
 *  Determine the rotation of the button.
 */
    c = cos(Theta);
    s = sin(Theta);

/*
 *  Now work through, applying the rotation and converting the
 *  position to pixel units.
 */
    for (i=0 ; i<12 ; i++)
    {
        x0 = xb[i]; y0 = yb[i];
/*
 *      Rotate as required.
 */
        xb[i] = c*x0 - s*y0 + X;
        yb[i] = c*y0 + s*x0 + Y;
/*
 *      Convert to pixel units.
 */
        xb[i] = (*CvtX)(CvtData,xb[i], zoom);
        yb[i] = (*CvtY)(CvtData,yb[i], zoom);
    }
/*
 *  Determine the fibre end.  This is where the tail ends. so
 *  average the x and y values of each side of the tail end to
 *  get the fibre end.
 */
    *FibX = (xb[5]+xb[6])*0.5;
    *FibY = (yb[5]+yb[6])*0.5;
   
/*
 *  Create our Tcl/Tk canvas create arguments - to create the polygon.
 */
    strcpy(tempString,"polygon ");

    for (i=0;i<12;i++)
    {
        sprintf(buffer,"%f %f ",xb[i],yb[i]);
        strcat(tempString,buffer);
    }

/*
 *  Copy into the output string, validating the length as we go.
 */
    if ((strlen(tempString)+1) >= ButGrapLen)
        *status = TDFXY__SPRINTF;
    else
        strncpy(ButGraphic1, tempString, ButGrapLen);

}


/*
 * Internal Function name:
     TdfFpilFull___SetupFullStructs

 * Description:

  
 * History:
     15-Sep-2004 - TJF - Original version, based on original TdfFpilFullInit() but
                         stripped down to allow a 400 and 404 fibre version.  
     22-Jun-2008 - TJF - Support clone.
     25-Feb-2013 - TJF - Renamed from  SetupFullStructs() to
                           TdfFpilFull___SetupFullStructs() so it can be called
                           by extracted AAOmega and HERMES code. Make extern.
 */
extern void TdfFpilFull___SetupFullStructs(
    FpilCollisionChecksType *routines,
    FpilInstParamsType      *params,
    FpilModelRoutinesType   *modelRoutines)
{
    TdfFpil___SetupMinStructs(routines,params);

    routines->AccessCons    = AccessCons400;
    routines->WhichSpec     = WhichSpec400;
    routines->DrawButFib    = DrawButFib;

    params->NumSpec         = 3;
    params->SpecNames       = SpecNames;
        
    
    modelRoutines->Access  = AccessModel;
    modelRoutines->Free    =  0;   /* No free routine needed */
    modelRoutines->CvtInit = CvtInit;
    modelRoutines->Rd2Xy   = Rd2Xy;
    modelRoutines->Xy2Pos  = Xy2Pos;
    modelRoutines->Xy2Rd   = Xy2Rd;
    modelRoutines->Pos2Xy  = Pos2Xy;
    modelRoutines->QuickAppToObs = QuickAppToObs;
    modelRoutines->Clone   = CloneModel;

    modelRoutines->NumParams = 6;
    modelRoutines->ParamNames = ParamNames;
    modelRoutines->ParamDescrs = ParamDescrs;
    modelRoutines->ParamDefaults = ParamDefaults;   
}


 

/*
 *+		        T d f F p i l F u l l I n i t 4 0 0

 * Function name:
      TdfFpilFullInit400

 * Function:
      Initialise Fpil with the 2dF instrument model, full FPIL model

 * Description:
      Calls FpilInit with appropiate values for 2dF (400 fibre version).  
      Returns an appropiate FpilType variable to be used in calls to Fpil routines.

      When you are finished with the "inst" varaible, you should pass it
      to FpilFree(3) to free allocated resources.

      The instrument name string will be "2dF".  The Telescope
      name string will be "AAT".

 * Language:
      C

 * Call:
      (void) TdfFpilFullInit400(inst)

 * Parameters:   (">" input, "!" modified, "W" workspace, "<" output)
      (<) inst (FpilType *) The FPIL instrument description to be used
                            in calls to Fpil routines.


 * Include files: tdffpil.h

 * See Also: {references}

 * Support: Tony Farrell, AAO

 *-

 * History:
     07-Dec-1999 - TJF - Original version.
      5-Dec-2001 - KS  - Now considers guide fibres to be just another
                         fibre type, of which 2dF has three.
     14-Dec-2004 - TJF - Rename from TdfFpilFullInit() to TdfFpilFullInit400() and
                         split to create SetupFullStruct(), such that
                         we can create TdfFpilFullInit404().
                        
                        
 */


extern void TdfFpilFullInit400(FpilType *inst)
{
    FpilCollisionChecksType routines;
    FpilInstParamsType      params;
    FpilModelRoutinesType   modelRoutines;

    TdfFpilFull___SetupFullStructs(&routines, &params, &modelRoutines);
    FpilInit(0, FPIL_VERSION, &routines, &params, &modelRoutines, inst);
}





/*
 *+		        T d f F p i l F u l l I n i t 4 0 4

 * Function name:
      TdfFpilFullInit404

 * Function:
      Initialise Fpil with the 2dF instrument model, full FPIL model - 404 fibre versions.

 * Description:
      Calls FpilInit with appropiate values for 2dF (404 fibre version).  
      Returns an appropiate FpilType variable to be used in calls to Fpil routines.

      This version allows AAO Configure to read in configuration files
      set up on the old 2dF.

      When you are finished with the "inst" varaible, you should pass it
      to FpilFree(3) to free allocated resources.

      The instrument name string will be "2dF (old - 404 fibres)".  The Telescope
      name string will be "AAT".

 * Language:
      C

 * Call:
      (void) TdfFpilFullInit404(inst)

 * Parameters:   (">" input, "!" modified, "W" workspace, "<" output)
      (<) inst (FpilType *) The FPIL instrument description to be used
                            in calls to Fpil routines.


 * Include files: tdffpil.h

 * See Also: {references}

 * Support: Tony Farrell, AAO

 *-

 * History:
     15-Dec-2004 - TJF - Original version, based on TdfFpilFullInit()
     04-Jan-2004 - TJF - Rename to 2dF-old-404, rather then having spaces.
                        
                        
 */


extern void TdfFpilFullInit404(FpilType *inst)
{
    FpilCollisionChecksType routines;
    FpilInstParamsType      params;
    FpilModelRoutinesType   modelRoutines;

    TdfFpilFull___SetupFullStructs(&routines, &params, &modelRoutines);

    /*
     * Modify things such that the resulting instrument is appropiate for
     * the old 2dF - with 404 fibres.
     */
    routines.WhichSpec     = WhichSpec404;
    routines.AccessCons    = AccessCons404;
    params.NumPivots       = 404;
    strncpy(params.Instrument,"2dF-old-404", sizeof(params.Instrument));
                                                              /* Max 20 characters */
    params.Instrument[sizeof(params.Instrument)-1] = '\0';

    FpilInit(0, FPIL_VERSION, &routines, &params, &modelRoutines, inst);
}

