/*+		T D F F P I L F U L L _ A A O M E G A
 * Module name:
     TDFFPILFULL_AAOMEGA

 * Function:
      AAOmega specific versions of routines needed by TdfFpilFullInit(3).

 * Description:
      This contains the routine TdfFpilFullInitAAOmega() used to initialise
      access to the AAOmega FPIL library, full mode.
     
 * Language:
      C

 * Support:
      Tony Farrell, AAO
      Keith Shortridge, AAO

 *-

 *     @(#) $Id: ACMM:2dFutil/tdffpilfull_aaomega.c,v 5.2 10-Sep-2013 09:12:11+10 tjf $

 * History:
     25-Feb-2013 - TJF - Original version, extracted from tdffpilfull.c in
                          prep for HERMES work.

 */
/*
 *  Include files
 */
#define __EXTENSIONS__   /* To get PATH_MAX on solaris */


static char *rcsId="@(#) $Id: ACMM:2dFutil/tdffpilfull_aaomega.c,v 5.2 10-Sep-2013 09:12:11+10 tjf $";
static void *use_rcsId = (0 ? (void *)(&use_rcsId) : (void *) &rcsId);

#include <limits.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <stdio.h>
#include "drama.h"
#include "sds.h"

#define TDFFPIL_BUILD  /* Used by the include file */
#include "tdffpil.h"

#include "tdfxy.h"
#include "slalib.h"
#include "tdFbuttonDims.h"
#include "Ers.h"
#include "tdfxy_err.h"

#ifndef PATH_MAX
#define PATH_MAX 1024
#endif

#ifdef DEBUG
#undef DEBUG
#endif
#define DEBUG printf

/* 
 * Constant definitions
 */
#define TDFCONSFILE400 "tdFconstants400.sds"      /* 2dF constants file - 400 fibre ver */
#define TDFCONSFILEDF "tdFconstantsDF.sds"      /* 2dF constants file - 400 dual-fibre (AAOmega/HERMES) version */

/*
 * Macro function definitions
 */

/*
 * type definitions
 */
typedef struct {
    FpilModelType model;
    TdfDistType   dist;    /* Set by AccessModel using TdfGetDist() */
    TdfLinType    linpars; /* Set by AccessModel using TdfGetLin()  */
    TdfXyType     xypars;  /* Set by CvtInit using TdfXyInit()      */
} TdfModel, *TdfModelPnt;

/*
 * Global variables and functions imported (but not declared in include files)
 */

/*
 * Global variables exported (definitions and initialization of globals
 * declared in the associated include file)
 */

/*
 * Variables private to this module (static variables)
 */

/*
 * Spectrograh names, used by WhichSpec() and TdfFpilFullInit()
 */


static const char * const AAOmegaSpecNames[] = { "Guide", "AAOmega"};

/* Originally, it was not clear what the pivot numbers for the AAOmega plates
 * would be, so the code was written to use a look-up table that could be
 * changed later. The number of pivots is deduced from the size of this
 * array.
 */
static const int AAOmegaGuidePivs[] = { 50, 100, 150, 200, 250, 300, 350, 400 };



/*
 * Internal Function name:
     TdfFpilFull_CreateCombinedPlate
  
 * Description:
     Given a source SDS structure describing 2dF plates 0 and 1, create
     data for plate 2 - a representation of a combined configuration that
     should be suitable for configuring either plate.


     Note - if sourceId is an external SDS item, it will be copied and
     then deleted in the presumition it was read using SdsRead().  If it 
     is NOT an external item, it will be modified in place.
 
     In either case, on return, *id points to the resulting structure.


  
 * History:
     26-Feb-2013 - TJF - Original version, extracted from AccessConsAAOmega
 */

extern void TdfFpilFull_CreateCombined(
    const SdsIdType sourceId,
    SdsIdType * const id,
    StatusType * const status)
{
    int sourceExtern;
    SdsIdType field0InUseId;
    SdsIdType field1InUseId;
    SdsIdType field2InUseId;
    SdsIdType field0ItemId;
    SdsIdType field1ItemId;
    SdsIdType field2ItemId;
    SdsIdType field0Id;
    SdsIdType field1Id;
    SdsIdType field2Id;
    unsigned short *data1;
    unsigned short *data0;
    unsigned short *data2;
    int Item;
    int *IntData0;
    int *IntData1;
    int *IntData2;
    double *DoubleData0;
    double *DoubleData1;
    double *DoubleData2;
    unsigned long actlen;
    unsigned long Dims[1];
    int Elements;
    int Elem1;
    int i;
    
    static char* IntItemNames[] = { "xPiv", "yPiv", "xPark", "yPark", 
                                                         "graspX", "graspY" };
    static int IntItems = sizeof(IntItemNames)/sizeof(char*);
    static char* DoubleItemNames[] = { "tPark" };
    static int DoubleItems = sizeof(DoubleItemNames)/sizeof(char*);
    


    if (*status != STATUS__OK) return;

    /*  If the structure  has been read in directly by SdsRead(), it
     *  will be an external Sds structure and cannot be modified.  (We
     *  could change the values, but not the structure itself).  We want
     *  to add a new field2 that represents the combination of both plates. 
     *  So we first, we copy the structure to an internal structure.
     *
     *  In some cases, sourceId may not be external (HERMES case) in which case
     *  we can modify it in place.
     */
    SdsIsExternal(sourceId, &sourceExtern, status);
    if (sourceExtern) {
        SdsCopy(sourceId,id,status);
        SdsReadFree(sourceId, status);
        SdsFreeId(sourceId, status);
    } else {
        *id = sourceId;
    }

    
    /*  Now we create an empty field2 structure that we can fill up item 
     *  by item. (My original code to do this used SdsCopy to copy field0,
     *  then SdsInsert to insert it into the structure, but this seemed to
     *  cause a number of intermittent segmentation errors, as though some
     *  undefined variable was being used in memory accesses. Doing it by
     *  an item-wise copy of the structure does seem to work. Moreover,
     *  that original code made field2 a copy of field0 and then changed
     *  field0 so it represented the merged fields. It was eventually
     *  decided field2 should be the merged field.)
     *
     *  (Added later - 29/6/06: It may be that the problem with the use of
     *  SdsInsert was due to the external Id being freed too early, but I've
     *  not tried to re-instate the old code to check.)
     */
     
    SdsNew(*id,"field2",0,0,SDS_STRUCT,0,0,&field2Id,status);
    SdsFind(*id, "field1", &field1Id, status);
    SdsFind(*id, "field0", &field0Id, status);

    /*
     *  Find the inUse items on each plate. Create the new inUse item 
     *  for plate 2, and get pointers to all three inUse arrays.
     */
     
    SdsFind(field0Id, "inUse", &field0InUseId, status);
    SdsPointer(field0InUseId, (void *)&data0, &actlen, status);
    Elements = actlen / sizeof(data0[0]);
    SdsFind(field1Id, "inUse", &field1InUseId, status);
    SdsPointer(field1InUseId, (void *)&data1, &actlen, status);
    Elem1 = actlen / sizeof(data1[0]);
    if (Elements > Elem1) Elements = Elem1;
    Dims[0] = Elements;
    SdsNew(field2Id, "inUse", 0, 0, SDS_USHORT,1,Dims,&field2InUseId, status);
    SdsPointer(field2InUseId, (void *)&data2, &actlen, status);
    
    /*  And now, also set to zero the field 2 in use flags for any fibre
     *  flagged in either field as not in use. Field 2 now has the superset of
     *  the field 0 and field 1 in use flags.
     */
    
    if (*status == 0) { 
        for (i = 0; i < Elements; i++) {
            data2[i] = data0[i];
            if (data1[i] == 0) data2[i] = 0;
        }
    }
    
    SdsFreeId(field2InUseId,status);
    SdsFreeId(field1InUseId,status);
    SdsFreeId(field0InUseId,status);
    
    /*  Now handle the rest of the items in the field. We set the xPiv,
     *  yPiv, xPark, yPark, graspX and graspY arrays in field 2 to be
     *  the average of the two real fields. We can do all the integer
     *  arrays in one loop.
     */
    
    for (Item = 0; Item < IntItems; Item++) {
        SdsFind(field0Id, IntItemNames[Item], &field0ItemId, status);
        SdsPointer(field0ItemId, (void *)&IntData0, &actlen, status);
        Elements = actlen / sizeof(IntData0[0]);
        SdsFind(field1Id, IntItemNames[Item], &field1ItemId, status);
        SdsPointer(field1ItemId, (void *)&IntData1, &actlen, status);
        Elem1 = actlen / sizeof(IntData1[0]);
        if (Elements > Elem1) Elements = Elem1;
        SdsNew(field2Id, IntItemNames[Item], 0, 0, SDS_INT,1,Dims,
                                              &field2ItemId, status);
        SdsPointer(field2ItemId, (void *)&IntData2, &actlen, status);
    
        if (*status == 0) {
            for (i = 0; i < Elements; i++) {
                IntData2[i] = (IntData0[i] + IntData1[i]) / 2;
            }
        }
    
        SdsFreeId(field2ItemId,status);
        SdsFreeId(field1ItemId,status);
        SdsFreeId(field0ItemId,status);
    }
    
    /*  And then we can do the double arrays in a second loop - this
     *  only does tPark, but we may as well use the same structure as
     *  for the int arrays.
     */
     
    for (Item = 0; Item < DoubleItems; Item++) {  
        SdsFind(field0Id, DoubleItemNames[Item], &field0ItemId, status);
        SdsPointer(field0ItemId, (void *)&DoubleData0, &actlen, status);
        Elements = actlen / sizeof(DoubleData0[0]);
        SdsFind(field1Id, DoubleItemNames[Item], &field1ItemId, status);
        SdsPointer(field1ItemId, (void *)&DoubleData1, &actlen, status);
        Elem1 = actlen / sizeof(DoubleData1[0]);
        if (Elements > Elem1) Elements = Elem1;
        SdsNew(field2Id, DoubleItemNames[Item], 0, 0, SDS_DOUBLE,1,Dims,
                                                 &field2ItemId, status);
        SdsPointer(field2ItemId, (void *)&DoubleData2, &actlen, status);
        
        if (*status == 0) { 
            for (i = 0; i < Elements; i++) {
                DoubleData2[i] = (DoubleData0[i] + DoubleData1[i]) * 0.5;
            }
        }

        SdsFreeId(field2ItemId,status);
        SdsFreeId(field1ItemId,status);
        SdsFreeId(field0ItemId,status);
    }
    
    /*  Now tidy up. Release any remaining SDS Ids */
        
    SdsFreeId(field2Id,status);
    SdsFreeId(field1Id,status);
    SdsFreeId(field0Id,status);



}


/*
 * Internal Function name:
      AccessConsAAOmega
  
 * Description:
      Accesses the 2dF positioner constants file (AAOmega version)
  
 * History:
      5-Mar-2005 -  KS - Original version, based on AccessCons400().
     17-Nov-2005 -  KS - Now makes field 2 the superset field, rather
                         than field 0.
     25-Feb-2013 - TJF - As part of HERMES support, we first try for the
                          file tdFconstantsDF.sds before trying for
                          tdFconstants400.sds.
                         AccessCons() is now  TdfFpilFull___AccessCons()
     26-Feb-2013 - TJF - Extract creation of plate 2 into new routine
                             TdfFpilFull_CreateCombined() above.
      
 */
static void AccessConsAAOmega (
    void * clientData DUNUSED,
    const char *dirOverride,
    const char *dirFallback,
    FpilRptRoutineType RptRoutine,
    void * RptArg,
    SdsIdType *id,
    StatusType *status)
{

    if (*status != STATUS__OK) return;

    /*  We set ourselves up as for the standard 400 fibre 2dF, but then
     *  modify the result to suit the way AAOmega is used.
     */
     
    TdfFpilFull___AccessCons(TDFCONSFILEDF, 
                             dirOverride, dirFallback, RptRoutine, RptArg,
                                                                  id, status);
    if (*status != STATUS__OK)
    {
        ErsOut(0, status,  "Trying for old format constants file");
        ErsFlush(status);
        TdfFpilFull___AccessCons(TDFCONSFILE400, 
                                 dirOverride, dirFallback, RptRoutine, RptArg,
                                 id, status);
    }


    /*
     * Create plate 2 - the combined plate.  Note, this will delete
     * the current *id value and create a new one.
     */
    TdfFpilFull_CreateCombined(*id, id, status);


}


/*
 * Internal Function name:
      WhichSpecAAOmega
  
 * Description:
      Determine the spectrograph associated with a pivot for AAOmega.

      For AAOmega, there is only one spectrograph, so we only have to
      distinguish between guide fibres and spectrograph fibres.
  
 * History:
      7-Mar-2005 -  KS - Original version, based on WhichSpec400.
     23-Mar-2007 -  KS - Modified use of 'unsigned' to avoid compiler warning.
 */
static void WhichSpecAAOmega (
    void * clientData DUNUSED,
    unsigned field    DUNUSED,
    unsigned pivot,
    SdsIdType Constants DUNUSED,
    unsigned *spec,
    const char ** SpecName,
    StatusType *status)
{
    unsigned int i;

    if (*status != STATUS__OK) return;

/*
 *  There are 400 (NUM_PIVOTS_2DF) fibre positions, numbered 1 to 400, 
 *  just as for 2dF (in 400 fibre mode), on 40 retractors. Some of these
 *  fibre positions are not used for AAOmega and so should be flagged as
 *  disabled. There are eight guide fibres - the code looks these up in
 *  a table, to allow the pivot numbers to be changed later.
 */
    *spec = 1;
    for (i = 0; i < sizeof(AAOmegaGuidePivs)/sizeof(AAOmegaGuidePivs[0]); i++) {
       if (pivot == (unsigned int)AAOmegaGuidePivs[i]) {
          *spec = 0;
          break;
       }
    }
    *SpecName = AAOmegaSpecNames[*spec];
}




/*
 *+		        T d f F p i l F u l l I n i t A A O m e g a

 * Function name:
      TdfFpilFullInitAAOmega

 * Function:
      Initialise Fpil with the 2dF instrument model, for AAOmega.

 * Description:
      Calls FpilInit with appropiate values for 2dF (AAOmega version).  
      Returns an appropiate FpilType variable to be used in calls to Fpil routines.

      This version allows AAO Configure to support AAOmega. Note that the
      differences between 2dF and AAOmega (or, strictly, 2dF in its AAOmega
      mode) are restricted mainly to questions as to whether the instrument
      is treated as having two distinct field plates, or whether one 
      combined filed plate (which should be good for either actual plate)
      is used.

      When you are finished with the "inst" varaible, you should pass it
      to FpilFree(3) to free allocated resources.

      The instrument name string will be "2dF-AAOmega".  The Telescope
      name string will be "AAT".

 * Language:
      C

 * Call:
      (void) TdfFpilFullInitAAOmega(inst)

 * Parameters:   (">" input, "!" modified, "W" workspace, "<" output)
      (<) inst (FpilType *) The FPIL instrument description to be used
                            in calls to Fpil routines.


 * Include files: tdffpil.h

 * See Also: {references}

 * Support: Tony Farrell, AAO

 *-

 * History:
     07-Mar-2005 -  KS - Original version, based on TdfFpilFullInit404(). 
     25-Feb-2013 - TJF - SetupFullStructs() is now  TdfFpilFull___SetupFullStructs()
                        
 */


extern void TdfFpilFullInitAAOmega(FpilType *inst)
{
    FpilCollisionChecksType routines;
    FpilInstParamsType      params;
    FpilModelRoutinesType   modelRoutines;

    TdfFpilFull___SetupFullStructs(&routines, &params, &modelRoutines);

    /*
     *  Modify things that are different between AAOmega and the base 2dF.
     *  AAOmega only has one spectrograph, so WhichSpec has to be handled
     *  differently, (the guides are treated as a spectrograph type, which
     *  is why NumSpec is 2 (it's 3 for the original 2dF)). We treate AAOmega
     *  as having three field plates - the real 1 and 2, with 0 (the default)
     *  being a combination of the two.
     */
     
    routines.WhichSpec     = WhichSpecAAOmega;
    routines.AccessCons    = AccessConsAAOmega;
    strncpy(params.Instrument,"2dF-AAOmega", sizeof(params.Instrument));
                                                      /* Max 20 characters */
    params.Instrument[sizeof(params.Instrument)-1] = '\0';

    params.NumSpec         = 2;
    params.SpecNames       = AAOmegaSpecNames;
    params.NumFields       = 3;

    FpilInit(0, FPIL_VERSION, &routines, &params, &modelRoutines, inst);
}

