/*+		T D F F P I L F U L L _ H E R M E S
 * Module name:
     TDFFPILFULL_HERMES

 * Function:
      HERMES specific versions of routines needed by TdfFpilFullInit(3).

 * Description:
      This contains the routine TdfFpilFullInitAAOmega() used to initialise
      access to the AAOmega FPIL library, full mode.
     
 * Language:
      C

 * Support:
      Tony Farrell, AAO
      Keith Shortridge, AAO

 *-

 *     @(#) $Id: ACMM:2dFutil/tdffpilfull_hermes.c,v 5.2 10-Sep-2013 09:12:12+10 tjf $

 * History:
     25-Feb-2013 - TJF - Original version, extracted from tdffpilfull.c in
                          prep for HERMES work.

 */
/*
 *  Include files
 */
#define __EXTENSIONS__   /* To get PATH_MAX on solaris */


static char *rcsId="@(#) $Id: ACMM:2dFutil/tdffpilfull_hermes.c,v 5.2 10-Sep-2013 09:12:12+10 tjf $";
static void *use_rcsId = (0 ? (void *)(&use_rcsId) : (void *) &rcsId);

#include <limits.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <stdio.h>
#include "drama.h"
#include "sds.h"

#define TDFFPIL_BUILD  /* Used by the include file */
#include "tdffpil.h"

#include "tdfxy.h"
#include "slalib.h"
#include "tdFbuttonDims.h"
#include "Ers.h"
#include "tdfxy_err.h"

#ifndef PATH_MAX
#define PATH_MAX 1024
#endif

#ifdef DEBUG
#undef DEBUG
#endif
#define DEBUG printf

/* 
 * Constant definitions
 */

#define TDFCONSFILEDF "tdFconstantsDF.sds"      /* 2dF constants file - 400 dual-fibre (AAOmega/HERMES) version */


/*
 * Macro function definitions
 */

/*
 * Global variables and functions imported (but not declared in include files)
 */

/*
 * Global variables exported (definitions and initialization of globals
 * declared in the associated include file)
 */

/*
 * Variables private to this module (static variables)
 */

/*
 * Spectrograh names, used by WhichSpec() and TdfFpilFullInit()
 */


static const char * const HERMESSpecNames[] = { "Guide", "HERMES"};

/* Originally, it was not clear what the pivot numbers for the AAOmega plates
 * would be, so the code was written to use a look-up table that could be
 * changed later. The number of pivots is deduced from the size of this
 * array.
 */
static const int HERMESGuidePivs[] = { 50, 100, 150, 200, 250, 300, 350, 400 };

/*
 * Internal Function name:
     HermesToStdLayout
  
 * Description:
     The constants file read for HERMES is the same file we use for AAOmega, since
     they are the same fibre buttons, but we now have two fibres per button.

     The standard xPark/yPark/graspX/graspY/inUse/rank items in the structure 
     refer  to the AAOmega fibres, while the HERMES items are in items named
     xPark_H/yPark_H/graspX_H/graspY_H/inUse_H/rank_H.

     In order to use the HERMES data in the FPIL system, we must removed the
     AAOmega items and rename the HERMES items to be the standard names.


     Note - if sourceId is an external SDS item, it will be copied and
     then deleted in the presumition it was read using SdsRead().  If it 
     is NOT an external item, it will be modified in place.

     In either case, on return, *id points to the resulting structure.


  
 * History:
     26-Feb-2013 - TJF - Original version
 */

static void HermesToStdLayout(
    const SdsIdType sourceId,
    SdsIdType * const id,
    StatusType * const status)
{
    int sourceExtern;


    static char* StdItemNames[] = { "xPark", "yPark", "graspX", "graspY", 
                                    "inUse", "rank" };
    static char* HERMESItemNames[] = { "xPark_H", "yPark_H", 
                                       "graspX_H", "graspY_H", 
                                       "inUse_H", "rank_H" };

    static char* FieldNames[] = { "field0", "field1" };

    static int NumStdItems    = sizeof(StdItemNames)/sizeof(StdItemNames[0]);
    static int NumHERMESItems = sizeof(HERMESItemNames)/sizeof(HERMESItemNames[0]);
    static int NumFields = sizeof(FieldNames)/sizeof(FieldNames[0]);

    int f;


    if (*status != STATUS__OK) return;

    /*
     * Validate the arrays above.
     */
    if (NumStdItems != NumHERMESItems) {
        fprintf(stderr, "\nPROGRAMMING ERROR - Item numbers don't match\n");
        fprintf(stderr, " In %s/HermesToStdLayout()\n", __FILE__);
        fprintf(stderr, " Sorry - not safe to continue for HERMES instrument\n");
        fprintf(stderr, " Please report this error to AAO software support\n\n");
        exit(1);
    }

    /*  If the structure  has been read in directly by SdsRead(), it
     *  will be an external Sds structure and cannot be modified.  (We
     *  could change the values, but not the structure itself).  We want
     *  to add a new field2 that represents the combination of both plates. 
     *  So we first, we copy the structure to an internal structure.
     *
     *  In some cases, sourceId may not be external (HERMES case) in which case
     *  we can modify it in place.
     */
    SdsIsExternal(sourceId, &sourceExtern, status);
    if (sourceExtern) {
        SdsCopy(sourceId,id,status);
        SdsReadFree(sourceId, status);
        SdsFreeId(sourceId, status);
    } else {
        *id = sourceId;
    }

    /*
     * Work through each field.
     */
    for (f = 0; f < NumFields ; ++f)
    {
        SdsIdType fieldId = 0;
        StatusType ignore = STATUS__OK;
        /*
         * Find the field
         */
        SdsFind(*id, FieldNames[f], &fieldId, status);
        if (*status == STATUS__OK)
        {
            int i;
            /*
             * Work through each item.
             */
            for (i = 0; i < NumStdItems ; ++i)
            {
                /*
                 * For each item, we find delete the AAOmega item, we
                 * then find the HERMES item and rename it to the standard
                 * name.
                 */
                SdsIdType stdItemId = 0;
                SdsIdType hermesItemId = 0;
                SdsFind(fieldId, StdItemNames[i], &stdItemId, status);
                SdsDelete(stdItemId, status);
                SdsFreeId(stdItemId, status);

                SdsFind(fieldId, HERMESItemNames[i], &hermesItemId, status);
                SdsRename(hermesItemId, StdItemNames[i], status);
                SdsFreeId(hermesItemId, status);

                if (*status != STATUS__OK)
                {
                    ErsRep(0, status, "Error renaming HERMES item to standard item");
                    break;
                }
            }
            if (*status != STATUS__OK)
                break;
        }
        else
        {
            ErsRep(0, status, "Error finding field %s in constants file",
                   FieldNames[f]);
            break;
        }
        SdsFreeId(fieldId, &ignore);
    }

}
/*^L
 * Internal Function name:
      AccessConsHERMES

 * Description:
      Accesses the 2dF positioner constants file (HERMES version)

 * History:
      4-Apr-2012 -  MVV - Original version, based on AccessConsAAOmega().
     25-Feb-2013 -  TJF - Start on real HERMES constants file version.
     26-Feb-2013 - TJF - Extract creation of plate 2 into new routine
                             TdfFpilFull_CreateCombined().  Add in call
                             to new HermeToStdLayout().  Now works as
                             required.
 */
static void AccessConsHERMES (
    void * clientData DUNUSED,
    const char *dirOverride,
    const char *dirFallback,
    FpilRptRoutineType RptRoutine,
    void * RptArg,
    SdsIdType *id,
    StatusType *status)
{

    if (*status != STATUS__OK) return;

    /*  We set ourselves up as for the standard 400 fibre 2dF, but then
     *  modify the result to suit the way HERMES is used.
     */
     
    TdfFpilFull___AccessCons(TDFCONSFILEDF, 
                             dirOverride, dirFallback, RptRoutine, RptArg,
                                                                  id, status);


    /*
     * Convert from the HERMES layout to the standard layout.
     */

    HermesToStdLayout(*id, id, status);

    /*
     * Create plate 2 - the combined plate.  Note, this will delete
     * the current *id value and create a new one.
     */
    TdfFpilFull_CreateCombined(*id, id, status);

}

/*
 * Internal Function name:
      WhichSpecHERMES
 
 * Description:
      Determine the spectrograph associated with a pivot for HERMES.

      For HERMES, there is only one spectrograph, so we only have to
      distinguish between guide fibres and spectrograph fibres.
 
 * History:
      4-Apr-2012 -  MVV - Original version, based on WhichSpecAAOmega.
 */
static void WhichSpecHERMES (
    void * clientData DUNUSED,
    unsigned field    DUNUSED,
    unsigned pivot,
    SdsIdType Constants DUNUSED,
    unsigned *spec,
    const char ** SpecName,
    StatusType *status)
{
    unsigned int i;

    if (*status != STATUS__OK) return;

/*
 *  There are 400 (NUM_PIVOTS_2DF) fibre positions, numbered 1 to 400,
 *  just as for 2dF (in 400 fibre mode), on 40 retractors. Some of these
 *  fibre positions are not used for AAOmega and so should be flagged as
 *  disabled. There are eight guide fibres - the code looks these up in
 *  a table, to allow the pivot numbers to be changed later.
 */
    *spec = 1;
    for (i = 0; i < sizeof(HERMESGuidePivs)/sizeof(HERMESGuidePivs[0]); i++) {
       if (pivot == (unsigned int)HERMESGuidePivs[i]) {
          *spec = 0;
          break;
       }
    }
    *SpecName = HERMESSpecNames[*spec];
}



/*^L
 *+                     T d f F p i l F u l l I n i t H E R M E S

 * Function name:
      TdfFpilFullInitHERMES

 * Function:
      Initialise Fpil with the 2dF instrument model, for HERMES.

 * Description:
      Calls FpilInit with appropiate values for 2dF (HERMES version).
      Returns an appropiate FpilType variable to be used in calls to Fpil routines.

      This version allows AAO Configure to support HERMES.

      When you are finished with the "inst" varaible, you should pass it
      to FpilFree(3) to free allocated resources.

      The instrument name string will be "2dF-HERMES".  The Telescope
      name string will be "AAT".

 * Language:
      C

 * Call:
      (void) TdfFpilFullInitHERMES(inst)

 * Parameters:   (">" input, "!" modified, "W" workspace, "<" output)
      (<) inst (FpilType *) The FPIL instrument description to be used
                            in calls to Fpil routines.


 * Include files: tdffpil.h

 * See Also: {references}

 * Support: Tony Farrell, AAO

 *-

 * History:
     04-Apr-2012 -  KS - Original version, based on TdfFpilFullInitAAOmega().

 */
extern void TdfFpilFullInitHERMES(FpilType *inst)
{
    FpilCollisionChecksType routines;
    FpilInstParamsType      params;
    FpilModelRoutinesType   modelRoutines;

    TdfFpilFull___SetupFullStructs(&routines, &params, &modelRoutines);

    /*
     *  Modify things that are different between AAOmega and HERMES.
     *  AAOmega only has one spectrograph, so WhichSpec has to be handled
     *  differently, (the guides are treated as a spectrograph type, which
     *  is why NumSpec is 2 (it's 3 for the original 2dF)). We treate AAOmega
     *  as having three field plates - the real 1 and 2, with 0 (the default)
     *  being a combination of the two.
     */

    routines.WhichSpec     = WhichSpecHERMES;
    routines.AccessCons    = AccessConsHERMES;
    strncpy(params.Instrument,"2dF-HERMES", sizeof(params.Instrument));
                                                      /* Max 20 characters */
    params.Instrument[sizeof(params.Instrument)-1] = '\0';

    params.NumSpec         = 2;
    params.SpecNames       = HERMESSpecNames;
    params.NumFields       = 3;

    FpilInit(0, FPIL_VERSION, &routines, &params, &modelRoutines, inst);
}

