/*+		T D F F P I L M I N

 * Module name:
     TDFFPILMIN

 * Function:
      Wrap up of 2dF collision detection using new Fpil interface.

 * Description:
      This module implements TdfFpilMinInit(3), a routine which invokes
      FpilInit(3) with appropiate values for 2dF in FPIL's minimal mode.  
      (collision detection etc.) Since the original
      2dF collision detection routines don't quite meet the required
      function prototypes, we provide wrappers (which can be removed
      when the original 2dF routines are modified, but since that
      requires changes to a lot of code, it won't be done for a while).

      In additon, the routine TdfFpil___SetupMinStructs() is used to 
      setup structures for the minimal FPIL mode, such that 
      TdfFpilFullInit(3) can fetch such structures and complete the setup
      for the full FPIL mode.

 * Language:
      C

 * Support: Tony Farrell, AAO

 *-

 *     @(#) $Id: ACMM:2dFutil/tdffpilmin.c,v 5.2 10-Sep-2013 09:12:12+10 tjf $

 * History:
     07-Dec-1999 - TJF - Original version
     28-Jan-2000 - TJF - Some of the callback routine argument lists have
                         been modified by fpil changes.  Implemented.
     10-May-2006 - TJF - Change use_rcsId to standards compatible form.
     19-Aug-2009 - KS  - Added fibre type argument to InvPos routine for
                         compatability with ESO version. 
 */
/*
 *  Include files
 */


static char *rcsId="@(#) $Id: ACMM:2dFutil/tdffpilmin.c,v 5.2 10-Sep-2013 09:12:12+10 tjf $";
static void *use_rcsId = (0 ? (void *)(&use_rcsId) : (void *) &rcsId);

#include <math.h>
#include <string.h>
#include "tdffpil.h"
#include "drama.h"
#include "tdFcollision.h"
#include "tdFbuttonDims.h"
#include "tdFfieldDims.h"

#include <stdio.h>
#include <stdlib.h>

/* 
 * Constant definitions
 */
#define PI 3.14159265359

/*
 * Macro function definitions
 */

/*
 * type definitions
 */

/*
 * Global variables and functions imported (but not declared in include files)
 */

/*
 * Global variables exported (definitions and initialization of globals
 * declared in the associated include file)
 */

/*
 * Variables private to this module (static variables)
 */


/*
 * Internal Internal Function name:
     ColButFib
  
 * Description:
     Hook for 2dF Button-Fibre collision check routine.
  
 * History:
     07-Dec-1999 - TJF - Original version.
  
 */
static int  ColButFib (
        void * clientData DUNUSED,
        double    butX, double    butY, double    theta,
        double    fvpX, double    fvpY, double    pivX, double    pivY,
        long int  fibreClear,
        double    fibreClearSqrd DUNUSED)
{
    return tdFcollisionButFib( butX, butY, theta, 
                                fvpX, fvpY, pivX, pivY,
                                fibreClear);
}


/*
 * Internal Function name:
     ColButBut
  
 * Description:
     Hook for 2dF Button-Button collision check routine.
  
 * History:
     07-Dec-1999 - TJF - Original version.
  
 */
static int  ColButBut (
        void * clientData DUNUSED,
        double    butXa, double    butYa, double    thetaa,
        double    butXb, double    butYb, double    thetab,
        long int  buttClear,
        double    buttClearSqrd DUNUSED)
{
    return tdFcollisionButBut(butXa, butYa, thetaa,
                              butXb, butYb, thetab,
                              buttClear);
}
/*
 * Internal Function name:
     ColFidBut
  
 * Description:
     Hook for 2dF Fiducial-Button collision check routine.
  
 * History:
     07-Dec-1999 - TJF - Original version.
  
 */
static int  ColFidBut (
        void * clientData DUNUSED,
        long int  butx, long int  buty, double    theta, long int  markx,
        long int  marky, 
        long int  fidTol DUNUSED,
        double    fidTolSqrd)
{
    return tdFcollisionMarkBut(butx, buty, theta, markx, marky, fidTolSqrd);
}
 
/*
 * Internal Function name:
     ColInvPos
  
 * Description:
     Hook for 2dF Invalid position routine (only screw positions
     are invalid).
  
 * History:
     07-Dec-1999 - TJF - Original version.
     19-Apr-2000 - TJF - Add plate argument, but this is not used.
     11-Aug-2005 - TJF - Returns false all the time - there are no longer
                         any invalid positions on 2dF (the screws disappeared
                         with the new field plates on Jan 2005).
     19-Aug-2009 - KS  - Added fibreType argument. 

   
 */
static int  ColInvPos (
        void * clientData DUNUSED,
        unsigned  plate DUNUSED,
        int       fibreType DUNUSED,
        long int  butx DUNUSED,
        long int  buty DUNUSED,
        double    theta DUNUSED,
        long int  butClear DUNUSED,
        double    butClearSqrd DUNUSED)
{
    return 0;
}


/*
 * Internal Function name:
     GetVirPiv
  
 * Description:
     Given a fibre location, return the virtual pivot point for the
     fibre.
  
 * History:
     07-Dec-1999 - TJF - Original version, based on the routine
                         ConfigGetVirtPiv() in tdFconfig library,
                         config.c module (By KS).
     09-Sep-2013 - TJF - The old virtual pivot point is no longer used
                           by 2dF, the fibre location is used instead.
			   Set the enviornment variable  AAOCONFIGURE_OLD_VPIVOT
				to use old technique (temp only).
  
 */
static void GetVirPiv (
        void * clientData DUNUSED,
        double OriginX,
        double OriginY,
        double Theta,
        double *VirtPivX,
        double *VirtPivY)
{
   double PIby2;                 /* For convenience, 90 degrees in radians */

   PIby2 = PI / 2;
   if (Theta < PIby2) {
      *VirtPivX = OriginX - (BUTTON_FVP * sin(Theta));
      *VirtPivY = OriginY + (BUTTON_FVP * cos(Theta));
   } else {
      Theta -= PIby2;
      if (Theta < PIby2) {
         *VirtPivX = OriginX - (BUTTON_FVP * cos(Theta));
         *VirtPivY = OriginY - (BUTTON_FVP * sin(Theta));
      } else {
         Theta -= PIby2;
         if (Theta < PIby2) {
            *VirtPivX = OriginX + (BUTTON_FVP * sin(Theta));
            *VirtPivY = OriginY - (BUTTON_FVP * cos(Theta));
         } else {
            Theta -= PIby2;
            *VirtPivX = OriginX + (BUTTON_FVP * cos(Theta));
            *VirtPivY = OriginY + (BUTTON_FVP * sin(Theta));
         }
      }
   }
   if (!getenv("AAOCONFIGURE_OLD_VPIVOT"))
   {
        *VirtPivX = OriginX;
        *VirtPivY = OriginY;
   }

}

/*
 *+		        T d f F p i l M i n I n i t

 * Function name:
      TdfFpilMinInit

 * Function:
      Initialise Fpil with the 2dF instrument model, minimal FPIL model

 * Description:
      Calls FpilInit with appropiate values for 2dF.  Returns an appropiate
      FpilType variable to be used in calls to Fpil routines.  Fpil
      constants access and coordinate transformations are not supported.

      When you are finished with the "inst" varaible, you should pass it
      to FpilFree(3) to free allocated resources.

      The instrument name string will be "2dF".  The Telescope
      name string will be "AAT".

 * Language:
      C

 * Call:
      (void) TdfFpilMinInit(inst)

 * Parameters:   (">" input, "!" modified, "W" workspace, "<" output)
      (<) inst (FpilType *) The FPIL instrument description to be used
                            in calls to Fpil routines.


 * Include files: tdffpil.h

 * See Also: {references}

 * Support: Tony Farrell, AAO

 *-

 * History:
     07-Dec-1999 - TJF - Original version.
 */
extern void TdfFpilMinInit(FpilType *inst)
{
    FpilCollisionChecksType routines;
    FpilInstParamsType      params;

    TdfFpil___SetupMinStructs(&routines,&params);

    FpilInit(0, FPIL_VERSION, &routines, &params, 0, inst);
}



/*
 *+		   T d f F p i l _ _  _ S e t u p M i n S t r u c t s

 * Function name:
      TdfFpil___SetupMinStructs

 * Function:
      Initialise minimal mode Fpil structures with the 2dF instrument model

 * Description:
      Sets up the structures routines and params ready for calling
      FpilInit() such that they support the minimal FPIL mode.

 * Language:
      C

 * Call:
      (void) = TdfFpil___SetupMinStructs(rouintes, params);

 * Parameters:   (">" input, "!" modified, "W" workspace, "<" output)
      (<) routines (FpilCollisionChecksType *) Will be setup with the
                     collision detection routines for 2dF.  (AccessCons
                     is NOT set).
      (<) params   (FpilInstParamsType *) Will be set up with the 2dF
                     instrument parameters

 * Include files: tdffpil.h

 * See Also: {references}

 * Support: Tony Farrell, AAO

 *-

 * History:
     07-Dec-1999 - TJF - Original version.
     04-Feb-1999 - TJF - Set NumFields and NumPivots entries.
     15-Mar-2000 - TJF - Set new FieldRadIncPark field.
     19-Apr-2000 - TJF - Add ParkMayCollide field.
     14-Dec-2004 - TJF - The number of pivots, fields and fiducials is now
                         based on macros from tdFfieldDims.h and the
                         number of pivots is now 400 (2dF field plate change
                         in Jan 2005).
 */
extern void  TdfFpil___SetupMinStructs(
    FpilCollisionChecksType *routines,
    FpilInstParamsType      *params)
{
    routines->ColButFib  = ColButFib;
    routines->ColButBut  = ColButBut;
    routines->ColFidBut  = ColFidBut;
    routines->ColInvPos  = ColInvPos;
    routines->GetVirPiv  = GetVirPiv;
    routines->AccessCons = 0;

    params->ButtonClear = BUTTON_CLEAR;
    params->FibreClear  = FIBRE_CLEAR;
    params->FidClear    = 1000;
    params->FieldRad    = FIELD_RAD;
    params->FieldRadIncPark = FIELD_RAD_INC_PARK;
    params->Extension   = EXTENSION;
    params->MaxPivAngle = MAX_PIV_ANGLE;
    params->MaxFibAngle = MAX_BUT_ANGLE;
    params->FibAngleVar = 1;
    params->MaxSize     = BUTTON_FVP - (BUTTON_PLE);
    params->NumFields   = NUM_FIELDS_2DF;
    params->NumPivots   = NUM_PIVOTS_2DF;
    params->NumFids     = NUM_FIDS_2DF;
    params->ParkMayCollide = 0;
    strcpy(params->Telescope,"AAT");
    strcpy(params->Instrument,"2dF");

}



