

static char *rcsId="@(#) $Id: ACMM:2dFutil/tdfxyfit.c,v 5.2 10-Sep-2013 09:12:12+10 tjf $";
static void *use_rcsId = (0 ? (void *)(&use_rcsId) : (void *) &rcsId);


#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "status.h"
#include "slalib.h"
#include "slamac.h" 
#include "sds.h"


#include "tdfxy.h"
#include "tdfxyfit.h"


/*+				T d f F u l l F i t

 *  Function name:
      TdfFullFit

 *  Function:
      Fit 2dF model to a set of data

 *  Description:
      Given a set of xy measurements of stars of known positions a model of
      the 2dF distortion and linear transformation parameters is fitted by
      least squares.
      
 *  Language:
      C

 *  Declaration:
          TdfFullFit(int nobs, double *rd, double *xy, double rc, double dc, 
                TdfXyType *xypars, TdfDistType *dist, double mjd, double coeffs[6],
                int fixdist[6], int fixlin[6], double *xyp, long int *status);
      
 *  Parameters:   (">" input, "!" modified, "W" workspace, "<" output)
      (>) nobs      (int)  Number of observations
      (>) rd        (double[nobs,2]) Array of apparent ra,dec (radians)
      (>) xy        (double[nobs,2]) Array of measured x,y (microns)
      (>) rc        (double) apparent RA of field centre.
      (>) dc        (double) apparent Dec of field centre.
      (!) xypars    (TdfXyType *) Transformation parameters set up by
                             a call to TdfXyInit.
      (!) dist      (TdfDistType *) Distortion parameters.
      (>) mjd       (double)   UT date and time as MJD.
      (!) coeffs    (double[6]) Coefficients of linear transformation.
      (>) fixdist   (int[6]) set TRUE for each distortion coefficient to
                             be fitted.
      (>) fixlin    (int[6] set TRUE for each linear coefficient to be fitted.
      (<) xyp       (double[nobs,2] Fitted values of x,y
      (!) status    (StatusType*) Modified status. 


 *  Support: Jeremy Bailey, {AAO}

 *  Version date: 06-Feb-94
 
 *  Modifications:
      20-Feb-1998 - KS - Changed status argument declaration to StatusType
                         to match that in the include file.
 *-
 */



void TdfFullFit(int nobs, double *rd, double *xy, char *ex, double rc, double dc, 
                TdfXyType *xypars, TdfDistType *dist, double mjd, double coeffs[6],
                int fixdist[6], int fixlin[6], double *xyp,
                double pvar[12], double *rms, double *psd, void progress(float), StatusType *status)
                
{
   const int MAXPAR = 20;
   double *aumat;
   double *bmat;
   double *wmat;
   double *vmat;
   double *wk;
   double *xy0;
   double pars[12];
   int pfit[12];
   int nfit;
   int ifit;
   int i,j;
   double pars2[12];
   double xvec[20];
   int jstat;
   double cov[20][20];
   double sumsq,var;
   int npts,jf;
   
   
   /*  Get Work arrays  */
   
   aumat = (double*)malloc(sizeof(double)*2*nobs*MAXPAR);  /* design matrix */
   bmat = (double*)malloc(sizeof(double)*2*nobs);          /* error vector */
   wmat = (double*)malloc(sizeof(double)*MAXPAR);          /* SVD matrices */
   vmat = (double*)malloc(sizeof(double)*MAXPAR*MAXPAR);
   wk = (double*)malloc(sizeof(double)*MAXPAR);
   xy0 = (double*)malloc(sizeof(double)*2*nobs);           /* Current xy vector */

   /*  Get initial parameter values  */
   
   for (i=0;i<6;i++)
     {
        pars[i] = coeffs[i];
        pfit[i] = fixlin[i];
        pfit[i+6] = fixdist[i];
     }
   pars[6] = dist->a*1e-3;
   pars[7] = dist->b*1e-6;
   pars[8] = dist->c*1e-6;
   pars[9] = dist->d*1e-6;
   pars[10] = dist->x0;
   pars[11] = dist->y0;
   
   nfit=0;
   for (i=0;i<12;i++) if (pfit[i] == 0) nfit++;
   
   /*  Initial vector  */
   
   
   TdfXyVec(nobs,rc,dc,xypars,pars,mjd,rd,xy0,status);
   
   /*  loop over parameters - setting up design matrix */
   
   ifit = 0;
   for (i=0;i<12;i++)
     {
        if (pfit[i] == 0)
          {
              pars[i] = pars[i] + 1.0;
              TdfXyVec(nobs,rc,dc,xypars,pars,mjd,rd,xyp,status);
              pars[i] = pars[i] - 1.0;
              jf = 0;
              for (j=0;j<nobs;j++)
                {
                  if (ex[j] == 0)
                    {
                      aumat[MAXPAR*2*jf+ifit] = xyp[2*j] - xy0[2*j];
                      aumat[MAXPAR*(2*jf+1)+ifit] = xyp[2*j+1] - xy0[2*j+1];
                      jf++;
                    }  
                }  
              ifit++;
              progress(((float)ifit)/((float)(nfit+4))*100.0);
           }
     }
     
   /*  set up error vector  */
     
   jf = 0;  
   for (j=0;j<nobs;j++) 
     {
       if (ex[j] == 0)
         {
           bmat[2*jf] = xy[2*j] - xy0[2*j];
           bmat[2*jf+1] = xy[2*j+1] - xy0[2*j+1];
           jf++;
           npts = jf;
         }  
     }
   
   /*  Do singular value decomposition of design matrix  */
   
   slaSvd(2*npts,nfit,2*nobs,MAXPAR,aumat,wmat,vmat,wk,&jstat);  
   progress(((float)nfit+2)/((float)(nfit+4))*100.0);
   
   slaSvdsol(2*npts,nfit,2*nobs,MAXPAR,bmat,aumat,wmat,vmat,wk,xvec);
   
   slaSvdcov(nfit,MAXPAR,MAXPAR,wmat,vmat,wk,(double*)cov);
   progress(((float)nfit+3)/((float)(nfit+4))*100.0);
   
   j=0;
   for (i=0;i<12;i++)
     {
        if (pfit[i] == 0)
          {
             pars2[i] = pars[i]+xvec[j];
             pvar[i] = cov[j][j];
             j++;
          }
        else
          {
             pars2[i] = pars[i];
             pvar[i] = 0.0;
          }
     }    
   TdfXyVec(nobs,rc,dc,xypars,pars2,mjd,rd,xyp,status);
   progress(100.0);
   for (i=0;i<6;i++)
     {
        coeffs[i] = pars2[i];
     }
   dist->a = pars2[6]*1e3;
   dist->b = pars2[7]*1e6;
   dist->c = pars2[8]*1e6;
   dist->d = pars2[9]*1e6;
   dist->x0 = pars2[10];
   dist->y0 = pars2[11];
   xypars->dist = *dist;
   
   sumsq = 0.0;
   for (i=0;i<nobs;i++)
     {
        if (ex[i] == 0)
          {
            sumsq = sumsq + (xyp[2*i]-xy[2*i])*(xyp[2*i]-xy[2*i]);
            sumsq = sumsq + (xyp[2*i+1]-xy[2*i+1])*(xyp[2*i+1]-xy[2*i+1]);
          }
     }
   var = sumsq/(double)npts;
   *rms = sqrt(var);
   *psd = sqrt(var*(double)npts/((double)npts-(double)nfit));

   for (i=0;i<12;i++)
     {
        pvar[i] = sqrt(var*pvar[i]);   
     }
   pvar[6] = pvar[6]*1e3;
   pvar[7] = pvar[7]*1e6;
   pvar[8] = pvar[8]*1e6;
   pvar[9] = pvar[9]*1e6;
   /*  free workspace  */
   
   free(aumat);
   free(bmat);
   free(wmat);
   free(vmat);
   free(wk);
   free(xy0);  
}


void TdfXyVec(int nobs,double rc,double dc,TdfXyType *xypars,double pars[12],
               double mjd, double *rd, double *xy, StatusType *status) 
               
{

   TdfDistType dist;
   TdfXyType xyp2;
   double x,y;
   int i;
   double coeffs[6];
   double r,d;
   
   xyp2 = *xypars;
   for (i=0;i<6;i++)
     {
        coeffs[i] = pars[i];
     }
   dist.a = pars[6]*1e3;
   dist.b = pars[7]*1e6;
   dist.c = pars[8]*1e6;
   dist.d = pars[9]*1e6;
   dist.x0 = pars[10];
   dist.y0 = pars[11];
   xyp2.dist = dist;
   
   for (i=0;i<nobs;i++)
     {
       r = rd[2*i];
       d = rd[2*i+1];
       TdfRd2xy(&xyp2,rc,dc,r,d,mjd,&x,&y,status);
       xy[2*i] = coeffs[0] + coeffs[1]*x + coeffs[2]*y;
       xy[2*i+1] = coeffs[3] + coeffs[4]*x + coeffs[5]*y;
     }
}
