/*
 * Simple test program which runs the 2dF RA/Dec to X/Y conversion code
 * over a series of concentric circles and outputs the results.  
 */
static const char *rcsId="@(#) $Id: ACMM:2dFutil/tdfxytest.c,v 5.2 10-Sep-2013 09:12:12+10 tjf $";
static void *use_rcsId = (0 ? (void *)(&use_rcsId) : (void *) &rcsId);
#include <stdio.h>

#include "drama.h"
#include "status.h"
#include "sds.h"
#include "tdfxy.h"
#include "mess.h"
#include "sds_err_msgt.h"
#include "tdfxy_err_msgt.h"
#include "slamac.h"

#define MA 30.3
#define ME 33.5
#define NP 0.24
#define CH -43.4
#define HF -6.53
#define ZGT 0.0
#define TEMP 290
#define PRES 900
#define HUMID 0.7
#define WAVE  0.6
#define ADC_A 0
#define ADC_B 0
#define DUT 0

/*
 * These are from a field used by 2dFfpImg - the raw0_07aug.sds file.
 */

#define MJD 54319.39728 
#define CENRA   4.188790205
#define CENDEC  -0.5235987756


static void DoPlate(unsigned plate, StatusType *status);
static void DoPoint(TdfXyType *xypars, TdfLinType *lin, 
                    double ra, double dec, int angle , double radius, 
                    StatusType *status);
int main()
{
  
    StatusType status = STATUS__OK;

    MessPutFacility(&MessFac_SDS);
    MessPutFacility(&MessFac_TDFXY);

    DoPlate(0,&status);
    DoPlate(1,&status);
    if (status != STATUS__OK)
    {
        char buffer[200];
        MessGetMsg(status, 0, sizeof(buffer), buffer);
        fprintf(stderr,"tdfxy test failed - %s\n", buffer);
        return 1;
    }
    return 0;
}

/*
 * The radius to which we will generate positions - 1 degree (2dF is
 * about 2 degrees in total
 */
#define RADIUS_MAX (DPI/180.0)
/*
 * Our step size - roughly the same as in distortion map.
 */
#define STEP_SIZE (RADIUS_MAX/23.0)

static void DoPlate(unsigned plate, StatusType *status)
{
    TdfDistType dist;
    TdfLinType  lin;
    TdfXyType   xypars;
    double radius;

    if (*status != STATUS__OK) return;
    TdfGetLin((plate==0 ? "./tdFlinear0.sds": "./tdFlinear1.sds"),
              &lin, status);
    TdfGetDist((plate==0 ? "tdFdistortion0.sds": "tdFdistortion1.sds"),
               &dist, status);
    if (*status != STATUS__OK) 
    {
        fprintf(stderr,"Failed to read files for plate %d\n", plate);
        return;
    }

    TdfXyInit(MJD, DUT, TEMP, PRES, HUMID, WAVE, WAVE,
              MA, ME, NP, CH, HF, ZGT, ADC_A, ADC_B, dist,
              &xypars, status);
    if (*status != STATUS__OK) 
    {
        fprintf(stderr,"Failed to initialise model for plate %d\n", plate);
        return;
    }
    printf("Plate %d\n", plate);
    printf("Radius  Angle  RA     Dec      X         Y\n");
   
    for (radius = 0; radius <= RADIUS_MAX ; radius+=STEP_SIZE)
    {
        double ra, dec;
        double cos_radius = radius*0.707106781186548;

        /*
         * create points around the circle at the specified radius.
         *
         * 0 degree point.
         */
        ra =  CENRA + radius;
        dec = CENDEC;
        DoPoint(&xypars, &lin, ra, dec, 0, radius, status);
        if (radius == 0) continue;

        /* 45 degrees */
        ra  = CENRA  + cos_radius;
        dec = CENDEC + cos_radius;
        DoPoint(&xypars, &lin, ra, dec, 45, radius, status);

        /* 90 degrees */
        ra  = CENRA;
        dec = CENDEC + cos_radius;
        DoPoint(&xypars, &lin, ra, dec, 90, radius, status);

        /* 135 degrees */
        ra  = CENRA  - cos_radius;
        dec = CENDEC + cos_radius;
        DoPoint(&xypars, &lin, ra, dec, 135, radius, status);

        /* 180 degrees */
        ra  = CENRA  - cos_radius;
        dec = CENDEC;
        DoPoint(&xypars, &lin, ra, dec, 180, radius, status);

        /* 225 degrees */
        ra  = CENRA  - cos_radius;
        dec = CENDEC - cos_radius;
        DoPoint(&xypars, &lin, ra, dec, 225, radius, status);

        /* 270 degrees */
        ra  = CENRA;
        dec = CENDEC - cos_radius;
        DoPoint(&xypars, &lin, ra, dec, 270, radius, status);

        /* 315 degrees */
        ra  = CENRA  + cos_radius;
        dec = CENDEC - cos_radius;
        DoPoint(&xypars, &lin, ra, dec, 315, radius, status);


    }
    printf("-------------------------------------\n");
}


static void DoPoint(
    TdfXyType *xypars, 
    TdfLinType *lin, 
    double ra, 
    double dec, 
    int angle , 
    double radius, 
    StatusType *status)
{
    double x, y;
    double xp, yp;
    if (*status != STATUS__OK) return;

    TdfRd2xy(xypars, CENRA, CENDEC, ra, dec,
            MJD, &x, &y, status);
    TdfXy2pos(lin, x, y, &xp, &yp);

    printf("%5.3f    %3d  %6.2f %6.2f %9.1f %9.1f\n",
           radius*DR2D,  /* Coverted from radians to degrees */
           angle,        /* Already in degrees */
           ra*DR2D,      /* Coverted from radians to degrees */
           dec*DR2D,     /* Coverted from radians to degrees */
           xp, yp);
           
}
