/*
 * Simple test program to help debug tdFcollisionButFib()
 *
 * Arguments:
 *    Button X (microns)
 *    Button Y (microns)
 *    Button T (radians)
 *    Fibre X  (microns)
 *    Fibre Y  (microns)
 *    Pivot X  (microns)
 *    Pivot Y  (microns).
 *
 * The output is the distance at which a collision occurs.
 *   Range 10 microns to 10000 microns.
 *
 * Button X/Y is validate to be on the 2dF field plate.
 * Fibre and Pivot positions must be on the 2dF field plate (including park 
 *   positions, since the pivot position is just inside the park position).
 *
 * "@(#) $Id: ACMM:2dFutil/testbutfib.c,v 5.2 10-Sep-2013 09:12:11+10 tjf $"
 */

static const char *rcsId="@(#) $Id: ACMM:2dFutil/testbutfib.c,v 5.2 10-Sep-2013 09:12:11+10 tjf $";
static void *use_rcsId = (0 ? (void *)(&use_rcsId) : (void *) &rcsId);

#include "tdFcollision.h"
#include "tdFfieldDims.h"

#include <stdlib.h>
#include <stdio.h>
#include <math.h>

static void usage(const char *argv0)
{
    fprintf(stderr,"Usage:%s ButX ButY ButT FibX FibY PivX PivY\n", argv0);
    exit(1);
}
static void stuff(double    butX, double    butY, double    theta,
                  double    fvpX, double    fvpY, double    pivX, double    pivY);



#define CLEAR_MIN 1
#define CLEAR_MAX 10000

int main(int argc, char *argv[])
{
    int butX=0;
    int butY=0;
    double butT=0;
    int fibX=0;
    int fibY=0;
    int pivX=0;
    int pivY=0;
    int clearance=CLEAR_MIN;
    if (argc != 8)
    {
        fprintf(stderr, "Wrong number of arguments\n");
        usage(argv[0]);
        return 1;
    }

    butX = atoi(argv[1]);
    butY = atoi(argv[2]);
    butT = atof(argv[3]);
    fibX = atoi(argv[4]);
    fibY = atoi(argv[5]);
    pivX = atoi(argv[6]);
    pivY = atoi(argv[7]);

    /*
     * Validate arguments (But ignoring theta for the moment).
     */
    if ((butX < -FIELD_RAD)||(butX > FIELD_RAD))
    {
        fprintf(stderr,"Button X (%d) is outside field plate\n", butX);
        return 1;
    }
    if ((butY < -FIELD_RAD)||(butY > FIELD_RAD))
    {
        fprintf(stderr,"Button Y (%d) is outside field plate\n", butY);
        return 1;
    }
    if ((fibX < -FIELD_RAD_INC_PARK)||(fibX > FIELD_RAD_INC_PARK))
    {
        fprintf(stderr,"Fibre X (%d) is outside field plate\n", fibX);
        return 1;
    }
    if ((fibY < -FIELD_RAD_INC_PARK)||(fibY > FIELD_RAD_INC_PARK))
    {
        fprintf(stderr,"Fibre Y (%d) is outside field plate\n", fibY);
        return 1;
    }
    if ((pivX < -FIELD_RAD_INC_PARK)||(pivX > FIELD_RAD_INC_PARK))
    {
        fprintf(stderr,"Pivot X (%d) is outside field plate\n", pivX);
        return 1;
    }
    if ((pivY < -FIELD_RAD_INC_PARK)||(pivY > FIELD_RAD_INC_PARK))
    {
        fprintf(stderr,"Pivot Y (%d) is outside field plate\n", pivY);
        return 1;
    }

    printf("Checking button at %d,%d,%g against fibre from %d,%d to %d,%d\n",
           butX,butY,butT,fibX,fibY,pivX,pivY);

    /*
     * stuff() outputs some potentially interesting details.
     */
    stuff(butX,butY,butT, fibX, fibY, pivX, pivY);

    /*
     * Work through 
     */
    while (clearance <= CLEAR_MAX)
    {
        int result = tdFcollisionButFib(butX,butY,butT, fibX, fibY, pivX, pivY,
                                        clearance);

        if (result)
        {
            printf("  Clearance is %d microns\n", clearance-1);
            return 0;
        }

        ++clearance;
    }
    printf("  Clearance is at least %d microns\n", clearance);
    return 0;

}
#define SQRD(x)           ((double)(x)*(double)(x))

/*
 * This function just outputs some interesting stuff about the
 * combination, basically code used in 6dF.
 */
static void stuff(
    double    butX,                  /* Fibre-end - x ordinate            */
    double    butY,                  /*           - y ordinate            */
    double    theta,                 /*           - theta                 */
    double    fvpX,                  /* Fibre - x at button end           */
    double    fvpY,                  /*       - y at button end           */
    double    pivX,                  /*       - x at pivot                */
    double    pivY)                  /*       - y at pivot                */
{
    /*
     *  a is the length of the fibre.
     *  b is the distance from the button to one end of the line.
     *  c is the distance from the button to the other end of the line.
     */

    double a = sqrt(SQRD(fvpX - pivX) + SQRD(fvpY - pivY));

    double b = sqrt(SQRD((fvpX - butX)) + SQRD((fvpY - butY))) ;
    
    double c = sqrt(SQRD((pivX - butX)) + SQRD((pivY - butY))) ;

    printf(" Fibre length = %.0f\n", a);
    printf(" Distance from button to fibre end = %.0f\n", b);
    printf(" Distance from button to pivot     = %.0f\n", c);

    if ((a <= b)&&(a <= c))
    {
        /*
         * WARNING - resuming button is a point
         */
        printf(" Collision not possible due to distance from button to ends being less then length (Presuming button is a point)\n");
        return;
    }
 
    if (fvpX == pivX)
    {
        /*
         * Fibre parallel to the Y axis
         */
        printf(" Fibre is parrallel to Y axis, distance to button (as a point) is %d microns\n",
                abs(fvpX - butX));
        return;
    }
/*
 *      We can now use code which determines the distance between a line
 *      and a point since
 *        1.  The first check ensures a perpendicular from line to the
 *            point is within the line-segment of concern
 *        2.  fvpX - pivX 
 *
 *     Given m is the gradient of the line between the fibre ends and
 *     b is the intercept, then, d the distance between the point x1,y1 and
 *     the line is given by
 *
 *      d = |  m.x1 - y + b    |   (Note, absolute value only)
 *          |------------------|
 *          | +/- sqrt(m^2 + 1)|
 *
 *
 *      Determine m and b.
 */
    {
        double m = (double)(pivY - fvpY)/(double)(pivX - fvpX);
        double b = pivY - m*pivX;
        
/*
 *      Determine the denominator and then the numerator.  Note, sqrt()
 *      returns the positive square root only.
 */
        double dd = sqrt(m*m + 1);
        double dn = m * butX - butY + b;
        
        double d;
/*
 *      Determine the distance.  (Note, dd is always greater then 1, therefore
 *      no divide by zero is possible)
 */
        if (dn < 0)
            d = dn * -1 / dd;
        else
            d = dn / dd;         

        printf(" Distance from button (as a point) to fibre is %.0f\n", d);
    }
}
