/*+                S I X D F C O L L I S I O N

 *  Module name:
      sixdFcollision

 *  Function:
      Contains collision detection support for 6dF. 

 *  Description:
      This module provides routines used to support collision detection
      within 6dF.  The FPIL library provides the interface, this module
      provides  SixdfFpilMinInit() which is used to initialise a variable
      of type FpilType.  It also provides SixdfFpil___SetupMinStructs(),
      used by SixdfFpilFullInit() to access the collision detection details.

      The entry points are  sixdfFpil___SetupMinStruct() and
      sixdfFpilMinInit().  The former is only meant to be called
      internally within this module or from sixdfFpilMaxInit().

      Note on angles.
          The positioner software uses theta = 0 along the positive y
          axis, instead of the positive x.   So to convert to mathematical
          orientation, we must rotate by  negative PI/2.   Our internal
          angles are all mathematical to ensure the matematics we do
          is correct, so the conversion is done just before the various 
          types representing line seguments etc. are set up.


 *  Language:
      C

 *  Support: Tony Farrell, AAO

 *-

 *     @(#) $Id: ACMM:6dFutil/sixdFcollision.c,v 4.3 06-Mar-2013 08:32:35+11 tjf $


 *  History:
     18-Jan-2000 - TJF - Original version.
     28-Jan-2000 - TJF - Support changes to FPIL function prototypes.
     24-Feb-2000 - TJF - Remove RCS header items and add Sccs Id.
     06-Apr-2000 - TJF - Debug and get working.  Some revamping.
     19-Aug-2009 - KS  - ColInvPos() now has a fibre type argument.
      {@change entry@}


 */

/*
 *  Include files
 */


static char *rcsId="@(#) $Id: ACMM:6dFutil/sixdFcollision.c,v 4.3 06-Mar-2013 08:32:35+11 tjf $";
static void *use_rcsId = (0 ? (void *)(&use_rcsId) : (void *) &rcsId);


#include <math.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "drama.h"
#include "sixdffpil.h"
#include "sixdFbuttonDims.h"
#include "sixdFfieldDims.h"



#define PI 3.14159265358979323846

/*
 *  This allows functions to be inlines under GNU C.
 */

#if defined(__GNUC__) || defined(GNUC)
#   define INLINE __inline__ static
#else
#   define INLINE static
#endif
/*
 *  Definitions
 */
#define SQRD(x)           ((double)(x)*(double)(x))

/*
 *
 *  We have here a bunch of types and associated routines for describing
 *  points, lines etc.
 *
 */


/*
 *  Type describing a point 
 */
typedef struct {
    double x;     /* X ordinate */
    double y;     /* Y ordinate */
} pointType;

/*
 *  Initialise a point from x and y.
 */
INLINE pointType * NewPoint(double x, double y, pointType *point)
{
    point->x = x;
    point->y = y;
    return point;
}

/*
 *  Type describing a line segment using start and end points but for which 
 *  the angle of the line and length of the line are known.  
 */
typedef struct {
    pointType start;  /* Start point */
    pointType end;    /* End point   */
    double angle;     /* Angle line lays along */
    double length;    /* Length of line from start point */
} lineSegType;

/*
 *  Create a lineSegType from a start point, angle and length.
 *
 */
INLINE lineSegType * NewLineSegAngle(pointType start, 
                                      double angle, 
                                      double length,
                                      lineSegType *line)
{
    double cosine;
    double sine;


/*
 *  Copy in basic values.
 */
    line->start  = start;
    line->angle  = angle;
    line->length = length;


    cosine = cos(angle);
    sine   = sin(angle);
/*
 *  Determine other end of the line.
 */
    line->end.x  = start.x + length * cosine;

    line->end.y  = start.y + length * sine;

    return line;

    
}
/*
 *  Create a lineSegType from a start and end points.
 */
INLINE lineSegType * NewLineSegEnds(pointType start, 
                                     pointType end,
                                     lineSegType *line)
{
    double x_len = start.x - end.x;
    double y_len = start.y - end.y;

    line->start  = start;
    line->end    = end;
   
    line->start  = start;

    line->length = sqrt(x_len*x_len + y_len*y_len);

    line->angle  = atan2(y_len, x_len);
    return line;
  
}



/*
 *  Type describing a circle.
 */
typedef struct {
    pointType center;
    double    radius;
} circleType;
/*
 *  Initialise a circle from x, y and radius
 */
INLINE circleType * NewCircleXYR(double x, double y, double radius, 
                             circleType *circle)
{
    circle->center.x = x;
    circle->center.y = y;
    circle->radius   = radius;
    return circle;
}
/*
 *  Initialise a circle from center point and radius
 */
INLINE circleType * NewCircleP(pointType center, double radius, 
                             circleType *circle)
{
    circle->center = center;
    circle->radius = radius;
    return circle;
}

/*
 * Internal Internal Function name:
      MathsTheta
  
 * Description:
      Converts from a positioner system theta value (0 = y axis) to
      a mathematical theta value (0 = x axis).

      
  
 * History:
     18-Jan-2000 - TJF - Original version.
  
 */
INLINE double MathsTheta(double theta)
{
/*
 *   Rotate and then normalize on the way out.
 */
    theta += (PI/2);
    return fmod(theta, (2*PI));
}

/*
 * Internal Internal Function name:
      CirclesOverlap
  
 * Description:
      Determines if two circles overlap.
  
 * History:
     18-Jan-2000 - TJF - Original version.
  
 */
INLINE int  CirclesOverlap (
    const circleType *c1,  /* Circle one */
    const circleType *c2)  /* Circle two */
    
{
/*
 *  This is quite easy.  Just determine the distance between the
 *  two points which given the origin's of the circles.  If that
 *  distance is less then the combined radii, they overlap.
 */
    if ((double)(SQRD(c1->center.x - c2->center.x) +
                 SQRD(c1->center.y - c2->center.y)) 
        <= SQRD(c1->radius + c2->radius))
        return 1;
    else
        return 0;
                                           
}


/*
 * Internal Internal Function name:
      CircleLineOverlap
  
 * Description:
      Determines if a circle and line segement overlap

      To handle non-zero width lines, just expand the circle
      radius by the width of the line

 * History:
     18-Jan-2000 - TJF - Original version.
  
 */
INLINE int  CircleLineOverlap (
    const circleType  *circle,  /* Circle details */
    const lineSegType *line)    /* Line details   */
    
{
    double radSqrd = SQRD(circle->radius);
/*
 *  a is the length of the line squared
 *  b is the distance from the circle to one end of the line squared
 *  c is the distance from the circle to the other end of the line squared.
 */

    double a = SQRD(line->length);

    double b = SQRD((line->start.x - circle->center.x)) +
               SQRD((line->start.y - circle->center.y)) ;

    double c = SQRD((line->end.x   - circle->center.x)) +
               SQRD((line->end.y   - circle->center.y));
/*
 *  If the distance from either end of the line is longer then the line
 *  length, then there can be no collision.
 */
    if (((a+radSqrd) <= b) || ((a+radSqrd) <= c))
        return 0;
    
    else if (line->start.x == line->end.x)
    {
/*
 *      Line parallel to the Y axis  (the next bit of code would get
 *      a divide by zero error) and y of point is within the length
 *      of the fibre (ensured by the above check) 
 */
        long int d = circle->center.x - line->end.x;
        return ((d <= circle->radius)? 1: 0);        
    }
    else
    {        
/*
 *      We can now use code which determines the distance between a line
 *      and a point since
 *        1.  The first check ensures a perpendicular from line to the
 *            point is within the line-segement of concern
 *        2.  line->end.x - line->start.x != 0
 *
 *     Given m is the gradient of the line between the fibre ends and
 *     b is the intercept, then, d the distance between the point x1,y1 and
 *     the line is given by
 *
 *      d = |  m.x1 - y + b    |   (Note, absolute value only)
 *          |------------------|
 *          | +/- sqrt(m^2 + 1)|
 *
 *
 *      Determine m and b.
 */
        double m = (double)(line->end.y - line->start.y)/
                   (double)(line->end.x - line->start.x);
        double b = line->end.y - m*line->end.x;
        
/*
 *      Determine the denominator and then the numerator.  Note, sqrt()
 *      returns the positive square root only.
 */
        double dd = sqrt(m*m + 1);
        double dn = m * circle->center.x - circle->center.y + b;
        
        double d;
/*
 *      Determine the distance.  (Note, dd is always greater then 1, therefore
 *      no divide by zero is possible)
 */
        if (dn < 0)
            d = dn * -1 / dd;
        else
            d = dn / dd;         
        
        return ((d <= circle->radius) ? 1: 0);
    }
    
}
/*
 * Internal Internal Function name:
      RotatePoint
  
 * Description:
      Rotate a given point around the origin using the sine and cosine of
      the rotation angle.

 * History:
     18-Jan-2000 - TJF - Original version.
  
 */

INLINE void RotatePoint (
    const pointType *point,
    double    sine,
    double    cosine,
    pointType *result )
{
    result->x =      point->x * cosine + point->y * sine;
    result->y = -1 * point->x * sine   + point->y * cosine;
}
  

typedef struct {
    pointType botLeft;   /* Bottom left */
    pointType topLeft;   /* Top left */
    pointType topRight;  /* Top right    */
    pointType botRight;  /* Bottom Right */
} rectangleType;

/*
 * Make a rectangle desribing a think line parallel to the  X axis.  Don't
 * apply it's angle, but save it for later use.
 */
INLINE void MakeRectangle(const lineSegType *line, 
                          double width,
                          rectangleType *rec)
{
    double yoff,xoff;
    double sine,cosine;
    double halfWidth = width/2;
    rectangleType PreRotation;


/*
 *  
 *  If the angle is so small as to be insignificant, then assume an
 *  angle of zero and just construct it.
 *      return immediately
 */        
    if ((line->angle <= 0.0001)&&(line->angle > -0.0001))
    {
        rec->botLeft.x   = line->start.x;
        rec->botLeft.y   = line->start.y - halfWidth;
        
        rec->topRight.x  = line->start.x + line->length;
        rec->topRight.y  = line->start.y + halfWidth;
        
        
        rec->topLeft.x   = rec->botLeft.x;
        rec->topLeft.y   = rec->topRight.y;
        
        rec->botRight.x  = rec->topRight.x;
        rec->botRight.y  = rec->botLeft.y;
        return;
    }
/*
 *  We need to rotate the rectangle.  Contruct a rectangle of the
 *  right size, but with the bottom left cornet at the origin and
 *  the bottom layed along the X axis.
 */
    PreRotation.botLeft.x = 0;
    PreRotation.botLeft.y = 0;

    PreRotation.topLeft.x = 0;
    PreRotation.topLeft.y = width;

    PreRotation.topRight.x = line->length;
    PreRotation.topRight.y = width;
    
    PreRotation.botRight.x = line->length;
    PreRotation.botRight.y = 0;
    
/*
 *  Determine sine and cosine.  (We use the negative angle, since
 *  we are effectivly removing a previous rotation.  
 */
    sine   = sin(-1*line->angle);
    cosine = cos(-1*line->angle);

/*
 *  So now point botLeft is at the origin.  Rotate the
 *  other three points by the rectangles angle (the one which put it
 *  on the X axis)
 */
    rec->botLeft = PreRotation.botLeft;
    RotatePoint(&PreRotation.topLeft,  sine, cosine, &rec->topLeft);
    RotatePoint(&PreRotation.topRight, sine, cosine, &rec->topRight);
    RotatePoint(&PreRotation.botRight, sine, cosine, &rec->botRight);
    
/*
 *  So now the rectangle is rotated correctly, we need to offset 
 *  it correctly.  The origin is given by the line start point, except
 *  that y allows for the width.
 */
    xoff =  line->start.x;
    yoff =  line->start.y - width/2;
/*
 *  Offset orgin back to where it should be
 */
    rec->botLeft.y   = yoff;
    rec->topLeft.y  += yoff;
    rec->botRight.y += yoff;
    rec->topRight.y += yoff;   
    

    rec->botLeft.x   = xoff;
    rec->topLeft.x  += xoff;
    rec->botRight.x += xoff;
    rec->topRight.x += xoff;
       
/*
 *  We are done.
 */

}

/*
 *  Rotate a rectangle by a specified angle, with the specified point
 *  to become the origin.
 */
INLINE void RotateRectangle(const rectangleType * const in,
/*                            double        originX,
                              double        originY,*/
                            double        sine,
                            double        cosine,
                            rectangleType *out)
{
/*
    pointType botRight;
    pointType topLeft;
    pointType topRight;
    pointType botRight;
*/
/*
 *  
 */    
/*
    botRight.x = in->botRight.x - originX;
    botRight.y = in->botRight.y - originY;

    topRight.x = in->topRight.x - originX;
    topRight.y = in->topRight.y - originY;

    topLeft.x  = in->topLeft.x  - originX;
    topLeft.y  = in->topLeft.y  - originY;

    botLeft.x  = in->botLeft.x  - originX;
    botLeft.y  = in->botLeft.y  - originY;
*/

/*
 *  Rotate all points.
 */
    RotatePoint(&in->botLeft,  sine, cosine, &out->botLeft);
    RotatePoint(&in->topLeft,  sine, cosine, &out->topLeft);
    RotatePoint(&in->topRight, sine, cosine, &out->topRight);
    RotatePoint(&in->botRight, sine, cosine, &out->botRight);
}

/*
 *  Determine the minimum value of four values.
 */
INLINE double FourPointMin(double a, double b, double c, double d)
{
    double min = a;
    if (b < min)
        min = b;
    if (c < min)
        min = c;
    if (d < min)
        min = d;
    return min;
}
/*
 *  Determine the maximum value of four values.
 */
INLINE double FourPointMax(double a, double b, double c, double d)
{
    double max = a;
    if (b > max)
        max = b;
    if (c > max)
        max = c;
    if (d > max)
        max = d;
    return max;
}

/*
 *  A type that represents the extremes of a shape - that is, the left most,
 *  right most, top most and bottom values.
 */
typedef struct {
    double left;
    double right;
    double top;
    double bottom;
} extremesType;

/*
 *  Initialise an Extremes types from a rectangle.
 */
INLINE void Extremes(const rectangleType *r, extremesType *e)
{

    e->left   = FourPointMin(r->topLeft.x,  r->botLeft.x, 
                           r->topRight.x, r->botRight.x);
    e->right  = FourPointMax(r->topLeft.x,  r->botLeft.x, 
                           r->topRight.x, r->botRight.x);
    
    e->bottom = FourPointMin(r->topLeft.y,  r->botLeft.y, 
                           r->topRight.y, r->botRight.y);
    e->top    = FourPointMax(r->topLeft.y,  r->botLeft.y, 
                           r->topRight.y, r->botRight.y);

}


/*
 *  Given variables supplying the extreme values of two objects, determine
 *  if the objects might have overlap.
 */
INLINE int IsOverlap(
    const extremesType *a, 
    const extremesType *b)
{
/*
 *  If b is totally to the left of a, no overlap is possible
 */
    if (b->right < a->left)
        return 0;
/*
 *  If a is totally to the left of b, no overlap is possible
 */
    else if (a->right < b->left)
        return 0;
/*
 *  If b is totally above a, no overlap is possible
 */
    else if (b->bottom > a->top)
        return 0;
/*
 *  If a is total above b, no overlap is possible
 */
    else if (a->bottom > b->top)
        return 0;
/*
 *  There is some overlap.
 */
    return 1;

}
/*
 * Internal Internal Function name:
      RectanglesCollide
  
 * Description:
      Determines if two rectangles collide

      Straede's conjecture - Two rectangles overlap if

          1.  Rotate Rectangles such that  A is aligned with X axis.  
              There must be overlap in either X or Y directions
          2.  Rotate Rectangles such that B is aligned with the X axis.
              There must be overlap in either X or Y directions

     Note that our rectangles are not normalized, meaning the "angle"
     member gives the angle the rectangle is at relative to the
     X axis and the points of the rectangle have not been adjuects for this.

 * History:
     19-Jan-2000 - TJF - Original version.
  
 */
INLINE int RectanglesCollide (
    const rectangleType *ra,
    const double raAngle,
    const rectangleType *rb,
    const double rbAngle)
{
    rectangleType raInb;        
    rectangleType raInOwn;
    rectangleType rbIna;
    rectangleType rbInOwn;
    extremesType  raExtremes;
    extremesType  rbExtremes;
    extremesType  targetExtremes;
    double        sine;
    double        cosine;


/*
 *  Rotate both rectangles to rectangle a's coorindate system.
 */
    sine   = sin(raAngle);
    cosine = cos(raAngle);
    RotateRectangle(ra, sine, cosine, &raInOwn);
    RotateRectangle(rb, sine, cosine, &rbIna);

/*
 *  Get extremes of rectangle a (in own system) and the rotated rectangle b)
 */
    Extremes(&raInOwn, &raExtremes);
    Extremes(&rbIna,   &targetExtremes);

    if (IsOverlap(&raExtremes, &targetExtremes))
    {
/*
 *      Rotate both rectangles to rectangle a's coorindate system.
 */
        sine   = sin(rbAngle);
        cosine = cos(rbAngle);

        RotateRectangle(rb, sine, cosine, &rbInOwn);
        RotateRectangle(ra, sine, cosine, &raInb);
/*
 *      Get extremes of rectangle b (normalized) and the rotated rectangle a)
 */
        Extremes(&rbInOwn, &rbExtremes);
        Extremes(&raInb,   &targetExtremes);

/*
 *      If we have any overlap, then we have a collision
 */
        return (IsOverlap(&rbExtremes, &targetExtremes));
    }
    return 0;

}
/*
 * Internal Internal Function name:
      ThickLinesCollide
  
 * Description:
      Determines if two line segements of non-zero width cross.


 * History:
     13-Dec-1999 - TJF - Original version.
  
 */
INLINE int ThickLinesCollide (
    
    const lineSegType *line1,   /* Line 1                           */
    double line1width,          /* Line 1 width                     */

    const lineSegType *line2,   /* Line 2                           */
    double line2width)          /* Line 2 width                     */

{   

    rectangleType r1;
    rectangleType r2;

/*
 *  Represent the thick lines as rectangles and then check 
 *  for collision between thoses rectangles.
 */
    MakeRectangle(line1, line1width, &r1);
    MakeRectangle(line2, line2width, &r2);

    return RectanglesCollide(&r1, line1->angle, &r2, line2->angle);
    

}

/*
 * Internal Internal Function name:
      AdjustPivot
  
 * Description:
      In 6dF, the pivot point of a fibre tends to move as a result of
      the retractor spring emerging from the retractor.  This routine
      takes the details of a fibre (button and pivot end and theta)
      and adjust the pivot point appropriately according to a formular
      worked out by Will Sanders.

   P[x]= min( 164.5, 162.5 - ((x-132.5)-Sqrt[(x-132.5.)^2 - 4(10 x-1475)])/2)

     where P is pivot radius a a function of extension x and x is measured from
     the original pivot radius at 164.5mm.

     The equation is a quadratic soluation of
         (to be filled in by Will)

     132.5 = The clear length of a fibre (ie. not including spring covered
               length).

     If p1 is less then the pivot radius (164.5mm) then we must re-calculate
     the pivot position.


 * History:
     01-Mar-2001 - TJF - Original version.
  
 */
INLINE void AdjustPivot (
    const pointType   * const buttonEnd,  /* Button end of fibre */
    pointType         * const pivotEnd)   /* pivot end of fibre - adjusted
                                             when returned */
{   
    const double NoSpringLength = 132500;
    double x_dist = buttonEnd->x - pivotEnd->x;
    double y_dist = buttonEnd->y - pivotEnd->y;

    /*
     * Determine the fibre extension.
     */
    double fibreLength = sqrt(SQRD(x_dist) + SQRD(y_dist));

    /*
     * This is the length the spring may be out (x-132.5 above)
     * Note, will be negative when spring is not out
     *    
     */
    double springOutLength = fibreLength - NoSpringLength;


    double p1;
    double rightPart1;
    double rightPart2;

    rightPart1 =  4*(10 * fibreLength - 1475000);
    rightPart2 = sqrt(SQRD(springOutLength) - rightPart1);

    p1 = 162500 - (springOutLength-rightPart2)/2;
    

    if (p1 < FIELD_RAD_INC_PVT)
    {
        /*
         * If we get here, we have to adjust the pivot points for a
         * new radius.  The pivot point is assumed to move out
         * on the line between the pivot point and the center of the plate,
         * until the radius is p1.
         */
        pointType newPivot;
        double angle = atan2(pivotEnd->y, pivotEnd->x);
        newPivot.x = p1 * cos(angle);
        newPivot.y = p1 * sin(angle);

        *pivotEnd = newPivot;
          
    }
}


/*
 * Internal Internal Function name:
     ColButFib
  
 * Description:
     6dF Button-Fibre collision check routine.

     Here we are checking for two things. 

        1.  A collision between
            a circle of radius BUTTON_RAD centered at butX, butY and
            the fibre of radius FIBRE_DIA.  The circle represents the
            fibre button itself.

        We could also check for a collision between two thick line segments, 
        one representing the fibre and the other the tail of
        the button.  Currently I believe this is unnecessary, as it
        actually represents a fibre/fibre collision (we can lay a fibre
        across this tail and put the tail across the fibre)  But, 
        A.  This should be checked to make sure it does not cause problems and
        B.  This may not be the same for OzPoz.
        So the relevant code is #ifdef-ed out, rather then deleted. Note that 
        it has not been tested at all.
        
  
 * History:
     13-Dec-1999 - TJF - Original version.
     01-Mar-2001 - TJF - Adjust the pivot point for the effect caused
                         as the spring comes out of the retractor.
     01-Nov-2001 - TJF - Reduced the size of the button circle used - it was
                         being overly pessimistic (based on fpcolBase.c change
                         by KS of 2-Feb-2001.

     
  
 */
static int  ColButFib (
        void * clientData DUNUSED,
        double    butX,              /* Button X ordinate */
        double    butY,              /* Button Y ordinate */
        double    theta DUNUSED,     /* Button rotation   */
        double    fvpX,              /* Fibre button end  X ordinate */
        double    fvpY,              /* Fibre button end, Y ordinate */ 
        double    pivX,              /* Fibre pivot point, X ordinate */
        double    pivY,              /* Fibre pivot point, Y ordinate */
        long int  fibreClear,        /* Clearance to allow between them */
        double    fibreClearSqrd DUNUSED)/* fibreClear squared */
{
    circleType  button;       /* Used to represent the button itself      */
    lineSegType fibre;        /* Used to represent the fibre              */


    pointType buttonLocation;
    pointType fibrePivotEnd;
    pointType fibreButtonEnd;

/*
 *  Create an object representing the button location point,
 *  and fibre end points
 */
    NewPoint(butX, butY, &buttonLocation);
    NewPoint(fvpX, fvpY, &fibreButtonEnd);
    NewPoint(pivX, pivY, &fibrePivotEnd);
/*
 *  Adjust the pivot point if need be.
 */
    AdjustPivot(&fibreButtonEnd, &fibrePivotEnd);
/*
 *  Contruct the circle object representing the fibre.  We increase the
 *  diameter by the fibre clearance when we do the check.
 */
    NewCircleP(buttonLocation, (BUTTON_RAD + fibreClear), &button);

/*
 *  Generate a line segment description object, given the ends of the fibre.
 */
    NewLineSegEnds(fibrePivotEnd, fibreButtonEnd, &fibre);

/*
 *  First check.  Is there a collsion between the fibre button and the 
 *  fibre
 */
    if (CircleLineOverlap(&button, &fibre))
        return 1;
#if 0  /* No longer do this */
    else
    {
        lineSegType  tail;     /* Used to represent the tail of the button */

/*
 *      Now we must check the tail of the button against the fibre.
 *
 *      Create a line structure representing the tail
 */
        NewLineSegAngle(buttonLocation, MathsTheta(theta), BUTTON_TAIL, &tail);
/*
 *      Do the check.
 */
        return ThickLinesCollide(&tail, BUTTON_TD, 
                                 &fibre, (FIBRE_DIA + fibreClear));
    }
#else
    return 0;
#endif
    
}


/*
 * Internal Function name:
     ColButBut
  
 * Description:
     6dF Button-Button collision check routine.

     Here we are checking for four things

     1.  An overlap of two circles, representing the two buttons
     2.  A collision between two lines, represent the tails of
         each button
     3.  A collision between a circle and a line segment - fibre A's button
         and fibre B's tail.
     4.  A collision between a circle and a line segment - fibre B's button
         and fibre A's tail.
  
 * History:
     13-Dec-1999 - TJF - Original version.
  
 */
static int  ColButBut (
        void * clientData DUNUSED,
        double    butXa,        /* Button A - X ordinate */
        double    butYa,        /* Button A - Y ordinate */
        double    thetaa,       /* Button A rotation     */
        double    butXb,        /* Button B - X ordinate */
        double    butYb,        /* Button B - Y ordinate */
        double    thetab,       /* Button B rotation     */
        long int  buttClear,    /* Tolerance             */
        double    buttClearSqrd DUNUSED)/* buttClear squared     */
{
    circleType buttonA;           /* Circle representing button a */
    circleType buttonB;           /* Circle representing button a */

    pointType  buttonLocationA;   /* Point for button a           */
    pointType  buttonLocationB;   /* Point for button b           */

    
    lineSegType  tailA;           /* Tail of button a             */
    lineSegType  tailB;           /* Tail of button b             */
/*
 *  Create the point types for each button
 */
    NewPoint(butXa, butYa, &buttonLocationA);
    NewPoint(butXb, butYb, &buttonLocationB);
    
/*
 *  Create the circles representing each button.  We add the button 
 *  clearance to B's radius to account for the clearnace.
 */
    NewCircleP(buttonLocationA, BUTTON_RAD, &buttonA);
    NewCircleP(buttonLocationB, BUTTON_RAD+buttClear, &buttonB);

/*
 *  Check for overlap between the buttons themselves.
 */
    if (CirclesOverlap(&buttonA, &buttonB))
        return 1;
/*
 *  We now need to check button A against B's tail.  This is a circle to
 *  thick line check.  We change the radius of the circle to account
 *  for the clearance and tail width, and then we can use a circle/line segment
 *  overlap check.
 */   
    buttonA.radius = BUTTON_RAD + buttClear + BUTTON_THW;

    NewLineSegAngle(buttonLocationB, MathsTheta(thetab), BUTTON_TAIL, &tailB);

    if (CircleLineOverlap(&buttonA, &tailB))
        return 1;

/*
 *  We now need to check button B against A's tail.  This is a circle to
 *  thick line check.  Agin We change the radius of the circle to account
 *  for the clearance and tail width and use the circle/line segument overlap
 *  check.
 */   
    buttonB.radius = BUTTON_RAD + buttClear + BUTTON_THW;

    NewLineSegAngle(buttonLocationA, MathsTheta(thetaa), BUTTON_TAIL, &tailA);

    if (CircleLineOverlap(&buttonB, &tailA))
        return 1;
         
/*
 *  Now we check for a collision between the two tails.  We add the button
 *  clearance to the sizes to allow for it.
 */
    return (ThickLinesCollide(&tailA, BUTTON_TD+buttClear, 
                              &tailB, BUTTON_TD+buttClear));

}
/*
 * Internal Function name:
     ColFidBut
  
 * Description:
     6dF Fiducial-Button collision check routine.

     Here we are checking two things

     1.  For an overlap between two circles, one representing the
         button and the other the fiducial.

     2.  For a collision between a line segment representing the button
         tail and a circle representing a fiducial.
  
 * History:
     13-Dec-1999 - TJF - Original version.
  
 */
static int  ColFidBut (
        void * clientData DUNUSED,
        long int  butx,      /* The button X ordinate   */
        long int  buty,      /* The button Y ordinate   */
        double    theta,     /* The button theta        */
        long int  markx,     /* The fiducial X ordinate */
        long int  marky,     /* The fiducial Y ordinate */
        long int  fidClear,  /* The tolerance           */
        double    fidClearSqrd DUNUSED)/* Tolerance squared       */
{
    circleType button;           /* Circle representing button   */
    circleType fiducial;         /* Circle representing fiducial */

    pointType  buttonLocation;   /* Point for button a           */
    pointType  fiducialLocation; /* Point for button b           */

    
/*
 *  Create the point types for the button and fiducial.
 */
    NewPoint(butx,  buty,  &buttonLocation);
    NewPoint(markx, marky, &fiducialLocation);
    
/*
 *  Create the circles representing each object.  We add the tolerance
 *  to the fiducial radius to account for the required clearance.
 */
    NewCircleP(buttonLocation,   BUTTON_RAD,       &button);
    NewCircleP(fiducialLocation, FID_RAD+fidClear, &fiducial);

/*
 *  Check for overlap between the buttons themselves.
 */
    if (CirclesOverlap(&button, &fiducial))
        return 1;

    else
    {
/*
 *      Create a line describing the button tail.
 */
        lineSegType  tail;           
        NewLineSegAngle(buttonLocation, MathsTheta(theta), BUTTON_TAIL, &tail);
/*
 *      Check for overlap between the tail and the fiducial
 */        
        return (CircleLineOverlap(&fiducial, &tail));
    }
    
}
 
/*
 * Internal Function name:
     ColInvPos
  
 * Description:
     6dF Invalid position routine 

     The only invalid position is at the origin of the plate, as the
     gripper is offset from the R arm, it cannot reach the exact center
     of the plate.

     We do this by checking for overlap between two circles, one
     for the plate center and one for the button.
  
 * History:
     13-Dec-1999 - TJF - Original version.
     19-Apr-2000 - TJf - Add plate argument, but this is not used.
     19-Aug-2009 - KS  - Add fibre type argument, also unused. 
  
 */
static int  ColInvPos (
        void * clientData DUNUSED,
        unsigned plate DUNUSED,   /* Plate to check on */
        int       fibreType DUNUSED, /* Fibre type code */
        long int  butx,           /* Button X position */
        long int  buty,           /* Button Y position */
        double    theta  DUNUSED, /* Button theta (Not needed */
        long int  tolerance,      /* Tolerance to allow (button tolerance) */
        double    tolSqrd DUNUSED)/* Tolerance squared */
        
{
    circleType button;
    circleType invPos;

    NewCircleXYR(butx, buty, tolerance, &button);
    NewCircleXYR(0,    0,    PLATE_CEN_INV, &invPos);

    return CirclesOverlap(&button, &invPos);
    
}


/*
 * Internal Function name:
     GetVirPiv
  
 * Description:
     Given a fibre location, return the virtual pivot point for the
     fibre.
  
     Since 6dF buttons can't be rotated, the fibre position is the
     virtual pivot position.

 * History:
     13-Dec-1999 - TJF - Original version.
  
 */
static void GetVirPiv (
        void * clientData DUNUSED,
        double OriginX,
        double OriginY,
        double Theta      DUNUSED,
        double *VirtPivX,
        double *VirtPivY)
{
    *VirtPivX = OriginX;
    *VirtPivY = OriginY;
}


/*
 *+		        s i x d f F p i l M i n I n i t

 * Function name:
      sixdfFpilMinInit

 * Function:
      Initialise Fpil with the 6dF instrument model, minimal FPIL model

 * Description:
      Calls FpilInit with appropiate values for 6dF.  Returns an appropiate
      FpilType variable to be used in calls to Fpil routines.  Fpil
      constants access and coordinate transformations are not supported.

      When you are finished with the "inst" varaible, you should pass it
      to FpilFree(3) to free allocated resources.

      The instrument name string will be "6dF".  The Telescope
      name string will be "UKSCHMIDT".

 * Language:
      C

 * Call:
      (void) sixdfFpilMinInit(inst)

 * Parameters:   (">" input, "!" modified, "W" workspace, "<" output)
      (<) inst (FpilType *) The FPIL instrument description to be used
                            in calls to Fpil routines.


 * Include files: sixdffpil.h

 * See Also: {references}

 * Support: Tony Farrell, AAO

 *-

 * History:
     13-Dec-1999 - TJF - Original version.
 */
extern void sixdfFpilMinInit(FpilType *inst)
{
    FpilCollisionChecksType routines;
    FpilInstParamsType      params;

    sixdfFpil___SetupMinStructs(&routines,&params);

    FpilInit(0, FPIL_VERSION, &routines, &params, 0, inst);
}



/*
 *+		   s i x d f F p i l _ _  _ S e t u p M i n S t r u c t s

 * Function name:
      sixdfFpil___SetupMinStructs

 * Function:
      Initialise minimal mode Fpil structures with the 2dF instrument model

 * Description:
      Sets up the structures routines and params ready for calling
      FpilInit() such that they support the minimal FPIL mode.

 * Language:
      C

 * Call:
      (void) = sixdfFpil___SetupMinStructs(rouintes, params);

 * Parameters:   (">" input, "!" modified, "W" workspace, "<" output)
      (<) routines (FpilCollisionChecksType *) Will be setup with the
                     collision detection routines for 2dF.  (AccessCons
                     is NOT set).
      (<) params   (FpilInstParamsType *) Will be set up with the 2dF
                     instrument parameters

 * Include files: sixdffpil.h

 * See Also: {references}

 * Support: Tony Farrell, AAO

 *-

 * History:
     13-Dec-1999 - TJF - Original version.
     04-Feb-2000 - TJF - Add NumFields and NumPivots entries 
     17-Feb-2000 - TJF - Add FieldRadIncPark and NumFids fields.
     03-Apr-2000 - TJF - Telescope name must match slaObs() name.
                         FIELD_RAD_INC_PARK changed to FIELD_RAD_INC_PVT.
     19-Apr-2000 - TJF - Add ParkMayCollide field.
     12-Dec-2007 - TJF - Support 3rd field plate, but only if a particular
                         environment variable is not set.
 */
extern void  sixdfFpil___SetupMinStructs(
    FpilCollisionChecksType *routines,
    FpilInstParamsType      *params)
{
    routines->ColButFib  = ColButFib;
    routines->ColButBut  = ColButBut;
    routines->ColFidBut  = ColFidBut;
    routines->ColInvPos  = ColInvPos;
    routines->GetVirPiv  = GetVirPiv;
    routines->AccessCons = 0;

    params->ButtonClear = BUTTON_CLEAR;
    params->FibreClear  = FIBRE_CLEAR;
    params->FidClear    = FID_DIA+FIBRE_CLEAR;
    params->FieldRad    = FIELD_RAD;
    params->FieldRadIncPark = FIELD_RAD_INC_PVT;
    params->Extension   = EXTENSION;
    params->MaxPivAngle = MAX_PIV_ANGLE;
    params->MaxFibAngle = 0;
    params->FibAngleVar = 0;
    params->MaxSize     = BUTTON_TAIL + BUTTON_RAD;
    if (getenv(SIXDFFPIL_TWO_PLATES_ENV_VAR_NAME) == 0)
        params->NumFields   = 3; 
    else
        params->NumFields   = 2; 
    params->NumPivots   = 154; /* 150 Object, 4 guide */
    params->NumFids     = 12;
    params->ParkMayCollide = 1; 
    strcpy(params->Telescope,"UKST"); /* Must match slaObs() name */
    strcpy(params->Instrument,"6dF");

}



