/*+		S I X D F F P I L F U L L
 * Module name:
     sixdFfpifull.c

 * Function:
      Wrap up 6dF constants file access and telescope model access.

 * Description:
      This module implements sixdfFpilFullInit(3), a routine which invokes
      FpilInit(3) with appropiate values for 6dF in FPIL's full mode.  
      (collision detection, constants access and coorindate transformation) 

      We make use of sixdfFpil___SetupMinStructs to fetch the FPIL Mininal
      mode setup (collision detection, basic instrument details etc.)

 * Language:
      C

 * Support: Tony Farrell, AAO

 *-

 *     @(#) $Id: ACMM:6dFutil/sixdFfpifull.c,v 4.3 06-Mar-2013 08:32:34+11 tjf $

 * History:
      20-Jan-2000 - TJF - Original version
      24-Feb-2000 - TJF - Remove RCS header items and add Sccs Id.
      30-Jan-2001 - KS  - Made the philosophical change that we are interested
                          in the number of different fibre types rather than
                          just the number of spectrographs. This is much more
                          flexible, and allows Fpil to accomodate complex
                          instruments with many fibre types. Accordingly, now
                          treat a guide fibre not as a special case, but as 
                          just a different fibre type. So added 'guide' to
                          SpecNames array and made corresponding changes to
                          sixdfFpilFullInit() and WhichSpec(). I'd like to 
                          see WhichSpec() renamed to WhichFibreType(), and
                          all the corresponding routine name changes, but
                          that can be a job for later.
      05-Jun-2002 - KS  - PATH_MAX defined even if missed in include files.
      12-Mar-2003 - CR  - Changed Spectrograph name from FLAIR to 6dF
      03-Apr-2003 - CR  - Changed directory to look for constants file
      {@change entry@}
 
 */

/*
 *  Include files
 */
#define __EXTENSIONS__   /* To get PATH_MAX on solaris */


static char *rcsId="@(#) $Id: ACMM:6dFutil/sixdFfpifull.c,v 4.3 06-Mar-2013 08:32:34+11 tjf $";
static void *use_rcsId = (0 ? (void *)(&use_rcsId) : (void *) &rcsId);


#include <limits.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>

#include "drama.h"
#include "sds.h"
#include "Ers.h"
#include "sixdffpil.h"
#include "sixdfpos.h"
#include "sixdFbuttonDims.h"
#include "sixdf_util_err.h"

#ifndef PATH_MAX
#define PATH_MAX 1024
#endif

/* 
 * Constant definitions
 */
#define SIXDFCONSFILE_OLD "sixdFconstants.sds"     /* 6dF constants file 2 plates */
#define SIXDFCONSFILE3 "sixdFconstants3.sds"       /* 6dF constants file 3 plates*/
#define SIXDFCONSDIR  "/home/6dFrobot/positioner"  /* Directory at the AAT/AAO */

/*
 * Define the guide pivot numbers, which are fixed.
 */
#define GUIDE_PVT_1 28
#define GUIDE_PVT_2 70
#define GUIDE_PVT_3 105
#define GUIDE_PVT_4 147

/*
 * Macro function definitions
 */

/*
 * type definitions
 */

/*
 * Global variables and functions imported (but not declared in include files)
 */

/*
 * Global variables exported (definitions and initialization of globals
 * declared in the associated include file)
 */

/*
 * Variables private to this module (static variables)
 */

/*
 * Spectrograh name, used by WhichSpec() and sixdfFpilFullInit
 */

static const char * const SpecNames[] = {
   "GUIDE",
   "6dF"
};


/*

 * Telescope model parameter names (currently based on AAT ones, but not
 * yet used)
 */
static const char * const ParamNames[] = {
    "MA",
    "ME",
    "NP",
    "CH"
};

/*
 * Telescope model parameter descriptions.
 */
static const char * const ParamDescrs[] = {
    "Azimuth misalignment - MA",
    "Elevation  misalignment - ME",
    "Non-perpendicularity - NP",
    "Collimation error in HA - CH",
};

#define MA 0
#define ME 0
#define NP 0
#define CH 0

/*
 * Telescope parameter default values.
 */
static const double ParamDefaults[] = {
    MA, ME, NP, CH
};




/*
 * Internal Function name:
      AccessCons
  
 * Description:
      Accesses the 6dF positioner constants file from the
      specified directory.
  
 * History:
      20-Jan-2000  TJF   Original version
      10-Mar-2000 - TJF - Add extra arguments to support search paths
                         and reporting of which file was opended and
                         implemented that functionality.
      12-Dec-2007 - TJF - Warning if we don't have the constants for the 
                           3rd field plate, but the relevant environment
                           variable to disable it is not set.
      17-Dec-2007 - TJF - 3 plate version revamped such that plate numbers
                          are 1 based rather then 0 based and the file name
                          changes.  So we check the environment variable before
                          looking for the file.  We must renamed the structures
                          to field0 through field2, so that they work with
                          the config routines which expect them to be zero
                          based.
  
 */
static void AccessCons (
    void * clientData DUNUSED,
    const char *dirOverride,
    const char *dirFallback,
    FpilRptRoutineType RptRoutine,
    void * RptArg,
    SdsIdType *id,
    StatusType *status)
{
    char filename[PATH_MAX]="";
    char *baseName = SIXDFCONSFILE3;
    int twoPlate = 0;

    if (*status != STATUS__OK) return;

/*
 *  Are we configuring for the 2 plate version of 6dF or the 3 plate version.
 *  The fail name and format changes.
 */
    if (getenv(SIXDFFPIL_TWO_PLATES_ENV_VAR_NAME) != 0)
    {
        baseName = SIXDFCONSFILE_OLD;
        twoPlate = 1;
        fprintf(stderr,"Configuring for 2 Plate version of 6dF\n");
        fprintf(stderr,
                " To configure for 3 plate version, unset the environment\n   variable %s\n",
                SIXDFFPIL_TWO_PLATES_ENV_VAR_NAME);
    }


    ErsPush();
/*
 *  First attempt, using the override directory if we have one.
 */
    if (dirOverride)
    {
        ErsSPrintf(sizeof(filename), filename, "%s/%s", 
                   dirOverride, baseName);
        SdsRead(filename, id, status);
    }
    if ((!dirOverride)||(*status == SDS__FOPEN)||(*status == SDS__FREAD))
    {
/*
 *      It is not in the override directory, try the /instsoft/2dF/positioner
 *      directory.
 */

        ErsAnnul(status);
        ErsSPrintf(sizeof(filename), filename, "%s/%s", 
                   SIXDFCONSDIR, baseName);
        SdsRead(filename, id, status);
        if (((*status == SDS__FOPEN)||(*status == SDS__FREAD))&&dirFallback)
        {
/*
 *           It is not in the default directory, the fallback
 *           directory if these is one.
 */
            ErsAnnul(status);
            ErsSPrintf(sizeof(filename), filename, "%s/%s", 
                       dirFallback, baseName);
            SdsRead(filename, id, status);
            
        }
    }
    ErsPop();
    if ((*status == SDS__FOPEN)||(*status == SDS__FREAD))
    {
        ErsRep(0,status,"Failed to open 6dF constants file \"%s\"",
               baseName);

        if (dirOverride)
            ErsRep(0,status,"   Tried looking in \"%s\"",dirOverride);

        ErsRep(0,status,"   Tried looking in \"%s\"",SIXDFCONSDIR);

        if (dirFallback)
            ErsRep(0,status,"   Tried looking in \"%s\"",dirFallback);


        if (twoPlate)
        {
            ErsRep(0, status, "NOTE - If you want to configure for the 3 plate version of 6dF, plate unset the environment variable %s.\n",
                   SIXDFFPIL_TWO_PLATES_ENV_VAR_NAME);
        }
        else
        {
            ErsRep(0, status, "NOTE - If you want to configure for the 2 plate version of 6dF, plate set the environment variable %s to any value.\n",
                   SIXDFFPIL_TWO_PLATES_ENV_VAR_NAME);
        }
    }
    else if (*status != STATUS__OK)
    {
        ErsRep(0,status,"Failed to open 6dF constants file \"%s\"",
               baseName);
        ErsRep(0, status, "General failure, status code %lx",
               (long int )*status);
    }
    else if (RptRoutine)
    {
        if (!twoPlate)
        {
            /*
             * 3 field plate 6dF mode.  At the moment a lot of
             * the configure code relies on the fields being named
             * 0, 1 and 2, rather then 1, 2 and 3.  Rename them here to
             * avoid modifying a lot of other code.
             */
            SdsIdType tid = 0;
            SdsFind(*id, "field1", &tid, status);
            SdsRename(tid, "field0", status);
            SdsFreeId(tid, status);

            SdsFind(*id, "field2", &tid, status);
            SdsRename(tid, "field1", status);
            SdsFreeId(tid, status);

            SdsFind(*id, "field3", &tid, status);
            SdsRename(tid, "field2", status);
            SdsFreeId(tid, status);

            if (*status != STATUS__OK)
            {
                ErsRep(0,status,"6dF constants file rename failed\"%s\" = %ld",
                       baseName,
                       (long int )*status);
                
            }
        }


        (*RptRoutine)(RptArg,filename,status);
    }
    
}

/*
 * Internal Function name:
      WhichSpec

 * Description:
      Determine the spectrograph associated with a pivot.

      6dF has only one spectrogrph (FLAIR), but we must indicate the
      Guide fibres.

 * History:
     17-Mar-2000 - TJF - Original DUMMY version.  Does not indicate guide
                         fibres.
     22-Jun-2000 - TJF - We now know which are the guide fibres, so
                         implement.
     30-Jan-2001 - KS  - SpecNames now includes the guide fibres, so make use
                         of that.
 */


static void WhichSpec (
    void * clientData DUNUSED,
    unsigned field DUNUSED,
    unsigned pivot,
    SdsIdType Constants DUNUSED,
    unsigned *spec,
    const char ** SpecName,
    StatusType *status)
{
    if (*status != STATUS__OK) return;

    switch (pivot)
    {
        case GUIDE_PVT_1:
        case GUIDE_PVT_2:
        case GUIDE_PVT_3:
        case GUIDE_PVT_4:
            *spec = 0;
            break;
        default:
            *spec = 1;
    }
    *SpecName = SpecNames[*spec];
  
}
/*
 * Internal Function name:
      DrawButFib

 * Description:
      Return the appropiate information for Drawing a button and fibre.

      The Tcl/Tk cavans widget graphic we 

  
 * History:
     17-Mar-2000 - TJF - Original version
     03-Apr-2000 - TJF - Completed implementation.
  
 */
static void DrawButFib  (
        void * clientData DUNUSED,
        double scaling DUNUSED,
        double zoom,            
        FpilCvtRoutineType CvtX,  /* For converting Microns in X to pixels */
        FpilCvtRoutineType CvtY,  /* For converting Microns in Y to pixels */
        void * CvtData,           /* To be passed to CvtX and CvtY         */
        double X,                 /* Fibre position in X, microns on plate */
        double Y,                 /* Fibre position in Y, microns on plate */
        double Theta,             /* Button theta in radians               */
        int Type DUNUSED,         /* Fibre type                            */
        unsigned ButGrapLen,      /* Length of ButGrapic string            */
        double *FibX,             /* Fibre X position in GUI returned here */
        double *FibY,             /* Fibre Y position in GUI returned here */
        char *ButGraphic1,        /* Tcl/Tk cavnas create command arugment 
                                     for drawing the button is returned 
                                     here                                  */
        char *ButGraphic2,        /* As per ButGraphic2, for second graphic*/
        FpilStatusType *status)
{
    char tempString[500];

    float bottomLeftX, bottomLeftY; /* Bottom left of circle defintion */
    float topRightX, topRightY;     /* Top right of circle definition  */
#   define NUM_POINTS 4             /* Number of points in polygon     */
    float xb[NUM_POINTS];           /* Tail X ordinates                */
    float yb[NUM_POINTS];           /* Tail Y ordinates                */
    float c,s;                      /* For cosine and sine of theta    */
    float tailYStart;               /* Y start position of tail        */
    register unsigned i;

    if (*status != STATUS__OK) return;
    
/*
 *  See sixdFbuttonDims.h for a discription of a 6dF button, but rotate
 *  the diagram by 90 degress clockwise to get the coorindate system
 *  used here.   
 *
 *  The orgin of a button is the fibre positon (X.Y).
 *
 *  We use both the Tcl/Tk grapic items, graphic 1 is used for a circle
 *  centered at the fibre and of radius BUTTON_RAD.  graphic 2 is used
 *  for the tail of the fibre, which is of length BUTTON_TAIL and width 
 *  (BUTTON_THW * 2).
 *
 *  The point where the tail leaves the fibre is given by the
 *  intercept between the circle  X^2 + Y^2 = BUTTON_RAD^2 and
 *  the line X = BUTTON_THW (remember, rotated by 90 over sixdFbuttonDims.h), 
 *  which is
 *
 *     Y^2 = BUTTON_RAD^2 - BUTTON_THW^2
 *
 *
 * We use a circle to represent the button itself and use graphic one
 * to draw this.  Here we need the coorindates of a bounding square around
 * the circle, adjusted for the fibre position and converted appropiately.
 *
 * Note, we don't need to take acount of the button rotation, as we would
 * be rotating a circle.
 */

    bottomLeftX = (*CvtX)(CvtData, (-1*BUTTON_RAD)+X, zoom);
    bottomLeftY = (*CvtY)(CvtData, (-1*BUTTON_RAD)+Y, zoom);
    topRightX   = (*CvtX)(CvtData, BUTTON_RAD+X,      zoom);
    topRightY   = (*CvtY)(CvtData, BUTTON_RAD+Y,      zoom);

    sprintf(tempString,"oval %f %f %f %f",
            bottomLeftX,
            bottomLeftY,
            topRightX,
            topRightY);

/*
 *  Check we have space and then copy the string to the output string.
 */
    if ((strlen(tempString)+1) >= ButGrapLen)
    {
        *status = SIXDF_UTIL__GRAPSTRLEN;
        return;
    }
    else
    {
        strcpy(ButGraphic1, tempString);
    }

/*
 *  Now we need to draw the tail (into ButGraphic2) and determine
 *  the Fibre position where it leaves the tail (into FibX, FibY).
 *
 *  We set up an array of the four positions which define the tail
 */
    tailYStart = sqrt((BUTTON_RAD*BUTTON_RAD) - (BUTTON_THW * BUTTON_THW));
    
    xb[0] = -BUTTON_THW;
    yb[0] = tailYStart;

    xb[1] = -BUTTON_THW;
    yb[1] = BUTTON_TAIL;
    
    xb[2] = BUTTON_THW;
    yb[2] = BUTTON_TAIL;

    xb[3] = BUTTON_THW;
    yb[3] = tailYStart;

/*
 *  Determine the rotation of the button
 */
    c = cos(Theta);
    s = sin(Theta);
   
/*
 *  Now work through, applying the rotation and converting the
 *  position to pixel units.
 */
    for (i=0 ; i < NUM_POINTS ; i++)
    {
        float x0,y0;
        x0 = xb[i]; 
        y0 = yb[i];
/*
 *      Rotate as required.
 */
        xb[i] = c*x0 - s*y0 + X;
        yb[i] = c*y0 + s*x0 + Y;
/*
 *      Convert to pixel units.
 */
        xb[i] = (*CvtX)(CvtData,xb[i], zoom);
        yb[i] = (*CvtY)(CvtData,yb[i], zoom);
    }
/*
 *  Determine the fibre end.  This is where the tail ends. so
 *  average the x and y values of each side of the tail end to
 *  get the fibre end.
 */
    *FibX = (xb[1]+xb[2])*0.5;
    *FibY = (yb[1]+yb[2])*0.5;

    
/*
 *  Create our Tcl/Tk canvas create arguments - to create the polygon.
 */
    sprintf(tempString, "polygon %f %f %f %f %f %f %f %f",
            xb[0],yb[0],
            xb[1],yb[1],
            xb[2],yb[2],
            xb[3],yb[3]);

/*
 *  Copy into the output string, validating the length as we go.
 */
    if ((strlen(tempString)+1) >= ButGrapLen)
        *status = SIXDF_UTIL__GRAPSTRLEN;
    else
        strncpy(ButGraphic2, tempString, ButGrapLen);

}






/*
 *+		        S i x d f F p i l F u l l I n i t

 * Function name:
      sixdfFpilFullInit

 * Function:
      Initialise Fpil with the 6dF instrument model, full FPIL model

 * Description:
      Calls FpilInit with appropiate values for 6dF.  Returns an appropiate
      FpilType variable to be used in calls to Fpil routines.

      When you are finished with the "inst" varaible, you should pass it
      to FpilFree(3) to free allocated resources.

      The instrument name string will be "6dF".  The Telescope
      name string will be "UKSCHMIDT".

 * Language:
      C

 * Call:
      (void) sixdfFpilFullInit(inst)

 * Parameters:   (">" input, "!" modified, "W" workspace, "<" output)
      (<) inst (FpilType *) The FPIL instrument description to be used
                            in calls to Fpil routines.


 * Include files: sixdffpil.h

 * See Also: {references}

 * Support: Tony Farrell, AAO

 *-

 * History:
      20-Jan-2000  TJF   Original version
      17-Mar-2000  TJF   Add WhichSpec, DrawButFib and telescope model
                         parameter stuff.
      30-Jan-2001  KS    Now treats 'number of spectrographs' as 'number of
                         fibre types' and so reports this as 2 not 1 (ie it
                         now includes guide fibres as a type).
      27-Feb-2013  TJF   Add Clone function.
 */
extern void sixdfFpilFullInit(FpilType *inst)
{
    FpilCollisionChecksType routines;
    FpilInstParamsType      params;
    FpilModelRoutinesType   modelRoutines;

/*
 *  Set up the basic instrument details and collision detection
 */
    sixdfFpil___SetupMinStructs(&routines,&params);

/*
 *  Set up the access to the contants file
 */
    routines.AccessCons    = AccessCons;
    routines.WhichSpec     = WhichSpec;
    routines.DrawButFib    = DrawButFib;

    params.NumSpec         = 2;
    params.SpecNames       = SpecNames;

    
/*
 *  Set up access to the telescope model.
 */
    modelRoutines.Access  = SixdfposTransAccess;
    modelRoutines.Free    = SixdfposFree;
    modelRoutines.CvtInit = SixdfposCvtInit;
    modelRoutines.Rd2Xy   = SixdfposRd2Xy;
    modelRoutines.Xy2Pos  = SixdfposXy2Pos;
    modelRoutines.Xy2Rd   = SixdfposXy2Rd;
    modelRoutines.Pos2Xy  = SixdfposPos2Xy;
    modelRoutines.QuickAppToObs = SixdfposQuickAppToObs;
    modelRoutines.Clone   = Sixdfpos___CloneModel;

    modelRoutines.NumParams = 4;
    modelRoutines.ParamNames = ParamNames;
    modelRoutines.ParamDescrs = ParamDescrs;
    modelRoutines.ParamDefaults = ParamDefaults;
    

    FpilInit(0, FPIL_VERSION, &routines, &params, &modelRoutines, inst);
}





