/*
 *+                                s i x d F s a v e
  
 * Function name:
      SixdfSave
  
 * Function:
	Saves 6dF configuration file after optionally saving a backup.
  
 * Description:
  	In many places, the 6dF positioner software writes Sds files
  	out to disk to save configuration files.  This routine does
  	this and if compiled withe CREATE_BACKUP macro defined to be 1,
        saves a copy of the file in a sub-directory the name 
  	of which is based on the current time.      
  
 * Language:
      C
  
 * Call:
     (void) = SixdfSave (file,usePath,pathName,sdsId,status)
  
 * Parameters:   (">" input, "!" modified, "W" workspace, "<" output)
 	(>) file (const char *) The base name of the file involved.  
 	(>) usePath (int) If true, then save in the directory refered to
 				by pathName, otherwise, save in the current
 				directory
 	(>) pathName (const char *) Only used if usePath is true, in which
 				case it indicates the directory in which
 				to save the files.
 	(>) sdsId	(SdsIdType) The Sds data to save 
	(>) show	(int) If true, log writing of file
	(>) message	(const char *) If show is true, a message to be output
			if the write successed and show it true.  It can 
			contain one %s character to indicate where the
			filename should be formated into the string.
	(!) status	(StatusType *) Modified status
        
  
 * Include files: tdFutil.h
  
 * External functions used:
        
  
 * External values used:
        
  
 * Prior requirements:
        
  
 * Support: Tony Farrell, AAO
  
 *-

 * "@(#) $Id: ACMM:6dFutil/sixdFsave.c,v 4.3 06-Mar-2013 08:32:34+11 tjf $"
  
 * History:
      10-Sep-1996 - TJF - Original version
      27-Oct-1999 - TJF - Don't need the backup anymore.  Disable it by
                          wrapping the code up in a macro.
      22-Feb-2000 - TJF - Create 6dF version, which does need the back
                          up stuff for the moment.  No VxWorks support
                          required.
      24-Feb-2000 - TJF - Rename SIXDFXY__SPRINTF and SIXDFXY__DIRCREERR
                          to SIXDF_UTIL equivalents.  Include sixdf_util_err.h.
      02-Mar-2001 - TJF - Intead of creating a new directory every minutes,
                          create one every 15 minutes, that should
                          be sufficent
  
 */



static char *rcsId="@(#) $Id: ACMM:6dFutil/sixdFsave.c,v 4.3 06-Mar-2013 08:32:34+11 tjf $";
static void *use_rcsId = (0 ? (void *)(&use_rcsId) : (void *) &rcsId);


#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
/* Most unix versions don't declare this in include files, but linux and
 * MacOsX do.
 * Need to find out what the standards say about it
 */
#    if !defined(__linux__) && !defined(__APPLE__)
    extern char *sys_errlist[];
#endif
#include <time.h>

#include "DitsTypes.h"
#include "sds.h"
#include "DitsMsgOut.h"
#include "Ers.h"
#include "sixdFutil.h"
#include "sixdf_util_err.h"

#ifdef DFUDGE_PROTOTYPES_UNIX
    extern int strftime(char *buf, int bufsize, const char *fmt, 
    		const struct tm *tm);
	/*
    extern int mkdir(const char *path, mode_t mode);
	*/
    extern int rename (const char *oldname, const char *newname);

#endif
#ifdef DFUDGE_PROTOTYPES_C
   extern time_t time(time_t *);
#endif

/*
 *  Define this to be 1 if you want to create the backup in the
 *  date and time based directory.
 */
#ifndef CREATE_BACKUP
#define CREATE_BACKUP 1
#endif

extern void SixdfSave (const char * file, int usePath, const char *pathName,
                       SdsIdType sdsId, int show,
                       const char *message, StatusType *status)
{
    char fname[200];	   /* Name for actual file	*/
#   if CREATE_BACKUP
        char sname[200];   /* Name for saved version of file */
        char datedir[200]; /* Component of saved file name with just date */
        char timedir[200]; /* Component of savedfile name with data and time*/
    
        time_t t;
        struct tm timeBuffer;
    
        char day[10];
        char curtime[10];
#   endif    

    if (*status != STATUS__OK) return;
    


#   if CREATE_BACKUP
/*
 *  Get the current time.
 */
        t = time(0);
        timeBuffer = *gmtime(&t);

/*
 *      Format a date string of the form
 *  		yymmdd
 */
        strftime(day,sizeof(day),"%y%m%d",&timeBuffer);
    
/*
 *      Format the time in the form.  We create a new directory
 *      every 15 minutes.
 *  
 *  		hh:mm:ss
 */
        if (timeBuffer.tm_min < 15) 
            strftime(curtime,sizeof(curtime),"%H:00:00",&timeBuffer);
        else if (timeBuffer.tm_min < 30) 
            strftime(curtime,sizeof(curtime),"%H:15:00",&timeBuffer);
        else if (timeBuffer.tm_min < 45) 
            strftime(curtime,sizeof(curtime),"%H:30:00",&timeBuffer);
        else
            strftime(curtime,sizeof(curtime),"%H:45:00",&timeBuffer);

#   endif /* CREATE_BACKUP */    
    if (usePath)
    {
        if (ErsSPrintf(sizeof(fname),fname,"%s/%s.sds",pathName,file) == EOF)
                *status = SIXDF_UTIL__SPRINTF;
     
#     if CREATE_BACKUP           
        else if (ErsSPrintf(sizeof(datedir),datedir,"%s/%s",pathName,day)
        					== EOF)
                *status = SIXDF_UTIL__SPRINTF;
                
        else if (ErsSPrintf(sizeof(timedir),timedir,"%s/%s/%s",pathName,day,
        		curtime) == EOF)
                *status = SIXDF_UTIL__SPRINTF;
                
        else if (ErsSPrintf(sizeof(sname),sname,"%s/%s/%s/%s.sds",pathName,day,
        		curtime,file) == EOF)
                *status = SIXDF_UTIL__SPRINTF;
#     endif
    }
    else
    {
        if (ErsSPrintf(sizeof(fname),fname,"%s.sds",file) == EOF)
                *status = SIXDF_UTIL__SPRINTF;
#     if CREATE_BACKUP                
        else if (ErsSPrintf(sizeof(datedir),datedir,"%s",day)
        					== EOF)
                *status = SIXDF_UTIL__SPRINTF;
                
        else if (ErsSPrintf(sizeof(timedir),timedir,"%s/%s",day,
        		curtime) == EOF)
                *status = SIXDF_UTIL__SPRINTF;
                
        else if (ErsSPrintf(sizeof(sname),sname,"%s/%s/%s.sds",day,
        		curtime,file) == EOF)
                *status = SIXDF_UTIL__SPRINTF;
#     endif
    }
    
    if (*status != STATUS__OK) return;
    
/*
 *  Make the directory corresponding to today's date.
 */
#   if CREATE_BACKUP
        if ((mkdir(datedir,S_IRWXG|S_IRWXU) == -1)&&(errno != EEXIST))
        {
            *status = SIXDF_UTIL__DIRCREERR;
            ErsRep(0,status,"Error creating directory \"%s\", errno = %x",
                   datedir,errno);
            ErsRep(0,status,"Error text is \"%s\"",sys_errlist[errno]);
            return;
        }
        errno = 0;
    
/*
 *  Make the directory corresponding to the current time.
 */
        if ((mkdir(timedir,S_IRWXG|S_IRWXU) == -1)&&(errno != EEXIST))
        {
            *status = SIXDF_UTIL__DIRCREERR;
            ErsRep(0,status,"Error creating directory \"%s\", errno = %x",
                   timedir,errno);
            ErsRep(0,status,"Error text is \"%s\"",sys_errlist[errno]);
            return;
        }
       errno = 0;
#   endif /* CREATE_BACKUP */    

/*
 *  Write file twice, once in the spot where it will be picked up next 
 *  time and once in the saved location.
 */
    SdsWrite(sdsId,fname,status);
    if (*status != STATUS__OK)
	ErsRep(0,status,"Error writing file \"%s\", errno = %d",fname,errno);
#   if CREATE_BACKUP
        else
        {
            SdsWrite(sdsId,sname,status);
            
            if ((*status == STATUS__OK)&&(show))
                MsgOut(status,message,fname);
            else if (*status != STATUS__OK)
                ErsRep(0,status,"Error writing file \"%s\", errno = %d",
                       sname,errno);   
        }
#   endif
    
}
