/*
 *+			S I X D F P O S

 * Module name:
      sixdfpos

 * Function:
      6dF position transformation library.

 * Description:
      This library implements routines to provide for transformation of
      sky positions in RA/DEC coordinates to 6dF positioner coorindates.

      It implements functions required by the FPIL library to support
      coordinate transformations.

      NOTE: Correct use of telescope parameters not defined.

 * Language:
      C


 * Support: Tony Farrell, Ken Russell, AAO

 *-


 *     @(#) $Id: ACMM:6dFutil/sixdfpos.c,v 4.3 06-Mar-2013 08:32:34+11 tjf $

 * History:
     03-Dec-1999 - TJF - Original version
     24-Feb-2000 - TJF - Include sixdf_util_err.h, rename SIXDFPOS__MALLOCERR
                         to SIXDF_UTIL__MALLOCERR.
     21-Jun-2000 - TJF - Implement basic telescope model based on 2dF code
                         but without distortion modeling as yet.
     04-Mar-2001 - TJF - Use slaPcd to apply radial distortion.  Still need
                         to investigate telescope model parameters and their
                         usage
     30-Apr-2002 - TJF - Add field centre wavelength to CvtInit.
     27-Feb-2013 - TJF - Add Clone function.
     
 */

/*
 *  Include files
 */
#define __EXTENSIONS__  /* For M_PI under solaris */


static char *rcsId="@(#) $Id: ACMM:6dFutil/sixdfpos.c,v 4.3 06-Mar-2013 08:32:34+11 tjf $";
static void *use_rcsId = (0 ? (void *)(&use_rcsId) : (void *) &rcsId);


#include <stdlib.h>     /* For malloc and free */
#include <math.h>
#include "sixdfpos.h"
#include "sixdFfieldDims.h"
#include "sixdf_util_err.h"
#include "drama.h"
#include "slalib.h"
#include "slamac.h"
/* 
 * Constant definitions
 *
 * THREE_DEGREES is 3 degress in radians.
 */
#define THREE_DEGREES (((double)2.0*M_PI*(double)3.0)/(double)360.0)

/*
 * Telescope parameter index values.  See  sixdFfpifull.c.  
 */
#define MA_INDEX 0
#define ME_INDEX 1
#define NP_INDEX 2
#define CH_INDEX 3
/*
 * Telescope focus length in microns (3.0723 m)
 */
#define UKST_FOCUS_LENGTH (3.0723*1.0e6)

/*
 * For use with SLA_PCD and SLA_UNPCD
 */
#define SCHMIDT_DISCO (-1.0/3.0)
/*
 * type definitions
 */

/*
 * This structure contains the parameters of the distortion model 
 * read/created by SixdfposTransAccess()
 */
typedef struct {
/*
    double a;
    double b;
    double c;
    double d;
    double x0;
    double y0;
    */
} SixdfDistType;


/*
 * This structure contains the parameters of the linear model 
 * read/created by SixdfposTransAccess() and it's inverse.
 */
typedef struct  {
    double coeffs[6];      /* Forward direction coefficents */
    double invCoeffs[6];   /* Reverse direction  coefficents */
} SixdfLinType;


/*
 * This structure contains details passed from SixdfposCvtInit to the
 * various conversion routines.
 */
typedef struct  {
    double cenAoprms[14];     /*  Apparent to observed parameters  field cen */
    double obsAoprms[14];     /*  Apparent to observed parameters  obs */
    double rot[3][3];      /*  Rotation matrix, Ax,El to mount frame */
    double cenLambda;         /*  Telelscope pointed wavelength  */
    double obsLambda;         /*  Observing wavelength  */
} SixdfXyType;


/*
 *  This structure contains the position transformation model details.
 */

typedef struct {
    FpilModelType model;        /* Must be first, see FpilModelAccess() */
    SixdfDistType   dist;       /* Set by SixdfposTransAccess  */
    SixdfLinType    linpars;    /* Set by SixdfposTransAccess  */
    SixdfXyType     xypars;     /* Set by SixdfposCvtInit      */
} SixdfposModel, *SixdfposModelPnt;

static void App2mount(int cen, SixdfXyType *xypars, double ra, double dec,
                      double *mha, double *mdec, StatusType *status);



/*
 *+	      	S i x d f p o s T r a n s A c c e s s

 * Function name:
      SixdfposTransAccess

 * Function:
      Access the 6dF transformation details.

 * Description:
      This routine initialises access to the 6dF transformation
      details.  It is called by FpilModelAccess() and must
      fit  the prototype type defintion FpilModelAccessRoutineType

      This routine should access the model and 
      malloc and return the address of a structure of type
      FpilModelType.  By actually mallocing a different structure
      (of type SixdfModelType) which contains as its first item a structure 
      of type FpilModelType, the 6dF telescope model implementation routines
      can keep there own information in this structure.  It just
      casts to and from FpilModelType as required (this can be considered
      deriving from FpilModelType, in the O-O sense).
      
      See FpilModelAccess for details on the usage of the directory
      releated arguments.

 * Language:
      C

 * Call:
      (void) = SixdfposTransAccesss(clientData, dirOverride, 
               dirFallback, RptRoutine, RptArg,  field, model,  status)

 * Parameters:   (">" input, "!" modified, "W" workspace, "<" output)
      (>) clientData      (void *) Was passed to FpilInit.
      (>) dirOverride     (const char *)  If non-null, a directory to
                                   search for the constant details file.
      (>) dirFallback     (const char *)  If non-null, a fall back
                                   directory to search for the constants
                                   details file.
      (>) RptRoutine      (FpilRptRoutineType) A routine to be invoked
                                   to report which file has been opened.
      (>) RptArg          (void *) Argument passed to RptRoutine.
      (>) field     (unsigned int) The field plate number.
      (<) model     (FpilModelType **) The address of the model is returned
                                   here.
      (!) status    (StatusType *) modified status.

 * Include files: sixdfpos.h

 *-

 * History:
     09-Dec-1999 - TJF - Original version - template only at this stage
     10-Mar-2000 - TJF - Add extra arguments to support search paths
                         and reporting of which file was opended.
     10-Jul-2000 - TJF - Had not been swapping X and Y in field plate
			 to positioner units conversion as I should
			 have been.
     06-Mar-2001 - TJF - Correct field plate to position units conversion 
	                 inserted.
     30-Sep-2002 - TJF - Don't report opening file - there is none.

 */
extern void SixdfposTransAccess (
    void * clientData DUNUSED,
    const char *dirOverride DUNUSED,  /* Unused for temp version only */
    const char *dirFallback DUNUSED,  /* Unused for temp version only */
    FpilRptRoutineType RptRoutine DUNUSED,
    void * RptArg DUNUSED ,
    unsigned int field DUNUSED,       /* Unused for temp version only */
    FpilModelType **model,
    StatusType *status)
{
    SixdfposModelPnt MyModel;
    double a,b,c,d,e,f;
    double det;

    if (*status != STATUS__OK) return;

    *model = 0;
/*
 *  Malloc data structure
 */
    if ((MyModel = (SixdfposModelPnt)malloc(sizeof(SixdfposModel))) == 0)
    {
        *status = SIXDF_UTIL__MALLOCERR;
        return;
    }

/*
 *  This is where we should read distortion and linear model parameters
 *  if need be.  We don't yet have a distorition model.  
 */
    
/*
 *  Currenty, the linear transformation from field plate units to
 *  positioner coordinates is simple.  The field plate x axis is the
 *  positioner y axies.  At some stage this must be done correctly, probably
 *  by loading it from a file.
 */
    a = MyModel->linpars.coeffs[0] = 0.0;
    b = MyModel->linpars.coeffs[1] = -1.0;
    c = MyModel->linpars.coeffs[2] = 0.0;
    d = MyModel->linpars.coeffs[3] = 0.0;
    e = MyModel->linpars.coeffs[4] = 0.0;
    f = MyModel->linpars.coeffs[5] = 1.0;

/*
 *  Determine the inverse linear model - to save us working it out
 *  for each conversion. (NOTE, no check for lack of inverse (which should
 *  never happen)
 */ 
    det = b * f - c * e;

    MyModel->linpars.invCoeffs[0] = ( c * d - a * f ) / det;
    MyModel->linpars.invCoeffs[1] = f/det;
    MyModel->linpars.invCoeffs[2] = -1 * c/det;
    MyModel->linpars.invCoeffs[3] = ( a * e - b * d ) / det;
    MyModel->linpars.invCoeffs[4] = -1 * e / det;
    MyModel->linpars.invCoeffs[5] = b / det;

/*
 *  Return.
 */
    if (*status == STATUS__OK)
        *model = (FpilModelType *)MyModel;
    else
        free(MyModel);


}

/*
 *+			S i x d f p o s C v t I n i t

 * Function name:
      SixdfposCvtInit

 * Function:
      Initialise for conversions at a specified time etc.


 * Description:
      This routine initialises the telescope model conversion routines
      for a particular time, wavelength and telescope model details.  This
      allows a sequence of transformation (for one 6dF) field to
      be done efficently.

      This routine is called by FpilModelCvtInit() and must meet
      the prototype given by the type defintion FpilModelCvtInitRoutineType.


 * Language:
      C

 * Call:
      (void) = SixdfposCvtInit(clientData,  mjd, dut, temp, pres, 
                                   humid, cenWave, obsWave, tel, status)

 * Parameters:   (">" input, "!" modified, "W" workspace, "<" output)
      (>) model   (FpilModelType *model) Contains the telescope model details
                           obtained by SixdfosTransAccess(3).  This routine
                           may or may not have been previously invoked
                           on this model.
      (>) mjd     (double) UTC data and time as a modified julian date.
      (>) dut     (double) Delta UT (UT1 - UTC) seconds.
      (>) temp    (double) Atmospheric temperature (K).
      (>) press   (double) Atmospheric pressure (mB).
      (>) humid   (double) Atmospheric humidity (0 to 1.0).
      (>) cenWave (double) Wavelength of telescope pointed using (micons).
      (>) obsWave (double) Wavelength of observation (micons).
      (>) params  (double *) Telescope model parameters.
      (!) status  (StatusType *) modified status.
                           

 * Returned value:

 * Include files: sixdfpos.h

 *-

 * History:
     09-Dec-1999 - TJF - Original version - template only at this stage
     03-Apr-2000 - TJF - Template expanded to use slaObs() and slaAoppa()
                         to set xypars.aoparms for SixdfposQuickAppToObs().
                         This is temp only, taken from 2dF equivalent and
                         should be reconsidered for 6dF.
     20-Jun-2000 - TJF - Implement, based on TdfXyInit(). We don't
                         yet use CH and NP parameters and they are ignored.
 */
extern void SixdfposCvtInit (
    FpilModelType *model,
    double mjd,
    double dut,
    double temp,
    double press,
    double humid,
    double cenWave,
    double obsWave,
    const double *params,
    StatusType *status)
{
    double longit,lat,hm;
    char name[80];
    double a,b;
    SixdfposModelPnt MyModel = (SixdfposModelPnt)model;

    if (*status != STATUS__OK) return;
/*
 *  Get observatory parameters  
 */
    slaObs(-1,"UKST",name,&longit,&lat,&hm);
/*  
 * Determine apparent to observed place parameters  
*/
   slaAoppa(mjd,dut,-longit,lat,hm,0.0,0.0,temp,press,humid,
            cenWave,0.0065, MyModel->xypars.cenAoprms);
   slaAoppa(mjd,dut,-longit,lat,hm,0.0,0.0,temp,press,humid,
            obsWave,0.0065, MyModel->xypars.obsAoprms);
/*
 * Set up rotation matrix to mount Ha,Dec  
 */
   a = lat - DPIBY2 - DAS2R*params[ME_INDEX];
   b = DAS2R*params[MA_INDEX];

   MyModel->xypars.rot[0][0] = cos(a);
   MyModel->xypars.rot[0][1] = 0;
   MyModel->xypars.rot[0][2] = -sin(a);
   MyModel->xypars.rot[1][0] = -sin(a)*sin(b);
   MyModel->xypars.rot[1][1] = cos(b);
   MyModel->xypars.rot[1][2] = -cos(a)*sin(b);
   MyModel->xypars.rot[2][0] = sin(a)*cos(b);
   MyModel->xypars.rot[2][1] = sin(b);
   MyModel->xypars.rot[2][2] = cos(a)*cos(b);

/*
 *  Save the wavelengths.
 */
   MyModel->xypars.cenLambda = cenWave;
   MyModel->xypars.obsLambda = obsWave;
}

/*
 *+			S i x d f p o s R d 2 X y 

 * Function name:
      SixdfposRd2Xy

 * Function:
      Determine 6df x/y from apparent RA and Dec.

 * Description:
      Determines the x,y coordinates in the field plate, from the apparent
      RA and Dec of a source and the field center.

      This routine is called by FpilModelRd2Xy() and must meet
      the prototype given by the type defintion FpilModelRd2XyRoutineType.

 * Language:
      C

 * Call:
      (void) = SixdfposRd2Xy(clientData, cra, cdec, ra, dec, 
                                         mjd, x, y, status)

 * Parameters:   (">" input, "!" modified, "W" workspace, "<" output)
      (>) model   (const FpilModelType *model) Contains the telescope 
                           model details  obtained by SixdfosTransAccess(3).
                           It Can be assumed that the SixdfposCvtInit 
                           routine has been  invoked on this.
      (>) cra     (double) Apparent RA of field center.
      (>) cdec    (double) Apparent DEC of field center.
      (>) ra      (double) Apparent RA of source.
      (>) dec     (double) Apparent DEC of source.
      (>) mjd     (double) UTC time and date expressed as Modified Julian Date
      (<) x       (double *) X position of source on field plate (microns).
      (<) y       (double *) Y position of source on field plate (microns).
      (!) status  (StatusType *) modified status.
 
 * Include files: fpi.h

 *-

 * History:
     09-Dec-1999 - TJF - Original version
     20-Jun-2000 - TJF - Implement real code based on 2dF TdfRd2xy routine.
     04-Mar-2001 - TJF - Apply radial distortion using slaUnpcd after
                         projecting from tanget plane.
 */        
extern void SixdfposRd2Xy (
    const FpilModelType *model,
    double cra,
    double cdec,
    double ra,
    double dec,
    double mjd DUNUSED,     /* Unused for temp version */
    double *x,
    double *y,
    StatusType *status)
{
    const SixdfposModelPnt MyModel = (const SixdfposModelPnt)model;
    double cmha,cmdec;        /* -ha,dec of field centre on AAT mount system */
    double mha,mdec;          /* -ha,dec of source on AAT mount system  */
    double xi,eta;            /* tangent plane coordinates  */
    int jstat;

    if (*status != STATUS__OK) return;

/*
 *  This is based on the 2dF version but without,
 *  at this stage, proper distortion mapping.
 *
 *  Validate arguments
 */
    if ((cra > D2PI) || (cra < 0.0))
    {
        *status = SIXDF_UTIL__ILLRA;
    }
    if ((cdec > DPI) || (cdec < (-DPI)))
    {
        *status = SIXDF_UTIL__ILLDEC;
    }
    if ((ra > D2PI) || (ra < 0.0))
    {
        *status = SIXDF_UTIL__ILLRA;
    }
    if ((dec > DPI) || (dec < (-DPI)))
    {
        *status = SIXDF_UTIL__ILLDEC;
    }
/*  
 *  Update Apparent to observed parameters  
 */
    slaAoppat(mjd, MyModel->xypars.cenAoprms);
    slaAoppat(mjd, MyModel->xypars.obsAoprms);
/*  
 *  Mount frame position of field centre  
 */
    App2mount(1, &MyModel->xypars,cra,cdec,&cmha,&cmdec,status);
/*
 *  Mount frame position of source  
 */
    App2mount(0, &MyModel->xypars,ra,dec,&mha,&mdec,status);

    if (*status != STATUS__OK) return;
/*  
 *  Project to tangent plane.  Remember to catch errors.
 */
    slaDs2tp(mha,mdec,cmha,cmdec,&xi,&eta,&jstat);

    if (jstat != 0) 
    {
        switch (jstat)
        {
        case 1:
            *status = SIXDF_UTIL__SLA_ERR_1;
            break;
        case 2:
            *status = SIXDF_UTIL__SLA_ERR_2;
            break;
        case 3:
            *status = SIXDF_UTIL__SLA_ERR_3;
            break;
        default:
            *status = SIXDF_UTIL__SLA_ERR_UNKN;
            break;
        }
        return;
    }
/*
 * Apply Radial Distortion 
 */
    slaPcd(SCHMIDT_DISCO, &xi, &eta);
/*
 * Multiply by Schmidt focal length to get x and y.  
 */
    *x = xi *  UKST_FOCUS_LENGTH;
    *y = eta * UKST_FOCUS_LENGTH;
   
}

/*
 *+			S I X D F P O S X y 2 P o s 

 * Function name:
      SixdfposXy2Pos

 * Function:
      Convert 6dF field plate coordinates to positioner coordinates.

 * Description:
      Convert the x/y coordinates of a star on the field plate, as output
      by SixdfposRd2Xy to the coordinates required for input to the 6dF
      robot positioner system.   

      This scheme of two x/y coorindate systems allows the robot to
      use a coordinate system which makes sense mechanically, whilst
      the values output by SixdfposRd2Xy make more sense from an
      astromical point of view.  This scheme also takes account of
      non-linearities etc. in the positioner system).  

      This routine is called by FpilModelXy2Ps() and must meet
      the prototype given by the type defintion FpilModelXy2PosRoutineType.

 * Language:
      C

 * Call:
      (void) = SixdfposXy2Pos(clientData, x, y, xp, yp, status)

 * Parameters:   (">" input, "!" modified, "W" workspace, "<" output)
      (>) model   (const FpilModelType *model) Contains the telescope 
                           model details  obtained by SixdfosTransAccess(3).
                           It Can be assumed that the SixdfposCvtInit routine 
                           has been invoked on this.
      (>) x       (double) Field plate x coordinate of object (microns).
      (>) y       (double) Field plate y coordinate of object (microns).
      (>) *xp     (double) Positioner X coordinate of object (microns).
      (>) *yp     (double) Positioner Y coordinate of object (microns).
      (!) status  (StatusType *) modified status.
                        
 * Include files: fpi.h

 *-

 * History:
     09-Dec-1999 - TJF - Original version, based on 2dF equivalent.
 */
extern void SixdfposXy2Pos (
        const FpilModelType *model,
        double x,
        double y,
        double *xp,
        double *yp,
        StatusType *status)
{
    double a,b,c,d,e,f;
    const SixdfposModelPnt MyModel = (const SixdfposModelPnt)model;

    if (*status != STATUS__OK) return;

/*
 *  Extract the transformation coefficent details.
 */  
    a = MyModel->linpars.coeffs[0];
    b = MyModel->linpars.coeffs[1];
    c = MyModel->linpars.coeffs[2];
    d = MyModel->linpars.coeffs[3];
    e = MyModel->linpars.coeffs[4];
    f = MyModel->linpars.coeffs[5];
  
/*
 *  Apply the transformation.
 */
    *xp = a + b*x + c*y;
    *yp = d + e*x + f*y;

}

/*
 *+			S i x d f p o s X y 2 R d 

 * Function name:
      SixdfposXy2Rd

 * Function:
      Determine apparent RA and Dec from 6dF x and y.

 * Description:
      Determine the apparent RA an Dec of a source from the x,y coorindate
      in the field plate and the field center.

      This routine is called by FpilModelXy2Rd() and must meet
      the prototype given by the type defintion FpilModelXy2RdRoutineType.

 * Language:
      C

 * Call:
      (void) = SixdfposXy2Rd(clientData, cra, cdec, x, y, 
                                         mjd, ra, dec, status)

 * Parameters:   (">" input, "!" modified, "W" workspace, "<" output)
      (>) model   (const FpilModelType *model) Contains the telescope 
                           model details  obtained by SixdfosTransAccess(3).
                           It Can be assumed that the SixdfposCvtInit 
                           routine has been invoked on this.
      (>) cra     (double) Apparent RA of field center.
      (>) cdec    (double) Apparent DEC of field center.
      (>) x       (double) X position of source on field plate (microns).
      (>) y       (double) Y position of source on field plate (microns).
      (>) mjd     (double) UTC time and date expressed as Modified Julian Date
      (<) ra      (double *) Apparent RA of source.
      (<) dec     (double *) Apparent DEC of source.
      (!) status  (StatusType *) modified status.


 * Include files: fpi.h

 *-

 * History:
     09-Dec-1999 - TJF - Original version
     20-Jun-2000 - TJF - Implement real code based on 2dF TdfXy2rd routine.
     04-Mar-2001 - TJF - Remove radial distortion using slaUnpcd before
                         deprojecting from tanget plane.
 */
extern void SixdfposXy2Rd (
    const FpilModelType *model,
    double cra,
    double cdec,
    double x,
    double y,
    double mjd DUNUSED,   /* In temp version only */
    double *ra,
    double *dec,
    StatusType *status)
{
    const SixdfposModelPnt MyModel = (const SixdfposModelPnt)model;
    double cmha,cmdec;        /* -ha,dec of field centre on AAT mount system */
    double mha,mdec;          /* -ha,dec of source on AAT mount system  */
    double xi,eta;            /* tangent plane coordinates  */
    double obs[3];            /* Observed Az, El cartesian vector  */
    double mount[3];          /* Mount -HA, Dec cartesian vector  */
    double az,el;             /* Observed Az, El */
    double zd;                /* Observed Zd */

    if (*status != STATUS__OK) return;
/* 
 *  This based on the 2dF version but without
 *  at this stage, proper distortion mapping.
 */

/*  
 * Update Apparent to observed parameters  
 */
    slaAoppat(mjd,MyModel->xypars.cenAoprms);
    slaAoppat(mjd,MyModel->xypars.obsAoprms);
/*  
 * Mount frame position of field centre  
 */
    App2mount(1, &MyModel->xypars,cra,cdec,&cmha,&cmdec,status);
/*
 * Convert from focal plane to tange plate co-ordinates
 */
    xi  = x/UKST_FOCUS_LENGTH;
    eta = y/UKST_FOCUS_LENGTH;
/*
 * Remove Radial Distortion 
 */
    slaUnpcd(SCHMIDT_DISCO, &xi, &eta);
/*  
 * Deproject from tangent plane to give mount -ha, dec  
 */
    slaDtp2s(xi,eta,cmha,cmdec,&mha,&mdec);
/*  
 * Convert to cartesian  
 */
    slaDcs2c(mha,mdec,mount);
/*  
 * Rotate to observed Az, El  (AZ south through east) 
 */
    slaDimxv(MyModel->xypars.rot,mount,obs);
/*  
 *Convert to spherical  
 */
    slaDcc2s(obs,&az,&el);
/*  
 * Change to standard Az, Zd convention  
 */
    az = DPI - az;
    zd = DPIBY2 - el;
/*  
 * Observed to Apparent  
 */
    slaOapqk("A",az,zd,MyModel->xypars.obsAoprms,ra,dec);

}

/*
 *+			S i x d f p o s P o s 2 X y 

 * Function name:
      SixdfposPos2Xy

 * Function:
      Convert positioner coorindates to field plate coordinates.  

 * Description:
      Convert the x,y coorindates of a star on the coorindate system 
      used by the 6dF positioner task to the x,y coordinates on the
      field plate as used by SixdfposXy2Rd.

      This scheme of two x/y coorindate systems allows the robot to
      use a coordinate system which makes sense mechanically, whilst
      the values output by SixdfposRd2Xy make more sense from an
      astromical point of view.  This scheme also takes account of
      non-linearities etc. in the positioner system.  

      This routine is called by FpilModelPos2Xy() and must meet
      the prototype given by the type defintion FpilModelPos2XyRoutineType.


 * Language:
      C

 * Call:
      (void) = SixdfposPos2Xy(clientData,  xp, yp, x, y, status)

 * Parameters:   (">" input, "!" modified, "W" workspace, "<" output)
      (>) model   (const FpilModelType *model) Contains the telescope 
                           model details  obtained by SixdfosTransAccess(3).
      (>) xp     (double) Positioenr X coordinate of object (microns).
      (>) yp     (double) Positioenr Y coordinate of object (microns).
      (>) *x      (double) Field plate x coordinate of object (microns).
      (>) *y      (double) Field plate y coordinate of object (microns).
      (!) status  (StatusType *) modified status.

 * Include files: fpi.h

 *-

 * History:
     09-Dec-1999 - TJF - Original version, based on 2dF equivalent.
 */
void SixdfposPos2Xy (
    const FpilModelType *model,
    double xp,
    double yp,
    double *x,
    double *y,
    StatusType *status)
{
    const SixdfposModelPnt MyModel = (const SixdfposModelPnt)model;
    double a,b,c,d,e,f;

    if (*status != STATUS__OK) return;

/*
 *  Extract inversion transformation coefficents from the model details.
 */
    a = MyModel->linpars.invCoeffs[0];
    b = MyModel->linpars.invCoeffs[1];
    c = MyModel->linpars.invCoeffs[2];
    d = MyModel->linpars.invCoeffs[3];
    e = MyModel->linpars.invCoeffs[4];
    f = MyModel->linpars.invCoeffs[5];

/*
 *  Apply the transformation.
 */
   *x = a + b*xp + c*yp;
   *y = d + e*xp + f*yp;


}

/*
 *+			S i x d f p o s F r e e 

 * Function name:
      SixdfposFree

 * Function:
      Free the 6dF positioner transformation model structure.

 * Description:
      Free and tidy up after a call to SixdfposTransAccess(3).  


 * Language:
      C

 * Call:
      (void) = Sixdfpos(model,status)

 * Parameters:   (">" input, "!" modified, "W" workspace, "<" output)
      (>) model   (FpilModelType *model) Contains the telescope model details
                           obtained by SixdfosTransAccess(3).  The 
                           SixdfposCvtInit routine
                           may or may not have been previously invoked
                           on this model.
      (!) status  (StatusType *) modified status.
                        

 * Include files: fpi.h

 *-

 * History:
     09-Dec-1999 - TJF - Original version
 */
extern void SixdfposFree (
    const FpilModelType *model,
    StatusType *status)

{
/*    SixdfposModelPnt MyModel = (SixdfposModelPnt)model;*/

    if (*status != STATUS__OK) return;
    free((void *)model);
}




/*^L
 *+                 S i x d f p o s Q u i c k A p p t o O b s

 * Function name:
      SixdfposQuickAppToObs

 * Function:
      Perfrom a quick apparent to observered place conversion.

 * Description:
      Perfrom a quick apparent to observered place conversion for 
      the 6dF on the UK Schmidt.

 * Language:

      C

 * Call:
        (void) = SixdfposQuickAppToObs(model, ra, dec,
                                        aob, zob, hob, dob, rob,
                                        status)

 * Parameters:   (">" input, "!" modified, "W" workspace, "<" output)
      (>) model   (const FpilModelType *model) Contains the telescope model
                       details obtained by FpilModelAccess(3) and on which
                           FpilModelCvtInit(3) has been invoked.
      (>) ra      (double) Apparent center RA in radians.
      (>) dec     (double) Apparent Center Declination in radians.
      (<) aob     (double *) Observed azimuth (radians, N=0, E = 90degrees)
      (<) zob     (double *) Observed zenith distance (radians)
      (<) hob     (double *) Observed Hour Angle (radians)
      (<) dob     (double *) Observed Declination (radians)
      (<) rob     (double *) Observed Right Ascension (radians)
      (!) status  (StatusType *) modified status.


 * Include files: fpil.h

 * See Also: {references}

 * External functions used:

 * Support: Tony Farrell, AAO

 * History:
     13-Mar-200 - TJF - Original DUMMY version
 */
extern void SixdfposQuickAppToObs  (
    const FpilModelType *model,
    double ra,
    double dec,
    double *aob,
    double *zob,
    double *hob,
    double *dob,
    double *rob,
    FpilStatusType *status)
{
    const SixdfposModelPnt MyModel = (const SixdfposModelPnt)model;
    if (*status != STATUS__OK) return;


    slaAopqk(ra,dec,MyModel->xypars.obsAoprms,aob,zob,hob,dob,rob);
}


/*
 * Internal Internal Function name:
      App2mount

 * Description:
       Convert apparent ra,dec to -ha,dec on the UKST mount system -
       i.e. the UKST polar axis is the pole of the coordinate system.


       Based on Tdf___App2mount in tdfxy.c

 * History:
     20-Jun-2000 - TJF - Original version.
     30-Sep-2002 - TJF - Add cen argument to support selection of which
                         set of aoprms we used (field center or object).

 */
static void App2mount(int cen, SixdfXyType *xypars, double ra, double dec,
                      double *mha, double *mdec, StatusType *status)
{
   double a;                 /* Observed azimuth  */
   double zd;                /* Observed zenith distance  */
   double h;                 /* Observed hour angle  */
   double d;                 /* Observed declination  */
   double r;                 /* Observed right ascension  */
   double e;                 /* Observed elevation */
   double obs[3];            /* Observed Az,El - cartesian vector  */
   double mount[3];          /* Mount frame position - cartesian vector */


   if (*status != STATUS__OK) return;

/*  
 * Apparent to Observed  
 */
   slaAopqk(ra,dec,
            (cen ? xypars->cenAoprms : xypars->obsAoprms),
            &a,&zd,&h,&d,&r);
/*  
 * Check ZD  
 */
   if (zd > 70*DD2R) 
   {
       *status = SIXDF_UTIL__ZDERR;
       return;
   }
/*  
 * Convert to Az,El on Wallace convention (Keck Report section 5.1)  
 */
   a = DPI - a;
   e = DPIBY2 - zd;
/*  
 * Convert to cartesian  
 */
   slaDcs2c(a,e,obs);
/*  
 * Rotate to -HA,Dec  
 */
   slaDmxv(xypars->rot,obs,mount);
/*  
 * Convert to spherical  
 */
   slaDcc2s(mount,mha,mdec);
}

/*
 * Internal Function name:
     Sixdfpos___CloneModel
  
 * Description:
      Mallocs spec for a 6dF model data structure and copies in data from 
      a source model.

  
 * History:
     27-Feb-2012 - TJF - Original version.
 */
extern void Sixdfpos___CloneModel(
    const FpilModelType * modelSource,
    FpilModelType ** modelClone,
    StatusType *status)
{

    SixdfposModelPnt MyModelSource = (SixdfposModelPnt)(modelSource);
    SixdfposModelPnt MyModelClone;

    

    if (*status != STATUS__OK) return;

    *modelClone = 0;
    if ((MyModelClone = malloc(sizeof(SixdfposModel))) == 0)
    {
        *status = SIXDF_UTIL__MALLOCERR;
        return;
    }
    *MyModelClone = *MyModelSource;
    *modelClone = (FpilModelType *)MyModelSource;
    
}
