/*                          s e t _ a l l i n u s e . C
  
 * Program name:
        set_allinuse
  
 * Function:
        Set all fibres in a "constants" file to be in-use.
  
 * Synopsis:
        set_allinuse <contants_file> [<constants_file>...]

 * Description:
        This program will set all the fibres in a "constants" file
        to  be in use.  It can deal with all file formats handled
        by configure (currently 2dF/6dF/OzPoz).

        Note that this program relies on a the SDS library version
        r1_4_3_21 (ACMM Version 3.21) or later.

 * Language:
      C++
  
 * Support: Tony Farrell, AAO
  
 * History:
      15-Apr-2005 - TJF - Original version
      
 *  Copyright (c) Anglo-Australian Telescope Board, 2005.
    Not to be used for commercial purposes without AATB permission.
  
 *  @(#) $Id: ACMM:aaoConfigure/extras/set_allinuse.C,v 12.7 04-Nov-2013 11:28:52+11 tjf $
  
 */


static char *rcsId="@(#) $Id: ACMM:aaoConfigure/extras/set_allinuse.C,v 12.7 04-Nov-2013 11:28:52+11 tjf $";
static void *use_rcsId = ((void)&use_rcsId,(void *) &rcsId);


#define DRAMA_ALLOW_CPP_STDLIB  /* Tell sds.h to allow template stuff */
#include "drama.h"
#include "sds.h"
#include "mess.h"
#include "sds_err_msgt.h"
#include <iostream>
#include <string>
#include <vector>

/*
 *  The processor class is based on SdsId and is used to process
 *  a single configuration file.
 */
class Processor : private SdsId
{
private:
    std::string fileName;     /* Name of the configuration file */
public:
    /*
     * Constructors
     */
    Processor(const char *file,
              StatusType *status);
    /*
     * Process - processes the entire file.
     */
    void Process(StatusType *status);
    /*
     * ProcessField - process the named field.
     */
    void ProcessField(std::string name, StatusType *status);
    /*
     * Write the file back out.
     */
    void Write(StatusType *status);
};


/*
 * Constructor - most work done in the SdsId constructor.
 */
Processor::Processor(
    const char *file,
    StatusType *status) :SdsId(file, status),  fileName(file)
{
    if (*status != STATUS__OK)
    {
        std::cerr << "Failure opening file " << file << std::endl;
    }
   
}
/*
 * Process method - process the entire file.
 */
void Processor::Process(StatusType *status)
{
    char name[SDS_C_NAMELEN] = "";
    if (*status != STATUS__OK) return;
    //std::cout << "Processing file " << fileName << std::endl;
    //List(status);

    /*
     * Get the name of the file's top-level Sds structure 
     */
    GetName(name, status);

    /*
     * A vector of the fields names in this file.
     */
    std::vector<std::string> fieldNames;
    
    
    //std::cout << "Structure name = " << name << std::endl;
    /*
     * For each type of file - put the field names into the fieldNames
     * vector.
     */
    if (strcmp(name, "OzPozInstStat") == 0)
    {
        std::cout << "Processing a OzPoz contants file -" << fileName 
                  << std::endl;
        fieldNames.push_back("Plate1_Data");
        fieldNames.push_back("Plate2_Data");
        // Ignore for now.
        //fieldNames.push_back("Plate3_Data");
        //fieldNames.push_back("Plate4_Data");
    }
    else if (strcmp(name, "sixdFconstants") == 0)
    {
        std::cout << "Processing a 6dF constants file - " << fileName
                  << std::endl;
        fieldNames.push_back("field0");
        fieldNames.push_back("field1");
    }
    else if (strcmp(name, "tdFconstants") == 0)
    {
        std::cout << "Processing an old (404 fibre) 2dF constants file -" 
                  << fileName << std::endl;
        fieldNames.push_back("field0");
        fieldNames.push_back("field1");
    }
    else if (strcmp(name, "tdFconstants400") == 0)
    {
        std::cout << "Processing a 2dF (400 fibre) constants file - " 
                  << fileName << std::endl;
        fieldNames.push_back("field0");
        fieldNames.push_back("field1");
    }
    else
    {
        std::cerr << "Top level file structure name " 
                 << name 
                 << " is invalid, not a 2dF/6dF/OzPoz configuration file"
                 << std::endl;
        *status = SDS__INVSTRUCTDEF;
        return;
    }

    /*
     * Process each field.
     */
    std::vector<std::string>::iterator field;
    for(field = fieldNames.begin(); field != fieldNames.end(); ++field)
    {
        ProcessField(*field,status);
    }
    
    
}

/*
 * Create a name for the new file and write it.
 */
void Processor::Write(StatusType *status)
{
    if (*status != STATUS__OK) return;
    std::string newName = fileName;
    newName += ".new";

    SdsId::Write(newName.c_str(), status);
    if (*status == STATUS__OK)
    {
        std::cerr << "Write new file " << newName << std::endl;
    }
    else
    {
        std::cerr << "Error writing file " << newName << std::endl;
    }
}
/*
 *  Process the named fiel.
 */
void Processor::ProcessField(std::string name, StatusType *status)
{
    if (*status != STATUS__OK) return;
    std::cout << "Processing field named " << name << std::endl;
    
    /*
     * Access the field and then the inUse structure for this field.
     */
    SdsId field(*this, name.c_str(), status);
    SdsId inUse(field, "inUse", status);

    long ndims;
    unsigned long dims[7];
    /*
     * Must have one and only one dimension
     */
    inUse.Dims(&ndims, dims, status);
    if ((*status == STATUS__OK)&&(ndims != 1))
    {
        *status = SDS__INVDIMS;
        std::cerr << "Field \"" << name << "\" in file \"" << fileName
                  << "\" has "   << ndims << " dimensions rather then one "
                  << std::endl;
    }
    
    /*
     * Grab the code (SDS type) since the inUse fields can be either
     * short or unsigned short.
     */
    SdsCodeType code;
    inUse.Code(&code, status);

    if (*status != STATUS__OK)
    {
        /* Bad status - nothing to do here just now */
    }
    else if (code == SDS_SHORT)
    {
        /*
         * Access the array and set all flags to true.
         */
        SdsArrayShort value;
        inUse.ArrayAccess(&value, ndims, dims, status);
        unsigned long count = value.Size();
        if (*status == STATUS__OK)
        {
            for (unsigned long i = 0; i < count ; ++i)
                value[i] = 1;
        }
    }
    else if (code == SDS_USHORT)
    {
        /*
         * Access the array and set all flags to true.
         */
         SdsArrayUshort value;
        inUse.ArrayAccess(&value, ndims, dims, status);
        unsigned long count = value.Size();
        if (*status == STATUS__OK)
        {
            for (unsigned long i = 0; i < count ; ++i)
                value[i] = 1;
        }
    }
    else
    {
        *status = SDS__TYPE;
        std::cerr << "Field \"" << name << "\" in file \"" << fileName
                  << "\" has and unexpected SDS type of "
                  << SdsTypeToStr(code, 1)
                  << std::endl;
        
    }
 


    if (*status != STATUS__OK)
    {
        std::cerr << "Error processing field " 
                  << name
                  << " in file \""
                  << fileName 
                  << "\""
                  << std::endl;
        
    }
}

/*
 * The main program.
 */
int main(int argc, const char *argv[]) 
{
    if (argc <= 1)
    {
        std::cerr << argv[0] 
                  << "::Usage::" 
                  << argv[0] 
                  << " <files...>" 
                  << std::endl;
        return 1;

    }
    /*
     * Allow us to translate error codes
     */
    MessPutFacility(&MessFac_SDS);

    /*
     * Process each file on the command line.
     */
    StatusType status = STATUS__OK;
    for (int i = 1; (i < argc)&&(status == STATUS__OK) ; ++i)
    {
        Processor file(argv[i], &status);
        file.Process(&status);
        if (status != STATUS__OK)
            std::cerr << "Failure processing file " << argv[i] << std::endl;
        file.Write(&status);
    }
    if (status != STATUS__OK)
    {
        char buffer[200];
        MessGetMsg(status, 0, sizeof(buffer), buffer);
        std::cerr << "   Status = " << buffer  << std::endl;
        return 1;
    }
    return 0;
}
