/* -------------------------------------------------------------------------- */

/*               A l l o c a t i o n  S u b - S y s t e m
 *
 *  Function:
 *     Keeps track of allocations and which allocations are valid.
 *
 *  Description:
 *     This module provides a set of routines for use with versions of
 *     the method-specific configuration code that need to do a large
 *     number of allocations and deallocations. A simulated annealing
 *     algorithm would be one such algorithm. These algorithms are 
 *     prepared to trade off a significant setup time against a very
 *     fast test for whether an allocation is possible, given the current
 *     set of allocations.
 *
 *     The key routine is IsAllocAble(), which tests to see if a given
 *     allocation - indentified by an 'allocation index' (a number arbitrarily
 *     assigned between 0 and N-1 where N is the total number of physically
 *     possible allocations) - is possible, given the current set of
 *     allocations. To make this work, all allocations must be recorded
 *     through calls to AllocateIndex() and all deallocations must be
 *     recorded through calls to DeallocateIndex(). AllocationGlobalInit()
 *     must be called once at the start of the program, and AllocationInit()
 *     needs to be called to set up the routines for a new configuration of
 *     pivots and targets. AllocationInit() is the time-consuming routine.
 *
 *  Authors:
 *     Keith Shortridge, AAO.
 *     Brent Miszalski, MQ/AAO.
 *
 *  History:
 *     22nd Jul 2005. Original version. KS/AAO.
 *     23rd Jul 2005. Reversed argument order for AllocationIndex(), added
 *                    AllocationGlobalRelease(). Added temporary code to
 *                    save and restore sparse matrix using a disk file. KS/AAO.
 *     23rd Jul 2005. Added PivotOfTarget() and TargetOfPivot()
 *                    which use auxiliary arrays AllocPivots and AllocTargets
 *                    respectively. The auxiliary arrays are located in 
 *                    CONFIG_ALLOC_Global and are initialised in
 *                    AllocationInit().
 *                    -Swapped parameter order of AllocationIndex().
 *                    -Added check if the target or pivot to be allocated in 
 *                    IsAllocAble() is already allocated in IsAllocable().
 *                    BM.
 *     24th Jul 2005. Some reformatting to remove TABs. Reworked initialisation
 *                    code to use symmetry of the matrix. KS.
 *     27th Jul 2005. Removed use of linked list to hold allocations in favour
 *                    of just using the sorted array directly. This should
 *                    be rather faster. Improved some comments, added not
 *                    on trade-offs. KS.
 *     28th Jul 2005. Added the generic fibre swapping routine Swap(). BM.
 *     29th Jul 2005. Fixed bug in new allocation code - pivot and target
 *                    allocations were not being updated. Moved Swap() into
 *                    the method-specific code. KS.
 *     30th Jul 2005. Moved to a simple flags array to record allocations. KS.
 *     28th Sep 2005. Added some invalid range checking in TargetOfPivot 
 *                    and PivotOfTarget. BM.
 *     25th Feb 2006. Added some code to AllocationInit as a first attempt at
 *                    a reallocation strategy. Also added collision matrix 
 *                    filename to reporting of collision matrix generation. BM.
 *      1st Mar 2006. Saved file now has details of all pivot and target
 *                    positions. This should be a sufficient check to prevent 
 *                    its being used for inappropriate fields. Prior
 *                    allocations are now handled properly (except that
 *                    because the target order can be changed when allocations
 *                    are made, the saved matrix is being recalculated
 *                    unnecessarily in this case). Reallocation following
 *                    a complete deallocation now works, as a consequence of
 *                    the improved saved file validity checking. KS/AAO.
 *      6th Mar 2006. Although previous fix allowed a reallocation of a field,
 *                    it still required a re-calculation of the collision
 *                    matrix. Since only the order of the targets is changed,
 *                    this is now overcome by sorting the target list into a
 *                    fixed order based on target coordinates and using this
 *                    internally, Since this does not change on reallocation,
 *                    the collision matrix no longer needs to be recalculated.
 *                    The downside is an increase in internal complexity, as
 *                    the code now has to juggle two types of target index,
 *                    internal indices (into the sorted target list) and
 *                    external indices (into the target list as passed to the
 *                    sub-system).  It took a while to get this right. I also
 *                    took the liberty of editing the 25th Feb change comments,
 *                    where these discussed a possible reallocation strategy
 *                    that in the end was replaced by this sorted target
 *                    list. (And I reformatted some code to fit into 80
 *                    column lines, and removed some comments from the trade-off
 *                    appendix that had become outdated.) KS/AAO.
 *     15th Jun 2006. MaxAllocations is now a percentage. Fixed problem with 
 *                    value of PossibleAllocations being set incorrectly when
 *                    matrix read from disk (caused a crash). KS.
 *     20th Jun 2006. Re-ordered the sparse sampling code so that it doesn't
 *                    need to recalculate the fibre lists for each target each
 *                    time through the loop. Code no longer gives guide and
 *                    sky fibres immunity from sampling. KS.
 *      2nd Nov 2006. A significant reworking of the way lists of incompatible
 *                    allocations are stored. The single IncompatibleAllocations
 *                    array has been replaced by a set of malloc'ed arrays,
 *                    one for each allocation. This allows the various lists
 *                    to be manipulated independently, and so allows the
 *                    'on-the-fly' mode of operation, as well as the saving
 *                    of individual lists without needing the whole set of
 *                    incompatabilities to be computed first. It also
 *                    opens up the possibility of multi-thread calculation
 *                    of the lists, but that's the next step.. The only
 *                    external change needed was the introduction of
 *                    AllocationFinish(), although some way of controlling
 *                    the 'on-the-fly' option is going to be needed. KS.
 *      8th Nov 2006. Added support for multi-threaded allocation. As many
 *                    background threads as requested are used to fill in
 *                    the incompatible allocation lists, either in on-the-fly
 *                    mode or all at the start, in AllocationInit(). KS.
 *      9th Nov 2006. Fixed bug causing background tasks to exit prematurely
 *                    on occasion. KS.
 *     30th Nov 2006. Get number of processors correctly on MacOsX (using 
 *                    sysctlbyname, sysconf does not work).  Handle other
 *                    cases where _SC_NPROCESSORS_ONLN is not defined.  Allow
 *                    the environment variable CONFIGURE_ONE_CPU_ONLY to turn
 *                    use of multiple CPU's off and output a message to stdout
 *                    indicating the number of processors that were found. TJF.
 *      5th Sep 2007. Calls to pthread_join() were passing the address of the
 *                    pthread_t thread identifier rather than its value.
 *                    Corrected. Note that there are a number of warnings
 *                    to do with the signedness of the array of incompatible
 *                    allocations (see IncompArrayAddr) which is declared as
 *                    unsigned but was treated as signed in places, and
 *                    the design has it holding a -1 flag at the end, which
 *                    surely needs to be fixed. KS.
 *     25th May 2009. Discovered that Mac OS X does not support the sem_init()
 *                    call used to create the semaphore used in multi-threaded
 *                    operation, with the result that errors were being 
 *                    flagged, particularly on Macs that could run a large
 *                    number of threads. The semaphore code has been revised,
 *                    based on the DRAMA GitLoger code, which had solved this
 *                    problem some time ago. KS.
 *     26th May 2009. Modified the debug line giving the time taken for the
 *                    initial setup to show clock-on-wall time as well as
 *                    CPU time. Added additional include files needed to
 *                    compile the modified semaphore code under Linux. KS.
 *      1st Jun 2009. Modified SaveArrays() so it writes a summary text file
 *                    that can be used for regression testing. Considered
 *                    removing AllocArray entirely, but decided to leave it in
 *                    just to keep AllocEntryDetails() efficient. Instead,
 *                    changed the comments so they reflect its actual role in
 *                    the code. KS
 *      4th Jun 2009. Reviewed the use of signed values for index numbers in
 *                    the unsigned IncompArrayAddr arrays. Now compiles, at 
 *                    least under OS X, without warnings. KS.
 *     21st Aug 2012. Was getting errors from the OS X compiler about the use 
 *                    of integers as arguments for the background thread (which
 *                    pthread_create() declares as void* values). Adding 'long'
 *                    casts fixes this, but I suspect this may not be true for
 *                    all compilers. KS.
 */
 
/* -------------------------------------------------------------------------- */

static const char *rcsId="@(#) $Id: ACMM:aaoConfigure/src/AllocSubSystem.c,v 12.7 04-Nov-2013 11:28:43+11 tjf $";
static void *use_rcsId = (0 ? (void *)(&use_rcsId) : (void *) &rcsId);

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <time.h>
#include <semaphore.h>
#include <pthread.h>
#include <unistd.h>
#include <errno.h>
#ifdef __APPLE__
#include <sys/sysctl.h>
#endif

#include "AllocSubSystem.h"

#include "config_err.h"

#ifdef DEBUG
#undef DEBUG
#endif
#define DEBUG printf

/* -------------------------------------------------------------------------- */

/*                              C o n s t a n t s
 *
 *  The lists of incompatible allocation indices held in arrays whose
 *  addresses are held in the IncompArrayAddr elements of the allocation
 *  structures are terminated by an entry of all ones. (At one time the
 *  value used was simply -1, but the arrays are actually declared as
 *  unsigned, so this was not legitimate.) This defines that flag value
 *  properly.
 */
 
#define ALL_ONES ~0

/* -------------------------------------------------------------------------- */

/*                              T y p e s
 */
 
/*  An allocation is defined by a pivot number and a target number */

typedef struct CONFIG_METHOD_Allocation {
   unsigned short Pivot;                  /* Pivot number (from 0) */
   unsigned short IntTarget;              /* Internal target number (from 0) */
   double Theta;                          /* Theta value for this allocation */
   unsigned int Incompatibles;            /* # of incompatible allocations */
   unsigned int* IncompArrayAddr;         /* Address of array of incompatible
                                           * allocations. */
} CONFIG_METHOD_Allocation;

extern void sdprand (int seed);
extern double dprand (void);


/* -------------------------------------------------------------------------- */

/*                              G l o b a l s
 */
 
/*  All the global information needed by the allocation sub-system is held
 *  in this one CONFIG_ALLOC_Global structure.
 *
 *  PossibleAllocations     is the number of physically possible allocations.
 *                          This is derived by looking at the entries in the
 *                          targets/pivots matrix and counting the valid ones.
 *
 *  Allocations             Is an array of structures, one for each possible
 *                          allocation, giving details for that possible
 *                          allocation. (Note, these are not the current 
 *                          allocations, they are the full set of possible
 *                          ones.) Each entry gives the pivot and target index
 *                          for the allocation, and its theta value (all that's
 *                          needed to define the allocation), and the number of
 *                          other possible allocations that are incompatible
 *                          with this allocation. There is, most importantly,
 *                          a component called IncompArrayAddr, which is
 *                          the address of a malloced array containing a 
 *                          list of the index numbers for all the allocations
 *                          that are incompatible with the allocation in
 *                          question, terminated by a flag value of all ones.
 *                          If IncompArrayAddr is null, this list has yet to
 *                          be computed.
 *
 *  IndexLookup             is an array giving the allocation index values
 *                          for each combination of pivot and target. If an
 *                          allocation is not physically possible the array
 *                          holds -1.
 *
 *  AllocFlags              is an array of flags indicating which of the
 *                          possible allocations are currently made.
 *
 *  AllocArraySize          is the number of elements in AllocArray - the number
 *                          of elements for which storage was allocated.
 *
 *  AllocEntries            is the number of allocations currently made.
 *
 *  AllocArray              is an array holding the index numbers of all the
 *                          current allocations, ending with a -1 entry.
 *                          This is now used only as workspace by the routine
 *                          AllocEntryDetails(), which sets it up when it is
 *                          called.
 *
 *  AllocArrayValid         is a flag indicating that AllocArray is up to date.
 *                          This is set by AllocEntryDetails() and is cleared by
 *                          any change to the current allocations.
 *
 *  MethodInfo              is a pointer to the CONFIG_MethodInfo structure
 *                          maintained by the main configuration code. It
 *                          is something used in a number of routines and
 *                          having it here means it doesn't have to be passed
 *                          to routines such as AllocationIndex().
 *
 *  AllocPivots             is an array that holds the companion target index
 *                          for each pivot if the pivot is allocated. 
 *                          If a pivot is not allocated the item value is -1. 
 *                          This is used by TargetOfPivot.
 *             
 *  AllocTargets            is an array that holds the companion pivot index
 *                          for each target if the target is allocated. 
 *                          If a target is not allocated the item value is -1. 
 *                          This is used by PivotOfTarget.
 *
 *  TargetIndex             is an array holding the sorted target index values.
 *                          This allows the index for a target in the original,
 *                          external, target list to be determined quickly from
 *                          its index in the internal target list.
 *                          ExternalIndex = TargetIndex[InternalIndex].
 *
 *  ReverseIndex            is an array reversing the sorted target index values
 *                          This allows the index for a target in the internal,
 *                          sorted, target list to be determined quickly from 
 *                          its index in the external, original, target list.
 *                          InternalIndex = ReverseIndex[ExternalIndex].
 *
 *  SortedTargetInfo        is and array of target info structures that
 *                          is formed by sorting the original target list
 *                          into coordinate order.
 *
 *  Matrix                  is the address of the Matrix argument passed
 *                          to AllocationInit(). Note that we assume this
 *                          remains valid for the whole allocation.
 *
 *  OriginalAllocations     is the number of physically possible allocations
 *                          given the target field. We save this in common
 *                          so that SaveArrays() can write it to the saved
 *                          file as an additional check that the save data is
 *                          valid for the current configuration.
 *             
 *  ActualAllocations       is the number of allocations actually considered,
 *                          after any sparse sampling. Saved for the same
 *                          reason as OriginalAllocations.
 *
 *  MatrixFilename          is the name used for the saved data file as
 *                          passed to AllocationInit().
 *
 *  FAvailAllBase           is an array used to hold flags used to indicate 
 *                          that each possible allocation is actually
 *                          available (as opposed to having been flagged
 *                          out of consideration by the sparse sampling). It
 *                          is global so it can be saved by SaveArrays().
 *
 *  OnTheFly                is a flag set false if all the collision 
 *                          calculations are to be performed during 
 *                          initialisation. If not, they will be performed
 *                          as needed. Setting OnTheFly false makes the
 *                          allocation run much faster at the expense of
 *                          a longer initialisation period.
 *
 *  GlobalSemaphore         controls multi-thread access to crucial bits
 *                          of the global data, mainly the Allocation
 *                          structure array with its lists of incompatible
 *                          allocations. (This is only used if unnamed
 *                          semaphores are supported.)
 *
 *  GlobalSemaphoreName     on systems that don't support unnamed POSIX
 *                          semaphores, this gives the name created for the
 *                          semaphore.
 *
 *  GlobalSemaphoreAddr     the address of the structure describing the
 *                          global semaphore in use. This can be used for
 *                          both named and unnamed semaphores.
 *
 *  AllocationComplete      is set by the main thread to tell any other
 *                          threads to exit.
 *
 *  ThreadCount             is the number of background threads (0,1,2,or 3)
 *                          being used.
 *
 *  MaxThreads              the maximum number of threads supported.
 *
 *  Threads                 is the set of background threads in use. This
 *                          is an array malloced at global initialisation 
 *                          time, when we work out how many processors we
 *                          have available.
 *
 *  IncompListCount         is the number of incompatability lists
 *                          generated so far.
 *
 *  One of the most important points about these globals is shown by the
 *  following crucial sequence. 
 *
 *  o  If we start with a target and pivot number, then we can use the 
 *     IndexLookup array to see if this target can physically
 *     be allocated to the pivot. IndexLookup will have -1 for an allocation
 *     that is physically not possible. If the allocation is possible, it
 *     contains the index number used for the allocation.
 *
 *  o  This index number is an index into the Allocations global array.
 *     The entry for this possible allocation in the Allocations array
 *     gives the number of allocations incompatible with this allocation.
 *     And it gives the address (IncompArrayAddr) of a list of allocations
 *     that are incompatible with this allocation.
 *
 *  o  This is the set of index numbers for all the physically possible 
 *     allocations that are incompatible with this allocation. The list ends
 *     with a flag value of all ones.
 *
 *  o  If any of these index numbers are flagged as already allocated, then
 *     the current set of allocations is not compatible with the allocation 
 *     we're looking at. That pivot cannot be allocated to that target.
 *     If all allocation is currently made, its entry in AllocFlags is set.
 *  
 */
 
static struct CONFIG_ALLOC_GlobalType {
   int PossibleAllocations;              /* Number of possible allocations */
   volatile CONFIG_METHOD_Allocation* Allocations;/* Details of each allocation */
   int* IndexLookup;                     /* Alloc indexes by pivot & target */ 
   int AllocArraySize;                   /* Size allocated for AllocArray */
   int* AllocArray;                      /* Current allocations, sorted */
   int AllocEntries;                     /* Number of current allocations */
   char* AllocFlags;                     /* Current allocations, flagged */
   int AllocArrayValid;                  /* AllocArray is up to date */
   const CONFIG_MethodInfo* MethodInfo;  /* Overall housekeeping info */
   int* AllocPivots;                     /* Flags each pivot as allocated */
   int* AllocTargets;                    /* Flags each target as allocated */
   long* TargetIndex;                    /* Target index array */
   long* ReverseIndex;                   /* The inverse of TargetIndex */
   CONFIG_TargetInfo* SortedTargetInfo;  /* Sorted target information */
   CONFIG_METHOD_MatrixElement **Matrix; /* Matrix of possible allocations */
   int OriginalAllocations;              /* Number of possible allocations */
   int ActualAllocations;                /* Number of allocations considered */
   char MatrixFilename[1024];            /* Name of saved data file */
   int* FAvailAllBase;                   /* Flags used for sparse sampling */
   int OnTheFly;                         /* True for on-the-fly calculation */
   sem_t GlobalSemaphore;                /* Guards this global structure */
   sem_t* GlobalSemaphoreAddr;           /* Address of semaphore structure */
   char GlobalSemaphoreName[1024];       /* Name of the semaphore, if needed */
   volatile int AllocationComplete;      /* Set to tell threads to exit */
   pthread_t* Threads;                   /* Set of background threads */
   int MaxThreads;                       /* Maximum number of threads */
   int ThreadCount;                      /* Number of background threads */
   int IncompListCount;                  /* Number of calculated lists */
} CONFIG_ALLOC_Global;

/* -------------------------------------------------------------------------- */

/*            I n t e r n a l  R o u t i n e  D e c l a r a t i o n s
 *
 *  This is the point to declare any internal routines. They should all
 *  be static.
 */

static void GenerateIncompatList (
   unsigned int I1,
   int* CollisionsFound,
   unsigned int WorkArray[],
   StatusType* Status);

static void SaveArrays (void);

static int AllocBinSearch (              /* (<) Index of match, or -1 */
   const unsigned int Array[],           /* (>) Array to be searched */
   int Size,                             /* (>) Number of integers in Array */
   unsigned int Target);                 /* (>) Value to be searched for */

static void ListAllocArray (void);

static void AllocSortTargets (void);

static void* BackgroundCalc (void* Argument);

static void SemCreate (const CONFIG_MethodInfo *MethodInfo);

static void SemDelete (void);


/* -------------------------------------------------------------------------- */

/*                 A l l o c a t i o n  G l o b a l  I n i t
 *
 *  This routine must be called once at the start of the program. It
 *  does any necessary initialisation of the allocation sub-system.
 */
 
void AllocationGlobalInit (void)
{
   int Nprocessors;               /* Number of available processors */
   
   CONFIG_ALLOC_Global.PossibleAllocations = 0;
   CONFIG_ALLOC_Global.Allocations = NULL;
   CONFIG_ALLOC_Global.IndexLookup = NULL;
   CONFIG_ALLOC_Global.AllocArray = NULL;
   CONFIG_ALLOC_Global.AllocFlags = NULL;
   CONFIG_ALLOC_Global.AllocEntries = 0;
   CONFIG_ALLOC_Global.AllocArraySize = 0;
   CONFIG_ALLOC_Global.AllocPivots = NULL;
   CONFIG_ALLOC_Global.AllocTargets = NULL;
   CONFIG_ALLOC_Global.MethodInfo = NULL;
   CONFIG_ALLOC_Global.AllocArrayValid = CONFIG_FALSE;
   CONFIG_ALLOC_Global.TargetIndex = NULL;
   CONFIG_ALLOC_Global.ReverseIndex = NULL;
   CONFIG_ALLOC_Global.SortedTargetInfo = NULL;
   CONFIG_ALLOC_Global.Matrix = NULL;
   CONFIG_ALLOC_Global.OriginalAllocations = 0;
   CONFIG_ALLOC_Global.ActualAllocations = 0;
   CONFIG_ALLOC_Global.MatrixFilename[0] = '\0';
   CONFIG_ALLOC_Global.GlobalSemaphoreAddr = 
                          &CONFIG_ALLOC_Global.GlobalSemaphore;
   CONFIG_ALLOC_Global.GlobalSemaphoreName[0] = '\0';
   CONFIG_ALLOC_Global.FAvailAllBase = NULL;
   CONFIG_ALLOC_Global.AllocationComplete = CONFIG_FALSE;
   CONFIG_ALLOC_Global.ThreadCount = 0;
   CONFIG_ALLOC_Global.IncompListCount = 0;
   CONFIG_ALLOC_Global.OnTheFly = CONFIG_TRUE;
   
   /*  We need to know how many processors we have available so we can 
    *  set up the Threads array.  The way we do this depends on the
    *  machine.
    */
#ifdef __APPLE__
   {
       int     count ;
       size_t  size=sizeof(count) ;
       
       if (sysctlbyname("hw.ncpu",&count,&size,NULL,0))
           Nprocessors = 1;
       else
           Nprocessors = count; 
   }
#elif !defined(_SC_NPROCESSORS_ONLN)
#  warning "Don't know how to work out the number of processors on this machine"
   Nprocessors = 1;
#else
   Nprocessors = sysconf(_SC_NPROCESSORS_ONLN);
#endif
   if (Nprocessors <= 0) Nprocessors = 1;
   if (getenv("CONFIGURE_ONE_CPU_ONLY"))
   {
       printf("Alloc:Configure will use one CPU out of %d due to env variable\n",
           Nprocessors);
       Nprocessors = 1;
   }
   else if (Nprocessors == 1)
       printf("Alloc:Configure will use 1 CPU\n");
   else
       printf("Alloc:Configure will use %d CPUs\n", Nprocessors);

   CONFIG_ALLOC_Global.Threads = (pthread_t*) malloc (
                              (Nprocessors - 1) * sizeof(pthread_t));
   if (CONFIG_ALLOC_Global.Threads) {
      CONFIG_ALLOC_Global.MaxThreads = Nprocessors - 1;
   } else {
      CONFIG_ALLOC_Global.MaxThreads = 0;
   }

}

/* -------------------------------------------------------------------------- */

/*             A l l o c a t i o n  G l o b a l  R e l e a s e
 *
 *  This routine should be called at the end of the program, to release
 *  any resources allocated by the allocation sub-system. It can also
 *  be called internally to reset the sub-system, for example when a
 *  new field is being configured.
 */
 
void AllocationGlobalRelease (void)
{
   int Index;
   
   /*  The lists of incompatible allocations in the Allocations structure
    *  are themselves malloced arrays, and so need to be released first.
    */
    
   for (Index = 0; Index < CONFIG_ALLOC_Global.PossibleAllocations; Index++) {
      if (CONFIG_ALLOC_Global.Allocations[Index].IncompArrayAddr) {
         free (CONFIG_ALLOC_Global.Allocations[Index].IncompArrayAddr);
      }
   }
   
   /*  Now start to reset the global data. */
   
   CONFIG_ALLOC_Global.PossibleAllocations = 0;
   CONFIG_ALLOC_Global.AllocArraySize = 0;
   CONFIG_ALLOC_Global.AllocEntries = 0;
   CONFIG_ALLOC_Global.AllocationComplete = CONFIG_FALSE;
   
   /*  Free any allocated memory. */
   
   if (CONFIG_ALLOC_Global.Allocations != NULL) {
      free ((void*) CONFIG_ALLOC_Global.Allocations);
      CONFIG_ALLOC_Global.Allocations = NULL;
   }
   if (CONFIG_ALLOC_Global.IndexLookup != NULL) {
      free (CONFIG_ALLOC_Global.IndexLookup);
      CONFIG_ALLOC_Global.IndexLookup = NULL;
   }
   if (CONFIG_ALLOC_Global.AllocArray != NULL) {
      free (CONFIG_ALLOC_Global.AllocArray);
      CONFIG_ALLOC_Global.AllocArray = NULL;
   }
   if (CONFIG_ALLOC_Global.AllocFlags != NULL) {
      free (CONFIG_ALLOC_Global.AllocFlags);
      CONFIG_ALLOC_Global.AllocFlags = NULL;
   }
   if (CONFIG_ALLOC_Global.AllocPivots != NULL) {
      free (CONFIG_ALLOC_Global.AllocPivots);
      CONFIG_ALLOC_Global.AllocPivots = NULL;
   }
   if (CONFIG_ALLOC_Global.AllocTargets != NULL) {
      free (CONFIG_ALLOC_Global.AllocTargets);
      CONFIG_ALLOC_Global.AllocTargets = NULL;
   }
   if (CONFIG_ALLOC_Global.TargetIndex != NULL) {
      free (CONFIG_ALLOC_Global.TargetIndex);
      CONFIG_ALLOC_Global.TargetIndex = NULL;
   }
   if (CONFIG_ALLOC_Global.ReverseIndex != NULL) {
      free (CONFIG_ALLOC_Global.ReverseIndex);
      CONFIG_ALLOC_Global.ReverseIndex = NULL;
   }
   if (CONFIG_ALLOC_Global.SortedTargetInfo != NULL) {
      free (CONFIG_ALLOC_Global.SortedTargetInfo);
      CONFIG_ALLOC_Global.SortedTargetInfo = NULL;
   }
   if (CONFIG_ALLOC_Global.FAvailAllBase != NULL) {
      free (CONFIG_ALLOC_Global.FAvailAllBase);
      CONFIG_ALLOC_Global.FAvailAllBase = NULL;
   }
   
   /*  MethodInfo isn't allocated by the allocation sub-system, so we
    *  don't release it.
    */
    
   CONFIG_ALLOC_Global.MethodInfo = NULL;
   
   /*  Close down the global semaphore, if one is being used */
   
   SemDelete();
}


/* -------------------------------------------------------------------------- */

/*                       A l l o c a t i o n  I n i t
 *
 *  This routine must be called when a new configuration is to be computed.
 *  It performs all necessary field-dependent and pivot-dependent 
 *  initialisation of the allocation sub-system. Its arguments are the
 *  same as those for ConfigMethodInit(), and it is intended to be called
 *  from ConfigMethodInit(). It makes use of the matrix of possible allocations
 *  set up by ConfigMethodInit(), and so should not be called until that
 *  matrix is set up. This routine is performs a lot of calculation as it
 *  sets up the arrays used by the allocation sub-system, particularly the
 *  sparse incompatability matrix that conceptually matches each possible
 *  allocation with every other possible allocation and determines if they
 *  are compatible with each other or not.
 */
 
void AllocationInit (
   CONFIG_METHOD_MatrixElement
                              **Matrix,/* (>) Matrix of possible allocations */
   const CONFIG_MethodInfo *MethodInfo,/* (>) Method details */
   int MaxAllocPercent,                /* (>) Max allocations: 0 => no limit */
   CONFIG_ReportFunction
            ReportFunction,            /* (>) Feedback routine to call */
   void *FunctionArg,                  /* (>) Caller-supplied argument */
   AllocationPrefs* Preferences,       /* (>) Preferences. */
   StatusType *Status)                 /* (!) Inherited status */
{
   /*  Local variables */
   
   int Allocation;           /* Index used to set up IndexLookup array */
   int ActualAllocations;    /* Number of allocations actually considered */
   int Bytes;                /* Size of allocated memory */
   int ComputeSparseMatrix;  /* True if sparse matrix needs to be calculated */
   int IPiv;                 /* Loop index through pivots */
   int ITarget;              /* Loop index through targets */
   int Elements;             /* Elements in an allocated array */
   float ReportPercent;      /* Percentage progress to report at */
   float LastPercentReported;/* Percentage for last progress report */
   int OriginalExcess;       /* Original allocations in excess of max allowed */
   float PercentDone;        /* Percentage used for reporting progress */
   int PossibleAllocations;  /* Number of allocations after sampling */
   int OriginalAllocations;  /* Number of physically possible allocations */
   int I1;                   /* Outer index through possible allocations */
   int TotalCollisions;      /* Total number of collisions detected */
   int MaxCollisions;        /* Maximum collisions for any allocation */
   int Combinations;         /* Count of incompatability entries tested */
   clock_t cpu_start;        /* CPU when sparse matrix computation started */
   clock_t cpu_time;         /* CPU used for matrix computation */
   time_t clock_start;       /* Time when sparse matrix computation started */
   time_t clock_time;        /* Clock on wall time for matrix computation */
   const CONFIG_PivotInfo *PivotInfo;  /* Local pointer to MethodInfo element */
   CONFIG_TargetInfo* SortedTargetInfo;/* Local pointer to global array */
   long* ReverseIndex;       /* Local pointer to global array */
   long* TargetIndex;        /* Local pointer to global array */
   int ReportAllocs;         /* Number of tested allocations per report */
   int MaxPossibleAllocations; /*Maximum Possible Allocations before we have 
                                to start sparse-sampling the number of fibres 
                                available to each per target */
   
   int* NFAvail = NULL;      /* Number of fibres that can reach each target */
   int** FAvail = NULL;      /* Used during sparse sampling */
   int* FAvailBase = NULL;   /* Actual array accessed through FAvail */
   int** FAvailAll = NULL;   /* Flags each allocation as available or not */
   double* W = NULL;         /* Target weights used in sparse sampling */
   FILE* SavedFd = NULL;     /* Used to read the saved incompatability matrix */
   short ConfigFileValid;    /* True if we can use the saved matrix file */
   int seed;                 /* Used to set the random numbers for sparse
                              * sampling. */
   int Index;                /* A general-purpose index variable */
   int RandIndex;            /* Used in sparse sampling code */
   int WMaxIndex;            /* Used in sparse sampling code */
   int ThreadCount;          /* Number of background threads to use */
   int ThreadIndex;          /* Index through background threads */
   int ThreadOffset;         /* Start allocation index for background threads */
   void* Result;             /* Thread completion status - ignored */

   seed = time(NULL);
   sdprand (seed);
   
   /*  Release any global arrays allocated by previous calls. Note that this
    *  nulls out the global MethodInfo address, so we need to do this
    *  call before we set MethodInfo.
    */
    
   AllocationGlobalRelease(); 
   
   /*  Initialise the global semaphore that guards the global data. This
    *  is only needed if we have multiple threads accessing the global 
    *  data. If we don't it will always be available when requested. We
    *  initialise it to 'available'. Because this gets messy if unnamed
    *  semaphores aren't supported, we have a separate routine to do this.
    */
   
   SemCreate(MethodInfo);
      
   strncpy(CONFIG_ALLOC_Global.MatrixFilename,Preferences->MatrixFilename,
                                  sizeof(CONFIG_ALLOC_Global.MatrixFilename));
                                  
   /*  First, we allocate space for a variety of local workspace arrays we
    *  need for help in setting up the allocation sub-system. Most of these
    *  are connected with the sparse sampling of allocations, and are only
    *  actually used if sparse sampling is needed.
    */
   
   /*  FAvailAll is a 2D array of target vs pivot. This would all be a lot 
    *  easier if we allowed ourselves the use of variable size arrays (as in 
    *  C89), and just declated FAvailAll as [Targets][Matrix], but - 
    *  annoyingly - not all compilers support this yet. As it is, we have 
    *  to allocate FAvailAll as an array of indices to another allocated
    *  array (as used by, say Numerical Recipies in C, or indeed by the
    *  Matrix code we use). 
    *
    *  Actually, the above comments don't really apply any more, since
    *  FAvailAllBase turns out to be needed by SaveArrays() and so has
    *  to be held as a global variable.
    */
   
   Bytes = sizeof(int) * MethodInfo->NumberTargets * MethodInfo->NumberPivots;
   CONFIG_ALLOC_Global.FAvailAllBase = (int*) malloc(Bytes);
   if (CONFIG_ALLOC_Global.FAvailAllBase == (int*) NULL) {
       char s[100];
       sprintf(s,"Unable to allocate memory for FAvailAllBase array (%d bytes)",
                                                                      Bytes);
       MethodInfo->ErrRepRoutine(0,s);
       *Status = CONFIG__NO_MEMORY;
       goto Exit;
   }
   Bytes = sizeof(int*) * MethodInfo->NumberTargets;
   FAvailAll = (int**) malloc(Bytes);
   if (FAvailAll == (int**) NULL) {
       char s[100];
       sprintf(s,"Unable to allocate memory for FAvailAll array (%d bytes)",
                                                                      Bytes);
       MethodInfo->ErrRepRoutine(0,s);
       *Status = CONFIG__NO_MEMORY;
       goto Exit;
   }
   for (ITarget = 0; ITarget < MethodInfo->NumberTargets; ITarget++) {
      FAvailAll[ITarget] = CONFIG_ALLOC_Global.FAvailAllBase + 
                                   (ITarget * MethodInfo->NumberPivots);
   }
   
   /*  FAvail is a similarly sized array to FAVailAll, so similar comments
    *  about allocation and indexing apply. This is only used if we do go
    *  into sparse sampling, but we allocate it anyway.
    *
    *  FAvail is a temporary matrix that is updated each time the possible
    *  allocations are reduced. It stores the fibre indices that each target
    *  can be reached by. It exists for the purpose of simple random fibre 
    *  selection (e.g. RandIndex = ...). It may be possible to re-write this 
    *  as some kind of linked-list where once a possible allocation is 
    *  removed from FAvailAll then that corresponding allocation is just
    *  excised from FAvail.
    */
    
   Bytes = sizeof(int) * MethodInfo->NumberTargets * MethodInfo->NumberPivots;
   FAvailBase = (int*) malloc(Bytes);
   if (FAvailBase == (int*) NULL) {
       char s[100];
       sprintf(s,"Unable to allocate memory for FAvailBase array (%d bytes)",
                                                                      Bytes);
       MethodInfo->ErrRepRoutine(0,s);
       *Status = CONFIG__NO_MEMORY;
       goto Exit;
   }
   Bytes = sizeof(int*) * MethodInfo->NumberTargets;
   FAvail = (int**) malloc(Bytes);
   if (FAvail == (int**) NULL) {
       char s[100];
       sprintf(s,"Unable to allocate memory for FAvail array (%d bytes)",
                                                                      Bytes);
       MethodInfo->ErrRepRoutine(0,s);
       *Status = CONFIG__NO_MEMORY;
       goto Exit;
   }
   for (ITarget = 0; ITarget < MethodInfo->NumberTargets; ITarget++) {
      FAvail[ITarget] = FAvailBase + (ITarget * MethodInfo->NumberPivots);
   }
   
   /*  NFAvail is used to record the number of fibres that can reach 
    *  each of the targets.
    */
    
   Bytes = sizeof(int) * MethodInfo->NumberTargets;
   NFAvail = (int*) malloc(Bytes);
   if (NFAvail == (int*) NULL) {
       char s[100];
       sprintf(s,"Unable to allocate memory for NFAvail array (%d bytes)",
                                                                      Bytes);
       MethodInfo->ErrRepRoutine(0,s);
       *Status = CONFIG__NO_MEMORY;
       goto Exit;
   }
   
   /*  W is a weight array used in sparse sampling.
    */
    
   Bytes = sizeof(double) * MethodInfo->NumberTargets;
   W = (double*) malloc(Bytes);
   if (W == (double*) NULL) {
       char s[100];
       sprintf(s,"Unable to allocate memory for W array (%d bytes)",Bytes);
       MethodInfo->ErrRepRoutine(0,s);
       *Status = CONFIG__NO_MEMORY;
       goto Exit;
   }
      
   /*  End of setup of local workspace arrays */
    
   /*  Record the address of the MethodInfo structure in our allocation
    *  sub-system common, since we need it quite a bit.
    */
    
   CONFIG_ALLOC_Global.MethodInfo = MethodInfo;

   /* Initialise the AllocPivots and AllocTargets arrays. */
   
   Bytes = sizeof(int) * MethodInfo->NumberPivots;
   CONFIG_ALLOC_Global.AllocPivots = (int*) malloc(Bytes);
   if (CONFIG_ALLOC_Global.AllocPivots == (int*) NULL) {
       char s[100];
       sprintf(s,
               "Unable to allocate memory for AllocPivots array (%d bytes)",
               Bytes);
       MethodInfo->ErrRepRoutine(0,s);
       *Status = CONFIG__NO_MEMORY;
       goto Exit;
   }
   else {
      /* No allocations yet, updated later on ... BM */
      
      for(IPiv=0;IPiv < MethodInfo->NumberPivots;IPiv++)
      {
         CONFIG_ALLOC_Global.AllocPivots[IPiv] = -1;
      }
   }
   
   Bytes = sizeof(int) * MethodInfo->NumberTargets;
   CONFIG_ALLOC_Global.AllocTargets = (int*) malloc(Bytes);
   if (CONFIG_ALLOC_Global.AllocTargets == (int*) NULL) {
       char s[100];
       sprintf(s,
               "Unable to allocate memory for AllocTargets array (%d bytes)",
               Bytes);
       MethodInfo->ErrRepRoutine(0,s);
       *Status = CONFIG__NO_MEMORY;
       goto Exit;
   }
   else {
      /* No allocations yet, updated later on ... BM */
      for(ITarget=0;ITarget < MethodInfo->NumberTargets;ITarget++)
      {
         CONFIG_ALLOC_Global.AllocTargets[ITarget] = -1;
      }
   }
   
   /*  Set up the arrays needed to handle the sorting of the target list
    *  into ascending coordinate order.
    */
       
   Bytes = sizeof(long) * MethodInfo->NumberTargets;
   CONFIG_ALLOC_Global.TargetIndex = (long*) malloc(Bytes);
   if (CONFIG_ALLOC_Global.TargetIndex == (long*) NULL) {
       char s[100];
       sprintf(s,
               "Unable to allocate memory for TargetIndex array (%d bytes)",
               Bytes);
       MethodInfo->ErrRepRoutine(0,s);
       *Status = CONFIG__NO_MEMORY;
       goto Exit;
   }
   
   Bytes = sizeof(long) * MethodInfo->NumberTargets;
   CONFIG_ALLOC_Global.ReverseIndex = (long*) malloc(Bytes);
   if (CONFIG_ALLOC_Global.ReverseIndex == (long*) NULL) {
       char s[100];
       sprintf(s,
               "Unable to allocate memory for ReverseIndex array (%d bytes)",
               Bytes);
       MethodInfo->ErrRepRoutine(0,s);
       *Status = CONFIG__NO_MEMORY;
       goto Exit;
   }
   
   Bytes = sizeof(CONFIG_TargetInfo) * MethodInfo->NumberTargets;
   CONFIG_ALLOC_Global.SortedTargetInfo = (CONFIG_TargetInfo*) malloc(Bytes);
   if (CONFIG_ALLOC_Global.SortedTargetInfo == (CONFIG_TargetInfo*) NULL) {
       char s[100];
       sprintf(s,
            "Unable to allocate memory for SortedTargetInfo array (%d bytes)",
               Bytes);
       MethodInfo->ErrRepRoutine(0,s);
       *Status = CONFIG__NO_MEMORY;
       goto Exit;
   }
   
   /*  Now sort the target arrays into ascending coordinate order.
    *  From here on the code should use the global SortedTargetInfo 
    *  array rather than MethodInfo->TargetInfo, and needs to allow 
    *  for the differences in internal and external target indices,
    *  which it needs to do by passing target indices from outside
    *  through ReverseIndex and passing internal target indices through
    *  TargetIndex to get the exernal index numbers.
    */
   
   AllocSortTargets();
   SortedTargetInfo = CONFIG_ALLOC_Global.SortedTargetInfo;
   ReverseIndex = CONFIG_ALLOC_Global.ReverseIndex;
   TargetIndex = CONFIG_ALLOC_Global.TargetIndex;
   
   /*  We now do a pass through the matrix, counting the number of possible
    *  allocations. The sizes of almost all the arrays we allocate are
    *  based on this. We assume that the matrix lists all the physically
    *  possible allocations as possible - that is, none are flagged as
    *  'already allocated' - we could check for that as well, and so include
    *  all the physically possible allocations even if some have already
    *  been made, but in fact we leave that until later.
    */
   
   CONFIG_ALLOC_Global.Matrix = Matrix;
   PossibleAllocations = 0;
   for(IPiv=0;IPiv<MethodInfo->NumberPivots;IPiv++){
      for (ITarget = 0; ITarget < MethodInfo->NumberTargets; ITarget++) {
         if (Matrix[IPiv][ITarget].PathStatus == CONFIG_METHOD_POSSIBLE) {
            PossibleAllocations++;
         }
      }
   }
   
   /*  We want to know how many of the possible allocations we are
    *  actually going to work with. If a limit has been set, then we
    *  can reduce the number of allocations we use by sparse sampling.
    *  Note that if this results in a number of allocations less than
    *  those used in a saved matrix file, then we can't use that file.
    */
    
   OriginalAllocations = PossibleAllocations;
   DEBUG ("Before sampling, PossibleAllocations %d\n",PossibleAllocations);   
   if (MaxAllocPercent > 0) {
   
      /*  There is a limit on the number of allocations. Work out
       *  how many we will actually use, after any sparse sampling.
       */
      
      if (MaxAllocPercent > 100) MaxAllocPercent = 100;
      MaxPossibleAllocations = 
              ((double)MaxAllocPercent / 100.0) * PossibleAllocations;
      ActualAllocations = PossibleAllocations;
      if (PossibleAllocations > MaxPossibleAllocations) {
         ActualAllocations = MaxPossibleAllocations;
      }
   } else {
   
      /*  No limit on the number of allocations. Setting this means we don't 
       *  have to go through the sparse sampling loop.
       */
       
      MaxPossibleAllocations = PossibleAllocations;
      ActualAllocations = PossibleAllocations;
   }
   
   /*  Save allocation counts in common for use by SaveArrays(). */
   
   CONFIG_ALLOC_Global.ActualAllocations = ActualAllocations;
   CONFIG_ALLOC_Global.OriginalAllocations = OriginalAllocations;
   
   /*  Once we have the pivot details and the target array sorted into
    *  a standard order, we are able to see if this is the same configuration 
    *  as the one where we last computed the collision matrix. If so,
    *  we don't need to recompute the collision matrix, and we don't
    *  need to go through the process of sparse sampling the allocations
    *  either. In fact, we really don't want to do that, as sampling
    *  the allocations differently will invalidate the collision matrix
    *  just as much as changing the target set.
    *
    *  See if the saved incompatability matrix data is applicable to
    *  the configuration we're dealing with. It is important to make sure
    *  the saved data does match the current set of targets and pivots.
    *  When the matrix is saved, we save the whole set of coordinates 
    *  and types of the pivots and targets as well as the collision matrix.
    *  Now we check that those saved values match those saved. We also
    *  check that the total number of possible allocations matches the
    *  number we've just calculated, as an additional check.
    */

   ComputeSparseMatrix = CONFIG_TRUE;
   ConfigFileValid = CONFIG_FALSE;
   SavedFd = fopen(Preferences->MatrixFilename,"r");
   if (SavedFd) {
      int FileAllocs;
      int NumberPivots,NumberTargets;
      CONFIG_TargetInfo TargetDetails;
      CONFIG_PivotInfo PivotDetails;
      short MisMatch = CONFIG_FALSE;

      /*  Check that the pivot information matches */

      fread(&NumberPivots,sizeof(NumberPivots),1,SavedFd);
      if (NumberPivots != MethodInfo->NumberPivots) {
         printf ("Number of pivots in saved file does not match.\n");
         MisMatch = CONFIG_TRUE;
      } else {
         int IPiv;
         for (IPiv = 0; IPiv < NumberPivots; IPiv++) {
            fread (&PivotDetails,sizeof(PivotDetails),1,SavedFd);
            if ((PivotDetails.X != MethodInfo->PivotInfo[IPiv].X) ||
                (PivotDetails.Y != MethodInfo->PivotInfo[IPiv].Y) ||
                (PivotDetails.FibreType != 
                                MethodInfo->PivotInfo[IPiv].FibreType)) {
               printf ("Pivot details in saved file do not match.\n");
               MisMatch = CONFIG_TRUE;
               break;
            }
         }
      }

      /*  Check that the target information matches */

      if (!MisMatch) {
         fread(&NumberTargets,sizeof(NumberTargets),1,SavedFd);
         if (NumberTargets != MethodInfo->NumberTargets) {
            printf ("Number of targets in saved file does not match.\n");
            MisMatch = CONFIG_TRUE;
         } else {
            int ITarget;
            for (ITarget = 0; ITarget < NumberTargets; ITarget++) {
               fread (&TargetDetails,sizeof(TargetDetails),1,SavedFd);
               if ((TargetDetails.X != SortedTargetInfo[ITarget].X) ||
                   (TargetDetails.Y != SortedTargetInfo[ITarget].Y) ||
                   (TargetDetails.Type != 
                                     SortedTargetInfo[ITarget].Type)) {
                  printf ("Target details in saved file do not match.\n");
                  MisMatch = CONFIG_TRUE;
                  break;
               }
            }
         }
      }
      
      /*  Check that the number of allocations matches - this was the
       *  original test used, but it isn't enough if the order of
       *  targets has changed. We just use it as an additional check.
       */

      if (!MisMatch) {
         fread(&FileAllocs,sizeof(FileAllocs),1,SavedFd);
         if (FileAllocs != OriginalAllocations) {
            printf (
              "Number of possible allocations in saved file does not match.\n");
            MisMatch = CONFIG_TRUE;
         }
      }
      
      /*  We also need to be sure that the actual number of allocations
       *  used when the file was saved matches the limit we are using.
       */
       
      if (!MisMatch) {
         fread(&FileAllocs,sizeof(FileAllocs),1,SavedFd);
         if ((FileAllocs != ActualAllocations)) {
            printf (
              "Number of actual allocations in saved file does not match.\n");
            MisMatch = CONFIG_TRUE;
         }
      }
      if (!MisMatch) {
         ConfigFileValid = CONFIG_TRUE;
      }
      
      /*  If we come out of this section with ConfigFileValid set, then
       *  the saved matrix file is suitable for use. This will let us
       *  read some compute-intensive stuff from the file instead of
       *  recalculating it. If the file is valid, then SavedFd describes
       *  a valid, open, file positioned just after the possible allocation
       *  count.
       */
   }
   
   /*  We use the FAvailAll array to hold a record of which allocations
    *  are available for us to use. Originally, we set this to to show
    *  true for any fibre/target combination that has a Matrix path
    *  status of 'possible'. If we have too many possible allocations and
    *  need to reduce these by sparse sampling, we clear the flags for
    *  those culled allocations in the FAvailAll array, and they will not
    *  be used in the rest of the code - it will be as if they had been
    *  flagged as 'impossible' in the Matrix. (We could simply set their
    *  path status flags to a new incompatible code, but the matrix is
    *  something extermal to the allocation sub-system and so we shouldn't
    *  mess with it.)
    */
   
   for(ITarget=0;ITarget<MethodInfo->NumberTargets;ITarget++) {
      int IntITarget = ITarget;
      int ExtITarget = TargetIndex[IntITarget];
      for(IPiv=0;IPiv<MethodInfo->NumberPivots;IPiv++) {
         FAvailAll[IntITarget][IPiv] = CONFIG_FALSE;
         if(Matrix[IPiv][ExtITarget].PathStatus == CONFIG_METHOD_POSSIBLE) 
            FAvailAll[IntITarget][IPiv] = CONFIG_TRUE;
      }
   }
   
   /*  This is the point where we condiser whether we need to reduce the
    *  overhead of the incompability matrix by in some way pruning the
    *  number of possible allocations. This section of code works through
    *  the possible allocations trying to flag as many as necessary as
    *  'unavailable'. It does this by changing the FAvailAll entry for
    *  the unavailable allocations so that it indicates 'false'.
    *
    *  If the saved file was valid, then we can - and should - bypass 
    *  this step and just read the saved number of possible allocations
    *  (as recorded after sparse selection) and the saved copy of the
    *  FAvailAll array. (Note that what we actually have to read is the
    *  FAvailAllBase array, since that is the real data - FAvailAll is
    *  just an access array for it.)
    */
    
   if (ConfigFileValid) {
      int Elements;
      int Entries;
      Elements = MethodInfo->NumberPivots * MethodInfo->NumberTargets;
      Entries = fread(CONFIG_ALLOC_Global.FAvailAllBase,
                                          sizeof(int),Elements,SavedFd);
      if (Entries != Elements) {
         perror ("Reading error");
         printf ("I/O error reading availability array from file.\n");
         goto Exit;
      }
      PossibleAllocations = ActualAllocations;
      
      /*  So, if ConfigFileValid is set, we now have SavedFd pointing to
       *  the position in the file just after the availability data. We
       *  also have PossibleAllocations set to the saved value, and
       *  FAvailAll filled up with the availability data we calculated
       *  last time.
       */
       
   } else {
   
      /*  Reduce the number of possible allocations by sparse sampling 
       *  the fibres. We use a preliminary measure of 
       *  SUM(Number_of_available_fibres/target_priority) to
       *  give a weight to each target. We then find the max weight
       *  in the field. For each max weight, we randomly select
       *  a fibre that the target can be allocated by and deny
       *  that fibre the ability to allocate the target by 
       *  settings FAvailAll[ITarget][IPiv] to false.
       *  We continue this until the number of possible allocations
       *  is equal to MaxPossibleAllocations.  Ideally, people should not
       *  try to configure fields with too many targets, but this
       *  isn't always so simple, and in any case was not a restriction
       *  with previous versions of Configure.
       *  This aims to drastically reduce the size (and thus
       *  computational overhead/time associated with) of the
       *  collision matrix.
       *
       *  Note that this section contains all the sparse secection logic.
       *  This could be replaced by some other algorithm that modified
       *  the value of PossibleAllocations and sets the values in
       *  FAvailAll accordingly. (Other algorithms may need different
       *  work arrays, in which case the code allocating FAVail, NFAvail,
       *  W, etc, may also need changing.
       */

      OriginalExcess = PossibleAllocations - MaxPossibleAllocations;
      ReportPercent = 2.0;
      LastPercentReported = 0.0;
      WMaxIndex = 0;
      
      /*  This loop fills up the FAvail and NFAvail arrays. NFAvail holds
       *  the number of pivots that can reach each target, and FAvail holds
       *  the actual pivot numbers. As we sample down the possible allocations,
       *  these two arrays get modified.
       */
       
      for(ITarget=0;ITarget<MethodInfo->NumberTargets;ITarget++){
         int KPiv;
         int IntITarget = ITarget;
         W[IntITarget] = 0.0;
         KPiv = 0;
         NFAvail[IntITarget] = 0;
         for(IPiv=0;IPiv<MethodInfo->NumberPivots;IPiv++) {
            FAvail[IntITarget][IPiv] = -1;
            if(FAvailAll[IntITarget][IPiv]) 
            {
               FAvail[IntITarget][KPiv] = IPiv;
               NFAvail[IntITarget] += 1;
               KPiv++;
            }  
         }
      }
      
      /*  This loop does the sparse sampling. As we run through it, we
       *  identify a target that can most afford to loose an allocation
       *  (on the grounds that it is a combination of low priority and
       *  has many fibres that can reach it). Then we randomly remove 
       *  from consideration one of the fibres that can reach it.
       */
       
      while(PossibleAllocations > MaxPossibleAllocations)
      {
         int flag=0;
         if (ReportFunction != NULL) {
            int CurrentExcess = PossibleAllocations - MaxPossibleAllocations;
            PercentDone = 100.0 * 
                          (1.0 - (float)CurrentExcess/(float)OriginalExcess);
            if ((PercentDone - LastPercentReported) > ReportPercent) {
               if ((*ReportFunction)(FunctionArg,'P',
                                     "Sparse sampling allocations",
                                             0,0,0,0,0,0.0,PercentDone) != 0) {
                   MethodInfo->InfoRoutine(MethodInfo->InfoArgument,
                                        "Cancel request received");
                   *Status = CONFIG__CANCELLED;
                   goto Exit;
               }
               LastPercentReported = PercentDone;
            }
         }

         for(ITarget=0;ITarget<MethodInfo->NumberTargets;ITarget++){
            int IntITarget = ITarget;
            
            /*  At this point, it is possible to make sky and/or guide
             *  fibres immune from sampling, but this can leave you with
             *  nothing but those targets, so this has been disabled.
             */
            
            /* 
            if(SortedTargetInfo[IntITarget].Type == 'S' || 
                              SortedTargetInfo[IntITarget].Type == 'F'){
               continue;
            }
            */
            
            /*This should work OK (get the right priority from the 
              target info)?   */
            W[IntITarget] = 
                 NFAvail[IntITarget]*1.0/SortedTargetInfo[IntITarget].Priority;
            /*printf("%d %.2f\n",ITarget,W[ITarget]);*/
            if(flag == 0){
               WMaxIndex = IntITarget;
               flag = 1;
            } else {
               if(W[IntITarget] > W[WMaxIndex]){
                  WMaxIndex = IntITarget;
               }
            }
         }
         
         /*  At this point, WMaxIndex is the index of the target that we
          *  think can most afford to loose one of its allocations.
          */
          
         /*  Sanity check - if no targets remaining have any possible
          *  fibres, we quit now, or the loop will never terminate. (This
          *  can happen if sky and/or guide targets are made immune from 
          *  sampling, since you can find yourself left with just those
          *  targets.)
          */
         
         if (NFAvail[WMaxIndex] <= 0) break;
         
         /*  Randomly select a fibre that can reach the WMaxIndex target
          *  and eliminate that allocation. Modify NFAvail for this index,
          *  and remove the target from the array in FAvail.
          */
         
         RandIndex = (int)(dprand()*1.0*NFAvail[WMaxIndex]+0.5);
         if(FAvailAll[WMaxIndex][FAvail[WMaxIndex][RandIndex]])
         {
            int Index;
            FAvailAll[WMaxIndex][FAvail[WMaxIndex][RandIndex]] = CONFIG_FALSE;
            NFAvail[WMaxIndex]--;
            for (Index = RandIndex; Index < NFAvail[WMaxIndex]; Index++) {
               FAvail[WMaxIndex][Index] = FAvail[WMaxIndex][Index + 1];
            }
            PossibleAllocations--;
         }
      }
   }

   DEBUG("After sampling, PossibleAllocations %d\n",PossibleAllocations);
      
   /*  Now that we know how many possible allocations there are, set up
    *  an array to record their details (just pivot and target numbers at
    *  the moment. We'll add the values connected with the incompatabile
    *  allocations array in a moment.) We also allocate the array used to
    *  provide a quick look-up of allocation index given pivot and
    *  target number. Since the IncompArrayAddr fields record the addresses
    *  of malloced memory, we clear them right now to make sure they are
    *  clearly initialised to show no memory allocated yet.
    */
   
   CONFIG_ALLOC_Global.PossibleAllocations = PossibleAllocations;
   Bytes = sizeof(CONFIG_METHOD_Allocation) * PossibleAllocations;
   CONFIG_ALLOC_Global.Allocations = (CONFIG_METHOD_Allocation*) malloc(Bytes);
   if (CONFIG_ALLOC_Global.Allocations == (CONFIG_METHOD_Allocation*) NULL) {
       char s[100];
       sprintf(s,
               "Unable to allocate memory for allocations array (%d bytes)",
               Bytes);
       MethodInfo->ErrRepRoutine(0,s);
       *Status = CONFIG__NO_MEMORY;
       goto Exit;
   }
   for (Index = 0; Index < PossibleAllocations; Index++) {
      CONFIG_ALLOC_Global.Allocations[Index].IncompArrayAddr = NULL;
   }
   Bytes = sizeof(int) * MethodInfo->NumberTargets * MethodInfo->NumberPivots;
   CONFIG_ALLOC_Global.IndexLookup = (int*) malloc(Bytes);
   if (CONFIG_ALLOC_Global.IndexLookup == (int*) NULL) {
       char s[100];
       sprintf(s,
               "Unable to allocate memory for index lookup array (%d bytes)",
               Bytes);
       MethodInfo->ErrRepRoutine(0,s);
       *Status = CONFIG__NO_MEMORY;
       goto Exit;
   }

   /*  The fast lookup array that records which allocations have
    *  been made so far. (This could be a flag in Allocations, but I
    *  suspect it's faster to look up in a simple char array.)
    */

   Bytes = sizeof(char) * PossibleAllocations;
   CONFIG_ALLOC_Global.AllocFlags = (char*) malloc(Bytes);
   if (CONFIG_ALLOC_Global.AllocFlags == (char*) NULL) {
       char s[100];
       sprintf(s,
               "Unable to allocate memory for allocated flags array (%d bytes)",
               Bytes);
       MethodInfo->ErrRepRoutine(0,s);
       *Status = CONFIG__NO_MEMORY;
       goto Exit;
   }
   for (Allocation = 0; Allocation < PossibleAllocations; Allocation++) {
      CONFIG_ALLOC_Global.AllocFlags[Allocation] = 0;
   }

   /*  And now set up the allocations and index lookup array. Remember that
    *  the matrix was set up externally, and so uses external target indices.
    */
   
   Allocation = 0;
   for (IPiv = 0; IPiv < MethodInfo->NumberPivots; IPiv++) {
      for (ITarget = 0; ITarget < MethodInfo->NumberTargets; ITarget++) {
         int IntITarget = ITarget;
         int ExtITarget = TargetIndex[IntITarget];
         int LIndex = (IntITarget * MethodInfo->NumberPivots) + IPiv;
         CONFIG_ALLOC_Global.IndexLookup[LIndex] = -1;
         /* Only use the allocations that have survived sparse sampling */
         if (FAvailAll[IntITarget][IPiv]) {
            CONFIG_ALLOC_Global.Allocations[Allocation].IntTarget = IntITarget;
            CONFIG_ALLOC_Global.Allocations[Allocation].Pivot = IPiv;
            CONFIG_ALLOC_Global.Allocations[Allocation].Theta = 
                                              Matrix[IPiv][ExtITarget].Theta;
            CONFIG_ALLOC_Global.Allocations[Allocation].Incompatibles = 0;
            CONFIG_ALLOC_Global.Allocations[Allocation].IncompArrayAddr = NULL;
            CONFIG_ALLOC_Global.IndexLookup[LIndex] = Allocation;
            Allocation++;
         }
      }
   }
     
      
   /*  We also allocate an array to hold the allocated index numbers in 
    *  ascending order. We allow space for a -1 index to signal the end 
    *  of the list. This array used to be used during the allocation process
    *  to test if an allocation had been made, and so needed to be kept up
    *  to date as the code was run. Having it in ascending order made it
    *  reasonably quick to search it to find if an allocation had been made.
    *  However, this role is now taken by the AllocFlags array, which is much
    *  more straightforward and far faster. This array is now only really used
    *  in the final stages of an allocation, and is now just workspace
    *  used to make AllocEntryDetails() more efficient. AllocEntryDetails()
    *  fills up this array quickly when it is called and finds it no longer
    *  valid, and then can use it for subsequent calls. This is efficient if
    *  what happens is just a series of calls to AllocEntryDetails() for all
    *  the allocations right at the end of the allocation process, with no
    *  intervening changes to the allocation.
    */

   Elements = MethodInfo->NumberPivots + 1;
   Bytes = sizeof(int) * Elements;
   CONFIG_ALLOC_Global.AllocArray = (int*) malloc(Bytes);
   if (CONFIG_ALLOC_Global.AllocArray == (int*) NULL) {
       char s[100];
       sprintf(s,
           "Unable to allocate memory for ordered allocations array (%d bytes)",
               Bytes);
       MethodInfo->ErrRepRoutine(0,s);
       *Status = CONFIG__NO_MEMORY;
       goto Exit;
   }
   
   /*  Initialise the array. We set it as invalid, and AllocEntryDetails()
    *  can fill it up as needed.
    */
   
   CONFIG_ALLOC_Global.AllocEntries = 0;
   CONFIG_ALLOC_Global.AllocArraySize = Elements;
   CONFIG_ALLOC_Global.AllocArray[0] = -1;
   CONFIG_ALLOC_Global.AllocArrayValid = CONFIG_FALSE;
      
   /*  If we had a compatible saved matrix file, we don't need to compute
    *  the incompatability data. Instead we read it it, allocation by
    *  allocation. See comments in SaveArrays() for the format used.
    *  Basically, there is an allocation index (with -1 indicating the
    *  end of the data), then a collision count, then the list of
    *  allocations incompatible with the specified allocation index.
    */
             
   CONFIG_ALLOC_Global.IncompListCount = 0;
   if (ConfigFileValid) {
      int FileIndex;
      (void) fread(&FileIndex,sizeof(FileIndex),1,SavedFd);
      while (FileIndex >= 0) {
      
         /*  Index number is not -1, so there is more data. Read 
          *  number of collisions, allocate space for them (including the
          *  -1 at the end of the list) and read them in.
          */
         int Collisions;
         unsigned int* ListPtr;
         fread(&Collisions,sizeof(Collisions),1,SavedFd);
         ListPtr = 
            (unsigned int*) malloc ((Collisions + 1) * sizeof(unsigned int));
         if (ListPtr == (unsigned int*) 0) {
            printf (
              "Unable to allocate %d entries for index %d collision list\n",
                                                      Collisions,FileIndex);
         } else {
            int Entries = 
                 fread(ListPtr,sizeof(unsigned int),Collisions + 1,SavedFd);
            if (Entries != (Collisions + 1)) {
               printf ("Error (%d) reading %d entries from matrix file\n",
                                                   Entries,Collisions + 1);
            } else {
               CONFIG_ALLOC_Global.Allocations[FileIndex].IncompArrayAddr =
                                                                       ListPtr;
               CONFIG_ALLOC_Global.Allocations[FileIndex].Incompatibles =
                                                           Collisions;
               CONFIG_ALLOC_Global.IncompListCount++;
            }
         }
         
         /*  Then read the next index number. */
         
         (void) fread(&FileIndex,sizeof(FileIndex),1,SavedFd);
      }
      fclose(SavedFd);
      SavedFd = (FILE*) NULL;
   
   }
   
   /*  Note that even if we've read in a saved copy of the incompatability
    *  data, it may not be exhaustive. If we are operating in an on-the-fly
    *  mode, all we have to do now is start up the background threads (if
    *  we're using them) to give them a chance to fill it up. If the
    *  data is complete, the threads will just complete almost at once.
    *  If we are forcing a pre-computation of the incompatability
    *  data, we still run the background threads, but we run the same
    *  calculation ourselves in the foreground and wait for the whole
    *  data to be computed.
    */
    
   /*  Either way, we start any background threads going now. Each thread
    *  starts calculating the incompatability lists for allocations starting
    *  at a specified index. This index is split evenly between the threads,
    *  so they start off separated sensibly, and after that they're on 
    *  their own.
    */
   
   ThreadCount = Preferences->MaxThreads;
   if (ThreadCount < 0) ThreadCount = 0;
   if (ThreadCount > CONFIG_ALLOC_Global.MaxThreads) {
      ThreadCount = CONFIG_ALLOC_Global.MaxThreads;
   }
   CONFIG_ALLOC_Global.ThreadCount = ThreadCount;
   ThreadOffset = 0;
   if (ThreadCount > 0) {
      ThreadOffset = PossibleAllocations / ThreadCount;
   }
   for (ThreadIndex = 0; ThreadIndex < ThreadCount; ThreadIndex++) {
      void* Argument = (void*)((long)(ThreadOffset * ThreadIndex));
      pthread_create (&(CONFIG_ALLOC_Global.Threads[ThreadIndex]),(void*)0,
                                                  BackgroundCalc,Argument);
   } 
   
   /*  We also see if we are to complete all incompatability calculations
    *  before starting, or to calculate them on the fly. (There doesn't
    *  seem much advantage to not using 'on the fly', but this is how
    *  the system used to work, so it remains an option.)
    */
      
   CONFIG_ALLOC_Global.OnTheFly = Preferences->OnTheFly;    
   if (!CONFIG_ALLOC_Global.OnTheFly) {
       
      /*  We are going to make sure the incompatability data is all
       *  calculated before we start.
       */
      char repbuf[80];
      sprintf(repbuf,"Generating collision matrix (%s)",
                                       Preferences->MatrixFilename);
      TotalCollisions = 0;
      MaxCollisions = 0;
      Combinations = 0;
      cpu_start = clock();
      cpu_time = 0;
      clock_start = time(NULL);
      clock_time = 0;
      PivotInfo = MethodInfo->PivotInfo;
      ReportAllocs = PossibleAllocations / 100;
      for (I1 = PossibleAllocations - 1; I1 >= 0; I1--) {
         int NewCollisions;

         /*  For each possible allocation, we use GenerateIncompatList()
          *  to produce the full list of incompatible allocations for
          *  that allocation. We do it backwards, starting at the highest
          *  numbered, in order to take advantage of the symmetry of the 
          *  process - see comments in GenerateIncompatList().
          */

         if (ReportFunction != NULL) {
            if ((I1 % ReportAllocs) == 0) {
               PercentDone = 
                   ((float) CONFIG_ALLOC_Global.IncompListCount / 
                                   (float) PossibleAllocations) * 100.0;
               if ((*ReportFunction)(FunctionArg,'P',
                       repbuf,0,0,0,0,0,0.0,PercentDone) != 0) {

                   /*  We respond to a cancel request by aborting the initial
                    *  pre-calculation, but for the moment we leave the
                    *  background threads (if any) running.
                    */

                   MethodInfo->InfoRoutine(MethodInfo->InfoArgument,
                                        "Cancel request received");
                   *Status = CONFIG__CANCELLED;
                   goto Exit;
               }
            }
         }
         if (CONFIG_ALLOC_Global.Allocations[I1].IncompArrayAddr ==
                                                     (unsigned int*) 0) {
            GenerateIncompatList(I1,&NewCollisions,0,Status);
            if (*Status != 0) goto Exit;
         }
      }
      
      /*  At this point we should wait for any background tasks to complete,
       *  which they should pretty much immediately.
       */
      
      DEBUG ("Waiting for threads\n");
      for (ThreadIndex = 0; ThreadIndex < CONFIG_ALLOC_Global.ThreadCount;
                                                               ThreadIndex++) {
         pthread_join((CONFIG_ALLOC_Global.Threads[ThreadIndex]),&Result);
         DEBUG ("Thread %d completes\n",ThreadIndex + 1);
      }

      /*  As the collision data is now known completely, this is
       *  a good place to save it.
       */

      SaveArrays ();
      cpu_time = clock() - cpu_start;
      clock_time = time(NULL) - clock_start;
      DEBUG ("Initial setup took %d seconds (%d CPU sec)\n",(int)clock_time,
                                      (int)(cpu_time / CLOCKS_PER_SEC));

   }
   
   /*  Finally, run through all the allocated pivots and record the
    *  allocations. The AllocInfo target index is external (not sorted)
    *  and this is what AllocationIndex() expects.
    */
   
   for (IPiv = 0; IPiv < MethodInfo->NumberPivots; IPiv++) {
      if (MethodInfo->AllocInfo[IPiv].State == 'A') {
         int Index;
         int ExtITarget; 
         ExtITarget = MethodInfo->AllocInfo[IPiv].TargetIndex;
         Index = AllocationIndex (IPiv,ExtITarget);
         if (Index >= 0) AllocateIndex(Index);
      }
   }
   
   /*  This is only here to prevent a compiler warning about a diagnostic
    *  routine that isn't being used at the moment. Clearly, it doesn't
    *  actually call the routine.
    */
    
   if (0) ListAllocArray();
   
Exit:;

   /*  Release any arrays we needed as workspace */
   
   if (FAvailAll) free(FAvailAll);
   if (FAvail) free(FAvail);
   if (FAvailBase) free(FAvailBase);
   if (NFAvail) free(NFAvail);
   if (W) free(W);
   
   /*  Close down the matrix file if it's still open. (This will be
    *  the result of an error.)
    */
   
   if (SavedFd) fclose(SavedFd);
}

/* -------------------------------------------------------------------------- */

/*                    A l l o c a t i o n  F i n i s h
 *
 *  This routine must be called when a new configuration has been computed.
 *  It provides a place to finish things up for a given allocation, and
 *  forms a logical pair with AllocationInit(). For example, if collision
 *  data has been calculated on the fly rather than during AllocationInit(),
 *  there needs to be a place such as this to write out the saved data file.
 */
 
void AllocationFinish (
   StatusType *Status)                 /* (!) Inherited status */
{
   void* Result;                       /* Return code from thread - ignored */
   int ThreadIndex;                    /* Index through background threads */
   
   /*  Set the allocation complete flag. If any threads are still 
    *  running, they're supposed to be testing this and stopping
    *  when they see it set. So then the threads should complete,
    *  and we wait for them to do so. We should only need to wait like
    *  this if we're doing on the fly calculation of allocations.
    */
    
   CONFIG_ALLOC_Global.AllocationComplete = CONFIG_TRUE;
   if (CONFIG_ALLOC_Global.OnTheFly) {
      DEBUG ("Waiting for threads\n");
      for (ThreadIndex = 0; ThreadIndex < CONFIG_ALLOC_Global.ThreadCount;
                                                               ThreadIndex++) {
         pthread_join((CONFIG_ALLOC_Global.Threads[ThreadIndex]),&Result);
         DEBUG ("Thread %d completes\n",ThreadIndex + 1);
      }
   }
   
   /*  If we were doing OnTheFly calculation of incompatible allocations,
    *  now is the time to write the saved data. (Otherwise, it will have
    *  been saved in AllocationInit(), as soon as it was calculated.)
    */
    
   if (CONFIG_ALLOC_Global.OnTheFly) {
      SaveArrays();
   }
   
}

/* -------------------------------------------------------------------------- */

/*                        S a v e  A r r a y s
 *
 *  An internal routine of the allocation sub-system. This saves the
 *  arrays already calculated for the configuration. This is really
 *  intended to save the incompatable allocation arrays that have
 *  been calculated, so they don't need working out again.
 *
 *  In order to run regression tests on this code, to make sure that
 *  we are getting the same results even after the code is changed, we
 *  would ideally compare the collision matrix file that gets created with
 *  one created by an earlier run. However, this is too large to be saved
 *  with the code files, so instead we write a summary text file that
 *  should contain enough data that we can be reasonably sure we have the
 *  same results. We don't record the actual collision matrix, but we do
 *  record the individual checksums for the different possible allocations.
 */
 
void SaveArrays (void)
{
   FILE* SavedFd = NULL;   /* Used to write the saved incompatability matrix */
   FILE* ListFd = NULL;    /* Used to write the text file used for testing */
   char TextFile[1024];    /* Name of the text file */
   int NameLen;            /* Characters in TextFile */
   
   /*  We open the file to be used for the actual matrix, and also the text
    *  file used for regression testing.
    */

   SavedFd = fopen(CONFIG_ALLOC_Global.MatrixFilename,"w");
   
   strncpy(TextFile,CONFIG_ALLOC_Global.MatrixFilename,sizeof(TextFile));
   TextFile[sizeof(TextFile) - 1] = '\0';
   NameLen = strlen(TextFile);
   strncat(TextFile,".lis",sizeof(TextFile) - NameLen - 1);
   ListFd = fopen(TextFile,"w");
   if (ListFd) {
      fprintf (ListFd,
        "\nOnly used for regression testing of the allocation sub-system\n\n");
   }
   
   if (SavedFd) {
      int NumberPivots,NumberTargets;
      int IPiv,ITarget;
      int Entries,Elements;
      int I1;
      NumberPivots = CONFIG_ALLOC_Global.MethodInfo->NumberPivots;
      NumberTargets = CONFIG_ALLOC_Global.MethodInfo->NumberTargets;

      /*  Save the pivot details at the start of the file */

      if (ListFd) (void) fprintf (ListFd,"Number pivots = %d\n",NumberPivots);
      (void) fwrite(&NumberPivots,sizeof(NumberPivots),1,SavedFd);
      for (IPiv = 0; IPiv < NumberPivots; IPiv++) {
         (void) fwrite(&(CONFIG_ALLOC_Global.MethodInfo->PivotInfo[IPiv]),
            sizeof(CONFIG_ALLOC_Global.MethodInfo->PivotInfo[IPiv]),1,SavedFd);
      }

      /*  Then save the target details */

      if (ListFd) (void) fprintf (ListFd,"Number targets = %d\n",NumberTargets);
      (void) fwrite(&NumberTargets,sizeof(NumberTargets),1,SavedFd);
      for (ITarget = 0; ITarget < NumberTargets; ITarget++) {
         (void) fwrite(&(CONFIG_ALLOC_Global.SortedTargetInfo[ITarget]),
            sizeof(CONFIG_ALLOC_Global.SortedTargetInfo[ITarget]),1,SavedFd);
      }

      /*  Write the number of original allocations possible, followed
       *  by the number actually used, then the contents of the FAvailAll 
       *  array (taken from its base array, which contains the actual 
       *  data we need)
       */

      (void) fwrite(&CONFIG_ALLOC_Global.OriginalAllocations,
                    sizeof(CONFIG_ALLOC_Global.OriginalAllocations),1,SavedFd);
      (void) fwrite(&CONFIG_ALLOC_Global.ActualAllocations,
                      sizeof(CONFIG_ALLOC_Global.ActualAllocations),1,SavedFd);
      Elements = CONFIG_ALLOC_Global.MethodInfo->NumberPivots * 
                              CONFIG_ALLOC_Global.MethodInfo->NumberTargets;
      Entries = fwrite(CONFIG_ALLOC_Global.FAvailAllBase,sizeof(int),
                                                      Elements,SavedFd);
      if (Entries != Elements) {
         printf ("Error (%d) writing FAvailAll array of %d entries\n",
                                              Entries,Elements);
      }
      if (ListFd) {
         (void) fprintf (ListFd,"Original allocations = %d\n",
                                 CONFIG_ALLOC_Global.OriginalAllocations);
         (void) fprintf (ListFd,"Actual allocations   = %d\n",
                                 CONFIG_ALLOC_Global.ActualAllocations);
         (void) fprintf (ListFd,"Possible allocations = %d\n",
                                 CONFIG_ALLOC_Global.PossibleAllocations);
         (void) fprintf (ListFd,"Checksums for each possible allocation:\n");
      }                       

      /*  And now, for each of the allocations for which incompatability
       *  lists have been calculated, write the lists into the file. In
       *  each case we write the index of the allocation, the number
       *  of incompatible allocations, and then the list of
       *  incompatable allocations, including the terminating all ones flag.
       *  Finally, we write a allocation number of -1.
       */

      for (I1 = 0; I1 < CONFIG_ALLOC_Global.PossibleAllocations; I1++) {
         int Collisions;
         if (CONFIG_ALLOC_Global.Allocations[I1].IncompArrayAddr) {
            (void) fwrite (&I1,sizeof(I1),1,SavedFd);
            Collisions = CONFIG_ALLOC_Global.Allocations[I1].Incompatibles;
            (void) fwrite (&Collisions,sizeof(Collisions),1,SavedFd);
            Entries = fwrite (
                CONFIG_ALLOC_Global.Allocations[I1].IncompArrayAddr,
                                sizeof(unsigned int),Collisions + 1,SavedFd);
            if (Entries != (Collisions + 1)) {
               printf ("Error (%d) writing array of %d entries\n",
                                              Entries,Collisions + 1);
               break;
            }
            
            /*  For the regression test text file, we write just a simple
             *  exclusive OR checksum for each collision array.
             */
             
            if (ListFd) {
               unsigned int Checksum = 0;
               int Index;
               for (Index = 0; Index < Collisions; Index++) {
                  Checksum ^= 
                     CONFIG_ALLOC_Global.Allocations[I1].IncompArrayAddr[Index];
               }
               (void) fprintf (ListFd,"%d\n",Checksum);
            }
         }
      }
      I1 = -1;
      (void) fwrite (&I1,sizeof(I1),1,SavedFd);
      fclose(SavedFd);
      SavedFd = (FILE*) NULL;
   } else {
      printf ("Error creating saved matrix file\n");
   }
   if (ListFd) {
      fclose (ListFd);
   } else {
      printf ("Error creating regression test text file\n");
   }
}

/* -------------------------------------------------------------------------- */

/*                 G e n e r a t e  I n c o m p a t  L i s t
 *
 *  This routine generates the list of incompatable allocations
 *  for a specified allocation. I1 gives the index number of the
 *  allocation in question. WorkArray should be an int array of
 *  dimensionns at least CONFIG_ALLOC_Global.PossibleAllocations + 1
 *  (the +1 allows for the terminating all ones flag appended at the end of
 *  an incompatible allocation list). If WorkArray is passed as null,
 *  this routine malloc's a suitable array and releases it when
 *  finished - but this is a bit inefficient.
 */
 
void GenerateIncompatList (
   unsigned int I1,
   int* CollisionsFound,
   unsigned int WorkArray[],
   StatusType* Status)
{
   int IntTarget1 = CONFIG_ALLOC_Global.Allocations[I1].IntTarget;
   int ExtTarget1 = CONFIG_ALLOC_Global.TargetIndex[IntTarget1];
   int Pivot1 = CONFIG_ALLOC_Global.Allocations[I1].Pivot;
   int Target1X = CONFIG_ALLOC_Global.SortedTargetInfo[IntTarget1].X;
   int Target1Y = CONFIG_ALLOC_Global.SortedTargetInfo[IntTarget1].Y;
   int Pivot1X = CONFIG_ALLOC_Global.MethodInfo->PivotInfo[Pivot1].X;
   int Pivot1Y = CONFIG_ALLOC_Global.MethodInfo->PivotInfo[Pivot1].Y;
   double Theta1 = CONFIG_ALLOC_Global.Matrix[Pivot1][ExtTarget1].Theta;
   int Collisions = 0;
   int I2;
   int Bytes;
   int FreeWorkArray = CONFIG_FALSE;
   
   if (WorkArray == (unsigned int*) 0) {
      Bytes = (CONFIG_ALLOC_Global.PossibleAllocations + 1) * 
                                                       sizeof(unsigned int);
      WorkArray = (unsigned int*) malloc (Bytes);
      if (WorkArray == (unsigned int*) 0) {
         *Status = CONFIG__NO_MEMORY;
         goto Exit;
      }
      FreeWorkArray = CONFIG_TRUE;
   }
   for (I2 = 0; I2 < CONFIG_ALLOC_Global.PossibleAllocations; I2++) {

      /*  Because the matrix is diagonally symmetrical, we may
       *  already know the answer. If I2 == I1, then there is
       *  a collision - an allocation has to be incompatible with
       *  itself. In all cases, if I1 is in I2's list, then I2 is
       *  in I1's list. If we have already calculated I2's list, then
       *  we can look up the result in it. Note that when this is
       *  called with the matrix being built up completely before
       *  any allocation begins, then we fill up with I1 starting at
       *  its maximum value and working down, which means that we will
       *  always have the value already if I2 > I1.
       */

      CONFIG_Boolean Collision = CONFIG_FALSE;

      if (I2 == I1) {
         Collision = CONFIG_TRUE;
         
      } else if (CONFIG_ALLOC_Global.Allocations[I2].IncompArrayAddr) {

         /*  In this case, we have already worked out the full list
          *  of incompatible allocations for I2 (since there is a non-null
          *  address for the array holding its list). So the 
          *  question is simply, does the list of incompatible
          *  allocations for I2 include I1? 
          */

         int I2Size = CONFIG_ALLOC_Global.Allocations[I2].Incompatibles;
         unsigned int* I2Array = 
                    CONFIG_ALLOC_Global.Allocations[I2].IncompArrayAddr;
         int Entry = AllocBinSearch(I2Array,I2Size,I1);
         if ((Entry >= 0) && (I2Array[Entry] == I1)) {
            Collision = CONFIG_TRUE;
         }
      } else {


         /*  We don't have the list for I2 yet, so we have to calculate 
          *  for ourselves whether allocations I1 and I2 are mutually 
          *  incompatible.
          */

         int IntTarget2 = CONFIG_ALLOC_Global.Allocations[I2].IntTarget;
         int ExtTarget2 = CONFIG_ALLOC_Global.TargetIndex[IntTarget2];
         int Pivot2 = CONFIG_ALLOC_Global.Allocations[I2].Pivot;
         int Target2X = CONFIG_ALLOC_Global.SortedTargetInfo[IntTarget2].X;
         int Target2Y = CONFIG_ALLOC_Global.SortedTargetInfo[IntTarget2].Y;
         int Pivot2X = CONFIG_ALLOC_Global.MethodInfo->PivotInfo[Pivot2].X;
         int Pivot2Y = CONFIG_ALLOC_Global.MethodInfo->PivotInfo[Pivot2].Y;
         double Theta2 = CONFIG_ALLOC_Global.Matrix[Pivot2][ExtTarget2].Theta;
         Collision = CONFIG_ALLOC_Global.MethodInfo->Collision(
                                   CONFIG_ALLOC_Global.MethodInfo, 
                                   Target1X, Target1Y,
                                   Theta1,
                                   Pivot1X, Pivot1Y,
                                   Target2X, Target2Y, 
                                   Theta2, 
                                   Pivot2X, Pivot2Y);            
      }

      if (Collision) {
      
         /*  There has been a collision, so record it in the work array */
         
         WorkArray[Collisions++] = I2;
      }
   }

   /*  Terminate the incompatability list for this allocation with
    *  a -1 flag. Record the number of incompatible entries for this 
    *  allocation in the allocations array.
    */

   WorkArray[Collisions] = ALL_ONES;
   
   /*  Now that we know how long the work array is, we can allocate 
    *  space for it and copy it into that space, recording the location
    *  in the Allocations structure. We make sure we get exclusive access
    *  to the Allocations structure before we write into it.
    */
   
   sem_wait (CONFIG_ALLOC_Global.GlobalSemaphoreAddr);
   Bytes = (Collisions + 1) * sizeof(unsigned int);
   if (CONFIG_ALLOC_Global.Allocations[I1].IncompArrayAddr) {
      
      /*  If while we were calculating the incompatible list, the
       *  list has appeared, then another thread must have beaten us 
       *  to it. That's OK, but it means we've wasted a bit of time.
       */
       
      /* DEBUG ("Concurrent calculation of list for index %d\n",I1); */
      if (CONFIG_ALLOC_Global.Allocations[I1].Incompatibles != Collisions) {
         DEBUG ("And they got a different answer! (%d, %d) That's serious..\n",
              CONFIG_ALLOC_Global.Allocations[I1].Incompatibles,Collisions);
      }
   } else {
   
     /*  Note that we set CONFIG_ALLOC_Global.Allocations[I1].IncompArrayAddr
      *  as the last step in this process, once the array has been filled
      *  properly. This is because the code earlier in this routine that checks
      *  to see if the already calculated incompatible allocations include
      *  the one we're looking for might otherwise end up testing an array
      *  that hasn't yet been filled.
      */
      
      unsigned int* IncompArrayAddr = (unsigned int*) malloc(Bytes);
      if (IncompArrayAddr) {
         memcpy (IncompArrayAddr,WorkArray,Bytes);
      } else {
         char s[100];
         sprintf(s,"Unable to allocate entry %d array (%d bytes)",I1,Bytes);
         CONFIG_ALLOC_Global.MethodInfo->ErrRepRoutine(0,s);
         *Status = CONFIG__NO_MEMORY;
      }
      CONFIG_ALLOC_Global.Allocations[I1].Incompatibles = Collisions;
      CONFIG_ALLOC_Global.IncompListCount++;
      CONFIG_ALLOC_Global.Allocations[I1].IncompArrayAddr = IncompArrayAddr;
   }
   sem_post (CONFIG_ALLOC_Global.GlobalSemaphoreAddr);
   *CollisionsFound = Collisions;
   
Exit:;

   /*  On the way out, release WorkArray if we had to allocate it */
   
   if (FreeWorkArray) free (WorkArray);
   
}

/* -------------------------------------------------------------------------- */

/*                   B a c k g r o u n d  C a l c
 *
 *  This is an internal routine of the allocation sub-system. It is the
 *  main routine for a background thread, which has the job of running 
 *  around through the possible allocations calculating any incompatible
 *  allocation lists that have not so far been created. Any number of
 *  these threads can run at once, all working together to fill in
 *  the gaps, but there is only a practical gain if there are spare
 *  processors available to run them on.
 */
 
void* BackgroundCalc (void* Argument)
{
   unsigned int Index;
   unsigned int* WorkArray = NULL;
   int Bytes;
   int ListsDone;
   int PossibleAllocations;
   StatusType Status = STATUS__OK;
   
   /*  We take argument as the index value at which we should start
    *  to fill in - this prevents multiple threads all doing the
    *  same calculation at once. It would be nice to have some sort
    *  of noticeboard array the threads could use to reserve an index.
    */
   
   Index = (unsigned long) Argument;
   DEBUG ("Background thread starting on index %d\n",Index);

   ListsDone = 0;
   
   /*  First, we need a work array, so that GenerateIncompatList()
    *  doesn't need to allocate its own workspace.
    */
   
   PossibleAllocations = CONFIG_ALLOC_Global.PossibleAllocations;
   Bytes = (PossibleAllocations + 1) * sizeof(unsigned int);
   WorkArray = (unsigned int*) malloc (Bytes);
   if (WorkArray == (unsigned int*) 0) {
      Status = CONFIG__NO_MEMORY;
      goto Exit;
   }
   
   for (;;) {
   
      /*  Check to see if the 'complete' flag has been set by the 
       *  main thread. If so, this is a request to quit now.
       */
   
      if (CONFIG_ALLOC_Global.AllocationComplete) break;
      
      /*  Each time we wrap round the top of the array, we see if
       *  there are any entries left that need to be filled in.
       *  If not, we quit with a job well done. Since we're just 
       *  looking at the global array, we don't need to take the
       *  semaphore at this point. That can be left to
       *  GenerateIncompatList().
       */
      
      if (Index >= PossibleAllocations) {
         int StillNeeded = CONFIG_FALSE;
         int CheckIndex;
         Index = 0;
         for (CheckIndex = 0; CheckIndex < PossibleAllocations; CheckIndex++) {
            if (CONFIG_ALLOC_Global.Allocations[CheckIndex].IncompArrayAddr ==
                                                     (unsigned int*) 0) {
               StillNeeded = CONFIG_TRUE;
               break;
            }
            if (CONFIG_ALLOC_Global.AllocationComplete) break;
         }
         if (!StillNeeded) break;
      }
      if (CONFIG_ALLOC_Global.Allocations[Index].IncompArrayAddr == 
                                                    (unsigned int*) 0) {
      
         /*  This is one that needs doing. */
         
         int Collisions;
         GenerateIncompatList (Index,&Collisions,WorkArray,&Status);
         if (Status != STATUS__OK) goto Exit;
         ListsDone++;
      }
      
      /*  Now, see if the next one needs doing. */
         
      Index++;
   }
   
Exit:;

   DEBUG ("Thread exiting, status %ld, lists done  %d\n",
          (long)Status,ListsDone);
   if (WorkArray) free(WorkArray);
   return 0;
}

/* -------------------------------------------------------------------------- */

/*                   A l l o c a t i o n  I n d e x
 *
 *  Given a pivot number and a target number, this returns the index 
 *  number for the corresponding allocation. If the allocation is not
 *  physically valid, this returns -1. The target number is an external
 *  target number.
 */
 
int AllocationIndex (
   int IPivot,
   int ExtITarget)
{
   
   /*  Note: This calculation for the index - ie the order for pivots
    *  and targets in the 2D IndexLookup array - must match that used
    *  in AllocationInit() to set up the IndexLookup array.
    */
    
   int Index = -1;
   if ((ExtITarget < 0) || (IPivot < 0) ||
        (ExtITarget >= CONFIG_ALLOC_Global.MethodInfo->NumberTargets) ||
        (IPivot >= CONFIG_ALLOC_Global.MethodInfo->NumberPivots)) {
      Index = -1;
   } else {
      int IntITarget = CONFIG_ALLOC_Global.ReverseIndex[ExtITarget];
      Index = CONFIG_ALLOC_Global.IndexLookup[
         (IntITarget * CONFIG_ALLOC_Global.MethodInfo->NumberPivots) + IPivot];
   }
   return Index;         
}

/* -------------------------------------------------------------------------- */

/*                     A l l o c a t e  I n d e x
 *
 *  This routine must be called to tell the allocation sub-system that
 *  an allocation with the specified index has been made. It returns
 *  true (non-zero) if it recorded the allocation OK, false otherwise.
 *  (It will only fail if more allocations are made than the sub-system
 *  expects, and it allows for every pivot to be allocated, so it 
 *  shouln't fail.)
 */


int AllocateIndex (                    /*  Returns true if index handled OK */
   int Index)                          /*  Index of allocation being made */
{
   int Result = 1;
   
   /*  Check Index is valid */
   
   if ((Index < 0) || (Index >= CONFIG_ALLOC_Global.PossibleAllocations)) {
      Result = 0;
      goto Exit;
   }
      
   /*  Check the allocation hasn't already been made. If so, that's it. */
   
   if (CONFIG_ALLOC_Global.AllocFlags[Index]) goto Exit;
   
   /*  Flag the allocation as made, and increment the allocation count */
   
   CONFIG_ALLOC_Global.AllocFlags[Index] = CONFIG_TRUE;
   
   CONFIG_ALLOC_Global.AllocEntries++;
   
   /* Now update AllocPivots and AllocTargets. */

   CONFIG_ALLOC_Global.AllocPivots
        [CONFIG_ALLOC_Global.Allocations[Index].Pivot] = 
                              CONFIG_ALLOC_Global.Allocations[Index].IntTarget;
   CONFIG_ALLOC_Global.AllocTargets
        [CONFIG_ALLOC_Global.Allocations[Index].IntTarget] = 
                              CONFIG_ALLOC_Global.Allocations[Index].Pivot;
   /*
   DEBUG ("Allocating target %d to %d\n",
                              CONFIG_ALLOC_Global.Allocations[Index].Target,
                              CONFIG_ALLOC_Global.Allocations[Index].Pivot);
    */
                               
   /*  Flag that AllocArray is no longer up to date */
   
   CONFIG_ALLOC_Global.AllocArrayValid = CONFIG_FALSE;
      
Exit:;

   return Result;
       
}

/* -------------------------------------------------------------------------- */

/*                     D e a l l o c a t e  I n d e x
 *
 *  This routine must be called to tell the allocation sub-system that
 *  a previous allocation with the specified index has been deallocated. 
 *  It returns true (non-zero) if it recorded the deallocation OK, false 
 *  otherwise. (It will only fail if it had no record of the specified
 *  allocation.)
 */
 
int DeallocateIndex (                  /*  Returns true if index handled OK */
   int Index)                          /*  Index of deallocation being made */
{
   int Result = 1;
   
   /*  Check Index is valid */
   
   if ((Index < 0) || (Index >= CONFIG_ALLOC_Global.PossibleAllocations)) {
      Result = 0;
      goto Exit;
   }
      
   /*  Check the allocation has already been made. If not, report error. */
   
   if (!(CONFIG_ALLOC_Global.AllocFlags[Index])) {
      Result = 0;
      goto Exit;
   }
   
   /*  Flag the allocation as cleared */
   
   CONFIG_ALLOC_Global.AllocFlags[Index] = CONFIG_FALSE;
      
   /* Now update AllocPivots and AllocTargets. */

   CONFIG_ALLOC_Global.AllocPivots
                        [CONFIG_ALLOC_Global.Allocations[Index].Pivot] = -1;
   CONFIG_ALLOC_Global.AllocTargets
                        [CONFIG_ALLOC_Global.Allocations[Index].IntTarget] = -1;
                      
   /*  Update the entry count and set to return good status */
   
   CONFIG_ALLOC_Global.AllocEntries--;
   Result = 1;
   
   /*  Flag that AllocArray is no longer up to date */
   
   CONFIG_ALLOC_Global.AllocArrayValid = CONFIG_FALSE;
         
Exit:;

   return Result;
       
}

/* -------------------------------------------------------------------------- */

/*                     L i s t  A l l o c  A r r a y
 *
 *  An internal diagnostic routine for the allocation sub-system that
 *  lists the current recorded allocations. Note that this no longer
 *  uses the global AllocArray (which used to hold a sorted list of the
 *  current allocations), but now uses AllocFlags directly.
 */
 
void ListAllocArray(void)
{
   int I;
   for (I = 0; I < CONFIG_ALLOC_Global.PossibleAllocations; I++) {
      if (CONFIG_ALLOC_Global.AllocFlags[I]) printf ("%d ",I);
   }
   printf ("\n");
}  

/* -------------------------------------------------------------------------- */

/*                        A l l o c  E n t r i e s
 *
 *  Returns the number of current allocations - those reported via 
 *  AllocateIndex() and DeallocateIndex().
 */
 
int AllocEntries(void)
{
   /*  This is simple. The value required is held in the sub-system
    *  global structure.
    */
   
   return CONFIG_ALLOC_Global.AllocEntries;
    
}  

/* -------------------------------------------------------------------------- */

/*                    A l l o c  E n t r y  D e t a i l s
 *
 *  Returns the details of one of the current allocations - those reported via 
 *  AllocateIndex() and DeallocateIndex(). The allocation in question is
 *  given by a number which should be in the range 1 up to and including the
 *  value returned by AllocEntries(). If the number is invalid, this routine
 *  returns false, otherwise it returns true. The target number returned is 
 *  an external target index.
 */
 
int AllocEntryDetails(
   int Number,                 /* The number of the allocation in question */
   int* Index,                 /* The allocation index for the allocation */
   int* Pivot,                 /* The pivot number - from zero up */
   int* ExtTarget,             /* The target number - from zero up */
   double* Theta)              /* The theta value for this allocation */
{
   int OK = CONFIG_TRUE;
   int Ind;
   int I;
   
   /*  Get AllocArray up to date if necessary. This is now the only use
    *  of AllocArray in this code, and it was suggested in the comments
    *  to earlier versions that it could be dispensed with and this routine
    *  could simply work through AllocFlags each time it was called. That
    *  is true, but it seems more efficient to only have to sweep through 
    *  AllocFlags once and save the results in AllocArray. This means that
    *  AllocArray is now reduced really to just private workspace for this
    *  routine. Note that any change to the current allocation (as reported 
    *  to AllocateIndex() or DeallocateIndex()) clears the valid flag.
    */
   
   if (!CONFIG_ALLOC_Global.AllocArrayValid) {
      Ind = 0;
      for (I = 0; I < CONFIG_ALLOC_Global.PossibleAllocations; I++) {
         if (CONFIG_ALLOC_Global.AllocFlags[I]) {
            CONFIG_ALLOC_Global.AllocArray[Ind++] = I;
            if (Ind >= CONFIG_ALLOC_Global.AllocArraySize) {
               OK = CONFIG_FALSE;
               goto Exit;
            }
         }
      }
      CONFIG_ALLOC_Global.AllocArray[Ind] = -1;
      CONFIG_ALLOC_Global.AllocArrayValid = CONFIG_TRUE;
   }
            
   if ((Number < 1) || (Number > CONFIG_ALLOC_Global.AllocEntries)) {
      OK = CONFIG_FALSE;
   } else {
      int IntTarget;
   
      /*  Note that the target index returned is the external one,
       *  referring to the original unsorted target list passed to the
       *  allocation sub-system.
       */
       
      OK = CONFIG_TRUE;
      *Index = CONFIG_ALLOC_Global.AllocArray[Number - 1];
      IntTarget = CONFIG_ALLOC_Global.Allocations[*Index].IntTarget;
      *ExtTarget = CONFIG_ALLOC_Global.TargetIndex[IntTarget];
      *Pivot = CONFIG_ALLOC_Global.Allocations[*Index].Pivot;
      *Theta = CONFIG_ALLOC_Global.Allocations[*Index].Theta;
   }

Exit:;

   return OK;
}  

/* -------------------------------------------------------------------------- */

/*                         I s  A l l o c  A b l e
 *
 *  Possibly the most important routine in the allocation sub-system. This
 *  returns true if the allocation with the specified index (as returned
 *  by AllocationIndex() is valid given the current set of allocations
 *  (as reported via AllocateIndex() and DeallocateIndex().)
 */
 
int IsAllocAble (             /* Returns true if this allocation is valid */
   int Index)                 /* Index of the allocation to be tested */
{
   unsigned int  Allocation;  /* Allocation incompatible with the test one */
   int  IncompatIndex;        /* Index into the sparse incompatability array */
   int  Allowed;              /* True if allocation is allowed */
   int  IntITarget;           /* Internal target index for allocation */
   int  ExtITarget;           /* External target index for allocation */
   unsigned int* IncompatArray;/* Array holding incompatible allocations list */
   StatusType LocalStatus;    /* Local inherited status variable */

   /*  This routine works by comparing the numbers in two arrays. One is
    *  the list of allocations incompatible with this allocation, the other
    *  is the array flagging the current allocations. Clearly, if one of the
    *  allocations incompatible with this one is flagged as made, the 
    *  allocation is disallowed.
    */
   
   LocalStatus = 0;
   Allowed = CONFIG_TRUE;
   
   /*  First, check that Index is valid */
   
   if ((Index < 0) || (Index >= CONFIG_ALLOC_Global.PossibleAllocations)) {
      Allowed = CONFIG_FALSE;
      goto Exit;
   }
   
   /*  Secondly, we want to test if the Target or Pivot we want to allocate 
    *  (as indexed by Index) is already allocated. If it is then we 
    *  return false. (Note that PivotOfTarget() works with external
    *  target indices.)
    */

   IntITarget = CONFIG_ALLOC_Global.Allocations[Index].IntTarget;
   ExtITarget = CONFIG_ALLOC_Global.TargetIndex[IntITarget];
   if (PivotOfTarget(ExtITarget) >= 0
        || TargetOfPivot(CONFIG_ALLOC_Global.Allocations[Index].Pivot) >= 0)
   { 
      Allowed = CONFIG_FALSE;
      goto Exit;
   }

   /*  We want the details of this allocation. In particular we want to
    *  know the address of the array that holds the list of incompatible
    *  allocations for this allocation.
    */
    
   IncompatArray = CONFIG_ALLOC_Global.Allocations[Index].IncompArrayAddr;
   
   /*  If the list hasn't been generated yet, do it on the fly. */
   
   if (IncompatArray == (unsigned int*) 0) {
      int Collisions;
      GenerateIncompatList(Index,&Collisions,0,&LocalStatus);
      IncompatArray = CONFIG_ALLOC_Global.Allocations[Index].IncompArrayAddr;
   }
   IncompatIndex = 0;
   
   /*  Run through all the incompatible allocations and see if any
    *  have been made. The list of allocations is terminated by an all ones
    *  flag.
    */
   
   if (LocalStatus == 0) {
      for (;;) {
         Allocation = IncompatArray[IncompatIndex++];
         if (Allocation == ALL_ONES) break;
         if (CONFIG_ALLOC_Global.AllocFlags[Allocation]) {
            Allowed = CONFIG_FALSE;
            break;
         }
      }
   }
   
Exit:;
 
   return Allowed;
    
}

/* -------------------------------------------------------------------------- */

/*                        P i v o t O f T a r g e t 
 *
 *  Used to find out the pivot of the fibre allocated to a target
 *  if the allocation exists. The target number passed is an external
 *  target number.
 *
 *  Returns the pivot index if ITarget is allocated. Returns -1 otherwise.
 */
 
int PivotOfTarget(int ExtITarget)
{
   int IPivot;
   if (ExtITarget < 0 || ExtITarget >= 
                     CONFIG_ALLOC_Global.MethodInfo->NumberTargets) {
      IPivot = -1;
   } else {
      int IntITarget = CONFIG_ALLOC_Global.ReverseIndex[ExtITarget];
      IPivot = CONFIG_ALLOC_Global.AllocTargets[IntITarget];
   }
   return IPivot;
}

/* -------------------------------------------------------------------------- */

/*                        T a r g e t O f P i v o t 
 *
 *  Used to find out the target of the fibre allocated from a pivot
 *  if the allocation exists.
 *
 *  Returns the external target index if IPivot is allocated. 
 *  Returns -1 otherwise.
 */
 
int TargetOfPivot(int IPivot)
{
   int ExtITarget;             /* Returned target number - in external order */
   
   if(IPivot < 0 || IPivot >= CONFIG_ALLOC_Global.MethodInfo->NumberPivots) {
      ExtITarget = -1;
   } else {
      int IntITarget = CONFIG_ALLOC_Global.AllocPivots[IPivot];
      if (IntITarget >= 0) {
         ExtITarget = CONFIG_ALLOC_Global.TargetIndex[IntITarget];
      } else {
         ExtITarget = -1;
      }
   }
   return ExtITarget;
}
/* -------------------------------------------------------------------------- */

/*                        S w a p
 *
 *  Swaps two existing allocations with no net loss of allocations.
 *  No implementation was provided in the old configure API.
 *  This generic routine is useful for:
 *  (i)  Post configuration swapping routines and
 *  (ii) Algorithm developers.
 *  and was developed as part of the randomisation process
 *  for the simulated annealing algorithm written by BM/MQ_AAO.
 *  
 *  The idea is that we want to swap the allocations:
 *  
 *  [PivotA,TargetA] and [PivotB,TargetB]
 *            to
 *  [PivotA,TargetB] and [PivotB,TargetA]
 *
 *  This is achieved by first deallocating both
 *  
 *  [PivotA,TargetA] and [PivotB,TargetB]
 *  
 *  and then trying to allocate
 *  
 *  [PivotA,TargetB] followed by [PivotB,TargetA]
 *
 *  If this fails we try to swap the other way around
 *
 *  [PivotB,TargetA] followed by [PivotA,TargetB]
 *
 *  If this still fails we try to restore the original state
 *  by trying to allocate 
 *  
 *  [PivotA,TargetA] followed by [PivotB,TargetB]
 *          or
 *  [PivotB,TargetB] followed by [PivotA,TargetA]
 *
 *  Returns 1 is the swap was successful, -1 if unsuccessful 
 *
 *  Author: BM.
 */

int Swap(int PivotA,int TargetA,int PivotB,int TargetB)
{
   int IndexAA,IndexAB,IndexBA,IndexBB;   /* Indices used in swapping process,
                     format is Index[IPivot][ITarget] */
   int nalloc;          /* Keeps track of the number of allocations
                     we are making in the swapping process */

   /*  Note that the target indices used throughout this routine are
    *  in terms of the external target list, not the internal sorted
    *  target list. This is a routine that works entirely using 
    *  the external interface of the allocation sub-system - indeed,
    *  it started life as an external routine, and was later packaged
    *  within the allocation sub-system code.
    */

   /* We can't swap with the same target or pivot */
   if(TargetA == TargetB || PivotA == PivotB)
      return -1;
   /* The case if some components aren't present isn't handled here.
    * We rely on both allocations existing, hence this test.
    */
   if(PivotA < 0 || TargetA < 0 || PivotB < 0 || TargetB < 0)
      return -1;
   /* Check if the allocations already exist */
   if(TargetOfPivot(PivotA) != TargetA || TargetOfPivot(PivotB) != TargetB)
      return -1;
   /* Initialise indices used in swapping process */
   
   IndexAA = AllocationIndex(PivotA,TargetA);
   IndexAB = AllocationIndex(PivotA,TargetB);
   IndexBA = AllocationIndex(PivotB,TargetA);
   IndexBB = AllocationIndex(PivotB,TargetB);
   
   /* Deallocate existing allocations to give us some 
    * room to make the swap.
    */
   DeallocateIndex(IndexAA);
   DeallocateIndex(IndexBB);
   
   /* Try to allocate [PivotA,TargetB] first */
   nalloc = 0;
   if(IsAllocAble(IndexAB))
   {
      AllocateIndex(IndexAB);
      nalloc++;
   }
   /* If successful, we finish the swap */
   if(nalloc ==1 && IsAllocAble(IndexBA))
   {
      AllocateIndex(IndexBA);
      nalloc++;
      return 1;
   }
   /* Failed to finish swap, clean up tracks for next try */
   else if(nalloc == 1)
   {
      DeallocateIndex(IndexAB);
      nalloc = 0;
   }
   /* Try to allocate [PivotB,TargetA] first */
   if(IsAllocAble(IndexBA))
   {
      AllocateIndex(IndexBA);
      nalloc++;
   }
   /* If successful, we finish the swap */
   if(nalloc ==1 && IsAllocAble(IndexAB))
   {
      AllocateIndex(IndexAB);
      nalloc++;
      return 1;
   }
   /* Failed to finish swap, clean up tracks for recovery */
   else if(nalloc == 1)
   {
      DeallocateIndex(IndexBA);
      nalloc = 0;
   }
   /* Try to restore the original state of the allocations */
   if(IsAllocAble(IndexAA))
   {
      AllocateIndex(IndexAA);
      if(IsAllocAble(IndexBB))
      {
         AllocateIndex(IndexBB);
      }
   }
   /* Try the other way if first way failed */
   else if(IsAllocAble(IndexBB))
   {
      AllocateIndex(IndexBB);
      if(IsAllocAble(IndexAA))
      {
         AllocateIndex(IndexAA);
      }
   }
   /*else 
   {
       May need some extra kind of return type
            here to indicate that the original state
       could not be restored.

       As it stands it is just -1 (unsuccessful swap)

       I don't think this code would ever execute.
   }*/
   return -1;
}


/* -------------------------------------------------------------------------- */

/*                   A l l o c  B i n  S e a r c h
 *
 *  A binary search through an array of integers for an entry equal to or
 *  lower than a specified value. If there is no entry lower than the
 *  specified value, it returns -1. Otherwise, it returns the index of
 *  the array element (starting from 0) that is either the target value
 *  or the nearest value that is lower than the target value.
 *
 *  Credit: This code is a C re-working of a very nice Java implementation
 *  I found on www.tbray.org, but Googling for 'binary search'. I don't
 *  know who TBray is, but he should get the credit for this. The example
 *  given is a straight binary search for a specified value, but by
 *  removing a final test this I turned it into a routine that did what
 *  I need for the allocation sub-system. Normally, this routine has a
 *  test at the end for (Low >= 0) && (Array[Low] == Target) in which
 *  case it returns Low. In other cases it returns -1 to indicate the entry
 *  is not found. Taking that test out allows AllocateIndex() to 
 *  use this routine to find the entry above which it should insert its
 *  new entry. However, it means that DeallocateIndex() and AllocationInit()
 *  have to test explicitly to see if an exact match was found or not.
 */

int AllocBinSearch (                     /* (<) Index of match, or -1 */
   const unsigned int Array[],           /* (>) Array to be searched */
   int Size,                             /* (>) Number of integers in Array */
   unsigned int Target)                  /* (>) Value to be searched for */
{
   int High = Size;
   int Low = -1;
   int Probe;
   while ((High - Low) > 1) {
      Probe = (High + Low) / 2;
      if (Array[Probe] > Target) {
         High = Probe;
      } else {
         Low = Probe;
      }
   }
   return Low;
}

/* -------------------------------------------------------------------------- */

/*                   A l l o c  T a r g e t  C o m p a r
 *
 *  This routine is the comparison routine passed to qsort() to sort 
 *  the target list into ascending order of X,y position. What is being
 *  sorted is a list of target indices into the TargetInfo[] array passed
 *  originally as part of the MethodInfo structure whose address is held
 *  globally. So this routine has to compare the appropriate coordinate
 *  values for the index elements whose addresses it is passed.
 */

int AllocTargetCompar (
   const void* Ptr1,
   const void* Ptr2)
{
   /*  The pointers are in fact pointers to long values and these values
    *  are themselves index values into the TargetInfo array.
    */
    
   int Compar;                         /* Returned comparision value */
    
   const CONFIG_MethodInfo *MethodInfo;/* Local pointer to Method details */
   const CONFIG_TargetInfo *TargetInfo;/* Local pointer to MethodInfo element */
   
   long* IndexPtr1 = (long*) Ptr1;
   long* IndexPtr2 = (long*) Ptr2;
   long Index1 = *IndexPtr1;
   long Index2 = *IndexPtr2;
   
   /*  From those, pick up the coordinates and compare them. The return
    *  values (-1,0, or +1) are as required by the standard qsort(3) routine.
    */

   MethodInfo = CONFIG_ALLOC_Global.MethodInfo;
   TargetInfo = MethodInfo->TargetInfo;
   
   /*  Compare the Y coordinates first. Only if these are equal
    *  do we need to look at the X coordinates.
    */
   {    
       long Y1 = TargetInfo[Index1].Y;
       long Y2 = TargetInfo[Index2].Y;
       if (Y1 < Y2) {
           Compar = -1;
       } else if (Y1 > Y2) {
           Compar = 1;
       } else {
           long X1 = TargetInfo[Index1].X;
           long X2 = TargetInfo[Index2].X;
           if (X1 < X2) {
               Compar = -1;
           } else if (X1 > X2) {
               Compar = 1;
           } else {
               
               /*  If the X and Y coordinates are both the same, this is
                *  probably worth indicating to the user. We fall back on
                *  comparing the object name in this case, just to try to
                *  get a reproducible sorting order.
                */
               
               DEBUG ("Warning: Targets '%s' and '%s' are in the same place\n",
                      TargetInfo[Index1].ObjName,TargetInfo[Index2].ObjName);
               Compar = strcmp(TargetInfo[Index1].ObjName,TargetInfo[Index2].ObjName);
           }
       }
   }
   
   return Compar;
}
   

/* -------------------------------------------------------------------------- */

/*                   A l l o c  S o r t  T a r g e t s
 *
 *  This routine sorts the target list into ascending order of X,y position.
 *  The intention is that this gets around the problems when the same
 *  target list is passed to the allocation sub-system, but with the targets
 *  in a changed order. The collision matrix would normally need to be
 *  recalculated in these circumstances, but this can be avoided if the
 *  allocation sub-system works internally always with a sorted target list.
 *  The problem is to hide this resorting from the higher levels of the code.
 *  This assumes that the TargetIndex, ReverseIndex and
 *  SortedTargetInfo global arrays have already been allocated but not filled,
 *  and that the TargetInfo in the MethodInfo structure pointed to by
 *  the MethodInfo global variable is valid. It fills up SortedTargetInfo
 *  with a sorted version of the MethodInfo.TargetInfo structure array,
 *  and fills TargetIndex and ReverseIndex to provide a quick mapping from
 *  the index values used by the higher level code and the index values into
 *  SortedTargetInfo. The code internally should use SortedTargetInfo
 *  rather than the supplied MethodInfo.TargetInfo array now that it has
 *  been calculated.
 */

void AllocSortTargets (void)
{
   /*  Local variables */
   
   const CONFIG_MethodInfo *MethodInfo;/* Local pointer to Method details */
   const CONFIG_TargetInfo *TargetInfo;/* Local pointer to MethodInfo element */
   CONFIG_TargetInfo* SortedTargetInfo;/* Local pointer to global array */
   long* TargetIndex;                  /* Local pointer to global array */
   long* ReverseIndex;                 /* Local pointer to global array */
   
   int  ITarget;                       /* Loop index through targets */
   int  NTargets;                      /* Number of targets */
   
   /*  To save constant appearances of 'CONFIG_ALLOC_Global.' in the code,
    *  set local pointers to the global arrays.
    */
   
   MethodInfo = CONFIG_ALLOC_Global.MethodInfo;
   TargetInfo = MethodInfo->TargetInfo;
   TargetIndex = CONFIG_ALLOC_Global.TargetIndex;
   ReverseIndex = CONFIG_ALLOC_Global.ReverseIndex;
   SortedTargetInfo = CONFIG_ALLOC_Global.SortedTargetInfo;
   NTargets = MethodInfo->NumberTargets;
      
   /*  Initialise the TargetIndex array, prior to sorting it on the
    *  basis of the corrsponding target coordinates.
    */
   
   for (ITarget = 0; ITarget < NTargets; ITarget++) {
      TargetIndex[ITarget] = ITarget;
   }
         
   /*  We do the actual sort of the TargetIndex array. The AllocTargetCompar
    *  function we pass to qsort() has access to the global TargetInfo
    *  array and so can compare the coordinates of the elements whose
    *  index values it is passed.
    */
   
   qsort(TargetIndex,NTargets,sizeof(long),AllocTargetCompar);
   
   /*  Now set the various arrays into the sorted order, and set the
    *  index arrays to reflect this. TargetIndex array is now sorted
    *  into ascending position order of the targets that it indexes.
    */
    
   for (ITarget = 0; ITarget < NTargets; ITarget++) {
      int TIndex = TargetIndex[ITarget];
      ReverseIndex[TIndex] = ITarget;
      memcpy (&(SortedTargetInfo[ITarget]),&(TargetInfo[TIndex]),
                                           sizeof(CONFIG_TargetInfo));
   }
   
   /*  And that's that. */
   
}

/* -------------------------------------------------------------------------- */

/*                        S e m  C r e a t e 
 *
 *  This routine creates the global semaphore used to control access to the
 *  global allocation structure. This would normally be a simple call to
 *  sem_init(), but some systems (that's you, Mac OS X!) don't support the
 *  unnamed semaphores created by sem_init and on such systems sem_init()
 *  returns an error. If the semaphore cannot be created, this routine sets the
 *  maximum number of threads to zero, which means the allocation can continue,
 *  but without using multi-threading to speed things up. The MethodInfo
 *  argument is used only to provide a mechanism for reporting errors.
 *
 *  If this call succeeds, the address of the semaphore to use is set into
 *  GlobalSemaphoreAddr in the global variables. If a named semaphore had to
 *  be used, its name is set in GlobalSemaphoreName.
 */
 
void SemCreate (
   const CONFIG_MethodInfo *MethodInfo)
{
   int Failed = CONFIG_FALSE;
   
   /*  First of all, we try it the easy way. */
   
   CONFIG_ALLOC_Global.GlobalSemaphoreAddr =
                          &CONFIG_ALLOC_Global.GlobalSemaphore;
   if (sem_init (CONFIG_ALLOC_Global.GlobalSemaphoreAddr,0,1) < 0) {
   
      /*  It failed. Look for the trade-mark of a system that doesn't
       *  support unnamed semaphores.
       */
       
      if (errno == ENOSYS) {
      
         /*  We need a named semaphore. We need one in a writable directory,
          *  with a unique name. We use a file name based on the pid of this
          *  task, and put it in /tmp.
          */
          
         sprintf (CONFIG_ALLOC_Global.GlobalSemaphoreName,"/tmp/Config_%ld.sem",
                                                                (long)getpid());
                  
         /*  Now create it . If the semaphore already exists, the existing
          *  semaphore will be used, but it will be initialised to the specified
          *  value - 1, ie available.
          */
         
         CONFIG_ALLOC_Global.GlobalSemaphoreAddr = sem_open(
                       CONFIG_ALLOC_Global.GlobalSemaphoreName, 
                                      O_CREAT|O_EXCL ,S_IWUSR | S_IRUSR , 1);
                                      
         if (CONFIG_ALLOC_Global.GlobalSemaphoreAddr == (sem_t *)SEM_FAILED) {
            Failed = CONFIG_TRUE;
         }
      }
   }
   
   if (Failed) {
   
      /*  If there was a failure, we report it, and set the global variables
       *  to what should be safe values. (This sets it as though an unnamed
       *  semaphore was used, but it won't have been initialised properly.
       *  In this case calls to sem_get() and sem_post() will return errors.
       */
       
      MethodInfo->ErrRepRoutine(0,"Failed to create thread-control semaphore.");
      MethodInfo->ErrRepRoutine(0,"Will fall back to single thread operation.");
      CONFIG_ALLOC_Global.MaxThreads = 0;
      CONFIG_ALLOC_Global.GlobalSemaphoreName[0] = '\0';
      CONFIG_ALLOC_Global.GlobalSemaphoreAddr = 
                       &CONFIG_ALLOC_Global.GlobalSemaphore;
   }
}

/* -------------------------------------------------------------------------- */

/*                        S e m  D e l e t e 
 *
 *  This routine deletes the global semaphore used to control access to the
 *  global allocation structure.
 */
 
void SemDelete (void)
{
   /*  If the semaphore address is null, it wasn't created properly, so
    *  we have nothing to do,
    */
    
   if (CONFIG_ALLOC_Global.GlobalSemaphoreAddr != NULL) {
   
      /*  We have to handle this differently for named and unnamed semaphores */
      
      if (CONFIG_ALLOC_Global.GlobalSemaphoreName[0] != '\0') {
         
         /*  This is a named semaphore. We need to close it and delete it. 
          *  This is the reverse of the sem_open() call that created it.
          */
         
         (void) sem_close(CONFIG_ALLOC_Global.GlobalSemaphoreAddr);
         (void) sem_unlink(CONFIG_ALLOC_Global.GlobalSemaphoreName);
         
      } else {
      
         /*  This is an unnamed semaphore. We destroy it, reversing the
          *  sem_init() call that created it.
          */
          
         (void) sem_destroy (CONFIG_ALLOC_Global.GlobalSemaphoreAddr);
      }
      
      CONFIG_ALLOC_Global.GlobalSemaphoreAddr = NULL;
      CONFIG_ALLOC_Global.GlobalSemaphoreName[0] = '\0';
   }
}
   
/* -------------------------------------------------------------------------- */

/*    A p p e n d i x :  A  N o t e  O n  T r a d e - O f f s
 *
 
   There are a number of trade-offs in this code, mainly connected with the
   way information is stored in the global arrays. 

   The largest array is the incompatability matrix. In principle, we allocate
   an index number to each possible allocation and then create a square matrix
   PossibleAllocations by PossibleAllocations with each entry a flag showing
   whether the two allocations represented by that entry are incompatible.
   That is, if one allocation will cause a collision with the other, then
   the matrix entry should show that this combination is incompatible.

   However, there are two reasons why I didn't use a straightforward
   2D array for this. One is connected with the size, the other with the
   way it is used.

   A typical 2dF field has about 400 pivots and perhaps
   400 possible targets. (Sometimes, considerably more targets have to be
   considered. We have worked with fields with up to 1000 targets - after 
   that, the higher levels of the configuration code usually filter them
   by magnitude.) So there might be 160,000 possible allocations on the 
   simplistic assumption that each pivot might be allocated to any
   of the targets. However, most of these allocations are physically 
   impossible. Typically, only about 10% are physically possible. But this
   would still make the matrix 16000 square, in the 400 target, 400 pivot
   2dF case. Held as bytes, this is about 256Mbytes. This isn't out of the
   question, but it's big, and will increase enormously if the number of
   targets increases.

   Such a matrix could be reduced in size by using just one bit for each
   element, but this introduces overheads in access time, because of the
   matrix addressing and bit shuffling needed. It is also the case that
   the matrix is diagonally-symmetric; if allocation I is incompatible
   with allocation J, then allocation J is incompatible with allocation I.
   This would reduce the size of a bit implementation of the matrix from
   256Mbytes in the example case to 16Mbytes. So this would be a possible
   implementation, the downside being the bit-wise access required. These
   figures are affected significantly if a crowded field is involved - one
   test field for AAOmega had over 65,000 physically possible allocations,
   and the memory usage skyrockets. The 'sparse-sampling' of the allocations
   (and that needs to be described in more detail) can help a lot here.

   The implementation actually used in this code makes use of the sparseness
   of the matrix. In practice, most allocations are compatible with most 
   other allocations. In the 2dF test field used for initial testing, out of
   about 256 million matrix elements, only about 4 million represented
   incompatible combinations. The current implementation holds, for each of
   the possible allocations, a sorted array giving the index numbers of
   the allocations incompatible with that allocation. This makes, in the 
   example case, an array of 4 million elements. With each held as a
   4 byte integer, this also takes 16Mbytes.

   (Note that a 4 byte integer is used, but this could be cut down to a
   2 byte integer if the allocation index numbers never exceed the 16bit
   limit. A -1 value is used to flag the end of each set of numbers, but
   an unsigned 16-bit implementation could just test for 0xffff instead
   of -1, so 16 bits could be used if there were fewer than 65535 possible
   allocations. But this would require that we enforce sparse sampling
   if there were more than 65535 possible allocations.)

   One reason this implementation was chosen over the bit array option
   was that this makes it easier to code IsAllocAble(), which needs to
   be able to check that the current set of allocations is compatible with any 
   proposed new allocation. In this implementation, the allocations are
   recorded just in the AllocFlags array, which can be indexed by
   allocation index, and IsAllocAble() just works through the list of
   incompatible allocations for the proposed allocation, seeing if any
   are flagged as current. For this to be fast, we need the incompatible
   index numbers available in an array. The array doesn't need to be sorted,
   but it is, just as a consequence of the way it is built up.

   It is also the case that the setting up of the incompatability matrix is
   an inherrently parallel computation, since the value of each element is
   independent on the value of any other element. By making each of the
   incompatability lists a separately allocated array, it becomes possible
   to calculate them independently, and simulataneously, and this 
   implementation uses multi-threading to get the most out of a multi-
   processor system.
   
   'On-the-fly' calculation of the incompatable allocations was first
   tried on a single allocation basis - have we been asked about this
   allocation already, and if so what was the answer? This was less
   successful than hoped. The amount of information as to which allocations
   collide was relatively easy to store, but the information as to which
   combinations had already been tested was huge. The current scheme only
   works well because IsAllocAble() knows that its list of the incompatible
   allocations for any one allocation is complete. (I experimented
   with holding the 'already tested' information in Judy arrays, and even
   in up to a Gigabyte bit array, but it always ran slowly, partly because
   it required maintaining an array listing the current allocations in
   order to test the allocation in question against each current allocation.
   Having a simple, complete (any usually short) list of incompatible
   allocations and a flag array in which each of these can be checked to
   see if it is already allocated is much faster. At least, I couldn't
   do better. The reader is welcome to try.)

   The upshot of all this is that there are loads of ways this could all be
   coded, and without actually trying an idea in practice it's hard to
   guess how efficient it will be.

   Improving this has been left as an exercise for the reader...

*/
