/* -------------------------------------------------------------------------- */

/*                        C h e c k  M a t r i c e s
 *
 *  This is a test routine developed during regression testing to diagnose
 *  problems with the collision matrix code. It compares the contents of
 *  one saved collision matrix file with those of another, and lists any
 *  discrepancies. (It's pretty crude and has little error handling - if
 *  the number of targets and pivots don't match, for example, it probably
 *  crashes. But it does pick up cases where one run of the allocation 
 *  system has classed two allocations as incompatible but another run thought
 *  they were OK, and that's what it was written for.)
 *
 *  Compiling and linking:
 *     This is a self-contained, stand-alone program.
 *     gcc -o CheckMatrices CheckMatrices.c  is all that's needed.
 *
 *  Invocation:
 *     CheckMatrices file1 file2
 *
 *     where file1 and file2 are two saved collision matrix files.
 *
 *  Author(s): Keith Shortridge, AAO.
 *
 *  History:
 *     12th July 2009. Original version, KS.
 */

#include <stdio.h>
#include <stdlib.h>

typedef long int CONFIG_Boolean;     /* Used for logical values */

typedef struct CONFIG_PivotInfo {    /* Details about each pivot */
   long X;                           /* X-coordinate of pivot, in microns */
   long Y;                           /* Y-coordinate of pivot, in microns */
   long ParkX;                       /* X-coordinate of park position */
   long ParkY;                       /* Y-coordinate of park position */
   int FibreType;                    /* A code indicating the fibre type - this
                                      * is instrument-dependent. */
   unsigned int FibreLength;         /* Fibre length in microns */
   double Theta;                     /* Theta value of park position */
   int AllowAllocChange;             /* Indicates if the allocation algorithm
                                      * is allowed to change the allocation
                                      * of this pivot.  If this is false, then
                                      * NoteDeallocation and NoAllocation will
                                      * return errors for this pivot. (This
                                      * field may be true even though FibreType
                                      * indicates a broken fibre, so both fields
                                      * need to be checked.) */
} CONFIG_PivotInfo;

typedef struct CONFIG_TargetInfo {   /* Details about each target */
   long X;                           /* X-coordinate of target, in microns */
   long Y;                           /* Y-coordinate of target, in microns */
   int OriginalIndex;                /* Index in original target list */
   CONFIG_Boolean PreAllocated;      /* True if explicitly allocated */
   char Type;                        /* Type of object. This is instrument
                                      * dependent. */
   char Priority;                    /* Priority value, 0..9, 9 = highest. */
   char Spect;                       /* Combines with 'Type' to provide extra
                                      * information about the target - 
                                      * usually the spectrograph to which it
                                      * should be allocated, hence the name. */
   char *ObjName;                    /* Pointer to string giving the name of
                                      * the Object */
} CONFIG_TargetInfo;
 
int main (
   int argc,
   char* argv[])
{
   FILE* SavedFd = NULL;   /* Used to read the saved incompatability matrix */
   FILE* SavedFd2 = NULL;  /* Used to read the saved incompatability matrix */
   
   SavedFd = fopen(argv[1],"r");
   SavedFd2 = fopen(argv[2],"r");
   
   if (SavedFd && SavedFd2) {
      int NumberPivots,NumberPivots2,NumberTargets,NumberTargets2;
      int IPiv,ITarget;
      CONFIG_TargetInfo TargetDetails,TargetDetails2;
      CONFIG_PivotInfo PivotDetails,PivotDetails2;
      int FileAllocs,FileAllocs2;
      int Entries,Elements;
      int I1;
      int* FAvailAllBase;
      int Bytes;
      int FileIndex,FileIndex2;
      
      fread(&NumberPivots,sizeof(NumberPivots),1,SavedFd);
      fread(&NumberPivots2,sizeof(NumberPivots2),1,SavedFd2);
      if (NumberPivots != NumberPivots2) {
         printf ("Number of pivots differs, %d %d\n",
                                   NumberPivots,NumberPivots2);
      }

      for (IPiv = 0; IPiv < NumberPivots; IPiv++) {
         fread (&PivotDetails,sizeof(PivotDetails),1,SavedFd);
         fread (&PivotDetails2,sizeof(PivotDetails2),1,SavedFd2);
         if (PivotDetails.X != PivotDetails2.X ||
                PivotDetails.Y != PivotDetails2.Y) {
            printf ("Pivot discrepancy, %d, %f, %f %f %f\n",IPiv,
               PivotDetails.X,PivotDetails.Y,PivotDetails2.X,PivotDetails2.Y);
         }
      }
      
      fread(&NumberTargets,sizeof(NumberTargets),1,SavedFd);
      fread(&NumberTargets2,sizeof(NumberTargets2),1,SavedFd2);
      if (NumberTargets != NumberTargets2) {
         printf ("Number of targets differs, %d %d\n",
                                   NumberTargets,NumberTargets2);
      }
      
      for (ITarget = 0; ITarget < NumberTargets; ITarget++) {
         fread (&TargetDetails,sizeof(TargetDetails),1,SavedFd);
         fread (&TargetDetails2,sizeof(TargetDetails2),1,SavedFd2);
         if (TargetDetails.X != TargetDetails2.X ||
                TargetDetails.Y != TargetDetails2.Y) {
            printf ("Target discrepancy, %d, %f, %f %f %f\n",IPiv,
               TargetDetails.X,TargetDetails.Y,TargetDetails2.X,TargetDetails2.Y);
         }
      }
      fread(&FileAllocs,sizeof(FileAllocs),1,SavedFd);
      fread(&FileAllocs2,sizeof(FileAllocs2),1,SavedFd2);
      if (FileAllocs != FileAllocs2) {
         printf ("Original allocation discrepancy %d %d\n",
                                                  FileAllocs,FileAllocs2);
      }
      fread(&FileAllocs,sizeof(FileAllocs),1,SavedFd);
      fread(&FileAllocs2,sizeof(FileAllocs2),1,SavedFd2);
      if (FileAllocs != FileAllocs2) {
         printf ("Actual allocation discrepancy %d %d\n",
                                                  FileAllocs,FileAllocs2);
      }
      
      Elements = NumberPivots * NumberTargets;
      Bytes = sizeof(int) * Elements;
      FAvailAllBase = (int*) malloc(Bytes);
      Entries = fread(FAvailAllBase,sizeof(int),Elements,SavedFd);
      Entries = fread(FAvailAllBase,sizeof(int),Elements,SavedFd2);


      (void) fread(&FileIndex,sizeof(FileIndex),1,SavedFd);
      (void) fread(&FileIndex2,sizeof(FileIndex2),1,SavedFd2);
      while (FileIndex >= 0 || FileIndex2 > 0) {
         unsigned int* ListPtr;
         unsigned int* ListPtr2;
         int Collisions,Collisions2;
         if (FileIndex != FileIndex2) {
            printf ("File index discrepancy, %d %d\n",FileIndex,FileIndex2);
         }
         fread(&Collisions,sizeof(Collisions),1,SavedFd);
         fread(&Collisions2,sizeof(Collisions2),1,SavedFd2);
         if (Collisions != Collisions2) {
            printf ("Collisions discrepancy, %d %d %d\n",
                                        FileIndex,Collisions,Collisions2);
         }
         
         ListPtr = 
            (unsigned int*) malloc ((Collisions + 1) * sizeof(unsigned int));
         ListPtr2 = 
            (unsigned int*) malloc ((Collisions2 + 1) * sizeof(unsigned int));
         fread(ListPtr,sizeof(unsigned int),Collisions + 1,SavedFd);
         fread(ListPtr2,sizeof(unsigned int),Collisions2 + 1,SavedFd2);
         if (Collisions != Collisions2) {
            int Index,Index2;
            for (Index = 0; Index < Collisions; Index++) {
               int Found = 0;
               int Entry = ListPtr[Index];
               for (Index2 = 0; Index2 < Collisions2; Index2++) {
                  if (ListPtr2[Index2] == Entry) {
                     Found = 1;
                     break;
                  }
               }
               if (!Found) {
                  printf ("Entry %d is in first file but not in second\n",
                                                                    Entry);
               }
            }
            for (Index2 = 0; Index2 < Collisions2; Index2++) {
               int Found = 0;
               int Entry = ListPtr2[Index2];
               for (Index = 0; Index < Collisions; Index++) {
                  if (ListPtr[Index] == Entry) {
                     Found = 1;
                     break;
                  }
               }
               if (!Found) {
                  printf ("Entry %d is in second file but not in first\n",
                                                                    Entry);
               }
            }
         }
         (void) fread(&FileIndex,sizeof(FileIndex),1,SavedFd);
         (void) fread(&FileIndex2,sizeof(FileIndex2),1,SavedFd2);
         free (ListPtr);
         free (ListPtr2);
      }
      fclose(SavedFd);
      fclose(SavedFd2);
   }
   return 0;
}
