/*
 *         T c l _ H a s h  T a b l e   -    E m u l a t o r
 *
 *  Function:
 *      Emulates the Tcl hash table routines used by the configure.
 *
 *  Description:
 *      To allow the standalone 'test harness' code used to exercise
 *      the method code used by 'configure' - which we prefer to link
 *      without the Tcl libraries - this file provides an cheap and
 *      cheerful emulation of the Tcl hash table routines used by
 *      'configure'. These don't actually use a hash table at all, they
 *      just do a linear search - for the test harness, this shouldn't
 *      matter.
 *
 *  Author:
 *      Keith Shortridge, AAO.
 *
 * "@(#) $Id: ACMM:aaoConfigure/src/HashEmulator.c,v 12.7 04-Nov-2013 11:28:48+11 tjf $"
 *
 *  History:
 *      24th Feb 2005. Original version, KS.
 *      24th Aug 2006. Need string.h rather then strings.h to get all 
 *              prototypes. Need stdlib.h rather then malloc.h.  TJF.
 */

static const char *rcsId="@(#) $Id: ACMM:aaoConfigure/src/HashEmulator.c,v 12.7 04-Nov-2013 11:28:48+11 tjf $";
static void *use_rcsId = (0 ? (void *)(&use_rcsId) : (void *) &rcsId);

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "HashEmulator.h"

void Tcl_InitHashTable (
   Tcl_HashTable* HashTable,
   int TypeCode)
{
   if (TypeCode != TCL_STRING_KEYS) {
      *HashTable = (Tcl_HashTable) 0;
   } else {
      *HashTable = (Tcl_HashTable) malloc (sizeof(Tcl_HashTableInstance));
      if (*HashTable) {
         int I;
         Tcl_HashTableInstance* HashTablePtr = *HashTable;
         for (I = 0; I < TCL_EM_MAX_ENTRIES; I++) {
            HashTablePtr->EntryList[I].Name[0] = '\0';
         }
         HashTablePtr->HighestEntry = 0;
      }
   }
}

void Tcl_DeleteHashTable (
   Tcl_HashTable* HashTable)
{
   if (*HashTable) free (*HashTable);
}

Tcl_HashEntry* Tcl_CreateHashEntry (
   Tcl_HashTable* HashTable,
   char* Name,
   int* New)
{
   int Found = 0;
   Tcl_HashEntry* Entry = (Tcl_HashEntry*) 0;
   Tcl_HashEntry* EmptyEntry = (Tcl_HashEntry*) 0;
   if (HashTable) {
      int I;
      Tcl_HashTableInstance* HashTablePtr = *HashTable;
      char Char0 = Name[0];
      for (I = 0; I < HashTablePtr->HighestEntry; I++) {
         char HashChar0 = HashTablePtr->EntryList[I].Name[0];
         if (HashChar0) {
            if (HashChar0 == Char0) {
               if (!strcmp(Name,HashTablePtr->EntryList[I].Name)) {
                  Found = 1;
                  Entry = &(HashTablePtr->EntryList[I]);
                  break;
               }
            }
         } else {
            if (!EmptyEntry) EmptyEntry = &(HashTablePtr->EntryList[I]);
         }
      }
      if (!Found) {
         if (!EmptyEntry) {
            if (HashTablePtr->HighestEntry < TCL_EM_MAX_ENTRIES) {
               EmptyEntry = 
                   &(HashTablePtr->EntryList[HashTablePtr->HighestEntry]);
               HashTablePtr->HighestEntry++;
            }
         }
         if (EmptyEntry) {
            strncpy(EmptyEntry->Name,Name,sizeof(EmptyEntry->Name));
            EmptyEntry->Name[sizeof(EmptyEntry->Name) - 1] = '\0';
            EmptyEntry->Value = 0;
            Entry = EmptyEntry;
         }
      }
   }
   *New = (!Found && EmptyEntry);
   return Entry;
}
       
void Tcl_SetHashValue (
   Tcl_HashEntry* Entry,
   int Value)
{
   if (Entry) {
      Entry->Value = Value;
   }
}
              
Tcl_HashEntry* Tcl_FindHashEntry (
   Tcl_HashTable* HashTable,
   char* Name)
{
   Tcl_HashEntry* Entry = (Tcl_HashEntry*) 0;
   if (HashTable) {
      int I;
      char Char0 = Name[0];
      Tcl_HashTableInstance* HashTablePtr = *HashTable;
      for (I = 0; I < HashTablePtr->HighestEntry; I++) {
         char HashChar0 = HashTablePtr->EntryList[I].Name[0];
         if (HashChar0) {
            if (HashChar0 == Char0) {
               if (!strcmp(Name,HashTablePtr->EntryList[I].Name)) {
                  Entry = &(HashTablePtr->EntryList[I]);
                  break;
               }
            }
         }
      }
   }
   return Entry;
}

int Tcl_GetHashValue (
   Tcl_HashEntry* Entry)
{
   int Value = 0;
   if (Entry) Value = Entry->Value;
   return Value;
}

void TclEmListHashTable (        
   Tcl_HashTable* HashTable)
{
   if (*HashTable) {
      int I;
      Tcl_HashTableInstance* HashTablePtr = *HashTable;
      for (I = 0; I < HashTablePtr->HighestEntry; I++) {
         printf ("'%s' : %d\n",
             HashTablePtr->EntryList[I].Name,HashTablePtr->EntryList[I].Value);
      }
   }
}

/*  Test program, only compiled if 'TEST' is defined. */

#ifdef TEST

int main ()
{
   Tcl_HashTable HashTable;
   Tcl_HashEntry* EntryPtr;
   int I;
   int New;
   Tcl_InitHashTable (&HashTable,TCL_STRING_KEYS);
   for (I = 1; I <= 20; I++) {
      char Name[32];
      sprintf (Name,"Entry number %d",I);
      EntryPtr = Tcl_CreateHashEntry (&HashTable,Name,&New);
      Tcl_SetHashValue (EntryPtr,I);
      if (!New) printf ("Error, entry %d not new\n",I);
      if (!EntryPtr) printf ("Error, no entry %d made\n",I);
   }
   for (I = 20; I >= 1; I--) {
      char Name[32];
      sprintf (Name,"Entry number %d",I);
      EntryPtr = Tcl_CreateHashEntry (&HashTable,Name,&New);
      Tcl_SetHashValue (EntryPtr,I * 100);
      if (New) printf ("Error, entry %d new on second pass\n",I);
      if (!EntryPtr) printf ("Error, no entry %d on second pass\n",I);
   }
   TclEmListHashTable(&HashTable);
   for (I = 1; I <= 20; I++) {
      char Name[32];
      sprintf (Name,"Entry number %d",I);
      EntryPtr = Tcl_FindHashEntry (&HashTable,Name);
      if (!EntryPtr) {
         printf ("Error, no entry %d found\n",I);
      } else {
         int Value = Tcl_GetHashValue(EntryPtr);
         if (Value != (I * 100)) {
            printf ("Error, Entry %d has value %d, not %d\n",I,Value,I * 100);
         }
      }
   }
   EntryPtr = Tcl_FindHashEntry (&HashTable,"Dummy");
   if (EntryPtr) printf ("Error, non-existent entry found\n");
   Tcl_DeleteHashTable (&HashTable);

   return 0;
}

#endif        
