/*+
 *               C O N F I G _ M E T H O D . A N N E A L
 *
 *  Function:
 *     A simulated annealing approach to 2dF field configuration.
 *
 *  Description:
 *     This file contains a set of routines that between them provide the
 *     default fibre instrument configuration algorithm. These are intended 
 *     to be called from the 'housekeeping and interface' (config.c) 
 *     layer of routines described in the document 'Fibre Instrument 
 *     Configuration Algorithm Routines'. 
 *     
 *     These routines constitute the new simulated annealing based field
 *     configuration algorithm (FCA) written by Brent Miszalski. This work was for his 
 *     Honours thesis with Macquarie University (MU) under a recently established
 *     Honours scholarship scheme with MU to encourage research in instrumentation
 *     science. The algorithm is described in depth in the following references.
 *     
 *     The two main components are the ConfigMethodAllocate and ConfigMethodEnergy 
 *     functions. The former constitutes the annealing schedule which calls the latter
 *     to determine the energy or merit of allocations proposed by the annealing schedule.
 *     The annealing schedule then uses this information, in conjunction with the Metropolis
 *     algorithm (Metropolis et al. 1953) to guide the algorithm towards an optimal solution
 *     in a probabilistic fashion.
 *     
 *     The SA FCA is strongly dependent on the allocation sub-system written by KS
 *     and BM. The allocation sub-system implements a novel collision matrix
 *     that enables orders of magnitudes greater swaps to occur than previously
 *     possible. This enables a much larger extent of the parameter space of the problem
 *     to be explored by the Metropolis algorithm.
 *
 *     Further information is provided in the references below:
 *
 *  References:   Miszalski, B. 2005, Honours thesis, Macquarie University,
 *                Soon to be in AAO Epping library,
 *                Digital Copy: http://www.aao.gov.au/local/www/brent/thesis.ps.gz
 *                
 *   README FIRST Miszalski et al. 2006, AAO Newsletter 109, page 24,
 *                A new field configuration algorithm for AAOmega. 
 *
 *                Miszalski et al. 2006, submitted, MNRAS,
 *                Digital Copy: http://www.aao.gov.au/local/www/brent/MNRAS_SA.ps.gz
 *                
 *  Version date: 16th September 2013.
 *
 *  Support: K. Shortridge, AAO.
 *           B. Miszalski, Macquarie University/AAO.
 *-
 *  "@(#) $Id: ACMM:aaoConfigure/src/config_method.anneal.c,v 12.7 04-Nov-2013 11:28:47+11 tjf $"
 *
 *  History:
 *     10th Mar 2005.  Original version as config_method.template.c. KS/AAO.
 *     21st Feb 2006.  Brent modified the original skeleton code to produce
 *                     the original simulated annealing code as used for his
 *                     thesis work. This is now being fed back into the main
 *                     code tree as config_method.anneal.c. KS/AAO.
 *     21st Feb 2006.  Removed all TAB characters. ConfigMethodMatrix.h now
 *                     explicitly included - I think this makes things 
 *                     clearer. KS/AAO.
 *     22nd Feb 2006.  Added explicit code to release all memory arrays before
 *                     starting a new allocation - fixing memory leaks when
 *                     repeated allocations are performed. Added new routine
 *                     ConfigMethodUpdate() to provide feedback to GUI.
 *                     Attempt to linearise percentage complete during 
 *                     annealing by using estimated iterations. KS/AAO.
 *     25th Feb 2006.  Major revision for upcoming Configure release.
 *                     This includes:
 *                     
 *                     -Much improved comments here and there...
 *                     
 *                     -Objective function parameter abstraction:
 *                     The current algorithm parameters are sufficient
 *                     for most fields. Three different levels of 
 *                     annealing are provided: quick, standard and 
 *                     thorough. The parameters defined for each level
 *                     are essentially guesses of what each should be. 
 *                     The main point of difference is choice of MaxIters
 *                     and dT. The advantage of this approach is that it
 *                     is very simple for users. This could backfire for
 *                     more advanded users (e.g. Will Saunders once he becomes
 *                     familiar with the algorithm parameters) that could 
 *                     be catered for in future versions by having 
 *                     all parameters available for modification 
 *                     in expert mode. Such a feature would
 *                     require some of the current parameters to be 
 *                     non-existant in expert mode. This is currently
 *                     not possible within the existing parameter framework,
 *                     as such it has not been implemented. This may be 
 *                     necessary after an initial release of the new configure.
 *
 *                     -Added collision matrix filename parameter.
 *                     This may be useful for users that are configuring
 *                     different fields from one configure install.
 *                     TODO: Use field.matrix for field.fld instead of 
 *                     collision.matrix (current default).
 *                     NB: To get the relevant parameter to work properly
 *                     a two calls to free(*.StringAddress) for NValues == 0
 *                     were commented (in ConfigMethodSetParam and 
 *                     ConfigMethodReleaseMem - though the latter may not be needed).
 *                     Further work was required in configure.tcl to enable
 *                     such 'C' parameters with NValues == 0. This was necessary
 *                     because such parameters were previously unused in configure.
 *
 *                     -Realised the close-pair separation calculation was fundamentally flawed
 *                     for real-world observers. Should replace X,Y distance with proper Ra,Dec
 *                     distance. Would need input of Scott Croom?
 *                     Maybe we just ASSUME that this will give the desired result?
 *
 *     28th Feb 2006.  Fixed issues with default matrix file name. Added code
 *                     for a progress report that's linear with time, as 
 *                     opposed to linear with progress. Which to use is a
 *                     matter of taste. KS/AAO.
 *      2nd Mar 2006.  Fixed problem with reporting of allocations in the
 *                     case where tragets had been pre-allocated. Updates 
 *                     were being sent too often - fixed update time code
 *                     so it only reports progress twice a second and
 *                     makes a full allocation update every two seconds. KS.
 *      6th Mar 2006.  New version of code obtained from BM that has support
 *                     for CBS parameters. My only change was to add this
 *                     comment. KS.
 *     16th May 2006.  Now supports a limit on the number of allocations
 *                     considered by the code. Reworked code in a couple of
 *                     places to eliminate compile warnings. KS.
 *     15th Jun 2006.  MaxAllocations is now expressed as a percentage. KS.
 *      7th Jul 2006.  Corrected code for fiducial edge weighting. KS.
 *      1st Nov 2006.  Added call to AllocationFinish(). KS.
 *      5th Nov 2006.  Added MaxThreads and 'On-the-fly' parameters. KS.
 *      7th Nov 2006.  Default for 'On-the-fly' parameter now false. KS.
 *     30th Nov 2006.  Get number of processors correctly on MacOsX (using 
 *                     sysctlbyname, sysconf does not work).  Handle other
 *                     cases where _SC_NPROCESSORS_ONLN is not defined.  Allow
 *                     the environment variable CONFIGURE_ONE_CPU_ONLY to turn
 *                     use of multiple CPU's off and output a message to stdout
 *                     indicating the number of processors that were found. TJF.
 *     27th Aug 2007.  Changed defaults: default straightening is now moderate,
 *                     instead of never, and 'weight peripheral fiducials' and
 *                     'enforce sky quota' are now true. KS.
 *      5th Sep 2007.  Some changes to support the concept of 'locked' fibres.
 *                     The initial calculation of the matrix now assumes that
 *                     any existing allocations must be valid - this was
 *                     causing problems with testing, where allocations were
 *                     made by hand with more tolerant values for the bend
 *                     angles, and then were being flagged as impossible in
 *                     the matrix. The use of AllowAllocChange has changed 
 *                     and is now incorporated in the annealing code, so that
 *                     the annealing leaves locked fibres alone, rather than
 *                     simply setting the ignore flag in the matrix for all
 *                     allocations involving such fibres, which was not having
 *                     the required effect. KS.
 *       9th Feb 2009. Modified the description of the algorithm in the
 *                     name string, now that this is likely to be displayed
 *                     by the GUI. KS.
 *      19th Aug 2009. The invalid position routines (*ColInvPos) now take
 *                     a fibre type argument. KS.
 *      16th Sep 2013. Now takes into account affectation data (crosstalk and
 *                     ghost data to try to minimise the effects of having
 *                     data from bright targets spill over into fibres being
 *                     used for fainter targets. KS.
 *      17th Sep 2013. No checks to see if affectation data is available, and
 *                     modifies the parameter name and value if not in order
 *                     to indicate this in the allocation dialogue. KS.
 */
 
/*                            O v e r v i e w
 *
 *  The first routine called from outside will be ConfigMethodDetails().
 *  In this we set up the fields that will allow the rest of the algorithm
 *  routines to be called. We have to set up:
 *
 *  GlobalInit : A global initialisation routine, called just once to
 *               perform necessary one-off initialisation.
 *               We set this to ConfigMethodGlobalInit.
 *  ReleaseMem : A close down routine, called when the program is to be
 *               shut down. We set this to ConfigMethodReleaseMem.
 *  Init       : A routine, called when the upper layers have the 
 *               field plate details (pivot positions etc) and the target
 *               details (position, types etc). We set this to 
 *               ConfigMethodInit.
 *  Allocate   : The routine called when the upper layers want an allocation
 *               to be performed. We set this to ConfigMethodAllocate.
 *
 *  There are also a number of utility routines connected mainly with
 *  parameters that we set to a set of routines whose code is taken directly
 *  from the original config_method.c code. These are all driven by
 *  tables set up a the start of this file. These tables differ from
 *  algorithm to algorithm, but the code to handle the parameters can be
 *  the same for all algorithms.
 *
 *  There are also two routines used by the upper levels to control some
 *  details of the allocation. In this file we supply dummy versions.
 *  
 *  SetSky        : A routine called when the upper layers want to explicitly
 *                  control the allocation of sky targets to different types
 *                  of fibre. We set this to ConfigMethodSetSky.
 *  SetImportFile : A routine used to make the algorithm read in a set of
 *                  pre-determined allocations from a text file.
 * 
 *  Once those are set up, the upper layers will next call the GlobalInit
 *  routine, here ConfigMethodGlobalInit.  In this code, that routine
 *  does a bit of general housekeeping, mainly connected with the default
 *  values for the parameters.
 *
 *  The upper layers may then call some of the parameter routines, but the
 *  next important call will be to the Init routine when the plate details 
 *  are known. Here we use this call to set up a matrix of pivots versus
 *  targets, giving details of whether each possible pivot/target allocation
 *  is possible and if so some additional information such as the fibre
 *  theta angle for the allocation.
 *
 *  The next call will be to the Allocate routine. Here we don't do anything
 *  clever. We just go through each pivot in turn and allocate it to the
 *  first possible target. When we make an allocation, we update the
 *  matrix so that any allocations shown in there as possible but which 
 *  would now collide with the newly allocated fibre are now no longer
 *  flagged as possible. When we make an allocation we make a callback
 *  to the upper layer's 'NoteAllocation' routine so that the allocation
 *  can be recorded.
 *
 *  And that's it.
 *  
 *  BM: For an overview of the algorithm itself see the AAO Newsletter article
 *  referenced in the header of this file.
 */

/* -------------------------------------------------------------------------- */

/*                       I n c l u d e  F i l e s
 */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/time.h>
#include <unistd.h>
#ifdef __APPLE__
#include <sys/sysctl.h>
#endif

/* THESE SHOULD BE OBTAINED FROM THE PROPER SPOT 
 * SO THAT IT IS INSTRUMENT INDEPENDENT, NOT JUST 2DF
 * N.B. The only thing in the SA FCA that uses 
 * this info is the detection of
 * close pairs. Please ask me about this Keith. BM.
 * This is also dep. on 1.05, field rad in degrees
 * of the 2dF field, again - not good.
 */
#define FIELD_RAD        253000  
#define FIELD_RAD_INC_PARK 285000

/*  DRAMA include files, only status.h needed */

#include "status.h"

/* Allocation Sub-System - required by SA FCA*/
#include "AllocSubSystem.h"

#include "configure_subs.h"

/*
 * Prototypes for the random number generator from dprand.c that we use,
 * if we are going to use one. This makes sure we get the same results on
 * all platforms.
 */
 
extern void sdprand (int seed);
extern double dprand (void);

/*
 * DUNUSED Flag which under GCC, prevents warnings about an argument being
 * unused - should be added where we know it will be unused
 */
 
#ifndef DUNUSED
#if defined(__GNUC__) || defined(GNUC)
#   define DUNUSED /* Default definition */
#   ifdef __GNUC_MINOR__
#   if __GNUC_MINOR__ >= 7
#       undef  DUNUSED
#       define DUNUSED __attribute__ ((unused))
#   endif
#   endif
#else
#   define DUNUSED /* */
#endif
#endif /* DUNUSED */


/*  Include files specific to CONFIG itself */

#include "tdFconfigMethod.h"
#include "config_err.h"

/*  Used to highlight debugging lines */

#ifdef DEBUG
#undef DEBUG
#endif
#define DEBUG printf

static int DEBUG_AffectCalcs = 0;
static double DEBUG_TotalAffect = 0.0;
static double DEBUG_MaxAffect = 0.0;
static double DEBUG_MinAffect = 0.0;
static double DEBUG_TotalPri = 0.0;
static double DEBUG_MaxPri = 0.0;
static double DEBUG_MinPri = 0.0;


/* -------------------------------------------------------------------------- */

/*            M e t h o d - s p e c i f ic  d e c l a r a t i o n s
 *
 *  'Method' in this case refers to anything specific to the actual allocation
 *  used.
 */

/*  Note that there are a set of codes used in these routines that all have
 *  names like CONFIG_METHOD_POSSIBLE, CONFIG_METHOD_ALLOCATED, etc, that
 *  describe the possible states for the allocation of a given fibre to a
 *  given target. These used to be declared in this file, but because the
 *  allocation sub-system uses them as well, they have been moved to the
 *  file ConfigMethodMatrix.h.
 */

#include "ConfigMethodMatrix.h"

/*  Method parameter constants. Note that having these statically defined
 *  like this makes it very easy to implement ConfigMethodParamDetails(),
 *  and also helps to code ConfigMethodSetParam(). However, it is important
 *  that the information held here is consistent - each parameter has a
 *  symbolic name that equates to an integer index, and the entries in
 *  the ConfigMethodParamXxxx arrays must match those indices. It isn't quite
 *  possible to make the whole parameter business 'table-driven' (controlled
 *  entirely by the entries in here). Some explicit code is needed, but the 
 *  more we can do from the definitions here the easier the code is to modify.
 *  At present, to add a new parameter, 1) Increase the value of CONFIG_METHOD-
 *  _NPARMS, 2) Define a new symbolic value for its index (a new parameter
 *  called CONFIG_METHOD_xxxx_PARM, 3) Add new entries to all the CONFIG_METHOD
 *  _ParmXxxx tables. 4) For a character parameter with a fixed set of possible
 *  values, define a new array of the possible values (see CONFIG_METHOD_-
 *  SelectChoices) and symbolic indices for the values, 5) Add initialisation
 *  code to ConfigMethodGlobalInit(), 6) Add whatever code is needed to make
 *  use of the new parameter.
 */
/*  This code defines two parameters for the template allocation algorithm,
 *  purely for demonstration purposes. The first is an integer parameter
 *  which is used to set the number of a pivot that is not be allocated.
 *  The template code will deallocate this if it allocates it - this is
 *  in purely to exercise the deallocation code, and in passing it shows how
 *  to use a parameter. The second parameter is a character string that can
 *  have only one of a limited set of values, and is included here simply
 *  to demonstrate how to set up such a string - it isn't used by the code.
 */
 
#define CONFIG_METHOD_NPARMS 15           /*  Number of parameters supported */

/*  A array of structures like this goes into the method-specific global
 *  variables. For each parameter, there is a double precision value (used
 *  only for floating point parameters) and an integer value. For logical 
 *  parameters the integer value will be 0 (false) or 1 (true). For integer
 *  parameters the integer value is that of the parameter. For character
 *  parameters with an unlimited number of possible values, when the parameter
 *  is set to a new value an array is allocated in memory (using malloc())
 *  for a static copy of that new value string, and the address of this
 *  allocated string is held in common. For character parameters with a
 *  limited set of possible values, all possible values are held in a static
 *  array and the address of the value currently set is held in common. In
 *  this case, the integer value is also used to hold the index number of
 *  the current value being used.
 */

typedef struct CONFIG_METHOD_ParamValue {
   int IntegerValue;                         /* Integer value - see above */
   char *StringAddress;                      /* Address of string */
   double DoubleValue;                       /* Floating point value */
} CONFIG_METHOD_ParamValue;

/* Choose one of three levels of `thoroughness'
 * for the randomisation of the annealing schedule
 */

#define CONFIG_METHOD_ANNEALING 0
   #define CONFIG_METHOD_ANNEALING_QUICK 0
   #define CONFIG_METHOD_ANNEALING_STANDARD 1
   #define CONFIG_METHOD_ANNEALING_THOROUGH 2

/* Enables close pair weighting constraint.
 * Pairs with separations THETA in the range
 * CONFIG_METHOD_THETA_MIN <= THETA <= CONFIG_METHOD_THETA_MAX
 * are weighted.
 */

#define CONFIG_METHOD_CLOSE_PAIRS 1
#define CONFIG_METHOD_THETA_MIN 2
#define CONFIG_METHOD_THETA_MAX 3

/* Enables weighting of CBS pairs included in .fld as
 * target-cbspair for each target named `target'.
 */ 

#define CONFIG_METHOD_CROSS_BEAM_SWITCHING 4

/* Enables constraint on fibre straightness.
 * Fibre straightness is achieved at the expense of target
 * priority yield (more lower priority targets allocated).
 * Also less CBS pairs may be achieved in this mode.
 */ 

#define CONFIG_METHOD_STRAIGHTEN_FIBRES 5
   #define CONFIG_METHOD_STRAIGHTEN_FIBRES_NEVER 0
   #define CONFIG_METHOD_STRAIGHTEN_FIBRES_MODERATE 1
   #define CONFIG_METHOD_STRAIGHTEN_FIBRES_HEAVY 2

/* Collision matrix filename. Text entry parameter */

#define CONFIG_METHOD_COLLISION_MATRIX_FNAME 6

/* Enforce sky quota */

#define CONFIG_METHOD_ENFORCE_SKY_QUOTA 7

/* Weight fiducial targets near edge */

#define CONFIG_METHOD_FID_EDGE 8

/* Weight fiducial target pair */

#define CONFIG_METHOD_FID_PAIRS 9


/* An expert parameter that determines the
 * seed used for PRNG dprand()
 */

#define CONFIG_METHOD_RANDOM_SEED 10

/*  An expert parameter that controls the maximum number of 
 *  allocations considered by the code.
 */
 
#define CONFIG_METHOD_MAX_ALLOCATIONS 11

/*  A parameter that controls the maximum number of 
 *  background threads used by the code.
 */
 
#define CONFIG_METHOD_MAX_THREADS 12

/*  A parameter that controls whether incompatible allocations
 *  should be calculated 'on-the-fly' or all precalculated before
 *  annealing starts.
 */
 
#define CONFIG_METHOD_ON_THE_FLY 13

/*  A parameter that controls the 'affectation' of fainter targets by
 *  having brighter targets allocated in such a way that the light from
 *  the brighter spills over (crossover, ghosting, etc) onto the 
 *  fainter. The details of this parameter are still experimental.***
 */

#define CONFIG_METHOD_AFFECTATION 14

/*  
 *  The following arrays contain the parameter details as used by
 *  ConfigMethodParamDetails() and ConfigMethodSetParam().
 *  CONFIG_METHOD_ParmNames   lists the name of the parameters.
 *  CONFIG_METHOD_ParmTypes   lists their types (I,C,F or L).
 *  CONFIG_METHOD_ParmLimits  lists their limits, if they are numeric
 *  CONFIG_METHOD_ParmNValues lists their restricted number of possible
 *                            string values, if they have such a restriction.
 *  CONFIG_METHOD_ParmValues  lists those possible string values, if they
 *                            have them.
 *
 *  Note that the affectation parameter name can be modified during the
 *  initialisation stage if there is no affectation data available.
 */

static char *CONFIG_METHOD_ParmNames[CONFIG_METHOD_NPARMS] = {
      "Annealing",                          /* CONFIG_METHOD_ANNEALING */
      "Weight Close Pairs",                 /* CONFIG_METHOD_CLOSE_PAIRS */
      "ThetaMin (arcsec)",                  /* CONFIG_METHOD_THETA_MIN */
      "ThetaMax (arcsec)",                  /* CONFIG_METHOD_THETA_MAX */      
      "CrossBeamSwitching",                 /* CONFIG_METHOD_CROSS_BEAM_SWITCHING */
      "Straighten Fibres",                  /* CONFIG_METHOD_STRAIGHTEN_FIBRES */
      "Collision matrix filename",          /* CONFIG_METHOD_COLLISION_MATRIX_FNAME */
      "Enforce sky quota",                  /* CONFIG_METHOD_ENFORCE_SKY_QUOTA */
      "Weight peripheral fiducial targets", /* CONFIG_METHOD_FID_EDGE */
      "Weight fiducial target pairs",              /* CONFIG_METHOD_FID_PAIRS */
      "Random Seed (expert)",               /* CONFIG_METHOD_RANDOM_SEED */
      "Percentage of allocations sampled (expert)",
                                            /* CONFIG_METHOD_MAX_ALLOCATIONS */
      "Number of background threads to use",/* CONFIG_METHOD_MAX_THREADS */
      "On-the-fly collision calculation",   /* CONFIG_METHOD_ON_THE_FLY */
      "Affectation parameter"               /* CONFIG_METHOD_AFFECTATION */
    };
static char *CONFIG_METHOD_AnnealingChoices[] = {
        "Quick",                            /* CONFIG_METHOD_ANNEALING_QUICK */
        "Standard",                         /* CONFIG_METHOD_ANNEALING_STANDARD */
        "Thorough"                          /* CONFIG_METHOD_ANNEALING_THOROUGH */
};
static char *CONFIG_METHOD_StraightenFibresChoices[] = {
        "never",                            /* CONFIG_METHOD_STRAIGHTEN_FIBRES_NEVER */
        "moderate",                         /* CONFIG_METHOD_STRAIGHTEN_FIBRES_MODERATE */
        "heavy"                             /* CONFIG_METHOD_STRAIGHTEN_FIBRES_HEAVY */
};
static char CONFIG_METHOD_ParmTypes[CONFIG_METHOD_NPARMS] = {
      'C',                                  /* CONFIG_METHOD_ANNEALING */
      'L',                                  /* CONFIG_METHOD_CLOSE_PAIRS */
      'I',                                  /* CONFIG_METHOD_THETA_MIN */
      'I',                                  /* CONFIG_METHOD_THETA_MAX */      
      'L',                                  /* CONFIG_METHOD_CROSS_BEAM_SWITCHING */
      'C',                                  /* CONFIG_METHOD_STRAIGHTEN_FIBRES */
      'C',                                  /* CONFIG_METHOD_COLLISION_MATRIX_FNAME */
      'L',                                  /* CONFIG_METHOD_ENFORCE_SKY_QUOTA */
      'L',                                  /* CONFIG_METHOD_FID_EDGE */
      'L',                                  /* CONFIG_METHOD_FID_PAIRS */
      'I',                                  /* CONFIG_METHOD_RANDOM_SEED */
      'I',                                  /* CONFIG_METHOD_MAX_ALLOCATIONS */
      'I',                                  /* CONFIG_METHOD_MAX_THREADS */
      'L',                                  /* CONFIG_METHOD_ON_THE_FLY */
      'F'                                   /* CONFIG_METHOD_AFFECTATION */
    };
/* NEED TO FIX UP THESE RANGES - THIS IS A BIT COARSE */
static double CONFIG_METHOD_ParmLimits[CONFIG_METHOD_NPARMS][2] = {
    {0.0,0.0},                               /* CONFIG_METHOD_ANNEALING */
    {0,0.0},                                 /* CONFIG_METHOD_CLOSE_PAIRS */
    {0,10000},                               /* CONFIG_METHOD_THETA_MIN */
    {0,10000},                               /* CONFIG_METHOD_THETA_MAX */      
    {0,0.0},                                 /* CONFIG_METHOD_CROSS_BEAM_SWITCHING */
    {0,0.0},                                 /* CONFIG_METHOD_STRAIGHTEN_FIBRES */
    {0,0.0},                                 /* CONFIG_METHOD_COLLISION_MATRIX_FNAME */
    {0,0.0},                                 /* CONFIG_METHOD_ENFORCE_SKY_QUOTA */
    {0,0.0},                                 /* CONFIG_METHOD_FID_EDGE */
    {0,0.0},                                 /* CONFIG_METHOD_FID_PAIRS */
    {0,10000},                               /* CONFIG_METHOD_RANDOM_SEED */
    {0,100},                                 /* CONFIG_METHOD_MAX_ALLOCATIONS */
    {0,3},                                   /* CONFIG_METHOD_MAX_THREADS */
    {0,0.0},                                 /* CONFIG_METHOD_ON_THE_FLY */
    {0.0,500.0}                              /* CONFIG_METHOD_AFFECTATION */
    }; 

static int CONFIG_METHOD_ParmNValues[CONFIG_METHOD_NPARMS] ={
      3,                                     /* CONFIG_METHOD_ANNEALING */
      0,                                     /* CONFIG_METHOD_CLOSE_PAIRS */
      0,                                     /* CONFIG_METHOD_THETA_MIN */
      0,                                     /* CONFIG_METHOD_THETA_MAX */      
      0,                                     /* CONFIG_METHOD_CROSS_BEAM_SWITCHING */
      3,                                     /* CONFIG_METHOD_STRAIGHTEN_FIBRES */
      0,                                     /* CONFIG_METHOD_COLLISION_MATRIX_FNAME */
      0,                                     /* CONFIG_METHOD_ENFORCE_SKY_QUOTA */
      0,                                     /* CONFIG_METHOD_FID_EDGE */
      0,                                     /* CONFIG_METHOD_FID_PAIRS */
      0,                                     /* CONFIG_METHOD_RANDOM_SEED */      
      0,                                     /* CONFIG_METHOD_MAX_ALLOCATIONS */
      0,                                     /* CONFIG_METHOD_MAX_THREADS */
      0,                                     /* CONFIG_METHOD_ON_THE_FLY */
      0                                      /* CONFIG_METHOD_AFFECTATION */
    };

static char **CONFIG_METHOD_ParmValues[CONFIG_METHOD_NPARMS] = {
      CONFIG_METHOD_AnnealingChoices,         /* CONFIG_METHOD_ANNEALING */                        
      0,                                      /* CONFIG_METHOD_CLOSE_PAIRS */
      0,                                      /* CONFIG_METHOD_THETA_MIN */
      0,                                      /* CONFIG_METHOD_THETA_MAX */      
      0,                                      /* CONFIG_METHOD_CROSS_BEAM_SWITCHING */
      CONFIG_METHOD_StraightenFibresChoices,  /* CONFIG_METHOD_STRAIGHTEN_FIBRES */
      0,                                      /* CONFIG_METHOD_COLLISION_MATRIX_FNAME */
      0,                                      /* CONFIG_METHOD_ENFORCE_SKY_QUOTA */
      0,                                      /* CONFIG_METHOD_FID_EDGE */
      0,                                      /* CONFIG_METHOD_FID_PAIRS */
      0,                                      /* CONFIG_METHOD_RANDOM_SEED */
      0,                                      /* CONFIG_METHOD_MAX_ALLOCATIONS */
      0,                                      /* CONFIG_METHOD_MAX_THREADS */
      0,                                      /* CONFIG_METHOD_ON_THE_FLY */
      0                                       /* CONFIG_METHOD_AFFECTATION */
    };
    
/*  The default value used for the parameter shown here as
 *  CONFIG_METHOD_COLLISION_MATRIX_FNAME - the collision matrix file name.
 */
 
#define CONFIG_METHOD_DEFAULT_FNAME "collision.matrix"

/* -------------------------------------------------------------------------- */

/*  The following defines any global variables used by the method-specific
 *  code. This defines a single global variable, called CONFIG_METHOD_Global
 *  which is a structure containing all the global information needed by
 *  the allocation algorithm. This makes it clear when we look at the code
 *  when a global variable is being used - it will be in the structure
 *  called CONFIG_METHOD_Global.
 *
 *  At the moment, this contains the matrix used by the allocation
 *  algorithm (which requires two arrays, one the 2D array of matrix elements,
 *  one the 1D array that holds the addresses of the matrix element at
 *  the start of each array (this is Matrix, and having it makes it possible
 *  to use a dynamically allocated array as if it were a statically
 *  declared 2D array - a trick used, for example, by the C version of
 *  Numerical Recipies).
 *
 *  It also contains the parameter values.
 */

typedef struct CONFIG_METHOD_GlobalType {
   CONFIG_METHOD_MatrixElement **Matrix;       
                                 /* Address of memory allocated to hold addrs
                                  * of starts of rows of allocation matrix. */
   CONFIG_METHOD_MatrixElement *MatrixAddr;
                                 /* Address of memory allocated for matrix */
   CONFIG_METHOD_MatrixElement **PhysicalMatrix;       
                                 /* Address of memory allocated to hold addrs
                                  * of starts of rows of allocation matrix. */
   CONFIG_METHOD_MatrixElement *PhysicalMatrixAddr;
                                 /* Address of memory allocated for matrix */
   CONFIG_METHOD_ParamValue ParameterValues[CONFIG_METHOD_NPARMS];
                                 /* Current parameter values */
   int SkyToAlloc;               /* Number of sky targets to allocate */
   int** InAssoc;                /* List of targets that surround each target
                                  * within a certain distance 
                                  */  
   int InAssocSize;              /* The number of elements in InAssoc, 
                                  * remembered so it can be released properly.
                                  */
   double AlphaMax;              /* Maximum Non-radial pivot angle */
   double** FibreAlphas;         /* Pre-calculated Non-radial pivot angles 
                                  * for all possible allocations 
                                  */
   int FibreAlphaSize;           /* The number of elements in FibreAlphas, 
                                  * remembered so it can be released properly.
                                  */
   int* NAssoc;                  /* The number of targets reachable by each target,
                                  * used in conjunction with InAssoc.
                                  */
   int* NAffectedFibres;         /* The number of fibres affected by light (ghosts,
                                  * crosstalk etc) from each fibre. This is an 
                                  * array with one element for each pivot and
                                  * gives the number of entries in the corresponding
                                  * entry in AffectedFibres and AffectedWeights.
                                  */
   int** AffectedFibres;         /* Set of arrays, one for each pivot, giving the
                                  * number of the pivots affected by light from
                                  * the target assigned to that pivot.
                                  */
   float** AffectedWeights;      /* Set of arrays, one for each pivot, giving the
                                  * degree to which the light from the target 
                                  * assigned to that pivot affects the pivot listed
                                  * in the corresponding entry in AffectedFibres.
                                  */
   double PriWeights[9];         /* Pre-calculated Beta^Priority */
   int** InReach;                /* A list of targets that each pivot can reach */
   int InReachSize;              /* The number of elements in InReach */   
   int* NReach;                  /* The number of targets that each pivot can reach
                                  * used in conjunction with InReach 
                                  */
   double* TargetRadial;         /* Pre-calculated target distance from the centre */
   int* ReportedAllocations;     /* Array used to record the last allocations as
                                  * reported to the housekeeping layers. This
                                  * allows changges to the allocations to be
                                  * reported as batches, without needing individual
                                  * allocations to be reported - which is too
                                  * much for an algorithm that does a lot of
                                  * tentative allocation & deallocation.
                                  */


} CONFIG_METHOD_GlobalType;

 
CONFIG_METHOD_GlobalType CONFIG_METHOD_Global;
                                 /* Actual global structure used */
                                 
                             
/* -------------------------------------------------------------------------- */

/*            I n t e r n a l  R o u t i n e  D e c l a r a t i o n s
 *
 *  This is the point to declare any internal routines. They should all
 *  be static. (Actually, all routines in this file are static, since even
 *  those called by the housekeeping layer are called through the structure
 *  set up by ConfigMethodDetails).
 */

static double ConfigMethodEnergy(const CONFIG_MethodInfo* MethodInfo,
                                                         int IPiv,int ITarget);
static void ConfigMethodUpdate (const CONFIG_MethodInfo *MethodInfo,
                                                           StatusType *Status);
   
static void ConfigMethodGetParam (const CONFIG_MethodInfo *MethodInfo,
                       char *Name, char *Value, int VSize, StatusType *Status);

static void ConfigMethodCheckAffected (const CONFIG_MethodInfo *MethodInfo);

/* -------------------------------------------------------------------------- */

/*              C o n f i g  M e t h o d  G l o b a l  I n i t
 *
 *  Function:
 *     Performs any global initialisation reqired for allocation.
 *
 *  Description:
 *     This routine, which is allocation algorithm specific, is called
 *     just once, during program startup, to allow the algorithm to perform 
 *     any necessary initialisation of its global structures.
 */

static void ConfigMethodGlobalInit (const CONFIG_MethodInfo *MethodInfo DUNUSED)
{
   
   /*  Local variables  */
   
   short AffectKnown;                 /* True if affectation data available */
   int Index;                         /* General loop index */
   int Size;                          /* Length of default parameter strings */
   char* StringAddress;               /* Used for default parameter strings */
   int Nprocessors;                   /* Number of processors available */
      
   /*  At the moment, all we do is initialise the parameters. However, one thing we
    *  want to do is see if we have affectation (crosstalk, ghost) data available,
    *  because if not we will modify the parameter details to make this clearer in
    *  the allocation dialogues. We try to get the affectation data for each pivot
    *  (we don't need it all, so dummy 10 long arrays will do) and if we get any,
    *  then clearly it's known.
    */
   
   AffectKnown = CONFIG_FALSE;
   for (Index = 0; Index < MethodInfo->NumberPivots; Index++) {
      int NAff = 0;
      int Affected[10];
      float Weights[10];
      MethodInfo->FibreAffectData(MethodInfo,MethodInfo->Field,Index + 1,
                                               10,Affected,Weights,&NAff);
      if (NAff > 0) {
         AffectKnown = CONFIG_TRUE;
         break;
      }
   }
   if (!AffectKnown) {
      CONFIG_METHOD_ParmNames[CONFIG_METHOD_AFFECTATION] = "(No affectation data available)";
   }

   /*  If there are any 'any value' character parameters, set their addresses
    *  to null. Because if any new values are set, any memory already allocated
    *  to such a parameter will be freed.
    */

   for (Index = 0; Index < CONFIG_METHOD_NPARMS; Index++) {
      if (CONFIG_METHOD_ParmTypes[Index] == 'C') {
         if (CONFIG_METHOD_ParmNValues[Index] == 0) {
            CONFIG_METHOD_Global.ParameterValues[Index].StringAddress = NULL;
         }
      }
   }
   
   /*  We need to know how many processors we have available so we can 
    *  set up the limits and sensible default for the number of background
    *  threads.  The way we do this depends on the
    *  machine.
    */
#ifdef __APPLE__
   {
       int     count ;
       size_t  size=sizeof(count) ;
       
       if (sysctlbyname("hw.ncpu",&count,&size,NULL,0))
           Nprocessors = 1;
       else
           Nprocessors = count; 
   }
#elif !defined(_SC_NPROCESSORS_ONLN)
#  warning "Don't know how to work out the number of processors on this machine"
   Nprocessors = 1;
#else
   Nprocessors = sysconf(_SC_NPROCESSORS_ONLN);
#endif
   if (Nprocessors <= 0) Nprocessors = 1;
   if (getenv("CONFIGURE_ONE_CPU_ONLY"))
   {
       printf("Annealing:Configure will use one CPU out of %d due to env variable\n",
           Nprocessors);
       Nprocessors = 1;
   }
   else if (Nprocessors == 1)
       printf("Annealing:Configure will use 1 CPU\n");
   else
       printf("Annealing:Configure will use %d CPUs\n", Nprocessors);

   CONFIG_METHOD_ParmLimits[CONFIG_METHOD_MAX_THREADS][1] = Nprocessors - 1;
   
   /*  The parameter values have to be initialised here, since they may never
    *  be changed at all and so need sensible default values. This section
    *  has to be modified if a new parameter is introduced. Note that 
    *  limited range character parameters have to have both their string
    *  addresses and integer values set. Logical and integer parameters need
    *  their integer values set, floating point values need their double
    *  value set, and character parameters that can take any value need to
    *  have their string addresses either left at NULL (from the previous loop)
    *  or set to point to a static string used as a default value.
    */

     CONFIG_METHOD_Global.ParameterValues[CONFIG_METHOD_ANNEALING].IntegerValue = CONFIG_METHOD_ANNEALING_STANDARD;
     CONFIG_METHOD_Global.ParameterValues[CONFIG_METHOD_ANNEALING].StringAddress = CONFIG_METHOD_AnnealingChoices[CONFIG_METHOD_ANNEALING_STANDARD];
     
     CONFIG_METHOD_Global.ParameterValues[CONFIG_METHOD_CLOSE_PAIRS].IntegerValue = CONFIG_FALSE;
     CONFIG_METHOD_Global.ParameterValues[CONFIG_METHOD_THETA_MIN].IntegerValue = 0;   
     CONFIG_METHOD_Global.ParameterValues[CONFIG_METHOD_THETA_MAX].IntegerValue = 60;  
     
    
     CONFIG_METHOD_Global.ParameterValues[CONFIG_METHOD_STRAIGHTEN_FIBRES].IntegerValue = 
                                            CONFIG_METHOD_STRAIGHTEN_FIBRES_MODERATE;
     CONFIG_METHOD_Global.ParameterValues[CONFIG_METHOD_STRAIGHTEN_FIBRES].StringAddress = 
                                            CONFIG_METHOD_StraightenFibresChoices[CONFIG_METHOD_STRAIGHTEN_FIBRES_MODERATE];
     
     CONFIG_METHOD_Global.ParameterValues[CONFIG_METHOD_CROSS_BEAM_SWITCHING].IntegerValue = CONFIG_FALSE;
    
     CONFIG_METHOD_Global.ParameterValues[CONFIG_METHOD_ENFORCE_SKY_QUOTA].IntegerValue = CONFIG_TRUE;
     CONFIG_METHOD_Global.ParameterValues[CONFIG_METHOD_FID_EDGE].IntegerValue = CONFIG_TRUE;
     CONFIG_METHOD_Global.ParameterValues[CONFIG_METHOD_FID_PAIRS].IntegerValue = CONFIG_FALSE;
     
     CONFIG_METHOD_Global.ParameterValues[CONFIG_METHOD_RANDOM_SEED].IntegerValue = 0; 
     CONFIG_METHOD_Global.ParameterValues[CONFIG_METHOD_MAX_ALLOCATIONS].IntegerValue = 0; 
    
     CONFIG_METHOD_Global.ParameterValues[CONFIG_METHOD_MAX_THREADS].IntegerValue =
                                                                         Nprocessors - 1; 
     CONFIG_METHOD_Global.ParameterValues[CONFIG_METHOD_ON_THE_FLY].IntegerValue
                                                                         = CONFIG_FALSE;
     CONFIG_METHOD_Global.ParameterValues[CONFIG_METHOD_AFFECTATION].DoubleValue = 
                                                                        (AffectKnown ? 1.0 : 0.0);
       

   /*  String parameters that have arbitrary values are assumed to
    *  have their values in malloc'ed memory. This is because the code that
    *  sets new values releases the memory used for previous values to
    *  avoid leaks. So the default values for any such parameters must be
    *  held in allocated memory, not just in static storage.
    */
   
   CONFIG_METHOD_Global.ParameterValues[CONFIG_METHOD_COLLISION_MATRIX_FNAME].
                                                              IntegerValue = 0;
   Size = strlen(CONFIG_METHOD_DEFAULT_FNAME) + 1;
   StringAddress = (char*) malloc (Size);
   if (StringAddress) {
      strcpy (StringAddress,CONFIG_METHOD_DEFAULT_FNAME);
      CONFIG_METHOD_Global.ParameterValues
        [CONFIG_METHOD_COLLISION_MATRIX_FNAME].StringAddress = StringAddress;
   }
     
   /*  Initialise the addresses of the matrix-related global items */
   
   CONFIG_METHOD_Global.Matrix = NULL;
   CONFIG_METHOD_Global.MatrixAddr = NULL;
   CONFIG_METHOD_Global.PhysicalMatrix = NULL;
   CONFIG_METHOD_Global.PhysicalMatrixAddr = NULL;
   
   CONFIG_METHOD_Global.InAssoc = NULL;
   CONFIG_METHOD_Global.NAssoc = NULL;
   CONFIG_METHOD_Global.NAffectedFibres = NULL;
   CONFIG_METHOD_Global.AffectedFibres = NULL;
   CONFIG_METHOD_Global.AffectedWeights = NULL;
   CONFIG_METHOD_Global.InReach = NULL;
   CONFIG_METHOD_Global.NReach = NULL;
   CONFIG_METHOD_Global.TargetRadial = NULL;
   CONFIG_METHOD_Global.InReachSize = 0;
   CONFIG_METHOD_Global.FibreAlphas = NULL;  
   CONFIG_METHOD_Global.FibreAlphaSize = 0;
   CONFIG_METHOD_Global.InAssocSize = 0;

   CONFIG_METHOD_Global.ReportedAllocations = NULL;
   

   for(Index=0;Index<9;Index++)
   {
      CONFIG_METHOD_Global.PriWeights[Index] = 0;     
   }
   /* This may be causing no skys ever to be allocated if executed
    * after ConfigMethodSetSky. BM. */
   CONFIG_METHOD_Global.SkyToAlloc = 0;
   /*  Any global initialisation needed by the allocation sub-system */
   
   AllocationGlobalInit();

}
   
/* -------------------------------------------------------------------------- */

/*                    C o n f i g  M e t h o d  I n i t
 *
 *  Function:
 *     Performs any necessary initialisation reqired for allocation.
 *
 *  Description:
 *     This routine, which is allocation algorithm specific, is called
 *     to allow the algorithm to perform any necessary initialisation when
 *     a new target field is set up.
 *
 *     It is passed a pointer to the method information structure, which
 *     amongst other things, contains an array of Pivot information structures,
 *     which give the details of the pivots (home position, whether broken
 *     or not, which spectrograph they are associated with, etc), a pointer to
 *     the array of Target information structures which give the details of
 *     the possible targets (position, associated spectrograph, etc.), and
 *     a pointer to the array of structures describing the current allocations,
 *     if any. The report function should be called at regular intervals
 *     if initialisation takes a significant time, to allow a user interface
 *     to provide some feedback to the user.
 */

static void ConfigMethodInit (
   const CONFIG_MethodInfo *MethodInfo,
   CONFIG_ReportFunction
            ReportFunction,          /* (>) Feedback routine to call */
   void *FunctionArg,                /* (>) Caller-supplied argument */
   StatusType *Status)               /* (!) Inherited status */
{
   /*  Local variables  */
   
   double Alpha;                     /* Non-radial pivot angle for fibre */
   int Bytes;                        /* Bytes of dynamic memory required */
   double DeltaX;                    /* X displacement of fibre */
   double DeltaY;                    /* Y displacement of fibre */
   double DistanceSquared;           /* Square of required fibre length */
   int FibreType;                    /* Type of fibre - guide or otherwise */
   int IPiv;                         /* Index through pivots */
   int ITarget;                      /* Index through targets */
   int JPiv;                         /* Second index through pivots */
   int MaxAllocPercent;              /* Limit on allocations to consider */
   double MaxExtensionSquared;       /* Square of maximum fibre extension */
   int NPivots;                      /* Number of pivots (ie # of fibres) */
   double ParkDistanceSquared;       /* Square of fibre length when parked */
   double PercentDone;               /* Percentage of current task completed */
   long PivotX;                      /* Pivot home position in X */
   long PivotY;                      /* Pivot home position in Y */
   AllocationPrefs Preferences;      /* Allocation sub-system preferences */
   int ReportPivs;                   /* Report after processing this many */
   char TargetSpect;                 /* Spectrograph code for target */
   char TargetType;                  /* Type of target - guide or otherwise */
   double TargetX;                   /* Target position in X */
   double TargetY;                   /* Target position in Y */
   double Theta;                     /* Theta value for possible allocation */
   double ThetaFib;                  /* Theta for the fibre - from the pivot */
   
   /*  Variables used to access the allocation matrix */
   
   CONFIG_METHOD_MatrixElement **Matrix;
                                     /* Local synonym for .Matrix in globals */
   CONFIG_METHOD_MatrixElement *MatrixAddr;
                                     /* Start of memory allocated for matrix */
   CONFIG_METHOD_MatrixElement **MatrixLineAddr;   
                                     /* Addr of memory allocated to hold 
                                      * addresses of lines of matrix */
   CONFIG_METHOD_MatrixElement *MatrixPtr;
                                     /* Points to matrix line addresses */
                                     
   /*  And a similar set for the 'physical' matrix. */
   
   CONFIG_METHOD_MatrixElement **PhysicalMatrix;
                                     /* Local synonym for .Matrix in globals */
   CONFIG_METHOD_MatrixElement *PhysicalMatrixAddr;
                                     /* Start of memory allocated for matrix */
   CONFIG_METHOD_MatrixElement **PhysicalMatrixLineAddr;   
                                     /* Addr of memory allocated to hold 
                                      * addresses of lines of matrix */
   CONFIG_METHOD_MatrixElement *PhysicalMatrixPtr;
                                     /* Points to matrix line addresses */
                                        
   NPivots = MethodInfo->NumberPivots;
   
   /*  Check all the dynamically allocated memory addresses, and free
    *  any that are still allocated from previous calls.
    */
    
   if (CONFIG_METHOD_Global.Matrix != NULL) {
      free (CONFIG_METHOD_Global.Matrix);
      CONFIG_METHOD_Global.Matrix = NULL;
   }
   if (CONFIG_METHOD_Global.MatrixAddr != NULL) {
      free (CONFIG_METHOD_Global.MatrixAddr);
      CONFIG_METHOD_Global.MatrixAddr = NULL;
   }
   if (CONFIG_METHOD_Global.ReportedAllocations != NULL) {
      free (CONFIG_METHOD_Global.ReportedAllocations);
      CONFIG_METHOD_Global.ReportedAllocations = NULL;
   }
   if (CONFIG_METHOD_Global.AffectedFibres) {
   
      /*  AffectedFibres is an array of arrays, each of which must be
       *  released before AffctedFibres itself is released. The same is true of
       *  AffectedWeights. 
       */
       
      for(IPiv=0;IPiv<NPivots;IPiv++) {
         if (CONFIG_METHOD_Global.AffectedFibres[IPiv]) {
            free (CONFIG_METHOD_Global.AffectedFibres[IPiv]);
            CONFIG_METHOD_Global.AffectedFibres[IPiv] = (int*)NULL;
         }
      }
      free (CONFIG_METHOD_Global.AffectedFibres);
      CONFIG_METHOD_Global.AffectedFibres = (int**)NULL;
   }
   
   if (CONFIG_METHOD_Global.AffectedWeights) {
      for(IPiv=0;IPiv<NPivots;IPiv++) {
         if (CONFIG_METHOD_Global.AffectedWeights[IPiv]) {
            free (CONFIG_METHOD_Global.AffectedWeights[IPiv]);
            CONFIG_METHOD_Global.AffectedWeights[IPiv] = (float*)NULL;
         }
      }
      free (CONFIG_METHOD_Global.AffectedWeights);
      CONFIG_METHOD_Global.AffectedWeights = (float**)NULL;
   }

   if (CONFIG_METHOD_Global.NAffectedFibres) {
      free (CONFIG_METHOD_Global.NAffectedFibres);
      CONFIG_METHOD_Global.NAffectedFibres = (int*)NULL;
   }
   
   /*  The reported allocations array has an element for each pivot,
    *  giving the target number to which that pivot has been allocated.
    *  A -1 value indicates that it has not been allocated. If a pivot
    *  has already been allocated, then the housekeeping layer knows 
    *  about it already, so treat it as if we've already reported it.
    */
    
   Bytes = MethodInfo->NumberPivots * sizeof(int);
   if ((CONFIG_METHOD_Global.ReportedAllocations = 
                                       (int*) malloc (Bytes)) == NULL) {
       char s[100];
       sprintf(s,
               "Unable to allocate memory for reported allocations (%d bytes)",
               Bytes);
       MethodInfo->ErrRepRoutine(0,s);
       *Status = CONFIG__NO_MEMORY;
       goto Exit;
   }
   for (IPiv = 0; IPiv < MethodInfo->NumberPivots; IPiv++) {
      ITarget = -1;
      if (MethodInfo->AllocInfo[IPiv].State == 'A') {
         ITarget = MethodInfo->AllocInfo[IPiv].TargetIndex;
      }
      CONFIG_METHOD_Global.ReportedAllocations[IPiv] = ITarget;
   }

   /*  For the purposes of reporting progress, we pick a suitable number
    *  of pivots. Here we report every 10% of changes, since this should
    *  be a pretty rapid operation.
    */

   ReportPivs = (int) ((float) MethodInfo->NumberPivots / 10.0);
   if (ReportPivs < 0) ReportPivs = 1;
    
   /*  We allocate the matrix that holds all the possible allocations,
    *  there being one entry for each pivot-target combination - ie the
    *  matrix is an array NPiv by NTargets, each element simply indicating
    *  whether or not the corresponding allocation is possible. So we
    *  start by setting it up on the basis simply of 'can the fibre from
    *  this pivot reach this target?'. Then as we start to make allocations
    *  each new allocation will invalidate other possible allocations, and
    *  these have to be worked out as we go.
    */
    
   Bytes = MethodInfo->NumberPivots * 
           MethodInfo->NumberTargets * 
           sizeof(CONFIG_METHOD_MatrixElement);

   if ((MatrixAddr = (CONFIG_METHOD_MatrixElement *) malloc (Bytes)) == NULL) {
      char s[100];
      sprintf(s,
              "Unable to allocate memory for allocation matrix of %d bytes",
              Bytes);
      MethodInfo->ErrRepRoutine(0,s);

      *Status = CONFIG__NO_MEMORY;
      goto Exit;
   }
   Bytes = MethodInfo->NumberPivots * sizeof(CONFIG_METHOD_MatrixElement *);
   if ((MatrixLineAddr = 
                   (CONFIG_METHOD_MatrixElement **) malloc (Bytes)) == NULL) {

       char s[100];
       sprintf(s,
               "Unable to allocate memory for matrix addresses (%d bytes)",
               Bytes);
       MethodInfo->ErrRepRoutine(0,s);
       *Status = CONFIG__NO_MEMORY;
       goto Exit;
   }
   
   /*  Set up the matrix line addresses in the global array used for this
    *  purpose. Note that the effect is that .Matrix can be accessed as if
    *  it were a 2D array, ie .Matrix[IPiv][ITarget] and this generates
    *  references to the appropriate elements of the allocation matrix.
    */

   CONFIG_METHOD_Global.Matrix = MatrixLineAddr;
   CONFIG_METHOD_Global.MatrixAddr = MatrixAddr;
   
   /*  For convenience, we access the Matrix from now on through the
    *  'Matrix' variable, which operates syntatically just like a 
    *  2D array (although that isn't what it is). 
    */

   Matrix = CONFIG_METHOD_Global.Matrix;
   
   MatrixPtr = MatrixAddr;
   for (IPiv = 0; IPiv < MethodInfo->NumberPivots; IPiv++) {
      CONFIG_METHOD_Global.Matrix[IPiv] = MatrixPtr;
      MatrixPtr += MethodInfo->NumberTargets;
   }
   
   /*  Now we repeat this for the physical matrix. This is the matrix that
    *  lists only those allocations that are possible or impossible on
    *  the basis of physical constraints, and so is unchanged throughout
    *  an allocation.
    */
   
   Bytes = MethodInfo->NumberPivots * 
           MethodInfo->NumberTargets * 
           sizeof(CONFIG_METHOD_MatrixElement);

   if ((PhysicalMatrixAddr = 
                   (CONFIG_METHOD_MatrixElement *) malloc (Bytes)) == NULL) {
      char s[100];
      sprintf(s,
              "Unable to allocate physical allocation matrix of %d bytes",
              Bytes);
      MethodInfo->ErrRepRoutine(0,s);

      *Status = CONFIG__NO_MEMORY;
      goto Exit;
   }
   Bytes = MethodInfo->NumberPivots * sizeof(CONFIG_METHOD_MatrixElement *);
   if ((PhysicalMatrixLineAddr = 
                   (CONFIG_METHOD_MatrixElement **) malloc (Bytes)) == NULL) {

       char s[100];
       sprintf(s,
               "Unable to allocate physical matrix addresses (%d bytes)",
               Bytes);
       MethodInfo->ErrRepRoutine(0,s);
       *Status = CONFIG__NO_MEMORY;
       goto Exit;
   }
   
   CONFIG_METHOD_Global.PhysicalMatrix = PhysicalMatrixLineAddr;
   CONFIG_METHOD_Global.PhysicalMatrixAddr = PhysicalMatrixAddr;
   
   PhysicalMatrix = CONFIG_METHOD_Global.Matrix;
   
   PhysicalMatrixPtr = PhysicalMatrixAddr;
   for (IPiv = 0; IPiv < MethodInfo->NumberPivots; IPiv++) {
      CONFIG_METHOD_Global.PhysicalMatrix[IPiv] = PhysicalMatrixPtr;
      PhysicalMatrixPtr += MethodInfo->NumberTargets;
   }
   
   /*  Initially, we run through all the elements of the array calculating
    *  whether or not an allocation is ever possible, independent of any
    *  other allocations. So this checks for invalid bend angles, excessive
    *  fibre lengths, incompatible targets & pivots, etc.
    */  
   for (IPiv = 0; IPiv < MethodInfo->NumberPivots; IPiv++) {
      if (ReportFunction != NULL) {
         if ((IPiv % ReportPivs) == 0) {
            PercentDone = ((float) IPiv / 
                                 (float) MethodInfo->NumberPivots) * 100.0;
            if ((*ReportFunction)(FunctionArg,'P',"Setting up path matrix",
                 0,0,0,0,0,0.0,PercentDone) != 0) {
                MethodInfo->InfoRoutine(MethodInfo->InfoArgument,
                                        "Cancel request received");
               *Status = CONFIG__CANCELLED;
               goto Exit;
            }
         }
      } 
      PivotX = MethodInfo->PivotInfo[IPiv].X;
      PivotY = MethodInfo->PivotInfo[IPiv].Y;
      MaxExtensionSquared = (double)MethodInfo->PivotInfo[IPiv].FibreLength;
      MaxExtensionSquared = MaxExtensionSquared * MaxExtensionSquared;
      FibreType = MethodInfo->PivotInfo[IPiv].FibreType;
      for (ITarget = 0; ITarget < MethodInfo->NumberTargets; ITarget++) {
         Matrix[IPiv][ITarget].CollidingPivot = 0;
         TargetX = (double) MethodInfo->TargetInfo[ITarget].X;
         TargetY = (double) MethodInfo->TargetInfo[ITarget].Y;
         DeltaX = (double) PivotX - TargetX;
         DeltaY = (double) PivotY - TargetY;
         DistanceSquared = (DeltaX * DeltaX) + (DeltaY * DeltaY);
         DeltaX = (double) PivotX - MethodInfo->PivotInfo[IPiv].ParkX;
         DeltaY = (double) PivotY - MethodInfo->PivotInfo[IPiv].ParkY;
         ParkDistanceSquared = (DeltaX * DeltaX) + (DeltaY * DeltaY);
         if ((DistanceSquared < MaxExtensionSquared) && 
                      (DistanceSquared > ParkDistanceSquared)) {
            TargetType = MethodInfo->TargetInfo[ITarget].Type;
            TargetSpect = MethodInfo->TargetInfo[ITarget].Spect;
            
            /*  We see if this target/pivot combination is an existing
             *  allocation. If so, we assume it must be possible. (This
             *  gets round questions such as allocations made by hand with
             *  more tolerant restrictions on the alpha angle, for example.)
             */
             
            if ((MethodInfo->AllocInfo[IPiv].State == 'A') && 
                      (MethodInfo->AllocInfo[IPiv].TargetIndex == ITarget)) {
                      
                Matrix[IPiv][ITarget].PathStatus = CONFIG_METHOD_POSSIBLE;
                     

            /*  There used to be a test here to see if the AllowAllocChange
             *  flag was set for the pivot, in which case the path status 
             *  was set to CONFIG_METHOD_IGNORE. The SA algorithm has been
             *  modified to explicitly leave such 'locked' targets alone,
             *  and so this test - which isn't strictly a test for whether
             *  the allocation is physically possible, and so was unlike the
             *  other tests in this section - has been removed.
             */
             
            } else if (!MethodInfo->FibreTargetCompatible(MethodInfo,
                                           TargetType,TargetSpect,FibreType)) {
               /* 
                *  See if the pivot and the target are of incompatible types 
                */
               Matrix[IPiv][ITarget].PathStatus = CONFIG_METHOD_INCOMPATIBLE;

            } else {
               /*
                * Check if the bend angle would allow the allocation
                */
                    /*THIS SEEMS A BIT INEFFICIENT, SHOULD USE ARRAY OF PREDETERMINED ALPHAS?*/
               if (MethodInfo->CheckAlpha(
                      MethodInfo,
                      &(MethodInfo->TargetInfo[ITarget]),
                      &(MethodInfo->PivotInfo[IPiv]),&Alpha)) {
                  /*
                   * Bend angle would allow the allocation.  Caculate the
                   * theta angle we would get and then check against invalid
                   * positions
                   */
                  ThetaFib = MethodInfo->ThetaFib (
                      MethodInfo,
                      &(MethodInfo->TargetInfo[ITarget]),
                      &(MethodInfo->PivotInfo[IPiv]));
                  Theta = (ThetaFib > PI) ? ThetaFib - PI : ThetaFib + PI;
                  if (!MethodInfo->ColInvPos(MethodInfo,FibreType,
                                            TargetX,TargetY,Theta)){
                                            
                      /*  Almost there. We need to see if the allocation would
                       *  result in a fibre crossing an adjacent parked pivot
                       *  position. Note that for some systems (2dF) this isn't
                       *  a problem - and the test routine always returns OK.
                       *  For others (OzPoz, 6dF) we don't want to cover a
                       *  parked button, or EVEN to cover the position where a
                       *  pivot goes when parked, since this can make things
                       *  almost impossible for the delta task, even if the
                       *  configuration is technically feasible.
                       */
                      
                      if (MethodInfo->AdjPark(MethodInfo,IPiv,TargetX,
                                                       TargetY,Theta)) {
                         Matrix[IPiv][ITarget].PathStatus = 
                                                   CONFIG_METHOD_PARK_CLASH;
                      } else {
                      
                          FibreType = MethodInfo->PivotInfo[IPiv].FibreType;
                          if (MethodInfo->FibreIsBroken(MethodInfo,
                                                             FibreType)) {
                             Matrix[IPiv][ITarget].PathStatus = 
                                                     CONFIG_METHOD_BROKEN;
                          } else {
                             Matrix[IPiv][ITarget].PathStatus = 
                                                     CONFIG_METHOD_POSSIBLE;
                          }

                          /*
                           * Position is ok.  But ensure we don't do this if 
                           * the fibre is broken;
                           */
                     /*commented by BM     Matrix[IPiv][ITarget].PathStatus = 
                                                     CONFIG_METHOD_POSSIBLE;*/
                          /*  Just for the template code, 
                           *  I removed this, because I don't think its
                           *  needed for the basic matrix - but I'm
                           *  not sure I've not removed the only test for
                           *  a broken fibre (KS).
                          Matrix[IPiv][ITarget].Alpha = Alpha;
                          FibreType = MethodInfo->PivotInfo[IPiv].FibreType;
                          if (!MethodInfo->FibreIsBroken(MethodInfo,
                                                             FibreType)) {
                             TargetMatrix[ITarget].Pivot[TargetMatrix[ITarget].
                                                              NumPivots]=IPiv;
                             TargetMatrix[ITarget].NumPivots++;
                          }
                          */
                      }
                   } else {
                      /* 
                       * Position would collide with an invalid positon
                       */
                      Matrix[IPiv][ITarget].PathStatus = CONFIG_METHOD_SCREWED;
                      /*
                       * If already allocated, output a warning.
                       */
                      if ((MethodInfo->AllocInfo[IPiv].State == 'A') && 
                         (MethodInfo->AllocInfo[IPiv].TargetIndex == ITarget)){
                         /*  This printf() shouldn't be here - either don't do
                          *  the test or produce a proper error message! *** */
                         printf("Invalid position (screw ?)");
                         printf("  Collision set for allocation: %d %d\n",
                                                              IPiv,ITarget);
                      }
                   }
               } else {
                  /*
                   *  Bend angle would prevent allocation.
                   */
                  Matrix[IPiv][ITarget].PathStatus = CONFIG_METHOD_BAD_ALPHA;
                  /*
                   *  If already allocated, output a warning. (Now that we
                   *  test for allocations earlier, this isn't going to
                   *  happen and could be dropped.)
                   */
                  if ((MethodInfo->AllocInfo[IPiv].State == 'A') && 
                      (MethodInfo->AllocInfo[IPiv].
                                      TargetIndex == ITarget)){
                    printf("Bad Alpha Angle set for allocation: %d %d\n"
                           ,IPiv,ITarget);
                  }
                }
             }
          } else {
            /*
             * Pivot won't reach target.
             */
            Matrix[IPiv][ITarget].PathStatus = CONFIG_METHOD_WONT_REACH;
            /*
             *  If already allocated, output a warning.
             */
            if ((MethodInfo->AllocInfo[IPiv].State == 'A') && 
                (MethodInfo->AllocInfo[IPiv].TargetIndex == ITarget)){
                printf("Bad Extension set for allocation: %d %d\n"
                       ,IPiv,ITarget);
            }

          }
       }
    }
    
   /*  We now do a pass through the matrix, setting the theta values for
    *  those allocations that are possible. In each case, we assume no twist
    *  of the pivot, and so simply calculate the theta value that corresponds
    *  to a straight button, in line with the fibre.
    */

   for (IPiv = 0; IPiv < MethodInfo->NumberPivots; IPiv++) {
      if (ReportFunction != NULL) {
         if ((IPiv % ReportPivs) == 0) {
            PercentDone = ((float) IPiv / 
                                 (float) MethodInfo->NumberPivots) * 100.0;
            if ((*ReportFunction)(FunctionArg,'P',
                       "Setting radial button angles",0,0,0,0,0,0.0,
                                                     PercentDone) != 0) {
                MethodInfo->InfoRoutine(MethodInfo->InfoArgument,
                                        "Cancel request received");
                *Status = CONFIG__CANCELLED;
                goto Exit;
            }
         }
      } 
      for (ITarget = 0; ITarget < MethodInfo->NumberTargets; ITarget++) {
         Theta = 0.0;
         if (Matrix[IPiv][ITarget].PathStatus == CONFIG_METHOD_POSSIBLE) {
            ThetaFib = MethodInfo->ThetaFib (
                         MethodInfo,
                         &(MethodInfo->TargetInfo[ITarget]),
                                       &(MethodInfo->PivotInfo[IPiv]));
            Theta = (ThetaFib > PI) ? ThetaFib - PI : ThetaFib + PI;
         }
         Matrix[IPiv][ITarget].Theta = Theta;           
      }
   }
   
   /*  Limit the maximum number of allocations to be considered. (This isn't
    *  the maximum to be made, it's used to set a limit on the overall
    *  number of possible allocations that the code chooses between, and
    *  can be used to limit the size of the arrays used by the allocation
    *  sub-system, and it time it takes to set up.)
    */
    
   MaxAllocPercent = CONFIG_METHOD_Global.ParameterValues[
                        CONFIG_METHOD_MAX_ALLOCATIONS].IntegerValue;

   /* Initialise the allocation sub-system for this configuration
    * Using CONFIG_METHOD_COLLISION_MATRIX_FNAME to supply 
    * collision matrix filename
    * (Ideally would use field.matrix for field.fld).
    */
   
   Preferences.OnTheFly = CONFIG_METHOD_Global.ParameterValues[
                              CONFIG_METHOD_ON_THE_FLY].IntegerValue;
   Preferences.MaxThreads = CONFIG_METHOD_Global.ParameterValues[
                             CONFIG_METHOD_MAX_THREADS].IntegerValue;
   Preferences.MatrixFilename = 
     CONFIG_METHOD_Global.ParameterValues[CONFIG_METHOD_COLLISION_MATRIX_FNAME].
                                                                  StringAddress;
   AllocationInit (Matrix,MethodInfo,MaxAllocPercent,ReportFunction,FunctionArg,
                   &Preferences,Status);

   if (*Status != 0) goto Exit;         
   
 
   /*  Finally, at this initialisation point, the data held in the physical
    *  matrix should be exactly the same as the data for the dynamic matrix,
    *  so we can just copy one into the other.
    */
    
   Bytes = MethodInfo->NumberPivots * 
           MethodInfo->NumberTargets * 
           sizeof(CONFIG_METHOD_MatrixElement);

   memcpy (CONFIG_METHOD_Global.PhysicalMatrixAddr,
                           CONFIG_METHOD_Global.MatrixAddr,Bytes);

   /* At this point we set up the AffectedFibres and AffectedWeights arrays
    * that tell us which other fibres are affected by light from targets 
    * assigned to each fibre. This is due to effects like ghosts and cross-talk
    * between adjacent fibres. AffectedFibres is a set of arrays, one for each
    * pivot, that list the fibres affected by light from that pivot. 
    * AffectedWeights is a set of arrasy of exactly the same size that give the
    * degree to which the other fibres are affected by the light from that 
    * pivot. The ghost from one HERMES fibre, for example, will potentially 
    * fall on more than one fibre, affecting each to a different degree.
    * For simplicity, we allow ourselves to be profligate with memory here 
    * and allocate arrays large enough to allow for each fibre to affect 
    * every other fibre. We could be cleverer about this.
    */
    
   if((CONFIG_METHOD_Global.AffectedFibres = 
                           (int**)malloc(NPivots*sizeof(int*))) == NULL){
      MethodInfo->ErrRepRoutine(0,
                         "Unable to allocate memory for AffectedFibres array");
      *Status = CONFIG__NO_MEMORY;
      return;
   }
   for(IPiv=0;IPiv<NPivots;IPiv++) {
      CONFIG_METHOD_Global.AffectedFibres[IPiv] = NULL;
   }
   
   if((CONFIG_METHOD_Global.AffectedWeights = 
                         (float**)malloc(NPivots*sizeof(float*))) == NULL){
      MethodInfo->ErrRepRoutine(0,
                        "Unable to allocate memory for AffectedWeights array");
      *Status = CONFIG__NO_MEMORY;
      return;
   }
   for(IPiv=0;IPiv<NPivots;IPiv++) {
      CONFIG_METHOD_Global.AffectedWeights[IPiv] = NULL;
   }
   
   if((CONFIG_METHOD_Global.NAffectedFibres = 
                              (int*)malloc(NPivots*sizeof(int))) == NULL){
      MethodInfo->ErrRepRoutine(0,
                        "Unable to allocate memory for NAffectedFibres array");
      *Status = CONFIG__NO_MEMORY;
      return;
   }
   for(IPiv=0;IPiv<NPivots;IPiv++) {
      CONFIG_METHOD_Global.NAffectedFibres[IPiv] = 0;
   }
   
   for(IPiv=0;IPiv<NPivots;IPiv++) {
      CONFIG_METHOD_Global.NAffectedFibres[IPiv] = 0;
      if((CONFIG_METHOD_Global.AffectedFibres[IPiv] = 
                              (int*)malloc(NPivots*sizeof(int))) == NULL){
         MethodInfo->ErrRepRoutine(0,
                    "Unable to allocate memory for AffectedFibres sub-array");
         *Status = CONFIG__NO_MEMORY;
         return;
      }
      for(JPiv=0;JPiv<NPivots;JPiv++) {
         CONFIG_METHOD_Global.AffectedFibres[IPiv][JPiv] = -1;
      }
      if((CONFIG_METHOD_Global.AffectedWeights[IPiv] = 
                      (float*)malloc(NPivots*sizeof(float))) == NULL){
         MethodInfo->ErrRepRoutine(0,
                   "Unable to allocate memory for AffectedWeights sub-array");
         *Status = CONFIG__NO_MEMORY;
         return;
      }
      for(JPiv=0;JPiv<NPivots;JPiv++) {
         CONFIG_METHOD_Global.AffectedWeights[IPiv][JPiv] = 0.0;
      }
   }
   
   /*  That's allocated the arrays - now we have to fill them up. 
    *  Note that FibreAffectData() numbers fibres from 1 up, and
    *  we have to allow for that.
    */
    
   for (IPiv = 0; IPiv < NPivots; IPiv++) {
      int NAff = 0;
      MethodInfo->FibreAffectData(MethodInfo,MethodInfo->Field,IPiv + 1,
          NPivots,CONFIG_METHOD_Global.AffectedFibres[IPiv],
          CONFIG_METHOD_Global.AffectedWeights[IPiv],&NAff);
      for (JPiv = 0; JPiv < NAff; JPiv++) {
         CONFIG_METHOD_Global.AffectedFibres[IPiv][JPiv]--;
      }
      CONFIG_METHOD_Global.NAffectedFibres[IPiv] = NAff;
   }
                           
Exit:;

}

/* -------------------------------------------------------------------------- */

/*              C o n f i g  M e t h o d  S e t  S k y
 *
 *  Function:
 *     Sets the number of fibres of a given type to be allocated to sky.
 *
 *  Description:
 *     This routine allows the higher levels to exercise precise control
 *     over the number of fibres of each type that ought to be allocated
 *     to sky targets. This is a requirement that all allocation algorithms
 *     need to support, so it is more appropriate to handle it in this way
 *     than through the method-specific parameters. The settings laid down
 *     by a call to this routine should be taken into account by the next
 *     call to ConfigMethodAllocate(). The settings made by this routine 
 *     are only relevant for fibre types that can be allocated to both
 *     sky and non-sky targets. 
 */

static void ConfigMethodSetSky (
   const CONFIG_MethodInfo *MethodInfo,
   int FibreType,                    /* (>) Fibre type code */
   int NumberSkyAllocations,         /* (>) Number of fibres for sky targets */
   StatusType *Status)               /* (!) Inherited status */
{
/*  Local variables */
   
   if (*Status != STATUS__OK) return;
   /* Just simply record the number we need */ 
   CONFIG_METHOD_Global.SkyToAlloc = NumberSkyAllocations;
      
}   
/* -------------------------------------------------------------------------- */

/*              C o n f i g  M e t h o d  E n e r g y
 *
 *  Function:
 *     Calculates the merit of an individual allocation [IPiv,ITarget].
 *     This function provides an elegant interface to influencing the
 *     properties of a configured field. The addition of simple terms, toggled
 *     by combinations of parameters, can achieve outcomes that may be 
 *     near-impossible to implement in traditional field configuration algorithms.
 *     
 *  Description:
 *     Returns (in latex):
 *       E = \left[ \beta^{p_i} + \delta \sum_j^\mathrm{NAssoc} \beta^{p_j} \right]
 *       \left(\frac{\alpha_i - \alpha_\mathrm{max}}{\alpha_\mathrm{max}}\right)^\gamma
 *
 *     where: 
 *     \beta^{p_i}           PriWeight, is 2.0 to the power of p_i - is the priority of ITarget. 
 *                           This term constitutes the primary weighting base
 *                           (which is the default objective function for most fields).
 *
 *     \delta                Delta, turns on/off the close-pairs constraint 
 *                           (required for CBS mode also)
 *                           
 *     \sum_j...             ClosePairs, sum of close pairs in InAssoc weighted by \beta^{p_j}, where
 *                           p_j is the priority of each close pair.
 *
 *     (second line)         Fibre straightness constraint. Can be detrimental to optimality
 *                           of configured field as it restricts the sample of targets a fibre
 *                           can select for allocation.
 *
 *     \alpha_\mathrm{max}   AlphaMax, maximum fibre deviation            
 *     \alpha_i              Alpha, deviation of [IPiv,ITarget] allocation
 *     
 *     In order to calculate E efficiently (ConfigMethodEnergy is called thousands of times)
 *     many of the above data are tabulated in ConfigMethod_Global arrays.
 *     Currently these are:
 *          FibreAlphas      Fibre deviation for each possible [IPiv,ITarget] allocation
 *          PriWeights       A simple array of \beta^{all priorities} to reduce exponential calculation
 *          InAssoc          Close pair target indices around each target (used for CBS mode as well)
 *          NAssoc           Number of close pairs tabulated in InAssoc
 *     
 *     For further information see the references given in the header of this file.
 *      
 *  Author: 
 *    Brent Miszalski AAO/MU.
 */
static double ConfigMethodEnergy(const CONFIG_MethodInfo* MethodInfo,int IPiv,int ITarget)
{
   double Affect;      /* Negative contribution from other affected fibres */
   double Alpha;     
   double AlphaMax;
   int PivotB,TargetB;  
   const CONFIG_TargetInfo* TargetInfo; 
   double ClosePairs;   
   int Priority;
   double PriWeight;    
   int NTargets;
   int IAffected;
   int j;
   double Gamma;
   double Delta;
   int EnforceSkyQuota;
   double SkyWeight;
   double Zeta;         /* Affectation weight - a bodge at the moment **** */
   double E;            /* Return value */

   int FidEdgeWeight;   /* If true we want to add extra weight to those fiducial targets near the field edge */
   double Rmax;         /* Maximum distance a target can have from field centre, i.e. FIELD_RAD */
   double Radius;       /* The distance ITarget is from the field centre*/

   int FidPairWeight;   /* If true we weight fiducial target pairs 
                         * i.e. include fiducial targets in CBS/close pair weighting
                         */
   
   /* Check if the allocation is "possible" */
   if(ITarget < 0 || IPiv < 0 || IPiv >= MethodInfo->NumberPivots || ITarget >= MethodInfo->NumberTargets)
      return 0;
   
   /* Get fibre deviation information for alpha constraint term*/
   AlphaMax = CONFIG_METHOD_Global.AlphaMax;
   Alpha =  CONFIG_METHOD_Global.FibreAlphas[IPiv][ITarget];
   
   /* Set nominal values of Gamma - for fibre straightness constraint */
   switch(CONFIG_METHOD_Global.ParameterValues[CONFIG_METHOD_STRAIGHTEN_FIBRES].IntegerValue)
   {
           case CONFIG_METHOD_STRAIGHTEN_FIBRES_NEVER:
                   Gamma = 0.0;
                   break;
           case CONFIG_METHOD_STRAIGHTEN_FIBRES_MODERATE:
                   Gamma = 0.1;
                   break;
           case CONFIG_METHOD_STRAIGHTEN_FIBRES_HEAVY:
                   Gamma = 2.0;
                   break;
           default:
                   Gamma = 0.0;
                   break;
   }
   /* Set nominal value of Delta to enable close pair constraint or CBS mode */
   Delta = 0.0;
   if(CONFIG_METHOD_Global.ParameterValues[CONFIG_METHOD_CROSS_BEAM_SWITCHING].IntegerValue == CONFIG_TRUE 
   || CONFIG_METHOD_Global.ParameterValues[CONFIG_METHOD_CLOSE_PAIRS].IntegerValue == CONFIG_TRUE)
           Delta = 1.0;
   
   /* Should we adding extra weight to sky targets to maximise the chances of the quota being
    * met?
    * Choose a nominal weighting of 2.0 times for sky targets over other targets.
    * This should be increased if users report the enforce sky quota option not working.
    */
   EnforceSkyQuota = CONFIG_METHOD_Global.ParameterValues[CONFIG_METHOD_ENFORCE_SKY_QUOTA].IntegerValue;
   SkyWeight = 2.0;
   
   FidEdgeWeight = CONFIG_METHOD_Global.ParameterValues[CONFIG_METHOD_FID_EDGE].IntegerValue;
   Rmax = FIELD_RAD;
   Radius = CONFIG_METHOD_Global.TargetRadial[ITarget];
   
   FidPairWeight = CONFIG_METHOD_Global.ParameterValues[CONFIG_METHOD_FID_PAIRS].IntegerValue;
  
   /* Get the weight of the current target (PriWeight - i.e. \beta^{p_i}) */ 
   TargetInfo = MethodInfo->TargetInfo;
   NTargets = MethodInfo->NumberTargets;
   Priority = TargetInfo[ITarget].Priority;
   if(Priority-1 >= 0 && Priority-1 <= 8)
      PriWeight =  CONFIG_METHOD_Global.PriWeights[Priority-1];
   else /* something went wrong somewhere? */
      PriWeight = 0;

   if(EnforceSkyQuota == CONFIG_TRUE && TargetInfo[ITarget].Type == 'S'){
         PriWeight *= SkyWeight;
   }

   /* Add any close pairs to the ClosePairs total */
   ClosePairs = 0.0;
   if(Delta != 0.0){
      for(j=0;j<CONFIG_METHOD_Global.NAssoc[ITarget];j++) {
         TargetB = CONFIG_METHOD_Global.InAssoc[ITarget][j];
         PivotB = PivotOfTarget(TargetB);
         if(PivotB >= 0 && TargetInfo[TargetB].Priority-1 >= 0 
            && TargetInfo[TargetB].Priority-1 <= 8){
            if(FidPairWeight == CONFIG_FALSE && TargetInfo[TargetB].Type == 'F') { continue; }
            if(EnforceSkyQuota == CONFIG_TRUE && TargetInfo[TargetB].Type == 'S'){
               ClosePairs+= SkyWeight*CONFIG_METHOD_Global.PriWeights[TargetInfo[TargetB].Priority-1];
            } else {
               ClosePairs+= CONFIG_METHOD_Global.PriWeights[TargetInfo[TargetB].Priority-1];
            }
         }
      }
   }
   
   /*  Take into account any affected fibres. */
   
   Affect = 0.0;
   Zeta = CONFIG_METHOD_Global.ParameterValues[CONFIG_METHOD_AFFECTATION].DoubleValue;
   if (Zeta > 0.0) {
      if (TargetInfo[ITarget].Type != 'S' && TargetInfo[ITarget].Type != 'F') {
         int NAffected = CONFIG_METHOD_Global.NAffectedFibres[IPiv];
         float MTarget = TargetInfo[ITarget].Magnitude;
         for (IAffected = 0; IAffected < NAffected; IAffected++) {
            int Affected = CONFIG_METHOD_Global.AffectedFibres[IPiv][IAffected];
            if (Affected >= 0) {
               TargetB = TargetOfPivot(Affected);
               if (TargetB >= 0) {
                  if (TargetInfo[TargetB].Type != 'S') {
                     float MagDiff = TargetInfo[TargetB].Magnitude - MTarget;
                     if (MagDiff > 0.0) {
                        Affect += pow(2.0,MagDiff) *
                                    CONFIG_METHOD_Global.AffectedWeights[IPiv][IAffected];
                     }
                  }
               }
            }
         }
      }
   }
   
   /*  Gather DEBUG statistics for Affectation. **** Remove later **** */
   DEBUG_TotalAffect += Affect;
   DEBUG_TotalPri += PriWeight;
   if (DEBUG_AffectCalcs == 0) {
      DEBUG_MaxAffect = Affect;
      DEBUG_MinAffect = Affect;
      DEBUG_MaxPri = PriWeight;
      DEBUG_MinPri = PriWeight;
   } else {
      if (Affect > DEBUG_MaxAffect) DEBUG_MaxAffect = Affect;
      if (Affect < DEBUG_MinAffect) DEBUG_MinAffect = Affect;
      if (PriWeight > DEBUG_MaxPri) DEBUG_MaxPri = PriWeight;
      if (PriWeight < DEBUG_MinPri) DEBUG_MinPri = PriWeight;
   }
   DEBUG_AffectCalcs++;
   Affect *= Zeta;
           
   /*E = (PriWeight+Delta*ClosePairs)*pow(fabs((Alpha-AlphaMax)/AlphaMax),Gamma);*/
   E = (PriWeight-Affect+Delta*ClosePairs)*pow(fabs((Alpha-AlphaMax)/AlphaMax),Gamma);
   if(FidEdgeWeight == CONFIG_TRUE && TargetInfo[ITarget].Type == 'F'){
           E *= 1.0/fabs((Radius-Rmax)/Rmax);
   }
                   
   return E;
}

/* -------------------------------------------------------------------------- */

/*              C o n f i g  M e t h o d  A l l o c a t e
 *
 *  Function:
 *     Performs a full allocation of fibres to targets using a Simulated Annealing
 *     algorithm. This is the heart of the algorithm, containing the annealing schedule,
 *     not to mention alot of setup required for the annealing 
 *     (i.e. establishing the parameter space of the problem that the Metropolis algorithm
 *     can explore efficiently).
 *
 *  Description:
 *     See the references given in the header of this file.
 *
 *  Author: 
 *    Brent Miszalski AAO/MU.
 */

static void ConfigMethodAllocate (
   const CONFIG_MethodInfo *MethodInfo,
   CONFIG_ReportFunction
            ReportFunction,          /* (>) Feedback routine to call */
   void *FunctionArg,                /* (>) Caller-supplied argument */
   StatusType *Status)               /* (!) Inherited status */
{
   /* MethodInfo related and essentials */
   const CONFIG_AllocInfo *AllocInfo;              /* Allocation info array of structures */
   const CONFIG_TargetInfo* TargetInfo;            /* Target Information array */
   const CONFIG_PivotInfo* PivotInfo;              /* Pivot Information array */
   int IPiv;                                       /* Loop index through pivots */
   int ITarget;                                    /* Loop index through targets */
   CONFIG_METHOD_MatrixElement **Matrix;           /* Local synonym for .Matrix in globals */
   CONFIG_METHOD_MatrixElement **PhysicalMatrix;   /* What pivots can reach (physically) what targets*/
   int NPiv;                                       /* Number of pivots */
   int NTargets;                                   /* Number of targets */
   int** InReach;                                  /* A list of targets physically reachable from each pivot */
   int* NReach;                                    /* Number of targets in each row of InReach */
   int** InAssoc;                                  /* A list of close targets,
                                                    * either separated by THETA, where ThetaMin <= THETA <= ThetaMax,
                                                    * or CBS pairs
                                                    */
   double ThetaMin,ThetaMax;                       /* As above */
   int* NAssoc;                                    /* Number of targets in each row of InAssoc */
   int IsCrossBeamSwitching;                       /* Are we in Cross Beam Switching mode? If so we setup InAssoc differently. */
   char crossbeam[80];                             /* Used in CBS mode to detect pairs */
   double TargetSeparation;                        /* Used to calculate InAssoc */

   /* A N N E A L I N G  S C H E D U L E */
   double T,Ti,Tf,dT;                              /* Temperature, Initial Temperature, Final Temperature, Temperature Decrement */
   double E1,E2;                                   /* Original Energy, Final Energy (Both E1,E2 > 0)*/
   double dE;                                      /* dE = (E2 - E1)*/
   
   int MaxIters;                                   /* Maximum number of iterations over all pivots per dT
                                                    * Also the max number of swaps each pivot can make per dT 
                                                    */
   int Iter;                                       /* Loop index for which Iter < MaxIters is the terminating condition */
   int i,j,k,z;                                    /* Loop indices (arbitrary) */
   
   /* R A N D O M I S A T I O N */
   int seed;                                       /* Seed for pseudo-random number generation with dprand() */
   int RandIndex;                                  /* Randomly selected target index (in InReach array) */ 
   
   int IndexAA,IndexAB,IndexBA,IndexBB;            /* AllocSubSystem indices used in randomisation process*/
   int TargetA,TargetB,PivotA,PivotB;              /* Indices of pivots and targets involved in randomisation process.
                                                    *  Used to generate Index** 
                                                    */
   int SkyAlloc;                                   /* Keeps track of sky targets allocated inside randomisation process.
                                                    * Makes sure we don't go over sky quota, etc. 
                                                    */

   /* P A R A M E T E R S */
   char Value[80];                                 /* Used to read in parameter values */
   
   double Beta;                                    /* Base weighting for priority of targets */
   double Alpha;                                   /* Used to obtain Non-radial pivot angle alpha */
   int SkyToAlloc;                                 /* Number of sky targets to allocate */
   int EnforceSkyQuota;                            /* If true we enforce SkyToAlloc */

   /* O T H E R */
   int sky_total;                                  /* Total number of sky targets */
   time_t MsecLastUpdate;                          /* Time of last full update */
   time_t MsecLastReport;                          /* Time of last progress report */
   long MsecNow;                                   /* Current time in msec - since 1970 */
   float Percent;                                  /* Percentage progress to date */
   int ExpectedIterations;                         /* Total times through main loop */
   int Iterations;                                 /* Times through main loop so far */
   struct timeval TimeNow;                         /* Current time */
   
   /* G E T / S E T  G L O B A L  P A R A M E T E R S */

   /* Includes many of the essentials and values to be used by ConfigMethodEnergy */
   Matrix = CONFIG_METHOD_Global.Matrix;
   PhysicalMatrix = CONFIG_METHOD_Global.PhysicalMatrix;
   NTargets = MethodInfo->NumberTargets;
   NPiv = MethodInfo->NumberPivots;
   AllocInfo = MethodInfo->AllocInfo;
   TargetInfo = MethodInfo->TargetInfo;
   PivotInfo = MethodInfo->PivotInfo;
   SkyToAlloc = CONFIG_METHOD_Global.SkyToAlloc;

   /* This needs to be changed if the algorithm is to deal with pre-configured fields. BM. */
   SkyAlloc=0;
   
   /* Establish nominal annealing schedule parameters for three different annealing options.
    * MaxIters is the dominant parameter in influencing annealing schedule run-time and so
    * it is doubled for each subsequent level of annealing.
    */
   switch (CONFIG_METHOD_Global.ParameterValues[CONFIG_METHOD_ANNEALING].IntegerValue) {
           case CONFIG_METHOD_ANNEALING_QUICK:
                   Ti = 100.0;
                   dT = 0.1;
                   Tf = 0.1;
                   MaxIters = 15.0;
                   break;
           case CONFIG_METHOD_ANNEALING_STANDARD:
                   /* This may be better set to 512 ?*/
                   Ti = 100.0;
                   dT = 0.01;
                   Tf = 0.1;
                   MaxIters = 30.0;
                   break;
           case CONFIG_METHOD_ANNEALING_THOROUGH:
                   Ti = 512.0;
                   dT = 0.01;
                   Tf = 0.1;
                   MaxIters = 60.0;
                   break;
           default:
                   Ti = 100.0;
                   dT = 0.01;
                   Tf = 0.1;
                   MaxIters = 30.0;
                   break;
   }

   /* At the moment we fix Beta, this value has been adequate for fields encountered so far */
   Beta = 2.0;
   
   IsCrossBeamSwitching =  CONFIG_METHOD_Global.ParameterValues[CONFIG_METHOD_CROSS_BEAM_SWITCHING].IntegerValue;
   
   CONFIG_METHOD_Global.AlphaMax = MethodInfo->MaxPivAngle(MethodInfo); 

   ConfigMethodGetParam(MethodInfo,"ThetaMin (arcsec)",Value,sizeof(Value),Status);
   ThetaMin=atoi(Value);
   ConfigMethodGetParam(MethodInfo,"ThetaMax (arcsec)",Value,sizeof(Value),Status);
   ThetaMax=atoi(Value);
   
   EnforceSkyQuota = CONFIG_METHOD_Global.ParameterValues[CONFIG_METHOD_ENFORCE_SKY_QUOTA].IntegerValue;

   
   ConfigMethodGetParam(MethodInfo,"Random Seed (expert)",Value,sizeof(Value),Status);
   seed = atoi(Value);
   
   for(i=0;i<9;i++)
   {
      CONFIG_METHOD_Global.PriWeights[i] = pow(Beta,i+1);
   }
         
   /* S E T U P  I N R E A C H */
   /* Tells us which pivots (fibres) can reach what targets */
   
   if (CONFIG_METHOD_Global.InReach) {
   
      /*  InReach is an array of arrays, each one of which is separately
       *  malloc'ed. CONFIG_METHOD_Global.InReachSize gives the number
       *  of elements of InReach. (We don't just assume it's the same as
       *  it is this time, although NPiv probably isn't going to change.)
       *  Before we release InReach itself, we need to release the sub-arrays.
       */
        
      for(i=0;i<CONFIG_METHOD_Global.InReachSize;i++) {
         if (CONFIG_METHOD_Global.InReach[i]) {
            free (CONFIG_METHOD_Global.InReach[i]);
            CONFIG_METHOD_Global.InReach[i] = (int*)NULL;
         }
      }
      free (CONFIG_METHOD_Global.InReach);
      CONFIG_METHOD_Global.InReach = (int**)NULL;
   }

   if((CONFIG_METHOD_Global.InReach = (int**)malloc(NPiv*sizeof(int*))) == NULL){
            MethodInfo->ErrRepRoutine(0,"Unable to allocate memory for InReach array");
         *Status = CONFIG__NO_MEMORY;
      return;
   }
   CONFIG_METHOD_Global.InReachSize = NPiv;

   if (CONFIG_METHOD_Global.NReach) {
      free (CONFIG_METHOD_Global.NReach);
      CONFIG_METHOD_Global.NReach = (int*)NULL;
   }

   if((CONFIG_METHOD_Global.NReach = (int*)malloc(NPiv*sizeof(int))) == NULL){
            MethodInfo->ErrRepRoutine(0,"Unable to allocate memory for NReach array");
         *Status = CONFIG__NO_MEMORY;
      return;
   }
   
   for(i=0;i<NPiv;i++) {
      if((CONFIG_METHOD_Global.InReach[i] = (int*)malloc(NTargets*sizeof(int))) == NULL){
               MethodInfo->ErrRepRoutine(0,"Unable to allocate memory for InReach sub-array");
            *Status = CONFIG__NO_MEMORY;
         return;
      }

      for(j=0;j<NTargets;j++){
         CONFIG_METHOD_Global.InReach[i][j] = -1;
         CONFIG_METHOD_Global.NReach[i] = 0;
      }
   }

   /* Fill the InReach and NReach arrays */
   for(i=0;i<NPiv;i++) {
      k=0;
      for(j=0;j<NTargets;j++) {
         if(PhysicalMatrix[i][j].PathStatus == CONFIG_METHOD_POSSIBLE 
         && TargetInfo[j].Priority != CONFIG_PRIO_INVALID
         && TargetInfo[j].Priority >= 1 && TargetInfo[j].Priority <= 9)
         {
            CONFIG_METHOD_Global.InReach[i][k] = j;
            CONFIG_METHOD_Global.NReach[i] += 1;
            k++;
         }  
      }
   }
   InReach = CONFIG_METHOD_Global.InReach;
   NReach = CONFIG_METHOD_Global.NReach;

   /* S E T U P  F I B R E A L P H A S */
   /* A matrix of fibre deviations for all possible allocations*/
   if (CONFIG_METHOD_Global.FibreAlphas) {
   
      /*  Like InReach, FibreAlphas is an array of arrays, each of which must be
       *  released before FibreAlphas itself is released.
       */
       
      for(i=0;i<CONFIG_METHOD_Global.FibreAlphaSize;i++) {
         if (CONFIG_METHOD_Global.FibreAlphas[i]) {
            free (CONFIG_METHOD_Global.FibreAlphas[i]);
            CONFIG_METHOD_Global.FibreAlphas[i] = NULL;
         }
      }
      free (CONFIG_METHOD_Global.FibreAlphas);
      CONFIG_METHOD_Global.FibreAlphas = NULL;
   }


   if((CONFIG_METHOD_Global.FibreAlphas  = (double**)malloc(NPiv*sizeof(double*))) == NULL){
            MethodInfo->ErrRepRoutine(0,"Unable to allocate memory for FibreAlphas array");
         *Status = CONFIG__NO_MEMORY;
      return;
   }
      
   CONFIG_METHOD_Global.FibreAlphaSize = NPiv;
   
   for(i=0;i<NPiv;i++) {
      if((CONFIG_METHOD_Global.FibreAlphas[i]  = (double*)malloc(NTargets*sizeof(double))) == NULL){
               MethodInfo->ErrRepRoutine(0,"Unable to allocate memory for FibreAlphas sub-array");
            *Status = CONFIG__NO_MEMORY;
         return;
      }
      for(j=0;j<NTargets;j++){
         CONFIG_METHOD_Global.FibreAlphas[i][j]  = -1.0;
      }
   }
   /* Fill the FibreAlphas matrix */
   for(i=0;i<NPiv;i++)
   {
      for(j=0;j<NReach[i];j++)
      {
         ITarget = InReach[i][j];
         MethodInfo->CheckAlpha(MethodInfo,&TargetInfo[ITarget],&PivotInfo[i],&Alpha);
         CONFIG_METHOD_Global.FibreAlphas[i][ITarget] = Alpha;
      }
   }
   

   /* S E T U P  T A R G E T R A D I A L*/
   /* An array of target distances from the centre of the field 
    * Only used for weighting fiducial targets near the edge of the field
    */
   
   if (CONFIG_METHOD_Global.TargetRadial) {
      free (CONFIG_METHOD_Global.TargetRadial);
      CONFIG_METHOD_Global.TargetRadial = (double*)NULL;
   }
   if((CONFIG_METHOD_Global.TargetRadial = (double*)malloc(NTargets*sizeof(double))) == NULL){
            MethodInfo->ErrRepRoutine(0,"Unable to allocate memory for TargetRadial array");
         *Status = CONFIG__NO_MEMORY;
      return;
   }
   /* Fill TargetRadial */
   for(ITarget=0;ITarget<NTargets;ITarget++){
		CONFIG_METHOD_Global.TargetRadial[ITarget] = sqrt(pow(TargetInfo[ITarget].X,2) + pow(TargetInfo[ITarget].Y,2));
	}
   
   /* S E T U P  I N A S S O C */
   /* Tells us which targets are within a given separation of other a specific target */
   /* N.B. For production code, in the case of Cross-beam switching, InAssoc must only
    * consist of Target pairs (or more than one target if specified).
    */
   if (CONFIG_METHOD_Global.InAssoc) {
   
      /*  Like InReach, InAssoc is an array of arrays, each of which must be
       *  released before InAssoc itself is released.
       */
       
      for(i=0;i<CONFIG_METHOD_Global.InAssocSize;i++) {
         if (CONFIG_METHOD_Global.InAssoc[i]) {
            free (CONFIG_METHOD_Global.InAssoc[i]);
            CONFIG_METHOD_Global.InAssoc[i] = (int*)NULL;
         }
      }
      free (CONFIG_METHOD_Global.InAssoc);
      CONFIG_METHOD_Global.InAssoc = (int**)NULL;
   }

   if((CONFIG_METHOD_Global.InAssoc = (int**)malloc(NTargets*sizeof(int*))) == NULL){
            MethodInfo->ErrRepRoutine(0,"Unable to allocate memory for InAssoc array");
         *Status = CONFIG__NO_MEMORY;
      return;
   }
   CONFIG_METHOD_Global.InAssocSize = NTargets;
   
   if (CONFIG_METHOD_Global.NAssoc) {
      free (CONFIG_METHOD_Global.NAssoc);
      CONFIG_METHOD_Global.NAssoc = (int*)NULL;
   }
   if((CONFIG_METHOD_Global.NAssoc = (int*)malloc(NTargets*sizeof(int))) == NULL){
            MethodInfo->ErrRepRoutine(0,"Unable to allocate memory for NAssoc array");
         *Status = CONFIG__NO_MEMORY;
      return;
   }
   for(i=0;i<NTargets;i++) {
      CONFIG_METHOD_Global.NAssoc[i] = 0;
      if((CONFIG_METHOD_Global.InAssoc[i] = (int*)malloc(NTargets*sizeof(int))) == NULL){
         MethodInfo->ErrRepRoutine(0,"Unable to allocate memory for InAssoc sub-array");
         *Status = CONFIG__NO_MEMORY;
         return;
      }
      for(j=0;j<NTargets;j++)
         CONFIG_METHOD_Global.InAssoc[i][j] = -1;
   }

   /* Fill the InAssoc and NAssoc arrays */

   /* This is where we detect CBS pairs from the .fld 
    * and incorporate them into InAssoc
    * N.B. This shows that CBS and close pairs are mutually exclusive
    * => we can't let the user enable both the relevant parameters.
    */
   for(i=0;i<NTargets;i++) {
      k=0;
      z=0;
      sprintf(crossbeam,"%s-cbspair",TargetInfo[i].ObjName);
      for(j=0;j<NTargets;j++)
      {
         if(i==j)
            continue;
         if(IsCrossBeamSwitching == CONFIG_TRUE)
         {
            if(strcmp(crossbeam,TargetInfo[j].ObjName) == 0)
            {
               CONFIG_METHOD_Global.InAssoc[i][k] = j;
               CONFIG_METHOD_Global.NAssoc[i] += 1;
               k++;
               /* make sure we add the very important complementary info */
               CONFIG_METHOD_Global.InAssoc[j][z] = i;
               CONFIG_METHOD_Global.NAssoc[j] += 1;
               z++;
            }
         }
         else
         {
            TargetSeparation = sqrt(pow(TargetInfo[i].X-TargetInfo[j].X,2)
                     + pow(TargetInfo[i].Y-TargetInfo[j].Y,2));
            /* This is probably VERY BAD - should use proper distance
             * in ra/dec. Ask Scott Croom?
             * These separations are only in approximate X,Y coordinates.
             */
            if(TargetSeparation >= (ThetaMin/3600)*(FIELD_RAD/1.05)
               && TargetSeparation <= (ThetaMax/3600)*(FIELD_RAD/1.05)) {
               CONFIG_METHOD_Global.InAssoc[i][k] = j;
               CONFIG_METHOD_Global.NAssoc[i] += 1;
               k++;
            }
         }
      }
      
   }
   
   InAssoc = CONFIG_METHOD_Global.InAssoc;
   NAssoc = CONFIG_METHOD_Global.NAssoc;
   
   /* S E T U P  S O M E  M O R E  T H I N G S */

   gettimeofday(&TimeNow,NULL);
   MsecNow = (TimeNow.tv_sec * 1000) + (TimeNow.tv_usec / 1000);
   MsecLastUpdate = MsecLastReport = MsecNow;
   
   /* Needed to keep track of sky targets
    * in their allocation process 
    */
   sky_total = 0;
   
   for(j=0;j<NTargets;j++)
   {
      if(TargetInfo[j].Type == 'S')
         sky_total++;
   }

   E1 = E2 = dE = 0.0;
   /* If IntegerValue of CONFIG_METHOD_RANDOM_SEED == 0, 
    * use time(NULL), otherwise use its value for the seed*/
   if(seed == 0)
           seed = time(NULL);
   sdprand (seed);

   /*  Determine the expected number of passes through the annealing
    *  loop. This can be used to provide a more linear estimate of
    *  progress - although it seems a matter of taste whether one prefers
    *  this to the non-linear estimate of progress based on temperature.
    *  (You don't feel so bad stopping the non-linear one as it spends ages
    *  at the end trying to get things just that little bit better.)
    */
    
   ExpectedIterations = 0;
   for(T=Ti;T>=Tf;T*=(1-dT)){
      ExpectedIterations++;
   }
   
   /* S T A R T I N G  A N N E A L I N G  S C H E D U L E */
   printf("Starting the Annealing Schedule\n");
   /* Temperature - used as control parameter */
   /* T typically Beta**9 */
   /* dT typically 0.01 to ensure slow cooling */
   
   Iterations = 0;
   for(T=Ti;T>=Tf;T*=(1.0-dT)){
           Iterations++;
      IPiv = 0;
   /* The idea here is that we try and make as many swaps as
    * possible per dT such that Thermal Equilibrium may be established at each T
    */
   for(Iter=0;Iter < MaxIters;Iter++){
   for(IPiv=0;IPiv < NPiv;IPiv++){
           gettimeofday(&TimeNow,NULL);
           MsecNow = (TimeNow.tv_sec * 1000) + (TimeNow.tv_usec / 1000);
           /* Report progress - take your pick as to which to use */        
           Percent = ((Ti-T)/(Ti-Tf))*100.0; /* Non-linear based on temp */
           Percent = ((float)Iterations/(float)ExpectedIterations)*100.0;
                                             /* Linear based on iterations */
           /*  Make an overall sumary report twice a second */
           if ((MsecNow - MsecLastReport) > 500) {
              MsecLastReport = MsecNow;
              if(ReportFunction != NULL) {
                   if((*ReportFunction)(FunctionArg,'P',
                   "Annealing...",0,0,0,0,0,0.0,Percent) != 0){
                           MethodInfo->InfoRoutine(MethodInfo->InfoArgument,
                                        "Cancel request received");
                   *Status = CONFIG__CANCELLED;
                   goto Exit;
                   }
              }
           }
           /*  Send a full update of allocations every 2 seconds. We might
            *  want to tune this a bit.
            */
           if ((MsecNow - MsecLastUpdate) > 2000) {
              ConfigMethodUpdate(MethodInfo,Status);
              MsecLastUpdate = MsecNow;
           }
      IndexAA = IndexAB = IndexBB = 0;
      TargetA = TargetB = PivotA = PivotB = 0;
      /* No targets to allocate */
      if(NReach[IPiv] <= 0)
         continue;
      /* Random target selection and swapping loop */
      for(k=0;k<NReach[IPiv];k++){
          E1 = 0;
          E2 = 0;
          PivotA = IPiv;
          /* R A N D O M I S A T I O N */ 
            RandIndex = (int)(dprand()*1.0*(NReach[PivotA]));
         
          TargetA = TargetOfPivot(PivotA);
          TargetB = InReach[PivotA][RandIndex];
          if(TargetB < 0)
             continue;
          PivotB = PivotOfTarget(TargetB);
          /*should make sure this doesn't exclude Case (iv) (defined below)*/
          if(TargetA == TargetB || PivotA == PivotB){
            if(!(TargetA < 0 && PivotB < 0))
            continue;
          }
          /* Make sure we don't mess with locked pivots */

          if (!PivotInfo[PivotA].AllowAllocChange) continue;
          if (PivotB >= 0 && !PivotInfo[PivotB].AllowAllocChange) continue;
          
          IndexAA = AllocationIndex(PivotA,TargetA);
          IndexAB = AllocationIndex(PivotA,TargetB);
          IndexBA = AllocationIndex(PivotB,TargetA);
          IndexBB = AllocationIndex(PivotB,TargetB);
          /* Handle various randomisation cases (there are four of them) */
          /* Allocated?|____PivotA____|____TargetB____|
           * Case (i)  |      Yes     |      Yes      |
           * Case (ii) |      Yes     |      No       |
           * Case (iii)|      No      |      Yes      | 
           * Case (iv) |      No      |      No       |
           * Case (v)  |         NOT USED (see MNRAS paper for details)
           */
          /* Aim is to get PivotA allocated to TargetB 
           * IF the Metropolis algorithm is satisfied 
           * WITHOUT any allocation attrition.
          */
          /* Case (i) */
          if(TargetA >=0 && PivotB >=0
          && PhysicalMatrix[PivotA][TargetB].PathStatus == CONFIG_METHOD_POSSIBLE
          && PhysicalMatrix[PivotB][TargetA].PathStatus == CONFIG_METHOD_POSSIBLE){
            E1 = ConfigMethodEnergy(MethodInfo,PivotA,TargetA)
               + ConfigMethodEnergy(MethodInfo,PivotB,TargetB);

            if(Swap(PivotA,TargetA,PivotB,TargetB) > 0){
            E2 = ConfigMethodEnergy(MethodInfo,PivotA,TargetB)
               + ConfigMethodEnergy(MethodInfo,PivotB,TargetA);
            break;
         }
         else{
            E2 = E1;
         }
               }
          
          /* Case (ii) */
          if(TargetA >= 0 && PivotB < 0
          && PhysicalMatrix[PivotA][TargetB].PathStatus == CONFIG_METHOD_POSSIBLE){
            /* we don't want to "deallocate" a sky fibre through a swap */
            if(TargetInfo[TargetA].Type == 'S' && TargetInfo[TargetB].Type != 'S')
               continue;
            if(TargetInfo[TargetA].Type != 'S' && TargetInfo[TargetB].Type == 'S'){
               /*allocate a new sky target through a swap if we aren't already over the limit */
               if(SkyToAlloc == 0)
                  continue;
               else if(SkyAlloc >= SkyToAlloc || (SkyToAlloc >= sky_total && SkyAlloc >= sky_total))
                  continue;
            }

         E1 = ConfigMethodEnergy(MethodInfo,PivotA,TargetA);

            DeallocateIndex(IndexAA);
         
         if(IsAllocAble(IndexAB) > 0){
            AllocateIndex(IndexAB);
            E2 = ConfigMethodEnergy(MethodInfo,PivotA,TargetB);
            if(SkyToAlloc > 0 && TargetInfo[TargetA].Type != 'S' && TargetInfo[TargetB].Type == 'S')
                  SkyAlloc++;
            break;
         }  
         else if(IsAllocAble(IndexAA) > 0){
            AllocateIndex(IndexAA);
            E2 = E1;
         }
          }
          /* Case (iii) */
          if(TargetA < 0 && PivotB >=0
          && PhysicalMatrix[PivotA][TargetB].PathStatus == CONFIG_METHOD_POSSIBLE){
         E1 = ConfigMethodEnergy(MethodInfo,PivotB,TargetB);
            DeallocateIndex(IndexBB);
   
         if(IsAllocAble(IndexAB) > 0){
            AllocateIndex(IndexAB);
            E2 = ConfigMethodEnergy(MethodInfo,PivotA,TargetB);
            break;
         }  
         else if(IsAllocAble(IndexBB) > 0){
            AllocateIndex(IndexBB);
            E2 = E1;
         }
          } 
          /* Case (iv) */
          if(TargetA < 0 && PivotB < 0){
         E1 = 0;

         if(TargetInfo[TargetB].Type == 'S') {
            if(SkyToAlloc == 0)
               continue;
            else if(SkyAlloc >= SkyToAlloc || (SkyToAlloc >= sky_total && SkyAlloc >= sky_total))
               continue;
         }
            if(IsAllocAble(IndexAB) > 0) {
            AllocateIndex(IndexAB);
            
            E2 = ConfigMethodEnergy(MethodInfo,PivotA,TargetB);

            if(TargetInfo[TargetB].Type == 'S') 
               SkyAlloc++;
            break;
         }
          }

      }/*end Target Selection loop*/

      /* M E T R O P O L I S  A L G O R I T H M */
      dE = (E2 - E1);
      if(dE < 0 && dprand()*1.0 >= exp(dE/T)){
         /* Swap performed above is not accepted
          * Proceeding to restore previous state 
          * (that corresponding to energy E1) */

         /* Case (i) */
         /* Want to reverse the previous swap */
         if(TargetA >=0 && PivotB >=0
         && TargetOfPivot(PivotA) == TargetB
         && TargetOfPivot(PivotB) == TargetA
         && PhysicalMatrix[PivotA][TargetA].PathStatus == CONFIG_METHOD_POSSIBLE
         && PhysicalMatrix[PivotB][TargetB].PathStatus == CONFIG_METHOD_POSSIBLE)
         {
            if(Swap(PivotA,TargetB,PivotB,TargetA) > 0)
            {
                    /* "Beware the Jabberwock, my son!
                     *  The jaws that bite, the claws that catch!
                     *  Beware the Jubjub bird, and shun
                     *  The frumious Bandersnatch!"
                     */
            }  
         }
      
         /* Case (ii) */
         /* Want to restore [PivotA,TargetB] -> [PivotA][TargetA] */
         if(TargetA >= 0 && PivotB < 0 
         && TargetOfPivot(PivotA) == TargetB
         && PhysicalMatrix[PivotA][TargetA].PathStatus == CONFIG_METHOD_POSSIBLE)
         {
            DeallocateIndex(IndexAB);
            if(IsAllocAble(IndexAA) > 0)
            {
               AllocateIndex(IndexAA);
            }
            else if(IsAllocAble(IndexAB) > 0)
            {
               AllocateIndex(IndexAB);
            }
         }
         /* Case (iii) */
         /* Want to restore [PivotA,TargetB] -> [PivotB,TargetB] */
         if(TargetA < 0 && PivotB >=0
         && TargetOfPivot(PivotA) == TargetB)
         {
            DeallocateIndex(IndexAB);
            if(IsAllocAble(IndexBB) > 0)
            {
               AllocateIndex(IndexBB);
            }
            else if(IsAllocAble(IndexAB) > 0)
            {
               AllocateIndex(IndexAB);
            }

         }
      
         /* Case (iv) */
         /* Want to restore to no allocations */
         if(TargetA < 0 && PivotB < 0 
         && TargetOfPivot(PivotA) == TargetB){
            DeallocateIndex(IndexAB);
         }
         /* Energy of system now restored to previous amount 
          * This is symbolic - doesn't actually need to happen. */
         E2 = E1;
         }
      /* E N D  M E T R O P O L I S  A L G O R I T H M */
   }
   /* E N D  L O O P  O V E R  P I V O T S */
   }
   /* E N D  M A X I T E R S  L O O P */

   /*if(DISPLAY == 1) ConfigMethodDisplay(MethodInfo,Status);*/

   }
Exit:;
   printf("Annealing Schedule Complete\n");
   
   ConfigMethodCheckAffected(MethodInfo);  /* DEBUG call for affected fibre calculations */
   
   AllocationFinish(Status);
   
   /*printf("Annealing Schedule Complete Iterations: %d\n",Iterations);*/
   /* E N D  O F  A N N E A L I N G  S C H E D U L E */
   
   /* Possible Routines to Run Post-Configuration should be called here */

   /*ConfigMethodUncrossFibres(MethodInfo);*/
        
   ConfigMethodUpdate(MethodInfo,Status);
      
   /*  A final call to the ReportFunction should update the allocation
    *  statistics.
    */
    
   if(ReportFunction != NULL) {
      (void)(*ReportFunction)(FunctionArg,'P',
                   "Annealing...",0,0,0,0,0,0.0,100.0);
   }

} 
/* -------------------------------------------------------------------------- */

/*              C o n f i g  M e t h o d  U p d a t e
 *
 *  Function:
 *     Updates the housekeeping layer with the progress of the allocation.
 *
 *  Description:
 *     With an algorithm that constantly allocates and deallocates, it
 *     is not possible to use the reporting functions the way the housekeeping
 *     interface originally envisaged - by reporting each individual
 *     allocation and deallocation (which is potentially displayed on
 *     the GUI as a result). Instead, this routine can be called
 *     at suitable intervals and it updates the housekeeping layer on changes
 *     since the last update.
 *     
 *  Author: 
 *    Keith Shortridge AAO.
 */

static void ConfigMethodUpdate (
   const CONFIG_MethodInfo *MethodInfo,
   StatusType *Status)                    /* (!) Inherited status */
{
   int AllIndex;
   int Index;
   int IPiv;
   int ITarget;
   double Theta;
   int NPiv;
   int NAlloc;
   int* AllocedPivs;
   
   NPiv = MethodInfo->NumberPivots;
   
   /*  First we want to see which pivots are no longer allocated. This is
    *  a little messy, as we can't get that list directly. So we allocate
    *  a flag array, one entry for each pivot. We start with each pivot
    *  flagged as not allocated, then we change that for each pivot that
    *  the alloc sub-systemm has in its allocations list.
    */
    
   NAlloc = AllocEntries();
   AllocedPivs = (int*) malloc (NPiv * sizeof(int));
   if (AllocedPivs) {
      for (IPiv = 0; IPiv < NPiv; IPiv++) {
         AllocedPivs[IPiv] = CONFIG_FALSE;
      }
      for (AllIndex = 1; AllIndex <= NAlloc; AllIndex++) {
         AllocEntryDetails(AllIndex,&Index,&IPiv,&ITarget,&Theta);
         AllocedPivs[IPiv] = CONFIG_TRUE;
      }
      
      /*  Now we go through all the pivots that we last reported as
       *  allocated, to see if any of them have been deallocated. If so,
       *  we report the deallocation.
       */
      
      for (IPiv = 0; IPiv < NPiv; IPiv++) {
         if (CONFIG_METHOD_Global.ReportedAllocations[IPiv] >= 0) {
            if (!AllocedPivs[IPiv]) {
               MethodInfo->NoteDeallocation (MethodInfo,IPiv,Status);
               CONFIG_METHOD_Global.ReportedAllocations[IPiv] = -1;
            }
         }
      }
      
      /*  Then we go through all the pivots that are now allocated, to
       *  see if they were previously reported as unallocated or allocated
       *  to a different target. If so, we have to report an update.
       */
      
      for (AllIndex = 1; AllIndex <= NAlloc; AllIndex++) {
         AllocEntryDetails(AllIndex,&Index,&IPiv,&ITarget,&Theta);
         if (CONFIG_METHOD_Global.ReportedAllocations[IPiv] != ITarget) {
         
            /*  Allocation status has changed */
            
            if (CONFIG_METHOD_Global.ReportedAllocations[IPiv] >= 0) {
            
               /*  Was last reported allocated to a different target. In this
                *  case we have to report its deallocation.
                */
                
               MethodInfo->NoteDeallocation (MethodInfo,IPiv,Status);
               CONFIG_METHOD_Global.ReportedAllocations[IPiv] = -1;
            }
         }
      }
      for (AllIndex = 1; AllIndex <= NAlloc; AllIndex++) {
         AllocEntryDetails(AllIndex,&Index,&IPiv,&ITarget,&Theta);
         if (CONFIG_METHOD_Global.ReportedAllocations[IPiv] != ITarget) {
         
            /*  Report allocation to new target */
      
            MethodInfo->NoteAllocation (MethodInfo,IPiv,ITarget,Theta,0,Status);
            CONFIG_METHOD_Global.ReportedAllocations[IPiv] = ITarget;
         }
      }
               
      /*  Release the allocated workspace */
               
      free (AllocedPivs);               
   }
   
}

/* -------------------------------------------------------------------------- */

/*             C o n f i g  M e t h o d  C h e c k  A f f e c t e d
 *
 *  Function:
 *     Quantifies the cross-fibre affectation for an allocation.
 *
 *  Description:
 *     This routine is a diagnostic tool to help develop the code that tries
 *     to minimise the affectation - the effects of bright targets being 
 *     assigned to fibres that affect other fibres assigned to fainter targets.
 */

static void ConfigMethodCheckAffected (
   const CONFIG_MethodInfo *MethodInfo)
{
   double Affect;                                  /* -ve contribution from affected fibres */
   const CONFIG_AllocInfo *AllocInfo;              /* Allocation info array of structures */
   int IAffected;
   double AlphaMax;
   double AlphaTotal;
   int IPiv;
   int NAlphas;
   int NMagDiffs;
   const CONFIG_TargetInfo* TargetInfo;
   double MaxMagDiff;
   double TotalMagDiffs; 

   TargetInfo = MethodInfo->TargetInfo;
   AllocInfo = MethodInfo->AllocInfo;
   Affect = 0.0;
   MaxMagDiff = 0.0;
   TotalMagDiffs = 0.0;
   NMagDiffs = 0;
   AlphaMax = 0.0;
   AlphaTotal = 0.0;
   NAlphas = 0;
   
   /*  This is calculating the same affectation statistic that the annealing is trying
    *  to minimise, and so the code is almost identical to that used in the Energy
    *  calculation in ConfigMethodEnergy().
    */
    
   for (IPiv = 0; IPiv < MethodInfo->NumberPivots; IPiv++) {
      if (AllocInfo[IPiv].State == 'A') {
         int ITarget = AllocInfo[IPiv].TargetIndex;
         if (ITarget >= 0) {
            double RadToDeg = 180.0 / 3.14159;
            double Alpha =  RadToDeg * fabs(CONFIG_METHOD_Global.FibreAlphas[IPiv][ITarget]);
            if (Alpha > AlphaMax) AlphaMax = Alpha;
            AlphaTotal += Alpha;
            NAlphas++;
            if (TargetInfo[ITarget].Type != 'S' && TargetInfo[ITarget].Type != 'F') {
               int NAffected = CONFIG_METHOD_Global.NAffectedFibres[IPiv];
               float MTarget = TargetInfo[ITarget].Magnitude;
               for (IAffected = 0; IAffected < NAffected; IAffected++) {
                  int Affected = CONFIG_METHOD_Global.AffectedFibres[IPiv][IAffected];
                  if (Affected >= 0) {
                     if (MethodInfo->AllocInfo[Affected].State == 'A') {
                        int TargetB = MethodInfo->AllocInfo[Affected].TargetIndex;
                        if (TargetB >= 0) {
                           if (TargetInfo[TargetB].Type != 'S') {
                              float MagDiff = TargetInfo[TargetB].Magnitude - MTarget;
                              if (MagDiff > 0.0) {
                                 if (MagDiff > MaxMagDiff) MaxMagDiff = MagDiff;
                                 TotalMagDiffs += MagDiff;
                                 NMagDiffs++;
                                 Affect += pow(2.0,MagDiff) * 
                                    CONFIG_METHOD_Global.AffectedWeights[IPiv][IAffected];
                              }
                           }
                        }
                     }
                  }
               }
            }
         }
      }
   }
   printf ("Affectation effect = %f\n",Affect);
   if (DEBUG_AffectCalcs > 0) {
      printf ("Over %d calculations:\n",DEBUG_AffectCalcs);
      printf ("Affect values: min %f, max %f mean %f\n",DEBUG_MinAffect,
                         DEBUG_MaxAffect,DEBUG_TotalAffect / DEBUG_AffectCalcs);
      printf ("Priority values: min %f, max %f mean %f\n",DEBUG_MinPri,
                         DEBUG_MaxPri,DEBUG_TotalPri / DEBUG_AffectCalcs);
   }
   if (NMagDiffs > 0) {
      printf ("Over %d comparisons:\n",NMagDiffs);
      printf ("Max mag diff %f, mean %f\n",MaxMagDiff,TotalMagDiffs / NMagDiffs);
   }
   if (NAlphas > 0) {
      printf ("Over %d allocations:\n",NAlphas);
      printf ("Max alpha %f, mean %f\n",AlphaMax,AlphaTotal / NAlphas);
   }
   DEBUG_MaxAffect = 0.0;
   DEBUG_MinAffect = 0.0;
   DEBUG_TotalAffect = 0.0;
   DEBUG_AffectCalcs = 0;
   DEBUG_MaxPri = 0.0;
   DEBUG_MinPri = 0.0;
   DEBUG_TotalPri = 0.0;
}


/* -------------------------------------------------------------------------- */

/*             C o n f i g  M e t h o d  R e l e a s e  M e m
 *
 *  Function:
 *     Releases any memory allocated by the allocation-specific code.
 *
 *  Description:
 *     This routine is called at the end of the program in order to give 
 *     it a chance to release any dynamic memory that may have been allocated.
 *     Any such memory should have its address recorded in the global data
 *     structures.
 */

static void ConfigMethodReleaseMem (
   const CONFIG_MethodInfo *MethodInfo DUNUSED)
{
   /*  Local variables  */

   int Index;                      /* Index through method parameters */

   /*  Check all the dynamically allocated memory addresses, and free
    *  any that are still allocated.
    */
    
   if (CONFIG_METHOD_Global.Matrix != NULL) {
      free (CONFIG_METHOD_Global.Matrix);
      CONFIG_METHOD_Global.Matrix = NULL;
   }
   if (CONFIG_METHOD_Global.MatrixAddr != NULL) {
      free (CONFIG_METHOD_Global.MatrixAddr);
      CONFIG_METHOD_Global.MatrixAddr = NULL;
   }
   if (CONFIG_METHOD_Global.PhysicalMatrix != NULL) {
      free (CONFIG_METHOD_Global.PhysicalMatrix);
      CONFIG_METHOD_Global.PhysicalMatrix = NULL;
   }
   if (CONFIG_METHOD_Global.PhysicalMatrixAddr != NULL) {
      free (CONFIG_METHOD_Global.PhysicalMatrixAddr);
      CONFIG_METHOD_Global.PhysicalMatrixAddr = NULL;
   } 
    if (CONFIG_METHOD_Global.InAssoc != NULL) {
      for(Index=0; Index < CONFIG_METHOD_Global.InAssocSize; Index++) {
         if (CONFIG_METHOD_Global.InAssoc[Index]) {
            free (CONFIG_METHOD_Global.InAssoc[Index]);
            CONFIG_METHOD_Global.InAssoc[Index] = (int*)NULL;
         }
      }
      free (CONFIG_METHOD_Global.InAssoc);
      CONFIG_METHOD_Global.InAssoc = NULL;
   }
    
if (CONFIG_METHOD_Global.FibreAlphas != NULL) {
      for(Index=0; Index < CONFIG_METHOD_Global.FibreAlphaSize; Index++) {
         if (CONFIG_METHOD_Global.FibreAlphas[Index]) {
            free (CONFIG_METHOD_Global.FibreAlphas[Index]);
            CONFIG_METHOD_Global.FibreAlphas[Index] = NULL;
         }
      }
      free (CONFIG_METHOD_Global.FibreAlphas);
      CONFIG_METHOD_Global.FibreAlphas = NULL;
   }

   if (CONFIG_METHOD_Global.NAssoc != NULL) {
      free (CONFIG_METHOD_Global.NAssoc);
      CONFIG_METHOD_Global.NAssoc = NULL;
   }

   if (CONFIG_METHOD_Global.TargetRadial != NULL) {
      free (CONFIG_METHOD_Global.TargetRadial);
      CONFIG_METHOD_Global.TargetRadial = NULL;
   }

   if (CONFIG_METHOD_Global.InReach != NULL) {
      for(Index = 0; Index < CONFIG_METHOD_Global.InReachSize; Index++) {
         if (CONFIG_METHOD_Global.InReach[Index]) {
            free (CONFIG_METHOD_Global.InReach[Index]);
            CONFIG_METHOD_Global.InReach[Index] = (int*)NULL;
         }
      }
      free (CONFIG_METHOD_Global.InReach);
      CONFIG_METHOD_Global.InReach = NULL;
   }

   if (CONFIG_METHOD_Global.NReach != NULL) {
      free (CONFIG_METHOD_Global.NReach);
      CONFIG_METHOD_Global.NReach = NULL;
   }
   if (CONFIG_METHOD_Global.ReportedAllocations != NULL) {
      free (CONFIG_METHOD_Global.ReportedAllocations);
      CONFIG_METHOD_Global.ReportedAllocations = NULL;
   }

   /*  If there are any 'any value' character parameters that have been set,
    *  free any memory allocated to them.
    */

   for (Index = 0; Index < CONFIG_METHOD_NPARMS; Index++) {
      if (CONFIG_METHOD_ParmTypes[Index] == 'C') {
         if (CONFIG_METHOD_ParmNValues[Index] == 0) {
            if (CONFIG_METHOD_Global.ParameterValues[Index].StringAddress
                                                                    != NULL) {
               free (
                 CONFIG_METHOD_Global.ParameterValues[Index].StringAddress);
            }
         }
      }
   }
   
}

/* -------------------------------------------------------------------------- */

/*         C o n f i g  M e t h o d  N u m b e r  P a r a m s
 *
 *  Function:
 *     Returns the number of method-specific parameters.
 *
 *  Description:
 *     This routine returns the number of method-specific parameters that
 *     can be used to control the details of the allocation algorithm.
 */

static int ConfigMethodNumberParams (
    const CONFIG_MethodInfo *MethodInfo DUNUSED)
{
   return (CONFIG_METHOD_NPARMS);
}

/* -------------------------------------------------------------------------- */

/*         C o n f i g  M e t h o d  P a r a m  D e t a i l s
 *
 *  Function:
 *     Returns details about the method-specific parameters.
 *
 *  Description:
 *     This routine returns details about the various method-specific 
 *     parameters that can be used to control the details of the 
 *     allocation algorithm. This includes their names, types, and
 *     possible values.
 */

static void ConfigMethodParamDetails (
   const CONFIG_MethodInfo *MethodInfo DUNUSED,
   int ParameterNumber,
   char **Name,
   char *Type,
   double Limits[],
   int *NValues,
   char ***Values,
   StatusType *Status)
{
   /*  Local variables  */

   int Index;                         /* Index into parameter arrays */

   if (*Status != STATUS__OK) return;

   if ((ParameterNumber < 1) || (ParameterNumber > CONFIG_METHOD_NPARMS)) {
       MethodInfo->ErrRepRoutine(0,
        "Attempt to get details about non-existent method-specific parameter");
      *Status = CONFIG__INVAL_PARM_NO;
   } else {
      Index = ParameterNumber - 1;
      *Name = CONFIG_METHOD_ParmNames[Index];
      *Type = CONFIG_METHOD_ParmTypes[Index];
      Limits[0] = CONFIG_METHOD_ParmLimits[Index][0];
      Limits[1] = CONFIG_METHOD_ParmLimits[Index][1];
      *NValues = CONFIG_METHOD_ParmNValues[Index];
      *Values = CONFIG_METHOD_ParmValues[Index];
   }

}

/* -------------------------------------------------------------------------- */

/*              C o n f i g  M e t h o d  S e t  P a r a m
 *
 *  Function:
 *     Sets the value of one of the method-specific parameters.
 *
 *  Description:
 *     This routine sets the value of one of the various method-specific 
 *     parameters that can be used to control the details of the 
 *     allocation algorithm. The parameter is specified by name, and
 *     the value is specified as a character string.
 */

static void ConfigMethodSetParam (
   const CONFIG_MethodInfo *MethodInfo DUNUSED,
   char *Name,                        /* (>) Name of parameter */
   char *Value,                       /* (>) Value of parameter */
   StatusType *Status)                /* (!) Inherited status value */
{
   /*  Local variables  */

   int Bytes;                         /* Number of bytes for character param */
   char *ExpertStr;                   /* Pointer to "(expert)" string */
   double FloatValue;                 /* Value of floating point param */
   int Index;                         /* Index into parameter arrays */
   long IntegerValue;                 /* Value of integer parameter */
   int Length;                        /* Chars in string to compare */
   double MaxLimit;                   /* Maximum value for parameter */
   double MinLimit;                   /* Minimum value for parameter */
   CONFIG_Boolean NameMatched;        /* True if Name is a valid name */
   int NValues;                       /* Number of values for char param */
   CONFIG_Boolean ParamOK;            /* True if char param is valid string */
   char **ParmValues;                 /* Address of possible char values */
   char *StringAddress;               /* Address of parameter value string */
   int StringIndex;                   /* Index through possible parm values */
   char Type;                         /* PArameter type 'I','F','C','L' */

   if (*Status != STATUS__OK) return;
   
   /*  Work through the names of the various parameters looking for
    *  a match with that specified.
    */

   NameMatched = CONFIG_FALSE;
   for (Index = 0; Index < CONFIG_METHOD_NPARMS; Index++) {
      if (!strcmp(Name,CONFIG_METHOD_ParmNames[Index])) {

         /*  Name matches, so pick up the type, limits (for numerical values)
          *  and number of possible values (for character parameters).
          */

         NameMatched = CONFIG_TRUE;
         Type = CONFIG_METHOD_ParmTypes[Index];
         MinLimit = CONFIG_METHOD_ParmLimits[Index][0];
         MaxLimit = CONFIG_METHOD_ParmLimits[Index][1];
         NValues = CONFIG_METHOD_ParmNValues[Index];

         /*  Now handle the parameter according to type  */

         if (Type == 'L') {

            /*  Logical parameters must be either 'T' or 'F' */

            if (!strcmp(Value,"T")) {
               CONFIG_METHOD_Global.ParameterValues[Index].IntegerValue = 
                                                                   CONFIG_TRUE;
            } else if (!strcmp(Value,"F")) {
               CONFIG_METHOD_Global.ParameterValues[Index].IntegerValue = 
                                                                  CONFIG_FALSE;
            } else {
                char s[100];
                sprintf(s,
                      "'%s' is a logical parameter and cannot be set to '%s'",
                        Name,Value);
                MethodInfo->ErrRepRoutine(0,s);
               *Status = CONFIG__INVAL_PARM_VALUE;
            }

         } else if (Type == 'I') {

            /*  Integer parameters must be valid integer values within
             *  the specified range.
             */

            if (sscanf(Value,"%ld",&IntegerValue) == 1) {
               if (((double) IntegerValue < MinLimit) ||
                    ((double) IntegerValue > MaxLimit)) {
                   char s[100];
                   sprintf(s,
                  "Parameter '%s' must be between %d and %d. %ld is invalid.",
                           Name,(int) MinLimit, (int) MaxLimit, IntegerValue);
                   MethodInfo->ErrRepRoutine(0,s);
                   *Status = CONFIG__PARM_OUT_RANGE;
               } else {
                  CONFIG_METHOD_Global.ParameterValues[Index].IntegerValue = 
                                                               IntegerValue;
               }
            } else {
                char s[100];
                sprintf(s,
                        "'%s' is not a valid value for integer parameter '%s'",
                        Value,Name);
                MethodInfo->ErrRepRoutine(0,s);
                *Status = CONFIG__INVAL_PARM_VALUE;
            }

         } else if (Type == 'F') {

            /*  Floating point parameters must be valid real values within
             *  the specified range.
             */

            if (sscanf(Value,"%lf",&FloatValue) == 1) {
               if ((FloatValue < MinLimit) || (FloatValue > MaxLimit)) {
                   char s[100];
                   sprintf(s,
                   "Parameter '%s' must be between %f and %f. %f is invalid.",
                           Name,MinLimit,MaxLimit,FloatValue);
                           MethodInfo->ErrRepRoutine(0,s);
                  *Status = CONFIG__PARM_OUT_RANGE;
               } else {
                  CONFIG_METHOD_Global.ParameterValues[Index].DoubleValue = 
                                                                 FloatValue;
               }
            } else {
                char s[100];
                sprintf(s,
                        "'%s' is not a valid value for parameter '%s'",
                        Value,Name);
                MethodInfo->ErrRepRoutine(0,s);
               *Status = CONFIG__INVAL_PARM_VALUE;
            }

         } else if (Type == 'C') {

            /*  Character values can have either any value, or one of a 
             *  specified set. 
             */

            if (NValues <= 0) {

               /*  This paramter can have any value, so it must be valid.
                *  We allocate memory to hold a copy of the new value (this
                *  allows the value to be any length at all), first freeing
                *  any previously allocated.
                */
                
               if (CONFIG_METHOD_Global.ParameterValues[Index].StringAddress
                                                                     != NULL) {
                  free (CONFIG_METHOD_Global.ParameterValues[Index]
                                                              .StringAddress);
               }
               Bytes = strlen(Value);
               if (Bytes == 0) {
                  CONFIG_METHOD_Global.ParameterValues[Index].StringAddress 
                                                                      = NULL;
               } else {
                  StringAddress = (char *) malloc(Bytes + 1);
                  strcpy (StringAddress,Value);
                  CONFIG_METHOD_Global.ParameterValues[Index].StringAddress
                                                             = StringAddress;
               }
            } else {

               /*  This character string must have one of a set of specified
                *  values. Check that this is one, and set the index in
                *  the global variables. Allow for the possibility that the
                *  "(expert)" may have been stripped off.
                */

               ParmValues = CONFIG_METHOD_ParmValues[Index];
               ParamOK = CONFIG_FALSE;
               for (StringIndex = 0; StringIndex < NValues; StringIndex++) {
                  if ((ExpertStr = 
                             strstr(ParmValues[StringIndex],"(expert)"))) {
                     Length = ExpertStr - ParmValues[StringIndex];
                  } else {
                     Length = strlen(Value);
                 }
                  if (!strncmp(Value,ParmValues[StringIndex],Length)) {
                     CONFIG_METHOD_Global.ParameterValues[Index].StringAddress
                                                    = ParmValues[StringIndex];
                     CONFIG_METHOD_Global.ParameterValues[Index].IntegerValue
                                                                = StringIndex;
                     ParamOK = CONFIG_TRUE;
                     break;
                  }
               }
               if (!ParamOK) {
                   char s[100];
                   sprintf(s,
                           "'%s' is not one of the possible values for '%s'",
                           Value,Name);
                   MethodInfo->ErrRepRoutine(0,s);
                   *Status = CONFIG__INVAL_PARM_VALUE;
               }
            }
         }

         /*  Finished processing a matched name, so break out of the loop
          *  through the various parameter names.
          */

         break;
      }
   }

   if (!NameMatched) {
       char s[100];
       sprintf(s,"'%s' is not a recognised parameter name",Name);
       MethodInfo->ErrRepRoutine(0,s);
      *Status = CONFIG__INVAL_PARM_NAME;
   }
}
                  
/* -------------------------------------------------------------------------- */

/*              C o n f i g  M e t h o d  G e t  P a r a m
 *
 *  Function:
 *     Gets the value of one of the method-specific parameters.
 *
 *  Description:
 *     This routine gets the value of one of the various method-specific 
 *     parameters that can be used to control the details of the 
 *     allocation algorithm. The parameter is specified by name, and
 *     the value is returned as a character string.
 */

static void ConfigMethodGetParam (
   const CONFIG_MethodInfo *MethodInfo DUNUSED,
   char *Name,                        /* (>) Name of parameter */
   char *Value,                       /* (<) Value of parameter */
   int VSize,                         /* (>) Maximum number of characters */
   StatusType *Status)                /* (!) Inherited status value */
{
   /*  Local variables  */

   int Index;                         /* Index into parameter arrays */
   CONFIG_Boolean NameMatched;        /* True if Name is a valid name */
   char Type;                         /* Parameter type 'I','F','C','L' */
   char *ValuePtr;                    /* Points to string to be returned */
   char ValueString[64];              /* Used to format numbers and logicals */

   if (*Status != STATUS__OK) return;

   /*  Work through the names of the various parameters looking for
    *  a match with that specified.
    */

   NameMatched = CONFIG_FALSE;
   ValuePtr = NULL;
   for (Index = 0; Index < CONFIG_METHOD_NPARMS; Index++) {
      if (!strcmp(Name,CONFIG_METHOD_ParmNames[Index])) {

         /*  Name matches, so pick up its type.
          */

         NameMatched = CONFIG_TRUE;
         Type = CONFIG_METHOD_ParmTypes[Index];

         /*  Now handle the parameter according to type  */

         if (Type == 'L') {

            /*  Logical parameters are returned as either 'T' or 'F' */

            if (CONFIG_METHOD_Global.ParameterValues[Index].IntegerValue) {
               ValuePtr = "T";
            } else {
               ValuePtr = "F";
            }

         } else if (Type == 'I') {

            /*  Integer parameters are formatted into a value string.
             */

            sprintf(ValueString,"%d",
                CONFIG_METHOD_Global.ParameterValues[Index].IntegerValue);
            ValuePtr = ValueString;

         } else if (Type == 'F') {

            /*  Floating point parameters are also formatted into a value
             *  string.
             */

            sprintf(ValueString,"%f",
                CONFIG_METHOD_Global.ParameterValues[Index].DoubleValue);
            ValuePtr = ValueString;

         } else if (Type == 'C') {

            /*  Character values can have either any value, or one of a 
             *  specified set. In either case, the address of the current
             *  value is held in the global variables.
             */

            ValuePtr = 
                 CONFIG_METHOD_Global.ParameterValues[Index].StringAddress;
         }

         /*  Finished processing a matched name, so break out of the loop
          *  through the various parameter names.
          */

         break;
      }
   }

   /*  After working through the various accepted parameter names, we either
    *  didn't find the name specified, in which case we have an error, or
    *  we have a pointer (ValuePtr) to a string giving the current value.
    */

   if (NameMatched) {
      if (ValuePtr == NULL) {
         Value[0] = '\0';
      } else {
         strncpy (Value,ValuePtr,VSize);
         Value[VSize - 1] = '\0';
      }
   } else {
       char s[100];
       sprintf(s,"'%s' is not a recognised parameter name",Name);
      MethodInfo->ErrRepRoutine(0,s);
      *Status = CONFIG__INVAL_PARM_NAME;
   }
}


/* -------------------------------------------------------------------------- */

/*          C o n f i g  M e t h o d  S e t  I m p o r t  F i l e
 *
 *  Function:
 *     Set import mode and the name of the file to import.
 *
 *  Description:
 *     This routine is called directly by ConfigSetImportFile(). This
 *     is not well defined at the moment but can be considered a way of
 *     importing allocation information into the allocation layer. This
 *     interface may be replaced in the re-working of the housekeeping
 *     layer to support additional facilities provided by different
 *     allocation algorithms.
 */

static void ConfigMethodSetImportFile(
   const CONFIG_MethodInfo *MethodInfo DUNUSED,
   char* name DUNUSED)
{
}


/* -------------------------------------------------------------------------- */

/*                C o n f i g  M e t h o d  D e t a i l s
 *
 *  The lone external routine in this module.  Allows the Config library
 *  to featch details of the method.
 */

extern void ConfigMethodDetails(CONFIG_MethodDetails *MethodDetails)
{
    MethodDetails->Version = CONFIG_VERSION;
    MethodDetails->RoutineCount = CONFIG_MTH_N_ROUTINES;
    MethodDetails->clientData = 0;  /* Not currently used */

    strncpy(MethodDetails->Name, "Simulated annealing algorithm", 
                                              CONFIG_MTH_NAME_LEN-1);
    MethodDetails->Name[CONFIG_MTH_NAME_LEN-1] = '\0';

    MethodDetails->GlobalInit    = ConfigMethodGlobalInit;
    MethodDetails->Init          = ConfigMethodInit;
    MethodDetails->ReleaseMem    = ConfigMethodReleaseMem;
    MethodDetails->SetSky        = ConfigMethodSetSky;
    MethodDetails->Allocate      = ConfigMethodAllocate;
    MethodDetails->NumParams     = ConfigMethodNumberParams;
    MethodDetails->ParamDetails  = ConfigMethodParamDetails;
    MethodDetails->SetParam      = ConfigMethodSetParam;
    MethodDetails->GetParam      = ConfigMethodGetParam;
    MethodDetails->SetImportFile = ConfigMethodSetImportFile;
}
/*
         1         2         3         4         5         6         7         8
12345678901234567890123456789012345678901234567890123456789012345678901234567890
*/
