/*+
 *               C O N F I G _ M E T H O D . T E M P L A T E
 *
 *  Function:
 *     An example of 'Algorithm-specific' routines for 2dF fibre allocations.
 *
 *  Description:
 *     This file contains a set of routines that between them provide the
 *     default fibre instrument configuration algorithm. These are intended 
 *     to be called from the 'housekeeping and interface' (config.c) 
 *     layer of routines described in the document 'Fibre Instrument 
 *     Configuration Algorithm Routines'. 
 *
 *     This module implements a skeleton allocation algorithm. It is not
 *     a good one, but it will perform an allocation. (Although at present
 *     it cannot handle the case where it is passed a set of targets some
 *     of which are already allocated.) The intention is that it can
 *     provide a template on which a decent algorithm can be built.
 *
 *  Version date: 10th March 2005.
 *
 *  Support: K. Shortridge, AAO.
 *-
 *  "@(#) $Id: ACMM:aaoConfigure/src/config_method.template.c,v 12.7 04-Nov-2013 11:28:48+11 tjf $"
 *
 *  History:
 *     10th Mar 2005.  Original version. KS/AAO.
 *      19th Aug 2009. The invalid position routines (*ColInvPos) now take
 *                     a fibre type argument. KS.
 *+
 */
 
/*                            O v e r v i e w
 *
 *  The first routine called from outside will be ConfigMethodDetails().
 *  In this we set up the fields that will allow the rest of the algorithm
 *  routines to be called. We have to set up:
 *
 *  GlobalInit : A global initialisation routine, called just once to
 *               perform necessary one-off initialisation.
 *               We set this to ConfigMethodGlobalInit.
 *  ReleaseMem : A close down routine, called when the program is to be
 *               shut down. We set this to ConfigMethodReleaseMem.
 *  Init       : A routine, called when the upper layers have the 
 *               field plate details (pivot positions etc) and the target
 *               details (position, types etc). We set this to 
 *               ConfigMethodInit.
 *  Allocate   : The routine called when the upper layers want an allocation
 *               to be performed. We set this to ConfigMethodAllocate.
 *
 *  There are also a number of utility routines connected mainly with
 *  parameters that we set to a set of routines whose code is taken directly
 *  from the original config_method.c code. These are all driven by
 *  tables set up a the start of this file. These tables differ from
 *  algorithm to algorithm, but the code to handle the parameters can be
 *  the same for all algorithms.
 *
 *  There are also two routines used by the upper levels to control some
 *  details of the allocation. In this file we supply dummy versions.
 *  
 *  SetSky        : A routine called when the upper layers want to explicitly
 *                  control the allocation of sky targets to different types
 *                  of fibre. We set this to ConfigMethodSetSky.
 *  SetImportFile : A routine used to make the algorithm read in a set of
 *                  pre-determined allocations from a text file.
 * 
 *  Once those are set up, the upper layers will next call the GlobalInit
 *  routine, here ConfigMethodGlobalInit.  In this code, that routine
 *  does a bit of general housekeeping, mainly connected with the default
 *  values for the parameters.
 *
 *  The upper layers may then call some of the parameter routines, but the
 *  next important call will be to the Init routine when the plate details 
 *  are known. Here we use this call to set up a matrix of pivots versus
 *  targets, giving details of whether each possible pivot/target allocation
 *  is possible and if so some additional information such as the fibre
 *  theta angle for the allocation.
 *
 *  The next call will be to the Allocate routine. Here we don't do anything
 *  clever. We just go through each pivot in turn and allocate it to the
 *  first possible target. When we make an allocation, we update the
 *  matrix so that any allocations shown in there as possible but which 
 *  would now collide with the newly allocated fibre are now no longer
 *  flagged as possible. When we make an allocation we make a callback
 *  to the upper layer's 'NoteAllocation' routine so that the allocation
 *  can be recorded.
 *
 *  And that's it.
 *   
 */

/* -------------------------------------------------------------------------- */

/*                       I n c l u d e  F i l e s
 */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

/*  DRAMA include files, only status.h needed */

#include "status.h"

/*  Debugging macro */

#ifdef DEBUG
#undef DEBUG
#endif
#define DEBUG (void) printf

/*
 * Prototypes for the random number generator from dprand.c that we use,
 * if we are going to use one. This makes sure we get the same results on
 * all platforms.
 */
 
extern void sdprand (int seed);
extern double dprand (void);

/*
 * DUNUSED Flag which under GCC, prevents warnings about an argument being
 * unused - should be added where we know it will be unused
 */
 
#ifndef DUNUSED
#if defined(__GNUC__) || defined(GNUC)
#   define DUNUSED /* Default definition */
#   ifdef __GNUC_MINOR__
#   if __GNUC_MINOR__ >= 7
#       undef  DUNUSED
#       define DUNUSED __attribute__ ((unused))
#   endif
#   endif
#else
#   define DUNUSED /* */
#endif
#endif /* DUNUSED */


/*  Include files specific to CONFIG itself */

#include "tdFconfigMethod.h"
#include "config_err.h"

/* -------------------------------------------------------------------------- */

/*            M e t h o d - s p e c i f ic  d e c l a r a t i o n s
 *
 *  'Method' in this case refers to anything specific to the actual allocation
 *  used.
 */
 
/*  These are the codes used for the PathStatus flag in the matrix. These
 *  indicate either that a given path is possible, or the reasons why it
 *  isn't. Originally, all elements are set either to POSSIBLE, WON'T_REACH
 *  BAD_ALPHA, BROKEN, PARK_CLASH or INCOMPATIBLE. Nothing can alter these
 *  conditions, which are physical. As the program progresses, fibres get 
 *  allocated and some of the POSSIBLEs change into IN_USE or ALLOCATED or 
 *  COLLISION, and may change back again if fibres are deallocated.
 *  OUTSIDE means the targets too far out and would be badly vignetted...
 *  SCREWED means it would fall on one of the plate screws, which 
 *  would be very bad news!  IGNORE means no attempt should be made to
 *  allocate the pivot. DISABLED means that this is an allocation that is
 *  possible, but is not being allowed at the moment (for example, to
 *  prevent any sky target allocations).
 */

#define CONFIG_METHOD_POSSIBLE 0
#define CONFIG_METHOD_ALLOCATED 1
#define CONFIG_METHOD_WONT_REACH 2
#define CONFIG_METHOD_BROKEN 3
#define CONFIG_METHOD_FIBRE_IN_USE 4
#define CONFIG_METHOD_INCOMPATIBLE 5
#define CONFIG_METHOD_BAD_ALPHA 6
#define CONFIG_METHOD_COLLISION 7
#define CONFIG_METHOD_TARGET_IN_USE 8
#define CONFIG_METHOD_OUTSIDE 9
#define CONFIG_METHOD_SCREWED 10
#define CONFIG_METHOD_IGNORE  11
#define CONFIG_METHOD_PARK_CLASH 12
#define CONFIG_METHOD_DISABLED 13

/*  Method parameter constants. Note that having these statically defined
 *  like this makes it very easy to implement ConfigMethodParamDetails(),
 *  and also helps to code ConfigMethodSetParam(). However, it is important
 *  that the information held here is consistent - each parameter has a
 *  symbolic name that equates to an integer index, and the entries in
 *  the ConfigMethodParamXxxx arrays must match those indices. It isn't quite
 *  possible to make the whole parameter business 'table-driven' (controlled
 *  entirely by the entries in here). Some explicit code is needed, but the 
 *  more we can do from the definitions here the easier the code is to modify.
 *  At present, to add a new parameter, 1) Increase the value of CONFIG_METHOD-
 *  _NPARMS, 2) Define a new symbolic value for its index (a new parameter
 *  called CONFIG_METHOD_xxxx_PARM, 3) Add new entries to all the CONFIG_METHOD
 *  _ParmXxxx tables. 4) For a character parameter with a fixed set of possible
 *  values, define a new array of the possible values (see CONFIG_METHOD_-
 *  SelectChoices) and symbolic indices for the values, 5) Add initialisation
 *  code to ConfigMethodGlobalInit(), 6) Add whatever code is needed to make
 *  use of the new parameter.
 */

/*  This code defines two parameters for the template allocation algorithm,
 *  purely for demonstration purposes. The first is an integer parameter
 *  which is used to set the number of a pivot that is not be allocated.
 *  The template code will deallocate this if it allocates it - this is
 *  in purely to exercise the deallocation code, and in passing it shows how
 *  to use a parameter. The second parameter is a character string that can
 *  have only one of a limited set of values, and is included here simply
 *  to demonstrate how to set up such a string - it isn't used by the code.
 */
 
#define CONFIG_METHOD_NPARMS 2           /*  Number of parameters supported */

/*  A array of structures like this goes into the method-specific global
 *  variables. For each parameter, there is a double precision value (used
 *  only for floating point parameters) and an integer value. For logical 
 *  parameters the integer value will be 0 (false) or 1 (true). For integer
 *  parameters the integer value is that of the parameter. For character
 *  parameters with an unlimited number of possible values, when the parameter
 *  is set to a new value an array is allocated in memory (using malloc())
 *  for a static copy of that new value string, and the address of this
 *  allocated string is held in common. For character parameters with a
 *  limited set of possible values, all possible values are held in a static
 *  array and the address of the value currently set is held in common. In
 *  this case, the integer value is also used to hold the index number of
 *  the current value being used.
 */

typedef struct CONFIG_METHOD_ParamValue {
   int IntegerValue;                         /* Integer value - see above */
   char *StringAddress;                      /* Address of string */
   double DoubleValue;                       /* Floating point value */
} CONFIG_METHOD_ParamValue;


/*  First parameter - index = CONFIG_METHOD_DEALLOC_PARM = 0.    
 *
 *  The first parameter is an integer that can be set to any value between 0
 *  and 404. It gives the number of a pivot that is not be be allocated.
 *  The value is interpreted as a pivot number counting from 1 up, so the
 *  default value of zero (set in ConfigMethodGlobalInit()) means that
 *  all pivots are elegible for allocation.
 */

#define CONFIG_METHOD_DEALLOC_PARM 0
                                        
/*  2nd parameter - index = CONFIG_METHOD_CHAR_PARM = 1.    
 *
 *  The second parameter is a character parameter with a number of
 *  different possible options. This parameter isn't used, and is here
 *  purely for show. We define symbolic constants for each of the
 *  possible values.
 */

#define CONFIG_METHOD_CHAR_PARM 1

#define CONFIG_METHOD_CHAR_FIRST 0
#define CONFIG_METHOD_CHAR_SECOND 1
#define CONFIG_METHOD_CHAR_THIRD 2

static char *CONFIG_METHOD_CharChoices[] = {
    "first choice",                          /* CONFIG_METHOD_CHAR_FIRST  */
    "second choice",                         /* CONFIG_METHOD_CHAR_SECOND   */
    "third choice"                           /* CONFIG_METHOD_CHAR_THIRD   */
    };

/*  
 *  The following arrays contain the parameter details as used by
 *  ConfigMethodParamDetails() and ConfigMethodSetParam().
 *  CONFIG_METHOD_ParmNames   lists the name of the parameters.
 *  CONFIG_METHOD_ParmTypes   lists their types (I,C,F or L).
 *  CONFIG_METHOD_ParmLimits  lists their limits, if they are numeric
 *  CONFIG_METHOD_ParmNValues lists their restricted number of possible
 *                            string values, if they have such a restriction.
 *  CONFIG_METHOD_ParmValues  lists those possible string values, if they
 *                            have them.
 */

static char *CONFIG_METHOD_ParmNames[CONFIG_METHOD_NPARMS] = {
    "Fibre not to allocate",                  /* CONFIG_METHOD_DEALLOC_PARM */
    "Demo character parameter",               /* CONFIG_METHOD_CHAR_PARM    */
    };
    
static char CONFIG_METHOD_ParmTypes[CONFIG_METHOD_NPARMS] = {
    'I',                                      /* CONFIG_METHOD_DEALLOC_PARM */
    'C',                                      /* CONFIG_METHOD_CHAR_PARM    */
    };

static double CONFIG_METHOD_ParmLimits[CONFIG_METHOD_NPARMS][2] = {
    { 0.0, 404.0},                            /* CONFIG_METHOD_DEALLOC_PARM */
    { 0.0, 0.0},                              /* CONFIG_METHOD_CHAR_PARM    */
    }; 

static int CONFIG_METHOD_ParmNValues[CONFIG_METHOD_NPARMS] ={
    0,                                        /* CONFIG_METHOD_DEALLOC_PARM */
    3,                                        /* CONFIG_METHOD_CHAR_PARM    */
    };

static char **CONFIG_METHOD_ParmValues[CONFIG_METHOD_NPARMS] = {
    0 ,                                       /* CONFIG_METHOD_DEALLOC_PARM */
    CONFIG_METHOD_CharChoices,                /* CONFIG_METHOD_CHAR_PARM    */
    }; 

/*  The template allocation code uses a matrix to keep track of which
 *  allocations are possible and which are not. At the start of the algorithm,
 *  its elements are set purely on the basis of physical constraints - can the
 *  fibre from this pivot reach that target, given the various restrictions
 *  on bend angle, fibre length, screw positions, etc, and whether the
 *  target and fibre are compatible. As the allocation algorithm runs and
 *  allocations are made, this matrix is updated to reflect allocations that
 *  have become invalid becuase they would collide with allocations already
 *  made. The matrix is a 2-D array of elements of type 
 *  CONFIG_METHOD_MatrixElement, defined here. The most important item
 *  is PathStatus, which shows whether the allocation is possible or not
 *  - it always has one of the values defined as CONFIG_METHOD_POSSIBLE
 *  through to CONFIG_METHOD_DISABLED. The Theta value is included 
 *  because it is used a lot in the code, and it is easier to calculate
 *  all the possible theta values right at the start of the allocation than
 *  to do so on the fly as the code runs. The CollidingPivot item makes it
 *  easier to handle the case when an allocation is cancelled - if a 
 *  pivot is no longer allocated, all allocations that show it as a colliding
 *  pivot may now have becoe possible again.
 */
 
typedef struct CONFIG_METHOD_MatrixElement {
   unsigned char PathStatus;     /* Flags state of the possible path */
   int CollidingPivot;           /* A fibre that collides with the path */
   double Theta;                 /* The theta value for the button */
} CONFIG_METHOD_MatrixElement;


/* -------------------------------------------------------------------------- */

/*  The following defines any global variables used by the method-specific
 *  code. This defines a single global variable, called CONFIG_METHOD_Global
 *  which is a structure containing all the global information needed by
 *  the allocation algorithm. This makes it clear when we look at the code
 *  when a global variable is being used - it will be in the structure
 *  called CONFIG_METHOD_Global.
 *
 *  At the moment, this contains the matrix used by the allocation
 *  algorithm (which requires two arrays, one the 2D array of matrix elements,
 *  one the 1D array that holds the addresses of the matrix element at
 *  the start of each array (this is Matrix, and having it makes it possible
 *  to use a dynamically allocated array as if it were a statically
 *  declared 2D array - a trick used, for example, by the C version of
 *  Numerical Recipies).
 *
 *  It also contains the parameter values.
 */

typedef struct CONFIG_METHOD_GlobalType {
   CONFIG_METHOD_MatrixElement **Matrix;       
                                 /* Address of memory allocated to hold addrs
                                  * of starts of rows of allocation matrix. */
   CONFIG_METHOD_MatrixElement *MatrixAddr;
                                 /* Address of memory allocated for matrix */
   CONFIG_METHOD_ParamValue ParameterValues[CONFIG_METHOD_NPARMS];
                                 /* Current parameter values */
} CONFIG_METHOD_GlobalType;

 
CONFIG_METHOD_GlobalType CONFIG_METHOD_Global;
                                 /* Actual global structure used */
                                 
                                 
/* -------------------------------------------------------------------------- */

/*            I n t e r n a l  R o u t i n e  D e c l a r a t i o n s
 *
 *  This is the point to declare any internal routines. They should all
 *  be static. (Actually, all routines in this file are static, since even
 *  those called by the housekeeping layer are called through the structure
 *  set up by ConfigMethodDetails).
 */

static void ConfigMethodAllocUpdate (
   const CONFIG_MethodInfo *MethodInfo,
   int Pivot,                     /* (>) The newly allocated pivot */
   int Target,                    /* (>) The newly allocated target */
   double Theta,                  /* (>) The theta angle for the button */
   StatusType *Status);           /* (!) Inherited status variable */
    
static void ConfigMethodDeallocUpdate (
   const CONFIG_MethodInfo *MethodInfo,
   int Pivot,                     /* (>) The number of the deallocated pivot */
   int Target,                    /* (>) Target allocated to this pivot */
   StatusType *Status);           /* (!) Inherited status */
   
/* -------------------------------------------------------------------------- */

/*              C o n f i g  M e t h o d  G l o b a l  I n i t
 *
 *  Function:
 *     Performs any global initialisation reqired for allocation.
 *
 *  Description:
 *     This routine, which is allocation algorithm specific, is called
 *     just once, during program startup, to allow the algorithm to perform 
 *     any necessary initialisation of its global structures.
 */

static void ConfigMethodGlobalInit (const CONFIG_MethodInfo *MethodInfo DUNUSED)
{
   
   /*  Local variables  */
   
   int Index;
   
   /*  At the moment, all we do is initialise the parameters */

   /*  If there are any 'any value' character parameters, set their addresses
    *  to null. Because if any new values are set, any memory already allocated
    *  to such a parameter will be freed.
    */

   for (Index = 0; Index < CONFIG_METHOD_NPARMS; Index++) {
      if (CONFIG_METHOD_ParmTypes[Index] == 'C') {
         if (CONFIG_METHOD_ParmNValues[Index] == 0) {
            CONFIG_METHOD_Global.ParameterValues[Index].StringAddress = NULL;
         }
      }
   }

   /*  The parameter values have to be initialised here, since they may never
    *  be changed at all and so need sensible default values. This section
    *  has to be modified if a new parameter is introduced. Note that 
    *  limited range character parameters have to have both their string
    *  addresses and integer values set. Logical and integer parameters need
    *  their integer values set, floating point values need their double
    *  value set, and character parameters that can take any value need to
    *  have their string addresses either left at NULL (from the previous loop)
    *  or set to point to a static string used as a default value.
    */

   CONFIG_METHOD_Global.ParameterValues[CONFIG_METHOD_DEALLOC_PARM]
                                                       .IntegerValue = 0;
          
   CONFIG_METHOD_Global.ParameterValues[CONFIG_METHOD_CHAR_PARM]
                                     .IntegerValue = CONFIG_METHOD_CHAR_SECOND;
   CONFIG_METHOD_Global.ParameterValues[CONFIG_METHOD_CHAR_PARM]
       .StringAddress = CONFIG_METHOD_CharChoices[CONFIG_METHOD_CHAR_SECOND];
}
   
/* -------------------------------------------------------------------------- */

/*                    C o n f i g  M e t h o d  I n i t
 *
 *  Function:
 *     Performs any necessary initialisation reqired for allocation.
 *
 *  Description:
 *     This routine, which is allocation algorithm specific, is called
 *     to allow the algorithm to perform any necessary initialisation when
 *     a new target field is set up.
 *
 *     It is passed a pointer to the method information structure, which
 *     amongst other things, contains an array of Pivot information structures,
 *     which give the details of the pivots (home position, whether broken
 *     or not, which spectrograph they are associated with, etc), a pointer to
 *     the array of Target information structures which give the details of
 *     the possible targets (position, associated spectrograph, etc.), and
 *     a pointer to the array of structures describing the current allocations,
 *     if any. The report function should be called at regular intervals
 *     if initialisation takes a significant time, to allow a user interface
 *     to provide some feedback to the user.
 */

static void ConfigMethodInit (
   const CONFIG_MethodInfo *MethodInfo,
   CONFIG_ReportFunction
            ReportFunction,          /* (>) Feedback routine to call */
   void *FunctionArg,                /* (>) Caller-supplied argument */
   StatusType *Status)               /* (!) Inherited status */
{
   /*  Local variables  */
   
   double Alpha;                     /* Non-radial pivot angle for fibre */
   int Bytes;                        /* Bytes of dynamic memory required */
   double DeltaX;                    /* X displacement of fibre */
   double DeltaY;                    /* Y displacement of fibre */
   double DistanceSquared;           /* Square of required fibre length */
   int FibreType;                    /* Type of fibre - guide or otherwise */
   int IPiv;                         /* Index through pivots */
   int ITarget;                      /* Index through targets */
   double MaxExtensionSquared;       /* Square of maximum fibre extension */
   double ParkDistanceSquared;       /* Square of fibre length when parked */
   double PercentDone;               /* Percentage of current task completed */
   long PivotX;                      /* Pivot home position in X */
   long PivotY;                      /* Pivot home position in Y */
   int ReportPivs;                   /* Report after processing this many */
   char TargetSpect;                 /* Spectrograph code for target */
   char TargetType;                  /* Type of target - guide or otherwise */
   double TargetX;                   /* Target position in X */
   double TargetY;                   /* Target position in Y */
   double Theta;                     /* Theta value for possible allocation */
   double ThetaFib;                  /* Theta for the fibre - from the pivot */
   
   CONFIG_METHOD_MatrixElement **Matrix;
                                     /* Local synonym for .Matrix in globals */
   CONFIG_METHOD_MatrixElement *MatrixAddr;
                                     /* Start of memory allocated for matrix */
   CONFIG_METHOD_MatrixElement **MatrixLineAddr;   
                                     /* Addr of memory allocated to hold 
                                      * addresses of lines of matrix */
   CONFIG_METHOD_MatrixElement *MatrixPtr;
                                     /* Points to matrix line addresses */
                                      
   /*  Check all the dynamically allocated memory addresses, and free
    *  any that are still allocated from previous calls.
    */
    
   if (CONFIG_METHOD_Global.Matrix != NULL) {
      free (CONFIG_METHOD_Global.Matrix);
      CONFIG_METHOD_Global.Matrix = NULL;
   }
   if (CONFIG_METHOD_Global.MatrixAddr != NULL) {
      free (CONFIG_METHOD_Global.MatrixAddr);
      CONFIG_METHOD_Global.MatrixAddr = NULL;
   }

   /*  For the purposes of reporting progress, we pick a suitable number
    *  of pivots. Here we report every 10% of changes, since this should
    *  be a pretty rapid operation.
    */

   ReportPivs = (int) ((float) MethodInfo->NumberPivots / 10.0);
   if (ReportPivs < 0) ReportPivs = 1;
    
   /*  We allocate the matrix that holds all the possible allocations,
    *  there being one entry for each pivot-target combination - ie the
    *  matrix is an array NPiv by NTargets, each element simply indicating
    *  whether or not the corresponding allocation is possible. So we
    *  start by setting it up on the basis simply of 'can the fibre from
    *  this pivot reach this target?'. Then as we start to make allocations
    *  each new allocation will invalidate other possible allocations, and
    *  these have to be worked out as we go.
    */
    
   Bytes = MethodInfo->NumberPivots * 
           MethodInfo->NumberTargets * 
           sizeof(CONFIG_METHOD_MatrixElement);

   if ((MatrixAddr = (CONFIG_METHOD_MatrixElement *) malloc (Bytes)) == NULL) {
      char s[100];
      sprintf(s,
              "Unable to allocate memory for allocation matrix of %d bytes",
              Bytes);
      MethodInfo->ErrRepRoutine(0,s);

      *Status = CONFIG__NO_MEMORY;
      goto Exit;
   }
   Bytes = MethodInfo->NumberPivots * sizeof(CONFIG_METHOD_MatrixElement *);
   if ((MatrixLineAddr = 
                   (CONFIG_METHOD_MatrixElement **) malloc (Bytes)) == NULL) {

       char s[100];
       sprintf(s,
               "Unable to allocate memory for matrix addresses (%d bytes)",
               Bytes);
       MethodInfo->ErrRepRoutine(0,s);
       *Status = CONFIG__NO_MEMORY;
       goto Exit;
   }
   
   /*  Set up the matrix line addresses in the global array used for this
    *  purpose. Note that the effect is that .Matrix can be accessed as if
    *  it were a 2D array, ie .Matrix[IPiv][ITarget] and this generates
    *  references to the appropriate elements of the allocation matrix.
    */

   CONFIG_METHOD_Global.Matrix = MatrixLineAddr;
   CONFIG_METHOD_Global.MatrixAddr = MatrixAddr;
   
   /*  For convenience, we access the Matrix from now on through the
    *  'Matrix' variable, which operates syntatically just like a 
    *  2D array (although that isn't what it is). 
    */

   Matrix = CONFIG_METHOD_Global.Matrix;
   
   MatrixPtr = MatrixAddr;
   for (IPiv = 0; IPiv < MethodInfo->NumberPivots; IPiv++) {
      CONFIG_METHOD_Global.Matrix[IPiv] = MatrixPtr;
      MatrixPtr += MethodInfo->NumberTargets;
   }
   
   /*  Initially, we run through all the elements of the array calculating
    *  whether or not an allocation is ever possible, independent of any
    *  other allocations. So this checks for invalid bend angles, excessive
    *  fibre lengths, incompatible targets & pivots, etc.
    */  
   for (IPiv = 0; IPiv < MethodInfo->NumberPivots; IPiv++) {
      if (ReportFunction != NULL) {
         if ((IPiv % ReportPivs) == 0) {
            PercentDone = ((float) IPiv / 
                                 (float) MethodInfo->NumberPivots) * 100.0;
            if ((*ReportFunction)(FunctionArg,'P',"Setting up path matrix",
                 0,0,0,0,0,0.0,PercentDone) != 0) {
                MethodInfo->InfoRoutine(MethodInfo->InfoArgument,
                                        "Cancel request received");
               *Status = CONFIG__CANCELLED;
               goto Exit;
            }
         }
      } 
      PivotX = MethodInfo->PivotInfo[IPiv].X;
      PivotY = MethodInfo->PivotInfo[IPiv].Y;
      MaxExtensionSquared = (double)MethodInfo->PivotInfo[IPiv].FibreLength;
      MaxExtensionSquared = MaxExtensionSquared * MaxExtensionSquared;
      FibreType = MethodInfo->PivotInfo[IPiv].FibreType;
      for (ITarget = 0; ITarget < MethodInfo->NumberTargets; ITarget++) {
         Matrix[IPiv][ITarget].CollidingPivot = 0;
         TargetX = (double) MethodInfo->TargetInfo[ITarget].X;
         TargetY = (double) MethodInfo->TargetInfo[ITarget].Y;
         DeltaX = (double) PivotX - TargetX;
         DeltaY = (double) PivotY - TargetY;
         DistanceSquared = (DeltaX * DeltaX) + (DeltaY * DeltaY);
         DeltaX = (double) PivotX - MethodInfo->PivotInfo[IPiv].ParkX;
         DeltaY = (double) PivotY - MethodInfo->PivotInfo[IPiv].ParkY;
         ParkDistanceSquared = (DeltaX * DeltaX) + (DeltaY * DeltaY);
         if ((DistanceSquared < MaxExtensionSquared) && 
                      (DistanceSquared > ParkDistanceSquared)) {
            TargetType = MethodInfo->TargetInfo[ITarget].Type;
            TargetSpect = MethodInfo->TargetInfo[ITarget].Spect;
            if (!MethodInfo->PivotInfo[IPiv].AllowAllocChange) {
               /*
                *  We are not allowed to change the allocation of this pivot.
                */
               Matrix[IPiv][ITarget].PathStatus = CONFIG_METHOD_IGNORE;

            } else if (!MethodInfo->FibreTargetCompatible(MethodInfo,
                                           TargetType,TargetSpect,FibreType)) {
               /* 
                *  See if the pivot and the target are of incompatible types 
                */
               Matrix[IPiv][ITarget].PathStatus = CONFIG_METHOD_INCOMPATIBLE;

            } else {
               /*
                * Check if the bend angle would allow the allocation
                */
               if (MethodInfo->CheckAlpha(
                      MethodInfo,
                      &(MethodInfo->TargetInfo[ITarget]),
                      &(MethodInfo->PivotInfo[IPiv]),&Alpha)) {
                  /*
                   * Bend angle would allow the allocation.  Caculate the
                   * theta angle we would get and then check against invalid
                   * positions
                   */
                  ThetaFib = MethodInfo->ThetaFib (
                      MethodInfo,
                      &(MethodInfo->TargetInfo[ITarget]),
                      &(MethodInfo->PivotInfo[IPiv]));
                  Theta = (ThetaFib > PI) ? ThetaFib - PI : ThetaFib + PI;
                  if (!MethodInfo->ColInvPos(MethodInfo,FibreType,
                                            TargetX,TargetY,Theta)){
                                            
                      /*  Almost there. We need to see if the allocation would
                       *  result in a fibre crossing an adjacent parked pivot
                       *  position. Note that for some systems (2dF) this isn't
                       *  a problem - and the test routine always returns OK.
                       *  For others (OzPoz, 6dF) we don't want to cover a
                       *  parked button, or EVEN to cover the position where a
                       *  pivot goes when parked, since this can make things
                       *  almost impossible for the delta task, even if the
                       *  configuration is technically feasible.
                       */
                      
                      if (MethodInfo->AdjPark(MethodInfo,IPiv,TargetX,
                                                       TargetY,Theta)) {
                         Matrix[IPiv][ITarget].PathStatus = 
                                                   CONFIG_METHOD_PARK_CLASH;
                      } else {
                      
                          /*
                           * Position is ok.  But ensure we don't do this if 
                           * the fibre is broken;
                           */
                          Matrix[IPiv][ITarget].PathStatus = 
                                                     CONFIG_METHOD_POSSIBLE;
                          /*  Just for the template code, 
                           *  I removed this, because I don't think its
                           *  needed for the basic matrix - but I'm
                           *  not sure I've not removed the only test for
                           *  a broken fibre (KS).
                          Matrix[IPiv][ITarget].Alpha = Alpha;
                          FibreType = MethodInfo->PivotInfo[IPiv].FibreType;
                          if (!MethodInfo->FibreIsBroken(MethodInfo,
                                                             FibreType)) {
                             TargetMatrix[ITarget].Pivot[TargetMatrix[ITarget].
                                                              NumPivots]=IPiv;
                             TargetMatrix[ITarget].NumPivots++;
                          }
                          */
                      }
                   } else {
                      /* 
                       * Position would collide with an invalid positon
                       */
                      Matrix[IPiv][ITarget].PathStatus = CONFIG_METHOD_SCREWED;
                      /*
                       * If already allocated, output a warning.
                       */
                      if ((MethodInfo->AllocInfo[IPiv].State == 'A') && 
                         (MethodInfo->AllocInfo[IPiv].TargetIndex == ITarget)){
                         /*  This printf() shouldn't be here - either don't do
                          *  the test or produce a proper error message! *** */
                         printf("Invalid position (screw ?)");
                         printf("  Collision set for allocation: %d %d\n",
                                                              IPiv,ITarget);
                      }
                   }
               } else {
                  /*
                   *  Bend angle would prevent allocation.
                   */
                  Matrix[IPiv][ITarget].PathStatus = CONFIG_METHOD_BAD_ALPHA;
                  /*
                   *  If already allocated, output a warning.
                   */
                  if ((MethodInfo->AllocInfo[IPiv].State == 'A') && 
                      (MethodInfo->AllocInfo[IPiv].
                                      TargetIndex == ITarget)){
                    printf("Bad Alpha Angle set for allocation: %d %d\n"
                           ,IPiv,ITarget);
                  }
                }
             }
          } else {
            /*
             * Pivot won't reach target.
             */
            Matrix[IPiv][ITarget].PathStatus = CONFIG_METHOD_WONT_REACH;
            /*
             *  If already allocated, output a warning.
             */
            if ((MethodInfo->AllocInfo[IPiv].State == 'A') && 
                (MethodInfo->AllocInfo[IPiv].
                 TargetIndex == ITarget)){
                printf("Bad Extension set for allocation: %d %d\n"
                       ,IPiv,ITarget);
            }

          }
       }
    }
    
   /*  We now do a pass through the matrix, setting the theta values for
    *  those allocations that are possible. In each case, we assume no twist
    *  of the pivot, and so simply calculate the theta value that corresponds
    *  to a straight button, in line with the fibre.
    */

   for (IPiv = 0; IPiv < MethodInfo->NumberPivots; IPiv++) {
      if (ReportFunction != NULL) {
         if ((IPiv % ReportPivs) == 0) {
            PercentDone = ((float) IPiv / 
                                 (float) MethodInfo->NumberPivots) * 100.0;
            if ((*ReportFunction)(FunctionArg,'P',
                       "Setting radial button angles",0,0,0,0,0,0.0,
                                                     PercentDone) != 0) {
                MethodInfo->InfoRoutine(MethodInfo->InfoArgument,
                                        "Cancel request received");
                *Status = CONFIG__CANCELLED;
                goto Exit;
            }
         }
      } 
      for (ITarget = 0; ITarget < MethodInfo->NumberTargets; ITarget++) {
         Theta = 0.0;
         if (Matrix[IPiv][ITarget].PathStatus == CONFIG_METHOD_POSSIBLE) {
            ThetaFib = MethodInfo->ThetaFib (
                         MethodInfo,
                         &(MethodInfo->TargetInfo[ITarget]),
                                       &(MethodInfo->PivotInfo[IPiv]));
            Theta = (ThetaFib > PI) ? ThetaFib - PI : ThetaFib + PI;
         }
         Matrix[IPiv][ITarget].Theta = Theta;           
      }
   }
   
Exit:;

}

/* -------------------------------------------------------------------------- */

/*              C o n f i g  M e t h o d  S e t  S k y
 *
 *  Function:
 *     Sets the number of fibres of a given type to be allocated to sky.
 *
 *  Description:
 *     This routine allows the higher levels to exercise precise control
 *     over the number of fibres of each type that ought to be allocated
 *     to sky targets. This is a requirement that all allocation algorithms
 *     need to support, so it is more appropriate to handle it in this way
 *     than through the method-specific parameters. The settings laid down
 *     by a call to this routine should be taken into account by the next
 *     call to ConfigMethodAllocate(). The settings made by this routine 
 *     are only relevant for fibre types that can be allocated to both
 *     sky and non-sky targets. 
 */

static void ConfigMethodSetSky (
   const CONFIG_MethodInfo *MethodInfo,
   int FibreType,                    /* (>) Fibre type code */
   int NumberSkyAllocations,         /* (>) Number of fibres for sky targets */
   StatusType *Status)               /* (!) Inherited status */
{
   /*  Local variables */
   
   if (*Status != STATUS__OK) return;
      
}   


/* -------------------------------------------------------------------------- */

/*              C o n f i g  M e t h o d  A l l o c a t e
 *
 *  Function:
 *     Performs a full allocation of fibres to targets.
 *
 *  Description:
 *     This is the method-specific layer routine that actually performs
 *     an allocation.
 */

static void ConfigMethodAllocate (
   const CONFIG_MethodInfo *MethodInfo,
   CONFIG_ReportFunction
            ReportFunction,          /* (>) Feedback routine to call */
   void *FunctionArg,                /* (>) Caller-supplied argument */
   StatusType *Status)               /* (!) Inherited status */
{
   /*  Local variables */
   
   const CONFIG_AllocInfo *AllocInfo;/* Allocation info array of structures */
   int DeallocFibre;                 /* Fibre to be deallocated */
   int IPiv;                         /* Loop index through pivots */
   int ITarget;                      /* Loop index through targets */
   CONFIG_METHOD_MatrixElement **Matrix;
                                     /* Local synonym for .Matrix in globals */
   int NPiv;                         /* Number of pivots */
   int NTargets;                     /* Number of targets */
   double Theta;                     /* Theta value for possible allocation */
   double ThetaFib;                  /* Theta for the fibre - from the pivot */
   
   Matrix = CONFIG_METHOD_Global.Matrix;
   NTargets = MethodInfo->NumberTargets;
   NPiv = MethodInfo->NumberPivots;
   AllocInfo = MethodInfo->AllocInfo;
   
   DeallocFibre = CONFIG_METHOD_Global.
                     ParameterValues[CONFIG_METHOD_DEALLOC_PARM].IntegerValue;

   /*  We go through each pivot in turn  */

   for (IPiv = 0; IPiv < NPiv; IPiv++) {

      /*  If the pivot is already allocated, or broken, we ignore it.  An
       *  eligible pivot will have a state given as 'U' for 'unallocated'
       */

      if (AllocInfo[IPiv].State == 'U') {

         /*  Look through the matrix for this pivot, looking for any
          *  possible targets. When we find one, we allocate it.
          */

         for (ITarget = 0; ITarget < NTargets; ITarget++) {
            if (Matrix[IPiv][ITarget].PathStatus == CONFIG_METHOD_POSSIBLE) {
               printf ("Allocating pivot %d to target %d\n",IPiv,ITarget);
               ThetaFib = MethodInfo->ThetaFib (MethodInfo,
                                   &(MethodInfo->TargetInfo[ITarget]),
                                           &(MethodInfo->PivotInfo[IPiv]));
               Theta = Matrix[IPiv][ITarget].Theta;
               ConfigMethodAllocUpdate(MethodInfo,IPiv,ITarget,Theta,Status);
               
               /*  Just to check that the deallocate routine works, we
                *  de-allocate the fibre if it's the one selected in
                *  the 'don't allocate' parameter. - This also shows
                *  how parameters can be used.
                */
                
               if ((IPiv + 1) == DeallocFibre) {
                  printf ("Deallocating pivot %d\n",IPiv);
                  ConfigMethodDeallocUpdate (MethodInfo,IPiv,ITarget,Status);
               }
               
               break;
            }
         }
      }
   }
   
   
} 

/* -------------------------------------------------------------------------- */

/*        C o n f i g  M e t h o d  A l l o c  U p d a t e
 *
 *  Function:
 *     Updates the method-specific global structures after a new allocation.
 *
 *  Description:
 *     After a new allocation of pivot and target is made, this routine 
 *     should be called to update the various method-specific global 
 *     structures. It sets the possible path matrix elements
 *     for the actual pivot and fibre involved (ie one row and one column),
 *     and checks for the effects of the new allocation upon the
 *     other possible paths. It makes the NoteAllocation() callback so that 
 *     the global allocation structure is updated with the new allocation.
 *
 *  Note:
 *     This routine is a combination of the code from ConfMethodSetAlloc()
 *     and ConfMethodNewAlloc() in the original config_method.c file.
 */

static void ConfigMethodAllocUpdate (
   const CONFIG_MethodInfo *MethodInfo,
   int Pivot,                     /* (>) The newly allocated pivot */
   int Target,                    /* (>) The newly allocated target */
   double Theta,                  /* (>) The theta angle for the button */
   StatusType *Status)            /* (!) Inherited status variable */
{

   /*  Local variables  */

   int Collision;                 /* True if there is a collision */
   int ITarget;                   /* Loop index through targets */
   int IPiv;                      /* Loop index through pivots */
   CONFIG_METHOD_MatrixElement **Matrix;
                                  /* Local synonym for .Matrix in globals */
   long NewTargetX;               /* X-coordinate of newly allocated target */
   long NewTargetY;               /* Y-coordinate of newly allocated target */
   long NewPivotX;                /* X-coordinate of newly allocated pivot */
   long NewPivotY;                /* Y-coordinate of newly allocated pivot */
   int NPiv;                      /* Number of pivots */
   int NTargets;                  /* Number of targets */
   const CONFIG_PivotInfo *PivotInfo;/* Pivot info array of structures */
   long PivotX;                   /* X-coordinate of allocated pivot */
   long PivotY;                   /* Y-coordinate of allocated pivot */
   int  fibreType;                /* fibre type of previously allocated pivot */
   const CONFIG_TargetInfo *TargetInfo; /* Target info array of structures */
   long TargetX;                  /* X-coordinate of allocated target */
   long TargetY;                  /* Y-coordinate of allocated target */

   if (*Status != STATUS__OK) return;
   
   /*  Pick up local pointers to the various information structures whose
    *  addresses are held in the method-specific global structure. Also
    *  pick up the number of possible targets and pivots.
    */

   Matrix = CONFIG_METHOD_Global.Matrix;
   NTargets = MethodInfo->NumberTargets;
   NPiv = MethodInfo->NumberPivots;
   PivotInfo = MethodInfo->PivotInfo;
   TargetInfo = MethodInfo->TargetInfo;

   /*  Call the general non-method specific utility routine to update the
    *  global allocation information structure array. (This also checks that
    *  Pivot and Target are valid, which is a useful internal check.)
    *  Note, the tentative flag in the allocation is not set.
    */
    
   MethodInfo->NoteAllocation (MethodInfo,Pivot,Target,Theta,0,Status);
   if (*Status != STATUS__OK) goto Exit;

   /*  We make the internal check that the allocation shown is one that the
    *  possible path matrix regards as possible.  If it doesn't, then there's
    *  a problem here.
    */

   if (Matrix[Pivot][Target].PathStatus != CONFIG_METHOD_POSSIBLE) {
       char s[100];
       MethodInfo->TentativeFix(MethodInfo, Pivot, 0, Status);
       sprintf(s,"Invalid allocation, pivot %d to target %d attempted",
               Pivot + 1, Target + 1);
       MethodInfo->ErrRepRoutine(0,s);
       *Status = CONFIG__INVAL_ALLOC;
       goto Exit;
   }

   /*  Since this pivot is now allocated, we have to modify all the entries
    *  in the possible allocation matrix to reflect this. Other allocations
    *  involving this pivot or this target are now no longer possible.
    */
   
   for (ITarget = 0; ITarget < NTargets; ITarget++) {
       if (Matrix[Pivot][ITarget].PathStatus == CONFIG_METHOD_POSSIBLE) {
               Matrix[Pivot][ITarget].PathStatus = CONFIG_METHOD_FIBRE_IN_USE;
       }
   }
   for (IPiv = 0; IPiv < NPiv; IPiv++) {
       if (Matrix[IPiv][Target].PathStatus == CONFIG_METHOD_POSSIBLE) {
           Matrix[IPiv][Target].PathStatus = CONFIG_METHOD_TARGET_IN_USE;
       }
   }
   
   /*  And we flag this allocation as made */
   
   Matrix[Pivot][Target].PathStatus = CONFIG_METHOD_ALLOCATED;
   Matrix[Pivot][Target].Theta = Theta;
   
   /*  Now we need to consider the effect of the new allocation on
    *  the other possible allocations given in the matrix. Any allocation
    *  flagged as possible in the matrix has to be tested to see if it
    *  is invalidated by this new allocation.
    */
    
   NewTargetX = TargetInfo[Target].X;
   NewTargetY = TargetInfo[Target].Y;
   NewPivotX = PivotInfo[Pivot].X;
   NewPivotY = PivotInfo[Pivot].Y;
   
   /*  We need to pass through every element of the matrix. In every case
    *  where an allocation is flagged as being possible, we have to consider
    *  whether this new allocation has invalidated it.
    */

   for (IPiv = 0; IPiv < NPiv; IPiv++) {
      PivotX = PivotInfo[IPiv].X;
      PivotY = PivotInfo[IPiv].Y;
      fibreType = PivotInfo[IPiv].FibreType;
      for (ITarget = 0; ITarget < NTargets; ITarget++) {
         if (Matrix[IPiv][ITarget].PathStatus == CONFIG_METHOD_POSSIBLE) {

            /*  This allocation is flagged as possible. We need to check
             *  whether there is a collision between the new allocation
             *  button and the potential allocation represented by the 
             *  matrix entry.
             */

            Collision = CONFIG_FALSE;
            TargetX = TargetInfo[ITarget].X;
            TargetY = TargetInfo[ITarget].Y;

            Collision = MethodInfo->Collision(MethodInfo, 
                                      NewTargetX, NewTargetY,
                                      Theta,
                                      NewPivotX, NewPivotY,
                                      TargetX, TargetY, 
                                      Matrix[IPiv][ITarget].Theta, 
                                      PivotX, PivotY);

            if (Collision) {
               Matrix[IPiv][ITarget].PathStatus = CONFIG_METHOD_COLLISION;
               Matrix[IPiv][ITarget].CollidingPivot = Pivot;
            }

            if (!Collision) {
              if (MethodInfo->ColInvPos(MethodInfo,fibreType,
                                        NewTargetX,NewTargetY,Theta)) {
                Matrix[IPiv][ITarget].PathStatus = CONFIG_METHOD_SCREWED;
                Collision = CONFIG_TRUE;
              }
            }
         }
      }
   }
   
Exit:;
   
}
                                                                      
/* -------------------------------------------------------------------------- */

/*          C o n f i g  M e t h o d  D e a l l o c  U p d a t e
 *
 *  Function:
 *     Revises the possible path matrix in the light of a deallocation.
 *
 *  Description:
 *     This routine is called whenever a deallocation is made to update
 *     the set of possible allocations maintained in the matrix of all
 *     possible paths between pivots and targets. When an allocation is
 *     cancelled, this may make valid some of the allocation paths, previously
 *     flagged as impossible because of collision with the now deallocated
 *     fibre.
 *
 *  Note:
 *     The code for this routine is essentially the same as that in
 *     ConfigMethodDealloc() in the original config_method.c, but it has
 *     been simplified by removing the report routine calls - it all
 *     happens too fast for this to be needed. Some other simplifications
 *     to do with removing Oxford algorithm-specific code have also been 
 *     made.
 */

static void ConfigMethodDeallocUpdate (
   const CONFIG_MethodInfo *MethodInfo,
   int Pivot,                     /* (>) The number of the deallocated pivot */
   int Target,                    /* (>) Target allocated to this pivot */
   StatusType *Status)            /* (!) Inherited status */
{

   /*  Local variables  */

   int CollidingPivot;           /* Pivot causing collision - ignored */
   int ITarget;                  /* Loop index through targets */
   int IPiv;                     /* Loop index through pivots */
   CONFIG_METHOD_MatrixElement **Matrix;
                                 /* Local synonym for .Matrix in globals */
   int NPiv;                     /* Number of pivots */
   int NTargets;                 /* Number of targets */

   if (*Status != STATUS__OK) return;

   /*  Pick up local pointers to the various information structures whose
    *  addresses are held in the method-specific global structure. Also
    *  pick up the number of possible targets and pivots.
    */

   Matrix = CONFIG_METHOD_Global.Matrix;
   NTargets = MethodInfo->NumberTargets;
   NPiv = MethodInfo->NumberPivots;

   /*  Call the housekeeping utility routine to record this deallocation
    *  in the allocation information structure array.    */

   MethodInfo->NoteDeallocation (MethodInfo,Pivot,Status);

   if (*Status != STATUS__OK) goto Exit;

   /*  We reset the matrix elements for the pivot and the target that it
    *  was assigned to.  Any that were flagged as 'in use' now may be
    *  possible again - we have to check them. The element for the allocation
    *  that has just been deallocated must be possible.
    */

   Matrix[Pivot][Target].PathStatus = CONFIG_METHOD_POSSIBLE;
   for (IPiv = 0; IPiv < NPiv; IPiv++) {
      if (Matrix[IPiv][Target].PathStatus == CONFIG_METHOD_TARGET_IN_USE) {
         if (MethodInfo->CheckAllocation(MethodInfo,IPiv,Target,
                 Matrix[IPiv][Target].Theta,&CollidingPivot)) {
             Matrix[IPiv][Target].PathStatus = CONFIG_METHOD_POSSIBLE;
         }
      }
   }
   for (ITarget = 0; ITarget < NTargets; ITarget++) {
      if (Matrix[Pivot][ITarget].PathStatus == CONFIG_METHOD_FIBRE_IN_USE) {
         if (MethodInfo->CheckAllocation(MethodInfo,Pivot,ITarget,
                            Matrix[Pivot][ITarget].Theta,&CollidingPivot)) {
            Matrix[Pivot][ITarget].PathStatus = CONFIG_METHOD_POSSIBLE;
         }
      }
   }

   /*  We run through all the possible paths given in the matrix. Some of
    *  these will be flagged as impossible because of a collision, and these
    *  are the ones we're interested in. For each of these, the number of
    *  the fibre for which the collision was detected is recorded in the
    *  matrix. If this colliding fibre is not the one we have just deallocated,
    *  then clearly the allocation is still invalid. If the colliding fibre
    *  recorded in the matrix is the one we've just deallocated, then this
    *  doesn't automatically mean that the path is now valid, since there
    *  may be other allocated fibres that it will still collide with (a path
    *  may collide with many fibres, but only one is recorded in the matrix).
    *  So in this case we need to do a full validity check on the allocation.
    */

   for (IPiv = 0; IPiv < NPiv; IPiv++) {
      for (ITarget = 0; ITarget < NTargets; ITarget++) {
         if (Matrix[IPiv][ITarget].PathStatus == CONFIG_METHOD_COLLISION) {
            if (Matrix[IPiv][ITarget].CollidingPivot == Pivot) {
               if (MethodInfo->CheckAllocation(MethodInfo,IPiv,ITarget,
                         Matrix[IPiv][ITarget].Theta,&CollidingPivot)) {
                  Matrix[IPiv][ITarget].PathStatus = CONFIG_METHOD_POSSIBLE;
               }
            }
         }
      }
   }

Exit:;

}

/* -------------------------------------------------------------------------- */

/*             C o n f i g  M e t h o d  R e l e a s e  M e m
 *
 *  Function:
 *     Releases any memory allocated by the allocation-specific code.
 *
 *  Description:
 *     This routine is called at the end of the program in order to give 
 *     it a chance to release any dynamic memory that may have been allocated.
 *     Any such memory should have its address recorded in the global data
 *     structures.
 */

static void ConfigMethodReleaseMem (
   const CONFIG_MethodInfo *MethodInfo DUNUSED)
{
   /*  Local variables  */

   int Index;                      /* Index through method parameters */

   /*  Check all the dynamically allocated memory addresses, and free
    *  any that are still allocated.
    */
    
   if (CONFIG_METHOD_Global.Matrix != NULL) {
      free (CONFIG_METHOD_Global.Matrix);
      CONFIG_METHOD_Global.Matrix = NULL;
   }
   if (CONFIG_METHOD_Global.MatrixAddr != NULL) {
      free (CONFIG_METHOD_Global.MatrixAddr);
      CONFIG_METHOD_Global.MatrixAddr = NULL;
   }

   /*  If there are any 'any value' character parameters that have been set,
    *  free any memory allocated to them.
    */

   for (Index = 0; Index < CONFIG_METHOD_NPARMS; Index++) {
      if (CONFIG_METHOD_ParmTypes[Index] == 'C') {
         if (CONFIG_METHOD_ParmNValues[Index] == 0) {
            if (CONFIG_METHOD_Global.ParameterValues[Index].StringAddress
                                                                    != NULL) {
               free (
                 CONFIG_METHOD_Global.ParameterValues[Index].StringAddress);
            }
         }
      }
   }
   
}

/* -------------------------------------------------------------------------- */

/*         C o n f i g  M e t h o d  N u m b e r  P a r a m s
 *
 *  Function:
 *     Returns the number of method-specific parameters.
 *
 *  Description:
 *     This routine returns the number of method-specific parameters that
 *     can be used to control the details of the allocation algorithm.
 */

static int ConfigMethodNumberParams (
    const CONFIG_MethodInfo *MethodInfo DUNUSED)
{
   return (CONFIG_METHOD_NPARMS);
}

/* -------------------------------------------------------------------------- */

/*         C o n f i g  M e t h o d  P a r a m  D e t a i l s
 *
 *  Function:
 *     Returns details about the method-specific parameters.
 *
 *  Description:
 *     This routine returns details about the various method-specific 
 *     parameters that can be used to control the details of the 
 *     allocation algorithm. This includes their names, types, and
 *     possible values.
 */

static void ConfigMethodParamDetails (
   const CONFIG_MethodInfo *MethodInfo DUNUSED,
   int ParameterNumber,
   char **Name,
   char *Type,
   double Limits[],
   int *NValues,
   char ***Values,
   StatusType *Status)
{
   /*  Local variables  */

   int Index;                         /* Index into parameter arrays */

   if (*Status != STATUS__OK) return;

   if ((ParameterNumber < 1) || (ParameterNumber > CONFIG_METHOD_NPARMS)) {
       MethodInfo->ErrRepRoutine(0,
        "Attempt to get details about non-existent method-specific parameter");
      *Status = CONFIG__INVAL_PARM_NO;
   } else {
      Index = ParameterNumber - 1;
      *Name = CONFIG_METHOD_ParmNames[Index];
      *Type = CONFIG_METHOD_ParmTypes[Index];
      Limits[0] = CONFIG_METHOD_ParmLimits[Index][0];
      Limits[1] = CONFIG_METHOD_ParmLimits[Index][1];
      *NValues = CONFIG_METHOD_ParmNValues[Index];
      *Values = CONFIG_METHOD_ParmValues[Index];
   }

}

/* -------------------------------------------------------------------------- */

/*              C o n f i g  M e t h o d  S e t  P a r a m
 *
 *  Function:
 *     Sets the value of one of the method-specific parameters.
 *
 *  Description:
 *     This routine sets the value of one of the various method-specific 
 *     parameters that can be used to control the details of the 
 *     allocation algorithm. The parameter is specified by name, and
 *     the value is specified as a character string.
 */

static void ConfigMethodSetParam (
   const CONFIG_MethodInfo *MethodInfo DUNUSED,
   char *Name,                        /* (>) Name of parameter */
   char *Value,                       /* (>) Value of parameter */
   StatusType *Status)                /* (!) Inherited status value */
{
   /*  Local variables  */

   int Bytes;                         /* Number of bytes for character param */
   char *ExpertStr;                   /* Pointer to "(expert)" string */
   double FloatValue;                 /* Value of floating point param */
   int Index;                         /* Index into parameter arrays */
   long IntegerValue;                 /* Value of integer parameter */
   int Length;                        /* Chars in string to compare */
   double MaxLimit;                   /* Maximum value for parameter */
   double MinLimit;                   /* Minimum value for parameter */
   CONFIG_Boolean NameMatched;        /* True if Name is a valid name */
   int NValues;                       /* Number of values for char param */
   CONFIG_Boolean ParamOK;            /* True if char param is valid string */
   char **ParmValues;                 /* Address of possible char values */
   char *StringAddress;               /* Address of parameter value string */
   int StringIndex;                   /* Index through possible parm values */
   char Type;                         /* PArameter type 'I','F','C','L' */

   if (*Status != STATUS__OK) return;

   /*  Work through the names of the various parameters looking for
    *  a match with that specified.
    */

   NameMatched = CONFIG_FALSE;
   for (Index = 0; Index < CONFIG_METHOD_NPARMS; Index++) {
      if (!strcmp(Name,CONFIG_METHOD_ParmNames[Index])) {

         /*  Name matches, so pick up the type, limits (for numerical values)
          *  and number of possible values (for character parameters).
          */

         NameMatched = CONFIG_TRUE;
         Type = CONFIG_METHOD_ParmTypes[Index];
         MinLimit = CONFIG_METHOD_ParmLimits[Index][0];
         MaxLimit = CONFIG_METHOD_ParmLimits[Index][1];
         NValues = CONFIG_METHOD_ParmNValues[Index];

         /*  Now handle the parameter according to type  */

         if (Type == 'L') {

            /*  Logical parameters must be either 'T' or 'F' */

            if (!strcmp(Value,"T")) {
               CONFIG_METHOD_Global.ParameterValues[Index].IntegerValue = 
                                                                   CONFIG_TRUE;
            } else if (!strcmp(Value,"F")) {
               CONFIG_METHOD_Global.ParameterValues[Index].IntegerValue = 
                                                                  CONFIG_FALSE;
            } else {
                char s[100];
                sprintf(s,
                      "'%s' is a logical parameter and cannot be set to '%s'",
                        Name,Value);
                MethodInfo->ErrRepRoutine(0,s);
               *Status = CONFIG__INVAL_PARM_VALUE;
            }

         } else if (Type == 'I') {

            /*  Integer parameters must be valid integer values within
             *  the specified range.
             */

            if (sscanf(Value,"%ld",&IntegerValue) == 1) {
               if (((double) IntegerValue < MinLimit) ||
                    ((double) IntegerValue > MaxLimit)) {
                   char s[100];
                   sprintf(s,
                  "Parameter '%s' must be between %d and %d. %ld is invalid.",
                           Name,(int) MinLimit, (int) MaxLimit, IntegerValue);
                   MethodInfo->ErrRepRoutine(0,s);
                   *Status = CONFIG__PARM_OUT_RANGE;
               } else {
                  CONFIG_METHOD_Global.ParameterValues[Index].IntegerValue = 
                                                               IntegerValue;
               }
            } else {
                char s[100];
                sprintf(s,
                        "'%s' is not a valid value for integer parameter '%s'",
                        Value,Name);
                MethodInfo->ErrRepRoutine(0,s);
                *Status = CONFIG__INVAL_PARM_VALUE;
            }

         } else if (Type == 'F') {

            /*  Floating point parameters must be valid real values within
             *  the specified range.
             */

            if (sscanf(Value,"%lf",&FloatValue) == 1) {
               if ((FloatValue < MinLimit) || (FloatValue > MaxLimit)) {
                   char s[100];
                   sprintf(s,
                   "Parameter '%s' must be between %f and %f. %f is invalid.",
                           Name,MinLimit,MaxLimit,FloatValue);
                           MethodInfo->ErrRepRoutine(0,s);
                  *Status = CONFIG__PARM_OUT_RANGE;
               } else {
                  CONFIG_METHOD_Global.ParameterValues[Index].DoubleValue = 
                                                                 FloatValue;
               }
            } else {
                char s[100];
                sprintf(s,
                        "'%s' is not a valid value for parameter '%s'",
                        Value,Name);
                MethodInfo->ErrRepRoutine(0,s);
               *Status = CONFIG__INVAL_PARM_VALUE;
            }

         } else if (Type == 'C') {

            /*  Character values can have either any value, or one of a 
             *  specified set. 
             */

            if (NValues <= 0) {

               /*  This paramter can have any value, so it must be valid.
                *  We allocate memory to hold a copy of the new value (this
                *  allows the value to be any length at all), first freeing
                *  any previously allocated.
                */

               if (CONFIG_METHOD_Global.ParameterValues[Index].StringAddress
                                                                     != NULL) {
                  free (CONFIG_METHOD_Global.ParameterValues[Index]
                                                              .StringAddress);
               }
               Bytes = strlen(Value);
               if (Bytes == 0) {
                  CONFIG_METHOD_Global.ParameterValues[Index].StringAddress 
                                                                      = NULL;
               } else {
                  StringAddress = (char *) malloc(Bytes + 1);
                  strcpy (StringAddress,Value);
                  CONFIG_METHOD_Global.ParameterValues[Index].StringAddress
                                                             = StringAddress;
               }
            } else {

               /*  This character string must have one of a set of specified
                *  values. Check that this is one, and set the index in
                *  the global variables. Allow for the possibility that the
                *  "(expert)" may have been stripped off.
                */

               ParmValues = CONFIG_METHOD_ParmValues[Index];
               ParamOK = CONFIG_FALSE;
               for (StringIndex = 0; StringIndex < NValues; StringIndex++) {
                  if ((ExpertStr = 
                             strstr(ParmValues[StringIndex],"(expert)"))) {
                     Length = ExpertStr - ParmValues[StringIndex];
                  } else {
                     Length = strlen(Value);
                  }
                  if (!strncmp(Value,ParmValues[StringIndex],Length)) {
                     CONFIG_METHOD_Global.ParameterValues[Index].StringAddress
                                                    = ParmValues[StringIndex];
                     CONFIG_METHOD_Global.ParameterValues[Index].IntegerValue
                                                                = StringIndex;
                     ParamOK = CONFIG_TRUE;
                     break;
                  }
               }
               if (!ParamOK) {
                   char s[100];
                   sprintf(s,
                           "'%s' is not one of the possible values for '%s'",
                           Value,Name);
                   MethodInfo->ErrRepRoutine(0,s);
                   *Status = CONFIG__INVAL_PARM_VALUE;
               }
            }
         }

         /*  Finished processing a matched name, so break out of the loop
          *  through the various parameter names.
          */

         break;
      }
   }

   if (!NameMatched) {
       char s[100];
       sprintf(s,"'%s' is not a recognised parameter name",Name);
       MethodInfo->ErrRepRoutine(0,s);
      *Status = CONFIG__INVAL_PARM_NAME;
   }
}
                  
/* -------------------------------------------------------------------------- */

/*              C o n f i g  M e t h o d  G e t  P a r a m
 *
 *  Function:
 *     Gets the value of one of the method-specific parameters.
 *
 *  Description:
 *     This routine gets the value of one of the various method-specific 
 *     parameters that can be used to control the details of the 
 *     allocation algorithm. The parameter is specified by name, and
 *     the value is returned as a character string.
 */

static void ConfigMethodGetParam (
   const CONFIG_MethodInfo *MethodInfo DUNUSED,
   char *Name,                        /* (>) Name of parameter */
   char *Value,                       /* (<) Value of parameter */
   int VSize,                         /* (>) Maximum number of characters */
   StatusType *Status)                /* (!) Inherited status value */
{
   /*  Local variables  */

   int Index;                         /* Index into parameter arrays */
   CONFIG_Boolean NameMatched;        /* True if Name is a valid name */
   char Type;                         /* Parameter type 'I','F','C','L' */
   char *ValuePtr;                    /* Points to string to be returned */
   char ValueString[64];              /* Used to format numbers and logicals */

   if (*Status != STATUS__OK) return;

   /*  Work through the names of the various parameters looking for
    *  a match with that specified.
    */

   NameMatched = CONFIG_FALSE;
   ValuePtr = NULL;
   for (Index = 0; Index < CONFIG_METHOD_NPARMS; Index++) {
      if (!strcmp(Name,CONFIG_METHOD_ParmNames[Index])) {

         /*  Name matches, so pick up its type.
          */

         NameMatched = CONFIG_TRUE;
         Type = CONFIG_METHOD_ParmTypes[Index];

         /*  Now handle the parameter according to type  */

         if (Type == 'L') {

            /*  Logical parameters are returned as either 'T' or 'F' */

            if (CONFIG_METHOD_Global.ParameterValues[Index].IntegerValue) {
               ValuePtr = "T";
            } else {
               ValuePtr = "F";
            }

         } else if (Type == 'I') {

            /*  Integer parameters are formatted into a value string.
             */

            sprintf(ValueString,"%d",
                CONFIG_METHOD_Global.ParameterValues[Index].IntegerValue);
            ValuePtr = ValueString;

         } else if (Type == 'F') {

            /*  Floating point parameters are also formatted into a value
             *  string.
             */

            sprintf(ValueString,"%f",
                CONFIG_METHOD_Global.ParameterValues[Index].DoubleValue);
            ValuePtr = ValueString;

         } else if (Type == 'C') {

            /*  Character values can have either any value, or one of a 
             *  specified set. In either case, the address of the current
             *  value is held in the global variables.
             */

            ValuePtr = 
                 CONFIG_METHOD_Global.ParameterValues[Index].StringAddress;
         }

         /*  Finished processing a matched name, so break out of the loop
          *  through the various parameter names.
          */

         break;
      }
   }

   /*  After working through the various accepted parameter names, we either
    *  didn't find the name specified, in which case we have an error, or
    *  we have a pointer (ValuePtr) to a string giving the current value.
    */

   if (NameMatched) {
      if (ValuePtr == NULL) {
         Value[0] = '\0';
      } else {
         strncpy (Value,ValuePtr,VSize);
         Value[VSize - 1] = '\0';
      }
   } else {
       char s[100];
       sprintf(s,"'%s' is not a recognised parameter name",Name);
      MethodInfo->ErrRepRoutine(0,s);
      *Status = CONFIG__INVAL_PARM_NAME;
   }
}


/* -------------------------------------------------------------------------- */

/*          C o n f i g  M e t h o d  S e t  I m p o r t  F i l e
 *
 *  Function:
 *     Set import mode and the name of the file to import.
 *
 *  Description:
 *     This routine is called directly by ConfigSetImportFile(). This
 *     is not well defined at the moment but can be considered a way of
 *     importing allocation information into the allocation layer. This
 *     interface may be replaced in the re-working of the housekeeping
 *     layer to support additional facilities provided by different
 *     allocation algorithms.
 */

static void ConfigMethodSetImportFile(
   const CONFIG_MethodInfo *MethodInfo DUNUSED,
   char* name DUNUSED)
{
}


/* -------------------------------------------------------------------------- */

/*                C o n f i g  M e t h o d  D e t a i l s
 *
 *  The lone external routine in this module.  Allows the Config library
 *  to featch details of the method.
 */

extern void ConfigMethodDetails(CONFIG_MethodDetails *MethodDetails)
{
    MethodDetails->Version = CONFIG_VERSION;
    MethodDetails->RoutineCount = CONFIG_MTH_N_ROUTINES;
    MethodDetails->clientData = 0;  /* Not currently used */

    strncpy(MethodDetails->Name, "Template scheme", CONFIG_MTH_NAME_LEN-1);
    MethodDetails->Name[CONFIG_MTH_NAME_LEN-1] = '\0';

    MethodDetails->GlobalInit    = ConfigMethodGlobalInit;
    MethodDetails->Init          = ConfigMethodInit;
    MethodDetails->ReleaseMem    = ConfigMethodReleaseMem;
    MethodDetails->SetSky        = ConfigMethodSetSky;
    MethodDetails->Allocate      = ConfigMethodAllocate;
    MethodDetails->NumParams     = ConfigMethodNumberParams;
    MethodDetails->ParamDetails  = ConfigMethodParamDetails;
    MethodDetails->SetParam      = ConfigMethodSetParam;
    MethodDetails->GetParam      = ConfigMethodGetParam;
    MethodDetails->SetImportFile = ConfigMethodSetImportFile;
}

/*
         1         2         3         4         5         6         7         8
12345678901234567890123456789012345678901234567890123456789012345678901234567890
*/
