/*+
 *                       C O N F I G U R E _ S U B S
 *
 *  Function:
 *     Utility C subroutines used by 'Configure'.
 *
 *  Description:
 *     This file contains a set of C-callable routines that perform a variety
 *     of specific tasks for the main 'Configure' code. These are generally
 *     connected with housekeeping for the main configuration structure. 
 *     These are intimately connected with the routines in configure.c
 *     (even using some global variables declared in that file), and were
 *     originally included in that file. They were moved into this separate
 *     file to make it easier to write programs such as 'testharness', used
 *     to provide a relatively simple, non-interactive, way of performing
 *     an allocation (in the case of testharness, the idea was to make it
 *     easy to experiment with new versions of the allocation algorithm).
 *     Since such programs need to manipulate the configuration structure
 *     in the same way as does 'configure', they need to be able to call
 *     the same subroutines. In general, all routines in configure.c that
 *     were independent of TCL have been moved into this file.
 *
 *  Compiling:
 *     Having said that these routines do not use Tcl, there is one bit of code -
 *     DoMakeAllocatedIndex() - that contains calls to the Tcl hash table
 *     routines. Rather than try to abstract these (although that's the best
 *     solution), a set of emulation routines have been provided in the
 *     HashEmulator code. Defining the pre-processor variable NO_TCL causes
 *     this code to be built against those instead of the real Tcl libraries.
 *
 *  Version date: 13th August 2013.
 *
 *  Support: K. Shortridge, Tony Farrell, AAO.
 *-
 *  "@(#) $Id: ACMM:aaoConfigure/src/configure_subs.c,v 12.7 04-Nov-2013 11:28:49+11 tjf $"
 *
 *  History:
 *      4th Mar 2005.  Original version. KS/AAO.
 *      4th Mar 2006.  Added ConfDoAddCBSPair to add CBS pairs to a field.
 *                     It may have memory issues.
 *                     BM/MacquarieUni_AAO
 *     24th Jul 2006.  DoMakeAllocated() now adds the wlenIndex item and
 *                     ConfGetIds() will return it.
 *     23rd Aug 2006.  ConfDoAddSkyPosition() and ConfDoAddCBSPair() must handle
 *                      wlenIndex.   
 *      1st Nov 2006.  Fixed an undefined variable problem in 
 *                     ConfDoXyConversion() that only showed up when compiled
 *                     with optimisation. KS.
 *      5th Aug 2009.  Added use of separate pmra and pmdec structure items
 *                     rather than pseudo-comments to handle proper motion
 *                     values. Fixed small bug in ConfDoXYConversion() that was
 *                     causing wavelength index values to be ignored. KS.
 *     14th Aug 2009.  ConfGetProperMotions() now has a ClearFlag argument and
 *                     clears the proper motion pseudo-comments from the
 *                     comment string if this is set. ConfListLine() now
 *                     outputs explicit proper motion columns. The routine
 *                     ConfListPAFTarget() now expects Sds id giving the proper
 *                     motion arrays in the various structures. KS.
 *     21st Aug 2012.  On OS X, had compiler warnings about the type of the
 *                     ClientData arguments in Tcl hash table calls. Added
 *                     'long' casts, and this fixed the problem, at least for
 *                     this compiler. KS.
 *     13th Aug 2013.  Added ConfFibreAffectData(). KS.
 *     16th Sep 2013,  Moved ConfFibreAffectData() to config.c, as the routine
 *                     ConfigFibreAffectData(). This conforms to the proper
 *                     rules for what routines allocation code can call. KS.
 *     22nd Oct 2013.  In ConfMakeAllocatedIndex(), added additional diagnostics
 *                     for non-matching Ra,Dec values. KS.
 *    
 */
 
#include "configure_subs.h"

/*  DEBUG is used to make diagnostic code stand out */

#ifdef DEBUG
#undef DEBUG
#endif
#define DEBUG printf

/*  Standard include files  */

#if defined WIN32
#include <windows.h>
#endif

#define __EXTENSIONS__  /* For solaris include files */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <ctype.h>
#include <limits.h>
#ifndef macintosh
#include <unistd.h>
#include <sys/param.h>
#endif

/*  Not all systems seem to declare readlink() properly, so we do so here.
 */

#ifndef __macosx__
int readlink (const char* path, char *buf, size_t bufsiz);
#endif
/*  Program-specific include files  */

#include "drama.h" 
#include "status.h"
#include "sds.h"
#include "Ers.h"
#include "mess.h"
#include "arg.h"
#include "arg_err.h" 
#include "tdFconfig.h"
#include "tdFconvert.h"
#include "convertPAF.h"
#include "slalib.h"
#include "slamac.h"
#include "conf_err.h"

/*  Defining the pre-processor variable NO_TCL causes this code to be built
 *  against the HashEmulator emulation of the TCL hash table routines
 *  instead of the real Tcl libraries.
 */
 
#ifdef NO_TCL
#include "HashEmulator.h"
#else
#include "tcl.h"
#endif

/*
 *  Not all operating systems define MAXPATHLEN so define it here if needed.
 */
 
#ifndef MAXPATHLEN
#define MAXPATHLEN 1024
#endif

/*
 *  Note that of two below, CMTLEN is always used for buffers which may
 *  take either a comment or name item, hence if their sizes changes.
 *  CMTLEN must remain the largest.
 */
 
#define NAMELEN 80              /* Length of "name" items in structures */
#define CMTLEN  80              /* Length of "comment" items in structures */

/*  PAF flags are used to control the output for different objects when a
 *  PAF file is being written. The only one needed is PAF_NO_PRIO which
 *  is used to suppress the priority information for VLT guide stars.
 */
 
#define PAF_NO_PRIO 1

/*  ------------------------------------------------------------------------- */

/*           G l o b a l  V a r i a b l e s  &  C o n s t a n t s
 */
 
/*
 *  When an instrument is selected, details are copied here. Note that most
 *  calls to Fpil routines take *Instrument as the first argument, so it is
 *  important that this be defined before any such calls are made. However,
 *  the way the program runs guarantees that the first thing that happens is
 *  that the instrument is selected. These external variables need to be
 *  defined in another program module. If linked as part of the main
 *  'configure' program, these are defined in configure.c - other programs
 *  such as testharness need to define them for themselves (usually by
 *  copying the definition from configure.c).
 */

extern FpilType *Instrument;                 /* Instrument details pointer */
extern FpilConstantsType InstConstDetails;   /* Used to access constants   */
extern FpilModelType *TelModel;              /* For telescope model        */
extern SdsIdType InstFieldDetails;           /* Sds id of field details    */
extern unsigned CurrentField;                /* Current field number       */
extern double *TelModelParams;               /* Telescope model parameters */
extern double *TelModelParamsUsed;           /* Telescope model parameters
                                              * used when the last coordinate
                                              * conversion was performed.  */
extern unsigned int TelModelNumParams;       /* Number of tel model params */
extern unsigned int InstNumFibreTypes;       /* Number of fibre types      */
extern int *UseFibreType;                    /* Points to array of values  
                                                indicating if a fibre type
                                                should be configured.      */
extern time_t FileTime;                      /* Time used in file name     */

extern double gtemp;                    /* Temperature in degrees kelvin */
extern double gpress;                   /* Pressure in mm Hg */
extern double ghumid;                   /* Humidity - as a fraction, ie 0 - 1 */
extern double gwave;                    /* Wavelength in microns for main 
                                         * target objects (non-guide objects) */
extern double gwave2;                   /* Wavelength in microns for main 
                                         * target objects (non-guide objects) 
                                         * for the 2nd 2df spectrograph */
/*  ------------------------------------------------------------------------- */

/*                          C h a n g e  M j d
 *
 *  Invoked to set a new configuration MJD.  This involves changing the
 *  apparent position of the field center, so is convenient to have
 *  packaged as a separate routine.
 */
 
void ChangeMjd (SdsIdType fieldId, double mjd, StatusType *status)
{
    double cenRa, cenDec;
    double appRa, appDec;

    if (*status != STATUS__OK) return;
    
    /*
     * Get the field centre mean position
     */
     
    ArgGetd(fieldId, "cenRa",  &cenRa,  status);
    ArgGetd(fieldId, "cenDec", &cenDec, status);

    /*
     * Convert the field centre mean position to apparent.
     */
     
    slaMap(cenRa, cenDec, 0 ,0,0,0,2000,  mjd, &appRa, &appDec);

    /*
     * Put all the new values in the SDS structure.
     */
     
    ArgPutd(fieldId, "configMjd", mjd,    status);
    ArgPutd(fieldId, "appRa",     appRa,  status);
    ArgPutd(fieldId, "appDec",    appDec, status);

}

/*  ------------------------------------------------------------------------- */

/*                      F r e e  A n d  F i n d
 *
 * This function does an Sds find, but before doing so, free's any old
 * SDS id being pointed to by the variable. 
 */
 
void FreeAndFind(const SdsIdType source, const char *const name,
    SdsIdType * const id, StatusType * const status)
{
    if (*status != STATUS__OK)
        return;

    if (*id)
        SdsFreeId(*id, status);
    if (*status != STATUS__OK) {
        ErsRep(0, status, "Error freeing id (%ld) for item \"%s\"",
            (long int) *id, name);
    } else {
        *id = 0;
        SdsFind(source, name, id, status);
        if (*status != STATUS__OK)
            ErsRep(0, status, "Error finding SDS item \"%s\"", name);
    }
}

/*  ------------------------------------------------------------------------- */

/*
 *                C o n f  G e t  P r o p e r  M o t i o n s
 *
 *  This routine looks through the comment string for a target object,
 *  looking for any proper-motion information encoded as a pseudo-comment.
 *  It is looking specifically for fields of the form PM-RA=value
 *  and PM-DEC=value. See ConfParseComment() for more detail on which
 *  formats are actually accepted. The proper motions are specified in
 *  the pseudo-comments in units of arcsec/year, but are returned by this
 *  routine in units of radians/year which is as required by the SLALIB 
 *  routine sla_mapqk(). If the ClearFlag argument is set true, the 
 *  proper motion pseudo-comments are removed from the comment string.
 */
 
int ConfGetProperMotions (            /* True if proper motion data found */
    char* Comment,                    /* Address at which to start parsing */
    short ClearFlag,                  /* True if pm data should be cleared */
    double* PmRa,                     /* Proper motion in RA */
    double* PmDec)                    /* Proper motion in Dec */
{
    int Found = FALSE;
    char Keyword[80];
    char Value[80];
    char *NextChar;
    int Key;
    
    *PmRa = 0.0;
    *PmDec = 0.0;
    
    /*  For efficiency, we do a quick scan for the string "PM-", which
     *  both keywords have in common. If this isn't there, there's no
     *  point doing the more exhaustive scan of the string.
     */
   
    if (strstr(Comment,"PM-")) {

       /*  This looks as if it might have proper motion pseudo-comments.
        *  We scan it in detail using ConfParseComment(). If this finds one
        *  of the proper motion keywords, we get the value associated
        *  with it. Note that ConfParseComment() returns the PAF keywords
        *  associated with the specification, which for PM-RA and PM-DEC
        *  are PM_RA and PM_DEC - underscores, not dashes.
        */
        
       NextChar = Comment;
       for (;;) {
           char* CommentPtr = NextChar;
           Key = ConfParseComment (CommentPtr, Keyword, Value, &NextChar);
           if (!Key) break;
           if (!strncmp(Keyword,"PM_RA",5)) {
               *PmRa = atof(Value) * PI / (60.0 * 60.0 * 180.0);
               Found = TRUE;
           } else if (!strncmp(Keyword,"PM_DEC",6)) {
               *PmDec = atof(Value) * PI / (60.0 * 60.0 * 180.0);
               Found = TRUE;
           }
           if (Found && ClearFlag) {
              char* Ptr = CommentPtr;
              while ((*Ptr++ = *NextChar++));
              NextChar = CommentPtr;
           }
       }
    }
    
    return (Found);
}

/*  ------------------------------------------------------------------------- */

/*
 *                    C o n f  P a r s e  C o m m e n t
 *
 *  This routine parses a comment field for a target object, looking for any
 *  of the auxiliary keywords that can be embedded in a comment string,
 *  particularly for FLAMES. It is passed the address of a starting character
 *  in the comment string, and - if it finds one of the auxiliary keywords -
 *  returns the auxiliary keyword, a value string for it, suitable for output
 *  to a PAF file, and the address of the next character in the comment string.
 *  The calling routine should continue to call this routine, each time passing
 *  it the address it returns, until it returns indicating that no keyword
 *  was found. The keyword is returned padded on the right to 8 characters.
 */
 
int ConfParseComment (                /* True if there was a keyword found */
    char* Comment,                    /* Address at which to start parsing */
    char* Keyword,                    /* Returned with Keyword string */
    char* Value,                      /* Returned with value string */
    char** NextChar)                  /* Address at which to start next */
{
    char* CharPtr;
    int I;
    int KeyIndex;
    int Length;
    short PossibleKey;
    short Leading;
    
    /*  The parsing provided by this routine is a little crude. It will
     *  correctly parse any correctly formatted strings, but may not
     *  pick up some incorrectly formatted ones. It's possible the 
     *  interface should actually include an error flag that can be set.
     */
    
    static char* Keywords[] = {
       "BAND", "B-V", "V-R", "PM-RA", "PM-DEC", "SNR", "CATEGORY", "DIAMETER" };
    static char* PAFKeywords[] = {
       "BAND", "B_V", "V_R", "PM_RA", "PM_DEC", "SNR", "CATEGORY", "DIAMETER" };
    static short KeywordIsChar[] = {
        TRUE,   FALSE, FALSE, FALSE,   FALSE,    FALSE, TRUE,       FALSE };
       
    static int NumKeywords = sizeof(Keywords)/sizeof(char*);
    
    /*  We allow both formats (Name)=value and (Name=value). In both cases,
     *  these begin with a '('. We skip any leading blanks, commas, and
     *  semi-colons.
     */
    
    CharPtr = Comment;
    PossibleKey = FALSE;
    KeyIndex = 0;
    for (;;) {
       char TheChar = *CharPtr;
       if (TheChar == '(') {
           CharPtr++;
           PossibleKey = TRUE;
           break;
       } else if ((TheChar == ' ') || (TheChar == ',') || (TheChar == ';')) {
           CharPtr++;
       } else {
           break;
       }
    }
    
    /*  So, if we have a startling '(', we can look through all the possible
     *  keywords. If we find a match, we check that the next characters are
     *  '=' or ')='. If they are, we have a match.
     */
    
    if (PossibleKey) {
        PossibleKey = FALSE;
        for (KeyIndex = 0; KeyIndex < NumKeywords; KeyIndex++) {
            Length = strlen(Keywords[KeyIndex]);
            if (!strncmp(CharPtr,Keywords[KeyIndex],Length)) {
                CharPtr += Length;
                if (*CharPtr == ')') CharPtr++;
                if (*CharPtr == '=') {
                    PossibleKey = TRUE;
                    CharPtr++;
                    break;
                }   
            }
        }
    }
        
    /*  If we do have a match, KeyIndex will have the index of the keyword
     *  that matches. We now need to locate the value. We ignore leading
     *  blanks in the value string.
     */
    
    if (PossibleKey) {
        strcpy (Keyword,PAFKeywords[KeyIndex]);
        Length = strlen(Keyword);
        for (I = 0; I < (8 - Length); I++) strcat(Keyword," ");
        if (KeywordIsChar[KeyIndex]) *Value++ = '"';
        Leading = TRUE;
        for (;;) {
            char TheChar = *CharPtr++;
            short Ignore = FALSE;            
            if (Leading && (TheChar == ' ')) Ignore = TRUE;
            if (TheChar != ' ') Leading = FALSE;
            if (!Ignore) {
                if ((TheChar == ' ') || (TheChar == ')') || (TheChar == '\0') ||
                                         (TheChar == ';') || (TheChar == ',')) {
                    if (!(Leading && (TheChar == ' '))) {
                        if (KeywordIsChar[KeyIndex]) *Value++ = '"';
                        *Value = '\0';
                        if (TheChar == '\0') CharPtr--;
                        *NextChar = CharPtr;
                        break;
                    }
                }
                *Value++ = TheChar;
            }
        }
        
    }
    if (!PossibleKey) *NextChar = Comment;
    
    return (PossibleKey);    
}        


/*  ------------------------------------------------------------------------- */

/*                        C o n f  G e t  I d s
 *
 *   Many routines in this program are passed the Sds Id of a substructure
 *   of the configuration data and then need to access all the items in
 *   that structure. This routine packages up access to the Ra, Dec, Name,
 *   Type, Spectrograph, Priority, Pid, Ra and Dec proper motions, Wavelength
 *   index and Comment array items. Of these, not all sub-structures have a 
 *   Spectrograph, wavelength index or proper motion items, so 
 *   the calling routine should be prepared to have an Sds Id of zero 
 *   returned for these, which indicates that the item was not found in the 
 *   sub-structure.
 */
 
void ConfGetIds(SdsIdType topid, SdsIdType * ra_id, SdsIdType * dec_id,
    SdsIdType * name_id, SdsIdType * type_id,
    SdsIdType * spect_id, SdsIdType * prio_id,
    SdsIdType * mag_id, SdsIdType * pid_id, SdsIdType * comment_id,
    SdsIdType * wlenIndex_id, SdsIdType * pmra_id, SdsIdType * pmdec_id,
    StatusType * status)
 {
    StatusType localStatus;

    SdsFind(topid, "ra", ra_id, status);
    SdsFind(topid, "dec", dec_id, status);
    SdsFind(topid, "name", name_id, status);
    SdsFind(topid, "type", type_id, status);
    SdsFind(topid, "priority", prio_id, status);
    SdsFind(topid, "magnitude", mag_id, status);
    localStatus = STATUS__OK;
    SdsFind(topid, "spectrograph", spect_id, &localStatus);
    if (localStatus != STATUS__OK) *spect_id = 0;
    SdsFind(topid, "pId", pid_id, status);
    SdsFind(topid, "comment", comment_id, status);

    localStatus = STATUS__OK;
    SdsFind(topid, "wlenIndex", wlenIndex_id, &localStatus);
    if (localStatus != STATUS__OK) *wlenIndex_id = 0;
    localStatus = STATUS__OK;
    SdsFind(topid, "pmra", pmra_id, &localStatus);
    if (localStatus != STATUS__OK) *pmra_id = 0;
    localStatus = STATUS__OK;
    SdsFind(topid, "pmdec", pmdec_id, &localStatus);
    if (localStatus != STATUS__OK) *pmdec_id = 0;

}


/*  ------------------------------------------------------------------------- */

/*                   C o n f  L i s t  P A F  T a r g e t
 *
 *  This routine outputs the details of one target object in the current
 *  sub-structure.  This just packages up a piece of code that would
 *  otherwise reccur in the code for WritePAF().
 */

void ConfListPAFTarget(FILE * listing, int i,
    char* prefix,
    SdsIdType ra_id,
    SdsIdType dec_id,
    SdsIdType name_id, SdsIdType type_id, SdsIdType spect_id,
    SdsIdType prio_id, SdsIdType mag_id, SdsIdType pid_id,
    SdsIdType pmra_id, SdsIdType pmdec_id, SdsIdType comment_id, 
    int fibre,
    int PAF_Flags,
    StatusType * status)
 {

    char name[NAMELEN], comment[CMTLEN];
    double mag;
    char otype;
    char ospect;
    short prio, pid;
    unsigned long actlen;
    double ra, dec;
    double pmra, pmdec;
    int rhmsf[4];
    int ddmsf[4];
    char sign[1];
    char typeString[32];
    char typeText[80];
    char keyword[80];
    char value[80];
    char *nextChar;
    int key;
    
    if (*status != STATUS__OK) return;
    
    SdsGet(name_id, NAMELEN, NAMELEN * i, name, &actlen, status);
    SdsGet(ra_id, sizeof(double), i, &ra, &actlen, status);
    SdsGet(dec_id, sizeof(double), i, &dec, &actlen, status);
    slaDr2tf(3, ra, sign, rhmsf);
    slaDr2af(2, dec, sign, ddmsf);
    SdsGet(type_id, sizeof(char), i, &otype, &actlen, status);
    if (spect_id) {
        SdsGet(type_id, sizeof(char), i, &ospect, &actlen, status);
    } else {
        ospect = 0;
    }
    FpilDecodeTargetType(*Instrument, otype, ospect, TRUE, -1, typeString,
                                                               typeText);
    SdsGet(prio_id, sizeof(short), i, &prio, &actlen, status);
    SdsGet(mag_id, sizeof(double), i, &mag, &actlen, status);
    SdsGet(pid_id, sizeof(short), i, &pid, &actlen, status);
    SdsGet(comment_id, CMTLEN, CMTLEN * i, comment, &actlen, status);
    if (*status != STATUS__OK) return;

    fprintf (listing,"%sNAME            \"%s\";\n",prefix,name);
    fprintf (listing,"%sRA              %2.2d%2.2d%2.2d.%2.2d;\n",
                    prefix,rhmsf[0], rhmsf[1], rhmsf[2], rhmsf[3] / 10);
    if (sign[0] == '-') {
       fprintf (listing,"%sDEC             -%2.2d%2.2d%2.2d.%1.1d;\n",
                    prefix,ddmsf[0], ddmsf[1], ddmsf[2], ddmsf[3] / 10);
    } else {
       fprintf (listing,"%sDEC             %2.2d%2.2d%2.2d.%1.1d;\n",
                    prefix,ddmsf[0], ddmsf[1], ddmsf[2], ddmsf[3] / 10);
    }  
    fprintf (listing,"%sTYPE            \"%s\";\n",prefix,typeString);
    if ((PAF_Flags & PAF_NO_PRIO) == 0) {
        fprintf (listing,"%sPRIOR           %d;\n",prefix,prio);
    }
    fprintf (listing,"%sMAG             %f;\n",prefix,mag);
    fprintf (listing,"%sPRGID           %d;\n",prefix,pid);
    if (fibre >= 0) {
        fprintf (listing,"%sFIBRE           %d;\n",prefix,fibre);
    }
    
    /*  It used to be that the proper motion values were held in pseudo-comments
     *  at the start of the comment string, and so were only written out if
     *  explicitly specified. We get more or less the same effect now by only
     *  writing them out if the values are known and are non-zero.
     */
     
    pmra = 0.0;
    pmdec = 0.0;
    if (pmra_id) SdsGet(pmra_id, sizeof(double), i, &pmra, &actlen, status);
    if (pmdec_id) SdsGet(pmdec_id, sizeof(double), i, &pmdec, &actlen, status);
    if (pmra != 0.0 && pmdec != 0.0) {
       double radToArcsec = (60.0 * 60.0 * 180.0) / PI;
       pmra *= radToArcsec;
       pmdec *= radToArcsec;
       fprintf (listing,"%sPM_RA           %f;\n",prefix,pmra);
       fprintf (listing,"%sPM_DEC          %f;\n",prefix,pmdec);
    }
    
    /*  Now we look in the comment string for any of the auxiliary keywords.
     *  We cannot have a null value, so if we're passed one, ignore the 
     *  keyword. (If the comment string includes PM-RA or PM-DEC pseudo-
     *  comments, ConfParseComment() will still find them, and we might
     *  end up with duplicate values in the file. However, the input routines
     *  are supposed to remove such strings. We check anyway.)
     */
    
    nextChar = comment;
    do {
        char* commentPtr = nextChar;
        key = ConfParseComment (commentPtr, keyword, value, &nextChar);
        if (key) {
            if (!strcmp(keyword,"PM_RA") && (pmra != 0.0)) key = FALSE;
            if (!strcmp(keyword,"PM_DEC") && (pmdec != 0.0)) key = FALSE;
            if (key) {
                if (strlen(value) > 0) {
                    fprintf (listing,"%s%s        %s;\n",prefix,keyword,value);
                }
            }
        }
    } while (key);
    fprintf (listing,"%sCOMMENT         \"%s\";\n",prefix,nextChar);
   
    
}

/*  ------------------------------------------------------------------------- */

/*                     C o n f  L i s t  L i n e
 *
 *  This routine outputs the details of one target object in the current
 *  sub-structure to the listing file.  This is a utility routine for use
 *  by ConfMakeListing(). Note that the format changed in Aug 2009 to 
 *  include the explicit proper motion columns.
 */

void ConfListLine(
    FILE * listing, 
    int i, 
    SdsIdType ra_id,
    SdsIdType dec_id,
    SdsIdType name_id, 
    SdsIdType type_id, 
    SdsIdType spect_id,
    SdsIdType prio_id, 
    SdsIdType mag_id, 
    SdsIdType pid_id,
    SdsIdType comment_id, 
    SdsIdType wlenIndex_id,
    SdsIdType pmra_id,
    SdsIdType pmdec_id,
    StatusType * status)
 {

    char name[NAMELEN], comment[CMTLEN];
    double mag;
    char otype;
    char ospect;
    short prio, pid;
    unsigned long actlen;
    double ra, dec;
    int rhmsf[4];
    int ddmsf[4];
    char sign[1];
    char typeString[32];
    char typeText[80];
    short wlenIndex = -1;
    double pmdec = 0.0;
    double pmra = 0.0;

    SdsGet(name_id, NAMELEN, NAMELEN * i, name, &actlen, status);
    SdsGet(ra_id, sizeof(double), i, &ra, &actlen, status);
    SdsGet(dec_id, sizeof(double), i, &dec, &actlen, status);
    slaDr2tf(3, ra, sign, rhmsf);
    slaDr2af(2, dec, sign, ddmsf);
    SdsGet(type_id, sizeof(char), i, &otype, &actlen, status);
    if (spect_id) {
        SdsGet(spect_id, sizeof(char), i, &ospect, &actlen, status);
    } else {
        ospect = 0;
    }
    if (wlenIndex_id) {
        SdsGet(wlenIndex_id, sizeof(short), i, &wlenIndex, &actlen, status);
    } else {
        wlenIndex = -1;
    }
    
    FpilDecodeTargetType(*Instrument, otype, ospect, TRUE, wlenIndex, 
                         typeString, typeText);
    SdsGet(prio_id, sizeof(short), i, &prio, &actlen, status);
    SdsGet(mag_id, sizeof(double), i, &mag, &actlen, status);
    SdsGet(pid_id, sizeof(short), i, &pid, &actlen, status);
    SdsGet(comment_id, CMTLEN, CMTLEN * i, comment, &actlen, status);
    
    if (pmra_id) SdsGet(pmra_id, sizeof(double), i, &pmra, &actlen, status);
    if (pmdec_id) SdsGet(pmdec_id, sizeof(double), i, &pmdec, &actlen, status);
    
    if (*status != STATUS__OK)
        return;

    /* Proper motions need converting from radians/year to arcsec/year */
    
    pmra *= (60.0 * 60.0 * 180.0) / PI;
    pmdec *= (60.0 * 60.0 * 180.0) / PI;
    
    fprintf(listing,"%12s %2.2d %2.2d %2.2d.%3.3d  %c%2.2d %2.2d %2.2d.%2.2d ",
        name, rhmsf[0], rhmsf[1], rhmsf[2], rhmsf[3], sign[0],
        ddmsf[0], ddmsf[1], ddmsf[2], ddmsf[3]);
    fprintf(listing,"%s %d %6.2f %d %8.4f %8.4f %s\n",
        typeString, prio, mag, pid, pmra, pmdec, comment);
}

/*  ------------------------------------------------------------------------- */

/*                     C o n f  L i s t  D S S  L i n e
 *
 *  This routine outputs the details of one target object in the current
 *  sub-structure to the listing file as required by the DSS format.  This
 *  is a utility routine for use by ConfMakeDSSListing().
 */

void ConfListDSSLine(int skyonly,
    FILE * listing, int i, SdsIdType ra_id,
    SdsIdType dec_id,
    SdsIdType name_id, SdsIdType type_id, SdsIdType prio_id,
    SdsIdType mag_id, SdsIdType pid_id, SdsIdType comment_id,
    StatusType * status)
 {

    char name[NAMELEN], comment[CMTLEN];
    double mag;
    char otype;
    short prio, pid;
    unsigned long actlen;
    double ra, dec;
    int rhmsf[4];
    int ddmsf[4];
    char sign[1];


    SdsGet(type_id, sizeof(char), i, &otype, &actlen, status);
    if ((skyonly) && (!FpilIsTargetSky(*Instrument, otype)))
        return;

    SdsGet(name_id, NAMELEN, NAMELEN * i, name, &actlen, status);
    SdsGet(ra_id, sizeof(double), i, &ra, &actlen, status);
    SdsGet(dec_id, sizeof(double), i, &dec, &actlen, status);
    slaDr2tf(3, ra, sign, rhmsf);
    slaDr2af(2, dec, sign, ddmsf);
    SdsGet(prio_id, sizeof(short), i, &prio, &actlen, status);
    SdsGet(mag_id, sizeof(double), i, &mag, &actlen, status);
    SdsGet(pid_id, sizeof(short), i, &pid, &actlen, status);
    SdsGet(comment_id, CMTLEN, CMTLEN * i, comment, &actlen, status);
    if (*status != STATUS__OK)
        return;

    if (i >= 99) {
        fprintf(listing, "S%3d ", i + 1);
    } else {
        if (i >= 9) {
            fprintf(listing, "S0%2d ", i + 1);
        } else {
            fprintf(listing, "S00%1d ", i + 1);
        }
    }
    fprintf(listing,
        "%2.2d %2.2d %2.2d.%3.3d  %c%2.2d %2.2d %2.2d.%2.2d 2. 2.\n",
        rhmsf[0], rhmsf[1], rhmsf[2], rhmsf[3], sign[0],
        ddmsf[0], ddmsf[1], ddmsf[2], ddmsf[3]);
}

/*  ------------------------------------------------------------------------- */

/*                      C o n f  S d s  C h e c k
 *
 *  This is a small utility that packages up a common sequence of calls
 *  used to report an error accessing an item in an Sds structure. It
 *  converts the Sds status code passed into a text message, the outputs
 *  that message and the text passed in the call.
 */
 
void ConfSdsCheck(char *text, StatusType * status)
 {
    char buff[100];

    if (*status != STATUS__OK) {
        MessGetMsg(*status, 0, 100, buff);
        ErsRep(0, status, "%s", buff);
        ErsRep(0, status, "%s", text);
    }
}

/*  ------------------------------------------------------------------------- */

/*                   C o n f  D o  X Y  C o n v e r s i o n
*
 *  This is a utility routine used only by ConfConvertXy().  It can be
 *  called to work on the 'unallocated guide' structure (in which case each
 *  entry will be a target object classified as a guide), or on the 
 *  'unallocated object' structure (in which case each entry will be a target
 *  object classified as non-guide), or on the 'objects' structure. In the 
 *  last case, the fields in the structure relating to position realy refer 
 *  both to the targets and the fibres allocated to them, and in particular 
 *  note that the 'type' field can be either the unallocated type code or the
 *  type of the object to which the fibre has been allocated.
 *  gmap contains the mean to apparent parameters, and is now a parameter
 *  rather than a global variable (which it was in earlier versions of the
 *  code).
 *
 *  Modified:
 *    17-Oct-2002. FpilModelCvtInit() now takes two wavelength parameters. KS.
 *     4-Nov-2002. Now allows for proper motions included in pseudo-comments.
 *                 It would be better to have these in the SDS structure, but
 *                 that's too dramatic a change to make at this stage. KS.  
 *     1-Nov-2006. In some circumstances wid was left undefined. Fixed. KS.
 *     5-Aug-2009. Proper motions ar enow expected to be held in the optional
 *                 items 'pmra' and 'pmdec'. Also fixed a bug in the wlenIndex
 *                 code that had the effect of ignoring the index. KS.  
 */

void ConfDoXyConversion(
    const SdsIdType id, 
    const double cra, 
    const double cdec,
    const double mjd, 
          double* const gmap, 
    const wavelensType * wlens,
    StatusType * const status)

{
    char name[16];
    SdsCodeType code;
    long ndims;
    unsigned long dims[7];
    SdsIdType rid, did, tid, xid, yid, wid;
    SdsIdType pmraid, pmdecid;
    double ra, dec;
    unsigned long actlen;
    double xp, yp;
    int xp2, yp2;
    char type;
    double dhdx, dhdy, dddx, dddy;
    SdsIdType tempid;
    char obj[NAMELEN];
    StatusType stat;
    double Wavelength;
    double PrevWavelength;
    double PointingWavelength;
    unsigned int i;
    double ara, adec, xf, yf;
    double ara2, adec2, ara3, adec3;
    double pmra, pmdec;
    SdsIdType nid;
    char text[80];
    StatusType LocalStatus;

    if (*status != STATUS__OK)
        return;
    
    SdsFind(id, "ra", &rid, status);
    ConfSdsCheck("SDS Item ra not found", status);
    SdsFind(id, "dec", &did, status);
    ConfSdsCheck("SDS Item dec not found", status);
    SdsFind(id, "type", &tid, status);
    ConfSdsCheck("SDS Item type not found", status);
    SdsFind(id, "x", &xid, status);
    ConfSdsCheck("SDS Item x not found", status);
    SdsFind(id, "y", &yid, status);
    ConfSdsCheck("SDS Item y not found", status);
    wid = 0;
    if (*status == STATUS__OK)
    {
        SdsFind(id, "wlenIndex", &wid, status);
        if (*status == SDS__NOITEM) {
            wid = 0;  /* Should be unallocGuide array */
            *status = STATUS__OK;
        }
        ConfSdsCheck("SDS Item wlenIndex not found", status);
    }
    
    /*  Pick up the proper motions - which may not be present. These used
     *  to be picked up from the pseudo-comments in the comment entry, but
     *  now they are items in their own right, and even if they originally
     *  come from a comment field they are turned into the individual
     *  items when the input .fld file is processed.
     */
     
    LocalStatus = STATUS__OK;
    SdsFind (id,"pmra",&pmraid,&LocalStatus);
    if (LocalStatus == SDS__NOITEM) {
       pmraid = 0;
       LocalStatus = STATUS__OK;
    }
    ConfSdsCheck("SDS Item pmra not found", &LocalStatus);
    LocalStatus = STATUS__OK;
    SdsFind (id,"pmdec",&pmdecid,&LocalStatus);
    if (LocalStatus == SDS__NOITEM) {
       pmdecid = 0;
       LocalStatus = STATUS__OK;
    }
    ConfSdsCheck("SDS Item pmdec not found", &LocalStatus);
    
    /*  Get the main 'pointing' wavelength, which is the same for all
     *  fibre types.
     */
     
    PointingWavelength = FpilGetPointingWavelength (*Instrument);
    
    /*  Reset to an unlikely wavelength, to force initialisation of the
     *  conversion routines.
     */

    PrevWavelength = 0.0;

    /*  We get the number of items in the structure from the dimensions of
     *  one of the arrays ("ra" will do), and process each item in turn.
     */

    SdsInfo(rid, name, &code, &ndims, dims, status);
    for (i = 0; (i < dims[0]) && (*status == STATUS__OK); i++) {
        SdsGet(tid, sizeof(char), i, &type, &actlen, status);
        
        /*  If this is a guide star, we need to do the coordinate conversion
         *  for a different wavelength to that used for the ordinary
         *  target stars. So we need to see what wavelength to use, and
         *  if this has changed, we need to re-initialise the coordinate
         *  conversion.
         */
        
        Wavelength = FpilGetFibreWavelength(*Instrument,type);

        /*
         * It is also possible for the wavelength to be assoicated with
         * the object itself - here we look for if the wlenIndex is within the
         * range of wavelength indicies we support.  If it is, then we fetch
         * and use the wavelength.
         */
        if (wid) {
            short wlenIndex;
            SdsGet(wid, sizeof(short), i, &wlenIndex, &actlen, status);
            if ((wlenIndex >= 0)&&(wlenIndex < MAX_WLENS)) {
                tdFwlensGet(wlens, wlenIndex, &Wavelength, status);
            }
                
        }
        if (Wavelength != PrevWavelength) {
            FpilModelCvtInit(TelModel, mjd, 0.0, gtemp, gpress, ghumid, 
			     PointingWavelength, Wavelength, TelModelParams, status);
            PrevWavelength = Wavelength;
        }
        
        /*  See if this is an unallocated pivot (only when we process the
         *  "objects" structure will this be the case).
         */

        if (type != FpilUnallocatedType(*Instrument)) {
        
            /*  This is an allocated pivot, so we calculate the X,Y
             *  position from the Ra,Dec of the target to which it is
             *  allocated.
             */
            
            pmra = 0.0;
            pmdec = 0.0;
            SdsGet(rid, sizeof(double), i, &ra, &actlen, status);
            SdsGet(did, sizeof(double), i, &dec, &actlen, status);
            if (pmraid) SdsGet (pmraid, sizeof(double), 
                                          i, &pmra, &actlen, status); 
            if (pmdecid) SdsGet (pmdecid, sizeof(double), 
                                          i, &pmdec, &actlen, status); 
            
            /*  If proper motions are specified we need to convert to 
             *  apparent RA/Dec using slaMapqk(). If not, we can use the
             *  simpler slaMapqkz(). The proper motions should be in units
             *  of radians/year.
             */
             
            if (pmdec != 0.0 || pmra != 0.0) {
               slaMapqk(ra,dec,pmra,pmdec,0.0,0.0,gmap,&ara,&adec);
            } else {
               slaMapqkz(ra, dec, gmap, &ara, &adec);
            }
            FpilModelRd2Xy(TelModel, cra, cdec, ara, adec, mjd, &xf, &yf,
                status);
            if (*status != STATUS__OK) {
                goto Exit;
            }
            FpilModelXy2Pos(TelModel, xf, yf, &xp, &yp, status);
            xp2 = (int) xp;
            yp2 = (int) yp;
            SdsPut(xid, sizeof(INT32), i, &xp2, status);
            SdsPut(yid, sizeof(INT32), i, &yp2, status);

            /*  If this is a guide object (which can be the case when we
             *  process either the unallocated guide structure or the
             *  "objects" structure), work out the local derivatives for the
             *  ra and dec with respect to x and y.
             */

            if (FpilIsTargetGuide(*Instrument, type)) {
                FpilModelXy2Rd(TelModel, cra, cdec, xf, yf, mjd, &ara, &adec,
                    status);
                FpilModelXy2Rd(TelModel, cra, cdec, xf + 1000.0, yf, mjd, &ara2,
                    &adec2, status);
                FpilModelXy2Rd(TelModel, cra, cdec, xf, yf + 1000.0, mjd, &ara3,
                    &adec3, status);
                if (*status == STATUS__OK) {
                    dhdx = (ara - ara2) / 1000.0;
                    dddx = (adec2 - adec) / 1000.0;
                    dhdy = (ara - ara3) / 1000.0;
                    dddy = (adec3 - adec) / 1000.0;
                    SdsFind(id, "dhdx", &tempid, status);
                    ConfSdsCheck("SDS Item dhdx not found", status);
                    SdsPut(tempid, sizeof(double), i, &dhdx, status);
                    SdsFreeId(tempid, status);
                    SdsFind(id, "dddx", &tempid, status);
                    ConfSdsCheck("SDS Item dddx not found", status);
                    SdsPut(tempid, sizeof(double), i, &dddx, status);
                    SdsFreeId(tempid, status);
                    SdsFind(id, "dhdy", &tempid, status);
                    ConfSdsCheck("SDS Item dhdy not found", status);
                    SdsPut(tempid, sizeof(double), i, &dhdy, status);
                    SdsFreeId(tempid, status);
                    SdsFind(id, "dddy", &tempid, status);
                    ConfSdsCheck("SDS Item dddy not found", status);
                    SdsPut(tempid, sizeof(double), i, &dddy, status);
                    SdsFreeId(tempid, status);
                }
            }
            if (*status != STATUS__OK) {
                stat = STATUS__OK;
                SdsFind(id, "name", &nid, &stat);
                SdsGet(nid, NAMELEN, NAMELEN * i, obj, &actlen, &stat);
                SdsFreeId(nid, &stat);
                sprintf(text, "Error converting to xy for object %d - %s", i,
                    obj);
                ConfSdsCheck(text, status);
                goto Exit;
            }
        } else {
        
            /*  This is an unallocated pivot.  In this case, we do the 
             *  reverse, and calculate the Ra,Dec from the parked position
             *  in X,Y. (This might seem to be something we shouldn't care
             *  about, but for some instruments - 6dF - parked fibres are
             *  in the field, and if they record data an observer might want
             *  to know where it came from.)
             */
             
            SdsGet(xid, sizeof(INT32), i, &xp2, &actlen, status);
            SdsGet(yid, sizeof(INT32), i, &yp2, &actlen, status);
            xp = (double) xp2;
            yp = (double) yp2;
            FpilModelXy2Rd(TelModel, cra, cdec, xp, yp, mjd, &ra, &dec,status);
            SdsPut(rid, sizeof(double), i, &ra, status);
            SdsPut(did, sizeof(double), i, &dec, status);
            if (*status != STATUS__OK) {
                sprintf(text, "Error converting from xy for fibre %d", i);
                ConfSdsCheck(text, status);
                goto Exit;
            }
        }
            
    }
    
Exit:;

    /*  On exit, play safe and reset conversions to the main observing
     *  wavelength.
     */
     
    LocalStatus = 0;
    FpilModelCvtInit(TelModel, mjd, 0.0, gtemp, gpress, ghumid, gwave, gwave,
                                               TelModelParams, &LocalStatus);
   
    SdsFreeId(rid, status);
    SdsFreeId(did, status);
    SdsFreeId(tid, status);
    SdsFreeId(xid, status);
    SdsFreeId(yid, status);
}

/*  ------------------------------------------------------------------------- */

/*+
 *                     D o  M a k e  A l l o c a t e d
 *
 *   DoMakeAllocated() is the main routine used by ConfMakeAllocated() to
 *   create the 'objects' structure. It is packaged so that it can also be
 *   called directly from other C routines such as the PAF input routines.
 *   It is passed the Sds Id's for the configuration data structure and
 *   for the pivot data structure.
 *-
 */
 
void DoMakeAllocated (
   SdsIdType topid,
   SdsIdType pivid,
   StatusType* status)
{   
    SdsIdType objid;
    SdsIdType name_id, ra_id, dec_id, fibre_id, x_id, y_id, theta_id, type_id;
    SdsIdType prio_id, mag_id, spect_id, pid_id, dhdx_id, dhdy_id, dddx_id,
        dddy_id;
    SdsIdType comment_id;
    SdsIdType xpid, ypid, tpid;
    SdsIdType wlenIndex_id, pmra_id, pmdec_id;
    char name[NAMELEN];
    SdsCodeType code;
    long ndims;
    unsigned long dims[7], dims2[7];
    unsigned int i;

    unsigned long actlen;

    if (*status != STATUS__OK) return;

    SdsFind(topid, "objects", &objid, status);
    ConfSdsCheck("Error finding objects structure", status);

    if (*status == STATUS__OK) 
    {
        StatusType ignore = STATUS__OK;
        SdsFreeId(objid, &ignore);
        return;
    }

    ErsAnnul(status);

    SdsNew(topid, "objects", 0, NULL, SDS_STRUCT, 0, NULL, &objid, status);

    SdsFind(pivid, "xPark", &xpid, status);
    SdsInfo(xpid, name, &code, &ndims, dims, status);
    dims2[0] = NAMELEN;
    dims2[1] = dims[0];
    SdsNew(objid, "name", 0, NULL, SDS_CHAR, 2, dims2, &name_id, status);
    SdsNew(objid, "ra", 0, NULL, SDS_DOUBLE, 1, dims, &ra_id, status);
    SdsNew(objid, "dec", 0, NULL, SDS_DOUBLE, 1, dims, &dec_id, status);
    SdsNew(objid, "fibre", 0, NULL, SDS_SHORT, 1, dims, &fibre_id, status);
    SdsNew(objid, "x", 0, NULL, SDS_INT, 1, dims, &x_id, status);
    SdsNew(objid, "y", 0, NULL, SDS_INT, 1, dims, &y_id, status);
    SdsNew(objid, "theta", 0, NULL, SDS_DOUBLE, 1, dims, &theta_id, status);
    SdsNew(objid, "type", 0, NULL, SDS_CHAR, 1, dims, &type_id, status);
    SdsNew(objid, "priority", 0, NULL, SDS_SHORT, 1, dims, &prio_id, status);
    SdsNew(objid, "magnitude", 0, NULL, SDS_DOUBLE, 1, dims, &mag_id, status);
    SdsNew(objid, "spectrograph", 0, NULL, SDS_CHAR, 1, dims, &spect_id,
        status);
    SdsNew(objid, "pId", 0, NULL, SDS_SHORT, 1, dims, &pid_id, status);
    SdsNew(objid, "dhdx", 0, NULL, SDS_DOUBLE, 1, dims, &dhdx_id, status);
    SdsNew(objid, "dhdy", 0, NULL, SDS_DOUBLE, 1, dims, &dhdy_id, status);
    SdsNew(objid, "dddx", 0, NULL, SDS_DOUBLE, 1, dims, &dddx_id, status);
    SdsNew(objid, "dddy", 0, NULL, SDS_DOUBLE, 1, dims, &dddy_id, status);
    SdsNew(objid, "comment", 0, NULL, SDS_CHAR, 2, dims2, &comment_id, status);
    SdsNew(objid, "wlenIndex", 0, NULL, SDS_SHORT, 1, dims, &wlenIndex_id, 
           status);
    SdsNew(objid, "pmra", 0, NULL, SDS_DOUBLE, 1, dims, &pmra_id, status);
    SdsNew(objid, "pmdec", 0, NULL, SDS_DOUBLE, 1, dims, &pmdec_id, status);

    SdsFind(pivid, "yPark", &ypid, status);
    SdsFind(pivid, "tPark", &tpid, status);

    if (*status == STATUS__OK) {
        double d;
        int l;
        short s;
        char c;
        for (i = 0; i < dims[0]; i++) {
            strcpy(name, "Parked");
            SdsPut(name_id, NAMELEN, NAMELEN * i, name, status);
            d = 0.0;
            SdsPut(ra_id, sizeof(double), i, &d, status);
            SdsPut(dec_id, sizeof(double), i, &d, status);
            s = 0;
            SdsPut(fibre_id, sizeof(short), i, &s, status);
            SdsGet(xpid, sizeof(INT32), i, &l, &actlen, status);
            SdsPut(x_id, sizeof(INT32), i, &l, status);
            SdsGet(ypid, sizeof(INT32), i, &l, &actlen, status);
            SdsPut(y_id, sizeof(INT32), i, &l, status);
            SdsGet(tpid, sizeof(double), i, &d, &actlen, status);
            SdsPut(theta_id, sizeof(double), i, &d, status);
            c = FpilUnallocatedType(*Instrument);
            SdsPut(type_id, sizeof(char), i, &c, status);
            s = 1;
            SdsPut(prio_id, sizeof(short), i, &s, status);
            d = 0.0;
            SdsPut(mag_id, sizeof(double), i, &d, status);
            c = ' ';
            SdsPut(spect_id, sizeof(char), i, &c, status);
            s = 0;
            SdsPut(pid_id, sizeof(short), i, &s, status);
            SdsPut(dhdx_id, sizeof(double), i, &d, status);
            SdsPut(dhdy_id, sizeof(double), i, &d, status);
            SdsPut(dddx_id, sizeof(double), i, &d, status);
            SdsPut(dddy_id, sizeof(double), i, &d, status);
            SdsPut(comment_id, CMTLEN, CMTLEN * i, name, status);
            s = -1;
            SdsPut(wlenIndex_id, sizeof(short), i, &s, status);
            SdsPut(pmra_id, sizeof(double), i, &d, status);
            SdsPut(pmdec_id, sizeof(double), i, &d, status);
        }
    }

    SdsFreeId(xpid, status);
    SdsFreeId(ypid, status);
    SdsFreeId(tpid, status);
    SdsFreeId(name_id, status);
    SdsFreeId(ra_id, status);
    SdsFreeId(dec_id, status);
    SdsFreeId(fibre_id, status);
    SdsFreeId(x_id, status);
    SdsFreeId(y_id, status);
    SdsFreeId(theta_id, status);
    SdsFreeId(type_id, status);
    SdsFreeId(prio_id, status);
    SdsFreeId(mag_id, status);
    SdsFreeId(spect_id, status);
    SdsFreeId(pid_id, status);
    SdsFreeId(dhdx_id, status);
    SdsFreeId(dhdy_id, status);
    SdsFreeId(dddx_id, status);
    SdsFreeId(dddy_id, status);
    SdsFreeId(comment_id, status);
    SdsFreeId(wlenIndex_id, status);
    SdsFreeId(pmra_id, status);
    SdsFreeId(pmdec_id, status);
    SdsFreeId(objid, status);

}

/*  ------------------------------------------------------------------------- */

/*        P r e l i m i n a r y   D e f i n i t i o n s   f o r
 *
 *                         C o n f  M a k e  A l l o c a t e d  I n d e x
 *
 *  ConfMakeAllocatedIndex is invoked if the objects->index, 
 *  unallocGuide->allocated and unallocObject->allocated items have just 
 *  been created by  ConfAddIndex().  It is possible that some items in the
 *  "objects" array are in the unallocObj/unallocGui arrays, as the
 *  structure may be the result of a previous run of the program - but
 *  the "index" item may or may not be correct.
 *
 *  In addition, we should add any objects not in unallocObject/unallocGuide
 *  to the appropiate component to ensure the structure is consistent with
 *  the way the config library will use it.
 * 
 *  We use a Tcl Hash table to help us here.  We first get access to
 *  the appropiate components of the unallocGuide and unallocObject structures
 *  (name, ra, dec, allocated).  We use the name as the hash table key.  The
 *  hash table value is the index for unallocObject objects and
 *  the negative of the index+1 for unallocGuide objects.  This avoids us
 *  setting up structures to maintain this information and should make
 *  this quite efficent.  Each object must have a unique name.
 *
 *  Then, we walk through the objects structure.  For every allocated
 *  object, we look for in the hash table for an object of the same 
 *  name.  
 *
 *  If the object exists in the hash table, then we ensure it is marked
 *  as allocated.   If not, then our structure is corrupted.  IF the
 *  existing index for the object is -1, then we set that index.  If the
 *  index for an allocated object was not -1, then is must correspond to the
 *  index for correspnding the unallocObject/unallocGuide item, otherwise
 *  our structure is corrupted.
 *
 *  If the object does not exist in the hash table, then its index should
 *  be minus 1.  We  create a list of such objects.
 *
 *  After the above, we will have the number of indexes of each object
 *  in the allocated are which are not in the unallocate area.  We then
 *  expand those compoenets and copy the details in.  We set the index
 *  for each object and the allocated area appropiately.
 */
 
typedef struct {
    unsigned size;
    SdsIdType name;
    char *namePtr;
    SdsIdType ra;
    double *raPtr;
    SdsIdType dec;
    double *decPtr;
    SdsIdType type;
    char *typePtr;
    SdsIdType other;            /* index/allocated */
    void *otherPtr;
} structIds;

const structIds IdsInitVal = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };

/*
 * Macros used to converting the guide array lists index to a hash table
 * value and the reverse.  The idea is to represent the guide array index
 * as a negative number, to avoid confusion with the object array lists, which
 * use the positive value as the hash value.
 */
 
#define GuideIdToHash(_i_) ((((int)(_i_))+1)*-1)
#define HashToGuideId(_i_) ((unsigned)(((_i_)*-1)-1))

/*
 * We use this hash table value to indicate items added to the hash table
 * as they were found in the "objects" component.  
 */

#define EXTRA_VALUE INT_MIN

#define CONF_C_ALLOC_EXISTS 0   /* Object exists in unallocGuide or
                                   unallocObject array or is not
                                   allocated */
#define CONF_C_ALLOC_GUIDE 1    /* Allocated guide object which does
                                   not exist in unallocGuide array */
#define CONF_C_ALLOC_OBJECT 2   /* Allocated program object which 
                                   does not exist in unallocObject
                                   array */

/*  ------------------------------------------------------------------------- */

/*        U t i l i t y  R o u t i n e s   f o r
 *
 *                         C o n f  M a k e  A l l o c a t e d  I n d e x
 */
 
/*  ------------------------------------------------------------------------- */

/*                   C o n f  I n d e x  I d s  G e t
 *
 *  Access each of the items of interest in a object array structure.
 *  If "alloc" is ture, we are dealing with the array of allocated
 *  structures ("objects").  Otherwise, it is one of the unallocated
 *  item arrays.
 */
 
void ConfIndexIdsGet(SdsIdType topid,
    int alloc, structIds * ids, StatusType * status)
{
    unsigned long length;
    char name[SDS_C_NAMELEN];
    SdsCodeType code;
    long ndims;
    unsigned long dims[7];

/*
 *  Set initial values.
 */
    *ids = IdsInitVal;

    if (*status != STATUS__OK)
        return;

    SdsFind(topid, "ra", &ids->ra, status);
    SdsFind(topid, "dec", &ids->dec, status);
    SdsFind(topid, "name", &ids->name, status);
    SdsFind(topid, "type", &ids->type, status);
/*
 *  If this is from the allocated (objects) array, find the index.  If this
 *  is from the unallocGuide/unallocObject arrays, find the allocated item.
 */
    if (alloc)
        SdsFind(topid, "index", &ids->other, status);
    else
        SdsFind(topid, "allocated", &ids->other, status);

/*
 *  Get points to the data for each item.
 */
    SdsPointer(ids->name, (void **) &ids->namePtr, &length, status);
    SdsPointer(ids->ra, (void **) &ids->raPtr, &length, status);
    SdsPointer(ids->dec, (void **) &ids->decPtr, &length, status);
    SdsPointer(ids->type, (void **) &ids->typePtr, &length, status);
    SdsPointer(ids->other, (void **) &ids->otherPtr, &length, status);

/*
 *  Get the length of each array.
 */
    SdsInfo(ids->ra, name, &code, &ndims, dims, status);
    ids->size = dims[0];

}

/*  ------------------------------------------------------------------------- */

/*                   C o n f  I n d e x  I d s  F r e e
 *
 *   Free the ids created by ConfIndexIdsGet().
 */
 
void ConfIndexIdsFree(structIds * ids)
{
    StatusType ignore = STATUS__OK;
    if (ids->ra)
        SdsFreeId(ids->ra, &ignore);
    if (ids->dec)
        SdsFreeId(ids->dec, &ignore);
    if (ids->name)
        SdsFreeId(ids->name, &ignore);
    if (ids->type)
        SdsFreeId(ids->type, &ignore);
    if (ids->other)
        SdsFreeId(ids->other, &ignore);

    *ids = IdsInitVal;
}


/*  ------------------------------------------------------------------------- */

/*            C o n f  M a k e  A l l o c a t e d  I n d e x
 *
 *  This is the routine called from ConfAddIndex() (which implements 
 *  AddIndex{}) to generate the index and allocated cross-reference items.
 */
 
void ConfMakeAllocatedIndex(SdsIdType topid, StatusType * status)
{
    SdsIdType objectsId = 0;    /* Id to "objects" array    */
    SdsIdType unallocGuiId = 0; /* Id to "unallocGui" array */
    SdsIdType unallocObjId = 0; /* Id to "unallocObj" array */

    structIds objectsStruct;    /* Parts of objects array    */
    structIds unallocGuiStruct; /* Parts of unallocGui array */
    structIds unallocObjStruct; /* Parts of unallocObj array */

    Tcl_HashTable hashTable;    /* The hash table of unallocGui and
                                   unallocObj names  */

    register unsigned i;

    int *objectNotExist = 0;    /* Points to array of malloc data
                                   which are flags.  with one of the
                                   three values above
                                 */

    unsigned allocNEGuideCnt = 0;
                                /* Number of allocated guide not
                                   in unallocGuide array */
    unsigned allocNEObjCnt = 0; /* Number of allocated program objects
                                   not in unallocObject array */

    int duplicates = 0;         /* Do we have duplicate objects */
    

    StatusType ignore = STATUS__OK;
    if (*status != STATUS__OK)
        return;

/*
 *  Access the structures
 */
    SdsFind(topid, "objects", &objectsId, status);
    SdsFind(topid, "unallocGuide", &unallocGuiId, status);
    SdsFind(topid, "unallocObject", &unallocObjId, status);

/*
 *  And the components of the structures
 */
    ConfIndexIdsGet(objectsId, 1, &objectsStruct, status);
    ConfIndexIdsGet(unallocGuiId, 0, &unallocGuiStruct, status);
    ConfIndexIdsGet(unallocObjId, 0, &unallocObjStruct, status);

/*
 *  Initialise the TCL hash table.  The key is a string (the name).
 *  
 */
    Tcl_InitHashTable(&hashTable, TCL_STRING_KEYS);

    if (*status != STATUS__OK)
        goto Exit;

/*
 *  Add the unallocObject array to the hash table.  The hash value is just
 *  the index, the key is the name.
 */

    for (i = 0; i < unallocObjStruct.size; ++i) {
        Tcl_HashEntry *entryPtr;
        int new;
        char *name = &unallocObjStruct.namePtr[NAMELEN * i];
        entryPtr = Tcl_CreateHashEntry(&hashTable, name, &new);


        if (!new) {
            ErsRep(0, status,
                "Unallocated Program Object %s is not unique.", name);
            duplicates = 1;
        } else
            Tcl_SetHashValue(entryPtr, (long)i);

    }
/*
 *  likewise for the unallocGuide array.  Here the hash value is different.
 */
    for (i = 0; i < unallocGuiStruct.size; ++i) {
        Tcl_HashEntry *entryPtr;
        int new;
        char *name = &unallocGuiStruct.namePtr[NAMELEN * i];
        int value = GuideIdToHash(i);
        entryPtr = Tcl_CreateHashEntry(&hashTable, name, &new);


        if (!new) {
            ErsRep(0, status,
                "Unallocated Guide Object %s is not unique.", name);
            duplicates = 1;
        } else
            Tcl_SetHashValue(entryPtr, (long)value);
    }

    if (duplicates) {
        ErsRep(0, status, "Configure cannot handle duplicate object names.");
        ErsRep(0, status, "Please remove them and try again.");
        *status = CONF__DUPLICATES;
        goto Exit;
    }
    
/*
 *  The hash table is now set up.  
 * 
 *  Malloc some space to keep track of which allocated objects do not
 *  exist in unallocGuide/unallocObject
 */
 
    objectNotExist = malloc(sizeof(*objectNotExist) * objectsStruct.size);

/*
 *  Walk through the objects array
 */
 
    for (i = 0; i < objectsStruct.size; ++i) {
        Tcl_HashEntry *entryPtr;
/*
 *      By default, assume that if the object is allocated, it exists in
 *      the other parts of the structure.
 */
        objectNotExist[i] = CONF_C_ALLOC_EXISTS;
/*
 *      If this object is allocated, See if this object has a hash table entry
 */
        if (objectsStruct.typePtr[i] != FpilUnallocatedType(*Instrument)) {
            char *name = &objectsStruct.namePtr[NAMELEN * i];
            short *indexPtr = objectsStruct.otherPtr;

            entryPtr = Tcl_FindHashEntry(&hashTable, name);
            if (entryPtr) {
                structIds *ids;
                int j;
                char *type;
                char *allocated;
/*
 *              We have an entry for this name.  Get details of the
 *              structure we are dealing with and validate the type.
 */
                j = (long) Tcl_GetHashValue(entryPtr);
                if (j == EXTRA_VALUE) {
                    fprintf(stderr,
                        "Multiple definition of object of name %s ", name);
                    fprintf(stderr, "in \"objects\" structure\n");
                    fprintf(stderr,
                        "%s:%d Assertion failed\n", __FILE__, __LINE__);
                    fprintf(stderr, "Talk to tjf@aaoepp.aao.gov.au\n");
                    exit(-1);
                } else if (j < 0) {
                    type = "unallocGuide";
                    j = HashToGuideId(j);
                    ids = &unallocGuiStruct;

                    if (!FpilIsTargetGuide(*Instrument,
                            objectsStruct.typePtr[i])) {
                        fprintf(stderr,
                            "Object %d, name %s, name corresponds to guide ",
                            i, name);
                        fprintf(stderr,
                            "object %d, but the allocation does not\n", j);
                        fprintf(stderr,
                            "Configure's data structures are corrupt\n");
                        fprintf(stderr,
                            "%s:%d Assertion failed\n", __FILE__, __LINE__);
                        fprintf(stderr, "Talk to tjf@aaoepp.aao.gov.au\n");
                        exit(-1);
                    }
                } else {
                    type = "unallocObject";
                    ids = &unallocObjStruct;
                    if (FpilIsTargetGuide(*Instrument,
                            objectsStruct.typePtr[i])) {
                        fprintf(stderr,
                            "Object %d, name %s, name corresponds to program ",
                            i, name);
                        fprintf(stderr,
                            "object %d, but the allocation does not\n", j);
                        fprintf(stderr,
                            "Configure's data structures are corrupt\n");
                        fprintf(stderr,
                            "%s:%d Assertion failed\n", __FILE__, __LINE__);
                        fprintf(stderr, "Talk to tjf@aaoepp.aao.gov.au\n");
                        exit(-1);
                    }

                }
/*
 *              ids now points to the correct one of unallocObjStruct/
 *              unallocGuiStruct.  The type is correct
 *
 *              Validate the same ra and dec.  There are other things we
 *              could check but this would seem to be sufficent.
 */
                if (objectsStruct.raPtr[i] != ids->raPtr[j]) {
                    fprintf(stderr, "Object %d - %s, %s %d.  RA not the same\n",
                        i, name, type, j);
                    fprintf(stderr,"RA values: %.20f %.20f\n",
                            objectsStruct.raPtr[i],ids->raPtr[j]);
                    fprintf(stderr,
                        "Configure's data structures are corrupt\n");
                    fprintf(stderr,
                        "%s:%d Assertion failed\n", __FILE__, __LINE__);
                    fprintf(stderr, "Talk to tjf@aaoepp.aao.gov.au\n");
                    exit(-1);
                }

                if (objectsStruct.decPtr[i] != ids->decPtr[j]) {
                    fprintf(stderr,
                        "Object %d - %s, %s %d.  DEC not the same\n",
                        i, name, type, j);
                    fprintf(stderr,"Dec values: %.20f %.20f\n",
                            objectsStruct.decPtr[i],ids->decPtr[j]);
                    fprintf(stderr,
                        "Configure's data structures are corrupt\n");
                    fprintf(stderr,
                        "%s:%d Assertion failed\n", __FILE__, __LINE__);
                    fprintf(stderr, "Talk to tjf@aaoepp.aao.gov.au\n");
                    exit(-1);
                }

/*
 *              The unallocGuide/unallocObject array entry for this
 *              object should have the "allocated" flag set.
 */
                allocated = ids->otherPtr;
                if (!allocated[j]) {
                    fprintf(stderr,
                        "Object %d - %s, %s %d.  allocated flag not set\n",
                        i, name, type, j);
                    fprintf(stderr,
                        "Configure's data structures are corrupt\n");
                    fprintf(stderr,
                        "%s:%d Assertion failed\n", __FILE__, __LINE__);
                    fprintf(stderr, "Talk to tjf@aaoepp.aao.gov.au\n");
                    exit(-1);

                }
/*
 *              Now if the index entry is mimus 1, we should set it.  Otherwise
 *              it should be the same value as j.
 */
                if (indexPtr[i] == -1) {
                    indexPtr[i] = j;
                } else if (indexPtr[i] != j) {
                    fprintf(stderr,
                        "Object %d - %s, %s %d.  Index flag invalid - %d\n",
                        i, name, type, j, indexPtr[i]);
                    fprintf(stderr,
                        "Configure's data structures are corrupt\n");
                    fprintf(stderr,
                        "%s:%d Assertion failed\n", __FILE__, __LINE__);
                    fprintf(stderr, "Talk to tjf@aaoepp.aao.gov.au\n");
                    exit(-1);
                }

            } else {

                int new;
/*
 *              No hash table entry.  Validate index before noting this object
 *              so that later we can add it to the appropiate array.
 *              We also add it to the hash table to help us in picking
 *              up duplicates.
 *
 *              index should be minus -1.  Otherwise
 */
                if (indexPtr[i] != -1) {
                    fprintf(stderr,
                        "Allocated object %d (%s) is not in the hash table,",
                        i, name);
                    fprintf(stderr, " but has a index of %d\n", indexPtr[i]);
                    fprintf(stderr,
                        "Configure's data structures are corrupt\n");
                    fprintf(stderr, "%s:%d Assertion failed\n", __FILE__,
                        __LINE__);
                    fprintf(stderr, "Talk to tjf@aaoepp.aao.gov.au\n");
                    exit(-1);
                }
                if (FpilIsTargetGuide(*Instrument, objectsStruct.typePtr[i])) {
                    objectNotExist[i] = CONF_C_ALLOC_GUIDE;
                    ++allocNEGuideCnt;
                } else {
                    objectNotExist[i] = CONF_C_ALLOC_OBJECT;
                    ++allocNEObjCnt;
                }

                entryPtr = Tcl_CreateHashEntry(&hashTable, name, &new);


                if (!new) {
                    fprintf(stderr,
                        "CreateHashEntry indicated object %s is not unique,",
                        name);
                    fprintf(stderr,
                        " but it was not found so should be unique\n");
                    fprintf(stderr,
                        "%s:%d Assertion failed\n", __FILE__, __LINE__);
                    fprintf(stderr, "Talk to tjf@aaoepp.aao.gov.au\n");
                    exit(-1);
                }
                Tcl_SetHashValue(entryPtr, EXTRA_VALUE);

            }
        }
    }

/*
 *  Now I need to add each of the allocated objects not in the appropiate
 *  array to the end of the array.
 */
    if (allocNEObjCnt || allocNEGuideCnt) {
        if (allocNEObjCnt)
            ConfResizeAllItems(unallocObjId,
                unallocObjStruct.size + allocNEObjCnt, status);

        if (allocNEGuideCnt)
            ConfResizeAllItems(unallocGuiId,
                unallocGuiStruct.size + allocNEGuideCnt, status);

        for (i = 0; i < objectsStruct.size; ++i) {
            SdsIdType uid;
            int obj_no;
            short *indexPtr = objectsStruct.otherPtr;

            if (objectNotExist[i] == CONF_C_ALLOC_EXISTS)
                continue;

/*
 *          Copy in the details, code based on code in ConfDeallocateFibre
 */
            if (objectNotExist[i] == CONF_C_ALLOC_OBJECT) {
                obj_no = unallocObjStruct.size++;
                uid = unallocObjId;
                ConfArrayCopy(objectsId, uid, "spectrograph", sizeof(char),
                    i, obj_no, status);
                ConfArrayCopy(objectsId, uid, "wlenIndex", sizeof(short),
                    i, obj_no, status);

            } else {            /* (objectNotExist[i] == CONF_C_ALLOC_GUIDE) */

                obj_no = unallocGuiStruct.size++;
                uid = unallocGuiId;
                ConfArrayCopy(objectsId, uid, "dhdx", sizeof(double), i,
                    obj_no, status);
                ConfArrayCopy(objectsId, uid, "dhdy", sizeof(double), i,
                    obj_no, status);
                ConfArrayCopy(objectsId, uid, "dddx", sizeof(double), i,
                    obj_no, status);
                ConfArrayCopy(objectsId, uid, "dddy", sizeof(double), i,
                    obj_no, status);

            }

            ConfArrayCopy(objectsId, uid, "name", NAMELEN, NAMELEN * i,
                NAMELEN * obj_no, status);
            ConfArrayCopy(objectsId, uid, "ra", sizeof(double), i, obj_no,
                status);
            ConfArrayCopy(objectsId, uid, "dec", sizeof(double), i, obj_no,
                status);
            ConfArrayCopy(objectsId, uid, "x", sizeof(INT32), i, obj_no, status);
            ConfArrayCopy(objectsId, uid, "y", sizeof(INT32), i, obj_no, status);
            ConfArrayCopy(objectsId, uid, "type", sizeof(char), i, obj_no,
                status);
            ConfArrayCopy(objectsId, uid, "priority", sizeof(short), i, obj_no,
                status);
            ConfArrayCopy(objectsId, uid, "magnitude", sizeof(double), i,
                obj_no, status);
            ConfArrayCopy(objectsId, uid, "pId", sizeof(short), i, obj_no,
                status);
            if (*status == STATUS__OK) {
               ConfArrayCopy(objectsId, uid, "pmra", sizeof(double), i, obj_no,
                                                                       status);
               ConfArrayCopy(objectsId, uid, "pmdec", sizeof(double), i, obj_no,
                                                                       status);
               if (*status == SDS__NOITEM) *status = STATUS__OK;
            }
            ConfArrayCopy(objectsId, uid, "comment", CMTLEN, CMTLEN * i,
                CMTLEN * obj_no, status);

/*
 *          Indicate object is allocated and set index.
 */
            ConfArrayPutc(uid, "allocated", obj_no, 1, status);
            indexPtr[i] = obj_no;


        }
    }

/*
 *  Clean up.
 */

  Exit:
    if (objectNotExist)
        free(objectNotExist);
    Tcl_DeleteHashTable(&hashTable);

    if (objectsId)
        SdsFreeId(objectsId, &ignore);

    if (unallocGuiId)
        SdsFreeId(unallocGuiId, &ignore);

    if (unallocObjId)
        SdsFreeId(unallocObjId, &ignore);

    ConfIndexIdsFree(&objectsStruct);
    ConfIndexIdsFree(&unallocGuiStruct);
    ConfIndexIdsFree(&unallocObjStruct);

}

/*  ------------------------------------------------------------------------- */

/*                    C o n f  W h i c h  S p e c
 *
 *  Determine which fibre type a specified pivot is associated with.
 *  If the Descr argument is non-null, it will be set to a string 
 *  description of the fibre type. This is mainly a wrap-up of the
 *  FPIL routine FpilConWhichSpec(). The name is now slightly misleading.
 *  Originally (in 2dF) the only important distinctions between fibres
 *  (other than guide fibres, which were handled separately) was which
 *  of two possible spectrographs they were associated with. Now, particularly
 *  with FLAMES, where there are a number of different fibre types (not
 *  necessarily all associated with different spectrographs as such),
 *  this routine is really concerned with the general case of what type
 *  a fibre is, and this is no longer just a simple case of having
 *  different spectrographs. However, the name has persisted.
 *
 *  Note. this function takes pivot from 0 to NPiv-1, whilst FPIL uses
 *  1 to NPiv.
 */
 
int ConfWhichSpec(int IPiv, const char **Descr)
{
    unsigned Spec;
    const char *SpecDescr;
    StatusType status = STATUS__OK;

    FpilConWhichSpec(*Instrument, InstConstDetails, CurrentField, IPiv + 1,
        &Spec, &SpecDescr, &status);

    if (Descr)
        *Descr = SpecDescr;

    return Spec;

}


/*  ------------------------------------------------------------------------- */

/*                    C o n f  R e m o v e  A l l o c
 *
 *   This is a utility routine used by ConfCompressUnalloc(). It removes
 *   the items flagged as 'allocated' from the unallocated sub-structures
 *   of the configuration data structure.
 *
 *   Modified:
 *      20th Jan 2003. KS. No longer returns bad status if allocated 
 *                         array not present.
 *      26th Jul 2006. TJF.Improve commenting.
 */
 
void ConfRemoveAlloc(SdsIdType id, StatusType * status)
 {
    SdsIdType allid;
    char c;
    unsigned int i;
    unsigned int j;
    unsigned int nalloc;  /* Number NOT allocated */
    SdsCodeType code;
    char name[16];
    long ndims;
    unsigned long dims[7];
    unsigned long actlen;
    char *dataPnt;


    /*  Failing to find an "allocated" array should not be regarded as a 
     *  problem. It just indicates that this structure has already been
     *  compressed by this routine.
     */
     
    SdsFind(id, "allocated", &allid, status);
    if (*status != STATUS__OK) {
        *status = STATUS__OK;
        return;
    }
    SdsInfo(allid, name, &code, &ndims, dims, status);
    nalloc = 0;
    /*
     * Work out the number that are NOT allocated
     */
    SdsPointer(allid, (void **)&dataPnt, &actlen, status);
    for (i = 0; i < dims[0]; i++) {
        if (dataPnt[i] == 0)
            nalloc++;
    }
    j = nalloc - 1; 
    /*
     * Working through the fibre entries.  Remember, we are
     * removing the "allocated" objects.  We want to unallocated
     * items from the end of the array (indicies nalloc onwards) to the
     * available entries in the first part of the array (items which are
     * currently marked as allocated, and hence are unwanted)
     */
    for (i = 0; i < nalloc; i++) {
        /*
         * Work out if this item is allocated.
         */
        SdsGet(allid, sizeof(char), i, &c, &actlen, status);
        if (c == 1) {
            /*
             * Yes it is.  Increment j to point to the next
             * non-allocated object.
             */
            do {
                j++;
                SdsGet(allid, sizeof(char), j, &c, &actlen, status);
            }
            while (c == 1);
            /*
             * At this point, j is the index to the item we want to copy (an
             * unallocated item in the part of the array we want to trim
             * off).    i is the index of the next "allocated" item (and there
             * for available item) in the first part of the structure.
             *
             * Now copy this item.
             */
            ConfCopyAllItems(id, j, i, status);
        }
    }
    SdsDelete(allid, status);
    SdsFreeId(allid, status);
    ConfResizeAllItems(id, nalloc, status);


}

/*  ------------------------------------------------------------------------- */

/*                      C o n f  C o p y  A l l  I t e m s
 *
 *  Within the specified SDS structure, we have a series of arrays.
 *  We want to copy item "source" to item "destin" in each element
 *  of the array. This is a utility routine used by ConfRemoveAlloc().
 */
 
void ConfCopyAllItems(SdsIdType id,
    int source, int destin, StatusType * status)
 {
    SdsIdType array_id;
    int index;
    char value[CMTLEN];
    SdsCodeType code;
    char name[16];
    long ndims;
    unsigned long dims[7];
    unsigned long actlen;
    unsigned long size = 0;

    index = 1;
    do {
        /*
         * Find the next item in the structure and get its details.
         */
        SdsIndex(id, index, &array_id, status);
        SdsInfo(array_id, name, &code, &ndims, dims, status);
        switch (code) {
        case SDS_CHAR:
        case SDS_BYTE:
        case SDS_UBYTE:
            size = sizeof(char);
            break;
        case SDS_USHORT:
        case SDS_SHORT:
            size = sizeof(short);
            break;
        case SDS_UINT:
        case SDS_INT:
            size = sizeof(INT32);
            break;
        case SDS_FLOAT:
            size = sizeof(float);
            break;
        case SDS_DOUBLE:
            size = sizeof(double);
            break;
        }
        if (ndims == 1) {
            /*
             *  There is only one dimension to this item.  We are just
             *  copying a scalar item.
             */
            SdsGet(array_id, size, source, value, &actlen, status);
            SdsPut(array_id, size, destin, value, status);
        } else {                /* ndims == 2 */

            /*
             *  There are two dimensions, dims[0] specifies the size of
             *  each dimension.  We assume CMTLEN is the largets value 
             *  for dims[0], but don't check this
             */
            SdsGet(array_id, size * dims[0], source * dims[0], value, &actlen,
                status);
            SdsPut(array_id, size * dims[0], destin * dims[0], value, status);
        }
        index++;
        /*
         * Finished with this id.
         */
        SdsFreeId(array_id, status);
    }
    while (*status == STATUS__OK);
    if (*status == SDS__NOITEM)
        ErsAnnul(status);
}



/*  ------------------------------------------------------------------------- */

/*                 C o n f  R e s i z e  A l l  I t e m s
 *
 *   This is a utility routine used by ConfRemoveAlloc().  When allocated
 *   items have been removed from a sub-structure of the configuration
 *   data structure, this is called to resize the SDS arrays to their
 *   new, smaller, size.
 */
 
void ConfResizeAllItems(SdsIdType id, int size, StatusType * status)
 {
    SdsIdType id2;
    int index;
    SdsCodeType code;
    char name[16];
    long ndims;
    unsigned long dims[7];

    index = 1;
    do {
        SdsIndex(id, index, &id2, status);
        SdsInfo(id2, name, &code, &ndims, dims, status);
        if (ndims == 1) {
            dims[0] = size;
            SdsResize(id2, ndims, dims, status);
        } else {
            dims[1] = size;
            SdsResize(id2, ndims, dims, status);
        }
        index++;
        SdsFreeId(id2, status);
    }
    while (*status == STATUS__OK);
    if (*status == SDS__NOITEM)
        ErsAnnul(status);
}


/*  ------------------------------------------------------------------------- */

/*                    C o n f  A r r a y  G e t s
 *
 *   This is a utility routine that gets the num'th item from the array
 *   of shorts given by the name parameter within the SDS structure whose
 *   id is supplied.
 */
 
void ConfArrayGets(SdsIdType id, char *name, long num,
    short *s, StatusType * status)
 {
    SdsIdType tid;
    unsigned long actlen;

    if (*status != STATUS__OK)
        return;
    SdsFind(id, name, &tid, status);
    SdsGet(tid, sizeof(short), num, s, &actlen, status);
    SdsFreeId(tid, status);
    if (*status != STATUS__OK) {
        char buff[80];
        sprintf(buff, "Error getting array item %s", name);
        ConfSdsCheck(buff, status);
    }
}

/*  ------------------------------------------------------------------------- */

/*                    C o n f  A r r a y  P u t s
 *
 *   This is a utility routine that sets the num'th item from the array
 *   of shorts given by the name parameter within the SDS structure whose
 *   id is supplied to the supplied value.
 */
 
void ConfArrayPuts(SdsIdType id, char *name, long num,
    short s, StatusType * status)
 {
    SdsIdType tid;

    if (*status != STATUS__OK)
        return;
    SdsFind(id, name, &tid, status);
    SdsPut(tid, sizeof(short), num, &s, status);
    SdsFreeId(tid, status);
    if (*status != STATUS__OK) {
        char buff[80];
        sprintf(buff, "Error putting array item %s", name);
        ConfSdsCheck(buff, status);
    }
}

/*  ------------------------------------------------------------------------- */

/*                    C o n f  A r r a y  G e t c
 *
 *   This is a utility routine that gets the num'th item from the array
 *   of char given by the name parameter within the SDS structure whose
 *   id is supplied.
 */

void ConfArrayGetc(SdsIdType id, char *name, long num,
    char *c, StatusType * status)
 {
    SdsIdType tid;
    unsigned long actlen;

    if (*status != STATUS__OK)
        return;
    SdsFind(id, name, &tid, status);
    SdsGet(tid, sizeof(char), num, c, &actlen, status);
    SdsFreeId(tid, status);
    if (*status != STATUS__OK) {
        char buff[80];
        sprintf(buff, "Error getting array item %s", name);
        ConfSdsCheck(buff, status);
    }
}

/*  ------------------------------------------------------------------------- */

/*                    C o n f  A r r a y  P u t c
 *
 *   This is a utility routine that sets the num'th item from the array
 *   of char given by the name parameter within the SDS structure whose
 *   id is supplied to the supplied value.
 */
 
void ConfArrayPutc(SdsIdType id, char *name, long num,
    char c, StatusType * status)
 {
    SdsIdType tid;

    if (*status != STATUS__OK)
        return;
    SdsFind(id, name, &tid, status);
    SdsPut(tid, sizeof(char), num, &c, status);
    SdsFreeId(tid, status);
    if (*status != STATUS__OK) {
        char buff[80];
        sprintf(buff, "Error putting array item %s", name);
        ConfSdsCheck(buff, status);
    }
}

/*  ------------------------------------------------------------------------- */

/*                    C o n f  A r r a y  G e t i
 *
 *   This is a utility routine that gets the num'th item from the array
 *   of ints given by the name parameter within the SDS structure whose
 *   id is supplied.
 */
 
void ConfArrayGeti(SdsIdType id, char *name, long num,
    INT32 *i, StatusType * status)
 {
    SdsIdType tid;
    unsigned long actlen;

    if (*status != STATUS__OK)
        return;
    SdsFind(id, name, &tid, status);
    SdsGet(tid, sizeof(INT32), num, i, &actlen, status);
    SdsFreeId(tid, status);
    if (*status != STATUS__OK) {
        char buff[80];
        sprintf(buff, "Error getting array item %s", name);
        ConfSdsCheck(buff, status);
    }
}

/*  ------------------------------------------------------------------------- */

/*                    C o n f  A r r a y  P u t i
 *
 *   This is a utility routine that sets the num'th item from the array
 *   of ints given by the name parameter within the SDS structure whose
 *   id is supplied to the supplied value.
 */
 
void ConfArrayPuti(SdsIdType id, char *name, long num,
    INT32 i, StatusType * status)
 {
    SdsIdType tid;

    if (*status != STATUS__OK)
        return;
    SdsFind(id, name, &tid, status);
    SdsPut(tid, sizeof(INT32), num, &i, status);
    SdsFreeId(tid, status);
    if (*status != STATUS__OK) {
        char buff[80];
        sprintf(buff, "Error putting array item %s", name);
        ConfSdsCheck(buff, status);
    }
}

/*  ------------------------------------------------------------------------- */

/*                    C o n f  A r r a y  P u t d
 *
 *   This is a utility routine that sets the num'th item from the array
 *   of doubles given by the name parameter within the SDS structure whose
 *   id is supplied to the supplied value.
 */
 
void ConfArrayPutd(SdsIdType id, char *name, long num,
    double d, StatusType * status)
 {
    SdsIdType tid;

    if (*status != STATUS__OK)
        return;
    SdsFind(id, name, &tid, status);
    SdsPut(tid, sizeof(double), num, &d, status);
    SdsFreeId(tid, status);
    if (*status != STATUS__OK) {
        char buff[80];
        sprintf(buff, "Error putting array item %s", name);
        ConfSdsCheck(buff, status);
    }
}

/*  ------------------------------------------------------------------------- */

/*                C o n f  A r r a y  P u t  S t r i n g
 *
 *   This is a utility routine that sets the num'th item from the array
 *   of strings given by the name parameter within the SDS structure whose
 *   id is supplied to the supplied string. Note the assumption that the
 *   strings in question are all 80 characters long.
 */
 
void ConfArrayPutString(SdsIdType id, char *name, long num,
    char *string, StatusType * status)
 {
    SdsIdType tid;

    if (*status != STATUS__OK)
        return;
    SdsFind(id, name, &tid, status);
    SdsPut(tid, 80, 80 * num, string, status);
    SdsFreeId(tid, status);
    if (*status != STATUS__OK) {
        char buff[80];
        sprintf(buff, "Error putting array item %s", name);
        ConfSdsCheck(buff, status);
    }
}

/*  ------------------------------------------------------------------------- */

/*                    C o n f  A r r a y  C o p y
 *
 *  Function to copy data from one SDS array element to another.  The
 *  two elements must have equivalent types and names within their parent
 *  SDS structure.
 */

void ConfArrayCopy(
    SdsIdType id1,       /* SDS id of structure containing array to be copied */
    SdsIdType id2,       /* SDS id of structure into which array is copied */
    char *name,          /* Name of item in the above structures to be copied */
    long size,           /* Amount of data to copy (bytes) */
    long offset1,        /* Offset (array element units) in from array */
    long offset2,        /* Offset (array element units) in to array */
    StatusType* status)
 {
    SdsIdType tid1, tid2;
    unsigned long actlen;
    char buffer[CMTLEN];
    char *bufpnt;

    if (*status != STATUS__OK)
        return;

/*
 * We have an internal buffer which is normally big enough, but
 * if it is not, malloc space for the buffer.
 */
    if (size > (long) sizeof(buffer))
        bufpnt = malloc(size);
    else
        bufpnt = buffer;

/*
 * Find source and destination items.
 */
    SdsFind(id1, name, &tid1, status);
    SdsFind(id2, name, &tid2, status);
/*
 * Copy the data.
 */
    SdsGet(tid1, size, offset1, bufpnt, &actlen, status);
    SdsPut(tid2, size, offset2, bufpnt, status);

/*
 * Tidy up.
 */
    SdsFreeId(tid1, status);
    SdsFreeId(tid2, status);

    if (size > (long) sizeof(buffer))
        free(bufpnt);

    if (*status != STATUS__OK) {
        char buff[80];
        sprintf(buff, "Error copying item %s", name);
        ConfSdsCheck(buff, status);
    }
}
/*  ------------------------------------------------------------------------- */

/*
 *                C o n f  D o  A d d  C B S  P a i r
 *
 *  This is a utility routine that is used to implement SetCBSPairs{}.
 *  It adds a program target pair as specified in the calling parameters.
 *  May have memory issues.
 *
 *  Author: Brent Miszalski.
 */

void ConfDoAddCBSPair(
    SdsIdType topid, 
    char name[], 
    double ra, 
    double dec,
    char type,
    short priority,
    double magnitude,
    short pId,
    short wlenIndex,
    double pmra,
    double pmdec,
    char comment[],
    StatusType * status)
{
    SdsIdType fid,uid,tid;
    double x, y;
    int xx,yy;
    short s;
    double ra_cen,dec_cen,mjd;
    char buf[NAMELEN];
    SdsCodeType code;
    long ndims, obj_no;
    char spectType = 0;
    unsigned long dims[7];

    
    /* To start with we need to find the new X and Y coords of the target
     * (RA/DEC are already altered to taste before the call to ConfDoAddCBSPair)
     * To do this we use FpilModelRd2Xy.
     */
    SdsFind(topid, "fieldData", &fid, status);
    
    ArgGetd(fid, "appRa", &ra_cen, status);
    ArgGetd(fid, "appDec", &dec_cen, status);
    ArgGetd(fid, "configMjd", &mjd, status);

    FpilModelRd2Xy(TelModel, ra_cen, dec_cen, ra, dec, mjd, &x, &y, status);

    if(type == 'P'){
      SdsFind(topid, "unallocObject", &uid, status);
    } else if (type == 'F'){
      SdsFind(topid, "unallocGuide", &uid, status);
    }
    /* Now we have to enlarge the unallocated objects section
       and insert the new target data into the newly created space  */

    /* The targetType is now passed in by the caller. The spectType
     * is not actually used and is hard wired to 0.
     */
     /* FpilTargetSkyType(*Instrument, &targetType, &spectType); */
    
    SdsFind(uid, "ra", &tid, status);
    SdsInfo(tid, buf, &code, &ndims, dims, status);
    SdsFreeId(tid, status);
    ConfResizeAllItems(uid, dims[0] + 1, status);
    
    obj_no = dims[0];

    SdsFind(uid, "name", &tid, status);
    SdsPut(tid, NAMELEN, NAMELEN * obj_no, name, status);
    SdsFreeId(tid, status);
    
    SdsFind(uid, "ra", &tid, status);
    SdsPut(tid, sizeof(double), obj_no, &ra, status);
    SdsFreeId(tid, status);
    
    SdsFind(uid, "dec", &tid, status);
    SdsPut(tid, sizeof(double), obj_no, &dec, status);
    SdsFreeId(tid, status);
    
    SdsFind(uid, "x", &tid, status);
    xx = (int)x;
    SdsPut(tid, sizeof(int), obj_no, &x, status);
    SdsFreeId(tid, status);
    
    SdsFind(uid, "y", &tid, status);
    yy = (int)y;
    SdsPut(tid, sizeof(int), obj_no, &y, status);
    SdsFreeId(tid, status);
    
    SdsFind(uid, "type", &tid, status);
    SdsPut(tid, sizeof(char), obj_no, &type, status);
    SdsFreeId(tid, status);
    
    SdsFind(uid, "priority", &tid, status);
    SdsPut(tid, sizeof(short), obj_no, &priority, status);
    SdsFreeId(tid, status);
    
    SdsFind(uid, "magnitude", &tid, status);
    SdsPut(tid, sizeof(double), obj_no, &magnitude, status);
    SdsFreeId(tid, status);
    
    SdsFind(uid, "pId", &tid, status);
    SdsPut(tid, sizeof(short), obj_no, &pId, status);
    SdsFreeId(tid, status);
    
    SdsFind(uid, "comment", &tid, status);
    SdsPut(tid, 80, 80 * obj_no,comment, status);
    SdsFreeId(tid, status);
    
    /*  We allow for the possibility that pmra and pmdec may be missing */
    
    if (*status == STATUS__OK) {
       SdsFind(uid, "pmra", &tid, status);
       SdsPut (tid, sizeof(double), obj_no, &pmra, status);
       SdsFreeId (tid, status);
       *status = STATUS__OK;
    }
    if (*status == STATUS__OK) {
       SdsFind(uid, "pmdec", &tid, status);
       SdsPut (tid, sizeof(double), obj_no, &pmdec, status);
       SdsFreeId (tid, status);
       *status = STATUS__OK;
    }
    
    if (type != 'F'){
       SdsFind(uid, "spectrograph", &tid, status);
       SdsPut(tid, sizeof(char), obj_no, &spectType, status);
       SdsFreeId(tid, status);

       SdsFind(uid, "wlenIndex", &tid, status);
       SdsPut(tid, sizeof(short), obj_no, &wlenIndex, status);
       SdsFreeId(tid, status);

    }
    /*Don't know what this is for? */
    ConfArrayPutc(uid, "allocated", obj_no, 0, status);
    
    if (*status != STATUS__OK) {
        ConfSdsCheck("Error adding to unalloc objects structure", status);
    }

    if (*status == STATUS__OK) {
        if(type == 'P'){
           ArgGets(fid, "unallocObj", &s, status);
           s++;
           ArgPuts(fid, "unallocObj", s, status);
        } else if (type == 'F'){
           ArgGets(fid, "unallocGui", &s, status);
           s++;
           ArgPuts(fid, "unallocGui", s, status);
        }
    }
    
    SdsFreeId(uid, status);
    SdsFreeId(fid, status);
}
/*  ------------------------------------------------------------------------- */

/*
 *                C o n f  D o  A d d  S k y  P o s i t i o n
 *
 *  This is a utility routine that is used to implement AddSkyPosition{},
 *  but is also called directly from other code in this file. It adds
 *  a sky target as specified in the calling parameters.
 */
  
void ConfDoAddSkyPosition(SdsIdType topid, double xsky, double ysky,
    char title[], char targetType, long *object, StatusType * status)

{
    SdsIdType uid, fid;
    short s;
    char c;
    char name[NAMELEN];
    SdsCodeType code;
    long ndims, obj_no;
    short l;
    int x, y;
    double xx, yy;
    double appra, appdec, mjd;
    double rasky, decsky, ara, adec;
    /* char targetType; */
    char spectType = 0;
    SdsIdType tid;
    double gmap[21];         /* Mean to apparent parameters  */
    unsigned long dims[7];

    double lx, ly, jx, jy, jxx, jyy;

    SdsFind(topid, "fieldData", &fid, status);

    ArgGetd(fid, "appRa", &appra, status);
    ArgGetd(fid, "appDec", &appdec, status);
    ArgGetd(fid, "configMjd", &mjd, status);

    FpilModelPos2Xy(TelModel, xsky, ysky, &xx, &yy, status);
    FpilModelXy2Rd(TelModel, appra, appdec, xx, yy, mjd, &ara, &adec, status);

    FpilModelXy2Pos(TelModel, xx, yy, &lx, &ly, status);
    FpilModelRd2Xy(TelModel, appra, appdec, ara, adec, mjd, &jxx, &jyy, status);
    FpilModelXy2Pos(TelModel, jxx, jyy, &jx, &jy, status);

    slaMappa(2000.0, mjd, gmap);

    slaAmpqk(ara, adec, gmap, &rasky, &decsky);

    SdsFind(topid, "unallocObject", &uid, status);

    /*
     * Now we have to enlarge the unallocated objects section
     * and put the new data into it  
     */

    /*
     * The targetType is now passed in by the caller. The spectType
     * is not actually used and is hard wired to 0.
     */
     /* FpilTargetSkyType(*Instrument, &targetType, &spectType); */

    /*
     * We need the dimensions.
     */
    SdsFind(uid, "ra", &tid, status);
    SdsInfo(tid, name, &code, &ndims, dims, status);
    SdsFreeId(tid, status);

    ConfResizeAllItems(uid, dims[0] + 1, status);
    obj_no = dims[0];
    *object = obj_no;
    SdsFind(uid, "name", &tid, status);
    SdsPut(tid, NAMELEN, NAMELEN * obj_no, title, status);
    SdsFreeId(tid, status);

    SdsFind(uid, "ra", &tid, status);
    SdsPut(tid, sizeof(double), obj_no, &rasky, status);
    SdsFreeId(tid, status);

    SdsFind(uid, "dec", &tid, status);
    SdsPut(tid, sizeof(double), obj_no, &decsky, status);
    SdsFreeId(tid, status);

    SdsFind(uid, "x", &tid, status);
    x = (int) xsky;
    SdsPut(tid, sizeof(int), obj_no, &x, status);
    SdsFreeId(tid, status);

    SdsFind(uid, "y", &tid, status);
    y = (int) ysky;
    SdsPut(tid, sizeof(int), obj_no, &y, status);
    SdsFreeId(tid, status);

    SdsFind(uid, "type", &tid, status);
    c = targetType;
    SdsPut(tid, sizeof(char), obj_no, &c, status);
    SdsFreeId(tid, status);

    SdsFind(uid, "priority", &tid, status);
    l = 9;
    SdsPut(tid, sizeof(short), obj_no, &l, status);
    SdsFreeId(tid, status);


    SdsFind(uid, "magnitude", &tid, status);
    xx = 23.0;
    SdsPut(tid, sizeof(double), obj_no, &xx, status);
    SdsFreeId(tid, status);


    SdsFind(uid, "wlenIndex", &tid, status);
    l = -1;
    SdsPut(tid, sizeof(short), obj_no, &l, status);
    SdsFreeId(tid, status);
    
    if (*status == STATUS__OK) {
       SdsFind (uid,"pmra", &tid, status);
       xx = 0.0;
       SdsPut (tid, sizeof(double), obj_no, &xx, status);
       SdsFreeId (tid, status);
       *status = STATUS__OK;
    }
    if (*status == STATUS__OK) {
       SdsFind (uid,"pmdec", &tid, status);
       xx = 0.0;
       SdsPut (tid, sizeof(double), obj_no, &xx, status);
       SdsFreeId (tid, status);
       *status = STATUS__OK;
    }

    SdsFind(uid, "pId", &tid, status);
    l = 0;
    SdsPut(tid, sizeof(short), obj_no, &l, status);
    SdsFreeId(tid, status);


    SdsFind(uid, "comment", &tid, status);
    SdsPut(tid, 80, 80 * obj_no, "Additional Sky Position", status);
    SdsFreeId(tid, status);


    SdsFind(uid, "spectrograph", &tid, status);
    c = spectType;
    SdsPut(tid, sizeof(char), obj_no, &c, status);
    SdsFreeId(tid, status);


    ConfArrayPutc(uid, "allocated", obj_no, 0, status);
    if (*status != STATUS__OK) {
        ConfSdsCheck("Error adding to unalloc objects structure", status);
    }

    /* Now update the numbers for the GUI */
    if (*status == STATUS__OK) {
        ArgGets(fid, "unallocSky", &s, status);
        s++;
        ArgPuts(fid, "unallocSky", s, status);
    }


    {
        int x, y;
        ConfArrayGeti(uid, "x", obj_no, &x, status);
        ConfArrayGeti(uid, "y", obj_no, &y, status);

        /*  printf("DoAddSky:new object %d position is at %d,%d\n",
           obj_no,x,y); */

    }


    SdsFreeId(uid, status);
    SdsFreeId(fid, status);

}


/* -------------------------------------------------------------------------- */

/*           I m p  Z  E x e  D i r  -  P r e l i m i n a r i e s
 *
 *  To provide the configuration program with the means to determine which
 *  directory it is being run from, the utility routine ImpZExeDir() has been
 *  extracted from the code for the IMP message system used by DRAMA. This
 *  was the most conveniently packaged version of such a routine.
 *  
 *  The ImpZExeDir function and its dependencies have been copied
 *  verbatim from the impz_unix.c file. This short section sets the various
 *  constants used by IMP that are normally defined in IMP include files
 *  and which are used by the ImpZExeDir() code.
 */
 
#define IMPZ_PROTOTYPES_OK 1
#define IMP_FALSE 0
#define IMP_TRUE 1
#if defined(unix) || defined(__unix__) || defined(__unix) || \
                                                      defined(__macosx__)

/* -------------------------------------------------------------------------- */


/*                         I m p  Z  E x e  D i r
 *
 *  Routine:
 *     ImpZExeDir
 *
 *  Function:
 *     Gives the name of the directory from which the current program was run.
 *
 *  Description:
 *     In systems where individual tasks are started by loading executable
 *     programs from disk, it is often useful for a program to be able to
 *     locate the directory from which it was loaded. This enables it to
 *     find related files, which are often held in the same directory, 
 *     in a straightforward way that doesn not require the use of logical
 *     names (VMS), environment variables (UNIX), or some other such scheme.
 *
 *     The interface to this routine requires the zeroth argument passed to
 *     the program when it was started up. This is needed by the UNIX 
 *     implementation, although not by the VMS version. It returns the full
 *     pathname or file specification of the directory containing the
 *     program currently being run. This will be formatted such that it
 *     can be prepended to a filename to get a valid full file specification.
 *     (So, for example, under UNIX it will end with a '/', under VMS it will
 *     end with a ']'.) If the system is such that the concept does not make
 *     sense (embedded systems like VxWorks), or where it simply cannot be
 *     implemented), a 'Known' flag argument will be returned set false.
 *    
 *  Implementation details for UNIX:
 *     Most of the work is done by the routine findx(), originally written
 *     by Greg Limes (limes@eng.sun.com) which is appended to this file. It
 *     works by looking through the directories specified by the PATH
 *     environment variable looking for the program name passed as the
 *     zeroth argument (although there's a number of additional wrinkles).
 */

#ifdef IMPZ_PROTOTYPES_OK
static int findx(char *cmd, char *cwd, char *dir, char **pgm, char **run,
    char *path);
static int resolve(             /* get link resolution name */
    char *indir, char *cmd, char *dir, char **run);

void ImpZExeDir(
    char *Arg0,                 /* Zeroth argument to program */
    char *Directory,            /* Name of 'exe' directory */
    int LDir,                   /* Number of chars in Directory */
    int *Known)                 /* True if Directory is valid */
#else

static int findx();
static int resolve();

void ImpZExeDir(Arg0, Directory, LDir, Known)
char *Arg0;                     /* Zeroth argument to program */
char *Directory;                /* Name of 'exe' directory */
int LDir;                       /* Number of chars in Directory */
int *Known;                     /* True if Directory is valid */
#endif

/*  The following is my (KS) wrap-up of findx(), producing the UNIX version of
 *  ImpZExeDir(). This is a  routine which a) sets up a call to findx()
 *  to get the name of the 'exe' directory, b) manipulates the result
 *  slightly for aesthetic reasons and to make sure the result ends with a
 *  '/' character, and c) returns it in the nul-terminated character string
 *  Directory. The manipulation of the string returned by findx() is because
 *  the findx() algorithm tends to produce strings such as '/home/aatssb/ks/.'
 *  or '/home/aatssb/ks/work/../scrap/.' which are valid but messy if the
 *  result is ever to be output. (This is a small modification of the Figaro
 *  Fortran-callable utility GEN_EXEDIR, which was canibalised to produce
 *  this code, which is probably over fussy for this IMP application.)
 */

{
    /*  Local variables  */

    int Back;                    /* Number of '/' chars to work back through */
    char Char;                   /* Character in directory name */
    char CurrentDir[MAXPATHLEN] = ""; 
                                 /* Work space used for current directory */
    int Dots;                    /* Dot count between '/' characters */
    char ExeDir[MAXPATHLEN] = "";/* Receives the exe directory from findx() */
    int Ichar;                   /* Index into ExeDir */
    int Ioutchar;                /* Index into returned directory string */
    char LastChar;               /* Last character written into Directory */

    /*  Functions used  */


    /*  Set up the call to findx() to get the directory name  */

    *Known = IMP_FALSE;
    Ioutchar = 0;
    if (findx(Arg0, CurrentDir, ExeDir, (char **) 0,
            (char **) 0, getenv("PATH")) == 0) {

        /*  The call to findx() worked, so we now have a directory string. We
         *  combine the copying of it into the Directory string and the tidying
         *  up. We copy it character by character, keeping a count of the number
         *  of dots after each '/', clearing the count on an ordinary character.
         *  If we come to a new '/' or the end of the string with a positive 
         *  dot count, we skip back by a number of '/' chars equal to the 
         *  number of dots we found.  This has the effect of turning a '/./'
         *  sequence into a '/' and a 'xx/../yy' sequence into 'yy'.
         */

        Char = '\0';
        Ichar = 0;
        Dots = 0;
        do {
            Back = 0;
            LastChar = Char;
            Char = ExeDir[Ichar++];
            Directory[Ioutchar] = Char;
            if (Char == '\0') {
                Back = Dots;
                Ioutchar--;
            } else if (Char == '/') {
                Back = Dots;
                Dots = 0;
            } else if (Char == '.') {
                Dots++;
            } else {
                Dots = 0;
            }
            if (Back)
                LastChar = '/';
            while (Back) {
                if (Directory[--Ioutchar] == '/')
                    Back--;
                if (Ioutchar <= 0)
                    break;
            }
            if (Ioutchar < LDir)
                Ioutchar++;
        } while (Char);
        if (LastChar != '/')
            Directory[Ioutchar++] = '/';
        if (Ioutchar < LDir) {
            Directory[Ioutchar] = '\0';
            *Known = IMP_TRUE;
        }
    }

}




/* -------------------------------------------------------------------------- */

/*                               F i n d x
 *
 *  The following is the body of the code of the routine findx() as written
 *  by Greg Lines. I wanted to include it to keep this file self-contained,
 *  but I didn't want to dive into it making minor changes and so propagate
 *  a 'slightly' different version if I could help it. I have added a
 *  'static' to the declaration of findx() itself in order to hide the name
 *  and keep it a local routine. Other than that this is the code as supplied.
 *  The original release includes a man page which documents the routine fully,
 *  but I have not included that. Note that the main() routine declared in the
 *  code is not normally compiled.
 *
 *  Having said that in the original version of this routine, I have now made
 *  a more significant modification to the findx() code. I have introduced the
 *  variables dp and dstr, making it copy the individual path elements into
 *  dstr rather than work on them in place - the original code modified the
 *  PATH string it was passed, and so could not be called more than once. I
 *  have also removed the now unused variable 's'.  Subsequently, I have also
 *  eliminated the BSD function rindex() to allow this to run under SOLARIS,
 *  by using the following Rindex() function and using it instead. I have
 *  also changed the name of the symlink variable to symblink to prevent lint
 *  complaining about it hiding the name of a system function. Moreover, I
 *  have removed the TAB characters from the original code, since the TAB
 *  settings being used didn't match those in the rest of this file and so
 *  made the code layout very odd indeed.
 */

#ifdef IMPZ_PROTOTYPES_OK

char *Rindex(char *s, char c)
#else

char *Rindex(s, c)
char *s, c;

#endif

{
    char *ptr = s;
    char *cptr = (char *) 0;
    while (*ptr) {
        if (*ptr == c)
            cptr = ptr;
        ptr++;
    }
    return (cptr);
}

/*  ------------------------------------------------------------------------- */

#include <errno.h>
#include <sys/param.h>

#ifndef   X_OK
#define   X_OK   1              /* for "access" */
#endif

#ifndef   MAXPATHLEN
#define   MAXPATHLEN   1024
#endif

#ifndef   ENAMETOOLONG
#define   ENAMETOOLONG   EINVAL
#endif

#ifdef   FINDX_TESTMAIN
/*
 * An example of how to use this routine.
 */
extern char *getenv();          /* read value from environment */
char *pn = (char *) 0;          /* program name */
char *rn = (char *) 0;          /* run name */
char rd[MAXPATHLEN];            /* run directory */
char wd[MAXPATHLEN] = ".";      /* working directory */

int main(argc, argv)
int argc;
char **argv;
{
    findx(*argv, wd, rd, &pn, &rn, getenv("PATH"));
    printf("%s: %s running in %s from %s\n", pn, rn, wd, rd);
    return 0;
}
#endif

static int findx(char *cmd, char *cwd, char *dir, char **pgm,
    char **run, char *path)
{
    int rv = 0;
    char dstr[MAXPATHLEN], *dp, *f;
    if (!cmd || !*cmd || !cwd || !dir) {
        errno = EINVAL;         /* stupid arguments! */
        return -1;
    }
    if (!path || !*path)        /* missing or null path */
        path = ".";             /* assume sanity */

    if (*cwd != '/')
        if (!(getcwd(cwd, MAXPATHLEN))) {
            return -1;          /* cant get working directory */
        }

    f = Rindex(cmd, '/');
    if (pgm)                    /* user wants program name */
        *pgm = f ? f + 1 : cmd;

    if (dir) {                  /* user wants program directory */
        rv = -1;
        if (*cmd == '/')        /* absname given */
            rv = resolve("", cmd + 1, dir, run);
        else if (f)             /* relname given */
            rv = resolve(cwd, cmd, dir, run);
        else if ((f = path)) {  /* from searchpath */
            rv = -1;
            errno = ENOENT;     /* errno gets this if path empty */
            while (*f && (rv < 0)) {
                dp = dstr;
                while (*f && (*f != ':'))
                    *dp++ = *f++;
                if (*f) {
                    f++;
                }
                *dp = '\0';
                if (dstr[0] == '/')
                    rv = resolve(dstr, cmd, dir, run);
                else {
                    char abuf[MAXPATHLEN];
                    sprintf(abuf, "%s/%s", cwd, dstr);
                    rv = resolve(abuf, cmd, dir, run);
                }
            }
        }
    }
    return rv;
}

/*
 * resolve - check for specified file in specified directory. sets up dir,
 * following symlinks. returns zero for success, or -1 for error (with
 * errno set properly)
 */
static int resolve(char *indir, /* search directory */
    char *cmd,                  /* search for name */
    char *dir,                  /* directory buffer */
    char **run)
{                               /* resultion name ptr ptr */
    char *p;
#ifdef   ELOOP
    int sll;
    char symblink[MAXPATHLEN + 1];
#endif

    errno = ENAMETOOLONG;
    if ((int) (strlen(indir) + strlen(cmd) + 2) > MAXPATHLEN) {
        return -1;
    }

    sprintf(dir, "%s/%s", indir, cmd);
    if (access(dir, X_OK) < 0) {
        return -1;              /* not an executable program */
    }
#ifdef   ELOOP
    while ((sll = readlink(dir, symblink, MAXPATHLEN)) >= 0) {
        symblink[sll] = 0;
        if (*symblink == '/')
            strcpy(dir, symblink);
        else
            sprintf(Rindex(dir, '/'), "/%s", symblink);
    }
    if (errno != EINVAL) {
        return -1;
    }
#endif

    p = Rindex(dir, '/');
    *p++ = '\0';
    if (run)                    /* user wants resolution name */
        *run = p;
    return 0;
}
#elif defined (WIN32)
/*
 *  Implementation details for WIN32:
 *      We can get the full path name of Arg0 using GetFullPathName().  That
 *      routine also returns a pointer to the name component, so we can
 *      work out what the directory component is and return it.
 */

void ImpZExeDir(char *Arg0,     /* Zeroth argument to program */
    char *Directory,            /* Name of 'exe' directory */
    int LDir,                   /* Number of chars in Directory */
    int *Known)
 {                              /* True if Directory is valid */
    char FullName[MAX_PATH + 1];/* Space for full file name */
    char *NamePtr;              /* Pointer to name component */
/*
 *  Get the full path name of the executable.
 */
    if (GetFullPathName(Arg0, sizeof(FullName), FullName, &NamePtr) == 0) {
        *Known = FALSE;
        goto Exit;
    }
/*
 *  NamePtr points to the the first character of the name component of the
 *  file name.  We null terminate the string here so that we only have
 *  the directory.  We then copy it to the result string and ensure that
 *  is null terminated.
 */
    *NamePtr = '\0';
    strncpy(Directory, FullName, LDir);
    Directory[LDir - 1] = '\0';

    *Known = TRUE;

  Exit:;
}
#endif

/*
         1         2         3         4         5         6         7         8
12345678901234567890123456789012345678901234567890123456789012345678901234567890
*/





