#ifndef CONFIGURE_SUBS_INC
#define CONFIGURE_SUBS_INC
/*+
 *                       C O N F I G U R E _ S U B S . H
 *
 *  Function:
 *     Include file for the utility C subroutines used by 'Configure'.
 *
 *  Description:
 *     This file defines a set of C-callable routines that perform a variety
 *     of specific tasks for the main 'Configure' code. For details, see
 *     the comments at the start of configure_subs.c
 *
 *  Version date: 13th August 2013.
 *
 *  Support: K. Shortridge, Tony Farrell, AAO.
 *-
 *  "@(#) $Id: ACMM:aaoConfigure/src/configure_subs.h,v 12.7 04-Nov-2013 11:28:47+11 tjf $"
 *
 *  History:
 *      4th Mar 2005.  Original version. KS/AAO.
 *      5th Aug 2009.  Added pmra, pmdec arguments to ConfListLine,ConfGetIds,
 *                     and ConfDoAddCBSPair. KS/AAO.
 *     14th Aug 2009.  ConfGetProperMotions() now has a ClearFlag argument, KS.
 *     17th Aug 2009.  ConfListPAFTarget() now takes proper motion ra and
 *                     dec Sds Id arguments. KS.
 *     13th Aug 2013.  Added ConfFibreAffectData(). KS.
 *     16th Sep 2013.  Definition of __EXTENSIONS__ now made conditional.
 *                     Moved ConfFibreAffectData() to config.c, as the routine
 *                     ConfigFibreAffectData(). This conforms to the proper
 *                     rules for what routines allocation code can call. KS. 
 */
 
#ifndef __EXTENSIONS__ 
#define __EXTENSIONS__  /* For solaris include files */
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <ctype.h>
#include <limits.h>

#include "drama.h" 
#include "status.h"
#include "sds.h"
#include "tdFconvert.h"

#ifdef __cplusplus
extern "C" {
#endif


/*  Prototypes for the routines defined in configure_subs.h */

void ConfDoXyConversion(SdsIdType id, double cra, double cdec,
                        double mjd, double *gmap, const wavelensType * wlens, 
                        StatusType * status);
void ConfArrayGets(SdsIdType id, char *name, long num,
    short *s, StatusType * status);
void ConfArrayGetc(SdsIdType id, char *name, long num,
    char *s, StatusType * status);
void ConfArrayPuts(SdsIdType id, char *name, long num,
    short s, StatusType * status);
void ConfArrayPutc(SdsIdType id, char *name, long num,
    char s, StatusType * status);
void ConfArrayPuti(SdsIdType id, char *name, long num,
    INT32 i, StatusType * status);
void ConfArrayGeti(SdsIdType id, char *name, long num,
    INT32 *i, StatusType * status);
void ConfArrayPutd(SdsIdType id, char *name, long num,
    double d, StatusType * status);
void ConfArrayPutString(SdsIdType id, char *name, long num,
    char *string, StatusType * status);
void ConfArrayCopy(SdsIdType id1, SdsIdType id2, char *name, long size,
    long num1, long num2, StatusType * status);
void ConfCopyAllItems(SdsIdType id, int j, int i, StatusType * status);
void ConfResizeAllItems(SdsIdType id, int size, StatusType * status);
void ConfRemoveAlloc(SdsIdType id, StatusType * status);
void ConfListLine(FILE * listing, int i,
                  SdsIdType ra_id, SdsIdType dec_id,
                  SdsIdType name_id, SdsIdType type_id,
                  SdsIdType spect_id, SdsIdType prio_id,
                  SdsIdType mag_id, SdsIdType pid_id, SdsIdType comment_id,
                  SdsIdType wlenIndex_id, SdsIdType pmra_id, SdsIdType pmdec_id,
                  StatusType * status);
void ConfListPAFTarget(FILE * listing, int i,
    char* prefix,
    SdsIdType ra_id,
    SdsIdType dec_id,
    SdsIdType name_id, SdsIdType type_id, SdsIdType spect_id,
    SdsIdType prio_id, SdsIdType mag_id, SdsIdType pid_id,
    SdsIdType pmra_id, SdsIdType pmdec_id, SdsIdType comment_id, 
    int fibre,
    int PAF_Flags,
    StatusType * status);
int ConfGetProperMotions ( 
       char* Comment, short ClearFlag, double* PmRa,double* PaDec);                    /* Proper motion in Dec */
void ConfListDSSLine(int skyonly,
    FILE * listing, int i, SdsIdType ra_id,
    SdsIdType dec_id,
    SdsIdType name_id, SdsIdType type_id, SdsIdType prio_id,
    SdsIdType mag_id, SdsIdType pid_id, SdsIdType comment_id,
    StatusType * status);
void ConfGetIds(SdsIdType topid, SdsIdType * ra_id, SdsIdType * dec_id,
    SdsIdType * name_id, SdsIdType * type_id, SdsIdType * spect_id,
    SdsIdType * prio_id,
    SdsIdType * mag_id, SdsIdType * pid_id, SdsIdType * comment_id,
    SdsIdType * wlenIndex_id, SdsIdType * pmra_id, SdsIdType * pmdec_id,
    StatusType * status);
int ConfParseComment (char* Comment, char* Keyword, char* Value,
    char** NextChar);    
int ConfWhichSpec(int IPiv, const char **Descr);
void DoMakeAllocated (
   SdsIdType topid,
   SdsIdType pivid,
   StatusType* status);
void ConfMakeAllocatedIndex(SdsIdType topid, StatusType * status);
void ConfSdsCheck(char *text, StatusType * status);
void ChangeMjd (SdsIdType fieldId, double mjd, StatusType *status);
void ConfDoAddSkyPosition(SdsIdType topid, double xsky, double ysky,
    char title[], char targetType, long *object, StatusType * status);
void FreeAndFind(const SdsIdType source, const char *const name,
    SdsIdType * const id, StatusType * const status);
void ConfDoAddCBSPair(
    SdsIdType topid, 
    char name[], 
    double ra, 
    double dec,
    char type,
    short priority,
    double magnitude,
    short pId,
    short wlenIndex,
    double pmra,
    double pmdec,
    char comment[],
    StatusType * status);

/*  ImpZExeDir() returns the specification for the directory from which the
 *  executable program is run.  This is a direct copy of the internal DRAMA
 *  routine of the same name. It happens that this is a very low level routine
 *  and can be used stand-alone in this way.
 */

void ImpZExeDir(char *Arg0, char *Directory, int LDir, int *Known);



#ifdef __cplusplus
}
#endif

#endif /* CONFIGURE_SUBS_INC */
