#ifndef _CVTINFO_INC
#define _CVTINFO_INC
/*
 * Used by moviedata.C to support access to telescope model etc.
 *
 * "@(#) $Id: ACMM:aaoConfigure/src/cvtinfo.h,v 12.7 04-Nov-2013 11:28:49+11 tjf $"
 */


/*
 * Refraction parameters used by configure
 */
const double TEMP=285;
const double PRESS=900;
const double HUMID=0.5;
const double WAVE=0.60;


class CvtInfoType {


public:
    
    CvtInfoType(
        FpilType inst,
        const int plate,
        const SdsId &configId,
        const bool timeSet, 
        const double utTime, 
        StatusType *status);



     

    double GetUtc() const {
        return _utc;
    }
    double GetCenMeanRA() const {
        return _cenMeanRA;
    }
    double GetCenMeanDec() const {
        return _cenMeanDec;
    }

    double SpecWavelen() const {
        return _spectrographWavelength;
    }
    double AGWavelen() const {
        return _autoguiderWavelength;
    }
    bool HaveWaveLenArray() const {
        return !_WlensArray.empty();
    }
    short NumWaveLens() const {
        return _WlensArray.size();
    }
    double GetWavelen(unsigned index) const {
         return _WlensArray[index];
    }
    unsigned long FieldRadius() const {
        return FpilGetFieldRadius(_instrument);
    }



    /*
     * Clone our telescope model and do a conversion init on it for the specfied
     * wavelength and utc.  Return the result.
     */
    FpilModelType *ModelInit(double waveLength, double utc, StatusType *status) const {
        FpilModelType *model = 0;
        FpilModelClone(_telModel, &model, status);
        FpilModelCvtInit(model, utc, _dut, 
                         _temperature,_pressure,  _humidity, _spectrographWavelength, waveLength,
                         &_telModelParams[0], 
                         status); 

        return model;
    }
    /*
     * Update an existing model
     */
    void ModelInit(double waveLength, double utc, FpilModelType *model, StatusType *status) const {
        FpilModelCvtInit(model, utc, _dut, 
                         TEMP,PRESS,  HUMID, WAVE, waveLength,
                         &_telModelParams[0], 
                         status); 
    }


    // Static initialisation options.
    static void Init() {
        _spectrographWavelength = WAVE;
        _autoguiderWavelength   = WAVE;
        _temperature = TEMP;
        _pressure = PRESS;
        _humidity = HUMID;
    }
    static void SetSpecWave(double w) {
        _spectrographWavelength = w;
    }
    static void SetAgWave(double w) {
        _autoguiderWavelength = w;
    }
    static void SetTemp(double t) {
        _temperature = t;
    }

    static void SetPress(double t) {
        _pressure = t;
    }
    static void SetHumidity(double h) {
        _humidity = h;
    }


private:
    WlensArrayType      _WlensArray;
    double              _configMjd;
    double              _cenMeanRA;
    double              _cenMeanDec;
    double              _utc;

    double              _dut;
    double              _spWavelen;
    double              _agWavelen;
 
    FpilType            _instrument;
    std::vector<double> _telModelParams;
    FpilModelType       *_telModel;


    /* Assignment and Copy operators, made private to avoid misuse */
    CvtInfoType& operator=(const CvtInfoType & a); /* Assignment */
    CvtInfoType(const CvtInfoType&); /* Copy */

    static double _spectrographWavelength;
    static double _autoguiderWavelength;
    static double _temperature;
    static double _pressure;
    static double _humidity;


    
};

class CvtType {


public:
    void SetUtc(double utc) {
        _utc = utc;
        slaMap( _cenMeanRA, _cenMeanDec , 0,0,0,0, 2000, _utc,
                &_cenRA, &_cenDec);

    }
    /*
     * Use the conversion information to initialise a model for a specified
     * wavelength.
     */
    CvtType(double wavelength, CvtInfoType & cvtInfo, StatusType *status) 
        : _cenMeanRA(cvtInfo.GetCenMeanRA()), _cenMeanDec(cvtInfo.GetCenMeanDec()), _cvtInfo(&cvtInfo) {
        
        SetUtc(cvtInfo.GetUtc());
        _telModel = cvtInfo.ModelInit(wavelength, _utc, status);
    }

    void Rd2Pos(double ra, double dec, 
                double *x, double *y, 
                StatusType *status) const {
        
        
        if (*status != STATUS__OK) return;
        double xFp, yFp;
        FpilModelRd2Xy( _telModel, _cenRA, _cenDec, ra, dec, _utc,&xFp ,&yFp,status);
        FpilModelXy2Pos(_telModel, xFp, yFp, x, y, status);

    }

    /*
     * Update an existing model for a new wavelength.
     */
    void ModelInit(double waveLength, StatusType *status) {
        _cvtInfo->ModelInit(waveLength, _utc, _telModel, status);
    }


    double GetUtc() const {
        return _utc;
    }


private:
    double              _cenMeanRA;
    double              _cenMeanDec;
    double              _cenRA;
    double              _cenDec;

    CvtInfoType    *    _cvtInfo;
    FpilModelType  *    _telModel;
    double              _utc;
    

    /* Assignment and Copy operators, made private to avoid misuse */
    CvtType& operator=(const CvtType &a); /* Assignment */
    CvtType(const CvtType&); /* Copy */




};





#endif
