/*+        		M O V I E D A T A    

 *  Module name:
	MOVIEDATA
     
 *  Function:
        Given a 2dF configuration file and an observation duration - produce 
        movie data for use by the moviedata.tcl script.

 *  Synopsis:
        moviedata [options] <2dF_config_sds_file> <duration>


 *  Options:
        -r              => Don't do rotation correction of data.  (The rotation
                            correction presumes the guider is being used to
                            correct field rotation errors.
        -p <plate>      => Plate number, 0 or 1.  Defaults to plate 0.
        -l <time>       => Time of observation - Local time format.
        -u <time>       => Time of observation - UT time format.
        -m <mjd>        => As a floating point MJD.
        -f <filename>   => The base component of the output file name.
        -i <instrument> => The instrument name.  Defaults to 2df (which is
                             2dF-AAOmega.  Other options are 6df and ozpoz.
        -s <wlen>       => Spectrograph wavelength in microns
        -g <wlen>       => Guider wavelength in microns
        -t <temp>       => Observation temp in kelvins
        -h <humid>      => Observation humidity  - 0 to 1.
        -a <press>      => Observation pressure in millibars.

        If the time of observation is not provided - the config file in the 
        config file is used.

        If the -f filename option is not used, the config file name (and directory) 
        is used.

 *  Description:
         This program uses code similar to tweak optimization code used by the 
         2dF control task "tweak" code - to deterime the position of the fibres 
         and the position of the objects throughout an observation.  

         Three files are created
             <filename>.basic_tweak.csv -> A comma seperated list format file 
                   containing the basic X/Y positions of each (used) fibre from a 
                   basic tweak at the specified time.

             <filename>.optimized.csv -> A comma seperated list format file containing
                   the X/Y positions of each (used) fibre from an optimized tweak for
                   the specified duration centered about the specified time.

             <filename>.movie_data.csv -> A command seperated list format file 
                   containing details from the optimization process - allowing a 
                   movie for how the object position moves to be constructed.  
                   See moviedata.tcl for an example of how to display this.

         Instrument names.
              2df   => 2dF AAOmega mode, on the AAT
              6df   => 6dF on the UK Schmidt
              ozpoz => OzPoz/Flames on the VLT UT4.


 *  Language:
      C

 *  Support: Tony Farrell, AAO

 *-

 *  History:
      28-Jul-2008 - TJF - Original version.
      22-Oct-2008 - TJF - Many changes since last comment - now works nicely with 
                            moviedata.tcl script to display result.
     23-Sep-2009 - TJF - Include assert.h.  Needed for MacOsX Snow Lepoard.
     14-Jan-2010 - TJF - Fix compilation problems picked up by GCC 4.5.2, and
                          make const correct (which is maily the ImpZExeDir() code
			  that had to be changed, and now diverges from other versions
			  of this code).
     07-Aug-2012 - TJF - Fix some comment flaws.

 */




static const char *rcsId="@(#) $Id: ACMM:aaoConfigure/src/moviedata.C,v 12.7 04-Nov-2013 11:28:49+11 tjf $";
static void *use_rcsId = (0 ? (void *)(&use_rcsId) : (void *) &rcsId);

#define DRAMA_ALLOW_CPP_STDLIB
#define _GNU_SOURCE_

#include <assert.h>
#include "sds.h"
#include "arg.h"
#include "fpil.h"

extern "C" {
#include "tdfxy.h"      /* 2dF  - AAT */
#include "sixdffpil.h"  /* 6dF  - UKSCHMIDT */
#include "fpcolFpil.h"  /* OzPoz - VLT - 2 */
#include "slalib.h"
}
#include <ctype.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/param.h>  /* For MAXPATHLEN */
#include <errno.h>
#include <string.h>
#include "tdffpil.h"

#include <vector>
#include "slamac.h"
#include "slalib.h"
typedef std::vector<double> WlensArrayType;



#include <iostream>
#include "cvtinfo.h"

/*
 * cvtinfo.h needs these defined.
 */
double CvtInfoType::_spectrographWavelength;
double CvtInfoType::_autoguiderWavelength;
double CvtInfoType::_temperature;
double CvtInfoType::_pressure;
double CvtInfoType::_humidity;


#include "conf_err.h"

#include "timeUnixMjd.h"
using namespace timeUnixMjd;

static FpilType instrument = 0;


void ImpZExeDir(const char *Arg0, char *Directory, int LDir, int *Known);


/*
 * If the observation duration is greater then this value (in seconds) then
 * trigger the optimization code.  Should reflect same value in 2dFctrlTask/Tweak.C
 */
const double OBS_DUR_OPTIMIZE_TRIGGER = 1800;
/*
 * When running the optimization code - this is the value which determines
 * the time points we examine the data in.  Value is in seconds.
 *
 *  Don't yet know what this should be probably 10 minutes will do, but we
 *  will start off at a minute so we can test the code. 
 */
const double OBS_DUR_OPTIMIZE_DELTA   = 600;





static void TweakObject(
    const unsigned int i,
    const double InputRA,
    const double InputDec,
    const char   InputType,

    const CvtInfoType & cvtInfo,
    const short  * const WlenIndex,

    CvtType * const AgXyPars,
    CvtType * const SpecXyParsDef,
    CvtType * const SpecXyParsAlt,
    INT32  * const specificWavelensUsed,
    double * const altWavelength,

    double * const OutputX,
    double * const OutputY,
    short  * const OutputFibre,
    double * const OutputWlen,
    CvtType ** const modelUsed, 
    StatusType * const status);

static void TweakWithOptimize(
    FILE *dataFile,
    const bool dontRotate, 
    double obsDuration,
    unsigned ObjectCount,
    const double * const InputRA,
    const double * const InputDec,
    const char   * const InputType,

    const CvtInfoType & cvtInfo,
    const short  * const WlenIndex,

    CvtType * const AgXyPars,
    CvtType * const SpecXyParsDef,
    CvtType * const SpecXyParsAlt,
    INT32  * const specificWavelensUsed,
    double * const altWavelength,

    double * const OutputX,
    double * const OutputY,
    short  * const OutputFibre,
    double * const OutputWlen,

    StatusType * const status);


			  
static int CreateData( 
    FpilType inst, bool dontRotate,  int plate,  bool timeSet, 
    double utTime, const std::string &configFile,
    const std::string &outFilePrefix,
    double obsDuration);
/*
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
 */
const double MAX_DURATION = 60.0*60.0*14.0;

static std::string ConfigFilesDir;
#ifndef MAXPATHLEN
#define MAXPATHLEN 1024
#warning "Defining own MAXPATHLEN"
#endif
static char ExecutableDir[MAXPATHLEN];

int main(int argc, char *argv[])
{

    int plate  = 0;
    double utTime = 0;
    bool haveUtTime = false;
    int c;
    std::string filePrefix;
    bool haveFilePrefix = false;
    bool haveInstrument = false;
    bool dontRotate = false;

    CvtInfoType::Init();

    opterr = 1;
    while ((c = getopt(argc, argv, "rp:u:l:f:m:i:s:g:t:h:a:")) != -1)
    {
        switch (c)
        {
        case 'p': /* Plate number specified */
            plate = atoi(optarg);
            if (haveInstrument)
            {
                if ((plate < 0)||(plate > (int)FpilGetNumFields(instrument)))
                {
                    fprintf(stderr, "Invalid field number - %d, must be between 0 and %d\n",
                            plate, FpilGetNumFields(instrument));
                    exit(1);
                }
                
            }
            else if ((plate < 0)||(plate > 1))
            {
                fprintf(stderr, "Invalid field number - %d, must be 0 or 1\n", plate);
                fprintf(stderr, " Note - presumed instrument 2dF, if not, use -i before -p\n");
                exit(1);
            }
            
            break;
        case 'u': /* UT time of observation specified */
            utTime = S2Mjd(optarg, false);
            if (utTime <= UNIXSTART)
                exit(1);
            haveUtTime = true;
            break;
        case 'l': /* Local time of observation specified */
            utTime = S2Mjd(optarg, true);
            if (utTime == 0)
                exit(1);
            haveUtTime = true;
            break;
            
        case 'f':  /* File name specified */
            filePrefix = optarg;
            haveFilePrefix = true;
            break;
        case 'm':  /* MJD of observation specified */
            utTime = atof(optarg);
            if (utTime <= UNIXSTART)
            {
                fprintf(stderr, "The specified MJD (%.1f) is before the start of the UNIX clock - 1 Jan 1970 (MJD %.1f)\n",
                        utTime, UNIXSTART);
                exit(1);
            }
            haveUtTime = true;
            break;
        case 'i':
            haveInstrument = true;
            if (strcasecmp(optarg, "2df") == 0)
            {
                TdfFpilFullInitAAOmega(&instrument);
                plate = 0;
                fprintf(stderr, "Selected 2dF-AAOmega instrument\n");
            }
            else if (strcasecmp(optarg, "6df") == 0)
            {
                sixdfFpilFullInit(&instrument);
                fprintf(stderr, "Selected 6dF instrument\n");
            }
            else if (strcasecmp(optarg, "ozpoz") == 0)
            {
                fpcolFpilInit(&instrument);
                fprintf(stderr, "Selected 6dF instrument\n");
            }
            else 
            {
                fprintf(stderr, "Instrument \"%s\" is unknown\n", optarg);
                fprintf(stderr, "  Must be one of 2df/6df/ozpoz\n");
                exit(1);
            }
            break;
                
        case 's':
        {
            double value = atof(optarg);
            if (value > 0)
            {
                CvtInfoType::SetSpecWave(value);
                fprintf(stderr,"Spectrograph wavelength set to %.1f microns\n", value);
            }
            else
                fprintf(stderr, "Invalid wavelength %.1f\n", value);
            break;
        }
        case 'g':
        {
            double value = atof(optarg);
            if (value > 0)
            {
                CvtInfoType::SetAgWave(value);
                fprintf(stderr,"Autoguider wavelength set to %.1f microns\n", value);
            }
            else
                fprintf(stderr, "Invalid wavelength %.1f\n", value);
            break;
        }
            
        case 't':
        {
            double value = atof(optarg);
            if (value > 0)
            {
                CvtInfoType::SetTemp(value);
                fprintf(stderr,"Temperature set to %.1f kevins\n", value);
            }
            else
                fprintf(stderr, "Invalid temperature %.1f\n", value);
            break;
        }
        case 'h':
        {
            double value = atof(optarg);
            if ((value > 0)&&(value <= 1))
            {
                CvtInfoType::SetHumidity(value);
                fprintf(stderr,"Humidity set to %.1f\n", value);
            }
            else
                fprintf(stderr, "Invalid humidity %.1f\n", value);
            break;
        }
        case 'a':
         {
            double value = atof(optarg);
            if ((value > 0)&&(value < 2000))
            {
                CvtInfoType::SetSpecWave(value);
                fprintf(stderr,"Pressure set to %.1f millibars\n", value);
            }
            else
                fprintf(stderr, "Invalid pressure %.1f\n", value);
            break;
        }
        case 'r':
        {
            dontRotate = true;
            break;
        }
        case ':':
            fprintf(stderr, "Argument not supplied to option %c\n", optopt);
            exit(1);
        case '?':
            fprintf(stderr, "Invalid option %c\n", optopt);
            exit(1);
        default:
            fprintf(stderr, "Unexpected return from getopt()\n");
            exit(1);
            
        }
    }


    if ((optind + 2) > argc)
    {
        fprintf(stderr, "Insufficent arguments - must be two\n");
        exit(1);
    }
    std::string configFile(argv[optind++]);
    int duration(atoi(argv[optind++]));
    if ((duration <= 0)||(duration > MAX_DURATION))
    {
        fprintf(stderr, "Duration (second arg) must be and integer value greater then zero or greater then %d - duration in seconds\n", (int)(MAX_DURATION));
        exit(1);
    }

    
    if (!haveFilePrefix)
    {
        typedef unsigned int size_type;  /* Should be std::size_type */
        filePrefix = configFile;
	std::string::size_type last_slash = filePrefix.rfind('/');
	std::string::size_type last_period = filePrefix.rfind('.');
        if (last_period == std::string::npos)
        {
            // No period in config file name - use config file name as base of output files.
            //  Nothing to do.
        }
        else if (((last_slash == std::string::npos)&&
		  (last_period != std::string::npos))||
            (last_period > last_slash))
        {
            //   No slash in filename - but we do have a period.
            // OR
            //   Slash and period found, period is after slash.

            // In both cases, remove everything from the period onwards.
            filePrefix.resize(last_period);
        }
        else 
        {
            // last slash is after last period. USe conifg file name as base of output file
            //  Nothign to do.
        }
    }
    fprintf(stderr, "Output file basename is \"%s\"\n", filePrefix.c_str());

    /*
     *  Find where are being run from so we can find other files.
     */
     
    int known;
    ImpZExeDir(argv[0], ExecutableDir, sizeof(ExecutableDir), &known);

    /*  See if the CONFIG_FILE environment variable has a value and if
     *  so, grab it.
     */
     
    if (getenv("CONFIG_FILES"))
        ConfigFilesDir = getenv("CONFIG_FILES");



    /*
     * Set up the default fibre positioner instrument
     */
    if (!haveInstrument)
        TdfFpilFullInitAAOmega(&instrument);


    
    /*
     * Check we have a clone function.
     *
     * Cheat - using internal knowledge about the instrument structure.  
     *
     * This does NOT always work.
     */
    if (!  ((Fpil___InstType *)(instrument))->model.Clone)
    {
        fprintf(stderr,"Instrument %s does not yet support the Clone function in its model structure\n",
                FpilGetInstName(instrument));
        fprintf(stderr,"  Hence we can't run this program\n");
        exit(1);
    }



    return (CreateData(instrument, dontRotate, plate, haveUtTime, utTime, 
                       configFile.c_str(), filePrefix,  duration));
    

}

static void WriteFileHeader(
    FILE *f,
    const std::string &configFile,
    double utc,
    double obsDuration)
{
    fprintf(f,"Instrument name,%s\n",  FpilGetInstName(instrument));
    fprintf(f,"Input File,%s\n", configFile.c_str());
    fprintf(f,"Observation UT Time,%.6f,%s\n", utc, Mjd2Ut(utc).c_str());
    fprintf(f,"Observation Local Time,,%s\n", Mjd2Local(utc).c_str());
    fprintf(f,"Observation Duration (Seconds),%d\n", int(obsDuration));
    fprintf(f,"\n");
}


static int CreateData(
    FpilType inst,                  /* FPIL Instrument structure */
    const bool dontRotate,          /* Don't do rotation correction */
    const int plate,                /* Plate to configure for */
    const bool timeSet,             /* Has the user set the ut time */
    double utTime,                  /* User's ut time */
    const std::string &configFile,  /* Name of configuration (.sds) file */
    const std::string &outFilePrefix,
    const double obsDuration)       /* Observation duration in seconds */
{

    assert(instrument);
    assert((plate >= 0)||(plate <= 4));   
    assert(timeSet ? (utTime > UNIXSTART) : (utTime == 0));
    assert((obsDuration >=0)&&(obsDuration <= MAX_DURATION));

    StatusType status = STATUS__OK;

    /*
     * Read the configuration file.
     */

    SdsId configId(configFile.c_str(), &status);
    if (status != STATUS__OK)
    {
        fprintf(stderr, "Error reading config file \"%s\" - \"%s\"",
                configFile.c_str(), strerror(errno));
        return 1;
    }

    /*
     *  Get access to all the conversion information.
     */
    CvtInfoType cvtInfo(inst, plate, configId, timeSet, utTime,  &status);
    /*
     *  Note about the wavelengths - the first wavelength to CvtInit is what
     *  is presumed to be the field center and the second is the observation
     *  wavelength.  The first one must always be the same, but it doesn't
     *  really matter which one of the  wavelenghts (spectrograph
     *  or autoguider it is) it is - we just want the relationship to be correct.
     */

    /* Default spectrograph model */
    CvtType SpecXyParsDef(cvtInfo.SpecWavelen(), cvtInfo,  &status);

    /* Alternative model - used for when the  wavelength is specified */
    double altWavelength = cvtInfo.SpecWavelen();
    CvtType SpecXyParsAlt(altWavelength, cvtInfo,  &status);


    /* Note - Autoguider central wavelength is the spectrograph
       wavelength, that is what we point the telescope using */
    CvtType AgXyPars(cvtInfo.AGWavelen(), cvtInfo, &status);

    if (status != STATUS__OK)
    {
        fprintf(stderr,
                "Failed to initialise x/y to RA/DEC transformation\n");
        fprintf(stderr,
                "Error code %lx\n", (long)status);
        return 1;
    }


    /*
     *  Get Sds ids to the required information in the Objects structure.
     */ 
    SdsId InputObjectsId(configId,"objects",&status);
    SdsId InputRAId(InputObjectsId,"ra",&status);
    SdsId InputDecId(InputObjectsId,"dec",&status);
    SdsId InputTypeId(InputObjectsId,"type",&status);

    /*
     *  Optional item.  Only later versions of configure put this in.  
     *    The wavelength array index.
     */
    SdsId InputWlenIndexId;
    if (status == STATUS__OK)
    {
        SdsId Id(InputObjectsId, "wlenIndex", &status);
        if (status == STATUS__OK)
            InputWlenIndexId.ShallowCopy(Id);
        else if (status == SDS__NOITEM)
            status = STATUS__OK;
    }
    
    /*
     *  Get the object count, which should be the number of dimensions of
     *  any of the fields within the object field
     */
    unsigned long ObjectCount; /* Count of objects */
    char name[16];
    SdsCodeType code;
    long ndims;
    unsigned long dims[7];
    InputRAId.Info(name,&code,&ndims,dims,&status);    
    if ((status == STATUS__OK)&&
     	((ndims != 1)||(code != SDS_DOUBLE)))
    {
    	status = CONF__FIELD;
    }
    if (status != STATUS__OK)
    {
        fprintf(stderr,
         	"Invalid configuration file format ");
        fprintf(stderr,
                "Error code %lx\n", (long)status);
        return 1;
    }
    ObjectCount = dims[0];

    /*
     *  Get pointers to everything we need
     */
    double *InputRA;
    InputRAId.Pointer((void **)&InputRA,&status);
    
    double *InputDec;
    InputDecId.Pointer((void **)&InputDec,&status);
    
    char *InputType;
    InputTypeId.Pointer((void **)&InputType,&status);
    

    short *WlenIndex = 0;
    if (InputWlenIndexId) {
        InputWlenIndexId.Pointer((void **)&WlenIndex, &status);
    }


    if (status != STATUS__OK)
    {
        fprintf(stderr,
                "Error setting up for tweak\n");
        fprintf(stderr,
                "Error code %lx\n", (long)status);
        return 1;
    }
    
    std::vector<double> OutputX;
    std::vector<double> OutputY;
    std::vector<double> OutputWlen;
    std::vector<short>  OutputFibre;
    OutputX.reserve(ObjectCount);
    OutputY.reserve(ObjectCount);
    OutputWlen.reserve(ObjectCount);
    OutputFibre.reserve(ObjectCount);

    INT32 specificWavelensUsed = 0;
//#define DEBUG_TWEAK
#ifdef DEBUG_TWEAK
#   warning "DEBUG_TWEAK code enabled"
#endif

    const std::string BASIC_SUFFIX(".basic_tweak.csv");
    const std::string OPTIM_SUFFIX(".optimized.csv");
    const std::string OBJECT_SUFFIX(".movie_data.csv");

    const char UnallocatedTargetType = FpilUnallocatedType(instrument);

    
    std::string fileName = outFilePrefix + BASIC_SUFFIX;
    {
        /*
         * Code block just to ensure "basic" is not reused.
         */
        FILE *basic = fopen(fileName.c_str(), "w");
        if (basic == 0)
        {
            fprintf(stderr,"Failed to open output file \"%s\" - %s\n",
                    fileName.c_str(),strerror(errno));
            exit(1);
        }
        fprintf(stderr,"Writing basic X/Y information to file \"%s\"\n", fileName.c_str());
        fprintf(basic,"Object Movie Data File for instrument,%s\n",
                FpilGetInstName(instrument));
        fprintf(basic,"BASIC TWEAK RESULTS FILE\n");
        WriteFileHeader(basic, configFile, cvtInfo.GetUtc(), obsDuration);
        fprintf(basic,",,Fibre,X,Y,Type\n");
        
        /*
         * Tweak the positoin of every input object.
         */ 
        for (unsigned int i = 0; (i < ObjectCount) && (status == STATUS__OK) ; ++i)
        {
            TweakObject(i, 
                        InputRA[i], InputDec[i], InputType[i],
                        cvtInfo, WlenIndex, 
                        &AgXyPars, &SpecXyParsDef, &SpecXyParsAlt,
                        &specificWavelensUsed,
                        &altWavelength,
                        &OutputX[i], &OutputY[i], &OutputFibre[i],
                        &OutputWlen[i],  0,
                        &status);
            
            /*
             * Write out details for all used objects.
             */
            if (InputType[i] != UnallocatedTargetType)
            {
                char targetString[32];
                char targetText[64];
                FpilDecodeTargetType(instrument, InputType[i], ' ', 0, WlenIndex[i], 
                                     targetString, targetText);
                fprintf(basic, ",,%d,%.1f,%.1f,%s\n",
                        i+1, OutputX[i], OutputY[i], targetText);
            }
        }
        fclose(basic);
    }
   
    if ((obsDuration > OBS_DUR_OPTIMIZE_TRIGGER)&&(status == STATUS__OK))
    {
        /*
         * Run the tweak optizatoin code.
         */

        fileName = outFilePrefix + OBJECT_SUFFIX;
        FILE *dataFile = fopen(fileName.c_str(), "w");
        if (dataFile == 0)
        {
            fprintf(stderr,"Failed to open output file \"%s\" - %s\n",
                    fileName.c_str(),strerror(errno));
            exit(1);
        }
        fprintf(stderr,"Writing object movie X/Y information to file \"%s\"\n", fileName.c_str());
        fprintf(dataFile,"Object Movie Data File for instrument,%s\n",
                FpilGetInstName(instrument));
        fprintf(dataFile,"OBJECT POSITIONS OVER OBSERVATION FILE\n");
        fprintf(dataFile,"VERSION,%d\n", 1);
        fprintf(dataFile,"Max Field Radius Microns,%ld\n",
                cvtInfo.FieldRadius());
        WriteFileHeader(dataFile, configFile, cvtInfo.GetUtc(), obsDuration);



        
        TweakWithOptimize(dataFile, dontRotate, obsDuration, ObjectCount, 
                          InputRA, InputDec, InputType, 
                          cvtInfo, WlenIndex, 
                          &AgXyPars, &SpecXyParsDef, &SpecXyParsAlt,
                          &specificWavelensUsed,
                          &altWavelength,
                          &OutputX[0], &OutputY[0], &OutputFibre[0],
                          &OutputWlen[0], 
                          &status);

        fclose(dataFile);

        /*
         * Now write out the resulting object positions.
         */
        fileName = outFilePrefix + OPTIM_SUFFIX;
        FILE *objectPosFile = fopen(fileName.c_str(), "w");
        if (objectPosFile == 0)
        {
            fprintf(stderr,"Failed to open output file \"%s\" - %s\n",
                    fileName.c_str(),strerror(errno));
            exit(1);
        }
        fprintf(stderr,"Writing object actual position X/Y information to file \"%s\"\n", fileName.c_str());
        fprintf(objectPosFile,"Object Movie Data File for instrument,%s\n",
                FpilGetInstName(instrument));
        fprintf(objectPosFile,"OPTIMIZED OBJECT POSITIONS FILE\n");
        WriteFileHeader(objectPosFile, configFile, cvtInfo.GetUtc(), obsDuration);
        fprintf(objectPosFile,",,Fibre,X,Y,Type\n");


        for (unsigned int i = 0;  (i < ObjectCount) && (status == STATUS__OK) ; ++i)
        {
            
            if (InputType[i] != UnallocatedTargetType)
            {
                char targetString[32];
                char targetText[64];
                FpilDecodeTargetType(instrument, InputType[i], ' ', 0, WlenIndex[i], 
                                     targetString, targetText);
                fprintf(objectPosFile, ",,%d,%.1f,%.1f,%s\n",
                        i+1, OutputX[i], OutputY[i], targetText);
            }
        }
        fclose(objectPosFile);

        
    }
    else if (status == STATUS__OK)
    {
        fprintf(stderr, "Observation duration of %d seconds is too low to bother with tweak optimization\n",
                (int)obsDuration);
    }

    if (status != STATUS__OK)
    {
        fprintf(stderr, 
                "Error doing tweak process\n");
        fprintf(stderr,
                "Error code %lx\n", (long)status);
        return 1;
    }

    return 0;
}



/*
 * Initialise the information needed to convert from RA/Dec to X/Y.
 */

CvtInfoType::CvtInfoType(
    FpilType inst,
    const int plate,
    const SdsId &configId,
    const bool timeSet, 
    const double utTime, 
    StatusType *status) :
    _configMjd(0), _cenMeanRA(0), _cenMeanDec(0),
    _utc(0), _dut(0), /*_spWavelen(0), _agWavelen(0),*/ _instrument(inst)
{
    if (*status != STATUS__OK) return;



    /*
     * Access the telescope model data files.
     */
    FpilModelAccess(_instrument, ConfigFilesDir.c_str(), ExecutableDir,
                    0, 0, plate, &_telModel, status);
    if (*status != STATUS__OK) return;
    
/*
 *  Get details of the telescope model parameters.
 */
    const char *const *ParamNames=0;
    const char *const *ParamDescrs=0;
    const double *ParamDefaults=0;
    unsigned int TelModelNumParams;

    FpilModelParams(_telModel, &TelModelNumParams,
                    &ParamNames, &ParamDescrs, &ParamDefaults, status);
    if (*status != STATUS__OK) return;

    _telModelParams.reserve(TelModelNumParams);
    for (unsigned int i = 0; i < TelModelNumParams; ++i) {
        _telModelParams[i] = ParamDefaults[i];
    }

/*
 *  Get the field center RA and DEC and config MJD
 */
    Arg InputFieldId(configId,"fieldData",status);

    InputFieldId.Get("cenRa",     &_cenMeanRA,  status);
    InputFieldId.Get("cenDec",    &_cenMeanDec, status);
    InputFieldId.Get("configMjd", &_configMjd,  status);

/*
 *  If the field has a wavelength array (it may not) we need to access it.
 */

    _WlensArray.clear();  // Clear the array at this point.
 
    if (*status == STATUS__OK)
    {
        /*
         * We use the SDS array access method to get at the array and then
         * copy it to the WlensArray item (a vector).
         */
        SdsArrayDouble data;
        SdsId Id(InputFieldId, "wlens", status);
        Id.ArrayAccess(&data, status);
        if (*status == STATUS__OK)
        {
            const unsigned long count = data.Size();

            for (unsigned long i = 0; i < count ; ++i)
            {
                _WlensArray.push_back(data[i]);
            }
        }
        else if (*status == SDS__NOITEM)
            *status = STATUS__OK;
    }
/*
 */
    if (timeSet)
        _utc = utTime;
    else
    	_utc = _configMjd;

    fprintf(stderr,
            "Tweaking for utc of %.6f, mean pos %.6f,%.6f\n",
            _utc,
            _cenMeanRA, 
            _cenMeanDec);

}
    

/* 
 *  TweakObject -> Code to tweak the position of one object.
 */
static void TweakObject(
    const unsigned int fibIndex,       // (>) Index of fibre being tweaked 
    // This Input<xxx> set of parameters are from configure via the SDS file
    const double InputRA,              // (>) Object RA position
    const double InputDec,             // (>) Object Dec position
    const char   InputType,            // (>) Object type.

    const CvtInfoType & cvtInfo,  // (>) The position conversion information.
    const short  * const WlenIndex,    // (>) Array of wavelengths for each object

    CvtType * const AgXyPars,        // (!) Autoguider tel model conversion parameters.
    CvtType * const SpecXyParsDef,   // (!) Spectrograph tel model conversion pars
    CvtType * const SpecXyParsAlt,   // (!) Alt spectrograph model conversion pars
    INT32  * const specificWavelensUsed,// (!) count of wavelengths used
    double * const altWavelength,      // (!) wavelength of SpecXyParsAlt model

    double * const OutputX,            // (<) The resulting X position of the fibre
    double * const OutputY,            // (<) The resulting Y position of the fibre
    short  * const /*OutputFibre */,        // (<) Fibre number on spectrograph
    double * const OutputWlen,         // (<) The wavelength used for the conversion

    CvtType ** const modelUsed,       //(<) If non-null, then the address of the model
                                        //    which was used will be written here.  

    StatusType * const status) 
{
    if (*status != STATUS__OK) return;
    *OutputWlen = 0; /* Wavelength will be zero for parked and unused*/


    /*
     * Default value for the modelUsed output argument.
     */
    if (modelUsed)
        *modelUsed = 0;

    const char UnallocatedTargetType = FpilUnallocatedType(instrument);

    if (InputType != UnallocatedTargetType)
    {
        /*
         * fibre is used.
         */

        double appRA,appDec;    

        /*
         * Convert object position from mean 2000 to apparent place
         */
        slaMap( InputRA, InputDec , 0,0,0,0, 2000, 
                cvtInfo.GetUtc(),
                &appRA, &appDec);

        /*
         * Convert RA and DEC to field plate x and y.  Note that
         * if we are dealing with a fiducial, we use the autoguider 
         * parameters instead of the spectrograph parameters, since the
         * wavelengths used may be different.  It is also possible to have
         * different wavelength specs for each fibre.
         * spectrographs may have different wavelengths.
         */ 
        CvtType *model = 0;
        if (InputType == 'F')
        {
            model =  AgXyPars;
            *OutputWlen = cvtInfo.AGWavelen() * 1000.0; /* Cvt um to nm */
        }
        else 
        {
            model = SpecXyParsDef;
            *OutputWlen = cvtInfo.SpecWavelen() * 1000.0;
            /*
             * If we have WlensArray and WlenIndex, then we can check
             * to see if this object has a specific wavelength we should
             * use  instead of the default wavelength.
             */
            if ((cvtInfo.HaveWaveLenArray())&&(WlenIndex))
            {
                if ((WlenIndex[fibIndex] >= 0)&&
                    (WlenIndex[fibIndex] < cvtInfo.NumWaveLens() ))
                {
                    /* This will give us a wavelength in microns */
                    double wavelen = cvtInfo.GetWavelen(WlenIndex[fibIndex]);
                        
                    if (wavelen <= 0) {
                        fprintf(stderr, 
                                "Warning:Wavelength for fibre %d is %.1g microns, which is wrong.  Field was generated by buggy configure - using default value\n",
                               fibIndex+1, wavelen);
                        wavelen = cvtInfo.SpecWavelen();
                    }
                    
                    model = SpecXyParsAlt;
                    (*specificWavelensUsed)++;
                    if (wavelen != *altWavelength)
                    {
                        *altWavelength = wavelen;
                        model->ModelInit(wavelen, status);
                    }
                    *OutputWlen = wavelen * 1000.0;
                }
                else if (WlenIndex[fibIndex] >= cvtInfo.NumWaveLens())
                {
                    fprintf(stderr,"Tweak Warning:Object %d, wlen index %d, max value %d\n",
                            fibIndex+1, WlenIndex[fibIndex], 
                            (cvtInfo.NumWaveLens()));
                }
            }
        }

        /*
         * Convert from RA/Dec to positioner X/Y.
         */
        model->Rd2Pos(appRA, appDec, OutputX, OutputY, status);

        /*
         * If the caller wants to know which model we used, return it.
         */
        if (modelUsed)
            *modelUsed = model;

    }
}



/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * A call used by the tweak opmization code to maintain details on each object.
 */
class ObjectData {
private:
    std::vector<double> _Xx;  // Object X position at each time point
    std::vector<double> _Yy;  // Object Y positoin at each time point.
    std::vector<double> _UT;  // UT Time at each tie point (MJD).
    bool                _IsUsed; // Is this object used.
    bool                _IsGuide;// Is this a guide object.
    unsigned int        _points;// How man time points used.
public:
    ObjectData()  : _IsUsed(false), _IsGuide(false),_points(0) { }
    void SetSize(unsigned points) {
        _points = points;
        _Xx.reserve(points);
        _Yy.reserve(points);
        _UT.reserve(points);
    }
    /*
     * Convert an RA/Dec to X/Y on the field plate for a given point.
     */
    void Rd2Pos(
        unsigned int pointIdx,       /* Point index */
        const CvtType & model,     /* Model for conversion */
        double appRA, double appDec, /* Object RA/Dec */
        StatusType *status) {
    
        if (*status != STATUS__OK) return;

        assert (pointIdx < _points);
        
        double tx,ty;
        model.Rd2Pos(appRA, appDec, &tx, &ty, status);

        if (*status == STATUS__OK)
        {
            _Xx[pointIdx] = tx;
            _Yy[pointIdx] = ty;
            _UT[pointIdx] = model.GetUtc();
            _IsUsed = true;
        }
        else
        {
            _Xx[pointIdx] = 0;
            _Yy[pointIdx] = 0;
            _UT[pointIdx] = 0;
        }
        
    }
    void SetIsGuide()  {
        _IsGuide = true;
    }
    bool IsGuide() const {
        return _IsGuide;
    }
    // This object is not used - fibre is probably parked or broken.
    void SetNotUsed() {
        for (unsigned int i = 0; i < _points ; ++i)
        {
            _Xx[i] = 0;
            _Yy[i] = 0;
            _UT[i] = 0;
        }
        _IsUsed = false;
        _IsGuide = false;

    }
    bool IsUsed() const {
        return _IsUsed;
    }
    double X(unsigned point) const {
        assert(point < _points);
        return _Xx[point];
    }
    double Y(unsigned point) const {
        assert(point < _points);
        return _Yy[point];
    }
    double UT(unsigned point) const {
        assert(point < _points);
        return _UT[point];
    }
    
    /*
     * Rotation the X/Y positon at the specified time point by the given amount in radians.
     */
    void Rotate(unsigned point, double angle) {
        assert(point < _points);
        double result_x = _Xx[point] * cos(angle) - _Yy[point] * sin(angle);
        double result_y = _Xx[point] * sin(angle) + _Yy[point] * cos(angle);
        _Xx[point] = result_x;
        _Yy[point] = result_y;
 
    }
    
};

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

class ObjectTypeCount {
    unsigned int _guides;
    unsigned int _skys;
    unsigned int _progObjs;
    unsigned int _used;
    unsigned int _notUsed;
    unsigned int _total;
public:
    ObjectTypeCount( const std::vector<ObjectData> &objects, 
             const char   * const InputType,
             unsigned long ObjectCount) :
        _guides(0), _skys(0), _progObjs(0), _used(0), _notUsed(0), _total(ObjectCount) {
     
        for (unsigned i = 0; i < ObjectCount ; ++i)
        {
            if (objects[i].IsUsed())
            {
                ++_used;
                if (InputType[i] == 'F')
                    ++_guides;
                else if (InputType[i] == 'S')
                    ++_skys;
                else
                    ++_progObjs;
            }
            else
                ++_notUsed;
        }
        assert((_guides + _skys + _progObjs) == _used);
        assert((_used + _notUsed) == _total);

    }

    unsigned int NumGuides() const {
        return _guides;
    }
    unsigned int NumSkys() const {
        return _skys;
    }
    unsigned int NumProgObj() const {
        return _progObjs;
    }
    unsigned int NumUsed() const {
        return _used;
    }
    unsigned int NumNotUsed() const {
        return _notUsed;
    }
    unsigned int Total() const {
        return _total;
    }
};

static void PrintSummaryOfType(
    FILE * const o,
    const char *typeStr,
    const unsigned int num,
    const unsigned int nTimePoints,
    const std::vector<double> &max,
    const std::vector<double> &sum)
{
    if (num)
    {
        fprintf(o, ",Ave Error %ss", typeStr);
        for (unsigned int timePoint=0; timePoint < nTimePoints; ++timePoint)
        {
            fprintf(o, ",%.1f", sum[timePoint]/num);
        }      
        fprintf(o, ",,%d,%ss\n", num, typeStr);
    }
    else
        fprintf(o, "No fibres with type %s\n", typeStr);

    fprintf(o, ",Max Error %ss", typeStr);
    for (unsigned int timePoint=0; timePoint < nTimePoints; ++timePoint)
    {
        fprintf(o, ",%.1f", max[timePoint]);
    }      
    fprintf(o, ",,%d,%ss\n", num, typeStr);

}

/*
 * Print the summary stats on the object and project objects.
 */
static void PrintSummary(
    FILE * const o,
    const ObjectTypeCount &typeCount,      // Counts of number of objects of different types
    const unsigned int nTimePoints,
    const std::vector<double> &maxObj,
    const std::vector<double> &sumObj,
    const std::vector<double> &maxGui,
    const std::vector<double> &sumGui 

    )
{

    fprintf(o, "Summary\n");

    PrintSummaryOfType(o, "object", typeCount.NumProgObj(), nTimePoints, maxObj, sumObj);
    PrintSummaryOfType(o, "guide",  typeCount.NumGuides(),  nTimePoints, maxGui, sumGui);


    fprintf(o, ",");
    for (unsigned int timePoint=0; timePoint < nTimePoints; ++timePoint)
    {
        fprintf(o, ",");
    }      
    fprintf(o, ",,%d,skys\n", typeCount.NumSkys());
    fprintf(o, ",");
    for (unsigned int timePoint=0; timePoint < nTimePoints; ++timePoint)
    {
        fprintf(o, ",");
    }      
    fprintf(o, ",,%d,fibres used\n", typeCount.NumUsed());

}
#define SQRD(_x_) ((_x_)*(_x_))

class ObjectPosByIndex {  // This is an abstract class - an interface.
public:
    virtual ~ObjectPosByIndex() { }  // 
    virtual double X(unsigned int i) const = 0;
    virtual double Y(unsigned int i) const = 0;
    virtual std::string Type() const = 0;
};

class ObjectNewPosByIndex : public ObjectPosByIndex {
private:
    const double *_x;
    const double *_y;
public:
    ObjectNewPosByIndex(const double * const X, const double *const Y) : _x(X), _y(Y) {
    }
    virtual ~ObjectNewPosByIndex() {}
    virtual double X(unsigned int i) const {
        return _x[i];
    }
    virtual double Y(unsigned int i) const {
        return _y[i];
    }
    virtual std::string Type() const {
        return "Optimized position";
    }
        
};
class ObjectOrigPosByIndex : public ObjectPosByIndex {
private:
    const std::vector<ObjectData> & _objects;
    unsigned int _midTimePoint;
    
public:
    ObjectOrigPosByIndex( const std::vector<ObjectData> &objects,
                          unsigned long nTimePoints) :  _objects(objects), _midTimePoint(nTimePoints/2) {
    }
    virtual ~ObjectOrigPosByIndex() {}
    virtual double X(unsigned int i) const {
        return _objects[i].X(_midTimePoint);
    }
    virtual double Y(unsigned int i) const {
        return _objects[i].Y(_midTimePoint);
    }
    virtual std::string Type() const {
        return "Original position";
    }
        
};

/*
 * Print a set of errors.  We have the array of X and Y and the psoObj item gives the
 * reference item we are determinig errors from.
 */
static void PrintErrorSet(
    FILE * const o,                        // Log file.
    const unsigned long nTimePoints,       // Number of type points
    const double /*startTime*/,                // Start time of tweak process
    const double /*intervalDays*/,             // Tweak optimization interval
    const std::vector<ObjectData> &objects,// Data used to doing tweak optimization
    const char   * const InputType,        // Object type
    const short  * const WlenIndex,        // Array of wavelengths for each object
    const double * const /* X */,                // The resulting X position of the fibre
    const double * const /* Y */,                // The resulting Y position of the fibre

    const ObjectTypeCount &typeCount,      // Counts of number of objects of different types

    const ObjectPosByIndex &posObj         // Provides access to correct referece X/Y

    )
{
    /*
     * Counters for max errors and sums (used for average).
     */
    std::vector<double> maxObj(nTimePoints,0.0);
    std::vector<double> sumObj(nTimePoints,0.0);
    std::vector<double> maxGui(nTimePoints,0.0);
    std::vector<double> sumGui(nTimePoints,0.0);


    unsigned long ObjectCount = typeCount.Total();
    /*
     * Write the header.
     */
    fprintf(o, "RMS Error at each time point (%s)\n", posObj.Type().c_str());
    fprintf(o, "Fibre,Type,");
    for (unsigned i = 1; i <= nTimePoints ; ++i)
        fprintf(o, "Err(%d),", i);
    fprintf(o, "\n");

    /*
     * Work through each object.
     */
    for (unsigned i = 0; i < ObjectCount ; ++i)
    {
        if ((o)&&(objects[i].IsUsed()))
        {

            /*
             * Write the object number and type.
             */
            char targetString[32];
            char targetText[64];
            FpilDecodeTargetType(instrument, InputType[i], ' ', 0, WlenIndex[i], 
                                 targetString, targetText);

            


            fprintf(o, "%d,%s", i+1, targetText);

            /*
             * Work through the time points outputing the diff from the reference position.
             */
            for (unsigned int timePoint=0; timePoint < nTimePoints; ++timePoint)
            {

                double err = sqrt(SQRD(posObj.X(i) - objects[i].X(timePoint)) + 
                                  SQRD(posObj.Y(i)- objects[i].Y(timePoint)));
                fprintf(o, ",%.1f", err);
                    
                /*
                 * Gather summary information on guide and non-sky objects.
                 */
                if (InputType[i] == 'F')
                {
                    if (err > maxGui[timePoint])
                        maxGui[timePoint] = err;
                    sumGui[timePoint]+= err;
                }
                else if (InputType[i] != 'S')
                {
                    if (err > maxObj[timePoint])
                        maxObj[timePoint] = err;
                    sumObj[timePoint]+= err;
                }
            }
            fprintf(o, "\n");
        }
            
    }
    /*
     * Print the sumary information.
     */
    PrintSummary(o, typeCount, nTimePoints, maxObj, sumObj, maxGui, sumGui);


    fprintf(o, "\n\n");

}

/*
 * Print the before and after error information.
 */
static void PrintErrors(
    FILE * const o,                        // Log file.
    const unsigned long ObjectCount,       // Number of objects
    const unsigned long nTimePoints,       // Number of type points
    const double midPoint,                 // Specified mid point of observation
    const double startTime,                // Start time of tweak process
    const double intervalDays,             // Tweak optimization interval
    const std::vector<ObjectData> &objects,// Data used to doing tweak optimization
    const char   * const InputType,        // Object type
    const short  * const WlenIndex,        // Array of wavelengths for each object
    const double * const X,                // The resulting X position of the fibre
    const double * const Y                 // The resulting Y position of the fibre
    )
{


    //
    // Write time points as strings.
    fprintf(o, "\n");
    fprintf(o, "Time Points (UT)\n,");
    double ut = startTime;
    for (unsigned timePoint = 0; timePoint < nTimePoints ; ++timePoint)
    {
        fprintf(o, ",%s", Mjd2Ut_time(ut).c_str());
        ut += intervalDays;
    }
    fprintf(o, "\n");

    ut = startTime;
    fprintf(o, "Time Points (Delta-Secs)\n,");
    for (unsigned timePoint = 0; timePoint < nTimePoints ; ++timePoint)
    {
        double delta = (ut - midPoint)*SPD;
        fprintf(o, ",%.0f", delta);
        ut += intervalDays;
    }
    fprintf(o, "\n\n\n");


    // typeCount will maintain the number of skys, guides, program objects etc 
    ObjectTypeCount typeCount(objects, InputType, ObjectCount);


    // Print the error information on the raw data.  It is the last argument here which gives
    // the reference point.
    PrintErrorSet(o, nTimePoints, startTime, intervalDays, objects, InputType, WlenIndex, X, Y, typeCount,
                  ObjectOrigPosByIndex(objects, nTimePoints) );

    // PRinter the error information on the new data.
    PrintErrorSet(o, nTimePoints, startTime, intervalDays, objects, InputType, WlenIndex, X, Y, typeCount,
                  ObjectNewPosByIndex(X,Y));



}

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

static void TweakOptimizationRotationCorrection(
    FILE *o,
    bool dontRotate, 
    const unsigned ObjectCount,    
    const unsigned int numUsed,    
    const unsigned int numGuide,   
    const unsigned int nTimePoints,
    const char   * const InputType,
    const double * const posX,
    const double * const posY,
    std::vector<ObjectData> * const objects,
    std::vector<unsigned>   * const ObjectIndex,
    StatusType *status);

static void TweakOptimizationOptimize(
    FILE *o,
    const unsigned int numUsed,   
    const unsigned int numGuide,   
    const unsigned int nTimePoints,
    const std::vector<ObjectData> & objects,
    const std::vector<unsigned>   & ObjectIndex,
    double * posX,
    double * posY,
    StatusType *status);


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Optimizing tweak code.  This is run if the obsevation duration is greater then
 * OBS_DUR_OPTIMIZE_TRIGGER seconds.  Its job is to find the optimial location of the
 * fibre through the full observation.
 */
static void TweakWithOptimize(
    FILE * const o,
    const bool dontRotate, 
    const double obsDuration,
    const unsigned ObjectCount,
    const double * const InputRA,
    const double * const InputDec,
    const char   * const InputType,

    const CvtInfoType & cvtInfo, 
    const short  * const WlenIndex,

    CvtType * const AgXyPars,
    CvtType * const SpecXyParsDef,
    CvtType * const SpecXyParsAlt,
    INT32  * const specificWavelensUsed,
    double * const altWavelength,

    double  * const OutputX,
    double  * const OutputY,
    short  * const OutputFibre,
    double * const OutputWlen,
    StatusType * const status)
{

    if (*status != STATUS__OK) return;


#if 0
//#warning "Debug code - fix duration"
//    cvtInfo._XyDetails.obsDuration = 18000;  // 5 hours;
//    cvtInfo._XyDetails.obsDuration = 9000;  // 2.5 hours;
    fprintf(stderr,"Exposure time = %.0f seconds, %.2f hours\n",
            obsDuration,
            obsDuration/(60.0*60.0));
#endif
    /*
     * We need to work out how many time points we should be worrying about.
     *
     * We make sure we have an odd number - so there is a middle point.
     */
    unsigned int nTimePoints = (unsigned)(ceil(obsDuration/OBS_DUR_OPTIMIZE_DELTA));
    if ((nTimePoints%2) == 0)
        nTimePoints += 1;
    fprintf(stderr, 
            "Will break this into %d points\n", nTimePoints);


    /*
     * Work out the starting UT time and the interval in days. SPD gives us
     * the seconds per day.
     *
     * THIS CALCULATION IS DIFFERENT FROM 2dF TWEAK.
     *  In the 2dF Tweak, the startTime is passed in UTC.  The midpoint is then calcuated by adding
     *  ((obsDuration/2)/SPD) to the Start time.;
     */
    double midPoint = cvtInfo.GetUtc(); 
    double startTime = midPoint - (obsDuration/2)/SPD; 
/*
    double startTime = cvtInfo.GetUtc(); 
    double midPoint = startTime + (obsDuration/2)/SPD; 
*/


    double intervalSecs = obsDuration/(double)(nTimePoints-1);
    double intervalDays = intervalSecs/SPD;

    if (o)
    {
        fprintf(o,  "Start time of range,%s,UT,%s,LOCAL,%.8f,MJD\n",
                Mjd2Ut(startTime).c_str(), 
                Mjd2Local(startTime).c_str(),
            startTime);
        fprintf(o, "Mid point of range,%s,UT,%s,LOCAL,%.8f,MJD\n",
                Mjd2Ut(midPoint).c_str(), 
                Mjd2Local(midPoint).c_str(),
                midPoint);
        fprintf(o, "time points,%d\n", nTimePoints);
        fprintf(o, "Spacing,%.1f,seconds,%.7f,days\n",
                intervalSecs, intervalDays);
#ifdef PRINT_RAW
        fprintf(o, "Following data is the raw x/y positions for each fibre in each time point\n\n");
        fprintf(o, "Fibre,");
        for (unsigned i = 1; i <= nTimePoints ; ++i)
            fprintf(o, "X(%d),Y(%d),", i, i);
        
        fprintf(o, "\n");
#endif
        
    }



    /*
     * Allocate some space for each object.  This structure can record the
     * details for each time point for each object.
     */
    std::vector<ObjectData> objects(ObjectCount);


    unsigned int numUsed=0;  /* Number of fibres which are used */
    unsigned int numGuide=0; /* Number of guide fibres */
    /*
     * First we need to run the telecope model for each fibre at each time 
     * point, to get the X and Y position of each object at that time point.
     *
     * These positions are stored in the objects vector - above.
     */
    for (register unsigned int i = 0; 
         (i < ObjectCount) && (*status == STATUS__OK) ; ++i)
    {
        AgXyPars->SetUtc(midPoint);
        SpecXyParsDef->SetUtc(midPoint);
        SpecXyParsAlt->SetUtc(midPoint);

        /*
         * We start off by running the original tweak code, which fills in all the
         * details - it is only OutputX[i] and OutputY[i] that we might later
         * change.  We do set the MoveWarningLimit argument very high so we don't
         * trigger that code.  One thing of note is that we retrive the model
         * that was actualy used (the &model argument) such that we can use it outselves.
         */
        CvtType *model;
        TweakObject(i, 
                    InputRA[i], InputDec[i], InputType[i], 
                    cvtInfo, &WlenIndex[0],
                    AgXyPars, SpecXyParsDef, SpecXyParsAlt,
                    specificWavelensUsed,
                    altWavelength,
                    &OutputX[i], &OutputY[i], 
                    &OutputFibre[i],
                    &OutputWlen[i], &model,
                    status);


        
        
        /*
         * Now we fill out the objects[i] item for this fibre.   Note - if we don't
         * have a model at this point then this fibre is not used.
         */
        objects[i].SetSize(nTimePoints);
        if ((model)&&(*status == STATUS__OK)) {
            ++numUsed;
            if (InputType[i] == 'F') {
                /* Is a guide object */
                objects[i].SetIsGuide();
                ++numGuide;
            }
            double ut = startTime;
            static bool first=true;

#ifdef PRINT_RAW
            if (o)
                fprintf(o, "%d,", i+1);
#endif
            /*
             * Work through the time points.
             */
            for (unsigned int timePoint=0; timePoint < nTimePoints; ++timePoint)
            {
                AgXyPars->SetUtc(ut);
                SpecXyParsDef->SetUtc(ut);
                SpecXyParsAlt->SetUtc(ut);

                double appRA;
                double appDec;
                slaMap( InputRA[i], InputDec[i] , 0,0,0,0, 2000, ut, &appRA, &appDec);


                objects[i].Rd2Pos(timePoint, *model, appRA, appDec, status);

#ifdef PRINT_RAW
                if (o)
                    fprintf(o, "%.1f,%.1f,", objects[i].X(timePoint), objects[i].Y(timePoint));
#endif


                /* Move to the next time point */
                ut+= intervalDays;
            }
            /*
             * Some first time debugging.
             */
            if (first)
            {
                first=false;
                ut -= intervalDays;
                fprintf(stderr,
                        "End point time of range will be UT %s, Local %s\n",
                        Mjd2Ut(ut).c_str(), 
                        Mjd2Local(ut).c_str());
            }
#ifdef PRINT_RAW
            if (o)
                fprintf(o, "\n");
#endif
        }
        else
        {
            objects[i].SetNotUsed();
        }
        
    }
    /* Restore these */
    AgXyPars->SetUtc(startTime);
    SpecXyParsDef->SetUtc(startTime);
    SpecXyParsAlt->SetUtc(startTime);


    if (*status != STATUS__OK)
    {
        fprintf(stderr, "TWEAK with optimize failed in main loop\n");
        return;
    }
    fprintf(stderr,
            "There are %d fibres used in this configuration, %d are guides\n", 
            numUsed, numGuide);

    if (nTimePoints <= 1) 
    {
        fprintf(stderr, "Actually have nothing to do - there was only one point");
        return;
    }

    /*
     * At this point, we have the X and Y for each fibre at each time point.  We 
     * want to find the best position of each fibre.  This is independent for 
     * each fibre, except that we first remove any rotation between the sets of 
     * fibres - we rotate each set to minimize the rotation between that set and 
     * the center set.  This also setes up the ObjectIndex vector -> which gives 
     * us the index in each of the object arrays for each object i such 
     * that  0 <= i < numUsed. 
     */
    std::vector<unsigned>   ObjectIndex(numUsed);
    TweakOptimizationRotationCorrection(o, dontRotate, ObjectCount, numUsed, numGuide, 
                                        nTimePoints, InputType, OutputX, OutputY, 
                                        &objects, &ObjectIndex, status);

    /*
     * So now we have the position of each object at each time point with any 
     * rotation between the time points removed that could be removed by the
     * guide removed.  
     *
     * Find the best position of that object.
     */
    TweakOptimizationOptimize(o, numUsed, numGuide, nTimePoints,
                              objects, ObjectIndex, 
                              OutputX, OutputY, status);

#define PRINT_PROCESSED
    if (o)
    {
        fprintf(o, "\n");
#ifdef PRINT_PROCESSED
        fprintf(o, "Processed data follows - %spoints and final value\n",
                (dontRotate ? "" : "rotated "));
        fprintf(o, "Fibre,");
        for (unsigned i = 1; i <= nTimePoints ; ++i)
        {
            fprintf(o, "X(%d),Y(%d), ", i, i);
        }
        
        fprintf(o, "X(o),Y(o)\n");

        for (unsigned i = 0; i < ObjectCount ; ++i)
        {
            
            if ((o)&&(objects[i].IsUsed()))
            {
                fprintf(o, "%d,", i+1);
                for (unsigned int timePoint=0; timePoint < nTimePoints; ++timePoint)
                {
                    fprintf(o, "%.1f,%.1f,", objects[i].X(timePoint), objects[i].Y(timePoint));
                    
                }
                fprintf(o, "%.1f,%.1f\n", OutputX[i], OutputY[i]);
                
                
            
            }
            
        }
#endif
        /*
         * Print the error infomation to the file.
         */
        PrintErrors(o, ObjectCount, nTimePoints, midPoint, startTime, intervalDays, 
                    objects, InputType, WlenIndex, OutputX, OutputY);

    }


}

/*
 * Work out the rotation between two sets of time Points.
 */
static double GetRotation(
    unsigned n,
    double D0[][2],
    double D1[][2],
    StatusType *status)
{
    double coeffs[6];
    int j;
    if (*status != STATUS__OK)  return 0;

    slaFitxy(4, n, D0, D1, coeffs, &j);

    assert(j == 0);
    
    double xs;
    double ys;
    double xz;
    double yz;
    double perp;
    double rotation;
    slaDcmpf(coeffs, &xz, &yz, &xs, &ys, &perp, &rotation);
 

    return rotation;
}


/*
 * The main point of this function is to optimize the object positions across the
 * time periods to remove any correction which can be done by the 2dF field plate
 * rotator.
 *
 * Such a correction is made using the guide fibres only - so we work out the rotation
 * of the guide fibres between each time point against the center of the field.  We then
 * remove this rotation from each X/Y position for each fibre at each time point.
 */
static void TweakOptimizationRotationCorrection(
    FILE *o,                        /* Log file file pointer */
    const bool dontRotate,          /* Write details - but don't do the rotation correction */
    const unsigned ObjectCount,     /* Number of objects in Output and Input arrays */
    const unsigned int numUsed,     /* Number of fibres which are used */
    const unsigned int numGuide,    /* Number of guide fibres */
    const unsigned int nTimePoints, /* Number of time points */
    const char   * const InputType, /* Type of each object */
    const double * const posX, /* The nominal X position of the object */
    const double * const  posY, /* The nominal Y position of the object */
    std::vector<ObjectData> * const objects, /* The object array - data for each point */
    std::vector<unsigned>   * const ObjectIndex, /* Index in Input/Output arrays
                                                    for each object from 0 to numUsed-1 */
    StatusType *status)
{
    if (*status != STATUS__OK) return;
    /*
     * At this point, we have the X and Y for each fibre at each time point in "objects".
     */
    std::vector<unsigned>   GuideIndex(numGuide); /*  Index in Input/Output arrays for each
                                                      guide object from 0 to numGuide-1 */

    /*
     * We need a variable declared like this
     *     double midPoint[numGuide][2] 
     * For passing to GetRotation() and hence to slaFitXy().    But Ansi C++ does
     * not allow variable sized array declarations - we get warnings in GCC over it.
     * I would prefer not to dynamically allocate the array - since cleaning it up
     * will be necessary - so I make use of the fact that what I want is nothing more
     * then numGuide*2 doubles in contiguous memory. The C++ STL says that
     * for a vector, the address of the first element can be treated as a standard
     * C style array, hence - if I set up a vector (midPointData) of the right size, then
     * take the address of the first element - I have what I want.
     */
    std::vector<double> midPointData(numGuide*2);
    double (*midPoint)[2] = (double (*)[2])(&midPointData[0]);/* Mid point X/Y of each guide object,
                                                      form suitable for slaFitXy call */
    /*
     * Initialise for the rotation calcs and fill out ObjectIndex and GuideIndex
     * vectors.
     */
    for (unsigned int i=0, j=0, k=0; i < ObjectCount; ++i)
    {
        if ((*objects)[i].IsUsed())
        {

            // j should never exceed numUsed - but just check.
            assert(j < numUsed);

            /*
             * Remember the object index
             *
             */
            (*ObjectIndex)[j] = i;
            ++j;
            if (InputType[i] == 'F') 
            {
                /*
                 * This is a guide object.
                 *
                 * Grab the mid point X in an easier to use structure.
                 *
                 * And remember the index.
                 */
                midPoint[k][0]        = posX[i];
                midPoint[k][1]        = posY[i];
                GuideIndex[k] = i;
                ++k;
            }

        }
    }

    /*
     * If the rotation result varies by more then 1/10 of a degree, then we are likely
     * doing something wrong (check this).
     *
     * We must have at least 3 guide bundles to do this calcuation.  Otherwise, slaFitXy()
     * can't sort out rotation from the other linear model effects.
     */
    const double oneTenthDegree = DD2R/10.0; /* One tenth of a degree in radians */
    const unsigned int MIN_GUIDES = 3;
    if (numGuide >= MIN_GUIDES)
    {
        std::vector<double>     Rotation(nTimePoints); /* The rotation at each time point 
                                                          Guide objects */
        bool rotationErr = false;  /* Do we have a problem? */

        /*
         * Space to store data at a time point.  
         */
        std::vector<double> TimePointData(numGuide*2);

        /*
         * Loop through each time point.
         */
        for (unsigned int timePoint=0; timePoint < nTimePoints; ++timePoint)
        {
            /*
             *  See comments for midPoint above for what and why I am doing this.
             */
            double (*P)[2] = (double (*)[2])(&TimePointData[0]);
            /*
             * Get the X and Y at this timePoint in the time slice for each object
             * we are working with.
             */
            for (unsigned k = 0; k < numGuide ; ++k)
            {
                assert(GuideIndex[k] < ObjectCount);
                P[k][0] = (*objects)[GuideIndex[k]].X(timePoint);
                P[k][1] = (*objects)[GuideIndex[k]].Y(timePoint);
            }
            /*
             * So now we have two sets off values - the mid time set and the
             * P set corresponding to time point "timePoint".  Each have "numGuide" 
             * entries.  We want to determine the  rotation between these two sets.
             */
            Rotation[timePoint] =  GetRotation(numGuide, P, midPoint, status);

            if (fabs(Rotation[timePoint]) > oneTenthDegree)
            {
                /*
                 * If we get here, then slaFitXy() has failed to fit correctly - it is
                 * likely that although we have, in theory, enought points, they are
                 * not spread nicely around the field.  We must bail out.
                 */
                fprintf(stderr, "Warning:Field rotation calculation result too large - %.1f degrees\n",
                       (Rotation[timePoint]/DD2R));
                fprintf(stderr, "   - during long duration exposure optimization.\n");
                fprintf(stderr, "  It is likely your guide fibres are clustered on one side of the plate\n");
                fprintf(stderr, "  This will make it hard to correct for field rotation errors.\n");
                rotationErr = true;
                break;
            }


        }

        /*
         * Now we can do the rotation - but only if we we did not get an error.
         */
        if (!rotationErr)
        {
            /*
             * Rotate the position of each object such that any rotation is removed - the
             * base rotation is the center position - which should ensure the maximum
             * rotator range is available.
             */
            if (o)
            {
                if (dontRotate)
                    fprintf(o, "Rotation,NOT APPLIED\n");
                else
                    fprintf(o, "Rotation,APPLIED\n");
                
                fprintf(o, "Rotation for each time point,");
            }
            for (unsigned int timePoint=0; timePoint < nTimePoints; ++timePoint)
            {
                if (!dontRotate)
                {
                    for (unsigned j = 0; j < numUsed ; ++j)
                    {
                        assert((*ObjectIndex)[j] < ObjectCount);
                        (*objects)[(*ObjectIndex)[j]].Rotate(timePoint,  Rotation[timePoint]);
                        
                    }
                }
                if (o)
                    fprintf(o, "%.7f, ,", Rotation[timePoint]);


            }   
            if (o)
                fprintf(o, "\n\n");
        }
    }
    else
    {
        fprintf(stderr, "Warning:Insufficent guide bundles in allocation (%d, need %d) for rotation calulation\n",
               numGuide, MIN_GUIDES);
        fprintf(stderr, "   - during long duration exposure optimization.\n");
        fprintf(stderr, "  This will make it hard to correct for field rotation errors.\n");
    }
        

}

/*
 *  A structure used to collect statistics.
 */
struct StatsStruct {
    int _maxErrorFibre;
    double _maxError;
    double _sum;
    double _average;

    StatsStruct() : _maxErrorFibre(-1), _maxError(0), _sum(0), _average(0) {
    }
    void AddPoint(
        const unsigned int fibIndex,
        const double origX,
        const double origY,
        const double newX,
        const double newY,
              double * const localDeltaMax=0,
              double * const thisDelta=0) {

        
        double deltaX =  origX - newX;
        double deltaY =  origY - newY;
        double delta = sqrt(deltaX * deltaX + deltaY * deltaY);
        if (delta > _maxError)
        {
            _maxError      = delta;
            _maxErrorFibre = fibIndex+1;
        }
        _sum += delta;
        if (thisDelta)
            *thisDelta = delta;

        if (localDeltaMax && (delta > *localDeltaMax))
            *localDeltaMax = delta;

    }
    void SetCount(unsigned int i) {
        if (i == 0)
            _average = 0;
        else
            _average = _sum/(double)i;
    }
};

/*
 * Do the position optimization across the exposure duration
 *
 * Any rotation between the time points that can be removed has been.
 *
 * 
 */
static void TweakOptimizationOptimize(
    FILE *o,
    const unsigned int numUsed,     /* Number of fibres which are used */
    const unsigned int numGuide,    /* Number of fibres which are guide fibre */
    const unsigned int nTimePoints, /* Number of time points */
    const std::vector<ObjectData> & objects, /* The object array - data for each point */
    const std::vector<unsigned>   & ObjectIndex, /* Index in Input/Output arrays
                                                    for each object from 0 to numUsed-1 */
          double * const posX, /* The nominal X position of the object */
          double * const posY, /* The nominal Y position of the object */
    StatusType *status)
{
    if (*status != STATUS__OK) return;

    /*
     * Variables used to collect statitical information - some of which goes
     * into the FITS header.
     */
    StatsStruct origObj;   /* Status on original positions of object fibres */
    StatsStruct origGui;   /* Status on original positions of guide fibres */
    StatsStruct newObj;    /* Status on new positions of object fibres */
    StatsStruct newGui;    /* Status on new positions of guide fibres */
    double origMaxErrSumGui=0; /* Sums for maximum error of each time */
    double origMaxErrSumObj=0;
    double newMaxErrSumGui=0;
    double newMaxErrSumObj=0;
    /*
     * Say I find the min and max of x and y for each object across the time 
     * points.   The point I want is the centre in x and y.
     *  
     * Work through all the used objects.
     */
    for (unsigned j = 0; j < numUsed ; ++j)
    {
        double minX =  100000000;
        double maxX = -100000000;
        double minY =  100000000;
        double maxY = -100000000;
        unsigned int i = ObjectIndex[j];

        double fibreMaxErrOrig = 0;
        double fibreMaxErrNew = 0;

        /*
         * Determine the box that will enclose all the points for this
         * fibre.
         */
        for (unsigned int timePoint=0; timePoint < nTimePoints; ++timePoint)
        {
            double delta;
            double x =  objects[i].X(timePoint);
            double y =  objects[i].Y(timePoint);


            if (x < minX)
                minX = x;
            if (x > maxX)
                maxX = x;
            if (y < minY)
                minY = y;
            if (y > maxY)
                maxY = y;

            /*
             * Add this entry to the stats for the original positions..
             */
            if (objects[i].IsGuide())
                origGui.AddPoint(i, posX[i], posY[i], x, y, &fibreMaxErrOrig,&delta);
            else
                origObj.AddPoint(i, posX[i], posY[i], x, y, &fibreMaxErrOrig,&delta);


        }
        if (objects[i].IsGuide())
            origMaxErrSumGui += fibreMaxErrOrig;
        else
            origMaxErrSumObj += fibreMaxErrOrig;


        /*
         * Now work out the optimal position of this fibre.
         */
        double rangeX = (maxX - minX) / 2;
        double rangeY = (maxY - minY) / 2;
        
        posX[i] = rangeX + minX;
        posY[i] = rangeY + minY;


        /*
         * Collect stats on the new positions.
         */
        for (unsigned int timePoint=0; timePoint < nTimePoints; ++timePoint)
        {
            double delta;
            double x =  objects[i].X(timePoint);
            double y =  objects[i].Y(timePoint);
            
            /*
             * Add this point to the statis.
             */
            if (objects[i].IsGuide())
                newGui.AddPoint(i, posX[i], posY[i], x, y, &fibreMaxErrNew, &delta);
            else
                newObj.AddPoint(i, posX[i], posY[i], x, y, &fibreMaxErrNew, &delta);


        }
        if (objects[i].IsGuide())
            newMaxErrSumGui += fibreMaxErrNew;
        else
            newMaxErrSumObj += fibreMaxErrNew;


    }
    /*
     * Set the count needed for the statistics calculations.
     */
    origObj.SetCount((numUsed - numGuide)*nTimePoints);
    newObj.SetCount((numUsed - numGuide)*nTimePoints);
    origGui.SetCount(numGuide*nTimePoints);
    newGui.SetCount(numGuide*nTimePoints);

    /*
     * Output statistics.
     */
    fprintf(o, "Optimization Statistics,,Fibre Num,Max Error,Average Error,Average Max\n");
    if ((numUsed-numGuide) > 0)
    {
        fprintf(o, "Object\n,Original Position,");
        fprintf(o,
                "%3d,%.1f,%.1f,%.1f\n",
                origObj._maxErrorFibre, 
                origObj._maxError,origObj._average,
                origMaxErrSumObj/(numUsed-numGuide));
        fprintf(o, ",New Position,");
        fprintf(o,
                "%3d,%.1f,%.1f,%.1f\n",
                newObj._maxErrorFibre, newObj._maxError,  newObj._average,
                newMaxErrSumObj/(numUsed-numGuide));

        
    }

    if (numGuide > 0)
    {

        fprintf(o, "Guide\n,Original Position,");
        fprintf(o,
                "%3d,%5.1f,%5.1f,%5.1f\n",
                origGui._maxErrorFibre, 
                origGui._maxError,origGui._average,
                origMaxErrSumGui/numGuide);
        fprintf(o, ",New Position,");
        fprintf(o,
                "%3d,%5.1f,%5.1f,%5.1f\n",
                newGui._maxErrorFibre, newGui._maxError,  newGui._average,
                newMaxErrSumGui/numGuide);

    }
    fprintf(o, "\n");
    
    
}






/*
 *  ***************************************************************************
 *  This should be in a separate file as it is also used by configure, but
 *  the version in configure_subs.c does not work well with this program, due
 *  to other dependencies on configure.
 */

/* -------------------------------------------------------------------------- */

/*           I m p  Z  E x e  D i r  -  P r e l i m i n a r i e s
 *
 *  To provide the configuration program with the means to determine which
 *  directory it is being run from, the utility routine ImpZExeDir() has been
 *  extracted from the code for the IMP message system used by DRAMA. This
 *  was the most conveniently packaged version of such a routine.
 *  
 *  The ImpZExeDir function and its dependencies have been copied
 *  verbatim from the impz_unix.c file. This short section sets the various
 *  constants used by IMP that are normally defined in IMP include files
 *  and which are used by the ImpZExeDir() code.
 */
 
#define IMPZ_PROTOTYPES_OK 1
#define IMP_FALSE 0
#define IMP_TRUE 1

/* -------------------------------------------------------------------------- */


/*                         I m p  Z  E x e  D i r
 *
 *  Routine:
 *     ImpZExeDir
 *
 *  Function:
 *     Gives the name of the directory from which the current program was run.
 *
 *  Description:
 *     In systems where individual tasks are started by loading executable
 *     programs from disk, it is often useful for a program to be able to
 *     locate the directory from which it was loaded. This enables it to
 *     find related files, which are often held in the same directory, 
 *     in a straightforward way that doesn not require the use of logical
 *     names (VMS), environment variables (UNIX), or some other such scheme.
 *
 *     The interface to this routine requires the zeroth argument passed to
 *     the program when it was started up. This is needed by the UNIX 
 *     implementation, although not by the VMS version. It returns the full
 *     pathname or file specification of the directory containing the
 *     program currently being run. This will be formatted such that it
 *     can be prepended to a filename to get a valid full file specification.
 *     (So, for example, under UNIX it will end with a '/', under VMS it will
 *     end with a ']'.) If the system is such that the concept does not make
 *     sense (embedded systems like VxWorks), or where it simply cannot be
 *     implemented), a 'Known' flag argument will be returned set false.
 *    
 *  Implementation details for UNIX:
 *     Most of the work is done by the routine findx(), originally written
 *     by Greg Limes (limes@eng.sun.com) which is appended to this file. It
 *     works by looking through the directories specified by the PATH
 *     environment variable looking for the program name passed as the
 *     zeroth argument (although there's a number of additional wrinkles).
 */

static int findx(const char *cmd, char *cwd, char *dir, const char **pgm, char **run,
    char *path);
static int resolve(             /* get link resolution name */
    const char *indir, const char *cmd, char *dir, char **run);

void ImpZExeDir(
    const char *Arg0,                 /* Zeroth argument to program */
    char *Directory,            /* Name of 'exe' directory */
    int LDir,                   /* Number of chars in Directory */
    int *Known)                 /* True if Directory is valid */

/*  The following is my (KS) wrap-up of findx(), producing the UNIX version of
 *  ImpZExeDir(). This is a  routine which a) sets up a call to findx()
 *  to get the name of the 'exe' directory, b) manipulates the result
 *  slightly for aesthetic reasons and to make sure the result ends with a
 *  '/' character, and c) returns it in the nul-terminated character string
 *  Directory. The manipulation of the string returned by findx() is because
 *  the findx() algorithm tends to produce strings such as '/home/aatssb/ks/.'
 *  or '/home/aatssb/ks/work/../scrap/.' which are valid but messy if the
 *  result is ever to be output. (This is a small modification of the Figaro
 *  Fortran-callable utility GEN_EXEDIR, which was canibalised to produce
 *  this code, which is probably over fussy for this IMP application.)
 */

{
    /*  Local variables  */

    int Back;                    /* Number of '/' chars to work back through */
    char Char;                   /* Character in directory name */
    char CurrentDir[MAXPATHLEN] = ""; 
                                 /* Work space used for current directory */
    int Dots;                    /* Dot count between '/' characters */
    char ExeDir[MAXPATHLEN] = "";/* Receives the exe directory from findx() */
    int Ichar;                   /* Index into ExeDir */
    int Ioutchar;                /* Index into returned directory string */
    char LastChar;               /* Last character written into Directory */

    /*  Functions used  */


    /*  Set up the call to findx() to get the directory name  */

    *Known = IMP_FALSE;
    Ioutchar = 0;
    if (findx(Arg0, CurrentDir, ExeDir, (const char **) 0,
            (char **) 0, getenv("PATH")) == 0) {

        /*  The call to findx() worked, so we now have a directory string. We
         *  combine the copying of it into the Directory string and the tidying
         *  up. We copy it character by character, keeping a count of the number
         *  of dots after each '/', clearing the count on an ordinary character.
         *  If we come to a new '/' or the end of the string with a positive 
         *  dot count, we skip back by a number of '/' chars equal to the 
         *  number of dots we found.  This has the effect of turning a '/./'
         *  sequence into a '/' and a 'xx/../yy' sequence into 'yy'.
         */

        Char = '\0';
        Ichar = 0;
        Dots = 0;
        do {
            Back = 0;
            LastChar = Char;
            Char = ExeDir[Ichar++];
            Directory[Ioutchar] = Char;
            if (Char == '\0') {
                Back = Dots;
                Ioutchar--;
            } else if (Char == '/') {
                Back = Dots;
                Dots = 0;
            } else if (Char == '.') {
                Dots++;
            } else {
                Dots = 0;
            }
            if (Back)
                LastChar = '/';
            while (Back) {
                if (Directory[--Ioutchar] == '/')
                    Back--;
                if (Ioutchar <= 0)
                    break;
            }
            if (Ioutchar < LDir)
                Ioutchar++;
        } while (Char);
        if (LastChar != '/')
            Directory[Ioutchar++] = '/';
        if (Ioutchar < LDir) {
            Directory[Ioutchar] = '\0';
            *Known = IMP_TRUE;
        }
    }

}




/* -------------------------------------------------------------------------- */

/*                               F i n d x
 *
 *  The following is the body of the code of the routine findx() as written
 *  by Greg Lines. I wanted to include it to keep this file self-contained,
 *  but I didn't want to dive into it making minor changes and so propagate
 *  a 'slightly' different version if I could help it. I have added a
 *  'static' to the declaration of findx() itself in order to hide the name
 *  and keep it a local routine. Other than that this is the code as supplied.
 *  The original release includes a man page which documents the routine fully,
 *  but I have not included that. Note that the main() routine declared in the
 *  code is not normally compiled.
 *
 *  Having said that in the original version of this routine, I have now made
 *  a more significant modification to the findx() code. I have introduced the
 *  variables dp and dstr, making it copy the individual path elements into
 *  dstr rather than work on them in place - the original code modified the
 *  PATH string it was passed, and so could not be called more than once. I
 *  have also removed the now unused variable 's'.  Subsequently, I have also
 *  eliminated the BSD function rindex() to allow this to run under SOLARIS,
 *  by using the following Rindex() function and using it instead. I have
 *  also changed the name of the symlink variable to symblink to prevent lint
 *  complaining about it hiding the name of a system function. Moreover, I
 *  have removed the TAB characters from the original code, since the TAB
 *  settings being used didn't match those in the rest of this file and so
 *  made the code layout very odd indeed.
 */


const char *RindexC(const char *s, char c)
{
    const char *ptr = s;
    const char *cptr = (const char *) 0;
    while (*ptr) {
        if (*ptr == c)
            cptr = ptr;
        ptr++;
    }
    return (cptr);
}

char *Rindex(char *s, char c)
{
    char *ptr = s;
    char *cptr = (char *) 0;
    while (*ptr) {
        if (*ptr == c)
            cptr = ptr;
        ptr++;
    }
    return (cptr);
}



/*  ------------------------------------------------------------------------- */

#include <errno.h>
#include <sys/param.h>

#ifndef   X_OK
#define   X_OK   1              /* for "access" */
#endif

#ifndef   MAXPATHLEN
#define   MAXPATHLEN   1024
#endif

#ifndef   ENAMETOOLONG
#define   ENAMETOOLONG   EINVAL
#endif

static char sanityPath[]=".";

static int findx(const char *cmd, char *cwd, char *dir, const char **pgm,
    char **run, char *path)
{
    int rv = 0;
    char dstr[MAXPATHLEN], *dp;
    const char *f;
    if (!cmd || !*cmd || !cwd || !dir) {
        errno = EINVAL;         /* stupid arguments! */
        return -1;
    }
    if (!path || !*path)        /* missing or null path */
        path = sanityPath;             /* assume sanity */

    if (*cwd != '/')
        if (!(getcwd(cwd, MAXPATHLEN))) {
            return -1;          /* cant get working directory */
        }

    f = RindexC(cmd, '/');
    if (pgm)                    /* user wants program name */
        *pgm = f ? f + 1 : cmd;

    if (dir) {                  /* user wants program directory */
        rv = -1;
        if (*cmd == '/')        /* absname given */
            rv = resolve("", cmd + 1, dir, run);
        else if (f)             /* relname given */
            rv = resolve(cwd, cmd, dir, run);
        else if ((f = path)) {  /* from searchpath */
            rv = -1;
            errno = ENOENT;     /* errno gets this if path empty */
            while (*f && (rv < 0)) {
                dp = dstr;
                while (*f && (*f != ':'))
                    *dp++ = *f++;
                if (*f) {
                    f++;
                }
                *dp = '\0';
                if (dstr[0] == '/')
                    rv = resolve(dstr, cmd, dir, run);
                else {
                    char abuf[MAXPATHLEN];
                    sprintf(abuf, "%s/%s", cwd, dstr);
                    rv = resolve(abuf, cmd, dir, run);
                }
            }
        }
    }
    return rv;
}

/*
 * resolve - check for specified file in specified directory. sets up dir,
 * following symlinks. returns zero for success, or -1 for error (with
 * errno set properly)
 */
static int resolve(
    const char *indir, /* search directory */
    const char *cmd,                  /* search for name */
    char *dir,                       /* directory buffer */
    char **run)
{                               /* resultion name ptr ptr (within dir) */
    char *p;

    errno = ENAMETOOLONG;
    if ((int) (strlen(indir) + strlen(cmd) + 2) > MAXPATHLEN) {
        return -1;
    }

    sprintf(dir, "%s/%s", indir, cmd);
    if (access(dir, X_OK) < 0) {
        return -1;              /* not an executable program */
    }


    p = Rindex(dir, '/');
    *p++ = '\0';
    if (run)                    /* user wants resolution name */
        *run = p;
    return 0;
}

/*
         1         2         3         4         5         6         7         8
12345678901234567890123456789012345678901234567890123456789012345678901234567890
*/





