#                M O V I E D A T A _ G U I . T C L
#
# Script: moviedata_gui.tcl
#
# Description:
#     Used by moviedata.tcl to implement the main GUI window.
#
#     Entry point is "ContructGui"
#
#     Data file should have been read by moviedata_reader.tcl into the
#     variable the name of which is passed in.
#
# Author: Tony Farrell, AAO
#
# "@(#) $Id: ACMM:aaoConfigure/src/moviedata_gui.tcl,v 12.7 04-Nov-2013 11:28:48+11 tjf $"
#
# History:
#     22-Oct-2008 - TJF - Intial experimental version.
#     07-Aug-2012 - TJF - Get TimeBarScrolled working on Mac Cocoa version
#                          of which where the scrollbar adds a moveto argument.
#                          
#
# "@(#) $Id: ACMM:aaoConfigure/src/moviedata_gui.tcl,v 12.7 04-Nov-2013 11:28:48+11 tjf $"
##
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

proc ConstructGui { w data } {

    # Access variable.
    upvar #0 $data d

    # Frames in main window.
    frame $w.header -border 2 -relief ridge
    frame $w.summary -border 2 -relief ridge
    frame $w.timePointSel -border 2 -relief ridge
    frame $w.errorOrig -border 2 -relief ridge
    frame $w.errorOptim -border 2 -relief ridge

    pack $w.header $w.summary $w.errorOrig $w.errorOptim  $w.timePointSel -fill x -expand 1 -padx 3 -pady 3


    # Contruct each part of the main window.

    ConstructHeader          $w.header       $data
    ConstructSummary         $w.summary      $data
    ConstructError Original  $w.errorOrig    $data
    ConstructError Optimized $w.errorOptim   $data

    # The time point selector is a scroll bar
    set selector [ConstructTimePntSel       $w.timePointSel $data]
    $selector configure -command "TimeBarScrolled $selector $data"

    set centre [expr int($d(NumTimePoints)/2)+1]
    SetTimePointDisplayed $data $centre

    # Contruct the mimic window.
    ConstructMimic $data
}

#
# Invoked when the Time point selection scroll bar is scrolled.
#   Note - sometimes, we see "cmd=moveto" and sometimes we don't
#   have the command argument at all.  This is something to do
#   with the scrollbar implementation. The moveto is seen on 
#   Macs running the Cocoa version of wish, where as the
#   lack if it seems to be seen on X11.  If the "moveto" item is
#   not seen, then pos is an integer position number.  If moveto 
#   is seen, then pos is a propotion of the scrollbar.
#
#   We can also see "scroll", in which case there is one more argument,
#   which is often pages and pos is -1 or 1.
#
proc TimeBarScrolled { w data cmd {pos ""} args } {
    #puts stderr "TimeBarScrolled $data $cmd $pos \"$args\""
    upvar #0 $data d

    #puts stderr "Number of data points:$d(NumTimePoints)"

    # If pos is blank, then the scrollbar supplied only one argument
    #  the position, and it is an integer number.
    if { $pos == "" } {
        set pos $cmd
    } elseif { $cmd == "moveto" } {
        # Caclulate an integer position number.
        set pos [expr int(($pos*$d(NumTimePoints) + 0.5))]
        #puts "Caclulated pos = $pos"
    } elseif { ($cmd == "scroll") && ($args == "page") } {
	#puts stderr "TimeBarScrolled scroll command \"$pos\" \"$args\""
	if { "$pos" == "-1" } {
	    #puts stderr "TimeBarScrolled moving to negative end"
	    set pos 0
	} elseif { "$pos" == "1" } {
	    #puts stderr "TimeBarScrolled moving to postivie end"
	    set pos $d(NumTimePoints)
        } else {
	    #puts stderr "TimeBarScrolled no change in scrollbar"
	    return ""
        }
    } else {
	puts stderr "TimeBarScrolled unexpected scrollbar command \"$cmd\""
	return ""
    }

    # Work out the next time point and the position within the scroll bar.
    set timePoint [expr $pos + 1]
    if { $timePoint > $d(NumTimePoints) } {
        set timePoint  $d(NumTimePoints)
        set pos [expr $d(NumTimePoints) - 1]
    } elseif { $pos < 0 } { 
        set pos 0 
        set timePoint 1
    } 

    # Set the new scroll bar positoin.
    $w set $d(NumTimePoints) $d(NumTimePoints) $pos $pos

    # Display the time point.
    SetTimePointDisplayed $data $timePoint
}


#
# Display the time point - set the various variables used to display this point.
#
proc SetTimePointDisplayed { data point } {
    #puts "SetTimePointDisplayed $point"
    upvar #0 $data d
    
    set d(CurrentTimePointNum) $point
 
    set d(CurrentTimePoint) $d(TimePoints,$point)

    # The rotation value is in radians - convert to milli-degrees for display.
    set d(CurrentTimePointRotation) "[format %.1f [expr $d(Rotation,$point)*180.0/3.14159265*1000.0]] mDeg"

    set d(Original-AveErrorObjects) $d(Original,AveErrObj,$point)
    set d(Original-MaxErrorObjects) $d(Original,MaxErrObj,$point)
    set d(Original-AveErrorGuides)  $d(Original,AveErrGui,$point)
    set d(Original-MaxErrorGuides)  $d(Original,MaxErrGui,$point)

    set d(Optimized-AveErrorObjects) $d(Optimized,AveErrObj,$point)
    set d(Optimized-MaxErrorObjects) $d(Optimized,MaxErrObj,$point)
    set d(Optimized-AveErrorGuides)  $d(Optimized,AveErrGui,$point)
    set d(Optimized-MaxErrorGuides)  $d(Optimized,MaxErrGui,$point)

    set delta $d(TimePointsDelta,$point)
    set delta [expr int($delta/60.0)]
    if { $delta > 0 } {
        set delta "+$delta"
    }
    
    if { $delta == 0 } {
        set d(CurrentTimePointDelta) "center"        
    } else {
        set d(CurrentTimePointDelta) "$delta min"        
    }
    
    # Update the mimic.
    after idle MimicUpdate $data

}

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
# Construct the header window - details from the file header.
#
proc ConstructHeader { w data } {

    
    upvar #0 $data d

    # We use the ConstructGridFromLists function to do this - which makes it 
    # relatively easy, we just create two lists, the first as the labels, and 
    #the second has the entry in the  $data array to get the value from.
    #

    set labels [list Instrument "Config File" "Observation Duration (secs)" "Center Time (UT)" "Num Time Points" "Spacing (Secs)" "Start Time (UT)" "Number Allocated Progam Objects" "Number Allocated Skys" "Number Allocated Guide Stars" "Total Allocated Fibres" "Rotation"]
    set values [list Instrument InputFile Duration ObsTimeUT NumTimePoints SpacingSeconds StartUT numObjects numSkys numGuide numTotal RotationMode]

    ConstructGridFromLists $w $data $labels $values

}
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
# Contruct summary area of mimic.
#
proc ConstructSummary { w data } {

    upvar #0 $data d

    set row 0
    label $w.title -text "Optimization Statistics" 

    grid $w.title -column 0 -row $row -columnspan 5 -sticky ew
    incr row

    label $w.labFibNum   -text "Fibre Number" -relief ridge -border 2
    label $w.labMaxError -text "Maximum Error" -relief ridge -border 2
    label $w.labAveError -text "Average Error"  -relief ridge -border 2
    label $w.labMaxAveErr -text "Average Max Error"  -relief ridge -border 2

    grid $w.labFibNum    -column 1 -row $row
    grid $w.labMaxError  -column 2 -row $row
    grid $w.labAveError  -column 3 -row $row
    grid $w.labMaxAveErr -column 4 -row $row
    incr row


    label $w.oLab -text "Project Object/Sky Object" -anchor w
    grid $w.oLab -column 0 -row $row  -sticky ew


    incr row
    
    label $w.oOrigRowLab -text "Original Position:" -anchor e
    grid $w.oOrigRowLab -column 0 -row $row  -sticky ew

    ConstructSummaryLine $w.oOrig $data "ErrStats,Object,OrigPos"  $row

    incr row

    label $w.oOptimRowLab -text "Optimized Position:" -anchor e
    grid $w.oOptimRowLab -column 0 -row $row  -sticky ew

    ConstructSummaryLine $w.oOptim $data "ErrStats,Object,NewPos"  $row

    incr row



    label $w.gLab -text "Guide Objects" -anchor w
    grid $w.gLab -column 0 -row $row  -sticky ew
    incr row

    label $w.gOrigRowLab -text "Original Position:" -anchor e
    grid $w.gOrigRowLab -column 0 -row $row  -sticky ew
    ConstructSummaryLine $w.gOrig $data "ErrStats,Guide,OrigPos"  $row

    incr row

    


    label $w.gOptimRowLab -text "Optimized Position:" -anchor e
    grid $w.gOptimRowLab -column 0 -row $row  -sticky ew


    ConstructSummaryLine $w.gOptim $data "ErrStats,Guide,NewPos"  $row

    incr row




}


#
# Contruct the summary lines
#
proc ConstructSummaryLine {  wp data item  row } {
    upvar #0 $data d

    label ${wp}_FibNum -text $d($item,Fibre)   -relief ridge -border 2 
    label ${wp}_MaxErr -text $d($item,Max)     -relief ridge -border 2 
    label ${wp}_AveErr -text $d($item,Ave)     -relief ridge -border 2 
    label ${wp}_AveMax -text $d($item,AveMax)  -relief ridge -border 2 


    grid ${wp}_FibNum -column 1 -row $row -sticky ew
    grid ${wp}_MaxErr -column 2 -row $row -sticky ew
    grid ${wp}_AveErr -column 3 -row $row -sticky ew
    grid ${wp}_AveMax -column 4 -row $row -sticky ew
}
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
# Contruct our time point selector - a scroll var
#
proc ConstructTimePntSel { w data } {
    upvar #0 $data d

    set d(CurrentTimePoint) "xxx"
    set d(CurrentTimePointNum) 1
    set d(CurrentTimePointRotation) "yyy"

    set variable1 ${data}(CurrentTimePoint)
    set variable2 ${data}(CurrentTimePointDelta)
    set variable3 ${data}(CurrentTimePointNum)
    set variable4 ${data}(CurrentTimePointRotation)
    set variable5 ${data}(RotationMode)

    frame $w.timePoint
    pack $w.timePoint -side top -expand 1 -fill x

    label $w.timePoint.lab   -text "Time Point" -width 15
    label $w.timePoint.valueT -textvariable $variable1 -relief ridge -border 2 -width 20
    label $w.timePoint.valueD -textvariable $variable2 -relief ridge -border 2 -width 10
    label $w.timePoint.valueN -textvariable $variable3 -relief ridge -border 2 -width 10

 

    grid $w.timePoint.lab $w.timePoint.valueT $w.timePoint.valueD $w.timePoint.valueN  -row 0 -stick ew


    label $w.timePoint.rotationLab  -text "Rotation" -width 15
    label $w.timePoint.rotationVal  -textvariable $variable4 -relief ridge -border 2 -width 10
    label $w.timePoint.rotationMode -textvariable $variable5 -relief ridge -border 2 -width 20

    grid $w.timePoint.rotationLab -sticky ew -row 1 -column 1
    grid $w.timePoint.rotationVal -sticky ew -row 1 -column 2
    grid $w.timePoint.rotationMode -sticky ew -row 1 -column 3



    scrollbar $w.timePointScrollBar -orient horizontal 
#    grid  $w.timePointScrollBar -column 0 -columnspan 2 -row $row -sticky ew

    pack $w.timePointScrollBar -side top -expand 1 -fill x

    set pos [expr int($d(NumTimePoints)/2)]

    $w.timePointScrollBar set $d(NumTimePoints) $d(NumTimePoints) $pos $pos

    return $w.timePointScrollBar 

}


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
# Contruct a widget to display the errors etc. at a given time point.
#
proc ConstructError { type w data } {

    upvar #0 $data d

    set row 0

    #puts "Contruct Error $type $w $data"

    label $w.i${type}_Label -text "$type data at specified time point"
    grid $w.i${type}_Label -column 0 -row $row -sticky ew -columnspan 2
    incr row
    

    set d(${type}-AveErrorObjects) "x1"
    label $w.i${type}_AveErrObjLab   -text "Average Error Objects" 
    set variable ${data}(${type}-AveErrorObjects)
    label $w.i${type}_AveErrObjValue -textvariable $variable  -relief ridge -border 2 -width 15


    set d(${type}-MaxErrorObjects) "x2"
    label $w.i${type}_MaxErrObjLab   -text "Max Error Objects"
    set variable ${data}(${type}-MaxErrorObjects)
    label $w.i${type}_MaxErrObjValue -textvariable $variable  -relief ridge -border 2 -width 15


    grid $w.i${type}_AveErrObjLab   -column 0 -row $row -sticky ew 
    grid $w.i${type}_AveErrObjValue -column 1 -row $row -sticky ew 
    grid $w.i${type}_MaxErrObjLab   -column 2 -row $row -sticky ew 
    grid $w.i${type}_MaxErrObjValue -column 3 -row $row -sticky ew 
    incr row

    set d(${type}-AveErrorGuides) "x3"
    label $w.i${type}_AveErrGuiLab   -text "Average Error Guides"
    set variable ${data}(${type}-AveErrorGuides)
    label $w.i${type}_AveErrGuiValue -textvariable $variable  -relief ridge -border 2

    set d(${type}-MaxErrorGuides) "x4"
    label $w.i${type}_MaxErrGuiLab   -text "Max Error Guides"
    set variable ${data}(${type}-MaxErrorGuides)
    label $w.i${type}_MaxErrGuiValue -textvariable $variable  -relief ridge -border 2


    grid $w.i${type}_AveErrGuiLab   -column 0 -row $row -sticky ew 
    grid $w.i${type}_AveErrGuiValue -column 1 -row $row -sticky ew 
    grid $w.i${type}_MaxErrGuiLab   -column 2 -row $row -sticky ew 
    grid $w.i${type}_MaxErrGuiValue -column 3 -row $row -sticky ew 
    incr row

    

}

# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
# Utility function - contruct a grid from lists of details on them
#
# w -> parent frame window.
# data -> data variable
# labels -> List of labels for each window.
# values -> The entry in the $data variable to find the value.
#
proc ConstructGridFromLists { w data labels values } {
    upvar #0 $data d

    #puts "Data = $data"
    #puts "Data(Instrument) = $d(Instrument)"
    

    set i 0
    foreach l $labels {
        label $w.glab_$i -text "$l :" -anchor e

        set val_index [lindex $values $i]
        label $w.gval_$i -text $d($val_index) -anchor w

        grid $w.glab_$i -column 0 -row $i -sticky ew -padx 3m -pady 1m
        grid $w.gval_$i -column 1 -row $i -sticky ew -padx 3m -pady 1m

        incr i
        
    }
}
