#         m o v i e d a t a _ r e a d e r . t c l
#
# Script: moviedata_reader.tcl
#
# Description:
#     Used by moviedata.tcl to implement read the moviedata CVS file.
#
#     Entry point is "ReadDataFile"
#
#
# Author: Tony Farrell, AAO
#
# "@(#) $Id: ACMM:aaoConfigure/src/moviedata_reader.tcl,v 12.7 04-Nov-2013 11:28:48+11 tjf $"
#
# History:
#     22-Oct-2008 - TJF - Intial experimental version.
#
# "@(#) $Id: ACMM:aaoConfigure/src/moviedata_reader.tcl,v 12.7 04-Nov-2013 11:28:48+11 tjf $"
#
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
# Code from http://wiki.tcl.tk/2215 - example of how to split
# CVS file lines into lists.  From "Andy Goth", beleive to be public
# domain.
#
# Some modifications by TJF.
#

proc csv_split {line} {
    #puts "\$line = \"$line\""
    # Process each input character.
    set result [list]
    set beg 0
    while {$beg < [string length $line]} {
       if {[string index $line $beg] eq "\""} {
	  incr beg
	  set quote false
	  set word ""
          #puts "\$line = \"$line\""
          #puts "\$beg  = \"$beg\""

           # The first version was the original - the {*} gave problems
           #foreach char [list {*}[split [string range $line $beg end] ""] ""] {  }
           foreach char [list [split [string range $line $beg end] ""] ""] {
             # Search forward for the closing quote, one character at a time.
	     incr beg
	     if {$quote} {
                 if {($char == ",") || ($char == "") } {
		   # Quote followed by comma or end-of-line indicates the end of
		   # the word.
		   break
		} elseif {$char eq "\""} {
		   # Pair of quotes is valid.
		   append word $char
		} else {
		   # No other characters can legally follow quote.  I think.
		   error "extra characters after close-quote"
		}
		set quote false
	     } elseif {$char eq ""} {
		# End-of-line inside quotes indicates embedded newline.
		error "embedded newlines not supported"
	     } elseif {$char eq "\""} {
		# Treat the next character specially.
		set quote true
	     } else {
		# All other characters pass through directly.
		append word $char
	     }
	  }
	  lappend result $word
       } else {
	  # Use all characters up to the comma or line ending.
	  regexp -start $beg {.*?(?=,|$)} $line word
	  lappend result $word
	  set beg [expr {$beg + [string length $word] + 1}]
       }
    }

    # If the line ends in a comma, add a blank word to the result.
    if {[string index $line end] eq ","} {
       lappend result ""
    }

    # Done.  Return the result list.
    return $result
}

#
#Examples:
#
# % csv_split {a b c,"d e f,",}
# {a b c} {d e f,} {}
# % csv_split {a b c,"d e f,",g}
# {a b c} {d e f,} g
# % csv_split {a b c,"d e""f,",g}
# {a b c} {d e"f,} g
# % csv_split {a b c,"d e""f,","g"}
# {a b c} {d e"f,} g
# % csv_split {a b c,"d e""f,",""}
# {a b c} {d e"f,} {}
# % csv_split {a b c,"d e""f,","",""}
# {a b c} {d e"f,} {} {}
# % csv_split {a b c,"d e""f,","","",","}
# {a b c} {d e"f,} {} {} ,
# % csv_split {a b c,"d e""f,","","",",",,,}
# {a b c} {d e"f,} {} {} , {} {} {}
#
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

#
# We are given a list. We want to convert that list into array items.
#
#  arrayName => Name of array we will use.  Accessed with "upvar".
#  arrayItem => First component if array item name.  Can be null string.
#  list      => List to convert.
#  lfirst    => Index of first item in list to convert.  Defaults to 0.
#  afirst    => First array item number.  Defaults to 1
#  incr      => List increment.  Default 1.
#
#
# $arrayName(${arrayItem}<n>) is set to [lindex $list <n>-1+$lfirst]
#
#


proc ListToArray { arrayName arrayItem list {lfirst 0 } {afirst 1} {lincr 1} } {

    upvar $arrayName array

    set index $lfirst;

    set listLength [llength $list]
    for { set i $afirst } { $index < $listLength } { incr i } {

        set array(${arrayItem}$i) [lindex $list $index]
        incr index $lincr
    }
}


#
# We are given a list contains pairs of data. We want to convert that list
#  into array items.
#
#  arrayName => Name of array we will use.  Accessed with "upvar".
#  arrayItem1 => First component if array item name for first item in each pair. 
#  arrayItem1 => First component if array item name for second item in each pair. 
#  list      => List to convert.
#  lfirst    => Index of first item in list to convert.  Defaults to 0.
#  afirst    => First array item number.  Defaults to 1
#
#
#
proc ListToArrayPair { arrayName arrayItem1 arrayItem2 list {lfirst 0 } {afirst 1}  } {

    upvar $arrayName array

    set index $lfirst;

    set listLength [llength $list]

    #puts "ListToArrayPair \"$list\""

    for { set i $afirst } { $index < $listLength } { incr i } {

        #puts "Setting array(${arrayItem1}$i) to [lindex $list $index]"
        
        set array(${arrayItem1}$i) [lindex $list $index]
        incr index 
        set array(${arrayItem2}$i) [lindex $list $index]
        incr index 
    }
}
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
# Read the specified CVS data file.
#
#
proc ReadDataFile { file data } {
    set f [open $file]
    

    ReadBasicHeader $f $data
    
    ReadDataHeader $f $data
    
    ReadRotation $f $data
    
    ReadStats $f $data

    ReadTimePositions $f $data

    ReadTimePoints $f $data    

    ReadError $f $data Original 
    ReadError $f $data Optimized 

    close $f
}


# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
#
# Read the next line in the file and return it as a list, presuming it is a CSV
# data file.
#
proc GetLineAsList { file } {
    set line [gets $file]
    if [eof $file] {
        error "Unexpected end-of-file whilst reading data file"
    }
    set line [csv_split $line]
    return $line
}

#
# Extract one item from the next line in the file and put into the data array.
#
# file => File to read
# data => Array variable
# item => Item within array variable
# which => Index of which item from line (first is 0)
#
proc ExtractOneItem { file data item which} {

    upvar #0 $data d
    set line [GetLineAsList $file]
    set d($item) [lindex $line $which]
}


# Extract multiple items from the next line in the file and put into the data array
#
# file => File to read
# data => Array variable
# first => Index of first item from line (first is 0)
# items...  => Names of items to extract.
#
proc ExtractItems { file data first args} {

    upvar #0 $data d

    set nitems [llength $args]
    set line [GetLineAsList $file]
    set index $first

    for { set i 0 } { $i < $nitems } { incr i } {

        set name     [lindex $args $i]
        set d($name) [lindex $line $index]
        incr index

    }
}

#
# Read the basic header of the data file.
#
proc ReadBasicHeader { file data } {
    upvar #0 $data d

    set line [GetLineAsList $file]

    set fileType [lindex $line 0]
    if { $fileType != "Object Movie Data File for instrument" } {
        puts stderr "Input file does not appear to be a object movie data file"
        error "Invalid input file"
    }
    set d(Instrument) [lindex $line 1]
    #puts "Movie data file for instrument \"$d(Instrument)\""


    # Skip line
    gets $file

    ExtractOneItem $file $data "FileVersion" 1
    if { $d(FileVersion) != 1 } {
        error "Unsupport movie data file version number - $d(FileVersion)"
    }
    #puts "Version num $d(FileVersion)"

    ExtractOneItem $file $data "MaxFieldRadius" 1
    #puts "Maximum field radius is $d(MaxFieldRadius)"

    gets $file
    
    ExtractOneItem $file $data "InputFile" 1
    #puts "SDS Configuration file was $d(InputFile)"

 

}
#
# Read the header of the data area of the file.
#
proc ReadDataHeader { file data } {
    upvar #0 $data d
    ExtractOneItem $file $data "ObsTimeUT"    2
    ExtractOneItem $file $data "ObsTimeLocal" 2
    ExtractOneItem $file $data "Duration"     1
    
    gets $file
    ExtractItems $file $data 1 StartUT junk StartLocal junk StartMjd
    ExtractItems $file $data 1 MidUT   junk MidLocal   junk MidMjd
    ExtractOneItem $file $data "NumTimePoints" 1
    ExtractItems $file $data 1 SpacingSeconds junk SpacingDays

    ExtractOneItem $file $data "RotationMode" 1

}

#
# Read the rotation.
#
proc ReadRotation { file data } {
    upvar #0 $data d

    set line [GetLineAsList $file]

    #puts "Rotation line is - \"$line\""

    ListToArray d "Rotation," $line 1 1 2
    
    gets $file
}

#
# Read the statstics
#
proc ReadStats { file data } {
    upvar #0 $data d

    gets $file  ;# Skip header line
    gets $file  ;# Skip Object line
    
    ExtractItems $file $data 2 "ErrStats,Object,OrigPos,Fibre"  "ErrStats,Object,OrigPos,Max" "ErrStats,Object,OrigPos,Ave" "ErrStats,Object,OrigPos,AveMax"
    

    #puts $d(ErrStats,Object,OrigPos,AveMax)

    ExtractItems $file $data 2 "ErrStats,Object,NewPos,Fibre"  "ErrStats,Object,NewPos,Max" "ErrStats,Object,NewPos,Ave" "ErrStats,Object,NewPos,AveMax"

    #puts $d(ErrStats,Object,NewPos,AveMax)


    gets $file ;# Skip Guider line

    ExtractItems $file $data 2 "ErrStats,Guide,OrigPos,Fibre"  "ErrStats,Guide,OrigPos,Max" "ErrStats,Guide,OrigPos,Ave" "ErrStats,Guide,OrigPos,AveMax"
    
    #puts $d(ErrStats,Guide,OrigPos,AveMax)

    ExtractItems $file $data 2 "ErrStats,Guide,NewPos,Fibre"  "ErrStats,Guide,NewPos,Max" "ErrStats,Guide,NewPos,Ave" "ErrStats,Guide,NewPos,AveMax"

    #puts $d(ErrStats,Guide,NewPos,AveMax)

    gets $file 

}

#
# Read details on the time points themselves.
#
proc ReadTimePoints { file data } {
    upvar #0 $data d
    gets $file 
    set line [GetLineAsList $file]
    #puts "ReadTimePoints - line = $line"
    ListToArray d "TimePoints," $line 2

    gets $file
    set line [GetLineAsList $file]
    ListToArray d "TimePointsDelta," $line 2
    
}

#
# Read error infomration (type indicates from Reference or Optimized position)
#
proc ReadError { file data type } {
    upvar #0 $data d

    # Currently - we are only reading the summary information.

    set i 0
    for { set line "" } { $line != "Summary" } { set line [GetLineAsList $file] } {
        incr i
    }
    #puts "Read $i lines before hiting point of interest"
    

    # Read average error object.
    set line [GetLineAsList $file]
    ListToArray d "$type,AveErrObj," $line 2

    # Read max error object.
    set line [GetLineAsList $file]
    ListToArray d "$type,MaxErrObj," $line 2
   

    # Read average error guide.
    set line [GetLineAsList $file]
    ListToArray d "$type,AveErrGui," $line 2

    # Read max error object.
    set line [GetLineAsList $file]
    ListToArray d "$type,MaxErrGui," $line 2

    #puts "$d($type,AveErrGui,1)"
    #puts "$d($type,AveErrGui,35)"


    # The number of object and guide fibres can be found in these arrays, at the index two past
    # the number of guides
    set count [expr $d(NumTimePoints)+2]
    set d(numObjects) "$d($type,AveErrObj,$count)"
    set d(numGuide)   "$d($type,AveErrGui,$count)"

    # Next two line ave the number of skys and total number of fibres
    incr count 
    ExtractOneItem $file $data numSkys  $count
    ExtractOneItem $file $data numTotalx $count

    if { $d(numTotal) != $d(numTotalx) } {
        error "Stated number of fibres is different then listed number of fibres"
    }
    #puts "Num of skys is $d(numSkys)"
    #puts "Num of fibres is $d(numTotal)"
}



#
# Read the positions at each time point.
#
proc ReadTimePositions { file data } {

    upvar #0 $data d
    
    set line [gets $file]
    while { $line == "" } {
        set line [gets $file]
    }
    if { [string range $line 0 24] != "Processed data follows - " } {
        error "Object position information header not found when expected - got \"$line\""
    }
    # Skip the next line.
    gets $file

    set line [gets $file]
    set i 0

    # Find the location of the fibre data for the reference position - the center time point
    set refTimePoint [expr $d(NumTimePoints)/2+1]
    # Find the location of the fibre data for the optimal positoin of the fibre - max time point + 1
    set optTimePoint [expr $d(NumTimePoints)+1]

    while { $line != "" } {
        #puts "Read fibre info $line"
        incr i
        
        set line [csv_split $line]

        set d(Fibre,$i,Num) [lindex $line 0]

        # Read the data lines
        ListToArrayPair d "Fibre,$i,X," "Fibre,$i,Y," $line 1

        #puts "[array names d Fibre*]"

        set d(Fibre,$i,X,RefPos) $d(Fibre,$i,X,$refTimePoint)
        set d(Fibre,$i,Y,RefPos) $d(Fibre,$i,Y,$refTimePoint)
        set d(Fibre,$i,X,OptimPos) $d(Fibre,$i,X,$optTimePoint)
        set d(Fibre,$i,Y,OptimPos) $d(Fibre,$i,Y,$optTimePoint)

        set line [gets $file]

        #puts "Fibre $i  $d(Fibre,$i,X,RefPos) ,$d(Fibre,$i,Y,RefPos)   $d(Fibre,$i,X,OptimPos),$d(Fibre,$i,Y,OptimPos)"
        
    }
    set d(numTotal) $i
    #puts "ReadTimePositions read $i fibres"
    #puts "Last line contains $line"

    
}
