/*+		 S E T _ F I B R E _ S T A T E

 * Program name:
       set_fibre_state

 * Function:
       Program to set the fibre state in to appropiate modes

 * Synopis:
       set_fibre_state <consts_file> shuf [0|1|both| 
       set_fibre_state <consts_file> file [0|1|both| <file1> [<file2>]
       set_fibre_state <const_file> restore <file1>

 * Description:
       Reads the specified tdFconstants.sds file and sets
       the fibre state. 

       Details for the first command format:

        Argument 1 => The tdFconstants.sds file to use.
        Argument 2 => Operation.  One of

             shuf     ->  Disable fibres appropiately for normal charge 
                           shuffling.
             restore  ->  Restores fibre states as per the specified file.
             file     ->  Disable fibres listed in the specified file.

        Argument 3, restore mode => The file written as 
                          tdFconstants_changes.txt by a previous run of this
                          program

        Argument 3, shuff/file mode =>  The plate on which to make the 
                          changes. 0 or 1 or both or


        Argument 4 => For "file" mode.  A file containing a list of fibres.
                          This is just a simple list of numbers, any line
                          starting with # is ignored.  This is for one plate
                          only.  If you have specified plate 0 or 1, it is
                          for that plate. Otherwise,  it will be for plate 0.

        Argument 5 => For "file"mode.  A file containing a list of fibres.
                          This is just a simple list of numbers, any line
                          starting with # is ignored.  This is for plate 1
                          only and only if the plate spec was "both". 


       The file spec_fibres.txt must be present in the current directory the
       an environment variable SPEC_FIBRES_FILE must point to this file.  This
       is the file normally stored in /instsoft/2dF/spec_fibres.txt and defines
       the relationship between spectrograph slit number (rank) and pivot number.


 * Language:
      C++

 * Support: Tony Farrell, AAO

 *-

 * @(#) $Id: ACMM:aaoConfigure/src/set_fibre_state.C,v 12.7 04-Nov-2013 11:28:47+11 tjf $

 * History:
     01-Dec-2006 - TJF - Original version
     22-Jul-2008 - TJF - Modified to support reading spec_fibres.txt in order to work
                         with fibre rank rather then pivot number, since this is what
                         we really want to work with.
     23-Sep-2009 - TJF - Include assert.h.  Needed for MacOsX Snow Leopord.
     14-Jan-2010 - TJF - Include string.h.  Needed for GCC 4.5.2.
     21-Aug-2012 -  KS - Added const to rcsId line to avoid warnings.


 *  Copyright (c) Anglo-Australian Telescope Board, 2006, 2008.
    Not to be used for commercial purposes without AATB permission.
    No warranty of any form unless subject to a specific agreement.

 */
static const char *rcsId="@(#) $Id: ACMM:aaoConfigure/src/set_fibre_state.C,v 12.7 04-Nov-2013 11:28:47+11 tjf $";
static void *use_rcsId = (0 ? (void *)(&use_rcsId) : (void *) &rcsId);

/*
 *  Include files
 */
#include <assert.h>
#include <stdlib.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include "sds.h"
#include <string>
#include <unistd.h>
#include <fstream>
#include <iostream>
#include <limits>
#include <vector>

#include "tdFfieldDims.h"  /* for NUM_PIVOTS_2DF */

/* 
 * Constant definitions
 *
 */
#define APP_ERROR 1        /* Used to set inherited status bad */
#define FIELD0 "field0"    /* Name of field strucutre for plate 0 */
#define FIELD1 "field1"    /* Name of field structure for plate 1 */
#define RESTORE_FILE "tdFconstants_changes.txt"  /* Details of changes written here */
#define CONSTS_FILE "tdFconstants400.sds"        /* 2dF constants file */
#define CONSTS_BCK  "tdFconstants400.sds.bck"    /* Backup file name */


const unsigned SPECTRUM_NUM_MAX=400;  /* Maximum spectrum number */


const char *specFibresFile = "./spec_fibres.txt";
const char *specFibresEnvVar = "SPEC_FIBRES_FILE";



/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Helper classes
 *
 * 
 * class DeleteOnExitOFStream
 *
 * This is a sub-class of ofstream which will delete the file
 * on exit if the appropiate flag has been set.  Otherwise, it works
 * the same as ofstream (other then having simpler constructor and
 * open call).  By default - we delete the file.
 */
class DeleteOnExitOFStream : public std::ofstream {
private:
    bool deleteOnExit;
    std::string fileName;
public:
    DeleteOnExitOFStream() : std::ofstream(), deleteOnExit(true)  { 
    }
    // Constructor with open via file name - must remember the name.
    DeleteOnExitOFStream(const char *name)
        : std::ofstream(name), deleteOnExit(true), fileName(name)  { 
    }
    // Invoke this to indicate the object should be deleted on exit.
    void ClearDeleteOnExit() {
        deleteOnExit = false;
    }
    // Open call.  Do normal open but remember the file name.
    void open(const char *name) {
        std::ofstream::open(name);
        fileName = name;
    }
    // Close - if the file is open close it.  If it is to be deleted,
    //   delete it.
    void close() {
        if (this->is_open())
            std::ofstream::close();

        if ((deleteOnExit)&&(!fileName.empty()))
        {
            unlink(fileName.c_str());
            deleteOnExit = false;  // Don't do this agai.
        }
    }

    virtual ~DeleteOnExitOFStream() {
        // Ensure the file is closed.
        if (this->is_open())
            this->close();
    }
};  /* class DeleteOnExitOFStream */

/*
 * class InUseFlags
 *
 * A class used to read/write the inUse flags of the tdFconstants400
 * structure
 */
class InUseFlags {
private : 
    unsigned field;   /* Which field */
    SdsId inUseID;    /* The SDS ID of the structure item the data is in */
    short *data;      /* The data itself is here */


    /* Assignment and Copy operators, made private to avoid misuse */
    void  operator=(const InUseFlags& ); /* Assignment */
    InUseFlags(const InUseFlags&);       /* Copy */


public :
    /*
     * Constructor 
     *
     *  fileID    The SDS ID of the tdFconstants400 file structure.  
     *  field     The field (plate) number - 0 or 1.
     *  status    Modified status.
     *  
     */
    InUseFlags(const SdsId &fileID, unsigned field, StatusType *status);

    /*
     * Destructor.  Ensure the data is flushed.
     */
    virtual ~InUseFlags() { 
        StatusType ignore = STATUS__OK;
        if (inUseID)
            inUseID.Flush(&ignore);
    }

    /*
     * Get the value of an in-use flag.
     */
    bool Get(const unsigned num) const {
        assert ((data)&&(num < NUM_PIVOTS_2DF));
        return (bool)data[num];
    }
    /*
     * Set an in-use flag value
     */
    void SetInUse(const unsigned num) {
        assert ((data)&&(num < NUM_PIVOTS_2DF));
        data[num] = (short)1;
    }
    /*
     * Clear an in-use flag value.
     */
    void ClearInUse(const unsigned num) {
        assert ((data)&&(num < NUM_PIVOTS_2DF));
        data[num] = (short)0;
    }
}; /* class InUseFlags */


/*
 * Construct an InUseFlags item.
 */
InUseFlags::InUseFlags(
    SdsId const &fileID, 
    unsigned field, 
    StatusType *status) : data(0)
{
    /*
     * Valid field.
     */
    if ((field != 0)&&(field != 1))
    {
        fprintf(stderr, "Invalid field number - %d.\n", field);
        *status = SDS__INVITEM;
        return;
    }
    /*
     * We need to find the field (plate) structure and then find
     * the inUse array within it.
     */
    SdsId tmpFieldID(fileID, (field == 0 ? FIELD0 : FIELD1), status);
    if (*status != STATUS__OK)
    {
        fprintf(stderr, "Failed to find \"%s\" item within file structure.\n",
                (field == 0 ? FIELD0 : FIELD1));
        return;
    }
    SdsId tmpInUseID(tmpFieldID, "inUse", status);
    if (*status != STATUS__OK)
    {
        fprintf(stderr, 
                "Failed to find \"inUse\" item within field structure.\n");
        return;
    }
    /*
     * Check the dimensions of the inUse structure.  IF this is wrong, then
     * we probably have a file for the wrong instrument.
     */
    long ndims;
    unsigned long dims[7];

    tmpInUseID.Dims(&ndims, dims, status);
    if (ndims != 1)
    {
        fprintf(stderr, "The inUse field has %ld instead of 1 dimension.\n",
                ndims);
        *status = SDS__INVDIMS;
        return;
    }
    if (dims[0] != NUM_PIVOTS_2DF)
    {
        fprintf(stderr, 
                "The inUse field has a dimension of %ld instead of %d.\n",
                dims[0], NUM_PIVOTS_2DF);
        *status = SDS__INVDIMS;
        return;
    }
    /*
     * Should be ok now - access the data.
     */
    tmpInUseID.Pointer((void **)&data, status);
    if (*status != STATUS__OK)
    {
        fprintf(stderr, "Failed to get pointer to data.\n");
        return;
    }
    /*
     * At this point the structure is valid and we have a pointer to it.
     * We just need to keep the inUseID item about.
     */
    inUseID.ShallowCopy(tmpInUseID, true);
}


/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Class which parses and makes available the spec_fibres.txt file.
 */

class specFibres {
public:
    /*
     * Constructor - reads the file.
     */
    specFibres();

    /*
     * Given a 2dF spectrum number (1 to SPECTRUM_NUM_MAX), return the 2dF pivot index  
     *   (0 to (NUM_PIVOTS_2DF-1)).
     */
    int pivotIndex(
        const unsigned plate,
        const unsigned spectrumNum) const;
    /*
     * Given a 2dF pivot number (1 to NUM_PIVOTS_2DF), return the spectrum number 
     */
    int spectrumNum(
        const unsigned plate,
        const unsigned pivotNum) const;

private:
    /* Assignment and copy operators - made private to avoid misuse */
    specFibres & operator=(const void *); /* Assignment */
    specFibres(const specFibres&);        /* Copy */


    void ReadFileIntoArrays(FILE *f);
    static bool ReadLine(FILE *file, int *line, int *plate,
                         unsigned *pivot, int *rank);

    /* 
     * 2dF pivot index for each spectrum number
     * Indexed by spectrum number - 0 to (SPECTRUM_NUM_MAX+1)
     * -1 indicates invalid item 
     */  
    std::vector<int> pivotIndexPlate0;  
    std::vector<int> pivotIndexPlate1;  
};

specFibres::specFibres()
{
    std::string fileName;
    const char *envVarVal = getenv(specFibresEnvVar);
    if (envVarVal)
        fileName = envVarVal;
    else
        fileName = specFibresFile;


    FILE *f = fopen(fileName.c_str(), "r");
    if (!f)
    {
        fprintf(stderr,"Failed to open file \"%s\" (fibres to rank file)\n",
                fileName.c_str());
        fprintf(stderr,"  %s\n",
                strerror(errno));
        fprintf(stderr," Default name is \"%s\"\n", specFibresFile);
        fprintf(stderr," Can also be set via \"%s\" environment variable\n",
                specFibresEnvVar );
        exit(1);
    }
    /*
     * Initialise the arrays.
     */
    pivotIndexPlate0.reserve(SPECTRUM_NUM_MAX);
    pivotIndexPlate0.insert(pivotIndexPlate0.begin(), SPECTRUM_NUM_MAX, -1);


    pivotIndexPlate1.reserve(SPECTRUM_NUM_MAX);
    pivotIndexPlate1.insert(pivotIndexPlate1.begin(), SPECTRUM_NUM_MAX, -1);


    ReadFileIntoArrays(f);


    fclose (f);

    printf("Have read file \"%s\" to get fibre ordering\n",
           fileName.c_str());
}

void specFibres::ReadFileIntoArrays(
    FILE *f)
{
    int line  = 0;   
    int plate;
    unsigned pivot;
    int rank;

    while (ReadLine(f,&line,&plate,&pivot,&rank))
    {
/*
 *      We have read a valid line from the file.  
 *      The rank indicates the order on the CCD, i.e. the spectrum number.
 */
        if (((plate == 0)||(plate == 1))&&(pivot > 0)&&(pivot <= NUM_PIVOTS_2DF))
        {
            int fibre = rank-1;
            
            if ((fibre >= 0)&&(fibre < (int)(SPECTRUM_NUM_MAX)))
            {
                if (plate == 0)
                    pivotIndexPlate0[fibre] = pivot-1;
                else
                    pivotIndexPlate1[fibre] = pivot-1;

            }
            else
            {
                fprintf(stderr, 
                        "Fibre's data base file line %d, gave an invalid rank (%d), must be less then %d",
                        line, rank, NUM_PIVOTS_2DF);
            }
        }
        else if ((plate != 0)&&(plate != 1))
        {
            fprintf(stderr, 
                    "Fibre's data base file line %d, gave an invalid plate - %d, must be 0 or 1",
                    line, plate);

        }
        else
        {
            fprintf(stderr, 
                    "Fibre's data base file line %d, gave an invalid pivot - %d, must be 1 to %d",
                    line, pivot, NUM_PIVOTS_2DF);
        }
    }
/*
    for (i = 0 ; i < FIBRES_PER_SPECT ; ++i)
    {
        if (!details[i].ok)
            fprintf(stderr,"Fibre %d not ok\n",i+1);
    }
 */

 


}

bool specFibres::ReadLine(
    FILE *file, 
    int *line, 
    int *plate,
    unsigned *pivot, 
    int *rank)
{
    char buffer[500];
    while (fgets(buffer,sizeof(buffer),file))
    {
        ++(*line);
/*
 *      Skip any lines which start with "Plate", which indicates a table header
 *      line
 */
        if (strncmp(buffer,"Plate",5) == 0)
            continue;

        int count;
        int platePos;  /* Position of retractor within plate */
        int retractorNum;
        char condition[20];
        if ((count = sscanf(buffer,"%d %d %d %u %d %s",
                            plate,&platePos,&retractorNum,pivot,rank,
                            condition)) != 6)
            /* Change for ErsRep error later */
            fprintf(stderr,"Spec Fibres File:Failed to format line %d, count %d\n",
                    *line,count);

        else if ((*plate != 0)&&(*plate != 1))
            fprintf(stderr,"Spec Fibres File:Invalid Plate %d number on line %d\n", 
                    *plate, *line);

        else if ((*pivot <= 0)||(*pivot > NUM_PIVOTS_2DF))
            fprintf(stderr,"SpecFibresFile:Invalid Pivot number %d on line %d\n",
                    *pivot, *line);

        else if ((*rank <= 0)||(*rank > int(SPECTRUM_NUM_MAX)))
            fprintf(stderr,"SpecFibresFile:Invalid RANK %d on line %d\n",
                    *rank, *line);

        else
        {
#if 0 // debugging            
            if ((int)(*pivot) != *rank)
                printf("*** Plate %d, Fibre pivot number %d, spectrum rank %d\n",
                       *plate, *pivot, *rank);
#endif
            return true;
        }
    }
    return false;

}



/*
 * Given a 2dF spectrum number (1 to SPECTRUM_NUM_MAX), return the 2dF pivot index  
 *   (0 to (NUM_PIVOTS_2DF-1)).
 *
 * Returns -1 for no valid pivot for this spectra.
 */
int specFibres::pivotIndex(
    const unsigned plate,
    const unsigned spectrumNum) const
{
    assert((plate == 0 || plate == 1));
    assert((spectrumNum > 0)||(spectrumNum <=  SPECTRUM_NUM_MAX));
    if (plate == 0)
        return pivotIndexPlate0[spectrumNum-1];
    else
        return pivotIndexPlate1[spectrumNum-1];
}

/*
 * Given a 2dF pivot number (1 to NUM_PIVOTS_2DF), return the spectrum number 
 *   (1 to SPECTRUM_NUM_MAX)
 *
 * Returns -1 for no valid spectra for number for this pivot.
 */
int specFibres::spectrumNum(
        const unsigned plate,
        const unsigned pivotNum) const
{
    assert((plate == 0 || plate == 1));
    assert((pivotNum > 0)||(pivotNum <=  NUM_PIVOTS_2DF));

    int pivotIndex = pivotNum - 1;

    if (plate == 0)
    {
        for (unsigned s = 1; s < SPECTRUM_NUM_MAX ; ++s)
        {
            if (pivotIndexPlate0[s] == pivotIndex)
                return s+1;
        }
    }
    else
    {
        for (unsigned s = 0; s <= SPECTRUM_NUM_MAX ; ++s)
        {
            if (pivotIndexPlate1[s] == pivotIndex)
                return s+1;
        }
    }
    return -1;
}

 

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * Static function prototypes.
 */
static void Usage(bool full, const char *argv0);
static void SetForShuffle(
    const specFibres &fibreInfo,
    const bool evenFibres,
    const int plate,
    SdsId *ID,
    std::ofstream &restoreStream,
    StatusType *status);
  
static void RestoreFile(SdsId *ID,
                        const char *dataFile,
                        StatusType *status);
static void SetFromFile(
    const specFibres &fibreInfo,
    const int plate,
    SdsId *ID,
    std::ofstream &restoreStream,
    const char *file,
    StatusType *status);    

/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * 
 */
int main(int argc, char *argv[])
{
    StatusType status = STATUS__OK;


    if (argc < 4)
    {
        fprintf(stderr,"Insufficent arguments.\n");
        Usage(true, argv[0]);
        return 1;
    }
    /*
     * Check for access to the constants file.
     */
    if ((access(argv[1], R_OK)))
    {
        fprintf(stderr, "Cannot read input constants file %s - %s.\n",
                argv[1], strerror(errno));
        exit(1);
    }
    /*
     *Ensure we can write our output file without clobbering an older
     * back up file.
     */
    if ((access(CONSTS_FILE, F_OK) == 0)&&(access(CONSTS_BCK, F_OK) == 0))
    {
        fprintf(stderr, "Both \"%s\" and \"%s\" exist in the current\n directory.\n",
                CONSTS_FILE, CONSTS_BCK);
        fprintf(stderr, "Please rename/delete one of these before trying again.\n");
        exit(1);
    }
    /*
     * Read the constants file.
     */
    SdsId ID(argv[1], &status);
    if (status != STATUS__OK)
    {
        fprintf(stderr, "Failed to open constants file \"%s\" - \n", argv[1]);
        if (status == SDS__NOMEM)
            fprintf(stderr, " Insufficent memory.\n");
        else if ((status == SDS__FOPEN)||(status == SDS__FREAD))
            fprintf(stderr, " Error %d - \"%s\".\n",
                    errno, strerror(errno));
        else if (status == SDS__NOTSDS)
            fprintf(stderr, " File does not contain an SDS structure.\n");
        else
            fprintf(stderr," Unexpected SDS error code %ld.\n", (long)status);
        exit(1);
    }


    printf("Will work from constants file \"%s\"\n", argv[1]);
   
  

    if (strcmp(argv[2], "restore") == 0)
    {
        /*
         * "restore" command - next argument is the restore file name.
         */
        RestoreFile(&ID, argv[3], &status);
    }
    else 
    {
        specFibres fibreInfo;
        /*
         * shuf/file command.  Need to check we will not clobber
         * an older restore file.
         */
        if (access(RESTORE_FILE, F_OK) == 0)
        {
            fprintf(stderr, "The save file \"%s\" already exists - please rename/delete this before trying again.\n",
                    RESTORE_FILE);
            exit(1);
        }
        /*
         * This is an ofstream which also will delete the file when the
         * destructor is run, if the ClearDeleteOnExit() method has not been 
         * invoked. This allows use to delete the restore file if
         * the command fails.  NOT - relies on destructor running, so
         * exit using "return" rather then "exit".  
         */
        DeleteOnExitOFStream restoreStream;
        restoreStream.open(RESTORE_FILE);
        if (!restoreStream)
        {
            fprintf(stderr, "Failed to open restore file \"%s\" - \"%s\".\n", 
                    RESTORE_FILE, strerror(errno));
            exit(1);
        }
        /*
         * Restore file header comments.
         */
        restoreStream << "#tdFconstants state change restore file" << std::endl;
        time_t now;
        time(&now);
        restoreStream << "# Written at " << ctime(&now);
        restoreStream << "# This is a list of fibres that were disabled" 
                      << std::endl;


        /*
         * Now handle the plate argument - common to all the modes below.
         */
        int plate = -1;  /* If minus one, do both, otherwise plate 0/1 */
        if (strcmp(argv[3],"both") == 0)
        {
            restoreStream << "# Both plates modified" << std::endl;
            plate = -1;
        }
        else if (strcmp(argv[3],"0") == 0)
        {
            restoreStream << "# Plate 0 only modified" << std::endl;
            plate = 0;
        }
        else if (strcmp(argv[3],"1") == 0)
        {
            restoreStream << "# Plate 1 only modified" << std::endl;
            plate = 1;
        }
        else
        {
            fprintf(stderr, "Invalid plate - \"%s\" - must be 0, 1 or both.\n",
                argv[3]);
            Usage(false, argv[0]);
            return 1;
        }
        /*
         * Handle the mode argument and invoke the action. Now that
         * -shuf is like "shuf" except that it does the alternative
         * fibres.
         */
        if (strcmp(argv[2],"shuf") == 0)
        {
            restoreStream << "#   Shuffling mode (even disabled)" << std::endl;
            SetForShuffle(fibreInfo, true, plate, &ID, restoreStream, &status);
        }
        else if (strcmp(argv[2],"-shuf") == 0)
        {
            restoreStream << "#   Shuffling mode (odd disabled)" << std::endl;
            SetForShuffle(fibreInfo, false, plate, &ID, restoreStream, &status);
        }
        else if (((strcmp(argv[2],"file") == 0)&&(argc == 5))&&
                 ((plate == 0)||(plate == 1)))
            SetFromFile(fibreInfo, plate, &ID, restoreStream, argv[4], &status);
        else if ((strcmp(argv[2],"file") == 0)&&(argc == 6)&&(plate == -1))
        {
            SetFromFile(fibreInfo, 0, &ID, restoreStream, argv[4], &status);
            SetFromFile(fibreInfo, 1, &ID, restoreStream, argv[5], &status);
        }
        else
        {
            fprintf(stderr,"Invalid mode argument \"%s\"\n.",argv[2]);
            Usage(true, argv[0]);
            return 1;
        }
        /*
         * If we have good status at this point - then we want
         * to keep out restore file.
         */
        if (status == STATUS__OK)
            restoreStream.ClearDeleteOnExit();
        
    }
    if (status == APP_ERROR)
    {
        fprintf(stderr, "Error setting usage flags.\n");
        return 1;
    }  
    else if (status != STATUS__OK)
    {
        fprintf(stderr, "SDS Error setting usage flags - code %lx.\n",
                (long)status);
        return 1;
    }
    /*
     * We write the new constants file out to the current directory.
     *  But first rename any existing file.
     */
    if (access(CONSTS_FILE, F_OK) == 0)
    {
        if (rename(CONSTS_FILE, CONSTS_BCK) != 0)
        {
            fprintf(stderr, "Error renaming old version of \"%s\" to \"%s\" - %s.\n",
                    CONSTS_FILE, CONSTS_BCK, strerror(errno));
            exit(1);
            
        }
        else
            printf("Have renamed old version of \"%s\" to \"%s\".\n",
                   CONSTS_FILE, CONSTS_BCK);
    }
    ID.Write(CONSTS_FILE, &status);
    if (status != STATUS__OK)
    {
        fprintf(stderr, "Failed to write constants file \"%s\" -\n", CONSTS_FILE);
        if (status == SDS__NOMEM)
            fprintf(stderr, " Insufficent memory for output buffer.\n");
        else if ((status == SDS__FOPEN)||(status == SDS__FWRITE))
            fprintf(stderr, " Error %d - %s.\n",
                    errno, strerror(errno));
        else
            fprintf(stderr," Unexpected SDS error code %ld.\n", (long)status);
        exit(1);
    }
    /*
     * Done.
     */
    if (strcmp(argv[2], "restore") == 0)
    {
        printf("Have created \"%s\" - restored using data from file \"%s\"\n",
               CONSTS_FILE, argv[3]);
    }
    else
    {
        printf("Have created \"%s\" - \n  restore data written to \"%s\"\n",
               CONSTS_FILE, RESTORE_FILE);
    }
    exit(0);
}
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *  Helper functions
 *
 * Usage() - output command help.
 */
static void Usage(bool full, const char *argv0)
{
    fprintf(stderr, "2dF set fibre state program - Usage:\n");
    fprintf(stderr, " %s <constants_file> shuf  0|1|both\n", argv0);
    fprintf(stderr, " %s <constants_file> file  0|1      <file>\n", argv0);
    fprintf(stderr, " %s <constants_file> file  both     <file1> [<file2>]\n", 
            argv0);
    fprintf(stderr,  "%s <constants_file> restore file\n\n", argv0);
    
    if (!full) return;

    fprintf(stderr, 
            "This program will disable 2dF fibres by modifying \"%s\".  You \n",
            CONSTS_FILE);
    fprintf(stderr, 
            " would normally do this prior to configuring a field.  It is NOT normal\n");
    fprintf(stderr, 
            " procedure to use the resulting file with the robot, only with \"configure\".\n\n");
    fprintf(stderr,
            "Access to the file \"spec_fibres.txt\" is required - since fibres are disabled\n");

    fprintf(stderr," based on their rank (slit number) and this defines the pivot number\n");
    fprintf(stderr, " for each slit number.  Ask your support astronmer for this file.\n\n");
    fprintf(stderr, 
            "The file \"%s\" is generated to record the actual changes\n",
            RESTORE_FILE);
    fprintf(stderr, 
            " made and can be used with the \"restore\" command to restore the\n");
    fprintf(stderr, " \"%s\" file, but allowing other disabled fibres to remain\n",
            CONSTS_FILE);
    fprintf(stderr, " disabled.\n\n");

    fprintf(stderr, 
            "The new \"%s\" file is always written to the current directory.\n",
            CONSTS_FILE);
    fprintf(stderr, 
            " If a file of this name already exists int the current directory then it\n");
    fprintf(stderr, " is renamed to \"%s\"\n\n", CONSTS_BCK);
            
    fprintf(stderr, "Program Modes:\n");
    fprintf(stderr, 
            " \"shuff\" mode configures appropiately for charge shuffling.\n");
    fprintf(stderr, " \"file\" mode  reads file for each plate contain a list of fibres to\n");
    fprintf(stderr, "    be disabled.  Comments start with \"#\" characters.  Any number of fibres\n");
    fprintf(stderr, "    (space/tab separated) per line.\n");

}

/*
 *  Return true if the specified fibre is a guide fibre.
 */
static bool IsGuide(unsigned n)
{
    switch (n)
    {
    case 50:
    case 100:
    case 150:
    case 200:
    case 250:
    case 300:
    case 350:
    case 400:
        return true;
    default:
        break;
    }
    return false;
}

 



/*
 * Disable fibres appropiate for normal mode shuffling - every second 
 * second fibre is disabled. 
 */
static void SetForShuffle(
    const specFibres &fibreInfo,
    const bool evenFibres,
    const int plate,               /* Plate number, could be 0/1/-1 (for both) */
    SdsId *ID,                     /* Input tdFconstants structure */
    std::ofstream &restoreStream,  /* restore stream */
    StatusType *status)
{
    if (*status != STATUS__OK) return;
    if (plate == -1)
    {
        /*
         * Handle both plates by calling this routine recursively.
         */
        SetForShuffle(fibreInfo, evenFibres, 0, ID, restoreStream, status);
        SetForShuffle(fibreInfo, evenFibres, 1, ID, restoreStream, status);
        return;
    }
    /*
     * Access the inuse flags structure.
     */
    InUseFlags flags(*ID, plate, status);
    if (*status != STATUS__OK)
    {
        fprintf(stderr,"Error accessing inUse flags structure - %ld\n",
                (long) *status);
        return;
    }
    restoreStream << "Plate " << plate << std::endl;

    int first = (evenFibres ? 2 : 1);

    /*
     * work thorugh the fibres by spectrum number, every second fibre.
     */
    for (unsigned s = first; s <= SPECTRUM_NUM_MAX ; s+=2)
    {
        // Need the pivot number, -1 = invalid, otherwise will be valid range.
        int pivotIndex = fibreInfo.pivotIndex(plate, s); 
        if (IsGuide(pivotIndex+1))
            ;
        else if (pivotIndex >= 0)
        {
            assert(pivotIndex < NUM_PIVOTS_2DF);  // Should be certain
            bool currentState = flags.Get(pivotIndex);
            if (currentState)
            {
                // Current fibre is enabled and is not a guide
                // Write the restore file entry
                // and disable this fibre.
                restoreStream << (pivotIndex+1) << std::endl;
                flags.ClearInUse(pivotIndex);
            }
        }
        else
            fprintf(stderr,"Spectrum number %d does not have a pivot number\n",
                    s);
    }
}
  
/*
 * Read a list of disabled fibres from a file.
 */
static void SetFromFile(
    const specFibres &fibreInfo,   /* spectrum number to pivot info */
    const int plate,               /* plate number, 0 or 1 only */
    SdsId *ID,                     /* Input tdFconstants structure */
    std::ofstream &restoreStream,  /* restore stream */
    const char *file,              /* File with list of disabled fibres */
    StatusType *status)
{
    if (*status != STATUS__OK) return;

    /*
     * Access the inUse flags.
     */
    InUseFlags flags(*ID, plate, status);
    if (*status != STATUS__OK)
    {
        fprintf(stderr,"Error accessing inUse flags structure - %ld\n",
                (long) *status);
        return;
    }
    /*
     * Access the list of files.
     */
    std::ifstream inFile(file);
    if (!inFile)
    {
        *status = APP_ERROR;
        fprintf(stderr, "Error opening input file %s - %s\n",
                file, strerror(errno));
        return;
    }
    restoreStream << "# Using file " << file << " to disable fibres on plate "
                  << plate << std::endl;
    restoreStream << "Plate " << plate << std::endl;

    unsigned entryCount = 0;    /* Number of valid entries in the file */
    unsigned changedCount = 0;  /* Number of fibres we changed the state of */
    bool failed = false;        /* Failure occured */ 
    int line = 1;               /* line number. */

    /* Work through the file */
    while (inFile)
    {
        /*
         * Ignore any white space.
         */
        while ((inFile)&&(inFile.peek() != '\n')&&(isspace(inFile.peek())))
        {
            inFile.ignore(1);
        }

        /*
         * If the next character is a # - ignore until the end of line.  Similary
         * if the next character is the new line.  In both cases, we
         * increment the line count and go around again.
         */
        if ((inFile.peek() == '#')||(inFile.peek() == '\n'))
        {
            inFile.ignore(std::numeric_limits<int>::max(), '\n');
            ++line;
            continue;
        }
        /*
         * If we have advanced to the end of file - bail out now.
         */
        if (inFile.eof()) break;
        /*
         * At this point, we don't have white space or comments.  We should
         * just have a number.
         */
        unsigned rank;
        inFile >> rank;
        if (inFile.eof()) break;
        else if (inFile.fail())
        {
            fprintf(stderr, "Failed to read integer from line %d\n", 
                    line);
            failed = true;
            break;
        }
        /*
         * We have a fibre number as an unsigned int.
         */
        if ((rank < 1)||(rank > SPECTRUM_NUM_MAX))
        {
            failed = true;
            fprintf(stderr, "Invalid fibre spectrum number (rank) %d on line %d\n", 
                    rank, line);
            fprintf(stderr, "Must be in the range 1 to %d\n",
                     SPECTRUM_NUM_MAX);
            break;

        }
        ++entryCount;
        int pivotIndex = fibreInfo.pivotIndex(plate, rank);
        if (flags.Get(pivotIndex))
        {
            ++changedCount;
            flags.ClearInUse(pivotIndex);
            restoreStream << (pivotIndex+1) << std::endl;

            if (IsGuide(pivotIndex+1)) {
                fprintf(stderr,"WARNING:You have disabled guide fibre number %d on plate %d\n:",
                        pivotIndex+1, plate);
            }
        }
        
    } /* while() */
    if (failed)
    {
        *status = APP_ERROR;
        return;
    }
    printf("Plate %d - %d fibre entries, %d flags changed, %d already unused\n",
           plate, entryCount, changedCount, entryCount-changedCount);
}

/*
 * RestoreFile - invoked to restore a tdFconstants.sds file to its original state.  
 *  This is needed as fibres that were disabled before the command need to remain
 *  disabled.  It is particuarlly important if we are actually working with
 *  the robot files themselves. 
 */
static void RestoreFile(
    SdsId *ID,            /* Input tdFconstants structure */
    const char *dataFile, /* restore file name */
    StatusType *status)
{
    if (*status != STATUS__OK) return;
    InUseFlags *flags = 0;         /* Pointer to current set of inUse flags */

    /*
     * Access the restore data file.
     */
    std::ifstream dataStream(dataFile);
    if (!dataStream)
    {
        *status = APP_ERROR;
        fprintf(stderr, "Error opening input file %s - %s\n",
                dataFile, strerror(errno));
        return;
    }
    unsigned entryCount[2]   = {0, 0};/*Number of valid entries in the file by plate*/
    unsigned changedCount[2] = {0, 0};/* Number of fibres we changed the state of */
    bool failed = false;              /* Failure occured */ 
    int line = 1;                     /* line number. */
    int plate = -1;                   /* current plate number -1 = no current plate */
    while (dataStream)
    {
        /*
         * Ignore any white space.
         */
        while ((dataStream)&&(dataStream.peek() != '\n')&&
               (isspace(dataStream.peek())))
        {
            dataStream.ignore(1);
        }

        /*
         * If the next character is a # - ignore until the end of line.  Similary
         * if the next character is the new line.  In both cases, we
         * increment the line count and go around again.
         */
        if ((dataStream.peek() == '#')||(dataStream.peek() == '\n'))
        {
            dataStream.ignore(std::numeric_limits<int>::max(), '\n');
            ++line;
            continue;
        }
        /*
         * If we have advanced to the end of file - bail out now.
         */
        if (dataStream.eof()) break;
        /*
         * At this point, we don't have white space or comments.
         *  We might have a fibre  number or we might have a 
         *  plate spec.
         */
        if (dataStream.peek() == 'P')
        {
            /* A line starting with 'P', hopefully this is the "Plate n" spec
             * where n = 0 or 1.
             */
            std::string word;
            dataStream >> word;
            if (word != "Plate")
            {
                fprintf(stderr, "Failed to plate spec from line %d\n", 
                        line);
                failed = true;
                break;
            }
            unsigned newPlate;
            dataStream >> newPlate;
            if (dataStream.fail())
            {
                fprintf(stderr, "Failed to get plate number integer from line %d\n", 
                        line);
                failed = true;
                break;
            }
            if (newPlate > 1)
            {
                fprintf(stderr, "Invalid plate number %d on line %d\n", 
                        newPlate, line);
                failed = true;
                break;
            }
            /*
             * If we are here, we have a valid plate number.  It is most likely
             * a new plate number, but don't complain if it is not.  
             * 
             * If it is a new plate number, we must access the relevant plate
             * inUse flags.
             */
            if ((int)newPlate != plate)
            {
                if (flags)
                    delete flags;
                plate = newPlate;
                flags = new InUseFlags(*ID, plate, status);
                if (*status != STATUS__OK)
                {
                    failed = true;
                    break;
                }
            }
            continue;
        }
        /*
         * By this point the next item should be a fibre number.  IF we don't
         * yet have access to inUse flags - we have a problem.
         */
        if (!flags)
        {
            fprintf(stderr, "File format error - line %d- expected to have plate by now\n",
                    line);
        }
        unsigned fib;
        dataStream >> fib;
        if (dataStream.eof()) 
            break;
        if (dataStream.fail())
        {
            fprintf(stderr, "Failed to read integer from line %d\n", 
                    line);
            failed = true;
            break;
        }
        /*
         * We have a fibre number as an unsigned int.
         */
        if ((fib < 1)||(fib > NUM_PIVOTS_2DF))
        {
            failed = true;
            fprintf(stderr, "Invalid fibre number %d on line %d\n", fib, line);
            fprintf(stderr, "Must be in the range 1 to %d\n",
                    NUM_PIVOTS_2DF);
            break;

        }
        ++entryCount[plate];
        if (!flags->Get(fib-1))
        {
            ++changedCount[plate];
            flags->SetInUse(fib-1);
        }
        
    } /* while() */
    if (flags)
        delete flags;
    if (failed)
    {
        *status = APP_ERROR;
        return;
    }
    for (plate = 0; plate < 2 ; ++plate) {
        printf("Plate %d - %d fibre entries, %d flags changed, %d already set inuse\n",
               plate, 
               entryCount[plate], 
               changedCount[plate], 
               entryCount[plate]-changedCount[plate]);
    }

}
