/*+
                           t d F  w l e n s

 *  Module name:
      tdFwlens.c

 *  Function:

 *  Description:

 *  Language:
      C

 *  Authors: Tony Farrell, AAO (TJF).
 *-

 * "@(#) $Id: ACMM:aaoConfigure/src/tdFwlens.c,v 12.7 04-Nov-2013 11:28:49+11 tjf $"

 *  History:
      20-Jul-2006 TJF  Original version
      {@change entry@}

 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <time.h>
#include "status.h"
#include "Ers.h"
#include "sds.h"
#include "arg.h"
#include "tdFconvert.h"
#include "conf_err.h"


/*  ------------------------------------------------------------------------- */

/*+        T D F W L E N S I N I T

 *  Function name:
      tdFwlensInit

 *  Function:
      Initialise a structure of type wavelensType.

 *  Description:
      Just initialises this struture to be empty.

 *  Language:
      C

 *  Call:
      (void) = tdFwlens (wlens,status)

 *  Parameters:  (">" input, "!" modified, "W" workspace, "<" output)
      (<) wlens          (wavelensType *)  Wavelength structure.
      (!) status         (StatusType *) Modified status.

 *  Proir Requirements:

 *  Support: Tony Farrell, AAO

 *-

 *  History:
      20-Jul-2006  TJF  Original version
      {@change entry@}
 */
extern void tdFwlensInit(
    wavelensType * const wlens, 
    StatusType   * const status)
{
    if (*status != STATUS__OK) return;
    memset((void *)wlens, 0, sizeof(*wlens));
}
 


/*+        T D F W L E N S A D D

 *  Function name:
      tdFwlensAdd

 *  Function:
        Add a new wavelength to the list maintained in this wlens structure.

 *  Description:

 *  Language:
      C

 *  Call:
      (void) = tdFwlensAdd (index,wlen,wlens,status)

 *  Parameters:  (">" input, "!" modified, "W" workspace, "<" output)
      (>) index          (unsigned) index for this wavelength.must be<MAX_WLENS
      (>) wlen           (double ) The wavelength.
      (!) wlens          (wavelensType *)  Wavelength structure.
      (!) status         (StatusType *) Modified status.

 *  Proir Requirements:

 *  Support: Tony Farrell, AAO

 *-

 *  History:
      20-Jul-2006  TJF  Original version
      {@change entry@}
 */
extern void tdFwlensAdd(
    const unsigned int index,
    const double wlen, 
    wavelensType * const wlens, 
    StatusType   * const status)
{
    if (*status != STATUS__OK) return;
    if ((index > MAX_WLENS)||(wlen <= 0))
    {
        *status = CONF__SPECIAL;
        return;
    }
    /*
     * Add the new value.
     */
    wlens->wlen[index] = wlen;
}

/*+        T D F W L E N S G E  T

 *  Function name:
      tdFwlensGet

 *  Function:
        Return a wavelength from a wavelensType structure, specified by index.

 *  Description:
 

 *  Language:
      C

 *  Call:
      (void) = tdFwlens (wlens, index, wlen, ,status)

 *  Parameters:  (">" input, "!" modified, "W" workspace, "<" output)
      (>) wlens          (const wavelensType *)  Wavelength structure.
      (>) index          (unsigned) Index in the wavelength structure.
      (<) wlen           (double *) The wavelength.
      (!) status         (StatusType *) Modified status.

 *  Proir Requirements:

 *  Support: Tony Farrell, AAO

 *-

 *  History:
      20-Jul-2006  TJF  Original version
      {@change entry@}
 */
extern void tdFwlensGet(
    const wavelensType *wlens, 
    const unsigned index, 
    double * const wlen, 
    StatusType * const status)
{
    if (*status != STATUS__OK) return;
    if (index >= MAX_WLENS)
    {
        *status = CONF__SPECIAL;
        return;
    }
    *wlen = wlens->wlen[index];
}

/*+        T D F W L E N S T O S D S

 *  Function name:
      tdFwlensToSds

 *  Function:
      Put a wavelength structure into an SDS structure.

 *  Description:
      Creates a item SDS arra of double name "wlens" within the specified
      parent structure. The values from the wlens structure will be
      written there.

 *  Language:
      C

 *  Call:
      (void) = tdFwlensToSds (wlens,parent,status)

 *  Parameters:  (">" input, "!" modified, "W" workspace, "<" output)
      (>) wlens          (const wavelensType *)  Wavelength structure.
      (>) parent         (SdsIdType) Sds ID of parent structure (structure
                                      will be modified.
      (!) status         (StatusType *) Modified status.

 *  Proir Requirements:

 *  Support: Tony Farrell, AAO

 *-

 *  History:
      20-Jul-2006  TJF  Original version
      {@change entry@}
 */
extern void tdFwlensToSds(const wavelensType *wlens, SdsIdType parent, 
                          StatusType *status)
{
    SdsIdType id = 0;
    unsigned long dims[1];
    if (*status != STATUS__OK) return;

    /*
     * We don't bother adding this element if wlens == 0, it makes the
     * structure confusing.
     */
    if (wlens == 0) return;
    /*
     * Create the structure and put the data into it.
     */
    dims[0] = MAX_WLENS;
    SdsNew(parent, "wlens", 0, 0, SDS_DOUBLE, 1, dims, &id, status);
    SdsPut(id, MAX_WLENS*sizeof(double),  0, (void *)&wlens->wlen[0], status);
    SdsFreeId(id, status);
}
/*+        T D F W L E N S F R O M S D S

 *  Function name:
      tdFwlensFromSds

 *  Function:
       Read a wavelensType structure from an SDS strucutre.

 *  Description:
       The "parent" SDS Id may or may not contain an item of type
       wlens.  If it does, it is used to fill out the "wlens" structure.

       If it does not, then the wlens structure is initialised to empty.

 *  Language:
      C

 *  Call:
      (void) = tdFwlensFromSds (parent,wlens,status)

 *  Parameters:  (">" input, "!" modified, "W" workspace, "<" output)
      (>) parent         (SdsIdType) Sds ID of parent structure.
      (<) wlens          (wavelensType *)  Wavelength structure.
      (!) status         (StatusType *) Modified status.

 *  Proir Requirements:

 *  Support: Tony Farrell, AAO

 *-

 *  History:
      20-Jul-2006  TJF  Original version
      {@change entry@}
 */
extern void tdFwlensFromSds(
    const SdsIdType parent, 
    wavelensType * const wlens,
    StatusType * const status)
{
    StatusType ignore = STATUS__OK;
    SdsIdType id = 0;
    long dims[1] = { MAX_WLENS };
    if (*status != STATUS__OK) return;

    /* Start off by initialising the structure */
    tdFwlensInit(wlens, status);

    SdsFind(parent, "wlens", &id, status);
    if (*status == SDS__NOITEM)
    {
        *status = 0;
        return;
    }
    /*
     * Check if this is a double array with 1 dimension.
     */
    ArgCheckItem(id, SDS_DOUBLE, 1, dims, status);

    if (*status == STATUS__OK)
    {
        unsigned long actlen;
        SdsGet(id, sizeof(wlens->wlen), 0, (void *)wlens->wlen, &actlen,
               status);
    }
    SdsFreeId(id, &ignore);
}







/*+        T D F W L E N S S D S T O L I S T I N G H E A D E R

 *  Function name:
      tdFwlensSdsToListingHeader

 *  Function:
       Writes the WLENS header to a list file.

 *  Description:
       The "parent" SDS Id may or may not contain an item of type
       wlens.  If it does, it is used to write a header to a listing
       file.

 *  Language:
      C

 *  Call:
      (void) = tdFwlensSdsToListingHeader (parent,listingFile,status)

 *  Parameters:  (">" input, "!" modified, "W" workspace, "<" output)
      (>) parent         (SdsIdType) Sds ID of parent structure.
      (>) listingFile    (FILE *)  The file to write the entries too.
      (!) status         (StatusType *) Modified status.

 *  Proir Requirements:

 *  Support: Tony Farrell, AAO

 *-

 *  History:
      26-Jul-2006  TJF  Original version
      {@change entry@}
 */

extern void tdFwlensSdsToListingHeader(
    const SdsIdType parent, 
    FILE * const listingFile,
    StatusType * const status)
{
    wavelensType wlens;
    unsigned int index;
    if (*status != STATUS__OK) return;

    /*
     * Read from the SDS structure and then write
     * any valid entries to the file.
     */
    tdFwlensFromSds(parent, &wlens, status);
    for (index = 0; index < MAX_WLENS ; ++index)
    {
        if (wlens.wlen[index] > 0) {
            fprintf(listingFile, "WLEN%d %.2f\n", 
                    index+1, wlens.wlen[index]*MICRONS_TO_ANGSTROMS);
        }
    }
}



