/*
 *                         T e s t h a r n e s s
 *
 *  Function:
 *      Provides a test framework for experimental configuration algorithms.
 *
 *  Description:
 *      This file contains what is supposed to be a simple test program
 *      that can be used to exercise the allocation algorithm used by
 *      configure without the need to run the whole configure
 *      tcl/tk program interactively. It takes a (text) field file with a
 *      list of targets, runs the allocation algorithm, and creates a
 *      .sds file with the allocation that can be read into the full
 *      configure program.
 *
 *   Usage:   testharness instrument filename
 *
 *      where 'instrument' is the name of the instrument being configured,
 *      eg '2dF', and filename is the name of the input field file. The
 *      program produces an output file called 'output.sds' which overwrites
 *      any existing file of the same name. It also writes a text file giving
 *      the allocation details, as 'output.lis' also overwriting any existing
 *      file, although this is mainly intended as an aid for regression 
 *      testing.
 *
 *   Author:
 *      Keith Shortridge, AAO.
 *
 * "@(#) $Id: ACMM:aaoConfigure/src/testharness.c,v 12.7 04-Nov-2013 11:28:49+11 tjf $"
 *
 *   History:
 *       1st Mar 2005.  Original version. KS.
 *      29th May 2009.  Sets the random number generator to the same seed
 *                      so results can be compared for regression testing.
 *                      (I'm sure ther must have been other changes between
 *                      March 2005 and May 2009! Someone's been slack with
 *                      this history section.) KS.
 *       1st Jun 2009.  Added a call to ConfMakeListing() and added a version
 *                      of that routine. KS.
 *      11th Aug 2009.  Modified ConfMakeListing() to support use of explicit
 *                      pmra and pmdec structure items for proper motion values
 *                      rather than using pseudo-comments. KS.
 *      14th Aug 2009.  ConfConvertAscii() ditto. KS.
 *      17th Sep 2013.  Now reads in affectation data on startup. Added HERMES
 *                      to the list of supported instruments. KS.
 *
 *   Linking:
 *      At present, under Solaris on aaoprog, linked as follows from the 
 *      aaoConfigure/src directory: (this can be cut and pasted into a
 *      terminal session)

    gcc -Wall -ansi -g -I/instsoft/extern/include -I../../sds -I../drama_src \
       -I../../slalib_o -I../../ozpozFpcol/ws/include -I../../include \
       -I../../fpil -I. -I../../2dFutil -I../../6dFutil -g -o testharness \
       testharness.c -DNO_TCL configure_subs.c HashEmulator.c \
       libconfigure.a -lm

 *  and run as, for example:

    setenv CONFIG_FILES ../data_files
    ./testharness 2dF ../fields/2dF/tdf150.fld

 *  if you want to use a different allocation algorithm, you can link it in
 *  to override the standard one in libconfigure.a. For example, to link
 *  with the template algorithm in config_method.template.c, you can use:
 
    gcc -Wall -ansi -g -I/instsoft/extern/include -I../../sds -I../drama_src \
       -I../../slalib_o -I../../ozpozFpcol/ws/include -I../../include \
       -I../../fpil -I. -I../../2dFutil -I../../6dFutil -g -o testharness \
       testharness.c config_method.template.c -DNO_TCL configure_subs.c \
       HashEmulator.c libconfigure.a -lm
 */

static const char *rcsId="@(#) $Id: ACMM:aaoConfigure/src/testharness.c,v 12.7 04-Nov-2013 11:28:49+11 tjf $";
static void *use_rcsId = (0 ? (void *)(&use_rcsId) : (void *) &rcsId);

#include "configure_subs.h"
 
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <math.h>

#include "drama.h"
#include "status.h"
#include "tdFconfig.h"
#include "tdFconfigMethod.h"
#include "tdFconfig.h"
#include "tdFconvert.h"
#include "Ers.h"
#include "mess.h"
#include "slalib.h"
#include "slamac.h"
#include "sds.h"
#include "arg.h"

#include "conf_err.h"

#include "HashEmulator.h"

/*
 *  Include FPIL interface headers for each supported instrument.
 */

#include "tdffpil.h"            /* 2dF - AAT */
#include "sixdffpil.h"          /* 6dF - UKSCHMIDT */
#include "fpcolFpil.h"          /* OzPoz - VLT-2 */

/*
 *  Note that of two below, CMTLEN is always used for buffers which may
 *  take either a comment or name item, hence if their sizes changes.
 *  CMTLEN must remain the largest.
 */
 
#define NAMELEN 80              /* Length of "name" items in structures */
#define CMTLEN  80              /* Length of "comment" items in structures */

/*
 *  Executable directory.  Not all operating systems define MAXPATHLEN so
 *  define it here if needed.
 */
 
#ifndef MAXPATHLEN
#define MAXPATHLEN 1024
#endif
static char ExecutableDir[MAXPATHLEN];

/*
 *  Config files directory (value of the CONFIG_FILES parameter)
 */
 
static char ConfigFilesDir[MAXPATHLEN];

/*
 *  Assuming each supported instrument has a function with the
 *  following typedef which does the FPIL Full initialisation, then
 *  we can set up an array of supported instruments here and
 *  otherwise don't have to worry about them - FPIL gives us all
 *  the details.
 */
typedef void (*ConfFpilInitRoutineType) (FpilType * inst);

static struct {
   FpilType Info;
   ConfFpilInitRoutineType InitRoutine;
   const char *inst_file;
   const char *doco_site;
} Instruments[] = {
   {
   0, TdfFpilFullInitHERMES,
   "ftp://ftp.aao.gov.au/pub/2df/latest_config_files",
   "http://www.aao.gov.au/aaomega"},
   {
   0, TdfFpilFullInit400, 
   "ftp://ftp.aao.gov.au/pub/2df/latest_config_files",
   "http://www.aao.gov.au/2df"},
   {
   0, TdfFpilFullInit404, 
   "ftp://ftp.aao.gov.au/pub/2df/latest_config_files",
   "http://www.aao.gov.au/2df"},
   {
   0, TdfFpilFullInitAAOmega, 
   "ftp://ftp.aao.gov.au/pub/2df/latest_config_files",
   "http://www.aao.gov.au/aaomega"},
   {
   0, fpcolFpilInit, 
   "http://www.eso.org/observing/p2pp/OSS/FPOSS/FPOSS-tool.html",
   "http://www.eso.org/observing/p2pp/OSS/FPOSS/FPOSS-tool.html" },
   {
   0, sixdfFpilFullInit, 
   0,
   "http://www.aao.gov.au/ukst/6df.html"}
};

/*  Forward declarations of routines defined later in this file */

void OpenAsciiFile (
   char* FileName,
   float LowMag,
   float HighMag,
   SdsIdType* FileId,
   StatusType *Status);
   
void ListParameters (void);
   
static void RptTelModelFile(void *argument DUNUSED, const char *string,
    StatusType * status);
    
int ReportRoutine(void *clientData, char Type DUNUSED, 
    char *Description DUNUSED,
    int pivot DUNUSED, int guide DUNUSED, int target DUNUSED,
    long TargetX DUNUSED, long TargetY DUNUSED,
    double theta DUNUSED, double PercentDone);
   
void ConfAllocRpt(void *clientData, int Alloc,
    int pivot, int guide, int target,
    long TargetX DUNUSED, long TargetY DUNUSED,
    double theta, StatusType * status);
    
static int AllocateFibre(
   SdsIdType Id,
   SdsIdType PivId,
   int Fibre,
   char* Type,
   int Object);
   
static int ConvertXy(
    SdsIdType Id);
    
static int CompressUnalloc(
   SdsIdType Id);
   
static int SetUTMeridian(SdsIdType ID, double Ha);
   
static int DeallocateFibre(
   SdsIdType Id,
   SdsIdType PivId,
   int Fibre,
   char* Type,
   int* Object);
   
static int AddIndex(SdsIdType Id);

static int ConfMakeListing(
   SdsIdType Id,
   char* File,
   char* Type);

/*  Not all systems seem to declare strcasecmp() properly, so we do so here.
 */

#ifndef __macosx__
int strcasecmp(const char *s1, const char *s2);
#endif

#define FALSE 0
#define TRUE 1
#define DEBUG printf

/*  General catch-all error code, for the moment */

#define ERROR_CODE 1

/*  ------------------------------------------------------------------------- */

/*                      R p t  C o n s t  F i l e
 *
 *  This callback routine is invoked when the instrument support layer
 *  opens its constants files.
 */
 
static void RptConstFile(void *argument DUNUSED,
    const char *string, StatusType * status)
{
    if (*status != STATUS__OK)
        return;
    printf("Opened constants file \"%s\"\n", string);
}

/*  ------------------------------------------------------------------------- */

/*                    C o n f  R p t  A f f e c t  F i l e
 *
 *  This callback routine is invoked when the instrument support layer
 *  opens its affectation files. Note that this can receive reports of
 *  failures to open files as well as reports of files opened. This returns
 *  zero in all cases, which will keep FpilReadAffectData() reading even if
 *  if fails to find all the affectation files. 
 */
 
static short RptAffectFile(void *argument DUNUSED,
    short Failed, const char *string, StatusType * status)
{
    if (*status != STATUS__OK) return 0;
    if (Failed) {
        fprintf (stdout,"Failed to open affectation file (ignored):\n");
        fprintf (stdout,"%s\n",string);
    } else {
        fprintf(stdout, "Opened affectation file \"%s\"\n", string);
    }
    return 0;
}

/*  ------------------------------------------------------------------------- */

/*                         I n i t  M e s s a g e
 *
 *  This routine is used internally and passed to ConfigInitialise() for
 *  outputing informational messages to the user during allocation.
 */
 
static void InitMessage(void *argument, const char *string)
{
    printf ("%s\n",string);
}

/*  ------------------------------------------------------------------------- */

/*           G l o b a l  V a r i a b l e s  &  C o n s t a n t s
 */
 
/*
 *  When an instrument is selected, details are copied here. Note that most
 *  calls to Fpil routines take *Instrument as the first argument, so it is
 *  important that this be defined before any such calls are made. However,
 *  the way the program runs guarantees that the first thing that happens is
 *  that the instrument is selected.
 *
 *  These few have been taken from the corresponding section in configure.c
 *  Some of the others defined there may need to be used here as well.
 */

FpilType *Instrument = 0;                    /* Instrument details pointer */
FpilConstantsType InstConstDetails;          /* Used to access constants   */
FpilModelType *TelModel;                     /* For telescope model        */
SdsIdType InstFieldDetails;                  /* Sds id of field details    */
unsigned CurrentField;                       /* Current field number       */
unsigned int InstNumFibreTypes = 0;          /* Number of fibre types      */
int *UseFibreType;                           /* Points to array of values  
                                                indicating if a fibre type
                                                should be configured.      */
double *TelModelParams;                      /* Telescope model parameters */
double *TelModelParamsUsed;                  /* Telescope model parameters
                                              * used when the last coordinate
                                              * conversion was performed.  */
unsigned int TelModelNumParams;              /* Number of tel model params */

/*  Defaults for Temp, Pressure, Humidity and Wavelength. There are three
 *  wavelengths - the main one used for ordinary target objects, the one
 *  used for the guide stars (which will usually depend on any filters in
 *  the guiding system and the response of the guider - we assume that these
 *  effects dominate the colour of the guide star itself), and the wavelength
 *  used for the telescope guider.
 *
 *  At present, the user interface allows the temperature, pressure and
 *  humidity to be changed, and allows the main observing wavelength to
 *  be changed. However, at present, the guide wavelengths are fixed.
 *  Of the supported instruments, only FLAMES makes use of telescope guide
 *  targets. Note that Initialise{} modifies the humidity default if the
 *  instrument is FLAMES, in which case it uses 0.2. It also changes the
 *  pressure for FLAMES to 742.
 */
 
#define TEMP 285
#define PRESS 900
#define HUMID 0.5
#define WAVE 0.60

double gtemp = TEMP;                    /* Temperature in degrees kelvin */
double gpress = PRESS;                  /* Pressure in mm Hg */
double ghumid = HUMID;                  /* Humidity - as a fraction, ie 0 - 1 */
double gwave = WAVE;                    /* Wavelength in microns for main 
                                         * target objects (non-guide objects) */
double gwave2 = WAVE;                   /* Wavelength in microns for main 
                                         * target objects (non-guide objects) 
                                         * for the 2nd 2df spectrograph */
/*
 *  Flag to see if we should ignore some errors GBD 17/6/98
 */
 
int ImportOnlyMode = 0;

/*  ------------------------------------------------------------------------- */

/*                          m a i n
 */
 
int main (
   int argc,
   char* argv[])
{
   SdsIdType Cfig;
   char *configFilesValPnt;
   unsigned i;                           /* General loop index */
   int Known;
   unsigned InstrumentIndex=0;
   StatusType Status = 0;
   StatusType Ignore = 0;
   const char *const *ParamNames;
   const char *const *ParamDescrs;
   const double *ParamDefaults;
   const char *const *SpecNames;
   
   if (argc < 3) {
      printf ("Usage: testharness instrument fieldFile\n");
      Status = 1;
      goto Exit;
   }

   CurrentField = 0;                    /*  Really should be specified */
   
   /*  Find where are being run from so we can find other files. */
     
   ImpZExeDir(argv[0], ExecutableDir, sizeof(ExecutableDir), &Known);
   ExecutableDir[0] = '\0';

   /*  See if the CONFIG_FILES environment variable has a value and if
    *  so, grab it.
    */
     
   configFilesValPnt = getenv("CONFIG_FILES");
   if (configFilesValPnt) {
      strncpy(ConfigFilesDir, configFilesValPnt, sizeof(ConfigFilesDir));
      ConfigFilesDir[sizeof(ConfigFilesDir) - 1] = '\0';
   } else {
      ConfigFilesDir[0] = '\0';
   }

   /*  Initialise each supported instrument.  */
    
   for (i = 0; i < sizeof(Instruments) / sizeof(Instruments[0]); ++i) {
      (*Instruments[i].InitRoutine) (&Instruments[i].Info);
   }
   
   if (strlen(argv[1]) != 0) {
      for (i = 0; i < sizeof(Instruments) / sizeof(Instruments[0]); ++i) {
         if (strcasecmp(argv[1],FpilGetInstName(Instruments[i].Info)) == 0) {
            InstrumentIndex = i;
            break;
         }
      }
      if (i >= sizeof(Instruments) / sizeof(Instruments[0])) {
          printf("%s:Instrument \"%s\" unknown\n", argv[0], argv[1]);
          Status = 1;
          goto Exit;
      }
   }
   Instrument = &Instruments[InstrumentIndex].Info;
   printf("Configuring fibre instrument \"%s\" for telescope \"%s\"\n",
             FpilGetInstName(*Instrument), FpilGetTelescope(*Instrument));
   FpilConsAccess(*Instrument, ConfigFilesDir, ExecutableDir,
                            RptConstFile, 0, &InstConstDetails, &Status);
   if (Status != 0) goto Exit;
   
   /*  Read in any affectation data */
   
   FpilReadAffectData (*Instrument,ConfigFilesDir, ExecutableDir,
                                        RptAffectFile, 0, &Status);
   if (Status != 0) goto Exit;   
                            
   /*  Allocate space and initialise the flags which indicate if
    *  given fibre types are to be used, and the structure that
    *  keeps the allocation statistics for each fibre type.
    */
    
   FpilConSpecInfo(*Instrument, &InstNumFibreTypes, &SpecNames,&Status);
   if (Status != 0) goto Exit;
   UseFibreType = malloc((InstNumFibreTypes) * sizeof(*UseFibreType));
   for (i = 0; i < InstNumFibreTypes; ++i) {
      UseFibreType[i] = 1;
   }

   /*  Field number valid.  Access details and read telescope model
    *  details for this field.
    */
    
   FpilConFieldId(InstConstDetails, CurrentField, &InstFieldDetails,
                                                          &Status);

   /*  Set up the telescope model parameters */

   FpilModelAccess(*Instrument, ConfigFilesDir, ExecutableDir,
             RptTelModelFile, 0, CurrentField, &TelModel, &Status);

   /*  Get details of the telescope model parameters. */
   
   FpilModelParams(TelModel, &TelModelNumParams,
                &ParamNames, &ParamDescrs, &ParamDefaults, &Status);

   if (Status != STATUS__OK) goto Exit;

   /*  Malloc two arrays to contain the parameter values and copy
    *  in the default parameter values. One copy is used for the
    *  current parameter values, one to record the values used
    *  the last time the XY-conversion of positions was performed
    *  (See ConvertXy).
    */
                
   if (TelModelNumParams) {
      TelModelParams = malloc(sizeof(double) * TelModelNumParams);
      TelModelParamsUsed = malloc(sizeof(double) * TelModelNumParams);
      for (i = 0; i < TelModelNumParams; ++i) {
          TelModelParams[i] = ParamDefaults[i];
          TelModelParamsUsed[i] = ParamDefaults[i];
      }
   } else {
      TelModelParams = 0;
   }
   
   /*  Note - configure's code at this point does rather more, which we
    *  ought to duplicate. But for the moment, this minimal code will do.
    */

   ConfigInitialise(*Instrument,InstConstDetails,InitMessage,0,&Status);
   if (Status != STATUS__OK) goto Exit;
   
   /*  List the current parameter options */
   
   /* ListParameters(); */
   
   /*  The next two lines set parameters for the allocation algorithms.
    *  The "target selection" setting makes the standard allocation code
    *  use the 'Taylor' algorithm. The "Random Seed (expert)" option
    *  is supported by the simulated annealing algorithm, and setting
    *  this means that the random number generator used by the algorithm
    *  will produce the same sequence in all runs. This means that
    *  the same input files should produce exactly the same results,
    *  and these results can be compared for regression testing. We don't
    *  know which code we're linked with (although we could find out), so
    *  we just ignore bad status from these calls.
    */
    
   Ignore = 0;
   ConfigSetParameter("Target selection","by hardest target",&Ignore);
   Ignore = 0;
   ConfigSetParameter("Random Seed (expert)","1234",&Ignore);
    
   /*  Read in a target ascii file */
   
   OpenAsciiFile (argv[2],0.0,0.0,&Cfig,&Status);
   SetUTMeridian(Cfig,0.0);
   DoMakeAllocated(Cfig,InstFieldDetails,&Status);
   ConvertXy(Cfig);
   AddIndex (Cfig);
   
   /*  The following sequence - ConfigSetTargets(), ConfigSetSpecPivotsFlag(),
    *  ConfigAllocate() - is the essential sequence carried out by
    *  ConfDoAllocation in configure.c
    */
    
   ConfigSetTargets (0,Cfig,CurrentField,ReportRoutine,0,&Status);
   if (Status != STATUS__OK) goto Exit;
   
   /*  Indicate which fibre type we want to use */

   for (i = 0; i < InstNumFibreTypes; ++i) {
      ConfigSetSpecPivotsFlag(i, UseFibreType[i], &Status);
   }
   
   /*  Set sky counts for each fibre type - these are the number of fibres of
    *  this type that are to be allocated to sky targets (if available) or
    *  left unallocated for later allocation to sky targets. We clear out all
    *  the sky counts, then set those we want to control.
    *  Configure does this using the FibreTypeStats[i].FibreSkyCount values
    *  set from the GUI. We don't have this information here, so we have
    *  to just pick an arbitrary number, and assume that fibre type zero
    *  is a guide fibre we don't want to mess with.
    */
    
    ConfigSetSky(-1,-1,&Status);
    for (i = 0; i < InstNumFibreTypes; ++i) {
        if (i > 0) {
            DEBUG ("Setting sky for fibre type %d to 22\n",i);
            ConfigSetSky(i,22,&Status);
        }
    }
   /*  Perform the allocation */
   
   ConfigAllocate(ReportRoutine, 0, ConfAllocRpt, (void*) Cfig, &Status);
   
   /*  Output the resulting configuration as an Sds file */
   
   CompressUnalloc (Cfig);
   SdsWrite (Cfig,"Output.sds",&Status);
   
   /*  And as a text file listing the allocations */
   
   ConfMakeListing (Cfig,"Output.lis","alloc");
   
Exit:;

   return Status;  
}

/*  ------------------------------------------------------------------------- */

/*  The rest of the code in this file comprises routines taken from 
 *  configure.c and bodged slightly so they can be called from C instead
 *  of from Tcl. Eventually the original routines will be restructured
 *  to make this easier.
 */

/*  ------------------------------------------------------------------------- */

/*                       O p e n  A s c i i  F i l e
 *
 *   Opens an ASCII input text file - a target list - and converts it
 *   to the SDS structure used by configure. This code is taken mainly
 *   from the ConfConvertAscii{} routine in configure.c
 */

void OpenAsciiFile (
   char* FileName,
   float LowMag,
   float HighMag,
   SdsIdType* FileId,
   StatusType* Status)
{
    tdFfieldType field;
    short fracDayFlag = FALSE;
    short options = 0;
    double longit, lat, hm;
    FILE *asciiFile;
    char oname[NAMELEN];
    double uttime = 0.5;
    char buff[80];
    objectType *list, *next;
    float maglimits[2];
    int lineCount = 0;
    wavelensType wlens;   /* explict wavelength values. */
    short explicitPms;    /* true if file format includes proper motion cols */
    /*  ARGUS items commented out
    double Angle;
    double Offset;
    short InUse;
    char Scale[16];
    */
    
    maglimits[0] = LowMag;
    maglimits[1] = HighMag;

    /*  We need to work out the uttime correction to allow for the difference
     *  in midnight UT and local time.
     */
     
    slaObs(-1, (char *) FpilGetTelescope(*Instrument), oname, &longit, &lat,
                                                                         &hm);
    if (strcmp(oname, "?") == 0) {
       uttime = 0.0;
    } else {
       uttime = -longit / (2.0 * 3.141593);
    }
    /* Initiailse the explicit wavelength array */
    tdFwlensInit(&wlens, Status);

    /*  Now open file and process the field data at the start of the file */
    
    if ((asciiFile = fopen(FileName, "r")) == NULL) {
        printf("Cannot open file '%s'\n", FileName);
        *Status = ERROR_CODE;
        goto Exit;
    }
    tdFparseFieldData(Instrument, asciiFile, &field.fieldData, uttime,
                      &fracDayFlag, &lineCount, &wlens, &explicitPms, Status);
                                          
    /*  If field.fieldData.argusSpec has been set, then ARGUS information 
     *  has been specified. We handle this through ConfCSetArgusData(), which
     *  will arrange to set the position angle as required.
     *
     *  For the moment, this has been commented out *******
     */
    
    /*
    ConfCGetArgusData(&InUse,&Angle,Scale,&Offset);
    Angle = field.fieldData.posangle + Offset;
    ConfCSetArgusData(field.fieldData.argusSpec,Angle,
                                        field.fieldData.scale,Offset);
    */    

    field.unallocGuide = field.unallocObject = NULL;
    tdFparseObjectData(Instrument, &wlens, asciiFile, &field,
                       &options, maglimits, explicitPms, &lineCount, Status);

    fclose(asciiFile);

    tdFconvertToSds("2df_field", &field, options, &wlens, FileId, Status);

    /*
     * Free the two lists.  This is done in spite of the status, as they
     * were initialised above and will need to be cleaned up if they
     * are no longer null
     */
     
    list = field.unallocObject;
    while (list) {
        next = list->next;
        free(list);
        list = next;
    }
    list = field.unallocGuide;
    while (list) {
        next = list->next;
        free(list);
        list = next;
    }

    if (*Status == STATUS__OK) {
    
        /*  At this point the configure code returns the fracDayFlag to
         *  the calling TCL. */
         
    } else {
    
        printf("Error reading file '%s'\n",FileName);
        MessGetMsg(*Status, 0, sizeof(buff), buff);
        printf ("%s\n",buff);
    }
    
Exit:;

}

/*  ------------------------------------------------------------------------- */

/*  tdFparse.c contains calls to some Tcl routines. We provide implementations
 *  here that suit our purpose.
 */

int Tcl_GetStdChannel(int Type DUNUSED)
{
    return 1;
}

void Tcl_Write (int Chan DUNUSED, char* Buffer, int Flag DUNUSED)
{
   printf ("%s\n",Buffer);
}

/*  ------------------------------------------------------------------------- */

/*                         R e p o r t  R o u t i n e
 *
 *  Called by  ConfSetTargets and ConfAllocate, within ConfDoAllocation, 
 *  to report on allocation progress.  Note that this has the same calling
 *  sequence as ConfAllocRpt() but since it is now used only to report
 *  the progress rather than actual allocation details, most of the arguments
 *  are unused.
 */
 
int ReportRoutine(void *clientData, char Type DUNUSED, 
    char *Description,
    int pivot DUNUSED, int guide DUNUSED, int target DUNUSED,
    long TargetX DUNUSED, long TargetY DUNUSED,
    double theta DUNUSED, double PercentDone)
{
    /* printf ("%s, (%d%%)\n",Description,(int)PercentDone); */
    return 0;
}

/*  ------------------------------------------------------------------------- */

/*                        C o n f  A l l o c  R p t
 *
 *  Called by ConfigAllocate(), within ConfDoAllocation() to indicate 
 *  allocation/deallocation of fibres.  We use the Tcl command AllocRpt to 
 *  do this. Note that it is this mechanism that is responsible for keeping the
 *  configuration data structures updated during an allocation.
 *  This version was taken from configure.c, but changed to use printf
 *  instead of Tcl to output the results. This version assumes the SdsId of
 *  the overall configuration is passed as the client data - which is not how
 *  the TCL version works,
 */
 
void ConfAllocRpt(void *clientData, int Alloc,
    int pivot, int guide, int target,
    long TargetX DUNUSED, long TargetY DUNUSED,
    double theta, StatusType * status)
 {
    SdsIdType Cfig = (SdsIdType) clientData;
    char* Type;
    
    if (*status != STATUS__OK)
        return;

    if (guide) {
        Type = "guide";
    } else {
        Type = "object";
    }

    if (Alloc) {
       AllocateFibre (Cfig,InstFieldDetails,pivot + 1,Type,target + 1);
    } else {
       char TypeString[16];
       int Object;
       DeallocateFibre (Cfig,InstFieldDetails,pivot + 1,TypeString,&Object);
    }
}

/*  ------------------------------------------------------------------------- */

/*                 R p t  T e l  M o d e l  F i l e
 *
 *  This routine is invoked when the instrument support layer opens
 *  its telescope model files.  
 *
 */
 
static void RptTelModelFile(void *argument DUNUSED, const char *string,
    StatusType * status)
{
    if (*status != STATUS__OK) return;
    printf ("Opened telescope model file \"%s\n", string);
}

/*  ------------------------------------------------------------------------- */

/*
 *                     C o n v e r t  X y  { }
 */
 
static int ConvertXy(
    SdsIdType Id)

/*+
 *
 *  Command name:
 *    ConvertXy
 *
 *  Function:
 *    Calculate xy from Ra and Dec
 *  
 *  Call:
 *    ConvertXy id 
 *
 *  Description:
 *    Calculate the x,y corresponding to the RA and Dec values in
 *    a configuration structure, for the current values of the telescope
 *    and refraction parameters.
 *   
 *   Parameters:
 *
 *    (>) id   (Sds id)  Sds Id of configuration structure.
 *
 *   Returns:
 *      "OK" if succesful, "error" if the conversion failed
 *
 *   Support:
 *     Jeremy Bailey, AAO
 *
 *   Version Date:
 *     6-Oct-1994
 *    21-May-2001. Removed epoch and getting of appEpoch. Made gmap a local
 *                 variable. KS.
 *    16-Aug-2001. Now records TelModelParams in TelModelParamsUsed. KS. 
 *    17-Oct-2002. FpilModelCvtInit() now takes two wavelength parameters. KS. 
 *
 *    This code has been modified to run outside a TCL environment, for
 *    the purposes of testharness. 
 *-
 */
{
    SdsIdType id, idf, idtemp;
    double cra, cdec, mjd;
    double gmap[21];                   /* Mean to apparent parameters  */
    StatusType status = STATUS__OK;
    int i;
    wavelensType wlens;                /* Wavelenghts - when specified explictly */

    id = Id;   /* configure.c takes this from argv */
    
    /*
     *  Get field centre details
     */

    SdsFind(id, "fieldData", &idf, &status);
    ArgGetd(idf, "appRa", &cra, &status);
    ArgGetd(idf, "appDec", &cdec, &status);
    ArgGetd(idf, "configMjd", &mjd, &status);
    /*
     *  Get the specific wavelength values
     */
    tdFwlensFromSds(idf, &wlens, &status);

    SdsFreeId(idf, &status);
    
    /*  We sometimes need to know if the parameters we may be about to use
     *  are different to those used the last time the calculation was done.
     *  (This is because if so, it may invalidate any allocation, and the
     *  user may have to confirm this.) So we record the parameters used.
     *  At the moment, we only record the telescope model parameters, but
     *  perhaps we should record everything - this was introduced with
     *  the ARGUS position angle code, where it was important to know if
     *  the position angle had changed from the last time.
     */
   
    for (i = 0; i < TelModelNumParams; i++) {
        TelModelParamsUsed[i] = TelModelParams[i];
    }

    /*
     *  Set up x,y conversion parameters. The wavelength values in this
     *  call shouldn't matter because ConfDoXYConversion() always resets
     *  them to the values appropriate to the fibre type. Indeed, this
     *  FpilModelCvtInit() initialisation call may be unnecessary.
     */

    slaMappa(2000.0, mjd, gmap);

    FpilSetObsWavelength(*Instrument,gwave);
    FpilSetObsWavelength2(*Instrument,gwave2);

    FpilModelCvtInit(TelModel, mjd, 0.0, gtemp, gpress, ghumid, gwave, gwave,
        TelModelParams, &status);

    SdsFind(id, "unallocGuide", &idtemp, &status);
    if (status == SDS__NOITEM) {
        ErsAnnul(&status);
    } else {
        ConfDoXyConversion(idtemp, cra, cdec, mjd, gmap, &wlens, &status);
        SdsFreeId(idtemp, &status);
    }

    if (status == STATUS__OK) {
        SdsFind(id, "unallocObject", &idtemp, &status);
        if (status == SDS__NOITEM) {
            ErsAnnul(&status);
        } else {
            ConfDoXyConversion(idtemp, cra, cdec, mjd, gmap, &wlens, &status);
            SdsFreeId(idtemp, &status);
        }
    }

    if (status == STATUS__OK) {
        SdsFind(id, "objects", &idtemp, &status);
        if (status == SDS__NOITEM) {
            ErsAnnul(&status);
        } else {
            ConfDoXyConversion(idtemp, cra, cdec, mjd, gmap, &wlens, &status);
            SdsFreeId(idtemp, &status);
        }
    }

    if (status == STATUS__OK) {
        return 0;
    } else {
        char buff[200];
        MessGetMsg(status, 0, sizeof(buff), buff);
        printf ("Error in Ra,Dec to x,y conversion\n");
        printf ("%s\n",buff);
        return ERROR_CODE;

    }
}
/*  ------------------------------------------------------------------------- */

/*
 *                     S e t  U T  M e r i d i a n  { }
 */
 
static int SetUTMeridian(SdsIdType Id, double Ha)
/*+
 *
 *  Command name:
 *    SetUTMeridian
 *
 *
 *  Function:
 *    Set the UT of a structure for a given Hour Angle
 *  
 *  Call:
 *    SetUTMeridian id [ha]
 *
 *  Description:
 *    Set the UT in a configuration structure to the value required to
 *    put the field at a specified hour angle for its configuration date.
 *   
 *
 *   Parameters:
 *
 *    (>) id   (Sds id)  Sds Id of configuration structure.
 *    (>) ha   (double)  Required hour angle - if omitted set to the meridian. 
 *
 *
 *   Support:
 *     Jeremy Bailey, AAO
 *
 *   Version Date:
 *     6-Oct-1994
 *    21-May-2001. Now uses ChangeMjd(). KS 
 *
 *    Modified from the configure version to work outside a tcl/tk environment. 
 *-
 */
{
    StatusType status = STATUS__OK;
    SdsIdType id = 0, idf = 0;
    double cra, mjd, lst;
    double longit, lat, hm;
    char oname[NAMELEN];
    char sign[1];
    int ihmsf[4];
    char buff[80];
    double ha;
    int iy, im, idd;
    double f;
    int jstat;
    char buffer[200];
    StatusType ignore = STATUS__OK;


    /*  Original tcl-system code had: - note cnversion to radians
    
    id = atol(argv[1]);
    if (argc < 3)
        ha = 0.0;
    else
        ha = DPI / 12 * (atof(argv[2]));
    */
    /*  New non-Tcl code takes id and ha from passed parameters */
    id = Id;
    ha = Ha;

    /* The discard of const (FpilGetTelescope) is safe as when
       the first argument is -1, this address is not written too
     */
    slaObs(-1, (char *) FpilGetTelescope(*Instrument), oname, &longit, &lat,
        &hm);
    if (strcmp(oname, "?") == 0) {
        printf(
            "Telescope \"%s\" is not recognized by the slaObs routine.\n",
                     FpilGetTelescope(*Instrument));
        return ERROR_CODE;
    }

    SdsFind(id, "fieldData", &idf, &status);
    ArgGetd(idf, "appRa", &cra, &status);
    ArgGetd(idf, "configMjd", &mjd, &status);

    lst = slaGmst(mjd) - longit + slaEqeqx(mjd);
    mjd = mjd + (slaDrange(cra - lst + ha)) / D2PI / 1.0027379;
    ChangeMjd (idf, mjd, &status);
    slaDd2tf(0, mjd - floor(mjd), sign, ihmsf);
    slaDjcl(mjd, &iy, &im, &idd, &f, &jstat);

    if (Ha == 0.0) {
        sprintf(buffer,
            "UT set to %.4d/%.2d/%.2d  %.2d:%.2d:%.2d to put field on meridian",
            iy, im, idd, ihmsf[0], ihmsf[1], ihmsf[2]);
    } else {
        sprintf(buffer,
          "UT set to %.4d/%.2d/%.2d %.2d:%.2d:%.2d to put field at HA %f hours",
            iy, im, idd, ihmsf[0], ihmsf[1], ihmsf[2], Ha);
    }
    printf ("%s\n",buffer);

    SdsFreeId(idf, &ignore);

    if (status == STATUS__OK) {
        return 0;
    } else {
        MessGetMsg(status, 0, sizeof(buff), buff);
        printf("Error setting UT to meridian value\n");
        printf ("%s\n",buff);
        return ERROR_CODE;
    }
}

/*  ------------------------------------------------------------------------- */

/*
 *                     A d d  I n d e x  { }
 */
 
static int AddIndex(SdsIdType Id)

/*+
 *  Command name:
 *    AddIndex
 *
 *  Function:
 *    Add the index and allocated components to a configuration structure
 *  
 *  Call:
 *    AddIndex id 
 *
 *  Description:
 *    Create the "index" component in the allocated objects section
 *    and the "allocated" component in the unallocated objects and
 *    guide sections of a configuration structure, if they do not
 *    already exist.
 *   
 *   Parameters:
 *
 *    (>) id  (Sds Id)  SDS identifier of configuration structure.
 *
 *   Support:
 *     Jeremy Bailey, AAO
 *
 *   Version Date:
 *     6-Oct-1994    
 *-
 */
{
    SdsIdType topid, objid, indid, uid, allid, ra_id;
    StatusType status = STATUS__OK;
    char name[NAMELEN];
    SdsCodeType code;
    long ndims;
    unsigned long dims[7];
    unsigned int i;
    short s;
    char c;
    int indexFound = 0;         /* Indicates if we found the "index" item */


    /* topid = atol(argv[1]);  original TCL */
    topid = Id;

/*
 * Find the objects structure, and look for "index".
 */
    SdsFind(topid, "objects", &objid, &status);
    if (status != STATUS__OK) {
        printf ("Error finding objects structure\n");
        return ERROR_CODE;
    }
    SdsFind(objid, "index", &indid, &status);
    if (status == SDS__NOITEM) {
/*
 *     "index" does not exist.  Create it and initialise the values
 *     to -1.
 */
        ErsAnnul(&status);
        SdsFind(objid, "ra", &ra_id, &status);
        SdsInfo(ra_id, name, &code, &ndims, dims, &status);
        SdsNew(objid, "index", 0, NULL, SDS_SHORT, 1, dims, &indid, &status);
        for (i = 0; i < dims[0]; i++) {
            s = -1;
            SdsPut(indid, sizeof(short), i, &s, &status);
        }
        SdsFreeId(indid, &status);
        SdsFreeId(ra_id, &status);
    } else {
        indexFound = 1;
    }
    SdsFreeId(objid, &status);

/*
 * Find "unallocGuide" and the "allocated" component within.
 */
    SdsFind(topid, "unallocGuide", &uid, &status);
    if (status != STATUS__OK) {
        printf( "Error finding unallocGuide structure\n");
        return ERROR_CODE;
    }
    SdsFind(uid, "allocated", &allid, &status);
    if ((status == SDS__NOITEM) && (!indexFound)) {
/*
 *     "allocated" does not exist.  Create it and initialise to 0, indicating
 *     objects are not allocated.
 */
        ErsAnnul(&status);
        SdsFind(uid, "ra", &ra_id, &status);
        SdsInfo(ra_id, name, &code, &ndims, dims, &status);
        SdsNew(uid, "allocated", 0, NULL, SDS_CHAR, 1, dims, &allid, &status);
        for (i = 0; i < dims[0]; i++) {
            c = 0;
            SdsPut(allid, sizeof(char), i, &c, &status);
        }
        SdsFreeId(allid, &status);
        SdsFreeId(ra_id, &status);
    }
    SdsFreeId(uid, &status);
/*
 * Find "unallocObject" and the "allocated" component within.
 */
    SdsFind(topid, "unallocObject", &uid, &status);
    if (status != STATUS__OK) {
        printf ("Error finding unallocObject structure\n");
        return 0;
    }
    SdsFind(uid, "allocated", &allid, &status);
    if ((status == SDS__NOITEM) && (!indexFound)) {
/*
 *     "allocated" does not exist.  Create it and initialise to 0, indicating
 *     objects are not allocated.
 */ 
        ErsAnnul(&status);
        SdsFind(uid, "ra", &ra_id, &status);
        SdsInfo(ra_id, name, &code, &ndims, dims, &status);
        SdsNew(uid, "allocated", 0, NULL, SDS_CHAR, 1, dims, &allid, &status);
        for (i = 0; i < dims[0]; i++) {
            c = 0;
            SdsPut(allid, sizeof(char), i, &c, &status);
        }
        SdsFreeId(allid, &status);
        SdsFreeId(ra_id, &status);
    }
    SdsFreeId(uid, &status);

/*
 * If the index and allocated items were not found in the original
 * structure, we need to set them up appropiately.
 */
    if (!indexFound)
        ConfMakeAllocatedIndex(topid, &status);

    if (status == STATUS__OK) {
        return 0;
    } else {
        char buff[200];
        MessGetMsg(status, 0, sizeof(buff), buff);
        printf ("Error Making Allocated Objects Structure\n");
        printf ("%s\n",buff);
        return ERROR_CODE;

    }
}

/*  ------------------------------------------------------------------------- */

/*
 *                     A l l o c a t e  F i b r e  { }
 *
 *   Based on ConfAllocateFibre from configure.c, but modified to run
 *   outside a Tcl environment.
 */
 
static int AllocateFibre(
   SdsIdType Id,
   SdsIdType PivId,
   int Fibre,
   char* Type,
   int Object)


/*+
 *  Command name:
 *    AllocateFibre
 *
 *  Function:
 *    Allocate an object to a fibre
 *  
 *  Call:
 *    AllocateFibre id pivid fibre type object
 *
 *  Description:
 *    Allocate an object to a fibre. This command checks that the fibre
 *    and object are not already allocated, and if not copies the object
 *    information to the allocated objects section of the structure. It sets
 *    the index entry for the fibre to point to the object location in the
 *    unallocated structure, and it sets the allocated flag in that structure.
 *
 *    The command does not check the validity of the resulting configuration.
 *    That can be done by a call to DoCheck.
 *   
 *   Parameters:
 *
 *    (>) id  (Sds Id)   SDS identifier of configuration structure.
 *    (>) pivid (Sds Id) SDS identifier of the pivots structure.
 *    (>) fibre (int)    Fibre (pivot) number - starting from one.
 *    (>) type  (char)   "object" or "guide".
 *    (>) object (int)   Object number - starting from one.
 *
 *   Returns:
 *     "OK" if succesful, "Error" otherwise
 *
 *   Support:
 *     Jeremy Bailey, AAO
 *
 *   Version Date:
 *      6-Oct-1994  Original version. JAB.
 *      8-Nov-2001  Now updates fibre type allocation statistics. KS.
 *      3-Sep-2002  Now updates available target allocation statistics. KS.   
 *-
 */
{
    SdsIdType topid, objid, uid, pivid, fid;
    long fibre_no;
    long obj_no;
    StatusType status = STATUS__OK;
    short s;
    char c;
    short IsSky;
    char guide;
    int x, y, xp, yp;
    double theta;
    int fibreType;
    /* int PivIndex;  commented out for testharness */
    int NumPivTypes;
    int PivTypes[10];

    /*  Original Tcl code 
    topid = (SdsIdType) (atol(argv[1]));
    pivid = (SdsIdType) (atol(argv[2]));
    fibre_no = atol(argv[3]) - 1;
    obj_no = atol(argv[5]) - 1;
    */
    topid = Id;
    pivid = PivId;
    fibre_no = Fibre - 1;
    obj_no = Object - 1;
    
    fibreType = ConfWhichSpec((int) fibre_no, 0);

    if ((fibre_no < 0) || (fibre_no >= (int) FpilGetNumPivots(*Instrument))) {
        printf (
            "Illegal Pivot Number %ld\nMust be between 1 and %d\n",
            fibre_no + 1, FpilGetNumPivots(*Instrument));
        return ERROR_CODE;
    } else if (!UseFibreType[fibreType]) {
        const char *FibreTypeDescr;
        ConfWhichSpec((int) fibre_no, &FibreTypeDescr);
        printf(
            "Attempt to allocate pivot %ld, with a fibre type (%s)",
            fibre_no + 1, FibreTypeDescr);
        printf (" whose allocation is disabled\n");
        return ERROR_CODE;
    } else if (strcmp(Type, "object") == 0) {
        SdsFind(topid, "unallocObject", &uid, &status);
        ConfSdsCheck("Error finding unallocObject structure", &status);
        guide = 0;
    } else if (strcmp(Type, "guide") == 0) {
        SdsFind(topid, "unallocGuide", &uid, &status);
        ConfSdsCheck("Error finding unallocGuide structure", &status);
        guide = 1;
    } else {
        printf ("Type must be object or guide\n");
        return ERROR_CODE;
    }
/*      This DEBUG section provides additional details about the object.
    {
        int x, y;
        SdsIdType tid;
        char string[128];
        unsigned long actlen;
        ConfArrayGeti(uid, "x", obj_no, &x, &status);
        ConfArrayGeti(uid, "y", obj_no, &y, &status);
        SdsFind(uid, "name", &tid, &status);
        SdsGet(tid, 80, 80 * obj_no, string, &actlen, &status);
        SdsFreeId(tid, &status);

        DEBUG("AllocateFibre:Object %d (%s) position is at %d,%d\n",
           obj_no,string,x,y);

    }
*/
    SdsFind(topid, "objects", &objid, &status);
    ConfSdsCheck("Error finding objects structure", &status);

/*  Check fibre is not already allocated  */

    ConfArrayGetc(objid, "type", fibre_no, &c, &status);
    if (c != FpilUnallocatedType(*Instrument)) {

        printf(
            "Fibre at pivot number %ld is already allocated\n", fibre_no + 1);
        return ERROR_CODE;
    }

/*  Check fibre is useable  */

    ConfArrayGets(pivid, "inUse", fibre_no, &s, &status);
    if (s != 1 && ImportOnlyMode == 0) {
        printf (
            "Fibre at pivot number %ld is not useable\n", fibre_no + 1);

        return ERROR_CODE;
    }

/*  Check object is not already allocated  */

    ConfArrayGetc(uid, "allocated", obj_no, &c, &status);
    if (c != 0) {
        printf ("Object %ld is already allocated\n", obj_no + 1);
        return ERROR_CODE;
    }

/*  Copy data from the unallocated object section to the objects
    section of the structure  */

    ConfArrayCopy(uid, objid, "name", NAMELEN, NAMELEN * obj_no,
        NAMELEN * fibre_no, &status);
    ConfArrayCopy(uid, objid, "ra", sizeof(double), obj_no, fibre_no, &status);
    ConfArrayCopy(uid, objid, "dec", sizeof(double), obj_no, fibre_no, &status);


    ConfArrayCopy(uid, objid, "x", sizeof(int), obj_no, fibre_no, &status);
    ConfArrayCopy(uid, objid, "y", sizeof(int), obj_no, fibre_no, &status);

    ConfArrayCopy(uid, objid, "type", sizeof(char), obj_no, fibre_no, &status);

    ConfArrayCopy(uid, objid, "priority", sizeof(short), obj_no, fibre_no,
        &status);

    ConfArrayCopy(uid, objid, "magnitude", sizeof(double), obj_no, fibre_no,
        &status);

    ConfArrayCopy(uid, objid, "pId", sizeof(short), obj_no, fibre_no, &status);

    ConfArrayCopy(uid, objid, "comment", CMTLEN, CMTLEN * obj_no,
        CMTLEN * fibre_no, &status);

    if (guide) {
        ConfArrayCopy(uid, objid, "dhdx", sizeof(double), obj_no, fibre_no,
            &status);
        ConfArrayCopy(uid, objid, "dhdy", sizeof(double), obj_no, fibre_no,
            &status);
        ConfArrayCopy(uid, objid, "dddx", sizeof(double), obj_no, fibre_no,
            &status);
        ConfArrayCopy(uid, objid, "dddy", sizeof(double), obj_no, fibre_no,
            &status);
    }

/*  Calculate theta so that the button is parallel to the fibre  */

    ConfArrayGeti(pivid, "xPiv", fibre_no, &xp, &status);
    ConfArrayGeti(pivid, "yPiv", fibre_no, &yp, &status);
    ConfArrayGeti(objid, "x", fibre_no, &x, &status);
    ConfArrayGeti(objid, "y", fibre_no, &y, &status);

    theta = slaDranrm(atan2((double) (yp - y), (double) (xp - x)) - DPIBY2);

    ConfArrayPutd(objid, "theta", fibre_no, theta, &status);
    
/* We need to update the allocation statistics held for this fibre type */

    ConfArrayGetc(objid, "type", fibre_no, &c, &status);
    IsSky = FpilIsTargetSky(*Instrument, c);
    /*  - for the testharness, we aren't interested in the fibre type statistics
    if ((fibreType >= 0) && (fibreType < InstNumFibreTypes)) {
        FibreTypeStats[fibreType].Unallocated--;
        if (FpilIsTargetGuide(*Instrument, c)) {
            FibreTypeStats[fibreType].AllocToGuide++;
        } else if (IsSky) {
            FibreTypeStats[fibreType].AllocToSky++;
        } else {
            FibreTypeStats[fibreType].AllocToObject++;
        }
    }
    */
    
/* Also, given this target type (in 'c') we need to calculate how this
 * allocation affects the possible sky and object targets for the different
 * fibre types - it might have been a target that could be allocated to a
 * number of different fibre types
 */
 
    FpilCompatiblePivotTypes (*Instrument,10,c,0,&NumPivTypes,PivTypes);
    /*  - for the testharness, we aren't interested in the fibre type statistics
    for (PivIndex = 0; PivIndex < NumPivTypes; PivIndex++) {
        fibreType = PivTypes[PivIndex];
        if (IsSky) {
            FibreTypeStats[fibreType].Skies--;
        } else {
            FibreTypeStats[fibreType].Targets--;
        }
    }
    */

/* We now need to update the field data structure */

    SdsFind(topid, "fieldData", &fid, &status);
    ConfSdsCheck("Error finding fieldData structure", &status);


    if (status == STATUS__OK) {
        if (guide) {
            ArgGets(fid, "unallocGui", &s, &status);
            s--;
            ArgPuts(fid, "unallocGui", s, &status);
            ArgGets(fid, "allocGui", &s, &status);
            if (status == SDS__NOITEM) {
                ErsAnnul(&status);
                s = 1;
            } else
                s++;
            ArgPuts(fid, "allocGui", s, &status);
        } else {
            ConfArrayGetc(objid, "type", fibre_no, &c, &status);
            if (!FpilIsTargetGuide(*Instrument, c) &&
                !FpilIsTargetSky(*Instrument, c)) {
                ArgGets(fid, "unallocObj", &s, &status);
                s--;
                ArgPuts(fid, "unallocObj", s, &status);
                ArgGets(fid, "allocObj", &s, &status);
                if (status == SDS__NOITEM) {
                    ErsAnnul(&status);
                    s = 1;
                } else
                    s++;
                ArgPuts(fid, "allocObj", s, &status);
            } else if (FpilIsTargetSky(*Instrument, c)) {
                char name[20];

                /* Update unallocSky and allocSky */
                
                ArgGets(fid, "unallocSky", &s, &status);
                ArgPuts(fid, "unallocSky", s - 1, &status);
                ArgGets(fid, "allocSky", &s, &status);
                if (status == SDS__NOITEM) {
                    ErsAnnul(&status);
                    s = 1;
                } else
                    s++;
                ArgPuts(fid, "allocSky", s, &status);

                /* Update allocSkyn where n = fibre type number */
                
                sprintf(name, "allocSky%d", fibreType);
                ArgGets(fid, "name", &s, &status);
                if (status == SDS__NOITEM) {
                    ErsAnnul(&status);
                    s = 1;
                } else
                    s++;
                ArgPuts(fid, name, s, &status);

            }
        }
    }
    ConfArrayPuts(objid, "index", fibre_no, obj_no, &status);
    ConfArrayPutc(uid, "allocated", obj_no, (char) 1, &status);
    SdsFreeId(objid, &status);
    SdsFreeId(uid, &status);
    SdsFreeId(fid, &status);

    if (status == STATUS__OK) {
        return 0;
    } else {
        char buff[200];
        printf(
            "Error Allocating Fibre %ld\n", fibre_no + 1);
        MessGetMsg(status, 0, sizeof(buff), buff);
        printf ("%s\n",buff);
        return ERROR_CODE;

    }

}

/*  ------------------------------------------------------------------------- */

/*
 *                    D e a l l o c a t e  F i b r e  { }
 */
 
static int DeallocateFibre(
   SdsIdType Id,
   SdsIdType PivId,
   int Fibre,
   char* Type,
   int* Object)

/*+
 *  Command name:
 *    DeAllocateFibre
 *
 *  Function:
 *    Deallocate a fibre
 *  
 *  Call:
 *    DeAllocateFibre id pivid fibre type object
 *
 *  Description:
 *    Deallocate a fibre. Return the index of the deallocated object
 *    in the unallocated objects or guide structure. 
 *
 *    If the fibre has an index entry set, this operation simply
 *    involves removing the index entry and unsetting the allocated
 *    flag for the object. However if there is no index entry, it
 *    is necessary to enlarge the unallocated objects structure and
 *    copy the object data into the new entry.
 *
 *    In either case the fibre position is set back to the parked
 *    position.
 *   
 *   Parameters:
 *
 *    (>) id  (Sds Id)   SDS identifier of configuration structure.
 *    (>) pivid (Sds Id) SDS identifier of the pivots structure.
 *    (>) fibre (int)    Fibre (pivot) number - starting from one.
 *    (<) type  (char)   "uobj" or "uguide".
 *    (<) object (int)   Object number - starting from one.
 *
 *   Returns:
 *     "OK" if succesful, "Error" otherwise
 *
 *   Support:
 *     Jeremy Bailey, AAO
 *
 *   Version Date:
 *      6-Oct-1994  Original version. JAB.    
 *      8-Nov-2001  Now updates fibre type allocation statistics. KS.   
 *      3-Sep-2002  Now updates available target allocation statistics. KS.   
 *-
 */
{
    SdsIdType topid, objid, uid, pivid, fid;
    long fibre_no;
    int fibreType;
    long obj_no;
    StatusType status = STATUS__OK;
    short s;
    char c;
    char guide;
    SdsIdType xpid, ypid, tpid;
    char name[NAMELEN];
    SdsCodeType code;
    long ndims;
    unsigned long dims[7];
    int l;
    double d;
    SdsIdType ra_id;
    unsigned long actlen;
    char ctype;
    /* char buf[20]; only needed to set Tcl variables */
    int spec;
    int NumPivTypes;
    int PivTypes[10];
    short IsSky;
    int PivIndex;

    /*  Original Tcl code 
    topid = (SdsIdType) (atol(argv[1]));
    pivid = (SdsIdType) (atol(argv[2]));
    fibre_no = atol(argv[3]) - 1;
    */
    topid = Id;
    pivid = PivId;
    fibre_no = Fibre - 1;
    
    spec = ConfWhichSpec((int) fibre_no, 0);

    if ((fibre_no < 0) || (fibre_no >= (int) FpilGetNumPivots(*Instrument))) {
        printf (
            "Illegal Pivot Number %ld\nMust be between 1 and %d\n",
            fibre_no + 1, FpilGetNumPivots(*Instrument));
        return ERROR_CODE;
    } else if (!UseFibreType[spec]) {
        if (spec == 0) {

            printf (
                "Attempt to deallocate guide pivot %ld,\n", fibre_no + 1);
            printf (
                " when guide pivot deallocation is disabled\n");
        } else {
            const char *SpecDescr;
            ConfWhichSpec((int) fibre_no, &SpecDescr);
            printf (
                "Attempt to deallocate pivot %ld, from spectrograph (%s)\n",
                fibre_no + 1, SpecDescr);
            printf ( " marked as not to be deallocated.\n");
        }
        return ERROR_CODE;
    }


    SdsFind(topid, "objects", &objid, &status);
    ConfSdsCheck("Error finding objects structure", &status);

/*  Check fibre is allocated  */

    ConfArrayGetc(objid, "type", fibre_no, &ctype, &status);

    if (status != STATUS__OK) {
        char buff[200];
        MessGetMsg(status, 0, sizeof(buff), buff);
        ErsFlush(&status);  /* Must do this before setting result */
        printf ("Error getting fibre details\n");
        printf ("%s\n",buff);
        return ERROR_CODE;
    }


    if (ctype == FpilUnallocatedType(*Instrument)) {
        printf(
            "Fibre at pivot number %ld is not allocated\n", fibre_no + 1);
        return ERROR_CODE;
    }

    if (FpilIsTargetGuide(*Instrument, ctype)) {
        guide = 1;
        SdsFind(topid, "unallocGuide", &uid, &status);
        strcpy (Type,"uguide");
    } else {
        guide = 0;
        SdsFind(topid, "unallocObject", &uid, &status);
        strcpy (Type,"uobj");
    }

/*  Get fibre index  */

    ConfArrayGets(objid, "index", fibre_no, &s, &status);
    if (s != -1) {

/*  The object has an index specified - This is the easy case.
    We just unset the allocated flag  */
        obj_no = s;
        c = 0;
        ConfArrayPutc(uid, "allocated", obj_no, c, &status);
        /*
        sprintf(buf, "%ld", obj_no + 1);
        Tcl_SetVar(interp, argv[5], buf, 0);
        */
        *Object = obj_no + 1;
    }

/*  Otherwise we have to enlarge the unallocated objects section
    and copy the data back to it  */

    else {

        printf("Enlarging unalloctated objects section ");
        printf("should not be needed now)\n");

        SdsFind(uid, "ra", &ra_id, &status);
        SdsInfo(ra_id, name, &code, &ndims, dims, &status);


        SdsFreeId(ra_id, &status);
        ConfResizeAllItems(uid, dims[0] + 1, &status);
        obj_no = dims[0];
        /*
        sprintf(buf, "%ld", obj_no + 1);
        Tcl_SetVar(interp, argv[5], buf, 0);
        */
        *Object = obj_no + 1;

        ConfArrayCopy(objid, uid, "name", NAMELEN, NAMELEN * fibre_no,
            NAMELEN * obj_no, &status);
        ConfArrayCopy(objid, uid, "ra", sizeof(double), fibre_no, obj_no,
            &status);
        ConfArrayCopy(objid, uid, "dec", sizeof(double), fibre_no, obj_no,
            &status);
        ConfArrayCopy(objid, uid, "x", sizeof(int), fibre_no, obj_no, &status);
        ConfArrayCopy(objid, uid, "y", sizeof(int), fibre_no, obj_no, &status);
        ConfArrayCopy(objid, uid, "type", sizeof(char), fibre_no, obj_no,
            &status);
        ConfArrayCopy(objid, uid, "priority", sizeof(short), fibre_no, obj_no,
            &status);
        ConfArrayCopy(objid, uid, "magnitude", sizeof(double), fibre_no, obj_no,
            &status);
        ConfArrayCopy(objid, uid, "pId", sizeof(short), fibre_no, obj_no,
            &status);
        ConfArrayCopy(objid, uid, "comment", CMTLEN, CMTLEN * fibre_no,
            CMTLEN * obj_no, &status);
        if (guide) {
            ConfArrayCopy(objid, uid, "dhdx", sizeof(double), fibre_no,
                obj_no, &status);
            ConfArrayCopy(objid, uid, "dhdy", sizeof(double), fibre_no,
                obj_no, &status);
            ConfArrayCopy(objid, uid, "dddx", sizeof(double), fibre_no,
                obj_no, &status);
            ConfArrayCopy(objid, uid, "dddy", sizeof(double), fibre_no,
                obj_no, &status);
        } else
            ConfArrayCopy(objid, uid, "spectrograph", sizeof(char),
                fibre_no, obj_no, &status);
        ConfArrayPutc(uid, "allocated", obj_no, 0, &status);

    }

/*  Now set the objects section entry back to the parked position */

    SdsFind(pivid, "xPark", &xpid, &status);
    SdsFind(pivid, "yPark", &ypid, &status);
    SdsFind(pivid, "tPark", &tpid, &status);

    strcpy(name, "Parked");
    ConfArrayPutString(objid, "name", fibre_no, name, &status);
    ConfArrayPutd(objid, "ra", fibre_no, 0.0, &status);
    ConfArrayPutd(objid, "dec", fibre_no, 0.0, &status);
    ConfArrayPuts(objid, "fibre", fibre_no, 0, &status);
    SdsGet(xpid, sizeof(int), fibre_no, &l, &actlen, &status);
    ConfArrayPuti(objid, "x", fibre_no, l, &status);
    SdsGet(ypid, sizeof(int), fibre_no, &l, &actlen, &status);
    ConfArrayPuti(objid, "y", fibre_no, l, &status);
    SdsGet(tpid, sizeof(double), fibre_no, &d, &actlen, &status);
    ConfArrayPutd(objid, "theta", fibre_no, d, &status);
    ConfArrayPutc(objid, "type", fibre_no,
        FpilUnallocatedType(*Instrument), &status);
    ConfArrayPuts(objid, "priority", fibre_no, 1, &status);
    ConfArrayPutd(objid, "magnitude", fibre_no, 0.0, &status);
    ConfArrayPutc(objid, "spectrograph", fibre_no, ' ', &status);
    ConfArrayPuts(objid, "pId", fibre_no, 1, &status);
    ConfArrayPutd(objid, "dhdx", fibre_no, 0.0, &status);
    ConfArrayPutd(objid, "dhdy", fibre_no, 0.0, &status);
    ConfArrayPutd(objid, "dddx", fibre_no, 0.0, &status);
    ConfArrayPutd(objid, "dddy", fibre_no, 0.0, &status);
    ConfArrayPutString(objid, "comment", fibre_no, name, &status);
    ConfArrayPuts(objid, "index", fibre_no, -1, &status);

    SdsFreeId(xpid, &status);
    SdsFreeId(ypid, &status);
    SdsFreeId(tpid, &status);

/* We need to update the allocation statistics held for this fibre type */

    fibreType = ConfWhichSpec((int) fibre_no, 0);
    IsSky = FpilIsTargetSky(*Instrument, ctype);
    /*  Disabled for the testharness case 
    if ((fibreType >= 0) && (fibreType < InstNumFibreTypes)) {
        FibreTypeStats[fibreType].Unallocated++;
        if (FpilIsTargetGuide(*Instrument, ctype)) {
            FibreTypeStats[fibreType].AllocToGuide--;
        } else if (IsSky) {
            FibreTypeStats[fibreType].AllocToSky--;
        } else {
            FibreTypeStats[fibreType].AllocToObject--;
        }
    }
    */
    
/* Also, given this target type (in 'ctype') we need to calculate how this
 * deallocation affects the possible sky and object targets for the different
 * fibre types - it might be a target that can be allocated to a
 * number of different fibre types
 */
 
    FpilCompatiblePivotTypes (*Instrument,10,ctype,0,&NumPivTypes,PivTypes);
    for (PivIndex = 0; PivIndex < NumPivTypes; PivIndex++) {
        fibreType = PivTypes[PivIndex];
        /*
        if (IsSky) {
            FibreTypeStats[fibreType].Skies++;
        } else {
            FibreTypeStats[fibreType].Targets++;
        }
        */
    }
    
    if (status == STATUS__OK) {
        SdsFind(topid, "fieldData", &fid, &status);
        ConfSdsCheck("Error finding fieldData structure", &status);
    }
    if (status == STATUS__OK) {
        if (guide) {
            ArgGets(fid, "unallocGui", &s, &status);
            s++;
            ArgPuts(fid, "unallocGui", s, &status);
            ArgGets(fid, "allocGui", &s, &status);
            if (status == SDS__NOITEM) {
                ErsAnnul(&status);
                s = 1;
            } else
                s--;
            ArgPuts(fid, "allocGui", s, &status);
        } else {
            if (!FpilIsTargetGuide(*Instrument, ctype) &&
                !FpilIsTargetSky(*Instrument, ctype)) {
                /*  Not sky and not a guide object, so this is a 'program'
                 *  object.
                 */

                ArgGets(fid, "unallocObj", &s, &status);
                s++;
                ArgPuts(fid, "unallocObj", s, &status);
                ArgGets(fid, "allocObj", &s, &status);
                if (status == SDS__NOITEM) {
                    ErsAnnul(&status);
                    s = 1;
                } else
                    s--;
                ArgPuts(fid, "allocObj", s, &status);
            } else if (FpilIsTargetSky(*Instrument, ctype)) {
                char name[20];

                /* Update unallocSky and allocSky */
                ArgGets(fid, "unallocSky", &s, &status);
                ArgPuts(fid, "unallocSky", s + 1, &status);
                ArgGets(fid, "allocSky", &s, &status);
                if (status == SDS__NOITEM) {
                    ErsAnnul(&status);
                    s = 0;
                } else
                    s--;
                ArgPuts(fid, "allocSky", s, &status);

                /* Update allocSkyn where n = spectrograph number */
                sprintf(name, "allocSky%d", spec);
                ArgGets(fid, "name", &s, &status);
                if (status == SDS__NOITEM) {
                    ErsAnnul(&status);
                    s = 0;
                } else
                    s--;
                ArgPuts(fid, name, s, &status);
            }
        }
    }
    SdsFreeId(objid, &status);
    SdsFreeId(uid, &status);
    SdsFreeId(fid, &status);

    if (status == STATUS__OK) {
        return 0;
    } else {

        char buff[200];
        printf ("Error Deallocating Fibre %ld\n", fibre_no + 1);

        MessGetMsg(status, 0, sizeof(buff), buff);
        printf ("%s\n",buff);
        return ERROR_CODE;
    }

}


/*  ------------------------------------------------------------------------- */

/*
 *                 C o m p r e s s  U n a l l o c  { }
 */
 
static int CompressUnalloc(
   SdsIdType Id)

/*+
 *  Command name:
 *    CompressUnalloc
 *
 *  Function:
 *    Compress the Unallocated objects section
 *  
 *  Call:
 *    CompressUnalloc id
 *
 *  Description:
 *    Compress the unallocated objects and guide sections of a 
 *    configuration structure by removing any objects which have
 *    their "allocated" flag set. Also removes the corresponding "index"
 *    section from the allocated objects structure.
 *   
 *
 *   Parameters:
 *
 *    (<) id  (Sds Id)   SDS id of configuration structure
 *
 *   Support:
 *     Jeremy Bailey, AAO
 *
 *   Version Date:
 *     6-Oct-1994 
 *
 *   Modified:
 *     20th Jan 2003. KS. Failure to find "index" no longer treated
 *                        as an error - just indicates this routine
 *                        has already been run on this structure.   
 *-
 *  Modified from the TCL version in configure.c
 */
{
    SdsIdType topid, uid, objid, indid;
    StatusType status = STATUS__OK;
    StatusType indexStatus = STATUS__OK;

    topid = Id;
    SdsFind(topid, "unallocObject", &uid, &status);
    ConfRemoveAlloc(uid, &status);
    SdsFreeId(uid, &status);
    SdsFind(topid, "unallocGuide", &uid, &status);
    ConfRemoveAlloc(uid, &status);
    SdsFreeId(uid, &status);
    SdsFind(topid, "objects", &objid, &status);
    SdsFind(objid, "index", &indid, &indexStatus);
    if (indexStatus == STATUS__OK) {
        SdsDelete(indid, &status);
        SdsFreeId(indid, &status);
    }
    SdsFreeId(objid, &status);

    if (status == STATUS__OK) {
        return 0;
    } else {
        printf (
            "Error Compressing the Unallocated Object Structures\n");
        return ERROR_CODE;
    }

}

/*  ------------------------------------------------------------------------- */

/*                        L i s t  P a r a m e t e r s
 *
 */
 
void ListParameters (void)
{
   StatusType Status;
   int I,IVal,Np,NValues;
   char *Name;
   char Type;
   char **Values;
   char Value[80];
   double Limits[2];
   
   Status = 0;
   Np = ConfigNumberParameters();
   for (I = 1; I <= Np; I++) {
      ConfigParameterDetails (I,&Name,&Type,Limits,&NValues,&Values,&Status);
      if (Status == 0) {
         printf ("Parameter '%s'",Name);
         if (Type == 'L') {
            printf (" is logical\n");
         } else if (Type == 'I') {
            printf (" is an integer between %d and %d\n",
                            (int) Limits[0], (int) Limits[1]);
         } else if (Type == 'F') {
            printf (" is a real number between %f and %f\n",
                                                Limits[0],Limits[1]);
         } else if (Type == 'C') {
            printf (" is a character parameter, with ");
            if (NValues == 0) {
               printf (" unrestricted values\n");
            } else {
               printf (" the following possible values:\n");
               for (IVal = 0; IVal < NValues; IVal++) {
                  printf ("'%s'\n",Values[IVal]);
               }
            }
         }
         ConfigGetParameter(Name,Value,sizeof(Value),&Status);
         if (Status == 0) {
            printf ("Current value is '%s'\n",Value);
         }
      }
   }
   
}

/*  ------------------------------------------------------------------------- */

/*
 *                  M a k e  L i s t i n g  { }
 */

static int ConfMakeListing(
   SdsIdType Id,
   char* File,
   char* Type)

/*+
 *  Command Name:
 *    MakeListing
 *
 *  Function:
 *    Output listing of objects and allocations
 *
 *  Call:
 *    MakeListing id file type
 *
 *  Description:
 *    Output a text file listing of the allocations or objects
 *    or both in an Sds structure. The format of the listing of
 *    unallocated objects is the same as the input format for
 *    the configuration. If the allocations are listed this listing takes
 *    the form of comments in the configuration file so the file will
 *    still be a valid configuration input file. The unalocated objects
 *    listing includes all fiducial stars and sky targets whether allocated
 *    or not.
 *
 *
 *  Parameters:
 *    (>)  id    (Sds Id)  Sds identifier of the structure
 *    (>)  file  (char)    Name of the file to be created
 *    (>)  type  (char)    Information to output
 *                   alloc  =>  List Allocations
 *                   obj    =>  List unallocated objects
 *                   all    =>  List unallocated objects and allocations
 *
 *   Returns:
 *     "Error" - if the file could not be opened.
 *
 *   Support:
 *     Jeremy Bailey, AAO
 *
 *   Version Date:
 *     6-Oct-1994
 *    26-Nov-2001 Now includes allocated sky targets with unallocated 
 *                objects. KS.    
 *    25-Jul-2006  Add the new wlenIndex_id argument for ConfGetIds,
 *                          althrough it is not used. TJF
 *    10-Aug-2009 Added proper motion arguments for ConfGetIds and for
 *                ConfListLine. KS.
 *-
 *  Modified from the TCL version in configure.c
 */
{
    SdsIdType topid, objid, uid, fid;
    StatusType status;
    short alloc_array;
    FILE *listing;
    double mjd;
    int iy, im, id;
    double fd;
    int jstat;
    double ra, dec;
    int rhmsf[4];
    int ddmsf[4];
    char sign[1];
    char buf[80];
    char iname[16];
    SdsCodeType code;
    long ndims;
    unsigned long dims[7];
    SdsIdType ra_id, dec_id, name_id, type_id, spect_id, prio_id, mag_id,
        pid_id, comment_id, wlenIndex_id, pmra_id, pmdec_id, temp_id;
    unsigned int i;
    char mbuf[100];
    char c, s;

    /*  Original Tcl code
    if (argc != 4) {
        interp->result = "wrong # args";
        return ERROR_CODE;
    }
    status = STATUS__OK;
    topid = (SdsIdType) (atol(argv[1]));
    */
    topid = Id;
    status = STATUS__OK;
    SdsFind(topid, "fieldData", &fid, &status);
    /*  Original Tcl code
    listing = fopen(argv[2], "w");
    if (listing == NULL) {
        ErsOut(0, &status, "Error Opening File %s", argv[2]);
        interp->result = "Error";
        return TCL_OK;
    }
    */
    listing = fopen(File, "w");
    if (listing == NULL) {
        ErsOut(0, &status, "Error Opening File %s", File);
        return 0;
    }
    if (Instrument == NULL) {
        ErsOut(0, &status, "No instrument selected");
        return 0;
    }

/*  Output the Header Section  */

    ArgGetString(fid, "label", sizeof(buf), buf, &status);
    fprintf(listing, "LABEL %s\n", buf);
    ArgGetd(fid, "configMjd", &mjd, &status);
    slaDjcl(mjd, &iy, &im, &id, &fd, &jstat);
    fprintf(listing, "UTDATE %4.4d %2.2d %2.2d\n", iy, im, id);
    ArgGetd(fid, "cenRa", &ra, &status);
    ArgGetd(fid, "cenDec", &dec, &status);
    slaDr2tf(3, ra, sign, rhmsf);
    slaDr2af(2, dec, sign, ddmsf);
    fprintf(listing,
        "CENTRE %2.2d %2.2d %2.2d.%3.3d  %c%2.2d %2.2d %2.2d.%2.2d\n",
        rhmsf[0], rhmsf[1], rhmsf[2], rhmsf[3], sign[0], ddmsf[0], ddmsf[1],
        ddmsf[2], ddmsf[3]);
    if (strcmp(FpilGetInstName(*Instrument), "FLAMES")) {
        fprintf(listing, "EQUINOX 2000.0\n");
    }

    tdFwlensSdsToListingHeader(fid, listing, &status);

    /*  Original Tcl code 
    if ((strcmp(argv[3], "alloc") == 0) || (strcmp(argv[3], "all") == 0)) {
    */
    if ((strcmp(Type, "alloc") == 0) || (strcmp(Type, "all") == 0)) {
        SdsFind(topid, "objects", &objid, &status);
        ConfGetIds(objid, &ra_id, &dec_id, &name_id, &type_id, &spect_id,
                   &prio_id, &mag_id, &pid_id, &comment_id,  
                   &wlenIndex_id, &pmra_id, &pmdec_id, &status);
        SdsInfo(ra_id, iname, &code, &ndims, dims, &status);
        fprintf(listing, "*\n");
        fprintf(listing, "*  Fibre Allocations\n");
        fprintf(listing, "*\n");
        fprintf(listing, "*%5s%12s%12s%12s\n", "Fibre", "Name", "R.A.", "Dec");
        for (i = 0; i < dims[0]; i++) {
            fprintf(listing, "*%4d ", i + 1);
            ConfListLine(listing, i, ra_id, dec_id, name_id, type_id, spect_id,
                         prio_id, mag_id, pid_id, comment_id, wlenIndex_id, 
                         pmra_id, pmdec_id, &status);
            if (status != STATUS__OK)
                break;
        }
        SdsFreeId(name_id, &status);
        SdsFreeId(ra_id, &status);
        SdsFreeId(dec_id, &status);
        SdsFreeId(type_id, &status);
        if (spect_id)
            SdsFreeId(spect_id, &status);
        SdsFreeId(prio_id, &status);
        SdsFreeId(mag_id, &status);
        SdsFreeId(pid_id, &status);
        SdsFreeId(comment_id, &status);
        SdsFreeId(objid, &status);
        if (wlenIndex_id)
            SdsFreeId(wlenIndex_id, &status);
    }
    /*  Original Tcl code 
    if ((strcmp(argv[3], "obj") == 0) || (strcmp(argv[3], "all") == 0)) {
    */
    if ((strcmp(Type, "obj") == 0) || (strcmp(Type, "all") == 0)) {
        SdsFind(topid, "unallocObject", &uid, &status);
        ConfGetIds(uid, &ra_id, &dec_id, &name_id, &type_id, &spect_id,
            &prio_id, &mag_id, &pid_id, &comment_id,  &wlenIndex_id, 
            &pmra_id, &pmdec_id, &status);
        SdsInfo(ra_id, iname, &code, &ndims, dims, &status);
        fprintf(listing, "*\n");
        fprintf(listing, "*  Unallocated Objects\n");
        fprintf(listing, "*\n");
        fprintf(listing, "*%12s%12s%12s\n", "Name", "R.A.", "Dec");
        alloc_array = TRUE;
        SdsFind (uid,"allocated",&temp_id,&status);
        if (status != 0) {
            alloc_array = FALSE;
            status = 0;
        } else {
            SdsFreeId (temp_id, &status);
        }
        for (i = 0; i < dims[0]; i++) {
            if (alloc_array) {
                ConfArrayGetc(uid, "allocated", i, &c, &status);
            } else {
                c = 0;
            }
            if (c == 0) {
                ConfListLine(listing, i, ra_id, dec_id, name_id, type_id,
                             spect_id, prio_id, mag_id, pid_id, comment_id, 
                             wlenIndex_id, pmra_id, pmdec_id, &status);
            }
            if (status != STATUS__OK)
                break;
        }
        SdsFreeId(name_id, &status);
        SdsFreeId(ra_id, &status);
        SdsFreeId(dec_id, &status);
        SdsFreeId(type_id, &status);
        if (spect_id)
            SdsFreeId(spect_id, &status);
        SdsFreeId(prio_id, &status);
        SdsFreeId(mag_id, &status);
        SdsFreeId(pid_id, &status);
        SdsFreeId(comment_id, &status);
        SdsFreeId(uid, &status);
        if (wlenIndex_id)
            SdsFreeId(wlenIndex_id, &status);
    }
    /*  Original Tcl code
    if ((strcmp(argv[3], "obj") == 0) || (strcmp(argv[3], "all") == 0)) {
    */
    if ((strcmp(Type, "obj") == 0) || (strcmp(Type, "all") == 0)) {
        SdsFind(topid, "unallocGuide", &uid, &status);
        ConfGetIds(uid, &ra_id, &dec_id, &name_id, &type_id, &spect_id,
                   &prio_id, &mag_id, &pid_id, &comment_id, &wlenIndex_id,
                   &pmra_id, &pmdec_id, &status);
        SdsInfo(ra_id, iname, &code, &ndims, dims, &status);
        fprintf(listing, "*\n");
        fprintf(listing, "*  Fiducial Stars\n");
        fprintf(listing, "*\n");
        fprintf(listing, "*%12s%12s%12s\n", "Name", "R.A.", "Dec");
        for (i = 0; i < dims[0]; i++) {
            ConfListLine(listing, i, ra_id, dec_id, name_id, type_id, spect_id,
                         prio_id, mag_id, pid_id, comment_id, 
                         pmra_id, pmdec_id, wlenIndex_id, &status);
            if (status != STATUS__OK)
                break;
        }
        SdsFreeId(name_id, &status);
        SdsFreeId(ra_id, &status);
        SdsFreeId(dec_id, &status);
        SdsFreeId(type_id, &status);
        if (spect_id)
            SdsFreeId(spect_id, &status);
        SdsFreeId(prio_id, &status);
        SdsFreeId(mag_id, &status);
        SdsFreeId(pid_id, &status);
        SdsFreeId(comment_id, &status);
        if (wlenIndex_id)
            SdsFreeId(wlenIndex_id, &status);
        SdsFreeId(uid, &status);

/*  We also need the fiducials and skies from the allocated objects section  */

        SdsFind(topid, "objects", &objid, &status);
        ConfGetIds(objid, &ra_id, &dec_id, &name_id, &type_id, &spect_id,
                   &prio_id, &mag_id, &pid_id, &comment_id, &wlenIndex_id, 
                   &pmra_id, &pmdec_id, &status);
        SdsInfo(ra_id, iname, &code, &ndims, dims, &status);
        for (i = 0; i < dims[0]; i++) {
            ConfArrayGetc(objid, "type", i, &c, &status);
            ConfArrayGetc(objid, "spectrograph", i, &s, &status);
            if (FpilIsTargetGuide(*Instrument, c) ||
                                     FpilIsTargetSky(*Instrument, c)) {
                ConfListLine(listing, i, ra_id, dec_id, name_id, type_id,
                             spect_id, prio_id, mag_id, pid_id, comment_id, 
                             wlenIndex_id, pmra_id, pmdec_id, &status);
            }
            if (status != STATUS__OK)
                break;
        }
        SdsFreeId(name_id, &status);
        SdsFreeId(ra_id, &status);
        SdsFreeId(dec_id, &status);
        SdsFreeId(type_id, &status);
        if (spect_id)
            SdsFreeId(spect_id, &status);
        SdsFreeId(prio_id, &status);
        SdsFreeId(mag_id, &status);
        SdsFreeId(pid_id, &status);
        SdsFreeId(comment_id, &status);
        if (wlenIndex_id)
            SdsFreeId(wlenIndex_id, &status);
        SdsFreeId(objid, &status);
    }

    SdsFreeId(fid, &status);

    fclose(listing);
    if (status == STATUS__OK) {
        return 0;
    } else {
        MessGetMsg(status, 0, 100, mbuf);
        printf("Error in MakeListing\n%s", mbuf);
        return ERROR_CODE;
    }
}
  
