
/*+        		t i m e U n i x M j d  

 *  Module name:
	timeUnixMjd
     
 *  Function:
         Provides conversions between unix times and MJD's.

 *  Synopsis:


 *  Description:
        This module has code extracted from the 2dF control task, which handles
        conversions between Unix Time and the MJD's used in astromony.

 *  Language:
      C

 *  Support: Tony Farrell, AAO

 *-


 * "@(#) $Id: ACMM:aaoConfigure/src/timeUnixMjd.C,v 12.7 04-Nov-2013 11:28:48+11 tjf $"

 *  History:
      28-Aug-2008 - TJF - Original version.
      23-Sep-2009 - TJF - Include assert.h.  Needed for MacOsX Snow Leopord.
      14-Jan-2011 - TJF - Include stdio.h and fix RCS statics for GCC 4.5.2


 */
static const char *rcsId="@(#) $Id: ACMM:aaoConfigure/src/timeUnixMjd.C,v 12.7 04-Nov-2013 11:28:48+11 tjf $";
static void *use_rcsId = (0 ? (void *)(&use_rcsId) : (void *) &rcsId);

#include <assert.h>
#include "timeUnixMjd.h"
#include <time.h>
#include <stdio.h>

namespace timeUnixMjd {


/*
 * Convert a MJD to a string (UT or Local)
 */
static void Mjd2S(const double mjd, const int len, char * const string, const bool local, const bool timeOnly=false)
{
    time_t unixtime;                /* For the time as a unix time  */
    double unixday;                 /* The number of days since unix 0 */
    struct tm *tm;                  /* unix tm structure            */
                                                                                /*
 *  Convert the MJD to the number of days since the begining of the
 *  unix clock.  It must be positive.
 */
    unixday = mjd - UNIXSTART;
    assert (unixday > 0.0);

/*
 *  Convert to seconds (unix time)
 */
    unixtime = (long int)(unixday * SPD);
/*
 *  We now have the time in a form which an be handled by standard unix
 *  routines, so convert it to a string.
 */

    if (local)
        tm = localtime(&unixtime);
    else
        tm = gmtime(&unixtime);

    if (timeOnly)
        strftime(string,len,"%H:%M:%S",tm);
    else
        strftime(string,len,"%d-%b-%Y %H:%M:%S",tm);

}

/*
 * Convert a MJD to a local time.
 */
extern std::string Mjd2Local(double mjd)
{
    char time_string[40];
    Mjd2S(mjd, sizeof(time_string), time_string, true);
    return std::string(time_string);
}


/*
 * Convert a MJD to a UT Time time.
 */
extern std::string Mjd2Ut(double mjd)
{
    char time_string[40];
    Mjd2S(mjd, sizeof(time_string), time_string, false);
    return std::string(time_string);
}

/*
 * Convert a MJD to a UT Time time, time string only (no date)
 */
extern std::string Mjd2Ut_time(double mjd)
{
    char time_string[40];
    Mjd2S(mjd, sizeof(time_string), time_string, false, true);
    return std::string(time_string);
}

/*
 *  Conversion of a time string in standard format to MJD
 */
extern double S2Mjd(
    const char *string, 
    bool local)
{
    
    struct tm *tm;		/* Unix tm structure */
    struct tm tmi;
    time_t now = time(0);       /* Get current time in secs    */
    time_t thetime;             /* Stores the time in unix format */

/*
 *  We put default values into tm before calling strptime.
 */
    tm = local ? localtime(&now) : gmtime(&now);
    tmi = *tm;
    if (!local)
        tmi.tm_isdst = 0;  // Just be sure this is not set

/*
 *  Try for format with year including century.  Seconds are optional.
 *  We may have a negative year hear 
 */

    if (((strptime((char *)string,"%e-%h-%Y %H:%M:%S",&tmi) != 0)||
         (strptime((char *)string,"%e-%h-%Y %H:%M",&tmi) != 0))&&
        (tmi.tm_year > 0))
    {
        /* Nothing to do */
    }
    /*
     * Try for format with year but without century.  In this
     * case, we will assume a year less then 70 is 2000+year.
     */
    else if ((strptime((char *)string,"%e-%h-%y %H:%M:%S",&tmi) != 0)||
             (strptime((char *)string,"%e-%h-%y %H:%M",&tmi) != 0))

    {
        if (tmi.tm_year < 70)
            tmi.tm_year += 100;
    }
    /*
     *    Try for format with time only (seconds optional)
     */
    else if ((strptime((char *)string,"%H:%M:%S",&tmi) != 0)||
             (strptime((char *)string,"%H:%M",&tmi) != 0))
    {
        /*
         * copy across date specific files from todays date.
         */
        tmi.tm_mday = tm->tm_mday;  /* Day of month */
        tmi.tm_mon  = tm->tm_mon;   /* Months since Jan */
        tmi.tm_year = tm->tm_year;  /* Years since 1900 */
        tmi.tm_wday = tm->tm_wday;  /* Days since Sunday */
        tmi.tm_yday = tm->tm_yday;  /* Days since Jan 1 */
        tmi.tm_isdst = tm->tm_isdst;/* Daylight savings time */
    }
    else
    {
        fprintf(stderr, "\"%s\" is an invalid time string\n",string);
        fprintf(stderr, "  The required format is [DD-MMM-[YY]YY] HH:MM[:SS]\n");
        fprintf(stderr, "  The year must be between %d and %d\n",
                UNIX_MIN_YEAR, UNIX_MAX_YEAR);
        return 0;
    }



/*
 *  Now convert the contents of tm to a time_t structure (time in seconds
 *  since UNIXSTART).  These routines appear to return -1 for times 
 *  greater then that which can be handled by the unix time routines.
 *
 *  There does not appear to be a SVR4 equivalent to timegm.
 */
#   if defined(__SVR4)
        if (local)
        {
            tmi.tm_isdst = -1;  /* Unclear if this works for solaris */
	    thetime = mktime(&tmi);
        }
	else
	{
	    fprintf(stderr, "Sorry, this system does not support input of UT Times\n");
	    return 0;
	}
#   else
        if (local)
        {
            /*
             * By setting isdst to -1 timelocal will try to work out
             * the daylight savings setting.  It was getting this right
             * for redhat E3, 16-Jan-2007.  
             */
            tmi.tm_isdst = -1;
            thetime = timelocal(&tmi);
        }
        else
            thetime = timegm(&tmi);
#   endif



    if (thetime == -1)
    {
        fprintf(stderr, "\"%s\" cannot be handled by system time routines\n", string);
        fprintf(stderr, "  The required format is [DD-MMM-[YY]YY] HH:MM[:SS]\n");
        fprintf(stderr, "  The year must be between %d and %d\n",
                UNIX_MIN_YEAR, UNIX_MAX_YEAR);
        return 0;
    }

    /*
     * Convert to days and and MJD at unix start;
     */
    return double(thetime)/SPD + UNIXSTART;
  
}

/*
 *  Return the Current local time as a string
 */
extern std::string  LocalTimeDate()
{
    time_t now = time(0);           /* Get current time in secs     */
    struct tm *tm;                  /* unix tm structure            */

    /* Convert to Local time and put in buffer */        
    tm = localtime(&now);

    char buffer[30];
    strftime(buffer,sizeof(buffer),"%d-%h-%Y %H:%M:%S",tm);
    return buffer;
}
/*
 *  Return the Current UT time as a string
 */
extern std::string UtTimeDate() 
{
    time_t now = time(0);            /* Get current time in secs     */
    struct tm *tm;                  /* unix tm structure            */
    
    /* Convert to Universal (GM) time and put in buffer */
    tm = gmtime(&now);
    char buffer[30];
    strftime(buffer,sizeof(buffer),"%d-%b-%Y %H:%M:%S",tm);
    return buffer;
}

/*
 *  Return the current time as an Mjd
 */
extern double Mjd() 
{
    time_t now = time(0);       /* Get current time in secs    */
    double mjd;
    mjd = (double)now/SPD;	/* Convert to days              */
    mjd += UNIXSTART;     	/* Add MJD at unix start        */
    return mjd;    
}



}; // namespace timeUnixMjd
