/*+		 X Y 2 F L D

 * Program name:
       xy2fld

 * Function:
        Given a set of x/y positions, create an FLD file.

 * Synopis:
        xy2fld <instrument> <xy_file> > fld_file.

 * Description:
        Read a file of x/y positions from and produces on stdout
        an aaoConfigure field file.

        The input file is specified below and includes the UTDATE of
        the observation, the instrument and the centre position. 

        Each X/Y position is converted to an RA/Dec position and the
        .FLD file is generated.  

        If the input file name is "-", then input is read from stdin.

 * Input File Format
        * Comments are introduced by * or #
        LABEL  "some string"              # Passed through to FLD file.
        UTDATE 2002 07 07[.x]             # UT Date for observation
        CENTRE HH MM SS.SS DD MM SS.SS    # Centre of observation
        EQUINOX J2000.0                   # Equinox - currently always J2000.0.

        * Data lines
        <name> <x> <y> <type> <priority> <comment>
        ...

        
        <name>      = object name, to be passed to .fld file.
        <x>         = x position on field plate, microns.
        <y>         = y position on field plate, microns
        <type>      = object type - to be passed to .fld file. (defaults to P)
        <priority>  = object priority - to be passed to .fld file. (defaults to 5)
        <comment>   = object comment - to be passed to .fld file. (defaults to blank)

        Trailing values from <type> onwards can be left out.



 * Language:
      C++

 * Support: Tony Farrell, AAO

 *-

 * @(#) $Id: ACMM:aaoConfigure/src/xy2fld.C,v 12.7 04-Nov-2013 11:28:48+11 tjf $

 * History:
       01-Sep-2011 - TJF - Original version


 *  Copyright (c) Australian Astronomical Observatory, 2011
    Not to be used for commercial purposes without permission.
    No warranty of any form unless subject to a specific agreement.

 */
static const char *rcsId="@(#) $Id: ACMM:aaoConfigure/src/xy2fld.C,v 12.7 04-Nov-2013 11:28:48+11 tjf $";
static void *use_rcsId = (0 ? (void *)(&use_rcsId) : (void *) &rcsId);

/*
 *  Include files
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <time.h>
#include <vector>
#include <string>
#include "slalib.h"
#include "slamac.h"

#include "status.h"
#include "tdffpil.h"            /* 2dF - AAT */
#include "sixdffpil.h"          /* 6dF - UKSCHMIDT */
#include "fpcolFpil.h"          /* OzPoz - VLT-2 */
#include "Ers.h"

extern "C" {
#include "tdFconvert.h"
}

#ifndef MAXPATHLEN
#define MAXPATHLEN 1024
#endif

/*
 * Configure default values for these.
 */ 
const double WLEN = 0.6;  /* Wavelength of observation (micrometers) */
const double TEMP = 285;   /* Kelvin */
const double PRES = 900;   /* Millibars */
const double HUMID = 0.5;  /* relative */



/*
 * A class for managing 2dF field data.
 */
class FieldData {
private:
    fieldType _data;
public:
    // Constructor.  Inst must be defined and inFile open.  utoffset is the
    ///  default offset from UT midnight to use, if not specified in file.
    //   lineCount will be incremented by the number of lines read.
    FieldData(const FpilType &Inst, FILE *inFile, double utoffset, int *lineCount);

    // Output the field details to stdout in .fld file format.
    void Output(const char *fileName, const char *InstName) const;

    double GetAppRa() const {
        return _data.appRa;
    }
    double GetAppDec() const {
        return _data.appDec;
    }    
    double GetMeanRa() const {
        return _data.cenRa;
    }
    double GetMeanDec() const {
        return _data.cenDec;
    }    
    double GetMjd() const {
        return _data.configMjd;
    }
};

class TelModel {
private:
    FpilModelType *_model;
    unsigned int _numParams;
    std::vector <double> _params;  /* Current telescope model parameters */

    double _gmap[21];      /* For slalib calls. No const for the magic number */
    double _cenAppRa;      /* Field centre RA */
    double _cenAppDec;     /* Field centre Dec */
    double _mjd;           /* Time of observation as MJD */

    static void Reporter(void *argument, const char *string, StatusType * status);


public:
    TelModel(const FpilType &Inst, double cenAppRa, double cenAppDec, double mjd);

    ~TelModel() {
        if (_model)
        {
            StatusType status = STATUS__OK;
            FpilModelFree(_model, &status);
        }
    }

    // Convert positioner X/Y to MEAN Ra/Dec;  returns 0 on success, 1 on failure.
    int PosToRd(double x, double y, double *ra, double *dec) const;
};


void ProcessDataLines(const TelModel &model, FILE *inFile, int *lineCount);


// Class which is used for reading data from the input file and
// then outputing the converted data.
class DataLine {
private:
    char _name[LINE_LENGTH];
    char _comment[LINE_LENGTH];
    char  _type[LINE_LENGTH];
    int  _priority;
    long _x;
    long _y;
    bool _valid;
public:
    DataLine(FILE *inFile, int *lineCount);
    void Output(const TelModel &model) const;
};


int main(int argc, char *argv[])
{
    FpilType Inst;
    const char *InstName=0;
    double uttime_offset = 0;

    if (argc != 3)
    {
        fprintf(stderr,"%s:Usage:%s <instrument> <file>\n", argv[0], argv[0]);
        exit(1);
    }
    /*
     * Access the instrument details.
     */
    if ((strcasecmp(argv[1],"2df") == 0)||
        (strcasecmp(argv[1],"aaomega") == 0)||
        (strcasecmp(argv[1],"hermes") == 0))
    {
        TdfFpilFullInitAAOmega(&Inst);
        InstName="2dF";
        uttime_offset = 0.5;
    }
    else if  (strcasecmp(argv[1],"6df") == 0) 
    {
        sixdfFpilFullInit(&Inst);
        InstName="6dF";
        uttime_offset = 0.5;
    }
    else if  (strcasecmp(argv[1],"ozpoz") == 0) 
    {
        fpcolFpilInit(&Inst);
        InstName="OzPoz";
        uttime_offset = -0.2;
    }
    else 
    {
        fprintf(stderr,"%s:Invalid Instrument name \"%s\"\n", argv[0], argv[1]);
        exit(1);
    }

    FILE *inFile = 0;
    if (strcmp(argv[2],"-") == 0)
    {
        inFile = stdin;
    }
    else
    {
        inFile = fopen(argv[2],"r");
        if (!inFile)
        {
            fprintf(stderr, "%s:Failed to open file \"%s\", %s\n",
                    argv[0], argv[2], strerror(errno));
            exit(1);
            
        }
    }
    /*
     * Parse the header of the file by constructing a FieldData object.
     */
    int lineCount = 0; 
    FieldData fieldData(Inst, inFile, uttime_offset, &lineCount);

    TelModel model(Inst, fieldData.GetAppRa(), fieldData.GetAppDec(), 
                   fieldData.GetMjd());
    //CreateTelescopeModel(&fieldData, &model)

    fieldData.Output(argv[2], InstName);

    /*
     * We convert an X/Y of 0,0 to RA/Dec.  This checks the have a valid
     * field etc.
     */
    double nra, ndec;
    if (model.PosToRd(0, 0, &nra, &ndec))
    {
        fprintf(stderr,"Failed to convert field centre 0.0 to RA/Dec\n");
        exit(1);
    }
    
    double deltaRa, deltaDec;
    deltaRa =  fabs(nra - fieldData.GetMeanRa()) / DS2R;
    deltaDec = fabs(ndec - fieldData.GetMeanDec()) / DS2R;

#if 0  /* Test code */
    fprintf(stderr, "Original Field Centre - %.8f, %.8f\n",
            fieldData.GetMeanRa(), fieldData.GetMeanDec());
    fprintf(stderr, "New      Field Centre - %.8f, %.8f\n",
            nra, ndec);
    fprintf(stderr, "%g, %g\n", deltaRa, deltaDec);

#endif


    /*
     * The field centre should have come up easily to the original value.
     * Check that it is with 0.01 arc-seconds.
     */
    if ((deltaRa > 0.01)||(deltaDec > 0.01))
    {
        fprintf(stderr,"Field Centre conversion check failed\n");
        fprintf(stderr,"  Position different by %.1f, %.1f arc-seconds\n",
                deltaRa, deltaDec);
        exit(1);
    }
    ProcessDataLines(model, inFile, &lineCount);

    fprintf(stderr,"Processed %d lines\n", lineCount);
    fclose(inFile);

    exit(0);
}


/*
 * This function is called to process lines from the input file
 * and to write lines to the output file (stdout).
 */
void ProcessDataLines(const TelModel &model, FILE *inFile, int *lineCount)
{

    while (!feof(inFile)&&!ferror(inFile))
    {
        // Read a line of data and then output it.
        DataLine data(inFile, lineCount);
        data.Output(model);

    }
}





/*
 * To avoid linking to Tcl, create our own versions of these.
 */
extern "C" void * Tcl_GetStdChannel(int /*which*/)
{
    return stderr;
}
extern "C" int Tcl_Write(void *channel, char *buffer, int)
{
    fprintf((FILE *)(channel), "%s", buffer);
    return 0;  /* TCL_OK*/
}
extern "C" void ConfGetProperMotions()
{
}






/* ------------------------------------------------------------------------- 
 *
 *  FieldData class definitions.
 *
 */

FieldData::FieldData(
    const FpilType &Inst, 
    FILE *inFile, 
    double utoffset, 
    int *lineCount) 
{
    short fractDayFlag=0;
    wavelensType wlens;
    short explicitPms = 0;
    StatusType status = STATUS__OK;

    tdFparseFieldData(&Inst, inFile, &_data, utoffset,
                      &fractDayFlag, lineCount, &wlens, &explicitPms,
                      &status);

    if (status != STATUS__OK)
    {
        fprintf(stderr,"Failed to parse file header\n");
        exit(1);
    }
}
void FieldData::Output(
    const char *fileName, 
    const char *InstName) const 
{
    time_t now;
    time(&now);
    printf("* File generated by xy2fld program from %s\n", fileName);
    printf("*   Generated at %s", ctime(&now));
    printf("*   Generated for %s instrument model\n", InstName);
    printf("*\n");
    printf("*   Presumed:\n");
    printf("*       Field Plate = 1 (might impact model, does for 2dF)\n");
    printf("*       Temperature = %.0f degrees kelvin\n", TEMP);
    printf("*       Humidity    = %.0f percent\n", HUMID*100);
    printf("*       Pressure    = %.0f millibars\n", PRES);
    printf("*       Wavelength  = %.0f Angstroms\n", WLEN*MICRONS_TO_ANGSTROMS);
    printf("*    To get exactly the intended resuls, configure with the above values (default values)\n");
    printf("*\n");
    printf("LABEL %s\n", _data.label);
    
    int iy, im, id;
    double fd;
    int jstat;
    slaDjcl(_data.configMjd, &iy, &im, &id, &fd, &jstat);
    printf("UTDATE %4.4d %2.2d %2.2d.%d\n", iy, im, id, (int)(fd*1000));
    
    int rhmsf[4];
    int ddmsf[4];
    char sign[1];
    slaDr2tf(3, _data.cenRa, sign, rhmsf);
    slaDr2af(2, _data.cenDec, sign, ddmsf);
    printf("CENTRE %2.2d %2.2d %2.2d.%3.3d  %c%2.2d %2.2d %2.2d.%2.2d\n",
           rhmsf[0], rhmsf[1], rhmsf[2], rhmsf[3], sign[0], ddmsf[0], ddmsf[1],
           ddmsf[2], ddmsf[3]);
    
    
    
    printf("EQUINOX J2000.0\n");
    printf("*End of header info\n");        
}

static const char * GetConfigFilesDir()
{
    static char ConfigFilesDir[MAXPATHLEN];
    static bool done = false;
    if (done)
        return ConfigFilesDir;

    char *configFilesValPnt = getenv("CONFIG_FILES");
    if (configFilesValPnt) {
        strncpy(ConfigFilesDir, configFilesValPnt, sizeof(ConfigFilesDir));
        ConfigFilesDir[sizeof(ConfigFilesDir) - 1] = '\0';
    } else
        ConfigFilesDir[0] = '\0';
    done = true;
    return ConfigFilesDir;
}


/* ------------------------------------------------------------------------- 
 *
 *  TelModel class definitions.
 *
 */
TelModel::TelModel(
    const FpilType &Inst,
    double cenAppRa, 
    double cenAppDec, 
    double mjd) : _model(0), _cenAppRa(cenAppRa), _cenAppDec(cenAppDec), _mjd(mjd)
{
    unsigned int field = 1; /* We don't care what field, 1 works on all instruments
                               we currently support */
    StatusType status = STATUS__OK;
    FpilModelAccess(Inst, GetConfigFilesDir(), ".",
                    Reporter, 0, field, &_model, &status);

    if (status != STATUS__OK)
    {
        fprintf(stderr, "Failed to access telescope model files.");
        exit(1);
    }
    const char *const *ParamNames;
    const char *const *ParamDescrs;
    const double *ParamDefaults;
    /*
     * Get details of the telescope model parameters.
     */
    FpilModelParams(_model, &_numParams,
                    &ParamNames, &ParamDescrs, &ParamDefaults, &status);
    
    _params.resize(_numParams,0);
    for (unsigned i = 0; i < _numParams; ++i) {
        _params[i] = ParamDefaults[i];
    }

    FpilModelCvtInit(_model, _mjd, 0.0, TEMP, PRES, HUMID, WLEN, WLEN,
                     &_params[0], &status);


    /*
     * Set up for Sla quick conversions - _gmap item.
     */
    slaMappa(2000.0, _mjd, _gmap);
}

/*
 * Covert an X/Y to a Mean RA/Dec
 */
int TelModel::PosToRd(
    const double xp, 
    const double yp, 
          double * const ra, 
          double * const dec) const
{

    /* 
     * Convert from positioner to focal plane  x/y.
     */
    double x;
    double y;
    StatusType status = STATUS__OK;
    FpilModelPos2Xy(_model, xp, yp, &x, &y, &status);
    if (status != STATUS__OK)
    {
        fprintf(stderr,"Initial conversion of %ld, %ld to focal plane x/y failed - status code %ld\n",
                (long)xp, (long)(yp), (long)(status));
        return 1;
                
    }

    /*
     * Now we have to convert from focal plane x/y to apparent
     * position. The conversion routine in not accurate, so iterate
     * to get this right.
     */
    FpilModelXy2Rd(_model, _cenAppRa, _cenAppDec, x, y, _mjd, 
                   ra, dec, &status);     // Initial attempt.

    if (status != STATUS__OK)
    {
        fprintf(stderr,"Initial conversion of %ld, %ld to Apparent RA/Dec failed - status code %ld\n",
                (long)xp, (long)(yp), (long)(status));
        return 1;
                
    }


    const double initialRA = *ra;
    const double initialDec = *dec;
    const unsigned maxIters = 20;   // How many goes
    const double tolerance = 5;   // to get this close (microns)

    for (unsigned i = 1; (status == STATUS__OK)&&(i <= maxIters); ++i)
    {
        double nx, ny;
        // Convert the calculated value back to x/y
        FpilModelRd2Xy(_model, _cenAppRa, _cenAppDec, *ra, *dec, _mjd, 
                       &nx, &ny, &status);
        // How far off are we?
        const double deltaX = nx - x;
        const double deltaY = ny - y;
        const double delta = hypot(deltaX, deltaY);
        if (delta <= tolerance)
            break;

        // Not close enough. 
        double nRA, nDec, deltaRA, deltaDec;
        FpilModelXy2Rd(_model, _cenAppRa, _cenAppDec,
                       (x + deltaX), (y + deltaY), _mjd, &nRA, &nDec, &status);

        deltaRA = initialRA - nRA;
        deltaDec = initialDec - nDec;

        *ra += deltaRA;
        *dec += deltaDec;

        if (i == maxIters)
        {
            fprintf(stderr,
                    "For X/Y (%ld, %ld), failed to get good RA/Dec after %d iterations\n", 
                    (long)(xp), (long)(yp), maxIters);
        }
    }

    if (status != STATUS__OK)
    {
        fprintf(stderr,"Conversion of %ld, %ld to Apparent RA/Dec failed whilst iterating - status code %ld\n",
                (long)xp, (long)(yp), (long)(status));
        return 1;
    }
    slaAmpqk(*ra, *dec, const_cast<double *>(_gmap), ra, dec);
    return 0;


}


/*
 * Invoked when the model file is opened.
 */
void TelModel::Reporter(
    void * /*argument*/, 
    const char *string, 
    StatusType * status)
{
    if (*status != STATUS__OK)
        return;
    fprintf(stderr, "Opened telescope model file \"%s\"\n", string);
}

/* ------------------------------------------------------------------------- 
 *
 *  DataLine class definitions.
 *
 */

DataLine::DataLine(
    FILE * const inFile, 
    int * const lineCount): _priority(5), _x(0), _y(0), _valid(false)
{
    _name[0] = 0;
    _comment[0] = 0;
    strcpy(_type,"P");  /* Default type is P */
    /*
     * Read and parse the next line of input.  We do this in a similar
     * way to tdFparseObjectData(), since the file is similar, but not
     * similar enough to use the same function.
     */
    char dataline[LINE_LENGTH];
    while (fgets(dataline,LINE_LENGTH,inFile) != NULL) 
    {
        ++(*lineCount);
        if ((dataline[0] == '*') || (dataline[0] == '#'))
            continue;
        /* 
         * Check if completely blank 
         */
        bool isblank = true;
        for(unsigned i=0; i<LINE_LENGTH && dataline[i]!=0; i++) 
        {
           if (!isspace( (int) dataline[i]) && dataline[i]!=0) 
           {
             isblank = false;
             break;
           }
        }
                
        if (isblank) 
           continue;

        int items;

        items = sscanf(dataline,
                      "%s %ld %ld %s %d %[^\n]",
                      _name,&_x, &_y, _type, &_priority, _comment);
        /*
         * If we have at least three items, we are ok.
         */
        if (items < 3)
        {
            fprintf(stderr,"Line %d, Syntax error\n", *lineCount);
            return;
        }
        /*
         * Done;
         */
        _valid = true;
        break;
    }


    
}
void DataLine::Output(
    const TelModel &model) const 
{
    if (!_valid) return;


    printf("%s ", _name);
 

    double nra=0, ndec=0;
    model.PosToRd(_x, _y, &nra, &ndec);

    int rhmsf[4];
    char sign[1];
    slaDr2tf(3, nra, sign, rhmsf);

    printf("%2.2d %2.2d %2.2d.%3.3d ",
           rhmsf[0], rhmsf[1], rhmsf[2], rhmsf[3]);

    int ddmsf[4];
    slaDr2af(2, ndec, sign, ddmsf);
           
    printf("%c%2.2d %2.2d %2.2d.%2.2d ",
           sign[0], ddmsf[0], ddmsf[1], ddmsf[2], ddmsf[3]);

    printf("%s %d 10.0 0 %s(%ld, %ld)\n",
           _type, _priority, _comment, _x, _y);
   
}
