/*
 *                      F p i l  A f f e c t . c
 *
 *  Function:
 *     Routines that handle the data relating to 'affected' fibres.
 *
 *  Description:
 *     In this context, an 'affected' fibre is a fibre that is affected by
 *     spillover or ghosts or some other effect that results in light from
 *     one target contaminating the spectrum of another. In general, we
 *     want to arrange things so that as far as possible we do not have
 *     light from bright targets contaminating the spectra of fainter 
 *     targets. To do this, we need to be able to find out which fibres are
 *     affected by any given fibre. Generally, a fibre will affect more than
 *     one other fibre - a ghost may cover a number of other fibres, 
 *     crosstalk will affect adjacent fibres and even, to a lesser effect,
 *     fibres adjacent to them and so on.
 *
 *     This file implements the main routine for accessing affected fibre
 *     data (FpilGetAffectData()) as well as the utility routines
 *     FpilAffectErrorText(), FpilClearAffectData() and FpilReleaseAffectData().
 *
 *     For more details, see FpilAffect.h
 *
 *  Testing:
 *     gcc -o FpilAffect -Wall -ansi -pedantic -DFPIL_AFFECT_TEST FpilAffect.c
 *     ./FpilAffect <filename>
 *
 *     where <filename> is the name of a suitable CSV format file. This will
 *     list the affectation data given in that file.
 *
 *  Support: Keith Shortridge, AAO
 *
 *  History:
 *     26th Jun 2013. Original version. KS.
 *     21st Jul 2013. First fully fleshed out implementation. KS.
 *      2nd Aug 2013. Added FpilSetAffectError(). KS.
 */

#include "FpilAffect.h"

#include <stdio.h>
#include <strings.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>

#undef FALSE
#define FALSE 0
#undef TRUE
#define TRUE 1

#define DEBUG printf

/*  ------------------------------------------------------------------------- */

/*                       E r r o r  H a n d l i n g
 */

/*  This code doesn't need to be re-entrant or thread safe, so it's easy
 *  enough to provide an error reporting scheme that just gets passed
 *  different bits of information (such as the filename in use) at various
 *  times and stores them in global variables. If there is an error,
 *  these can then be combined by FpilAffectErrorText() into a single
 *  error string.
 *
 */
 
static char G_Filename[1024];
static char G_ErrorText[1024];
static char G_ErrorLine[256];
static char G_ErrorDetails[256];
static short G_ErrorFlag;

/*  Clears all the error details */

static void Fpil__CSVClearErrors (void)
{
   G_Filename[0] = '\0';
   G_ErrorText[0] = '\0';
   G_ErrorLine[0] = '\0';
   G_ErrorDetails[0] = '\0';
   G_ErrorFlag = FALSE;
}

/*  Notes the file name being used */

static void Fpil__CSVErrSetFilename (const char* Filename)
{
   strncpy (G_Filename,Filename,sizeof(G_Filename));
   G_Filename[sizeof(G_Filename) - 1] = '\0';
}

/*  Called when there has been an error (and so sets the error flag) and is
 *  passed a description of what went wrong.
 */
 
static void Fpil__CSVErrSetText (const char* Text, const char* Line)
{
   strncpy (G_ErrorText,Text,sizeof(G_ErrorText));
   G_ErrorText[sizeof(G_ErrorText) - 1] = '\0';
   if (Line) {
      strncpy (G_ErrorLine,Line,sizeof(G_ErrorLine));
      G_ErrorLine[sizeof(G_ErrorLine) - 1] = '\0';
   } else {
      G_ErrorLine[0] = '\0';
   }
   G_ErrorDetails[0] = '\0';
   G_ErrorFlag = TRUE;
}

/*  Used to provide additional details about the error - for example, the
 *  errno text on an I/O error, or the item from a file that wasn't valid.
 *  This is what is output by FpilAffectErrorText() in brackets.
 */
 
static void Fpil__CSVErrSetDetails (const char* Details)
{
   strncpy (G_ErrorDetails,Details,sizeof(G_ErrorDetails));
   G_ErrorDetails[sizeof(G_ErrorDetails) - 1] = '\0';
}

/*  ------------------------------------------------------------------------- */

/*                          Fpil__SplitCSVLine()
 *
 *  Splits up a line read from a CSV file into separate items, trying to
 *  be cleverish about working out what the separator character is. (In
 *  principle, we ought to be able to assume that the separator is the same
 *  for all lines, but you might not be able to work it out from the first
 *  line, so that gets complicated too.) This fills out an array of pointers
 *  to the start of each item in Line. Line is modified to replace
 *  the separators by nuls, so that the caller will see each item as
 *  a separate terminated string.
 */
 
static short Fpil__SplitCSVLine (
   char* Line, int MaxItems, int* ItemCount, char* Items[])
{
   /*  Our main problem is that we don't know what the separator character 
    *  is. However, it's unlikely that we have tabs or commas in the actual
    *  values, so if we see those then that's probably the separator. Spaces
    *  are also possible. Restricting ourselves to those should be safe...
    *  If we don't see any of these, we set the separator to the terminating
    *  nul and treat the whole ine as one entry. 
    */
   
   short ReturnCode = TRUE;
   int ItemNumber = 0;
   int TabCount = 0;
   int CommaCount = 0;
   int SpaceCount = 0;
   char Separator = '\0';
   char QuoteChar = '"';
   short InQuote = FALSE;
   char Ch;
   char* LinePtr = Line;
   char* ItemStart = Line;
   short InItem = FALSE;
   
   /*  First pass through to see what's in there and decide what to treat
    *  as the separator.
    */
   
   while ((Ch = *LinePtr)) {
      if (InQuote) {
         if (Ch == QuoteChar) InQuote = FALSE;
      } else {
         if (Ch == '"' || Ch == '\'') {
            InQuote = TRUE;
            QuoteChar = Ch;
         } else {
            if (Ch == '\t') TabCount++;
            else if (Ch == ',') CommaCount++;
            else if (Ch == ' ') SpaceCount++;
         }
      }
      LinePtr++;
   }
   if (TabCount > 0) {
      Separator = '\t';
   } else if (CommaCount > 0) {
      Separator = ',';
   } else if (SpaceCount > 0) {
      Separator = ' ';
   }
   
   /*  Now run through and split up the line into items */
   
   InQuote = FALSE;
   LinePtr = Line;
   ItemStart = Line;
   InItem = FALSE;
   if (Separator == ' ') while (*LinePtr == ' ') LinePtr++;
   while ((Ch = *LinePtr)) {
      if (!InItem) {
         ItemStart = LinePtr;
         InItem = TRUE;
      }
      if (InQuote) {
         if (Ch == QuoteChar) InQuote = FALSE;
      } else {
         if (Ch == '"' || Ch == '\'') {
            InQuote = TRUE;
            QuoteChar = Ch;
         } else {
            if (Ch == Separator) {
               *LinePtr = '\0';
               if (MaxItems > ItemNumber) Items[ItemNumber++] = ItemStart;
               InItem = FALSE;
               if (Separator == ' ') while (*(LinePtr + 1) == ' ') LinePtr++;
            }
         }
      }
      LinePtr++;
   }
   if (InItem) {
      if (MaxItems > ItemNumber) Items[ItemNumber++] = ItemStart;
   }
   *ItemCount = ItemNumber;
       
   return ReturnCode;
}

/*  ------------------------------------------------------------------------- */

/*                          Fpil__ReadCSVLine()
 *
 *  Reads the next line from the input file. This is complicated slightly by
 *  not knowing exactly what character or combination of characters will have
 *  been used to terminate the line. This code assumes some combination of
 *  carriage return ('\r') and newline ('\n') characters. It assumes the first
 *  of these that it finds terminates the line. Since this may leave another
 *  such character still to come, the code ignores any such character that
 *  appears to be at the start of the next line, meaning it skips over any
 *  string of such characters. This may mean that if fails to pick up what
 *  are really blank lines, but in this case that doesn't matter.
 *
 *  Returns the characters read from the file for the next line in Line, up
 *  to a maximum of LineSize. A nul terminating character ('\0') will always
 *  be set as the last of the returned characters, even if this means the
 *  line has to be truncated. If the end of file is hit while reading the line,
 *  Eof is set true. The routine returns true if there is valid data in Line -
 *  this may be the case even if it returns Eof as true.
 */
  
static short Fpil__ReadCSVLine (
   FILE* InputFile, char* Line, int LineSize, short* Eof)
{
    
   int Index = 0;
   short InQuotes = FALSE;
   short ReturnCode = TRUE;
   char QuoteChar = '"';
   *Eof = FALSE;
   for (;;) {
      char Ch = getc(InputFile);
      if (Ch == EOF) {
         if (ferror(InputFile)) {
            Fpil__CSVErrSetText("Error reading from file",NULL);
            Fpil__CSVErrSetDetails(strerror(errno));
            ReturnCode = FALSE;
         } else {
            *Eof = TRUE;
            if (Index == 0) ReturnCode = FALSE;
         }
         break;
      }
      if (Ch == '\0') break;
      if (InQuotes) {
         if (Ch == QuoteChar) {
            InQuotes = FALSE;
         }
      } else {
         if (Ch == '\n' || Ch == '\r' || Ch == '\0') {
            if (Index > 0) break;
         } else {
            if (Ch == '"' || Ch == '\'') {
               InQuotes = TRUE;
               QuoteChar = Ch;
            }
         }
      }
      if (Index < LineSize) Line[Index++] = Ch;
   }
   if (Index < LineSize) Line[Index] = '\0';
   Line[LineSize - 1] = '\0';
   
   return ReturnCode;
}

/*  ------------------------------------------------------------------------- */

/*                          Fpil__CSVNullItem()
 *
 *  Returns true if the passed item is null. Empty strings or strings that
 *  are all blank are considered to be null.
 */
 
static short Fpil__CSVNullItem (const char* Item)
{
   short IsNull = TRUE;
   const char* ItemPtr = Item;
   char Ch;
   while ((Ch = *ItemPtr++)) if (Ch != ' ') IsNull = FALSE;
   return IsNull;
}

/*  ------------------------------------------------------------------------- */

/*                          Fpil__CSVLineisComment()
 *
 *  Returns true if the passed Line is a comment. A line whose first non-blank
 *  character is '#' is a comment.
 */
 
static short Fpil__CSVLineIsComment(const char* Line)
{
   short Comment = FALSE;
   const char* LinePtr = Line;
   char Ch;
   while ((Ch = *LinePtr++)) {
      if (Ch == '#') Comment = TRUE;
      else if (Ch != ' ') break;
   }
   return Comment;
}

/*  ------------------------------------------------------------------------- */

/*                          Fpil__ScanAffectFile()
 *
 *  Does an initial pass through a CSV file. This looks for any errors in the
 *  file, and returns any date and version data it might have together with
 *  the highest fibre number for which it provides data, and the largest
 *  number of affected fibres associated with any fibre.
 */

static short Fpil__ScanAffectFile (
   FILE* InputFile,
   int* NFib,
   int* MaxAffected,
   char* DateString,
   int MaxDateChars,
   char* VersionString,
   int MaxVersionChars)
{
   char Line[2048];
   short Eof = FALSE;
   short ReadOK = TRUE;
   short Error = FALSE;
   
   if (DateString && (MaxDateChars > 1)) *DateString = '\0';
   if (VersionString && (MaxVersionChars > 1)) *VersionString = '\0';
   *MaxAffected = 0;
   *NFib = 0;
   
   while (ReadOK && !Eof && !Error) {
      char* Items[1024];
      char LineStarts[80];
      int ItemCount;
      int MaxItems = sizeof(Items)/sizeof(char*);
      
      /*  Read the next line from the file */
      
      ReadOK = Fpil__ReadCSVLine (InputFile,Line,sizeof(Line),&Eof);
      if (!ReadOK) {
         if (!Eof) Error = TRUE;
      } else {
      
         /*  If the line isn't a comment, split it up into the component
          *  items - save a copy for error logging, as the split inserts
          *  nul terminators at the end of each item in Line.
          */
          
         if (!Fpil__CSVLineIsComment(Line)) {
            strncpy(LineStarts,Line,sizeof(LineStarts));
            LineStarts[sizeof(LineStarts) - 1] = '\0';
            ReadOK = Fpil__SplitCSVLine (Line,MaxItems,&ItemCount,Items);
            if (!ReadOK) {
               Error = TRUE;
            } else {
            
               /*  Now see what this line has. We try to pick as many errors
                *  as possible in the file at this stage. We ignore lines with
                *  no items - ie blank lines.
                */
                
               if (ItemCount > 0) {
               
                  /*  Check for a version or date specification. */
                  
                  if (!strcasecmp(Items[0],"Version")) {
                     if (ItemCount >= 2) {
                        strncpy(VersionString,Items[1],MaxVersionChars);
                        VersionString[MaxVersionChars - 1] = '\0';
                     } else {
                        Fpil__CSVErrSetText(
                                     "Version line has no data",LineStarts);
                     }
                  } else if (!strcasecmp(Items[0],"Date")) {
                     if (ItemCount >= 2) {
                        strncpy(DateString,Items[1],MaxDateChars);
                        DateString[MaxDateChars - 1] = '\0';
                     } else {
                        Fpil__CSVErrSetText("Date line has no data",LineStarts);
                     }
                  } else {
                  
                     /*  Anything else should be a fibre number followed by
                      *  a set of pairs of numbers giving affected fibres
                      *  and their associated weights. Parsing the file is
                      *  complicated by the tendency of CSV files to have
                      *  null values at the end of lines  - just sets of commas.
                      *  (This is getting rather scrunched up towards the
                      *  end of the lines, suggesting this should really be
                      *  a separate routine that parses a single fibre line.)
                      */
                      
                     char* EndPtr;
                     int Fibre = strtol(Items[0],&EndPtr,0);
                     if (*EndPtr != '\0') {
                        Fpil__CSVErrSetText("Invalid fibre number",LineStarts);
                        Fpil__CSVErrSetDetails(Items[0]);
                        Error = TRUE;
                     } else {
                        int Item = 1;
                        int NAffected = 0;
                     
                        /*  See if this is the largest fibre number so far */
                        
                        if (Fibre > *NFib) *NFib = Fibre;
                        
                        /*  Look at the pairs of values, & check for validity */
                        
                        while (Item < ItemCount) {
                        
                           /*  A null item is OK, but it needs to be followed
                            *  only by null items. If not, we don't know what
                            *  to make of it, and that's an error.
                            */
                            
                           if (Fpil__CSVNullItem(Items[Item])) {
                              Item++;
                              while (Item < ItemCount) {
                                 if (!Fpil__CSVNullItem(Items[Item++])) {
                                    Fpil__CSVErrSetText(
                                       "Data follows null item",LineStarts);
                                    Fpil__CSVErrSetDetails(Items[Item - 1]);
                                    Error = TRUE;
                                    break;
                                 }
                              }
                              break;
                           }
                           
                           /*  This should be the next affected fibre # */
                           
                           (void) strtol(Items[Item++],&EndPtr,0);
                           if (*EndPtr != '\0') {
                              Fpil__CSVErrSetText(
                                 "Invalid affected fibre number",LineStarts);
                              Fpil__CSVErrSetDetails(Items[Item - 1]);
                              Error = TRUE;
                              break;
                           }
                           
                           /*  And it must be followed by a valid weight */
                           
                           if (Item >= ItemCount) {
                              Fpil__CSVErrSetText(
                                 "Missing weight specification",LineStarts);
                              Error = TRUE;
                              break;
                           } else {
                              if (Fpil__CSVNullItem(Items[Item])) {
                                 Fpil__CSVErrSetText(
                                   "Null weight specification",LineStarts);
                                 Error = TRUE;
                                 break;
                              }
                              (void) strtod(Items[Item++],&EndPtr);
                              if (*EndPtr != '\0') {
                                 Fpil__CSVErrSetText(
                                  "Invalid weight specification",LineStarts);
                                 Fpil__CSVErrSetDetails(Items[Item - 1]);
                                 Error = TRUE;
                                 break;
                              } else {
                              
                                 /*  Fibre and weight specified properly, is
                                  *  this more affected fibres than we've seen
                                  *  so far?
                                  */
                                  
                                 NAffected++;
                                 if (NAffected > *MaxAffected) {
                                    *MaxAffected = NAffected;
                                 }
                              }
                           }
                        }
                     }
                  }
               }
            }
         }
      }
   }
   rewind (InputFile);
   
   return !Error;
}

/*  ------------------------------------------------------------------------- */

/*                          Fpil__GetAffectLine()
 *
 *   Gets the next set of affectation data from a file. This assumes the
 *   file has already been scanned by Fpil__ScanAffectFile(), so it doesn't
 *   do more than minimal error checking, just for safety. It skips version
 *   and date and comment lines, and reads data from the next line that
 *   describes the fibres affected by a single fibre. It returns that fibre
 *   number in Fibre, and in Affected it returns the numbers of the other
 *   fibres affected by that fibre, and in Weights it returns the weights
 *   for those various fibres. Affected and Weights need to be arrays large 
 *   enough to hold the maximum amount of data for any one fibre (which
 *   Fpil__ScanAffectFile() returns and which should be used to allocate
 *   memory as needed), and MaxAffected should give the size of those 
 *   arrays. NAffected is set to the number of affected fibres. Returns true
 *   if it found any new data to return - so should just be called
 *   repeatedly until it returns false.
 */
static short Fpil__GetAffectLine (
   FILE* InputFile,
   int MaxAffected,
   int* Fibre,
   int* NAffected,
   int *Affected,
   float *Weights)
{
   char Line[2048];
   short Eof = FALSE;
   short ReadOK = TRUE;
   short Error = FALSE;
   
   char* Items[1024];
   int ItemCount;
   int MaxItems = sizeof(Items)/sizeof(char*);

   /*  Read the next line from the file */

   short GotData = FALSE;
   
   while (!GotData && !Eof && !Error) {
   
      ReadOK = Fpil__ReadCSVLine (InputFile,Line,sizeof(Line),&Eof);
      if (!ReadOK) {
         if (!Eof) Error = TRUE;
      } else {
      
         /*  If the line isn't a comment, split it up into the component
          *  items.
          */

         if (!Fpil__CSVLineIsComment(Line)) {
            ReadOK = Fpil__SplitCSVLine (Line,MaxItems,&ItemCount,Items);
            if (!ReadOK) {
               Error = TRUE;
            } else {

               /*  Now see what this line has. If it isn't blank and starts
                *  with a valid integer, it's going to be the list of affected
                *  fibres for a given fibre. Anything else may be a date or
                *  version spec, but we don't care about them.
                */ 

               if (ItemCount > 0) {

                  char* EndPtr;
                  *Fibre = strtol(Items[0],&EndPtr,0);
                  if (*EndPtr == '\0') {

                     /*  Look at the pairs of values, & check for validity */

                     int Item = 1;
                     int AffCount = 0;
                     while (Item < ItemCount) {

                        int AffFibre;
                        float Weight;

                        /*  Break out on a null item. */

                        if (Fpil__CSVNullItem(Items[Item])) break;

                        /*  This should be the next affected fibre #. All the
                         *  breaks in the next bit should set Error, so it's
                         *  easier to set it now and clear it at the end.
                         */

                        Error = TRUE;
                        AffFibre = strtol(Items[Item++],&EndPtr,0);
                        if (*EndPtr != '\0') break;

                        /*  And it must be followed by a valid weight */

                        if (Item >= ItemCount) break;
                        Weight = strtod(Items[Item++],&EndPtr);
                        if (*EndPtr != '\0') break;

                        if (AffCount < MaxAffected) {
                           Affected[AffCount] = AffFibre;
                           Weights[AffCount] = Weight;
                           AffCount++;
                        }
                        Error = FALSE;
                     }
                     *NAffected = AffCount;
                     GotData = TRUE;
                  }
               }
            }
         }
      }
   }
   
   /*  We don't expect any errors, but if we get one we should report
    *  something. It indicates that Fpil__ScanAffectFile() has fallen down
    *  on the job. Because this shouldn't happen other than in testing, I've 
    *  not gone to a lot of trouble to record error details.
    */
    
   if (Error) Fpil__CSVErrSetText("Unexpected error reading from file",NULL);

   return GotData;
}

/*  ------------------------------------------------------------------------- */

/*                          FpilAffectErrorText()
 *
 *  If FpilGetAffectData() (or any other of the lower level routines) hits
 *  an error, they make calls to record the details of the error. This
 *  routine collects all these details and constructs a single string
 *  that can be used to make a useful error report. Error should be a char
 *  array large enough to hold the error report, and Length should be the
 *  number of charactrs it can hold (including a terminating nul). If necessary
 *  the string returned in Error will be truncated, but as much information as
 *  possible will be returned. Error will always be returned properly nul-
 *  terminated.
 */
 
void FpilAffectErrorText (int Length, char* Error)
{
   Error[0] = '\0';
   if (G_ErrorFlag) {
   
      /*  This code does a lot of calls to strlen(Error) which could be
       *  avoided with rather more work keeping track of the length of
       *  the error description as it builds up. But this isn't going
       *  to be called unless something bad has happened, so simple wins.
       */
      
      if (G_Filename[0] != '\0') {
         strncpy (Error,"Error reading from CSV file ",Length);
         strncat (Error,G_Filename,Length - strlen(Error) - 1);
         strncat (Error,". ",Length - strlen(Error) - 1);
      }
      if (G_ErrorText[0] != '\0') {
         strncat (Error,G_ErrorText,Length - strlen(Error) - 1);
      }
      if (G_ErrorDetails[0] != '\0') {
         strncat (Error," (",Length - strlen(Error) - 1);
         strncat (Error,G_ErrorDetails,Length - strlen(Error) - 1);
         strncat (Error,")",Length - strlen(Error) - 1);
      }
      if (G_ErrorLine[0] != '\0') {
         strncat (Error," in line starting: ",Length - strlen(Error) - 1);
         strncat (Error,G_ErrorLine,Length - strlen(Error) - 1);
      }
      Error[Length - 1] = '\0';
   }
}

/*  ------------------------------------------------------------------------- */

/*                          FpilSetAffectError()
 *
 *  FpilSetAffectError() allows a caller to explicitly set the text of the
 *  message returned by FpilAffectErrorText(), and is useful in some cases 
 *  where a routine, for example, doesn't call FpilGetAffectData() because it
 *  knows a file doesn't exist, and want to provide a more detailed mesage 
 *  (perhaps including search path details) than FpilGetAffectData() would 
 *  produce.
 */
 
void FpilSetAffectError (
   const char* Text)
{
   G_Filename[0] = '\0';
   G_ErrorLine[0] = '\0';
   G_ErrorDetails[0] = '\0';
   Fpil__CSVErrSetText(Text,NULL);
}

/*  ------------------------------------------------------------------------- */

/*                          FpilClearAffectData()
 *
 *  FpilClearAffectData() clears out the fields of an FpilAffectDataType
 *  structure. It assumes the structure is uninitialised, so assumes there
 *  are no arrays to release, unlike FpilReleaseAffectData().
 */
 
void FpilClearAffectData (
   FpilAffectDataType* AffectData)
{
   AffectData->Plate = 0;
   AffectData->Version[0] = '\0';
   AffectData->Date[0] = '\0';
   AffectData->NFib = 0;
   AffectData->MaxAffected = 0;
   AffectData->AffectedFibres = NULL;
   AffectData->AffectedWeights = NULL;
   AffectData->AffectedCounts = NULL;
}

/*  ------------------------------------------------------------------------- */

/*                          FpilReleaseAffectData()
 *
 *  FpilGetAffectData() allocates arrays to hold the affectation data it reads
 *  from the input file, and stores their addresses in the FpilAffectDataType
 *  structure it is passed. These will eventually need to be released, and
 *  this routine packages up what is needed to release them. This can be
 *  called multiple times for the same structure without causing problems,
 *  although nothing happens on the subsequent calls.
 */
 
void FpilReleaseAffectData (
   FpilAffectDataType* AffectData)
{
   if (AffectData->AffectedFibres) free(AffectData->AffectedFibres);
   if (AffectData->AffectedWeights) free(AffectData->AffectedWeights);
   if (AffectData->AffectedCounts) free(AffectData->AffectedCounts);
   FpilClearAffectData (AffectData);
}

/*  ------------------------------------------------------------------------- */

/*                          FpilGetAffectData()
 *
 *  This is the main routine in this code. It is passed the name of a file
 *  containing affectation data for a given plate, and it fills up a
 *  structure of type FpilAffectDataType with that data. The contents of
 *  the structure and the file format used are described in detail in
 *  the comments in FpilAffect.h. The structure should eventually be passed
 *  to FpilReleaseAffectData() to release any memory allocated by this
 *  routine. This routine returns true if it was able to read the data
 *  properly; if it returns false, call FpilAffectErrorText() to get a
 *  description of the problem.
 */
 
short FpilGetAffectData (
   int Plate, 
   const char* Filename,
   FpilAffectDataType* AffectData)
{
   float* LineWeights = (float*) NULL;
   int* LineFibres = (int*) NULL;
   FILE* InputFile = (FILE*) NULL;
   short ReturnOK = TRUE;
   
   FpilClearAffectData(AffectData);
   
   /*  Set up error logging and open the file */
   
   Fpil__CSVClearErrors();
   Fpil__CSVErrSetFilename(Filename);

   InputFile = fopen (Filename,"r");
   if (InputFile == NULL) {
      Fpil__CSVErrSetText ("Failed to open input file",NULL);
      Fpil__CSVErrSetDetails(strerror(errno));
      ReturnOK = FALSE;
   } else {
   
      /*  Do a first pass through the file to a) check that it parses properly
       *  b) find out the maximum size of the arrays we'll need, both locally
       *  to handle one line and globally in the AffectData structure and
       *  c) get any version and date information out of it.
       */
       
      int NFib;
      int MaxAffected;
      if (!Fpil__ScanAffectFile (InputFile,&NFib,&MaxAffected,AffectData->Date,
                                 sizeof(AffectData->Date),AffectData->Version,
                                                 sizeof(AffectData->Version))) {
         ReturnOK = FALSE;
      } else {
      
         /*  Allocate local work arrays that can hold the affected fibres and
          *  corresponding weight values from one line of the file (ie for one
          *  fibre), and the larger arrays to hold all the affected fibre 
          *  numbers and weights (which will be NFib times the size of the
          *  single line arrays).
          */
          
         LineWeights = (float*) malloc (NFib * sizeof(float));
         LineFibres = (int*) malloc (NFib * sizeof(int));
         AffectData->AffectedFibres = 
                           (int*) malloc (NFib * MaxAffected * sizeof(int));
         AffectData->AffectedCounts = 
                           (int*) malloc (NFib * MaxAffected * sizeof(int));
         AffectData->AffectedWeights = 
                          (float*) malloc (NFib * MaxAffected * sizeof(float));
         
         if (!LineWeights || !LineFibres || !AffectData->AffectedCounts ||
                 !AffectData->AffectedFibres || !AffectData->AffectedWeights) {
            Fpil__CSVErrSetText (
                    "Failed to allocate memory for affected fibre data",NULL);
            ReturnOK = FALSE;
         } else {
            int Fibre;
            int NAffected;
            
            /*  Initialise the fibre data in the AffectData structure. It should
             *  be enough just to set the number of fibres affected by each
             *  individual fibre to zero. Note that Fibre numbers start at 1.
             */
            
            AffectData->Plate = Plate;
            AffectData->NFib = NFib;
            AffectData->MaxAffected = MaxAffected;
            for (Fibre = 1; Fibre <= NFib; Fibre++) {
               AffectData->AffectedCounts[Fibre] = 0;
            }
            
            /*  Loop through the file, each time getting the data for one
             *  fibre, and copy that into the arrays in AffectData.
             */
             
            while (Fpil__GetAffectLine (InputFile,MaxAffected,&Fibre,
                                           &NAffected,LineFibres,LineWeights)) {
               if (Fibre >= 1 && Fibre <= NFib) {
                  int I;
                  for (I = 0; I < NAffected; I++) {
                     AffectData->AffectedFibres[(Fibre - 1) * MaxAffected + I] = 
                                                               LineFibres[I];
                     AffectData->AffectedWeights[(Fibre - 1) * MaxAffected + I] = 
                                                               LineWeights[I];
                  }
                  AffectData->AffectedCounts[Fibre - 1] = NAffected;
               }
            }
         }
      }
   }
   
   /*  Release any local resources. If things went wrong, also release anything
    *  allocated for the AffectData structure.
    */

   if (InputFile) fclose(InputFile);
   if (LineWeights) free(LineWeights);
   if (LineFibres) free(LineFibres);
   if (!ReturnOK) FpilReleaseAffectData(AffectData);
   
   return ReturnOK;
}

/*  ------------------------------------------------------------------------- */

/*                             main() - Test program
 *
 *  A very basic test program for the code in this file, only compiled if
 *  FPIL_AFFECT_TEST is defined. See the main comments in this file for
 *  a summary of how to run the test.
 */
 
#ifdef FPIL_AFFECT_TEST
                 
int main (int argc, char* argv[])
{
   if (argc < 2) {
      printf ("Needs filename as argument\n");
   } else {
      FpilAffectDataType AffectData;
      if (FpilGetAffectData (0,argv[1],&AffectData)) {
         int Fibre;
         int MaxAffected = AffectData.MaxAffected;
         for (Fibre = 1; Fibre <= AffectData.NFib; Fibre++) {
            if (AffectData.AffectedCounts[Fibre - 1] > 0) {
               printf ("Fibre %d affects: ",Fibre);
               int I;
               for (I = 0; I < AffectData.AffectedCounts[Fibre - 1]; I++) {
                  printf ("fibre %d (%.2f) ",
                     AffectData.AffectedFibres[(Fibre - 1) * MaxAffected + I],
                     AffectData.AffectedWeights[(Fibre - 1) * MaxAffected + I]);
               }
               printf ("\n");
            }
         }
      } else {
         char Error[256];
         FpilAffectErrorText (sizeof(Error),Error);
         printf ("%s\n",Error);
      }
      FpilReleaseAffectData(&AffectData);
         
   }
   return 0;
}          
        
#endif
