/*
 *                      F p i l  A f f e c t . h
 *
 *  Function:
 *     Routines that handle the data relating to 'affected' fibres.
 *
 *  Description:
 *     In this context, an 'affected' fibre is a fibre that is affected by
 *     spillover or ghosts or some other effect that results in light from
 *     one target contaminating the spectrum of another. In general, we
 *     want to arrange things so that as far as possible we do not have
 *     light from bright targets contaminating the spectra of fainter 
 *     targets. To do this, we need to be able to find out which fibres are
 *     affected by any given fibre. Generally, a fibre will affect more than
 *     one other fibre - a ghost may cover a number of other fibres, 
 *     crosstalk will affect adjacent fibres and even, to a lesser effect,
 *     fibres adjacent to them and so on.
 *
 *     Affected fibres are tabulated in text files in a CSV (comma separated
 *     value) format, most easily generated from a spreadsheet. This makes
 *     their maintenance relatively simple. There are (or at least, can be)
 *     separate files for each instrument and each plate. Each file has an 
 *     entry for each of the fibres for the instrument, giving a list of
 *     affected fibre numbers and a weight for each fibre.
 *
 *  Usage:
 *     This file defines the main routine for accessing affected fibre
 *     data (FpilGetAffectData()) as well as the utility routines
 *     FpilAffectErrorText(), FpilReleaseAffectData() and FpilSetAffectError().
 *
 *     FpilGetAffectData() is passed the name of a file containing affected
 *     fibre data - either from ghosts or from crosstalk (the format is the
 *     same for both), and fills up a structure of type FpilAffectDataType
 *     - also defined in this file - with the relevant information. 
 *   
 *     The FpilAffectDataType structure contains the addresses of arrays
 *     allocated to hold the data, and these need to be released eventually
 *     through a call to FpilReleaseAffectData(). 
 *
 *     If FpilGetAffectData() returns an error indication, a call to
 *     FpilAffectErrorText() will fill a supplied character string with
 *     a reasonably detailed description of the error. FpilSetAffectError()
 *     allows a caller to explicitly set the text of the message returned
 *     by FpilAffectErrorText(), and is useful in some cases where a routine,
 *     for example, doesn't call FpilGetAffectData() because it knows a file
 *     doesnt exist, and want to provide a more detailed mesage (perhaps 
 *     including search path details) than FpilGetAffectData() would produce.
 *
 *  File format:
 *     o  The CSV files should have lines with some combination of carriage 
 *        return or newline characters to indicate the end of each line. The
 *        actual separators can be tabs, commas or spaces. Excel will generate
 *        suitable files using any of the 'comma separated variables', 
 *        'space delimited text' or 'tab delimited text' output options. Each 
 *        line will be split up into items, delimited by the separator in use.
 *        If the separator is a space, multiple consecutive spaces are treated
 *        as a single separator.
 *     o  Any line that starts with '#' is treated as a comment.
 *     o  Any line that whose first item is "Version" (case insensitive) can 
 *        be used to specify the file version; the second item on the line 
 *        should be a string giving the version number.
 *     o  Any line that whose first item is "Date" (case insensitive) can be 
 *        used to specify the file date; the second item on the line should be
 *        a string giving the date. Any string can be used for version and date.
 *        If there are multiple date or version lines, the last is used.
 *     o  Any other line should have as its first item the number of a fibre
 *        (fibre numbers start at 1). It should be followed by pairs of items
 *        that give the number of an affected fibre and the weight for the
 *        affectation.
 *
 *     For example:
 *
 *     # Sample crosstalk data
 *     Version 1.1
 *     Date 23/07/2013
 *       1  2  0.5  404  0.5
 *       2  1  0.5    3  0.5
 *       9  8  0.5   10  0.5
 *      10  9  0.5   11  0.5
 *      12 11  0.5   13  0.5  10  0.1  14  0.1
 *
 *     Obviously, most files will have more data than this. This file shows
 *     fibres 1,2,9 and 10 as only affecting thir immediately adjacent
 *     fibres (1 is between fibre 2 and 404 in 2dF), while fibre 12 for some
 *     reason also spills slightly into fibre 10 and 14.   
 *
 *  Support: Keith Shortridge, AAO
 *
 *  History:
 *     26th Jun 2013. Original version. KS.
 *     21st Jul 2013. First fully fleshed out implementation. KS.
 *      2nd Aug 2013. Added FpilSetAffectError(). KS.
 */

#ifndef __FpilAffect__
#define __FpilAffect__

/*  There are separate structures of type FpilAffectDataType for ghosts
 *  and for cross-talk for each plate supported by the instrument. Each one
 *  is filled by a call to FpilGetAffectData(), specifying a file that has the
 *  relevant data. So there will be separate files for each set of affectation
 *  data - for each instrument, for each plate, and each affectation type.
 *
 *  Note that the format used is potentially a bit wasteful. Fibres are numbered
 *  from 1 up. A file may not have data for every fibre, and in any case the
 *  maximum fibre number depends on the instrument. The highest numbered fibre
 *  for which data was found in the file is set as NFib in the structure, and
 *  the arrays have entries for all fibres from 1 up to a including NFib. (Which
 *  would be wasteful if the file only had data for fibre numbers 403 and 404,
 *  but that probably isn't going to happen. Most of the time the files will
 *  have entries for each fibre.)
 *
 *  Different fibres potentially affect a varying number of other fibres, to
 *  varying extents. To simplify things, we assume the number of affected
 *  fibres won't vary too much, and we store the lists of affected fibres in 
 *  a 2D array with one row for each fibre, and with each row long enough to
 *  hold the maximum number of affected fibres found by examining the file.
 *  This maximum number of affected fibres is given by MaxAffected.
 *  AffectedCounts is an array of NFib elements, AffectedFibres is an array
 *  of NFib * MaxAffected elements, as is AffectedWeights.
 *
 *  So, to determine the affectation data for fibre N, say, in this structure
 *  AffectedCounts[N-1] will give the number of other fibres affected by
 *  fibre N. (It's N-1 because fibre nubers start at 1, not zero.) 
 *  AffectedFibres[(N-1)*MaxAffected] gives the number of the first of
 *  these affected fibres, and subsequent fibre numbers follow. In the same
 *  way, AffectedWeights[(N-1)*MaxAffected] gives the weighting of the
 *  affectation for the first affected fibre and subsequent entries give
 *  the weights for the other affected fibres.
 */
 
typedef struct {
   char Version[80];         /*  String giving version number for file */
   char Date[80];            /*  String giving date of file */
   int Plate;                /*  Plate number, starting from 0 */
   int MaxAffected;          /*  Largest number of affected fibres */
   int NFib;                 /*  Number of fibres. Fibres go from 1 to NFib */
   int* AffectedCounts;      /*  The number of affected fibres for each fibre */
   int* AffectedFibres;      /*  2D array listing affected fibres */
   float* AffectedWeights;   /*  2D array listing weights for affected fibres */
} FpilAffectDataType;

/*  FpilGetAffectData() is passed the number of the plate in question, the
 *  name of the file that contains the data about the affected fibres, and
 *  the address of an FpilAffectDataType() structure that it will fill with
 *  the data from the file. It returns true if it reads the file without
 *  problems, and false if something has gone wrong. If it returns false,
 *  a call to FpilAffectErrorText() will provide an error description that
 *  can be output. If it returns true, the caller needs to make use of the
 *  data in the returned structure, and then finally needs to call
 *  FpilReleaseAffectData() to release any allocated memory connected with
 *  the structure. (If FpilGetAffectData() returns an error indication, it
 *  will already have called FpilReleaseAffectData(), so the data in the
 *  structure must not be used if an error is signalled. Calling 
 *  FpilReleaseAffectData() twice for the same structure is safe, so it can 
 *  be called directly even after an error return from FpilGetAffectData() 
 *  if that happens to be convenient.) FpilClearAffectData() simply clears
 *  out all the fields of what is assumed to be an uninitialised structure.
 */
 
short FpilGetAffectData (
   int Plate, 
   const char* Filename,
   FpilAffectDataType* AffectData);

void FpilAffectErrorText (
   int Length,
   char* Error);

void FpilReleaseAffectData (
   FpilAffectDataType* AffectData);
   
void FpilClearAffectData (
   FpilAffectDataType* AffectData);

void FpilSetAffectError (
   const char* Text);

#endif
