#ifndef _FPCOLMDLCALCONVERT_H_
#define _FPCOLMDLCALCONVERT_H_

/*******************************************************************************
* E.S.O. - VLT project
*
* "@(#) $Id: ACMM:ozpozFpcol/ws/include/fpcolMdlCalConvert.h,v 1.63 07-Oct-2008 11:10:31+10 tjf $"
*
*
* who       when      what
* --------  --------  ----------------------------------------------
* tfarrell  12/04/02  Added raToString, decToString, mjdToString
*                     stringToMjd, stringToPosition, stringToObjectPos. 
* ssmedley  07/09/01  created
*
*/

/************************************************************************
*   NAME
*	toFractions toRadians meanToApparent -
*	Miscellaneous conversion routines used by the FLAMES/OzPoz
*	Calibration s/w.
*
*   SYNOPSIS
*       #include "fpcolMdlCalConvert.h"
*
*	double toFractions (int degrees, int minutes, double seconds);
*
*	double toRadians (double encoded);
*
*	void meanToApparent
*		(double		meanRA,
*		 double		meanDEC,
*		 double		properMotionRA,
*		 double		properMotionDEC,
*		 double		params[21],
*		 double		*apparentRA,
*		 double		*apparentDEC);
*
*       const char *raToString(double ra);
*       const char *decToString(double dec);
*       const char *mjdToString(double mjd);
*
*       int stringToMjd(char *buffer,
*                      double *mjd);
*       int stringToPosition(char *buffer,
*                            double *ra,
*                            double *dec);
*      int stringToObjectPos(char *buffer,
*                            double *pars,
*                            int    *button,
*                            double *meanRA,
*                            double *meanDEC,
*                            double *appRA,
*                            double *appDEC,
*                            double *pmRA,
*                            double *pmDEC);
*
*   DESCRIPTION
*	This library provides conversion routines used by the
*	FLAMES/OzPoz calibration s/w.
*
*	toFractions() converts <degrees> <minutes> <seconds> into fractions of
*	degrees. ie. 10d 30m 0s == 10.5 d
*	Note: NO range checking is performed.
*
*	toRadians() converts an input double of the form
*	DDMMSS.SS
*	DD = degrees
*	MM = arcminutes
*	SS.SSD = arcseconds
*	into radians
*
*	meanToApparent() provides a more user-friendly interface to the
*	mean-to-apparent conversion routine in Pat Wallace's slalib library.
*	<params[21]> must be initialsed with a call to <slaMappa> before
*	invoking this routine.
*
*       raToString() converts an RA in radians to a string representation.  Returns
*       a pointer to a static string.
*
*       decToString() converts a Declination in radians to a string representation. Returns
*       a pointer to a static string.
*
*       mjdToString() converts an Time and Date as an MJD to a string representation. Returns
*       a pointer to a static string.
*
*       stringToMjd() converts a string in the format returned by mjdToString() back into
*       a MJD
*
*       stringToPosition() converts a string containing an RA and DEC specification as 
*       produced by raToString() and decToString() back into radians.
*
*       stringToObjectPos() converts a string containing a button number, RA and DEC
*       into the required information (Proper motion currently set to zero).
*
*   ENVIRONMENT
*
*   CAUTIONS
*
*	No range checking is performed on toFractions().
*
*   EXAMPLES
*
*   SEE ALSO
*
*	slaMappa() & slaMapqk() in Pat Wallace's slalib library.
*
*   BUGS
*
*------------------------------------------------------------------------
*/



#include <slalib.h>	/* slaMappa slaMapqk */

/*
 * This function converts <degrees> <minutes> <seconds>
 * into fractions of degrees.
 * ie. 10d 30m 0s == 10.5 d
 * Note: NO range checking is performed.
 */
double toFractions (int degrees, int minutes, double seconds);

/*
 * This function converts an input double of the form
 * DDMMSS.SS
 * DD = degrees
 * MM = arcminutes
 * SS.SSD = arcseconds
 * into radians
 */
double toRadians (double encoded);


/* convert RA/DEC coordinates from mean to apparent */
void meanToApparent
	(double		meanRA,		/* radians */
	 double		meanDEC,	/* radians */
	 double		properMotionRA,		/* radians */
	 double		properMotionDEC,	/* radians */
	 double		params[21],
	 double		*apparentRA,
	 double		*apparentDEC);


const char *raToString(double ra);

const char *decToString(double dec);
const char *mjdToString(double mjd);

int stringToMjd(char *buffer,
               double *mjd);
int stringToPosition(char *buffer,
                    double *ra,
                    double *dec);
int stringToObjectPos(char *buffer,
                      double *pars,
                      int    *button,
                      double *meanRA,
                      double *meanDEC,
                      double *appRA,
                      double *appDEC,
                      double *pmRA,
                      double *pmDEC);


#endif
