/*******************************************************************************
* E.S.O. - VLT project
*
* "@(#) $Id: ACMM:ozpozFpcol/ws/src/fpcol.c,v 1.63 07-Oct-2008 11:10:26+10 tjf $"
* 
*
* who       when      what
* --------  --------  ----------------------------------------------
* rschmutz  23/12/02  PPRS8283: fpcolInvPos: fibre type arg. added.
* tfarrell  20/03/02  FiducialFibre was not accounting for the
*                     fiducial and fibre sizes.  Revamp and fix
*                     a bug which occurs in the fibre parallel to
*                     y axis case.
* tfarrell  24/08/01  Fix up MAN page generation.
* dpopovic  11/04/01  2 PI correction in fpcol___rta2xyt()
* tfarrell  22/06/00  created
*/

/************************************************************************
*   NAME
*       fpcolButFib fpcolButBut fpcolFiducial fpcolInvPos fpcolFibFib - 
*       library routines to perform fibre collision detectionn.
*
*   SYNOPSIS
*       #include "fpcol.h"
*
*
*       vltLOGICAL  fpcolButFib (
*                                IN vltDOUBLE butR,
*                                IN vltDOUBLE butTheta,
*                                IN vltDOUBLE butAlpha,
*                                IN vltDOUBLE fibR,
*                                IN vltDOUBLE fibTheta,
*                                IN vltDOUBLE pivR,
*                                IN vltDOUBLE pivTheta,
*                                IN vltINT32  tolerance);
*
*       vltLOGICAL  fpcolButBut (
*                                   IN vltDOUBLE butAR,
*                                   IN vltDOUBLE butATheta,
*                                   IN vltDOUBLE butAAlpha,
*                                   IN vltDOUBLE butBR,
*                                   IN vltDOUBLE butBTheta,
*                                   IN vltDOUBLE butBAlpha,
*                                   IN vltINT32  tolerance);
*
*       vltLOGICAL  fpcolFiducial (
*                                   IN vltDOUBLE fibR,
*                                   IN vltDOUBLE fibTheta,
*                                   IN vltDOUBLE pivR,
*                                   IN vltDOUBLE pivTheta,
*                                   IN vltDOUBLE markR,
*                                   IN vltDOUBLE markTheta,
*                                   IN vltINT32  tolerance);
*
*       vltLOGICAL  fpcolInvPos (
*                                   IN vltUINT32  plate ,
*				    IN vltINT32   fibType,
*                                   IN vltDOUBLE  fibR,
*                                   IN vltDOUBLE  fibTheta,
*                                   IN vltDOUBLE  fibAlpha,
*                                   IN vltINT32   tolerance);
*
*       vltLOGICAL  fpcolFibFib (
*                                IN vltDOUBLE  fibAR,
*                                IN vltDOUBLE  fibATheta,
*                                IN vltDOUBLE  pivAR,
*                                IN vltDOUBLE  pivATheta,
*                                IN vltDOUBLE  fibBR,
*                                IN vltDOUBLE  fibBTheta,
*                                IN vltDOUBLE  pivBR,
*                                IN vltDOUBLE  pivBTheta,
*                                IN vltDOUBLE  tolerance);
*
*
*
*
*   DESCRIPTION
*      These routines provide fibre collision detection facilities
*      for the FLAMES Fibre Positioner - OzPoz.  The underlying collision
*      routines are implemented by fpcolBase.c, and are common to OzPoz and
*      the AAO configure program used to prepare fields.  As a result,
*      this module is largely an interface layer between the two, as
*      the AAO configure program wants coorinates in X and Y rather then
*      R/Theta.
*
*      Where possilble, this module performs quick checks to determine
*      if a collision may be possible before continuing with a detailed
*      check.
*
*      Note on coordinate systems and angles.
*
*          The OzPoz robot uses an R/Theta coorindate system with an
*          angle "alpha" indicating the direction of the fibre from
*          the button.  The AAO configure software uses an X/Y/Theta
*          coorindate system, where Theta provides information equivalent 
*          to the OzPoz angle alpha,
*      
*          OzPoz alpha is defined with alpha = 0 along the positive
*          X axis, with the angle being the angle of the line from
*          the fibre pivot point to the fibre.
*           
*          The configure software uses theta = 0 along the positive Y
*          axis, instead of the positive X.   So to convert alpha OzPoz 
*          to configure  orientation, we must rotate by PI/2.
*      
*     Units.  Angles are in radians.  R positions are in microns.  Tolerances
*           are in microns.
*      
*     fpcolButFib - This checks for a collision between a fibre button
*          and a fibre.  The fibre coorindate of the button is given 
*          by butR/butTheta and the angle of which is butAlpha.  The button
*          end coorindate of the fibre is given by fibR/fibTheta whilst the
*          pivot end coorindate is given by pviR/pivTheta.   This function
*          returns true if a collision has occured within the specified
*          tolerance.
*
*
*     fpcolButBut - This checks for a collision between a two fibre buttons
*          The fibre coorindate of the first button is given 
*          by butAR/butATheta and the angle of which is butAAlpha whilst
*          the fibre coorindate of the second button is given by
*          by butBR/butBTheta and the angle of which is butBAlpha.
*          This function returns true if a collision has occured within the 
*          specified tolerance.
*
*     fpcolFiducial - This fchecks for the obstruction of a fiducal by
*          a fibre or fibre button.  The button coorindates of the fibre
*          are given by fibR/fibTheta, whilst the coorindates of the
*          pivot end of the fibre are given by pivR/PivTheta.  The coorindates
*          of the fiducal mark are given by markR/markTheta.  This function
*          returns true if the fibre obscures the fiducial wihin the
*          specified tolerance.
*
*     fpcolInvPos - This checks if specified fibre position is invalid.
*          This may be plate specific so the plate argument indicates
*          the plate involved, base 1.
*          The fibre button coorindates are given by fibR, fibTheta, with
*          the rotation given by fibAlpha.
*          This function returns true if the position is invalid within
*          the specified tolerance.
*
*     fpcolFibFib - This function checks for crossing fibres.  The two
*          end of the fibre fibre are given by fibAR/fibATheta and
*          pivAR,pivATheta.  Simularly for the second fibre.  This function
*          returns true if the position if the fibres cross within the
*          specified tolerance.
*
*
*      
***
*
*   RETURN VALUES 
*       TRUE or FALSE, TRUE indicates a collision has occured.
*
*   ENVIRONMENT
*       lopoz - Flames Fibre Positioner.
*
*   CAUTIONS 
*
*   EXAMPLES
*
*   SEE ALSO
*      fpnvr module
*   BUGS   
*
*------------------------------------------------------------------------
*/
/*
 * Define module id for error macros
 */
#define MODULE_ID     "fpcol"
#define lccMODULE_ID  "fpcol"

#define _POSIX_SOURCE 1

static const char *rcsId="@(#) $Id: ACMM:ozpozFpcol/ws/src/fpcol.c,v 1.63 07-Oct-2008 11:10:26+10 tjf $"; 
static void *use_rcsId = (0 ? (void *)(&use_rcsId) : (void *) &rcsId);

#include "vltPort.h"    /* Must be first header */

#include <stdio.h>
#include <math.h>
/* 
 * Local Headers
 */
#include "fpDefines.h"
#include "fpcol.h"
#include "fpcolBase.h"
#include "fpcolErrors.h"
#include "fpcolInstDescr.h"
#include "fppt.h"

/*
 *
 */

/*
 * Macro function definitions
 */
#define SQRD(x)           ((double)(x)*(double)(x))  

/*
 * Define the INLINE macro, use to make a function INLINE when compiling
 * under GCC
 */
#if defined(__GNUC__) || defined(GNUC)
#   define INLINE __inline__ static
#else
#   define INLINE static
#endif

/*
 * A DEBUG macro - FPCOL_DEBUG enables some checks but makes this
 * module run slower.
 */
/*#define FPCOL_DEBUG */
#ifdef FPCOL_DEBUG
#warning "FPCOL Debug enabled, checks will run much slower"
#endif

#define MAX_BUT_SIZE buttonClear + fibreClear + MaxSize

/*
 *  Type used with the Bounding Box checks
 */
typedef struct {
    double x1;
    double x2;
    double y1;
    double y2;
} BoundingBox;

/*
 *  Type given a position in x/y/theta terms.
 */
typedef struct {
    double x;
    double y;
    double theta;
} ButXYT;

/*
 * Function which creates a bounding box around a fibre.
 */
INLINE void fpcol___fibreBox(
    const ButXYT * const fibreEnd,  /* The fibre button end -  co-orindates, 
                                      theta ignored  */
    const ButXYT * const pivotEnd,  /* The fibre pivot  end -  co-orindates, 
                                      theta ignored  */
    BoundingBox  * const box)
{
    if (pivotEnd->x > fibreEnd->x)
    {
        box->x1 = fibreEnd->x - fpcolINST_MAX_BUT_SIZE_CLR;
        box->x2 = pivotEnd->x + fpcolINST_MAX_BUT_SIZE_CLR;
    }
    else
    {
        box->x1 = pivotEnd->x - fpcolINST_MAX_BUT_SIZE_CLR;
        box->x2 = fibreEnd->x + fpcolINST_MAX_BUT_SIZE_CLR;
    }
        
    if (pivotEnd->y > fibreEnd->y)
    {
        box->y1 = fibreEnd->y - fpcolINST_MAX_BUT_SIZE_CLR;
        box->y2 = pivotEnd->y + fpcolINST_MAX_BUT_SIZE_CLR;
    }
    else
    {
        box->y1 = pivotEnd->y - fpcolINST_MAX_BUT_SIZE_CLR;
        box->y2 = fibreEnd->y + fpcolINST_MAX_BUT_SIZE_CLR;
    }
        
}


/*
 *  Function which creates a bounding box about a fibre button.
 */
INLINE void fpcol___buttonBox(
    const ButXYT * const fibre,
    BoundingBox  * const box)
{
    box->x1 = fibre->x - fpcolINST_MAX_BUT_SIZE_CLR;
    box->x2 = fibre->x + fpcolINST_MAX_BUT_SIZE_CLR;
    box->y1 = fibre->y - fpcolINST_MAX_BUT_SIZE_CLR;
    box->y2 = fibre->y + fpcolINST_MAX_BUT_SIZE_CLR;
}

/*
 * Function which returns true if we have a collision between two
 * boxes
 */
INLINE int fpcol___boxCollision(
    const BoundingBox * const box1,
    const BoundingBox * const box2)
{
    int safe = 0;
   
    if (box1->x2 < box2->x1)
        safe = 1;                /* Box 1 is too far left of box 2 to overlap*/
    else if (box2->x2 < box1->x1)
        safe = 1;                /* Box 2 is too far left of box 1 to overlap*/
    if (box1->y2 < box2->y1)
        safe = 1;                /* Box 1 to too much lower of box 2 */
    else if (box2->y2 < box1->y1)
        safe = 1;                /* Box 2 is too much lower of box 1 */
             
    return !safe;                
}


/*
 * Function to convert an OzPoz button location in r/theta/alpha to
 * a configure style location in x/y/t
 */
static void fpcol___rta2xyt(
                            IN  vltDOUBLE r,
                            IN  vltDOUBLE theta,
                            IN  vltDOUBLE alpha,
                            OUT ButXYT    * const posXYT)
{

    fpPOS pointRT = { r, theta };
    fpPOS pointXY;
    /*
     * Convert the point to X/Y,.
     */
    fpptRT2XY(pointRT, &pointXY);
    posXYT->x = pointXY.x;
    posXYT->y = pointXY.y;
    /*
     * Convert OzPoz alpha to Configure theta.
     */
    posXYT->theta = alpha + M_PI_2;
    if (posXYT->theta >  2.0 * M_PI)
        posXYT->theta -= 2.0 * M_PI;
}

/*
 * Function to convert an OzPoz fibre button location in r/theta with the
 * pivot point given in r/theta to a configure style location in x/y/t.
 *
 * We also return the pivot end in x/y/t, since we calulcate this anyway
 * and it may be needed.
 */
static void fpcol___rtp2xyt(
                            IN  vltDOUBLE butR,
                            IN  vltDOUBLE butTheta,
                            IN  vltDOUBLE pivR,
                            IN  vltDOUBLE pivTheta,
                            OUT ButXYT    * const fibXYT,
                            OUT ButXYT    * const pivXYT)
{
    double tmpdx, tmpdy;  /* Temps with sign adjusted */
    vltINT32 dx;         /* Easy access to x and y values */
    vltINT32 dy;
    fpPOS pointRT = { butR, butTheta };
    fpPOS pointXY;
    double *t = &fibXYT->theta;
    /*
     * Convert the button end to X/Y,.
     */
    fpptRT2XY(pointRT, &pointXY);
    fibXYT->x = pointXY.x;
    fibXYT->y = pointXY.y;

    /*
     * Convert the pivot end end to X/Y.
     */
    pointRT.x = pivR;
    pointRT.y = pivTheta;
    fpptRT2XY(pointRT, &pointXY);

    pivXYT->x = pointXY.x;
    pivXYT->y = pointXY.y;



    
    /*
     * Determine configure theta.  
     * 
     */
    dx = pointXY.x;   /* Easy access to x and y values */
    dy = pointXY.y;

    tmpdx = (dx < 0)? -dx: dx;
    tmpdy = (dy < 0)? -dy: dy;

    if (dx>0 && dy==0)
        *t = 3*M_PI_2;
    else if (dx>0 && dy>0)
        *t = 3*M_PI_2 + atan((double)tmpdy/(double)tmpdx);
    else if (dx==0 && dy>0)
        *t = 0;
    else if (dx<0 && dy>0)
        *t = M_PI_2 - atan((double)tmpdy/(double)tmpdx);
    else if (dx<0 && dy==0)
        *t = M_PI_2;
    else if (dx<0 && dy<0)
        *t = M_PI_2 + atan((double)tmpdy/(double)tmpdx);
    else if (dx==0 && dy<0)
        *t = M_PI;
    else /* (dx>0 && dy<0) */
        *t = 3*M_PI_2 - atan((double)tmpdy/(double)tmpdx);
    

    pivXYT->theta = *t;
}


/*
 * Returns true if a fibre obscures a fiducial.  Based on the AAO's FPIL
 * library equivalent routine.
 */
INLINE int fpcol___FiducialFibre(
    const ButXYT * const butEnd,
    const ButXYT * const fibEnd,
    const ButXYT * const mark)
{
/*
 * Assume the radius of the fiducial is its actual radius plus
 * the required clearance.
 */
  double radius = fpcolINST_FIBRE_CLEAR +  fpcolINST_FID_RAD;
  double radSqrd = SQRD(radius);
/*
 *  a is the length of the fibre
 *  b is the distance from the fiducial to the pivot
 *  c is the distance from the fiduical to the fibre end.
 */

    double a = SQRD((double)(butEnd->x - fibEnd->x)) + 
               SQRD((double)(butEnd->y - fibEnd->y));

    double b = SQRD((double)(butEnd->x - mark->x)) + 
               SQRD((double)(butEnd->y - mark->y));

    double c = SQRD((double)(fibEnd->x - mark->x)) + 
               SQRD((double)(fibEnd->y - mark->y));
/*
 *  If the distance from the pivot to the fiducial is longer than the
 *  fibre length + fiducial radius, then there can be no collision.  
 *  Likewise for the fibre end to the fiducial.
 */
    if (((a+radSqrd) <= b) || ((a+radSqrd) <= c))
        return FALSE;
    
    else if (butEnd->x == fibEnd->x)
    {
/*
 *      Fibre is parallel to the  Y axis  (the next bit of code would get
 *      a divide by zero error) and the y of point is within the length
 *      of the fibre (ensured by the above check) 
 *
 *      Collision occurs if the distance to the line in x is
 *      less then the circle radius. 
 */
        double d = fabs(mark->x - fibEnd->x);
        return ((d <= radius)? TRUE: FALSE);        
    }
    else
    {        
/*
 *      We can now use code which determines the distance between a line
 *      and a point since
 *        1.  The first check ensures a perpendicular from line to the
 *            point is within the line-segement of concern
 *        2.  fibEnd->x - butEnd->x != 0
 *
 *     Given m is the gradient of the line between the fibre ends and
 *     b is the intercept, then, d the distance between the point x1,y1 and
 *     the line is given by
 *
 *      d = |  m.x1 - y + b    |   (Note, absolute value only)
 *          |------------------|
 *          | +/- sqrt(m^2 + 1)|
 *
 *
 *      Determine m and b.
 */
        double m = (fibEnd->y - butEnd->y)/(fibEnd->x - butEnd->x);
        double b = fibEnd->y - m*fibEnd->x;
        
/*
 *      Determine the denominator and then the numerator.  Note, sqrt()
 *      returns the positive square root only.
 */
        double dd = sqrt(m*m + 1);
        double dn = m * mark->x - mark->y + b;
        
        double d;
/*
 *      Determine the distance.
 */
        if (dn < 0)
            d = dn * -1 / dd;
        else
            d = dn / dd;         
        
        return ((d <= radius) ? TRUE: FALSE);
    }
    
}


/*
 * This function returns TRUE if there is a collision between the
 * specified fibre button and the specified fibre. 
 */
extern vltLOGICAL  fpcolButFib (
                                IN vltDOUBLE butR,
                                IN vltDOUBLE butTheta,
                                IN vltDOUBLE butAlpha,
                                IN vltDOUBLE fibR,
                                IN vltDOUBLE fibTheta,
                                IN vltDOUBLE pivR,
                                IN vltDOUBLE pivTheta,
                                IN vltINT32  tolerance)
{
    ButXYT button;
    ButXYT fibButEnd;
    ButXYT fibPivEnd;

    BoundingBox buttonBox;
    BoundingBox fibreBox;

    /*
     * Convert positions to x/y/t coorindate system.
     */
    fpcol___rta2xyt(butR, butTheta, butAlpha, &button);
    fpcol___rtp2xyt(fibR, fibTheta, pivR, pivTheta, 
                    &fibButEnd, &fibPivEnd);

    /*
     * Determine bounding boxes
     */
    fpcol___buttonBox(&button, &buttonBox);
    fpcol___fibreBox (&fibButEnd, &fibPivEnd, &fibreBox);

    /*
     *  Normally, only if these boxes collide do we bother trying the 
     *  (expensive) detailed check routine.
     */
    if (fpcol___boxCollision(&buttonBox, &fibreBox))
        /*
         * Call the underlying button-fibre collision detection routine
         */
        return fpcolBaseColButFib(0, 
                                  button.x, button.y, button.theta,
                                  fibButEnd.x, fibButEnd.y,
                                  fibPivEnd.x, fibPivEnd.y,
                                  tolerance,
                                  SQRD(tolerance));

    else
        {
        /*
         * If debuging, always all the underlying routine and output
         * a message if we have a conflict.
         */
#       ifdef FPCOL_DEBUG
            {
            if (fpcolBaseColButFib(0, 
                                  button.x, button.y, button.theta,
                                  fibButEnd.x, fibButEnd.y,
                                  fibPivEnd.x, fibPivEnd.y,
                                  tolerance,
                                  SQRD(tolerance)))
                {
                fprintf(stderr,
"FPCOL ColButFib error, bounding boxes don't collide but button/fibre does\n");
                return TRUE;
                }
            }
#       endif
        return FALSE;
        }                          
}

/*
 * This function returns TRUE if there is a collision between the two
 * specified fibre buttons.
 */
extern vltLOGICAL  fpcolButBut (
                                   IN vltDOUBLE butAR,
                                   IN vltDOUBLE butATheta,
                                   IN vltDOUBLE butAAlpha,
                                   IN vltDOUBLE butBR,
                                   IN vltDOUBLE butBTheta,
                                   IN vltDOUBLE butBAlpha,
                                   IN vltINT32  tolerance)
{
    ButXYT butA;
    ButXYT butB;
    BoundingBox butABox;
    BoundingBox butBBox;
    /*
     * Convert the positions to x/y/t.
     */
    fpcol___rta2xyt(butAR, butATheta, butAAlpha, &butA);
    fpcol___rta2xyt(butBR, butBTheta, butBAlpha, &butB);
    /*
     * Create bounding boxes
     */
    fpcol___buttonBox(&butA, &butABox);
    fpcol___buttonBox(&butB, &butBBox);


        /*
     *  Normally, only if these boxes collide do we bother trying the 
     *  (expensive) detailed check routine.
     */
    if (fpcol___boxCollision(&butABox, &butBBox))
        {
        /*
         * Call the underlying button-fibre collision detection routine
         */
        return fpcolBaseColButBut(0,
                                  butA.x, butA.y, butA.theta,
                                  butB.x, butB.y, butB.theta,
                                  tolerance,
                                  SQRD(tolerance));
        }
    else
        {
        /*
         * If debuging, always all the underlying routine and output
         * a message if we have a conflict.
         */
#       ifdef FPCOL_DEBUG
            {
            if (fpcolBaseColButBut(0,
                              butA.x, butA.y, butA.theta,
                              butB.x, butB.y, butB.theta,
                              tolerance,
                              SQRD(tolerance)))
                {
                fprintf(stderr,
"FPCOL ColButBut error, bounding boxes don't collide but buttons do\n");
                return TRUE;
                }
            }
#       endif
        return FALSE;
        }
}

/*
 * This function returns true if there is a collision between the
 * specified button (including its fibre)  and the specified fiducial
 * mark.
 */
extern vltLOGICAL  fpcolFiducial (
                                   IN vltDOUBLE fibR,
                                   IN vltDOUBLE fibTheta,
                                   IN vltDOUBLE pivR,
                                   IN vltDOUBLE pivTheta,
                                   IN vltDOUBLE markR,
                                   IN vltDOUBLE markTheta,
                                   IN vltINT32  tolerance)
{
    ButXYT fibButEnd;
    ButXYT fibPivEnd;
    ButXYT mark;

    /*
     * Convert positions to X/Y/T coorindate system.  For the Mark
     * position, the assume alpha is zero since we don't care about that.
     */
    fpcol___rtp2xyt(fibR, fibTheta, pivR, pivTheta, 
                    &fibButEnd, &fibPivEnd);

    fpcol___rta2xyt(markR, markTheta, 0, &mark);
    /*
     * First check for a collision between the fibre and the mark
     */
    if (fpcol___FiducialFibre(&fibButEnd, &fibPivEnd, &mark))
        return TRUE;

    /* 
     * Now check for a collision between the fibre button and the
     * mark using the underlying routine.
     */ 
    return fpcolBaseColFidBut(0,
                              fibButEnd.x, fibButEnd.y, fibButEnd.theta, 
                              mark.x, mark.y,
                              tolerance, SQRD(tolerance));
    
}


/*
 *  Returns true if the specified fibre position on the specified plate
 *  is invalid.
 */
extern vltLOGICAL  fpcolInvPos (
                                   IN vltUINT32  plate ,
				   IN vltINT32   fibType,
                                   IN vltDOUBLE  fibR,
                                   IN vltDOUBLE  fibTheta,
                                   IN vltDOUBLE  fibAlpha,
                                   IN vltINT32   tolerance)
{
    ButXYT posXYT;
    /*
     *  Convert the position to X/Y/T system.
     */
    fpcol___rta2xyt(fibR, fibTheta, fibAlpha, &posXYT);
    /*
     * Call the underlying invalid position routine.
     */
    return (fpcolBaseColInvPos(0, plate, fibType,
                       posXYT.x, posXYT.y, posXYT.theta,
                       tolerance,
                       SQRD(tolerance)));
    
}        

/*
 * Returns true if the two specified fibres cross.  
 *
 * This routine is based on the AAO's FPIL library equivalent routine, which 
 * itself was based on the AAO's original 2dF collision detection library.
 */
extern vltLOGICAL  fpcol___FibFib (
                                const ButXYT * const fibA,
                                const ButXYT * const pivA,
                                const ButXYT * const fibB,
                                const ButXYT * const pivB,
                                IN vltDOUBLE  tolerance)
{
    double  dxa, dya;
    double  ssa, sa;
    double  sina, cosa;
    double  xtb, ytb;
    double  xpb, ypb;
    double  xBr, yBr;
    double  xBPr, yBPr;
    double m;
    double xCross;
    /*
     * We are trying to determine if there is an intercept between
     * two line segments. Our approach is to rotate the origin
     * such that fibre A lies on the X axies.  This makes the caculations
     * easier
     */
    

    /*
     * Detemine rotation required to put fibre A on X axis.
     */
    dxa  = pivA->x - fibA->x;
    dya  = pivA->y - fibA->y;
    ssa  = dxa*dxa + dya*dya;
    sa   = sqrt(ssa);
    cosa = dxa/sa;
    sina = dya/sa;

    /*
     * Determine coorindates of fibre B assuming fibre A starts at origin.
     */
    xtb = fibB->x - fibA->x;
    ytb = fibB->y - fibA->y;
    xpb = pivB->x - fibA->x;
    ypb = pivB->y - fibA->y;

    /*
     *  Rotate fibre B coordinates such fibre A can be assumed to line on 
     *  the X axis.
     */
    xBr  = xtb*cosa + ytb*sina;
    yBr  = ytb*cosa - xtb*sina;
    xBPr = xpb*cosa + ypb*sina;
    yBPr = ypb*cosa - xpb*sina;
    /*
     *  If fibre B is either fully above or fully below the X axis, there
     *  can be no collision.
     */
    if (yBr*yBPr > 0.0)                      
        return (FALSE);
    /*
     *  At this point, one end of fibre B is above and one below the X axis.
     *
     *  If both X components of fibre B are to the left of fibre A, then
     *  there can be no collision
     */ 
    else if (xBr > sa && xBPr > sa)             
        return (FALSE); 
    /*
     *  Now if both X components of fibre B are greter then zero, then 
     *  it must be the case the there is a collision
     */
    else if (xBr > 0.0 && xBPr > 0.0)             
        return (TRUE);
    /*
     *  Both X ends of fibre B are to the left of the Y axis.  No collision.
     */
    else if (xBr < 0.0 && xBPr < 0.0)             
        return (FALSE);
    /*
     *  At this point fibre B passes from quarter 1 to 3 or quarter 2 to 4 (or 
     *  vis-a-versa).
     *
     *  It intercepts with A only if the point at which is crosses the
     *  X axis (Y = 0) happens with the value of X positive and less then 
     *  the length of fibre A.   The point at which it crosses the X axis is 
     *  given by
     *
     *  X = x2 - y2 / m
     *
     *  This is positive if  x2 > y2/m, hence
     */
    m = (xBPr - xBr)/(yBPr - yBr);
    xCross = xBr - yBr*m;
    if ((xCross > 0)&&(xCross < sa))
        return (TRUE);
    /*
     *  No collision can occur
     */
    return(FALSE);
}

/*
 * Returns true if the two specified fibres cross.  
 *
 * This routine is based on the AAO's FPIL library equivalent routine.
 */
extern vltLOGICAL  fpcolFibFib (
                                IN vltDOUBLE  fibAR,
                                IN vltDOUBLE  fibATheta,
                                IN vltDOUBLE  pivAR,
                                IN vltDOUBLE  pivATheta,
                                IN vltDOUBLE  fibBR,
                                IN vltDOUBLE  fibBTheta,
                                IN vltDOUBLE  pivBR,
                                IN vltDOUBLE  pivBTheta,
                                IN vltDOUBLE  tolerance)
{
    ButXYT  fibA, pivA;      
    ButXYT  fibB, pivB;
    BoundingBox a;
    BoundingBox b;

    /*
     * Convert the fibre and pivot positions to the X/Y coorindate system.
     */
    fpcol___rtp2xyt(fibAR, fibATheta, pivAR, pivATheta, &fibA, &pivA);
    fpcol___rtp2xyt(fibBR, fibBTheta, pivBR, pivBTheta, &fibB, &pivB);

    /*
     *  Create boxes about the fibres.
     */
    fpcol___fibreBox(&fibA, &pivA,  &a);
    fpcol___fibreBox(&fibB, &pivB,  &b);
    /*
     *  Normally, only if these boxes collide do we bother trying the 
     *  (expensive) detailed check routine.
     */
    if (fpcol___boxCollision(&a, &b))
        return fpcol___FibFib(&fibA, &pivA, &fibB, &pivB, tolerance);

    else
        {
        /*
         * If debuging, always all the underlying routine and output
         * a message if we have a conflict.
         */
#       ifdef FPIL_DEBUG
            {
            if (fpcol___FibFib(&fibA, &pivA, &fibB, &pivB, tolerance))
                {
                fprintf(stderr,
      "FPCOL ColFibFib error, bounding boxes don't collide but fibres do\n");
                return TRUE;
                }
            }
#       endif
        return FALSE;
        
    }
}



/*___oOo___*/
