/*******************************************************************************
* E.S.O. - VLT project
*
* "@(#) $Id: ACMM:ozpozFpcol/ws/src/fpcolMdlCal.c,v 1.63 07-Oct-2008 11:10:27+10 tjf $"
*
*
* who       when      what
* --------  --------  ----------------------------------------------
* tfarrell  19/10/02  Improve output from show residuals command.
* tfarrell  18/10/02  Add "set fitfun powell2" command and set the powell
*                     mode appropiatly.  Note, - this is an undocumented
*                     command intended for debugging.
*                     Fix a bug in "clear resabove"
* tfarrell  16/10/02  Add remmeanoff command.
* tfarrell  10/10/02  When normalizing model - ensure we check for
*                     bad status correctly.
* tfarrell  28/08/02  Add "mdlpars normalize" command.
* tfarrell  26/08/02  Add "plot" command.
* tfarrell  22/08/02  "show residuals" should show absolute residual.
*                     "clear resabove" should clear on absolute residual.
* tfarrell  22/08/02  Add "clear" command and support it in "show offsets".
* tfarrell  21/08/02  Add "show residuals" command.
* tfarrell  27/06/02  Add fpcolMdlCalCheckFitType() which warns about
*                     fitting parameter which don't make sense (mainly the
*                     indivual linear model parameters).  Invoke this
*                     from "set fit" and "show fit".  It is also invoked
*                     from "fit".
*                     Drop "create" command, functionality now implemented
*                     in "fit" command.
*                     Add the "extract" command.
# tfarrell  23/06/02  Command to set offsets was incorrectly displaying
#                     them always in radians, instead of arc-seconds.
* tfarrell  20/06/02  Dropping the last distortion parameter due to new
*                     model. Changing units of the new last one.
* tfarrell  19/06/02  Add ModelParUnits array and use it, expand usage
*                     of ModelParNames.  Add "set fit n=[no]fit" command.
*                     Add "show fit" and "show pars" commands.  Add
*                     some prompting about next command.
* tfarrell  18/06/02  Add ModelParNames array and use some of it 
*                     for the "mdlpars show" command and the show command.
* tfarrell  09/05/02  Support wavelength specific offsets and a current
*                     wavelength for each fibre type.  Add "set fit a=b"
*                     sub-command.  Offsets are now in arc-seconds.  Rename
*                     buttonOffsets to offsets.
* tfarrell  15/04/02  Fit command now supports specified sub-sets of
*                     things to set.
* tfarrell  12/04/02  Added saveHeadFn and loadHeadFn commands.  Improve
*                     output of information in the show function.
* tfarrell  30/03/02  Add parsFn() - mdlpars command.
* ssmedley  07/09/01  created
*
*/

/************************************************************************
*   NAME
*	fpcal - OzPoz calibration software used to create an Astrometric
*	Model.
*
*   SYNOPSIS
*       #include "fpcolMdlCal.h"
*
*	int offsetsFn
*		(int			nArgs,
*		 char			*args[],
*		 FpcalData		*pFpcalData)
*	int showFn
*		(int			nArgs,
*		 char			*args[],
*		 FpcalData		*pFpcalData)
*	int quitFn
*		(int			nArgs,
*		 char			*args[],
*		 FpcalData		*pFpcalData)
*	int loadFn
*		(int			nArgs,
*		 char			*args[],
*		 FpcalData		*pFpcalData)
*	int setFn
*		(int			nArgs,
*		 char			*args[],
*		 FpcalData		*pFpcalData)
*	int helpFn
*		(int			nArgs,
*		 char			*args[],
*		 FpcalData		*pFpcalData);
*
*       int parsFn
*		(int			nArgs,
*		 char			*args[],
*		 FpcalData		*pFpcalData);
*
*       int saveHeadFn
*		(int			nArgs,
*		 char			*args[],
*		 FpcalData		*pFpcalData);
*
*       int loadHeadFn
*		(int			nArgs,
*		 char			*args[],
*		 FpcalData		*pFpcalData);
*       int extractFn
*		(int			nArgs,
*		 char			*args[],
*		 FpcalData		*pFpcalData);
*       int clearFn
*		(int			nArgs,
*		 char			*args[],
*		 FpcalData		*pFpcalData);
**
**
*   DESCRIPTION
*
*	The source code in this file contains the main() routine
*	of the <fpcal> binary. The main() routine itself is just
*	a driver - it accepts input strings from stdin & looks up
*	a command-table to find an appropriate action (a la
*	callback functions). If an action exists, the callback is
*	invoked with any additional arguments specified on stdin.
*	Additional commands can easily be added
*	by adding a callback to this command-table. (Look for
*	the CommandAssoc array <assoc>.)
*
*	The intelligence in this s/w lies in the callbacks. Currently,
*	they are: offsetsFn() showFn() quitFn() loadFn()
*	setFn() helpFn() parsFn() & elsewhere fitFn().
*
*	Each of these callbacks interprets & validates the arguments
*	it receives to determine an appropriate action. Most of them
*	simply validate the arguments & (assuming no errors) call
*	an external function to do the guts of the work for them.
*	They typically alter the FpcalData structure which is also
*	passed as an argument to each callback which facilitates a
*	communication mechanism & allows each callback to determine
*	the state of the system.
*
*	offsetsFn() loads button offsets from a specified file & stores
*	them in the FpcalData structure (passed as an argument).
*
*	showFn() displays the value of all user-settable values in
*	the FpcalData structure.
*
*	quitFn() terminates the program.
*
*	loadFn() will read in all the relevant data contained in the
*	specified FITS file & store it in the FpcalData structure.
*
*	setFn() allows the user to override the 'fit' & 'parameters'
*	values stored in the FpcalData structure.
*
*       parsFn() allows for low level changes and display of the
*       telescope model parameters.
*
*
*       saveHeadFn() will save the data loaded by loadFn() or
*       loadHeadFn() in a form which can be edited and then
*       loaded again with loadHeadFn()
*
*       loadHeadFn() loads a file in the format saved by
*       saveHeadFn().
*
*       extractFn() writes the button no, wavelength, tanget
*       plane coorindates and offsets to a file.  This can be
*       used to run the data through an external fitting program.
*
*       clearFn() is used to clear offsets to they are not used 
*       in a fit.
*
*	helpFn() shows a succinct list of available commands.
*
*   ENVIRONMENT
*
*   CAUTIONS
*
*   EXAMPLES
*
*   SEE ALSO
*
*	For a complete description of how to use this s/w, refer to
*	the OzPos Calibration Software Manual. (INS-TRE-AUS-13721-0075)
*
*   BUGS
*
*------------------------------------------------------------------------
*/


static const char *rcsId="@(#) $Id: ACMM:ozpozFpcol/ws/src/fpcolMdlCal.c,v 1.63 07-Oct-2008 11:10:27+10 tjf $";

static void *use_rcsId = (0 ? (void *)(&use_rcsId) : (void *) &rcsId);

#define __EXTENSIONS__    /* Pick up extra include file stuff on solaris */
#define _HPUX_SOURCE      /* Pick up extra include file stuff on HP UX */


#include <stdio.h>
#include <errno.h>		/* errno */
#include <stdlib.h>		/* atof */
#include <assert.h>
#include <strings.h>	        /* strtok */
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>

#include "slalib.h"

#include "fpcolMdlCalLibs.h"
#include "fpcolMdl.h"
#include "fpcolMdlCalLoadOffsets.h"
#include "fpcolMdlCalModel.h"	/* createModel */
#include "fpcolMdlCalFit.h"			/* fitFn */
#include "fpcolMdlCalGetline.h"		/* getline */
#include "fpcolMdlCalConvert.h"         /* ToString functions */

static const char *argv0 = 0;    /* Copy of argv[0] to main */

/*
 *  Names for each of the parameters.  Note the marco's defined within which are
 * checked later against macros from fpcolMdl.h.
 */
char *ModelParNames[] = {
  "X offset      ",   /* 0 */
  "Lin Mdl Par b ",   /* 1 */
  "Lin Mdl Par c ",   /* 2 */
  "Y offset      ",   /* 3 */
  "Lin Mdl Par e ",   /* 4 */
  "Lin Mdl Par f ",   /* 5 */
#define LASTLIN 5
  "Extra Rotation",  /* 6 */
#define ROTPAR  6
  "Extra Scale   ",  /* 7 */
#define SCALEPAR 7
  "NW Dist P1    ",  /* 8 */
#define NW_DIST_FIRST 8
  "NW Dist P2    ",  /* 9 */
  "NW Dist P3    ",  /* 10 */
#define NW_DIST_LAST  10
  "WL Dist P1    ",  /* 11 */
#define W_DIST_FIRST 11
  "WL Dist P2    ",  /* 12 */
#define W_DIST_LAST  12
  "undefined in13",  /* 13 */
  "undefined in14",  /* 14 */
  "undefined in15",  /* 15 */
  "undefined in16",  /* 16 */
  "undefined in17",  /* 17 */
#define NONPERPPAR 18
  "Non-perpendic ",  /* 18 */
  "undefined in19"   /* 19 */
};

char *ModelParUnits[] = {
  "Microns       ",   /* 0 */
  "              ",   /* 1 */
  "              ",   /* 2 */
  "Microns       ",   /* 3 */
  "              ",   /* 4 */
  "              ",   /* 5 */
  "Degrees       ",  /* 6 */
  "              ",  /* 7 */
  "mm/degree     ",  /* 8 */
  "mm/degree^3   ",  /* 9 */
  "mm/degree^5   ",  /* 10 */
  "mm.nm/degree  ",  /* 11 */
  "mm.(nm/deg)^2 ",  /* 12 */
  "              ",  /* 13 */
  "              ",  /* 14 */
  "              ",  /* 15 */
  "              ",  /* 16 */
  "              ",  /* 17 */
  "Degrees       ",  /* 18 */
  "              "   /* 19 */
};

/*
 *  Check against fpcolMdl.h
 */
#if LASTLIN !=  fpcolMDL_PAR_LASTLIN
#error "Parameters definition changed"
#endif
#if ROTPAR  !=  fpcolMDL_PAR_ROT
#error "Parameters definition changed"
#endif
#if SCALEPAR != fpcolMDL_PAR_SCALE  
#error "Parameters definition changed"
#endif
#if NW_DIST_FIRST != fpcolMDL_PAR_NW_DIST_FIRST 
#error "Parameters definition changed"
#endif
#if NW_DIST_LAST != fpcolMDL_PAR_NW_DIST_LAST
#error "Parameters definition changed"
#endif
#if W_DIST_FIRST != fpcolMDL_PAR_W_DIST_FIRST 
#error "Parameters definition changed"
#endif
#if W_DIST_LAST != fpcolMDL_PAR_W_DIST_LAST 
#error "Parameters definition changed"
#endif
#if NONPERPPAR != fpcolMDL_PAR_NONPERP
#error "Parameters definition changed"
#endif


#define MAX_ARGS	(NUM_PARAMETERS+10)

#define NUM_ELEMENTS(array) (sizeof(array)/sizeof(array[0]))

extern void fpcolMdlCalCheckFitType(int *aFit)
{
    if ((!aFit[1])||(!aFit[2])||(!aFit[4])||(!aFit[5]))
    {
/*
        fprintf(stderr,"Warning:You should not fit the basic linear model parameters\n");
        fprintf(stderr,"  The results are unreliable\n");
        fprintf(stderr,"  Instead fit the scale, rotation and distortion parameters\n");
*/
    }
}

/*
 * load button offsets from specified file, then store the
 * appropriate offset in the FpcalData structure for each button.
 */
static int offsetsFn
	(int			nArgs,
	 char			*args[],
	 FpcalData		*pFpcalData)
{
	StatusType	status = STATUS__OK;

	if (nArgs < 2)
	{
		fprintf(stderr, "%s: invalid number of arguments\n", args[0]);
		return -1;
	}
	if (strcasecmp("help", args[1]) == 0)
	{
		printf("%s: usage: %s <file>\n", args[0], args[0]);
		return 0;
	}

        if (pFpcalData->bInitialised)
          {
            loadButtonOffsets(args[1], 
                              pFpcalData,
                              &pFpcalData->curWaveIndex, 
                              pFpcalData->nFibres_a, 
                              pFpcalData->buttons, 
                              &status);

            if (status == STATUS__OK)
                printf("Next command in normal sequence - \"fit\" or \"show\" or \"set\"\n");


            return (status == STATUS__OK ? 0 : -1);

          } 
        else 
          {
            fprintf(stderr, "%s: must load/loadhd fpcaldata first!\n", args[0]);
            return -1;
          }
}
/*
 * Clear command.
 */
static int clearFn
	(int			nArgs,
	 char			*args[],
	 FpcalData		*pFpcalData)
{
    int buttonNo;
    int residualCut;
    if ((nArgs < 2)||
        (nArgs >= 2 && strcasecmp("help", args[1]) == 0))
	{
        printf("%s: usage: %s <buttonNo(>0)> [<index>]\n", args[0], args[0]);
        printf("%s: usage: %s resabove <microns(>0)>\n", args[0], args[0]);
        return 0;
	}
    if (buttonOffsetsStatus() != BO_LOADED)
        {
        printf("\tNo offsets loaded as yet\n");
        return -1;
        }
    else if ((strcasecmp("resabove", args[1])==0)&&
             (nArgs >= 3)&&((residualCut = atol(args[2])) > 0))
        {
        int i;
        if (pFpcalData->fitted == 0)
            {
            fprintf(stderr, "No fit done as yet - hence no residuals\n");
            return -1;
            };
        for (i = 0; i < pFpcalData->nOffsets; i++)
            {
            double residualX = fabs(pFpcalData->residualsInfo[i].x - 
                             pFpcalData->residualsInfo[i].initialX);
            double residualY = fabs(pFpcalData->residualsInfo[i].y - 
                             pFpcalData->residualsInfo[i].initialY);
            double residual = sqrt(residualX * residualX + 
                                   residualY * residualY);

            if ((residual > residualCut))
                {
                int waveIndex = pFpcalData->offsetWaveIndex[i];
                buttonNo  = pFpcalData->offsetButtonNo[i];
                if (clearButtonOffset(buttonNo, waveIndex) != 0)
                    {
                    fprintf(stderr, "Unable to clear offset for button %d, wave index %d\n", 
                            buttonNo, waveIndex);
                    return -1;
                    }
                else
                    {
                    printf("Cleared offset for button %d, wavelength index %d\n", 
                           buttonNo, waveIndex);
                    }               
                }
            }
        
        }
    else if ((buttonNo = atol(args[1])) > 0)
        {
        int waveIndex = fpcalWaveIdxForFibre(&pFpcalData->curWaveIndex, buttonNo);
        if (nArgs >= 3)
            waveIndex = atol(args[2]);

        if (clearButtonOffset(buttonNo, waveIndex) != 0)
            {
            fprintf(stderr, "Unable to clear offset for button %d, wave index %d\n", 
                    buttonNo, waveIndex);
            return -1;
            }
        else
            {
            printf("Cleared offset for button %d, wavelength index %d\n", buttonNo, waveIndex);
            }
        }
    else
        {
        char *helpCmd[2] = {"clear", "help"};
        fprintf(stderr, "%s: invalid number/format of arguments\n", args[0]);
        clearFn(2,helpCmd, pFpcalData);
        return -1;
        
        }
        
    return 0;
}
     

/*
 * Remove mean offsets function - remmeanoff command.
 */
static int remMeanFn
	(int			nArgs,
	 char			*args[],
	 FpcalData		*pFpcalData)
{

    if (nArgs > 1) 
        {
        fprintf(stderr, "%s: usage: %s\n", args[0], args[0]);
        return 1;
        }

    if (buttonOffsetsStatus() != BO_LOADED)
      {
      fprintf(stderr, "No offsets loaded as yet\n");
      return -1;
      }
    if (removeMeanOffsets() != 0)
      {
        fprintf(stderr, "Error removing mean offsets\n");
        return -1;
      }
    return 0;
      

}
/*
 * Plot command - run residuals command to plot the residuals..
 */
static int plotFn
	(int			nArgs,
	 char			*args[],
	 FpcalData		*pFpcalData)
{
    char directoryName[1000];
    int  dirValid = 0;
    if (nArgs > 1) 
        {
        fprintf(stderr, "%s: usage: %s\n", args[0], args[0]);
        return 1;
        }
    if (pFpcalData->fitted == 0)
        {
        fprintf(stderr, "No fit done as yet - hence no plot is sensible\n");
        return -1;
        };


    fpcolMdlCalFindExe(argv0, directoryName, sizeof(directoryName), &dirValid);


    if (dirValid)
        {
#       define PLOTTER "residuals"
#       define FIT_OUT "residuals.out"
        char *exeFile;
        char *command;
        int result;
        /*
         * Get the name of the executable for the residuals command.
         */
        exeFile = (char *)malloc (strlen(directoryName)+strlen(PLOTTER)+1);
        strcpy(exeFile,directoryName);
        strcat(exeFile,PLOTTER);

        if (access(exeFile, X_OK)) 
            {
            if (access(exeFile, F_OK)) 
                fprintf(stderr, "Can't find file \"%s\" so can't plot\n", exeFile);
            else
                fprintf(stderr, "Can't execute file \"%s\" so can't plot\n", exeFile);
            free(exeFile);
            return 1;
            }
        /*
         * Create a command for "system()".
         *  Space is required for the following elements:
         *      command space fit_out space & null
         */
        command = malloc(strlen(exeFile)+strlen(FIT_OUT)+5);
        strcpy(command, exeFile);
        strcat(command, " ");
        strcat(command, FIT_OUT);
        strcat(command, " &");
        free(exeFile);  /* No longer needed */
        
        result = system(command);
        if (result != 0)
            {
            fprintf(stderr, "Failure executing command \"%s\"\n", command);
            fprintf(stderr, " Termination status is %d, errno = %d\n", result, errno);
            return 1;
            }

        }
    
    
    else
        {
        fprintf(stderr, "%s: Unable to find executable directory\n", args[0]);
        return 1;
        }

    return 0;
}

/*
 * show values of specified variables.
 */
static int showFn
	(int			nArgs,
	 char			*args[],
	 FpcalData		*pFpcalData)
{
	int bAll, nMatches = 0;

	if (nArgs >= 2 && strcasecmp("help", args[1]) == 0)
	{
		printf("%s: usage: %s <var>\n", args[0], args[0]);
		printf("%s: where <var> is one of:\n", args[0]);
		printf("%s:\tfpcalData\n", args[0]);
		printf("%s:\tmodel\n", args[0]);
		printf("%s:\tpars\n", args[0]);
		printf("%s:\tfit\n", args[0]);
		printf("%s:\toffsets\n", args[0]);
		printf("%s:\tresiduals\n", args[0]);
		printf("%s:\tall\n", args[0]);
		return 0;
	}

	if (nArgs != 2)
	{
            char *helpCmd[2] = {"show", "help"};
            fprintf(stderr, "%s: invalid number of arguments\n", args[0]);
            showFn(2,helpCmd, pFpcalData);
            return -1;
	}
        /*
         * First work out if we are showing everything.
         */
	bAll = (strcasecmp("all", args[1]) == 0);
        /*
         * Show the fpcalData.  (Also show pars and fit below when command is fpcalData).
         */
	if (bAll || (strcasecmp("fpcalData", args[1]) == 0))
	{
		printf("fpcalData:\n");
		if (!pFpcalData->bInitialised)
			printf("\tNOT INITIALISED\n");
		else
		{
			int i;
                        printf("\tSource File Name\t%s\n",pFpcalData->fileName);
			printf("\tatmTemp\t\t%10.4f Degrees Celsius\n", pFpcalData->atmTemp);
			printf("\tatmPressure\t%10.4f Millibars\n", pFpcalData->atmPressure);
			printf("\tatmHumidity\t%10.4f Percent\n", pFpcalData->atmHumidity);
			printf("\tposAngle\t%10.4f Degrees\n", pFpcalData->posAngle);
			printf("\tfieldPlate\t%5d\n", pFpcalData->fieldPlate);
			printf("\tobservationDate\t%10.4f (MJD) %s\n", 
                               pFpcalData->observationDate,
                               mjdToString(pFpcalData->observationDate));
			printf("\tequinox\t\t%10.4f Years AD\n", pFpcalData->equinox);
			printf("\tnFibres\t\t%5d\n", pFpcalData->nFibres_a);
			printf("\tfieldCentre\tra=%13.10f dec=%13.10f (Mean-Radians)\n",
                               pFpcalData->fieldCentreMeanRA, 
                               pFpcalData->fieldCentreMeanDEC);

                        printf("\t           \tra=%s  dec=%s\n",
                               raToString(pFpcalData->fieldCentreMeanRA),
                               decToString(pFpcalData->fieldCentreMeanDEC));

			printf("\tfieldCentre\tra=%13.10f dec=%13.10f (Apparent-Radians)\n",
                               pFpcalData->fieldCentreAppRA, 
                               pFpcalData->fieldCentreAppDEC);

                        printf("\t           \tra=%s  dec=%s\n",
                               raToString(pFpcalData->fieldCentreAppRA),
                               decToString(pFpcalData->fieldCentreAppDEC));


			printf("\tObjects positions (apparent)");
			if (pFpcalData->offsetRA != NULL)
				printf("/offset");
			printf(":\n");
 


			for (i = 0; i < pFpcalData->nFibres_a; i++)
			{
				printf("\t\t[%3d]  ra=%13.10f  dec=%13.10f\n",
                                       pFpcalData->buttons[i],
                                       pFpcalData->apparentRA[i],
                                       pFpcalData->apparentDEC[i]);
                                printf("\t\t          %s  %s\n",
                                       raToString( pFpcalData->apparentRA[i]),
                                       decToString(pFpcalData->apparentDEC[i]));

			}
                }
                ++nMatches;
        }

        /*
         * Show model parameters - shown for all command, fpcalData and par commands.
         */
	if (bAll || 
            (strcasecmp("fpcalData", args[1]) == 0)|| 
            (strcasecmp("pars", args[1]) == 0))
	{
		if (!pFpcalData->bInitialised)
			printf("\tNOT INITIALISED\n");
		else
		{
			int i;
                        const FpcalPars *pars = fpcolMdlCalGetPars(&pFpcalData->model);

			printf("\tTel Model parameters:\n");
			for (i = 0; i < NUM_PARAMETERS; i++)
			{
				printf("\t\t%2d (%s) %14.6f %s\n",
                                       i, ModelParNames[i],
                                       pars->p[i],
                                       ModelParUnits[i]);
			}
                }
                ++nMatches;
        }

        /*
         * Show fit settings for each parameters - if we are to fit that parameter or not.
         *  Shown for all command, fpcalData and fit commands.
         */
        if (bAll || 
            (strcasecmp("fpcalData", args[1]) == 0)|| 
            (strcasecmp("fit", args[1]) == 0))
	{
		if (!pFpcalData->bInitialised)
			printf("\tNOT INITIALISED\n");
		else
		{
			int i;
                        printf("\tParameter Fit Status: \n");
			for (i = 0; i < NUM_PARAMETERS; i++)
			{
				printf("\t\t%2d (%s) - %s\n",
                                       i, ModelParNames[i],
                                       pFpcalData->aFit[i] ? "don't fit" : "fit");
			}
                        fpcolMdlCalCheckFitType(pFpcalData->aFit);
		}
		nMatches++;
	}
        /*
         * Show details of the model - really just wavelengths for each spectrograph
         */
	if (bAll || (strcasecmp("model", args[1]) == 0))
	{
          fpcolMdlCalShowModels(&pFpcalData->model);
          fprintf(stderr, "Current Giraffe wavelength is index %d, wavelength %g\n",
                  pFpcalData->curWaveIndex.Giraffe,
                  fpcolMdlCalWavelength(&pFpcalData->model, pFpcalData->curWaveIndex.Giraffe));
          fprintf(stderr, "Current UVES    wavelength is index %d, wavelength %g\n",
                  pFpcalData->curWaveIndex.UVES,
                  fpcolMdlCalWavelength(&pFpcalData->model, pFpcalData->curWaveIndex.UVES));
          fprintf(stderr, "Current FACB    wavelength is index %d, wavelength %g\n",
                  pFpcalData->curWaveIndex.FACB,
                  fpcolMdlCalWavelength(&pFpcalData->model, pFpcalData->curWaveIndex.FACB));

          nMatches++;
	}

        /*
         * Show the offsets loaded if any.
         */
	if (bAll ||(strcasecmp("offsets", args[1]) == 0))
	{
		printf("Offsets:\n");
		if (buttonOffsetsStatus() != BO_LOADED)
			printf("\tNo offsets loaded as yet\n");
		else
		{
                  int numOffsets = getNumOffsets();
                  int offsetIndex;
                  for (offsetIndex = 0; offsetIndex < numOffsets ; ++offsetIndex)
                    {
                      const RaDec	*pOffset = NULL;
                      pOffset = getOffset(offsetIndex);
                      if (pOffset != NULL)
                          {
                          printf("\t[%3d][wi% 1d] ra=%4.1f, dec=%4.1f (%s)\n", 
                                 pOffset->button, pOffset->wIndex, 
                                 RAD2SEC(pOffset->ra), RAD2SEC(pOffset->dec), "arc-seconds");
                          }
                     }
                }
                ++nMatches;
        }
        /*
         * Show residuals - not output for "show all"
         */
	if (strcasecmp("residuals", args[1]) == 0)
	{
            printf("Residuals:\n");
            if (buttonOffsetsStatus() != BO_LOADED)
                printf("\tNo offsets loaded as yet - hence no residuals\n");
            else if (pFpcalData->fitted == 0)
                printf("\tNo fit done as yet - hence no residuals\n");
            else
                {
                int i;
                double rms_microns;
                double rms_arcsec;
                double sumsq = 0;
                StatusType status = STATUS__OK;
                FpilModelType *model = fpcolMdlCalWaveModel(&pFpcalData->model, 
                                                            pFpcalData->offsetWaveIndex[0]);
                printf("residuals (microns):\n             initial   fitted deviation  abs resid\n");
                for (i = 0; i < pFpcalData->nOffsets; i++)
                    {
                    double arcsec;
                    long residX = pFpcalData->residualsInfo[i].x - 
                        pFpcalData->residualsInfo[i].initialX;
                    long residY = pFpcalData->residualsInfo[i].y - 
                        pFpcalData->residualsInfo[i].initialY;
                

                    double residsq = (double)residX * (double)residX + 
                                     (double)residY * (double)residY;
                  
                    double resid = sqrt(residsq);

                    sumsq += residsq;

                    fpcolMdlMicrons2Arcsec(model, resid, &arcsec, &status);
 


                    printf("[%3d]: x: %9ld %9ld (%5ld)  %5ld  microns (%3.2f arcsec)\n",
                           pFpcalData->offsetButtonNo[i],
                           pFpcalData->residualsInfo[i].initialX,
                           pFpcalData->residualsInfo[i].fittedX,
                           residX, (long)rint(resid+0.5), arcsec);
                    printf("       y: %9ld %9ld (%5ld)\n",
                           pFpcalData->residualsInfo[i].initialY,
                           pFpcalData->residualsInfo[i].fittedY,
                           residY);                    
                    } /* for */
                /*
                 * Determine the RMS in microns and convert it to arc-seconds.
                 */
                rms_microns =  sqrt(sumsq/pFpcalData->nOffsets);
                fpcolMdlMicrons2Arcsec(fpcolMdlCalWaveModel(&pFpcalData->model, 
                                                            pFpcalData->offsetWaveIndex[0]),
                                       rms_microns, &rms_arcsec, &status);
 
                printf("\n Residual RMS=%8.4g microns, %5.2g arcseconds\n",
                      rms_microns, rms_arcsec);
                }
             ++nMatches;
	
        }

	if (nMatches == 0)
	{
            char *helpCmd[2] = {"show", "help"};
	    fprintf(stderr, "%s: unknown sub-command <%s>\n", args[0], args[1]);
            showFn(2,helpCmd, pFpcalData);
            return -1;
	}
	return 0;
}


/*
 * terminate program.
 */
static int quitFn
	(int			nArgs,
	 char			*args[],
	 FpcalData		*pFpcalData)
{
	if (nArgs >= 2 && strcasecmp("help", args[1]) == 0)
	{
		printf("%s: usage: %s\n", args[0], args[0]);
		return 0;
	}
	if (nArgs != 1)
	{
		fprintf(stderr, "%s: no arguments required\n", args[0]);
		return -1;
	}
	(void) pFpcalData;
	exit(0);
}


/*
 * load all the relevant fibre data for the specified FITS file.
 */
static int loadFn
	(int			nArgs,
	 char			*args[],
	 FpcalData		*pFpcalData)
{
	StatusType	status = STATUS__OK;

	if (nArgs >= 2 && strcasecmp("help", args[1]) == 0)
	{
		printf("%s: usage: %s <FITSfile>\n", args[0], args[0]);
		return 0;
	}
	if (nArgs != 2)
	{
		fprintf(stderr, "%s: invalid number of arguments\n", args[0]);
		return -1;
	}

	fpcalDestructor(pFpcalData);
	fpcalConstructor(pFpcalData);

	getFpcalData(args[1], pFpcalData, &status);

        if (status == STATUS__OK)
            printf("Next command in normal sequence - \"offsets\" or \"show\" or \"mdlpars show\"\n");

	return (status == STATUS__OK ? 0 : -1);
}

/*
 * this function allows the user to override some of the
 * fpcalData & offset variables.
 */
static int setFn
	(int			nArgs,
	 char			*args[],
	 FpcalData		*pFpcalData)
{
	if (nArgs >= 2 && strcasecmp("help", args[1]) == 0)
	{
		fprintf(stderr, "%s: usage: %s offset <button#> <RA> <DEC>\n", 
                        args[0], args[0]);
                fprintf(stderr, "%s: usage: %s wavelen giraffe|uves|facb <wavelength>\n",
                        args[0], args[0]);
                fprintf(stderr, "%s: usage: %s fitfun svd|powell\n", args[0], args[0]);
                fprintf(stderr, "%s: usage: %s powell tolerance|maxits val\n", args[0], args[0]);
		fprintf(stderr, "%s: usage: %s fit <0> <2> <3> <4> <5> <6>...<%d>\n", 
                       args[0], args[0], NUM_PARAMETERS-1);
                fprintf(stderr, "%s: usage: %s fit noall\n", args[0], args[0]);
                fprintf(stderr, "%s: usage: %s fit nolinear\n", args[0], args[0]);
                fprintf(stderr, "%s: usage: %s fit [no]origin\n", args[0], args[0]);
                fprintf(stderr, "%s: usage: %s fit [no]scale\n", args[0], args[0]);
                fprintf(stderr, "%s: usage: %s fit [no]nonperp\n", args[0], args[0]);
                fprintf(stderr, "%s: usage: %s fit [no]rotation\n", args[0], args[0]);
                fprintf(stderr, "%s: usage: %s fit [no]dist\n", args[0], args[0]);
                fprintf(stderr, "%s: usage: %s fit [no]wdist\n", args[0], args[0]);
                fprintf(stderr, "%s: usage: %s fit rod   (rotation, origin, distortion)\n",
                        args[0], args[0]);
                fprintf(stderr, "%s: usage: %s fit n=0|1|[no]fit\n", args[0], args[0]);
		fprintf(stderr, "%s: usage: %s parameters <0> <2> <3> <4> <5> <6>...<%d>\n", 
                       args[0], args[0], NUM_PARAMETERS-1);
		return 0;
	}
	if (nArgs < 3)
	{
          char *helpCmd[2] = {"set", "help"};
          fprintf(stderr, "%s: invalid number of arguments (%d)\n", args[0], nArgs);
          setFn(2,helpCmd, pFpcalData);
          return -1;
	}
	if (strcasecmp("offset", args[1]) == 0)
	{
		double	ra, dec;
		int		buttonNo;
                int waveIndex;

                if (!pFpcalData->bInitialised)
                  {
                    fprintf(stderr, "%s: must load/loadhd fpcaldata first!\n", 
                            args[0]);
                    return -1;
                  }

		if (nArgs != 5)
                  {
                    fprintf(stderr, "%s: set offset <button#> <RA> <DEC>\n", args[0]);
                    return -1;
                  }
		errno = 0;
		buttonNo = atoi(args[2]);

                if (!buttonFound(pFpcalData->nFibres_a, pFpcalData->buttons, buttonNo))
                  {
                    fprintf(stderr, "Button number %d is not one found in the header\n", buttonNo);
                    return -1;
                  }

		ra = atof(args[3]);
		dec = atof(args[4]);
		/* atof & strtod do not appear to be detecting errors! */
		if (errno != 0)
			fprintf(stderr, "%s: bugger\n", args[0]);

                waveIndex = fpcalWaveIdxForFibre(&pFpcalData->curWaveIndex, buttonNo);

		if (setButtonOffset(buttonNo, ra, dec, waveIndex) != 0)
			return -1;
		else
		{
			StatusType	status = STATUS__OK;
			const RaDec	*pOffset = getButtonOffset(buttonNo, waveIndex, &status);

			if (pOffset == NULL)
				return -1;
			printf("button %d, wavelength %g, offset set to ra=%8.4f dec=%8.4f %s\n",
                               buttonNo,
                               fpcolMdlCalWavelength(&pFpcalData->model, waveIndex),
#                              if OFFSET_IN_ARCSEC
                               RAD2SEC(pOffset->ra), RAD2SEC(pOffset->dec), "arc-seconds");
#                            else
                               pOffset->ra, pOffset->dec, "radians");
#                            endif

		}
	}
	else if (strcasecmp("wavelen", args[1]) == 0)
	{
          int wavelen = atol(args[3]);
          if ((nArgs == 4)&&(wavelen > 0)&&(strcasecmp("giraffe", args[2]) == 0))
            {
              int index = fpcolMdlCalNewWavelength(wavelen, pFpcalData);
              if (index < 0)
                {
                  fprintf(stderr, 
                          "Failure adding new GIRAFFE wavelength %d to internal tables\n", 
                          wavelen);
                  return -1;

                }
              pFpcalData->curWaveIndex.Giraffe = index;
            }
          else if ((nArgs == 4)&&(wavelen > 0)&&(strcasecmp("uves", args[2]) == 0))
            {
              int index = fpcolMdlCalNewWavelength(wavelen, pFpcalData);
              if (index < 0)
                {
                  fprintf(stderr, "Failure adding new UVES wavelength %d to internal tables\n", 
                          wavelen);
                  return -1;

                }
              pFpcalData->curWaveIndex.UVES = index;
             }
          else if ((nArgs == 4)&&(wavelen > 0)&&(strcasecmp("facb", args[2]) == 0))
            {
               int index = fpcolMdlCalNewWavelength(wavelen, pFpcalData);
              if (index < 0)
                {
                  fprintf(stderr, "Failure adding new FACB wavelength %d to internal tables\n", 
                          wavelen);
                  return -1;

                }
              pFpcalData->curWaveIndex.FACB = index;
            }
          else
            {
                fprintf(stderr, "%s: usage: %s wavelen giraffe|uves|facb <wavelength>\n",
                        args[0], args[0]);
                return -1;
            }
        }
	else if (strcasecmp("parameters", args[1]) == 0)
	{
		int i;
                FpcalPars pars;

		if (nArgs != NUM_PARAMETERS+2)
		{
			fprintf(stderr, "%s: set parameters <0> <2> <3> <4>...<%d>\n", 
                                args[0], NUM_PARAMETERS-1);
			return -1;
		}
		printf("aParameters set to: ");
		for (i = 0; i < NUM_PARAMETERS; i++)
		{
			pars.p[i] = atof(args[2+i]);
			printf("%8.4f ", pars.p[i]);
		}
		printf("\n");
                fpcolMdlCalSetPars(&pars, &pFpcalData->model);
	}
        else if (strcasecmp("powell", args[1]) == 0)
        {
            if (strncasecmp("tolerance", args[2], strlen(args[2])) == 0)
                {
                double oldTol = pFpcalData->powellTol;
                pFpcalData->powellTol = atof(args[3]);
                if (pFpcalData->powellTol <= 0)
                    {
                    pFpcalData->powellTol = oldTol;
                    fprintf(stderr,
                            "Invalid powell function tolernace, must be greater then zero\n");
                    }
                }
            else if (strncasecmp("maxits", args[2], strlen(args[2])) == 0)
                {
                int oldIts = pFpcalData->powellMaxIts;
                pFpcalData->powellMaxIts = atoi(args[3]);
                if (pFpcalData->powellMaxIts <= 0)
                    {
                    pFpcalData->powellMaxIts = oldIts;
                    fprintf(stderr,
                            "Invalid powell function max its, must be greater then zero\n");
                    }
                }
           else
               {
               fprintf(stderr,
                       "%s %s: Invalid sub-argument %s.  Should be \"maxits\" or \"tolerance\"\n",
                       args[0], args[1], args[2]);
               }
            printf("Powell function tolerance = %g, number iterations = %d\n",
                   pFpcalData->powellTol, pFpcalData->powellMaxIts);
        }
	else if (strcasecmp("fitfun", args[1]) == 0)
        {
            if (strncasecmp("svd", args[2], strlen(args[2])) == 0)
                {
                pFpcalData->fitFunc = fpcolMdlFit;
                printf("Using Singular Value Decomposition for fitting\n");
                }
            else if (strncasecmp("powell", args[2], strlen(args[2])) == 0)
                {
                pFpcalData->fitFunc = fpcolMdlPowell;
                pFpcalData->powellMode = POWELL_MODE_1;
                printf("Using Modified Powell Minimization for fitting (1)\n");
                }
            else if (strncasecmp("powell2", args[2], strlen(args[2])) == 0)
                {
                pFpcalData->fitFunc = fpcolMdlPowell;
                pFpcalData->powellMode = POWELL_MODE_2;
                 printf("Using Modified Powell Minimization for fitting (2)\n");
                }
            else
                {
                fprintf(stderr,
                        "%s %s: Invalid fit function \"%s\", must be \"svd\" or \"powell\"\n",
                        args[0], args[1], args[2]);
                }
        }
	else if (strcasecmp("fit", args[1]) == 0)
	{
		int i;
                int par;
                int setting;
                char flag[200];

                if (strcasecmp("linear", args[2]) == 0)
                    {
                    for (i = 0; i <= fpcolMDL_PAR_LASTLIN ; ++i)
                        pFpcalData->aFit[i] = 0;
                    }
                else if (strcasecmp("nolinear", args[2]) == 0)
                    {
                    for (i = 0; i <= fpcolMDL_PAR_LASTLIN ; ++i)
                        pFpcalData->aFit[i] = 1;
                    }
                else if (strcasecmp("origin", args[2]) == 0)
                  {
                    pFpcalData->aFit[0] = 0;
                    pFpcalData->aFit[3] = 0;
                  }
                else if (strcasecmp("noorigin", args[2]) == 0)
                  {
                    pFpcalData->aFit[0] = 1;
                    pFpcalData->aFit[3] = 1;
                   }
                else if (strcasecmp("rotation", args[2]) == 0)
                  {
                    pFpcalData->aFit[fpcolMDL_PAR_ROT] = 0;
                   }
                else if (strcasecmp("norotation", args[2]) == 0)
                  {
                    pFpcalData->aFit[fpcolMDL_PAR_ROT] = 1;
                  }
                else if (strcasecmp("scale", args[2]) == 0)
                  {
                    pFpcalData->aFit[fpcolMDL_PAR_SCALE] = 0;
                   }
                else if (strcasecmp("noscale", args[2]) == 0)
                  {
                    pFpcalData->aFit[fpcolMDL_PAR_SCALE] = 1;
                  }
                else if (strcasecmp("nonperp", args[2]) == 0)
                  {
                    pFpcalData->aFit[fpcolMDL_PAR_NONPERP] = 0;
                   }
                else if (strcasecmp("nononperp", args[2]) == 0)
                  {
                    pFpcalData->aFit[fpcolMDL_PAR_NONPERP] = 1;
                  }
                else if (strcasecmp("dist", args[2]) == 0)
                  {
                    for (i = fpcolMDL_PAR_NW_DIST_FIRST; i <= fpcolMDL_PAR_NW_DIST_LAST ; ++i)
                        pFpcalData->aFit[i] = 0;
                  }
                else if (strcasecmp("nodist", args[2]) == 0)
                  {
                    for (i = fpcolMDL_PAR_NW_DIST_FIRST; i <= fpcolMDL_PAR_NW_DIST_LAST ; ++i)
                        pFpcalData->aFit[i] = 1;
                  }
                else if (strcasecmp("wdist", args[2]) == 0)
                  {
                    for (i = fpcolMDL_PAR_W_DIST_FIRST; i <= fpcolMDL_PAR_W_DIST_LAST ; ++i)
                        pFpcalData->aFit[i] = 0;
                  }
                else if (strcasecmp("nowdist", args[2]) == 0)
                  {
                    for (i = fpcolMDL_PAR_W_DIST_FIRST; i <= fpcolMDL_PAR_W_DIST_LAST ; ++i)
                        pFpcalData->aFit[i] = 1;
                  }
/*
                else if (strcasecmp("all", args[2]) == 0)
                  {
                   for (i = 0; i < NUM_PARAMETERS ; ++i)
                      {
                        pFpcalData->aFit[i] = 0;
                      }
                  }
*/
                else if (strcasecmp("noall", args[2]) == 0)
                  {
                   for (i = 0; i < NUM_PARAMETERS ; ++i)
                      {
 
                        pFpcalData->aFit[i] = 1;
                      }
                  }
                /*
                 * Special case - rotation, origin and distortion.  Clear all the
                 * others and set these ones up.
                 */
               else if (strcasecmp("rod", args[2]) == 0)
                  {
                   for (i = 0; i < NUM_PARAMETERS ; ++i)
                        pFpcalData->aFit[i] = 1;                   

                   pFpcalData->aFit[0] = 0;
                   pFpcalData->aFit[3] = 0;
                   pFpcalData->aFit[fpcolMDL_PAR_ROT] = 0;
                   for (i = fpcolMDL_PAR_NW_DIST_FIRST; i <= fpcolMDL_PAR_NW_DIST_LAST ; ++i)
                        pFpcalData->aFit[i] = 0;

                  }
                else if (sscanf(args[2],"%d=%d", &par, &setting) == 2)
                  {
                    if ((par < NUM_PARAMETERS)&&((setting == 0)||(setting == 1)))
                      pFpcalData->aFit[par] = setting;
                    else
                      {
                         fprintf(stderr, 
                                "First value must be the model parameter index, range 0 to %d\n",
                                NUM_PARAMETERS-1);
                        fprintf(stderr, "Second value must be 1 or 0  (don't fit) or 0 (fit)\n");
                        fprintf(stderr, " or alternativly \"fit\" or \"nofit\"\n");
                       return -1;
                      }
                  }
                else if (sscanf(args[2],"%d=%s", &par, flag) == 2)
                  {
                    setting = -1;
                    if (strcasecmp(flag,"fit") == 0)
                        setting = 0;
                    else if (strcasecmp(flag,"nofit") == 0)
                        setting = 1;
                    if ((par < NUM_PARAMETERS)&&((setting == 0)||(setting == 1)))
                      pFpcalData->aFit[par] = setting;
                    else
                      {
                        fprintf(stderr, 
                                "First value must be the model parameter index, range 0 to %d\n",
                                NUM_PARAMETERS-1);
                         fprintf(stderr, "Second value must \"fit\" or \"nofit\"\n");
                        return -1;
                      }
                  }
                else
                  {
                    if ((nArgs-2) != NUM_PARAMETERS)
                      {
                        for (i = 0; i < NUM_PARAMETERS; i++)
                          {
                            fprintf(stderr, "<%d> ", i);
                          }
                        fprintf(stderr, "\n");
                        fprintf(stderr,"You supplied %d parameter%s instead of %d\n",
                                nArgs-2, 
                                (nArgs-2 == 1 ? "" : "s"),
                                NUM_PARAMETERS);
                        fprintf(stderr, "%s: set fit <0> <1> <2>...<%d>\n", args[0],
                                NUM_PARAMETERS-1);
                        fprintf(stderr, "%s: set fit [no]all\n", args[0]);
			fprintf(stderr, "%s: set fit [no]linear\n", args[0]);
			fprintf(stderr, "%s: set fit [no]origin\n", args[0]);
			fprintf(stderr, "%s: set fit [no]scale\n", args[0]);
			fprintf(stderr, "%s: set fit [no]rotation\n", args[0]);
			fprintf(stderr, "%s: set fit [no]dist\n", args[0]);
			fprintf(stderr, "%s: set fit [no]wdist\n", args[0]);
                        fprintf(stderr, "%s: set fit n=0|1|[no]fit\n", args[0]);

                        return -1;
                      }
                    for (i = 0; i < NUM_PARAMETERS; i++)
			pFpcalData->aFit[i] = atof(args[2+i]);

                  }
                printf("Fit set to :");
                for (i = 0; i < NUM_PARAMETERS; i++)
                  {
                    printf("%d ", pFpcalData->aFit[i]);
                  }
                printf("\n    (0=fit, 1=don't fit, use \"show fit\" for more detail)\n");
                printf("Next command in normal sequence - \"fit\" or \"set fit\" or \"show\"\n");

                fpcolMdlCalCheckFitType(pFpcalData->aFit);
        }
        else
	{
		fprintf(stderr, "%s: unknown sub-command <%s>\n", args[0], args[1]);
		return -1;
	}

	return 0;
}




/*
 * This function peforms load level operations on the telescope model parameters
 */
static int parsFn
         (int			nArgs,
	 char			*args[],
	 FpcalData		*pFpcalData)
{
    StatusType status = STATUS__OK;
    /*
     * Handle help command and insufficent arguments.  Only  "show" has
     * no arguments.
     */
    if ((nArgs >= 2 && strcasecmp("help", args[1]) == 0)||
        (nArgs < 2)||
        (nArgs == 2 && (
            (strcasecmp("show", args[1]) != 0)&&
            (strcasecmp("normalize", args[1]) != 0)&&
            (strncasecmp("save", args[1], 4) != 0))))
	{
        printf("%s: usage: %s loadsds <plate #> \n", args[0], args[0]);
        printf("%s: usage: %s loadvlt <plate #>\n", args[0], args[0]);
        printf("%s: usage: %s show\n", args[0], args[0]);
        printf("%s: usage: %s normalize\n", args[0], args[0]);
        printf("%s: usage: %s rotate <degrees>\n", args[0], args[0]);
        printf("%s: usage: %s scale <factor>\n", args[0], args[0]);
        printf("%s: usage: %s savesds\n", args[0], args[0]);
        printf("%s: usage: %s savevlt\n", args[0], args[0]);
        printf("%s: usage: %s set [par] [val]\n", args[0], args[0]);
        return 0;
	}
    if (strcasecmp("show", args[1]) == 0)
        {
        /*
         * Show details and analysis of the parameters
         */
        int i;
        const FpcalPars *pars;

        pars = fpcolMdlCalGetPars(&pFpcalData->model);
        printf("Parameters are for field plate %d\n", pFpcalData->fieldPlate);
        printf("Raw Linear model Parameters:\n");
        for (i = 0 ; i <= fpcolMDL_PAR_LASTLIN ; ++i)
          {
            printf("\t%14.6f (%s %s)\n", pars->p[i], ModelParNames[i], ModelParUnits[i]);
          } 
        fpcolMdlDecompLin(pars->p);

        printf("Extra rotation applied in addition to the above linear model = %.2g degrees\n", 
               pars->p[fpcolMDL_PAR_ROT]);
        printf("Extra scale applied in addition to the above linear model = %.4g\n", 
               pars->p[fpcolMDL_PAR_SCALE]);
        printf("Extra  nonperpendicularity applied in addition to the above lin model = %.4g\n", 
               pars->p[fpcolMDL_PAR_NONPERP]);


        printf("Non-wavelength dependent distortion parameters:\n");
        for (i = fpcolMDL_PAR_NW_DIST_FIRST; i <= fpcolMDL_PAR_NW_DIST_LAST ; ++i)
          printf("\t%14.6f (%s %s)\n", pars->p[i], ModelParNames[i], ModelParUnits[i]); 
      

        printf("Wavelength dependent distortion parameters:\n");
        for (i = fpcolMDL_PAR_W_DIST_FIRST; i <= fpcolMDL_PAR_W_DIST_LAST ; ++i)
          printf("\t%14.6f (%s %s)\n", pars->p[i], ModelParNames[i], ModelParUnits[i]); 
        

        printf("Remaining (currenlty undefined) parameters:\n");
        for (i =  fpcolMDL_PAR_W_DIST_LAST+1; i < NUM_PARAMETERS ; ++i)
          printf("\t%14.6f (%s %s)\n", pars->p[i], ModelParNames[i], ModelParUnits[i]); 
        printf("\n");

       }
    else if (strcasecmp("normalize", args[1]) == 0)
       {
        FpilModelType *newModel;
        FpcalPars pars;
        StatusType status = 0;
        if (pFpcalData->fieldPlate == 0)
          {
            fprintf(stderr,"%s:You have not yet read in a model\n", args[0]);
            return -1;
          }        
        /*
         * Fetch the current parameters and use them to create
         * a model we can play with.
         */
        pars = *fpcolMdlCalGetPars(&pFpcalData->model);
        fpcolMdlCreateModel(NUM_PARAMETERS,
                            pars.p,
                            pFpcalData->fieldPlate, /* field plate number */
                            &newModel,
                            &status);

        if (status != STATUS__OK)
          {
            fprintf(stderr, "Error creating model - status code %lX", (long int)status);
            return -1;
          }
       /*
        * Normalize the model.
        */
        fpcolMdlNormalizeModel(newModel,&status);
        if (status != STATUS__OK)
          {
            StatusType ignore = STATUS__OK;
            fpcolMdlFree(newModel, &ignore);
            fprintf(stderr, "Error normalizing model - status code %lX", (long int)status);
            return -1;
          }
        /*
         * Fetch the parameters from this model and store them into this program
         * using fpcolMdlCalSetPars().
         */
        fpcolMdlGetModel(newModel, NUM_PARAMETERS, pars.p,
                         &status);
        if (status != STATUS__OK)
          {
            StatusType ignore = STATUS__OK;
            fpcolMdlFree(newModel, &ignore);
            fprintf(stderr, "Error getting new model - status code %lX", (long int)status);
            return -1;     
          }
        fpcolMdlCalSetPars(&pars, &pFpcalData->model);
        fpcolMdlFree(newModel,&status);
 
       }
    else if (strcasecmp("rotate", args[1]) == 0)
        {
        /*
         * Apply a rotation.
         */
        double xz, yz, xs, ys, perp, orient;
        double rotation = atof(args[2]);
        double cosine;
        double sine;
        double a1, b1, c1, d1, e1, f1, b2, c2, e2, f2;
        FpcalPars pars;

        pars = *fpcolMdlCalGetPars(&pFpcalData->model);

        printf("Will apply a rotation of %.6f degrees to the model\n",
               rotation);

        cosine = cos(rotation/180.0*M_PI);
        sine   = sin(rotation/180.0*M_PI);
            
        slaDcmpf(pars.p, &xz, &yz, &xs, &ys, &perp, &orient);
        printf("current orientation = %.6f radians, %.1f degrees\n", 
               orient, orient*180.0/M_PI);


        /*
         * Short hand for existing parameter values.
         */
        a1 = pars.p[0];
        b1 = pars.p[1];
        c1 = pars.p[2];
        d1 = pars.p[3];
        e1 = pars.p[4];
        f1 = pars.p[5] ;

        /*
         * Short hand for rotation.
         */
        b2 = cosine;
        c2 = sine;
        e2 = -1 * sine;
        f2 = cosine;

        /*
         * The original model is
         *  x1 = a1 + b1.x0 + c1.y0
         *  y1 = d1 + e1.x0 + f1.y0
         *
         * We want to apply something like this.
         *
         * x2 = b2.x1 + c2.y1
         * y2 = e2.x1 + f2.y1
         *
         * Which is equivalent to this.
         *
         *  x2 = b2.(a1 + b1.x0 + c1.y0) + c2.(d1 + e1.x0 + f1.y0)
         *  y2 = e2.(a1 + b1.x0 + c1.y0) + f2.(d1 + e1.x0 + f1.y0)
         *
         * And working out the new model, we get.
         *  x1 = (b2.a1 + c2.d1) + (b2.b1 + c2.e1) x0 + (b2.c1 + c2.f1) y0
         *  y2 = (e2.a1 + f2.d1) + (e2.b1 + f2.e1) x0 + (e2.c1 + f2.f1) y0
         *
         * Hence we can see our new parameters.
         */
        pars.p[0] = b2*a1 + c2*d1;
        pars.p[1] = b2*b1 + c2*e1;
        pars.p[2] = b2*c1 + c2*f1;
        pars.p[3] = e2*a1 + f2*d1;
        pars.p[4] = e2*b1 + f2*e1;
        pars.p[5] = e2*c1 + f2*f1;



        slaDcmpf(pars.p, &xz, &yz, &xs, &ys, &perp, &orient);
        printf("new orientation = %.6f radians, %.1f degrees\n", 
               orient, orient*180.0/M_PI);
 

        /*
         * Save the new model details.
         */
        fpcolMdlCalSetPars(&pars, &pFpcalData->model);

        }
    else if (strcasecmp("scale", args[1]) == 0)
        {
        /*
         * Scale the model.
         */
        double xz, yz, xs, ys, perp, orient;
        double scale = atof(args[2]);
        FpcalPars pars;

        pars = *fpcolMdlCalGetPars(&pFpcalData->model);

        printf("Will apply a scale of %.2f degrees to the model\n",
               scale);
        slaDcmpf(pars.p,
                 &xz, &yz, &xs, &ys, &perp, &orient);
        printf("Current Scale (x/y)= %.6f, %.6f\n", xs, ys);

        pars.p[1] *= scale;
        pars.p[2] *= scale;
        pars.p[4] *= scale;
        pars.p[5] *= scale;

        slaDcmpf(pars.p,
                 &xz, &yz, &xs, &ys, &perp, &orient);
        printf("New Scale (x/y)= %.6f, %.6f\n", xs, ys);


        /*
         * Save the new model details.
         */
        fpcolMdlCalSetPars(&pars, &pFpcalData->model);

        }
    else if ((strcasecmp("loadsds", args[1]) == 0)||
             (strcasecmp("loadvlt", args[1]) == 0))

        {
        /*
         * Handle loading and saving the model (note, saving should not need
         * the plate number - we should work out where else we can get that)
         */
        int plate = atol(args[2]);
        FpilModelType *newModel;
        FpcalPars pars;

        if ((plate < 0)||(plate > 2)) 
            {
            printf("Plate number must be 1 or 2\n");
            return 1;
            }


        if (strcasecmp("loadvlt", args[1]) == 0)
            {
            /*
             * To Extract the model from the VLT OLDD execute the fpcolMdlExtract 
             * command, which extracts it to a file we can then as per loadsds.
             */
            char command[100];
            int returnVal;
            sprintf(command,"fpcolMdlExtract %d\n", plate);
            if ((returnVal = system(command)) != 0)
                {
                fprintf(stderr,"%s: Command to extract data from VLT database failed\n",
                        args[0]);
                fprintf(stderr,"%s: Command tried \"%s\", returned status %d\n",
                        args[0], command, returnVal);
                return 1;
                }
            printf("Sucessfully extracted parameters from VLT OLDB\n");
            }           
        
        /*
         * Read the new model.  
         */
        fpcolMdlReadModel("", "", 0, 0, plate, &newModel, &status);
        
        if (status != STATUS__OK)
            {
            fprintf(stderr, "Failure reading new model file, status code %lx\n",
                    (long)status);
            return 1;
            }
        printf("Successfull read of new model file for plate %d\n", plate);
        
        /*
         * Fetch the parameters from this model and store them into this program
         * using fpcolMdlCalSetPars().
         */
        fpcolMdlGetModel(newModel, NUM_PARAMETERS, pars.p,
                         &status);
        fpcolMdlCalSetPars(&pars, &pFpcalData->model);

        
        /*
         * Ensure the rest of the program knows the field plate
         */
        pFpcalData->fieldPlate = plate;
        
        }
    else if ((strcasecmp("savesds", args[1]) == 0)||
             (strcasecmp("savevlt", args[1]) == 0))
      {
        FpilModelType *newModel;
        FpcalPars pars;
        StatusType status = 0;
        if (pFpcalData->fieldPlate == 0)
          {
            fprintf(stderr,"%s:You have not yet read in a model\n", args[0]);
            return -1;
          }
        /*
         * Fetch the current parameters and use them to create
         * a model we can play with.
         */
        pars = *fpcolMdlCalGetPars(&pFpcalData->model);
        fpcolMdlCreateModel(NUM_PARAMETERS,
                            pars.p,
                            pFpcalData->fieldPlate, /* field plate number */
                            &newModel,
                            &status);

        if (status != STATUS__OK)
          {
            fprintf(stderr, "Error creating model - status code %lX", (long int)status);
            return -1;
          }
 
        /*
         * Save the model to a file.
         */
        fpcolMdlNormalizeModel(newModel, &status);
        fpcolMdlWriteModel(newModel, 0, &status);
        if (status != STATUS__OK)
            {
            fprintf(stderr,"Error writing model file - status code %lx\n",
                    (long)status);
            return 1;
            }
        fpcolMdlFree(newModel, &status);
        printf("Model parameters saved to plate specific file\n");
        
        if (strcasecmp("savevlt", args[1]) == 0)
            {
            /*
             * To load into the VLT software, execute the fpcolMdlLoad command,
             * which loads the model from the file we just created.
             */
            char command[100];
            int returnVal;
            sprintf(command,"fpcolMdlLoad %d\n", pFpcalData->fieldPlate);
            if ((returnVal = system(command)) != 0)
                {
                fprintf(stderr,"%s: Command to load data into VLT database failed\n",
                        args[0]);
                fprintf(stderr,"%s: Command tried \"%s\", returned status %d\n",
                        args[0], command, returnVal);
                return 1;
                }
            printf("Sucessfully loaded parameters into VLT OLDB for plate %d\n",
                   pFpcalData->fieldPlate);
            }
        
        }
    else if ((strcasecmp("set", args[1]) == 0)&&(nArgs == 4))
      {
        int par = atol(args[2]);
        double value = atof(args[3]);
        FpcalPars pars = *fpcolMdlCalGetPars(&pFpcalData->model);


        if ((par < 0)||(par >= NUM_PARAMETERS))
          {
            fprintf(stderr,"First argument should be the parameter model index\n");
            fprintf(stderr,"Valid range is 0 to %d\n", NUM_PARAMETERS-1);
            return -1;
          }

        printf("Setting tel model parameter %d value to %.8f\n", par, value);

        pars.p[par] = value;

        fpcolMdlCalSetPars(&pars, &pFpcalData->model);
        return 0;

      }
    else 
        {
        char *helpCmd[2] = {"mdlpars", "help"};
        fprintf(stderr,"%s unknown sub-command \"%s\"\n", args[0], args[1]);
        parsFn(1,helpCmd, pFpcalData);
        return -1;
        } 
    return 0;
}

/*
 * This function provides a way to save the header information ths program
 * has previously loaded with a load or loadhd function.  The loadhd function
 * can then reload this information.
 */
static int saveHeadfn(
               int            nArgs,
               char           *args[],
               FpcalData      *pFpcalData)
{
  int i;
  
  FILE *f = 0;
  const FpcalPars *pars;
  
  if (((nArgs == 2) && (strcasecmp("help", args[1]) == 0))||
      (nArgs != 2))
    {
      printf("%s: usage: %s <textFile>\n", args[0], args[0]);
      return 0;
    }
  if (!pFpcalData->bInitialised)
    {
      fprintf(stderr, "%s: must load fpcaldata first!\n", args[0]);
      return -1;
    }

  if ((f = fopen(args[1], "w"))  == 0)
    {
      fprintf(stderr,"%s: Failed to open file %s for output\n", args[0], args[1]);
      perror("");
      return -1;
    }
  pars = fpcolMdlCalGetPars(&pFpcalData->model);
  
  fprintf(f, "#FLAMES FP HEADER CALIBRATION DETAILS\n");
  fprintf(f, "%5d\t\t\t\t#Field Plate\n", pFpcalData->fieldPlate);
  fprintf(f, "%10.4f\t\t\t#Atm Temp, Degrees Celsius\n", pFpcalData->atmTemp);
  fprintf(f, "%10.4f\t\t\t#Atm Pres, Millibars\n", pFpcalData->atmPressure);
  fprintf(f, "%10.4f\t\t\t#Atm Humidity, Percent\n", pFpcalData->atmHumidity);
  fprintf(f, "%10.4f %10.4f %10.4f #Gir, UVES, FACB Wavelengths, Nanometers\n", 
          fpcolMdlCalWavelength(&pFpcalData->model, pFpcalData->curWaveIndex.Giraffe),
          fpcolMdlCalWavelength(&pFpcalData->model, pFpcalData->curWaveIndex.UVES),
          fpcolMdlCalWavelength(&pFpcalData->model, pFpcalData->curWaveIndex.FACB));
  fprintf(f, "%10.4f\t\t\t#Adc angle, Degrees\n", pFpcalData->posAngle);
  fprintf(f, "%s\t\t#Observation Date\n", mjdToString(pFpcalData->observationDate));
  fprintf(f, "%9.4f\t\t\t#Equinox, Years AD\n", pFpcalData->equinox);
  fprintf(f, "%s %s\t#Field centre - mean RA\n",
          raToString(pFpcalData->fieldCentreMeanRA),
          decToString(pFpcalData->fieldCentreMeanDEC));
  fprintf(f, "%5d\t\t\t\t#Number of fibres\n", pFpcalData->nFibres_a);
  
  
  for (i = 0; i < pFpcalData->nFibres_a; i++)
    {
      fprintf(f, "%3d %s %s\t#Object %3d buttn and mean position\n",
              pFpcalData->buttons[i],
              raToString( pFpcalData->meanRA[i]),
              decToString(pFpcalData->meanDEC[i]),i);
      
    }
  fprintf(f, "####\n");
  
  for (i = 0; i < NUM_PARAMETERS; i++)
    {
      fprintf(f, "%.6f\t\t\t#Tel model parameter %d\n",
              pars->p[i],i);
    }
  fprintf(f, "####\n");

  fclose(f);
  return 0;
}

/*
 * This function provides a way to load the header information ths program
 * has previously saved with a savehd function.  Since the failed saved
 * by savehd is a defined text format - the use can edit the values before
 * loading them.  Otherwise this function works as pert the "load" function.
 */
static int loadHeadfn(
               int            nArgs,
               char           *args[],
               FpcalData      *pFpcalData)
{
  if (((nArgs == 2) && (strcasecmp("help", args[1]) == 0))||
      (nArgs != 2))
    {
      printf("%s: usage: %s <textFile>\n", args[0], args[0]);
      return 0;
    }
  if (getFpcalDataText(args[1], args[2], pFpcalData) == 0)
      {
      printf("Next command in normal sequence - \"offsets\" or \"show\" or \"mdlpars show\"\n");
      return 0;
      }
  else
      return -1;
}

/*
 * Extract data which allows an external fit to be done.
 */
static int extractFn
	(int			nArgs,
	 char			*args[],
	 FpcalData		*pFpcalData)
{
    StatusType status = 0;
    int i;
    FILE *out;
    if (((nArgs == 2) && (strcasecmp("help", args[1]) == 0))||
        (nArgs != 2))
        {
        printf("%s: usage: %s <textFile>\n", args[0], args[0]);
        return 0;
        }
    if (!pFpcalData->bInitialised)
	{
        fprintf(stderr, "%s: must load fpcaldata first! Use \"load\"\n", args[0]);
        return -1;
	}

    if ((out = fopen(args[1], "w")) == 0)
        {
        fprintf(stderr, "%s: Failed to open file %s\n", args[0], args[1]);
        } 
    /*
     * Calculate offsets and create the model.
     */
    calculateOffsets(pFpcalData, &status);
    if (fpcolMdlCalCreateModels(pFpcalData) != 0) 
        return -1;
    fprintf(out, "fib wavelen     xi          eta   o-ra o-dec\n");
    /*
     *
     */
    for (i = 0; i < pFpcalData->nOffsets ; ++i)
        {
        double xi;
        double eta;
        int buttonIndex =  fpcolMdlCalFindButton(pFpcalData->nFibres_a,
                                                 pFpcalData->buttons,
                                                 pFpcalData->offsetButtonNo[i]);
        
        int waveIndex = pFpcalData->offsetWaveIndex[i];
        

        FpilModelType *telModel = fpcolMdlCalWaveModel(&pFpcalData->model, waveIndex);
        
        fpcolMdlRd2Tan(telModel,
                       pFpcalData->fieldCentreAppRA,
                       pFpcalData->fieldCentreAppDEC,
                       pFpcalData->apparentRA[buttonIndex],
                       pFpcalData->apparentDEC[buttonIndex],
                       pFpcalData->observationDate,
                       &xi,&eta,&status);

        fprintf(out, "%3d %5.1f %11.5f %11.5f %5.2f %5.2f\n",
                pFpcalData->offsetButtonNo[i],
                fpcolMdlCalWavelength(&pFpcalData->model, waveIndex),
                xi*180.0/M_PI,
                eta*180.0/M_PI,
                RAD2SEC(pFpcalData->offsetRA[i]),
                RAD2SEC(pFpcalData->offsetDEC[i]));
            
        }
    fclose(out);

    if (status != STATUS__OK)
        return 1;
    return 0;
}



/*
 * function prototype for <helpFn>.
 * We actually want to access <assoc> from the function,
 * hence we have to add it to <assoc> first & then after
 * the declaration we will be able to access it.
 */
static int helpFn
	(int			nArgs,
	 char			*args[],
	 FpcalData		*pFpcalData);


/*
 * This structure used to map user commands and the
 * appropriate function to handle each of them.
 */
typedef struct
{
	char *command;
	int (*fn) (int nArgs,
                   char *args[],
                   FpcalData *pFpcalData);
} CommandAssoc;

CommandAssoc	assoc[] = {{"load", loadFn},
                           {"fit", fitFn},
                           {"help", helpFn},
                           {"offsets", offsetsFn},
                           {"quit", quitFn},
                           {"exit", quitFn},
                           {"set", setFn},
                           {"show", showFn},
                           {"mdlpars",parsFn },
                           {"savehd", saveHeadfn },
                           {"loadhd", loadHeadfn },
                           {"extract", extractFn },
                           {"clear",  clearFn },
                           {"plot",   plotFn },
                           {"remmeanoff", remMeanFn }
};


/*
 * very basic help information.
 */
static int helpFn
	(int			nArgs,
	 char			*args[],
	 FpcalData		*pFpcalData)
{
	int i;

	printf("Available commands are:\n");
	for (i = 0; i < (int) NUM_ELEMENTS(assoc); i++)
	{
		printf("\t%s\n", assoc[i].command);
	}
	printf("Type \"<command> help\" for further help.\n");

	(void) nArgs;
	(void) args;
	(void) pFpcalData;
	return 0;
}


int main (int argc, char *argv[])
{
	FpcalData		fpcalData;
	int				i, n, bFound;
	char			*p = NULL, buffer[128];
        argv0 =                 argv[0]; /* This is needed elsewhere */
        
	fpcalConstructor(&fpcalData);

        if (((sizeof(ModelParNames)/sizeof(ModelParNames[0])) != NUM_PARAMETERS)||
            ((sizeof(ModelParUnits)/sizeof(ModelParUnits[0])) != NUM_PARAMETERS))
          {
            fprintf(stderr,"Number of parameters have changed\n");
            fprintf(stderr,"This program needs to be updated appropiately\n");
            fprintf(stderr,"%ld, %ld, %ld, %ld, %ld, %ld, %d\n",
                   (long)sizeof(ModelParNames),
                   (long)sizeof(ModelParNames[0]),
                   (long)(sizeof(ModelParNames)/sizeof(ModelParNames[0])),
                   (long)sizeof(ModelParUnits),
                   (long)sizeof(ModelParUnits[0]),
                   (long)(sizeof(ModelParUnits)/sizeof(ModelParUnits[0])),
                   NUM_PARAMETERS);
 
            exit(1);
          }

        printf("FP Calibration program:Use \"help\" for list of commands\n");
        printf("Next command in normal sequence - \"load\", \"loadhd\" or \"mdlpars loadsds/loadvlt\"\n");
	/* continually prompt for user commands */
	/* while ((n = read(0, buffer, sizeof(buffer))) > 0) */
	while ((n = getline(stdin, buffer, sizeof(buffer))) >= 0)
	{
		if (n == 0)
			continue;
		p = strtok(buffer, " ");
		bFound = 0;

		/* is there a function (handler) for the specified command? */
		for (i = 0; i < (int) NUM_ELEMENTS(assoc); i++)
		{
			if (strcasecmp(assoc[i].command, p) == 0)
			{
				int		nArgs = 0, rv;
				char	*args[MAX_ARGS];

				/* bundle up all arguments nicely, ala argc, argv */
				while (nArgs < MAX_ARGS && p != NULL)
				{
					args[nArgs++] = p;
					p = strtok(NULL, " ");
				}

				/* invoke the handler */
				rv = assoc[i].fn(nArgs, args, &fpcalData);
				if (rv != 0)
					fprintf(stderr, "ERROR(%d) handling command\n", rv);
				else
					printf("OK\n");
				bFound = 1;
				break;
			}
		}
		if (!bFound)
			fprintf(stderr, "Unknown Command - enter \"help\" for a list\n");
	}
	return 0;
}

