
/*******************************************************************************
* E.S.O. - VLT project
*
* "@(#) $Id: ACMM:ozpozFpcol/ws/src/fpcolMdlCalCentGauss.c,v 1.63 07-Oct-2008 11:10:27+10 tjf $"
*
*
* who       when      what
* --------  --------  ----------------------------------------------
* ssmedley  07/09/01  created from AAO's GCamCentroidGauss.
*
*/

/*
 *+			G C a m C e n t r o i d G a u s s

 *  Function name:
      GCamCentroidGauss

 *  Function:
      Centroid an image within a subset of the full image area using 
      Gaussian fitting.

 *  Description:
      This routine takes an image and obtains gets the center of the strongest
      star image in it by collapsing it in both directions and fitting a
      gaussian to the strongest gaussian-like profile in each dimension.

 *  Language:
      C

 *  Call:
      (Void) = GCamCentroidGauss (Window,data,bias,results,status)

 *  Parameters:   (">" input, "!" modified, "W" workspace, "<" output)
      (>) Window	(GCamWindowType) Window definition.  Note this is the
			definition of the window to be centroided within the
			data area.  The MaxX and MaxY elements should indicate
			the size of the data area, where as Xoffset,Yoffset,
			Xdim and Ydim indicates the area to be centroided.
			PixelSize indicates the size of pixels.  This is
			DIFFERENT from the window definiton used to obtain
			the image (say with GCamImage())
      (>) data		(Void *) The data area.  Should be an unsigned 
			integer array corresponding to the pixel size.
      (>) bias		(int) Ignored.
      (<) results	(double {[2]}) The results of the centroiding.  
			This value is within the centroiding window, not
			the original window.  The value within the original
			window is 
			results[0]+Window->Xoffset,results[1]+Window->Yoffset.
			results[2] is set to -1.
      (!) status	(Int *) Modified status. Possible failure codes are
			GCAM__CENT_ZERO => All values are zero (after bias 
                        removal).
			GCAM__CENT_RANGE => Center of gravity outside of
			the range.  The probably indicates a programming error
			in this routine.
			GCAM__CENT_FITERR => Can't fit the image.
			


 *  Include files: GCam.h

 *  External functions used:
      malloc	(CRTL)	Allocate memory
      ErsRep	(Ers)	Report an error

 *  External values used:
      none

 *  Prior requirements:
      none  (Indepentent of GCamActivate)

 *  Support: Tony Farrell, AAO

 *-
 * Sccs Id:     gcamcentgauss.c, Release 1.2, 05/04/98
  
 * Copyright (c) Anglo-Australian Telescope Board, 1995.
   Not to be used for commercial purposes without AATB permission.
  

 *  History:
      19-Jun-1994 - TJF - Original version
       6-Dec-1994 - KS  - Gaussian version, produced by combining the
                          code for GCamCentroid with the gaussian fitting
                          algorithm written by JOS for the Figaro EMLT program.
      18-Aug-1995 - TJF - Use status CENT_FITERR instead of CENT_ERROR when
				fit fails.
      21-Aug-1995 - TJF - Moved this routine from the module gcamcentroid.c
      				module to gcamcentgauss.c
      {@change entry@}
 */
/* #include "DitsTypes.h" */
#include "fpcolMdlCalCentGauss.h"
/* #include "gcam_err.h" */
/* #include "Ers.h" */
/* #include "DitsMsgOut.h" */
#include <stdio.h>

#ifdef VMS
#   include <unixlib.h>
#elif defined(VxWorks)
#   include <memLib.h>
#else
#   include <stdlib.h>  
#endif

static int fig_gfit__();

extern void GCamCentroidGauss (GCamWindowType *Window,
			  void  *data,
			  int bias ,
			  double *results,
			  StatusType *status)
{

    register int x,y,offset;

    float Center;              /* Profile center determined by fit */
    float Cont0;               /* Bias determined by fit - ignored */
    int   EndPix;              /* Pixel at end of window being fitted */
    float Fwhm;                /* Full width at half max of profile - ignored */
    int   Pixels;              /* Number of pixels in array */
    int   ProfEnd;             /* End of fitted profile */
    int   ProfStart;           /* Start of fitted profile */
    float S2nf;                /* Signal to noise from fit - ignored */
    int   StartPix;            /* Pixel at start of window being fitted */
    float Strength;            /* Strength of fitted profile */
    float StrongestCenter;     /* Centre of strongest profile fitted */
    float StrongestStrength;   /* Strength of strongest profile fitted */

    float *xtotals = (float *)malloc (Window->Xdim * sizeof(float));
    float *ytotals = (float *)malloc (Window->Ydim * sizeof(float));

	(void) bias;	/* unused parameter */

    if (*status != GCAM__OK) return;

    results[2] = -1;	/* We don't get fwhm */

    if ((!xtotals)||(!ytotals))
    {
	*status = GCAM__MALLOCERR;
	return;
    }
/*
 *  Zero the xtotals and ytotals arrays.
 */
    for (x = 0; x < Window->Xdim ; ++x)
	xtotals[x] = 0;    
    for (y = 0; y < Window->Ydim ; ++y)
	ytotals[y] = 0;    

/*
 *	For each valid y line.
 */
    
    for (y = 0 ; y < Window->Ydim ; ++y)
    {
/*
 *	Offset is the total offset in the data array to the next element.
 *	It is made up of the y line number (y+Window->Yoffset) times the
 *	the x dimension.
 */
	offset = (y+Window->Yoffset)*Window->MaxX;
/*
 *	Sum each valid x value into both the y and x totals as appropaite.
 *	Take note of pixel size here.
 */
	if (Window->PixelSize == 4)
	{
	    for (x = 0 ;x < Window->Xdim ; ++x)
	    {
			/* double d = (double)(((unsigned long *)(data))
				[x+Window->Xoffset+offset]); */
			double d = ((double *) data)[x+Window->Xoffset+offset];

	        ytotals[y] += d;
	        xtotals[x] += d;
            }
        }
	else if (Window->PixelSize == 2)
        {
	    for (x = 0 ;x < Window->Xdim ; ++x)
	    {
		long int  d = (long int)(((unsigned short *)(data))
				[x+Window->Xoffset+offset]);

	        ytotals[y] += d;
	        xtotals[x] += d;

            }
        }
	else    /* Assume Window->PixelSize == 1 */
        {
	    for (x = 0 ;x < Window->Xdim ; ++x)
	    {
		int d = (int)(((unsigned char *)(data))
				[x+Window->Xoffset+offset]);

	        ytotals[y] += d;
	        xtotals[x] += d;
            }
	}
    }

	/* TODO: SCoTT was here */
	for (y = 0; y < Window->Ydim; y++)
	{
		printf("ytotals[%d] = %f\n", y, ytotals[y]);
	}
	for (x = 0; x < Window->Xdim; x++)
	{
		printf("xtotals[%d] = %f\n", x, xtotals[x]);
	}

/*
 *  xtotals contains the image collapsed across x. We run through the array
 *  attempting to fit a gaussian at each point in it. Most of the time,
 *  the gaussian fitter (fig_gfit()) will not find any suitable gaussian
 *  profile. If it does, it returns a strength for the profile. We are 
 *  looking for the strongest profile, so we record the center of the strongest
 *  line and compare the strength of each line as we find it with the current
 *  strongets line. fig_gfit() is passed a range in which to find the profile.
 *  If it finds one, it returns the range in the array covered by the profile,
 *  so we can skip over that. 
 */

    StrongestCenter = 0.0;
    StrongestStrength = 0.0;
    StartPix = 3;
    EndPix = StartPix + 10;
    Pixels = Window->Xdim;
    while (EndPix < Pixels - 4) {

       /*  The call to fig_gfit() may look a bit odd - this is because this
        *  is a Fortran routine, converted to C using f2c. So the routine
        *  name follows f2c conventions - ending in '__' and all the parameters
        *  are passed by reference. Setting Fwhm to zero before the call
        *  causes the routine to determine the full width for itself rather
        *  than assume a specified value.
        */

       Fwhm = 0.0;
       Center = 0.0;
       fig_gfit__(&Center,&Fwhm,&Strength,&ProfStart,&ProfEnd,&Cont0,
                           &S2nf,xtotals,&Pixels,&StartPix,&EndPix);

       /*  If we found a profile, Center will be non-zero. We can move the
        *  search range past the profile just found. We also see if this
        *  is the strongest line found so far.
        */

       if (Center != 0)
	   {
          StartPix = ProfEnd + 1;
          EndPix = StartPix + 10;
          if (Strength > StrongestStrength)
		  {
             StrongestStrength = Strength;
             StrongestCenter = Center;
          }
       } else {
          StartPix++;
          EndPix++;
       }
    }	/* end loop */

    if (StrongestCenter == 0) {
       /* ErsRep (0,status, */
		fprintf(stderr, 
         "GCamCentroidGauss: Insufficient signal to fit profile in X\n");
       *status = GCAM__CENT_FITERR;
       results[0] = 0;
       results[1] = 0;
    } else {
       results[0] = StrongestCenter;
    }

    /*  Now we repeat the process in the Y direction. */

    if (*status == 0) {
       StrongestCenter = 0.0;
       StrongestStrength = 0.0;
       StartPix = 3;
       EndPix = StartPix + 10;
       Pixels = Window->Ydim;
       while (EndPix < Pixels - 4) {

          Fwhm = 0.0;
          fig_gfit__(&Center,&Fwhm,&Strength,&ProfStart,&ProfEnd,&Cont0,
                           &S2nf,ytotals,&Pixels,&StartPix,&EndPix);

          if (Center != 0) {
             StartPix = ProfEnd + 1;
             EndPix = StartPix + 10;
             if (Strength > StrongestStrength) {
                StrongestStrength = Strength;
                StrongestCenter = Center;
             }
          } else {
             StartPix++;
             EndPix++;
          }
       } 

       if (StrongestCenter == 0) {
          /* ErsRep (0,status, */
		   fprintf(stderr, 
              "GCamCentroidGauss: Insufficient signal to fit profile in Y\n");
          *status = GCAM__CENT_FITERR;
          results[1] = 0;
          results[0] = 0;
       } else {
          results[1] = StrongestCenter;
       }
   }

}


/*
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *
 *                          f i g - g f i t
 */

/*  The following code is the source of the Fortran function fig_gfit,
 *  as converted by f2c, with the following minor change to the prologue:
 *  the include of f2c.h has been removed, to simplify compilation on systems
 *  that do not have f2c installed, and the three definitions that this
 *  code requires from f2c.h - integer, real and doublereal - have been
 *  provided explicitly, and dabs and abs. (KS/AAO - 6th Dec 1994).
 */

/* fig_gfit.f -- translated by f2c (version 19940927).
   You must link the resulting object file with the libraries:
	-lf2c -lm   (in that order)
*/

typedef int integer;
typedef float real;
typedef double doublereal;
#define abs(x) ((x) >= 0 ? (x) : -(x))
#define dabs(x) (doublereal)abs(x)

/* #include "f2c.h" */

/* + */
/* Subroutine */static 
	int fig_gfit__(xcen, fwhm, stren, nlc, nhc, cont0, s2nf, y, 
	nx, nl, nh)
real *xcen, *fwhm, *stren;
integer *nlc, *nhc;
real *cont0, *s2nf, *y;
integer *nx, *nl, *nh;
{
    /* Initialized data */

    static real fourl2 = (float)2.7726;
    static real sfa = (float)2.44141e-4;
    static struct {
	real e_1;
	integer fill_2[2];
	} equiv_17 = { (float)1.0, {0,0} };


    /* System generated locals */
    integer i__1;
    static real equiv_2[3], equiv_11[9], equiv_14[3];

    /* Builtin functions */
    double sqrt(), log(), exp();

    /* Local variables */
    static real dcdx, c1on2, sneb, snfd, ybar, cont, ymax, test;
    static integer npts;
    static real c1den;
    static integer nhcm1;
    static real d2max;
    static integer nlcp1;
#define a (equiv_11)
#define c_ (equiv_14)
    static integer i, j;
    static real c2neg2;
#define v (equiv_2)
#define x ((real *)&equiv_17)
    static real d2ydx2, slope;
#define c0 (equiv_14)
#define c1 (equiv_14 + 1)
#define c2 (equiv_14 + 2)
    static real fwhms;
#define v0 (equiv_2)
#define v1 (equiv_2 + 1)
#define v2 (equiv_2 + 2)
#define x0 ((real *)&equiv_17)
#define x1 ((real *)&equiv_17 + 1)
#define x2 ((real *)&equiv_17 + 2)
    static real y0, d2maxc;
#define a00 (equiv_11)
#define a01 (equiv_11 + 3)
#define a02 (equiv_11 + 6)
#define a10 (equiv_11 + 1)
#define a11 (equiv_11 + 4)
#define a12 (equiv_11 + 7)
#define a20 (equiv_11 + 2)
#define a21 (equiv_11 + 5)
#define a22 (equiv_11 + 8)
    static real c0test, c1test;
    static integer n10, nc, n12, n14, n24, n16, n18;
    static real v00;
    static integer n26;
    static real dx, fx, snebfd, wx, weight, sumfit, sumerr, yxsonc, ym1, yp1, 
	    det;
    static integer n14l, n18h;
    static real sna;
    static integer n18l;
    static real ych, snb, snc, snd, ycl, sne, snf;
    static integer n14x;
    static real efx;
    static integer nhw, n18x;
    static real sig;
    static integer nlw;
    static real err, sum, yxs;
    static integer ncp1;
    static real a00a11, a01a10, a00a22, a01a12, a01a22, a02a11, a02a20, 
	    a02a21, a10a22, a11a22, a12a20, a12a21;


/*    F I G _ G F I T  -  LINE CENTRE DETERMINATION BY PSEUDO-GAUSSIAN FIT
TING*/


/*   ARGUMENTS : */

/*     XCEN     D  F  CENTRE POSITION OF LINE RELATIVE TO CENTRE OF FIRST 
*/
/*                       ELEMENT IN ARRAY Y. */

/*     FWHM     S  F  ZERO = FIT FULL-WIDTH HALF-MAXIMUM */
/*                    NON-ZERO = CONSTRAINED FULL-WIDTH HALF-MAXMIMUM */
/*              D  F  FITTED FULL-WIDTH HALF-MAXIMUM IF ZERO SOURCE */

/*     STREN    D  F  INTEGRATED STRENGTH OF LINE ALLOWING FOR CONTINUUM. 
*/

/*     NLC      D  I  LOW INDEX LIMIT OF REGION FOUND TO CONTAIN LINE. */

/*     NHC      D  I  HIGH INDEX LIMIT OF REGION FOUND TO CONTAIN LINE. */

/*     CONT0    D  F  CONTINUUM LEVEL AT BIN CONTAINING MAXIMUM VALUE */
/*                       IN THE RANGE INDEX NL TO INDEX NH. */

/*     S2NF     D  F  SIGNAL-TO-NOISE FIGURE  +1.0 = PURE SIGNAL */
/*                                            -1.0 = PURE NOISE */


/*     Y        S  F  SOURCE ARRAY IN WHICH LINE IS TO BE FOUND. */

/*     NX       S  I  NUMBER OF ELEMENTS IN ARRAY Y */

/*     NL       S  I  LOW INDEX LIMIT OF REGION IN ARRAY Y IN WHICH */
/*                       THE LINE IS TO BE FOUND. */

/*                       FOUND. */
/*     NH       S  I  HIGH INDEX LIMIT OF REGION IN ARRAY Y IN WHICH */
/*                       THE LINE IS TO BE FOUND. */


/*   THIS ROUTINE FITS A QUADRATIC POLYNOMIAL BY SOLVING THE */
/*   'NORMAL EQUATIONS' */

/*                 V = A.C */

/*   WHERE */

/*        V(J) IS THE SUM, OVER THE FITTED DATA, OF W(M)*Y(M)*X(M)**(J-1) 
*/
/*        A(J,K) IS THE SUM, OVER THE FITTED DATA, OF W(M)*X(M)**(J+K-1) 
*/
/*        W(M) IS THE WEIGHT GIVEN TO THE DATA POINT Y(M) */
/*        C(K) IS THE KTH COEFFICIENT OF THE FITTED POLYNOMIAL */

/*   THIS FIT IS DONE TO THE LOGARITHM OF THE EXCESS OF THE DATA OVER AN 
*/
/*   ESTIMATED CONTINUUM AND THE WEIGHTING IS CHOSEN SO AS TO MAKE THIS */
/*   EQUIVALENT TO FITTING A GAUSSIAN TO THE ORIGINAL DATA. */

/*   HISTORY - */

/*   THIS IS JOHN STRAEDE'S SDRSYS ROUTINE GFIT, MODIFIED FOR USE ON */
/*   THE VAX, PARTICULARLY FOR FIGARO, BY KS.  24 FEB 88. */
/*   THE FOLLOWING CHANGES WERE MADE: */

/*   A) ROUTINE RE-NAMED TO FIG_GFIT */
/*   B) ALL INTEGERS NOW INTEGER*4 (VAX DEFAULT) */
/*   C) INTERDATA SPECIFIC COMPILER FLAGS ($TITL ETC) REMOVED */
/*   D) PARAMETER NX ADDED, SO NO LONGER ASSUMES LENGTH OF NX ELEMENTS */
/*   E) ALL REFERENCES TO 2048, 2050 ETC CHANGED TO NX, NX+2 ETC */

/*   G F I T  -  ARRAYS, CONSTANTS */

/*   ARGUMENT ARRAY */

/*   EQUATION VALUES VECTOR */

/*   COEFFICIENTS OF SIMULTANEOUS EQUATIONS */

/*   COEFFICIENTS VECTOR (SOLUTION TO SIMULTANEOUS EQUATIONS) */

/*   POWERS OF X */



/*    0 */
/*   X */
    /* Parameter adjustments */
    --y;

    /* Function Body */

/*   4.LN(2) */


/*   G F I T  -  FIND SMOOTHED PEAK */

/*   RESET STRENGTH AND SIGNAL-TO-NOISE-FIGURE */
    *stren = (float)0.;
    *s2nf = (float)0.;

/*   CONSTRAINED FULL-WIDTH HALF-MAXIMUM */
    fwhms = *fwhm * *fwhm;
    *c2 = (float)0.;
    if (fwhms > (float)0.) {
	*c2 = -(doublereal)fourl2 / fwhms;
    }

/*   FIND MAXIMUM FOR LINE CENTRE */
    ymax = (float)0.;
    y0 = (float)0.;
    if (*nl > 1) {
	y0 = y[*nl - 1];
    }
    yp1 = y[*nl];

    i__1 = *nh;
    for (n10 = *nl; n10 <= i__1; ++n10) {

	ym1 = y0;
	y0 = yp1;
	yp1 = (float)0.;
	if (n10 < *nx) {
	    yp1 = y[n10 + 1];
	}
	ybar = (ym1 + y0 * (float)2. + yp1) / (float)4.;
	if (ybar < ymax) {
	    goto L10;
	}
/*                                       NOT MAXIMUM */
	ymax = ybar;
	nc = n10;
L10:
	;
    }

    *nhc = (*nl + *nh + 1) / 2;
    if (nc == *nh) {
	goto L30;
    }
/*                                NO LINE OR LINE AT RED END OF REGION */
    if (ymax <= (float)0.) {
	goto L30;
    }
/*                                   NO POSITIVE MAXIMUM */
/*   CHECK FOR LOCAL PEAK */
    if (y[nc - 1] >= y[nc] && y[nc] <= y[nc + 1]) {
	goto L30;
    }
/*                                                                NO PEAK 
*/
/*   G F I T  -  FIND LINE BOUNDARIES */

/*   RED BOUNDARY */
    *nhc = nc;
    ncp1 = nc + 1;

    i__1 = *nx - 3;
    for (n12 = ncp1; n12 <= i__1; ++n12) {

	slope = y[n12 - 1] + (y[n12] - y[n12 + 1]) * (float)2. - y[n12 + 2];
	if (slope < (float)0.) {
	    goto L121;
	}
/*                                        RED MINIMUM FOUND */
	*nhc = n12;
/* L12: */
    }

L121:
    if (*nhc == nc) {
	goto L30;
    }

/*   BLUE BOUNDARY */
    *nlc = nc;
    n14l = *nx + 2 - nc;

    i__1 = *nx - 3;
    for (n14 = n14l; n14 <= i__1; ++n14) {

	n14x = *nx + 1 - n14;
	slope = y[n14x + 1] + (y[n14x] - y[n14x - 1]) * (float)2. - y[n14x - 
		2];
	if (slope < (float)0.) {
	    goto L141;
	}
/*                                        BLUE MINIMUM FOUND */
	*nlc = n14x;
/* L14: */
    }

L141:
    if (*nlc == nc) {
	goto L30;
    }

/*   G F I T  -  GAUSSIAN CORE LIMITS AND CONTINUUM */

/*   FIND LIMITS OF GAUSSIAN CORE */
/*   BY MAXIMUMS OF SECOND DERIVATIVES */

    d2max = y[nc - 2] + y[nc - 1] - y[nc] * (float)4. + y[nc + 1] + y[nc + 2];
    d2maxc = d2max;
    ncp1 = nc + 1;
    nhw = ncp1;
    if (ncp1 > *nhc) {
	goto L161;
    }

    i__1 = *nhc;
    for (n16 = ncp1; n16 <= i__1; ++n16) {

	d2ydx2 = y[n16 - 2] + y[n16 - 1] - y[n16] * (float)4. + y[n16 + 1] + 
		y[n16 + 2];
	if (d2ydx2 <= d2max) {
	    goto L16;
	}

	nhw = n16;
	d2max = d2ydx2;

L16:
	;
    }

L161:
    d2max = d2maxc;
    nlw = nc - 1;
    n18l = *nx + 2 - nc;
    n18h = *nx + 1 - *nlc;
    if (n18l >= n18h) {
	goto L181;
    }

    i__1 = n18h;
    for (n18 = n18l; n18 <= i__1; ++n18) {

	n18x = *nx + 1 - n18;
	d2ydx2 = y[n18x + 2] + y[n18x + 1] - y[n18x] * (float)4. + y[n18x - 1]
		 + y[n18x - 2];
	if (d2ydx2 <= d2max) {
	    goto L18;
	}

	nlw = n18x;
	d2max = d2ydx2;

L18:
	;
    }

L181:

/*   CONTINUUM */
    ych = (y[*nhc - 1] + y[*nhc] * (float)3. + y[*nhc + 1]) / (float)5.;
    ycl = (y[*nlc - 1] + y[*nlc] * (float)3. + y[*nlc + 1]) / (float)5.;
    dx = (real) (*nhc - *nlc);
    dcdx = (ych - ycl) / dx;
    *cont0 = ycl + (real) (nc - *nlc) * dcdx;

/*   G F I T  -  CONSTRUCT NORMAL EQUATIONS */

/*   RESET (CLEAR) NORMAL EQUATIONS */
    for (i = 1; i <= 3; ++i) {
	v[i - 1] = (float)0.;
	for (j = 1; j <= 3; ++j) {
	    a[i + j * 3 - 4] = (float)0.;
/* L20: */
	}
    }

/*   INTEGRATE STRENGTH ABOVE CONTINUUM AND ACCUMULATE NORMAL EQUATIONS */
    nlcp1 = *nlc + 1;
    nhcm1 = *nhc - 1;
    npts = 0;
    sum = (float)0.;

    i__1 = nhcm1;
    for (n24 = nlcp1; n24 <= i__1; ++n24) {

	*x1 = (real) (n24 - nc);
	cont = *cont0 + dcdx * *x1;
	yxs = y[n24] - cont;
	sum += yxs;
	if (y[n24] <= (float)0.) {
	    goto L24;
	}
/*                                        NULL/NEGATIVE DATA */
	if (n24 < nlw || n24 > nhw) {
	    goto L24;
	}
/*                                                         OUTSIDE COR
E */
	yxsonc = yxs / (dabs(yxs) + dabs(cont));
	if (yxsonc < sfa) {
	    goto L24;
	}
/*                                        TERRIBLE DATA */
	weight = yxs / sqrt(y[n24]);
	yxs = log(yxs);
	*x2 = *x1 * *x1;
	yxs -= *c2 * *x2;

	for (i = 1; i <= 3; ++i) {

	    wx = weight * x[i - 1];
	    v[i - 1] += wx * yxs;

	    for (j = 1; j <= 3; ++j) {

		a[i + j * 3 - 4] += wx * x[j - 1];
/* L22: */
	    }
	}

	++npts;

L24:
	;
    }

/*   LINE INTEGRATED STRENGTH */
    *stren = sum;

/*   G F I T  -  SOLVE NORMAL EQUATIONS */

    if (*c2 < (float)0.) {
	goto L241;
    }
/*                                  FULL-WIDTH HALFMAXIMUM CONSTRAINED */
/*   FIT FULL-WIDTH HALFMAXIMUM */
    if (npts < 3) {
	goto L30;
    }
/*                                 INSUFFICIENT DATA */
/*   SOLVE NORMAL EQUATIONS */
    a00a22 = *a00 * *a22;
    a01a12 = *a01 * *a12;
    a01a22 = *a01 * *a22;
    a02a11 = *a02 * *a11;
    a02a20 = *a02 * *a20;
    a02a21 = *a02 * *a21;
    a10a22 = *a10 * *a22;
    a11a22 = *a11 * *a22;
    a12a20 = *a12 * *a20;
    a12a21 = *a12 * *a21;
    sna = *a22 * *v0 - *a02 * *v2;
    snb = a11a22 - a12a21;
    snc = *a22 * *v1 - *a12 * *v2;
    snd = a01a22 - a02a21;
    sne = a00a22 - a02a20;
    snf = a10a22 - a12a20;

    sneb = sne * snb;
    snfd = snf * snd;
    snebfd = sneb - snfd;
    c0test = dabs(snebfd) / (dabs(sneb) + dabs(snfd));
    if (c0test < sfa) {
	goto L30;
    }
/*                                     ILL CONDITIONED EQUATIONS */
    *c0 = (sna * snb - snc * snd) / snebfd;

    c1den = a01a12 - a02a11;
    c1test = dabs(c1den) / (dabs(a01a12) + dabs(a02a11));
    if (c1test < sfa) {
	goto L30;
    }
/*                                     ILL CONDITIONED EQUATIONS */
    v00 = *v0 - *a00 * *c0;
    *c1 = (*a12 * v00 - *a02 * (*v1 - *a10 * *c0)) / c1den;
    *c2 = (v00 - *a01 * *c1) / *a02;
    if (*c2 >= (float)0.) {
	goto L30;
    }
/*                                 NOT AN EMISSION LINE ! */
/*   LINE FULLWIDTH-HALFMAXIMUM */
    *fwhm = sqrt(-(doublereal)fourl2 / *c2);
    goto L259;

/*   G F I T  -  LINE CENTRE DETERMINATION */

/*   FULL-WIDTH HALF-MAXIMUM CONSTRAINED */
L241:
    if (npts < 2) {
	goto L30;
    }
/*                                 INSUFFICIENT DATA */
/*   SOLVE FOR TWO COEFFICIENTS */
    a00a11 = *a00 * *a11;
    a01a10 = *a01 * *a10;
    det = a00a11 - a01a10;
    test = dabs(det) / (dabs(a00a11) + dabs(a01a10));
    if (test < sfa) {
	goto L30;
    }
/*                                   ILL CONDITIONED EQUATIONS */
    *c0 = (*a11 * *v0 - *a01 * *v1) / det;
    *c1 = (*a00 * *v1 - *a10 * *v0) / det;

/*   LINE CENTRE POSITION */
L259:
    c1on2 = *c1 / (float)2.;
    c2neg2 = -(doublereal)(*c2) * (float)2.;
    if (c2neg2 < dabs(c1on2)) {
	goto L30;
    }
/*                                            MORE THAN 2 BINS FROM PEAK 
*/
    *xcen = (real) (nc - 1) - c1on2 / *c2;

/*   GOODNESS OF FIT */
    sumerr = (float)0.;
    sumfit = (float)0.;

    i__1 = nhw;
    for (n26 = nlw; n26 <= i__1; ++n26) {

	*x1 = (real) (n26 - nc);
	*x2 = *x1 * *x1;
	fx = *c0 + *c1 * *x1 + *c2 * *x2;
	efx = exp(fx);
	cont = *cont0 + dcdx * *x1;
	sig = y[n26] - cont;
	if (sig <= (float)0.) {
	    goto L26;
	}
/*                                     IGNORE NON-POSITIVE 'SIGNAL' */
	err = sqrt(efx) - sqrt(sig);
	sumerr += err * err;
	sumfit += efx;
L26:
	;
    }

    if (sumfit <= (float)0.) {
	goto L30;
    }
    *s2nf = (sumfit - sumerr) / (sumfit + sumerr);

L30:
    return 0;

} /* fig_gfit__ */

#undef a22
#undef a21
#undef a20
#undef a12
#undef a11
#undef a10
#undef a02
#undef a01
#undef a00
#undef x2
#undef x1
#undef x0
#undef v2
#undef v1
#undef v0
#undef c2
#undef c1
#undef c0
#undef x
#undef v
#undef c_
#undef a



