/*******************************************************************************
* E.S.O. - VLT project
*
* "@(#) $Id: ACMM:ozpozFpcol/ws/src/fpcolMdlCalCentGauss2d.c,v 1.63 07-Oct-2008 11:10:27+10 tjf $"
*
*
* who       when      what
* --------  --------  ----------------------------------------------
* ssmedley  07/09/01  created from AAO GCamCentroid2dGauss.
*
*/


/*
 *+			G C a m C e n t r o i d 2 d G a u s s

 *  Function name:
      GCamCentroid2dGauss

 *  Function:
      Centroid an image within a subset of the full image area using 
      two dimensional Gaussian fitting.

 *  Description:
      This routine takes an image and obtains gets the center of the strongest
      star image in it by locating the peak in a smoothed version of the 
      image and using this as the starting point for a 2d gaussian fit.

 *  Language:
      C

 *  Call:
      (Void) = GCamCentroid2dGauss (Window,data,bias,results,status)

 *  Parameters:   (">" input, "!" modified, "W" workspace, "<" output)
      (>) Window	(GCamWindowType) Window definition.  Note this is the
			definition of the window to be centroided within the
			data area.  The MaxX and MaxY elements should indicate
			the size of the data area, where as Xoffset,Yoffset,
			Xdim and Ydim indicates the area to be centroided.
			PixelSize indicates the size of pixels.  This is
			DIFFERENT from the window definiton used to obtain
			the image (say with GCamImage())
      (>) data		(Void *) The data area.  Should be an unsigned 
			integer array corresponding to the pixel size.
      (>) bias		(int) Ignored.
      (<) results	(double {[2]}) The results of the centroiding.  
			This value is within the centroiding window, not
			the original window.  The value within the original
			window is 
			results[0]+Window->Xoffset,results[1]+Window->Yoffset.
			results[2] is set to the full with half maximum.
      (!) status	(Int *) Modified status. Possible failure codes are
			GCAM__CENT_FITERR => Some error in fitting.
			


 *  Include files: GCam.h

 *  External functions used:
      malloc	(CRTL)	Allocate memory
      ErsRep	(Ers)	Report an error

 *  External values used:
      none

 *  Prior requirements:
      none  (Indepentent of GCamActivate)

 *  Support: Tony Farrell, AAO

 *  Bugs and limitations:
      There is no check on whether the successive fits actually converge
      or not. It might be nice to monitor the convergence, looking at the
      successive changes in X and Y positions, in order to a) quit the
      fit loop once the positions are settled to a fraction of a pixel,
      b) to indicate a problem if they do not converge, possibly falling back
      on the peak values determined earlier.

 *-
 * Sccs Id:     gcamcentg2d.c, Release 1.4, 05/15/98
  
 * Copyright (c) Anglo-Australian Telescope Board, 1995.
   Not to be used for commercial purposes without AATB permission.
  
 *  History:
      19-Jun-1994 - TJF - Original version
       9-Jun-1995 - KS  - 2d Gaussian version, produced by combining the
                          code for GCamCentroid with a 2d gaussian fitting
                          algorithm written by JOS.
      18-Aug-1995 - TJF - Use GCAM_CENT_FITERR code when we can't fit.
      21-Aug-1995 - TJF - Moved this routine from the module gcamcentroid.c
      				module to gcamcentg2d.c
      12-Aug-1996 - KS/ - Substantial modifications to gauss fitting routine,
                    JOS   and to the way it it called, to improve the
                          accuracy by improving the discrimination and
                          the initial estimate of peak width.
      13-Aug-1996 - KS  - Added definitions of TRUE and FALSE.
      15-Dec-1997 - KS/ - Modified algorithm used to try to reduce effects
                    JOS   of a noisy background. In particular, single pixel
                          spikes in the background should now be completely
                          ignored.
 */
/*
 *  Include files
 */
#include <stdio.h>
#include <assert.h>
#include "fpcolMdlCalCentGauss.h"

#ifdef VMS
#   include <unixlib.h>
#elif defined(VxWorks)
#   include <memLib.h>
#else
#   include <stdlib.h>
#endif

static void GCamPeakF (float *Array, int Nx, int Ny, int Npix, int *XCent,
   int *YCent, int *Status);

static void GCamWidthF (float *Array, int Nx, int Ny, int XCent, int YCent,
              double Background, double *Fwhm, int *Status);

static int GCamGaussfit (float *image, int rows, int cols, double aspectratio,
              int row, int col, double range, double *xcen, double *ycen,
              double *peak, double *fwhm, double *background, double *validity,
              int *Pixels, double *Noise);

static double GCamBackground (double *w, int nc, int nr, int c0, int r0,
                                                           int *c, int*r) ;

#ifndef TRUE
#define TRUE 1
#endif
#ifndef FALSE
#define FALSE 0
#endif

/*  These five lines are to make it easier to move code between this routine
 *  and the Figaro routine gen_gauss2d.c, which are almost identical, but
 *  since they operate in slightly different environments have some subtle
 *  differences - one is that Figaro code always starts variable names in
 *  upper case, hence the use of 'Results', and they have different error
 *  codes. Also the Figaro code takes offset parameters, which this code 
 *  does not.
 */

#define Results results
#define Status status
#define ERROR_CODE GCAM__CENT_FITERR
#define Xoff 0
#define Yoff 0

extern void GCamCentroid2dGauss (GCamWindowType *Window,
			  void  *data,
			  int bias,
			  double *results,
			  StatusType *status)
{

    float *Array;              /* Address of float version of image */
    float *ArrayLinePtr;       /* Addresses along a line of 'Array' */
    double Aspect;             /* Aspect ration of data - assumed square */
    double Backg;              /* Estimated background for gaussian */
    int CrudeFit;              /* True if a crude fit was made */
    double CrudeFwhm;          /* Approximate fullwidth value */
    double CrudeXCent;         /* Approximate X center value */
    double CrudeYCent;         /* Approximate Y center value */
    double Fwhm;               /* Full width at half max of profile - ignored */
    int IFit;                  /* Loop index through fit refinement */
    int IStat;                 /* Status of 2d fit */
    double Noise;              /* Noise estimated by fit - ignored */
    int NXPeak;                /* Integer position of peak in X (from 0) */
    int NYPeak;                /* Integer position of peak in Y (from 0) */
    int Offset;                /* Offset (in elements) into the data array */
    double OldXCent;           /* Previous position of peak in X */
    double OldYCent;           /* Previous position of peak in Y */
    double Peak;               /* Esitimated peak value for gaussian */
    int Pixels;                /* Number of pixels used in fit */
    double Range;              /* Range over which 2d fits is to be made */
    double Valid;              /* Validity estimate for fit - ignored */
    int X;                     /* Loop index in X through data */
    double XCent;              /* Position of peak in X (from 0) */
    double XDiff;              /* Difference between crude and final fit in X */
    int Xdim;                  /* Window dimensions in X */
    int Y;                     /* Loop index in Y through data */
    double YCent;              /* Position of peak in Y (from 0) */
    double YDiff;              /* Difference between crude and final fit in X */
    int Ydim;                  /* Window dimensions in Y */

    (void) bias;
    if (*status != GCAM__OK) return;

    results[2] = -1;	/* Default value */


    Array = (float *)malloc (Window->Xdim * Window->Ydim * sizeof(float));
    if (!Array) {
	*status = GCAM__MALLOCERR;
	return;
    }

    /*
     *	Copy the data within the window into the floating point array
     *   that the fitting routines will use.
     */
    
    for (Y = 0 ; Y < Window->Ydim ; Y++) {

       /*
        *  Offset is the total offset in the data array to the next element.
        *  It is made up of the y line number (y+Window->Yoffset) times the
        *  x dimension. ArrayLinePtr is the address of the line in question in
        *  the target floating point array.
        */

       Offset = (Y+Window->Yoffset)*Window->MaxX;
       ArrayLinePtr = Array + Y * Window->Xdim;

       /*  Allowing for the pixel size of the data, copy the data for this
        *  line into the target array.
        */

		if (Window->PixelSize == 8)
		{
			for (X = 0 ;X < Window->Xdim ; X++)
			{
				*ArrayLinePtr++ = (float)(((double *)(data))
				[X+Window->Xoffset+Offset]);
			}
		}
		else if (Window->PixelSize == 4) {
	  for (X = 0 ;X < Window->Xdim ; X++) {
             *ArrayLinePtr++ = (float)(((unsigned long *)(data))
				[X+Window->Xoffset+Offset]);

          }
       } else if (Window->PixelSize == 2) {
	  for (X = 0 ;X < Window->Xdim ; X++) {
	      *ArrayLinePtr++ = (float)(((unsigned short *)(data))
				[X+Window->Xoffset+Offset]);

          }
       } else {  /* Assume Window->PixelSize == 1 */
	  for (X = 0 ;X < Window->Xdim ; X++) {
	      *ArrayLinePtr++ = (float)(((unsigned char *)(data))
				[X+Window->Xoffset+Offset]);

          }
       }
   }

   /*  Set the window dimensions in Xdim, Ydim. From now on, the code
    *  should be just the same as that for the Figaro routine gen_gauss2d.
    *  (This is why, for example, we have ERROR_CODE defined rather than
    *  just using the GCAM code explicitly.)
    */

   Xdim = Window->Xdim;
   Ydim = Window->Ydim;

   /*  Now find the position of the peak in a smoothed version of the
    *  image.
    */

   GCamPeakF (Array,Xdim,Ydim,5,&NXPeak,&NYPeak,Status);

   /*  Using this as a starting point, call the 2d gaussian fitter
    *  to get a precise fit to the data. Experience shows that the
    *  fitter works best with only a limited subset of the data
    *  to look at, making the Range parameter important. The routine
    *  GCamWidthF() can produce a reasonable estimate of the range, 
    *  but it needs a good estimate of the background to work with.
    *  What we do, therefore, is call GCamGaussfit() once mainly 
    *  to get a good estimate of the background. We then use 
    *  GCamWidthF() to estimate the fullwidth of the peak, given this
    *  background value. We then call GCamGaussfit() again, this
    *  time with a much restricted range based on the estimated
    *  fullwidth. The background estimate made on the first pass is
    *  probably going to be better than one made with a more restricted
    *  range, so we leave that unchanged in the second call. We could
    *  continue to call GCamGaussfit() to hone the results, but
    *  experience seems to indicate that this is not necessary. The
    *  sequence is coded as a loop, however, to allow for this - all
    *  that would be needed is to change the loop limit from 2 to a 
    *  higher value.
    */

   Aspect = 1.0;
   Range = (Xdim > Ydim) ? Xdim / 2 : Ydim / 2;
   XCent = NXPeak;
   YCent = NYPeak;
   Backg = -1.0;
   CrudeFit = FALSE;
   for (IFit = 0; IFit < 2; IFit++) {
      OldXCent = XCent;
      OldYCent = YCent;
      if (Range > NXPeak - 2) Range = NXPeak - 2;
      if (Range > NYPeak - 2) Range = NYPeak - 2;
      if (Range > Xdim - NXPeak - 2) Range = Xdim - NXPeak - 2;
      if (Range > Ydim - NYPeak - 2) Range = Ydim - NYPeak - 2;
      Pixels = -1; Noise = -1.0;
      IStat = GCamGaussfit (Array,Ydim,Xdim,Aspect,NYPeak,NXPeak,Range,
                &XCent,&YCent,&Peak,&Fwhm,&Backg,&Valid,&Pixels,&Noise);
	printf("Backg=%f\n", Backg);
      /*  On the first pass through, we can live with just getting a background
       *  estimate, even if for some reason GCamGaussfit fails. Error code 3
       *  indicates a failure to fit a proper gaussian, but the routine will
       *  still return sensible values for the fit.
       */

      if (IStat == 3) IStat = 0;
      if (IStat != 0) {
      	if ((IFit == 0) && (Backg > 0)) IStat = 0;
      }
      if (IStat != 0) {
         *Status = ERROR_CODE;
         XCent = OldXCent; YCent = OldYCent;
         goto Exit;
      }
      if (IFit == 0) {
         GCamWidthF (Array,Xdim,Ydim,NXPeak,NYPeak,Backg,&Fwhm,Status);
	 printf("Fwhm=%f\n", Fwhm);
	 if (*Status == 0) {
            CrudeFit = TRUE;
            CrudeXCent = NXPeak;
            CrudeYCent = NYPeak;
            CrudeFwhm = Fwhm;
        }
      } else {
         NXPeak = XCent + 0.5;
         NYPeak = YCent + 0.5;
      }
      Range = 2.5 * Fwhm;
   }

   /*  If we got a result, perform a brief sanity check. The final result
    *  should be reasonably close to the crude result we get from GCamPeakF()
    *  and GCamWidthF(). If it isn't, we've probably seen a perverse profile
    *  that the 2D gaussian algorithm has had trouble with. In this case, we
    *  could either choose to prefer the crude result, or we can report a
    *  failure to fit. For the moment, we do the former.
    */

   if (CrudeFit) {
      XDiff = CrudeXCent - XCent;
      if (XDiff < 0.0) XDiff = -XDiff;
      YDiff = CrudeYCent - YCent;
      if (YDiff < 0.0) YDiff = -YDiff;
      if ((XDiff > CrudeFwhm) || (YDiff > CrudeFwhm)) {
         XCent = CrudeXCent;
         YCent = CrudeYCent;
         Fwhm = CrudeFwhm;
      }
   }


   /*  Return the results */

Exit:;

   Results[0] = XCent + (float) Xoff;
   Results[1] = YCent + (float) Yoff;
   Results[2] = Fwhm;

   if (Array != 0) (void) free(Array);
}



/* -------------------------------------------------------------------- */
/*                                                                      */
/*                     G C a m  G a u s s f i t                         */
/*                                                                      */
/*  Fast gaussian profile fit.						*/
/*									*/
/*  This routine attempts to fit a 2-D gaussian profile to a section of	*/
/*  an image.  It does this by mathematical trickery described below:	*/
/*									*/
/*  1.  An estimate of the background is made.  The accuracy critically	*/
/*	depends upon this.  This estimate is subtracted from the data.	*/
/*  2.  The logarithm is taken.  This converts a gaussian profile into	*/
/*	a paraboloid.							*/
/*  3.  A 2-D quadratic is fitted using a weighting of			*/
/*									*/
/*			    Z - background				*/
/*									*/
/*      to the logarithms.  This is equivalent to using equal weighting */
/*      if we were fitting a gaussian directly to the original data.	*/
/*  4.  The gaussian is now recovered from the quadratic fit.		*/
/*									*/
/*  Note that fwhm, background, and noise are input/output arguments.	*/
/*  If these are passed as zero or negative values, the routine will	*/
/*  produce estimates of these for itself. If positive values are 	*/
/*  passed, these will be used by the routine.				*/
/*									*/
/*  									*/
/*									*/
/*  Restrictions:							*/
/*									*/
/*  The choice of initial position (row,col) should be such that an	*/
/*  ellipse of x-radius range and y-radius range/aspectratio should fit */
/*  entirely within the data image.					*/
/*									*/
/*  If there are multiple emission features in the data, this routine	*/
/*  relies on being passed a reasonably accurate value for the Range	*/
/*  parameter in order to discriminate against the other features. It	*/
/*  is a feature of this algorithm that its fit is thrown quite badly	*/
/*  if the thing it fits is not a single peak.	(Comment added by KS.)	*/
/*									*/
/*  This routine will work only for an emission feature.		*/
/*									*/
/*									*/
/*  John Straede.							*/
/*  AAO.  July,  1996.							*/
/*									*/
/*  History: 								*/
/*  24th July 1996. Original version released May, 1995. Modifications	*/
/*		    by KS and TJF incorporated.  These included changes	*/
/*		    to return codes, removing check on free, declaring 	*/
/*		    slaDmat as static, removing printf,including an	*/
/*		    omitted line of code defining rr in a loop.		*/
/*  30th July 1996. Now only calculates background and noise if these   */
/*                  are passed with zero or -ve values.                 */

#include <math.h>


/* Routine to solve simultaneous equations.				*/
   static void
       slaDmat (int n, double *A, double *V, double *det, int *f, int *w) ;



static int GCamGaussfit
	     (float *image,		/* 2-D array containing image	*/
	      int rows, int cols,	/* Size of image		*/
	      double aspectratio,	/* y/x shape of pixel		*/

	      int row, int col,		/* Expected centre of gaussian	*/
	      double range,		/* radius of fit, x pixels	*/

	      double *xcen, double *ycen, /* Fitted centre		*/
	      double *peak,		/* Fitted peak above background */
	      double *fwhm,		/* Fitted fullwidth-halfmaximum */
	      double *background,	/* Estimated background		*/
	      double *validity,		/* Estimated validity of result */
	      int *Pixels,		/* Number of pixels in fit	*/
	      double *Noise)		/* Background noise (RMS)	*/

/*  Returned:	0 if a successful fit was obtained, otherwise a number	*/
/*		indicating an unrecoverable error condition.		*/
/*									*/
/*		1 = expected centre not within image			*/
/*		2 = malloc failed to get workspace			*/
/*		3 = couldn't solve the normal equations			*/
/*		4 = error in sample size				*/
/*		5 = insufficient range (< 9 pixels)			*/
/*									*/


{
   int i, j ;		/* indices					*/
   int wp ;		/* index into workspace				*/
   double *w ;		/* workspaces into which portion of image is	*/
			/* copied.  Also used for fitting weights.	*/
   double xsize, ysize; /* extents of examined area			*/
   int nr, nc ;		/* rows & columns in workspace			*/
   int mrp, mcp ;	/* high row & column extents of examined area	*/
   int mcm, mrm ;	/* low row & column extents of examined area	*/
   int nw ;		/* number of pixels in examined area		*/
   double A[4][4], V[4] ; /* Used to define simutaneous eqns V = Ax	*/
   double ltb, rtb, lbb, rbb ;  /* Background estimates of 4 corners of	*/
				/* examined area.			*/
   double lb, rb ;		/* Background estimates at ends of row	*/
   double x, xx, y, yy ;	/* Coordinates and squares		*/
   double rr ;			/* radius squared from current centre	*/
   double RR ;			/* critical radius (range) squared	*/
   double noise, noise2 ;	/* used to determine fit validity	*/
   double elements ;		/* pixels in noise estimation		*/
   int work[4] ;		/* work space for equation solving	*/
   double det ;			/* determinant				*/
   int flag ;			/* 0 = simultaneous equations solved	*/
   double drb ;			/* background slope along row		*/
   double s, ss ;		/* sums of fit error & error squared	*/
   double e ;			/* Laplacian of 1 pixel used for noise	*/
   double v ;			/* weight for fit			*/
   int xlt,ylt, xrt,yrt, xlb,ylb, xrb,yrb ; /* low pixel coordinates	*/
					    /* near the four corners	*/
   double sxt, sxb, syl, syr ;	/* background slopes along edges of	*/
				/* work area.				*/
   double W, sW ;		/* weight, sum of weights for noise	*/
   int PixelsFitted ;		/* number of pixels in Gaussian fit	*/
   int pixels ;			/* previous value of PixelsFitted	*/
   double circ ;		/* sum of four nearest pixels		*/
   double discrim ;		/* discrimination level			*/
   double x0, y0 ;		/* peak centre relative to work array	*/
   double cb ;                  /* interpolated background at centre    */
   double halfmax ;             /* half-maximum                         */
   double sum, dsum ;           /* sum and contribution for FWHM        */
   double dzdx ;                /* slope in x direction                 */
   double dzdy ;                /* slope in y direction                 */
   double excess ;              /* excess over half-maximum             */
   double temp ;                /* temporary storage for swap           */
   double noiseX4 ;             /* noise times 4 for discrimination     */
   double xsum, ysum ;          /* moment sums for centroid             */

/* Check for adequate range - must have at least 25 pixels		*/

   if (range < 2.0)   return (5) ;

/* Check that centre is within image.					*/

   if (row >= rows  ||  row < 0  ||  col >= cols  ||  col < 0)
   {
      return(1) ;
   }

   RR = range*range ;

/* Determine size of relevant section of image.				*/

/* Number of columns.							*/

   xsize = range ;
   mcp = (int)(xsize + 0.5) ;
   mcm = mcp ;

   if (col + mcp >= cols)   mcp = cols - col - 1 ;
   if (col < mcm)           mcm = col ;
   nc = mcp + mcm + 1 ;

/* Number of rows.							*/

   ysize = range/aspectratio ;
   mrp = (int)(ysize + 0.5) ;
   mrm = mrp ;

   if (row + mrp >= rows)   mrp = rows - row - 1 ;
   if (row < mrm)           mrm = row ;
   nr = mrp + mrm + 1 ;

/* Number of pixels.							*/

   nw = nr*nc ;

/* Get work space for image and fitting weights.			*/

   w = (double *)malloc(3*sizeof(double)*nw) ;

   if (!w)
   {
      return (2) ;
   }

/* Transfer relevant part of image to workspace.			*/

   wp = 0 ;

   printf("%dx%d ========================\n", nc, nr);
   for (i = row - mrm ;  i <= row + mrp ;  i++)
   {
      for (j = col - mcm ;  j <= col + mcp ;  j++)
      {
	 w[wp] = image[i*cols+j] ;
	 printf("%7.5f ", w[wp]);
	 wp++ ;
      }
      printf("\n");
   }

/* Estimate background by finding the value and position of the lowest	*/
/* pixel value near each corner.					*/

/* First get corner values.						*/

   ltb = GCamBackground (w, nc, nr, 1, 1, &xlt, &ylt) ;
   rtb = GCamBackground (w, nc, nr, nc-2, 1, &xrt, &yrt) ;
   lbb = GCamBackground (w, nc, nr, 1, nr-2, &xlb, &ylb) ;
   rbb = GCamBackground (w, nc, nr, nc-2, nr-2, &xrb, &yrb) ;

/*  Estimate background slopes near edges.				*/

   if (xrt == xlt)
      sxt = 0.0 ;
   else
      sxt = (rtb - ltb)/((double)(xrt - xlt)) ;

   if (xrb == xlb)
      sxb = 0.0 ;
   else
      sxb = (rbb - lbb)/((double)(xrb - xlb)) ;

   if (ylb == ylt)
      syl = 0.0 ;
   else
      syl = (lbb - ltb)/((double)(ylb - ylt)) ;

   if (yrb == yrt)
      syr = 0.0 ;
   else
      syr = (rbb - rtb)/((double)(yrb - yrt)) ;

/* Extrapolate backgound to corners.					*/

   ltb -= sxt*(double)xlt + syl*(double)ylt ;
   rtb += sxt*(double)(nc - xrt) - syr*(double)yrt ;
   lbb -= sxb*(double)xlb - syl*(double)(nr - ylb) ;
   rbb += sxb*(double)(nc - xrb) + syr*(double)(nr - yrb) ;
   cb = 0.25*(ltb + rtb + lbb + rbb) ;

   if (*background <= 0.0)
      *background = cb ;
   else
   {
      cb -= *background ;
      ltb -= cb ;
      rtb -= cb ;
      lbb -= cb ;
      rbb -= cb ;
   }

/* Subtract background.							*/
/* The corner values define a "twisted plane".				*/

   wp = 0 ;

   if (mrp + mrm == 0  ||  mcp + mcm == 0)
   {
      free (w) ;
      return (4) ;
   }

   for (i = 0 ;  i < nr ;  i++)
   {
      y = (double)(i - mrm);
      yy = y/(double)(mcm + mcp) ;
      lb = ltb + (lbb - ltb)*yy ;
      rb = rtb + (rbb - rtb)*yy ;
      drb = (rb - lb)/(double)(mcm + mcp) ;

      y = aspectratio*y ;
      yy = y*y ;

      for (j = 0 ;  j < nc ;  j++)
      {
	 x = (double)(j - mcm) ;
	 rr = yy + x*x ;

	 if (rr <= RR)
	 {
	    w[wp] -= lb +  drb*x ;
	 }

         wp++ ;
      }
   }


/* Estimate the noise.  This is done by taking a weighted R.M.S. of	*/
/* the deviation of the Laplacian.  The weight is the square of the	*/
/* reciprocal of the smoothed value.  This biases the result towards	*/
/* the background.							*/

   s = 0.0 ;
   ss = 0.0 ;
   sW = 0.0 ;
   elements = 0.0 ;
   pixels = 0 ;
   wp = nc + 1 ;

   for (i = 1 ;  i < nr-1 ;  i++)
   {
      for (j = 1 ;  j < nc-1 ;  j++)
      {
	 if (w[wp] > 0.0)
	 {
	    circ = (4.*(w[wp-nc] + w[wp+nc] + w[wp-1] + w[wp+1]) +
		     w[wp-nc-1] + w[wp-nc+1] + w[wp+nc-1] + w[wp+nc+1])/20. ;
	    e = circ - w[wp] ;
	    W = circ ;
	    W *= W ;

	    if (W > 0.0)
	    {
	          s += e/W ;
	          ss += e*e/W ;
	          sW += 1./W ;
	          pixels++ ;
                  w[wp+2*nw] = e*e ;
	    }
	 }
         wp++ ;
      }
      wp += 2 ;
   }

   if (pixels <= 1)
      noise = 0.0 ;	/* Need at least 2 pixels for noise calculation. */
   else
   {
      noise = (ss - s*s/sW)/sW ;
      if (noise < 0.0)   noise = 0.0 ;
      noise = sqrt(noise) ;
   }

   if (*Noise <= 0.0)
      *Noise = noise ;
   else
      noise = *Noise ;

/* Compute weights for fit.						*/

   noise2 = noise*noise ;

   wp = 0 ;
   *peak = 0.0 ;

   for (i = 0 ;  i < nr ;  i++)
   {
      for (j = 0 ; j < nc ;  j++)
      {
	 if (*peak < w[wp])   *peak = w[wp] ;
	 wp++ ;
      }

   }

/* Count pixels "above" half-maximum allowing for partial contributions. */

   wp = 0 ;
   sum = 0.0 ;
   xsum = 0.0 ;
   ysum = 0.0 ;
   halfmax = *peak/2.0 ;

   for (i = 0 ;  i < nr ;  i++)
   {
      y = (double)(i - mrm) ;

      for (j = 0 ;  j < nc ;  j++)
      {
	 x = (double)(j - mcm) ;
         dsum = 0.0 ;

         if (i > 0  &&  i < nr - 1  &&  j > 0  &&  j < nc - 1)
         {
            /* Compute slopes                                         */

            dzdx = (w[wp+1] - w[wp-1])*0.5
                       + (w[wp+nc+1] - w[wp+nc-1]
                                        + w[wp-nc+1] - w[wp-nc-1])*0.25 ;
            dzdy = (w[wp-nc] - w[wp+nc])*0.5
                       + (w[wp-nc+1] - w[wp+nc+1]
                                        + w[wp-nc-1] - w[wp+nc-1])*0.25 ;

            /* We don't need the signs                                */

            dzdx = fabs(dzdx) ;
            dzdy = fabs(dzdy) ;

            /* We are not really interested in which slopes these
               represent but it would be convenient to have them in
               order of size.
               */

            if (dzdy > dzdx)
            {
               temp = dzdy ;
               dzdy = dzdx ;
               dzdx = temp ;
            }

            if (w[wp] - 0.5*(dzdx + dzdy) > halfmax)
               sum += 1.0 ;
            else if (w[wp] - 0.5*(dzdx - dzdy) > halfmax)
            {
               excess = w[wp] - 0.5*(dzdx + dzdy) ;
               dsum = 1.0 - excess*excess/(dzdx*dzdy) ;
               sum += dsum ;
            }
            else if (w[wp] + 0.5*(dzdx - dzdy) > halfmax)
            {
               excess = w[wp] - halfmax ;
               dsum = excess/dzdx + 0.5 ;
               sum += dsum ;
            }
            else if (w[wp] + 0.5*(dzdx + dzdy) > halfmax)
            {
               excess = w[wp] + 0.5*(dzdx + dzdy) - halfmax ;
               dsum = excess*excess/(dzdx*dzdy) ;
               sum += dsum ;
            }

         }
         xsum += x*dsum ;
         ysum += y*dsum ;
         wp++ ;
      }

   }

   *xcen = (double)col + xsum/sum ;
   *ycen = (double)row + ysum/(sum*aspectratio) ;

   *fwhm = sqrt(1.273*sum) ;

   discrim = (*peak)/16. ;
   if (discrim < noise)   discrim = noise ;

   PixelsFitted = 0 ;
   pixels = -1 ;
   noiseX4 = noise*4.0 ;

   while (PixelsFitted < 5  &&  discrim >= noise)
   {
      PixelsFitted = 0 ;
      wp = 0 ;

      for (i = 0 ;  i < nr ;  i++)
      {
         for (j = 0 ;  j < nc ;  j++)
         {
            if (i > 0  &&  i < nr-1  &&  j > 0  &&  j < nc-1)
            {
	       if (w[wp] > discrim  &&  w[wp+2*nw] < noise*w[wp])
	       {
	          w[wp+nw] = w[wp] ;
	          PixelsFitted++ ;
	       }
	       else
	          w[wp+nw] = 0.0 ;
            }
            else
               w[wp+nw] = 0.0 ;

	    wp++ ;

         }

      }

      discrim *= 0.8409 ;
      pixels = PixelsFitted ;
   }


/* Take logarithm of background subtracted data.			*/

   wp = 0 ;

   for (i = 0 ;  i < nr ;  i++)
   {
      for (j = 0 ;  j < nc ;  j++)
      {
	 if (w[wp+nw] > 0.0)
	 {
	    w[wp] = log(w[wp]) ;
	 }

	 wp++ ;
      }

   }

/* Form simultaneous equations for fit.					*/
/*									*/
/*		V = A.x							*/
/*									*/
/*	(Note: x in the above has nothing to do with x as used in	*/
/*	 this subroutine! )						*/

   for (i = 0 ;  i < 4 ;  i++)
   {
      V[i] = 0.0 ;

      for (j = 0 ;  j < 4 ;  j++)   A[i][j] = 0.0 ;
   }

   wp = 0 ;
   PixelsFitted = 0 ;

   for (i = 0 ;  i < nr ;  i++)
   {
      y = aspectratio*(double)(i - mrm) ;
      yy = y*y ;

      for (j = 0 ;  j < nc ;  j++)
      {
	 x = (double)(j - mcm) ;
	 xx = x*x ;
	 rr = xx + yy ;


	 v = w[wp+nw] ;		/* Weight for fitting	*/

	 if (v > 0.0)
	 {

	    if (rr <= RR)
	    {
	       A[0][0] += v*rr*rr ;
	       A[0][1] += v*x*rr ;
	       A[0][2] += v*y*rr ;
	       A[0][3] += v*rr ;

	       A[1][1] += v*xx ;
	       A[1][2] += v*x*y ;
	       A[1][3] += v*x ;

	       A[2][2] += v*yy ;
	       A[2][3] += v*y ;

	       A[3][3] += v ;

	       V[0] += v*w[wp]*rr ;
	       V[1] += v*w[wp]*x ;
	       V[2] += v*w[wp]*y ;
	       V[3] += v*w[wp] ;

	       PixelsFitted++ ;
	    }

	 }

	 wp++ ;

      }
 
   }

   *Pixels = PixelsFitted ;


/* Fill in lower triangle of symmetrical matrix.			*/

   A[1][0] = A[0][1] ;
   A[2][0] = A[0][2] ;
   A[3][0] = A[0][3] ;
   A[2][1] = A[1][2] ;
   A[3][1] = A[1][3] ;
   A[3][2] = A[2][3] ;

/* Solve simultaneous equations.					*/
/* Solution will be returned in V					*/

   slaDmat (4, A[0], V, &det, &flag, work) ;
   free(w) ;
   if (V[0] > 0.0)   flag = -1 ;

   if (flag != 0)
   {
      return (3) ;
   }

/* Recover gaussian profile.						*/

   *peak = exp(V[3] - 0.25*(V[1]*V[1] + V[2]*V[2])/V[0]) ;
   *fwhm = 1.665/sqrt(-V[0]) ;
   x0 = - 0.5*V[1]/V[0] ;
   y0 = - 0.5*V[2]/V[0] ;
   *xcen = (double)col + x0 ;
   *ycen = (double)row + y0/aspectratio ;


   if (*peak > 0.0)   *validity = 1. - noise/(*peak + noise) ;
   else               *validity = *peak/noise ;


   return 0 ;
}

#define min(a,b) (((a) < (b)) ? (a) : (b))
#define max(a,b) (((a) > (b)) ? (a) : (b))

/*
 * This function returns the average value of all points surrounding
 * <row0> <col0>. It excludes the minimum & maximum values though.
 * (ie. they are treated as outliers)
 */
double GCamAvgLocal (
	double	*data,
	int		nCols,
	int		nRows,
	int		col0,
	int		row0)
{
	/* wmin, wmax: minimum and maximum in 3X3 block */
	double	wmin, wmax, sum = 0.0;
	int		r, c;

	wmin = wmax = data[nCols*row0+col0] ;

	for (r = -1; r <= 1; r++)
	{
		for (c = -1; c <= 1; c++)
		{
			int index = (nCols * (row0 + r)) + col0 + c;
			double val = data[index];
			assert(index >= 0 && index < (nRows * nCols));
			wmin = min(wmin, val);
			wmax = max(wmax, val);
			sum += val;
		}
	}
	sum -= (wmin + wmax);
	return sum/7.0;
}

void GCamCompare (int a, int b, int *r)
{
	if (a < b)
		*r = 1;
	else if (a == b)
		*r = 0;
	else
		*r = -1;
}

static double GCamBackground (
	double	*w,            /* data array                                 */
	int		cols,          /* number of columns in data array            */
	int		rows,          /* number of rows in data array               */
	int		col0,          /* column of initial minimum in data array    */
	int		row0,          /* row of initial minimum in data array       */
	int		*col,          /* column position of found minimum           */
	int		*row)          /* row position of found minimum              */

   /* Returns: value of found background for this position            */
{
	int		dcol, drow ;     /* increments for column and row search       */
	double	back;            /* background estimations                     */

	/* make sure col0 & row0 have an appropriate range */
	col0 = max(col0, 1);
	row0 = max(row0, 1);
	col0 = min(col0, cols - 2);
	row0 = min(row0, rows - 2);

	back = GCamAvgLocal(w, cols, rows, row0, col0);

	GCamCompare(col0, cols / 2, &dcol);
	GCamCompare(row0, rows / 2, &drow);
	if (drow == 0 && dcol == 0)
		goto exit;

	while (1)
	{
		double	backRow, backCol;

		/*
		 * iteratively move centre to different location until we no
		 * longer get a lower background-noise value.
		 */
		backRow = GCamAvgLocal(w, cols, rows, row0 + drow, col0);
		backCol = GCamAvgLocal(w, cols, rows, row0, col0 + dcol);

		if (backRow < back && backRow < backCol)
		{
			row0 += drow;
			back = backRow;
		}
		else if (backCol < back)
		{
			col0 += dcol;
			back = backCol;
		}
		else
			break;

		assert(row0 >= 0 && row0 < rows);
		assert(col0 >= 0 && col0 < cols);

		GCamCompare(col0, cols / 2, &dcol);
		GCamCompare(row0, rows / 2, &drow);
	}

exit:

	*col = col0;
	*row = row0;

	return back;
}


/*                    G C a m  P e a k  F
 *
 *  Function:
 *     Finds a peak in a crudely smoothed array.
 *
 *  Description:
 *     This routine effecively smooths a 2-dimensional array and then
 *     finds the position of the peak data element in the smoothed data.
 *     The smoothing is crude - it is a rectangular box filter of specified
 *     size - but fast. Actually, in the interests of speed, the routine
 *     does not in fact smooth the data. What it really does is work out
 *     the total for the data surrounding each pixel in the array (ie in a
 *     box centered on the test pixel with a size as specified) and reports
 *     the position of the pixel for which that total is the largest.
 */

static void GCamPeakF (
   float *Array,                     /* (>) Array of data in question */
   int Nx,                           /* (>) First dimension of array */
   int Ny,                           /* (>) Second dimension of array */
   int Npix,                         /* (>) Size of side of smoothing box */
   int *XCent,                       /* (<) X coord of peak - from 0 up */
   int *YCent,                       /* (<) Y coord of peak - from 0 up */
   int *Status)                      /* (!) Inherited status */
{
   /*  Coding note: This should be a very simple piece of code, and this
    *  may look like a very complicated way to do a simple job. The point,
    *  however, is to do the simple job very quickly (that's the reason for
    *  using such a crude filter, in the first place!). So the code slides
    *  a 'box' of the specified size over the data, starting at one corner
    *  of the array. Having summed the data in that box, the box is then
    *  moved over by one pixel in X (ie the column number increases by one).
    *  The total for the new position could be calculated by summing all the
    *  pixels in the box again, but it's simpler to subtract off the total 
    *  for the column that just moved out of the box and then add the total
    *  for the new column that moved into the box. This means that we have to
    *  maintain an array of column totals for each column in the box. The
    *  other complexity in the interests of speed is that an array of pointers
    *  to the data is kept, one for each line in the box. So the start of
    *  each line is messy, as the totals for each column are calculated
    *  and the line pointers are set up, but it then goes very quickly through
    *  the rest of the line.
    */

   /*  Local variables  */

   int       Col;               /* Index through columns in the data */
   double    ColTotal;          /* Total of data in one box column */
   double   *ColTotals;         /* Totals of data in each column in the 'box' */
   int       FirstCol;          /* Next ColTotals element to discard */
   float    *FirstLinePtr;      /* Address of start of first line in box */
   int       Ix;                /* Loop index through data in X */
   int       Iy;                /* Loop index through data in Y */
   int       Line;              /* Index through lines in the data */
   float    *LinePtr;           /* Address of a line in the box */
   float   **LinePtrs;          /* Array of pointers to lines in the array */
   int       LastX;             /* Last box corner in X used */
   int       LastY;             /* Last box corner in Y used */
   double    MaxTotal;          /* Largest box total to date */
   int       MaxX;              /* X start of box with largest total */
   int       MaxY;              /* Y start of box with largest total */
   double    Total;             /* Total of data within the box */
   

   if (*Status != 0) return;

   /*  Initial values */

   LinePtrs = NULL;
   ColTotals = NULL;
   MaxX = 0;
   MaxY = 0;

   /*  Make sure we have an odd number of pixels in the box - this makes
    *  things neater.
    */

   Npix |= 1;

   /*  Get workspace for the arrays containing the pointers to the array
    *  elements in each line we take at a time, and for the totals in each
    *  column of the 'box' Npix by Npix that we move through the data.
    */

   LinePtrs = (float **) malloc (Npix * sizeof(float *));
   ColTotals = (double *) malloc (Npix * sizeof(double));
   if ((LinePtrs == NULL) || (ColTotals == NULL)) {
      *Status = 1;
      goto Exit;
   }

   MaxTotal = *Array;
   MaxX = 0;
   MaxY = 0;
   LastY = Ny - Npix;
   LastX = Nx - Npix;
   FirstLinePtr = Array;
   for (Iy = 0; Iy <= LastY; Iy++) {
      LinePtr = FirstLinePtr;
      FirstLinePtr += Nx; 
      for (Line = 0; Line < Npix; Line++) {
         LinePtrs[Line] = LinePtr;
         LinePtr += Nx;
      }
      Total = 0;
      for (Col = 0; Col < Npix; Col++) {
         ColTotal = 0.0;
         for (Line = 0; Line < Npix; Line++) {
            ColTotal += *(LinePtrs[Line]++);
         }
         ColTotals[Col] = ColTotal;
         Total += ColTotal;
      }
      FirstCol = 0;
      

      /*  At this point we have the Npix by Npix box conceptually with
       *  one corner at the start of the line Iy. Total is the sum of
       *  the data within the box. ColTotals[] has the set of totals for
       *  each column in the box. LinePtrs[] contains the addresses of
       *  the next element in the data in each row of the box. FirstCol
       *  is the index into ColTotals[] of the element corresponding to
       *  the leftmost column in the box (ie the first one to be replaced
       *  as the box is shifted to the right). We look to see if this is 
       *  the largest total so far.
       */

      if (MaxTotal < Total) {
         MaxTotal = Total;
         MaxX = Ix;
         MaxY = Iy;
      }

      /*  Having got all that rather messy setup out of the way, we should
       *  be able to rush through the rest of the processing for the rest of
       *  this line. We conceptually mover the 'box' over the Npix lines
       *  we are looking at. Each time we step it, the total will be 
       *  decremented by the total in the Npix elements in the column now
       *  to the left (moving out of the box) and incremented by the total
       *  in the Npix elements in the column on the right that has just moved
       *  into the box. We have already done the Ix=0 case, so we start at 1.
       */

      for (Ix = 1; Ix < LastX; Ix++) {

         /*  Decrement the total by the value in the column just moved
          *  out of the box, then calculate the total for the new column
          *  that has now moved into the box. See if this is a new peak,
          *  and carry on for the rest of the line.
          */

         Total -= ColTotals[FirstCol];

         ColTotal = 0.0;
         for (Line = 0; Line < Npix; Line++) {
            ColTotal += *(LinePtrs[Line]++);
         }
         ColTotals[FirstCol++] = ColTotal;
         Total += ColTotal;
         if (FirstCol >= Npix) FirstCol = 0;

         if (MaxTotal < Total) {
            MaxTotal = Total;
            MaxX = Ix;
            MaxY = Iy;
         }
      }
   }

   *XCent = MaxX + Npix / 2;
   *YCent = MaxY + Npix / 2;

Exit:;

   /*  On the way out, release any allocated memory */

   if (ColTotals != NULL) (void) free (ColTotals);
   if (LinePtrs != NULL) (void) free (LinePtrs);

}

/*  This is a Fortran-callable version of GCamPeakF, for testing under
 *  UNIX.
 */

void gcampeakf_ (
   float *Array,                     /* (>) Array of data in question */
   int *Nx,                          /* (>) First dimension of array */
   int *Ny,                          /* (>) Second dimension of array */
   int *Npix,                        /* (>) Size of side of smoothing box */
   int *XCent,                       /* (<) X coord of peak - from 0 up */
   int *YCent,                       /* (<) Y coord of peak - from 0 up */
   int *Status)                      /* (!) Inherited status */
{
   GCamPeakF (Array,*Nx,*Ny,*Npix,XCent,YCent,Status);
}

/*  s l a D m a t     is an SLALIB routine used by GCamCentroid */

static void slaDmat(int n, double *a, double y[], double *d, int *jf, int iw[])
/*
**  - - - - - - - -
**   s l a D m a t
**  - - - - - - - -
**
**  Matrix inversion & solution of simultaneous equations.
**
**  (double precision)
**
**  For the set of n simultaneous equations in n unknowns:
**     a.y = x
**
**  where:
**     a is a non-singular n x n matrix
**     y is the vector of n unknowns
**     x is the known vector
**
**  slaDmat computes:
**     the inverse of matrix a
**     the determinant of matrix a
**     the vector of n unknowns
**
**  Arguments:
**
**     symbol  type dimension           before              after
**
**       n      int                  no. of unknowns       unchanged
**       a      double  [n][n]           matrix             inverse
**       y      double   [n]              vector            solution
**      *d      double                      -             determinant
**    > *jf     int                         -           singularity flag
**       iw     int      [n]                -              workspace
**
**
**    >  jf is the singularity flag.  If the matrix is non-singular,
**       jf=0 is returned.  If the matrix is singular, jf=-1 & d=0.0 are
**       returned.  In the latter case, the contents of array a on return
**       are undefined.
**
**  Algorithm:
**     Gaussian elimination with partial pivoting.
**
**  Speed:
**     Very fast.
**
**  Accuracy:
**     Fairly accurate - errors 1 to 4 times those of routines optimized
**     for accuracy.
**
**  J.O.Straede   AAO  April 1993
**  P.T.Wallace   Starlink   31 October 1993
*/

#define SFA 1e-20

{
  int k, imx, i, j, ki;
  double amx, t, yk;
 
/* Pointers to beginnings of rows in matrix a[n][n] */
 
   double *ak,    /* row k    */
          *ai,    /* row i    */
          *aimx;  /* row imx  */
 
   *jf = 0;
   *d = 1.0;
   for ( k = 0, ak = a; k < n; k++, ak += n ) {
      amx = fabs ( ak[k] );
      imx = k;
      aimx = ak;
      if ( k != n ) {
         for ( i = k + 1, ai = ak + n; i < n; i++, ai += n ) {
            t = fabs ( ai[k] );
            if ( t > amx ) {
               amx = t;
               imx = i;
               aimx = ai;
            }
         }
      }
      if ( amx < SFA ) {
         *jf = -1;
      } else {
         if ( imx != k ) {
            for ( j = 0; j < n; j++ ) {
               t = ak[j];
               ak[j] = aimx[j];
               aimx[j] = t;
            }
            t = y[k];
            y[k] = y[imx];
            y[imx] = t;
            *d = - *d;
         }
         iw[k] = imx;
         *d *= ak[k];
         if ( fabs ( *d ) < SFA ) {
            *jf = -1;
         } else {
            ak[k] = 1.0 / ak[k];
            for ( j = 0; j < n; j++ ) {
               if ( j != k ) {
                  ak[j] *= ak[k];
               }
            }
            yk = y[k] * ak[k];
            y[k] = yk;
            for ( i = 0, ai = a; i < n; i++, ai += n ) {
               if ( i != k ) {
                  for ( j = 0; j < n; j++ ) {
                     if ( j != k ) {
                        ai[j] -= ai[k] * ak[j];
                     }
                  }
                  y[i] -= ai[k] * yk;
               }
            }
            for ( i = 0, ai = a; i < n; i++, ai += n ) {
               if ( i != k ) {
                  ai[k] *= - ak[k];
               }
            }
         }
      }
   }
   if ( *jf != 0 ) {
      *d = 0.0;
   } else {
      for ( k = n; k-- > 0; ) {
         ki = iw[k];
         if ( k != ki ) {
            for ( i = 0, ai = a; i < n; i++, ai += n ) {
               t = ai[k];
               ai[k] = ai[ki];
               ai[ki] = t;
            }
         }
      }
   }
}

/*                    G C a m  W i d t h  F
 *
 *  Function:
 *     Estimates the width of the peak in an array.
 *
 *  Description:
 *     This routine is passed a floating point array and the location within
 *     it of the peak of the data and an estimate of the background. Given
 *     these, it attempts to estimate the full-width at half-maximum of the
 *     peak. It does this crudely - it's only supposed to be an estimate -
 *     by collapsing the data in both directions and in each case looking 
 *     for the points at which the profile falls to half the value of the
 *     peak, allowing for the background. The accuracy of the estimate
 *     depends a lot on the accuracy of the background estimate.
 */

static void GCamWidthF (
   float *Array,                     /* (>) Array of data in question */
   int Nx,                           /* (>) First dimension of array */
   int Ny,                           /* (>) Second dimension of array */
   int NXPeak,                       /* (>) X coord of peak - from 0 up */
   int NYPeak,                       /* (>) Y coord of peak - from 0 up */
   double Background,                /* (>) Some estimate of the background */
   double *Fwhm,                     /* (<) Returned crude width estimate */
   int *Status)                      /* (!) Inherited status */
{
   /*  Local variables */

   float *ArrayPtr;      /* Points to start of a line in the data array */
   int    Bytes;         /* Number of bytes of workspace needed */
   double HalfPeak;      /* Value of the collapsed data at half-max */
   int    Ix;            /* Index through the data in X */
   int    Iy;            /* Index through the data in Y */
   int    XLim1;         /* Position of first half-max point in X */
   int    XLim2;         /* Position of second half-max point in X */
   int    YLim1;         /* Position of first half-max point in Y */
   int    YLim2;         /* Position of second half-max point in Y */
   float *Wptr;          /* Address of allocated work array */

   if (*Status != 0) return;

   /*  Initial values  */

   Wptr = 0;

   /*  Get enough memory to collapse the data along either axis */

   if (Nx > Ny) Bytes = Nx * sizeof(float);
   else Bytes = Ny * sizeof(float);
   Wptr = (float *) malloc (Bytes);
   if (Wptr == NULL) {
      *Status = 1;
      goto Exit;
   }
   
   /*  Collapse the data in Y - ie along the X axis */

   for (Ix = 0; Ix < Nx; Wptr[Ix++] = 0);
   ArrayPtr = Array;
   for (Iy = 0; Iy < Ny; Iy++) {
      for (Ix = 0; Ix < Nx; Ix++) {
         Wptr[Ix] += ArrayPtr[Ix];
      }
      ArrayPtr += Nx;
   }

   /*  The target data value is half of the peak value minus the background */
   
   HalfPeak = ((Wptr[NXPeak] - (Background * Ny)) / 2.0) + (Background * Ny);

   /*  Start from the peak and work down, looking for the point at which we
    *  first pass the target value. Then repeat going up the collapsed array.
    */

   XLim1 = 0;
   for (Ix = NXPeak; Ix >= 0; Ix--) {
      if (Wptr[Ix] < HalfPeak) {
         XLim1 = Ix;
         break;
      }
   }
   XLim2 = Nx;
   for (Ix = NXPeak; Ix < Nx; Ix++) {
      if (Wptr[Ix] < HalfPeak) {
         XLim2 = Ix;
         break;
      }
   }

   /*  Now, repeat in X. Collapse the data in X - ie along the Y axis */

   for (Iy = 0; Iy < Ny; Wptr[Iy++] = 0);
   ArrayPtr = Array;
   for (Iy = 0; Iy < Ny; Iy++) {
      for (Ix = 0; Ix < Nx; Ix++) {
         Wptr[Iy] += ArrayPtr[Ix];
      }
      ArrayPtr += Nx;
   }

   /*  The target data value is half of the peak value minus the background */
   
   HalfPeak = ((Wptr[NYPeak] - (Background * Nx)) / 2.0) + (Background * Nx);

   /*  Start from the peak and work down, looking for the point at which we
    *  first pass the target value. Then repeat going up the collapsed array.
    */

   YLim1 = 0;
   for (Iy = NYPeak; Iy >= 0; Iy--) {
      if (Wptr[Iy] < HalfPeak) {
         YLim1 = Iy;
         break;
      }
   }
   YLim2 = Ny;
   for (Iy = NYPeak; Iy < Ny; Iy++) {
      if (Wptr[Iy] < HalfPeak) {
         YLim2 = Iy;
         break;
      }
   }

   /*  Now, combine all these values for a rough estimate of the actual
    *  halfwidth of the central peak.
    */

   *Fwhm = ((YLim2 - YLim1) + (XLim2 - XLim1)) * 0.5;

   /*  On the way out, release the memory allocated */

Exit:;

   if (Wptr) free (Wptr);

}
