

/*******************************************************************************
* E.S.O. - VLT project
*
* "@(#) $Id: ACMM:ozpozFpcol/ws/src/fpcolMdlCalCentroid.c,v 1.63 07-Oct-2008 11:10:27+10 tjf $"
*
*
* who       when      what
* --------  --------  ----------------------------------------------
* ssmedley  07/09/01  created
* tfarrell  21/08/02  Catch device by zero error in centroid().
*
*/

/************************************************************************
*   NAME
*	centroid2dGauss centroid centroidGauss kgGaussFit -
*	Miscellaneous centroiding routines for FLAMES/OzPoz
*	Calibration Software.
*
*   SYNOPSIS
*       #include "fpcolMdlCalCentroid.h"
*
*	void centroid (
*		double		*grid,
*		const int	nRows,
*		const int	nCols,
*		double		*pXCentroid,
*		double		*pYCentroid);
*
*	int centroid2dGauss (
*		double		*grid,
*		int		nRows,
*		int		nCols,
*		double		*pX,
*		double		*pY,
*		double		*pFWHM);
*
*	int centroidGauss (
*		double		*grid,
*		int		nRows,
*		int		nCols,
*		double		*pX,
*		double		*pY,
*		double		*pFWHM);
*
*	int kgGaussFit (
*		double		*grid,
*		int		nRows,
*		int		nCols,
*		double		*pX,
*		double		*pY,
*		double		*pFWHM);
*
*   DESCRIPTION
*
*	This library provides several centroiding routines for use with the
*	OzPoz calibration s/w.
*
*	centroid() is the most basic routine which finds the point (x,y) where
*	half the light is above x & half the light below x. Similarly for y.
*
*	centroid2dGauss() tries to find the centroid by fitting the data to a
*	2-dimensional Gaussian. As I (ss) did not write this s/w I know no more
*	about it. Check the implementation for more details.
*
*	centroidGauss() comes from a similar source to centroid2dGauss(). This
*	routine, however, collapses data in x & y & performs 2 1-dimensional
*	Gaussian fits.
*
*	kgGaussFit() also collapses data in x & y & performs 2 1-dimensional
*	Gaussian fits. This implementation comes from Karl Glazebrook's
*	PDL::Fit::Gaussian CPAN module. (converted to C where appropriate.)
*
*	A dummy program to test the accuracy of centroiding routines can
*	be compiled by an additional -DTEST_CENTROID compiler flag.
*
*	A dummy program to verify the accuracy of the gaussian()
*	routine can be compiled by an additional -DTEST_GAUSSIAN
*	compiler flag.
*
*   ENVIRONMENT
*
*   CAUTIONS
*
*	kgGaussFit() is by far the better (more accurate) routine & should
*	be preferred over all others.
*
*   EXAMPLES
*
*   SEE ALSO
*
*   BUGS
*
*------------------------------------------------------------------------
*/

#define __EXTENSIONS__    /* Pick up extra include file stuff on solaris */
#define _HPUX_SOURCE      /* Pick up extra include file stuff on HP UX */
#include <stdio.h>
#include <math.h>		/* pow hypot sqrt M_PI M_E */
#include <strings.h>	/* bzero */

#if (TEST_CENTROID == 1 || TEST_GAUSSIAN == 1)
#include <stdlib.h>	/* atof */
#endif

#include "fpcolMdlCalCentGauss.h"
#include "fpcolMdlCalCentroid.h"
#include "fpcolMdlCalKgGaussFit.h"


#ifdef TEST_CENTROID
#define GRID_SIZE	10
#endif

void centroid (
	double		*grid,
	const int	nRows,
	const int	nCols,
	double		*pXCentroid,
	double		*pYCentroid)
{
	int		x, y;
	double	intensity, numerator = 0.0, denominator = 0.0;

	for (x = 0; x < nCols; x++)
	{
		intensity = 0.0;
		for (y = 0; y < nRows; y++)
		{
			intensity += grid[y*nCols+x];
		}
		numerator += (x * intensity);
		denominator += intensity;
	}
        /*
         * This indicates that all the data is zero.
         */
        if (denominator == 0)
            {
            *pXCentroid = 0;
            *pYCentroid = 0;
            return;
             }
	*pXCentroid = numerator / denominator;

	numerator = 0.0;
	denominator = 0.0;
	for (y = 0; y < nRows; y++)
	{
		intensity = 0.0;
		for (x = 0; x < nCols; x++)
		{
			intensity += grid[y*nCols+x];
		}
		numerator += (y * intensity);
		denominator += intensity;
	}
	*pYCentroid = numerator / denominator;
}

int centroid2dGauss (
	double	*grid,
	int		nRows,
	int		nCols,
	double	*pX,
	double	*pY,
	double	*pFWHM)
{
	GCamWindowType	window;
	StatusType		status = GCAM__OK;
	double			result[3];

	window.MaxX = nCols;
	window.MaxY = nRows;
	window.Xoffset = 0;
	window.Yoffset = 0;
	window.Xdim = nCols;
	window.Ydim = nRows;
	window.PixelSize = sizeof(double);

	GCamCentroid2dGauss(&window, (void *) grid, -1, result, &status);
	if (status != GCAM__OK)
		return (int) status;

	*pX = result[0];
	*pY = result[1];
	if (pFWHM != NULL)
		*pFWHM = result[2];

	return 0;
}

int centroidGauss (
	double	*grid,
	int		nRows,
	int		nCols,
	double	*pX,
	double	*pY,
	double	*pFWHM)
{
	GCamWindowType	window;
	StatusType		status = GCAM__OK;
	double			result[3];

	window.MaxX = nCols;
	window.MaxY = nRows;
	window.Xoffset = 0;
	window.Yoffset = 0;
	window.Xdim = nCols;
	window.Ydim = nRows;
	window.PixelSize = 4; /* 4 means interpret <grid> as array of double */

	GCamCentroidGauss(&window, (void *) grid, -1, result, &status);
	if (status != GCAM__OK)
		return (int) status;

	*pX = result[0];
	*pY = result[1];
	if (pFWHM != NULL)
		*pFWHM = result[2];

	return 0;
}

int kgGaussFit (
        int     fibre,
	double	*grid,
	int		nRows,
	int		nCols,
	double	*pX,
	double	*pY,
	double	*pFWHM)
{
	double	col[nRows];
	double	row[nCols];
	int		x, y, rv;
	double	fwhmX, fwhmY, bg, peak;

	bzero(col, sizeof(col));
	bzero(row, sizeof(row));

	/*
	 * collapse each column to get a row (stored in <row>)
	 * collapse each row to get a column (stored in <col>)
	 */
	for (x = 0; x < nCols; x++)
	{
		for (y = 0; y < nRows; y++)
		{
			row[x] += grid[y * nCols + x];
			col[y] += grid[y * nCols + x];
		}
	}

	/* last arg == 0.0 means don't calculate/use bg noise */
	rv = PDLGaussianFit(fibre, nCols, row, pX, &fwhmX, &bg, &peak, 0.0);
	if (rv != 0)
		return rv;

	rv = PDLGaussianFit(fibre, nRows, col, pY, &fwhmY, &bg, &peak, 0.0);
	if (rv != 0)
		return rv;

	*pFWHM = (fwhmX > fwhmY ? fwhmX : fwhmY);

	return 0;
}

/* #if (TEST_CENTROID == 1 || TEST_GAUSSIAN == 1) */
double gaussian (double x, double mean, double stdDev)
{
	return (1.0 / (sqrt(2.0 * M_PI * (pow(stdDev, 2))))) *
		pow(M_E, (-(pow(x - mean, 2))) / (2.0 * (pow(stdDev, 2))));
}
/* #endif */

#ifdef TEST_GAUSSIAN
int main (int argc, char *argv[])
{
	double	mean = 0.0;
	double	stdDev = 1.0;
	if (argc < 2 || argc > 4)
	{
		fprintf(stderr, "usage: %s <x> [<stdDev>] [<mean>]\n", argv[0]);
		exit(1);
	}
	if (argc >= 3)
		stdDev = atof(argv[2]);
	if (argc == 4)
		mean = atof(argv[3]);
	printf("nmean=%f\nstdDev=%f\n", mean, stdDev);
	printf("%f\n", gaussian(atof(argv[1]), mean, stdDev));
	return 0;
}
#endif

#ifdef TEST_CENTROID
int main (int argc, char *argv[])
{
	double	grid[GRID_SIZE][GRID_SIZE];
	double	offsetX, offsetY, peakX, peakY, xCentroid, yCentroid, fwhm;
	int		x, y;

	if (argc != 3)
	{
		fprintf(stderr, "usage: %s <x.xxx> <y.yyy>\n", argv[0]);
		exit(1);
	}
	offsetX = atof(argv[1]);
	offsetY = atof(argv[2]);
	if (fabs(offsetX) > GRID_SIZE/2)
	{
		fprintf(stderr, "X offset places peak outside grid.\n");
		exit(1);
	}
	if (fabs(offsetY) > GRID_SIZE/2)
	{
		fprintf(stderr, "Y offset places peak outside grid.\n");
		exit(1);
}
	peakX = (GRID_SIZE-1.0)/2.0 + offsetX;
	peakY = (GRID_SIZE-1.0)/2.0 + offsetY;

	printf("peak is at (%f, %f)\n", peakX, peakY);

	for (y = 0; y < GRID_SIZE; y++)
	{
		for (x = 0; x < GRID_SIZE; x++)
		{
			grid[y][x] = gaussian(hypot(x - peakX, y - peakY),
								  0.0, /* mean */
								  1.0); /* std. dev. */
			printf("%7.5f ", grid[y][x]);
		}
		printf("\n");
	}

	centroid((double *) grid,
			 GRID_SIZE,
			 GRID_SIZE,
			 &xCentroid,
			 &yCentroid);

	printf("centroid: (%f, %f)\n", xCentroid, yCentroid);

	/* use an alternative routine so we can compare results */
	if (centroid2dGauss((double *) grid,
						GRID_SIZE,
						GRID_SIZE,
						&xCentroid,
						&yCentroid,
						&fwhm) == 0)
	{
		printf("gaussian fit, centroid: x=%f y=%f fwhm=%f\n",
			   xCentroid,
			   yCentroid,
			   fwhm);
	}
	return 0;
}
#endif
