

/*******************************************************************************
* E.S.O. - VLT project
*
* "@(#) $Id: ACMM:ozpozFpcol/ws/src/fpcolMdlCalConvert.c,v 1.63 07-Oct-2008 11:10:25+10 tjf $"
*
*
* who       when      what
* --------  --------  ----------------------------------------------
* tfarrell  12/04/02  Added raToString, decToString, mjdToString
*                     stringToMjd, stringToPosition, stringToObjectPos. 
* ssmedley  07/09/01  created
*
*/

/************************************************************************
*   NAME
*	toFractions toRadians meanToApparent -
*	Miscellaneous conversion routines used by the FLAMES/OzPoz
*	Calibration s/w.
*
*   SYNOPSIS
*       #include "fpcolMdlCalConvert.h"
*
*	double toFractions (int degrees, int minutes, double seconds);
*
*	double toRadians (double encoded);
*
*	void meanToApparent
*		(double		meanRA,
*		 double		meanDEC,
*		 double		properMotionRA,
*		 double		properMotionDEC,
*		 double		params[21],
*		 double		*apparentRA,
*		 double		*apparentDEC);
*
*       const char *raToString(double ra);
*       const char *decToString(double dec);
*       const char *mjdToString(double mjd);
*
*       int stringToMjd(char *buffer,
*                      double *mjd);
*       int stringToPosition(char *buffer,
*                            double *ra,
*                            double *dec);
*      int stringToObjectPos(char *buffer,
*                            double *pars,
*                            int    *button,
*                            double *meanRA,
*                            double *meanDEC,
*                            double *appRA,
*                            double *appDEC,
*                            double *pmRA,
*                            double *pmDEC);
*
*
*   DESCRIPTION
*	This library provides conversion routines used by the
*	FLAMES/OzPoz calibration s/w.
*
*	toFractions() converts <degrees> <minutes> <seconds> into fractions of
*	degrees. ie. 10d 30m 0s == 10.5 d
*	Note: NO range checking is performed.
*
*	toRadians() converts an input double of the form
*	DDMMSS.SS
*	DD = degrees
*	MM = arcminutes
*	SS.SSD = arcseconds
*	into radians
*
*	meanToApparent() provides a more user-friendly interface to the
*	mean-to-apparent conversion routine in Pat Wallace's slalib library.
*	<params[21]> must be initialsed with a call to <slaMappa> before
*	invoking this routine.
*
*       raToString() converts an RA in radians to a string representation.  Returns
*       a pointer to a static string.
*
*       decToString() converts a Declination in radians to a string representation. Returns
*       a pointer to a static string.
*
*       mjdToString() converts an Time and Date as an MJD to a string representation. Returns
*       a pointer to a static string.
*
*       stringToMjd() converts a string in the format returned by mjdToString() back into
*       a MJD
*
*       stringToPosition() converts a string containing an RA and DEC specification as 
*       produced by raToString() and decToString() back into radians.
*
*       stringToObjectPos() converts a string containing a button number, RA and DEC
*       into the required information (Proper motion currently set to zero).
*
*
*   ENVIRONMENT
*
*   CAUTIONS
*
*	No range checking is performed on toFractions().
*
*   EXAMPLES
*
*   SEE ALSO
*
*	slaMappa() & slaMapqk() in Pat Wallace's slalib library.
*
*   BUGS
*
*------------------------------------------------------------------------
*/

#include <stdio.h>

#include "fpcolMdlCalConvert.h"


#define MAX_DEGREES		359
#define MAX_ARCMINUTES	59
#define MAX_ARCSECONDS	59.9999

/*
 * This function converts <degrees> <minutes> <seconds>
 * into fractions of degrees.
 * ie. 10d 30m 0s == 10.5 d
 * Note: NO range checking is performed.
 */
double toFractions (int degrees, int minutes, double seconds)
{
	return (degrees + (((double) minutes + (seconds / 60.0)) / 60));
}


/*
 * This function converts an input double of the form
 * DDMMSS.SS
 * DD = degrees
 * MM = arcminutes
 * SS.SSS = arcseconds
 * into radians
 */
double toRadians (double encoded)
{
	int		bNegative = (encoded < 0.0);
	double	posEncoded = (bNegative ? -encoded : encoded);
	int		degrees = (posEncoded / 10000) * (bNegative ? -1 : 1);
	int		arcMinutes = (posEncoded - (degrees * 10000)) / 100;
	double	arcSeconds = (posEncoded - (degrees * 10000) - (arcMinutes * 100));
	double	radians;
	int		j;

	/*
	 * Despite being documented as so, <slaCaf2r> does not
	 * compute the result if there is a range-check violation.
	 * Hence, perform range checks here.
	 */
	if (degrees > MAX_DEGREES)
	{
		fprintf(stderr, "toRadians: WARNING: degrees range violation on %11.4f\n", encoded);
		degrees = MAX_DEGREES;
	}
	if (arcMinutes > MAX_ARCMINUTES)
	{
		fprintf(stderr, "toRadians: WARNING: arcMinutes range violation on %11.4f\n", encoded);
		arcMinutes = MAX_ARCMINUTES;
	}
	if (arcSeconds > MAX_ARCSECONDS)
	{
		fprintf(stderr, "toRadians: WARNING: arcSeconds range violation on %11.4f\n", encoded);
		arcSeconds = MAX_ARCSECONDS;
	}

	slaDaf2r(degrees, arcMinutes, arcSeconds, &radians, &j);

	if (bNegative)
		radians = -radians;
	if (j != 0)
		fprintf(stderr, "toRadians: WARNING: range check failed on <%.4f> (%d)\n", encoded, j);
	/*
	printf("%6.2f is %d degrees %d arcminutes %.2f arcseconds (%f)\n",
			encoded, degrees, arcMinutes, arcSeconds, radians);
	*/

	return radians;
}


/* convert RA/DEC coordinates from mean to apparent */
void meanToApparent
	(double		meanRA,		/* radians */
	 double		meanDEC,	/* radians */
	 double		properMotionRA,		/* radians */
	 double		properMotionDEC,	/* radians */
	 double		params[21],
	 double		*apparentRA,
	 double		*apparentDEC)
{
	slaMapqk(meanRA,		/* in: mean RA (radians) */
			 meanDEC,		/* in: mean DEC (radians)*/
			 properMotionRA,	/* in: proper motion RA (radians) */
			 properMotionDEC,	/* in: proper motion DEC (radians) */
			 0.0,			/* n/a: parallax */
			 0.0,			/* n/a: radial velocity */
			 params,
			 apparentRA,	/* out: apparent RA */
			 apparentDEC);	/* out: apparent DEC */
}



/*
 * Convert an RA value in radians to a string.  Note - returns pointer to static string.
 */
const char *raToString(double ra)
{
  static char buffer[40];

  char ra_sign;
  int hmsf[4];

  slaDr2tf(2, ra, &ra_sign, hmsf);
  sprintf(buffer, "%c%.2d:%.2d:%.2d.%d",
          ra_sign, hmsf[0], hmsf[1], hmsf[2], hmsf[3]);

  return buffer;
 
}
/*
 *  Convert a declination in radians to a string. Note - returns pointer to static string.
 */
const char *decToString(double dec)
{
  static char buffer[40];

  char dec_sign;
  int dmsf[4];

  slaDr2af(2, dec, &dec_sign, dmsf);

  sprintf(buffer, "%c%.2d:%.2d:%.2d.%.2d", 
          dec_sign, dmsf[0], dmsf[1], dmsf[2], dmsf[3]);

  return buffer;
}

/*
 * Convert a MJD to a string.  Note - returns poitner to a static string.
 */
const char *mjdToString(double mjd)
{
  static char buffer[40];
  char time_sign;
  int ymdf[4];
  int thmsf[4];
  int j;
  slaDjcal(2, mjd, ymdf, &j);
  slaCd2tf(0, mjd - (int)mjd, &time_sign, thmsf);

  sprintf(buffer, "%.4d/%.2d/%.2d %.2d:%.2d:%.2d", 
          ymdf[0], ymdf[1], ymdf[2], thmsf[0], thmsf[1], thmsf[2]);

  return buffer;
}

/*
 * Convert the simple date string written by mjdToSTring() to back to a mjd.
 */
int stringToMjd(char *buffer,
                      double *mjd)
{
  int year;
  int month;
  int day;
  int hour;
  int min;
  int sec;
  double date;
  double time;
  int status;

  if (sscanf(buffer, "%d/%d/%d %d:%d:%d", &year, &month, &day, &hour, &min, &sec) != 6)
      return -1;
  
  slaCaldj(year, month, day, &date, &status);
  if (status != 0)
    return -1;

  time = hour/24.0 + min/(24.0 * 60.0) + sec /(24.0 * 60.0 * 60.0);

  *mjd = date + time;

  return 0;
}
/*
 * Convert a string containing RA/Dec in the form written by the raToString()
 * decToString() back into radians.
 */
int stringToPosition(char *buffer,
                              double *ra,
                              double *dec)
{
  char   ra_sign;
  int    ra_hour;
  int    ra_min;
  float  ra_sec;

  char   dec_sign;
  int    dec_deg;
  int    dec_min;
  float  dec_sec;

  int j;
  
  if (sscanf(buffer, "%c%d:%d:%f %c%d:%d:%f", 
            &ra_sign,  &ra_hour, &ra_min,  &ra_sec,
            &dec_sign, &dec_deg, &dec_min, &dec_sec) != 8)
    return -1;
  
   slaDtf2r(ra_hour, ra_min, ra_sec, ra, &j);
   if ((j != 0)&&(ra_sign != '+'))
     {
       fprintf(stderr,"Failed to format RA from string \"%s\"\n", buffer);
       return -1;
     }
   
   slaDaf2r(dec_deg, dec_min, dec_sec, dec, &j);
   if (j != 0)
     {
       fprintf(stderr,"Failed to format DEC from string \"%s\"\n", buffer);
       return -1;
     }
   if (dec_sign == '-')
     *dec *= -1;



  return 0;
}
/*
 * Convert a string contianing a button number (3 characters), RA written by
 * raToString() and dec written by decToString() back into appropiate values.
 *
 * Intended to convert an object line written by the savehd command back into
 * object information
 *
 * At the moment we won't use proper motion - it is not written by savehd and
 * here it is assumed to be zero.
 */
int stringToObjectPos(char *buffer,
                      double *pars,
                      int    *button,
                      double *meanRA,
                      double *meanDEC,
                      double *appRA,
                      double *appDEC,
                      double *pmRA,
                      double *pmDEC)
{

  /*
   * Null terminate after the button number - so that we can easily format it
   */
  buffer[3] = '\0';
  if (sscanf(buffer,"%d", button) != 1)
    {
      fprintf(stderr, "Failed to read button number\n");
      return -1;
    }
  /*
   * Now grab the mean positions.
   */
  if (stringToPosition(&buffer[4], meanRA, meanDEC) == -1)
    return -1;


  /*
   * Currently no proper motions used.
   */
  *pmRA = 0;
  *pmDEC = 0;

  /*
   * convert to apparent positions.
   */
  meanToApparent(*meanRA, *meanDEC, *pmRA, *pmDEC, pars, appRA, appDEC);
  return 0;
  
}

