

/*******************************************************************************
* E.S.O. - VLT project
*
* "@(#) $Id: ACMM:ozpozFpcol/ws/src/fpcolMdlCalFit.c,v 1.63 07-Oct-2008 11:10:25+10 tjf $"
*
*
* who       when      what
* --------  --------  ----------------------------------------------
* tfarrell  10/18/02  Support extra parameters to fit function.  Output
*                     rms and psd as arcseconds as well as microns.
* tfarrell  08/10/02  Removed vltPort.h - this is not built as part
*                     of the VLT software but as a standalone program
* psantin   31/08/02  added vltPort.h
* tfarrell  24/08/02  Revamp the way the residual output stuff works, based
*                     on Matthew Colless's work in the area.  Corresponding
*                     change to residuals.tcl.
* tfarrell  24/08/02  Update comments about residual  calculations and
*                     subtract rater then add offset to initial position.
* tfarrell  21/08/02  Improve formating of variance output.  Save residuals
*                     insert of outputting them
* tfarrell  21/08/02  Fix bug I had introduced in wavelength model handling,
*                     which was causing problems in the residuals.out file output.
*                     Remove decimal places in residual output.
* tfarrell  27/06/02  Warn about fitting parameters which don't make sense
*                     by calling fpcolMdlCalCheckFitType() both before
*                     an after the fit.
*                     Remove the need for the "create" command by
*                     implementing its functionality here.
*                     findButton() extracted to fpcolMdlLibs.c and renamed
*                     as fpcolMdlCalFindButton().
* tfarrell  09/05/02  We support more offsets then fibre buttons, so
*                     that we can model wavelength.  Output should
*                     indicate if a parameter was fitted.
* tfarrell  12/04/02  Improve formating of output.
* ssmedley  07/09/01  created
*
*/

/************************************************************************
*   NAME
*	fitFn -
*	Fit a model, calculating residuals & handling error conditions
*	appropriately.
*
*   SYNOPSIS
*       #include "fpcolMdlCalFit.h"
*
*	int fitFn
*		(int			nArgs,
*		 char			*args[],
*		 FpcalData		*pFpcalData)
*
*
*   DESCRIPTION
*	This library routine fits the specified data to create an
*	astrometric model.
*
*	fitFn() accepts an already created model <ppModel> & tries to fit
*	the data in <pFpcalData> using fpcolMdlFit(). If successful, it
*	will output a file (currently residuals.out) with the results of
*	all appropriate residuals calculations. This file can then be
*	input into a Tcl/Tk script (currently residuals.tcl) to (graphically)
*	view the resulting astrometric model.
*
*
*   ENVIRONMENT
*
*   CAUTIONS
*
*   EXAMPLES
*
*   SEE ALSO
*
*   BUGS
*
*------------------------------------------------------------------------
*/

#define __EXTENSIONS__    /* Pick up extra include file stuff on solaris */
#define _HPUX_SOURCE      /* Pick up extra include file stuff on HP UX */

#include <math.h>		/* sqrt */
#include <strings.h>	/* strcasecmp */
#include <stdlib.h>
#include <assert.h>

#include "fpcolMdlCalFit.h"
#include "fpcolMdlCalModel.h"	/* createModel */
#include "fpcolMdlCalMacros.h"			/* RD2POS */


/* name of output file containing residual results */
#define RESIDUALS_OUT	"residuals.out"

static void CalcCompass( 
    FILE *outputFile,
    FpcalData *pFpcalData,
    StatusType *status);

/*
 * call <OzPosFit> with all relevant variables.
 * Note: must have 'create'd model
 *       & 'load'ed fpcaldata.
 */
int fitFn
	(int			nArgs,
	 char			*args[],
	 FpcalData		*pFpcalData)
{
	double			aParameterVariances[NUM_PARAMETERS], rms, psd;
	StatusType		status = STATUS__OK;
	int				i;
	FILE			*pFile = NULL;
        FpcalPars               origParams;
        FpcalPars               newParams;
        double                  *apparentRA;
        double                  *apparentDEC;
        double                  rms_arcsec;
        double                  psd_arcsec;

	if (nArgs >= 2 && strcasecmp("help", args[1]) == 0)
	{
		printf("%s: usage: %s\n", args[0], args[0]);
		return 0;
	}
	if (nArgs != 1)
	{
		fprintf(stderr, "%s: no arguments required\n", args[0]);
		return -1;
	}
	if (!pFpcalData->bInitialised)
	{
		fprintf(stderr, "%s: must load fpcaldata first! Use \"load\"\n", args[0]);
		return -1;
	}

        /*
         * Calculate the offsets and create the model.
         */
	calculateOffsets(pFpcalData, &status);
        if (fpcolMdlCalCreateModels(pFpcalData) != 0) 
            return -1;
        /*
         * Output any warning about fitting parameters which don't make sense.
         */
	fpcolMdlCalCheckFitType(pFpcalData->aFit);


        /*
         * For each offset, we need the corresponding apparent RA and Dec, which
         * we get from the button number assocaited with the offset.
         */
        apparentRA  = (double *)malloc(sizeof(double)*pFpcalData->nOffsets);
        assert(apparentRA != 0);
        apparentDEC = (double *)malloc(sizeof(double)*pFpcalData->nOffsets);
        assert(apparentDEC != 0);

        for (i = 0; i < pFpcalData->nOffsets ; ++i) 
          {
            int buttonIndex = fpcolMdlCalFindButton(pFpcalData->nFibres_a,
                                         pFpcalData->buttons, 
                                         pFpcalData->offsetButtonNo[i]);
            if (buttonIndex >= 0)
              {
                apparentRA[i] = pFpcalData->apparentRA[buttonIndex];
                apparentDEC[i] = pFpcalData->apparentDEC[buttonIndex];
              }
            else 
              {
                fprintf(stderr,"Can't find RA/DEC for button %d, for which we have an offset\n",
                        pFpcalData->offsetButtonNo[i]);
                fprintf(stderr," This is probably a programming error\n");
                return -1;
              }
          }

        /*
         * Save the existing parameter values.
         */
        origParams = *fpcolMdlCalGetPars(&pFpcalData->model);

	/* fit the model: Note that this will alter the parameters of the model */
	(*pFpcalData->fitFunc)(pFpcalData->nOffsets,
                               &pFpcalData->model,
                               pFpcalData->powellTol,
                               pFpcalData->powellMaxIts,
                               pFpcalData->powellMode,
                               pFpcalData->observationDate,
                               pFpcalData->fieldCentreAppRA,
                               pFpcalData->fieldCentreAppDEC,
                               apparentRA,
                               apparentDEC,
                               pFpcalData->offsetRA,
                               pFpcalData->offsetDEC,
                               pFpcalData->offsetWaveIndex,
                               pFpcalData->aFit,
                               aParameterVariances,
                               &rms,
                               &psd,
                               &status);

	if (status != STATUS__OK)
	{
          fpcolMdlCalSetPars(&origParams, &pFpcalData->model);
          fprintf(stderr, "%s: Couldn't set/fit model\n", args[0]);
          return -1;
	}
        newParams = *fpcolMdlCalGetPars(&pFpcalData->model);


	printf("new parameter values and variances:\n");
	for (i = 0; i < NUM_PARAMETERS; i++)
	{
		printf("%.2d: %12.8g +/- %12.8g  (%sfitted) %s - %s\n", 
                       i, newParams.p[i], sqrt(aParameterVariances[i]),
                       (pFpcalData->aFit[i] ? "not ": "    "),
                       ModelParNames[i], ModelParUnits[i]);
	}

        /*
         * Convert the rms and psd to arceconds before outputing in both
         * microns and arcseconds.  We can assume the wavelength index
         * for the first fibre.
         */
        fpcolMdlMicrons2Arcsec(fpcolMdlCalWaveModel(&pFpcalData->model, 
                                                    pFpcalData->offsetWaveIndex[0]),
                               rms, &rms_arcsec, &status);
        fpcolMdlMicrons2Arcsec(fpcolMdlCalWaveModel(&pFpcalData->model,
                                                    pFpcalData->offsetWaveIndex[0]),
                               psd, &psd_arcsec, &status);

	printf("rms=%8.4g microns, %5.2g arcseconds\n", rms, rms_arcsec);
        printf("psd=%8.4g microns, %5.2g arcseconds\n", psd, psd_arcsec);

	/*
	 * CALCULATE RESIDUALS:
	 *	 This involves determining the difference b/w the actual position of the
	 *	 fibre buttons & the positions given by our newly calculated model.
	 *	 The actual position is simply the apparent RA/DEC + offset RA/DEC.
	 *	 The position given by the model is returned by inputing soley the
	 *	 apparent RA/DEC.
	 *	 Note that the results will be converted to X/Y (microns) to make
	 *   the GUI implementation easier.
	 */

	if ((pFile = fopen(RESIDUALS_OUT, "w")) == NULL)
	{
		perror(RESIDUALS_OUT);
                fprintf(stderr, "%s: Cannot open output file\n", args[0]);
                fpcolMdlCalSetPars(&origParams, &pFpcalData->model);
		return -1;
	}

        fprintf(pFile, "#NAME=%s\n", pFpcalData->fileName);
        fpcolMdlCalSetPars(&origParams, &pFpcalData->model);
        CalcCompass(pFile, pFpcalData, &status);
	fprintf(pFile, "# The format of this file is:\n");
	fprintf(pFile, "# fibre x y initialX initialY fittedX fittedY\n");


	for (i = 0; i < pFpcalData->nOffsets; i++)
	{
		double	x, y, initialX, initialY, fittedX, fittedY;
                int waveIndex = pFpcalData->offsetWaveIndex[i];

		/* use initial model with RA/DEC - offset to get position. */
                /*
                 * We we use the initial model applied to the object sky position, to get
                 * the button plate position.
                 */
                fpcolMdlCalSetPars(&origParams, &pFpcalData->model);
		RD2POS(fpcolMdlCalWaveModel(&pFpcalData->model, waveIndex),
			   pFpcalData->fieldCentreAppRA,
			   pFpcalData->fieldCentreAppDEC,
			   apparentRA[i],
			   apparentDEC[i],
			   pFpcalData->observationDate,
			   &initialX,
			   &initialY,
			   &status);

		/* 
                 * Use the fitted model with just RA/DEC to get the position of 
                 * the object in plate coorindates as determined by the
                 * fitted model.
                 */
                fpcolMdlCalSetPars(&newParams, &pFpcalData->model);
		RD2POS(fpcolMdlCalWaveModel(&pFpcalData->model, waveIndex),
			   pFpcalData->fieldCentreAppRA,
			   pFpcalData->fieldCentreAppDEC,
			   apparentRA[i],
			   apparentDEC[i],
			   pFpcalData->observationDate,
			   &fittedX,
			   &fittedY,
			   &status);
                

		if (status != STATUS__OK)
		{
			fprintf(stderr, "[%3d]: ERROR\n", pFpcalData->offsetButtonNo[i]);
			continue;
		}

                /*
                 * Approximation to the true position of the button.
                 *
                 * The fitted model, applied to the RA/DEC minus offset
                 */
		RD2POS(fpcolMdlCalWaveModel(&pFpcalData->model, waveIndex),
			   pFpcalData->fieldCentreAppRA,
			   pFpcalData->fieldCentreAppDEC,
			   apparentRA[i]  - pFpcalData->offsetRA[i],
			   apparentDEC[i] - pFpcalData->offsetDEC[i],
			   pFpcalData->observationDate,
			   &x,
			   &y,
			   &status);
                /*
                 * Save the values here so that the "show residuals" command
                 * will work.
                 */
                pFpcalData->residualsInfo[i].initialX = initialX;
                pFpcalData->residualsInfo[i].initialY = initialY;
                pFpcalData->residualsInfo[i].fittedX  = fittedX;
                pFpcalData->residualsInfo[i].fittedY  = fittedY;
                pFpcalData->residualsInfo[i].x        = x;
                pFpcalData->residualsInfo[i].y        = y;
                /*
                 * Output details to the residuals files.
                 */
		fprintf(pFile,
				"%3d %9.0f %9.0f %9.0f %9.0f %9.0f %9.0f\n",
				pFpcalData->offsetButtonNo[i],
				x,
				y,
				initialX,
				initialY,
				fittedX,
				fittedY);
	}
	fclose(pFile);
        pFpcalData->fitted = 1;

	/* save the fitted astrometric model to disk.  We just use the
           wavelength index 0 model - they are all the same from this
           point of view.
         */
        fpcolMdlCalSetPars(&newParams, &pFpcalData->model); /* Make sure we have fitted model*/
        fpcolMdlNormalizeModel(fpcolMdlCalWaveModel(&pFpcalData->model, 0), &status);
	fpcolMdlWriteModel(fpcolMdlCalWaveModel(&pFpcalData->model, 0), "./", &status);
	if (status != STATUS__OK)
	{
		fprintf(stderr, "%s: Couldn't save fitted model\n", args[0]);
                fpcolMdlCalSetPars(&origParams, &pFpcalData->model);
		return -1;
	}
	printf("%s: Fitted Astrometric Model for plate %d has been successfully saved.\n", 
               args[0],  pFpcalData->fieldPlate);

        /*
         * Restore the original model parameters.
         */
        fpcolMdlCalSetPars(&origParams, &pFpcalData->model);

        /*
         * Free malloced data
         */
        free(apparentRA);
        free(apparentDEC);

        /*
         * Again output any warning about fitting parameters which don't make sense.
         */
	fpcolMdlCalCheckFitType(pFpcalData->aFit);

	return 0;
}

#define M_PI          3.14159265358979323846
#define R2AS ((double)((180.0*60.0*60.0/M_PI)))

/*
 * Calculated the size and orientation of a one arc-second vector
 * in plate coorindates in both north and east directions.
 */
static void CalcCompass(  
    FILE *outputFile,
    FpcalData *pFpcalData,
    StatusType *status)
{
    long northX; 
    long northY;
    long eastX; 
    long eastY;
    double fieldCentreX;
    double fieldCentreY;

    double fieldPlus1NX;
    double fieldPlus1NY;
    double fieldPlus1EX;
    double fieldPlus1EY;
    /*
     * Get the field center positoin X and Y (should be 0,0);
     */
    RD2POS(fpcolMdlCalWaveModel(&pFpcalData->model, 0),
           pFpcalData->fieldCentreAppRA,
           pFpcalData->fieldCentreAppDEC,
           pFpcalData->fieldCentreAppRA,
           pFpcalData->fieldCentreAppDEC,
           pFpcalData->observationDate,
           &fieldCentreX,
           &fieldCentreY,
           status);
    /*
     * Calculate the X and Y positions on the plate for 1 arc-second north
     * of the field centre.
     */
    RD2POS(fpcolMdlCalWaveModel(&pFpcalData->model, 0),
           pFpcalData->fieldCentreAppRA,
           pFpcalData->fieldCentreAppDEC,
           pFpcalData->fieldCentreAppRA,
           pFpcalData->fieldCentreAppDEC+(1/R2AS),
           pFpcalData->observationDate,
           &fieldPlus1NX,
           &fieldPlus1NY,
           status);
    

    /*
     * Calculate the X and Y positions on the plate for 1 arc-second north
     * of the field centre.
     */
    RD2POS(fpcolMdlCalWaveModel(&pFpcalData->model, 0),
           pFpcalData->fieldCentreAppRA,
           pFpcalData->fieldCentreAppDEC,
           pFpcalData->fieldCentreAppRA+(1/R2AS),
           pFpcalData->fieldCentreAppDEC,
           pFpcalData->observationDate,
           &fieldPlus1EX,
           &fieldPlus1EY,
           status);

    /*
     * Work out out vectors.
     */
    northX = fieldPlus1NX - fieldCentreX;
    northY = fieldPlus1NY - fieldCentreY;
    eastX  = fieldPlus1EX - fieldCentreX;
    eastY  = fieldPlus1EY - fieldCentreY;

    fprintf(outputFile, "#COMPASS=%ld %ld %ld %ld\n", northX, northY, eastX, eastY);

}
