
/*******************************************************************************
* E.S.O. - VLT project
*
* "@(#) $Id: ACMM:ozpozFpcol/ws/src/fpcolMdlCalFluxOffset.c,v 1.63 07-Oct-2008 11:10:26+10 tjf $"
*
*
* who       when      what
* --------  --------  ----------------------------------------------
* tfarrell  19/10/02  Improve debugging output.  Allow for cutting out offsets
*                     where the signal to noise is poor (-cut option).  
* tfarrell  22/08/02  Support subtracting min value from each grid and
*                     -nosubmin option.
* tfarrell  21/08/02  Add -gdhs option.  
* tfarrell  09/05/02  Support output in arc-seconds rather then radians.
* ssmedley  07/09/01  created
*
*/

/************************************************************************
*   NAME
*	fluxOffsets - Generate ASCII Offset Files from ASCII Fibre
*	Flux Files.
*
*   SYNOPSIS
*       #include "fpcolMdlCalFluxOffset.h"
*
*	int isPrime (int n)
*
*	void solveQuadratic (float a, float b, float c, float *p1, float *p2)
*
*	int getDimensions (int nFiles, char *aFiles[], Dimensions *pDim)
*
*	int getGridIndices (
*		RaDec			to,
*		const Dimensions	*pDim,
*		int			*xIndex,
*		int			*yIndex)
*
*	void display (
*		const double		*grid,
*		const int		nRows,
*		const int		nCols)
*
*	int getCentroid(
*               int             fibre
*		double		*grid,
*		int		nRows,
*		int		nCols,
*		int		fitType,
*		double		*pX,
*		double		*pY)
*
*   DESCRIPTION
*
*	The main() routine in this file is for the <fluxOffsets> utility
*	which comprises part of the OzPoz Calibration Software.
*
*	The fluxOffsets program will generate an ASCII Offset File from
*	a set of ASCII Fibre Fluxes Files. Typically, these Fibre Flux
*	Files come from processing UVES slit images with the <img2flux>
*	utility.
*
*	The program reads/loads the data contained in the Fibre Fluxes Files
*	& determines the offset of the telescope when the measured
*	flux/intensity is at its greatest, for each fibre button. It
*	then writes this information to disk in the format of an
*	ASCII Offset File.
*
*	isPrime() is a utility function of getDimensions(). It simply
*	identifies whether or not the passed parameter is a prime number.
*
*	solveQuadratic() solves quadratic equations of the form:
*	ay^2 + by + c
*
*	getDimensions() answers the question:
*	What are the dimensions of the fibre button grid? It can
*	deduce this from the number of Fibre Fluxes Files.
*
*	getGridIndices() returns the position in the grid for the telescope
*	offset specified by <to>. Note that the grid has origin (0, 0) in
*	the top-left corner, so it can be used as an array index.
*
*	display() displays the values in a grid on stdout.
*
*	getCentroid() finds the centroid within <grid> using the specified
*	centroiding algorithm. (settable via command-line argument)
*
*
*   ENVIRONMENT
*
*   CAUTIONS
*
*   EXAMPLES
*
*   SEE ALSO
*
*	For a description of how to use this program, refer to
*	OzPos Calibration Software Manual. (INS-TRE-AUS-13721-0075)
*
*   BUGS
*
*------------------------------------------------------------------------
*/


static const char *rcsId="@(#) $Id: ACMM:ozpozFpcol/ws/src/fpcolMdlCalFluxOffset.c,v 1.63 07-Oct-2008 11:10:26+10 tjf $";

static void *use_rcsId = (0 ? (void *)(&use_rcsId) : (void *) &rcsId);


#define __EXTENSIONS__    /* Pick up extra include file stuff on solaris */
#define _HPUX_SOURCE      /* Pick up extra include file stuff on HP UX */

#include <stdio.h>		/* printf perror */
#include <stdlib.h>		/* malloc */
#include <strings.h>	/* */
#include <math.h>		/* sqrt */
#include <assert.h>		/* assert */
#include <limits.h>
#include <float.h>

#include <slamac.h>		/* DR2AS */

#include "fpcolMdlCalGetline.h"
#include "fpcolMdlCalCentroid.h"
#include "fpcolMdlCalLibs.h"


/* each fibre has its own grid of fluxes */
typedef struct
{
	double	*grid[3];	/* flux in blue/central/red wavelengths */
	int		nData;
}Grid;

typedef struct
{
	double	ra;
	double	dec;
}RaDec;

/*
 * structure containing all relevant info pertaining to the dimensions
 * of the fibre grid.
 */
typedef struct
{
	RaDec	tl;
	RaDec	tr;
	RaDec	bl;
	RaDec	br;
	int		nRows;
	int		nCols;
	double	spacing;
}Dimensions;

#define MAX_FIBRES 512

/* global variables - eek! */
int bVerbose = 0;

int isPrime (int n)
{
	int i;

	if (n % 2 == 0)
		return 0;

	for (i = 3; i < (n - 1) / 2; i += 2)
	{
		if (n % i == 0.0)
			return 0;
	}
	return 1;
}

/*
 * My god! Yr 10 math *can* actually be used for something other
 * than academic/esoteric purposes!
 */
void solveQuadratic (float a, float b, float c, float *p1, float *p2)
{
	*p1 = (-b + sqrt((pow(b, 2) - (4 * a * c)))) / (2 * a);
	*p2 = (-b - sqrt((pow(b, 2) - (4 * a * c)))) / (2 * a);
}

/*
 * This function answers the question:
 * What are the dimensions of the fibre button grid?
 */
int getDimensions (int nFiles, char *aFiles[], int giraffeDHSfmt, Dimensions *pDim)
{
	int		i;

	if (nFiles < 4 || isPrime(nFiles))
	{
		fprintf(stderr, "Invalid number of files specified. Must form a rectangular grid.\n");
		return -1;
	}
	for (i = 0; i < nFiles; i++)
	{
		FILE	*pFile = fopen(aFiles[i], "r");
		char	buffer[128];
		int		n;
		double	ra, dec;


		if (pFile == NULL)
		{
			perror(aFiles[i]);
			return -1;
		}
		while ((n = getline(pFile, buffer, sizeof(buffer))) >= 0)
		{
			if (n == 0 || buffer[0] == '#')
				continue;	/* ignore blank lines & comments */
			/* first line is a telescope offset */
                        if (giraffeDHSfmt)
                        {
                            if ((n = sscanf(buffer, "OFFSET_RA=%lf OFFSETDEC=%lf", &ra, &dec)) != 2)
			    {
				fprintf(stderr, "Couldn't read telescope offset in file %s.\n", aFiles[i]);
				return -1;
			    }
                        } 
                        else
			    if ((n = sscanf(buffer, "%lf%lf", &ra, &dec)) != 2)
			    {
				fprintf(stderr, "Couldn't read telescope offset in file %s.\n", aFiles[i]);
				return -1;
			    }
			break;
		}
		fclose(pFile);
		pFile = NULL;
		/* find the corners of the grid */
		if ((i == 0) || (ra <= pDim->tl.ra && dec <= pDim->tl.dec))
		{
			pDim->tl.ra = ra;
			pDim->tl.dec = dec;
		}
		if ((i == 0) || (ra >= pDim->tr.ra && dec <= pDim->tr.dec))
		{
			pDim->tr.ra = ra;
			pDim->tr.dec = dec;
		}
		if ((i == 0) || (ra <= pDim->bl.ra && dec >= pDim->bl.dec))
		{
			pDim->bl.ra = ra;
			pDim->bl.dec = dec;
		}
		if ((i == 0) || (ra >= pDim->br.ra && dec >= pDim->br.dec))
		{
			pDim->br.ra = ra;
			pDim->br.dec = dec;
		}
	}

	if (bVerbose)
	{
		fprintf(stderr, "tl: %f, %f\n", pDim->tl.ra, pDim->tl.dec);
		fprintf(stderr, "tr: %f, %f\n", pDim->tr.ra, pDim->tr.dec);
		fprintf(stderr, "bl: %f, %f\n", pDim->bl.ra, pDim->bl.dec);
		fprintf(stderr, "br: %f, %f\n", pDim->br.ra, pDim->br.dec);
	}

	if ((pDim->tl.ra != pDim->bl.ra) ||
		(pDim->tl.dec != pDim->tr.dec) ||
		(pDim->br.ra != pDim->tr.ra) ||
		(pDim->br.dec != pDim->bl.dec))
	{
		fprintf(stderr, "Fibre grid does not appear to be rectangular.\n");
		return -1;
	}

	/* can assume fibres are equally spaced */
	{
		float	dummy, r;
		float	aspectRatio = (pDim->tr.ra - pDim->tl.ra) /
							  (pDim->br.dec - pDim->tr.dec);

		/*
		 * Given the following 2 simultaneous equations, we need to solve the
		 * resulting quadratic equation to find the dimensions of the
		 * fibre grid:
		 * (x - 1) / (y - 1) = aspectRatio
		 * x * y = nFiles
		 *
		 * solve: ay^2 - (a-1)y - n = 0
		 */
		solveQuadratic(aspectRatio, -(aspectRatio - 1), -nFiles, &r, &dummy);
		pDim->nRows = (int) r;
		pDim->nCols = nFiles / pDim->nRows;
		pDim->spacing = (pDim->tr.ra - pDim->tl.ra) / (pDim->nCols - 1);
	}
	return 0;
}

/*
 * floating-point equality (comparison) is bad!
 * just check that <a> & <b> are *almost* equal.
 */
#define NEAR_ENOUGH(a, b) (fabs((a) - (b)) < 0.0000001)

/*
 * return the position in the grid for the telescope offset specified
 * by <to>. Note that the grid has origin (0, 0) in the top-left corner,
 * so it can be used as an array index.
 */
int getGridIndices (
	RaDec			to,
	const Dimensions	*pDim,
	int			*xIndex,
	int			*yIndex)
{
	int i;

	/*
	 * we know the that grid is equally spaced & the spacing b/w each
	 * point in the grid is known <pDim->spacing>. Therefore, we just
	 * have to find the number of unit-spacings b/w <to> & the telescope
	 * offset at the top-left corner of the grid.
	 */
	*xIndex = -1;
	for (i = 0; i < pDim->nCols; i++)
	{
		if (NEAR_ENOUGH(to.ra, pDim->tl.ra + (pDim->spacing * i)))
		{
			*xIndex = i;
			break;
		}
	}
	if (*xIndex < 0)
		return -1;

	*yIndex = -1;
	for (i = 0; i < pDim->nRows; i++)
	{
		if (NEAR_ENOUGH(to.dec, pDim->tl.dec + (pDim->spacing * i)))
		{
			*yIndex = i;
			break;
		}
	}
	if (*yIndex < 0)
		return -1;

	return 0;
}

/*
 * print the grid on stderr
 */
void display (
	const double	*grid,
	const int		nRows,
	const int		nCols)
{
	int	x, y;

        fprintf(stderr, "         ");
        for (x = 0; x < nRows; x++)
            fprintf(stderr, "%3d       ", x+1);
        fprintf(stderr, "\n");

	for (y = 0; y < nRows; y++)
	{
                fprintf(stderr, "%3d:  ", y+1);
		for (x = 0; x < nRows; x++)
		{
			fprintf(stderr, "%9.1f ", grid[nCols*y+x]);
		}
		fprintf(stderr, "\n");
	}
}

/*
 * Find the centroid within the grid using the specified algorithm.
 */
int getCentroid(
        int    fibre,
	double *grid,
	int	nRows,
	int	nCols,
	int	fitType,
	double	*pX,
	double	*pY)
{
	double dummy;

	/*
	 * fitType:
	 * 0 = centroid,
	 * 1 = centroid2dGauss,
	 * 2 = centroidGauss
	 * 3 = kgGaussFit
	 */
	if (fitType == 3)
	{
            if (kgGaussFit(fibre, grid, nRows, nCols, pX, pY, &dummy) != 0)
                {
                return -1;
                }
	}
	else if (fitType == 2)
	{
		if (centroidGauss(grid,
				  nRows,
				  nCols,
				  pX,
				  pY,
				  &dummy) != 0)
		{
			return -1;
		}

	}
	else if (fitType == 1)
	{
		if (centroid2dGauss(grid,
					nRows,
					nCols,
					pX,
					pY,
					&dummy) != 0)
		{
			return -1;
		}
	}
	else
	{
		centroid(grid,
			 nRows,
			 nCols,
			 pX,
			 pY);
	}
        if ((*pX <= 0)||(*pX >= nCols)||(*pY <= 0)||(*pY >= nRows))
            fprintf(stderr, "Warning: Centroid at end of or outside raster for fibre %d\n",
                    fibre);

	return 0;

}
/*
 * Return the x and y value of the peak pixel and
 * the value of the pixel.
 */
void getPeak(double *grid, 
             int	nRows,
             int	nCols,
             double     *peakVal,
             int        *peakX,
             int        *peakY)
{
    int x,y;
    *peakVal = -1*DBL_MAX;
    *peakX   = -1;
    *peakY   = -1;
    for (x = 0; x < nCols; x++)
        {
        for (y = 0; y < nRows; y++)
            {
            if (grid[y*nCols+x] > *peakVal)
                {
                *peakVal = grid[y*nCols+x];
                *peakX = x;
                *peakY = y;
                }
            }
        }
}
/*
 * Subtract the min value from every element in the array.  We also
 * happen to return the max value.
 */
void subtractMinValue(
	double *grid,
	int	nRows,
	int	nCols,
        double  *maxFlux)
{
    double minFlux = DBL_MAX;
    double valToSub;
    int i;
    int minAt = -1;
    int maxAt = -1;

    *maxFlux = -1*DBL_MAX;
    /*
     * First pass through - work out the min and max value.
     */
    for (i= 0; i < (nRows*nCols) ; ++i)
        {
        if (grid[i] < minFlux)
            {
            minFlux = grid[i];
            minAt   = i;
            }
        if (grid[i] > *maxFlux)
            {
            *maxFlux = grid[i];
            maxAt    = i;
            }
        }


    /*
     * If the min flux is less then zero, then set it to zero - this
     * indicates a problem in the calibration during the data reduction but
     * not considered unusual so we just handle it.
     */
    if (minFlux < 0)
        valToSub = 0;
    else
        valToSub = minFlux;



    if (bVerbose)
        {
        int row, col;
        row = (minAt / nRows) + 1;
        col = (minAt % nRows) + 1;
        fprintf(stderr, "Subtracting value of %8.1f, min is %8.1f (at %d,%d)\n", 
                valToSub, minFlux, col, row);

        row = (maxAt / nRows) + 1;
        col = (maxAt % nRows) + 1;
        fprintf(stderr, "Max flux value is %8.1f (%8.1f above min) (at %d,%d)\n",
                *maxFlux, *maxFlux-minFlux, col, row);
        }
    /*
     * Lower max by what we are removing.
     */
    *maxFlux -= valToSub;

    /*
     * Second pass - subtract the min value we worked out.
     */
    for (i= 0; i < (nRows*nCols) ; ++i)
        {
        grid[i] -= minFlux;
        }   
}

void usage (const char *pName)
{
	fprintf(stderr, "usage: %s [-option] file1 file2 ... fileN\n", pName);
	fprintf(stderr, "\noptions:\n");
        fprintf(stderr, "-gdhs          input files use GIRAFFE DHS format\n");
	fprintf(stderr, "-rgb           input files contain fluxes in RGB wavelengths.\n");
	fprintf(stderr, "-verbose       Extra logging information.\n");
        fprintf(stderr, "-nosubmin      Subtract the minimum value from each value in grid\n");
	fprintf(stderr, "-o <ouputFile> Specify location of ASCII offset file. [stdout]\n");
        fprintf(stderr, "-peakoff <file> Create a peak offsets file of the given name\n");
	fprintf(stderr, "-gaussfit      Use gaussian fitting routine to find centroid.\n");
	fprintf(stderr, "-gaussfit2     Use alternate gaussian fitting routine to find centroid.\n");
	fprintf(stderr, "-kgfit         Use KG's gaussian fitting routine to find centroid.\n");
	exit(1);
}

int main (int argc, char *argv[])
{
    int		i, j, nFiles = 0, bRGB = 0, rv = 0;
    int		fitType = 0; /* 0 = centroid,
                                1 = centroid2dGauss
                                2 = centroidGauss */
    int         subtractMin = 1;
    double      cut    = 0;  /* If not zero, throw out offsets if max flux is below
                                this value. */
                                   
    Dimensions	dim;
    Grid		aGrids[MAX_FIBRES];
    char		*aFiles[argc];
    FILE		*pOutput = NULL;
    FILE                *pPeakFile = NULL;
    int             giraffeDHSfmt = 0;  /* =1 means we are reading a GIRAFFE DHS formated
                                           input file */

    int             cutCount;           /* Number of offsets cut */
    int             validCount;         /* Number of valid offsets */

    memset(aGrids,0, MAX_FIBRES * sizeof(Grid));
    assert(aFiles != NULL);

    for (i = 1; i < argc; i++)
	{
        if (argv[i][0] != '-')
            {
            aFiles[nFiles++] = argv[i];
            continue;
            }
        if (strcmp(argv[i], "-rgb") == 0)
            bRGB = 1;
        else if (strcmp(argv[i], "-verbose") == 0)
            bVerbose = 1;
        else if (strcmp(argv[i], "-o") == 0)
            {
            if (i + 1 > argc)
                fprintf(stderr, "ignoring superfluous -o argument\n");
            else
                {
                pOutput = fopen(argv[++i], "w");
                if (pOutput == NULL)
                    {
                    perror(argv[i]);
                    exit(1);
                    }
                }
            }
        else if (strcmp(argv[i], "-peakoff") == 0)
            {
            if (i + 1 > argc)
                fprintf(stderr, "ignoring superfluous -peakoff argument\n");
            else
                {
                pPeakFile = fopen(argv[++i], "w");
                if (pPeakFile == NULL)
                    {
                    perror(argv[i]);
                    exit(1);
                    }
                }
            }
        else if (strcasecmp(argv[i], "-gaussfit") == 0)
            fitType = 1;
        else if (strcasecmp(argv[i], "-gaussfit2") == 0)
            fitType = 2;
        else if (strcasecmp(argv[i], "-kgfit") == 0)
            fitType = 3;
        else if (strcasecmp(argv[i], "-gdhs") == 0)
            giraffeDHSfmt = 1;
        else if (strcasecmp(argv[i], "-nosubmin") == 0)
            subtractMin = 0;
        else if (strcasecmp(argv[i], "-cut") == 0)
            {
            if (i + 1 > argc)
                {
                fprintf(stderr, "ignoring superfluous -cut argument\n");
                continue;
                }
            cut = atof(argv[++i]);
            }
        else
            usage(argv[0]);
	} /* for */

    if ((cut != 0)&&(!subtractMin))
        {
        fprintf(stderr, "At the moment, you cannot combine -cut and -nosubmin\n");
        exit(1);
        }

    if (giraffeDHSfmt && bRGB)
        {
        fprintf(stderr,"At the moment we don't support -rgb and -gdhs together\n");
        exit(1);
        }

    if (pOutput == NULL)
        pOutput = stdout;

    /* find out what the dimensions of the grid are */
    if (getDimensions(nFiles, aFiles, giraffeDHSfmt, &dim) != 0)
        {
        fprintf(stderr, "bugger: couldn't get dimensions\n");
        exit(1);
        }
    if (bVerbose)
	{
        fprintf(stderr, "dimensions are: %dx%d\n", dim.nCols, dim.nRows);
        fprintf(stderr, "spacing is: %f\n", dim.spacing);
	}

    /*
     * iterate over all the files, loading the fluxes for each
     * fibre into a grid. ie. 1 grid per fibre.
     */
    for (i = 0; i < nFiles; i++)
	{
        FILE	*pFile = fopen(aFiles[i], "r");
        char	buffer[128];
        int		n, x, y;
        int		bReadTO = 0; /* read telescope offset yet? */
        int     line = 0;
        RaDec	to;

        if (pFile == NULL)
            {
            perror(argv[i]);
            continue;
            }

        if (bVerbose)
            fprintf(stderr, "Reading %s\n", aFiles[i]);

        while ((n = getline(pFile, buffer, sizeof(buffer))) >= 0)
            {
            int		fibre;
            double	blue, central, red;

            ++line;
            if (n == 0 || buffer[0] == '#')
                continue;	/* ignore blank lines & comments */
            else if ((giraffeDHSfmt)&&
                     ((strncmp(buffer," Sequence", 9) == 0)||
                      (strncmp(buffer," --------", 9) == 0)))
                continue;       /* ignore header lines from GIRAFFE DHS files*/ 
            if (bReadTO == 0)
                {
                if (giraffeDHSfmt)
                    {
                    if ((n = sscanf(buffer, "OFFSET_RA=%lf OFFSETDEC=%lf", &to.ra, &to.dec)) != 2)
                        {
                        fprintf(stderr, "Couldn't read telescope offset, file %s, line %d.\n",
                                aFiles[i], line);
                        return -1;
                        }
                    } 
                else				/* first line is a telescope offset */
                    {
                    if ((n = sscanf(buffer, "%lf%lf", &to.ra, &to.dec)) != 2)
                        {
                        fprintf(stderr, "Couldn't read telescope offset, file %s, line %d.\n",
                                aFiles[i], line);                      
                        exit(1);
                        }
                    }
		/*
                 * find out what position this telescope offset has in
                 * the fibre grid.
                 */
                if (getGridIndices(to, &dim, &x, &y) != 0)
                    {
                    fprintf(stderr, 
                            "File %s does not appear to contain a telescope offset in the grid\n",
                            aFiles[i]);
                    exit(1);
                    }
                if (bVerbose)
                    fprintf(stderr, "File %s has grid coords %d %d\n", aFiles[i], x, y);
                assert(x < dim.nCols && y < dim.nRows);
                bReadTO = 1;
                continue;
                }

            /* get fibre number & associated flux */
            if (giraffeDHSfmt)
                {
                int seq;    /* Sequence number */
                int slit;   /* Slit number     */
                if (sscanf(buffer, "%d %d %d %lf", &seq, &slit, &fibre, &central) == 4)
                    n = 1;
                else if (sscanf(buffer, "%d %d * %lf", &seq, &slit, &central) == 3)
                    continue ;  /* Calibration fibre, not from OzPoz */
                else if (sscanf(buffer, "%d %d %d *", &seq, &slit, &fibre) == 3)
                    continue ;  /* Fibre does not fall on the CCD */
                else
                    n = 0;      /* Format failed */ 
                }
            else if (bRGB)
                {
                n = sscanf(buffer,
                           "%d%lf%lf%lf",
                           &fibre,
                           &red,
                           &central,
                           &blue) == 4;
                if (n)
                    {
                    if (red > central)
                        fprintf(stderr, "Warning: red flux > central flux for fibre %d in file %s.\n", fibre, aFiles[i]);
                    if (blue > central)
                        fprintf(stderr, "Warning: blue flux > central flux for fibre %d in file %s.\n", fibre, aFiles[i]);
                    }
                }
            else
                {
                n = sscanf(buffer, "%d%lf", &fibre, &central) == 2;
                }

            if (!n)
                {
                fprintf(stderr, "Couldn't read fibre/flux in %s, line %d.\n", aFiles[i], line);
                continue;
                }

            if (fibre >= MAX_FIBRES)
                {
                fprintf(stderr, "Fibre number %d exceeds maximum. (file %s, line %d)\n", 
                        fibre, aFiles[i], line);
                continue;
                }
            assert(fibre >= 0);
            /*
             * make sure we haven't seen a flux for this fibre at this
             * position in the grid before.
             */
            if (aGrids[fibre].nData == 0)
                {
				/*
				 * no fluxes recorded at all for this fibre.
				 * Allocate memory for grid.
				 */
                const int nBytes = dim.nRows * dim.nCols * sizeof(double);
                for (j = 0; j < (bRGB ? 3 : 1); j++)
                    {
                    aGrids[fibre].grid[j] = (double *) malloc(nBytes);
                    assert(aGrids[fibre].grid[j] != NULL);
                    memset(aGrids[fibre].grid[j],0, nBytes);
                    }
                }
            else if (aGrids[fibre].grid[0][y*dim.nCols+x] != 0.0)
                {
                fprintf(stderr, "multiple fluxes for fibre %d at (%d, %d)(file %s, line %d) \n",
                        fibre, x, y, aFiles[i], line);
                exit(1);
                }
            /* add the flux(es) to the grid for this fibre */
            if (bRGB)
                {
                aGrids[fibre].grid[1][y*dim.nCols+x] = blue;
                aGrids[fibre].grid[2][y*dim.nCols+x] = red;
                }
            aGrids[fibre].grid[0][y*dim.nCols+x] = central;
            aGrids[fibre].nData++;
            } /* while */
        fclose(pFile);
	}
    /*
     * 
     */
    for (i = 0; i < MAX_FIBRES; i++)
	{
        double dataSum = 0;
        int dataElem = 0;
        
        if (aGrids[i].nData == 0)	/* unused fibre */
            continue;

        else if (aGrids[i].nData != (dim.nCols * dim.nRows))
            {
            /* grid for this fibre was not entirely filled */
            fprintf(stderr, "Only found %d/%d fluxes for fibre %d.\n",
                    aGrids[i].nData,
                    dim.nCols * dim.nRows,
                    i);
            aGrids[i].nData = 0;  /* Ensure this is not used below */
            rv = 1;
            continue;
            }
        for (dataElem = 0; dataElem < aGrids[i].nData ; ++dataElem)
            {
            dataSum += aGrids[i].grid[0][dataElem];
            }
        if (dataSum <= 0)
            {
            if (dataSum == 0)
                fprintf(stderr, "Fibre %.3d had no flux values (probably wasn't used).\n",
                        i);
            else
                 fprintf(stderr, "Fibre %.3d had all negative fluxes (flat fielding error?).\n",
                        i);
               
            aGrids[i].nData = 0;  /* Ensure this is not used below */
            }
        }
   /*
     * At this point, we have a grid of fluxes for each fibre.
     * We can simply perform a centroid on each grid to find the
     * position of greatest flux (in grid coordinates).
     * A simple mapping will then convert this grid coordinate,
     * back into a telescope offset Ra/Dec.
     */
    validCount = 0;
    cutCount = 0;
    for (i = 0; i < MAX_FIBRES; i++)
	{
        double	centralXCentroid, centralYCentroid; /* in grid coordinates */
        double  maxFlux; /* max flux */
        RaDec	centralTO; /* centroid in telescope offset */

        if (aGrids[i].nData == 0)	/* unused fibre */
            continue;

        if (bVerbose)
            {
            fprintf(stderr, "**** Processing fibre %d *****\n", i);
            fprintf(stderr, "Flux grid for fibre %d is:\n", i);
            display(aGrids[i].grid[0], dim.nRows, dim.nCols);
            }

        if (subtractMin)
            {
            subtractMinValue(aGrids[i].grid[0], dim.nRows, dim.nCols, &maxFlux);        
            if ((cut > 0)&&(maxFlux < cut))
                {
                ++cutCount;
                fprintf(stderr,"Fibre %d removed due to poor signal to noise (%g)\n",
                        i, maxFlux);
                continue;
                }
            }
        if (getCentroid(i, aGrids[i].grid[0],
                        dim.nRows,
                        dim.nCols,
                        fitType,
                        &centralXCentroid,
                        &centralYCentroid) != 0)
            {
            fprintf(stderr, "Failed to calculate centroid for fibre %d\n", i);
            rv = 1;
            continue;
            }
        /*
         * ok, got the centroid in grid coordinates, now to find this
         * position in terms of the telescope offset (Ra/Dec).
         */
        centralTO.ra = dim.tl.ra + (centralXCentroid * dim.spacing);
        centralTO.dec = dim.tl.dec + (centralYCentroid * dim.spacing);
        if (bVerbose)
            {
            fprintf(stderr, "fibre %d: Centroid - grid (%.2f, %.2f) <-> offset (%.2f, %.2f)\n",
                   i,
                   centralXCentroid+1,
                   centralYCentroid+1,
                   centralTO.ra,
                   centralTO.dec);
            /*display(aGrids[i].grid[0], dim.nRows, dim.nCols);*/ /* Display with min subtracted*/
            }
        ++validCount;
        fprintf(pOutput, "%3d %5.2f %5.2f\n", i, 
                centralTO.ra, centralTO.dec);


        /*
         * IF we want the peak offsets file - calculate it.
         */
        if (pPeakFile)
            {
            int     centralXPeak, centralYPeak;
            double  ra,dec;
            double  peakVal;


            getPeak(aGrids[i].grid[0], dim.nRows, dim.nCols, &peakVal, 
                    &centralXPeak, &centralYPeak);

            ra =  dim.tl.ra  + ((double)centralXPeak * dim.spacing);
            dec = dim.tl.dec + ((double)centralYPeak * dim.spacing);
            fprintf(pPeakFile, "%3d %5.2f %5.2f  (%g)\n", i, 
                    ra, dec, peakVal);
            }
        /*
         * In theory, the red & blue wavelength centroid should be
         * VERY close to the central wavelength centroid. Check to see
         * that it is, & if it's not, warn the user.
         */
        for (j = 1; bRGB && j <= 2; j++)
        {
        double	xCentroid, yCentroid, deltaX, deltaY;
        RaDec	to;
        double  maxFlux; /* max flux */

        if (subtractMin)
            {
            subtractMinValue(aGrids[i].grid[j], dim.nRows, dim.nCols, &maxFlux);
            if ((cut > 0)&&(maxFlux < cut))
                {
                ++cutCount;
                fprintf(stderr,"Fibre %d removed due to poor signal to noise (%g) xxx\n",
                        i, maxFlux);
                continue;
                }
            }
        if (getCentroid(i, aGrids[i].grid[j],
                        dim.nRows,
                        dim.nCols,
                        fitType,
                        &xCentroid,
                        &yCentroid) != 0)
            {
            fprintf(stderr, "Failed to calculate %s-wavelength centroid for fibre %d\n", (j == 1 ? "blue" : "red"), i);
            rv = 1;
            continue;
            }
        to.ra = dim.tl.ra + (xCentroid * dim.spacing);
        to.dec = dim.tl.dec + (yCentroid * dim.spacing);
        /* DR2AS == number of arcseconds in a radian (currenlty assuming input is in arc-sec) */
        deltaX = (centralTO.ra - to.ra)/* * DR2AS*/;
        deltaY = (centralTO.dec - to.dec)/* * DR2AS*/;
        /* if out by more than 1 arc-second, warn user */
        if (deltaX >= 1.0 || deltaY >= 1.0)
            {
            fprintf(stderr,
                    "fibre %d: %s-wavelength centroid varies from central-wavelength\ncentroid by (%f, %f). In arcseconds (%f, %f)\n",
                    i,
                    (j == 1 ? "blue" : "red"),
                    centralTO.ra - to.ra,
                    centralTO.dec - to.dec,
                    deltaX,
                    deltaY);
            rv = 1;
            }
        }
        }
    if ((cut)&&(cutCount)) 
        {
        fprintf(stderr, 
                "Remove %d offsets due to poor signal to noise (worse then %g counts)\n",
                cutCount, cut);
        }

    fprintf(stderr, "Wrote %d offsets\n", validCount);

    if (pPeakFile)
        fclose(pPeakFile);
    if (pOutput != stdout)
        fclose(pOutput);
    return rv;
}

