
/*******************************************************************************
* E.S.O. - VLT project
*
* "@(#) $Id: ACMM:ozpozFpcol/ws/src/fpcolMdlCalGenOffsets.c,v 1.63 07-Oct-2008 11:10:26+10 tjf $"
*
*
* who       when      what
* --------  --------  ----------------------------------------------
* tfarrell  15/10/02  Include unitstd.h for getopt().
* tfarrell  10/10/02  Drop include of fitsio.h, in is picked up by Libs.h.
* tfarrell  20/06/02  Dropping the last distortion parameter due to new
*                     model.
* tfarrell  18/06/02  Change some of the ModelParsNames values.
* tfarrell  09/05/02  Support multiple wavelengths code and output
*                     in arc-seconds rather then radians.  Add a file
*                     header.
* ssmedley  07/09/01  created
*
*/

/************************************************************************
*   NAME
*	genOffsets - Generate test/dummy ASCII Offset files.
*
*   SYNOPSIS
*       #include "fpcolMdlCalGenOffsets.h"
*
*	void delta
*		(FpcalData	pFpcalData,
*		 double		ra,
*		 double		dec,
*		 double		*deltaRA,
*		 double		*deltaDEC,
*		 StatusType	*pStatus)
*
*	void createOffsets
*		(const char	*pFitsFilename,
*		 const char	*pOffsetFilename,
*		 double		deviation,
*		 double		aParameters[NUM_PARAMETERS],
*		 int		*pStatus)
*
*
*   DESCRIPTION
*
*	The genOffsets utility will generate a dummy/test ASCII Offset file
*	from a specified FITS file. This is useful for testing and ensuring
*	the integrity of the Astrometric Model output by the <fpcal> program.
*
*	delta() determines the offset b/w Ra/Dec coordinates in 2 different
*	models.
*
*	createOffsets() creates a file <pOffsetFilename> of random
*	RA/DEC offsets for each of the fibres specified in the
*	binary table of <pFitsFilename>. The offset values lie in the
*	range +/- <deviation> (arc seconds).
*	This offset file can then be used by the <point> utility to
*	simulate "noise" in the positions of the fibres.
*
*
*   ENVIRONMENT
*
*   CAUTIONS
*
*   EXAMPLES
*
*   SEE ALSO
*
*	For a description of how to use this program, refer to
*	OzPos Calibration Software Manual. (INS-TRE-AUS-13721-0075)
*
*   BUGS
*
*------------------------------------------------------------------------
*/


static const char *rcsId="@(#) $Id: ACMM:ozpozFpcol/ws/src/fpcolMdlCalGenOffsets.c,v 1.63 07-Oct-2008 11:10:26+10 tjf $";

static void *use_rcsId = (0 ? (void *)(&use_rcsId) : (void *) &rcsId);


#define __EXTENSIONS__    /* Pick up extra include file stuff on solaris */
#define _HPUX_SOURCE      /* Pick up extra include file stuff on HP UX */

#include <stdlib.h>	/* atof getopt */
#include <unistd.h>     /* getopt sometimes */
#include <errno.h>	/* perror */
#include <strings.h>	/* strcmp */

#include "fpcolMdlCalConvert.h"
#include "fpcolMdlCalFields.h"
#include "fpcolMdlCalLibs.h"			/* FpcalData getFpcalData */
#include "fpcolMdlCalModel.h"	              /* Model management */
#include "slamac.h"			/* DR2AS */
#include "fpcolMdlCalMacros.h"			/* POS2RD RD2POS */
#include "fpcolMdlCalGetline.h"		/* getline */

#ifdef NON_DETERMINISTIC
#include <sys/types.h>	/* time */
#include <time.h>		/* time */
#endif


/*
 * determine the offset b/w the 2 models.
 */
void delta( 
           FpcalData  *pFpcalData,
           FpcalPars  *trueModelPars,
           FpcalPars  *origModelPars,
           int        object,
           double     *deltaRA,
           double     *deltaDEC,
           StatusType		*pStatus)
{
	double	xPos, yPos, ra2, dec2;

        fpcolMdlCalSetPars(origModelPars, &pFpcalData->model);
	/* use first model to convert RA/DEC -> x/y position */
	RD2POS(fpcolMdlCalWaveModel(&pFpcalData->model,0),
		   pFpcalData->fieldCentreAppRA,
		   pFpcalData->fieldCentreAppDEC,
		   pFpcalData->apparentRA[object],
		   pFpcalData->apparentDEC[object],
		   pFpcalData->observationDate,
		   &xPos,
		   &yPos,
		   pStatus);

	/* now use second model to convert x/y position -> RA/DEC */
        fpcolMdlCalSetPars(trueModelPars, &pFpcalData->model);
	POS2RD(fpcolMdlCalWaveModel(&pFpcalData->model,0),
		   pFpcalData->fieldCentreAppRA,
		   pFpcalData->fieldCentreAppDEC,
		   xPos,
		   yPos,
		   pFpcalData->observationDate,
		   &ra2,
		   &dec2,
		   pStatus);

	/* our new RA/DEC will be slightly different to the original */
	*deltaRA =  pFpcalData->apparentRA[object] - ra2;
	*deltaDEC = pFpcalData->apparentDEC[object] - dec2;
}


/*
 * this function creates a file <pOffsetFilename> of random
 * RA/DEC offsets for each of the fibres specified in the
 * binary table of fits <pFilename> or text file <pFilename> (based
 * on the flag <fits>). The offset values lie in the
 * range +/- <deviation> (arc seconds).
 * This offset file can then be used by the <point> utility to
 * simulate "noise" in the positions of the fibres.
 */
void createOffsets(
                   FpcalData      *fpcalData,
                   char           *pOffsetFilename,
                   double	  deviation,
                   FpcalPars      *trueModelPars,
                   FpcalPars      *origModelPars,
                   StatusType	  *pStatus)
{
	StatusType		status = STATUS__OK;
	double			offsetRA, offsetDEC, r;
	FILE			*pFile;
	int				i;

	if (*pStatus != STATUS__OK)
		return;


	if (strcmp("-", pOffsetFilename) == 0)
		pFile = stdout;
	else
	{
		pFile = fopen(pOffsetFilename, "w");
		if (pFile == NULL)
		{
			perror(pOffsetFilename);
			*pStatus = -1;
			return;
		}
	}
        fprintf(pFile, "#Offset file generated by genOffsets ");
#       if OFFSET_IN_ARCSEC
          fprintf(pFile, "(in arc-seconds)\n");
#       else
          fprintf(pFile, "(in radians)\n");
#       endif
        fprintf(pFile, "GIRAFFE_WLEN %d\n", 
                (int)fpcolMdlCalWavelength(&fpcalData->model, fpcalData->curWaveIndex.Giraffe));
        fprintf(pFile, "UVES_WLEN %d\n",
                (int)fpcolMdlCalWavelength(&fpcalData->model, fpcalData->curWaveIndex.UVES));
        fprintf(pFile, "FACB_WLEN %d\n", 
                (int)fpcolMdlCalWavelength(&fpcalData->model, fpcalData->curWaveIndex.FACB));
 
	for (i = 0; status == STATUS__OK && i < fpcalData->nFibres_a; i++)
	{
		/* determine the offset in position b/w the 2 models */
		delta(fpcalData, trueModelPars, origModelPars, i, &offsetRA, &offsetDEC, &status);

                if (status != STATUS__OK)
                  {
                  fprintf(stderr,"Failed to generate delta for object %d, status %lx\n", 
                          i+1, (long)status);
                  break;
                  }

		/*
		 * generate some artificial noise
		 * in the range +/- <deviation> arc seconds.
		 * DR2AS == number of arcseconds in a radian.
		 */
		r = (double) rand() / (double) RAND_MAX;
		offsetRA += deviation / DR2AS * (r - 0.5);
		r = (double) rand() / (double) RAND_MAX;
		offsetDEC += deviation / DR2AS * (r - 0.5);

		fprintf(pFile, "%3d %18.15f %18.15f\n", 
                        fpcalData->buttons[i], 
#               if OFFSET_IN_ARCSEC
                        RAD2SEC(offsetRA), RAD2SEC(offsetDEC));
#               else
#                   warning "Offset output in radians"
                        offsetRA, offsetDEC);
#               endif
	}

	if (strcmp("-", pOffsetFilename) != 0)
		fclose(pFile);

	if (status != STATUS__OK)
          *pStatus = -1;
}

void usage (const char *pName)
{
	fprintf(stderr, "usage: %s [-options] <file>\n", pName);
	fprintf(stderr, "\navailable options are:\n");
	fprintf(stderr, "-h             help\n");
	fprintf(stderr, "-n <noise>     number of arc seconds to vary the offset by [0.0]\n");
	fprintf(stderr, "-o <file>      output filename [-]\n");
	fprintf(stderr, "-p				prompt for parameters to model\n");
        fprintf(stderr, "-t             <file> is text rather then FITS\n");
	fprintf(stderr, "<file> is a FITS file containing a binary table with fibre button\n");
	fprintf(stderr, "numbers in it or a text file written by fpcal.\n");
}

/*
 *  Names for each of the parameters.  Note the marco's defined within which are
 * checked later against macros from fpcolMdl.h.
 */
char *ModelParNames[] = {
  "X offset      ",   /* 0 */
  "Lin Mdl Par b ",   /* 1 */
  "Lin Mdl Par c ",   /* 2 */
  "Y offset      ",   /* 3 */
  "Lin Mdl Par e ",   /* 4 */
  "Lin Mdl Par f ",   /* 5 */
#define LASTLIN 5
  "Extra Rotation",  /* 6 */
#define ROTPAR  6
  "Extra Scale   ",  /* 7 */
#define SCALEPAR 7
  "NW Dist P1    ",  /* 8 */
#define NW_DIST_FIRST 8
  "NW Dist P2    ",  /* 9 */
  "NW Dist P3    ",  /* 10 */
#define NW_DIST_LAST  10
  "WL Dist P1    ",  /* 11 */
#define W_DIST_FIRST 11
  "WL Dist P2    ",  /* 12 */
#define W_DIST_LAST  12
  "undefined in13 ",  /* 13 */
  "undefined in14",  /* 14 */
  "undefined in15",  /* 15 */
  "undefined in16",  /* 16 */
  "undefined in17",  /* 17 */
  "Non-perpendic ",  /* 18 */
  "undefined in19"   /* 19 */
};

/*
 *  Check against fpcolMdl.h
 */
#if LASTLIN !=  fpcolMDL_PAR_LASTLIN
#error "Parameters definition changed"
#endif
#if ROTPAR  !=  fpcolMDL_PAR_ROT
#error "Parameters definition changed"
#endif
#if SCALEPAR != fpcolMDL_PAR_SCALE  
#error "Parameters definition changed"
#endif
#if NW_DIST_FIRST != fpcolMDL_PAR_NW_DIST_FIRST 
#error "Parameters definition changed"
#endif
#if NW_DIST_LAST != fpcolMDL_PAR_NW_DIST_LAST
#error "Parameters definition changed"
#endif
#if W_DIST_FIRST != fpcolMDL_PAR_W_DIST_FIRST 
#error "Parameters definition changed"
#endif
#if W_DIST_LAST != fpcolMDL_PAR_W_DIST_LAST 
#error "Parameters definition changed"
#endif


int main (int argc, char *argv[])
{
	int		i, c, bDefaultParams = 1;
        StatusType status = STATUS__OK;
	double	noise = 0.0;
	char	*pOutputFilename = "-";	/* stdout */
        FpcalPars       origModelPars;
        FpcalPars       trueModelPars;
        int     fits=1;
	FpcalData		fpcalData;

               
        if ((sizeof(ModelParNames)/sizeof(ModelParNames[0])) != NUM_PARAMETERS)
          {
            fprintf(stderr,"Number of parameters have changed\n");
            fprintf(stderr,"This program needs to be updated appropiately\n");
            fprintf(stderr,"%ld, %ld, %ld, %d\n",
                   (long)sizeof(ModelParNames),
                   (long)sizeof(ModelParNames[0]),
                   (long)(sizeof(ModelParNames)/sizeof(ModelParNames[0])),
                   NUM_PARAMETERS);
 
            exit(1);
          }

	while ((c = getopt(argc, argv, "hn:o:pt")) != EOF)
	{
		switch (c)
		{
			case 'h':
				usage(argv[0]);
				exit(0);
				break;
			case 'n':
				noise = atof(optarg);
				break;
			case 'o':
				pOutputFilename = optarg;
				break;
			case 'p':
				bDefaultParams = 0;
				break;
                        case 't':
                                fits = 0;
                                break;
			case '?':
				usage(argv[0]);
				exit(1);
				break;
                                
		}
	}
	if (argc - optind != 1)
	{
		fprintf(stderr, "Specify 1 FITS file only\n");
		usage(argv[0]);
		exit(1);
	}

#ifdef NON_DETERMINISTIC
	srand(time(NULL));
#endif

        /* Construct the fpcal object */
	fpcalConstructor(&fpcalData);
        
	/* load all relevant information from FITS file */
        if (fits)
          {
            getFpcalData(argv[optind], &fpcalData, &status);
            if (status != STATUS__OK)
              {
                printf("Error reading fits file\n");
                exit(1);
              }
          }
        /* Or from a text file */
        else if (getFpcalDataText(argv[optind], "genOffsets", &fpcalData))
          {
            printf("Error reading text file\n");
            exit(1);
          }

        


	/* create initial model */
        
        status = fpcolMdlCalCreateModels(&fpcalData);
        if (status)
          {
            fprintf(stderr, "Failed to create initial telescope model\n");
            exit(1);
          }


        /*
         * Get the original model parameters and a copy (the trueModelPars).
         */
        trueModelPars = origModelPars = *fpcolMdlCalGetPars(&fpcalData.model);

	if (!bDefaultParams)
	{
		char	buffer[64];

		for (i = 0; i < NUM_PARAMETERS; i++)
                  {
                    printf("Please enter PARS%.02d (%s)[%.6f]: ", i, ModelParNames[i], 
                           trueModelPars.p[i]);
                    if (getline(stdin, buffer, sizeof(buffer)) < 0)
                      break; /* EOF */
                    if (strlen(buffer) != 0)
                      trueModelPars.p[i] = atof(buffer);
                  }
	}
	else
	{
		/* set default arbitrary model parameters  - an adjustment of existing model*/
		trueModelPars.p[0] += 100.5;
		trueModelPars.p[1] *=   0.9999;
		trueModelPars.p[2] *=   0.00005;
		trueModelPars.p[3] +=   0.7;
		trueModelPars.p[4] *=  -0.00005;
		trueModelPars.p[5] *=   0.9999;
	}

	for (i = 0; i < NUM_PARAMETERS; i++)
	{
		printf("PARS%.02d (%s) = %.6f\n", i, ModelParNames[i], trueModelPars.p[i]);
	}

	createOffsets(&fpcalData,
                      pOutputFilename,
                      noise,
                      &trueModelPars,
                      &origModelPars,
                      &status);

        if (status)
          fprintf(stderr, "Program completed with errror\n");
	return status;
}


