
/*******************************************************************************
* E.S.O. - VLT project
*
* "@(#) $Id: ACMM:ozpozFpcol/ws/src/fpcolMdlCalGffo.c,v 1.63 07-Oct-2008 11:10:27+10 tjf $"
*
*
* who       when      what
* --------  --------  ----------------------------------------------
* tfarrell  09/06/02  Support multiple wavelength code changes.
* ssmedley  07/09/01  created
*
*/

/************************************************************************
*   NAME
*	gffo - Generate ASCII Flux Files From ASCII Offset File.
*
*   SYNOPSIS
*       #include "fpcolMdlCalGffo.h"
*
*
*   DESCRIPTION
*
*	The gffo utility will generate a set of ASCII Fibre Flux
*	Files from an ASCII Offset file. It achieves this by creating
*	a 3x3 grid of pixel values about the fibre position. The values
*	in the grid are derived from a gaussian distribution offset
*	from the centre pixel by the specified offset for the fibre.
*
*   ENVIRONMENT
*
*   CAUTIONS
*
*   EXAMPLES
*
*   SEE ALSO
*
*   BUGS
*
*------------------------------------------------------------------------
*/


static const char *rcsId="@(#) $Id: ACMM:ozpozFpcol/ws/src/fpcolMdlCalGffo.c,v 1.63 07-Oct-2008 11:10:27+10 tjf $";

static void *use_rcsId = (0 ? (void *)(&use_rcsId) : (void *) &rcsId);

#define __EXTENSIONS__    /* Pick up extra include file stuff on solaris */
#define _HPUX_SOURCE      /* Pick up extra include file stuff on HP UX */

#include <stdio.h>
#include <math.h>
#include <string.h>	/* strncmp */
#include <stdlib.h>	/* atoi */
#include <assert.h>
#include <slamac.h>	/* DR2AS */

#include "fpcolMdlCalCentroid.h"
#include "fpcolMdlCalLibs.h"
#include "fpcolMdlCalLoadOffsets.h"


void usage (const char *pName)
{
	fprintf(stderr, "usage: %s -options\n", pName);
	fprintf(stderr, "-options:\n");
	fprintf(stderr, "-i <file>    file specifying offsets for each fibre.\n");
	fprintf(stderr, "-pre <pre>   prefix of output flux files. [p]\n");
	fprintf(stderr, "-post <post> postfix of output flux files.[none]\n");
	fprintf(stderr, "-s <spacing> spacing b/w points in grid (arcseconds) [1.0].\n");
	fprintf(stderr, "-v           verbose output mode.\n");
	fprintf(stderr, "-x <x>       number of points in grid in x direction [3].\n");
	fprintf(stderr, "-y <y>       number of points in grid in x direction [3].\n");
	exit(1);
}

int bVerbose = 0;

int main(int argc, char *argv[])
{
	int		xDim = 3, yDim = 3;
	int		i, x, y;
	double	centreRA, centreDEC, stdDev, spacing = 1.0 / DR2AS;
	char	*pre = "p", *post = "";


	/* parse command-line arguments */
	for (i = 1; i < argc; i++)
	{
		if (strcmp(argv[i], "-v") == 0)
		{
			bVerbose = 1;
			continue;
		}
		else if (strcmp(argv[i], "-h") == 0)
			usage(argv[0]);

		if (i + 1 > argc)
			fprintf(stderr, "ignoring superfluous %s argument.\n", argv[i]);

		if (strcmp(argv[i], "-x") == 0)
			xDim = atoi(argv[++i]);
		else if (strcmp(argv[i], "-y") == 0)
			yDim = atoi(argv[++i]);
		else if (strcmp(argv[i], "-s") == 0)
		{
			/*
			 * spacing specified in arcseconds on command-line, but
			 * stored internally in radians.
			 */
			spacing = atof(argv[++i]) / DR2AS;
		}
		else if (strcmp(argv[i], "-pre") == 0)
			pre = argv[++i];
		else if (strcmp(argv[i], "-post") == 0)
			post = argv[++i];
		else if (strcmp(argv[i], "-i") == 0)
		{
			StatusType	status = STATUS__OK;
                        FpcalCurWaveIdx curWaveIndex = { 0, 0, 0};
			loadButtonOffsets(argv[++i], 0, &curWaveIndex, 0, 0, &status);
			if (status != STATUS__OK)
			{
				fprintf(stderr, "error reading offset file: %s\n", argv[i]);
				exit(1);
			}
		}
		else
		{
			fprintf(stderr, "unknown command %s.\n", argv[i]);
			usage(argv[0]);
		}
	}

	if (buttonOffsetsStatus() != BO_LOADED)
	{
		fprintf(stderr, "Please specify offsets file with '-i'.\n");
		exit(1);
	}

	/* determine the centre of the grid in RA/Dec. */
	centreRA = (spacing * (xDim - 1)) / 2;
	centreDEC = (spacing * (yDim - 1)) / 2;
	/* 4 std dev covers 99% of gaussian. */
	stdDev = (((xDim > yDim ? xDim : yDim) - 1) * spacing) / 4;
	if (bVerbose)
		printf("stdDev=%f\n", stdDev);

	for (y = 0; y < yDim; y++)
	{
		for (x = 0; x < xDim; x++)
		{
			FILE		*pFile = NULL;
			char		buffer[128];
			double		toRA, toDEC; /* telescope offset */
			const RaDec	*pOffset = NULL;
                        int             numOffsets;
			/* we're going to be creating xDim*yDim files */
			sprintf(buffer, "%s%d%s", pre, xDim * y + x + 1, post);
			pFile = fopen(buffer, "w");
			if (pFile == NULL)
			{
				perror(buffer);
				exit(1);
			}
			if (bVerbose)
				printf("Writing %s ...\n", buffer);

			/* telescope offset is specified in first line of file */
			toRA = (x * spacing) - centreRA;
			toDEC = (y * spacing) - centreDEC;
			fprintf(pFile, "%f %f\n", toRA, toDEC);

                        numOffsets = getNumOffsets();
                        for (i = 0; i < numOffsets ; ++i);
			while ((pOffset = getOffset(i)) != NULL)
			{
				double	centroidRA = centreRA + pOffset->ra;
				double	centroidDEC = centreDEC + pOffset->dec;
				double	p, flux;

				if ((x == 0 && y == 0) &&	/* only make this check once */
					(fabs(pOffset->ra) > centreRA ||
					 fabs(pOffset->dec) > centreDEC))
				{
					fprintf(stderr, "centroid for fibre %d is outside grid.\n", i);
				}

				/* <p> is distance from centroid to grid coordinate <x>, <y> */
				p = hypot((x * spacing) - centroidRA,
						  (y * spacing) - centroidDEC);
				flux = gaussian(p,
								0.0, /* mean */
								stdDev);	/* std. dev. */
				assert(flux >= 0.0);
				fprintf(pFile, "%d %f\n", i, flux);
				if (bVerbose && 0)
				{
					if (x== 0 && y == 0)
						printf("oidRA=%f oidDEC=%f cRA=%f cDEC=%f\n",
							centroidRA, centroidDEC, centreRA, centreDEC);
				}
			}
			fclose(pFile);
			pFile = NULL;
		}
	}

	return 0;
}
