/*******************************************************************************
* E.S.O. - VLT project
*
* "@(#) $Id: ACMM:ozpozFpcol/ws/src/fpcolMdlCalImg2flux.c,v 1.63 07-Oct-2008 11:10:26+10 tjf $"
*
*
* who       when      what
* --------  --------  ----------------------------------------------
* tfarrell  15/10/02  Include unitstd.h for getopt().
* tfarrell  10/10/02  A report from Matthew Colless seems to indicate
*                     that we sometimes need sys/types.h to get fitsio.h
*                     to compile.
* ssmedley  07/09/01  created
*
*/

/************************************************************************
*   NAME
*	img2flux - Convert a UVES slit image to a set of ASCII Fibre Fluxes
*	Files
*
*   SYNOPSIS
*       #include "fpcolMdlCalImg2Flux.h"
*
*	void image_constructor (Image *pI)
*
*	void image_destructor (Image *pI)
*
*	int getImage (
*		fitsfile	*pFile,
*		const Point	tl,
*		const Point	br,
*		Image		*pImage,
*		int		*pFitsStatus)
*
*	int imageIterator(
*		const Image	*pI,
*		int		(*pFn)(int pixelNo,
*		int		pixel,
*		void		*pData),
*		void		*pData)
*
*	int imageIntensityFn (
*		int		pixelNo,
*		int		pixel,
*		void		*pData)
*
*	int imageIntensity (const Image *pI)
*
*	int showImageFn (
*		int		pixelNo,
*		int		pixel,
*		void		*pData)
*
*	void showImage (const Image *pI)
*
*	int removeBackgroundNoiseFn (
*		int		pixelNo,
*		int		pixel,
*		void		*pData)
*
*	void removeBackgroundNoise(
*		Image		*pImage,
*		int		background)
*
*
*   DESCRIPTION
*
*	This program will extract a UVES slit image embedded in a
*	FITS file & use it to create a set of ASCII Fibre Fluxes
*	Files.
*
*	The intensity/flux for a fibres is determined by the
*	intensity of the fibre in the image. The positions of
*	fibres in the image is specified by the contents of an
*	image positions file. This file also contains the
*	positions/regions of areas within the image to use in
*	background/noise calculations.
*
*	image_constructor() initialises an Image object. Note:
*	should call image_destructor() before using the object for
*	a new/separate image, or when finished with the object.
*
*	image_destructor() cleans-up an Image object. Note that
*	image_constructor() should be called before this object
*	is used again.
*
*	getImage() extracts a sub-image from within a UVES Slit
*	image. <tl> specifies the top-left corner of the region
*	for the sub-image & <br> the bottom-right.
*
*	imageIterator() iterates over entire image, passing each pixel
*	to the function specified by <pFn>. If the return value of
*	<pFn> < 0, then we immediately abort o/wise we return the last
*	return value of <pFn>.
*
*	imageIntensityFn() is the work function of imageIntensity().
*
*	imageIntensity() is simply the sum of all the pixels in the
*	image specified by <pI>.
*
*	showImageFn() is the work function of showImage().
*
*	showImage() uses imageIterator() to display the pixel values
*	in the image specified by <pI>.
*
*	removeBackgroundNoiseFn() is the work function of
*	removeBackgroundNoise().
*
*	removeBackgroundNoise() uses imageIterator() to subtract
*	<background> from every pixel in the image specified by <pImage>.
*
*
*   ENVIRONMENT
*
*   CAUTIONS
*
*   EXAMPLES
*
*   SEE ALSO
*
*   BUGS
*
*------------------------------------------------------------------------
*/


static const char *rcsId="@(#) $Id: ACMM:ozpozFpcol/ws/src/fpcolMdlCalImg2flux.c,v 1.63 07-Oct-2008 11:10:26+10 tjf $";

static void *use_rcsId = (0 ? (void *)(&use_rcsId) : (void *) &rcsId);


#define __EXTENSIONS__    /* Pick up extra include file stuff on solaris */
#define _HPUX_SOURCE      /* Pick up extra include file stuff on HP UX */

/*
 * The resulting binary from this source code accepts a set of FITS
 * files (each containing a UVES Slit image) & generates the appropriate
 * Fibre Fluxes files.
 */
#include <sys/types.h>
#include <assert.h>
#include <stdlib.h>		/* getopt */
#include <strings.h>	/* strncmp */
#include <unistd.h>     /* getopt sometimes */
#include "fitsio.h"

#include "fpcolMdlCalGetline.h"
#include "fpcolMdlCalLoadRegions.h"	/* FibreRegion */

#ifndef bool
typedef int bool;
#endif
#ifndef true
#define true 1
#define false 0
#endif

#define MAX_FIBRES 16

#define NPIXELS(i) ((i).width * (i).height)

typedef struct
{
	int		width;
	int		height;
	int		*aPixels;
	bool	bLoaded;
}Image;

void image_constructor (Image *pI)
{
	pI->width = 0;
	pI->height = 0;
	pI->aPixels = NULL;
	pI->bLoaded = false;
}

void image_destructor (Image *pI)
{
	if (pI->bLoaded && pI->aPixels != NULL)
		free(pI->aPixels);
}

/*
 * Get a sub-image from within the UVES Slit image.
 */
int getImage (
	fitsfile	*pFile,
	const Point	tl,
	const Point	br,
	Image		*pImage,
	int			*pFitsStatus)
{
	int	nullValue = 0;
	int	dummy = -1;
	int firstPixel;
	int	y;
	int	nCols = -1;	/* not yet determined */

	if (*pFitsStatus != 0)
		return 0;

	assert(tl.y >= br.y);
	assert(tl.x <= br.x);
	assert(pImage != NULL);

	if (pImage->bLoaded == true)
	{
		image_destructor(pImage);
		image_constructor(pImage);
	}

	assert(pImage->bLoaded == false);
	assert(pImage->aPixels == NULL);

	/* determine the xdimension of the FITS image */
	fits_read_key(pFile, TINT, "NAXIS1", &nCols, NULL, pFitsStatus);
	if (*pFitsStatus != 0)
		return 2;

	assert(nCols > 0);
	assert(br.x <= nCols);

	pImage->width = br.x - tl.x + 1;
	pImage->height = tl.y - br.y + 1;
	pImage->aPixels = (int *) malloc(pImage->width *
									 pImage->height *
									 sizeof(int));
	pImage->bLoaded = true;

	/*
	printf("nCols is %d\nw is %d\nh is %d\n",
		   nCols,
		   pImage->width,
		   pImage->height);
	*/

	for (y = tl.y; *pFitsStatus == 0 && y >= br.y; y--)
	{
		/* firstPixel = (y * nCols) + tl.x; */
		firstPixel = xy2index(tl.x, y, nCols);
		fits_read_img(pFile,
					  TINT,
					  firstPixel,
					  pImage->width,
					  &nullValue,
					  &(pImage->aPixels[(tl.y - y) * pImage->width]),
					  &dummy,
					  pFitsStatus);
	}

	return 0;
}

/*
 * This function iterates over entire image, passing each pixel
 * to the function specified by <pFn>.
 * If the return value of <pFn> < 0, then we immediately abort.
 * o/wise we return the last return value of <pFn>.
 */
int imageIterator(
	const Image	*pI,
	int			(*pFn)(int			pixelNo,
					   int			pixel,
					   void			*pData),
	void		*pData)
{
	int x, y, pixelNo, rv;

	assert(pI != NULL);
	assert(pI->bLoaded == true);
	assert(pI->aPixels != NULL);
	assert(pI->width > 0);
	assert(pI->height > 0);

	for (y = 0; y < pI->height; y++)
	{
		for (x = 0; x < pI->width; x++)
		{
			pixelNo = y * pI->width + x;
			rv = (*pFn)(pixelNo, pI->aPixels[pixelNo], pData);
			if (rv < 0)
				return rv;
		}
	}
	return rv;
}

/*
 * work function of imageIntensity.
 */
int imageIntensityFn (
	int			pixelNo,
	int			pixel,
	void		*pData)
{
	int *pII = (int *) pData;

	*pII += pixel;
	(void) pixelNo;
	return 0;
}

/*
 * The image intensity is simply the sum of all the pixels in the image.
 */
int imageIntensity (const Image *pI)
{
	int rv;
	int	ii = 0;	/* variable storing image intensity */

	rv = imageIterator(pI, imageIntensityFn, (void *) &ii);
	if (rv < 0)
		return rv;

	return ii;
}

/*
 * work function of showImage.
 */
int showImageFn (
	int			pixelNo,
	int			pixel,
	void		*pData)
{
	Image *pImage = (Image *) pData;
	printf("%04d ", pixel);
	if (pixelNo % pImage->width == pImage->width - 1)
		printf("\n");
	return 0;
}

/*
 * Display a textual representation of the image on stdout.
 */
void showImage (const Image *pI)
{
	imageIterator(pI, showImageFn, (void *) pI);
}

/*
 * work function of removeBackgroundNoise.
 */
int gv;	/* TODO: global variable - ick! */

int removeBackgroundNoiseFn (
	int		pixelNo,
	int		pixel,
	void	*pData)
{
	Image	*pImage = (Image *) pData;

	pImage->aPixels[pixelNo] -= gv;
	(void) pixel;
	return 0;
}

/*
 * Remove background noise from all pixels in an image.
 */
void removeBackgroundNoise(
	Image	*pImage,
	int		background)
{
	gv = background;
	imageIterator(pImage, removeBackgroundNoiseFn, (void *) pImage);
}

void usage (const char *pName)
{
	fprintf(stderr, "usage: %s [-options] <FITSFile1> ... <FITSFileN>\n", pName);
	fprintf(stderr, "\noptions:\n");
	fprintf(stderr, "-h           help\n");
	fprintf(stderr, "-i <file>    file containing image positions [img.positions]\n");
	fprintf(stderr, "-p <prefix>  output filename prefix [f]\n");
	fprintf(stderr, "-p <postfix> output filename postfix [none]\n");
	fprintf(stderr, "-v           verbose output\n");
}

int main (int argc, char *argv[])
{
	fitsfile	*pFile = NULL;
	int			fitsStatus = 0;
	int			i;
	Image		image;
	int			c;
	char		*pPositionsFile = "img.positions";
	FILE		*pOutputFile = stdout;
	/*
	 * a fibre region is the area within the image in which the
	 * fibre is contained.
	 */
	FibreRegion	aFibreRegions[MAX_FIBRES];
	FibreRegion	bkgRegion1, bkgRegion2;
	int			nRegions;
	int			background;
	bool		bVerbose = false;
	char		*pre = "f";
	char		*post = "";

	while ((c = getopt(argc, argv, "i:hp:P:v")) != EOF)
	{
		switch (c)
		{
			case 'i':
				pPositionsFile = optarg;
				break;
			case 'h':
				usage(argv[0]);
				exit(0);
				break;
			case 'p':
				pre = optarg;
				break;
			case 'P':
				post = optarg;
				break;
			case 'v':
				bVerbose = true;
				break;
			case '?':
				usage(argv[0]);
				exit(1);
				break;
		}
	}

	if (optind >= argc)
	{
		fprintf(stderr, "Must specify at least 1 FITS file.\n");
		usage(argv[0]);
		exit(2);
	}

	image_constructor(&image);

	/* load all image regions, including background region */
	if (loadImageLocations(pPositionsFile,
						   aFibreRegions,
						   &bkgRegion1,
						   &bkgRegion2,
						   &nRegions) != 0)
	{
		fprintf(stderr, "Couldn't load image regions from %s.\n", pPositionsFile);
		return 2;
	}

	assert(nRegions <= MAX_FIBRES);
	printf("Loaded %d fibre regions\n", nRegions);

	for (i = optind; fitsStatus == 0 && i < argc; i++)
	{
		int				r;
		unsigned int	bpp;
		char			filename[128];
		int				avgIntensity1, avgIntensity2;

		/* open the FITS file */
		fits_open_file(&pFile, argv[optind], READONLY, &fitsStatus);

		/*
		 * all of this code assumes that the 'int' type is able to
		 * contain the pixel values. A simple check will ensure that
		 * this is true.
		 */
		fits_read_key(pFile, TINT, "BITPIX", &bpp, NULL, &fitsStatus);
		assert(bpp <= sizeof(int) * 8);

		/* extract telescope offset from comment fields of header. */
		/* COMMENT TELESCOPE OFFSET rrr.rr ddd.dd
		TODO
		while (1)
		{
			char	comment[128], dummy[128];
			fits_read_key(pFile, TSTRING, "COMMENT", dummy, comment, &fitsStatus);
			if (fitsStatus != 0)
			{
				fprintf(stderr, "Couldn't find COMMENT containing telescope offset\n");
				break;
			}
			printf("read comment: %s ... %s\n", comment, dummy);
		}
		*/
		sprintf(filename, "%s%d%s", pre, i - optind + 1, post);
		pOutputFile = fopen(filename, "w");
		if (pOutputFile == NULL)
		{
			perror(filename);
			return 4;
		}
		printf("Writing %s ...\n", filename);
		fprintf(pOutputFile, "0 0 TELESCOPE OFFSET NYI\n");

		/* get the first background sub-image */
		getImage(pFile,
				 bkgRegion1.tl,
				 bkgRegion1.br,
				 &image,
				 &fitsStatus);
		if (fitsStatus != 0)
		{
			fprintf(stderr, "Could not find image in FITS file.\n");
			break;
		}
		avgIntensity1 = imageIntensity(&image) / NPIXELS(image);
		printf("Background 1 is %d\n", avgIntensity1);

		/* get the second background sub-image */
		getImage(pFile,
				 bkgRegion2.tl,
				 bkgRegion2.br,
				 &image,
				 &fitsStatus);
		avgIntensity2 = imageIntensity(&image) / NPIXELS(image);
		printf("Background 2 is %d\n", avgIntensity2);

		/*
		 * The purpose of having 2 background regions is that
		 * we're trying 2 cater for any vertical gradient in the
		 * background noise.
		 * So, take the average value.
		 */
		background = (avgIntensity1 + avgIntensity2) / 2;
		printf("Average background is %d\n", background);

		for (r = 0; r < nRegions; r++)
		{
			int ii;

			/* get the sub-image for the fibre */
			getImage(pFile,
					 aFibreRegions[r].tl,
					 aFibreRegions[r].br,
					 &image,
					 &fitsStatus);
			/* remove background noise from all pixels */
			removeBackgroundNoise(&image, background);
			if (bVerbose)
				showImage(&image);

			/* determine image intensity within sub-image */
			ii = imageIntensity(&image);
			printf("image intensity for fibre %d is: %d\n", aFibreRegions[r].fibreNo, ii);
			fprintf(pOutputFile, "%d %d\n", aFibreRegions[r].fibreNo, ii);
		}
		fits_close_file(pFile, &fitsStatus);
	}

	if (fitsStatus != 0)
	{
		fits_report_error(stderr, fitsStatus);
		return 3;
	}

	image_destructor(&image);

	return 0;
}
