
/*******************************************************************************
* E.S.O. - VLT project
*
* "@(#) $Id: ACMM:ozpozFpcol/ws/src/fpcolMdlCalKgGaussFit.c,v 1.63 07-Oct-2008 11:10:26+10 tjf $"
*
*
* who       when      what
* --------  --------  ----------------------------------------------
* tfarrell  19/10/02  Added fibre number for error messages.
* ssmedley  07/09/01  created
*
*/

/************************************************************************
*   NAME
*	PDLGaussianFit - Fit a set of 1d data points to a Gaussian.
*
*   SYNOPSIS
*       #include "fpcolMdlCalKgGaussFit.h"
*
*	int PDLGaussianFit (
*               const int               fibre,
*		const int		nPoints,
*		const double		*aData,
*		double			*pCentre,
*		double			*pFWHM,
*		double			*pBackground,
*		double			*pPeak,
*		double			bgNoise)
*
*
*   DESCRIPTION
*
*	Taken from Karl Glazebrook's PDL::Fit::Gaussian module.
*
*	Fit's a 1D radial Gaussian robustly free parameters are the peak
*	height, FWHM. Centre is assumed to be X=0 (i.e. start of piddle).
*	The background is NOT fit, because I find this is generally
*	unreliable, rather a median is determined in the 'outer' 10% of
*	pixels (i.e. those at the end of the data piddle). The initial
*	estimate of the FWHM is the length of the piddle/3, so it might fail
*	if the piddle is too long. (This is non-robust anyway). Most data
*	does just fine and this is a good default gaussian fitter.
*
*	If <bgNoise> < 0 then this routine will try to determine the
*	background noise itself. O/wise, the value of <bgNoise> is used.
*
*	A test gaussian fitting program is available if this file is
*	compiled with a -DTEST_KGGAUSSFIT compilation flag.
*
*
*   ENVIRONMENT
*
*   CAUTIONS
*
*   EXAMPLES
*
*   SEE ALSO
*
*   BUGS
*
*------------------------------------------------------------------------
*/


static const char *rcsId="@(#) $Id: ACMM:ozpozFpcol/ws/src/fpcolMdlCalKgGaussFit.c,v 1.63 07-Oct-2008 11:10:26+10 tjf $";

static void *use_rcsId = (0 ? (void *)(&use_rcsId) : (void *) &rcsId);



#include <math.h>

#include "fpcolMdlCalGauss.c"
#include "fpcolMdlCalKgGaussFit.h"

#ifdef TEST_KGGAUSSFIT
#define MAX_DATA	1024
#endif

/*
 * If <bgNoise> < 0 then this routine will try to determine the
 * background noise itself. O/wise, the value of <bgNoise> is used.
 */

int PDLGaussianFit (
        const int               fibre,
	const int		nPoints,
	const double	*aData,
	double			*pCentre,
	double			*pFWHM,
	double			*pBackground,
	double			*pPeak,
	double			bgNoise)
{
	int		i, nb, rv;
	double	ymax = -1e-30, val, xcenguess, bkg;
	double	par[NPAR], a[NPAR][NPAR];
	double	ytmp[nPoints];
	double	yytmp[nPoints];
	double	rtmp[nPoints];
	double	sig[nPoints];
	double	xval[nPoints];
	double	aDatafit[nPoints];

	for (i = 0; i < nPoints; i++)
	{
		ytmp[i] = aData[i];
		xval[i] = i;
		sig[i] = 1.0; /* Room for expansion */

		if (aData[i] > ymax)   /* peak value */
			ymax = aData[i];
	}

	if (bgNoise < 0.0)
	{
		/* Find background points - outer 10% */
		nb = 0;
		for (i = (int) (0.9 * (double) nPoints); i < nPoints; i++)
		{
			yytmp[nb] = ytmp[i];
			nb++;
		}

		/* Estimate background */
		bkg = 0.0;
		if (nb > 0)
		{
			lqsortD(yytmp, 0, nb - 1);
			i = (nb - 1) / 2;
			bkg = yytmp[i]; /* Median */
		}
	}
	else
		bkg = bgNoise;

	/* Estimate xcentroid */
	val = 0.0;
	xcenguess = 0.0;
	for (i = 0; i < nPoints; i++)
	{
		ytmp[i] -= bkg;	/* subtract background noise */
		xcenguess += ytmp[i] * i;
		val += ytmp[i];
	}
	xcenguess /= val;

	par[2] = xcenguess;
	par[1] = ymax - bkg;
	par[0] = nPoints / 3; /* FWHM guesstimate: 1/3 of given box */

	/* Do the fit */
	rv = marquardt(nPoints, 3, xval, ytmp, sig, par, rtmp, a);
	if (rv != 0)
		fprintf(stderr, "marquardt terminated with return value ***%d***, fibre %d\n", 
                        rv, fibre);

	*pCentre = par[2];
	*pFWHM = (fabs(par[0])) * 2.0 * sqrt(log(2.0)); /* Ret Values */

	*pPeak = par[1];
	*pBackground = bkg;

	for (i = 0; i < nPoints; i++)
	{
		val = ( (double) i - par[2] ) / par[0];
		aDatafit[i] = par[1] * exp (- val * val) + bkg;
	}

	/*
	 * printf("centre=%f\nfwhm=%f\n", *pCentre, *pFWHM);
	 * printf("peak_ht=%f\nbackground=%f\n", *pPeak, *pBackground);
	 */

	return rv;
}

#ifdef TEST_KGGAUSSFIT
int main (int argc, char *argv[])
{
	double	data[MAX_DATA];
	int		i, nPoints = 0;
	double	centre, fwhm, peak, bg;


	if (argc == 1)
	{
		printf("usage: %s 1.11 2.22 3.33 4.44 .. n.nn\n", argv[0]);
		exit(0);
	}

	for (i = 1; i < argc; i++)
	{
		data[nPoints++] = atof(argv[i]);
	}

	for (i = 0; i < 10000; i++)
	{
		if (PDLGaussianFit(nPoints,
					   data,
					   &centre,
					   &fwhm,
					   &bg,
					   &peak,
					   0.0) != 0)
		{
			exit(1);
		}
	}

	printf("centre=%f\n", centre);
	printf("fwhm=%f\n", fwhm);
	printf("bg=%f\n", bg);
	printf("peak=%f\n", peak);

	return 0;
}
#endif
