
/*******************************************************************************
* E.S.O. - VLT project
*
* "@(#) $Id: ACMM:ozpozFpcol/ws/src/fpcolMdlCalLibs.c,v 1.63 07-Oct-2008 11:10:25+10 tjf $"
*
*
* who       when      what
* --------  --------  ----------------------------------------------
* tfarrell  20/10/02  Go back to SVD as the default fit method.
* tfarrell  18/10/02  Support powellMode in constructor.
* tfarrell  10/10/02  Drop include of fitsio.h and status.h, they are picked up by Libs.h.
* tfarrell  22/08/02  In calculateOffsets() - If an offset entry does not exist, 
*                     then don't create an empty entry for it, instead just decrement
*                     the number of offsets.  This supports clearing offsets.
* tfarrell  21/08/02  Support residualsInfo for each offset.
* tfarrell  27/06/02  Add fpcolMdlCalCheckFitType() fpcolMdlCalFindButton().
* tfarrell  23/06/02  Header now has Ra/Dec in simple units of degrees
*                     rather then the ddmmss.dd format.
* tfarrell  09/05/02  Support mutliple wavelengths.  Fix some bugs.
* tfarrell  12/04/02  Add getFpcalData and support functions.
* ssmedley  07/09/01  created
*
*/

/************************************************************************
*   NAME
*	fpcalConstructor fpcalDestructor getFpcalData calculateOffsets -
*	Module to store/load/manipulate data used during FLAMES/OzPoz
*	calibration & subsequently to create an astrometric model.
*
*   SYNOPSIS
*       #include "fpcolMdlCalLibs.h"
*
*	struct { ... }FpcalData;
*
*	void fpcalConstructor (FpcalData *pFpcalData);
*
*	void fpcalDestructor (FpcalData *pFpcalData);
*
*	void getFpcalData (
*		const char	*pFitsFilename,
*		FpcalData	*pFpcalData,
*		StatusType	*pStatus);
*
*	void calculateOffsets (
*		FpcalData	*pFpcalData,
*		StatusType	*pStatus);
*
*       int int getFpcalDataText(
*                     char *fileName,
*                     char *cmdName,
*                     FpcalData	*pFpcalData);
*
*   DESCRIPTION
*
*	fpcalConstructor() will initialise an FpcalData object. You
*	should call fpcalDestructor() whenever you are done with it, or
*	use the object for another set of data.
*
*	fpcalDestructor() tidies up a used FpcalData object. After a
*	call to this function, the object should no longer be used until
*	fpcalConstructor() has been invoked for the object.
*
*	getFpcalData() extracts all relevant data from the FITS file
*	specified by <pFitsFilename> for creating an astrometric model.
*	Note that the FITS fields that are required are specified by
*	#define's in the fpcolMdlCalFields.h header file. If successful,
*	the appropriate data will reside in <pFpcalData>.
*
*	calculateOffsets() will calculate/set the offsets for each fibre
*	button in <pFpcalData>. Note that the offsets are (usually) specified
*	in a separate text file & must be loaded beforehand. (see
*	loadButtonOffsets() in fpcolMdlCalLoadOffsets.h).
*
*   ENVIRONMENT
*
*   CAUTIONS
*
*   EXAMPLES
*
*   SEE ALSO
*
*   BUGS
*
*------------------------------------------------------------------------
*/


#include <stdlib.h>	/* malloc */
#include <stdio.h>	/* fopen */
#include <errno.h>
#include <strings.h>	/* strncmp */
#include <slalib.h>	/* slaMappa slaMapqk */
#include <assert.h>

#include "fpcolMdlCalLibs.h"
#include "fpcolMdl.h"
#include "fpcolMdlCalLoadOffsets.h"
#include "fpcolMdlCalConvert.h"	/* meanToApparent toRadians */
#include "fpcolMdlCalFields.h"
#include "fpcolMdlCalModel.h"


#define NUM_ELEMENTS(array) (sizeof(array)/sizeof(array[0]))

/*
 * Mallocs the data arrays used to store the object information.
 */
static int fpcalMallocObjectArrays(int nFibres, FpcalData *pFpcalData)
{
  const unsigned int	arraySpace = nFibres * sizeof(double);
  pFpcalData->apparentRA       = (double *) malloc(arraySpace);
  pFpcalData->apparentDEC      = (double *) malloc(arraySpace);
  pFpcalData->meanRA           = (double *) malloc(arraySpace);
  pFpcalData->meanDEC          = (double *) malloc(arraySpace);
  pFpcalData->properMotionRA   = (double *) malloc(arraySpace);
  pFpcalData->properMotionDEC  = (double *) malloc(arraySpace);
  pFpcalData->buttons          = (int *)    malloc(nFibres  * sizeof(int));

  /* 
   * check for memory allocation errors 
   */
  if (pFpcalData->apparentRA      == NULL ||
      pFpcalData->apparentDEC     == NULL ||
      pFpcalData->meanRA          == NULL ||
      pFpcalData->meanDEC         == NULL ||
      pFpcalData->properMotionRA  == NULL ||
      pFpcalData->properMotionDEC == NULL ||
      pFpcalData->buttons         == NULL)
    {
      fprintf(stderr, "Could not malloc\n");

      /*
       * Tidy up anything we actually got.
       */
      if (pFpcalData->apparentRA)  { free(pFpcalData->apparentRA)  ; pFpcalData->apparentRA = 0;  }
      if (pFpcalData->apparentDEC) { free(pFpcalData->apparentDEC) ; pFpcalData->apparentDEC = 0; }
      if (pFpcalData->meanRA)      { free(pFpcalData->meanRA)      ; pFpcalData->meanRA = 0;      }
      if (pFpcalData->meanDEC)     { free(pFpcalData->meanDEC)     ; pFpcalData->meanDEC = 0;     }
      if (pFpcalData->buttons)     { free(pFpcalData->buttons)     ; pFpcalData->buttons = 0;     }
      if (pFpcalData->properMotionRA)
        { 
          free(pFpcalData->properMotionRA) ; 
          pFpcalData->properMotionRA = 0;
        }
      if (pFpcalData->properMotionDEC) 
        { 
          free(pFpcalData->properMotionDEC) ; 
          pFpcalData->properMotionDEC = 0; 
        }

      pFpcalData->nFibres_a = -1;
      pFpcalData->nOffsets = 0;
      return -1;
    }
  return 0;
}



/*
 * Return the index in buttons of the specified buttonNo.
 */
extern int fpcolMdlCalFindButton(
                        const int nButtons, 
                        const int *buttons, 
                        const int buttonNo)
{
  int  i;
  if ((nButtons == 0)||(!buttons))
    return -1;

  for (i = 0; i < nButtons ;  ++i)
    {
      if (buttons[i] == buttonNo)
        return i;
    }
  return -1;
}
/*
 * This function sets the offsets for each fibre button.
 * Note: make sure button offsets have been loaded before
 *       calling this function.
 */
void calculateOffsets (FpcalData *pFpcalData, StatusType *pStatus)
{
        int numOffsets;
        int offsetIndex;
        int notFound;
	const RaDec	*pOffset = NULL;

	if (*pStatus != STATUS__OK)
		return;

#ifdef AAO_DEBUG
	/* make sure button offsets have been loaded. */
	if (buttonOffsetsStatus() != BO_LOADED)
	{
		fprintf(stderr, "error: button offsets not loaded!\n");
		*pStatus = -1;
		return;
	}
#endif

	if (pFpcalData->offsetRA != NULL)
	{
		free(pFpcalData->offsetRA);
		free(pFpcalData->offsetDEC);
                free(pFpcalData->offsetWaveIndex);
                free(pFpcalData->offsetButtonNo);
                free(pFpcalData->residualsInfo);
	}

        /*
         * Get the number of offsets we have.
         */
        pFpcalData->nOffsets = numOffsets = getNumOffsets();

	pFpcalData->offsetRA  = (double *) malloc(numOffsets * sizeof(double));
	pFpcalData->offsetDEC = (double *) malloc(numOffsets * sizeof(double));
        pFpcalData->offsetWaveIndex = (int *) malloc(numOffsets * sizeof(int));
        pFpcalData->offsetButtonNo  = (int *) malloc(numOffsets * sizeof(int));
        pFpcalData->residualsInfo   = (FpcalResidualsInfo *) malloc(numOffsets * 
                                                                    sizeof(FpcalResidualsInfo));
        /* Zero the residuals - they won't be set here */
        memset(pFpcalData->residualsInfo, 0, (numOffsets * sizeof(FpcalResidualsInfo)));
        pFpcalData->fitted = 0;  /* Indicates residuals are not valid */

        notFound = 0;            /* Offsets not found */

        for (offsetIndex = 0; offsetIndex < numOffsets ; ++offsetIndex)
          {
          /* retrieve offset RA/DEC for each fibre button */
          pOffset = getOffset(offsetIndex);
          if (pOffset == NULL)
              {
              /*
               * We don't want these in the system - probably happened due to an
               * offset being cleared.  offsetIndex needs to continue increasing so
               * that getOffset works, but we don't want to increase the index
               * in the pFpcalData offset arrays, so increment the notFound counter
               * which allows this to work.
               */
              fprintf(stderr, "Offset %d not found - will ignore.\n",offsetIndex);
              ++notFound;
              continue;
              }
          pFpcalData->offsetRA[offsetIndex-notFound]        = pOffset->ra;
          pFpcalData->offsetDEC[offsetIndex-notFound]       = pOffset->dec;
          pFpcalData->offsetWaveIndex[offsetIndex-notFound] = pOffset->wIndex;
          pFpcalData->offsetButtonNo[offsetIndex-notFound]  = pOffset->button;
          }
        /*
         * Reduce the number of offsets by the number not found;
         */
        pFpcalData->nOffsets -= notFound;

        if (pFpcalData->nOffsets <= 0)
            {
            fprintf(stderr, "All offsets have been cleared\n");
            *pStatus = -1;
            }
}

#warning "Fetching of proper montion from binary table currenlty disabled"
#define NOPMS

int workFn
	(long			totalN,
	 long			offset,
	 long			firstN,
	 long			nValues,
	 int			nArrays,
	 iteratorCol	*data,
	 void			*pUserData)
{
        double		*aRA = NULL, *aDEC = NULL;
        int		i, *aButtons = NULL, abSeen[512];
	FpcalData	*pFpcalData = (FpcalData *) pUserData;
#ifndef NOPMS
        double *aPMRA = NULL, *aPMDEC = NULL;
#endif

	/*
	 * note that first element of array is a sentinel value
	 * denoting empty field & is not used
	 */
	memset(abSeen, 0, sizeof(abSeen));
	pFpcalData->nFibres_a = 0;

	assert(totalN == nValues);
	assert(firstN == 1);
#ifdef NOPMS
	assert(nArrays == 3); /* No proper motions - we expect only 3 columns */
#else
	assert(nArrays == 5);
#endif

	/* 
         * allocate memory to store RA/DEC coordinates 
         */
        if (fpcalMallocObjectArrays(nValues, pFpcalData) != 0)
          return -1;

	aRA = fits_iter_get_array(&data[0]);
	aDEC = fits_iter_get_array(&data[1]);
	aButtons = fits_iter_get_array(&data[2]);
#ifndef NOPMS
	aPMRA = fits_iter_get_array(&data[3]);
	aPMDEC = fits_iter_get_array(&data[4]);
#endif

	assert(aRA != NULL);
	assert(aDEC != NULL);
	assert(aButtons != NULL);
#ifndef NOPMS
	assert(aPMRA != NULL);
	assert(aPMDEC != NULL);
#endif

	/*
	 * Note that in the following calculations we are
	 * _excluding_ the sentinel value in the first element
	 * of the array.
	 */
	for (i = 1; i <= nValues; i++)
	{
		int index = pFpcalData->nFibres_a;

		if (abSeen[aButtons[i]] != 0)
		{
			fprintf(stderr, "Multiple entries for button fibre #%d - ignoring\n", aButtons[i]);
			continue;
		}
		abSeen[aButtons[i]] = 1;

		/* according to JAB: highest known PM is 10 arcsec/yr! */
#ifdef NOPMS
                pFpcalData->properMotionRA[index] = 0;
                pFpcalData->properMotionDEC[index] = 0;
#else
		assert(aPMRA[i] < 60.0);
		assert(aPMDEC[i] < 60.0);
		/* proper motion values are in arcseconds. convert to radians */
		pFpcalData->properMotionRA[index] = toFractions(0, 0, aPMRA[i]);
		pFpcalData->properMotionDEC[index] = toFractions(0, 0, aPMDEC[i]);
#endif
                /*pFpcalData->meanRA[index]  = toRadians(aRA[i]);
                  pFpcalData->meanDEC[index] = toRadians(aDEC[i]);*/
                pFpcalData->meanRA[index]  = aRA[i]/180.0*M_PI;
                pFpcalData->meanDEC[index] = aDEC[i]/180.0*M_PI;

		/* convert RA/DEC coordinates from mean to apparent */
		meanToApparent(aRA[i]/180.0*M_PI,	/* in: mean RA */
                               aDEC[i]/180.0*M_PI,	/* in: mean DEC */
                               pFpcalData->properMotionRA[index],
                               pFpcalData->properMotionDEC[index],
                               pFpcalData->meanAppParams,
                               &(pFpcalData->apparentRA[index]),	/* out: apparent RA */
                               &(pFpcalData->apparentDEC[index]));/* out: apparent DEC */

		pFpcalData->buttons[index] = aButtons[i];
		pFpcalData->nFibres_a++;
	}

	/* stop compiler warnings */
	(void) totalN;
	(void) offset;
	(void) firstN;
	(void) nArrays;
	return 0;
}


/*
 * Read a single keyword and report any failure
 */
static void readKeyword(
                        fitsfile  *pFile,
                        int       typeCode,
                        int       ignoreError,
                        char      *fieldName,
                        void      *data,
                        int       *firstFailure)
{
  int fitsStatus = 0;
  fits_read_key(pFile, typeCode, fieldName, data,
                NULL, &fitsStatus);
  if ((fitsStatus != 0)&&(!ignoreError))
    {
      fprintf(stderr,"Error reading keyword \"%s\"\n", fieldName);
      if (*firstFailure == 0)
        *firstFailure = fitsStatus;
    }
}
/*
 * read relevant header information into <FpcalData> structure.
 */
void readHeader
	(fitsfile	*pFile,
	 FpcalData	*pFpcalData,
	 int		*pStatus)
{
	int i;
        FpcalPars pars;
        int fitsStatus = 0;
        if (pFpcalData->fileName)
            free(pFpcalData->fileName);
        fits_read_key_longstr(pFile, FILENAME_FIELD,  &pFpcalData->fileName, 
                              0, &fitsStatus);
        if (fitsStatus != 0)
            {
            fprintf(stderr, "Error reading keyword \"%s\"\n", FILENAME_FIELD);
            *pStatus = fitsStatus;
            }
        else
            {
            printf("File name is \"%s\"\n", pFpcalData->fileName);
            }
	/*
	 * now extract state information about the data.
	 * atmospheric conditions, telescope position, etc.
	 */
	readKeyword(pFile, TDOUBLE, 0 , TEMP_FIELD, &(pFpcalData->atmTemp),
				  pStatus);
	readKeyword(pFile, TDOUBLE, 0, PRESSURE_FIELD, &(pFpcalData->atmPressure),
				  pStatus);
	readKeyword(pFile, TDOUBLE, 0, HUMIDITY_FIELD, &(pFpcalData->atmHumidity),
				  pStatus);
	readKeyword(pFile, TDOUBLE, 0, EQUINOX_FIELD, &(pFpcalData->equinox),
				  pStatus);
	readKeyword(pFile, TDOUBLE, 0, OBS_DATE_FIELD, &(pFpcalData->observationDate),
				  pStatus);
	readKeyword(pFile, TINT, 0, PLATE_FIELD, &(pFpcalData->fieldPlate),
				  pStatus);
	readKeyword(pFile, TDOUBLE, 0, CENTRE_RA_FIELD, &(pFpcalData->fieldCentreMeanRA),
				  pStatus);
	readKeyword(pFile, TDOUBLE, 0, CENTRE_DEC_FIELD, &(pFpcalData->fieldCentreMeanDEC),
				  pStatus);

	/*
	 * convert the field centre RA/DEC to radians & then
	 * convert from mean to apparent.
	 */
	if (*pStatus == 0)
	{
		/*
		 * the <meanToApparent> routine requires <meanAppParams> so they must
		 * be set up beforehand.
		 */
        /*
          pFpcalData->fieldCentreMeanRA  = toRadians(pFpcalData->fieldCentreMeanRA);
          pFpcalData->fieldCentreMeanDEC = toRadians(pFpcalData->fieldCentreMeanDEC);
        */
          pFpcalData->fieldCentreMeanRA  = pFpcalData->fieldCentreMeanRA/180.0*M_PI;
          pFpcalData->fieldCentreMeanDEC = pFpcalData->fieldCentreMeanDEC/180.0*M_PI;
          slaMappa(pFpcalData->equinox,
                   pFpcalData->observationDate,
                   pFpcalData->meanAppParams);
          
          meanToApparent(pFpcalData->fieldCentreMeanRA,
                         pFpcalData->fieldCentreMeanDEC,
                         0.0,	/* no proper motion for field centre */
                         0.0,
                         pFpcalData->meanAppParams,
                         &(pFpcalData->fieldCentreAppRA),
                         &(pFpcalData->fieldCentreAppDEC));
	}
        
        /*
         *  ADC position angle is only output when the ADC is used.
         */ 
        if (*pStatus == 0)
          {
            readKeyword(pFile, TDOUBLE, 1, POSANGLE_FIELD, &(pFpcalData->posAngle),
                        pStatus);
            if (*pStatus != 0) 
              {
                pFpcalData->posAngle = 0;
                *pStatus = 0;
                fits_clear_errmsg();
              }
          }
        if (*pStatus == 0)
          {
            /*
             * Read each of the three wavelengths in the file.
             */
            double wavelength;
            int index;
            char *fields[] = {GIRA_WLEN_FIELD, UVES_WLEN_FIELD, FACB_WLEN_FIELD};
            int  *curIndex[] = { &pFpcalData->curWaveIndex.Giraffe,
                                 &pFpcalData->curWaveIndex.UVES,
                                 &pFpcalData->curWaveIndex.FACB };

            for (i = 0; i < 3; ++i) {
              readKeyword(pFile, TDOUBLE, 0, fields[i], &wavelength, pStatus);
              if (*pStatus != 0)
                {
                  fprintf(stderr, 
                          "Current ignoring failure to read wavelength keyword %s\n", 
                          fields[i]);
                  wavelength = 450;
                  *pStatus = 1;
                }
              if ((index = fpcolMdlCalNewWavelength(wavelength, pFpcalData)) < 0)
                {
                  fprintf(stderr, "Failure adding wavelength %g to internal tables\n", wavelength);
                  *pStatus = 1;
                  return;
                }
              /*
               * This becomes the new current wavelength.
               */
              *curIndex[i] = index;
            }
          }

	/* pars (model parameters) values */
	for (i = 1; i <= NUM_PARAMETERS; i++)
	{
		char	fieldName[16];
		sprintf(fieldName, "%s%d", PARS_PREFIX, i);

		readKeyword(pFile, TDOUBLE, 0, fieldName, &pars.p[i-1], pStatus);

                if (*pStatus != 0)
                  {
                    fprintf(stderr,
                      "Warning - telescope model parameter %s not found in header, set to zero\n",
                            fieldName);
                    pars.p[i-1] = 0;
                    *pStatus = 0;
                    fits_clear_errmsg();
                  }


	}
        fpcolMdlCalSetPars(&pars, &pFpcalData->model);

}

void getFpcalData
	(const char *pFitsFilename,
	 FpcalData	*pFpcalData,
	 StatusType	*pStatus)
{
	fitsfile	*pFile = NULL;
	int			fitsStatus = 0;
#ifdef NOPMS
	iteratorCol	aIters[3];
#else
	iteratorCol	aIters[5];
#endif

	if (*pStatus != STATUS__OK)
		return;

	/* open the FITS file */
	if (fits_open_file(&pFile, pFitsFilename, READONLY, &fitsStatus) != 0)
	{
		fits_report_error(stderr, fitsStatus);
		*pStatus = -1;
		return;
	}

	/*
	 * If <pFpcalData> has been used before, we need to
	 * deallocate previously allocated memory for it.
	 */
	if (pFpcalData->apparentRA != NULL)
	{
		fpcalDestructor(pFpcalData);
		fpcalConstructor(pFpcalData);
	}


	/* move to the (binary) table we're interested in */
	fits_movnam_hdu(pFile, BINARY_TBL, EXT_NAME, 0, &fitsStatus);

#ifdef AAO_DEBUG
	/* check that the units of the RA/DEC columns are in degrees */
	if (fitsStatus == 0)
	{
		char	*aColNames[] = {RA_FIELD, DEC_FIELD};
		char	unit[32];
		int		colNum;

		for (i = 0; i < (int) NUM_ELEMENTS(aColNames); i++)
		{
			fits_get_colnum(pFile, CASEINSEN, aColNames[i], &colNum, &fitsStatus);
			fits_get_bcolparms(pFile, colNum, NULL, unit, NULL, NULL,
							   NULL, NULL, NULL, NULL, &fitsStatus);
			AAO_ASSERT(strncmp(unit, "degrees", 7) == 0);
			/* printf("%s field has unit type %s\n", aColNames[i], unit); */
		}
	}
#endif

        if (fitsStatus == 0)
          {
            readHeader(pFile, pFpcalData, &fitsStatus);
            if (fitsStatus != 0)
              fprintf(stderr, "Error reading FITS header keywords\n");
          }

	/* specify the columns in the binary table that we're interested in */
        fits_iter_set_by_name(&aIters[0], pFile, RA_FIELD, TDOUBLE, InputCol);
        fits_iter_set_by_name(&aIters[1], pFile, DEC_FIELD, TDOUBLE, InputCol);
        fits_iter_set_by_name(&aIters[2], pFile, BUTTON_FIELD, TINT, InputCol);
#ifndef NOPMS
        fits_iter_set_by_name(&aIters[3], pFile, PM_RA_FIELD, TDOUBLE, InputCol);
        fits_iter_set_by_name(&aIters[4], pFile, PM_DEC_FIELD, TDOUBLE, InputCol);
#endif

	/* get all the data (in the form of a callback to <workFn> */
	/* -1 means pass all data to <workFn> in one hit */
        if (fitsStatus == 0)
          {
            fits_iterate_data(NUM_ELEMENTS(aIters), aIters, 0, -1, workFn, pFpcalData, 
                              &fitsStatus);
            if (fitsStatus != 0)
              fprintf(stderr,"Error reading binary table data\n");
          }

	fits_close_file(pFile, &fitsStatus);

	if (fitsStatus != 0)
	{
		fits_report_error(stderr, fitsStatus);
		*pStatus = -1;
	}
	else
		pFpcalData->bInitialised = 1;
}



/*
 * read relevant header information into <FpcalData> structure from a file
 * which has been saved using savehd and possibly modified by hand.
 */
int getFpcalDataText(
                      const char *fileName,
                      const char *cmdName,
                      FpcalData	*pFpcalData)
{
  int i;
  char buffer[200];
  int errors;
  int arraySpace;
  FILE *f = 0;
  FpcalPars pars;
  double wlenUVES, wlenFACB, wlenGiraffe;

  if ((f = fopen(fileName, "r")) == 0) 
    {
      fprintf(stderr,"%s: Failed to open file %s for input\n", cmdName, fileName);
      perror("");
      return -1;
    }
  if (pFpcalData->fileName)
      free(pFpcalData->fileName);
  pFpcalData->fileName = (char *)malloc(strlen(fileName)+1);
  strcpy(pFpcalData->fileName, fileName);
  
  /* 
   * Read header line.
   */
  if (!fgets(buffer, sizeof(buffer), f))
    {
      fprintf(stderr,"%s: Failed to read header lline from file %s\n", cmdName, fileName);
      fclose(f);
      return -1;
      
    }
  
  /*
   * Check the file header is ok
   */
#define HEADER "#FLAMES FP HEADER CALIBRATION DETAILS"
  if (strncmp(buffer,HEADER, strlen(HEADER)) != 0)
    {
      fprintf(stderr,"%s: Head of file %s is not correct\n", cmdName, fileName);
      fclose(f);
      return -1;
    }

  /*
   * If <pFpcalData> has been used before, we need to
   * deallocate previously allocated memory for it.
   */
  if (pFpcalData->apparentRA != NULL)
    {
      fpcalDestructor(pFpcalData);
      fpcalConstructor(pFpcalData);
    }

  errors = 0;
  /*
   * Start reading the simple items.
   */
  if (!fgets(buffer, sizeof(buffer), f))
    {
      fprintf(stderr,"%s: Failed to read field plate line  from file %s\n", cmdName, fileName);
      fclose(f);
      return -1;
    }
  if (sscanf(buffer,"%d", &pFpcalData->fieldPlate) != 1)
    {
      ++errors;
      fprintf(stderr,"%s: Error formating field plate from line \"%s\"\n", cmdName, buffer);
      
    }


  if (!fgets(buffer, sizeof(buffer), f))
    {
      fprintf(stderr,"%s: Failed to read temperature line  from file %s\n", cmdName, fileName);
      fclose(f);
      return -1;
    }
  if (sscanf(buffer,"%lg", &pFpcalData->atmTemp) != 1)
    {
      ++errors;
      fprintf(stderr,"%s: Error formating temperature from line \"%s\"\n", cmdName, buffer);
      
    }


  if (!fgets(buffer, sizeof(buffer), f))
    {
      fprintf(stderr,"%s: Failed to read pressure line  from file %s\n", cmdName, fileName);
      fclose(f);
      return -1;
    }
  if (sscanf(buffer,"%lg", &pFpcalData->atmPressure) != 1)
    {
      ++errors;
      fprintf(stderr,"%s: Error formating pressure from line \"%s\"\n", cmdName, buffer);
      
    }

  if (!fgets(buffer, sizeof(buffer), f))
    {
      fprintf(stderr,"%s: Failed to read humidity line  from file %s\n", cmdName, fileName);
      fclose(f);
      return -1;
    }
  if (sscanf(buffer,"%lg", &pFpcalData->atmHumidity) != 1)
    {
      ++errors;
      fprintf(stderr,"%s: Error formating humidty from line \"%s\"\n", cmdName, buffer);
      
    }


  if (!fgets(buffer, sizeof(buffer), f))
    {
      fprintf(stderr,"%s: Failed to read wavelength line  from file %s\n", cmdName, fileName);
      fclose(f);
      return -1;
    }
  wlenGiraffe = 450;
  wlenUVES    = 500;
  wlenFACB    = 650;
  if (sscanf(buffer,"%lg %lg %lg", &wlenGiraffe, &wlenUVES, &wlenFACB) != 3)
    {
      ++errors;
      fprintf(stderr,"%s: Error formating wavelength from line \"%s\"\n", cmdName, buffer);
      
    }
  /*
   * The wavelength needs to be added to the internal table.  This wavelength
   * becomes the current wavelength.
   */
  pFpcalData->curWaveIndex.Giraffe = fpcolMdlCalNewWavelength(wlenGiraffe, pFpcalData);
  pFpcalData->curWaveIndex.UVES    = fpcolMdlCalNewWavelength(wlenUVES,    pFpcalData);
  pFpcalData->curWaveIndex.FACB    = fpcolMdlCalNewWavelength(wlenFACB,    pFpcalData);
  if (( pFpcalData->curWaveIndex.Giraffe < 0)||
      ( pFpcalData->curWaveIndex.UVES < 0) ||
      ( pFpcalData->curWaveIndex.FACB < 0))
    {
      fprintf(stderr,"%s: Failure adding new wavelengths to internal tables\n", cmdName);
    }

  if (!fgets(buffer, sizeof(buffer), f))
    {
      fprintf(stderr,"%s: Failed to read ADC angle line  from file %s\n", cmdName, fileName);
      fclose(f);
      return -1;
    }
  if (sscanf(buffer,"%lg", &pFpcalData->posAngle) != 1)
    {
      ++errors;
      fprintf(stderr,"%s: Error formating ADC angle from line \"%s\"\n", cmdName, buffer);
      
    }
 if (!fgets(buffer, sizeof(buffer), f))
    {
      fprintf(stderr,"%s: Failed to observation date line from file %s\n", cmdName, fileName);
      fclose(f);
      return -1;
    }
 if (stringToMjd(buffer, &pFpcalData->observationDate) == -1) 
   {
     fprintf(stderr,"%s: Failed to convert observation date string \"%s\" to MJD\n", 
             cmdName, buffer);
     ++errors;
   }


  if (!fgets(buffer, sizeof(buffer), f))
    {
      fprintf(stderr,"%s: Failed to read equinox line  from file %s\n", cmdName, fileName);
      fclose(f);
      return -1;
    }
  if (sscanf(buffer,"%lg", &pFpcalData->equinox) != 1)
    {
      ++errors;
      fprintf(stderr,"%s: Error formating equinox from line \"%s\"\n", cmdName, buffer);
      
    }


  if (!fgets(buffer, sizeof(buffer), f))
    {
      fprintf(stderr,"%s: Failed to field centre line from file %s\n", cmdName, fileName);
      fclose(f);
      return -1;
    }

  if (stringToPosition(buffer, 
                      &pFpcalData->fieldCentreMeanRA,
                      &pFpcalData->fieldCentreMeanDEC) == -1) 
    {
     fprintf(stderr,"%s: Failed to convert filter center string \"%s\" to radians\n", 
             cmdName, buffer);
     ++errors;
    }

  /*
   * convert the field centre RA/DEC from mean to apparent.
   *
   * the <meanToApparent> routine requires <meanAppParams> so they must
   * be set up beforehand.
   */
  slaMappa(pFpcalData->equinox,
           pFpcalData->observationDate,
           pFpcalData->meanAppParams);
  
  meanToApparent(pFpcalData->fieldCentreMeanRA,
                 pFpcalData->fieldCentreMeanDEC,
                 0.0,	/* no proper motion for field centre */
                 0.0,
                 pFpcalData->meanAppParams,
                 &(pFpcalData->fieldCentreAppRA),
                 &(pFpcalData->fieldCentreAppDEC));



  if (!fgets(buffer, sizeof(buffer), f))
    {
      fprintf(stderr,"%s: Failed to read number of fibres from file %s\n", cmdName, fileName);
      fclose(f);
      return -1;
    }
  if (sscanf(buffer,"%d", &pFpcalData->nFibres_a) != 1)
    {
      ++errors;
      fprintf(stderr,"%s: Error formating pressure from line \"%s\"\n", cmdName, buffer);
      
    }
  printf("Field contains %d fibres\n", pFpcalData->nFibres_a);
  arraySpace =  pFpcalData->nFibres_a+1* sizeof(double);

  
  /* 
   * allocate memory to store RA/DEC coordinates 
   */
  if (fpcalMallocObjectArrays( pFpcalData->nFibres_a, pFpcalData) != 0)
    return -1;

  for (i = 0; i < pFpcalData->nFibres_a ; ++i)
    {
      if (!fgets(buffer, sizeof(buffer), f))
        {
          fprintf(stderr,"%s: Failed to read details of object %d\n", cmdName, i+1);
          fclose(f);
          return -1;
        }
      if (stringToObjectPos(buffer, 
                           pFpcalData->meanAppParams,
                           &pFpcalData->buttons[i],
                           &pFpcalData->meanRA[i],
                           &pFpcalData->meanDEC[i],
                           &pFpcalData->apparentRA[i],
                           &pFpcalData->apparentDEC[i],
                           &pFpcalData->properMotionRA[i],
                           &pFpcalData->properMotionDEC[i]) == -1) 
        {
          ++errors;
          fprintf(stderr,"%s: Error formating object %d details from line \"%s\"\n", 
                  cmdName, i, buffer);
        }
      
    }

  /*
   * After the object list we should have seperator line.  Complain if we don't get it.
   */
  if (!fgets(buffer, sizeof(buffer), f))
    {
      fprintf(stderr,"%s: Failed to read object seperator line from file %s\n", cmdName, fileName);
      fclose(f);
      return -1;
    }
  if (buffer[0] != '#')
    {
      fprintf(stderr,
              "%s: Expected end of objects but didn't get it - line \"%s\"\n", 
              cmdName, fileName);
      fclose(f);
      return -1;
    }



  for (i = 0; i < NUM_PARAMETERS; i++)
    {
           
      if (!fgets(buffer, sizeof(buffer), f))
        {
          fprintf(stderr,"%s: Failed to read details of tel parameter index %d\n", cmdName, i);
          fclose(f);
          return -1;
        }

      if (sscanf(buffer,"%lg", &pars.p[i]) != 1)
        {
          ++errors;
          fprintf(stderr,"%s: Error formating tel parameter index %d from line \"%s\"\n", 
                  cmdName, i, buffer);
          
        }
    }
  /*
   * Set the telescope model parameters.
   */
  fpcolMdlCalSetPars(&pars, &pFpcalData->model);


  /*
   * After the tel pars list we should have seperator line.  Complain if we don't get it.
   */
  if (!fgets(buffer, sizeof(buffer), f))
    {
      fprintf(stderr,
              "%s: Failed to read tel pars seperator line from file %s\n", cmdName, fileName);
      fclose(f);
      return -1;
    }
  fclose(f);
  if (buffer[0] != '#')
    {
      fprintf(stderr,
              "%s: Expected end of tel pars but didn't get it - line \"%s\"\n", 
              cmdName, fileName);
      return -1;
    }


  pFpcalData->bInitialised = 1;

  return 0;
}



void fpcalConstructor (FpcalData *pFpcalData)
{
  int i;

  memset(pFpcalData, 0, sizeof(*pFpcalData));
        
  /*
   * No default fit.  The use must say what to do.
   */
  for (i = 0; i < NUM_PARAMETERS ; ++i)
    pFpcalData->aFit[i] = 1;

  /*
   * Set up the fit function.
   */
  pFpcalData->fitFunc = fpcolMdlFit;
  pFpcalData->powellTol = 0.000001;
  pFpcalData->powellMaxIts = 200;
  pFpcalData->powellMode = POWELL_MODE_1;  /* Default powell mode */
}


void fpcalDestructor (FpcalData *pFpcalData)
{
  FpcalMdl *model = &pFpcalData->model;
  int index;

  if (pFpcalData == NULL)
    return;

  if (pFpcalData->fileName)
      free(pFpcalData->fileName);

  if (pFpcalData->apparentRA)
    {
      free(pFpcalData->apparentRA);
      free(pFpcalData->apparentDEC);
      free(pFpcalData->meanRA);
      free(pFpcalData->meanDEC);
      free(pFpcalData->properMotionRA);
      free(pFpcalData->properMotionDEC);
    }
  if (pFpcalData->offsetRA != NULL)
    {
      free(pFpcalData->offsetRA);
      free(pFpcalData->offsetDEC);
      free(pFpcalData->offsetWaveIndex);
      free(pFpcalData->offsetButtonNo);
      free(pFpcalData->residualsInfo);
    }
  
  if (model->created)
    {
      StatusType status = STATUS__OK;
      for (index = 0; index < model->wavelenCnt ; ++index)
        fpcolMdlFree(model->models[index], &status);
    }
        

}


/*
 * Returns true if the specified fibre is a UVES fibre.  Taken from
 * fpistTest.c.
 */
int fpcalIsUVES(unsigned int fibre)
{
  if ((fibre == 3  )||(fibre == 37 )||(fibre == 69 )||
      (fibre == 103)||(fibre == 135)||(fibre == 169)||
      (fibre == 201)||(fibre == 235))
    return 1;
  else
    return 0;
}
/*
 * Returns true if the specified fibre is an FACB fibre.  Taken from
 * fpistTest.c
 */
int fpcalIsFACB(unsigned int fibre)
{
  if ((fibre == 17)||(fibre == 85)||(fibre == 149)||(fibre == 213))
    return 1;
  else 
    return 0;

}
/*
 * Returns true if the specified fibre is a Giraffe fibre.
 *
 *   We assume Giraffe if not UVES and not FACB (we don't support
 *   unused etc).
 */
int fpcalIsGiraffe(unsigned int fibre)
{
  return (!fpcalIsFACB(fibre)&&!fpcalIsUVES(fibre));
}

extern int fpcalWaveIdxForFibre(const FpcalCurWaveIdx *curWaveIndex, unsigned int fibre)
{
  if (fpcalIsFACB(fibre))
    return curWaveIndex->FACB;
  else if (fpcalIsUVES(fibre))
    return curWaveIndex->UVES;
  else
    return curWaveIndex->Giraffe;
}

extern void fpcolMdlDecompLin(
    const double *pars)
{
    double xz, yz, xs, ys, perp, orient;
    slaDcmpf((double *)pars,/* Cast to discard const - slaDcmf not prototyped correctly*/
             &xz, &yz, &xs, &ys, &perp, &orient);
    printf("Zero point = %.1f, %.1f (microns)\n", xz, yz);
    printf("Scale (x/y)= %.6f, %.6f (%.6f)\n", xs, ys, (xs*ys));
    printf("nonperpendicularity = %.8f (radians), %.3f degrees\n", 
           perp, perp*180.0/M_PI);
    printf("orientation = %.8f radians, %.3f degrees\n", 
           orient, orient*180.0/M_PI);
 
}
