
/*******************************************************************************
* E.S.O. - VLT project
*
* "@(#) $Id: ACMM:ozpozFpcol/ws/src/fpcolMdlCalLoadOffsets.c,v 1.63 07-Oct-2008 11:10:26+10 tjf $"
*
*
* who       when      what
* --------  --------  ----------------------------------------------
* tfarrell  16/10/02  Add removeMeanOffsets() function.
* tfarrell  26/08/02  MAX_BUTTONS is the limit, not one less.
* tfarrell  22/08/02  Add clearButtonOffset() function.
* tfarrell  09/05/02  Revamp to support multiple wavelengths.
* ssmedley  07/09/01  created
*
*/
/************************************************************************
*   NAME
*	loadButtonOffsets getNextOffset getButtonOffset buttonOffsetsStatus
*	setButtonOffset - A module for reading/storing the contents of ASCII
*	Offset Files.
*
*   SYNOPSIS
*       #include "fpcolMdlCalLoadOffsets.h"
*
*       int buttonFound(
*                       const int nButtons, 
*                      const int *buttons, 
*                      const int buttonNo);
*
*	void loadButtonOffsets
*		(const char	 *pOffsetFilename,
*                FpcalCurWaveIdx curWaveIndex,
*		 StatusType	 *pStatus);
*
*	const RaDec *getOffset (int index);
*
*	const RaDec *getButtonOffset
*		(int	     buttonNo,
*                int         waveIndex,
*		 StatusType  *pStatus);
*
*	int buttonOffsetsStatus (void);
*
*       int getNumOffsetWavelengths(void);
*
*	int setButtonOffset (int buttonNo, double ra, double dec, int waveIndex);
*
*       int getNumOffsets();
*
*       int clearButtonOffset(int buttonNo, waveIndex);
*
*       int  removeMeanOffsets();
*
*   DESCRIPTION
*       buttonFound is a utility function.  Return true if button buttonNo is found 
*       in the list of size nButtons specified by buttons.  Also returns true if the list
*       is empty.
*
*	loadButtonOffsets() will read in the contents of an ASCII Offset
*	File & store the values in an internal buffer for querying/modification.  The
*       current wavelength index for each fibre type must be supplied and may
*       be modified,   This is used to indicate the wavelength
*       index to be applied when there is no wavelength in the file.
*
*	The format of an entry in an ASCII Offset File is:
*	FIBRE_NUMBER OFFSET_RA OFFSET_DEC\n
*
*       or one of the following lines to set the wavelengths for the next set of
*       fibres.  Any line starting with ! or # is considered a comment.
*
*       GIRAFFE_WLEN wlen\n
*       UVES_WLEN wlen\n
*       FACB_WLEN wlen\n
*
*
*	getOffset() is an access function to return an offset of a given index.
*
*	getButtonOffset() returns the offset for the fibre specified by
*	<buttonNo>.
*
*	buttonOffsetsStatus() returns <true> if button offsets have been
*	loaded from an ASCII Offset File *OR* at least 1 offset has been
*	specified with the setButtonOffset() routine.
*
*	setButtonOffset() will set the offset Ra/Dec for the specified fibre
*	in <buttonNo> and the specified wavelength index.
*
*       getNumOffsets() returns the number of offsets.  Except that if offsets have
*       been cleared, this may be wrong in which get getOffset() will return NULL
*       for an offset which has been cleared.
*
*       clearButtonOffset() clears the offset assoicated with the given button
*       and wavelength index.
*
*       removeMeanOffsets() will remove the mean offset from all offsets.
*       Return 0 on success and -1 if no offsets have been loaded.
*
*	Note that there also exists another set of some the above routines with
*	an '_o' postfix. These routines accept an additional parameter, an
*	Offsets buffer in which all data is stored. They perform the same
*	functionality as above, only they apply to the buffer passed as a
*	parameter.
*
*   ENVIRONMENT
*
*   CAUTIONS
*
*   EXAMPLES
*
*   SEE ALSO
*
*   BUGS
*
*------------------------------------------------------------------------
*/


#include <stdio.h>	/* fopen fscanf fclose */
#include <stdlib.h>	/* malloc */
#include <strings.h>	/* */
#include <assert.h>

#include "fpcolMdlCalLoadOffsets.h"
#include "fpcolMdlCalModel.h"

static Offsets	offsets;

static int setButtonOffset_o (
	Offsets	*pO,
	int		buttonNo,
	double	ra,
	double	dec,
        int     waveIndex);



/*
 * Return true if button buttonNo is found in the list of size
 * nButtons specified by buttons.  Also returns true if the list
 * is empty.
 */
extern int buttonFound(
                       const int nButtons, 
                       const int *buttons, 
                       const int buttonNo)
{
  int  i;
  if ((nButtons == 0)||(!buttons))
    return 1;

  for (i = 0; i < nButtons ;  ++i)
    {
      if (buttons[i] == buttonNo)
        return 1;
    }
  return 0;
}

/* The only _o form used outside */
static void loadButtonOffsets_o (
        const char	*pOffsetFilename,  /* File name */
        FpcalData	*pFpcalData,       /* Optional point ot pFpcalData, only needed
                                            * if setting new wavelengths is required */
        FpcalCurWaveIdx *curWaveIndex,     /* Indexs for current wavelenghts of each fibre type*/
        const int       nButtons,          /* Number of items in buttons array */
        const int       *buttons,          /* List of acceptable button numbers.  If null
                                              or nButtons is zero, accept all */
	Offsets		*pO,               /* Structure to put stuff in */
	StatusType	*pStatus)
{
	FILE	*pFile = NULL;
	int		n, buttonNo;
	double	ra, dec;
        char    buffer[200];
        int lineNo;

	assert(pOffsetFilename != NULL);
	assert(pO != NULL);

	if (*pStatus != 0)
		return;

	/* initialize all elements of array to zero */
	memset(pO, 0, sizeof(Offsets));
	assert(pO->bLoaded == BO_NOT_LOADED);
	assert(pO->offset[MAX_OFFSETS-1].ra == 0.0);
        assert(pO->numOffsets == 0);


	if ((pFile = fopen(pOffsetFilename, "r")) == NULL)
	{
		perror(pOffsetFilename);
		*pStatus = -1;
		pO->bLoaded = 0;
		return;
	}
        lineNo = 0;

        while (fgets(buffer, sizeof(buffer), pFile))
          {
            int wavelen;
            ++lineNo;
            if (strlen(buffer) == 0)
              ;
            else if ((buffer[0] == '#')||(buffer[0] == '!'))
              ;
            else if ((n = sscanf(buffer, "%d%lf%lf", &buttonNo, &ra, &dec)) == 3)
              {
                if (!buttonFound(nButtons, buttons, buttonNo))
                  {
                    fprintf(stderr,"Offset found for button %d on line %d\n", buttonNo, lineNo);
                    fprintf(stderr,"  Don't have information from header for this button\n");
                    fprintf(stderr,"  Offset ignored\n");
                    
                  }
                else if (setButtonOffset_o(pO, buttonNo, ra, dec, 
                                           fpcalWaveIdxForFibre(curWaveIndex, buttonNo)) != 0)
                  {
                    *pStatus = -1;
                    pO->bLoaded = 0;
                    goto close;
                  }
              }
            else if ((pFpcalData)&&((n = sscanf(buffer, "GIRAFFE_WLEN %d", &wavelen)) == 1))
              {
                if ((curWaveIndex->Giraffe = fpcolMdlCalNewWavelength(wavelen, pFpcalData)) < 0)
                  {
                    fprintf(stderr,"Failed to add new Giraffe wavelength %d from line %d\n", 
                            wavelen, lineNo);
                    *pStatus = -1;
                    pO->bLoaded = 0;
                    goto close;
                  }
              }
            else if ((pFpcalData)&&((n = sscanf(buffer, "UVES_WLEN %d", &wavelen)) == 1))
              {
                if ((curWaveIndex->UVES = fpcolMdlCalNewWavelength(wavelen, pFpcalData)) < 0)
                  {
                    fprintf(stderr,"Failed to add new UVES wavelength %d from line %d\n", 
                            wavelen, lineNo);
                    *pStatus = -1;
                    pO->bLoaded = 0;
                    goto close;
                  }
               }
            else if ((pFpcalData)&&((n = sscanf(buffer, "FACB_WLEN %d", &wavelen)) == 1))
              {
                if ((curWaveIndex->FACB = fpcolMdlCalNewWavelength(wavelen, pFpcalData)) < 0)
                  {
                    fprintf(stderr,"Failed to add new FACB wavelength %d from line %d\n", 
                            wavelen, lineNo);
                    *pStatus = -1;
                    pO->bLoaded = 0;
                    goto close;
                  }
               }
            else 
              {
                fprintf(stderr,"Offset file line %d is invalid\n", lineNo);
              }
}

	fprintf(stderr, "Loaded %d offsets from %s.\n",
			pO->numOffsets,
			pOffsetFilename);
	pO->bLoaded = 1;
close:
	fclose(pFile);
}

const Offsets *loadButtonOffsets (
	const char	*pOffsetFilename,
        FpcalData	*pFpcalData,
        FpcalCurWaveIdx *curWaveIndex,
        const int       nButtons,
        const int       *buttons,
	StatusType	*pStatus)
{
	loadButtonOffsets_o(pOffsetFilename, pFpcalData,
                            curWaveIndex, nButtons, buttons, &offsets, pStatus);
        return &offsets;
}


/*
 * Return an offset by index.
 */

static const RaDec *getOffset_o (
	const Offsets	*pO,
	int		index)
{
  if ((index < pO->numOffsets)&&(pO->offset[index].bSet))
    return &(pO->offset[index]);
  else
    return NULL;
}

const RaDec *getOffset (int index)
{
	return getOffset_o(&offsets, index);
}


/*
 * Return an offset by button+waveIndex;
 */
const RaDec *getButtonOffset_o(
                             const Offsets *pO,
                             const int	buttonNo,
                             const int      waveIndex,
                             StatusType	*pStatus)
{
  int i;
  if (pO->bLoaded != BO_LOADED ||
      buttonNo <= 0            ||
      buttonNo > MAX_BUTTONS  ||
      waveIndex < 0            ||
      waveIndex >= MAX_WLENS)
    {
      *pStatus = -1;
      return NULL;
    }
  for (i = 0; i < MAX_OFFSETS ; ++i)
    {
      if ((pO->offset[i].bSet)&&
          (pO->offset[i].button == buttonNo)&&
          (pO->offset[i].wIndex == waveIndex))
        {
          return &pO->offset[i];
        }
    }
  *pStatus = -1;
  return NULL;
}
const RaDec *getButtonOffset
	(const int	buttonNo,
         const int      waveIndex,
	 StatusType	*pStatus)
{
  return getButtonOffset_o(&offsets, buttonNo, waveIndex, pStatus);
}


static int buttonOffsetsStatus_o (Offsets *pO)
{
	return pO->bLoaded;
}


int buttonOffsetsStatus (void)
{
	return buttonOffsetsStatus_o(&offsets);
}


static int setButtonOffset_o (
	Offsets	*pO,
	int		buttonNo,
	double	ra,
	double	dec,
        int     waveIndex)
{
  int i;
  int offsetIndex = pO->numOffsets;  /* Index of next offset */

#if OFFSET_IN_ARCSEC
  ra  = SEC2RAD(ra);
  dec = SEC2RAD(dec);
#else
#warning "Offset Inputs in radians"
#endif

  if (buttonNo > MAX_BUTTONS)
    {
      fprintf(stderr, "%d exceeds maximum button identifier of %d\n", buttonNo, 
              MAX_BUTTONS);
      return -1;
    }
  for (i = 0; i < pO->numOffsets ; ++i)
    {
      if ((pO->offset[i].bSet)&&
          (pO->offset[i].button == buttonNo)&&
          (pO->offset[i].wIndex == waveIndex))
        {
          offsetIndex = i;
          break;
        }
    }
  
  
  if (offsetIndex >= MAX_OFFSETS)
    {
      fprintf(stderr, "Exceeded the maximum number of offsets - %d", MAX_OFFSETS);
      return -1;
    }
  /*
   * OVerwrite existing value for this button and wavelength combination or create
   * a new element.
   */
  pO->offset[offsetIndex].bSet = 1;
  pO->offset[offsetIndex].ra = ra;
  pO->offset[offsetIndex].dec = dec;
  pO->offset[offsetIndex].wIndex = waveIndex;
  pO->offset[offsetIndex].button = buttonNo;
  pO->bLoaded = 1;
  /*
   * If creating new element, increment number of offsets.
   */
  if (offsetIndex == pO->numOffsets)
    ++pO->numOffsets;
  return 0;
}

int setButtonOffset (int buttonNo, double ra, double dec, int waveIndex)
{
	return setButtonOffset_o(&offsets, buttonNo, ra, dec, waveIndex);
}



int getNumOffsets(void)
{
  return offsets.numOffsets;
}

/*
 * Clear a button offset - note, this will mean that getNumOffsets is wrong and that
 * and offset of the given index will return 0 when we use getButtonOffset()
 */

static int clearButtonOffset_o (
	Offsets	*pO,
	int		buttonNo,
        int     waveIndex)
{
    int i;
    int offsetIndex;

    if (buttonOffsetsStatus() != BO_LOADED)
        {
        printf("\tNo offsets loaded as yet\n");
        return -1;
        }

    if (buttonNo > MAX_BUTTONS)
    {
      fprintf(stderr, "%d exceeds maximum button identifier of %d\n", 
              buttonNo, MAX_BUTTONS);
      return -1;
    }
    for (i = 0; i < pO->numOffsets ; ++i)
        {
        if ((pO->offset[i].bSet)&&
            (pO->offset[i].button == buttonNo)&&
            (pO->offset[i].wIndex == waveIndex))
            {
            offsetIndex = i;
            break;
            }
        }
  
  
  if (offsetIndex >= pO->numOffsets)
    {
      fprintf(stderr, "No such offset - button %d, waveIndex %d", buttonNo, waveIndex);
      return -1;
    }
  /*
   * OVerwrite existing value for this button and wavelength combination or create
   * a new element.
   */
  pO->offset[offsetIndex].bSet = 0;
  pO->offset[offsetIndex].ra = 0;
  pO->offset[offsetIndex].dec = 0;
  pO->offset[offsetIndex].wIndex = 0;
  pO->offset[offsetIndex].button = 0;
  return 0;  
}

int clearButtonOffset(int buttonNo, int waveIndex)
{
    return clearButtonOffset_o(&offsets, buttonNo, waveIndex);
}


int removeMeanOffsets()
{
  int i;
  double totalRa  = 0;
  double totalDec = 0;
  double meanRa   = 0;
  double meanDec  = 0;
  int numOffsets    = 0;
  
  if (offsets.bLoaded != BO_LOADED)
    {
      return -1;
    }

  for (i = 0; i < MAX_OFFSETS ; ++i)
    {
      if (offsets.offset[i].bSet)
        {
          totalRa  += offsets.offset[i].ra;
          totalDec += offsets.offset[i].dec;
          ++numOffsets;
        }    
    }

  /*printf("Total RA = %g, Total Dec = %g\n", totalRa, totalDec);*/
  if (numOffsets != 0)
    {
      meanRa  = totalRa/numOffsets;
      meanDec = totalDec/numOffsets;
    }

  printf("Removing mean offset ra=%6.3f, dec=%6.3f (%s) from %d offsets\n",
         RAD2SEC(meanRa), RAD2SEC(meanDec), "arc-seconds", numOffsets);

  for (i = 0; i < MAX_OFFSETS ; ++i)
    {
      if (offsets.offset[i].bSet)
        {
          offsets.offset[i].ra  -= meanRa;
          offsets.offset[i].dec -= meanDec;
        }
    
    }

  return 0;
}
