
/*******************************************************************************
* E.S.O. - VLT project
*
* "@(#) $Id: ACMM:ozpozFpcol/ws/src/fpcolMdlCalLoadRegions.c,v 1.63 07-Oct-2008 11:10:27+10 tjf $"
*
*
* who       when      what
* --------  --------  ----------------------------------------------
* ssmedley  07/09/01  created
*
*/

/************************************************************************
*   NAME
*	loadImageLocations xy2index -
*	Load image positions from file.
*
*   SYNOPSIS
*       #include "fpcolMdlCalLoadRegions.h"
*
*	int loadImageLocations (
*		const char	*pPositionsFile,
*		FibreRegion	*aFibreRegions,
*		FibreRegion	*pBackground1,
*		FibreRegion	*pBackground2,
*		int		*pnRegions);
*
*	int xy2index (
*		int		x,
*		int		y,
*		int		nCols);
*
*   DESCRIPTION
*
*	This module is used to read in a set of image positions from disk.
*	The image positions specify the regions of an image that are used to
*	calculate fibre fluxes & background noise. A single region is denoted
*	by 2 coordinates of a rectangle (top-left & bottom-right) which should
*	appear within an UVES slit image.
*
*	The format of the image regions/positions file is:
*	REGION_NO TOP_LEFT_X TOP_LEFT_Y BOTTOM_RIGHT_X BOTTOM_RIGHT_Y\n
*
*	REGION_NO is negative if the region is used for determining background
*	noise. (Note, can only accept 2 background regions).
*
*	loadImageLocations() loads these regions from disk. The 2 background
*	regions are stored in <pBackground1> & <pBackground2>. All other
*	regions are stored in the <aFibreRegions> array & <pnRegions> set to
*	the number of regions this array contains.
*
*	xy2index() is a simple utility function for mapping a pixel (at x,y) into
*	it's position in a 1 dimensional structure. (ie. it allows a 2-dimensional
*	structure to be represented as a 1-dimensional array in C.)
*
*   ENVIRONMENT
*
*   CAUTIONS
*
*   EXAMPLES
*
*   SEE ALSO
*
*   BUGS
*
*------------------------------------------------------------------------
*/


#include <stdio.h>		/* perror */
#include <strings.h>	/* strncmp */
#include <assert.h>

#include "fpcolMdlCalLoadRegions.h"
#include "fpcolMdlCalGetline.h"

/*
 * read in a file containing image positions used in calculating
 * fibre fluxes.
 */
int loadImageLocations (
	const char	*pPositionsFile,
	FibreRegion	*aFibreRegions,
	FibreRegion	*pBackground1,
	FibreRegion	*pBackground2,
	int			*pnRegions)
{
	FILE	*pFile = fopen(pPositionsFile, "r");
	int		n;
	int		rv = 0;
	char	buffer[128];
	int		bBackground1 = 0, bBackground2 = 0;

	if (pFile == NULL)
	{
		perror(pPositionsFile);
		return 2;
	}

	*pnRegions = 0;

	while ((n = getline(pFile, buffer, sizeof(buffer))) >= 0)
	{
		FibreRegion	*p = NULL;

		if (n == 0 || buffer[0] == '#')
			continue;	/* empty line or comment */

		/* the background region has a fibre number of -1 */
		if (strncmp(buffer, "-1", 2) == 0)
		{
			p = pBackground1;
			assert(bBackground1 == 0);
			bBackground1 = 1;
		}
		else if (strncmp(buffer, "-2", 2) == 0)
		{
			p = pBackground2;
			assert(bBackground2 == 0);
			bBackground2 = 1;
		}
		else
		{
			p = &(aFibreRegions[*pnRegions]);
			(*pnRegions)++;
		}

		if ((n = sscanf(buffer,
						"%d%d%d%d%d",
						&(p->fibreNo),
						&(p->tl.x),
						&(p->tl.y),
						&(p->br.x),
						&(p->br.y))) != 5)
		{
			fprintf(stderr, "Couldn't read fibre region in %s.\n", pPositionsFile);
			rv = -1;
			break;
		}
		if (p->tl.y < p->br.y || p->tl.x > p->br.x)
		{
			fprintf(stderr, "Region %d is invalid\n", *pnRegions);
			rv = -1;
			break;
		}
	}

	fclose(pFile);
	if (rv == 0)
		if (bBackground1 != 1 || bBackground2 != 1)
		{
			fprintf(stderr, "Must specify 2 background regions.\n");
			rv = -1;
		}

	return rv;
}

/*
 * Map a pixel (at x,y) into it's position in a 1 dimensional structure.
 * <nCols> is number of columns in image.
 */

int xy2index (
	int		x,
	int		y,
	int		nCols)
{
	/* x, y are 1-based coordinates */
	assert(x < nCols);
	return ((y - 1) * nCols) + x;
}
