
/*******************************************************************************
* E.S.O. - VLT project
*
* "@(#) $Id: ACMM:ozpozFpcol/ws/src/fpcolMdlCalMkImg.c,v 1.63 07-Oct-2008 11:10:27+10 tjf $"
*
*
* who       when      what
* --------  --------  ----------------------------------------------
* tfarrell  15/10/02  Include unitstd.h for getopt().
* ssmedley  07/09/01  created
*
*/

/************************************************************************
*   NAME
*	mkImg - Generate artificial star images.
*
*   SYNOPSIS
*
*   DESCRIPTION
*
*	This program, mkImg, will generate an artificial image that can
*	be used to test centroiding. It uses a gaussian profile to simulate
*	the intensity/flux of light coming from a star.
*
*	The pixel values (of dimension specified by -d) of a star image
*	centred about the position specified by -c are output on stdout
*	(or as specified by -o). The program uses a gaussian distribution
*	magnified by -m with random noise added in the range specified
*	by -n.
*
*   ENVIRONMENT
*
*   CAUTIONS
*
*   EXAMPLES
*
*   SEE ALSO
*
*   BUGS
*
*------------------------------------------------------------------------
*/


static const char *rcsId="@(#) $Id: ACMM:ozpozFpcol/ws/src/fpcolMdlCalMkImg.c,v 1.63 07-Oct-2008 11:10:27+10 tjf $";

static void *use_rcsId = (0 ? (void *)(&use_rcsId) : (void *) &rcsId);

#define __EXTENSIONS__    /* Pick up extra include file stuff on solaris */
#define _HPUX_SOURCE      /* Pick up extra include file stuff on HP UX */

#include <stdio.h>	/* perror sscanf printf */
#include <stdlib.h>	/* atof rand */
#include <math.h>	/* hypot */
#include <slamac.h>	/* DR2AS */
#include <unistd.h>     /* getopt sometimes */

#include "fpcolMdlCalCentroid.h"

void usage (const char *pName)
{
	fprintf(stderr, "usage: %s [-options]\n", pName);
	fprintf(stderr, "options:\n");
	fprintf(stderr, "-c <centre> Centre of gaussian within image.\n");
	fprintf(stderr, "-d <dim>    Dimensions of image [3x3]\n");
	fprintf(stderr, "-h          help\n");
	fprintf(stderr, "-m <factor> Magnification factor [1.0]\n");
	fprintf(stderr, "-n <noise>  Random photon noise to add to system [0.0]\n");
	fprintf(stderr, "-o <file>   Output file for image. [stdout]\n");
	fprintf(stderr, "-s <scale>  spacing b/w pixels in image (arcseconds) [1.0]\n");
}

int main (int argc, char *argv[])
{
	double	noise = 0.0;
	int		xDim = 3;
	int		yDim = 3;
	int		bPeakSpecified = 0;
	double	peakX, peakY;
	int		c, x, y;
	double	stdDev;
	double	magnification = 1.0;
	double	spacing = 1.0 / DR2AS;	/* 1 arcsecond */
	FILE	*pOutput = stdout;

	while ((c = getopt(argc, argv, "c:d:hm:n:o:s:")) != EOF)
	{
		switch (c)
		{
			case 'c':
				if (sscanf(optarg, "%lf,%lf", &peakX, &peakY) != 2)
				{
					fprintf(stderr, "Invalid centre value <%s>.\n", optarg);
					exit(1);
				}
				bPeakSpecified = 1;
				break;
			case 'd':
				if (sscanf(optarg, "%dx%d", &xDim, &yDim) != 2)
				{
					fprintf(stderr, "Invalid image dimensions <%s>.\n", optarg);
					exit(1);
				}
				break;
			case 'm':
				magnification = atof(optarg);
				break;
			case 'n':
				noise = atof(optarg);
				break;
			case 'o':
				pOutput = fopen(optarg, "w");
				if (pOutput == NULL)
				{
					perror(optarg);
					exit(2);
				}
				break;
			case 's':
				spacing = atof(optarg) / DR2AS;
				break;
			case 'h':
				usage(argv[0]);
				exit(0);
				break;
			case '?':
				usage(argv[1]);
				exit(1);
				break;
		}
	}

	if (bPeakSpecified == 0)
	{
		peakX = (xDim - 1.0) / 2.0;
		peakY = (yDim - 1.0) / 2.0;
	}
	else
	{
		if (peakX < 0 || peakX > (xDim -1) || peakY < 0 || peakY > (yDim - 1))
		{
			fprintf(stderr, "Centre point is outside image!\n");
			exit(2);
		}
	}

	/* 4 std dev covers 99% of gaussian. */
	stdDev = (((xDim > yDim ? xDim : yDim) - 1) * spacing) / 4;

	fprintf(stderr, "peak: (%f, %f)\n", peakX, peakY);
	fprintf(stderr, "image dimensions: (%d, %d)\n", xDim, yDim);
	fprintf(stderr, "noise: %f\n", noise);
	fprintf(stderr, "spacing: %f\n", spacing);
	fprintf(stderr, "standard deviation: %f\n", stdDev);
	fprintf(stderr, "magnification: %f\n", magnification);

	{
		double	grid[yDim][xDim];
		double	p;

		for (y = 0; y < yDim; y++)
		{
			for (x = 0; x < xDim; x++)
			{
				p = hypot((x - peakX) * spacing, (y - peakY) * spacing);
				grid[y][x] = gaussian(p,
									  0.0, /* mean */
									  stdDev); /* std. dev. */
				grid[y][x] *= magnification;
				if (noise != 0.0)
				{
					/* add random noise in range: -noise .. noise */
					double factor = (rand() - 16384.0) / 16384.0;
					grid[y][x] += factor * noise;
				}
				printf("%7.5f ", grid[y][x]);
			}
			printf("\n");
		}
	}
	return 0;
}
