
/*******************************************************************************
* E.S.O. - VLT project
*
* "@(#) $Id: ACMM:ozpozFpcol/ws/src/fpcolMdlCalRegion.c,v 1.63 07-Oct-2008 11:10:27+10 tjf $"
*
*
* who       when      what
* --------  --------  ----------------------------------------------
* tfarrell  15/10/02  Include unitstd.h for getopt().
* tfarrell  10/10/02  A report from Matthew Colless seems to indicate
*                     that we sometimes need sys/types.h to get fitsio.h
*                     to compile.
* ssmedley  07/09/01  created
*
*/

/************************************************************************
*   NAME
*	region - verify image positions within FITS image.
*   SYNOPSIS
*
*
*   DESCRIPTION
*
*	This program, region, takes a FITS file containing a UVES
*	slit image & an image positions file & outputs a second
*	FITS file containing an image. The output FITS image, however,
*	has the image positions "drawn" on it. This is to verify
*	that the image positions are correct for the slit image.
*
*	Use -i commanad-line argument to specify image positions
*	file.
*
*   ENVIRONMENT
*
*   CAUTIONS
*
*   EXAMPLES
*
*   SEE ALSO
*
*   BUGS
*
*------------------------------------------------------------------------
*/


static const char *rcsId="@(#) $Id: ACMM:ozpozFpcol/ws/src/fpcolMdlCalRegion.c,v 1.63 07-Oct-2008 11:10:27+10 tjf $";

static void *use_rcsId = (0 ? (void *)(&use_rcsId) : (void *) &rcsId);
#define __EXTENSIONS__    /* Pick up extra include file stuff on solaris */
#define _HPUX_SOURCE      /* Pick up extra include file stuff on HP UX */

#include <assert.h>
#include <stdlib.h>	/* system */
#include <sys/types.h>
#include <unistd.h>     /* getopt sometimes */
#include "fitsio.h"

#include "fpcolMdlCalLoadRegions.h"	/* FibreRegion loadImageLocations */

/*
 * draw a box in a FITS image.
 * boundaries of box specified by <r>.
 */
void drawBoundingBox(
	fitsfile	*pFile,
	FibreRegion	r,
	int			nCols)
{
	int	aValues[nCols];
	int	i;
	int	width = r.br.x - r.tl.x + 1;
	int height = r.tl.y - r.br.y + 1;
	int fitsStatus = 0;

	assert(width > 0);
	assert(height > 0);

	for (i = 0; i < nCols; i++)
	{
		aValues[i] = 1000;
	}

	/* top line */
	fits_write_img(pFile,
				   TINT,
				   xy2index(r.tl.x, r.tl.y, nCols),	/* first element */
				   width,	/* number elements */
				   aValues,
				   &fitsStatus);
	/* bottom line */
	fits_write_img(pFile,
				   TINT,
				   xy2index(r.tl.x, r.br.y, nCols),	/* first element */
				   width,	/* number elements */
				   aValues,
				   &fitsStatus);

	/* left & right lines */
	for (i = 1; i < height - 1; i++)
	{
		/* left */
		fits_write_img(pFile,
					   TINT,
					   xy2index(r.tl.x, r.br.y + i, nCols),	/* first element */
					   1,	/* number elements */
					   aValues,
					   &fitsStatus);
		/* right */
		fits_write_img(pFile,
					   TINT,
					   xy2index(r.br.x, r.br.y + i, nCols),	/* first element */
					   1,	/* number elements */
					   aValues,
					   &fitsStatus);
	}

	if (fitsStatus != 0)
		fits_report_error(stderr, fitsStatus);
}


void usage (const char *pName)
{
	fprintf(stderr, "usage: %s [-options] <FITSfile>\n", pName);
	fprintf(stderr, "options:\n");
	fprintf(stderr, "-h        help\n");
	fprintf(stderr, "-i <file> file containing image positions [img.positions]\n");
	fprintf(stderr, "-o <file> output FITS file [region.fits]\n");
	fprintf(stderr, "-v        verbose\n");
}

int main (int argc, char *argv[])
{
	fitsfile	*pFile = NULL;
	fitsfile	*pOutput = NULL;
	int			fitsStatus = 0;
	char		*outputFilename = "region.fits";
	int			i;
	char		command[512];
	char		*pPositionsFile = "img.positions";
	int			nCols = -1;
	FibreRegion	aFibreRegions[32];
	FibreRegion	bkgRegion1, bkgRegion2;
	int			nRegions = 0;
	int			c;
	int			bVerbose = 0;

	while ((c = getopt(argc, argv, "i:ho:v")) != EOF)
	{
		switch (c)
		{
			case 'i':
				pPositionsFile = optarg;
				break;
			case 'h':
				usage(argv[0]);
				exit(0);
				break;
			case 'o':
				outputFilename = optarg;
				break;
			case 'v':
				bVerbose = 1;
				break;
			case '?':
				usage(argv[0]);
				exit(1);
				break;
		}
	}

	if (optind + 1 != argc)
	{
		fprintf(stderr, "Must specify 1 FITS file only.\n");
		usage(argv[0]);
		exit(1);
	}

	fits_open_file(&pFile, argv[optind], READONLY, &fitsStatus);
	if (fitsStatus != 0)
	{
		fits_report_error(stderr, fitsStatus);
		exit(2);
	}

	sprintf(command, "cp %s %s", argv[optind], outputFilename);
	if (system(command) != 0)
	{
		fprintf(stderr, "Couldn't copy FITS file to %s.\n", outputFilename);
		fits_close_file(pFile, &fitsStatus);
		exit(2);
	}

	fits_open_file(&pOutput, outputFilename, READWRITE, &fitsStatus);
	if (fitsStatus != 0)
	{
		fits_report_error(stderr, fitsStatus);
		fits_close_file(pFile, &fitsStatus);
		exit(2);
	}

	/* get coordinates of background & fibre regions */
	if (loadImageLocations(pPositionsFile,
						   aFibreRegions,
						   &bkgRegion1,
						   &bkgRegion2,
						   &nRegions) != 0)
	{
		fprintf(stderr, "Couldn't load image regions from %s.\n", pPositionsFile);
		fits_close_file(pFile, &fitsStatus);
		fits_close_file(pOutput, &fitsStatus);
		exit(3);
	}


	/* determine the xdimension of the FITS image */
	fits_read_key(pFile, TINT, "NAXIS1", &nCols, NULL, &fitsStatus);
	if (fitsStatus != 0)
	{
		fprintf(stderr, "FITS file does not appear to have UVES slit image.\n");
		fits_report_error(stderr, fitsStatus);
		exit(3);
	}
	assert(nCols > 0);

	/* draw bounding box for background region */
	drawBoundingBox(pOutput, bkgRegion1, nCols);
	drawBoundingBox(pOutput, bkgRegion2, nCols);
	/* draw bounding box for all fibre regions */
	for (i = 0; i < nRegions; i++)
	{
		drawBoundingBox(pOutput, aFibreRegions[i], nCols);
	}

	fits_close_file(pFile, &fitsStatus);
	fits_close_file(pOutput, &fitsStatus);

	printf("Output file is: %s\n", outputFilename);
	exit(0);
}
