/*******************************************************************************
* E.S.O. - VLT project
*
* "@(#) $Id: ACMM:ozpozFpcol/ws/src/fpcolMdlExtract.c,v 1.63 07-Oct-2008 11:10:26+10 tjf $"
* 
*
* who       when      what
* --------  --------  ----------------------------------------------
* tfarrell  13/08/01  created
*                  
*/

/************************************************************************
*   NAME
*       fpcolMdlExtract - Extract FLAMES/OzPoz Telescope model from robot DB.
*
*   SYNOPSIS
*       fpcolMdlExtract plate 
*
*   DESCRIPTION
*       This program will extract the OzPoz telescope model for the specified
*       plate from the FLAMES FP Database.  The model is written to the file
*       named "OzPozModel%d.sds" where %d = the plate number (1 to 4).
*      
*
***
*
*   ENVIRONMENT
*       lopoz - Flames Fibre Positioner.
*
*   CAUTIONS 
*
*   EXAMPLES
*
*   SEE ALSO
*       fpcolMdl module, fpcolMdlLoad
*   BUGS   
*
*------------------------------------------------------------------------
*/


/*
 * Define module id for error macros
 */

#define MODULE_ID     "fpcol"
#define lccMODULE_ID     "fpcol"

#define _POSIX_SOURCE 1

static const char *rcsId="@(#) $Id: ACMM:ozpozFpcol/ws/src/fpcolMdlExtract.c,v 1.63 07-Oct-2008 11:10:26+10 tjf $";

static void *use_rcsId = (0 ? (void *)(&use_rcsId) : (void *) &rcsId);


/*
 *  Include files
 */
#include "vltPort.h"
#include "fpcolMdl.h"
#include "fpcolMdl_err.h"
#include "fpcol.h"

#include <stdlib.h>

/*
 * Print error details.
 */
static void printError(const char *context, ccsERROR *error)
{
    fprintf(stderr,"%s, Err Num %d, Location %s, Module %s\n",
            context,
            error->errorNumber,
            error->location,
            error->moduleId);
    fprintf(stderr,"%s\n", error->runTimePar);
}


extern int main(int argc, char *argv[])
{
    int plate;
    ccsERROR error;
    FpilModelType *model = 0;
    StatusType status = STATUS__OK;
    double modelPars[fpMDL_NUM_PARS];
    vltBYTES64 attrName;
    dbTYPE      dbType[dbMAX_FIELD_CNT];
    vltINT32    dbActual;
    vltUINT16   dbRecCnt;
    dbATTRTYPE  dbAttrType;

/*
 *  Validate arguments.
 */
    if (argc != 2) 
        {
        fprintf(stderr, "%s:Wrong number of arguments (%d instead of 1)\n",
                argv[0], argc-1);
        fprintf(stderr, "Usage:%s plateNo(1-4)\n", argv[0]);
        return 1;
        }

    plate = atoi(argv[1]);
    if ((plate < 1) || (plate > fpNUM_PLATES))
        {
        fprintf(stderr,"%s:Invalid plate number %d\n",argv[0], plate);
        fprintf(stderr, "Usage:%s plateNo(1-4)\n", argv[0]);
        return 1;
        }

/*
 *  Initialise access to the telescope.
 */
    if (ccsInit("fpcolMdlLoadModel", 0, NULL, NULL, &error) != SUCCESS)
        {
        fpcolMdlFree(model, &status);
        printError("Error initialising access to CCS DB", &error);
        return 1;
        }

/*
 *  Read from the Data Base.
 *
 *  Create the name of the DB item.
 */
    sprintf((char *)attrName, "<alias>PCC:plate_%d.telMdlPars", plate);
/*
 *  Read the attribute.
 */
    if (dbReadSymbolic(attrName, 
                       dbEMPTY, 
                       dbType, 
                       (char *)modelPars,
                       sizeof(double)*fpMDL_NUM_PARS,
                       &dbActual, 
                       &dbRecCnt,
                       &dbAttrType, 
                       &error) != SUCCESS)
    {
        printError("Error reading from CCS DB", &error);
        ccsExit(&error);
        return 1;
    }

  
/*
 *  Tidy up CCS access.
 */
    ccsExit(&error);
/*
 *  Read the telescope model file.
 */
    fpcolMdlCreateModel(fpMDL_NUM_PARS, modelPars, plate, &model, &status);

    if (status != STATUS__OK)
        {
        fpcolDRAMAErrorToVLTError(status, __FILE__,  &error);
        printError("Error creating telescope model", &error);
        return 1;
        }
/*
 *  Write the telescope model to the file.
 */
    fpcolMdlWriteModel(model, 0, &status);
    if (status != STATUS__OK)
        {
        fpcolMdlFree(model, &status);
        fpcolDRAMAErrorToVLTError(status, __FILE__,  &error);
        printError("Error writing telescope model parameters file", &error);
        return 1;
        }

/*
 *  Tidy up.
 */
    fpcolMdlFree(model, &status);

    return(0);

}

