/*******************************************************************************
* E.S.O. - VLT project
*
* "@(#) $Id: ACMM:ozpozFpcol/ws/src/fpcolMdlLoad.c,v 1.63 07-Oct-2008 11:10:25+10 tjf $"
* 
*
* who       when      what
* --------  --------  ----------------------------------------------
* rschmutz  24/04/02  -> lflics1.
* tfarrell  17/04/02  Support writing to the WS OLDB.
* tfarrell  13/08/01  created
*                  
*/

/************************************************************************
*   NAME
*       fpcolMdlLoad - Load FLAMES/OzPoz Telescope model into Robot DB.
*
*   SYNOPSIS
*       fpcolMdlLoad plate 
*
*   DESCRIPTION
*       This program will load the OzPoz telescope model for the specified
*       plate into the FLAMES FP Database.  The model is read from the file
*       named "OzPozModel%d.sds" where %d = the plate number (1 to 4) and
*       normally written to the LCU (lflics1) OLDB (which is scanned back to
*       the WS (wflames).
*    
*       If the plate number is specified as negative, then we want to
*       write to the WS (wflames) OLDB.  This is only sensible if scanning
*       of the LCU is not active.
*
***
*
*   ENVIRONMENT
*       lopoz - Flames Fibre Positioner.
*
*   CAUTIONS 
*
*   EXAMPLES
*
*   SEE ALSO
*       fpcolMdl module, fpcolMdlExtract
*   BUGS   
*
*------------------------------------------------------------------------
*/


/*
 * Define module id for error macros
 */

#define MODULE_ID     "fpcol"
#define lccMODULE_ID     "fpcol"

#define _POSIX_SOURCE 1

static const char *rcsId="@(#) $Id: ACMM:ozpozFpcol/ws/src/fpcolMdlLoad.c,v 1.63 07-Oct-2008 11:10:25+10 tjf $";

static void *use_rcsId = (0 ? (void *)(&use_rcsId) : (void *) &rcsId);


/*
 *  Include files
 */
#include "vltPort.h"
#include "fpcolMdl.h"
#include "fpcolMdl_err.h"
#include "fpcol.h"

#include <stdlib.h>

/*
 * Print error details.
 */
static void printError(const char *context, ccsERROR *error)
{
    fprintf(stderr,"%s, Err Num %d, Location %s, Module %s\n",
            context,
            error->errorNumber,
            error->location,
            error->moduleId);
    fprintf(stderr,"%s\n", error->runTimePar);
}

static void reportRoutine (
    void * clientData,
    const char *string,
    StatusType *status)
{
    if (*status != STATUS__OK) return;

    printf("%s\n", string);
}



extern int main(int argc, char *argv[])
{
    int plate;
    ccsERROR error;
    FpilModelType *model = 0;
    StatusType status = STATUS__OK;
    double modelPars[fpMDL_NUM_PARS];
    vltBYTES64 attrName;
    vltINT32 actual;
    dbTYPE   dataTypeDouble[2] = {dbDOUBLE, dbUNDEFINED};
    vltLOGICAL writeToWS =  ccsFALSE;
/*
 *  Validate arguments.
 */
    if (argc != 2) 
        {
        fprintf(stderr, "%s:Wrong number of arguments (%d instead of 1)\n",
                argv[0], argc-1);
        fprintf(stderr, "Usage:%s plateNo(1-4)\n", argv[0]);
        return 1;
        }

    plate = atoi(argv[1]);

/*
 * If we have a negative plate number we will write to the
 * Workstation OLDB rather then the LCU OLDB.  We don't normally
 * want to do this as the parameters are scaned from the LCU to WS,
 * so THE WS OLDB attriburtes are overwritten.  But if we don't have
 * the LCU and are simulating the system, then we will want to write
 * to the WS.
 */
    if (plate < 0)
      {
        writeToWS = ccsTRUE;
        plate *= -1;
      }

    if ((plate < 1) || (plate > fpNUM_PLATES))
        {
        fprintf(stderr,"%s:Invalid plate number %d\n",argv[0], plate);
        fprintf(stderr, "Usage:%s plateNo(1-4)\n", argv[0]);
        return 1;
        }
/*
 *  Read the telescope model file.
 */
    fpcolMdlReadModel(0, 0, reportRoutine, 0, plate, &model, &status);

    if (status != STATUS__OK)
        {
        fpcolDRAMAErrorToVLTError(status, __FILE__,  &error);
        printError("Error reading telescope model from file", &error);
        return 1;
        }
/*
 *  Read the parameters from the telescope model.
 */
    fpcolMdlGetModel(model, fpMDL_NUM_PARS, modelPars, &status);
    if (status != STATUS__OK)
        {
        fpcolMdlFree(model, &status);
        fpcolDRAMAErrorToVLTError(status, __FILE__,  &error);
        printError("Error getting telescope model parameters", &error);
        return 1;
        }

/*
 *  Initialise access to the telescope.
 */
    if (ccsInit("fpcolMdlLoadModel", 0, NULL, NULL, &error) != SUCCESS)
        {
        fpcolMdlFree(model, &status);
        printError("Error initialising access to CCS DB", &error);
        return 1;
        }


/*
 * Write into the Data Base - note that this item is scaned from the
 * the LCU to the WS so we must write through to the LCU for this
 * to take effect.
 *
 * Create the name of the DB item.
 */
    if (writeToWS)
      {
        printf("Warning - writing to \"wflames\" OLDB rather then \"lflics1\"\n");
        printf("          this only makes sense if scanning is not active\n");
        sprintf((char *)attrName, "@wflames<alias>PCC:plate_%d.telMdlPars", plate);
      }
    else
      sprintf((char *)attrName, "@lflics1<alias>PCC:plate_%d.telMdlPars", plate);
/*
 * Write the attribute.
 */
    if (dbWriteSymbolic(attrName, 
                        dbEMPTY,
                        dataTypeDouble, 
                        (char *)modelPars,
                        sizeof(double)*fpMDL_NUM_PARS,
                        &actual,
                        fpMDL_NUM_PARS, 
                        dbVECTOR, 
                        &error) != SUCCESS)
    {
        printError("Error writing to CCS DB", &error);
    } 


/*
 *  Tidy up.
 */
    fpcolMdlFree(model, &status);

    ccsExit(&error);
    return(0);

}

