/*******************************************************************************
* E.S.O. - VLT project
*
* "@(#) $Id: ACMM:ozpozFpcol/ws/src/fpcolMdlPowell.c,v 1.63 07-Oct-2008 11:10:26+10 tjf $
* 
*
* who       when      what
* --------  --------  ----------------------------------------------
* tfarrell   19/10/02  Another correction to the function based on
*                      analysis by Mathew Colless.
* tfarrell   18/10/02  Add tolerance, maxIts and mode parameters.  mode is
*                      used to select the function to minimize - since
*                      the sense of the function being minimize is unclear.
*                      The default (mode 1) is described below.
 tfarrell   28/08/02  Created.
*/

/************************************************************************
*   NAME
*       fpcolMdlPowell - Model fit for OzPoz using Powell technique.
*
*   SYNOPSIS
*       #include "fpcolMdl.h"
*
*
*        void fpcolMdlPowell(
*                      int nobs, 
*                      FpilModelType *model, 
*                      double tolerance,
*                      int maxIts,
*                      int mode, 
*                      double mjd, 
*                      double rc, 
*                      double dc,
*                      double *ra, 
*                      double *dec, 
*                      double *dra, double *ddec,
*                      int *pfit, 
*                      double *pvar, 
*                      double *rms, 
*                      double *psd, 
*                      StatusType *status);
*
*   DESCRIPTION
*     Ths routine does a Powell minimization of the following formular
*
*
*     sqrt|sumof(j=1 -> nobs)(f([Pi],[SKYj] - f([Pf],[SKYj]-[OFFSETj]))^2)|
*         |----------------------------------------------------------------|
*         |                   nobs                                         |
*
*     Where
*            [Pi]       = Our initial model parameters
*            [Pf]       = Is the fitted parameters were are trying to determine
*              f        = Our linear and distortion model we want parameters for.
*            [SKYj]     = The sky coorindates for each observation
*            [OFFSETj]  = The offset required to centre each observation
*         
*     "tolerance" is the minimum change the fit is looking for. (microns).
*     "maxIts" is the maximum number of iterations to try.
*     "mode" should normally be 1.  Mode = 2 is used for debugging purposes (currenlty
*     it minimizes a slight change to the formular above.
*
*     "nobs" is the number of fibres  used.  "model" contains the telescope model. 
*     On input this must have  the same parameters as that used to configure the 
*     fibres for the test.  On output 
*     it will contain the fitted parameters.  "rc" should contain the field 
*     centre apparent RA (radians), and "dec" the Declination (radians).
*     "dra" and "ddec" are array sof size nobs containing the apparent RA
*     and Declination offsets needed to centre the star on the fibre.
*     "pfit" is an array specifying which parameters are to be fitted. 
*     It should have a value of zero if the corresponding parameter is
*     to be fitted. one to fix the parameter at its initial value.
*
*     "pvar" is an array of variances on the fitted parameters.
*     "rms" is the RMS of the fit in microns.
*     "psd " is the estimated Population Standard Deviation of the fit in microns.  
*           This is only valid if the number of objects exceeds the number of parameters being 
*           fited.  Otherwise it is set to 0
*           
*     "status" is a DRAMA style modified status argument.
*
*     The scheme for a powell minimization comes from Numerical Recipies, as does
*     the required code.
*
***
*   ENVIRONMENT
*
*   CAUTIONS 
*
*   EXAMPLES
*
*   SEE ALSO
*       fpcolMdlFit (least squares fit), fpcol module, AAO FPIL Library.
*   BUGS   
*"
*------------------------------------------------------------------------
*/

/*
 * Define module id for error macros
 */

#define MODULE_ID     "fpcol"

static const char *rcsId="@(#) $Id: ACMM:ozpozFpcol/ws/src/fpcolMdlPowell.c,v 1.63 07-Oct-2008 11:10:26+10 tjf $";

static void *use_rcsId = (0 ? (void *)(&use_rcsId) : (void *) &rcsId);


#define _POSIX_SOURCE 1
#define _XOPEN_SOURCE 1

#include <stdlib.h>     /* For malloc and free */
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <limits.h>
#include <float.h>
/*
 * Before including fpcolMdl.h we must define this macro to ensure
 * the correct include files are brought in.
 */
#define FPCOL_MDL_MODULE
#include "fpcolMdlCalLibs.h"
#include "fpcolMdlCalModel.h"
#include "fpDefines.h" /* For fpMDL_NUM_PARS */

/*
 * Slalib include files.
 */
#include "slalib.h"
#include "slamac.h"
static void scaleUpSet(double *p);
static double FunctionToMinimize(double *pars);
static void powell(double p[], double **xi, int n, double ftol, int *iter, double *fret,
                   double (*func)(double []));
static double **matrix(long nrl, long nrh, long ncl, long nch);
static void free_matrix(double **m,long nrl, long nrh, long ncl, long nch);
static double *vector(long nl, long nh);
static void free_vector(double *v, long nl, long nh);

static int powellMaxIts;
static int powellMode;
/*
 *
 *
 * 
 */

typedef struct { 
    double scaleFactor;
    int    index;
} ParDetails;

struct {
    double *origModelX;  /* Should be set to array of length nobs and */
    double *origModelY;  /*     field in with the results of the orignil model */
    double *ra;         /* Pointer to ra values for each obs */
    double *dec;        /* Pointer to dec values for each obs */
    double *ora;         /* Pointer to ra offset for each obs */
    double *odec;        /* Pointer to dec offset for each obs */
    double cra;          /* Field centre ra */
    double cdec;         /* Field centre dec */
    int    *widx;        /* Pointer to the wavelength windex for each offset */
    double mjd;          /* MJD of observation */
    int    nobs;
    int    nFitPars;     /* Number of parameters being fitted */
    FpcalMdl   *models;  /* Models pointer */
    FpcalPars origPars;  /* Original model parameters */
    ParDetails *parDetails; /* Pointer to details of parameters we are fitting */
} WorkStruct;  

extern void fpcolMdlPowell(int nobs, FpcalMdl *models, 
                           double tolerance, int maxIts, int mode, 
                           double mjd, double cra, double cdec,
                           double *ra, double *dec, double *dra, double *ddec, int *widx,
                           int *pfit, double *pvar, double *rms, double *psd, 
                           StatusType *status)
              
{
 
    register int i,j;

    double *parVector;
    double **dirMatrix;
    int iter;
    double fret;

    /*
     * These two values are used deep in the code - so we copy them to
     * static variables.
     */ 
    powellMaxIts = maxIts;
    powellMode   = mode;
    /*
     * Get some space for the results applied to the original model.
     */
    WorkStruct.origModelX = (double *) calloc(nobs, sizeof(double));
    WorkStruct.origModelY = (double *) calloc(nobs, sizeof(double));
    /*
     * Copy things needed to run the function we want to minmize.
     */
    WorkStruct.ra   = ra;
    WorkStruct.dec  = dec;
    WorkStruct.ora  = dra;
    WorkStruct.odec = ddec;
    WorkStruct.cra  = cra;
    WorkStruct.cdec = cdec;
    WorkStruct.widx = widx;
    WorkStruct.mjd  = mjd;
    WorkStruct.nobs = nobs;
    WorkStruct.models = models;
    /*
     * Get the original parameters.
     */
    WorkStruct.origPars = *fpcolMdlCalGetPars(models);

    /*
     * Work out the plate positions for all objects using the original model.
     */
    for (i = 0 ; i < nobs ; ++i)
        {
        double x,y;
        /* Index to telescope model for wavelength assoicated with this offset*/
        FpilModelType *telModel = fpcolMdlCalWaveModel(models, widx[i]);
        
        if (mode == POWELL_MODE_2)
            fpcolMdlRd2Xy(telModel, cra, cdec, 
                          (ra[i]  + WorkStruct.ora[i]), 
                          (dec[i] + WorkStruct.odec[i]),
                          mjd,&x,&y,status);
        else
             fpcolMdlRd2Xy(telModel, cra, cdec, ra[i], dec[i],mjd,&x,&y,status);
           
        fpcolMdlXy2Pos(telModel, x, y, 
                       &WorkStruct.origModelX[i], 
                       &WorkStruct.origModelY[i], 
                       status);
        }
    

    /*
     * count the number of parameters we are fitting and then malloc space to
     * store info about them.
     */
    WorkStruct.nFitPars = 0;  
    for (i = 0; i < fpMDL_NUM_PARS ; ++i)
        {
        if (pfit[i] == 0)
            ++WorkStruct.nFitPars;
        }
    printf("We will fit %d parameter%s\n", WorkStruct.nFitPars,
           ((WorkStruct.nFitPars != 1) ? "s" : "" ));
    WorkStruct.parDetails = (ParDetails *)malloc(WorkStruct.nFitPars * sizeof(ParDetails));


    /*
     * Set up the vector and matrix for the powell function
     */
    parVector = vector(1,  WorkStruct.nFitPars );
    dirMatrix = matrix(1,  WorkStruct.nFitPars, 1,  WorkStruct.nFitPars);

    /*
     * Now for each parameter we are fitting, we want to set up WorkStruct.parDetails
     * to contain the scale factor and index into WorkStruct.origPars.  We also need
     * parVector to contain the inital parameter - which is zero for items which
     * are zero or one otherwise.
     */
    j = 0;
    for (i = 0; i < fpMDL_NUM_PARS ; ++i)
        {
        if (pfit[i] == 0)
            {
            WorkStruct.parDetails[j].index = i;
            if (WorkStruct.origPars.p[i] == 0)
                {
                WorkStruct.parDetails[j].scaleFactor = 1;
                parVector[j+1] = 0;
                }
            else 
                {
                WorkStruct.parDetails[j].scaleFactor = WorkStruct.origPars.p[i];
                parVector[j+1] = 1;
                }            
            ++j;
            }
        
        }

    /*
     *  The direction matrix should start out with WorkStruct.nFitPars unit vectors.
     */
    for (i = 1 ; i <=   WorkStruct.nFitPars  ; ++i)
        {
        for (j = 1; j <=   WorkStruct.nFitPars  ; ++j)
            {
            if (i == j)
                dirMatrix[i][j] = 1;
            else
                dirMatrix[i][j] = 0;
            }
        }


    if ( WorkStruct.nFitPars)
        {
        powell(parVector, dirMatrix,  WorkStruct.nFitPars , 
               tolerance, &iter, &fret, FunctionToMinimize);
        printf("Took %d iteractions\n", iter);
        }
    else
        {
        /*
         * Nothing to fit - we just want the result.  The Set up function should
         * just give us the existing model in this case.
         */
        fret = FunctionToMinimize(parVector);
        }
    /*
     *  Scale up the result and set as the new model.
     */
    scaleUpSet(parVector);
    /*
     *  Set the variances.
     */
    for (i = 0; i < fpMDL_NUM_PARS ; ++i)
        {
        pvar[i] = 0;  /* FUDGE = I don't know how to do this yet */
        }
    /*
     * Get population standard deviation, if possible.
     */
    if (nobs > WorkStruct.nFitPars)
        {
        double dnobs = nobs;
        double var = fret * fret;   /* Reverse squareroot used in getting fret result */
        *psd = sqrt(var*dnobs/(nobs - WorkStruct.nFitPars));
        }
    else
        *psd = 0;

    /*
     * The fret returned value is the rms.
     */
    *rms = fret;

    printf("Powell Fit completed (%d)\n", mode);

     /*
     * Tidy up.
     */
    free(WorkStruct.origModelX);
    free(WorkStruct.origModelY);
    free(WorkStruct.parDetails);

    free_vector(parVector, 1,  WorkStruct.nFitPars );
    free_matrix(dirMatrix, 1,  WorkStruct.nFitPars, 1,  WorkStruct.nFitPars);


}

/*
 * Create a telescope model parameters array based on the original parameters
 * (in WorkStruct.origPars) where we are not fitting a parameter and a combination
 * of the value in the "p" vector and the scaling factor for a parameter for the parameters
 * we are fitting.  Then set that model.
 *
 * Note - p is a Numerical Recipes vector, 1 based, rather then 0 based.
 */
static void scaleUpSet(double *p)
{
    int i;
    FpcalPars pars;
    pars = WorkStruct.origPars;
    for (i = 0 ; i <   WorkStruct.nFitPars  ; ++i)
        {
        pars.p[WorkStruct.parDetails[i].index] =
            p[i+1] * WorkStruct.parDetails[i].scaleFactor;
        }
    fpcolMdlCalSetPars(&pars, WorkStruct.models);
}

static double FunctionToMinimize(double *pars)
{
    double sumOfSquares = 0;
    register int i;
    StatusType status = STATUS__OK;

    /*
     * Use the supplied parameters to set the telescope model.
     */
    scaleUpSet(pars);

    /*
     * Work out the X and Y for each observation using this model.
     */
    for (i = 0 ; i < WorkStruct.nobs ; ++i)
        {
        double x,y;
        double newX, newY;
        double xDiff, yDiff;
        /* Index to telescope model for wavelength assoicated with this offset*/
        FpilModelType *telModel = fpcolMdlCalWaveModel(WorkStruct.models, WorkStruct.widx[i]);
        if (powellMode == POWELL_MODE_2)
            fpcolMdlRd2Xy(telModel, WorkStruct.cra, WorkStruct.cdec,
                          WorkStruct.ra[i], WorkStruct.dec[i],
                          WorkStruct.mjd,&x,&y,&status);
        else
            fpcolMdlRd2Xy(telModel, WorkStruct.cra, WorkStruct.cdec,
                          (WorkStruct.ra[i]  - WorkStruct.ora[i]), 
                          (WorkStruct.dec[i] - WorkStruct.odec[i]),
                          WorkStruct.mjd,&x,&y,&status);
 
        fpcolMdlXy2Pos(telModel, x, y, &newX, &newY, &status);
        
        xDiff = newX - WorkStruct.origModelX[i];
        yDiff = newY - WorkStruct.origModelY[i];
        sumOfSquares += (xDiff*xDiff)+(yDiff*yDiff);
        }
    
    return  sqrt(sumOfSquares/WorkStruct.nobs);
}


/*
 *--------------------------
 *  The functions from Numerical Recipes in C required to run the Powell method, including
 * Powell itself.
 *
 */

#define NRANSI

#define BRITMAX 100
#define CGOLD 0.3819660
#define ZEPS 1.0e-10
#define SHFT(a,b,c,d) (a)=(b);(b)=(c);(c)=(d);

#define GOLD 1.618034
#define GLIMIT 100.0
#define TINY 1.0e-20
#define TOL 2.0e-4

static double sqrarg;
#define SQR(a) ((sqrarg=(a)) == 0.0 ? 0.0 : sqrarg*sqrarg)
#define SIGN(a,b) ((b) >= 0.0 ? fabs(a) : -fabs(a))
static double maxarg1,maxarg2;
#define FMAX(a,b) (maxarg1=(a),maxarg2=(b),(maxarg1) > (maxarg2) ?\
        (maxarg1) : (maxarg2))
#define NR_END 1
#define FREE_ARG char*

static double brent(double ax, double bx, double cx,
                   double (*f)(double), double tol, double *xmin);
static double f1dim(double x);
static void mnbrak(double *ax, double *bx, double *cx, double *fa, double *fb,
                   double *fc, double (*func)(double));
static void linmin(double p[], double xi[], int n, double *fret,
                   double (*func)(double []));

static void nrerror(error_text)
char error_text[];
/* Numerical Recipes standard error handler */
{
	void exit();

	fprintf(stderr,"Numerical Recipes run-time error...\n");
	fprintf(stderr,"%s\n",error_text);
	fprintf(stderr,"...now exiting to system...\n");
	exit(1);
}



static double *vector(long nl, long nh)
/* allocate a double vector with subscript range v[nl..nh] */
{
	double *v;

	v=(double *)malloc((size_t) ((nh-nl+1+NR_END)*sizeof(double)));
	if (!v) nrerror("allocation failure in vector()");
	return v-nl+NR_END;
}

static void free_vector(double *v, long nl, long nh)
/* free a double vector allocated with vector() */
{
    free((FREE_ARG) (v+nl-NR_END));
    nh = 0; /* Just to get rid of unused variable warning */
}

static int ncom;
static double *pcom,*xicom,(*nrfunc)(double []);

static double f1dim(double x)
{
	int j;
	double f,*xt;

	xt=vector(1,ncom);
	for (j=1;j<=ncom;j++) xt[j]=pcom[j]+x*xicom[j];
	f=(*nrfunc)(xt);
	free_vector(xt,1,ncom);
	return f;
}




static void linmin(double p[], double xi[], int n, double *fret, double (*func)(double []))
{

	int j;
	double xx,xmin,fx,fb,fa,bx,ax;

	ncom=n;
	pcom=vector(1,n);
	xicom=vector(1,n);
	nrfunc=func;
	for (j=1;j<=n;j++) {
		pcom[j]=p[j];
		xicom[j]=xi[j];
	}
	ax=0.0;
	xx=1.0;
	mnbrak(&ax,&xx,&bx,&fa,&fx,&fb,f1dim);
	*fret=brent(ax,xx,bx,f1dim,TOL,&xmin);
	for (j=1;j<=n;j++) {
		xi[j] *= xmin;
		p[j] += xi[j];
	}
	free_vector(xicom,1,n);
	free_vector(pcom,1,n);
}

static void mnbrak(double *ax, double *bx, double *cx, double *fa, double *fb, double *fc,
                   double (*func)(double))
{
	double ulim,u,r,q,fu,dum;

	*fa=(*func)(*ax);
	*fb=(*func)(*bx);
	if (*fb > *fa) {
		SHFT(dum,*ax,*bx,dum)
		SHFT(dum,*fb,*fa,dum)
	}
	*cx=(*bx)+GOLD*(*bx-*ax);
	*fc=(*func)(*cx);
	while (*fb > *fc) {
		r=(*bx-*ax)*(*fb-*fc);
		q=(*bx-*cx)*(*fb-*fa);
		u=(*bx)-((*bx-*cx)*q-(*bx-*ax)*r)/
			(2.0*SIGN(FMAX(fabs(q-r),TINY),q-r));
		ulim=(*bx)+GLIMIT*(*cx-*bx);
		if ((*bx-u)*(u-*cx) > 0.0) {
			fu=(*func)(u);
			if (fu < *fc) {
				*ax=(*bx);
				*bx=u;
				*fa=(*fb);
				*fb=fu;
				return;
			} else if (fu > *fb) {
				*cx=u;
				*fc=fu;
				return;
			}
			u=(*cx)+GOLD*(*cx-*bx);
			fu=(*func)(u);
		} else if ((*cx-u)*(u-ulim) > 0.0) {
			fu=(*func)(u);
			if (fu < *fc) {
				SHFT(*bx,*cx,u,*cx+GOLD*(*cx-*bx))
				SHFT(*fb,*fc,fu,(*func)(u))
			}
		} else if ((u-ulim)*(ulim-*cx) >= 0.0) {
			u=ulim;
			fu=(*func)(u);
		} else {
			u=(*cx)+GOLD*(*cx-*bx);
			fu=(*func)(u);
		}
		SHFT(*ax,*bx,*cx,u)
		SHFT(*fa,*fb,*fc,fu)
	}
}


static double brent(double ax, double bx, double cx, double (*f)(double), double tol,
	double *xmin)
{
	int iter;
	double a,b,d,etemp,fu,fv,fw,fx,p,q,r,tol1,tol2,u,v,w,x,xm;
	double e=0.0;

	a=(ax < cx ? ax : cx);
	b=(ax > cx ? ax : cx);
	x=w=v=bx;
	fw=fv=fx=(*f)(x);
	for (iter=1;iter<=BRITMAX;iter++) {
		xm=0.5*(a+b);
		tol2=2.0*(tol1=tol*fabs(x)+ZEPS);
		if (fabs(x-xm) <= (tol2-0.5*(b-a))) {
			*xmin=x;
			return fx;
		}
		if (fabs(e) > tol1) {
			r=(x-w)*(fx-fv);
			q=(x-v)*(fx-fw);
			p=(x-v)*q-(x-w)*r;
			q=2.0*(q-r);
			if (q > 0.0) p = -p;
			q=fabs(q);
			etemp=e;
			e=d;
			if (fabs(p) >= fabs(0.5*q*etemp) || p <= q*(a-x) || p >= q*(b-x))
				d=CGOLD*(e=(x >= xm ? a-x : b-x));
			else {
				d=p/q;
				u=x+d;
				if (u-a < tol2 || b-u < tol2)
					d=SIGN(tol1,xm-x);
			}
		} else {
			d=CGOLD*(e=(x >= xm ? a-x : b-x));
		}
		u=(fabs(d) >= tol1 ? x+d : x+SIGN(tol1,d));
		fu=(*f)(u);
		if (fu <= fx) {
			if (u >= x) a=x; else b=x;
			SHFT(v,w,x,u)
			SHFT(fv,fw,fx,fu)
		} else {
			if (u < x) a=u; else b=u;
			if (fu <= fw || w == x) {
				v=w;
				w=u;
				fv=fw;
				fw=fu;
			} else if (fu <= fv || v == x || v == w) {
				v=u;
				fv=fu;
			}
		}
	}
	nrerror("Too many iterations in brent");
	*xmin=x;
	return fx;
}

static void powell(double p[], double **xi, int n, double ftol, int *iter, double *fret,
	double (*func)(double []))
{
	int i,ibig,j;
	double del,fp,fptt,t,*pt,*ptt,*xit;

	pt=vector(1,n);
	ptt=vector(1,n);
	xit=vector(1,n);
	*fret=(*func)(p);
	for (j=1;j<=n;j++) pt[j]=p[j];
	for (*iter=1;;++(*iter)) {
		fp=(*fret);
		ibig=0;
		del=0.0;
		for (i=1;i<=n;i++) {
			for (j=1;j<=n;j++) xit[j]=xi[j][i];
			fptt=(*fret);
			linmin(p,xit,n,fret,func);
			if (fabs(fptt-(*fret)) > del) {
				del=fabs(fptt-(*fret));
				ibig=i;
			}
		}
		if (2.0*fabs(fp-(*fret)) <= ftol*(fabs(fp)+fabs(*fret))) {
			free_vector(xit,1,n);
			free_vector(ptt,1,n);
			free_vector(pt,1,n);
			return;
		}
		if (*iter == powellMaxIts) nrerror("powell exceeding maximum iterations.");
		for (j=1;j<=n;j++) {
			ptt[j]=2.0*p[j]-pt[j];
			xit[j]=p[j]-pt[j];
			pt[j]=p[j];
		}
		fptt=(*func)(ptt);
		if (fptt < fp) {
			t=2.0*(fp-2.0*(*fret)+fptt)*SQR(fp-(*fret)-del)-del*SQR(fp-fptt);
			if (t < 0.0) {
				linmin(p,xit,n,fret,func);
				for (j=1;j<=n;j++) {
					xi[j][ibig]=xi[j][n];
					xi[j][n]=xit[j];
				}
			}
		}
	}
}

static double **matrix(long nrl, long nrh, long ncl, long nch)
/* allocate a double matrix with subscript range m[nrl..nrh][ncl..nch] */
{
	long i, nrow=nrh-nrl+1,ncol=nch-ncl+1;
	double **m;

	/* allocate pointers to rows */
	m=(double **) malloc((unsigned int)((nrow+NR_END)*sizeof(double*)));
	if (!m) nrerror("allocation failure 1 in matrix()");
	m += NR_END;
	m -= nrl;

	/* allocate rows and set pointers to them */
	m[nrl]=(double *) malloc((unsigned int)((nrow*ncol+NR_END)*sizeof(double)));
	if (!m[nrl]) nrerror("allocation failure 2 in matrix()");
	m[nrl] += NR_END;
	m[nrl] -= ncl;

	for(i=nrl+1;i<=nrh;i++) m[i]=m[i-1]+ncol;

	/* return pointer to array of pointers to rows */
	return m;
}

static void free_matrix(double **m,long nrl, long nrh, long ncl, long nch)
/* free a double matrix allocated by matrix() */
{
	free((FREE_ARG) (m[nrl]+ncl-NR_END));
	free((FREE_ARG) (m+nrl-NR_END));

        nrh = nch = 0;  /* Just to get rid of unused variable warning */
}
