/*******************************************************************************
* E.S.O. - VLT project
*
* "@(#) $Id: ACMM:ozpozFpcol/ws/test/fpcolTest.c,v 1.63 07-Oct-2008 11:10:30+10 tjf $"
*
*
* who       when      what
* --------  --------  ----------------------------------------------
* tfarrell  20/03/02  In doing fibre/button collision checks, should
*                     use fibre clearance rather then button clearance.
* tfarrell  07/07/00  created, Initial ws tests complete.
*/

/************************************************************************
*   NAME
*       fpcolTest - Test program for fpcol librayr.
*
*   SYNOPSIS
*       fpcolTest
*
*   DESCRIPTION
*
*  Program to test out the FPCOL collision detection routines
*
*  We have an array of fibre positions (Button end X/Y/Theta and Pivot 
*  position X/Y) (read from the file "fibre_table.txt") and another of 
*  fiducial mark positions (X/Y) (hard coded).
*
*  We have a matrix indicating where we expect a collision between 
*  fibres (three types, button-button, button-fibre and fibre-fibre) and
*  another matrix giving details of collisions between button/fibres and
*  fiducial marks.
*
*  All these arrays are initialised by hand to the correct values.
*  
*  We then run through each fibre and check for collisions against each
*  other fibre and each mark.  Where the result is different then
*  expected, we output an error message.
*
*  The reason for having the initial positions in X/Y rather then
*  R/T is that it forces a check on the versions in both directions and
*  is easier to use in conjuction with the configure program.
*
*  When you run this program, it should output
*
*     Collision detection test passed
*
*
*  Ohteriwse, it outputs the details of the test which failed.
*
***
*
*   CAUTIONS
*
*   EXAMPLES
*
*   SEE ALSO
*
*   BUGS
*
*------------------------------------------------------------------------
*/

#include "vltPort.h"    /* Must be first header */

#include <stdio.h>
#include <math.h>
#include <string.h>
#include "fppt.h"
#include "fpcol.h"
#include "fpcolInstDescr.h"
#include "fpcolBase.h"

/*
 * Define a structure used to hold details about a fibre button.
 */
typedef struct {
    double butX;           /* Button position in X */
    double butY;           /* Button position in Y */
    double butT;           /* Button position in theta (configure theta) */
    double pivX;           /* Fibre pivot position in X */
    double pivY;           /* Fibre pivot position in Y */
    vltLOGICAL invalidPos; /* Set true if this position should be invalid */

    double butR;           /* Button position in R */
    double butTheta;       /* Button position in (FLAMES) Theta */
    double butAlpha;       /* Button angle alpha  */
    double pivR;           /* Button pivot position in R */
    double pivTheta;       /* Button pivot position in (FLAMES) Theta */

    char   type;           /* IF = 'U', ignore.  This allows us to read
                              details from the "fibre_table.txt" file, 
                              generated by configure and to refer back to it */

} fibreType;

/*
 * Define a structure used to indicate if a collision occurs for
 * fibre-fibre combination
 */
typedef struct {
    vltLOGICAL butFib;  /* Button fibre collision occurs */
    vltLOGICAL butBut;  /* Button-Button collision occurs */
    vltLOGICAL fibFib;  /* Fibre-Fibre collision occurs */
} fibColType;

/*
 * Define a structure representing fiducial marks.
 */
typedef struct {
    double x;    /* Mark position in X */  
    double y;    /* Mark position in Y */
    double r;    /* Mark position in R */
    double t;    /* Mark position in Theta */
} markType;
        

/*
 *  Define the number of fibres and fiduical marks we are going to check
 */
#define NUM_FIBRES 504
#define NUM_MARKS  80

/*
 * Array of fibre details.
 */
static fibreType fibres[NUM_FIBRES];

/* 
 * These fiducial mark positions are generated from one of the sky grid 
 * rings and therefore there is a good change of some collisions etc. and
 * we can check that by displaying the configuration and seeing if the
 * fibre in obsure a sky mark at the specified position.  Note that I
 * beleive I only have an equivalence of the first 80 positions from
 * the outer ring of sky grid 
 * positions.  Note that fiducial marks are very small so a collision
 * between the fibre and a mark is unlikely and hard to see in configure.
 */
static markType  marks[NUM_MARKS] = {
    {378442, 2377},   /* 1 */
    {377546, 26135},
    {375160, 49790},
    {371293, 73248},
    {365961, 96417},  /* 5 */
    {359185, 119206},
    {350991, 141524},
    {341412, 163284},
    {330486, 184399},
    {318255, 204787}, /* 10 */
    {304768, 224366},
    {290079, 243060},
    {274244, 260795},
    {257328, 277500},
    {239396, 293110},
    {220519, 307564},
    {200771, 320803},
    {180232, 332777},
    {158981, 343437},
    {137103, 352742}, /* 20 */
    {114683, 360655},
    {91811, 367144},
    {68577, 372184},
    {45072, 375756},
    {21389, 377845},
    {-2377, 378442},
    {-26135, 377546},
    {-49790, 375160},
    {-73248, 371293},
    {-96417, 365961}, /* 30 */
    {-119206, 359185},
    {-141524, 350991},
    {-163284, 341412},
    {-184399, 330486},
    {-204787, 318255},
    {-224366, 304768},
    {-243060, 290079},
    {-260795, 274244},
    {-277500, 257328},
    {-293110, 239396}, /* 40 */
    {-307564, 220519},
    {-320803, 200771},
    {-332777, 180232},
    {-343437, 158981},
    {-352742, 137103},
    {-360655, 114683},
    {-367144, 91811},
    {-372184, 68577},
    {-375756, 45072},
    {-377845, 21389}, /* 50 */
    {-378442, -2377},
    {-377546, -26135},
    {-375160, -49790},
    {-371293, -73248},
    {-365961, -96417},
    {-359185, -119206},
    {-350991, -141524},
    {-341412, -163284},
    {-330486, -184399},
    {-318255, -204787}, /* 60 */
    {-304768, -224366},
    {-290079, -243060},
    {-274244, -260795},
    {-257328, -277500},
    {-239396, -293110},
    {-220519, -307564},
    {-200771, -320803},
    {-180232, -332777},
    {-158981, -343437},
    {-137103, -352742}, /* 70 */
    {-114683, -360655},
    {-91811, -367144},
    {-68577, -372184},
    {-45072, -375756},
    {-21389, -377845},
    {2377, -378442},
    {26135, -377546},
    {49790, -375160},
    {73248, -371293},
    {96417, -365961} /* 80 */
};

/*
 * fibCol indicates if a particular part of fibres collides and in which way.
 */
static fibColType   fibCol[NUM_FIBRES][NUM_FIBRES];
/*
 * markCol indicates if a particular fibre collides with a particular mark.
 */
static vltLOGICAL   markCol[NUM_FIBRES][NUM_MARKS];


/*
 * Macro which given a fibre number returns an index
 */
#define FIBI(_i_) (_i_-1)
#define MARKI(_i_) (_i_-1)

/*
 * Marcos which set the various collision flags for two fibres.
 */
#define FIBFIBCOL(_f1_, _f2_)  fibCol[FIBI(_f1_)][FIBI(_f2_)].fibFib = TRUE; \
                               fibCol[FIBI(_f2_)][FIBI(_f1_)].fibFib = TRUE

#define BUTBUTCOL(_f1_, _f2_)  fibCol[FIBI(_f1_)][FIBI(_f2_)].butBut = TRUE; \
                               fibCol[FIBI(_f2_)][FIBI(_f1_)].butBut = TRUE

#define BUTFIBCOL(_f1_, _f2_)  fibCol[FIBI(_f1_)][FIBI(_f2_)].butFib = TRUE; \
                               fibCol[FIBI(_f2_)][FIBI(_f1_)].butFib = TRUE


#define FIBFIBCOL1(_f1_, _f2_) fibCol[FIBI(_f1_)][FIBI(_f2_)].fibFib = TRUE

#define BUTBUTCOL1(_f1_, _f2_) fibCol[FIBI(_f1_)][FIBI(_f2_)].butBut = TRUE

#define BUTFIBCOL1(_f1_, _f2_) fibCol[FIBI(_f1_)][FIBI(_f2_)].butFib = TRUE

#define MARKCOL(_f_, _m_)  markCol[FIBI(_f_)][MARKI(_m_)] = TRUE


static void init();
static int  check();


/*
 * Main function.
 */
extern int main()
{
    /*
     * Initaiilse.
     */
    init();

    /*
     * Do that check and report the result.
     */
    if (check() == FAILURE)
        {
        fprintf(stderr, "Collision detection test failed\n");
        return 1;
        }
    else
        {
        printf("Collision detection test passed\n");
        return 0;
        }


}

/*
 * init() function.  Initalises the arrays.  The fibres array is
 * initialised from the file "fibre_table.txt".
 *
 * Creating the file "fibre_table.txt".
 *  This contains a large number of fibre allocation details.  It is
 *  generated from a combination of the output of "configure" and
 *  information from the file "fpInsConfig.sds".  Having generated
 *  a field using configure, assuming the file is ozpoz.sds, the
 *  sequence is as follows
 *
 *  sdsdump ozpoz.sds objects.x              > alloc_x
 *  sdsdump ozpoz.sds objects.y              > alloc_y
 *  sdsdump ozpoz.sds objects.theta          > alloc_theta
 *  sdsdump ozpoz.sds objects.type           > alloc_type
 *  sdsdump fpInsConfig.sds Plate1_Data.xPiv > piv_x
 *  sdsdump fpInsConfig.sds Plate1_Data.yPiv > piv_y
 *
 *  Now edit alloc_type to put each character on a separate line.
 *  Edit alloc_x, alloc_y and alloc_theta to add a line at the begining
 *    of the file, indicating X/Y/Theta.
 *
 *  Edit piv_x and piv_y and replace the first line by one indicating
 *    piv_x/piv_y.
 *
 *  Ensure the following command
 *
 *  paste type alloc_x alloc_y piv_x piv_y alloc_theta > table
 * 
 *  Manually add the following two lines to the begining of the
 *  file

Used by fpcolTest.c.  First three lines ignored.
"@(#) $Id: ACMM:ozpozFpcol/ws/test/fpcolTest.c,v 1.63 07-Oct-2008 11:10:30+10 tjf $"


 *  The file is now ready.  You may need to adjust NUM_FIBRES in this
 *  program to reflect the file.
 * 
 *  Note that "sdsdump" is part of the Standard AAO sds release.
 *
 *  When fibre numbers are output, add four to the number to get
 *  the line in the file (since they are one based and the file
 *  has three  extra lines at the begining).  The fibre number does not
 *  corresponded to the number in configure since lines have
 *  been removed.
 *
 *  Having used this approch, you can use "configure" to check fibre
 *  locations on the mimic.  
 *
 *  NOTE: NUM_FIBRES (and the number of entries on "table") can be more
 *  then the number of fibres in the positioner and often will be so that
 *  we can add fibres which collide with others.
 *  
 */

static void init()
{
    FILE *table = NULL;
    char buffer[100];
    int fibre = 0;

    /*
     * Open the file.
     */
    if ((table = fopen("fibre_table.txt", "r")) == NULL)
        {
        fprintf(stderr,"Failed to open file \"fibre_table.txt\"\n");
        exit(1);
        }
    /*
     * Read and forget the first three lines
     */
    fgets(buffer, sizeof(buffer), table);
    fgets(buffer, sizeof(buffer), table);
    fgets(buffer, sizeof(buffer), table);

    /*
     * Read the file.
     */
    while (fgets(buffer, sizeof(buffer), table))
        {
        if (fibre >= NUM_FIBRES)
            {
            break;
            }
        

        if (sscanf(buffer,"%c %lg %lg %lg %lg %lg",
                   &fibres[fibre].type,
                   &fibres[fibre].butX,
                   &fibres[fibre].butY,
                   &fibres[fibre].pivX,
                   &fibres[fibre].pivY,
                   &fibres[fibre].butT) != 6)
            {
            fprintf(stderr,"Line %d in file \"table\" invalid\n", fibre+2);
            exit(1);
            }

        fibres[fibre].invalidPos = FALSE;

        ++fibre;
        }
    if (fibre != NUM_FIBRES)
        {
        fprintf(stderr,"Insufficent entries (%d) in file \"table\"\n", fibre+2);
        exit(1);
        }
    /*
     * fibCol and markCol arrays.  Set all to zero and then fill in
     * by hand on the ones which are collisions.  We cheat a bit
     * by using memset to zero the arrays, which may not be safe in
     * all cases in theory but is safe in practice on the target
     * machines.
     */
    memset(fibCol,  0, sizeof(fibCol));
    memset(markCol, 0, sizeof(markCol));

    /*
     *  Indicate the collisons we expect.  This is just added by
     *  hand.  The Macro FIBFIBCOL fills out fibCol appropriately.
     */
    FIBFIBCOL(50,   55);
    FIBFIBCOL(105, 175);
    FIBFIBCOL(125, 135);
    FIBFIBCOL(125, 175);
    FIBFIBCOL(130, 135);
    FIBFIBCOL(130, 140);
    FIBFIBCOL(130, 145);
    FIBFIBCOL(130, 175);
    FIBFIBCOL(135, 175);
    FIBFIBCOL(145, 175);
    FIBFIBCOL(155, 175);
    FIBFIBCOL(160, 170);
    FIBFIBCOL(160, 175);
    FIBFIBCOL(165, 170);
    FIBFIBCOL(165, 175);
    FIBFIBCOL(170, 175);
    FIBFIBCOL(205, 210);
    FIBFIBCOL(215, 225);
    FIBFIBCOL(215, 220);
    FIBFIBCOL(235, 260);
    FIBFIBCOL(240, 260);
    FIBFIBCOL(245, 260);
    FIBFIBCOL(245, 249);
    FIBFIBCOL(248, 249);
    FIBFIBCOL(248, 260);
    FIBFIBCOL(248, 260);
    FIBFIBCOL(249, 260);
    FIBFIBCOL(250, 260);
    FIBFIBCOL(255, 260);
    FIBFIBCOL(265, 275);
    FIBFIBCOL(310, 311);
    FIBFIBCOL(350, 355);
    FIBFIBCOL(350, 374);
    FIBFIBCOL(360, 365);
    FIBFIBCOL(372, 374);
    FIBFIBCOL(372, 375);
    FIBFIBCOL(374, 375);
    FIBFIBCOL(390, 395);
    FIBFIBCOL(390, 430);
    FIBFIBCOL(404, 430);
    FIBFIBCOL(415, 430);
    FIBFIBCOL(420, 430);
    FIBFIBCOL(425, 430);
    FIBFIBCOL(470, 480);
    FIBFIBCOL(470, 485);
    FIBFIBCOL(480, 485);
    FIBFIBCOL(496, 500);
    FIBFIBCOL(499, 500);

    /*
     * Now the specially added entries meant to trigger collision detection
     *
     * For fibre-fibre and button-button collisoins, normally we want
     * the two-way checks done by FIBFIBCOL and BUTBUT collisions.  Sometimes
     * due to rounding errors, we only want a one way check.  For 
     * Button-fibre collisions, one way is the default.
     *
     *
     * Fibre 501 is a clone of fibre 101, but at the location
     * of 175, leading to collisions with both 101 and 175 since they
     * are very close.
     */
    FIBFIBCOL1(501, 101);  /* Due to rounding, one way check only */

    BUTBUTCOL(101, 501);
    BUTBUTCOL(175, 501);
       
    BUTFIBCOL1(101, 501);

    BUTFIBCOL1(175, 501);
    BUTFIBCOL1(501, 175);

    /*
     * Fibre 502 is a clone of 295 but at the position of 290.  It
     * should only collide with 290.
     */
    FIBFIBCOL1(295, 502);  /* Due to rounding, one way check only */
    
    BUTBUTCOL(290, 502);
    BUTFIBCOL(290, 502);

    /*
     * Fibre 503 is a clone of 190 but at the position of 195.  
     *
     * Note the unusual button-fibre between 190 and 503, due to
     * 503 running close to the button of 190.
     */
    FIBFIBCOL(190, 503);
    BUTBUTCOL(195, 503);
    BUTFIBCOL(195, 503);
    BUTFIBCOL1(190, 503);

    /*
     * Fibre 504 is a clone of fibre 45 but placed in the center of the plate
     * It should have no collisions but the position is invalid.
     */
    fibres[FIBI(504)].invalidPos = TRUE;


    /*
     * Initialisation of fiducial mark position collisions with fibres.
     */
    MARKCOL(20,  5);
    MARKCOL(185, 37);
    MARKCOL(195, 39);
    MARKCOL(230, 47);
    MARKCOL(285, 59);
    MARKCOL(303, 62);
    MARKCOL(315, 65);
    MARKCOL(320, 66);
    MARKCOL(355, 71);
    MARKCOL(365, 74);
    MARKCOL(370, 75);
    MARKCOL(375, 76);
    MARKCOL(380, 78);
    MARKCOL(385, 79);
    MARKCOL(395, 80);
    MARKCOL(503, 39);

}

static void cvtFibres(void);
static void cvtMarks(void);
static int checkFibre(unsigned fibre);
static int checkFibre2Fibres(unsigned fibre);
static int checkFibre2Marks(unsigned fibre);

static int check1Mark(unsigned fibre, unsigned mark);
static int check1Fibre(unsigned fibre, unsigned otherFibre);

/*
 *  Function to do the checks.
 */
static int  check()
{
    register unsigned fibre = 0;
    int result = SUCCESS;

    /*
     * Convert the X/Y/Theta coorindates to R/Theta/Alpha.
     */
    cvtFibres();
    cvtMarks();

    /*
     * Do each of the checks for each fibre.
     */
    for (fibre = 0 ; fibre < NUM_FIBRES ; ++fibre)
        {

        /*
         * Check for invalid position.
         */
        if (checkFibre(fibre) != SUCCESS)
            result = FAILURE;

        /*
         * Check this fibre against all other fibres
         */
        if (checkFibre2Fibres(fibre) != SUCCESS)
            result = FAILURE;

        /*
         * Check this fibre against the fiduical marks.
         */
        if (checkFibre2Marks(fibre) != SUCCESS)
            result = FAILURE;
        }
    return result;

}

/*
 *  Convert the fibre  X/Y/Theta positions to
 *  OzPoz R/Theta/Alpha positions.
 */
static void cvtFibres(void)
{
    register unsigned fibre = 0;
    for (fibre = 0 ; fibre < NUM_FIBRES ; ++fibre)
        {
        fpPOS butXY;
        fpPOS pivXY;
        fpPOS butRT;
        fpPOS pivRT;

        /*
         *  Set up the fpPOS variables for the conversion
         */
        butXY.x = fibres[fibre].butX;
        butXY.y = fibres[fibre].butY;
        pivXY.x = fibres[fibre].pivX;
        pivXY.y = fibres[fibre].pivY;

        /*
         * Convert X/Y positions to R/Theta postions
         */
        fpptXY2RT(butXY, &butRT);
        fpptXY2RT(pivXY, &pivRT);

        /*
         * Save the values.
         */
        fibres[fibre].butR     = butRT.x;
        fibres[fibre].butTheta = butRT.y;

        fibres[fibre].pivR     = pivRT.x;
        fibres[fibre].pivTheta = pivRT.y;

        /*
         * Now we need alpha  (IS THIS RIGHT?)
         */
        /*
        fibres[fibre].butAlpha = atan2((butRT.y - pivRT.y),
                                    (butRT.x - pivRT.x));
        */
        fibres[fibre].butAlpha = fibres[fibre].butT - M_PI_2;
        if (fibres[fibre].butAlpha < (-1*M_2_PI))
            fibres[fibre].butAlpha += M_2_PI;

        }
}

/*
 *  Convert the mark X/Y positons to OzPoz R/Theta positions
 */
static void cvtMarks(void)
{
    register unsigned mark = 0;
    for (mark = 0 ; mark < NUM_MARKS ; ++mark)
        {
        fpPOS XY;
        fpPOS RT;

        /*
         *  Set up the fpPOS variables for the conversion
         */
        XY.x = marks[mark].x;
        XY.y = marks[mark].y;

        /*
         * Convert X/Y positions to R/Theta postions
         */
        fpptXY2RT(XY, &RT);
        /*
         * Save the values.
         */
        marks[mark].r = RT.x;
        marks[mark].t = RT.y;

        }
}

/*
 *  For the specified fibre, perform the invalid position check.
 */
static int checkFibre(unsigned fibre)
{
    int result = SUCCESS;
    int rtCol  = FALSE;
    int xytCol = FALSE;
    int expCol = FALSE;

    /*
     * Do the check in both R/T and X/Y coorindate systems.
     */
    rtCol = fpcolInvPos(0,
                        fibres[fibre].butR,
                        fibres[fibre].butTheta,
                        fibres[fibre].butAlpha,
                        fpcolINST_BUTTON_CLEAR);

    xytCol = fpcolBaseColInvPos(0, 0,
                             fibres[fibre].butX,
                             fibres[fibre].butY,
                             fibres[fibre].butT,
                             fpcolINST_BUTTON_CLEAR,
                             (fpcolINST_BUTTON_CLEAR*fpcolINST_BUTTON_CLEAR));

    expCol = fibres[fibre].invalidPos;
    /*
     * We ensure the R/T check and X/Y/T check return the same result and
     * check the result against what we expected.
     */
    if (rtCol != xytCol)
        {
        fprintf(stderr,
                "ERROR: Fibre %d checking invalid position\n", fibre+1);
        fprintf(stderr,
     "       Difference between based R/T (%s) and X/Y/T (%s) based checks\n",
            (rtCol  ? "true" : "false"),
            (xytCol ? "true" : "false"));
        result = FAILURE;
        }
    if (rtCol != expCol)
        {
        fprintf(stderr,
                "ERROR: Fibre %d checking invalid position\n", fibre+1);
        fprintf(stderr,"       Result was not expected, expected %s, got %s\n",
            (expCol ? "true" : "false"),
            (rtCol  ? "true" : "false"));        
        result = FAILURE;
        }

    return result;
}

/*
 *  For the specified fibre, check it against all other fibres
 */
static int checkFibre2Fibres(unsigned fibre)
{
    register unsigned otherFibre = 0;
    int result = SUCCESS;

    for (otherFibre = 0 ; otherFibre < NUM_FIBRES ; ++otherFibre)
        {
        if (fibre == otherFibre) continue;
        if (check1Fibre(fibre, otherFibre) == FAILURE)
            {
            result = FAILURE;
            }
        }
    return result;
}

/*
 *  For the specified fibre, check it against the fiducial marks.
 */
static int checkFibre2Marks(unsigned fibre)
{
    register unsigned mark = 0;
    int result = SUCCESS;

    for (mark = 0 ; mark < NUM_MARKS ; mark++)
        {
        if (check1Mark(fibre, mark) == FAILURE)
            {
            result = FAILURE;
            }
        }
    return result;
}

/*
 *  Report errors from fibre-fibre checks.
 */
static void errorRTdifXYT(int fibre, int other, char *test, int rt, int xyt)
{
    fprintf(stderr,"ERROR: Fibre %d checking against %d for %s collisions\n",
            fibre+1, other+1, test);
    fprintf(stderr,
     "       Difference between based R/T (%s) and X/Y/T (%s) based checks\n",
            (rt  ? "true" : "false"),
            (xyt ? "true" : "false"));       
            
}
static void errorRTdifExp(int fibre, int other, char *test, 
                          int expected, int got)
{
    fprintf(stderr,"ERROR: Fibre %d checking against %d for %s collisions\n",
            fibre+1, other+1, test);
    fprintf(stderr,"       Result was not expected, expected %s, got %s\n",
            (expected ? "true" : "false"),
            (got      ? "true" : "false"));
            
}

/*
 * Check one fibre against another.
 */
static int check1Fibre(unsigned fibre, unsigned other)
{
    int result = SUCCESS;
    int rtCol  = FALSE;
    int xytCol = FALSE;
    int expCol = FALSE;

    /*
     * Don't check unallocated fibres.
     */
    if ((fibres[fibre].type == 'U')||
        (fibres[other].type == 'U')) return result;

    /*
     * First check for a Button/Button Collision between these
     * two fibres
     */    
    rtCol = fpcolButBut(fibres[fibre].butR, fibres[fibre].butTheta,
                        fibres[fibre].butAlpha,
                        fibres[other].butR, fibres[other].butTheta,
                        fibres[other].butAlpha, fpcolINST_BUTTON_CLEAR);

    xytCol = fpcolBaseColButBut(0, fibres[fibre].butX, fibres[fibre].butY,
                             fibres[fibre].butT,
                             fibres[other].butX, fibres[other].butY,
                             fibres[other].butT,
                             fpcolINST_BUTTON_CLEAR,
                             (fpcolINST_BUTTON_CLEAR*fpcolINST_BUTTON_CLEAR));

    expCol = fibCol[fibre][other].butBut;

    /*
     * We ensure the R/T check and X/Y/T check return the same result and
     * check the result against what we expected.
     */
    if (rtCol != xytCol)
        {
        errorRTdifXYT(fibre,other,"Button Button", rtCol, xytCol);
        result = FAILURE;
        }
    if (rtCol != expCol)
        {
        errorRTdifExp(fibre,other,"Button Button", expCol, rtCol);
        result = FAILURE;
        }


    /*
     * Now check this fibre button against the other fibre
     */
    rtCol = fpcolButFib(fibres[fibre].butR, fibres[fibre].butTheta,
                        fibres[fibre].butAlpha,
                        fibres[other].butR, fibres[other].butTheta,
                        fibres[other].pivR, fibres[other].pivTheta,
                        fpcolINST_FIBRE_CLEAR);

    xytCol = fpcolBaseColButFib(0, fibres[fibre].butX, fibres[fibre].butY,
                             fibres[fibre].butT,
                             fibres[other].butX, fibres[other].butY,
                             fibres[other].pivX, fibres[other].pivY,
                             fpcolINST_FIBRE_CLEAR,
                             (fpcolINST_FIBRE_CLEAR*fpcolINST_FIBRE_CLEAR));

    expCol = fibCol[fibre][other].butFib;

    /*
     * We ensure the R/T check and X/Y/T check return the same result and
     * check the result against what we expected.
     */
    if (rtCol != xytCol)
        {
        errorRTdifXYT(fibre,other,"Button Fibre", rtCol, xytCol);
        result = FAILURE;
        }
    if (rtCol != expCol)
        {
        errorRTdifExp(fibre,other,"Button Fibre", expCol, rtCol);
        result = FAILURE;
        }



    /*
     * The final check is for a fibre crossover.  Here these is no
     * X/Y/T equivalent to check
     */
    rtCol = fpcolFibFib(fibres[fibre].butR, fibres[fibre].butTheta,
                        fibres[fibre].pivR, fibres[fibre].pivTheta,
                        fibres[other].butR, fibres[other].butTheta,
                        fibres[other].pivR, fibres[other].pivTheta,
                        fpcolINST_FIBRE_CLEAR);

    expCol = fibCol[fibre][other].fibFib;

    /*
     * Check against what we expected
     */
    if (rtCol != expCol)
        {
        errorRTdifExp(fibre,other,"Fibre Fibre", expCol, rtCol);
        result = FAILURE;
        }

    return result;
}
/*
 *  Check one fibre against one mark.
 */ 
static int check1Mark(unsigned fibre, unsigned mark)
{

    int result = SUCCESS;
    int rtCol  = FALSE;
    int xytCol = FALSE;
    int expCol = FALSE;

    if (fibres[fibre].type == 'U')
        return result;


    rtCol = fpcolFiducial(fibres[fibre].butR, fibres[fibre].butTheta,
                          fibres[fibre].pivR, fibres[fibre].pivTheta,
                          marks[mark].r, marks[mark].t,
                          fpcolINST_FID_CLEAR);

    xytCol =fpcolBaseColFidBut(0, fibres[fibre].butX, fibres[fibre].butY,
                               fibres[fibre].butT,
                               marks[mark].x, marks[mark].y, 
                               fpcolINST_FID_CLEAR,
                               (fpcolINST_FID_CLEAR*fpcolINST_FID_CLEAR));


    expCol = markCol[fibre][mark];

    /*
     * The X/Y/T check only checks for button/fiduical collision, not
     * fibre/collision check as is allso done by the R/T equivalent.  So
     * if xytCol is true, then rtCol should be true but no other
     * version is required.
     */
    if ((xytCol)&&(!rtCol))
        {
        fprintf(stderr,"ERROR: Checking fibre %d against fiducial %d\n",
            fibre+1, mark+1);
        fprintf(stderr,
     "       Collision using X/Y/T but not using R/T\n");
        result = FAILURE;
        }
    /*
     * Ensure the collision is what we expected.
     */
    if (rtCol != expCol)
        {
        fprintf(stderr,"ERROR: Checking Fibre %d against fiducial %d\n",
            fibre+1, mark+1);
        fprintf(stderr,"       Result was not expected, expected %s, got %s\n",
            (expCol ? "true" : "false"),
            (rtCol  ? "true" : "false"));

        result = FAILURE;
        }
    return result;

}
