/*******************************************************************************
* E.S.O. - VLT project
*
* "@(#) $Id: ACMM:ozpozFpcol/ws/test/fpcolTestButFib.c,v 1.63 07-Oct-2008 11:10:30+10 tjf $"
*
*
* who       when      what
* --------  --------  ----------------------------------------------
* tjf       20/03/02  created
*/

/************************************************************************
*   NAME
*       fpcolTestButFib - Test program button/fibre collisions
*
*   SYNOPSIS
*       fpcolTestButFib "xy" butX butY butAlpha fibX fibY pivX pivX
*       fpcolTestButFib "rt" butR butT butAlpha fibR fibT pivR pivT
*
*
*   DESCRIPTION
*
*  Program to test out the FPCOL collision detection routines for
*  a particular button and fibre.
*
*  If the first argument is "xy" then positions are given in X/Y.  If
*  the first argument is "rt" then positions are given in R/Theta.
*
*  We check for a collision between a button at the specific position and
*  alpha angle with a fibre with the given fibre and pivot end positions.
*
*  butX = button X position in microns
*  butY = button Y position in microns
*  fibX = fibre  X position in microns
*  fibY = fibre  Y position in microns
*  pivX = pivot  X position in microns
*  pivY = pivot  Y position in microns
*  butR = button R position in microns
*  butT = button Theta in radians.
*  fibR = fibre  R position in microns
*  fibT = fibre  Theta in radians.
*  pivR = pivot  R position in microns
*  pivT = pivot  Theta in radians.
*  buttonAlpha = alpha angle of button.
*
*  The check is done in both R/Theta and X/Y. 
****
*
*   CAUTIONS
*
*   EXAMPLES
*
*   SEE ALSO
*
*   BUGS
*
*------------------------------------------------------------------------
*/

#include "vltPort.h"    /* Must be first header */

#include <stdio.h>
#include <math.h>
#include <string.h>
#include "fppt.h"
#include "fpcol.h"
#include "fpcolBase.h"
#include "fpcolInstDescr.h"

static void usage(const char *cmd)
{
      fprintf(stderr,"Usage:%s \"xy\" butX butY butAlpha fibX fibY pivX pivX\n", cmd);
      fprintf(stderr,"     :%s \"rt\" butR butT butAlpha fibR fibT pivR pivT\n", cmd);

}

/*
 * Convert from X/Y to R/Theta
 */
static void cvtXY2RT(double x, double y, double *r, double *t)
{
  /*
   * If origin, return defined theta
   */
  if ((x == 0) && (y == 0))
    {
      *r = 0;
      *t = 0;
    }
  else
    {
      *r = sqrt((x * y) + (y * y));
      *t = atan2(y, x);
      if (*t < 0.0)
	*t += 2.0 * M_PI;
    }
}

/*
 * Convert from R/Theta to X/Y
 */
static void cvtRT2XY(double r, double t, double *x, double *y)
{
  *x = r * cos(t);
  *y = r * sin(t);
}

int main(int argc, char *argv[])
{
  double butR;
  double butTheta;
  double butAlpha;
  double fibR;
  double fibTheta;
  double pivR;
  double pivTheta;
  double butX;
  double butY;
  double fibX;
  double fibY;
  double pivX;
  double pivY;

  vltLOGICAL rtCol;
  vltLOGICAL xyCol;
 
  /*
   * Validate arguments
   */
  if (argc != 9)
    {
      fprintf(stderr,"Wrong number of arguments (%d)\n", argc);
      usage(argv[0]);
      return 1;
    }
  /*
   * Input is in x/y
   */
  if (strcmp(argv[1],"xy") == 0)
    {
      butX     = atof(argv[2]);
      butY     = atof(argv[3]);
      butAlpha = atof(argv[4]);
      fibX     = atof(argv[5]);
      fibY     = atof(argv[6]);
      pivX     = atof(argv[7]);
      pivY     = atof(argv[8]);

      cvtXY2RT(butX, butY, &butR, &butTheta);
      cvtXY2RT(fibX, fibY, &fibR, &fibTheta);
      cvtXY2RT(pivX, pivY, &pivR, &pivTheta);
    }
  /*
   * Input is in R/Theta
   */
  else if (strcmp(argv[1],"rt") == 0)
    {
      butR     = atof(argv[2]);
      butTheta = atof(argv[3]);
      butAlpha = atof(argv[4]);
      fibR     = atof(argv[5]);
      fibTheta = atof(argv[6]);
      pivR     = atof(argv[7]);
      pivTheta = atof(argv[8]);

      cvtRT2XY(butR, butTheta, &butX, &butY);
      cvtRT2XY(fibR, fibTheta, &fibX, &fibY);
      cvtRT2XY(pivR, pivTheta, &pivX, &pivY);
    }
  else
    {
      fprintf(stderr,"Invalid flag argument \"%s\"\n", argv[1]);
      usage(argv[0]);
      return 1;
    }

  /*
   *Output details
   */
  printf("Button/Fibre collision detection check\n");
  printf("  Positions in R/Theta\n");
  printf("    Button: %.1f %.6f\n", butR, butTheta);
  printf("    Fibre : %.1f %.6f\n", fibR, fibTheta);
  printf("    Pivot : %.1f %.6f\n", pivR, pivTheta);
  printf("  Positions in X/Y\n");
  printf("    Button: %.1f %.1f\n", butX, butY);
  printf("    Fibre : %.1f %.1f\n", fibX, fibY);
  printf("    Pivot : %.1f %.1f\n", pivX, pivY);

  /*
   * Now check this button against the fibre in both R/Theta and X/Y systems.
   */
    rtCol = fpcolButFib(butR, butTheta,
                        butAlpha,
                        fibR, fibTheta,
                        pivR, pivTheta,
                        fpcolINST_FIBRE_CLEAR);

    xyCol = fpcolBaseColButFib(0, butX, butY,
			       butTheta,
			       fibX, fibY,
			       pivX, pivY,
			       fpcolINST_FIBRE_CLEAR,
			       (fpcolINST_FIBRE_CLEAR*fpcolINST_FIBRE_CLEAR));


    /*
     * We ensure the R/T check and X/Y/T check return the same result and
     * check the result against what we expected.
     */
    if (rtCol != xyCol)
        {
	  fprintf(stderr,"X/Y and R/T check returned different values\n");
	  fprintf(stderr,"  Maybe this is rounding, but it could indicate a bug\n");
	  fprintf(stderr,"  X/Y check = \"%s\", R/T check = \"%s\"\n",
		  (xyCol ? "collision" : "no collision"),
		  (rtCol ? "collision" : "no collision"));
	  return 1;
        }

    else if (rtCol)
      printf("COLLISON\n");
    else
      printf("No Collision\n");

    return 0;
}
