/*
 *                    C O N F I G U R E   ( F P O S S )
 * 
 *  Program name:
 *     Configure
 * 
 *  Function:
 *     Main C code module for the AAO Fibre Positioner Configuration program.
 *  
 *  Description:
 *     The 'configure' program is part of the observation preparation
 *     software used by a number of instruments built by the AAO. These
 *     include the original '2dF', the more recent '6dF' and the 'OzPoz'
 *     fibre positioner built for the 'FLAMES' instrument for ESO. The
 *     ESO version of this program is known as FPOSS - Fibre Positioner
 *     Observation Support Software. For more details about the overall
 *     structure of the code and what the program has to achieve, see the
 *     comments at the start of the file 'configure.tcl' or the FPOSS
 *     internals manual written for ESO.
 *     
 *     The comments at the start of configure.tcl emphasise that the
 *     structure of the configuration program is that of a Tcl/Tk program
 *     with a large number of Tcl commands being implemented in C. This
 *     file - configure.c - is the location of the interface code for all
 *     these Tcl commands. In many cases, the work for these Tcl commands is
 *     done through calls to other layers of the system, but in all cases
 *     a C-implemented Tcl command ends up invoking one of the routines in
 *     this file.
 *     
 *     This file also includes the main() entry point for the program, and
 *     in that sense this is the main code module for the program. However,
 *     the main() code does little other than initialise Tcl/Tk, causing
 *     them to be initialised in such a way that the routine ConfAddCommands()
 *     is invoked. This adds all the C-implemented Tcl commands to the
 *     Tcl interpreter. Control then passes on to the main-line code in
 *     configure.tcl, which puts up the main GUI for the program. After that,
 *     the program operates like a conventional (if quite complex) Tcl/Tk
 *     program.
 *
 *  Synopsis:
 *     configure [options]
 *     
 *     For a full description of all the options, see the Configure/FPOSS
 *     user guide.  In general use, very few of these options are ever used,
 *     particularly when used in the ESO context as FPOSS.
 *
 *  Language:
 *     C
 *
 *  Authors:
 *     Jeremy Bailey, AAO. (JAB)
 *     Gavin Dalton, University of Oxford (GBD)
 *     Tony Farrell, AAO (TJF)
 *     Keith Shortridge, AAO (KS)
 *     Chris Ramage, AAO (CR)  
 *     Brent Miszalski, Macquarie University/AAO (BM)
 *
 *  Program overview: 
 *     A relatively full description of the operation of the configuration
 *     program is provided in the FPOSS Internals manual written as part of
 *     the OzPoz project. This should be consulted to provide an overall
 *     picture of the operation of this code. In particular, it contains
 *     details of the SDS data structures manipulated by the program, and
 *     a walk-through what the code does as it performs a basic configuration.
 *     
 *  Pre-processor variables:
 *     FLAMES_ONLY  is defined if the program is to be built with support only
 *                  for FLAMES. Normally, it will be built with support for
 *                  all of 2dF, 6dF and FLAMES.
 *    
 *     Some additional pre-processor variables are used to control details
 *     of the compilation under different operating systems, but these will
 *     all be set by the program development environments for these machines.
 *     (For example, macintosh is defined when compiling on a Macintosh, 
 *     WIN32 is defines when compiling for Windows.)
 *
 * "@(#) $Id: ACMM:aaoConfigure/src/configure.c,v 12.7 04-Nov-2013 11:28:47+11 tjf $"
 *     
 *  History:
 *     The original version of this program (only supporting 2dF) dates from
 *     1993. The history section of the code that dates from 1993 to 2001
 *     is long and - given the substantial rewriting to support multiple
 *     instruments such as 6dF and FLAMES - somewhat irrelevant. It has been
 *     removed from this file and saved in the file configure.history, which
 *     can be consulted as a historical document. The following section
 *     contains only a brief summary of changes up to 2001, and the detailed
 *     history starts with the release of the program with support for all of 
 *     2dF, 6dF and FLAMES.
 *     
 *     18-Oct-1993 - JAB - Original version
 *     15-Nov-1996 - GBD - Released with changes to support the 'Oxford' 
 *                         allocation algorithm.
 *     04-Nov-1998 - TJF - Reworked as a Tcl program instead of a DRAMA
 *                         program.
 *     20-Apr-2000 - TJF - Added support for 6dF.
 *     03-Jul-2000 - KS  - Initial version with OzPoz support.
 *     26-Sep-2001 - KS  - First release with proper OzPoz support. Commenting
 *                         revised.
 *     30-Oct-2001 - KS  - Added TargetDescr{}. Instrument info now passed to
 *                         PAF input conversion routines.
 *     05-Nov-2001 - KS  - Implemented ExpertAllowed{}.
 *     09-Nov-2001 - KS  - Implemented UpdateFibreStats{} and GetFibreStats{}.
 *     20-Nov-2001 - KS  - Added SetSkyCounts{} and support for explicit control
 *                         over the number of sky targets allocated for each
 *                         fibre type.
 *     21-Nov-2001 - KS  - SetSkyCounts{} was returning an undefined Tcl code
 *                         when called with no arguments. Fixed.
 *     26-Nov-2001 - KS  - The confusing CONF_SPECIAL code discontinued. A 
 *                         listing of unallocated objects now includes sky
 *                         targets as well as fiducial targets.
 *     28-Nov-2001 - KS  - Minor changes to get a clean compilation under
 *                         Linux.
 *     10-Dec-2001 - KS  - Blanks in PAF file items that don't allow them
 *                         now changed to underscores. Leading spaces are
 *                         now allowed in the values for items such as B-V
 *                         included in comment strings, and such items with
 *                         null values are not written to PAF files.
 *     19-Dec-2001 - KS  - GetFibre{} now makes sure it doesn't return fibre
 *                         type descriptions with embedded blanks.
 *     03-Jan-2002 - KS  - DoCheck{} now accepts a 'positioner tolerances'
 *                         optional argument. PAF files now include the 
 *                         INS.OSS.VER keyword, generated using the new
 *                         routine RCSVersion().
 *     30-Jan-2002 - KS  - Added SkycatListing{}.
 *     05-Feb-2002 - KS  - Added support for PAF file checksums. This is
 *                         disabled at present. To enable, change the value
 *                         in the #define for USE_CHECKSUM. For FLAMES, the
 *                         default humidity is now 20% instead of 50%.
 *     06-Feb-2002 - KS  - Skycat listing format improved - now defines
 *                         plotting symbols properly.
 *     12-Feb-2002 - KS  - Use of local tcl and tk directories modified to 
 *                         make it easier to change tcl/tk versions.
 *     25-Mar-2002 - KS  - ConfDoConversion() now uses FpilGetFibreWavelength().
 *                         This supports use of different wavelength 
 *                         combinations when checking configurations.
 *     11-Apr-2002 - KS  - Added PivotInUse{}. Supports -use_all_fibres command
 *                         line option.
 *     04-Jun-2002 - KS  - PAFFileName and WritePAF now incorporate the changes
 *                         suggested by Andreas Kaufer to incorporate the
 *                         time and mode into the file name and the PAF.ID and
 *                         PAF.NAME fields.
 *     05-Jun-2002 - KS  - Declaration of putenv() removed to compile properly
 *                         under Linux RedHat 7.2.
 *     30-Jul-2002 - KS  - Modified vertex set used to draw the VLT guide 
 *                         probe. See comments in the code. KS.
 *     08-Aug-2002 - KS  - Corrected field number passed to FpilModelAccess()
 *                         by SetPlate{}. Revised use of 'macintosh' pre-
 *                         processor variable to allow compilation on OS X.
 *     21-Aug-2002 - KS  - TEL.TARG.NAME now included in PAF files.
 *     22-Aug-2002 - TJF - Modify WritePAF to write the SKYANGLE correctly.
 *                         (Originally made at Paranal 20/8/02, incorporated
 *                         into this version by KS.)
 *     28-Aug-2002 - KS  - ReadPAF{} now returns a code giving the fibre
 *                         combination.
 *      5-Sep-2002 - KS  - Fixed bug where VLT guide stars were not being
 *                         identified properly in WritePAF{} because the
 *                         positions had changes. Stars now identified by name.
 *      2-Oct-2002 - KS  - Modified ArgusOffset from 90 to 0. This probably 
 *                         still isn't the right answer, but it makes things
 *                         less complicated while we try to find the right
 *                         answer.
 *     17-Oct-2002 - KS  - FpilModelCvtInit() now takes two wavelength 
 *                         parameters. ArgusOffset set to -90, which is now
 *                         believed to be the correct value - although this
 *                         may change.
 *     19-Oct-2002 - KS  - Incorporated two changes coming from Paranal during
 *                         commissioning. Code to decide which VLT guide probe
 *                         position is 'POS' and which 'NEG' has been modified,
 *                         and the atmospheric pressure used for FLAMES has
 *                         been reduced. The Argus offset is no longer
 *                         subtracted from the angle written as INS.ARGS.ANGLE.
 *                         Unused variables removed.
 *     22-Oct-2002 - KS  - VLT guide probe code comments tidied up.
 *     05-Nov-2002 - KS  - Now uses proper motion information in pseudo-comments
 *                         in .fld files to correct apparent RA/DEC.  PAF file
 *                         no longer has TEL.ROT.OFFANGLE.
 *     20-Jan-2003 - KS  - Fixed bug where successive calls to CompressUnalloc{}
 *                         on the same structure caused an error. (Usually 
 *                         showed up when 'List' was being used.)
 *     12-Mar-2003 - CR  - Field plates for 6dF are 1 and 2, not 0 and 1. 
 *                         Changes made to translate the internally use 0 and 1
 *                         into 1 and 2 whereever the information is visible to
 *                         the user.
 *     16-Apr-2003 - CR  - New function FixHistoricalSDS added to provide a 
 *                         convenient place to fix legacy SDS files so that
 *                         They can still be used by configure. Currently the
 *                         only fix included is to enable 6dF configure to read
 *                         SDS files that use an incorrect character to indicate
 *                         parked fibres.
 *                         Bug fixed in DoMakeAllocated. status variable was not
 *                         dereferenced correctly, leading to erroneous "objects"
 *                         structures being added to the SDS file each time one
 *                         was opened.
 *                         Three error messages that included a report of fibre
 *                         number were missing a "+ 1" hence were reporting an
 *                         incorrect fibre number.
 *     16-Sep-2003 - CR  - Changed scanf to correctly pick up acmm version number
 *                         from the header file.
 *     13-Oct-2003 - CR  - Change to support different wavelengths for the two 
 *                         2dF spectrigraphs
 *     02-Aug-2004 - TJF - Warnings should go to stderr rather then stdout.
 *                          Handling of -tk_version option should use write 
 *                          rather then printf to avoid MacOsX problems.
 *     04-Jan-2005 - TJF - 2dF has changed to have 400 rather then 404 fibres.
 *                          Since we must support old and new configurations,
 *                          there are now 2dF instruments.  TdfFpilFullInit() 
 *                          is replaced by TdfFpilFullInit404() and 
 *                          TdfFpilFullInit400().  
 *     06-Jan-2005 - TJF - Add CheckSds command.
 *                         Fix a few cases where error messages were lost due
 *                          to flushing after writing to interp->result.
 *                         Add AllocAddToUnalloc command (adaption of
 *                          of AllocToUnalloc.
 *                         Add ConfigSystemVer to allow fetching of system 
 *                           version information.
 *                         Support maintance of the number of not in use fibres
 *                           of each time in FibreStats, updated by UpdateFibreStats
 *                           and returned by GetFiberStats.
 *     14-Feb-2005 - TJF - Support references to instrument file location
 *                          and documentation site.  
 *     01-Mar-2005 -  KS - Split code up so that most directly C-callable
 *                         routines are in configure_subs.c, leaving this
 *                         file mainly for routines intended to be called from
 *                         Tcl. Global variables no longer static to support
 *                         reference in configure_subs.c. 
 *     08-Mar-2005 -  KS - Added AAOmega as an option.
 *     03-Aug-2005 - TJF - CheckSds now returns a different message for
 *                          wrong plate, allowing the plate to be changed by
 *                          the GUI.
 *     25-Feb-2006 - BM  - Fixed a bug in ConfParameterDetails where if 
 *                         nvalus == 0, the loop doesn't execute 
 *                         to pass on the remaining args (argv[7]).
 *     04-Mar-2006 - BM  - Added ConfSetCBSPairs to generates a 
 *                         cross-beam switching (CBS) pair 
 *                         for each program target in a similar fashion to
 *                         ConfSetSkyGrid.
 *                         N.B. There may be some memory issues.
 *                         It goes through all program target info,
 *                         calculates the offset in ra/dec for each target,
 *                         then adds a new target using ConfDoAddCBSPair
 *                         in configure_subs.c
 *     29-Mar-2006 - KS  - Re-ordered instruments so AAOmega comes first.
 *     30-Jun-2006 - KS  - Added safety checks on fibre type codes before 
 *                         using them as array indices. Compilation warnings
 *                         about signedness avoided by proper declarations. 
 *     23-Aug-2006 - TJF   Found some cases where wlenIndex was not being handled,
 *                          in particular - allocating sky grid and setting up for
 *                          cross beam switching.
 *                         Ensure that when reading old SDS files, we insert
 *                          the wavelength index stuff.
 *     05-Sep-2007 - KS  - Introduced the concept of locked fibres, using the
 *                         AllocLockedFlags array. Added FindObjectByName{},
 *                         LockAllocations{},ClearLocking{}, LockedFibres{},
 *                         FibreLocked{}.
 *     25-Oct-2007 - KS  - Fixed bug in FindObjectByName{}, which was matching 
 *                         names where the test name matched the start of an
 *                         object name, even thought the object name was
 *                         longer - eg 'Prog51' matched 'Prog511'.
 *     17-Dec-2007 - TJF - ConfInitialise modified to support 3rd 6dF field plate.
 *                          Also reformated somewhat.
 *     20-Dec-2007 - TJF - Added extra arguments to ConfSetSkyGrid() such that
 *                          we can generate only one of the three rings and
 *                          that we can generate program objects.  This makes
 *                          it easy to set up fields for debugging purposes.
 *                          Also evened up the spread of the points in the
 *                            inner two circles - there were not being done 
 *                            well.
 *     09-Feb-2008 -  KS - Added AlgorithmString{}.
 *     10-Aug-2009 -  KS - Added support for holding proper motion values in
 *                         explicit pmra,pmdec structure items rather than as
 *                         pseudo-comments. KS.
 *     17-Aug-2009 -  KS - GetObjectData{} now returns proper motion values.
 *                         ConfListPAFTarget() now takes proper motion Sds id
 *                         values.
 *     11-Apr-2012 - MVV - Added HERMES and its initialisation routine to the 
 *                         Instruments array.
 *     08-Aug-2012 - TJF - Tcl and Tk version specific libraries now found 
 *                         within tcl_dir/tk_dir, so that we can keep older
 *                         version sub-directories, due to Tcl now needing
 *                         us to do that.
 *     22-Aug-2012 -  KS - ConfigFieldCheck() now takes arguments allowing it
 *                         to report invalid pivots. DoCheck{} now uses this
 *                         to optionally return an array of invalid pivot
 *                         numbers. Minor changes to avoid some compiler 
 *                         warnings.
 *     31-Oct-2012 -  KS - ConfCheckSds() now no longer fails if there is
 *                         no 'objects' structure present. This was causing
 *                         it to reject files created using AllocToUnalloc.
 *     16-Sep-2013 -  KS - Now reads in affectation data (crosstalk and ghost
 *                         data) as part of initialisation for instruments
 *                         that support it.
 *     22-Oct-2013 -  KS - Implemented MergeCheckDuplicates{}.
 *     04-Nov-2013 - TJF - ConfAddIndex() was failing for the files actually
 *                          sent to the robot, due to those not having an
 *                          unallocObject component in the SDS file (deleted
 *                          by the 2dF control task to save space).  Modified
 *                          ConfAddIndex() to create an empty structure.  This 
 *                          all least allows us to view such files (allocations
 *                          may crash).
 */

static const char *rcsId="@(#) $Id: ACMM:aaoConfigure/src/configure.c,v 12.7 04-Nov-2013 11:28:47+11 tjf $";
static void *use_rcsId = (0 ? (void *)(&use_rcsId) : (void *) &rcsId);

/*  ------------------------------------------------------------------------- */

/*                       P r e l i m in a r i e s
 *
 *  The usual stuff in any file before we get to the code proper. Assorted
 *  definitions, include files, forward declarations of routines..  To
 *  find the start of the real code, search forward for "M a i n".
 */

/*  DEBUG is used to make diagnostic code stand out */

#ifdef DEBUG
#undef DEBUG
#endif
#define DEBUG printf

/*  By default, at present, PAF files do not have checksums included. To
 *  enable the writing of checksums change USE_CHECKSUM to a non-zero value.
 */
 
#define USE_CHECKSUM 0

/*  The following define the file names used for the initial Tcl script
 *  and the directory we try to read it from.
 */

#define CONF_TCLDIR "CONFIGURE_DIR"
#define CONF_TCLFILE "configure.tcl"
#define CONF_SCRIPT "CONFIGURE_SCRIPT"

/*  Standard include files  */

#if defined WIN32
#include <windows.h>
#endif

#define __EXTENSIONS__  /* For solaris include files */
#ifndef _GNU_SOURCE     /* for things like putenv on linux */
#define _GNU_SOURCE
#endif



#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <ctype.h>
#include <limits.h>
#ifndef macintosh
#include <unistd.h>
#include <sys/param.h>
#endif


/*  Not all systems seem to declare strcasecmp() properly, so we do so here.
 */

#ifndef __macosx__
int strcasecmp(const char *s1, const char *s2);
#endif

/*  At the moment, the Mac OS 9 version only has support for FLAMES */

#ifdef macos9
#ifndef FLAMES_ONLY
#define FLAMES_ONLY 1
#endif
#endif

/*  Program-specific include files  */

#include "drama.h" 
#include "status.h"
#include "sds.h"
#include "Ers.h"
#include "mess.h"
#include "sdstcl.h"
#include "arg.h"
#include "arg_err.h" 
#include "tcl.h"
#include "tdFconfig.h"
#include "tk.h"
#include "tdFconvert.h"
#include "convertPAF.h"
#include "slalib.h"
#include "slamac.h"
#include "sds_err_msgt.h"
#include "arg_err_msgt.h"
#include "conf_err_msgt.h"
#include "config_err_msgt.h"
#include "tdfxy_err_msgt.h"
#include "sixdf_util_err_msgt.h"
#include "fpil_err_msgt.h"
#include "conf_err.h"
#include "configure_subs.h"

/*
 *  Needed for Macintosh.
 */
 
#ifdef macos9
#  include <unistd.h>
   extern int MacintoshInit _ANSI_ARGS_((void));
   extern int SetupMainInterp _ANSI_ARGS_((Tcl_Interp * interp));
#endif

/*
 *  Include FPIL interface headers for each supported instrument.
 */

#ifndef FLAMES_ONLY
  #include "tdffpil.h"            /* 2dF - AAT */
  #include "sixdffpil.h"          /* 6dF - UKSCHMIDT */
#endif


#include "fpcolFpil.h"          /* OzPoz - VLT-2 */


/*
 *  Assuming each supported instrument has a function with the
 *  following typedef which does the FPIL Full initialisation, then
 *  we can set up an array of supported instruments here and
 *  otherwise don't have to worry about them - FPIL gives us all
 *  the details.
 */
typedef void (*ConfFpilInitRoutineType) (FpilType * inst);

/*  This should be made more general, but for the moment it's convenient
 *  to be able to build the system just for FLAMES. We should also be
 *  able to build one just for 2dF and one just for 6dF, I suppose - or
 *  maybe one for AAO use that supports both 2dF and 6dF but not FLAMES?
 */
 
#if defined(FLAMES_ONLY)
    static struct {
        FpilType Info;
        ConfFpilInitRoutineType InitRoutine;
        const char *inst_site;
        const char *doco_site;
    } Instruments[] = {
        {
        0, fpcolFpilInit, 
        "http://www.eso.org/observing/p2pp/OSS/FPOSS/FPOSS-tool.html",
        "http://www.eso.org/observing/p2pp/OSS/FPOSS/FPOSS-tool.html" },
    };
#else
    static struct {
        FpilType Info;
        ConfFpilInitRoutineType InitRoutine;
        const char *inst_file;
        const char *doco_site;
    } Instruments[] = {
        {
        0, TdfFpilFullInitHERMES,
        "ftp://ftp.aao.gov.au/pub/2df/latest_config_files",
        "http://www.aao.gov.au/aaomega"},
        {
        0, TdfFpilFullInitAAOmega, 
        "ftp://ftp.aao.gov.au/pub/2df/latest_config_files",
        "http://www.aao.gov.au/aaomega"},
        {
        0, TdfFpilFullInit400, 
        "ftp://ftp.aao.gov.au/pub/2df/latest_config_files",
        "http://www.aao.gov.au/2df"},
        {
        0, TdfFpilFullInit404, 
        "ftp://ftp.aao.gov.au/pub/2df/latest_config_files",
        "http://www.aao.gov.au/2df"},
        {
        0, fpcolFpilInit, 
        "http://www.eso.org/observing/p2pp/OSS/FPOSS/FPOSS-tool.html",
        "http://www.eso.org/observing/p2pp/OSS/FPOSS/FPOSS-tool.html" },
        {
        0, sixdfFpilFullInit, 
        0,
        "http://www.aao.gov.au/ukst/6df.html"}
    };
#endif

/*  A structure of type FibreStatStruct contains all the relevant statistics
 *  for a given fibre type. The global variable FibreTypeStats is a pointer
 *  to a malloc'd array of such a structure.
 */

typedef struct FibreStatStruct {
    char Name[80];                   /* Name describing fibre type */
    int Fibres;                      /* Total # fibres of this type */
    int AllocToSky;                  /* # fibres allocated to sky */
    int AllocToGuide;                /* # fibres allocated to guide stars */
    int AllocToObject;               /* # fibres allocated to objects */
    int Unallocated;                 /* # fibres not allocated to anything */
    int FibreSkyCount;               /* # fibres to be allocated to sky */
    int Targets;                     /* # compatible unallocated targets */
    int Skies;                       /* # compatible unallocated sky targets */
    int NotInUse;                    /* # Number of Fibres marked as not in use 
                                          (Reduces number Unallocated) */
} FibreStatStruct;

/*  ------------------------------------------------------------------------- */

/*           G l o b a l  V a r i a b l e s  &  C o n s t a n t s
 */
 
/*
 *  When an instrument is selected, details are copied here. Note that most
 *  calls to Fpil routines take *Instrument as the first argument, so it is
 *  important that this be defined before any such calls are made. However,
 *  the way the program runs guarantees that the first thing that happens is
 *  that the instrument is selected.
 */

FpilType *Instrument = 0;                    /* Instrument details pointer */
FpilConstantsType InstConstDetails;          /* Used to access constants   */
FpilModelType *TelModel;                     /* For telescope model        */
SdsIdType InstFieldDetails;                  /* Sds id of field details    */
unsigned CurrentField;                       /* Current field number       */
double *TelModelParams;                      /* Telescope model parameters */
double *TelModelParamsUsed;                  /* Telescope model parameters
                                              * used when the last coordinate
                                              * conversion was performed.  */
unsigned int TelModelNumParams;              /* Number of tel model params */
unsigned int InstNumFibreTypes = 0;          /* Number of fibre types      */
unsigned int InstNumPivots = 0;              /* Number of fibres/pivots    */
int *UseFibreType;                           /* Points to array of values  
                                                indicating if a fibre type
                                                should be configured.      */
unsigned char* AllocLockedFlags;             /* Points to array with one entry
                                                for each pivot, indicating if
                                                that fibre is allowed to have 
                                                its allocation changed.    */
FibreStatStruct* FibreTypeStats;             /* Allocation statistics for  
                                                each fibre type.           */
time_t FileTime;                             /* Time used in file name     */

#define MAX_ERRORS 10           /* Maximum allocation errors */
#define ALLOCATE_NORMAL 0       /* ClientData item value to ConfDoAllocation */
#define ALLOCATE_BATCH  1       /* ClientData item value to ConfDoAllocation */

/*
 *  Note that of two below, CMTLEN is always used for buffers which may
 *  take either a comment or name item, hence if their sizes changes.
 *  CMTLEN must remain the largest.
 */
 
#define NAMELEN 80              /* Length of "name" items in structures */
#define CMTLEN  80              /* Length of "comment" items in structures */

/*       M i m i c  W i n d o w  S i z i n g  a n d  S c a l i n  g
 *
 *  Scaling is all controlled by the relationship between the field size
 *  in microns and the number of pixels used for the mimic window. At a
 *  zoom setting of 1, the mimic window shows just a little more than
 *  the whole available field.  The mimic window size in pixels is placed in the
 *  variable MimicWindowSize by the ConfInitialise() function, which
 *  determines it based on a percentage (MIMIC_PROPORTION) of the smallest 
 *  dimension of the display.
 *  
 *  A circle is drawn some pixels (MIMIC_CIRCLE_OUTSIDE) in from this to 
 *  represent the available field.  The radius of this circle is placed in 
 *  MimicCircleRadius and the center of this circle in relation to the mimic 
 *  window is placed in MimicCircleOffset. (For OzPoz, the code allows an
 *  additional number of pixels (MIMIC_GUIDE_EXTRA) to allow for the need to
 *  show the guide probe. For 6dF, we add an additional amount to allow for
 *  the pivot positions being some way outside the field (MIMIC_6DF_EXTRA)).
 *
 *  So all scaling is controlled by the relative values of the field
 *  size (fetched from FPIL using FpilGetFieldRadiusIncPark()) and the value
 *  of the variable MimicCircleRadius (the radius in pixels of the circle
 *  drawn in the   window to represent the field). 
 *
 *  Tcl commands are avaialble to fetch these values.
 */

unsigned int MimicWindowSize = 0;
unsigned int MimicCircleRadius = 0;
unsigned int MimicCircleOffset = 0;

#define MIMIC_PROPORTION 0.8
#define MIMIC_CIRCLE_OUTSIDE 10
#define MIMIC_GUIDE_EXTRA 60
#define MIMIC_6DF_EXTRA 10

/*  useGrey is a global flag that indicates that a grey-scale set of colours
 *  is being used. This is intended as an option mainy for colour-blind users,
 *  since more contrasting hues of grey can be selected if this flag is set.
 *  The tcl command useGrey gets the value of this flag, and the tcl
 *  command setGrey can set it.
 */
 
static short useGrey = 0;

/*  PAF flags are used to control the output for different objects when a
 *  PAF file is being written. The only one needed is PAF_NO_PRIO which
 *  is used to suppress the priority information for VLT guide stars.
 */
 
#define PAF_NO_PRIO 1

/*
 *  Imported from the configversion.c module.
 */
 
extern const char *const ConfigVersion;
extern const char *const ConfigDate;
extern const char *const ConfigSystemVersion;

/*
 * WIN32 compatibility stuff
 */
 
#ifdef WIN32

#define MAXPATHLEN MAX_PATH

/* Symbolic constants for the "access" routine: */

#define R_OK    4               /* Test for Read permission */
#define W_OK    2               /* Test for Write permission */
#define X_OK    1               /* Test for eXecute permission */
#define F_OK    0               /* Test for existence of File */

/*
 * WIN32 does not provide the "access" function. Here is a
 * cheap and nasty equivalent - just checks for existence of the file.
 */
 
static int access(const char *path, int amode)
{
    DWORD attributes = GetFileAttributes(path);
    if (attributes == 0xFFFFFFFF)
        return -1;              /* File does not exist */
    else {
        return 0;               /* File exists, but not sure how to check 
                                 * for readability or executability */
    }
}

#endif

/*  End of Win 32 compatibility stuff */

/*  Sds Ids - kept globally to speed up drawing of displays  */

static SdsIdType gxid = 0, gyid = 0, gthetaid = 0, gxPivid = 0, gyPivid =
    0, gtype = 0;

static SdsIdType gtopid = 0, gpivotid = 0, gxpid = 0, gypid = 0, gthetapid = 0;

/*
 *  XErrorCount is the number of X errors so far
 *  MAX_XERRORS is the maximum number we allow (to avoid error loops)
 */
 
static int XErrorCount = 0;
#define MAX_XERRORS 10

/*
 *  Flag to see if we should ignore some errors GBD 17/6/98
 */
 
static int ImportOnlyMode = 0;

/*  Flag to see if expert mode is allowed */

static int ExpertAllowed = 0;

/*  Flag to control enabling of all fibres, if requested by the -use_all_fibres
 *  command line option.
 */

static int EnableAllFibres = 0;

/*
 *  Executable directory.  Not all operating systems define MAXPATHLEN so
 *  define it here if needed.
 */
 
#ifndef MAXPATHLEN
#define MAXPATHLEN 1024
#endif
static char ExecutableDir[MAXPATHLEN];

/*
 *  Config files directory (value of the CONFIG_FILES parameter)
 */
 
static char ConfigFilesDir[MAXPATHLEN];

/*  Defaults for Temp, Pressure, Humidity and Wavelength. There are three
 *  wavelengths - the main one used for ordinary target objects, the one
 *  used for the guide stars (which will usually depend on any filters in
 *  the guiding system and the response of the guider - we assume that these
 *  effects dominate the colour of the guide star itself), and the wavelength
 *  used for the telescope guider.
 *
 *  At present, the user interface allows the temperature, pressure and
 *  humidity to be changed, and allows the main observing wavelength to
 *  be changed. However, at present, the guide wavelengths are fixed.
 *  Of the supported instruments, only FLAMES makes use of telescope guide
 *  targets. Note that Initialise{} modifies the humidity default if the
 *  instrument is FLAMES, in which case it uses 0.2. It also changes the
 *  pressure for FLAMES to 742.
 */
 
#define TEMP 285
#define PRESS 900
#define HUMID 0.5
#define WAVE 0.60

double gtemp = TEMP;                    /* Temperature in degrees kelvin */
double gpress = PRESS;                  /* Pressure in mm Hg */
double ghumid = HUMID;                  /* Humidity - as a fraction, ie 0 - 1 */
double gwave = WAVE;                    /* Wavelength in microns for main 
                                         * target objects (non-guide objects) */
double gwave2 = WAVE;                   /* Wavelength in microns for main 
                                         * target objects (non-guide objects) 
                                         * for the 2nd 2df spectrograph */

/*  Flag to indicate batch mode operation */

static short batchMode = FALSE;

/*  ------------------------------------------------------------------------- */

/*                   F u n c t i o n  P r o t y p e s
 *
 *  Function prototypes for routines defined in this file.
 */

static int ConfAddCommands(Tcl_Interp * interp);
static int ConfGetArgusData(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc, char *argv[]);
static int ConfSetArgusData(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc, char *argv[]);
static void RCSVersion(char* Version, int Length);
#ifndef NO_X11  /* Don't need this if we don't have X11 */
static int XError(ClientData clientData, XErrorEvent * errorEventPtr);
#endif
static void ErsHandler(void *outArg DUNUSED, unsigned int count,
    const ErsMessageType messages[], StatusType * status);
static void ConfMessage(void *argument, const char *string);
static void ConfRptConstFile(void *argument, const char *string,
    StatusType * status);
static short ConfRptAffectFile(void *argument, short Failed, 
    const char *string, StatusType * status);
static void ConfRptTelModelFile(void *argument, const char *string,
    StatusType * status);

/*  The following routines are made externally visible, so they can be
 *  used by modules such as the PAF input routines.
 */
 
void DoMakeAllocated (
   SdsIdType topid,
   SdsIdType pivid,
   StatusType* status);
   
void ConfCSetArgusData(
    short InUse,
    double Angle,
    char* Scale,
    double Offset);

void ConfCGetArgusData(
    short* InUse,
    double* Angle,
    char* Scale,
    double* Offset);

/*  ------------------------------------------------------------------------- */

/*                 U t i l i t y   R o u t i n e s
 *
 *  Utility routines for use by functions defined in this file. Some of these
 *  are declarations for routines defined later, other short routines are
 *  defined here.
 */

/*  checksum() calculates the checksum of a character array, and 
 *  char_encode() encodes an unsigned integer into a printable ASCII
 *  string. 
 */
 
static void checksum (char* buf, int length, unsigned short* sum16, 
                                              unsigned int* sum32);
                
static void char_encode (unsigned int value, char* ascii, int nbytes,
                                                          int permute);

/*  LinesIntersect() tests whether or not two vectors intersect. */

static int LinesIntersect(double VecX1, double VecY1, double VecX2,
    double VecY2, double X1, double Y1, double X2, double Y2);

/*  Sds Id leak checking routines. If Sds Ids are not released once they
 *  are no longer needed, subtle memory leaks and fragmentation can result.
 *  These routines can be used to test for Sds Id leaks. The basic principle
 *  used here is that if Sds is leaking Ids then Sds Id values (Sds Ids are
 *  just integers) will constantly increase - if Ids are not leaking then
 *  most will be reused as the program progresses.
 */
 
/*
 * Invoked to prepare for a Sds id leak test.  The value returned from
 * this function should be passed to LeakCheck to check for Sds leaks
 * between the two calls.
 */
 
typedef long int LeakCheckType;

static LeakCheckType LeakPrep(void)
{
    StatusType ignore = STATUS__OK;
    SdsIdType id = 0;
    ArgNew(&id, &ignore);
    ArgDelete(id, &ignore);
    if (ignore == STATUS__OK)
        return (LeakCheckType) id;
    else
        return -1;
}

/*
 * Check for Sds leaks.  The origVal should be a value returned
 * from LeakPrep() above prior to the sequences of Sds calls we
 * want to check for leaks.
 *
 * If a leak has occured, a message is written to stderr about it.
 *
 *  routine -> string output in message to indicate where the check was done.
 *  allow   -> We allow this many Sds id's to allocated in this routine
 *             as part of it's job.
 */
 
static void LeakCheck(const char *const routine,
    LeakCheckType origVal, int allowance)
{
    LeakCheckType newVal = LeakPrep() - allowance;
    if ((newVal == -1) || (origVal == -1)) {
        fprintf(stderr, "%s:Error checking for SDS leak\n", routine);
    } else if (newVal != origVal) {
        fprintf(stderr, "### %s:Leaked %ld Sds id's\07\n", routine,
            (newVal - origVal));
    } else {
        fprintf(stderr, "### %s:Leaked NO Sds id's\n", routine);
    }
}

/*  Tcl interfaces to LeakPrep() and LeakCheck() */

static int TclLeakPrep(ClientData clientdata DUNUSED, Tcl_Interp * interp,
    int argc DUNUSED, char *argv[]DUNUSED)
{
    sprintf(interp->result, "%ld", (long int) (LeakPrep()));
    return TCL_OK;
}

static int TclLeakCheck(ClientData clientdata DUNUSED, Tcl_Interp * interp,
    int argc DUNUSED, char *argv[]DUNUSED)
{

    if (argc != 4) {
        sprintf(interp->result, "%s:Wrong number of arguments", argv[0]);
        return TCL_ERROR;
    }
    LeakCheck(argv[1], atol(argv[2]), atol(argv[3]));
    return TCL_OK;
}

/*
 * By setting a break point in this routine, we can cause the debugger
 * to be triggered from Tcl code.
 */
 
static int ConfInvokeDebugger(ClientData clientdata DUNUSED,
    Tcl_Interp * interp DUNUSED, int argc DUNUSED, char *argv[])
{
    printf("%s invoked\n", argv[0]);
    return TCL_OK;
}

/*  ------------------------------------------------------------------------- */

/*
 *                               M a i n
 *
 *  This is the main routine for Configure. It performs any necessary
 *  initialisations prior to starting up Tcl/Tk, and then passes control
 *  over to Tcl/Tk.  Note that Configure is best thought of as a Tcl/Tk
 *  program that makes use of a number of Tcl commands that are implemented
 *  in C. So this routine has the job of starting up Tcl/Tk in such a
 *  way that these additional commands are defined, and doing any necessary
 *  initialisation needed by the C routines that implement the Tcl commands.
 */

extern int main(int argc, char *argv[])
{
    /*  Local variables  */
    
    char filename[MAXPATHLEN];
    char library[MAXPATHLEN];
    char *configFilesValPnt;
    char tcl_library_env[MAXPATHLEN + 40] = "TCL_LIBRARY=";
    char tk_library_env[MAXPATHLEN + 40] = "TK_LIBRARY=";
    StatusType status = STATUS__OK;
    int known;
    short sysgenMode = FALSE;
    register unsigned i;

    char **newArgv;                      /* Modified command line arguments */
    
    /*  Initialise the Mess routines. These decode status return codes,
     *  and need to be told which sets of status codes are expected. 
     */
    
    MessPutFacility(&MessFac_SDS);
    MessPutFacility(&MessFac_ARG);
    MessPutFacility(&MessFac_CONF);
    MessPutFacility(&MessFac_CONFIG);
    MessPutFacility(&MessFac_TDFXY);
    MessPutFacility(&MessFac_SIXDF_UTIL);
    MessPutFacility(&MessFac_FPIL);
    
    /*  Assume we're not in batch mode. We'll check the command arguments
     *  later. Ditto system generation mode.
     */
     
    batchMode = FALSE;
    sysgenMode = FALSE;
    
#ifndef macos9

    /*  This is the usual start-up for Windows or Unix.  The Mac (as usual)
     *  is significantly different and is handled in the #else clause, at
     *  least in its old OS 9 version. OS X is more or less standard UNIX.
     */

    /*  See if we are being run from a program that includes the name
     *  'expert'. If we are, allow expert mode.
     */
     
    if (strstr(argv[0], "expert")) ExpertAllowed = 1;
    
    /*
     *  A quick pass through the command line arguments to see if we are
     *  running in batch mode - one of the -a or -d flags should be specified
     *  for batch operation.
     *  
     *  We also look to see if one of the tcl_library or tk_library flags
     *  is specified, in which case this is just being run as part of the
     *  system generation process. Ditto -tk_version and -tcl_version.
     *  -tk_version is reported here, as there isn't an [info tkversion]
     *  option.
     *
     *  We also look for the -use_all_fibres option. This is handled entirely
     *  by the C layer and allows an expert user to force all fibres to be
     *  regarded as in use. This is really only for use during commissioning,
     *  when the fibres available changes rapidly and many configurations
     *  are done by hand anyway.
     */
 
    for (i = 1; i < (unsigned) argc; ++i) {
        if (!(strncmp(argv[i],"-a",2)) || !(strncmp(argv[i],"-d",2))) {
           batchMode = TRUE;
        }
        if (!(strncmp(argv[i],"-tk_library",11))  || 
            !(strncmp(argv[i],"-tcl_library",12)) ||
            !(strncmp(argv[i],"-tcl_version",12))) {
           sysgenMode = TRUE;
        }
        if (!(strncmp(argv[i],"-tk_version",11))) {
           sysgenMode = TRUE;
           /*
            * It is a point but we can't use printf here - the Mac 
            * stdout library adds a \r before the \n, which messes
            * up the makefile.  (And this only happens if the
            * output is not going to a file).
            */
           write(STDOUT_FILENO, TK_VERSION, strlen(TK_VERSION));
           write(STDOUT_FILENO, "\n", 1);
        }
        if (!(strncmp(argv[i],"-use_all_fibres",15))) {
           if (!ExpertAllowed) {
              printf (
                "-use_all_fibres is only allowed if expert mode is enabled\n");
              return (0);
           } else {
              EnableAllFibres = TRUE;
              printf ("All fibres will be treated as in use\n");
           }
        }
    }
        
    /*
     *  Find where are being run from so we can find other files.
     */
     
    ImpZExeDir(argv[0], ExecutableDir, sizeof(ExecutableDir), &known);

    /*  See if the CONFIG_FILE environment variable has a value and if
     *  so, grab it.
     */
     
    configFilesValPnt = getenv("CONFIG_FILES");
    if (configFilesValPnt) {
        strncpy(ConfigFilesDir, configFilesValPnt, sizeof(ConfigFilesDir));
        ConfigFilesDir[sizeof(ConfigFilesDir) - 1] = '\0';
    } else
        ConfigFilesDir[0] = '\0';

    /*  See if the Tcl library files are in the directory we are being
     *  run from.  If so, we must set TCL_LIBRARY and TK_LIBRARY. We expect
     *  these to be in the directories 'tcl_dir' and 'tk_dir'. As a check,
     *  the makefiles that create these should have included a file in
     *  each giving the version number. We check for that - it's important
     *  that it match the version we've been linked with.
     */
     
    strcpy(library, ExecutableDir);
    strcat(library, "tcl_dir");
    if (access(library, R_OK | X_OK) == 0) {

        /*  Note, the strange putenv function requires that tcl_library_env
         *  be kept about until the variable is overwritten (or in this case
         *  until the program completes).
         */
         
        strcat(tcl_library_env, library);
        strcat(tcl_library_env, "/tcl");
        strcat(tcl_library_env, TCL_VERSION);
        putenv(tcl_library_env);
        
        /*  Check the version - we ar elooking for a file called 
         *  TCL_VERSION_m.n where m.n is the current tcl version.
         */
        
        strcat(library,"/TCL_VERSION_");
        strcat(library,TCL_VERSION);
        if (access(library, R_OK) != 0) {
            fprintf (stderr,
                     "Warning - local tcl_dir may not be for Tcl version %s\n",
                                                                TCL_VERSION);
        }
    }

    /*  Now tk. */
    
    strcpy(library, ExecutableDir);
    strcat(library, "tk_dir");
    if (access(library, R_OK | X_OK) == 0) {
     
        /*  Note, the strange putenv function requires that tk_library_env
         *  be kept about until the variable is overwritten (or in this case
         *  until the program completes).
         */
         
        strcat(tk_library_env, library);
        strcat(tk_library_env, "/tk");
        strcat(tk_library_env, TK_VERSION);
        putenv(tk_library_env);
        
        /*  Check the version - we ar elooking for a file called 
         *  TK_VERSION_m.n where m.n is the current tk version.
         */
        
        strcat(library,"/TK_VERSION_");
        strcat(library,TK_VERSION);
        if (access(library, R_OK) != 0) {
            fprintf (stderr,
                     "Warning - local tk_dir may not be for Tk version %s\n",
                                                                TK_VERSION);
        }
    }

    /*  Find the configure tcl script.  We first look for the environment
     *  variable defined by CONF_SCRIPT.  If that is not defined, then we
     *  look for configure.tcl in the directory CONFIGURE_DIR.  Finally, we
     *  look for configure.tcl in the executable directory.
     */
     
    if (getenv(CONF_SCRIPT)) {
        strcpy(filename, getenv(CONF_SCRIPT));
        printf("%s:Using script file \"%s\"\n", argv[0], filename);
    } else if (getenv(CONF_TCLDIR)) {
        strcpy(filename, getenv(CONF_TCLDIR));
        strcat(filename, "/");
        strcat(filename, CONF_TCLFILE);
    } else {
        strcpy(filename, ExecutableDir);
        strcat(filename, CONF_TCLFILE);
    }

    /*  We have to add the Tcl script filename as the first argument.  This
     *  is required by Tk_Main.  Thus, we have to malloc space for a new
     *  argv array.
     */
     
    newArgv = malloc(sizeof(char *) * (argc + 1));
    newArgv[0] = argv[0];
    newArgv[1] = filename;
    for (i = 1; i < (unsigned) argc; ++i) {
        newArgv[i + 1] = argv[i];
    }
    
    /*  Make sure the global FileTime variable contains a sensible time */
    
    FileTime = time(0);
    
#else /* defined macos9 */

    /*
     *  Macintosh OS9  is rather different from windows and unix. We can really
     *  only make use of the executable directory. We also need to run
     *  the Mac-specific initialisation code provided by Tcl.
     */
     
    if (MacintoshInit() != TCL_OK) Tcl_Exit(1);
    
    ExpertAllowed = 1;
    argc = 0;
    newArgv = malloc(sizeof(char *) * (argc + 1));
    newArgv[0] = "configure";
    newArgv[1] = NULL;
    if (getcwd(ExecutableDir, sizeof(ExecutableDir)) == NULL) {
        ExecutableDir[0] = '\0';
    } else {
        ExecutableDir[strlen(ExecutableDir) - 1] = '\0';
    }
    ConfigFilesDir[0] = '\0';
    
#endif                          /* macintosh */

    /*  Initialise each supported instrument. We don't have to do this in
     *  system generation mode, and it confuses the output if we do.
     */
    
    if (!sysgenMode) {
       for (i = 0; i < sizeof(Instruments) / sizeof(Instruments[0]); ++i) {
           (*Instruments[i].InitRoutine) (&Instruments[i].Info);
       }
    }
       
/*
 *  Initialise Tcl/Tk and run the program. From here on, Tcl/Tk takes
 *  control, and we only get back into C when ConfAddCommands() is called
 *  to define the C-implemented Tcl commands, and when one of those
 *  commands is invoked.
 */
    if (batchMode) {
       Tcl_Main(argc + 1, newArgv, ConfAddCommands);
    } else {
       Tk_Main(argc + 1, newArgv, ConfAddCommands);
    }

/*
 *  The Tk documentation says the above never returns, but just in case,
 *  do our tidying up.
 */
    free(newArgv);
    ErsStop(&status);

    return (0);
}

/*  ------------------------------------------------------------------------- */

/*
 *                        T c l  C o m m a n d s
 *
 *   The rest of the code in this file provides the implementation of the
 *   various new Tcl commands defined in ConfAddCommands().
 */

/*  ------------------------------------------------------------------------- */

/*
 *                       I n i t i a l i s e { }
 */
 
static int ConfInitialise(ClientData clientData DUNUSED,
                          Tcl_Interp * interp, int argc, char *argv[])

/*+
 *  Command Name:
 *    Initialise
 *
 *  Function:
 *    Initialise configuration algorithms
 *
 *  Call:
 *    Initialise overrideDir defaultDir instrument field x y
 *
 *  Description:
 *    Initialise the configuration algorithm routines by calling selecting
 *    the instrument, the initial field and calling ConfigInitialise
 *
 *    Instrument details are read at this time.  You cannot change
 *    the instrument after this call.
 *
 *    The field details are read at this time, but this may be changed
 *
 *
 *  Parameters:
 *    (>)  instrument  (string)  Name of the instrument to enable
 *                               If supplied as an empty string, the first 
 *                               instrument known by the software is used.
 *    (>)  field      (int)      The initial plate/field to configure for.
 *    (>)  x          (int)      Display x dimension.
 *    (>)  y          (int)      Display y dimension.
 *
 *
 *   Support:
 *     Jeremy Bailey, AAO
 *     Tony Farrell, AAO
 *
 *   Version Date:
 *     15-Mar-2000    
 *-
 */
{

    StatusType status = STATUS__OK;
    unsigned InstrumentIndex = 0;
    int field;
    int fielddisplay; 
    int x;
    int y;

    if (argc != 5) {
        sprintf(interp->result, "%s:Wrong number of arguments", argv[0]);
        return TCL_ERROR;
    }

/*
 *  Parse the integer arguments.
 */
    if (Tcl_GetInt(interp, argv[2], &field) != TCL_OK) {
        return TCL_ERROR;
    }
    if (Tcl_GetInt(interp, argv[3], &x) != TCL_OK) {
        return TCL_ERROR;
    }
    if (Tcl_GetInt(interp, argv[4], &y) != TCL_OK) {
        return TCL_ERROR;
    }
/*
 *  If an instrument name has been specified, attempt to find that instrument.
 */
    if (strlen(argv[1]) != 0) {
        register unsigned i;
        for (i = 0; i < sizeof(Instruments) / sizeof(Instruments[0]); ++i) {
            if (strcasecmp(argv[1], 
                           FpilGetInstName(Instruments[i].Info)) == 0) {
                InstrumentIndex = i;
                break;
            }
        }
        if (i >= sizeof(Instruments) / sizeof(Instruments[0])) {
            sprintf(interp->result,
                    "%s:Instrument \"%s\" unknown", argv[0], argv[1]);
            return TCL_ERROR;
        }
    }

/*
 *  Set up the instrument
 */
    Instrument = &Instruments[InstrumentIndex].Info;

    printf("Configuring fibre instrument \"%s\" for telescope \"%s\"\n",
           FpilGetInstName(*Instrument), FpilGetTelescope(*Instrument));

    FpilConsAccess(*Instrument, ConfigFilesDir, ExecutableDir,
                   ConfRptConstFile, 0, &InstConstDetails, &status);

    if (status != STATUS__OK)
    {
        ErsRep(0, &status, "Failed to access instrument configuration file.");
        ErsRep(0, &status, "This file normally needs to be downloaded.");
        if (Instruments[InstrumentIndex].inst_file)            
            ErsRep(0, &status, "Please goto %s for this file",
                   Instruments[InstrumentIndex].inst_file);
        ErsRep(0, &status, "Documentation should be available from %s",
               Instruments[InstrumentIndex].doco_site);
        exit(1);
    }
    
/*
 *  Read in any affectation data - data about how light from one fibre can
 *  spill over (ghosting, crosstalk, etc) onto another. Note that if the
 *  reporting routine clears any bad status - which it can do, and in fact
 *  does at present - FpilReadAffectData() will not return bad status, even
 *  if no affectation data could be found.
 */
 
    FpilReadAffectData (*Instrument,ConfigFilesDir, ExecutableDir,
                                        ConfRptAffectFile, 0, &status);
    if (status != STATUS__OK)
    {
        char ErrorDetails[1024];
        FpilReadAffectError(sizeof(ErrorDetails),ErrorDetails);
        ErsRep (0,&status,"Failed to access all affectation data files.");
        ErsRep (0,&status,"%s",ErrorDetails);
        exit(1);
    }

/*
 *  Modify the humidity and pressure defaults, now that we know the instrument.
 *  The VLT is in a somehat drier and higher location than are 2dF and 6dF.
 */
    if (!strcmp(FpilGetInstName(*Instrument), "FLAMES")) {
        ghumid = 0.2;
        gpress = 742;
    }

/*
 *  Validate initial value number.
 */
    fielddisplay = field;
    if (!strcmp(FpilGetInstName(*Instrument), "6dF")) 
    {
        /*
         * In 6dF, the user supplies a 1 based field plate number.  We must
         * convert it to zero based.
         */
        int numFields = FpilGetNumFields(*Instrument);
        fprintf(stderr, "Configuring for 6dF - with %d field plates - field %d\n",
                numFields, field);

        field -= 1;
        if ((field < 0)||(field >= numFields))
        {
            sprintf(interp->result,
                    "%s:Field %d not valid, %s, plate range is 1 to %d\n",
                    argv[0],
                    fielddisplay,
                    FpilGetInstName(*Instrument),
                    numFields);
            return TCL_ERROR;
        }
    } else if (field >= FpilGetNumFields(*Instrument)) {
        sprintf(interp->result,
                "%s:Field %d not valid, %s has only %d field plates\n",
                argv[0],
                fielddisplay,
                FpilGetInstName(*Instrument), FpilGetNumFields(*Instrument));
        return TCL_ERROR;
    } 
    {
        const char *const *ParamNames;
        const char *const *ParamDescrs;
        const double *ParamDefaults;
        const char *const *SpecNames;
        register unsigned i;

        CurrentField = field;
/*
 *          Field number valid.  Access details and read telescope model
 *          details for this field.
 */
        FpilConFieldId(InstConstDetails, CurrentField, &InstFieldDetails,
                       &status);
            
/*          At this point, if the -use_all_fibres command line flag was given
 *          and allowed, we enable all the fibres, irrespective of the actual
 *          inUse flag values. This code is rather nasty and low-level, but
 *          it sets all the inUse flags for all the fields.
 */
        if (EnableAllFibres) {
            
            int NumFields;         /* Number of fields in structure */
            int Field;             /* Loop index through fields */
            SdsIdType FieldID;     /* Sds id for field structure */
            SdsIdType inUseID;     /* Sds id for inUse array */
            char ItemName[20];     /* Name of item in structure - ignored */
            SdsCodeType Code;      /* Type code for item - ignored */
            unsigned long Dims[10];/* Dimensions of structure object */
            long NDims;            /* Number of dimensions for object */
            unsigned long Length;  /* Size of structure item in bytes */
            void *VPointer;        /* Receives pointer to array data */
            short *SPtr;           /* Start of inUse flags */
            int NPiv;              /* # of pivots - ie size of inUse array */
            int I;                 /* Loop index through inUse flags */
               
            NumFields = FpilGetNumFields(*Instrument);
            for (Field = 0; Field <  NumFields; Field++) {
                FpilConFieldId(InstConstDetails, Field, &FieldID, &status);
                SdsFind (FieldID,"inUse",&inUseID,&status);
                SdsInfo (inUseID,ItemName,&Code,&NDims,Dims,&status);
                SdsPointer (inUseID,&VPointer,&Length,&status);
                NPiv = Dims[0];
                SPtr = (short *) VPointer;
                for (I = 0; I < NPiv; I++) {
                    *SPtr++ = 1;
                }
                SdsFreeId (inUseID,&status);
                SdsFreeId (FieldID,&status);
            }
        }
            
/*          Allocate space for the 'locked allocation' flags. These
 *          are used to indicate that fibre is not to have its allocation
 *          changed.
 */
        InstNumPivots = FpilGetNumPivots(*Instrument);
        AllocLockedFlags = 
            (unsigned char*) malloc(InstNumPivots *  sizeof(unsigned char));
        for (i = 0; i < InstNumPivots; i++) {
            AllocLockedFlags[i] = CONFIG_FALSE;
        }

/*          Allocate space and initialise the flags which indicate if
 *          given fibre types are to be used, and the structure that
 *          keeps the allocation statistics for each fibre type.
 */
        FpilConSpecInfo(*Instrument, &InstNumFibreTypes, &SpecNames,
                        &status);
        UseFibreType = malloc((InstNumFibreTypes) * sizeof(*UseFibreType));
        FibreTypeStats = malloc((InstNumFibreTypes) * 
                                sizeof(FibreStatStruct));
        for (i = 0; i < InstNumFibreTypes; ++i) {
            UseFibreType[i] = 1;
            FibreTypeStats[i].Name[0]       = '\0';
            FibreTypeStats[i].Fibres        = 0;
            FibreTypeStats[i].AllocToSky    = 0;
            FibreTypeStats[i].AllocToGuide  = 0;
            FibreTypeStats[i].AllocToObject = 0;
            FibreTypeStats[i].Unallocated   = 0;
            FibreTypeStats[i].FibreSkyCount = -1;
            FibreTypeStats[i].NotInUse      = 0;
        }                     
        FpilModelAccess(*Instrument, ConfigFilesDir, ExecutableDir,
                        ConfRptTelModelFile, 0, CurrentField, &TelModel, &status);

/*
 *          Get details of the telescope model parameters.
 */
        FpilModelParams(TelModel, &TelModelNumParams,
                        &ParamNames, &ParamDescrs, &ParamDefaults, &status);

        if (status == STATUS__OK) {
/*
 *              Malloc two arrays to contain the parameter values and copy
 *              in the default parameter values. One copy is used for the
 *              current parameter values, one to record the values used
 *              the last time the XY-conversion of positions was performed
 *              (See ConvertXY).
 */
            if (TelModelNumParams) {
                TelModelParams = malloc(sizeof(double) * TelModelNumParams);
                TelModelParamsUsed = 
                    malloc(sizeof(double) * TelModelNumParams);
                for (i = 0; i < TelModelNumParams; ++i) {
                    TelModelParams[i] = ParamDefaults[i];
                    TelModelParamsUsed[i] = ParamDefaults[i];
/*
  printf("Parameter %d, name %s, descr %s, default %g\n",
  i, ParamNames[i], ParamDescrs[i], 
  ParamDefaults[i]);
*/
                }
            } else {
                TelModelParams = 0;
            }
        }
    }

/*
 *  Work out mimic sizes.  x and y represent the size of the display.
 *  We work out the smallest of the two values and then the mimic
 *  window size is set to MIMIC_PROPORTION of this.
 */
    if (x < y) {
        MimicWindowSize = (unsigned int) ((float) x * MIMIC_PROPORTION);
    } else {
        MimicWindowSize = (unsigned int) ((float) y * MIMIC_PROPORTION);
    }
/*
 *  Now work out the position of the circle center and the size of the circle.
 */
    MimicCircleOffset = MimicWindowSize / 2;
    MimicCircleRadius = MimicCircleOffset - MIMIC_CIRCLE_OUTSIDE;
    if (!strcmp(FpilGetInstName(*Instrument), "FLAMES")) {
        MimicCircleRadius -= MIMIC_GUIDE_EXTRA;
    }
    if (!strcmp(FpilGetInstName(*Instrument), "6dF")) {
        MimicCircleRadius -= MIMIC_6DF_EXTRA;
    }

/*  
 *  Initialize Config package  
 */

    ConfigInitialise(*Instrument, InstConstDetails, ConfMessage,
                     interp, &status);


    if (status != STATUS__OK) {
        char mbuf[100];
        MessGetMsg(status, 0, sizeof(mbuf), mbuf);
        sprintf(interp->result, "%s:Error initialising configure -\n", mbuf);
        return TCL_ERROR;
    }
/*  
 *  Ers had not yet been started, so any previous errors will goto
 *  to stderr.  We are now ready to start ERS if status is ok.
 *  Initialise ERS.  From now on, ERS errors will goto the GUI.
 */
    ErsStart(ErsHandler, interp, 0, 0, &status);

    return TCL_OK;
}

/*  ------------------------------------------------------------------------- */

/*              F i e l d  P l a t e  T o  D i s p  X / Y
 *
 *  Functions to convert a field plate position in x/y in microns to
 *  the corresponding x/y value on the display, taking account of 
 *  the zoom factor. Note that the function signature must match
 *  the FpilCvtRoutineType defined in fpil.h.
 */
 
static int FieldPlateToDispX(void *clientData DUNUSED,
    double Microns, double Zoom)
{
    double Result;
    double circleOffset = MimicCircleOffset; /* Just to convert int to double */
    double scale = (double) MimicCircleRadius /
        (double) FpilGetFieldRadiusIncPark(*Instrument);

    Result = (scale * Microns + circleOffset) * Zoom + 0.5;
    return ((int) Result);
}

static int FieldPlateToDispY(void *clientData DUNUSED,
    double Microns, double Zoom)
{
    double Result;
    double circleOffset = MimicCircleOffset; /* Just to convert int to double */
    double scale = (double) MimicCircleRadius /
        (double) FpilGetFieldRadiusIncPark(*Instrument);

    Result = (-1.0 * scale * Microns + circleOffset) * Zoom + 0.5;
    return ((int) Result);
}

/*  ------------------------------------------------------------------------- */

/*              D i s p  T o  F i e l d  P l a t e  X / Y
 *
 *  These are the reverse of the previous conversions. Note that these
 *  are not a mirror image of the previous routines, since these have an
 *  offset value in their parameters. This just happens to be more convenient
 *  given the way the rest of the code uses these routines.
 */

static int DispToFieldPlateX(double Pixels, /* Pixel units to convert */
    double Offset,           /* Offset of window in pixels (scroll position) */
    double Zoom)             /* Current zoom factor */
{ 
    double Result;
    double circleOffset = MimicCircleOffset - Offset;
    double scale = (double) FpilGetFieldRadiusIncPark(*Instrument) /
        (double) MimicCircleRadius;

    Result = ((Pixels / Zoom) - circleOffset) * scale + 0.5;
    return ((int) Result);
}

static int DispToFieldPlateY(double Pixels, /* Pixel units to convert */
    double Offset,           /* Offset of window in pixels (scroll position) */
    double Zoom)             /* Current zoom factor */
{
    double Result;
    double circleOffset = MimicCircleOffset - Offset;
    double scale = (double) FpilGetFieldRadiusIncPark(*Instrument) /
        (double) MimicCircleRadius;

    Result = (circleOffset - (Pixels / Zoom)) * scale + 0.5;
    return ((int) Result);
}
 
/*  ------------------------------------------------------------------------- */

/*
 *                     I n s t r u m e n t s  { }
 */
 
static int ConfInstruments(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc DUNUSED, char *argv[]DUNUSED)


/*+
 *  Command Name:
 *    Instruments
 *
 *  Function:
 *    Returns the list of supported instruments.
 *
 *  Call:
 *   Instruments
 *
 *  Description:
 *   Returns the list of supported fibre positioner instrumetns.
 *
 *
 *  Parameters: none
 *
 *
 *   Support:
 *     Tony Farrell, AAO
 *
 *   Version Date:
 *     13-Mar-2000    
 *-
 */
{

    register unsigned i;

/*
 *  We assume there is at least one instrument.
 */
    Tcl_AppendElement(interp, (char *)FpilGetInstName(Instruments[0].Info));

    for (i = 1; i < sizeof(Instruments) / sizeof(Instruments[0]); ++i) {
/*
 *      Append a space and then the next instrument name.
 */
        Tcl_AppendElement(interp,(char *)FpilGetInstName(Instruments[i].Info));
    }
    return TCL_OK;
}

/*  ------------------------------------------------------------------------- */

/*
 *                     I n s t r u m e n t  { }
 */
 
static int ConfInstrument(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc DUNUSED, char *argv[]DUNUSED)


/*+
 *  Command Name:
 *    Instrument
 *
 *  Function:
 *    Returns the name of the current instrument.
 *
 *  Call:
 *   Instrument
 *
 *  Description:
 *   Returns the name of the instrument being configure.
 *
 *
 *  Parameters: none
 *
 *
 *   Support:
 *     Tony Farrell, AAO
 *
 *   Version Date:
 *     13-Mar-2000    
 *-
 */
 
{

    if (Instrument == 0) {
        sprintf(interp->result, "%s:Configure not yet initialised", argv[0]);
        return TCL_ERROR;
    }
    sprintf(interp->result, "%s", FpilGetInstName(*Instrument));

    return TCL_OK;
}

/*  ------------------------------------------------------------------------- */

/*
 *                         T e l e s c o p e  { }
 */

static int ConfTelescope(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc DUNUSED, char *argv[]DUNUSED)


/*+
 *  Command Name:
 *    Telescope
 *
 *  Function:
 *    Returns the name of the Telescope..
 *
 *  Call:
 *   Telescope
 *
 *  Description:
 *   Returns the name of the Telescope for the instrument being configured.
 *
 *
 *  Parameters: none
 *
 *
 *   Support:
 *     Tony Farrell, AAO
 *
 *   Version Date:
 *     13-Mar-2000    
 *-
 */
{

    if (Instrument == 0) {
        sprintf(interp->result, "%s:Configure not yet initialised", argv[0]);
        return TCL_ERROR;
    }
    sprintf(interp->result, "%s", FpilGetTelescope(*Instrument));

    return TCL_OK;
}

/*  ------------------------------------------------------------------------- */

/*
 *                      N u m  P i v o t s  { }
 */

static int ConfNumPivots(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc DUNUSED, char *argv[]DUNUSED)


/*+
 *  Command Name:
 *    NumPivots
 *
 *  Function:
 *    Returns the number of pivots in the current instrument.
 *
 *  Call:
 *   NumPivots
 *
 *  Description:
 *   Returns the number of pivots in the current fibre instrument
 *
 *
 *  Parameters: none
 *
 *
 *   Support:
 *     Tony Farrell, AAO
 *
 *   Version Date:
 *     15-Mar-2000    
 *-
 */
{

    if (Instrument == 0) {
        sprintf(interp->result, "%s:Configure not yet initialised", argv[0]);
        return TCL_ERROR;
    }
    sprintf(interp->result, "%d", FpilGetNumPivots(*Instrument));

    return TCL_OK;
}

/*  ------------------------------------------------------------------------- */

/*
 *                      N u m  F i e l d s  { }
 */

static int ConfNumFields(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc DUNUSED, char *argv[]DUNUSED)


/*+
 *  Command Name:
 *    NumFields
 *
 *  Function:
 *    Returns the number of field plates in the current instrument.
 *
 *  Call:
 *    NumPivots
 *
 *  Description:
 *    Returns the number of field plates in the current fibre instrument
 *
 *  Parameters: none
 *
 *   Support:
 *     Tony Farrell, AAO
 *
 *   Version Date:
 *     15-Mar-2000    
 *-
 */
{

    if (Instrument == 0) {
        sprintf(interp->result, "%s:Configure not yet initialised", argv[0]);
        return TCL_ERROR;
    }
    sprintf(interp->result, "%d", FpilGetNumFields(*Instrument));

    return TCL_OK;
}

/*  ------------------------------------------------------------------------- */

/*
 *                      C l a s s i f y  F i l e  { }
 */

static int ConfClassifyFile (ClientData clientData DUNUSED,
                      Tcl_Interp * interp, int argc, char *argv[])

/*+
 *  Command Name:
 *     ClassifyFile
 *
 *  Function:
 *     Attempts to work out what sort of a file has been specified.
 *
 *  Call:
 *     ClassifyFile Filename
 *
 *  Description:
 *     Returns one of: "PAF", "SDS", "FLD" or "UNKNOWN"
 *
 *  Parameters: 
 *     Filename  (string) Full name of the file in question. 
 *    
 *  Support:
 *     Keith Shortridge, AAO
 *
 *  Version Date:
 *     17-May-2001    
 *-
 */
{
   FILE* theFile;
   char DataLine[256];
   int LineCount;
   char* FileType;
   char theChar;
   char* DotPtr;
   char* FilePtr;
   char* LinePtr;
   char* FoldPtr;
   short Known;

   if (argc != 2) {
       interp->result = "wrong # args";
       return TCL_ERROR;
   }

   Known = FALSE;
   FileType = "UNKNOWN";
   
   /*  If the file exists, we try to look at its contents rather than
    *  rely on the file name.
    */
    
   if (access(argv[1], R_OK) == 0) {
   
      if ((theFile = fopen(argv[1], "r")) == NULL) {
          sprintf(interp->result, "Cannot Open file %s", argv[1]);
          return TCL_ERROR;
      }

      LineCount = 0;

      /*  We look at the first 10 non-comment lines in the file.  By then
       *  we ought to have found either a "LABEL" line in an ASCII field file,
       *  or a "PAF.HDS.START" line in a PAF file, or something unprintable,
       *  in which case we assume an SDS file.
       */

      while (LineCount < 10) {

         if (fgets(DataLine,sizeof(DataLine),theFile) == NULL) break;

         /*  Look for the first character that isn't a blank.  If we hit an 
          *  end of line, then go on to the next line.
          */

         LinePtr = DataLine;
         while ((theChar = *LinePtr) != 0) {
            if (theChar == '\n') break;
            if (theChar == ' ') {
               LinePtr++;
               continue;
            }
            if (!isprint((int)theChar)) {

               /*  We've got a non-printable character. This can't be a
                *  PAF and standard ASCII file. It's a binary file, so we
                *  assume it's an SDS file.
                */

               FileType = "SDS";
               Known = TRUE;
               break;

            } else {

               /*  We've got a non-space, printable character. If this is
                *  a '#' we're looking at a comment in a PAF file. If it's
                *  a '*' we're looking at a comment in an ASCII field file.
                *  We don't judge the file on that basis, however. We want a
                *  real line to look at.
                */

               if ((theChar == '#') || (theChar == '*')) break;

               /*  We've now got the start of a non-comment line. Look
                *  for a characteristic line for a file. First we fold the
                *  rest of the line to upper case.
                */

               FoldPtr = LinePtr;
               while ((theChar = *FoldPtr) != 0) {
                  if (!isprint((int)theChar)) break;
                  if (islower((int)theChar)) *FoldPtr = toupper(theChar);
                  FoldPtr++;
               } 
               if (!strncmp(LinePtr,"LABEL",5)) {
                  FileType = "FLD";
                  Known = TRUE;
                  break;
               } else if (!strncmp(LinePtr,"PAF.HDR.START",13)) {
                  FileType = "PAF";
                  Known = TRUE;
                  break;
               } else {
                  LineCount++;
                  break;
               }
            }
            LinePtr++;
         }
         if (Known) break;
      }
      fclose(theFile);
   }
   
   /*  If after that, we don't have a classification - which will be the
    *  case if the file didn't exist - we fall back on looking at the
    *  file extension. Find the last '.' in the filename, and look at
    *  what follows.
    */
   
   if (!Known) {
      DotPtr = (char*) NULL;
      FilePtr = argv[1];
      while ((theChar = *FilePtr) != 0) {
         if (theChar == '.') DotPtr = FilePtr;
         FilePtr++;
      }
      if (DotPtr) {
         if (!strcmp(DotPtr,".sds")) {
            FileType = "SDS";
         } else if (!strcmp(DotPtr,".paf")) {
            FileType = "PAF";
         } else if (!strcmp(DotPtr,".ins")) {
            FileType = "PAF";
         } else if (!strcmp(DotPtr,".fld")) {
            FileType = "FLD";
         }
      }
   }

   strcpy(interp->result,FileType);

   return TCL_OK;
}

/*  ------------------------------------------------------------------------- */

/*
 *                      T a r g e t  D e s c r  { }
 */

static int ConfTargetDescr (ClientData clientData DUNUSED,
                      Tcl_Interp * interp, int argc, char *argv[])

/*+
 *  Command Name:
 *     TargetDescr
 *
 *  Function:
 *     Given a target type code character, returns a fuller description.
 *
 *  Description:
 *     This routine is passed the single character code used to classify a
 *     target object, and returns the longer text description of the
 *     target type. For example, 'M' will return 'Medusa target'.
 *
 *  Call:
 *     TargetDescr Type
 *
 *  Parameters: 
 *     Type  (char) Single character giving the object type. 
 *
 *  Returns:
 *     Description of object type
 *   
 *  Support:
 *     Keith Shortridge, AAO
 *
 *  Version Date:
 *     30-Oct-2001    
 *-
 */
{
    char TargetString[32];
    char TargetText[128];
    char TargetType;

    if (argc != 2) {
        interp->result = "wrong # args";
        return TCL_ERROR;
    }
    if (Instrument == 0) {
        sprintf(interp->result, "%s:Configure not yet initialised", argv[0]);
        return TCL_ERROR;
    }

    TargetType = argv[1][0];
    if (FpilDecodeTargetType(*Instrument,TargetType,' ',0,-1, TargetString,
                                                                TargetText)) {

       strcpy(interp->result,TargetText);
    } else {
       strcpy(interp->result,"Unknown type");
    }
    return TCL_OK;
}

/*  ------------------------------------------------------------------------- */

/*
 *                      F i b r e  T y p e s  { }
 */

static int ConfFibreTypes(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc DUNUSED, char *argv[]DUNUSED)

/*+
 *  Command Name:
 *    FibreTypes
 *
 *  Function:
 *    Returns the fibre type names.
 *
 *  Call:
 *   FibreTypes
 *
 *  Description:
 *   Returns a list giving the names of the the fibre types supported
 *   by the instrument.
 *
 *  Parameters: none
 *
 *   Support:
 *     Tony Farrell, AAO
 *
 *   Version Date:
 *     15-Mar-2000    
 *-
 */
{

    unsigned int NumFibreTypes;
    const char *const *FibreTypeNames;
    StatusType status = STATUS__OK;
    register unsigned i;
    Tcl_DString Names;

    if (Instrument == 0) {
        sprintf(interp->result, "%s:Configure not yet initialised", argv[0]);
        return TCL_ERROR;
    }

    FpilConSpecInfo(*Instrument, &NumFibreTypes, &FibreTypeNames, &status);
    if (status != STATUS__OK) {
        char mbuf[100];
        MessGetMsg(status, 0, sizeof(mbuf), mbuf);
        sprintf(interp->result,
            "%s:Error getting fibre type details - %s\n", argv[0], mbuf);
        return TCL_ERROR;
    }

    Tcl_DStringInit(&Names);
    for (i = 0; i < NumFibreTypes; ++i) {
        Tcl_DStringAppendElement(&Names, FibreTypeNames[i]);
    }
    Tcl_DStringResult(interp, &Names);
    Tcl_DStringFree(&Names);

    return TCL_OK;
}

/*  ------------------------------------------------------------------------- */

/*
 *                     T a r g e t  C r i t e r i a  { }
 */

static int ConfTargetCriteria(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc DUNUSED, char *argv[]DUNUSED)

/*+
 *  Command Name:
 *    TargetCriteria
 *
 *  Function:
 *    Returns the target selection criteria.
 *
 *  Call:
 *    TargetCriteria
 *
 *  Description:
 *   Returns a list giving the names of the the target selection criteria
 *   supported by the instrument.
 *
 *
 *  Parameters: none
 *
 *
 *   Support:
 *     Keith Shortridge, AAO
 *
 *   Version Date:
 *     21-Aug-2000    
 *-
 */
{

    unsigned int NumCriteria;
    const char *const *TargetCriteria;
    register unsigned i;
    Tcl_DString Names;

    if (Instrument == 0) {
        sprintf(interp->result, "%s:Configure not yet initialised", argv[0]);
        return TCL_ERROR;
    }

    FpilTargetCriteria(*Instrument, &NumCriteria, &TargetCriteria);

    Tcl_DStringInit(&Names);
    for (i = 0; i < NumCriteria; ++i) {
        Tcl_DStringAppendElement(&Names, TargetCriteria[i]);
    }
    Tcl_DStringResult(interp, &Names);
    Tcl_DStringFree(&Names);

    return TCL_OK;
}

/*  ------------------------------------------------------------------------- */

/*
 *                     F i b r e  C o m b o s  { }
 */

static int ConfFibreCombos(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc DUNUSED, char *argv[]DUNUSED)

/*+
 *  Command Name:
 *    FibreCombos
 *
 *  Function:
 *    Returns the set of possible fibre combinations.
 *
 *  Call:
 *    FibreCombos
 *
 *  Description:
 *    Returns a list giving the names of the the possible fibre combinations
 *    supported by the instrument.
 *
 *  Parameters: none
 *
 *   Support:
 *     Keith Shortridge, AAO
 *
 *   Version Date:
 *     05-Feb-2001
 *-
 */
{

    int NumCombos;
    char* ComboNames[16];
    char* ModeNames[16];
    register unsigned i;
    Tcl_DString Names;

    if (Instrument == 0) {
        sprintf(interp->result, "%s:Configure not yet initialised", argv[0]);
        return TCL_ERROR;
    }

    FpilFibreCombos(*Instrument, sizeof(ComboNames)/sizeof(char*),
                                          &NumCombos, ComboNames, ModeNames);

    Tcl_DStringInit(&Names);
    for (i = 0; i < NumCombos; ++i) {
        Tcl_DStringAppendElement(&Names, ComboNames[i]);
    }
    Tcl_DStringResult(interp, &Names);
    Tcl_DStringFree(&Names);

    return TCL_OK;
}

/*  ------------------------------------------------------------------------- */

/*
 *                  F i b r e  C o m b o  M o d e  { }
 */

static int ConfFibreComboMode(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc, char *argv[])

/*+
 *  Command Name:
 *    FibreComboMode
 *
 *  Function:
 *    Returns the set of possible fibre combinations.
 *
 *  Call:
 *    FibreComboMode ComboId
 *
 *  Description:
 *   Returns a string giving the mode keyword associated with a
 *   given fibre combination.
 *
 *  Parameters:
 *     ComboId   (int) Index number of the combination in question. This 
 *                     should start from zero and should be the index into
 *                     the list of combinations returned by FibreCombos.
 *
 *   Support:
 *     Keith Shortridge, AAO
 *
 *   Version Date:
 *     08-Apr-2001
 *-
 */
{
    int NumCombos;
    int ComboId;
    char* ComboNames[16];
    char* ModeNames[16];

    if (Instrument == 0) {
        sprintf(interp->result, "%s:Configure not yet initialised", argv[0]);
        return TCL_ERROR;
    }
    if (argc != 2) {
        interp->result = "wrong # args";
        return TCL_ERROR;
    }
    ComboId = (atol(argv[1]));
    
    FpilFibreCombos(*Instrument, sizeof(ComboNames)/sizeof(char*),
                                          &NumCombos, ComboNames, ModeNames);
                                          
    if ((ComboId < 0) || (ComboId >= NumCombos)) {
        interp->result = "invalid combination index";
        return TCL_ERROR;
    }
    strcpy(interp->result,ModeNames[ComboId]);
    return TCL_OK;
}

/*  ------------------------------------------------------------------------- */

/*
 *                  F i b r e  C o m b o  T y p e s  { }
 */

static int ConfFibreComboTypes(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc, char *argv[])

/*+
 *  Command Name:
 *    FibreComboTypes
 *
 *  Function:
 *    Returns the set of fibre types included in a given combination.
 *
 *  Call:
 *    FibreComboTypes ComboId
 *
 *  Description:
 *    Returns a list giving the fibre type codes for each fibre type included
 *    in the specified combination.
 *
 *  Parameters:
 *     ComboId   (int) Index number of the combination in question. This 
 *                     should start from zero and should be the index into
 *                     the list of combinations returned by FibreCombos.
 *
 *   Support:
 *     Keith Shortridge, AAO
 *
 *   Version Date:
 *     06-Feb-2001
 *-
 */
{

    int ComboId;
    int TypeCodes[16];
    int NumCodes;
    register unsigned i;
    Tcl_DString Codes;

    if (Instrument == 0) {
        sprintf(interp->result, "%s:Configure not yet initialised", argv[0]);
        return TCL_ERROR;
    }
    if (argc != 2) {
        interp->result = "wrong # args";
        return TCL_ERROR;
    }
    ComboId = (atol(argv[1]));

    FpilComboFibreCodes(*Instrument, ComboId,sizeof(TypeCodes)/sizeof(int),
                                                      &NumCodes, TypeCodes);

    Tcl_DStringInit(&Codes);
    for (i = 0; i < NumCodes; ++i) {
        char CodeString[16];
        sprintf (CodeString,"%d",TypeCodes[i]);
        Tcl_DStringAppendElement(&Codes, CodeString);
    }
    Tcl_DStringResult(interp, &Codes);
    Tcl_DStringFree(&Codes);

    return TCL_OK;
}

/*  ------------------------------------------------------------------------- */

/*
 *                  S e t  I m p o r t  O n l y  M o d e  { }
 */

static int ConfSetImportOnlyMode(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc, char *argv[])
    
/*+
 *  Command Name:
 *    SetImportOnlyMode
 *
 *  Function:
 *    Sets a flag to suppress errors in 'import only' mode.
 *
 *  Call:
 *    SetImportOnlyMode Flag
 *
 *  Description:
 *    Sets a flag used by the allocation algorithm to suppress errors when
 *    used in 'import only' mode
 *
 *  Parameters:
 *     Flag   (int) Flag value to be used.
 *-
 */
{
    int value;
    if (argc != 2) {
        interp->result = "Wrong # args";
        return TCL_ERROR;
    }
    value = atoi(argv[1]);
    interp->result = "OK";
    ImportOnlyMode = value;
    ConfigSetImportOnlyMode(value);
    return TCL_OK;
}


/*  ------------------------------------------------------------------------- */

/*
 *                      S t a r t  U p d a t e  { }
 */

static int ConfStartUpdate(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc, char *argv[])

/*+
 *  Command Name:
 *    StartUpdate
 *
 *  Function:
 *    Start an update of the mimic display fibre information
 *
 *  Call:
 *    StartUpdate id pivid
 *
 *  Description:
 *    Called at the start of a mimic display update to get the Sds Ids
 *    of the various components - these are put in global variables
 *    for subsequent use by commands such as GetButton.
 *
 *  Parameters:
 *    (>)  id    (Sds Id)  Sds identifier of the structure
 *    (>)  pivid (Sds Id)  Sds identifier of the pivots structure
 *
 *   Support:
 *     Jeremy Bailey, AAO
 *
 *   Version Date:
 *     6-Oct-1994    
 *-
 */
{
    StatusType status;
    SdsIdType id;
    char mbuf[100];

    if (argc != 3) {
        interp->result = "wrong # args";
        return TCL_ERROR;
    }
    status = SDS__OK;
    id = (SdsIdType) (atol(argv[1]));
    FreeAndFind(id, "objects", &gtopid, &status);
    gpivotid = (SdsIdType) (atol(argv[2]));
    FreeAndFind(gpivotid, "xPiv", &gxPivid, &status);
    FreeAndFind(gpivotid, "yPiv", &gyPivid, &status);
    FreeAndFind(gtopid, "x", &gxid, &status);
    FreeAndFind(gpivotid, "xPark", &gxpid, &status);
    FreeAndFind(gpivotid, "yPark", &gypid, &status);
    FreeAndFind(gpivotid, "tPark", &gthetapid, &status);
    FreeAndFind(gtopid, "y", &gyid, &status);
    FreeAndFind(gtopid, "theta", &gthetaid, &status);
    FreeAndFind(gtopid, "type", &gtype, &status);

    if (status != SDS__OK) {
        MessGetMsg(status, 0, sizeof(mbuf), mbuf);
        ErsFlush(&status);
        sprintf(interp->result, "Error Starting Update\n%s", mbuf);
        return TCL_ERROR;
    } else
        return TCL_OK;
}

/*  ------------------------------------------------------------------------- */

/*
 *                       G e t  B u t t o n  { }
 */

static int ConfGetButton(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc, char *argv[])

/*+
 *
 *  Command name:
 *    GetButton
 *
 *
 *  Function:
 *    Get button information for mimic display
 *  
 *  Call:
 *    GetButton id pivid fibre index zoom button_graphic
 *
 *  Description:
 *    Return the information needed to draw a fibre and its button
 *    on the mimic display.
 *   
 *    This version does not use the id and pivid arguments, but rather
 *    uses ids previously set up with a call to StartUpdate.
 *
 *   Parameters:
 *
 *    (>) id      (Sds Id)  Sds identifier of the structure   
 *    (>) pivid   (Sds Id)  Sds identifier of the pivots structure
 *    (>) fibre   (int)     Fibre number
 *    (>) index   (int)     Index number in structure
 *    (>) zoom    (int)     Zoom factor
 *    (<) button_graphic1 (string)    A tcl variable in which will be placed
 *                 the arguments to a Tk canvas widget create command required
 *                 to draw a graphic of the button.  Only graphics which
 *                 accept a "-fill" option can be used since the Tcl code
 *                 sets colours appropiately depending on the type of object
 *                 and if it has been selected.  For example, this may contain
 *                 something like "polygon x1 y1 x2 y2 x2 y3..."
 *                 or             "oval 10 10 80 80"
 *    (<) button_graphic2 (string) A second such variable, used for the
 *                 same purpose if two graphics are required to represent
 *                 a button.  If not required, will be set to an empty string.
 *
 *   Returns:
 *     A list of 4 values giving the x,y values of the end points of the
 *     fibre line in a form suitable for use by the Tk canvas widget
 *     line command.
 *
 *   Support:
 *     Jeremy Bailey, AAO
 *     Tony Farrell, AAO
 *
 *   Version Date:
 *     16-Mar-2000
 *-
 */
{
    StatusType status;
    long fibre_no;
    long index_no;
    int x, y;                   /* Fibre position, microns              */
    int xa, ya;                 /* Fibre pivot end coordinates, microns */
    unsigned long actlen;
    double zoom;
    double xs, ys;              /* Fibre pivot end coordinates, pixels */
    double xf, yf;              /* Fibre button end coordinates, pixels */
    double theta;               /* Button theta                        */
    char graphic1[500];
    char graphic2[500];
    char mbuf[100];
    short type;                 /* Fibre Type                           */
    double scaling = (double) MimicCircleRadius /
        (double) FpilGetFieldRadiusIncPark(*Instrument);

    if (argc != 8) {
        interp->result = "wrong # args";
        return TCL_ERROR;
    }
    status = SDS__OK;
    fibre_no = atol(argv[3]) - 1;
    index_no = atol(argv[4]) - 1;
    zoom = atof(argv[5]);

/*
 * Validate the fibre number.
 */
    if ((fibre_no < 0) || (fibre_no >= (int) FpilGetNumPivots(*Instrument))) {
        sprintf(interp->result,
            "Illegal Pivot Number %ld\nMust be between 1 and %d",
            fibre_no + 1, FpilGetNumPivots(*Instrument));
        return TCL_ERROR;
    }

/*
 * Get button details
 *     Pivot position (x and y)
 *     Pivot Type
 *     Fibre Position (x, y and theta)
 *     
 */
    SdsGet(gxPivid, 4, fibre_no, &xa, &actlen, &status);
    SdsGet(gyPivid, 4, fibre_no, &ya, &actlen, &status);
    SdsGet(gtype, 2, fibre_no, &type, &actlen, &status);

    SdsGet(gxid, 4, index_no, &x, &actlen, &status);
    SdsGet(gyid, 4, index_no, &y, &actlen, &status);
    SdsGet(gthetaid, 8, index_no, &theta, &actlen, &status);

    if (status != SDS__OK) {
        MessGetMsg(status, 0, 100, mbuf);
        sprintf(interp->result, "Error Getting Fibre Data\n%s", mbuf);
        return TCL_ERROR;
    }

/*
 * Convert pivot position in microns on plate to pixels.
 */
    xs = FieldPlateToDispX(0, xa, zoom);
    ys = FieldPlateToDispY(0, ya, zoom);


    FpilDrawButFib(*Instrument, scaling, zoom,
        FieldPlateToDispX,
        FieldPlateToDispY,
        0,
        x, y, theta, type,
        sizeof(graphic1), &xf, &yf, graphic1, graphic2, &status);

    if (status != STATUS__OK) {
        MessGetMsg(status, 0, 100, mbuf);
        sprintf(interp->result, "Error Drawing Button\n%s", mbuf);
        return TCL_ERROR;
    }
/*
 * Write the line describing the fibre ends to result string.
 */
    sprintf(interp->result, "%f %f %f %f", xs, ys, xf, yf);
/*
 * Set the variables which are to contain the graphics
 */
    Tcl_SetVar(interp, argv[6], graphic1, 0);
    Tcl_SetVar(interp, argv[7], graphic2, 0);

    return TCL_OK;

}

/*  ------------------------------------------------------------------------- */

/*
 *                  M a k e  L i s t i n g  { }
 */

static int ConfMakeListing(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc, char *argv[])

/*+
 *  Command Name:
 *    MakeListing
 *
 *  Function:
 *    Output listing of objects and allocations
 *
 *  Call:
 *    MakeListing id file type
 *
 *  Description:
 *    Output a text file listing of the allocations or objects
 *    or both in an Sds structure. The format of the listing of
 *    unallocated objects is the same as the input format for
 *    the configuration. If the allocations are listed this listing takes
 *    the form of comments in the configuration file so the file will
 *    still be a valid configuration input file. The unalocated objects
 *    listing includes all fiducial stars and sky targets whether allocated
 *    or not.
 *
 *
 *  Parameters:
 *    (>)  id    (Sds Id)  Sds identifier of the structure
 *    (>)  file  (char)    Name of the file to be created
 *    (>)  type  (char)    Information to output
 *                   alloc  =>  List Allocations
 *                   obj    =>  List unallocated objects
 *                   all    =>  List unallocated objects and allocations
 *
 *   Returns:
 *     "Error" - if the file could not be opened.
 *
 *   Support:
 *     Jeremy Bailey, AAO
 *
 *   Version Date:
 *     6-Oct-1994
 *    26-Nov-2001 Now includes allocated sky targets with unallocated 
 *                objects. KS.    
 *    25-Jul-2006  Add the new wlenIndex_id argument for ConfGetIds,
 *                          althrough it is not used. TJF
 *    18-Aug-2009 Added proper motion arguments for ConfGetIds and for
 *                ConfListLine. Header now includes the PROPER_MOTION
 *                item. KS.
 *-
 */
{
    SdsIdType topid, objid, uid, fid;
    StatusType status;
    short alloc_array;
    FILE *listing;
    double mjd;
    int iy, im, id;
    double fd;
    int jstat;
    double ra, dec;
    int rhmsf[4];
    int ddmsf[4];
    char sign[1];
    char buf[80];
    char iname[16];
    SdsCodeType code;
    long ndims;
    unsigned long dims[7];
    SdsIdType ra_id, dec_id, name_id, type_id, spect_id, prio_id, mag_id,
        pid_id, comment_id, wlenIndex_id, pmra_id, pmdec_id, temp_id;
    unsigned int i;
    char mbuf[100];
    char c, s;

    if (argc != 4) {
        interp->result = "wrong # args";
        return TCL_ERROR;
    }
    status = STATUS__OK;
    topid = (SdsIdType) (atol(argv[1]));
    SdsFind(topid, "fieldData", &fid, &status);
    listing = fopen(argv[2], "w");
    if (listing == NULL) {
        ErsOut(0, &status, "Error Opening File %s", argv[2]);
        interp->result = "Error";
        return TCL_OK;
    }
    if (Instrument == NULL) {
        ErsOut(0, &status, "No instrument selected");
        interp->result = "Error";
        return TCL_OK;
    }

/*  Output the Header Section.  Note that from Aug 2009 this includes
 *  the PROPER_MOTIONS entry, indicating that ConfListLine() now outputs
 *  explicit proper motion columns instead of relying on these being
 *  in the pseudo-comments.
 */

    ArgGetString(fid, "label", sizeof(buf), buf, &status);
    fprintf(listing, "LABEL %s\n", buf);
    ArgGetd(fid, "configMjd", &mjd, &status);
    slaDjcl(mjd, &iy, &im, &id, &fd, &jstat);
    fprintf(listing, "UTDATE %4.4d %2.2d %2.2d\n", iy, im, id);
    ArgGetd(fid, "cenRa", &ra, &status);
    ArgGetd(fid, "cenDec", &dec, &status);
    slaDr2tf(3, ra, sign, rhmsf);
    slaDr2af(2, dec, sign, ddmsf);
    fprintf(listing,
        "CENTRE %2.2d %2.2d %2.2d.%3.3d  %c%2.2d %2.2d %2.2d.%2.2d\n",
        rhmsf[0], rhmsf[1], rhmsf[2], rhmsf[3], sign[0], ddmsf[0], ddmsf[1],
        ddmsf[2], ddmsf[3]);
    if (strcmp(FpilGetInstName(*Instrument), "FLAMES")) {
        fprintf(listing, "EQUINOX 2000.0\n");
    }
    fprintf(listing, "PROPER_MOTIONS\n");

    tdFwlensSdsToListingHeader(fid, listing, &status);

    if ((strcmp(argv[3], "alloc") == 0) || (strcmp(argv[3], "all") == 0)) {
        SdsFind(topid, "objects", &objid, &status);
        ConfGetIds(objid, &ra_id, &dec_id, &name_id, &type_id, &spect_id,
                   &prio_id, &mag_id, &pid_id, &comment_id,  
                   &wlenIndex_id, &pmra_id, &pmdec_id, &status);
        SdsInfo(ra_id, iname, &code, &ndims, dims, &status);
        fprintf(listing, "*\n");
        fprintf(listing, "*  Fibre Allocations\n");
        fprintf(listing, "*\n");
        fprintf(listing, "*%5s%12s%12s%12s\n", "Fibre", "Name", "R.A.", "Dec");
        for (i = 0; i < dims[0]; i++) {
            fprintf(listing, "*%4d ", i + 1);
            ConfListLine(listing, i, ra_id, dec_id, name_id, type_id, spect_id,
                         prio_id, mag_id, pid_id, comment_id, wlenIndex_id, 
                         pmra_id, pmdec_id, &status);
            if (status != STATUS__OK)
                break;
        }
        SdsFreeId(name_id, &status);
        SdsFreeId(ra_id, &status);
        SdsFreeId(dec_id, &status);
        SdsFreeId(type_id, &status);
        if (spect_id)
            SdsFreeId(spect_id, &status);
        SdsFreeId(prio_id, &status);
        SdsFreeId(mag_id, &status);
        SdsFreeId(pid_id, &status);
        SdsFreeId(comment_id, &status);
        SdsFreeId(objid, &status);
        if (wlenIndex_id)
            SdsFreeId(wlenIndex_id, &status);
    }
    if ((strcmp(argv[3], "obj") == 0) || (strcmp(argv[3], "all") == 0)) {
        SdsFind(topid, "unallocObject", &uid, &status);
        ConfGetIds(uid, &ra_id, &dec_id, &name_id, &type_id, &spect_id,
            &prio_id, &mag_id, &pid_id, &comment_id,  &wlenIndex_id, 
            &pmra_id, &pmdec_id, &status);
        SdsInfo(ra_id, iname, &code, &ndims, dims, &status);
        fprintf(listing, "*\n");
        fprintf(listing, "*  Unallocated Objects\n");
        fprintf(listing, "*\n");
        fprintf(listing, "*%12s%12s%12s\n", "Name", "R.A.", "Dec");
        alloc_array = TRUE;
        SdsFind (uid,"allocated",&temp_id,&status);
        if (status != 0) {
            alloc_array = FALSE;
            status = 0;
        } else {
            SdsFreeId (temp_id, &status);
        }
        for (i = 0; i < dims[0]; i++) {
            if (alloc_array) {
                ConfArrayGetc(uid, "allocated", i, &c, &status);
            } else {
                c = 0;
            }
            if (c == 0) {
                ConfListLine(listing, i, ra_id, dec_id, name_id, type_id,
                             spect_id, prio_id, mag_id, pid_id, comment_id, 
                             wlenIndex_id, pmra_id, pmdec_id, &status);
            }
            if (status != STATUS__OK)
                break;
        }
        SdsFreeId(name_id, &status);
        SdsFreeId(ra_id, &status);
        SdsFreeId(dec_id, &status);
        SdsFreeId(type_id, &status);
        if (spect_id)
            SdsFreeId(spect_id, &status);
        SdsFreeId(prio_id, &status);
        SdsFreeId(mag_id, &status);
        SdsFreeId(pid_id, &status);
        SdsFreeId(comment_id, &status);
        SdsFreeId(uid, &status);
        if (wlenIndex_id)
            SdsFreeId(wlenIndex_id, &status);
    }
    if ((strcmp(argv[3], "obj") == 0) || (strcmp(argv[3], "all") == 0)) {
        SdsFind(topid, "unallocGuide", &uid, &status);
        ConfGetIds(uid, &ra_id, &dec_id, &name_id, &type_id, &spect_id,
                   &prio_id, &mag_id, &pid_id, &comment_id, &wlenIndex_id,
                   &pmra_id, &pmdec_id, &status);
        SdsInfo(ra_id, iname, &code, &ndims, dims, &status);
        fprintf(listing, "*\n");
        fprintf(listing, "*  Fiducial Stars\n");
        fprintf(listing, "*\n");
        fprintf(listing, "*%12s%12s%12s\n", "Name", "R.A.", "Dec");
        for (i = 0; i < dims[0]; i++) {
            ConfListLine(listing, i, ra_id, dec_id, name_id, type_id, spect_id,
                         prio_id, mag_id, pid_id, comment_id, wlenIndex_id,
                         pmra_id, pmdec_id, &status);
            if (status != STATUS__OK)
                break;
        }
        SdsFreeId(name_id, &status); 
        SdsFreeId(ra_id, &status);
        SdsFreeId(dec_id, &status);
        SdsFreeId(type_id, &status);
        if (spect_id)
            SdsFreeId(spect_id, &status);
        SdsFreeId(prio_id, &status);
        SdsFreeId(mag_id, &status);
        SdsFreeId(pid_id, &status);
        SdsFreeId(comment_id, &status);
        if (wlenIndex_id)
            SdsFreeId(wlenIndex_id, &status);
        SdsFreeId(uid, &status);

/*  We also need the fiducials and skies from the allocated objects section  */

        SdsFind(topid, "objects", &objid, &status);
        ConfGetIds(objid, &ra_id, &dec_id, &name_id, &type_id, &spect_id,
                   &prio_id, &mag_id, &pid_id, &comment_id, &wlenIndex_id, 
                   &pmra_id, &pmdec_id, &status);
        SdsInfo(ra_id, iname, &code, &ndims, dims, &status);
        for (i = 0; i < dims[0]; i++) {
            ConfArrayGetc(objid, "type", i, &c, &status);
            ConfArrayGetc(objid, "spectrograph", i, &s, &status);
            if (FpilIsTargetGuide(*Instrument, c) ||
                                     FpilIsTargetSky(*Instrument, c)) {
                ConfListLine(listing, i, ra_id, dec_id, name_id, type_id,
                             spect_id, prio_id, mag_id, pid_id, comment_id, 
                             wlenIndex_id, pmra_id, pmdec_id, &status);
            }
            if (status != STATUS__OK)
                break;
        }
        SdsFreeId(name_id, &status);
        SdsFreeId(ra_id, &status);
        SdsFreeId(dec_id, &status);
        SdsFreeId(type_id, &status);
        if (spect_id)
            SdsFreeId(spect_id, &status);
        SdsFreeId(prio_id, &status);
        SdsFreeId(mag_id, &status);
        SdsFreeId(pid_id, &status);
        SdsFreeId(comment_id, &status);
        if (wlenIndex_id)
            SdsFreeId(wlenIndex_id, &status);
        SdsFreeId(objid, &status);
    }

    SdsFreeId(fid, &status);

    fclose(listing);
    if (status == STATUS__OK) {
        return TCL_OK;
    } else {
        MessGetMsg(status, 0, 100, mbuf);
        sprintf(interp->result, "Error in MakeListing\n%s", mbuf);
        return TCL_ERROR;
    }
}

/*  ------------------------------------------------------------------------- */

/*
 *                  S k y c a t   L i s t i n g  { }
 */

static int ConfSkycatListing(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc, char *argv[])

/*+
 *  Command Name:
 *    SkycatListing
 *
 *  Function:
 *    Output listing of allocated objects in skycat format
 *
 *  Call:
 *    SkycatListing id file
 *
 *  Description:
 *    Output a listing in skycat catalogue format of the allocated objects. 
 *
 *  Parameters:
 *    (>)  id    (Sds Id)  Sds identifier of the structure
 *    (>)  file  (char)    Name of the file to be created
 *
 *   Returns:
 *     "Error" - if the file could not be opened.
 *
 *   Support:
 *     Keith Shortridge, AAO
 *
 *   Version Date:
 *    29-Jan-2001 Original version, KS.
 *    25-Jul-2006  Add the new wlenIndex_id argument for ConfGetIds,
 *                          althrough it is not used. TJF
 *    10-Aug-2009 Added proper motion arguments to ConfGetIds() call. 
 *                At the moment these are not used. KS.
 *-
 */
{
    SdsIdType topid, objid, fid;
    StatusType status;
    FILE *listing;
    double mjd;
    int iy, im, id;
    double fd;
    int jstat;
    double ra, dec;
    int rhmsf[4];
    int ddmsf[4];
    char sign[1];
    char buf[80];
    char iname[16];
    SdsCodeType code;
    long ndims;
    unsigned long dims[7];
    SdsIdType ra_id, dec_id, name_id, type_id, spect_id, prio_id, mag_id,
        pid_id, comment_id, wlenIndex_id, pmra_id, pmdec_id;
    unsigned int i;
    char mbuf[100];
    unsigned long actlen;
    char typeString[32];
    char typeText[80];
    char name[NAMELEN];
    char otype;
    char ospect;
    char UnallocType;

    if (argc != 3) {
        interp->result = "wrong # args";
        return TCL_ERROR;
    }
    status = STATUS__OK;
    topid = (SdsIdType) (atol(argv[1]));
    SdsFind(topid, "fieldData", &fid, &status);
    listing = fopen(argv[2], "w");
    if (listing == NULL) {
        ErsOut(0, &status, "Error Opening File %s", argv[2]);
        interp->result = "Error";
        return TCL_OK;
    }
    if (Instrument == NULL) {
        ErsOut(0, &status, "No instrument selected");
        interp->result = "Error";
        return TCL_OK;
    }

    /*  Output the Header Section. First, just the header information
     *  included in the original input file about the field itself
     */

    ArgGetString(fid, "label", sizeof(buf), buf, &status);
    fprintf(listing, "#LABEL %s\n", buf);
    ArgGetd(fid, "configMjd", &mjd, &status);
    slaDjcl(mjd, &iy, &im, &id, &fd, &jstat);
    fprintf(listing, "#UTDATE %4.4d %2.2d %2.2d\n", iy, im, id);
    ArgGetd(fid, "cenRa", &ra, &status);
    ArgGetd(fid, "cenDec", &dec, &status);
    slaDr2tf(3, ra, sign, rhmsf);
    slaDr2af(2, dec, sign, ddmsf);
    fprintf(listing,
        "#CENTRE %2.2d %2.2d %2.2d.%3.3d  %c%2.2d %2.2d %2.2d.%2.2d\n",
        rhmsf[0], rhmsf[1], rhmsf[2], rhmsf[3], sign[0], ddmsf[0], ddmsf[1],
        ddmsf[2], ddmsf[3]);
        
    /*  Now the catalogue description information that skycat will use
     *  to plot the data. The important bit here is the symbol definition
     *  section, because this indicates how the targets will be displayed -
     *  and is probably the bit most people will have aesthetic opinions
     *  about!
     */
     
    fprintf(listing,"# Skycat configuration information\n");
    fprintf(listing,"serv_type: local\n");
    fprintf(listing,"long_name: %s\n",argv[2]);
    fprintf(listing,"short_name: %s\n",argv[2]);
    fprintf(listing,"url: %s\n",argv[2]);
    fprintf(listing,"%s%s\n","symbol: {Sky Ref} {circle white {} {} {} ",
                    "{$Sky==0 && $Ref==0}} {8 {}} \\" );
    fprintf(listing,"%s\n",
                    ": {Sky} {square white {} {} {} {$Sky==1}} {8 {}} \\");
    fprintf(listing,"%s\n",
                    ": {Ref} {diamond white {} {} {} {$Ref==1}} {8 {}}");
    fprintf(listing,"# End Skycat configuration information\n");
    
    /*  Now the column headings used by Skycat. Note that the Ref and Sky 
     *  column values are used in the conditions in the symbol descriptions
     *  in the configuration information that's just been written to control
     *  the way targets of different types are displayed.
     */

    fprintf(listing,"ID\tRA\tDEC\tSky\tRef\tType\n");
    fprintf(listing,"--\t--\t---\t---\t---\t----\n");

    /*  The first entry in the catalogue is the field center - this is not
     *  a real target, of course, but having it first will cause skycat to
     *  use this as the center of its field. Note that ra and dec still
     *  contain its coordinates, albeit in radians.
     */
     
    ra *= 180.0 / DPI;
    dec *= 180.0 / DPI;
    fprintf(listing,"%s\t%f\t%f\t%d\t%d\t%s\n",
                  "Field center",ra,dec,0,0,"Field center");
       
    /*  Now, we work through all the allocated objects, and output their
     *  details to the catalogue.
     */
     
    UnallocType = FpilUnallocatedType(*Instrument);
    SdsFind(topid, "objects", &objid, &status);
    ConfGetIds(objid, &ra_id, &dec_id, &name_id, &type_id, &spect_id,
               &prio_id, &mag_id, &pid_id, &comment_id, &wlenIndex_id, 
               &pmra_id, &pmdec_id, &status);
    SdsInfo(ra_id, iname, &code, &ndims, dims, &status);
    for (i = 0; i < dims[0]; i++) {
        SdsGet(name_id, NAMELEN, NAMELEN * i, name, &actlen, &status);
        SdsGet(ra_id, sizeof(double), i, &ra, &actlen, &status);
        SdsGet(dec_id, sizeof(double), i, &dec, &actlen, &status);
        ra *= 180.0 / DPI;
        dec *= 180.0 / DPI;
        SdsGet(type_id, sizeof(char), i, &otype, &actlen, &status);
        if (spect_id) {
            SdsGet(type_id, sizeof(char), i, &ospect, &actlen, &status);
        } else {
            ospect = 0;
        }
        FpilDecodeTargetType(*Instrument, otype, ospect, TRUE, -1, typeString,
                                                               typeText);
        if (status != STATUS__OK) break;
        if (otype != UnallocType) {
            fprintf(listing,"%s\t%f\t%f\t%d\t%d\t%s\n",
                  name,ra,dec,
                  FpilIsTargetSky(*Instrument,otype),
                  (FpilIsTargetGuide(*Instrument,otype) &&
                  !FpilIsTargetTelGuide(*Instrument,otype)),typeText);
        }
    }
    SdsFreeId(name_id, &status);
    SdsFreeId(ra_id, &status);
    SdsFreeId(dec_id, &status);
    SdsFreeId(type_id, &status);
    if (spect_id) SdsFreeId(spect_id, &status);
    SdsFreeId(prio_id, &status);
    SdsFreeId(mag_id, &status);
    SdsFreeId(pid_id, &status);
    SdsFreeId(comment_id, &status);
    if (wlenIndex_id)
        SdsFreeId(wlenIndex_id, &status);
    SdsFreeId(objid, &status);

    SdsFreeId(fid, &status);

    fclose(listing);
    if (status == STATUS__OK) {
        return TCL_OK;
    } else {
        MessGetMsg(status, 0, 100, mbuf);
        sprintf(interp->result, "Error in SkycatListing\n%s", mbuf);
        return TCL_ERROR;
    }
}

/*  ------------------------------------------------------------------------- */

/*
 *                        W r i t e  P A F  { }
 */

static int ConfWritePAF(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc, char *argv[])

/*+
 *
 *  Tcl Command Name:
 *    WritePAF
 *
 *  Function:
 *    Output PAF file listing of objects and allocations
 *
 *  Call:
 *    WritePAF id file selGuide guideName posneg mode expert
 *
 *  Description:
 *    Output a text file listing of the allocations or objects
 *    or both in an Sds structure. The format of the listing of
 *    unallocated objects is the standard PAF file format defined for
 *    the FLAMES instrument. It includes field parameters, the VLT
 *    guide stars, the allocated and unallocated targets, and the
 *    FACB guide stars.
 *
 *  Parameters:
 *    (>)  id       (Sds Id)  Sds identifier of the top level structure
 *    (>)  file     (char)    Name of the file to be created
 *    (>)  selGuide (int)     Zero if no VLT guide star has been selected, 
 *                            positive if one has.
 *    (>)  guideName(char)    Name of selected VLT guide star, if any.
 *    (>)  posneg   (char)    Orientation for guide pivot, if a guide
 *                            object has been selected. Must be "POS" or "NEG".
 *    (>)  mode     (char)    Keyword describing the current fibre combination.
 *    (>)  expert   (int)     Non-zero if in expert mode. (Unallocated targets
 *                            are only written to the PAF file in expert mode.)
 *
 *   Returns:
 *     "Error" - if the file could not be opened.
 *
 *   Support:
 *     Keith Shortridge, AAO
 *
 *   History:
 *     25th Sept 2000  KS  Original version
 *     18th Aug  2001  KS  Added ARGUS keywords
 *     22nd Aug  2001  KS  Added use of expert mode to control listing of
 *                         unallocated targets.
 *     10th Dec  2001  KS  Blanks in fieldName are now changed to underscores.
 *      3rd Jan  2002  KS  Now includes INS.OSS.VER keyword. 
 *      5th Feb  2002  KS  Now supports a PAF checksum. Variables commented.
 *      4th Jun  2002  KS  Format of PAF.ID and PAF.NAME changed accoring to
 *                         suggestion by Andreas Kaufer (ESO).
 *      1st Aug  2002  KS  Modified for new use of mode string, as requested
 *                         by Paranal.
 *      2nd Aug  2002  KS  Null mode strings now represented by '-'.
 *     21st Aug  2002  KS  TEL.TARG.NAME added.
 *     22nd Aug  2002  TJF SKYANGLE now calculated properly.
 *      5th Sep  2002  KS  Guide star now identified by name rather than by
 *                         x and y position.
 *     25th Jul  2006  TJF Add the new wlenIndex_id argument for ConfGetIds,
 *                          althrough it is not used.
 *     10th Aug 2009   KS  Added proper motion arguments to ConfGetIds call.
 *-
 */
{
    char* abbrNames[10];           /* Abbreviated PAF type names */
    SdsIdType topid;               /* Sds Id of top structure */
    SdsIdType objid;               /* Sds Id of objects structure */
    SdsIdType uid;                 /* Sds Id of unalloc guide structure */
    SdsIdType fid;                 /* Sds Id of field data structure */
    StatusType status;             /* Inherited status variable */
    FILE *listing;                 /* Used for the output PAF file */
    double mjd;                    /* Modified julian date */
    double ra;                     /* Central RA - mean */
    double dec;                    /* Central Dec - mean */
    int rhmsf[4];                  /* RA in hours minutes seconds */
    int ddmsf[4];                  /* Dec in degrees minutes seconds */
    char sign[1];                  /* Sign for RA and Dec */
    char iname[16];                /* Name of SDS item - ignored */
    SdsCodeType code;              /* SDS code type for an item */
    long ndims;                    /* Number of dimensions for SDS array */
    unsigned long dims[7];         /* Actual dimensions for SDS array */
    SdsIdType ra_id;               /* SDS id for RA array in structure */
    SdsIdType dec_id;              /* SDS id for Dec array in structure */
    SdsIdType name_id;             /* SDS id for Name array in structure */
    SdsIdType type_id;             /* SDS id for type array in structure */
    SdsIdType spect_id;            /* SDS id for spect array in structure */
    SdsIdType prio_id;             /* SDS id for priority array in structure */
    SdsIdType mag_id;              /* SDS id for magnitude array in structure */
    SdsIdType pid_id;              /* SDS id for pid array in structure */
    SdsIdType theta_id;            /* SDS id for theta array in structure */
    SdsIdType comment_id;          /* SDS id for comment array in structure */
    SdsIdType wlenIndex_id;        /* SDS id for wlenIndex array in structure */
    SdsIdType pmra_id;             /* SDS id for pmra array in structure */
    SdsIdType pmdec_id;            /* SDS id for pmdec array in structure */
    unsigned int i;                /* General index variable */
    int IGuide;                    /* Number of VLT guide targets */
    int unallocFACBTargets;        /* Number of unallocated reference stars */
    int Fibre;                     /* Allocated fibre for target -1 => none */
    char Prefix[16];               /* Prefix for PAF keywords for an object */
    char mbuf[100];                /* Error message buffer */
    char c;                        /* Single character used for target type */
    char s;                        /* Single char used for spectrograph */
    int numTypes;                  /* Number of different PAF types */
    char* typeNames[10];           /* Full names of PAF types */
    int iType;                     /* Index through types */
    int iFibre;                    /* Index through fibres */
    int fibreType;                 /* Fibre type code for a given fibre */
    short match;                   /* True if fibre type matches */
    int iFibType;                  /* Index through possible fibre types */
    int numTypeCodes;              /* Number of PAF fibre codes */
    int fibreTypeCodes[10];        /* PAF fibre codes; see FpilPAFFibreCodes()*/
    char insCode;                  /* Single letter for instrument */
    char unallocType;              /* Character used as "unallocated" code */
    int xusedCount;                /* Total unused targets */
    int usedCount;                 /* Total used non-sky targets */
    int totalUsedCount;            /* Total number of allocated targets */
    int skyCount;                  /* Total number of sky targets */
    int usedSkyCount;              /* Total number of allocated sky targets */
    int xTargetCount;              /* Total of unallocated targets */
    char fieldName[128];           /* Name associated with field */
    char timeString[32];           /* Formatted current time, ESO standard */
    char HhMmSs[8];                /* Formatted hours minutes secs, hhmmss. */
    struct tm* gmTimePtr;          /* Used to get current time */
    int selGuide;                  /* Value of selGuide parameter */
    int PAF_Flags;                 /* Used to pass info to ConfListPAFTarget */
    long nUnallocObj;              /* Number of unallocated objects */
    long nUnallocSky;              /* Number of unallocated sky targets */
    int nEntry;                    /* Entry count - used in PAF keyword names */
    double appEpoch;               /* Apparent epoch */
    double appRa;                  /* Field center RA -apparent */
    double appDec;                 /* Field center Dec -apparent */
    short InUse;                   /* True if ARGUS is being used */
    double Angle;                  /* ARGUS angle, if used */
    double Offset;                 /* ARGUS offset, if used */
    char Scale[32];                /* ARGUS scale, if used */
    short ExpertMode;              /* True if program is in expert mode */
    SdsIdType updateID  = 0;       /* ID for Update time in instrument file */
    StatusType localStatus;        /* Inherited status variable */
    int modTime;                   /* Instrument file modification time */
    unsigned long actlen;          /* Actual length of SDS item */
    struct tm* tmPtr;              /* Used to convert modification time */
    int length;                    /* Number of characters in string */      
    char Version[32];              /* RCS version number for FPOSS */
    short ChecksumEnabled;         /* True if checksum is to be included */
    int Blanks;                    /* Number of blanks to align checksum */
    int ChecksumOffset;            /* File offset for checksum string */
    char ModeString[64];           /* Local copy of mode argument */
    char* FileMode;                /* File name component of mode argument */
    char* InsMode;                 /* INS.MODE component of mode */
    char* GirafMode;               /* INS.GIRAF.MODE component of mode */
    char* UvesSlit;                /* INS.UVES.SLIT component of mode */
   
    /*  This routine is a bit of an oddity, having been introduced purely
     *  for the FLAMES instrument used at ESO. ESO like all output files
     *  from programs such as this to be in PAF format, and we're happy
     *  to comply. This means, however, that although this routine may
     *  look like a fairly general one - and there is no reason why someone
     *  shouldn't write out a 2dF configuration using this - there are
     *  places where it has definite ESO-isms. For example, at ESO the
     *  configure program is known as FP OSS, and that appears in the
     *  header.
     */
    
    /*  Initial values */
    
    localStatus = STATUS__OK;
    ChecksumEnabled =  USE_CHECKSUM;
    ChecksumOffset = 0;
      
    /*  Initial checks. Make sure we've been invoked properly */
    
    if (argc != 8) {
        interp->result = "wrong # args";
        return TCL_ERROR;
    }
    
    /*  See if we are in expert mode or not */
    
    ExpertMode = (atoi(argv[7]) != 0);
    
    /*  We need to know the unallocated type code used in the 'objects'
     *  structure for this instrument.
     */
     
    unallocType = FpilUnallocatedType(*Instrument);
    
    /*  Make sure there's an instrument selected */
    
    if (Instrument == NULL) {
        ErsOut(0, &status, "No instrument selected");
        interp->result = "Error";
        return TCL_OK;
    }
    
    /*  Open the file specified in the second argument */
    
    listing = fopen(argv[2], "w");
    if (listing == NULL) {
        ErsOut(0, &status, "Error Opening File %s", argv[2]);
        interp->result = "Error";
        return TCL_OK;
    }
    
    /*  Access the field data in the structure, using the SDS ID passed
     *  in the first argument. Get the field name, which we need for the
     *  PAF header.
     */
    
    status = STATUS__OK;
    topid = (SdsIdType) (atol(argv[1]));
    SdsFind(topid, "fieldData", &fid, &status); 
    ArgGetString(fid, "label", sizeof(fieldName), fieldName, &status);
    ArgGeti(fid,"unallocObj",&nUnallocObj,&status);
    ArgGeti(fid,"unallocSky",&nUnallocSky,&status);
    
    /*  Get the date and time, and format those in the ESO standard way. Note
     *  that we get these from the global FileTime, which should have been
     *  set by PAFFileName, which will have been used to generate the file
     *  name. This ensures that we have a common view of the time in both
     *  the file name and its contents.
     */
    
    gmTimePtr = gmtime(&FileTime);
    sprintf (timeString,"%4.4d-%2.2d-%2.2dT%2.2d:%2.2d:%2.2d",
         gmTimePtr->tm_year + 1900,gmTimePtr->tm_mon + 1,gmTimePtr->tm_mday,
         gmTimePtr->tm_hour,gmTimePtr->tm_min,gmTimePtr->tm_sec);
    sprintf (HhMmSs,"%2.2d%2.2d%2.2d",
         gmTimePtr->tm_hour,gmTimePtr->tm_min,gmTimePtr->tm_sec);
         
    /*  The strings written into PAF.ID and PAF.NAME must not contain
     *  spaces. So we have to pass through fieldName changing spaces etc to
     *  underscores. timeString won't contain spaces. Nor will the mode.
     *  we
     */
   
    length = strlen(fieldName);
    for (i = 0; i < length; i++) {
        char TheChar = fieldName[i];
        if ((TheChar < 'a') || (TheChar > 'z')) {
            if ((TheChar < 'A') || (TheChar > 'Z')) {
                if ((TheChar < '0') || (TheChar > '9')) {
                    fieldName[i] = '_';
                }
            }
        }
    }
    
    /*  The 'mode' string - which is generated by FpilFibreCombos() - has
     *  four separate slash-separated parts. We need to take these apart.
     *  Note that strtok() modifies its search string, so we take a local
     *  copy in ModeString.
     */
    
    FileMode = InsMode = GirafMode = UvesSlit = (char*) NULL;
    strncpy (ModeString,argv[6],sizeof(ModeString));
    ModeString[sizeof(ModeString) - 1] = '\0';
    FileMode = strtok(ModeString,"/");
    InsMode = strtok((char*)NULL,"/");
    GirafMode = strtok((char*)NULL,"/");
    UvesSlit = strtok((char*)NULL,"/");
    if (FileMode == (char*) NULL) FileMode = "";
    if (InsMode == (char*) NULL) InsMode = "";
    if (GirafMode == (char*) NULL) GirafMode = "";
    if (UvesSlit == (char*) NULL) UvesSlit = "";
    
    /*  Insert a warning at the start of the file */
    
    fprintf (listing,
               "#\n# This file has been generated by the FPOSS program.\n");
    fprintf (listing,
               "#\n# Do NOT attempt to edit it by hand.\n");
    fprintf (listing,
                  "# Modified files will almost certainly be rejected\n");
    fprintf (listing,
                  "# by the observing system.\n\n");
    
    /*  Write out the standard PAF header. 
     */
    
    fprintf (listing,"# PAF Header;\n");
    fprintf (listing,"PAF.HDR.START;\n");
    fprintf (listing,"PAF.TYPE          \"Paramfile\";\n");
    fprintf (listing,"PAF.ID            \"%s.%s.%s\";\n",fieldName,FileMode,
                                                                     HhMmSs);
    fprintf (listing,"PAF.NAME          \"%s.%s.%s\";\n",fieldName,FileMode,
                                                                     HhMmSs);
    fprintf (listing,"PAF.DESC          \"Output from FP OSS\";\n");
    fprintf (listing,"PAF.CRTE.NAME     \"FP OSS\";\n");
    fprintf (listing,"PAF.CRTE.DAYTIM   \"%s\";\n",timeString);
    fprintf (listing,"PAF.LCHG.NAME     \"\";\n");
    fprintf (listing,"PAF.LCHG.DAYTIM   \"\";\n");
    fprintf (listing,"PAF.CHCK.NAME     \"\";\n");
    fprintf (listing,"PAF.CHCK.DAYTIM   \"\";\n");
    
    /*  We treat the checksum line differently, depending on whether or not
     *  we are going to include a checksum. If we aren't, we just write a
     *  null entry which the reading software will ignore. If we are including
     *  a proper checksum, we write a zero value at this point and record
     *  its location in the file so that we can overwrite it with the
     *  calculated calue later. The checksum string has to start on a 4-byte
     *  boundary in the file, so we insert spaces as needed.
     */
     
    if (ChecksumEnabled) {
        fprintf (listing,"PAF.CHCK.CHECKSUM ");
        ChecksumOffset = ftell(listing) + 1;
        Blanks = 4 - (ChecksumOffset % 4);
        if (Blanks) {
            for (i = 0; i < Blanks; i++) fprintf (listing," ");
            ChecksumOffset += Blanks;
        }
        fprintf (listing,"\"0000000000000000\";\n");
    } else {
        fprintf (listing,"PAF.CHCK.CHECKSUM \"\";\n");
    }
    fprintf (listing,"PAF.HDR.END;\n");
    
    fprintf (listing,"\n");
    fprintf (listing,"TPL.FILE.DIRNAME  \"$INS_ROOT/SYSTEM/MISC\";\n");
    
    /*  Now write the various keywords that describe the field in general.
     *  If INS.ARG.USED is true, we also need INS.ARGS.ANGLE (0 - 360 deg)
     *  and INS.ARGS.SCALE (a string set to either "1:1.67" or "1:1"). The
     *  angle written to the PAF file is the same is the angle returned
     *  by ConfCGetArgusData - the same as the angle passed to the 
     *  astrometric code as the first telescope model parameter. (Earlier
     *  versions of this code subtracted the offset first, but this is
     *  not correct, given the way the first telescope model parameter happens
     *  to have been defined.)
     */
         
    fprintf (listing,"\n");
    fprintf (listing,"#Primary Keywords\n");
    fprintf (listing,"INS.EQUINOX       2000.0;\n");
    
    RCSVersion(Version,sizeof(Version));
    fprintf (listing,"INS.OSS.VER       \"%s\";\n",Version);
       
    ConfCGetArgusData (&InUse,&Angle,Scale,&Offset);
    if (InUse) {
       fprintf (listing,"INS.ARGS.USED     T;\n");
       fprintf (listing,"INS.ARGS.ANGLE    %f;\n",Angle);
       fprintf (listing,"INS.ARGS.SCALE    \"%s\";\n",Scale);
    } else {
       fprintf (listing,"INS.ARGS.USED     F;\n");
    }
    fprintf (listing,"\n");
    fprintf (listing,"#Telescope Category\n");
    
    ArgGetd(fid, "appRa", &appRa, &status);
    ArgGetd(fid, "appDec", &appDec, &status);
    ArgGetd(fid, "configMjd", &mjd, &status);
    ArgGetd(fid, "appEpoch", &appEpoch, &status);
    ArgGetd(fid, "cenRa", &ra, &status);
    ArgGetd(fid, "cenDec", &dec, &status);
    slaDr2tf(3, ra, sign, rhmsf);
    slaDr2af(2, dec, sign, ddmsf);
    
    /*  We duplicate the RA and DEC in both TEL and INS. This is
     *  deliberate, because the Telescope system needs this and so does
     *  the positioner software, and the easiest thing is to have them
     *  in twice.
     */
     
    fprintf (listing,"TEL.TARG.NAME     \"%s\";\n",fieldName);
    fprintf (listing,"TEL.TARG.RA       %2.2d%2.2d%2.2d.%2.2d;\n",
                    rhmsf[0], rhmsf[1], rhmsf[2], rhmsf[3] / 10);
    if (sign[0] == '-') {
       fprintf (listing,"TEL.TARG.DEC      -%2.2d%2.2d%2.2d.%1.1d;\n",
                    ddmsf[0], ddmsf[1], ddmsf[2], ddmsf[3] / 10);
    } else {
       fprintf (listing,"TEL.TARG.DEC      %2.2d%2.2d%2.2d.%1.1d;\n",
                    ddmsf[0], ddmsf[1], ddmsf[2], ddmsf[3] / 10);
    }
    fprintf (listing,"INS.TARG.RA       %2.2d%2.2d%2.2d.%2.2d;\n",
                    rhmsf[0], rhmsf[1], rhmsf[2], rhmsf[3] / 10);
    if (sign[0] == '-') {
       fprintf (listing,"INS.TARG.DEC      -%2.2d%2.2d%2.2d.%1.1d;\n",
                    ddmsf[0], ddmsf[1], ddmsf[2], ddmsf[3] / 10);
    } else {
       fprintf (listing,"INS.TARG.DEC      %2.2d%2.2d%2.2d.%1.1d;\n",
                    ddmsf[0], ddmsf[1], ddmsf[2], ddmsf[3] / 10);
    }
    
    /*  If we read the PAF file back into configure, we need the
     *  APPRA, APPDEC and MJD and EPOCH - actually, we need to think
     *  carefully about the way these are used, but they're in the
     *  structure so it makes sense to have them. Since they aren't
     *  in the FPOSS dictionary at the moment, we put them out as
     *  pseudo-comments.
     */
     
    slaDr2tf(3, appRa, sign, rhmsf);
    slaDr2af(2, appDec, sign, ddmsf);
    fprintf (listing,"#!INS.TARG.APPRA  %2.2d%2.2d%2.2d.%2.2d;\n",
                    rhmsf[0], rhmsf[1], rhmsf[2], rhmsf[3] / 10);
    if (sign[0] == '-') {
       fprintf (listing,"#!INS.TARG.APPDEC -%2.2d%2.2d%2.2d.%1.1d;\n",
                    ddmsf[0], ddmsf[1], ddmsf[2], ddmsf[3] / 10);
    } else {
       fprintf (listing,"#!INS.TARG.APPDEC %2.2d%2.2d%2.2d.%1.1d;\n",
                    ddmsf[0], ddmsf[1], ddmsf[2], ddmsf[3] / 10);
    }
    fprintf (listing,"#!INS.TARG.MJD    %f;\n",mjd);
    fprintf (listing,"#!INS.TARG.APPEPOCH  %f;\n",appEpoch);
    
    /*  Now the time stamp for the instrument status file. We look for
     *  an UPDATE_TIME item in the pivot details structure. If this is
     *  FLAMES, then such an item should be present.
     */
    
    localStatus = STATUS__OK;
    modTime = 0;
    SdsFind(InstFieldDetails, "UPDATE_TIME", &updateID, &localStatus);
    SdsGet(updateID,sizeof(int),0,&modTime,&actlen,&localStatus);
    SdsFreeId(updateID,&localStatus);
    tmPtr = gmtime((time_t*)&modTime);
    
    fprintf (listing,"\n");
    fprintf (listing,"#Instrument status file time stamp\n");
    fprintf (listing,"INS.STAT.SFSTAMP  \"%4d-%02d-%02dT%02d:%02d:%02d\";\n",
               1900 + tmPtr->tm_year, 1 + tmPtr->tm_mon,tmPtr->tm_mday,
                              tmPtr->tm_hour,tmPtr->tm_min,tmPtr->tm_sec);
    
    /*  We need to set the keywords for the current fibre combination mode */
    
    fprintf (listing,"INS.MODE          \"%s\";\n",InsMode);
    if ((GirafMode[0] != '\0') && (GirafMode[0] != '-')) {
        fprintf (listing,"INS.GIRAF.MODE    \"%s\";\n",GirafMode);
    }
    if ((UvesSlit[0] != '\0') && (UvesSlit[0] != '-')) {
        fprintf (listing,"INS.UVES.SLIT     \"%s\";\n",UvesSlit);
    }
        
    /*  The potential guide stars refer to those targets that can be used
     *  as VLT guide stars. These cannot be allocated, so will be found
     *  in the unallocated guide section of the structure. We search that
     *  structure for them. Note that one of these may have been selected
     *  and the guide probe position for it may also have been selected. In
     *  this case, the arguments used to invoke this command show this, and
     *  we can create an entry for the actual guide star to be used.
     */

    SdsFind(topid, "unallocGuide", &uid, &status);
    ConfGetIds(uid, &ra_id, &dec_id, &name_id, &type_id, &spect_id,
               &prio_id, &mag_id, &pid_id, &comment_id,  &wlenIndex_id, 
               &pmra_id, &pmdec_id, &status);
    SdsInfo(ra_id, iname, &code, &ndims, dims, &status);
    
    /*  The selected guide star, if there is one. We identify it by
     *  its name, as passed.
     */
    
    selGuide = atoi(argv[3]);
    if (selGuide > 0) {
        char* guideName = argv[4];
        char name[NAMELEN];
        fprintf (listing,"\n");
        fprintf (listing,"# Selected Instrument Guide Star\n");
        for (i = 0; i < dims[0]; i++) {
            ConfArrayGetc(uid, "type", i, &c, &status);
            if (FpilIsTargetTelGuide(*Instrument, c)) {
                SdsGet(name_id, NAMELEN, NAMELEN * i, name, &actlen, &status);
                if (!strcmp(name,guideName)) {
                    Fibre = -1;
                    PAF_Flags = PAF_NO_PRIO;
                    ConfListPAFTarget(listing, i, "TEL.GS.", ra_id, 
                        dec_id, name_id, type_id, spect_id, prio_id, mag_id,
                        pid_id, pmra_id, pmdec_id, comment_id, Fibre, 
                        PAF_Flags, &status);
                    fprintf (listing,"TEL.GS.PPOS            \"%s\";\n",
                                                                   argv[5]);
                    break;
                }
            }
        }
    }
    
    /*  And now all the others, the potential guide stars - note that these
     *  include the actual guide star selected, if there is one.
     *  As we go through, we note how many non-VLT guide targets there are
     *  in the unallocated guide section, since we need that later.
     */
     
    fprintf (listing,"\n");
    fprintf (listing,"# Instrument Potential Guide Stars\n");    
    IGuide = 0;
    unallocFACBTargets = 0;
    for (i = 0; i < dims[0]; i++) {
        ConfArrayGetc(uid, "type", i, &c, &status);
        if (FpilIsTargetTelGuide(*Instrument, c)) {
            IGuide++;
            Fibre = -1;
            sprintf (Prefix,"TEL.PGS%d.",IGuide);
            PAF_Flags = 0;
            ConfListPAFTarget(listing, i, Prefix, ra_id, dec_id, name_id,
                    type_id, spect_id, prio_id, mag_id, pid_id, 
                    pmra_id, pmdec_id, comment_id,
                    Fibre, PAF_Flags, &status);
        } else {
           unallocFACBTargets++;
        }
        if (status != STATUS__OK) break;
    }
    SdsFreeId(name_id, &status);
    SdsFreeId(ra_id, &status);
    SdsFreeId(dec_id, &status);
    SdsFreeId(type_id, &status);
    if (spect_id) {
        SdsFreeId(spect_id, &status);
    }
    SdsFreeId(prio_id, &status);
    SdsFreeId(mag_id, &status);
    SdsFreeId(pid_id, &status);
    SdsFreeId(comment_id, &status);
    if (wlenIndex_id)
        SdsFreeId(wlenIndex_id, &status);
    SdsFreeId(uid, &status);
    
    /*  And we need the central wavelength. This is maintained by the
     *  C layer in the global variable gwave, and is a value in microns.
     *  We convert it to Angstroms - is this the right unit?
     */
   
    fprintf (listing,"TEL.PCC.WLEN           %f;\n",gwave * MICRONS_TO_ANGSTROMS);
      
    /*  For each type of fibre, we want to put out a summary containing the
     *  following information:
     *  INS.type.USED.OBJECTS   is # of fibres allocated to non-sky targets
     *  INS.type.USED.SKY       is # of fibres allocated to sky targets
     *  INS.type.XOBJ           is # of potential targets this type of fibre
     *                          could have been allocated to, but which ended
     *                          up unallocated.
     *  INS.type.XUSE           is # of fibres of this type unallocated.
     *
     *  We also want to put out the details of each fibre that has been
     *  allocated.
     */
    
    /*  First of all we do this for the FACB guide stars. These have to be
     *  handled separately, since the information about them is scattered
     *  in different sections of the structure than that for the other types.
     *  Some of these will be unallocated, and will be found in the unallocated
     *  guide section of the structure. Some will have been allocated, and
     *  these will be found in the objects structure. We list here the ones
     *  that have been allocated, and leave the ones that haven't for the
     *  end of the file where we list all the unallocated targets.
     */
                                     
    fprintf (listing,"\n");
    fprintf (listing,"# FACBs - Allocated Reference Stars\n");    
    IGuide = 0;
    usedCount = usedSkyCount = xusedCount = 0;
    SdsFind(topid, "objects", &objid, &status);
    ConfGetIds(objid, &ra_id, &dec_id, &name_id, &type_id, &spect_id,
            &prio_id, &mag_id, &pid_id, &comment_id, &wlenIndex_id, 
            &pmra_id, &pmdec_id, &status);
            

    /*  Also pick up "theta" - which is not returned by ConfGetIds. */
     
    SdsFind(objid, "theta", &theta_id, &status);

    SdsInfo(ra_id, iname, &code, &ndims, dims, &status);
    for (i = 0; i < dims[0]; i++) {
    
        /*  What we are doing here is looking at each entry in the 'objects'
         *  structure. There is one entry for each fibre. For each, we see if
         *  this is a guide fibre. If it is, we see if it is allocated. If
         *  so, we see if it is allocated to sky or to something else. 
         *  (Actually, guide fibres can only be allocated to guide targets,
         *  as things stand, but that's something hidden in the Fpil
         *  compatability routine, so we shouldn't assume it here.) If it's
         *  allocated, we list it.
         */
         
        fibreType = ConfWhichSpec (i, 0);
        if (FpilIsFibreGuide(*Instrument,fibreType)) {
            ConfArrayGetc(objid, "type", i, &c, &status);
            if (c == unallocType) {
                xusedCount++;
            } else {
            
               /*  Get theta value - the button orientation */
               
                double theta;
                SdsGet(theta_id, sizeof(double), i, &theta, &actlen, &status);
                
                if (FpilIsTargetSky(*Instrument,c)) {
                    usedSkyCount++;
                } else {
                    usedCount++;
                }
                IGuide++;
                sprintf (Prefix,"INS.REF%d.",IGuide);
                Fibre = i + 1;
                PAF_Flags = 0;
                ConfListPAFTarget(listing, i, Prefix, ra_id, dec_id, name_id,
                    type_id, spect_id, prio_id, mag_id, pid_id, 
                    pmra_id, pmdec_id, comment_id,
                    Fibre, PAF_Flags, &status);
                    
                /*  We need to calculate the sky angle properly - this can be
                 *  derived from the button theta by adding PI/2 to it.
                 */
                 
                fprintf (listing,"%sSKYANGLE        %f;\n",
                                            Prefix, slaDranrm(theta+DPIBY2));
            }
        }
        if (status != STATUS__OK) break;
    }
    
    /*  At the end of that loop, IGuide is the number of guide fibres that
     *  have been allocated. usedCount is the number allocated to non-sky
     *  targets, usedSkyCount is the number (probably none) allocated to
     *  sky targets. unallocFACBTargets is the number of unallocated guide 
     *  targets, determined earlier. We now have all we need to output the
     *  FACB summary. 
     */
     
    fprintf (listing,"\n");
    fprintf (listing,"# FACB Fibre counts\n");
    fprintf (listing,"INS.FACB.USED.OBJECTS %d;\n",usedCount);
    fprintf (listing,"INS.FACB.USED.SKY     %d;\n",usedSkyCount);
    fprintf (listing,"INS.FACB.XOBJ         %d;\n",unallocFACBTargets);
    fprintf (listing,"INS.FACB.XUSE         %d;\n",xusedCount);
                                                                   
    
    /*  Release all the SDS Id we gathered in the process. */
                                                                 
    SdsFreeId(name_id, &status);
    SdsFreeId(ra_id, &status);
    SdsFreeId(dec_id, &status);
    SdsFreeId(type_id, &status);
    if (spect_id) {
        SdsFreeId(spect_id, &status);
    }
    SdsFreeId(prio_id, &status);
    SdsFreeId(mag_id, &status);
    SdsFreeId(pid_id, &status);
    SdsFreeId(comment_id, &status);
    if (wlenIndex_id)
        SdsFreeId(wlenIndex_id, &status);
    SdsFreeId(objid, &status);
        
    /*  Now we have to handle the targets for the various other fibre types.
     *  The allocated fibres are all in the 'objects' sub-structure. In
     *  an attempt to keep all this relatively instrument-independent -
     *  ie we don't want too much FLAMES-specific stuff here if we can
     *  help it - we do this by getting the number of different PAF types
     *  to use and then we go through each fibre in turn to see which
     *  PAF type it is.
     */ 
     
    SdsFind(topid, "objects", &objid, &status);
    ConfGetIds(objid, &ra_id, &dec_id, &name_id, &type_id, &spect_id,
               &prio_id, &mag_id, &pid_id, &comment_id, &wlenIndex_id, 
               &pmra_id, &pmdec_id, &status);
    SdsInfo(ra_id, iname, &code, &ndims, dims, &status);
    
    FpilPAFTypes (*Instrument,10,&numTypes,typeNames,abbrNames);
    totalUsedCount = 0;
    for (iType = 0; iType < numTypes; iType++) {    
        insCode = *(abbrNames[iType]);
        usedCount = usedSkyCount = xusedCount = 0;
        fprintf (listing,"\n");
        fprintf (listing,"# %s Allocated Fibres\n",typeNames[iType]);
        FpilPAFFibreCodes (*Instrument,iType,10,&numTypeCodes,fibreTypeCodes);
        nEntry = 0;
        for (iFibre = 0; iFibre < dims[0]; iFibre++) {
            ConfArrayGetc(objid, "type", iFibre, &c, &status);
            fibreType = ConfWhichSpec (iFibre, 0);
            match = FALSE;
            for (iFibType = 0; iFibType < numTypeCodes; iFibType++) {
                if (fibreTypeCodes[iFibType] == fibreType) {
                    match = TRUE;
                    break;
                }
            }
            if (match) {
                if (c == unallocType) {
                    xusedCount++;
                } else {
                    if (FpilIsTargetSky(*Instrument,c)) {
                        usedSkyCount++;
                    } else {
                        usedCount++;
                    }
                    Fibre = iFibre + 1;
                    PAF_Flags = 0;
                    nEntry++;
                    sprintf (Prefix,"INS.%cFIB%d.",insCode,nEntry);
                    ConfListPAFTarget(listing, iFibre, Prefix, ra_id, dec_id,
                        name_id, type_id, spect_id, prio_id, mag_id, pid_id, 
                        pmra_id, pmdec_id, comment_id, Fibre, PAF_Flags,
                        &status);
                }
                if (status != STATUS__OK) break;
            }
        }
        fprintf (listing,"\n");
        fprintf (listing,"# %s Fibre counts\n",typeNames[iType]);
        fprintf (listing,"INS.%s.USED.OBJECTS %d;\n",abbrNames[iType],
                                                                 usedCount);
        fprintf (listing,"INS.%s.USED.SKY     %d;\n",abbrNames[iType],
                                                                 usedSkyCount);
        totalUsedCount += (usedCount + usedSkyCount);
        
        SdsFind(topid, "unallocObject", &uid, &status);
        xTargetCount = 0;
        for (i = 0; i < (nUnallocObj + nUnallocSky); i++) {
            ConfArrayGetc(uid, "type", i, &c, &status);
            c = FpilEnabledTargetType (*Instrument,c);
            ConfArrayGetc(uid, "spectrograph", i, &s, &status);
            if (!FpilIsTargetSky(*Instrument,c)) {
                match = FALSE;
                for (iFibType = 0; iFibType < numTypeCodes; iFibType++) {
                    if (FpilTargetPivotCompatible(*Instrument, c, s,
                                        fibreTypeCodes[iFibType])) {
                        match = TRUE;
                        break;
                    }
                }
                if (match) xTargetCount++;
            }          
            if (status != STATUS__OK) break;
        }
        SdsFreeId(uid, &status);
        fprintf (listing,"INS.%s.XOBJ         %d;\n",abbrNames[iType],
                                                                 xTargetCount);
        fprintf (listing,"INS.%s.XUSE         %d;\n",abbrNames[iType],
                                                                   xusedCount);
    }
    SdsFreeId(name_id, &status);
    SdsFreeId(ra_id, &status);
    SdsFreeId(dec_id, &status);
    SdsFreeId(type_id, &status);
    if (spect_id) {
        SdsFreeId(spect_id, &status);
    }
    SdsFreeId(prio_id, &status);
    SdsFreeId(mag_id, &status);
    SdsFreeId(pid_id, &status);
    SdsFreeId(comment_id, &status);
    if (wlenIndex_id)
        SdsFreeId(wlenIndex_id, &status);
    SdsFreeId(objid, &status);
    
    /*  Total number of used fibres */
    
    fprintf (listing,"\n");
    fprintf (listing,"# TOTAL number of fibres in the configuration\n");    
    fprintf (listing,"# INS.PCC.NUMFIB =\n");    
    for (iType = 0; iType < numTypes; iType++) {
    
        /*  The only difference between these two cases is the final '+' in
         *  the line!
         */
         
        if (iType < (numTypes - 1)) {    
            fprintf (listing,"#   INS.%s.USED.OBJECTS + INS.%s.USED.SKY +\n",
                                         abbrNames[iType],abbrNames[iType]);
        } else {
            fprintf (listing,"#   INS.%s.USED.OBJECTS + INS.%s.USED.SKY\n",
                                         abbrNames[iType],abbrNames[iType]);
        }
    }
    fprintf (listing,"INS.PCC.NUMFIB         %d;\n",totalUsedCount);
       
    /*  And finally the unallocated targets. The unallocated program objects
     *  are all in the 'unallocObject' structure, and all the targets in there
     *  should be included in the listing. These are put out as pseudo-
     *  comments, with a #! in front of them - this means that standard ESO 
     *  PAF readers won't be overwhelmed by them, but configure itself can
     *  still read them. (This is a bit of a fudge, but it's effective.)
     *  This is only done in expert mode, to keep the size of the PAF file
     *  down. (This was a compromise reached with ESO, once it was realised
     *  that FPOSS probably would not usually be used to re-read a PAF file
     *  at the telescope.) We do the pass through, even if not in expert mode
     *  to get the counts right, but only write out the pseudo-comment 
     *  keyword in expert mode.
     */
     
    skyCount = 0;
    fprintf (listing,"\n");
    if (ExpertMode) fprintf (listing,"# Unallocated target objects\n");    
    SdsFind(topid, "unallocObject", &uid, &status);
    ConfGetIds(uid, &ra_id, &dec_id, &name_id, &type_id, &spect_id,
               &prio_id, &mag_id, &pid_id, &comment_id, &wlenIndex_id, 
               &pmra_id, &pmdec_id, &status);
    SdsInfo(ra_id, iname, &code, &ndims, dims, &status);
    for (i = 0; i < dims[0]; i++) {
        ConfArrayGetc(uid, "type", i, &c, &status);
        if (FpilIsTargetSky(*Instrument,c)) {
           skyCount++;
        }
        if (ExpertMode) {
            Fibre = -1;
            PAF_Flags = 0;
            sprintf (Prefix,"#!INS.XOBJ%d.",i + 1);
            ConfListPAFTarget(listing, i, Prefix, ra_id, dec_id, name_id,
                    type_id, spect_id, prio_id, mag_id, pid_id, 
                    pmra_id, pmdec_id, comment_id,
                    Fibre, PAF_Flags, &status);
        }
        if (status != STATUS__OK) break;
    }
    fprintf (listing,"\n");
    fprintf (listing,"INS.SKY.XOBJ        %d;\n",skyCount);
    
    SdsFreeId(name_id, &status);
    SdsFreeId(ra_id, &status);
    SdsFreeId(dec_id, &status);
    SdsFreeId(type_id, &status);
    if (spect_id) {
        SdsFreeId(spect_id, &status);
    }
    SdsFreeId(prio_id, &status);
    SdsFreeId(mag_id, &status);
    SdsFreeId(pid_id, &status);
    SdsFreeId(comment_id, &status);
    if (wlenIndex_id)
        SdsFreeId(wlenIndex_id, &status);
    SdsFreeId(uid, &status);

    /*  Then we have the unallocated guide objects. These will be in the
     *  'unallocGuide' section of the structure. We don't include the
     *  guide stars intended for use with the telescope guider, as these
     *  have already been listed.  Again, note the use of pseudo-comments.
     */
     
    if (ExpertMode) {
        fprintf (listing,"\n");
        fprintf (listing,"# Unallocated guide objects\n");    
        SdsFind(topid, "unallocGuide", &uid, &status);
        ConfGetIds(uid, &ra_id, &dec_id, &name_id, &type_id, &spect_id,
            &prio_id, &mag_id, &pid_id, &comment_id, &wlenIndex_id, 
            &pmra_id, &pmdec_id, &status);
        SdsInfo(ra_id, iname, &code, &ndims, dims, &status);
        IGuide = 0;
        for (i = 0; i < dims[0]; i++) {
            ConfArrayGetc(uid, "type", i, &c, &status);
            if ((FpilIsTargetGuide(*Instrument, c)) && 
                   !FpilIsTargetTelGuide(*Instrument, c)) {
                IGuide++;
                sprintf (Prefix,"#!INS.XREF%d.",IGuide);
                Fibre = -1;
                PAF_Flags = 0;
                ConfListPAFTarget(listing, i, Prefix, ra_id, dec_id, name_id,
                    type_id, spect_id, prio_id, mag_id, pid_id, 
                    pmra_id, pmdec_id, comment_id,
                    Fibre, PAF_Flags, &status);
            }
            if (status != STATUS__OK) break;
        }
        SdsFreeId(name_id, &status);
        SdsFreeId(ra_id, &status);
        SdsFreeId(dec_id, &status);
        SdsFreeId(type_id, &status);
        if (spect_id) {
            SdsFreeId(spect_id, &status);
        }
        SdsFreeId(prio_id, &status);
        SdsFreeId(mag_id, &status);
        SdsFreeId(pid_id, &status);
        SdsFreeId(comment_id, &status);
        if (wlenIndex_id)
            SdsFreeId(wlenIndex_id, &status);
        SdsFreeId(uid, &status);
    }
   
    SdsFreeId(fid, &status);
    
    fclose(listing);
    
    /*  Now, if we are to write a checksum into the file, we rewind it,
     *  and re-read it, calculating the checksum for it. We then overwrite
     *  the "0000000000000000" zero checksum it contains with the complement
     *  of the checksum calculated for it with that zero checksum included.
     *  The result will be a file that has a zero checksum (strictly, it
     *  has a -ve zero checksum of all 1's, which is a 1's complement
     *  -ve zero).
     */
     
    if (ChecksumEnabled) {
    
        char record[28800];
        int recsize;
        char ascii[32];

        unsigned short sum16 = 0;
        unsigned int sum32 = 0;

        /*  Re-read, accumulating the current checksum for the file */
        
        listing = fopen(argv[2], "r+");
        while (! feof (listing)) {
            if ((recsize = fread (record, sizeof(char), 
                                   sizeof(record)/sizeof(char), listing))) {
                checksum (record, recsize, &sum16, &sum32);
            }
        }
        
        /*  Calculate the required complement as an encoded ascii string */
        
        char_encode (~sum32, ascii, 4, 0);
        ascii[16] = '\0';
        
        /*  And overwrite the original zero checksum with the complement */
        
        fseek (listing,ChecksumOffset,SEEK_SET);
        fwrite (ascii,1,16,listing);
        fclose(listing);
    }
        
    
    if (status == STATUS__OK) {
        return TCL_OK;
    } else {
        MessGetMsg(status, 0, 100, mbuf);
        sprintf(interp->result, "Error in WritePAF\n%s", mbuf);
        return TCL_ERROR;
    }
}

/*  ------------------------------------------------------------------------- */

/*
 *                        P A F  F i l e  N a m e { }
 */

static int ConfPAFFileName (ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc, char *argv[])

/*+
 *
 *  Tcl Command Name:
 *    PAFFileName
 *
 *  Function:
 *    Generate the name to be used for an output PAF file.
 *
 *  Call:
 *    PAFFileName id mode
 *
 *  Description:
 *    For FLAMES, ESO mandate a form to be used for the PAF file name.
 *    This takes the form <label>.<mode>.<time>.ins where <label> is the
 *    label supplied by the astronomer for the field, <mode> is the 
 *    observing mode and <time> is the time the file was written in the
 *    form HHMMSS.
 *
 *  Parameters:
 *    (>)  id       (Sds Id)  Sds identifier of the top level structure
 *    (>)  mode     (char)    Keyword describing the current fibre combination.
 *
 *   Returns:
 *     The file name to be used.
 *
 *   Support:
 *     Keith Shortridge, AAO
 *
 *   History:
 *      4th Jun  2002  KS  Original version
 *      1st Aug  2002  KS  Modified for new use of mode string, as requested
 *                         by Paranal.
 *-
 */
{
          
    SdsIdType fid;                 /* Sds Id of field data structure */
    char fieldName[128];           /* Name associated with field */
    char* FileMode;                /* File name component of mode argument */
    struct tm* gmTimePtr;          /* Used to get current time */
    unsigned int i;                /* General index variable */
    int length;                    /* Number of characters in string */      
    char ModeString[64];           /* Local copy of mode argument */
    StatusType status;             /* Inherited status variable */
    SdsIdType topid;               /* Sds Id of top structure */
    
    /*  Initial checks. Make sure we've been invoked properly */
    
    if (argc != 3) {
        interp->result = "wrong # args";
        return TCL_ERROR;
    }
    
    /*  Make sure there's an instrument selected */
    
    if (Instrument == NULL) {
        ErsOut(0, &status, "No instrument selected");
        interp->result = "Error";
        return TCL_OK;
    }
        
    /*  Access the field data in the structure, using the SDS ID passed
     *  in the first argument. Get the field name.
     */
    
    status = STATUS__OK;
    topid = (SdsIdType) (atol(argv[1]));
    SdsFind(topid, "fieldData", &fid, &status); 
    ArgGetString(fid, "label", sizeof(fieldName), fieldName, &status);
    
    /*  The filed name must not contain spaces. So we have to pass through
     *  fieldName changing spaces and other non-alphabetic non-numeric
     *  characters to underscores.
     */
   
    length = strlen(fieldName);
    for (i = 0; i < length; i++) {
        char TheChar = fieldName[i];
        if ((TheChar < 'a') || (TheChar > 'z')) {
            if ((TheChar < 'A') || (TheChar > 'Z')) {
                if ((TheChar < '0') || (TheChar > '9')) {
                    fieldName[i] = '_';
                }
            }
        }
    }
    
    /*  Get the date and time. We get this into the global variable FileTime,
     *  so that WritePAF can be sure it is using the same time.
     */
    
    FileTime = time(0);
    gmTimePtr = gmtime(&FileTime);    
    
    /*  The current instrument mode is included in the file name, but this
     *  is now (since 1/8/02) a set of components as formatted by the
     *  routine FpilFibreCombos(). We now only need the first component. 
     */    
    
    FileMode = (char*) NULL;
    strncpy (ModeString,argv[2],sizeof(ModeString));
    ModeString[sizeof(ModeString) - 1] = '\0';
    FileMode = strtok(ModeString,"/");
    if (FileMode == (char*) NULL) FileMode = "";
    
    /*  Now format the resulting file name */
    
    sprintf (interp->result,"%s.%s.%2.2d%2.2d%2.2d.ins",fieldName,FileMode,
                  gmTimePtr->tm_hour,gmTimePtr->tm_min,gmTimePtr->tm_sec); 
                  
    return TCL_OK;
} 
    

/*  ------------------------------------------------------------------------- */

/*
 *                         R e a d  P A F  { }
 */

static int ConfReadPAF(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc, char *argv[])

/*+
 *  Tcl Command Name:
 *    ReadPAF
 *
 *  Function:
 *    Read a PAF file listing of objects and allocations
 *
 *  Call:
 *    ReadPAF
 *
 *  Description:
 *
 *  Parameters:
 *    (>)  file     (char)    Name of the file to be read
 *    (<)  selGuide (int)     Selected guide object - index into the 
 *                            unallocGuide structure, starting from 1. Zero
 *                            if no guide object selected. 
 *    (<)  posneg   (char)    Orientation for guide pivot, if a guide
 *                            object has been selected. Will be "POS" or "NEG".
 *    (<)  modeCode (int)     Integer code giving the mode - the fibre 
 *                            combination in use - for the file. If this 
 *                            cannot be determined, -1 is returned. Otherwise
 *                            this is an index into the set of combinations
 *                            returned by FpilFibreCombos().
 *
 *  Returns:
 *    An SDS id refering to the structure which contains the
 *    contants details for the current field. Returns an error
 *    message if the file could not be opened (in which case the
 *    C function returns TCL_ERROR).
 *
 *   Support:
 *     Keith Shortridge, AAO
 *
 *-
 *   History:
 *     25th Sept 2000.  Original version. KS.
 *     28th Aug  2002.  Mode code now returned. KS.
 */
{
    SdsIdType FileId;
    StatusType status;
    PAF_VLTGuideInfoType GuideInfo;
    int ModeCode;
        
    /*  Initial checks. Make sure we've been invoked properly */
    
    if (argc != 5) {
        interp->result = "wrong # args";
        return TCL_ERROR;
    }
    if (Instrument == NULL) {
        interp->result = "No instrument specified";
        return TCL_ERROR;
    }
      
    /*  Most of the work is done by the routines defined in convertPAF.c */
    
    status = STATUS__OK;
    PAF_CreateSds (Instrument,argv[1],InstFieldDetails,&FileId,&GuideInfo,
                                                           &ModeCode,&status);
    if (status == STATUS__OK) {
        char Buffer[64];
        sprintf(Buffer, "%d", GuideInfo.SelGuideIndex);
        Tcl_SetVar(interp, argv[2], Buffer, 0);
        Tcl_SetVar(interp, argv[3], GuideInfo.Posneg, 0);
        sprintf(Buffer, "%d", ModeCode);
        Tcl_SetVar(interp, argv[4], Buffer, 0);
        sprintf(interp->result, "%ld", (long int) FileId);
        return TCL_OK;
    } else {
        char Buff[100];
        MessGetMsg(status, 0, sizeof(Buff), Buff);
        sprintf(interp->result, "Error in readPAF\n%s", Buff);
        return TCL_ERROR;
    }
}


/*  ------------------------------------------------------------------------- */

/*
 *                  M a k e  D S S  L i s t i n g  { }
 */

static int ConfMakeDSSListing(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc, char *argv[])

/*+
 *  Command Name:
 *    MakeDSSListing
 *
 *  Function:
 *    Output listing positions in DSS format
 *
 *  Call:
 *    MakeDSSListing id file type
 *
 *  Description:
 *    Output a text file listing of the allocations or objects
 *    or both in an Sds structure. The format of the listing of
 *    unallocated objects is in DSS format.
 *
 *  Parameters:
 *    (>)  id    (Sds Id)  Sds identifier of the structure
 *    (>)  file  (char)    Name of the file to be created
 *    (>)  type  (char)    Information to output
 *                         dsssky =>  Allocated Skys only
 *                         dssall =>  All allocated objects
 *
 *   Returns:
 *     "Error" - if the file could not be opened.
 *
 *   Support:
 *     Tony Farrell, AAO
 *
 *   Modified:
 *     10th Aug 2009. Added proper motion arguments to ConfGetIds(). KS.
 *-
 */
{
    SdsIdType topid, objid, fid;
    StatusType status;
    FILE *listing;
    char iname[16];
    SdsCodeType code;
    long ndims;
    unsigned long dims[7];
    int skyonly;
    SdsIdType ra_id, dec_id, name_id, type_id, spect_id, prio_id, mag_id,
        pid_id, comment_id, wlenIndex_id, pmra_id, pmdec_id;
    unsigned long i;
    char mbuf[100];

    if (argc != 4) {
        interp->result = "wrong # args";
        return TCL_ERROR;
    }
    if (strcmp(argv[3], "dsssky") == 0)
        skyonly = 1;
    else
        skyonly = 0;

    status = STATUS__OK;
    topid = (SdsIdType) (atol(argv[1]));
    SdsFind(topid, "fieldData", &fid, &status);
    listing = fopen(argv[2], "w");
    if (listing == NULL) {
        ErsOut(0, &status, "Error Opening File %s", argv[2]);
        interp->result = "Error";
        return TCL_OK;
    }

    SdsFind(topid, "objects", &objid, &status);
    ConfGetIds(objid, &ra_id, &dec_id, &name_id, &type_id, &spect_id, &prio_id,
        &mag_id, &pid_id, &comment_id, &wlenIndex_id, 
        &pmra_id, &pmdec_id, &status);
    SdsInfo(ra_id, iname, &code, &ndims, dims, &status);
    for (i = 0; i < dims[0]; i++) {
        ConfListDSSLine(skyonly, listing, i, ra_id, dec_id, name_id, type_id,
            prio_id, mag_id, pid_id, comment_id, &status);
        if (status != STATUS__OK)
            break;
    }
    SdsFreeId(name_id, &status);
    SdsFreeId(ra_id, &status);
    SdsFreeId(dec_id, &status);
    SdsFreeId(type_id, &status);
    if (spect_id)
        SdsFreeId(spect_id, &status);
    SdsFreeId(prio_id, &status);
    SdsFreeId(mag_id, &status);
    SdsFreeId(pid_id, &status);
    SdsFreeId(comment_id, &status);
    if (wlenIndex_id)
        SdsFreeId(wlenIndex_id, &status);
    SdsFreeId(objid, &status);

    SdsFreeId(fid, &status);
    fclose(listing);
    if (status == STATUS__OK)
        return TCL_OK;
    else {
        MessGetMsg(status, 0, 100, mbuf);
        sprintf(interp->result, "Error in MakeDSSListing\n%s", mbuf);
        return TCL_ERROR;
    }
}


/*  ------------------------------------------------------------------------- */

/*
 *                        G e t  F i b r e  { }
 */

static int ConfGetFibre(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc, char *argv[])
    
/*+
 *  Command Name:
 *    GetFibre
 *
 *  Function:
 *    Return Information about a fibre
 *
 *  Call:
 *    GetFibre id fibre
 *
 *  Description:
 *    Return information associated with a fibre. 9 items are returned in the
 *    result string. These are as follows - the index numbers given are those
 *    that Tcl will use for lindex when it extracts these from the string.
 *
 *    0     x coordinate
 *    1     y coordindate
 *    2     Button angle (theta)
 *    3     The object type (target type code character, or the unallocated
 *                                                                    code)
 *    4     Fibre Type (the numeric code for the fibre type)
 *    5     Object Priority
 *    6     Object Name
 *    7     Fibre type (a descriptive string for the fibre type)
 *    8     Integer 0/1 indicating if the we are allowed to allocate/unallocate
 *                                                                  this fibre.
 *    9     Pivot position in x
 *   10     Pivot position in y
 *   11     Park position in x
 *   12     Park position in y
 *
 *  Note that the 'spect' code associated with the target to which the fibre
 *  is allocated (if it is indeed allocated) is not supplied by this code.
 *  This is because this is not included in the "objects" structure that is
 *  used by this routine.  (Once the target is allocated, this is regarded as
 *  being irrelevant, but it does mean that this routine strictly doesn't
 *  return enough information to reproduce the original object type exactly -
 *  for example, in 2dF the 'spect' field might have specified the spectrograph
 *  that matches the fibre type, or it might have been a zero code to indicate
 *  that either spectrograph would do - we can't tell this. 
 *
 *  Parameters:
 *    (>)  id    (Sds Id)  Sds identifier of the structure
 *    (>)  fibre (Sds Id)  Fibre (pivot) number - starting at one
 *
 *  Returns:
 *    List contaning the values.
 *
 *   Support:
 *     Jeremy Bailey, AAO
 *
 *   Version Date:
 *     6-Oct-1994
 *    13-Apr-2001  KS  Added pivot and park positions to the end of the list.
 *    19-Dec-2001  KS  Ensured that the type description doesn't contain blanks
 *                     as this confuses the tcl lindex operation.   
 *-
 */
{
    StatusType status;
    long fibre_no;
    SdsIdType xid = 0, yid = 0, thetaid = 0, typeid = 0;
    SdsIdType topid = 0;
    SdsIdType idf = 0;
    int x, y, fibreType;
    int xp, yp, xpark, ypark;
    short oprio;
    char oname[NAMELEN];
    char FibreTypeString[64];
    int length, i;
    double theta;
    unsigned long actlen;
    char mbuf[100];
    char type;
    const char *FibreTypeDescr;
    StatusType ignore = STATUS__OK;

    if (argc != 3) {
        interp->result = "wrong # args";
        return TCL_ERROR;
    }
    status = SDS__OK;
    idf = (SdsIdType) (atol(argv[1]));
    SdsFind(idf, "objects", &topid, &status);
    fibre_no = atol(argv[2]) - 1;

    if ((fibre_no < 0) || (fibre_no >= (int) FpilGetNumPivots(*Instrument))) {
        sprintf(interp->result,
            "Illegal Pivot Number %ld\nMust be between 1 and %d",
            fibre_no + 1, FpilGetNumPivots(*Instrument));
        return TCL_ERROR;
    }
    
    /*  We can get pivot information directly from the instrument field
     *  details structure.
     */
     
    ConfArrayGeti(InstFieldDetails, "xPiv", fibre_no, &xp, &status);
    ConfArrayGeti(InstFieldDetails, "yPiv", fibre_no, &yp, &status);
    ConfArrayGeti(InstFieldDetails, "xPark", fibre_no, &xpark, &status);
    ConfArrayGeti(InstFieldDetails, "yPark", fibre_no, &ypark, &status);

    /*  Now get the rest of the information needed */
    
    SdsFind(topid, "x", &xid, &status);
    SdsFind(topid, "y", &yid, &status);
    SdsFind(topid, "type", &typeid, &status);
    if (status != SDS__OK) {
        status = SDS__OK;
        SdsFind(topid, "xf", &xid, &status);
        SdsFind(topid, "yf", &yid, &status);
    }
    SdsFind(topid, "theta", &thetaid, &status);
    SdsGet(xid, 4, fibre_no, &x, &actlen, &status);
    SdsGet(yid, 4, fibre_no, &y, &actlen, &status);
    SdsGet(thetaid, 8, fibre_no, &theta, &actlen, &status);
    SdsGet(typeid, 1, fibre_no, &type, &actlen, &status);
    if (type != FpilUnallocatedType(*Instrument)) {
        SdsFind(topid, "name", &xid, &status);
        SdsGet(xid, NAMELEN, fibre_no * NAMELEN, &oname, &actlen, &status);
        SdsFind(topid, "priority", &xid, &status);
        SdsGet(xid, sizeof(short), fibre_no, &oprio, &actlen, &status);
    } else {
        oprio = 0;
        sprintf(oname, "\"\"");
    }
    fibreType = 0;
    fibreType = ConfWhichSpec((int) fibre_no, &FibreTypeDescr);

    SdsFreeId(topid, &ignore);
    SdsFreeId(xid, &ignore);
    SdsFreeId(yid, &ignore);
    SdsFreeId(typeid, &ignore);
    SdsFreeId(thetaid, &ignore);
    
    strncpy(FibreTypeString,FibreTypeDescr,sizeof(FibreTypeString));
    FibreTypeString[sizeof(FibreTypeString) - 1] = '\0';
    length = strlen(FibreTypeString);
    for (i = 0; i < length; i++) {
       if (FibreTypeString[i] == ' ') FibreTypeString[i] = '_';
    }

    if (status == SDS__OK) {
        sprintf(interp->result, "%d %d %f %c %d %d %s %s %d %d %d %d %d",
            x, y, theta, type, fibreType, oprio, oname, FibreTypeString,
            UseFibreType[fibreType],xp,yp,xpark,ypark);
        return TCL_OK;
    } else {
        MessGetMsg(status, 0, 100, mbuf);
        sprintf(interp->result, "Error Getting Fibre Data\n%s", mbuf);
        return TCL_ERROR;
    }
}

/*  ------------------------------------------------------------------------- */

/*
 *                        S e t  F i b r e  { }
 */
 
static int ConfSetFibre(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc, char *argv[])

/*+
 *  Command Name:
 *     SetFibre
 *
 *  Function:
 *     Set Information about a fibre
 *
 *  Call:
 *     SetFibre cid sid
 *
 *  Description:
 *     This routine was uncommented in the original code, and I (KS) believe
 *     that it is no longer used. It seems to expect a specific structure
 *     as a second argument with 'pivot', 'x', 'y' and 'theta' items. It 
 *     uses the pivot value as an index and sets the corresponding item in
 *     the 'objects' sub-structure of the configuration data structure to
 *     the x, y and theta values supplied in the second argument.
 *
 *  Parameters:
 *    (>)  icd    (Sds Id)  Sds identifier of the configuration data structure
 *    (>)  sid    (Sds Id)  Sds identifier of a structure containing 'pivot',
 *                          'x', 'y' and 'theta' items.
 *-
 */
{
    StatusType status;
    long fibre_no;
    SdsIdType xid = 0, yid = 0, thetaid = 0;
    SdsIdType topid = 0, idf = 0;
    SdsIdType arg;
    long x, y;
    int xi, yi;
    double theta;
    char mbuf[100];
    StatusType ignore = STATUS__OK;
    LeakCheckType LL;
    LL = LeakPrep();

    if (argc != 3) {
        interp->result = "wrong # args";
        return TCL_ERROR;
    }
    status = SDS__OK;
    idf = (SdsIdType) (atol(argv[1]));
    SdsFind(idf, "objects", &topid, &status);
    arg = atol(argv[2]);
    ArgGeti(arg, "pivot", &fibre_no, &status);
    ArgGeti(arg, "x", &x, &status);
    ArgGeti(arg, "y", &y, &status);
    xi = (int) x;
    yi = (int) y;
    fibre_no = fibre_no - 1;
    if ((fibre_no < 0) || (fibre_no >= (int) FpilGetNumPivots(*Instrument))) {
        sprintf(interp->result,
            "Illegal Pivot Number %ld\nMust be between 1 and %d",
            fibre_no + 1, FpilGetNumPivots(*Instrument));
        return TCL_ERROR;
    }

    ArgGetd(arg, "theta", &theta, &status);
    SdsFind(topid, "xf", &xid, &status);
    SdsFind(topid, "yf", &yid, &status);
    SdsFind(topid, "theta", &thetaid, &status);
    SdsPut(xid, 4, fibre_no, &xi, &status);
    SdsPut(yid, 4, fibre_no, &yi, &status);
    SdsPut(thetaid, 8, fibre_no, &theta, &status);

    SdsFreeId(topid, &ignore);
    SdsFreeId(xid, &ignore);
    SdsFreeId(yid, &ignore);
    SdsFreeId(thetaid, &ignore);
    LeakCheck("ConfSetFibre", LL, 0);

    if (status == SDS__OK) {
        return TCL_OK;
    } else {
        MessGetMsg(status, 0, 100, mbuf);
        sprintf(interp->result, "Error Setting Fibre Data\n%s", mbuf);
        return TCL_ERROR;
    }
}

/*  ------------------------------------------------------------------------- */

/*
 *                 S e t  B u t t o n  A n g l e  { }
 */
 
static int ConfSetButtonAngle(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc, char *argv[])

/*+
 *  Command Name:
 *    SetButtonAngle
 *
 *  Function:
 *    Set the angle of a button
 *
 *  Call:
 *    SetButtonAngle id pivid fibre angle
 *
 *  Description:
 *    Set the angle of a button to a specified value, or to the radial
 *    value.
 *
 *
 *  Parameters:
 *    (>)  id    (Sds Id)  Sds identifier of the structure
 *    (>)  pivid (Sds Id)  Sds identifier of the pivot structure
 *    (>)  fibre (int)     Fibre (pivot) number - starting at one
 *    (>)  theta (char)    Either the required button angle in radians,
 *                         or the word "radial".
 *
 *   Support:
 *     Jeremy Bailey, AAO
 *
 *   Version Date:
 *     6-Oct-1994    
 *-
 */
{
    StatusType status;
    long fibre_no;
    SdsIdType topid, idf, pivid;
    int xp, yp, x, y;
    double theta;
    char mbuf[100];

    if (argc != 5) {
        interp->result = "wrong # args";
        return TCL_ERROR;
    }
    status = SDS__OK;
    idf = (SdsIdType) (atol(argv[1]));
    pivid = (SdsIdType) (atol(argv[2]));
    SdsFind(idf, "objects", &topid, &status);
    fibre_no = atol(argv[3]);
    fibre_no = fibre_no - 1;
    if ((fibre_no < 0) || (fibre_no >= (int) FpilGetNumPivots(*Instrument))) {
        sprintf(interp->result,
            "Illegal Pivot Number %ld\nMust be between 1 and %d",
            fibre_no + 1, FpilGetNumPivots(*Instrument));
        return TCL_ERROR;
    }
    if (strcmp(argv[4], "radial") == 0) {

/*  Calculate theta so that the button is parallel to the fibre  */

        ConfArrayGeti(pivid, "xPiv", fibre_no, &xp, &status);
        ConfArrayGeti(pivid, "yPiv", fibre_no, &yp, &status);
        ConfArrayGeti(topid, "x", fibre_no, &x, &status);
        ConfArrayGeti(topid, "y", fibre_no, &y, &status);
        theta = slaDranrm(atan2((double) (yp - y), (double) (xp - x)) - DPIBY2);
    } else
        theta = atof(argv[4]);

    ConfArrayPutd(topid, "theta", fibre_no, theta, &status);
    SdsFreeId(topid, &status);

    if (status == SDS__OK) {
        return TCL_OK;
    } else {
        MessGetMsg(status, 0, 100, mbuf);
        sprintf(interp->result, "Error Setting Button Angle\n%s", mbuf);
        return TCL_ERROR;
    }
}

/*  ------------------------------------------------------------------------- */

/*
 *                 F i n d  O b j e c t  B y  N a m e  { }
 */
 
static int ConfFindObjectByName (ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc, char *argv[])

/*+
 *  Command Name:
 *    FindObjectByName
 *
 *  Function:
 *    Looks up a named object in an SDS sub-structure.
 *
 *  Call:
 *    FindObjectByName id name
 *
 *  Description:
 *    Looks up a named object in any of the three sections
 *    of the SDS structure (allocated objects, unallocated objects,
 *    unallocated guides) and returns its index in that structure.
 *
 *
 *  Parameters:
 *    (>)  id    (Sds Id)  Sds identifier of the substructure
 *    (>)  name  (char)    The object name.
 *
 *  Returns:
 *    The index of the named object in the sub-structure, starting from 1.
 *    If the object is not found, this routine returns zero.
 *
 *  Support:
 *     Keith Shortridge, AAO
 *
 *  Version Date:
 *     5-Sep-2007  KS  Original version.
 *    25-Oct-2007  KS  Comparision is now on the whole of the object name,
 *                     rather than just the length of the argument name.
 *                     The previous version would match 'Prog52' with
 *                     'Prog521' etc.
 *-
 */
{
    SdsIdType topid, id;
    StatusType status = STATUS__OK;
    unsigned long actlen;
    char name[NAMELEN];
    SdsCodeType code;
    long ndims;
    unsigned long dims[7];
    unsigned int i;
    int index_no;

    if (argc != 3) {
        interp->result = "wrong # args";
        return TCL_ERROR;
    }
    
    topid = (SdsIdType) (atol(argv[1]));
    index_no = 0;

    /*  We use the size of the "ra" field to get the number of objects. */
    
    SdsFind(topid, "ra", &id, &status);
    SdsInfo(id, name, &code, &ndims, dims, &status);
    SdsFreeId(id, &status);
    
    /*  Now go through all the objects in the sub-structure, looking for
     *  the name that matches
     */
     
    SdsFind(topid, "name", &id, &status);
    for (i = 0; i < dims[0]; i++) {
       SdsGet(id, NAMELEN, i * NAMELEN, name, &actlen, &status);
       if (!strcmp(argv[2],name)) {
          index_no = i + 1;
          break;
       }
    }
    SdsFreeId(id, &status);
    
    /*  Return the required index number */
    
    sprintf (interp->result,"%d",index_no);
    return TCL_OK;
}



/*  ------------------------------------------------------------------------- */

/*
 *                        G e t  O b j e c t  D a t a  { }
 */
 
static int ConfGetObjectData(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc, char *argv[])

/*+
 *  Command Name:
 *    GetObjectData
 *
 *  Function:
 *    Get information about an object
 *
 *  Call:
 *    GetObjectData id index name ra dec type spect priority mag pid 
 *                                              comment alloc pmra pmdec
 *
 *  Description:
 *    Return information about an object in any of the three sections
 *    of the SDS structure (allocated objects, unallocated objects,
 *    unallocated guides).
 *
 *
 *  Parameters:
 *    (>)  id    (Sds Id)  Sds identifier of the substructure
 *    (>)  index (int)     index number - starting at one
 *    (<)  name  (char)    The object name.
 *    (<)  ra    (double)  RA in radians.
 *    (<)  dec   (double)  Dec in radians.
 *    (<)  type  (char)    Type of object. See below for possible values.
 *    (<)  spect (char)    The spect code associated with the object. See below.
 *    (<)  priority (int)  Priority of object.
 *    (<)  mag   (double)  Magnitude of object.
 *    (<)  pid   (int)     Program id.
 *    (<)  text  (char)    Text describing the object type.
 *    (<)  comment (char)  Comment String.
 *    (<)  x     (int)     Position in X.
 *    (<)  y     (int)     Position in Y.
 *    (<)  alloc (int)     Flag indicating allocated.
 *    (<)  pmra  (double)  Proper motion in RA, in arcsec/year.
 *    (<)  pmdec (double)  Proper motion in Dec, in arcsec/year.
 *
 *    The value returned in the type field is one of the characters
 *    defined as possible target type codes for the instrument. This will
 *    be one of the type codes returned by FpilEncodeTargetType(), or (for
 *    items in the "objects" section, may be the 'unallocated' type code
 *    as returned by FpilUnallocatedType(). The value returned in the type
 *    field is the spect code returned by FpilEncodeTargetType() for the
 *    target. The 'alloc' value is only set if there is an 'allocated' 
 *    array in the structure - this is only the case under some circumstances
 *    for the unallocObject and unallocGuide structures. It should be assumed
 *    true for the 'objects' structure.
 *
 *   Support:
 *     Jeremy Bailey, AAO
 *
 *   Version Date:
 *     6-Oct-1994
 *    13-Oct-2001  KS  Added Y and Y positions to argument list.
 *    22-Jun-2006  KS  Added alloc flag to argument list.
 *    17-Aug-2009  KS  Added proper motion values to argument list. 
 *-
 */
{
    StatusType status, notfound;
    long index_no;
    SdsIdType id;
    SdsIdType topid;
    char name[NAMELEN];
    double ra, dec;
    char alloc;
    char type;
    char spect;
    short priority;
    double mag;
    short pid;
    short wlenIndex;
    int spectKnown;
    char comment[CMTLEN];
    unsigned long actlen;
    char mbuf[100];
    char buffer[80];
    char targetString[32];
    char targetText[80];
    int x,y;
    double pmra,pmdec;

    if (argc != 18) {
        interp->result = "wrong # args";
        return TCL_ERROR;
    }
    status = SDS__OK;
    topid = (SdsIdType) (atol(argv[1]));
    index_no = atol(argv[2]) - 1;

    SdsFind(topid, "name", &id, &status);
    SdsGet(id, NAMELEN, index_no * NAMELEN, name, &actlen, &status);
    SdsFreeId(id, &status);

    SdsFind(topid, "ra", &id, &status);
    SdsGet(id, sizeof(double), index_no, &ra, &actlen, &status);
    SdsFreeId(id, &status);

    SdsFind(topid, "dec", &id, &status);
    SdsGet(id, sizeof(double), index_no, &dec, &actlen, &status);
    SdsFreeId(id, &status);

    SdsFind(topid, "type", &id, &status);
    SdsGet(id, sizeof(char), index_no, &type, &actlen, &status);
    SdsFreeId(id, &status);


    notfound = 0;
    SdsFind(topid, "wlenIndex", &id, &notfound);
    if (notfound)
        wlenIndex = -1;
    else {
        SdsGet(id, sizeof(short), index_no, &wlenIndex, &actlen, &status);
        SdsFreeId(id, &status);
    }
  
    notfound = 0;
    spectKnown = FALSE;
    SdsFind(topid, "spect", &id, &notfound);
    if (notfound)
        spect = '\0';
    else {
        SdsGet(id, sizeof(char), index_no, &spect, &actlen, &status);
        SdsFreeId(id, &status);
        spectKnown = TRUE;
    }
    
    notfound = 0;
    SdsFind(topid, "priority", &id, &notfound);
    if (notfound)
        priority = 0;
    else {
        SdsGet(id, sizeof(short), index_no, &priority, &actlen, &status);
        SdsFreeId(id, &status);
    }

    notfound = 0;
    SdsFind(topid, "magnitude", &id, &notfound);
    if (notfound)
        mag = 0;
    else {
        SdsGet(id, sizeof(double), index_no, &mag, &actlen, &status);
        SdsFreeId(id, &status);
    }

    notfound = 0;
    SdsFind(topid, "pId", &id, &notfound);
    if (notfound)
        pid = 0;
    else {
        SdsGet(id, sizeof(short), index_no, &pid, &actlen, &status);
        SdsFreeId(id, &status);
    }

    notfound = 0;
    SdsFind(topid, "comment", &id, &notfound);
    if (notfound) {
        strcpy(comment, "No comment");
        actlen = 10;
    } else {
        SdsGet(id, CMTLEN, index_no * CMTLEN, comment, &actlen, &status);
        SdsFreeId(id, &status);
    }
    
    notfound = 0;
    SdsFind(topid, "x", &id, &notfound);
    if (notfound)
        x = 0;
    else {
        SdsGet(id, sizeof(x), index_no, &x, &actlen, &status);
        SdsFreeId(id, &status);
    }
    
    notfound = 0;
    SdsFind(topid, "y", &id, &notfound);
    if (notfound)
        y = 0;
    else {
        SdsGet(id, sizeof(y), index_no, &y, &actlen, &status);
        SdsFreeId(id, &status);
    }
    
    notfound = 0;
    SdsFind(topid, "allocated", &id, &notfound);
    if (notfound)
        alloc = 0;
    else {
        SdsGet(id, sizeof(alloc), index_no, &alloc, &actlen, &status);
        SdsFreeId(id, &status);
    }

    notfound = 0;
    SdsFind(topid, "pmra", &id, &notfound);
    if (notfound)
        pmra = 0;
    else {
        SdsGet(id, sizeof(double), index_no, &pmra, &actlen, &status);
        SdsFreeId(id, &status);
    }
    notfound = 0;
    SdsFind(topid, "pmdec", &id, &notfound);
    if (notfound)
        pmdec = 0;
    else {
        SdsGet(id, sizeof(double), index_no, &pmdec, &actlen, &status);
        SdsFreeId(id, &status);
    }
    
    /*  Convert proper motions from radians/year to arcsec/year */
    
    pmra *= (60.0 * 60.0 * 180.0) / DPI;
    pmdec *= (60.0 * 60.0 * 180.0) / DPI;
    
    
    FpilDecodeTargetType (*Instrument,type,spect,spectKnown, wlenIndex,
                          targetString, targetText);
    if (status == SDS__OK) {
        Tcl_SetVar(interp, argv[3], name, 0);
        sprintf(buffer, "%10f", ra);
        Tcl_SetVar(interp, argv[4], buffer, 0);
        sprintf(buffer, "%10f", dec);
        Tcl_SetVar(interp, argv[5], buffer, 0);
        sprintf(buffer, "%c", type);
        Tcl_SetVar(interp, argv[6], buffer, 0);
        sprintf(buffer, "%d", spect);
        Tcl_SetVar(interp, argv[7], buffer, 0);
        sprintf(buffer, "%d", priority);
        Tcl_SetVar(interp, argv[8], buffer, 0);
        sprintf(buffer, "%.2f", mag);
        Tcl_SetVar(interp, argv[9], buffer, 0);
        sprintf(buffer, "%d", pid);
        Tcl_SetVar(interp, argv[10], buffer, 0);
        Tcl_SetVar(interp, argv[11], comment, 0);
        Tcl_SetVar(interp, argv[12], targetText, 0);
        sprintf(buffer, "%d", x);
        Tcl_SetVar(interp, argv[13], buffer, 0);
        sprintf(buffer, "%d", y);
        Tcl_SetVar(interp, argv[14], buffer, 0);
        sprintf(buffer, "%d", alloc);
        Tcl_SetVar(interp, argv[15], buffer, 0);
        sprintf(buffer, "%8.4f", pmra);
        Tcl_SetVar(interp, argv[16], buffer, 0);
        sprintf(buffer, "%8.4f", pmdec);
        Tcl_SetVar(interp, argv[17], buffer, 0);

        return TCL_OK;
    } else {
        MessGetMsg(status, 0, 100, mbuf);
        sprintf(interp->result, "Error Getting Object Data\n%s", mbuf);
        return TCL_ERROR;
    }
}

/*  ------------------------------------------------------------------------- */

/*
 *                     G e t  A l l o c  E n t r i e s  { }
 */
 
static int ConfGetAllocEntries(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc, char *argv[])

/*+
 *  Command Name:
 *    GetAllocEntries
 *
 *  Function:
 *    Get the size of one of the sections of the SDS structure
 *
 *  Call:
 *    GetAllocEntries id
 *
 *  Description:
 *    Return the number of entries in any of the three sections
 *    of the SDS structure (allocated objects, unallocated objects,
 *    unallocated guides).
 *
 *
 *  Parameters:
 *    (>)  id    (Sds Id)  Sds identifier of the substructure
  *
 *   Support:
 *     Keith Shortridge, AAO
 *
 *   Version Date:
 *    21-Jun-2006  KS  Original version.
 *-
 */
{
    StatusType status;
    SdsIdType id;
    SdsIdType topid;
    SdsCodeType Code;
    char Name[NAMELEN];
    unsigned long Dims[10];
    long NDims;
    char String[32];

    if (argc != 2) {
        interp->result = "wrong # args";
        return TCL_ERROR;
    }
    status = SDS__OK;
    topid = (SdsIdType) (atol(argv[1]));

    SdsFind(topid, "type", &id, &status);
    SdsInfo(id, Name, &Code, &NDims, Dims, &status);
    SdsFreeId(id,&status);
    if (status == SDS__OK) {
        sprintf (String,"%ld",Dims[0]);
        Tcl_SetResult (interp,String,TCL_VOLATILE);
        return TCL_OK;
    } else {
        char mbuf[100];
        MessGetMsg(status, 0, 100, mbuf);
        sprintf(interp->result, "Error Getting number of entries\n%s", mbuf);
        return TCL_ERROR;
    }
}

/*  ------------------------------------------------------------------------- */

/*
 *                        G e t  O b j e c t  { }
 */
 
static int ConfGetObject(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc, char *argv[])

/*+
 *
 *  Command name:
 *    GetObject
 *
 *
 *  Function:
 *    Get object information for mimic display
 *  
 *  Call:
 *    GetObject id index zoom [tscale]
 *
 *  Description:
 *    Return the information needed to draw an object
 *    on the mimic display.
 *   
 *
 *   Parameters:
 *
 *    (>) id      (Sds Id)  Sds identifier of the substructure   
 *    (>) index   (int)     Index number in structure
 *    (>) zoom    (float)   Zoom factor for overall display
 *    (>) tscale  (float)   Scale factor for targets (optional -
 *                          defaults to 1.0).
 *
 *   Returns:
 *    The argument string of a Tk canvas create command which will
 *    draw the object on the mimic display.
 *
 *   Support:
 *     Jeremy Bailey, AAO
 *
 *   Version Date:
 *     6-Oct-1994    
 *-
 */
{
    StatusType status;
    StatusType spectStatus;
    long index_no;
    SdsIdType xid = 0, yid = 0, tid = 0, sid = 0;
    SdsIdType topid = 0;
    int x, y;
    unsigned long actlen;
    float zoom;
    char mbuf[100];
    char type;
    char spect;
    float tscale;
    SdsIdType allid = 0;
    char c;

    if (argc != 4 && argc != 5) {
        interp->result = "wrong # args";
        return TCL_ERROR;
    }
    if (Instrument == NULL) {
        interp->result = "No instrument specified";
        return TCL_ERROR;
    }
    
    status = SDS__OK;
    topid = (SdsIdType) (atol(argv[1]));
    index_no = atol(argv[2]) - 1;
    zoom = atof(argv[3]);
    tscale = 1.0;
    if (argc == 5) tscale = atof(argv[4]);
    SdsFind(topid, "allocated", &allid, &status);
    if (status == SDS__NOITEM)
        ErsAnnul(&status);
    else {
        SdsGet(allid, 1, index_no, &c, &actlen, &status);
        SdsFreeId(allid, &status);
        allid = 0;
        if (c == 1) {
            sprintf(interp->result, "unused");
            return TCL_OK;
        }
    }

    SdsFind(topid, "x", &xid, &status);
    SdsFind(topid, "y", &yid, &status);
    if (status == SDS__NOITEM) {
        ErsAnnul(&status);
        SdsFind(topid, "xf", &xid, &status);
        SdsFind(topid, "yf", &yid, &status);
    }
    SdsFind(topid, "type", &tid, &status);
    SdsGet(xid, 4, index_no, &x, &actlen, &status);
    SdsGet(yid, 4, index_no, &y, &actlen, &status);
    SdsGet(tid, 1, index_no, &type, &actlen, &status);
    if (status != SDS__OK) {
        StatusType ignore = STATUS__OK;
        MessGetMsg(status, 0, 100, mbuf);
        if (xid != 0)
            SdsFreeId(xid, &ignore);
        if (xid != 0)
            SdsFreeId(yid, &ignore);
        if (tid != 0)
            SdsFreeId(tid, &ignore);
        sprintf(interp->result, "Error Getting Fibre Data\n%s", mbuf);
        return TCL_ERROR;
    }

    /*  Not all targets have spectrograph info - guide targets don't, so we
     *  aren't surprised if we can't find a spectrograph code.
     */

    spectStatus = 0;
    SdsFind(topid, "spectrograph", &sid, &spectStatus);
    if (spectStatus == 0) {
        SdsGet(sid, 1, index_no, &spect, &actlen, &spectStatus);
        if (spectStatus != 0)
            spect = 0;
        SdsFreeId(sid, &spectStatus);
    }

    if (type == FpilUnallocatedType(*Instrument)) {
        sprintf(interp->result, "unused");
    } else {   

        FpilDrawTarget(*Instrument, type, spect, x, y, zoom, tscale,
              MimicCircleRadius, MimicCircleOffset, useGrey, interp->result);
                          
    }

    SdsFreeId(xid, &status);
    SdsFreeId(yid, &status);
    SdsFreeId(tid, &status);

    return TCL_OK;

}

/*  ------------------------------------------------------------------------- */

/*
 *                        C o n v e r t  A s c i i  { }
 */
 
static int ConfConvertAscii(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc, char *argv[])

/*+
 *
 *  Command name:
 *    ConvertAscii
 *
 *
 *  Function:
 *    Convert input text file to SDS
 *  
 *  Call:
 *    ConvertAscii file fracDayFlag [lowmag highmag]
 *
 *  Description:
 *    Convert a text file description of a configuration to its
 *    SDS representation and return the SDS identifier
 *   
 *   Parameters:
 *
 *    (>) file        (Char)  Name of the text file 
 *    (<) fracDayFlag (int)   True (1) if UTDATE in file specified a fractional
 *                            day value - ie if a specific day AND time was
 *                            made explicit in the file.
 *    (>) lowmag      (float) Lower magnitude limit
 *    (>) highmag     (float) Higher magnitude limit
 *
 *   Lowmag and highmag are optional, but if one is given both are required.
 *   If specified, the input targets will be filtered and only those within
 *   the magnitude limits will be included. 
 *
 *   Returns:
 *     The SDS Id of the created structure.
 *
 *   Support:
 *     Jeremy Bailey, AAO
 *
 *   Version Date:
 *     6-Oct-1994
 *    21-May-2001  KS  Added lowmag, highmag and fracDayFlag.
 *    14-Aug-2001  KS  Added position angle code.   
 *    21-Jul-2006  TJF Add wlens.
 *    14-Aug-2009  KS  Added explicitPms.
 *-
 */
{
    tdFfieldType field;
    StatusType status = STATUS__OK;
    SdsIdType fileId = 0;
    short fracDayFlag = FALSE;
    short options = 0;
    double longit, lat, hm;
    FILE *asciiFile;
    char oname[NAMELEN];
    double uttime = 0.5;
    char buff[80];
    objectType *list, *next;
    float maglimits[2];
    int lineCount = 0;
    double Angle;
    double Offset;
    short InUse;
    char Scale[16];
    wavelensType wlens;   /* explict wavelength values. */
    short explicitPms;    /* true if file format includes proper motion cols */

    maglimits[0] = maglimits[1] = 0.0;
    if (argc == 5) {
        maglimits[0] = atof(argv[3]);
        maglimits[1] = atof(argv[4]);
    } else {
        if (argc != 3) {
            interp->result = "wrong # args";
            return TCL_ERROR;
        }
    }

    /*  We need to work out the uttime correction to allow for the difference
     *  in midnight UT and local time.
     */
     
    slaObs(-1, (char *) FpilGetTelescope(*Instrument), oname, &longit, &lat,
                                                                         &hm);
    if (strcmp(oname, "?") == 0) {
       uttime = 0.0;
    } else {
       uttime = -longit / (2.0 * 3.141593);
    }
    /* Initiailse the explicit wavelength array */
    tdFwlensInit(&wlens, &status);
    
    /*  Now open file and process the field data at the start of the file */
    
    if ((asciiFile = fopen(argv[1], "r")) == NULL) {
        sprintf(interp->result, "Cannot Open file %s", argv[1]);
        return TCL_ERROR;
    }
    tdFparseFieldData(Instrument, asciiFile, &field.fieldData, uttime,
                      &fracDayFlag, &lineCount, &wlens, &explicitPms, &status);
                                          
    /*  If field.fieldData.argusSpec has been set, then ARGUS information 
     *  has been specified. We handle this through ConfCSetArgusData(), which
     *  will arrange to set the position angle as required.
     */
    
    ConfCGetArgusData(&InUse,&Angle,Scale,&Offset);
    Angle = field.fieldData.posangle + Offset;
    ConfCSetArgusData(field.fieldData.argusSpec,Angle,
                                        field.fieldData.scale,Offset);
        

    field.unallocGuide = field.unallocObject = NULL;
    tdFparseObjectData(Instrument, &wlens, asciiFile, &field,
                       &options, maglimits, explicitPms, &lineCount, &status);

    fclose(asciiFile);

    tdFconvertToSds("2df_field", &field, options, &wlens, &fileId, &status);

    /*
     * Free the two lists.  This is done in spite of the status, as they
     * were initialised above and will need to be cleaned up if they
     * are no longer null
     */
     
    list = field.unallocObject;
    while (list) {
        next = list->next;
        free(list);
        list = next;
    }
    list = field.unallocGuide;
    while (list) {
        next = list->next;
        free(list);
        list = next;
    }

    if (status == STATUS__OK) {
        if (fracDayFlag) {
            Tcl_SetVar(interp, argv[2], "1", 0);
        } else {
            Tcl_SetVar(interp, argv[2], "0", 0);
        }   
        sprintf(interp->result, "%d", (int) fileId);
        return TCL_OK;
    } else {
        ErsRep (0,&status,"Error reading file '%s'",argv[1]);
        MessGetMsg(status, 0, sizeof(buff), buff);
        Tcl_AppendResult(interp, "Error Converting Ascii File\n",
            (char *) NULL);
        Tcl_AppendResult(interp, buff, (char *) NULL);
        return TCL_ERROR;
    }
}

/*  ------------------------------------------------------------------------- */

/*
 *                  C o u n t  A s c i i  O b j e c t s  { }
 */
 
static int ConfCountAsciiObjects(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc, char *argv[])

/*+
 *
 *  Command name:
 *    CountAsciiObjects
 *
 *
 *  Function:
 *    Pass through input text file counting objects
 *  
 *  Call:
 *    ConvertAscii file [lowmag highmag]
 *
 *  Description:
 *    This routine performs processing on an ASCII input file similar to
 *    that performed by ConvertAscii(), but only returns the number of
 *    objects that will be read from the file, given the magnitude
 *    limits (if any).  
 *
 *   Parameters:
 *
 *    (>) file        (Char)  Name of the text file 
 *    (>) lowmag      (float) Lower magnitude limit
 *    (>) highmag     (float) Higher magnitude limit
 *
 *   Lowmag and highmag are optional, but if one is given both are required.
 *   If specified, the input targets will be filtered and only those within
 *   the magnitude limits will be included. 
 *
 *   Returns:
 *     The number of objects that will be read from the file.
 *
 *   Support:
 *     Keith Shortridge, AAO
 *
 *   Version Date:
 *    10-Jun-2001  KS  Original version, based on ConfConvertAscii().    
 *    21-Jul-2006  TJF Add wlens.
 *    14-Aug-2009  KS  Added explicitPms.
 *-
 */
{
    tdFfieldType field;
    StatusType status = STATUS__OK;
    short fracDayFlag = FALSE;
    FILE *asciiFile;
    double uttime = 0.0;
    char buff[80];
    float maglimits[2];
    int Count;
    int lineCount = 0;
    wavelensType wlens;   /* explict wavelength values. */
    short explicitPms;    /* true if file format includes proper motion cols */

    maglimits[0] = maglimits[1] = 0.0;
    if (argc == 4) {
        maglimits[0] = atof(argv[2]);
        maglimits[1] = atof(argv[3]);
    } else {
        if (argc != 2) {
            interp->result = "wrong # args";
            return TCL_ERROR;
        }
    }
    /* Initiailse the explicit wavelength array */
    tdFwlensInit(&wlens, &status);

    if ((asciiFile = fopen(argv[1], "r")) == NULL) {
        sprintf(interp->result, "Cannot Open file %s", argv[1]);
        return TCL_ERROR;
    }
    tdFparseFieldData(Instrument, asciiFile, &field.fieldData, uttime,
                      &fracDayFlag, &lineCount, &wlens, &explicitPms, &status);
    Count = tdFparseObjectCount(Instrument, asciiFile, maglimits, &status);

    fclose(asciiFile);

    if (status == STATUS__OK) {
        sprintf(interp->result, "%d", Count);
        return TCL_OK;
    } else {
        ErsRep (0,&status,"Error reading file '%s'",argv[1]);
        MessGetMsg(status, 0, sizeof(buff), buff);
        Tcl_AppendResult(interp, "Error Converting Ascii File\n",
            (char *) NULL);
        Tcl_AppendResult(interp, buff, (char *) NULL);
        return TCL_ERROR;
    }
}


/*  ------------------------------------------------------------------------- */

/*
 *                     S e t  U T  M e r i d i a n  { }
 */
 
static int ConfSetUTMeridian(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc, char *argv[])
/*+
 *
 *  Command name:
 *    SetUTMeridian
 *
 *
 *  Function:
 *    Set the UT of a structure for a given Hour Angle
 *  
 *  Call:
 *    SetUTMeridian id [ha]
 *
 *  Description:
 *    Set the UT in a configuration structure to the value required to
 *    put the field at a specified hour angle for its configuration date.
 *   
 *
 *   Parameters:
 *
 *    (>) id   (Sds id)  Sds Id of configuration structure.
 *    (>) ha   (double)  Required hour angle - if omitted set to the meridian. 
 *
 *
 *   Support:
 *     Jeremy Bailey, AAO
 *
 *   Version Date:
 *     6-Oct-1994
 *    21-May-2001. Now uses ChangeMjd(). KS  
 *-
 */
{
    StatusType status = STATUS__OK;
    SdsIdType id = 0, idf = 0;
    double cra, mjd, lst;
    double longit, lat, hm;
    char oname[NAMELEN];
    char sign[1];
    int ihmsf[4];
    char buff[80];
    double ha;
    int iy, im, idd;
    double f;
    int jstat;
    char buffer[200];
    StatusType ignore = STATUS__OK;

    if (argc < 2) {
        sprintf(interp->result, "%s:Wrong number of arguments", argv[0]);
        return TCL_ERROR;
    }

    id = atol(argv[1]);
    if (argc < 3)
        ha = 0.0;
    else
        ha = DPI / 12 * (atof(argv[2]));


    /* The discard of const (FpilGetTelescope) is safe as when
       the first argument is -1, this address is not written too
     */
    slaObs(-1, (char *) FpilGetTelescope(*Instrument), oname, &longit, &lat,
        &hm);
    if (strcmp(oname, "?") == 0) {
        sprintf(interp->result,
            "%s:Telescope \"%s\" is not recognized by the slaObs routine.",
            argv[0], FpilGetTelescope(*Instrument));
        strcat(interp->result, "  This must be fixed");
        return TCL_ERROR;
    }

    SdsFind(id, "fieldData", &idf, &status);
    ArgGetd(idf, "appRa", &cra, &status);
    ArgGetd(idf, "configMjd", &mjd, &status);

    lst = slaGmst(mjd) - longit + slaEqeqx(mjd);
    mjd = mjd + (slaDrange(cra - lst + ha)) / D2PI / 1.0027379;
    ChangeMjd (idf, mjd, &status);
    slaDd2tf(0, mjd - floor(mjd), sign, ihmsf);
    slaDjcl(mjd, &iy, &im, &idd, &f, &jstat);

    if (argc < 3) {
        sprintf(buffer,
            "UT set to %.4d/%.2d/%.2d  %.2d:%.2d:%.2d to put field on meridian",
            iy, im, idd, ihmsf[0], ihmsf[1], ihmsf[2]);
/*      DEBUG(
         "UT set to %.4d/%.2d/%.2d  %.2d:%.2d:%.2d to put field on meridian\n",
            iy,im,idd,ihmsf[0],ihmsf[1],ihmsf[2]);
            */
    } else {
        sprintf(buffer,
          "UT set to %.4d/%.2d/%.2d %.2d:%.2d:%.2d to put field at HA %f hours",
            iy, im, idd, ihmsf[0], ihmsf[1], ihmsf[2], atof(argv[2]));
/*
       DEBUG(
       "UT set to %.4d/%.2d/%.2d %.2d:%.2d:%.2d to put field at HA %f hours\n",
             iy,im,idd,ihmsf[0],ihmsf[1],ihmsf[2],atof(argv[2]));
             */
    }
    ConfMessage(interp, buffer);

    SdsFreeId(idf, &ignore);

    if (status == STATUS__OK) {
        return TCL_OK;
    } else {
        MessGetMsg(status, 0, sizeof(buff), buff);
        Tcl_AppendResult(interp, "Error setting UT to meridian value\n",
            (char *) NULL);
        Tcl_AppendResult(interp, buff, (char *) NULL);
        return TCL_ERROR;
    }
}


/*  ------------------------------------------------------------------------- */

/*
 *                        S e t  U T  { }
 */
 
static int ConfSetUT(ClientData clientData DUNUSED, Tcl_Interp * interp,
    int argc DUNUSED, char *argv[])
    
/*+
 *
 *  Command name:
 *    SetUT
 *
 *  Function:
 *    Set the UT Date of a configuration structure
 *  
 *  Call:
 *    SetUT id year month day
 *
 *  Description:
 *    Set the UT date in a configuration structure to a specified value.
 *   
 *   Parameters:
 *
 *    (>) id    (Sds id)  Sds Id of configuration structure.
 *    (>) year  (int)  Required date - year
 *    (>) month (int)  Required date - month 
 *    (>) day   (int/float) Required date - day. Note, this can have a
 *              fractional part, to allow a time of day to be specified.
 *
 *
 *   Support:
 *     Jeremy Bailey, AAO
 *
 *   Version Date:
 *     6-Oct-1994    
 *    21-May-2001. Now uses ChangeMjd(), and allows day to specify
 *                 a fractional part. KS.    
 *-
 */
{
    StatusType status = STATUS__OK;
    SdsIdType id, idf = 0;
    double mjd;
    double dd;
    int jstat;
    int iy, im, idd;
    char buff[80];

    id = atol(argv[1]);

    SdsFind(id, "fieldData", &idf, &status);
    iy = atoi(argv[2]);
    im = atoi(argv[3]);
    dd = atof(argv[4]);
    idd = (int) dd;
    slaCldj(iy, im, idd, &mjd, &jstat);
    mjd += dd - (double)idd;
    ChangeMjd (idf, mjd, &status);
    /*  Note - this used to have a + 0.5 added to mjd */

    SdsFreeId(idf, &status);

    if ((status == STATUS__OK) && (jstat == 0)) {
        return TCL_OK;
    } else {
        MessGetMsg(status, 0, sizeof(buff), buff);
        Tcl_AppendResult(interp, "Error setting UT\n", (char *) NULL);
        Tcl_AppendResult(interp, buff, (char *) NULL);
        return TCL_ERROR;
    }
}

/*  ------------------------------------------------------------------------- */

/*
 *                     Fix Historical SDS  { }
 */
 
static int ConfFixHistoricalSDS(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc, char *argv[])
/*+
 *
 *  Command name:
 *    FixHistoricalSDS
 *
 *
 *  Function:
 *    Adjust the SDS data structures to compensate for historical
 *    changes that have occurred.
 *  
 *  Call:
 *    SetHistoricalSDS id
 *
 *  Description:
 *    Correct the SDS data structure to allow for legacies of 
 *    previous versions of configure. 
 *    Presently, the only fix is to adjust old 6df files that
 *    used "X" to indicate unallocated rather than "U"
 *
 *   Parameters:
 *
 *    (>) id   (Sds id)  Sds Id of configuration structure.
 *
 *
 *   Support:
 *     Chris Ramage, AAO
 *
 *   Version Date:
 *    14-Feb-2003
 *-
 */
{
    StatusType status = STATUS__OK;
    SdsIdType id = 0;
    char buff[80];

    if (argc < 2) {
        sprintf(interp->result, "%s:Wrong number of arguments", argv[0]);
        return TCL_ERROR;
    }
    id = atol(argv[1]);


    /* For 6dF (and 2dF) the unallocated fibre type is 'U'
     * For flames, 'U' was used for something else, and 
     * hence flames uses 'X' for unallocated.
     * For a brief period during flames development, 'X' was 
     * used for all instruments, but this introduced a compatability 
     * problem with old SDS files. To prevent the problem, the
     * value was changed back to 'U' for both 6df and 2df.
     *
     * Sadly, a version of configure for 6dF escaped into the
     * wild with the unallocated value set to 'X'. This version
     * has been used for many months, so there are a lot of 
     * Incorrect SDS files out there.
     *
     * To compensate for the above, this routine detects the use
     * of 'X' in a 6dF file and modifies it to 'U'. This enables
     * the latest version of configure to read the incorrect SDS
     * files.
     */
    if (!strcmp(FpilGetInstName(*Instrument), "6dF")) {
        StatusType ignore = STATUS__OK;
        SdsIdType idf = 0, idt = 0;
        char typebuf[200];
        unsigned long typesize = 0;

        SdsFind(id, "objects", &idf, &status);
        SdsFind(idf,"type", &idt, &status);

        SdsGet(idt, sizeof(typebuf), 0, typebuf, &typesize, &status);

        if (status == STATUS__OK) {
            int i;
            char unalloc_char;
            int changemade = 0;

            unalloc_char = FpilUnallocatedType(*Instrument);

            for (i=0; i < typesize; i++) {
                if (typebuf[i] == 'X') {
                    typebuf[i] = unalloc_char;
                    changemade = 1;
                }
            }
 
            if (changemade) {
                SdsPut(idt, typesize, 0, typebuf, &status);
       }
        }
        SdsFreeId(idt, &ignore);
        SdsFreeId(idf, &ignore);
    }
    if (status == STATUS__OK) {
        /*
         * Add the wlens array to the field data structure and the
         * wlenIndex array to the objects and unallocObject arrays.
         * We do this if the field does not have the "wlens" array.
         * 
         */
        StatusType ignore = STATUS__OK;
        SdsIdType fieldId = 0;
        SdsFind(id, "fieldData", &fieldId, &status);
        if (status == STATUS__OK)
        {
            SdsIdType wlenId = 0;
            SdsFind(id, "wlen", &wlenId, &status);
            if (status == SDS__NOITEM)
            {
                wavelensType wlens;
                
                status = STATUS__OK;
                /*
                 * Create a wavelength structure and put it into
                 * the field structure.
                 */
                tdFwlensInit(&wlens, &status);
                tdFwlensToSds(&wlens, fieldId, &status);
                
                if (status == STATUS__OK)
                {
                    SdsIdType objId = 0;
                    SdsIdType tid = 0;
                    SdsFind(id, "objects", &objId, &status);
                    if (status == STATUS__OK)
                    {
                        char tname[SDS_C_NAMELEN];
                        SdsCodeType tcode;
                        unsigned long dims[7];
                        long ndims;
                        unsigned long actlen;
                        short *data;
                        unsigned i;
                        SdsFind(objId, "ra", &tid, &status);
                        SdsInfo(tid, tname, &tcode, &ndims, dims, &status);
                        SdsFreeId(tid, &status);
                        SdsNew(objId, "wlenIndex",  0, NULL, SDS_SHORT, 1, 
                               dims, &tid, &status);
                        SdsPointer(tid,  (void **)&data, &actlen, &status);
                        for (i = 0; (i < dims[0])&&(status == STATUS__OK); ++i)
                            data[i] = -1;
                        SdsFreeId(tid, &status);
                        SdsFreeId(objId, &status);
                    }
                    else if (status == SDS__NOITEM)
                    {
                        status = STATUS__OK;
                    }
                }
                if (status == STATUS__OK)
                {
                    SdsIdType unallocId = 0;
                    SdsIdType tid = 0;
                    SdsFind(id, "unallocObject", &unallocId, &status);
                    if (status == STATUS__OK)
                    {
                        char tname[SDS_C_NAMELEN];
                        SdsCodeType tcode;
                        unsigned long dims[7];
                        long ndims;
                        unsigned long actlen;
                        short *data;
                        unsigned i;
                        SdsFind(unallocId, "ra", &tid, &status);
                        SdsInfo(tid, tname, &tcode, &ndims, dims, &status);
                        SdsFreeId(tid, &status);
                        SdsNew(unallocId, "wlenIndex",  0, NULL, SDS_SHORT, 1, 
                               dims, &tid, &status);
                        SdsPointer(tid,  (void **)&data, &actlen, &status);
                        for (i = 0; (i < dims[0])&&(status == STATUS__OK); ++i)
                            data[i] = -1;
                        SdsFreeId(tid, &status);
                        SdsFreeId(unallocId, &status);
                    }
                    else if (status == SDS__NOITEM)
                    {
                        status = STATUS__OK;
                    }
                }
            }
            else
                SdsFreeId(wlenId, &status);
            SdsFreeId(fieldId, &ignore);
        } else {
            fprintf(stderr,"Failed to find field in SDS structure\n");
        }
        
    } /* status = STATUS__OK */

    if (status == STATUS__OK) {
        return TCL_OK;
    } else {
        MessGetMsg(status, 0, sizeof(buff), buff);
        Tcl_AppendResult(interp, "Error Fixing Historical SDS issues\n",
            (char *) NULL);
        Tcl_AppendResult(interp, buff, (char *) NULL);
        return TCL_ERROR;
    }
}

/*  ------------------------------------------------------------------------- */

/*
 *                        N e x t  D a y  { }
 */
 
static int ConfNextDay(ClientData clientData DUNUSED, Tcl_Interp * interp,
    int argc DUNUSED, char *argv[])

/*+
 *  Command name:
 *    NextDay
 *
 *  Function:
 *    Set the Date of a configuration structure to the next day
 *  
 *  Call:
 *    NextDay id 
 *
 *  Description:
 *    Set the UT date in a configuration structure the next day
 *    (i.e. increment it by one) .
 *   
 *   Parameters:
 *
 *    (>) id   (Sds id)  Sds Id of configuration structure.
 *
 *   Support:
 *     Jeremy Bailey, AAO
 *
 *   Version Date:
 *     6-Oct-1994    
 *    21-May-2001. Now uses ChangeMjd(). KS  
 *-
 */
{
    StatusType status = STATUS__OK;
    SdsIdType id, idf = 0;
    double mjd;
    char buff[80];

    id = atol(argv[1]);

    SdsFind(id, "fieldData", &idf, &status);
    ArgGetd(idf, "configMjd", &mjd, &status);
    mjd = mjd + 1;
    ChangeMjd (idf, mjd, &status);

    SdsFreeId(idf, &status);

    if (status == STATUS__OK) {
        return TCL_OK;
    } else {
        MessGetMsg(status, 0, sizeof(buff), buff);
        Tcl_AppendResult(interp, "Error setting UT\n", (char *) NULL);
        Tcl_AppendResult(interp, buff, (char *) NULL);
        return TCL_ERROR;
    }
}


/*  ------------------------------------------------------------------------- */

/*
 *                        G e t  P l a t e  { }
 */
 
static int ConfGetPlate(ClientData clientData DUNUSED, Tcl_Interp * interp,
    int argc DUNUSED, char *argv[]DUNUSED)

/*+
 *  Command name:
 *    GetPlate
 *
 *  Function:
 *    Get Field Plate
 *  
 *  Call:
 *    SetPlate
 *
 *  Description:
 *    Return thenumber of the field plate currently being configured
 *   
 *   Parameters: None
 *
 *   Returns:
 *      The field plate (0 or 1)
 *
 *   Support:
 *     Jeremy Bailey, AAO
 *
 *   Version Date:
 *     17-Nov-1994    
 *-
 */
{
  /* 6dF displays the plate numbers to the user as 1 and 2,
   * rather than 0 and 1, but 0 and 1 are used internally.
   */
    if (!strcmp(FpilGetInstName(*Instrument), "6dF")) 
    {
        sprintf(interp->result, "%d", CurrentField + 1);
    }
    else
    {
        sprintf(interp->result, "%d", CurrentField);
    }
    return TCL_OK;
}


/*  ------------------------------------------------------------------------- */

/*
 *                        G e t  W a v e l e n g t h  { }
 */
 
static int ConfGetWavelength(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc DUNUSED, char *argv[]DUNUSED)

/*+
 *  Command name:
 *    GetWavelength
 *
 *  Function:
 *    Get configuration wavelength
 *  
 *  Call:
 *    GetWavelength
 *
 *  Description:
 *    Return the current configuration wavelength in Angstroms
 *   
 *   Parameters: None
 *
 *   Returns:
 *      The configuration wavelength
 *
 *   Support:
 *     Jeremy Bailey, AAO
 *
 *   Version Date:
 *     1-Nov-1994    
 *-
 */
{
    sprintf(interp->result, "%8.2f", MICRONS_TO_ANGSTROMS * gwave);
    return TCL_OK;
}

/*  ------------------------------------------------------------------------- */

/*
 *                   S e t  W a v e l e n g t h  { }
 */
 
static int ConfSetWavelength(ClientData clientData DUNUSED,
    Tcl_Interp * interp DUNUSED, int argc DUNUSED, char *argv[])

/*+
 *  Command name:
 *    SetWavelength
 *
 *  Function:
 *    Set configuration wavelength
 *  
 *  Call:
 *    SetWavelength lambda
 *
 *  Description:
 *    Set the current configuration wavelength in Angstroms
 *   
 *   Parameters:
 *    (>) lambda   (double)  Configuration wavelength in Angstroms
 *
 *   Support:
 *     Jeremy Bailey, AAO
 *
 *   Version Date:
 *     15-Nov-1994    
 *-
 */
{
    /*  gwave is the global variable holding the current wavelength */

    double value;
    if ((argc != 2) && (argc != 3)) {
        interp->result = "wrong # args";
        return TCL_ERROR;
    }

    value = (atof(argv[1])) ;
    if ((value < MIN_WAVELENGTH)||(value > MAX_WAVELENGTH))
    {
        sprintf(interp->result, "Invalid wavelength value %.1f, must be between %d and %d (angstroms)",
                value, MIN_WAVELENGTH, MAX_WAVELENGTH);
        return TCL_ERROR;
    }
    gwave = value / MICRONS_TO_ANGSTROMS;
    gwave2 = gwave;

    if (argc == 3) {
        value = (atof(argv[2]));
        if ((value < MIN_WAVELENGTH)||(value > MAX_WAVELENGTH))
        {
            sprintf(interp->result, "Invalid wavelength value %.1f, must be between %d and %d (angstroms)",
                    value, MIN_WAVELENGTH, MAX_WAVELENGTH);
        return TCL_ERROR;
        }
        gwave2 = value / MICRONS_TO_ANGSTROMS;
    }

    return TCL_OK;
}


/*  ------------------------------------------------------------------------- */

/*
 *                     S e t  R e f  P a r s  { }
 */
 
static int ConfSetRefPars(ClientData clientData DUNUSED,
    Tcl_Interp * interp DUNUSED, int argc DUNUSED, char *argv[])

/*+
 *  Command name:
 *    SetRefPars
 *
 *  Function:
 *    Set Refraction Parameters
 *  
 *  Call:
 *    SetRefPars temp press humid
 *
 *  Description:
 *    Return the current values of the refraction parameters
 *   
 *   Parameters:
 *        temp    (double)  Temperature (K)
 *        press   (double)  Atmospheric pressure (mb)
 *        humid   (double)  Relative humidity 
 *
 *   Support:
 *     Jeremy Bailey, AAO
 *
 *   Version Date:
 *     16-Nov-1994    
 *-
 */
{
    /*  Set the global variables holding the parameters */
    
    gtemp = atof(argv[1]);
    gpress = atof(argv[2]);
    ghumid = atof(argv[3]);

    return TCL_OK;
}


/*  ------------------------------------------------------------------------- */

/*
 *                     G e t  R e f  P a r s  { }
 */
 
static int ConfGetRefPars(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc DUNUSED, char *argv[])

/*+
 *
 *  Command name:
 *    GetRefPars
 *
 *  Function:
 *    Get Refraction Parameters
 *  
 *  Call:
 *    GetRefPars temp press humid
 *
 *  Description:
 *    Return the current values of the refraction parameters
 *   
 *
 *   Parameters:
 *        temp    (Char)  Name of variable to receive the temperature
 *        press   (Char)  Name of variable to receive the atmospheric pressure
 *        humid   (Char)  Name of variable to receive the humidity
 *
 *   Support:
 *     Jeremy Bailey, AAO
 *
 *   Version Date:
 *     16-Nov-1994    
 *-
 */
{
    char buffer[40];

    sprintf(buffer, "%6.2f", gtemp);
    Tcl_SetVar(interp, argv[1], buffer, 0);
    if (gpress >= 1000.0)
        sprintf(buffer, "%7.2f", gpress);
    else
        sprintf(buffer, "%6.2f", gpress);
    Tcl_SetVar(interp, argv[2], buffer, 0);
    sprintf(buffer, "%8.4f", ghumid);
    Tcl_SetVar(interp, argv[3], buffer, 0);

    return TCL_OK;
}

/*  ------------------------------------------------------------------------- */

/*
 *                     S e t  T e l  P a r s  { }
 */
 
static int ConfSetTelPars(ClientData clientData DUNUSED,
    Tcl_Interp * interp DUNUSED, int argc DUNUSED, char *argv[])

/*+
 *
 *  Command name:
 *    SetTelPars
 *
 *  Function:
 *    Set Telescope Parameters
 *  
 *  Call:
 *    SetTelPars params
 *
 *  Description:
 *    Set the current values of the telescope pointing model parameters
 *   
 *   Parameters:
 *       (>) params (list) A list giving the telescope parameter values.  This
 *                         should be a list of real numbers, the number of
 *                         items being the same as the number returned in each
 *                         sub-list by GetTelPars.
 *
 *   Support:
 *     Tony Farrell, AAO
 *
 *   Version Date:
 *     14-Mar-2000
 *-
 */
{
    int listArgc;
    char **listArgv;
    unsigned int i;
    double *data;

    if (argc != 2) {
        sprintf(interp->result, "%s:Wrong number of arguments", argv[0]);
        return TCL_ERROR;
    }

/*
 *  Split the argument up into a list elements and confirm we have the
 *  correct number of elements.
 */
    if (Tcl_SplitList(interp, argv[1], &listArgc, &listArgv) != TCL_OK) {
        return TCL_ERROR;
    }
    if ((unsigned) listArgc != TelModelNumParams) {
        sprintf(interp->result,
            "%s:List has wrong number list elements, %d instead of %d",
            argv[0], listArgc, TelModelNumParams);
        return TCL_ERROR;
    }

/*
 *  Allocate temp storage space for the data, so that we can catch conversion 
 *  errors without clobbering the existing data items.
 */
    data = malloc(sizeof(double) * TelModelNumParams);

/*
 *  Work through converting each element.  IF we have an error, fall out
 *  at this point.
 */
    for (i = 0; i < TelModelNumParams; ++i) {
        if (Tcl_GetDouble(interp, listArgv[i], &data[i]) != TCL_OK) {
            free(data);
            return TCL_ERROR;
        }
    }
/*
 *  Copy the data into the correct spot.
 */
    for (i = 0; i < TelModelNumParams; ++i) {
        TelModelParams[i] = data[i];
    }

    free(data);
    return TCL_OK;
}

/*  ------------------------------------------------------------------------- */

/*
 *                     G e t  T e l  P a r s  { }
 */
 
static int ConfGetTelPars(ClientData clientData DUNUSED, Tcl_Interp * interp,
    int argc DUNUSED, char *argv[])

/*+
 *
 *  Command name:
 *    GetTelPars
 *
 *  Function:
 *    Get Telescope Parameters
 *  
 *  Call:
 *    GetTelPars 
 *
 *  Description:
 *    Return the current values of the telescope pointing model parameters
 *    as a list containing three lists.  The first of the three lists 
 *    contains the current telescope parameter values.  The second contains
 *    the simple names of each telescope parameter.  The third contains 
 *    a more descriptive name.  All three lists have the same number
 *    of elements.
 *
 *   Returns:
 *
 *   Support:
 *     Tony Farrell, AAO
 *
 *   Version Date:
 *     14-Mar-2000
 *-
 */
{
    unsigned int NumParams;
    const char *const *ParamNames;
    const char *const *ParamDescrs;
    const double *ParamDefaults;
    StatusType status = STATUS__OK;
    register unsigned i;

    Tcl_DString Names;
    Tcl_DString Descrs;
    Tcl_DString Values;
    Tcl_DString CommandResult;



    FpilModelParams(TelModel, &NumParams,
        &ParamNames, &ParamDescrs, &ParamDefaults, &status);

    if (status != STATUS__OK) {
        char mbuf[100];
        MessGetMsg(status, 0, sizeof(mbuf), mbuf);
        sprintf(interp->result,
            "%s:Error getting telescope model parameter details - %s\n",
            argv[0], mbuf);
        return TCL_ERROR;
    }

/*
 *  Validate number of telescope model parameters.
 */
    if (NumParams != TelModelNumParams) {
        fprintf(stderr, "%s:%d Assertion failed\n", __FILE__, __LINE__);
        fprintf(stderr, "Telescope model parameter count invalid\n");
        fprintf(stderr, "Talk to tjf@aaoepp.aao.gov.au\n");
        exit(-1);
    }

/*
 *  Initialise the various value strings
 */
    Tcl_DStringInit(&Names);
    Tcl_DStringInit(&Descrs);
    Tcl_DStringInit(&Values);
    Tcl_DStringInit(&CommandResult);
/*
 *  Put all the values into the strings
 */
    for (i = 0; i < TelModelNumParams; ++i) {
        char buffer[30];
        sprintf(buffer, "%g", TelModelParams[i]);
        Tcl_DStringAppendElement(&Values, buffer);
        Tcl_DStringAppendElement(&Names, ParamNames[i]);
        Tcl_DStringAppendElement(&Descrs, ParamDescrs[i]);
    }

/*
 *  Append the three sub-lists to the result string
 */
    Tcl_DStringAppendElement(&CommandResult, Tcl_DStringValue(&Values));
    Tcl_DStringAppendElement(&CommandResult, Tcl_DStringValue(&Names));
    Tcl_DStringAppendElement(&CommandResult, Tcl_DStringValue(&Descrs));
/*
 *  And put the result string  in the correct place.
 */
    Tcl_DStringResult(interp, &CommandResult);

/*
 *  Free resources.
 */
    Tcl_DStringFree(&Names);
    Tcl_DStringFree(&Descrs);
    Tcl_DStringFree(&Values);
    Tcl_DStringFree(&CommandResult);


    return TCL_OK;
}


/*  ------------------------------------------------------------------------- */

/*
 *                     S e t  P l a t e  { }
 */
 
static int ConfSetPlate(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc DUNUSED, char *argv[])

/*+
 *
 *  Command name:
 *    SetPlate
 *
 *  Function:
 *    Select the a new field plate
 *  
 *  Call:
 *    SetField field
 *
 *  Description:
 *    Select the field plate to operate on.  We validate the plate number and
 *    then access the constants details and telescope model appropiate for 
 *    this field plate
 *   
 *   Parameters:
 *    (>) plate       (int) Field Plate number, 0 to the number
 *                          supported by the instrument.
 *
 *   Support:
 *     Tony Farrell, AAO
 *
 *-
 */
{
    unsigned NewField;
    unsigned NewFieldDisplay; /* Untranslated value for user display. */
    char buffer[100];
    StatusType status = STATUS__OK;

    SdsIdType NewFieldDetails = 0;
    FpilModelType *NewTelModel = 0;

    if (argc != 2) {
        sprintf(interp->result, "%s:Wrong number of arguments", argv[0]);
        return TCL_ERROR;
    }

    NewField = atol(argv[1]);
    NewFieldDisplay = NewField;

    /* For 6dF, we need to translate from the displayed
     * field plate 1 and 2 to the internal 0 and 1.
     */
    if (!strcmp(FpilGetInstName(*Instrument), "6dF")) 
    {
      NewField -= 1;
    }

    if (NewField >= FpilGetNumFields(*Instrument)) {
        sprintf(interp->result,
            "%s:Field %d not valid, %s has only %d field plates\n",
            argv[0],
            NewFieldDisplay,
            FpilGetInstName(*Instrument), FpilGetNumFields(*Instrument));
        return TCL_ERROR;
    }
/*
 *  Field number valid.  Access details and read telescope model
 *  details for this field.
 */
    FpilConFieldId(InstConstDetails, NewField, &NewFieldDetails, &status);


    FpilModelAccess(*Instrument, ConfigFilesDir, ExecutableDir,
        ConfRptTelModelFile, interp, NewField, &NewTelModel, &status);


    if (status == STATUS__OK) {
/*
 *      Success.  Delete the old field details and copy in the new to
 *      the module level variables.
 */
        sprintf(buffer, "New field plate is %d", NewFieldDisplay);
        ConfMessage(interp, buffer);

        FpilModelFree(TelModel, &status);
        TelModel = NewTelModel;

        SdsFreeId(InstFieldDetails, &status);
        InstFieldDetails = NewFieldDetails;

        CurrentField = NewField;
        return TCL_OK;
    } else {
/*
 *      Error - tidy up, leaving current field as the valid field
 */
        StatusType ignore = STATUS__OK;
        char mbuf[100];

        if (NewTelModel)
            FpilModelFree(NewTelModel, &ignore);
        if (NewFieldDetails)
            SdsFreeId(NewFieldDetails, &ignore);

        MessGetMsg(status, 0, sizeof(mbuf), mbuf);
        sprintf(interp->result, "%s:Error select field plate %d - %s\n",
            argv[0], NewFieldDisplay, mbuf);
        return TCL_ERROR;

    }

}

/*  ------------------------------------------------------------------------- */

/*
 *                     P i v o t s  I d  { }
 */
 
static int ConfPivotsId(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc DUNUSED, char *argv[]DUNUSED)
/*+
 *
 *  Command name:
 *    PivotsId
 *
 *  Function:
 *    Return an Sds identifier for the field constants.
 *
 *  Call:
 *    PivotsId
 *
 *  Description:
 *    Returns an SDS id refering to the structure which contains the
 *    contants details for the current field.
 *   
 *    This id must not be free-ed, the structure must not be modified 
 *    or deleted.
 *
 *   Parameters:  none.
 *
 *   Support:
 *     Tony Farrell, AAO
 *
 *-
 */
{
    sprintf(interp->result, "%ld", (long int) InstFieldDetails);
    return TCL_OK;
}

/*  ------------------------------------------------------------------------- */

/*
 *  MakeSkyvals 
 * 
 * Comment by TJF, 17-Mar-2000.  I think this command counts the number
 * of sky values allocated to each spectrograph, and puts the
 * items allocSkyn (n = spectrograph number) into the field data with 
 * these values.
 *
 * Note -> Currently called directly by ConfDoAllocation(), which assumes
 * we have the same arguments as ConfDoAllocation().
 *
 */
static int ConfMakeSkyvals(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc DUNUSED, char *argv[])
{
    SdsIdType id, oid, idf;
    unsigned i;
    short *specSkys;
    char type;
    unsigned NumPivots = FpilGetNumPivots(*Instrument);
    StatusType status = STATUS__OK;


    id = atol(argv[1]);

/*
 *  Allocate space for counting the sky values.  Initialise them to zero.
 */
    specSkys = malloc(sizeof(*specSkys) * InstNumFibreTypes);

    for (i = 0; i < InstNumFibreTypes; ++i)
        specSkys[i] = 0;

/*
 *  Find the object data.
 */
    SdsFind(id, "fieldData", &idf, &status);

    SdsFind(id, "objects", &oid, &status);

/*
 *  Search through the pivots, increment the number of skys assocated
 *  with each fibre type. Note that the type will be the type of the
 *  target allocated to that pivot, or will be the special 'unallocated'
 *  code.
 */
    for (i = 0; i < NumPivots; i++) {
        ConfArrayGetc(oid, "type", i, &type, &status);
        if (FpilIsTargetSky(*Instrument, type)) {
            specSkys[ConfWhichSpec((int) i, 0)]++;
        }
    }
/*
 *  Now add the data to the field data.
 */
    for (i = 0; i < InstNumFibreTypes; ++i) {
        char buffer[20];
        sprintf(buffer, "allocSky%d", i + 1);
        ArgPuts(idf, buffer, specSkys[i], &status);
    }
/*
 *  Tidy up
 */

    SdsFreeId(idf, &status);
    SdsFreeId(oid, &status);
    free(specSkys);

    sprintf(interp->result, "OK");


    return TCL_OK;
}

/*  ------------------------------------------------------------------------- */

/*
 *                     C o n v e r t  X y  { }
 */
 
static int ConfConvertXy(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc DUNUSED, char *argv[])

/*+
 *
 *  Command name:
 *    ConvertXy
 *
 *  Function:
 *    Calculate xy from Ra and Dec
 *  
 *  Call:
 *    ConvertXy id 
 *
 *  Description:
 *    Calculate the x,y corresponding to the RA and Dec values in
 *    a configuration structure, for the current values of the telescope
 *    and refraction parameters.
 *   
 *   Parameters:
 *
 *    (>) id   (Sds id)  Sds Id of configuration structure.
 *
 *   Returns:
 *      "OK" if succesful, "error" if the conversion failed
 *
 *   Support:
 *     Jeremy Bailey, AAO
 *
 *   Version Date:
 *     6-Oct-1994
 *    21-May-2001. Removed epoch and getting of appEpoch. Made gmap a local
 *                 variable. KS.
 *    16-Aug-2001. Now records TelModelParams in TelModelParamsUsed. KS. 
 *    17-Oct-2002. FpilModelCvtInit() now takes two wavelength parameters. KS.  
 *-
 */
{
    SdsIdType id, idf, idtemp;
    double cra, cdec, mjd;
    double gmap[21];                   /* Mean to apparent parameters  */
    StatusType status = STATUS__OK;
    int i;
    wavelensType wlens;                /* Wavelenghts - when specified explictly */

    id = atol(argv[1]);
    
    /*
     *  Get field centre details
     */

    SdsFind(id, "fieldData", &idf, &status);
    ArgGetd(idf, "appRa", &cra, &status);
    ArgGetd(idf, "appDec", &cdec, &status);
    ArgGetd(idf, "configMjd", &mjd, &status);

    /*
     *  Get the specific wavelength values
     */
    tdFwlensFromSds(idf, &wlens, &status);

    SdsFreeId(idf, &status);
    
    /*  We sometimes need to know if the parameters we may be about to use
     *  are different to those used the last time the calculation was done.
     *  (This is because if so, it may invalidate any allocation, and the
     *  user may have to confirm this.) So we record the parameters used.
     *  At the moment, we only record the telescope model parameters, but
     *  perhaps we should record everything - this was introduced with
     *  the ARGUS position angle code, where it was important to know if
     *  the position angle had changed from the last time.
     */
   
    for (i = 0; i < TelModelNumParams; i++) {
        TelModelParamsUsed[i] = TelModelParams[i];
    }

    /*
     *  Set up x,y conversion parameters. The wavelength values in this
     *  call shouldn't matter because ConfDoXYConversion() always resets
     *  them to the values appropriate to the fibre type. Indeed, this
     *  FpilModelCvtInit() initialisation call may be unnecessary.
     */

    slaMappa(2000.0, mjd, gmap);

    FpilSetObsWavelength(*Instrument,gwave);
    FpilSetObsWavelength2(*Instrument,gwave2);

    FpilModelCvtInit(TelModel, mjd, 0.0, gtemp, gpress, ghumid, gwave, gwave,
        TelModelParams, &status);

    SdsFind(id, "unallocGuide", &idtemp, &status);
    if (status == SDS__NOITEM) {
        ErsAnnul(&status);
    } else {
        ConfDoXyConversion(idtemp, cra, cdec, mjd, gmap, &wlens, &status);
        SdsFreeId(idtemp, &status);
    }

    if (status == STATUS__OK) {
        SdsFind(id, "unallocObject", &idtemp, &status);
        if (status == SDS__NOITEM) {
            ErsAnnul(&status);
        } else {
            ConfDoXyConversion(idtemp, cra, cdec, mjd, gmap, &wlens, &status);
            SdsFreeId(idtemp, &status);
        }
    }

    if (status == STATUS__OK) {
        SdsFind(id, "objects", &idtemp, &status);
        if (status == SDS__NOITEM) {
            ErsAnnul(&status);
        } else {
            ConfDoXyConversion(idtemp, cra, cdec, mjd, gmap, &wlens, &status);
            SdsFreeId(idtemp, &status);
        }
    }

    if (status == STATUS__OK) {
        sprintf(interp->result, "OK");
        return TCL_OK;
    } else {
        char buff[200];
        MessGetMsg(status, 0, sizeof(buff), buff);
        Tcl_AppendResult(interp, "\nError in Ra,Dec to x,y conversion\n",
            (char *) NULL);
        Tcl_AppendResult(interp, buff, (char *) NULL);
        return TCL_ERROR;

    }
}

/*  ------------------------------------------------------------------------- */

/*
 *                     P a r a m s  C h a n g e d  { }
 */
 
static int ConfParamsChanged(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc DUNUSED, char *argv[] DUNUSED)

/*+
 *
 *  Command name:
 *    ParamsChanged
 *
 *  Function:
 *    See if parameters used by ConvertXY have changed.
 *  
 *  Call:
 *    ParamsChanged id
 *
 *  Description:
 *    Checks to see if the telescope model parameters used by ConvertXY
 *    have changed since the last time ConvertXY was used. If this is the
 *    case, using ConvertXY again may invalidate any allocation that has
 *    been made, and the user may need to be warned of this.
 *   
 *   Parameters: 
 *      (>) id   (Sds id)  Sds Id of configuration structure.
 *
 *   Returns:
 *      "1" if the parameters have changed, "0" if not.
 *
 *   Support:
 *     Keith Shortridge, AAO
 *
 *   History:
 *    16-Aug-2001. Original version KS.   
 *-
 */
{
    SdsIdType id, idf;
    double cra, cdec, mjd;
    StatusType status = STATUS__OK;
    int i;
    short Changed;

    /*  This code gets the central ra and dec and mjd, on the grounds
     *  that perhaps we ought to test them as well. However, for the 
     *  moment we don't actually use them. We should perhaps also
     *  use the humidity and other parameters?
     */
     
    id = atol(argv[1]);
    
    /*
     *  Get field centre details
     */

    SdsFind(id, "fieldData", &idf, &status);
    ArgGetd(idf, "appRa", &cra, &status);
    ArgGetd(idf, "appDec", &cdec, &status);
    ArgGetd(idf, "configMjd", &mjd, &status);
    SdsFreeId(idf, &status);
    
    /*  For the moment, we content ourselves with testing the telescope
     *  model parameters.
     */
    
    Changed = 0;
    for (i = 0; i < TelModelNumParams; i++) {
        if (TelModelParamsUsed[i] != TelModelParams[i]) Changed = 1;
    }

    sprintf(interp->result, "%d",Changed);
    
    return TCL_OK;
}

/*  ------------------------------------------------------------------------- */

/*
 *                     R a  2  S t r i n g  { }
 */
 
static int ConfRa2string(ClientData clientData DUNUSED, Tcl_Interp * interp,
    int argc DUNUSED, char *argv[])

/*+
 *
 *  Command name:
 *    Ra2string
 *
 *  Function:
 *    Convert an RA in radians to a string 
 *  
 *  Call:
 *    Ra2string ra
 *
 *  Description:
 *    Convert an RA in radians to a string in Hours, Minutes and Seconds
 *   
 *   Parameters:
 *
 *    (>) ra   (double)  RA in radians.
 *
 *   Returns:
 *     The string representation of the RA in hours, minutes and seconds
 *
 *   Support:
 *     Jeremy Bailey, AAO
 *
 *   Version Date:
 *     6-Oct-1994    
 *-
 */
{
    double d;
    int ihmsf[4];
    char sign[1];

    d = atof(argv[1]);
    slaDr2tf(2, d, sign, ihmsf);
    sprintf(interp->result, "%.2d %.2d %.2d.%.2d", ihmsf[0], ihmsf[1], ihmsf[2],
        ihmsf[3]);
    return TCL_OK;
}

/*  ------------------------------------------------------------------------- */

/*
 *                     D e c  2  S t r i n g  { }
 */
 
static int ConfDec2string(ClientData clientData DUNUSED, Tcl_Interp * interp,
    int argc DUNUSED, char *argv[])

/*+
 *
 *  Command name:
 *    Dec2string
 *
 *  Function:
 *    Convert a Dec in radians to a string 
 *  
 *  Call:
 *    Dec2string dec
 *
 *  Description:
 *    Convert a Dec in radians to a string in Degrees, Minutes and Seconds
 *   
 *   Parameters:
 *
 *    (>) dec  (double)  Declination in radians
 *
 *   Returns:
 *     The string representation of the Dec in degrees, minutes and seconds
 *
 *   Support:
 *     Jeremy Bailey, AAO
 *
 *   Version Date:
 *     6-Oct-1994    
 *-
 */
{
    double d;
    int ihmsf[4];
    char sign[1];

    d = atof(argv[1]);
    slaDr2af(1, d, sign, ihmsf);
    sprintf(interp->result, "%c%.2d %.2d %.2d.%.1d", sign[0],
        ihmsf[0], ihmsf[1], ihmsf[2], ihmsf[3]);
    return TCL_OK;
}

/*  ------------------------------------------------------------------------- */

/*
 *                     M j d  2  D a t e  { }
 */
 
static int ConfMjd2Date(ClientData clientData DUNUSED, Tcl_Interp * interp,
    int argc DUNUSED, char *argv[])

/*+
 *  Command name:
 *    Mjd2Date
 *
 *  Function:
 *    Convert an MJD to a string 
 *  
 *  Call:
 *    Mjd2Date mjd
 *
 *  Description:
 *    Convert a Modified Julian Date to a string giving the date in
 *     the form Year, Month, Day
 *   
 *   Parameters:
 *
 *    (>) mjd  (double)  Modified Julian Date
 *
 *   Returns:
 *     The string representation of the MJD as year, month, day
 *
 *   Support:
 *     Jeremy Bailey, AAO
 *
 *   Version Date:
 *     6-Oct-1994    
 *-
 */
{
    double d;
    int iy, im, id;
    double f;
    int j;

    d = atof(argv[1]);
    slaDjcl(d, &iy, &im, &id, &f, &j);
    sprintf(interp->result, "%.4d/%.2d/%.2d", iy, im, id);
    return TCL_OK;
}

/*  ------------------------------------------------------------------------- */

/*
 *                     M j d  2  T i m e  { }
 */
 
static int ConfMjd2Time(ClientData clientData DUNUSED, Tcl_Interp * interp,
    int argc DUNUSED, char *argv[])

/*+
 *  Command name:
 *    Mjd2Time
 *
 *  Function:
 *    Return UT time from an MJD 
 *  
 *  Call:
 *    Mjd2Time mjd
 *
 *  Description:
 *    Convert a Modified Julian Date to a string giving the fractional
 *    part of the MJD expressed as a UT time in hours, minutes and seconds.
 *   
 *   Parameters:
 *
 *    (>) mjd  (double)  Modified Julian Date
 *
 *   Returns:
 *     The string representation of the UT as hours, minutes, seconds
 *
 *   Support:
 *     Jeremy Bailey, AAO
 *
 *   Version Date:
 *     6-Oct-1994    
 *-
 */
{
    double d;
    int ihmsf[4];
    char sign[1];

    d = atof(argv[1]);
    slaDd2tf(0, d - floor(d), sign, ihmsf);
    sprintf(interp->result, "%.2d:%.2d:%.2d", ihmsf[0], ihmsf[1], ihmsf[2]);
    return TCL_OK;
}

/*  ------------------------------------------------------------------------- */

static int ConfRaDec2Ha(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc DUNUSED, char *argv[])

/*+
 *  Command name:
 *    RaDec2Ha
 *
 *  Function:
 *    Calculate HA from RA and Dec 
 *  
 *  Call:
 *    RaDec2Ha ra dec
 *
 *  Description:
 *    Calculate the observed Hour Angle corresponding to an apparent
 *    RA and Dec. The conversion is for the time and other parameters
 *    used in the last call to ConvertXy
 *   
 *   Parameters:
 *
 *    (>) ra  (double)  Apparent RA (radians)
 *    (>) dec (double)  Apparent Dec (radians)
 *
 *   Returns:
 *     The string representation of the observed HA as hours, minutes, seconds
 *
 *   Support:
 *     Jeremy Bailey, AAO
 *
 *   Version Date:
 *     6-Oct-1994    
 *-
 */
{
    double r, d;
    int ihmsf[4];
    char sign[1];
    double a, zd, ha, dd, rr;
    StatusType status = STATUS__OK;

    r = atof(argv[1]);
    d = atof(argv[2]);
    FpilModelQuickAppToObs(TelModel, r, d, &a, &zd, &ha, &dd, &rr, &status);
    slaDr2tf(0, ha, sign, ihmsf);
    sprintf(interp->result, "%c%.2d:%.2d:%.2d", sign[0],
        ihmsf[0], ihmsf[1], ihmsf[2]);
    return TCL_OK;
}

/*  ------------------------------------------------------------------------- */

/*
 *                     R a  D e c  2  Z d  { }
 */
 
static int ConfRaDec2Zd(ClientData clientData DUNUSED, Tcl_Interp * interp,
    int argc DUNUSED, char *argv[])

/*+
 *  Command name:
 *    RaDec2Zd
 *
 *  Function:
 *    Calculate ZD from RA and Dec 
 *  
 *  Call:
 *    RaDec2Zd ra dec
 *
 *  Description:
 *    Calculate the observed Zenith Distance corresponding to an apparent
 *    RA and Dec. The conversion is for the time and other parameters
 *    used in the last call to ConvertXy
 *   
 *   Parameters:
 *
 *    (>) ra  (double)  Apparent RA (radians)
 *    (>) dec (double)  Apparent Dec (radians)
 *
 *   Returns:
 *     The string representation of the observed ZD as degrees, minutes, seconds
 *
 *   Support:
 *     Jeremy Bailey, AAO
 *
 *   Version Date:
 *     6-Oct-1994    
 *-
 */
{
    double r, d;
    int ihmsf[4];
    char sign[1];
    double a, zd, ha, dd, rr;
    StatusType status = STATUS__OK;

    r = atof(argv[1]);
    d = atof(argv[2]);
    FpilModelQuickAppToObs(TelModel, r, d, &a, &zd, &ha, &dd, &rr, &status);
    slaDr2af(0, zd, sign, ihmsf);
    sprintf(interp->result, "%.2d:%.2d:%.2d", ihmsf[0], ihmsf[1], ihmsf[2]);
    return TCL_OK;
}

/*  ------------------------------------------------------------------------- */

/*
 *                     M a k e  A l l o c a t e d  { }
 */
 
static int ConfMakeAllocated(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc DUNUSED, char *argv[])

/*+
 *  Command name:
 *    MakeAllocated
 *
 *  Function:
 *    Create the Allocated objects section of a configuration structure 
 *  
 *  Call:
 *    MakeAllocated id pivid
 *
 *  Description:
 *    Create the allocated objects section of a configuration structure
 *    if it does not already exist. The new structure is initialized by
 *    setting all fibres to their parked positions.
 *   
 *   Parameters:
 *
 *    (>) id  (Sds Id)  SDS identifier of configuration structure.
 *    (>) pivid (Sds Id)  SDS identifier of pivots structure.
 *
 *   Support:
 *     Jeremy Bailey, AAO
 *
 *   Version Date:
 *     6-Oct-1994    
 *-
 */
{
    SdsIdType topid, pivid;
    StatusType status = STATUS__OK;
    char buff[100];
    
    topid = atol(argv[1]);
    pivid = atol(argv[2]);
    DoMakeAllocated (topid,pivid,&status);
    if (status == STATUS__OK) {
        return TCL_OK;
    } else {
        MessGetMsg(status, 0, sizeof(buff), buff);
        Tcl_AppendResult(interp, "\nError Making Allocated Objects Structure\n",
            (char *) NULL);
        Tcl_AppendResult(interp, buff, (char *) NULL);
        return TCL_ERROR;
    }
}
    

/*  ------------------------------------------------------------------------- */

/*
 *                     A d d  I n d e x  { }
 */
 
static int ConfAddIndex(ClientData clientData DUNUSED, Tcl_Interp * interp,
    int argc DUNUSED, char *argv[])

/*+
 *  Command name:
 *    AddIndex
 *
 *  Function:
 *    Add the index and allocated components to a configuration structure
 *  
 *  Call:
 *    AddIndex id 
 *
 *  Description:
 *    Create the "index" component in the allocated objects section
 *    and the "allocated" component in the unallocated objects and
 *    guide sections of a configuration structure, if they do not
 *    already exist.
 *   
 *   Parameters:
 *
 *    (>) id  (Sds Id)  SDS identifier of configuration structure.
 *
 *   Support:
 *     Jeremy Bailey, AAO
 *
 *   Version Date:
 *     6-Oct-1994    
 *-
 */
{
    SdsIdType topid, objid, indid, uid, allid, ra_id;
    StatusType status = STATUS__OK;
    char name[NAMELEN];
    SdsCodeType code;
    long ndims;
    unsigned long dims[7];
    unsigned int i;
    short s;
    char c;
    int indexFound = 0;         /* Indicates if we found the "index" item */


    topid = atol(argv[1]);

/*
 * Find the objects structure, and look for "index".
 */
    SdsFind(topid, "objects", &objid, &status);
    if (status != STATUS__OK) {
        sprintf(interp->result, "Error finding objects structure");
        return TCL_ERROR;
    }
    SdsFind(objid, "index", &indid, &status);
    if (status == SDS__NOITEM) {
/*
 *     "index" does not exist.  Create it and initialise the values
 *     to -1.
 */
        ErsAnnul(&status);
        SdsFind(objid, "ra", &ra_id, &status);
        SdsInfo(ra_id, name, &code, &ndims, dims, &status);
        SdsNew(objid, "index", 0, NULL, SDS_SHORT, 1, dims, &indid, &status);
        for (i = 0; i < dims[0]; i++) {
            s = -1;
            SdsPut(indid, sizeof(short), i, &s, &status);
        }
        SdsFreeId(indid, &status);
        SdsFreeId(ra_id, &status);
    } else {
        indexFound = 1;
    }
    SdsFreeId(objid, &status);

/*
 * Find "unallocGuide" and the "allocated" component within.
 */
    SdsFind(topid, "unallocGuide", &uid, &status);
    if (status != STATUS__OK) {
        sprintf(interp->result, "Error finding unallocGuide structure");
        return TCL_ERROR;
    }
    SdsFind(uid, "allocated", &allid, &status);
    if ((status == SDS__NOITEM) && (!indexFound)) {
/*
 *     "allocated" does not exist.  Create it and initialise to 0, indicating
 *     objects are not allocated.
 */
        ErsAnnul(&status);
        SdsFind(uid, "ra", &ra_id, &status);
        SdsInfo(ra_id, name, &code, &ndims, dims, &status);
        SdsNew(uid, "allocated", 0, NULL, SDS_CHAR, 1, dims, &allid, &status);
        for (i = 0; i < dims[0]; i++) {
            c = 0;
            SdsPut(allid, sizeof(char), i, &c, &status);
        }
        SdsFreeId(allid, &status);
        SdsFreeId(ra_id, &status);
    }
    SdsFreeId(uid, &status);
/*
 * Find "unallocObject" and the "allocated" component within.
 */
    SdsFind(topid, "unallocObject", &uid, &status);
    if (status == SDS__NOITEM) {
        /*
         * Failed to find the "unallocObj" structure.  This happens because
         * the 2dF observing software removes this structure (to save space)
         * when preparing it to send to the 2dF robot.  We generate a dummy 
         * version here. (through this does not mean the allocation will
         * succeed.)  This is based on code from tdFconvertToSds(), but that
         * should be broken up so we can access it directly.
         */
        unsigned long  dims[2];
        SdsIdType tmpId = 0;
       status = STATUS__OK;
        dims[1] = 0;
        dims[0] = NAME_LENGTH;
        SdsNew(topid,"unallocObject",0,NULL,SDS_STRUCT,0,NULL,&uid,&status);
        SdsNew(uid,"name",0,NULL,SDS_CHAR,2,dims,&tmpId,&status);
        SdsFreeId(tmpId, &status);
        SdsNew(uid,"ra",0,NULL,SDS_DOUBLE,1,&dims[1],&tmpId,&status);
        SdsFreeId(tmpId, &status);
        SdsNew(uid,"dec",0,NULL,SDS_DOUBLE,1,&dims[1],&tmpId,&status);
        SdsFreeId(tmpId, &status);
        SdsNew(uid,"x",0,NULL,SDS_INT,1,&dims[1],&tmpId,&status);
        SdsFreeId(tmpId, &status);
        SdsNew(uid,"y",0,NULL,SDS_INT,1,&dims[1],&tmpId,&status);
        SdsFreeId(tmpId, &status);
        SdsNew(uid,"type",0,NULL,SDS_CHAR,1,&dims[1],&tmpId,&status);
        SdsFreeId(tmpId, &status);
        SdsNew(uid,"spectrograph",0,NULL,SDS_CHAR,1,&dims[1],&tmpId,&status);
        SdsFreeId(tmpId, &status);
        SdsNew(uid,"priority",0,NULL,SDS_SHORT,1,&dims[1],&tmpId,&status);
        SdsFreeId(tmpId, &status);
        SdsNew(uid,"magnitude",0,NULL,SDS_DOUBLE,1,&dims[1],&tmpId,&status);
        SdsFreeId(tmpId, &status);
        SdsNew(uid,"pId",0,NULL,SDS_SHORT,1,&dims[1],&tmpId,&status);
        SdsFreeId(tmpId, &status);
        dims[0] = COMMENT_LENGTH;
        SdsNew(uid,"comment",0,NULL,SDS_CHAR,2,dims,&tmpId,&status);
        SdsFreeId(tmpId, &status);
        SdsNew(uid,"wlenIndex",0,NULL,SDS_SHORT,1,&dims[1],&tmpId,&status);
        SdsFreeId(tmpId, &status);
        SdsNew(uid,"pmra",0,NULL,SDS_DOUBLE,1,&dims[1],&tmpId,&status);
        SdsFreeId(tmpId, &status);
        SdsNew(uid,"pmdec",0,NULL,SDS_DOUBLE,1,&dims[1],&tmpId,&status);
        SdsFreeId(tmpId, &status);

        if (status != STATUS__OK) {
            sprintf(interp->result, "Error creating unallocObject structure");
            return TCL_ERROR;
        }
       
    }
    else if (status != STATUS__OK) {
        sprintf(interp->result, "Error finding unallocObject structure");
        return TCL_ERROR;
    }
    SdsFind(uid, "allocated", &allid, &status);
    if ((status == SDS__NOITEM) && (!indexFound)) {
/*
 *     "allocated" does not exist.  Create it and initialise to 0, indicating
 *     objects are not allocated.
 */ 
        ErsAnnul(&status);
        SdsFind(uid, "ra", &ra_id, &status);
        SdsInfo(ra_id, name, &code, &ndims, dims, &status);
        SdsNew(uid, "allocated", 0, NULL, SDS_CHAR, 1, dims, &allid, &status);
        for (i = 0; i < dims[0]; i++) {
            c = 0;
            SdsPut(allid, sizeof(char), i, &c, &status);
        }
        SdsFreeId(allid, &status);
        SdsFreeId(ra_id, &status);
    }
    SdsFreeId(uid, &status);

/*
 * If the index and allocated items were not found in the original
 * structure, we need to set them up appropiately.
 */
    if (!indexFound)
        ConfMakeAllocatedIndex(topid, &status);

    if (status == STATUS__OK) {
        return TCL_OK;
    } else {
        char buff[200];
        MessGetMsg(status, 0, sizeof(buff), buff);
        Tcl_AppendResult(interp, "\nError Making Allocated Objects Structure\n",
            (char *) NULL);
        Tcl_AppendResult(interp, buff, (char *) NULL);
        return TCL_ERROR;

    }
}

/*  ------------------------------------------------------------------------- */

/*
 *                     M a r k  F i b r e  T y p e  { }
 */
 
static int ConfMarkFibreType(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc, char *argv[])

/*+
 *  Command name:
 *    MarkFibreType
 *
 *  Function:
 *    Mark a fibre type as in use or not in use.
 *  
 *  Call:
 *    MarkFibreType flagArray
 *
 *  Description:
 *    The C layer (this file, configure.c) maintains a global array that
 *    flags which fibre type is to be used in configurations. This is the
 *    UseFibreTypes array. This is used when an allocation is performed.
 *    However, it is the Tcl/Tk layer that really knows which fibre types
 *    are to be used, and it maintains its own array of flags for each
 *    fibre type. When these are changed, MarkFibreType{} should be 
 *    invoked to keep the C layer's global array (UseFibreTypes) in sync
 *    with the Tcl layer.
 *   
 *   Parameters:
 *
 *    (>) flagArray  (Tcl array variable) This is the name of a Tcl array that
 *                   contains an entry for each fibre type (0 .. 
 *                   InstNumFibreTypes-1).  These entries are logical
 *                   values used to set the values in the UseFibreType array.
 *-
 */
 {
    register unsigned i;
    int *localUseFibreType;
    if (argc != 2) {
        sprintf(interp->result, "%s:Wrong # args", argv[0]);
        return TCL_ERROR;
    }

    /*
     * Malloc space for a local copy of the values.
     */
     
    localUseFibreType = malloc(InstNumFibreTypes * sizeof(*localUseFibreType));

    /*
     * Read each array element.
     */
    for (i = 0; i < InstNumFibreTypes; ++i) {
        char element[10];
        char *stringValue;
        sprintf(element, "%d", i);
        /*
         *  Read the array element
         */
        stringValue = Tcl_GetVar2(interp, argv[1], element, TCL_LEAVE_ERR_MSG);
        if (stringValue == NULL) {
            free(localUseFibreType);
            return TCL_ERROR;
        }

        /*
         * Get the value as a boolean.
         */
        if (Tcl_GetBoolean(interp, stringValue,
                &localUseFibreType[i]) == TCL_ERROR) {
            free(localUseFibreType);
            return TCL_ERROR;
        }
    }

    /*
     * Everything ok, make the value valid.  
     */
    for (i = 0; i < InstNumFibreTypes; ++i) {
        UseFibreType[i] = localUseFibreType[i];
    }
    /*
     * Tidy up and return.
     */
    free(localUseFibreType);
    return TCL_OK;

}

/*  ------------------------------------------------------------------------- */

/*
 *                 C h e c k  T a r g e t  C r i t e r i a  { }
 */
 
static int ConfCheckTargetCriteria(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc, char *argv[])

/*+
 *  Command name:
 *    CheckTargetCriteria
 *
 *  Function:
 *    Sees if a target meets a supplied set of criteria.
 *  
 *  Call:
 *    CheckTargetCriteria Flags NumCriteria TargetType TargetSpect SpectKnown 
 *
 *  Description:
 *    This routine is used as part of the code that highlights targets in
 *    the mimic display if they meet a user-specified set of criteria. It
 *    is little more than an interface to FpilTargetCriteriaMet() and that
 *    routine's comments should be consulted for more details. Note that
 *    the test is based purely on the type of the target. Additional
 *    criteria such as the target priority and whether or not it is allocated
 *    are handled separately. This routine is merely used to abstract the
 *    issues connected with target type.
 *   
 *  Parameters:
 *
 *    (>) Flags       (Tcl array)  Array of criteria flag values
 *    (>) NumCriteria (int)        Number of instrument-dependent criteria
 *    (>) TargetType  (string)     Type code for target to be checked
 *    (>) TargetSpect (string)     Spectrograph code for target to be checked
 *    (>) SpectKnown  (int)        True if the TargetSpect code can be used
 *
 *   Returns:
 *     The number of criteria met by the target.
 *
 *   Support:
 *     Keith Shortridge, AAO
 *
 *-
 */
 
{
    int     CriteriaMet;
    int*    Flags;
    short   FlagsMalloced;
    int     Icrit;
    int     LocalFlagArray[20];
    int     NumCriteria;
    int     ReturnValue;
    int     SpectKnown;
    char    TargetType;
    char    TargetSpect;

    Flags = (int*) NULL;
    FlagsMalloced = FALSE;
    ReturnValue = TCL_OK;
    CriteriaMet = 0;

    if (argc != 6) {
        sprintf(interp->result, "%s:Wrong # args", argv[0]);
        ReturnValue = TCL_ERROR;
        goto Exit;
    }
    NumCriteria = atoi(argv[2]);
    if (NumCriteria > 0) {
        if (NumCriteria <= (int)(sizeof(LocalFlagArray)/sizeof(int))) {
            Flags = LocalFlagArray;
            FlagsMalloced = FALSE;
        } else {

            /*  Malloc space for a local copy of the flag values if the
             *  predefined local flag array is too small.
             */
     
            Flags = malloc(NumCriteria * sizeof(int));
            if (Flags == (int*) 0) {
                sprintf(interp->result, "%s: Unable to malloc flag memory",
                                                                  argv[0]);
                ReturnValue = TCL_ERROR;
                goto Exit;
            }
            FlagsMalloced = TRUE;
        }

        /*  Get the value of each flag array element. */

        for (Icrit = 0; Icrit < NumCriteria; Icrit++) {
            char Element[10];
            char *StringValue;
            sprintf(Element, "%d", Icrit);

            /*  Read the Tcl array element  */

            StringValue = Tcl_GetVar2(interp, argv[1], Element,
                                                   TCL_LEAVE_ERR_MSG);
            if (StringValue == NULL) {
                ReturnValue = TCL_ERROR;
                goto Exit;
            }

            /*  Get the value as a boolean.  */

            if (Tcl_GetBoolean(interp, StringValue,
                                     &Flags[Icrit]) == TCL_ERROR) {
                ReturnValue = TCL_ERROR;
                goto Exit;
            }
        }

        /*  Now get the TargetType, TargetSpect and SpectKnown parameters */

        TargetType = *(argv[3]);
        TargetSpect = *(argv[4]);
        SpectKnown = atoi(argv[5]);

        CriteriaMet = FpilTargetCriteriaMet (*Instrument,Flags,NumCriteria,
                                         TargetType,TargetSpect,SpectKnown);
        
        sprintf(interp->result, "%d",CriteriaMet);
    }

Exit:;

    if (FlagsMalloced) free(Flags);
    return (ReturnValue);

}

/*  ------------------------------------------------------------------------- */

/*
 *                     A l l o c a t e  F i b r e  { }
 */
 
static int ConfAllocateFibre(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc, char *argv[])

/*+
 *  Command name:
 *    AllocateFibre
 *
 *  Function:
 *    Allocate an object to a fibre
 *  
 *  Call:
 *    AllocateFibre id pivid fibre type object
 *
 *  Description:
 *    Allocate an object to a fibre. This command checks that the fibre
 *    and object are not already allocated, and if not copies the object
 *    information to the allocated objects section of the structure. It sets
 *    the index entry for the fibre to point to the object location in the
 *    unallocated structure, and it sets the allocated flag in that structure.
 *
 *    The command does not check the validity of the resulting configuration.
 *    That can be done by a call to DoCheck.
 *   
 *   Parameters:
 *
 *    (>) id  (Sds Id)   SDS identifier of configuration structure.
 *    (>) pivid (Sds Id) SDS identifier of the pivots structure.
 *    (>) fibre (int)    Fibre (pivot) number - starting from one.
 *    (>) type  (char)   "object" or "guide".
 *    (>) object (int)   Object number - starting from one.
 *
 *   Returns:
 *     "OK" if succesful, "Error" otherwise
 *
 *   Support:
 *     Jeremy Bailey, AAO
 *
 *   Version Date:
 *      6-Oct-1994  Original version. JAB.
 *      8-Nov-2001  Now updates fibre type allocation statistics. KS.
 *      3-Sep-2002  Now updates available target allocation statistics. KS.   
 *-
 */
{
    SdsIdType topid, objid, uid, pivid, fid;
    long fibre_no;
    long obj_no;
    StatusType status = STATUS__OK;
    short s;
    char c;
    short IsSky;
    char guide;
    int x, y, xp, yp;
    double theta;
    int fibreType;
    int PivIndex;
    int NumPivTypes;
    int PivTypes[10];

    if (argc != 6) {
        interp->result = "Wrong # args";
        return TCL_ERROR;
    }

    topid = (SdsIdType) (atol(argv[1]));
    pivid = (SdsIdType) (atol(argv[2]));
    fibre_no = atol(argv[3]) - 1;
    obj_no = atol(argv[5]) - 1;
    fibreType = ConfWhichSpec((int) fibre_no, 0);

    if ((fibre_no < 0) || (fibre_no >= (int) FpilGetNumPivots(*Instrument))) {
        sprintf(interp->result,
            "Illegal Pivot Number %ld\nMust be between 1 and %d",
            fibre_no + 1, FpilGetNumPivots(*Instrument));
        return TCL_ERROR;
    } else if (!UseFibreType[fibreType]) {
        const char *FibreTypeDescr;
        ConfWhichSpec((int) fibre_no, &FibreTypeDescr);
        sprintf(interp->result,
            "Attempt to allocate pivot %ld, with a fibre type (%s) ",
            fibre_no + 1, FibreTypeDescr);
        strcat(interp->result, " whose allocation is disabled");
        return TCL_ERROR;
    } else if (strcmp(argv[4], "object") == 0) {
        SdsFind(topid, "unallocObject", &uid, &status);
        ConfSdsCheck("Error finding unallocObject structure", &status);
        guide = 0;
    } else if (strcmp(argv[4], "guide") == 0) {
        SdsFind(topid, "unallocGuide", &uid, &status);
        ConfSdsCheck("Error finding unallocGuide structure", &status);
        guide = 1;
    } else {
        interp->result = "Arg 4 must be object or guide";
        return TCL_ERROR;
    }
/*      This DEBUG section provides additional details about the object.
    {
        int x, y;
        SdsIdType tid;
        char string[128];
        unsigned long actlen;
        ConfArrayGeti(uid, "x", obj_no, &x, &status);
        ConfArrayGeti(uid, "y", obj_no, &y, &status);
        SdsFind(uid, "name", &tid, &status);
        SdsGet(tid, 80, 80 * obj_no, string, &actlen, &status);
        SdsFreeId(tid, &status);

        DEBUG("AllocateFibre:Object %d (%s) position is at %d,%d\n",
           obj_no,string,x,y);

    }
*/

    SdsFind(topid, "objects", &objid, &status);
    ConfSdsCheck("Error finding objects structure", &status);

/*  Check fibre is not already allocated  */

    ConfArrayGetc(objid, "type", fibre_no, &c, &status);
    if (c != FpilUnallocatedType(*Instrument)) {

        sprintf(interp->result,
            "Fibre at pivot number %ld is already allocated", fibre_no + 1);
        return TCL_ERROR;
    }

/*  Check fibre is useable  */

    ConfArrayGets(pivid, "inUse", fibre_no, &s, &status);
    if (s != 1 && ImportOnlyMode == 0) {
        sprintf(interp->result,
            "Fibre at pivot number %ld is not useable", fibre_no + 1);

        return TCL_ERROR;
    }

/*  Check object is not already allocated  */

    ConfArrayGetc(uid, "allocated", obj_no, &c, &status);
    if (c != 0) {
        sprintf(interp->result, "Object %ld is already allocated", obj_no + 1);
        return TCL_ERROR;
    }

/*  Copy data from the unallocated object section to the objects
    section of the structure  */

    ConfArrayCopy(uid, objid, "name", NAMELEN, NAMELEN * obj_no,
        NAMELEN * fibre_no, &status);
    ConfArrayCopy(uid, objid, "ra", sizeof(double), obj_no, fibre_no, &status);
    ConfArrayCopy(uid, objid, "dec", sizeof(double), obj_no, fibre_no, &status);


    ConfArrayCopy(uid, objid, "x", sizeof(int), obj_no, fibre_no, &status);
    ConfArrayCopy(uid, objid, "y", sizeof(int), obj_no, fibre_no, &status);

    ConfArrayCopy(uid, objid, "type", sizeof(char), obj_no, fibre_no, &status);

    ConfArrayCopy(uid, objid, "priority", sizeof(short), obj_no, fibre_no,
        &status);

    ConfArrayCopy(uid, objid, "magnitude", sizeof(double), obj_no, fibre_no,
        &status);

    ConfArrayCopy(uid, objid, "pId", sizeof(short), obj_no, fibre_no, &status);
    ConfArrayCopy(uid, objid, "pmra", sizeof(double), obj_no, fibre_no,
                                                                      &status);
    ConfArrayCopy(uid, objid, "pmdec", sizeof(double), obj_no, fibre_no,
                                                                      &status);

    ConfArrayCopy(uid, objid, "comment", CMTLEN, CMTLEN * obj_no,
        CMTLEN * fibre_no, &status);

    if (guide) {
        ConfArrayCopy(uid, objid, "dhdx", sizeof(double), obj_no, fibre_no,
            &status);
        ConfArrayCopy(uid, objid, "dhdy", sizeof(double), obj_no, fibre_no,
            &status);
        ConfArrayCopy(uid, objid, "dddx", sizeof(double), obj_no, fibre_no,
            &status);
        ConfArrayCopy(uid, objid, "dddy", sizeof(double), obj_no, fibre_no,
            &status);
        ConfArrayPuts(objid, "wlenIndex", fibre_no, -1, &status);

    } else {
        ConfArrayCopy(uid, objid, "wlenIndex", sizeof(short), obj_no, fibre_no,
            &status);
    }

/*  Calculate theta so that the button is parallel to the fibre  */

    ConfArrayGeti(pivid, "xPiv", fibre_no, &xp, &status);
    ConfArrayGeti(pivid, "yPiv", fibre_no, &yp, &status);
    ConfArrayGeti(objid, "x", fibre_no, &x, &status);
    ConfArrayGeti(objid, "y", fibre_no, &y, &status);

    theta = slaDranrm(atan2((double) (yp - y), (double) (xp - x)) - DPIBY2);

    ConfArrayPutd(objid, "theta", fibre_no, theta, &status);
    
/* We need to update the allocation statistics held for this fibre type */

    ConfArrayGetc(objid, "type", fibre_no, &c, &status);
    IsSky = FpilIsTargetSky(*Instrument, c);
    if ((fibreType >= 0) && (fibreType < InstNumFibreTypes)) {
        FibreTypeStats[fibreType].Unallocated--;
        if (FpilIsTargetGuide(*Instrument, c)) {
            FibreTypeStats[fibreType].AllocToGuide++;
        } else if (IsSky) {
            FibreTypeStats[fibreType].AllocToSky++;
        } else {
            FibreTypeStats[fibreType].AllocToObject++;
        }
    }
    
/* Also, given this target type (in 'c') we need to calculate how this
 * allocation affects the possible sky and object targets for the different
 * fibre types - it might have been a target that could be allocated to a
 * number of different fibre types
 */
 
    FpilCompatiblePivotTypes (*Instrument,10,c,0,&NumPivTypes,PivTypes);
    for (PivIndex = 0; PivIndex < NumPivTypes; PivIndex++) {
        fibreType = PivTypes[PivIndex];
        if ((fibreType >= 0) && (fibreType < InstNumFibreTypes)) {
            if (IsSky) {
                FibreTypeStats[fibreType].Skies--;
            } else {
                FibreTypeStats[fibreType].Targets--;
            }
        } else {
            DEBUG ("Invalid fibre type code %d returned\n",fibreType);
        }
    }

/* We now need to update the field data structure */

    SdsFind(topid, "fieldData", &fid, &status);
    ConfSdsCheck("Error finding fieldData structure", &status);


    if (status == STATUS__OK) {
        if (guide) {
            ArgGets(fid, "unallocGui", &s, &status);
            s--;
            ArgPuts(fid, "unallocGui", s, &status);
            ArgGets(fid, "allocGui", &s, &status);
            if (status == SDS__NOITEM) {
                ErsAnnul(&status);
                s = 1;
            } else
                s++;
            ArgPuts(fid, "allocGui", s, &status);
        } else {
            ConfArrayGetc(objid, "type", fibre_no, &c, &status);
            if (!FpilIsTargetGuide(*Instrument, c) &&
                !FpilIsTargetSky(*Instrument, c)) {
                ArgGets(fid, "unallocObj", &s, &status);
                s--;
                ArgPuts(fid, "unallocObj", s, &status);
                ArgGets(fid, "allocObj", &s, &status);
                if (status == SDS__NOITEM) {
                    ErsAnnul(&status);
                    s = 1;
                } else
                    s++;
                ArgPuts(fid, "allocObj", s, &status);
            } else if (FpilIsTargetSky(*Instrument, c)) {
                char name[20];

                /* Update unallocSky and allocSky */
                
                ArgGets(fid, "unallocSky", &s, &status);
                ArgPuts(fid, "unallocSky", s - 1, &status);
                ArgGets(fid, "allocSky", &s, &status);
                if (status == SDS__NOITEM) {
                    ErsAnnul(&status);
                    s = 1;
                } else
                    s++;
                ArgPuts(fid, "allocSky", s, &status);

                /* Update allocSkyn where n = fibre type number */
                
                sprintf(name, "allocSky%d", fibreType);
                ArgGets(fid, "name", &s, &status);
                if (status == SDS__NOITEM) {
                    ErsAnnul(&status);
                    s = 1;
                } else
                    s++;
                ArgPuts(fid, name, s, &status);

            }
        }
    }
    ConfArrayPuts(objid, "index", fibre_no, obj_no, &status);
    ConfArrayPutc(uid, "allocated", obj_no, (char) 1, &status);
    SdsFreeId(objid, &status);
    SdsFreeId(uid, &status);
    SdsFreeId(fid, &status);

    if (status == STATUS__OK) {
        interp->result = "OK";
        return TCL_OK;
    } else {
        char buff[200];
        ErsSPrintf(TCL_RESULT_SIZE, interp->result,
            "\nError Allocating Fibre %ld\n", fibre_no + 1);
        Tcl_AppendResult(interp, buff, (char *) NULL);

        MessGetMsg(status, 0, sizeof(buff), buff);
        Tcl_AppendResult(interp, buff, (char *) NULL);
        return TCL_ERROR;

    }

}

/*  ------------------------------------------------------------------------- */

/*                      L o c k e d  F i b r e s
 *
 *  There are a few routines here connected with the concept of 'locked'
 *  fibres. This was introduced quite late into the system, and is handled
 *  by a global array at this C level called AllocLockedFlags. The SDS
 *  'objects' structure has a flag called AllowAllocChange which might 
 *  seem to be exactly what is needed, but was being used to disable 
 *  allocations for specific fibre types, something also handled at this
 *  C level through the UseFibreType array. So the locked flags have been
 *  introduced in parallel with the fibre type use array and you can see
 *  these being used in combination at the start of the ConfDoAllocation}
 *  code to set the AllowAllocChange flags in the objects structure 
 *  just prior to an allocation. The allocation algorithm only knows about
 *  the AllowAllocChange flags, and this scheme allows this to be used
 *  for both purposes.
 */
 
/*  ------------------------------------------------------------------------- */

/*
 *                     L o c k  A l l o c a t i o n s  { }
 */
 
static int ConfLockAllocations (ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc, char *argv[])

/*+
 *  Command name:
 *    LockAllocations
 *
 *  Function:
 *    Lock all the current allocations.
 *  
 *  Call:
 *    LockAllocations id pivid
 *
 *  Description:
 *    This command sets the 'allocation locked' flags for all the
 *    pivots that are currently allocated. This means that all these
 *    allocations will be unchanged by the allocation algorithm. All
 *    currently unallocated pivots are unlocked by this call.
 *
 *   Parameters:
 *
 *    (>) id  (Sds Id)   SDS identifier of configuration structure.
 *
 *   Returns:
 *     "OK" if succesful, "Error" otherwise
 *
 *   Support:
 *     Keith Shortridge, AAO
 *
 *   Version Date:
 *      4-Sep-2007  Original version. JAB.
 *-
 */
{
    SdsIdType topid, objid;
    long fibre_no;
    StatusType status = STATUS__OK;
    char c;

    if (argc != 2) {
        interp->result = "Wrong # args";
        return TCL_ERROR;
    }

    topid = (SdsIdType) (atol(argv[1]));
    SdsFind(topid, "objects", &objid, &status);
    ConfSdsCheck("Error finding objects structure", &status);

/*  Set the 'locked' flag on or off for each fibre in turn  */

    for (fibre_no = 0; fibre_no < InstNumPivots; fibre_no++) {
       ConfArrayGetc(objid, "type", fibre_no, &c, &status);
       if (c == FpilUnallocatedType(*Instrument)) {
           AllocLockedFlags[fibre_no] = CONFIG_FALSE;
       } else {
           AllocLockedFlags[fibre_no] = CONFIG_TRUE;
       }
    }
    SdsFreeId(objid, &status);
    
    if (status == STATUS__OK) {
        interp->result = "OK";
        return TCL_OK;
    } else {
        char buff[200];
        ErsSPrintf(TCL_RESULT_SIZE, interp->result,
            "\nError locking allocations\n");
        Tcl_AppendResult(interp, buff, (char *) NULL);

        MessGetMsg(status, 0, sizeof(buff), buff);
        Tcl_AppendResult(interp, buff, (char *) NULL);
        return TCL_ERROR;

    }

}

/*  ------------------------------------------------------------------------- */

/*
 *                     C l e a r  L o c k i n g  { }
 */
 
static int ConfClearLocking (ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc, char *argv[])

/*+
 *  Command name:
 *    ClearLocking
 *
 *  Function:
 *    Clear all the current fibre locks.
 *  
 *  Call:
 *    ClearLocking
 *
 *  Description:
 *    This command clears the 'allocation locked' flags for all pivots.
 *
 *   Parameters: None.
 *
 *   Returns:
 *     "OK"
 *
 *   Support:
 *     Keith Shortridge, AAO
 *
 *   Version Date:
 *      4-Sep-2007  Original version. JAB.
 *-
 */
{
    int fibre_no;
    
    if (argc != 1) {
        interp->result = "Wrong # args";
        return TCL_ERROR;
    }

    for (fibre_no = 0; fibre_no < InstNumPivots; fibre_no++) {
       AllocLockedFlags[fibre_no] = CONFIG_FALSE;
    }
    
    interp->result = "OK";
    return TCL_OK;

}

/*  ------------------------------------------------------------------------- */

/*
 *                     L o c k e d  F i b r e s  { }
 */
 
static int ConfLockedFibres (ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc, char *argv[])

/*+
 *  Command name:
 *    LockedFibres
 *
 *  Function:
 *    Returns the number of fibres that are currently locked.
 *  
 *  Call:
 *    LockedFibres id [alloc]
 *
 *  Description:
 *    This command returns the number of fibres whose 'allocation locked' 
 *    flags have been set. Normally it returns the total number of locked
 *    fibres, but it can be asked to return the number of fibres that 
 *    have been locked and are allocated, or the number of locked fibres
 *    that are currently unallocated.
 *
 *   Parameters:
 *
 *    (>) id     (Sds Id)   SDS identifier of configuration structure.
 *    (>) alloc  (string)   Can be 'allocated' or 'unallocated'. Only
 *                          the first character is checked, and its case
 *                          is ignored. If omitted, LockedFibres} returns
 *                          the total number of locked fibres, independent
 *                          of their allocation status.
 *
 *   Returns: The number of locked fibres with the specified 
 *            allocation status.
 *
 *   Support:
 *     Keith Shortridge, AAO
 *
 *   Version Date:
 *      4-Sep-2007  Original version. JAB.
 *-
 */
{
    SdsIdType topid, objid;
    long fibre_no;
    StatusType status = STATUS__OK;
    char c;
    int Count = 0;
    
    /*  The version where we just count the locked fibres is trivial. We
     *  allow this to be called with or without the id argument, but
     *  ignore the id if passed.
     */
    
    if (argc == 1 || argc == 2) {
       for (fibre_no = 0; fibre_no < InstNumPivots; fibre_no++) {
          if (AllocLockedFlags[fibre_no]) Count++;
       }
       sprintf(interp->result,"%d",Count);
       
    } else if (argc == 3) {
    
       /*  If we are to take into account the allocation status, we
        *  need to look at the allocation flags in the configuration
        *  structure.
        */
       
       int AllocOnly = CONFIG_FALSE;
       if (argv[2][0] == 'a' || argv[2][0] == 'A') AllocOnly = CONFIG_TRUE;
       topid = (SdsIdType) (atol(argv[1]));
       SdsFind(topid, "objects", &objid, &status);
       ConfSdsCheck("Error finding objects structure", &status);

       for (fibre_no = 0; fibre_no < InstNumPivots; fibre_no++) {
          if (AllocLockedFlags[fibre_no]) {
             ConfArrayGetc(objid, "type", fibre_no, &c, &status);
             if (c == FpilUnallocatedType(*Instrument)) {
                if (!AllocOnly) Count++;
             } else {
                if (AllocOnly) Count++;
             }
          }
       }
       sprintf(interp->result,"%d",Count);
       SdsFreeId(objid, &status);
         
    } else {   
        interp->result = "Wrong # args";
        return TCL_ERROR;
    }
    
    return TCL_OK;

}

/*  ------------------------------------------------------------------------- */

/*
 *                     F i b r e  L o c k e d  { }
 */
 
static int ConfFibreLocked (ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc, char *argv[])

/*+
 *  Command name:
 *    FibreLocked
 *
 *  Function:
 *    Returns true if the specified fibre is locked.
 *  
 *  Call:
 *    FibreLocked fibre_no
 *
 *  Description:
 *    This command returns true (1) or false (0) depending on whether
 *    or not the specified fibre is locked - ie if its allocation may
 *    not be changed.
 *
 *   Parameters:
 *
 *    (>) fibre_no (int)   The index number of the fibre/pivot in question.
 *                         This goes from 1 up.
 *
 *   Returns: The locked status of the fibre: 1 for locked, 0 for not locked.
 *
 *   Support:
 *     Keith Shortridge, AAO
 *
 *   Version Date:
 *      4-Sep-2007  Original version. JAB.
 *-
 */
{
    long fibre_no;
    int Locked = 0;
    
    if (argc == 2) {
       fibre_no = atoi(argv[1]);
       Locked = AllocLockedFlags[fibre_no - 1] ? 1 : 0;
       sprintf (interp->result,"%d",Locked);
         
    } else {   
        interp->result = "Wrong # args";
        return TCL_ERROR;
    }
    
    return TCL_OK;

}

/*  ------------------------------------------------------------------------- */

/*
 *                     P i v o t  I n  U s e  { }
 */
 
static int ConfPivotInUse(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc, char *argv[])

/*+
 *  Command name:
 *    PivotInUse
 *
 *  Function:
 *    Examines the 'inUse' flag for a given pivot/fibre
 *  
 *  Call:
 *    PivotInUse pivid fibre
 *
 *  Description:
 *    This command returns the value of the 'inUse' flag for a given
 *    fibre/pivot.
 *
 *   Parameters:
 *
 *    (>) pivid (Sds Id) SDS identifier of the pivots structure.
 *    (>) fibre (int)    Fibre (pivot) number - starting from one.
 *
 *   Returns:
 *     1 if pivot is flagged as in use, 0 if not.
 *
 *   Support:
 *     Keith Shortridge, AAO
 *
 *   Version Date:
 *     11-Apr-2002  Original version. JAB.
 *-
 */
{
    SdsIdType pivid;
    long fibre_no;
    StatusType status = STATUS__OK;
    short s;

    if (argc != 3) {
        interp->result = "Wrong # args";
        return TCL_ERROR;
    }

    pivid = (SdsIdType) (atol(argv[1]));
    fibre_no = atol(argv[2]) - 1;

    if ((fibre_no < 0) || (fibre_no >= (int) FpilGetNumPivots(*Instrument))) {
        sprintf(interp->result,
            "Illegal Pivot Number %ld\nMust be between 1 and %d",
            fibre_no + 1, FpilGetNumPivots(*Instrument));
        return TCL_ERROR;
    }
    
    ConfArrayGets(pivid, "inUse", fibre_no, &s, &status);
    if (s != 1) {
        strcpy(interp->result,"0");
    } else {
        strcpy(interp->result,"1");
    }
    return TCL_OK;

}

/*  ------------------------------------------------------------------------- */

/*  The following global variables are used by ReportRoutine(), ConfAllocRpt()
 *  and ReportRoutineBatch(). gecount is a global error count, used to abort
 *  the allocation if errors become excessive. ginterp is used to keep a
 *  pointer to the current Tcl interpreter (but actually no longer seems
 *  to be used).
 */
 
static Tcl_Interp *ginterp;
static int gecount;

/*  ------------------------------------------------------------------------- */

/*                         R e p o r t  R o u t i n e
 *
 *  Called by  ConfSetTargets and ConfAllocate, within ConfDoAllocation, 
 *  to report on allocation progress.  Note that this has the same calling
 *  sequence as ConfAllocRpt() but since it is now used only to report
 *  the progress rather than actual allocation details, most of the arguments
 *  are unused.
 */
 
static int ReportRoutine(void *clientData, char Type DUNUSED, char *Description,
    int pivot DUNUSED, int guide DUNUSED, int target DUNUSED,
    long TargetX DUNUSED, long TargetY DUNUSED,
    double theta DUNUSED, double PercentDone)
 {
    Tcl_Interp *interp = clientData;
    char command[300];

    sprintf(command, "AllUpdate \"%s\" %f",Description, PercentDone);

/*
 *  Evaluate a command to report progress.  If we get an error, report
 *  it, but only abort if number of errors is above MAX_ERRORS.
 *
 *  Note, we only send the abort message once, after which we stop
 *  reporting messages to avoid flooding the screen.
 *
 */
    if (Tcl_Eval(interp, command) == TCL_ERROR) {

        if (gecount <= MAX_ERRORS) {
            char buffer[80];
            sprintf(buffer,
                "\n(Error when reporting allocation progress (%d))",
                gecount + 1);
            Tcl_AddErrorInfo(interp, buffer);
            Tcl_BackgroundError(interp);
        }
        gecount++;
        if (gecount == MAX_ERRORS) {
            StatusType status = STATUS__OK;
            ErsRep(ERS_M_BELL, &status,
                "Allocation Aborted due to too many (%d) errors", gecount);
            return 1;
        }
/*
 *  Command returned 1, this indicates a cancel is required.
 */
    } else if (strcmp(interp->result, "1") == 0) {
        return 1;
    }
    return 0;
}

/*  ------------------------------------------------------------------------- */

/*                   R e p o r t  R o u t i n e  B a t c h
 *
 *  Called by  ConfSetTargets and ConfAllocate, within ConfDoAllocationBatch, 
 *  to report on allocation progress.
 */
 
static int ReportRoutineBatch(void *clientData, char Type,
    char *Description,
    int pivot, int guide, int target,
    long TargetX DUNUSED, long TargetY DUNUSED,
    double theta, double PercentDone)
 {
    Tcl_Interp *interp = clientData;
    char command[300];

    if (guide)
        sprintf(command, "AllUpdateBatch %c %d %d uguide %f \"%s\" %f", Type,
            pivot + 1, target + 1, theta, Description, PercentDone);
    else
        sprintf(command, "AllUpdateBatch %c %d %d uobj %f \"%s\" %f", Type,
            pivot + 1, target + 1, theta, Description, PercentDone);

/*
 *  Evaluate a command to report progress.  If we get an error, report
 *  it, but only abort if number of errors is above MAX_ERRORS.
 *
 *  Note, we only send the abort message once, after which we stop
 *  reporting messages to avoid flooding the screen.
 *
 */
    if (Tcl_Eval(interp, command) == TCL_ERROR) {

        if (gecount <= MAX_ERRORS) {
            char buffer[80];
            sprintf(buffer,
                "\n(Error when reporting batch allocation progress (%d))",
                gecount + 1);
            Tcl_AddErrorInfo(interp, buffer);
            Tcl_BackgroundError(interp);
        }
        gecount++;
        if (gecount == MAX_ERRORS) {
            StatusType status = STATUS__OK;
            ErsRep(ERS_M_BELL, &status,
                "Allocation Aborted due to too many (%d) errors", gecount);
            return 1;
        }

/*
 *  Command returned 1, this indicates a cancel is required.
 */
    } else if (strcmp(interp->result, "1") == 0) {
        return 1;
    }
    return 0;
}


/*  ------------------------------------------------------------------------- */

/*                        C o n f  A l l o c  R p t
 *
 *  Called by ConfigAllocate(), within ConfDoAllocation() to indicate 
 *  allocation/deallocation of fibres.  We use the Tcl command AllocRpt to 
 *  do this. Note that it is this mechanism that is responsible for keeping the
 *  configuration data structures updated during an allocation.
 */
 
static void ConfAllocRpt(void *clientData, int Alloc,
    int pivot, int guide, int target,
    long TargetX DUNUSED, long TargetY DUNUSED,
    double theta, StatusType * status)
 {
    Tcl_Interp *interp = clientData;

    char command[300];

    if (*status != STATUS__OK)
        return;

    if (guide)
        sprintf(command, "AllocRpt %d %d %d uguide %f", Alloc,
            pivot + 1, target + 1, theta);
    else
        sprintf(command, "AllocRpt %d %d %d uobj %f", Alloc,
            pivot + 1, target + 1, theta);

/*
 *  Evaluate a command to report allocation/deallocation.  If we get an 
 *  error report it but only abort if number of errors is above MAX_ERRORS
 *
 *  Note, we only send the abort message once, after which we stop
 *  reporting messages to avoid flooding the screen.
 *
 */
    if (Tcl_Eval(interp, command) == TCL_ERROR) {

        if (gecount <= MAX_ERRORS) {
            char buffer[80];
            sprintf(buffer,
                "\n(Error when reporting allocations/deallocations (%d))",
                gecount + 1);
            Tcl_AddErrorInfo(interp, buffer);
            Tcl_BackgroundError(interp);
        }
        gecount++;
        if (gecount == MAX_ERRORS) {
            *status = CONF__RPTERR;
            ErsRep(ERS_M_BELL, status,
                "Allocation Aborted due to too many (%d) errors", gecount);
        }
    }
}


/*  ------------------------------------------------------------------------- */

/*
 *                     D o  A l l o c a t i o n  { }
 *
 *                 D o  A l l o c a t i o n  B a t c h  { }
 */
 
static int ConfDoAllocation(ClientData clientData DUNUSED, Tcl_Interp * interp,
    int argc DUNUSED, char *argv[])

/*+
 *  Command name:
 *    DoAllocation
 *    DoAllocationBatch
 *
 *  Function:
 *    Perform automatic allocation. The batch version does it in Batch mode
 *  
 *  Call:
 *    DoAllocation id
 *    DoAllocationBatch id
 *
 *  Description:
 *    Perform automatic allocation of objects to fibres by calling the
 *    routines ConfigSetTargets and ConfigAllocate. The Tcl procedure
 *    AllUpdate will be called to report the progress of the allocation.
 *
 *    Note that we make no change to the SDS structure, this is done
 *    by the AllUpdate command invoking commands such as AllocateFibre   
 *    SetButtonAngle and DeallocateFibre.
 *
 *    The clientData item is either ALLOCATE_NORMAL or ALLOCATE_BATCH 
 *    to indicate the mode.
 *
 *   Parameters:
 *
 *    (>) id  (Sds Id)   SDS identifier of configuration structure.
 *
 *   Support:
 *     Jeremy Bailey, AAO
 *
 *   Version Date:
 *     6-Oct-1994    
 *-
 */
{

    StatusType status = STATUS__OK;
    char mbuf[100];
    register unsigned i;
    CONFIG_ReportFunction RptFunction;

    if (clientData == (void *) ALLOCATE_BATCH)
        RptFunction = ReportRoutineBatch;
    else
        RptFunction = ReportRoutine;

    sprintf(mbuf, "Doing allocation for field %d", CurrentField);
    ConfMessage(interp, mbuf);

/*  Set Target SDS structure  */

    ConfigSetTargets(NULL, atol(argv[1]), CurrentField, RptFunction,
        (void *) interp, &status);

/*  Indicate which fibre type we want to use. This sets the 'allow
 *  allocation flag for each pivot of that type, so all these flags
 *  are set by this block of code.
 */

    for (i = 0; i < InstNumFibreTypes; ++i) {
        ConfigSetSpecPivotsFlag(i, UseFibreType[i], &status);
    }
    
/*  Set the 'allow allocation' flag on a per-pivot basis for each locked 
 *  fibre. Note that this has to come after the setting of these flags
 *  on the basis of fibre type.
 */
 
    for (i = 0; i < InstNumPivots; i++) {
       if (AllocLockedFlags[i]) {
          DEBUG ("Setting locked flag for fibre index %d\n",i);
          ConfigSetPivotFlag (i + 1, CONFIG_FALSE, &status);
       }
    }
    
/*  Set sky counts for each fibre type - these are the number of fibres of
 *  this type that are to be allocated to sky targets (if available) or
 *  left unallocated for later allocation to sky targets. We clear out all
 *  the sky counts, then set those for the fibre types for which a value
 *  was set using SetSkyCounts{}.
 */
    
    ConfigSetSky(-1,-1,&status);
    for (i = 0; i < InstNumFibreTypes; ++i) {
        if (FibreTypeStats[i].FibreSkyCount >= 0) {
            ConfigSetSky(i,FibreTypeStats[i].FibreSkyCount,&status);
        }
    }

/*  Do the allocation  */

    ginterp = interp;
    gecount = 0;

    ConfigAllocate(RptFunction, (void *) interp,
        ConfAllocRpt, (void *) interp, &status);

    if ((status == STATUS__OK) && (gecount < MAX_ERRORS)) {
/*
 *      Allocation ok.
 *      We need to update the sky values, which we can do by invoking
 *      the command used to implement MakeSkyvals.
 */
        return (ConfMakeSkyvals(clientData, interp, argc, argv));
    } else {
        char buff[200];
/*
 *   Note, we must return OK on error - just report details, as
 *   otherwise you are not allowed to fix what goes wrong, as 
 *   various bits of the code allow.  E.q. autorealocate (-q switch).
 */
        MessGetMsg(status, 0, sizeof(buff), buff);
        ErsRep(ERS_M_BELL, &status, "Error Performing Allocation:-");
        if (status != STATUS__OK)
            ErsRep(ERS_M_BELL, &status, "%s", buff);
        ErsFlush(&status);
        return TCL_OK;

    }

}

/*  ------------------------------------------------------------------------- */

/*
 *                          D o  C h e c k  { }
 */
 
static int ConfDoCheck(ClientData clientData DUNUSED, Tcl_Interp * interp,
    int argc, char *argv[])

/*+
 *  Command name:
 *    DoCheck
 *
 *  Function:
 *    Check validity of a configuration
 *  
 *  Call:
 *    DoCheck id [positionerTolerances [invFibs]]
 *
 *  Description:
 *    Perform a check of the validity of a configuration by calling the
 *    routine ConfigFieldCheck(). The Tcl procedure AllUpdate will be called 
 *    to report the progress of the checking.
 *   
 *   Parameters:
 *
 *    (>) id         Sds Id)   SDS identifier of configuration structure.
 *    (>) positionerTolerances (int) True (non-zero) if the check is to use
 *                   the same tolerances as the positioner itself. This is
 *                   the default. If the value is explicitly passed as zero
 *                   then the current tolerances (usually set through the
 *                   allocation dialogue) will be used.
 *    (<) invFibs    (array) If the configuration is invalid, then invFibs
 *                   will be returned as a Tcl array (with indices 0,1,..n)
 *                   containing the pivot numbers (starting at zero) of the
 *                   fibres involved. In principle, if these fibres were to be
 *                   deallocated, the resulting configuration would be valid.
 *                   Note that deallocating these fibres may not be the best
 *                   way to fix the configuration, but it should work, and may
 *                   be the most practical option if this routine is being
 *                   used to check a configuration over a range of hour angles.
 *
 *   Returns:
 *      "OK" if the configuration is valid, "Error" otherwise.
 *
 *   Support:
 *     Jeremy Bailey, AAO
 *     Keith Shortridge, AAO.
 *
 *   Version Date:
 *     6-Oct-1994 
 *     3-Jan-2002. Added PositionerTolerances optional argument.
 *    22-Aug-2012. Added invFibs optional argument.   
 *-
 */
{
    StatusType status = STATUS__OK;
    int PositionerTolerances = 1;
    
    /*  This array is used for testing the field check reporting option. 
     *  It needs to be large enough to hold all the invalid fibre numbers.
     *  If we have 400 invalid fibres, we have a serious problem! Note
     *  that even if there are more than this many invalid fibres, nothing
     *  will go seriously wrong - ConfigFieldCheck() returns once the
     *  supplied array is full, so all that happens is that the list of
     *  invalid fibres may be incomplete. Indeed, one possible way to run
     *  this might be to set the maximum number of invalid fibres in the
     *  ConfigFieldCheck() call to 1, in which case only one fibre at a time
     *  will be reported. Deallocating this may - possibly - mean that other
     *  fibres colliding with it will now become valid, and so repeated calls
     *  to this routine in this mode might be better than one call followed by
     *  deallocation of all fibres that show problems.
     */
    int InvalidFibres[400];
    int NInvPiv;
    
/*  Check on positioner tolerances argument, if supplied */

    if (argc >= 3) {
        if (atoi(argv[2]) == 0) PositionerTolerances = 0;
    }

/*  Set Target SDS structure  */

    ConfigSetTargets(NULL, atol(argv[1]), CurrentField,
        (CONFIG_ReportFunction) ReportRoutine, interp, &status);

/*  Do the check here explicitly, since we took the check line out of
 *  ConfigInitTargets to allow auto-reallocation ******GBD***** */

    ConfigFieldCheck((CONFIG_ReportFunction) ReportRoutine, interp, 
                    PositionerTolerances,400,InvalidFibres,&NInvPiv,&status);
    
/*  If there were invalid fibres, and if invFibs was specified in the call,
 *  fill the array with the fibre numbers. */
 
    if ((NInvPiv > 0) && (argc >= 4)) {
        int i;
        for (i = 0; i < NInvPiv; i++) {
            char Value[16];
            char Name[64];
            sprintf (Value,"%d",InvalidFibres[i]);
            sprintf (Name,"%s(%d)",argv[3],i);
            Tcl_SetVar(interp,Name,Value,0);
        }
    }

    if (status == STATUS__OK) {
        interp->result = "OK";
        return TCL_OK;
    } else {
        ErsFlush(&status);
        interp->result = "Error";
        return TCL_OK;
    }

}


/*  ------------------------------------------------------------------------- */

/*
 *                          D o  C h e c k  1  { }
 */
 
static int ConfDoCheck1(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc DUNUSED, char *argv[])

/*+
 *  Command name:
 *    DoCheck1
 *
 *  Function:
 *    Check validity of a configuration
 *  
 *  Call:
 *    DoCheck1 id
 *
 *  Description:
 *    Perform a check of the validity of a configurastion by calling the
 *    routines ConfigSetTargets. This is similar to DoCheck{} but is intended
 *    to be used when an initial allocation and check is run in batch mode.
 *    In fact the distinction seems to have been lost at some point in the
 *    development of the code, and this and ConfDoCheck() are now identical,
 *    except that this routine always uses the positioner tolerances. 
 *   
 *   Parameters:
 *
 *    (>) id  (Sds Id)   SDS identifier of configuration structure.
 *
 *   Returns:
 *      "OK" if the configuration is valid, "Error" otherwise.
 *
 *   Support:
 *     Jeremy Bailey, AAO
 *
 *   Version Date:
 *     6-Oct-1994    
 *-
 */
{
    StatusType status = STATUS__OK;

/*  Set Target SDS structure  */

    ConfigSetTargets(NULL, atol(argv[1]), CurrentField,
        (CONFIG_ReportFunction) ReportRoutine, interp, &status);

    ConfigFieldCheck((CONFIG_ReportFunction) ReportRoutine, interp, 1, 
                                              0, NULL, NULL, &status);

    if (status == STATUS__OK) {
        interp->result = "OK";
        return TCL_OK;
    } else {
        ErsFlush(&status);
        interp->result = "Error";
        return TCL_OK;
    }

}

/*  ------------------------------------------------------------------------- */

/*
 *                    D e a l l o c a t e  F i b r e  { }
 */
 
static int ConfDeallocateFibre(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc, char *argv[])

/*+
 *  Command name:
 *    DeAllocateFibre
 *
 *  Function:
 *    Deallocate a fibre
 *  
 *  Call:
 *    DeAllocateFibre id pivid fibre type object
 *
 *  Description:
 *    Deallocate a fibre. Return the index of the deallocated object
 *    in the unallocated objects or guide structure. 
 *
 *    If the fibre has an index entry set, this operation simply
 *    involves removing the index entry and unsetting the allocated
 *    flag for the object. However if there is no index entry, it
 *    is necessary to enlarge the unallocated objects structure and
 *    copy the object data into the new entry.
 *
 *    In either case the fibre position is set back to the parked
 *    position.
 *   
 *   Parameters:
 *
 *    (>) id  (Sds Id)   SDS identifier of configuration structure.
 *    (>) pivid (Sds Id) SDS identifier of the pivots structure.
 *    (>) fibre (int)    Fibre (pivot) number - starting from one.
 *    (<) type  (char)   "uobj" or "uguide".
 *    (<) object (int)   Object number - starting from one.
 *
 *   Returns:
 *     "OK" if succesful, "Error" otherwise
 *
 *   Support:
 *     Jeremy Bailey, AAO
 *
 *   Version Date:
 *      6-Oct-1994  Original version. JAB.    
 *      8-Nov-2001  Now updates fibre type allocation statistics. KS.   
 *      3-Sep-2002  Now updates available target allocation statistics. KS.   
 *-
 */
{
    SdsIdType topid, objid, uid, pivid, fid;
    long fibre_no;
    int fibreType;
    long obj_no;
    StatusType status = STATUS__OK;
    short s;
    char c;
    char guide;
    SdsIdType xpid, ypid, tpid;
    char name[NAMELEN];
    SdsCodeType code;
    long ndims;
    unsigned long dims[7];
    int l;
    double d;
    SdsIdType ra_id;
    unsigned long actlen;
    char ctype;
    char buf[20];
    int spec;
    int NumPivTypes;
    int PivTypes[10];
    short IsSky;
    int PivIndex;

    if (argc != 6) {
        interp->result = "Wrong # args";
        return TCL_ERROR;
    }

    topid = (SdsIdType) (atol(argv[1]));
    pivid = (SdsIdType) (atol(argv[2]));
    fibre_no = atol(argv[3]) - 1;
    spec = ConfWhichSpec((int) fibre_no, 0);

    if ((fibre_no < 0) || (fibre_no >= (int) FpilGetNumPivots(*Instrument))) {
        sprintf(interp->result,
            "Illegal Pivot Number %ld\nMust be between 1 and %d",
            fibre_no + 1, FpilGetNumPivots(*Instrument));
        return TCL_ERROR;
    } else if (!UseFibreType[spec]) {
        if (spec == 0) {

            sprintf(interp->result,
                "Attempt to deallocate guide pivot %ld,", fibre_no + 1);
            strcat(interp->result,
                " when guide pivot deallocation is disabled");
        } else {
            const char *SpecDescr;
            ConfWhichSpec((int) fibre_no, &SpecDescr);
            sprintf(interp->result,
                "Attempt to deallocate pivot %ld, from spectrograph (%s)",
                fibre_no + 1, SpecDescr);
            strcat(interp->result, " marked as not to be deallocated.");
        }
        return TCL_ERROR;
    }


    SdsFind(topid, "objects", &objid, &status);
    ConfSdsCheck("Error finding objects structure", &status);

/*  Check fibre is allocated  */

    ConfArrayGetc(objid, "type", fibre_no, &ctype, &status);

    if (status != STATUS__OK) {
        char buff[200];
        MessGetMsg(status, 0, sizeof(buff), buff);
        ErsFlush(&status);  /* Must do this before setting result */
        Tcl_AppendResult(interp, "\nError getting fibre details",
            (char *) NULL);
        Tcl_AppendResult(interp, buff, (char *) NULL);
        return TCL_ERROR;
    }


    if (ctype == FpilUnallocatedType(*Instrument)) {
        sprintf(interp->result,
            "Fibre at pivot number %ld is not allocated", fibre_no + 1);
        return TCL_ERROR;
    }

    if (FpilIsTargetGuide(*Instrument, ctype)) {
        guide = 1;
        SdsFind(topid, "unallocGuide", &uid, &status);
        Tcl_SetVar(interp, argv[4], "uguide", 0);
    } else {
        guide = 0;
        SdsFind(topid, "unallocObject", &uid, &status);
        Tcl_SetVar(interp, argv[4], "uobj", 0);
    }

/*  Get fibre index  */

    ConfArrayGets(objid, "index", fibre_no, &s, &status);

    if (s != -1) {

/*  The object has an index specified - This is the easy case.
    We just unset the allocated flag  */
        obj_no = s;
        c = 0;
        ConfArrayPutc(uid, "allocated", obj_no, c, &status);
        sprintf(buf, "%ld", obj_no + 1);
        Tcl_SetVar(interp, argv[5], buf, 0);
    }

/*  Otherwise we have to enlarge the unallocated objects section
    and copy the data back to it  */

    else {

        printf("Enlarging unalloctated objects section ");
        printf("should not be needed now)\n");

        SdsFind(uid, "ra", &ra_id, &status);
        SdsInfo(ra_id, name, &code, &ndims, dims, &status);


        SdsFreeId(ra_id, &status);
        ConfResizeAllItems(uid, dims[0] + 1, &status);
        obj_no = dims[0];
        sprintf(buf, "%ld", obj_no + 1);
        Tcl_SetVar(interp, argv[5], buf, 0);


        ConfArrayCopy(objid, uid, "name", NAMELEN, NAMELEN * fibre_no,
            NAMELEN * obj_no, &status);
        ConfArrayCopy(objid, uid, "ra", sizeof(double), fibre_no, obj_no,
            &status);
        ConfArrayCopy(objid, uid, "dec", sizeof(double), fibre_no, obj_no,
            &status);
        ConfArrayCopy(objid, uid, "x", sizeof(int), fibre_no, obj_no, &status);
        ConfArrayCopy(objid, uid, "y", sizeof(int), fibre_no, obj_no, &status);
        ConfArrayCopy(objid, uid, "type", sizeof(char), fibre_no, obj_no,
            &status);
        ConfArrayCopy(objid, uid, "priority", sizeof(short), fibre_no, obj_no,
            &status);
        ConfArrayCopy(objid, uid, "magnitude", sizeof(double), fibre_no, obj_no,
            &status);
        ConfArrayCopy(objid, uid, "pId", sizeof(short), fibre_no, obj_no,
            &status);
        ConfArrayCopy(objid, uid, "comment", CMTLEN, CMTLEN * fibre_no,
            CMTLEN * obj_no, &status);
        if (guide) {
            ConfArrayCopy(objid, uid, "dhdx", sizeof(double), fibre_no,
                obj_no, &status);
            ConfArrayCopy(objid, uid, "dhdy", sizeof(double), fibre_no,
                obj_no, &status);
            ConfArrayCopy(objid, uid, "dddx", sizeof(double), fibre_no,
                obj_no, &status);
            ConfArrayCopy(objid, uid, "dddy", sizeof(double), fibre_no,
                obj_no, &status);
        } else {
            ConfArrayCopy(objid, uid, "wlenIndex", sizeof(short), obj_no, fibre_no,
                          &status);
            ConfArrayCopy(objid, uid, "spectrograph", sizeof(char),
                          fibre_no, obj_no, &status);
        }
        ConfArrayPutc(uid, "allocated", obj_no, 0, &status);

    }

/*  Now set the objects section entry back to the parked position */

    SdsFind(pivid, "xPark", &xpid, &status);
    SdsFind(pivid, "yPark", &ypid, &status);
    SdsFind(pivid, "tPark", &tpid, &status);

    strcpy(name, "Parked");
    ConfArrayPutString(objid, "name", fibre_no, name, &status);
    ConfArrayPutd(objid, "ra", fibre_no, 0.0, &status);
    ConfArrayPutd(objid, "dec", fibre_no, 0.0, &status);
    ConfArrayPuts(objid, "fibre", fibre_no, 0, &status);
    SdsGet(xpid, sizeof(int), fibre_no, &l, &actlen, &status);
    ConfArrayPuti(objid, "x", fibre_no, l, &status);
    SdsGet(ypid, sizeof(int), fibre_no, &l, &actlen, &status);
    ConfArrayPuti(objid, "y", fibre_no, l, &status);
    SdsGet(tpid, sizeof(double), fibre_no, &d, &actlen, &status);
    ConfArrayPutd(objid, "theta", fibre_no, d, &status);
    ConfArrayPutc(objid, "type", fibre_no,
        FpilUnallocatedType(*Instrument), &status);
    ConfArrayPuts(objid, "priority", fibre_no, 1, &status);
    ConfArrayPutd(objid, "magnitude", fibre_no, 0.0, &status);
    ConfArrayPutc(objid, "spectrograph", fibre_no, ' ', &status);
    ConfArrayPuts(objid, "pId", fibre_no, 1, &status);
    ConfArrayPutd(objid, "dhdx", fibre_no, 0.0, &status);
    ConfArrayPutd(objid, "dhdy", fibre_no, 0.0, &status);
    ConfArrayPutd(objid, "dddx", fibre_no, 0.0, &status);
    ConfArrayPutd(objid, "dddy", fibre_no, 0.0, &status);
    ConfArrayPutString(objid, "comment", fibre_no, name, &status);
    ConfArrayPuts(objid, "wlenIndex", fibre_no, -1, &status);
    ConfArrayPuts(objid, "index", fibre_no, -1, &status);

    SdsFreeId(xpid, &status);
    SdsFreeId(ypid, &status);
    SdsFreeId(tpid, &status);

/* We need to update the allocation statistics held for this fibre type */

    fibreType = ConfWhichSpec((int) fibre_no, 0);
    IsSky = FpilIsTargetSky(*Instrument, ctype);
    if ((fibreType >= 0) && (fibreType < InstNumFibreTypes)) {
        FibreTypeStats[fibreType].Unallocated++;
        if (FpilIsTargetGuide(*Instrument, ctype)) {
            FibreTypeStats[fibreType].AllocToGuide--;
        } else if (IsSky) {
            FibreTypeStats[fibreType].AllocToSky--;
        } else {
            FibreTypeStats[fibreType].AllocToObject--;
        }
    }

/* Also, given this target type (in 'ctype') we need to calculate how this
 * deallocation affects the possible sky and object targets for the different
 * fibre types - it might be a target that can be allocated to a
 * number of different fibre types
 */
 
    FpilCompatiblePivotTypes (*Instrument,10,ctype,0,&NumPivTypes,PivTypes);
    for (PivIndex = 0; PivIndex < NumPivTypes; PivIndex++) {
        fibreType = PivTypes[PivIndex];
        if ((fibreType >= 0) && (fibreType < InstNumFibreTypes)) {
            if (IsSky) {
                FibreTypeStats[fibreType].Skies++;
            } else {
                FibreTypeStats[fibreType].Targets++;
            }
        } else {
            DEBUG ("Invalid fibre type returned, type %d\n",fibreType);
        }
    }
    
    if (status == STATUS__OK) {
        SdsFind(topid, "fieldData", &fid, &status);
        ConfSdsCheck("Error finding fieldData structure", &status);
    }
    if (status == STATUS__OK) {
        if (guide) {
            ArgGets(fid, "unallocGui", &s, &status);
            s++;
            ArgPuts(fid, "unallocGui", s, &status);
            ArgGets(fid, "allocGui", &s, &status);
            if (status == SDS__NOITEM) {
                ErsAnnul(&status);
                s = 1;
            } else
                s--;
            ArgPuts(fid, "allocGui", s, &status);
        } else {
            if (!FpilIsTargetGuide(*Instrument, ctype) &&
                !FpilIsTargetSky(*Instrument, ctype)) {
                /*  Not sky and not a guide object, so this is a 'program'
                 *  object.
                 */

                ArgGets(fid, "unallocObj", &s, &status);
                s++;
                ArgPuts(fid, "unallocObj", s, &status);
                ArgGets(fid, "allocObj", &s, &status);
                if (status == SDS__NOITEM) {
                    ErsAnnul(&status);
                    s = 1;
                } else
                    s--;
                ArgPuts(fid, "allocObj", s, &status);
            } else if (FpilIsTargetSky(*Instrument, ctype)) {
                char name[20];

                /* Update unallocSky and allocSky */
                ArgGets(fid, "unallocSky", &s, &status);
                ArgPuts(fid, "unallocSky", s + 1, &status);
                ArgGets(fid, "allocSky", &s, &status);
                if (status == SDS__NOITEM) {
                    ErsAnnul(&status);
                    s = 0;
                } else
                    s--;
                ArgPuts(fid, "allocSky", s, &status);

                /* Update allocSkyn where n = spectrograph number */
                sprintf(name, "allocSky%d", spec);
                ArgGets(fid, "name", &s, &status);
                if (status == SDS__NOITEM) {
                    ErsAnnul(&status);
                    s = 0;
                } else
                    s--;
                ArgPuts(fid, name, s, &status);
            }
        }
    }
    SdsFreeId(objid, &status);
    SdsFreeId(uid, &status);
    SdsFreeId(fid, &status);

    if (status == STATUS__OK) {
        interp->result = "OK";
        return TCL_OK;
    } else {

        char buff[200];
        sprintf(buff, "\nError Deallocating Fibre %ld\n", fibre_no + 1);
        Tcl_AppendResult(interp, buff, (char *) NULL);

        MessGetMsg(status, 0, sizeof(buff), buff);
        Tcl_AppendResult(interp, buff, (char *) NULL);
        return TCL_ERROR;
    }

}

/*  ------------------------------------------------------------------------- */

/*
 *                  U p d a t e  F i b r e  S t a t s  { }
 */
 
static int ConfUpdateFibreStats(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc, char *argv[])

/*+
 *  Command name:
 *    UpdateFibreStats
 *
 *  Function:
 *    Makes sure the fibre allocation statistics reflect the configuration.
 *  
 *  Call:
 *    UpdateFibreStats id
 *
 *  Description:
 *    The global array FibreTypeStats is used to maintain allocation 
 *    statistics for all the different fibre types. It is updated when
 *    a fibre is allocated or deallocated, but needs to be initialised
 *    both at the start of the program or when a configuration is read
 *    in. This routine performs this initialisation.
 *   
 *   Parameters:
 *
 *    (>) id  (Sds Id)   SDS identifier of configuration structure.
 *
 *   Returns:
 *     "OK" if succesful, some error indication otherwise
 *
 *   Support:
 *     Keith Shortridge, AAO
 *
 *   Version Date:
 *      6th Nov 2001. Original version. KS. 
 *     28th Nov 2001. Removed unused variable FibreInUse. KS.
 *      3rd Sep 2002. Now calculates available sky and object statistics for
 *                    each fibre type. KS.   
 *      7th Jan 2005. Not calculates number of not-in-use fibres, subtracting these
 *                     from the number of unallocated fibres.
 *-
 */
{
    const char *const *SpecNames;
    unsigned int NumFibreTypes;
    StatusType status = STATUS__OK;
    char name[NAMELEN];
    int Index;
    SdsIdType topid, objid, typeId, allocId, spectId, xId, yId;
    SdsCodeType code;
    short AllocArray;
    long ndims;
    unsigned long dims[7];
    int XPosn,YPosn;
    int NumFibres;
    int NumObjects;
    char Alloc;
    char Spect;
    short SpectKnown;
    unsigned long actlen;
    unsigned int FibreType;
    const char *FibreTypeDescr;
    char Type;
    char UnallocTypeCode;
    int ReturnCode;
    int PivIndex;
    int NumPivTypes;
    int PivTypes[10];
    short IsSky;
    int notInUseAllocated = 0; /* Flag, set true if any not in use are allocated*/
    
    if (argc != 2) {
        interp->result = "Wrong # args";
        return TCL_ERROR;
    }
    
    objid = 0;
    typeId = 0;

    /*  Clear out the fibre type statistics array. Clear out the counts.
     */
    
    FpilConSpecInfo(*Instrument, &NumFibreTypes, &SpecNames, &status);
    for (Index = 0; Index < NumFibreTypes; Index++) {
       FibreTypeStats[Index].Name[0]       = '\0';
       FibreTypeStats[Index].Fibres        = 0;
       FibreTypeStats[Index].AllocToSky    = 0;
       FibreTypeStats[Index].AllocToGuide  = 0;
       FibreTypeStats[Index].AllocToObject = 0;
       FibreTypeStats[Index].Unallocated   = 0;
       FibreTypeStats[Index].Targets       = 0;
       FibreTypeStats[Index].Skies         = 0;
       FibreTypeStats[Index].NotInUse      = 0;
    }
       
    /*  For all the types initialise the counts for things that are nothing
     *  to do with allocation, such as the total number of fibres of 
     *  each type, and the number of each type that are not in use (probably
     *  because they're broken). We start by assuming all fibres are 
     *  unallocated. Note that ConfWhichSpec() can return a code that is
     *  outside the range (0..NumFibreTypes-1) - it uses such codes for
     *  disabled fibres.
     */

    NumFibres = FpilGetNumPivots(*Instrument);
    for (Index = 0; Index < NumFibres; Index++) {
        FibreType = ConfWhichSpec(Index, &FibreTypeDescr);
        if ((FibreType >= 0) &&(FibreType < NumFibreTypes)) {
            if (FibreTypeStats[FibreType].Name[0] == '\0') {
                strncpy (FibreTypeStats[FibreType].Name,FibreTypeDescr,
                                     sizeof(FibreTypeStats[FibreType].Name));
                FibreTypeStats[FibreType].Name[
                       sizeof(FibreTypeStats[FibreType].Name) - 1] = '\0';
            }
            FibreTypeStats[FibreType].Fibres++;
            FibreTypeStats[FibreType].Unallocated++;
        }
    }
    
    /*  Get the Sds Id for the configuration data from the passed argument,
     *  then see if there is an objects section. If not, no targets have
     *  been allocated.
     */
   
    objid = typeId = allocId = spectId =  xId = yId = 0;
    topid = (SdsIdType) (atol(argv[1]));
    
    SdsFind(topid, "objects", &objid, &status);
    if (status == STATUS__OK) {
    
        /*  There is an "objects" section, so we can start to work through 
         *  it, building up the counts. Get the Sds Ids for the one array that
         *  we are going to make use of, namely the 'type' array.
         */

        SdsFind(objid, "type", &typeId, &status);
        SdsInfo(typeId, name, &code, &ndims, dims, &status);
        NumFibres = dims[0];

        UnallocTypeCode = FpilUnallocatedType(*Instrument);

        /*  Loop through each item in the objects structure. Remember that
         *  this contains information for one fibre, since there are is one
         *  item in the objects structure for each fibre on the instrument.
         */

        for (Index = 0; Index < NumFibres; Index++) {

            /*  For this fibre, we need to know which fibre type it represents.
             *  We could use the information in the pivot structure, assuming
             *  (knowing!) that there is a one to one mapping between items in
             *  there and in the objects section. However, it's probably better
             *  to use the usual routine for this purpose, namely a call to
             *  ConfWhichSpec().
             */

            FibreType = ConfWhichSpec(Index, &FibreTypeDescr);
            if ((FibreType >= 0) &&(FibreType <= NumFibreTypes)) {
                short inUse;
                SdsGet(typeId, sizeof(char), Index, &Type, &actlen, &status);
                if (Type != UnallocTypeCode) {

                    /*  If the fibre is allocated, see to what sort of object 
                     *  it is allocated. The "type" information is enough 
                     *  here. If we wanted a more complex set of statistics - 
                     *  some magnitude details, for example - we would have 
                     *  to use other items in the objects structure.
                     */

                    FibreTypeStats[FibreType].Unallocated--;
                    if (FpilIsTargetGuide(*Instrument,Type)) {
                        FibreTypeStats[FibreType].AllocToGuide++;
                    } else if (FpilIsTargetSky(*Instrument,Type)) {
                        FibreTypeStats[FibreType].AllocToSky++;
                    } else {
                        FibreTypeStats[FibreType].AllocToObject++;
                    }
                }
                /*
                 * Check the inUse flag.
                 */
                ConfArrayGets(InstFieldDetails, "inUse", Index, &inUse, &status);
                if (!inUse)
                {
                    /*
                     * This fibre is not in use.  
                     *
                     * If the fibre is actually allocated, whilst not being in use,
                     * then it probably indicates we have read in an old
                     * allocation (or have read in an allocation for a different
                     * plate from the original plate which was used). Report
                     * details.
                     *
                     * NOTE - this can result in Unallocated becomming negative.
                     */
                    if (Type != UnallocTypeCode)
                    {
                        char buffer[100];
                        sprintf(buffer,
                                "Warning - fibre %d not in use but allocated",
                                Index+1);
                        ConfMessage(interp, buffer);
                        ++notInUseAllocated;
                    }

                    FibreTypeStats[FibreType].Unallocated--;
                    FibreTypeStats[FibreType].NotInUse++;
                }
            }
        }
    }
    if (notInUseAllocated)
    {
        /*
         * Remove some more information on not in use fibres being allocated.
         */
        char buffer[100];
        sprintf(buffer,"WARNING:There are %d not in use fibres allocated.  You need to deallocate them.", notInUseAllocated);
        ConfMessage(interp, buffer);
        ConfMessage(interp, "   This could be due to reading in a field for the wrong plate.");
        ConfMessage(interp, "     or it  could be due to reading in an old field when the constants file has changed.");
        ConfMessage(interp, "   The \"Unallocated\" fibres will be wrong as a result of this and could end up negative");
    }
    
    /*  Release Sds Ids */
    
    if (objid) SdsFreeId(objid, &status);
    if (typeId) SdsFreeId(typeId, &status);
    objid = typeId = allocId = spectId = xId = yId = 0;
    
    /*  Now, we need to work out the unallocated targets compatible with 
     *  each fibre type. We have to look through the unallocated objects 
     *  and the unallocated guides to build up these statistics.
     */
     
    SdsFind(topid, "unallocObject", &objid, &status);
    if (status == STATUS__OK) {
    
        /*  There is an "unallocObject" section, so we can start to work through 
         *  it, building up the counts. Get the Sds Ids for the arrays that
         *  we are going to make use of, namely the 'type', 'spect' and
         *  'allocated' arrays - the allocated array may not exist, nor may
         *  the spect array. We also need the x and y positions to see if the
         *  target is in the field or not.
         */

        SdsFind(objid, "type", &typeId, &status);
        SdsInfo(typeId, name, &code, &ndims, dims, &status);
        NumObjects = dims[0];
        AllocArray = TRUE;
        SdsFind(objid, "allocated", &allocId, &status);
        if (status != 0) {
           AllocArray = FALSE;
           status = 0;
        }
        SpectKnown = TRUE;
        SdsFind(objid, "spectrograph", &spectId, &status);
        if (status != 0) {
           SpectKnown = FALSE;
           status = 0;
        }
        SdsFind(objid, "x", &xId, &status);
        SdsFind(objid, "y", &yId, &status);

        for (Index = 0; Index < NumObjects; Index++) {

            /*  For this target, we want to check that it isn't allocated,
             *  and then find out which pivot types it is compatible with.
             */

            SdsGet(typeId, sizeof(char), Index, &Type, &actlen, &status);
            SdsGet(xId, sizeof(int), Index, &XPosn, &actlen, &status);
            SdsGet(yId, sizeof(int), Index, &YPosn, &actlen, &status);
            Alloc = 0;
            if (AllocArray) {
               SdsGet(allocId, sizeof(char), Index, &Alloc, &actlen, &status);
            }
            Spect = 0;
            if (SpectKnown) {
               SdsGet(spectId, sizeof(char), Index, &Spect, &actlen, &status);
            }
            if (FpilOnField (*Instrument,XPosn,YPosn)) {
               if (!Alloc) {
                  IsSky = FpilIsTargetSky (*Instrument,Type);
                  FpilCompatiblePivotTypes (*Instrument,10,Type,Spect,
                                                     &NumPivTypes,PivTypes);
                  for (PivIndex = 0; PivIndex < NumPivTypes; PivIndex++) {
                     FibreType = PivTypes[PivIndex];
                     if (FibreType >= 0 && FibreType < InstNumFibreTypes) {
                         if (IsSky) {
                             FibreTypeStats[FibreType].Skies++;
                         } else {
                             FibreTypeStats[FibreType].Targets++;
                         }
                     } else {
                         DEBUG ("Invalid fibre type code %d\n",FibreType);
                     }
                  }
               }
            }
        }
        if (typeId) SdsFreeId(typeId, &status);
        if (spectId) SdsFreeId(spectId, &status);
        if (AllocArray) SdsFreeId(allocId, &status);
    }
    if (objid) SdsFreeId(objid, &status);
    objid = typeId = allocId = spectId = xId = yId = 0;
    
    SdsFind(topid, "unallocGuide", &objid, &status);
    if (status == STATUS__OK) {
    
        /*  There is an "unallocGuide" section, so we can start to work through 
         *  it, just as for the "unallocObject" section.
         */

        SdsFind(objid, "type", &typeId, &status);
        SdsInfo(typeId, name, &code, &ndims, dims, &status);
        NumObjects = dims[0];
        AllocArray = TRUE;
        SdsFind(objid, "allocated", &allocId, &status);
        if (status != 0) {
           AllocArray = FALSE;
           status = 0;
        }
        SpectKnown = TRUE;
        SdsFind(objid, "spectrograph", &spectId, &status);
        if (status != 0) {
           SpectKnown = FALSE;
           status = 0;
        }
        SdsFind(objid, "x", &xId, &status);
        SdsFind(objid, "y", &yId, &status);

        for (Index = 0; Index < NumObjects; Index++) {

            /*  For this target, we want to check that it isn't allocated,
             *  and then find out which pivot types it is compatible with.
             */

            SdsGet(typeId, sizeof(char), Index, &Type, &actlen, &status);
            SdsGet(xId, sizeof(int), Index, &XPosn, &actlen, &status);
            SdsGet(yId, sizeof(int), Index, &YPosn, &actlen, &status);
            Alloc = 0;
            if (AllocArray) {
               SdsGet(allocId, sizeof(char), Index, &Alloc, &actlen, &status);
            }
            Spect = 0;
            if (SpectKnown) {
               SdsGet(spectId, sizeof(char), Index, &Spect, &actlen, &status);
            }
            if (FpilOnField (*Instrument,XPosn,YPosn)) {
               if (!Alloc) {
                  FpilCompatiblePivotTypes (*Instrument,10,Type,Spect,
                                                     &NumPivTypes,PivTypes);
                  for (PivIndex = 0; PivIndex < NumPivTypes; PivIndex++) {
                     FibreType = PivTypes[PivIndex];
                     if (FibreType >= 0 && FibreType < InstNumFibreTypes) {
                         FibreTypeStats[FibreType].Targets++;
                     } else {
                         DEBUG ("Invalid fibre type %d returned\n",FibreType);
                     }
                  }
               }
            }
        }
       if (typeId) SdsFreeId(typeId, &status);
       if (spectId) SdsFreeId(spectId, &status);
       if (AllocArray) SdsFreeId(allocId, &status);
    }
    if (objid) SdsFreeId(objid, &status);
    
    /*  Report status */
    
    if (status == STATUS__OK) {
        interp->result = "OK";
        ReturnCode = TCL_OK;
    } else {
        interp->result = "Error";
        ReturnCode = TCL_ERROR;
    }
    
    return ReturnCode;
}


/*  ------------------------------------------------------------------------- */

/*
 *                    G e t  F i b r e  S t a t s  { }
 */
 
static int ConfGetFibreStats(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc, char *argv[])

/*+
 *  Command Name:
 *    GetFibreStats
 *
 *  Function:
 *    Get allocation statistics for a fibre type
 *
 *  Call:
 *    GetFibreStats index name enabled fibres skies guides 
 *                           objects unalloc sameAs mayShare targets skies notInUse
 *
 *  Description:
 *    Return information about the allocation statistics for a fibre type.
 *    Each fibre type has an index number, starting at one, corresponding to
 *    the index for the corresponding fibre type returned by FibreTypes{}.
 *
 *  Parameters:
 *    (>)  index (int)     index number - starting at zero
 *    (<)  name  (char)    The fibre type name.
 *    (<)  enabled (int)   True (non-zero) if the fibre type is currently
 *                         allocated.
 *    (<)  fibres (int)    Total number of fibres of this type.
 *    (<)  skies (int)     Number of fibres of this type allocated to sky.
 *    (<)  guides (int)    Number of fibres of this type allocated to guide
 *                         targets.
 *    (<)  objects (int)   Number of fibres of this type allocated to program
 *                         targets - ie not to sky or guide targets.
 *    (<)  unalloc (int)   Number of fibres of this type unallocated (but which
 *                          are in use and hence could be allocated).
 *    (<)  sameAs (int)    If the fibre type is treated as being the same as
 *                         a lower-numbered fibre type for the purposes of
 *                         classifying allocation results, this is the index
 *                         of that lower-numbered type. If negative, there
 *                         is no such lower index.
 *    (<)  mayShare (int)  If the fibre type is one that can be allocated to
 *                         both sky and non-sky targets, this is set to 1.
 *                         Otherwise, it is set to zero.
 *    (<)  targets (int)   Number of non-sky targets available to this fibre
 *                         type - these are unallocated non-sky targets 
 *                         within the field that are compatible with this type.
 *                         Note that many targets are compatible with a number
 *                         of different fibre types.
 *    (<)  skies (int)     Number of sky targets available to this fibre
 *                         type - these are unallocated sky targets 
 *                         within the field that are compatible with this type.
 *                         Note that many targets are compatible with a number
 *                         of different fibre types.
 *    (<) notInUse (int)   Number of fibres which are not in use.
 *
 *
 *   Support:
 *     Keith Shortridge, AAO
 *
 *   Version Date:
 *      8-Nov-2001  KS  Original version.
 *     19-Nov-2001  KS  Added sameAs and mayShare. 
 *      3-Sep-2002  KS  Added targets and skies. 
 *     07-Jan-2004  TJF Add notInUse.
 *-
 */
{
    char buffer[80];
    int fibreIndex;
    int Index;

    if (argc != 14) {
        interp->result = "wrong # args";
        return TCL_ERROR;
    }
    fibreIndex = atol(argv[1]);
    if ((fibreIndex < 0) || (fibreIndex >= InstNumFibreTypes)) {
        interp->result = "invalid index for GetFibreStats";
        return TCL_ERROR;
    }

    Tcl_SetVar(interp, argv[2], FibreTypeStats[fibreIndex].Name, 0);
    sprintf(buffer, "%d", UseFibreType[fibreIndex]);
    Tcl_SetVar(interp, argv[3], buffer, 0);
    sprintf(buffer, "%d", FibreTypeStats[fibreIndex].Fibres);
    Tcl_SetVar(interp, argv[4], buffer, 0);
    sprintf(buffer, "%d", FibreTypeStats[fibreIndex].AllocToSky);
    Tcl_SetVar(interp, argv[5], buffer, 0);
    sprintf(buffer, "%d", FibreTypeStats[fibreIndex].AllocToGuide);
    Tcl_SetVar(interp, argv[6], buffer, 0);
    sprintf(buffer, "%d", FibreTypeStats[fibreIndex].AllocToObject);
    Tcl_SetVar(interp, argv[7], buffer, 0);
    sprintf(buffer, "%d", FibreTypeStats[fibreIndex].Unallocated);
    Tcl_SetVar(interp, argv[8], buffer, 0);
    strcpy (buffer,"-1");
    for (Index = 0; Index < fibreIndex; Index++) {
        if (FpilPivotTypeMatch(*Instrument,Index,fibreIndex)) {
            sprintf(buffer, "%d", Index);
            break;
        }
    }
    Tcl_SetVar(interp, argv[9], buffer, 0);
    if (FpilPivotSharesSky(*Instrument,fibreIndex)) {
       Tcl_SetVar(interp, argv[10], "1", 0);
    } else {
        Tcl_SetVar(interp, argv[10], "0", 0);
    }      
    sprintf(buffer, "%d", FibreTypeStats[fibreIndex].Targets);
    Tcl_SetVar(interp, argv[11], buffer, 0);
    sprintf(buffer, "%d", FibreTypeStats[fibreIndex].Skies);
    Tcl_SetVar(interp, argv[12], buffer, 0);

    sprintf(buffer, "%d", FibreTypeStats[fibreIndex].NotInUse);
    Tcl_SetVar(interp, argv[13], buffer, 0);
    
    return TCL_OK;
}

/*  ------------------------------------------------------------------------- */

/*
 *                  S e t  S k y  C o u n t s  { }
 */
 
static int ConfSetSkyCounts(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc, char *argv[])

/*+
 *  Command name:
 *    SetSkyCounts
 *
 *  Function:
 *    Sets the number of sky targets to be allocated to a given fibre type.
 *  
 *  Call:
 *    SetSkyCounts
 *    SetSkyCounts Index Count
 *
 *  Description:
 *    The global array FibreTypeStats is used to maintain allocation 
 *    statistics for all the different fibre types. It also contains a 
 *    field FibreSkyCount which gives the number of fibres of this (or
 *    equivalent) types to be allocated to sky targets. (Or left
 *    unallocated for later allocation to sky, if there are no sky targets
 *    available.) If the form with no arguments is used, the counts for all
 *    types are set to -1, which indicates that no control is to be 
 *    exercised over the number of sky targets for the type in question.
 *   
 *   Parameters:
 *
 *    (>) Index  (int)   Fibre index, from 0 up. This should match a value
 *                       returned by FpilConfWhichSpect().
 *    (>) Count  (int)   Number of sky targets for this fibre type.
 *
 *   Returns:
 *     "OK" if succesful, some error indication otherwise
 *
 *   Support:
 *     Keith Shortridge, AAO
 *
 *   Version Date:
 *     19th Nov 2001. Original version. KS.    
 *-
 */
{
    const char *const *SpecNames;
    unsigned int NumFibreTypes;
    StatusType status = STATUS__OK;
    int Index;
    int ReturnCode;
    
    if ((argc != 1) && (argc != 3)) {
        interp->result = "Wrong # args";
        ReturnCode = TCL_ERROR;
    } else {
    
        FpilConSpecInfo(*Instrument, &NumFibreTypes, &SpecNames, &status);
        if (argc == 1) {

            /*  This is the form in which we simply clear out all the sky
             *  counts for each fibre type.
             */

            for (Index = 0; Index < NumFibreTypes; Index++) {
                FibreTypeStats[Index].FibreSkyCount = -1;
            }
            ReturnCode = TCL_OK;

        } else {

            /*  This is the form where we set the count for a given type */

            Index = atoi(argv[1]);
            if ((Index < 0) || (Index >= NumFibreTypes)) {
                sprintf(interp->result,"Index %d is invalid",Index);
                ReturnCode = TCL_ERROR;
            } else {
                FibreTypeStats[Index].FibreSkyCount = atoi(argv[2]);
                interp->result = "OK";
                ReturnCode = TCL_OK;
            }
        }
    }   
    
    return ReturnCode;
}

/*  ------------------------------------------------------------------------- */

/*
 *                     S d s  F i l e  N a m e  { }
 *
 *                  N e w  S d s  F i l e  N a m e  { }
 *
 *                   A s c i i  F i l e  N a m e  { }
 *
 *                  I m p o r t  F i l e  N a m e  { }
 *
 *                     D S S  F i l e  N a m e  { }
 *
 *                     L i s t  F i l e  N a m e  { }
 *
 */
 
static int ConfFileName(ClientData clientData, Tcl_Interp * interp,
    int argc DUNUSED, char *argv[])

/*+
 *
 *  Command names:
 *    SdsFileName
 *    NewSdsFileName
 *    AsciiFileName
 *    ImportFileName
 *    DSSFileName
 *    ListFileName
 *
 *  Function:
 *    Return name of Sds/Fld/Imp/Dss/list file corresponding to a 
 *    configuration file.
 *  
 *  Call:
 *    {cmd} name
 *
 *  Description:
 *    Return the name of file corresponding to the specified
 *    configuration file. This simply involves replacing whatever
 *    file extension is present with one of /sds/fld/imp/dss/lis, as
 *    specified in the client data.  For NewSdsFileName, it involves
 *    add _new.sds to the file name. So the differences between the way
 *    these different routines work is all 'hidden' in the way the commands
 *    are defined in ConfAddCommands().
 *   
 *   Parameters:
 *
 *    (<) name  (char)   Name of configuration file
 *
 *
 *   Support:
 *     Tony Farrell, AAO
 *
 *   Version Date:
 *     6-Oct-1994    
 *-
 */
{
    int l;
    if (argc != 2) {
        sprintf(interp->result, "%s:wrong # args", argv[0]);
        return TCL_ERROR;
    }
/*
 * Copy the input string to the result.
 */
    Tcl_SetResult(interp, argv[1], TCL_VOLATILE);

/*
 *  Find the period, indicating where the file type is.
 */
    l = strlen(argv[1]);
    for (l = strlen(argv[1]); l > 0; l--) {
        if (argv[1][l] == '.')
            break;
    }
/*
 *  Terminate the result so that we can add our desired type.
 *  (it is a bit naughty to assume we can write to interp->result[]
 *  here, it relies on an old behaviour of Tcl.  But is so hard to
 *  do it right, and the old behaviour really should keep working, there
 *  is too much old code out there.). In any case, this is done in many
 *  places in the code - although it would not be a huge job to update
 *  them to use a more strictly correct mechanism.
 */
    if (l > 0)
        interp->result[l] = 0;
/*
 * Append the desired file type.
 */
    Tcl_AppendResult(interp, (char *) (clientData), NULL);
    return TCL_OK;
}

/*  ------------------------------------------------------------------------- */

/*
 *                 C o m p r e s s  U n a l l o c  { }
 */
 
static int ConfCompressUnalloc(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc DUNUSED, char *argv[])

/*+
 *  Command name:
 *    CompressUnalloc
 *
 *  Function:
 *    Compress the Unallocated objects section
 *  
 *  Call:
 *    CompressUnalloc id
 *
 *  Description:
 *    Compress the unallocated objects and guide sections of a 
 *    configuration structure by removing any objects which have
 *    their "allocated" flag set. Also removes the corresponding "index"
 *    section from the allocated objects structure.
 *   
 *
 *   Parameters:
 *
 *    (<) id  (Sds Id)   SDS id of configuration structure
 *
 *   Support:
 *     Jeremy Bailey, AAO
 *
 *   Version Date:
 *     6-Oct-1994 
 *
 *   Modified:
 *     20th Jan 2003. KS. Failure to find "index" no longer treated
 *                        as an error - just indicates this routine
 *                        has already been run on this structure.   
 *-
 */
{
    SdsIdType topid, uid, objid, indid;
    StatusType status = STATUS__OK;
    StatusType indexStatus = STATUS__OK;

    topid = atol(argv[1]);
    SdsFind(topid, "unallocObject", &uid, &status);
    ConfRemoveAlloc(uid, &status);
    SdsFreeId(uid, &status);
    SdsFind(topid, "unallocGuide", &uid, &status);
    ConfRemoveAlloc(uid, &status);
    SdsFreeId(uid, &status);
    SdsFind(topid, "objects", &objid, &status);
    SdsFind(objid, "index", &indid, &indexStatus);
    if (indexStatus == STATUS__OK) {
        SdsDelete(indid, &status);
        SdsFreeId(indid, &status);
    }
    SdsFreeId(objid, &status);

    if (status == STATUS__OK) {
        return TCL_OK;
    } else {
        sprintf(interp->result,
            "Error Compressing the Unallocated Object Structures");
        return TCL_ERROR;
    }

}

/*  ------------------------------------------------------------------------- */

/*
 *                  N u m b e r  P a r a m e t e r s  { }
 */
 
static int ConfNumberParameters(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc DUNUSED, char *argv[]DUNUSED)

/*+
 *  Command name:
 *    NumberParameters
 *
 *  Function:
 *    Return the number of allocation parameters
 *  
 *  Call:
 *    NumberParameters
 *
 *  Description:
 *    Return the number of allocation parameters. This is a wrap-up routine
 *    to make ConfigNumberParameters() available to Tcl.
 *   
 *   Parameters:
 *
 *   Returns:
 *     The number of allocation parameters
 *
 *   Support:
 *     Jeremy Bailey, AAO
 *
 *   Version Date:
 *     6-Oct-1994    
 *-
 */
{
    int n;

    n = ConfigNumberParameters();
    sprintf(interp->result, "%d", n);
    return TCL_OK;
}

/*  ------------------------------------------------------------------------- */

/*
 *                  A l g o r i t h m  S t r i n g  { }
 */
 
static int ConfAlgorithmString(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc DUNUSED, char *argv[]DUNUSED)

/*+
 *  Command name:
 *    AlgorithmString
 *
 *  Function:
 *    Return a string describing the allocation algorithm.
 *  
 *  Call:
 *    AlgorithmString
 *
 *  Description:
 *    Return a string describing the allocation algorithm. This is a wrap-up
 *    routine to make ConfigAlgorithmString() available to Tcl.
 *   
 *   Parameters:
 *
 *   Returns:
 *     A description of the algorithm in use.
 *
 *   Support:
 *     Keith Shortridge, AAO
 *
 *   Version Date:
 *     9-Feb-2009    
 *-
 */
{

    sprintf(interp->result, "%s", ConfigAlgorithmString());
    return TCL_OK;
}

/*  ------------------------------------------------------------------------- */

/*
 *                     G e t  P a r a m e t e r  { }
 */
 
static int ConfGetParameter(ClientData clientData DUNUSED, Tcl_Interp * interp,
    int argc DUNUSED, char *argv[])

/*+
 *  Command name:
 *    GetParameter
 *
 *  Function:
 *    Return the value of an allocation parameter
 *  
 *  Call:
 *    GetParameter name
 *
 *  Description:
 *    Return the value of an allocation parameter.
 *   
 *   Parameters:
 *      (>) name (char)   The parameter name
 *
 *   Returns:
 *     The value of the allocation parameter
 *
 *   Support:
 *     Jeremy Bailey, AAO
 *
 *   Version Date:
 *     6-Oct-1994    
 *-
 */
{
    char value[100];
    StatusType status = STATUS__OK;

    ConfigGetParameter(argv[1], value, 100, &status);

    if (status == STATUS__OK) {
        sprintf(interp->result, "%s", value);
        return TCL_OK;
    } else {
        sprintf(interp->result, "Error Getting Parameter %s", argv[1]);
        return TCL_ERROR;
    }
}

/*  ------------------------------------------------------------------------- */

/*
 *                     S e t  P a r a m e t e r  { }
 */
 
static int ConfSetParameter(ClientData clientData DUNUSED, Tcl_Interp * interp,
    int argc DUNUSED, char *argv[])

/*+
 *  Command name:
 *    SetParameter
 *
 *  Function:
 *    Set the value of an allocation parameter
 *  
 *  Call:
 *    SetParameter name value
 *
 *  Description:
 *    Set the value of an allocation parameter.
 *   
 *   Parameters:
 *      (>) name (char)   The parameter name
 *      (>) value (char)  The parameter value
 *
 *   Support:
 *     Jeremy Bailey, AAO
 *
 *   Version Date:
 *     6-Oct-1994    
 *-
 */
{
    StatusType status = STATUS__OK;

    ConfigSetParameter(argv[1], argv[2], &status);

    if (status == STATUS__OK) {
        return TCL_OK;
    } else {
        sprintf(interp->result,
            "Error Setting Parameter %s to value %s", argv[1], argv[2]);
        return TCL_ERROR;
    }
}


/*  ------------------------------------------------------------------------- */

/*
 *                  P a r a m e t e r  D e t a i l s  { }
 */
 
static int ConfParameterDetails(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc DUNUSED, char *argv[])

/*+
 *  Command name:
 *    ParameterDetails
 *
 *  Function:
 *    Return details of an allocation parameter
 *  
 *  Call:
 *    ParameterDetails number name type low high nvalues values
 *
 *  Description:
 *    Set the value of an allocation parameter.
 *   
 *   Parameters:
 *      (>) number (char) The parameter number
 *      (<) name (char)   variable to receive the parameter name
 *      (<) type (char)   variable to receive the parameter type
 *      (<) low  (char)   variable to receive the low limit on the value
 *      (<) high (char)   variable to receive the high limit on the value
 *      (<) nvalues (char) variable to receive the number of possible values
 *      (<) values (char) variable to recive the list of possible values
 *
 *   Support:
 *     Jeremy Bailey, AAO
 *
 *   Version Date:
 *     6-Oct-1994    
 *-
 */
{
    StatusType status = STATUS__OK;
    int nvalues;
    double limits[2];
    char *name, type[2], **values;
    char buffer[80];
    int i;

    ConfigParameterDetails(atoi(argv[1]), &name, &(type[0]), limits,
        &nvalues, &values, &status);
    type[1] = 0;

    if (status == STATUS__OK) {
        Tcl_SetVar(interp, argv[2], name, 0);
        Tcl_SetVar(interp, argv[3], type, 0);
        if (type[0] == 'I') {
            sprintf(buffer, "%d", (int) (limits[0]));
            Tcl_SetVar(interp, argv[4], buffer, 0);
            sprintf(buffer, "%d", (int) (limits[1]));
            Tcl_SetVar(interp, argv[5], buffer, 0);
        } else {
            sprintf(buffer, "%f", limits[0]);
            Tcl_SetVar(interp, argv[4], buffer, 0);
            sprintf(buffer, "%f", limits[1]);
            Tcl_SetVar(interp, argv[5], buffer, 0);
        }
        sprintf(buffer, "%d", nvalues);
        Tcl_SetVar(interp, argv[6], buffer, 0);
        /* This wouldn't have been called in the following loop! BM */
        if(nvalues == 0) { Tcl_SetVar(interp, argv[7], buffer, 0); }
        else {
        Tcl_UnsetVar(interp, argv[7], 0);
        for (i = 0; i < nvalues; i++) {
            Tcl_SetVar(interp, argv[7], values[i],
                TCL_LIST_ELEMENT | TCL_APPEND_VALUE);
        }
        }
        return TCL_OK;
    } else {
        sprintf(interp->result, "Error Getting Parameter %s Details", argv[1]);
        return TCL_ERROR;
    }
}

/*  ------------------------------------------------------------------------- */

/*
 *                          G e t  T y p e  { }
 */
 
static int ConfGetType(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc, char *argv[])

/*+
 *  Command name:
 *    GetType
 *
 *  Function:
 *    Determines the colour to use to display a button based on its fibre type
 *  
 *  Description:
 *    This routine returns (to tcl, through the interpreter's result string)
 *    the colour to use when drawing a button. It does this on the basis of
 *    the fibre type associated with the pivot. If the fibre is disabled, it
 *    returns "Blank" and leaves the choice of how to display the fibre to
 *    the Tcl/Tk layer. The name is perhaps misleading - it gets a colour
 *    based on the type, not the type itself.
 *
 *   Call:
 *      GetType pivid fibre_no fibre_combo
 *
 *   Parameters:
 *
 *    (>) pivid  (Sds Id)   SDS identifier of pivot details structure. (This
 *                          is ignored now, but was needed in earlier versions
 *                          of the routine. Fpil now handles this structure.)
 *    (>) fibre_no (int)    Number of pivot in question - from 0 up.
 *    (>) fibre_combo (int) Index number of the selected fibre combination,
 *                          if such a selection has been made. -1 indicates
 *                          that no known configuration has been selected.
 *                          Values from 0 up are index values that match the
 *                          order returned by ConfFibreCombos().
 *
 *   Returns:
 *     "OK" if succesful, "Error" otherwise.
 *-
 */
{
    long fibre_no;
    int fibreType;
    int fibreCombo;
    char colourString[32];

    /*  Validate the arguments */

    if (argc != 4) {
        interp->result = "Wrong # args";
        return TCL_ERROR;
    }
    fibre_no = atol(argv[2]) - 1;
    if (fibre_no == -1)
        return TCL_OK;
    fibreCombo = atol(argv[3]);
    if ((fibre_no < 0) || (fibre_no >= (int) FpilGetNumPivots(*Instrument))) {
        sprintf(interp->result,
            "Illegal Pivot Number %ld\nMust be between 0 and %d",
            fibre_no + 1, FpilGetNumPivots(*Instrument));
        return TCL_ERROR;
    }

    /*  The colour choice is made by FpilFibreColour, and is based on the
     *  fibre type returned by ConfWhichSpec().  If the type is one that we're 
     *  not currently allocating (something Fpil doesn't know about) we 
     *  explicitly set it to Blank first.
     */

    fibreType = ConfWhichSpec((int) fibre_no, 0);
    if (!UseFibreType[fibreType]) {
        sprintf(interp->result, "Blank");
    } else {
        FpilFibreColour(*Instrument, fibreType, fibreCombo, useGrey,
                                       colourString, sizeof(colourString));
        sprintf(interp->result, "%s", colourString);
    }
    return TCL_OK;
}

/*  ------------------------------------------------------------------------- */

/*
 *                   R e a l l y  E x i s t s  { }
 */
 
static int ConfReallyExists(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc, char *argv[])

/*+
 *  Command name:
 *    ReallyExists
 *
 *  Function:
 *    Determines if a button really exists based on its fibre type
 *  
 *  Description:
 *    This routine returns (to tcl, through the interpreter's result string)
 *    a flag (0 for false, 1 for true) that indicates whether the button
 *    in question really exists. (Some instruments, such as FLAMES, have
 *    fibres included in the instrument descriptions that are just 
 *    placeholders for fibres that may be installed some time in the future.
 *
 *   Parameters:
 *
 *    (>) pivid  (Sds Id)   SDS identifier of pivot details structure. (This
 *                          is ignored now, but was needed in earlier versions
 *                          of the routine. Fpil now handles this structure.)
 *    (>) fibre_no (int)    Number of pivot in question - from 0 up.
 *
 *   Returns:
 *     "OK" if succesful, "Error" otherwise.
 *-
 */
{
    long fibre_no;
    int fibreType;

    /*  Validate the arguments */

    if (argc != 3) {
        interp->result = "Wrong # args";
        return TCL_ERROR;
    }
    fibre_no = atol(argv[2]) - 1;
    if (fibre_no == -1)
        return TCL_OK;
    if ((fibre_no < 0) || (fibre_no >= (int) FpilGetNumPivots(*Instrument))) {
        sprintf(interp->result,
            "Illegal Pivot Number %ld\nMust be between 0 and %d",
            fibre_no + 1, FpilGetNumPivots(*Instrument));
        return TCL_ERROR;
    }

    /*  The check for existence is made by FpilFibreReallyExists, and is
     *  based entirely on the fibre type.
     */

    fibreType = ConfWhichSpec((int) fibre_no, 0);

    if (FpilFibreReallyExists(*Instrument, fibreType)) {
        strcpy (interp->result,"1");
    } else {
        strcpy (interp->result,"0");
    }
    return TCL_OK;
}

/*  ------------------------------------------------------------------------- */

/*
 *                A d d  S k y  P o s i t i o n  { }
 */
 
static int ConfAddSkyPosition(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc, char *argv[])
    
/*+
 *  Command name:
 *    AddSkyPosition
 *
 *  Function:
 *    Add a new (user/method supplied) sky position to the unallocated
 *    objects sub-struct
 *  
 *  Description:
 *    Generate a new ra,dec,pair from a user-supplied x,y position
 *    Allocate the specified fibre to this new target. It returns the
 *    index of the new entry in the unallocated objects structure, which
 *    will have been enlarged in the process.
 *
 *  Call:
 *    AddSkyPosition id xsky ysky object zoom xoff yoff
 *
 *  Parameters:
 *
 *    (>) id   (Sds Id)  SDS identifier of configuration structure.
 *    (>) xsky (double)  input x-sky;
 *    (>) ysky (double)  input y-sky
 *    (<) object (int)   Object number - starting from one.
 *    (>) zoom (double)  The current zoom factor.
 *    (>) xoff (double)  Left window scroll position, range 0 to 1.
 *    (>) yoff (double)  Top window scroll position, range 0 to 1.
 *
 *   Returns:
 *     "OK" if succesful, "Error" otherwise
 *
 *   Support:
 *     Gavin Dalton, Oxford
 *
 *   Version Date:
 *     6-Oct-1994    
 *-
 */
 
{
    double xsky, ysky;
    double xoff, yoff;
    long object;
    SdsIdType topid;
    double zoom;
    char mbuf[100], label[NAMELEN];
    StatusType status = STATUS__OK;
    short allocSkyCnt, unallocSkyCnt;
    SdsIdType field;
    int is_2df_diff_wlen = 0;
    char targetType, spectType = 0;

    if (argc != 8) {
        interp->result = "Wrong # args";
        return TCL_ERROR;
    }
    topid = (SdsIdType) (atol(argv[1]));
    xsky = atof(argv[2]);
    ysky = atof(argv[3]);
    zoom = atof(argv[5]);
    xoff = atof(argv[6]);
    yoff = atof(argv[7]);

/*
 * Convert scroll position to pixel offsets. 
 */
    xoff *= MimicWindowSize;
    yoff *= MimicWindowSize;

/*
 *  Need the reverse of FieldPlateToDispX/Y ??
 */
    xsky = DispToFieldPlateX(xsky, xoff, zoom);
    ysky = DispToFieldPlateY(ysky, yoff, zoom);


/*
 * Need a unique name for the sky.
 */
    SdsFind(topid, "fieldData", &field, &status);
    ArgGets(field, "unallocSky", &unallocSkyCnt, &status);
    if (status == STATUS__OK) {
        allocSkyCnt = 0;
        ArgGets(field, "allocSky", &allocSkyCnt, &status);
        if (status == SDS__NOITEM)
            ErsAnnul(&status);
    }
    SdsFreeId(field, &status);


    is_2df_diff_wlen = (strncmp(FpilGetInstName(*Instrument), "2dF",3) == 0) &&
                            (gwave != gwave2); 

    FpilTargetSkyType(*Instrument, &targetType, &spectType); 

    if (is_2df_diff_wlen) {
        /* We are using 2df with different wavelengths of the two
    * spectrograps. Create two sky positions rather than just one, one
    * for each spectrpgraph. The target types are hard wired, which is
    * truly awful, but it did quickly get the job done.
    */
        sprintf(label, "User_Sky_%d_s1", unallocSkyCnt + allocSkyCnt + 1);           
        ConfDoAddSkyPosition(topid, xsky, ysky, label, 'B', &object, &status);   
        sprintf(label, "User_Sky_%d_s2", unallocSkyCnt + allocSkyCnt + 1);           
        ConfDoAddSkyPosition(topid, xsky, ysky, label, 'C', &object, &status);   
    } else {
        sprintf(label, "User_Sky_%d", unallocSkyCnt + allocSkyCnt + 1);
        ConfDoAddSkyPosition(topid, xsky, ysky, label, targetType, &object, &status);
    }

    sprintf(mbuf, "%ld", object + 1);
    Tcl_SetVar(interp, argv[4], mbuf, 0);


/*
printf("Adding user sky %s, unallocSkyCnt = %d, allocSkyCnt = %d\n",
       label, unallocSkyCnt, allocSkyCnt);
*/

    if (status != STATUS__OK) {
        char buff[200];
        MessGetMsg(status, 0, sizeof(buff), buff);
        Tcl_AppendResult(interp, "\nError Setting Sky Position\n",
            (char *) NULL);
        Tcl_AppendResult(interp, buff, (char *) NULL);
        return TCL_ERROR;


    }
    return TCL_OK;

}

/*  ------------------------------------------------------------------------- */

/*
 *                S e t  I m p o r t  F i l e  { }
 */
 
static int ConfSetImportFile(ClientData clientData DUNUSED,
    Tcl_Interp * interp DUNUSED, int argc DUNUSED, char *argv[]DUNUSED)
    
/*+
 *  Command name:
 *    SetImportFile
 *
 *  Function:
 *    Sets the file to be imported by the allocation algorithm.
 *  
 *  Description:
 *    This routine provides an interface from Tcl to the allocation
 *    algorithm routine ConfigSetImportFile(). It specifies the name
 *    of the file to be used to import a set of allocations.
 *
 *  Call:
 *    SetImportFile filename
 *
 *  Parameters:
 *
 *    (>) filename  (string)  Name of import file.
 *
 *   Support:
 *     Gavin Dalton, Oxford
 *-
 */
{
    ConfigSetImportFile(argv[1]);
    return (0);
}
/*  ------------------------------------------------------------------------- */

/*
 *                    S e t  C B S  P a i r s  { }
 */
 
static int ConfSetCBSPairs(ClientData clientData DUNUSED, Tcl_Interp * interp,
                           int argc DUNUSED, char *argv[])
    
/*+
 *  Command name:
 *   SetCBSPairs 
 *
 *  Function:
 *    Generates a cross-beam switching pair for each program target.
 *  
 *  Description:
 *    This routine creates a set of program targets 
 *    (one new target for each program target)
 *    in the unallocated objects section of the configuration data, 
 *    based on an offset in RA and DEC. 
 *    This is needed for fields to be observed using Nod & Shuffle with
 *    Cross-Beam Switching for AAOmega.
 *
 *  Call:
 *    SetCBSPairs id
 *
 *  Parameters:
 *
 *    (>) id  (Sds Id)  Id of top level of the configuration data structure.
 *
 *   Author:
 *     Brent Miszalski, AAO/Macquarie.
 *
 *   Modified:
 *     10th Aug 2009. Added proper motion arguments to calls to 
 *                    ConfDoAddCBSPair. KS.
 *-
 */
{
    /*char name[NAMELEN];*/
    StatusType status = STATUS__OK;
    SdsIdType topid,uid,fid;
    SdsIdType nameID,raID,decID,typeID,priorityID,
        magnitudeID,pIdID,commentID, wlenIndexID,pmraID,pmdecID;
    void* nameptr,*raptr,*decptr,
        *typeptr,*priorityptr,*magnitudeptr,*pIdptr,*commentptr,
        *wlenIndexptr,*pmraptr,*pmdecptr;
    int NTargets;
    double ra_cen,dec_cen;
    unsigned long buffer;
    int i;
    double* ra, *dec;
    double* rr, *dd;
    double* rapair, *decpair;
    double* pmra, *pmdec;
    char* type,*tt;
    short* priority,*pp;
    short* wlenIndex,*wlenpp;
    double* magnitude,*mm,*pmrap,*pmdecp;
    short* pId,*pi;
    int DeltaRaCBSArcsec;
    int DeltaDecCBSArcsec;
    double DeltaRaCBS;
    double DeltaDecCBS;
    int IncludeFid;

    /* Store & convert the offset values to degrees in radians */
    DeltaRaCBS = atoi(argv[2])*(PI/(3600*180));
    DeltaDecCBS = atoi(argv[3])*(PI/(3600*180));
    DeltaRaCBSArcsec = atoi(argv[2]);
    DeltaDecCBSArcsec = atoi(argv[3]);
    IncludeFid = atoi(argv[4]);
    printf("IncludeFid %d\n",IncludeFid);
    /*printf("SetCBSPairs DeltaRA: %d DeltaDEC: %d\n",DeltaRaCBS,DeltaDecCBS);*/

    topid = (SdsIdType) (atol(argv[1]));

    SdsFind(topid, "fieldData", &fid, &status);
    ArgGetd(fid, "appRa", &ra_cen, &status);
    ArgGetd(fid, "appDec", &dec_cen, &status);

    SdsFind(topid, "unallocObject", &uid, &status);
    SdsFind(uid, "name", &nameID, &status);
    SdsFind(uid, "ra", &raID, &status);
    SdsFind(uid, "dec", &decID, &status);
    SdsFind(uid, "type", &typeID, &status);
    SdsFind(uid, "priority", &priorityID, &status);
    SdsFind(uid, "magnitude", &magnitudeID, &status);
    SdsFind(uid, "wlenIndex", &wlenIndexID, &status);
    
    /*  We allow pmra and pmdec to be optional structure items. */
    
    pmraID = pmdecID = 0;
    if (status == STATUS__OK) {
       SdsFind(uid, "pmra", &pmraID, &status);
       SdsFind(uid, "pmdec", &pmdecID, &status);
       if (status != STATUS__OK) {
          pmdecID = 0;
          pmraID = 0;
          status = STATUS__OK;
       }
    }
    SdsFind(uid, "pId", &pIdID, &status);
    SdsFind(uid, "comment", &commentID, &status);
    
    SdsSize(raID,&buffer,&status);
    SdsPointer(raID,&raptr,&buffer,&status);
    NTargets = buffer/sizeof(double);
    ra = (double*)raptr;
    rapair = (double*)malloc(NTargets*sizeof(double));
    rr = (double*)malloc(NTargets*sizeof(double));
    memcpy(rr,ra,NTargets*sizeof(double));
    
    SdsSize(decID,&buffer,&status);
    SdsPointer(decID,&decptr,&buffer,&status);
    dec = (double*)decptr;
    decpair = (double*)malloc(NTargets*sizeof(double));
    dd = (double*)malloc(NTargets*sizeof(double));
    memcpy(dd,dec,NTargets*sizeof(double));
    
    SdsSize(typeID,&buffer,&status);
    SdsPointer(typeID,&typeptr,&buffer,&status);
    type = (char*)typeptr;
    tt = (char*)malloc(NTargets*sizeof(char));
    memcpy(tt,type,NTargets*sizeof(char));
    
    SdsSize(priorityID,&buffer,&status);
    SdsPointer(priorityID,&priorityptr,&buffer,&status);
    priority = (short*)priorityptr;
    pp = (short*)malloc(NTargets*sizeof(short));
    memcpy(pp,priority,NTargets*sizeof(short));
   
    SdsSize(magnitudeID,&buffer,&status);
    SdsPointer(magnitudeID,&magnitudeptr,&buffer,&status);
    magnitude = (double*)magnitudeptr;
    mm = (double*)malloc(NTargets*sizeof(double));
    memcpy(mm,magnitude,NTargets*sizeof(double));

    SdsSize(wlenIndexID,&buffer,&status);
    SdsPointer(wlenIndexID,&wlenIndexptr,&buffer,&status);
    wlenIndex = (short*)wlenIndexptr;
    wlenpp = (short*)malloc(NTargets*sizeof(short));
    memcpy(wlenpp,wlenIndex,NTargets*sizeof(short));

    pmrap = (double*)malloc(NTargets*sizeof(double));
    if (pmraID) {
       SdsSize(pmraID,&buffer,&status);
       SdsPointer(pmraID,&pmraptr,&buffer,&status);
       pmra = (double*)pmraptr;
       memcpy(pmrap,pmra,NTargets*sizeof(double));
    } else {
       for(i=0;i<NTargets;i++){
          pmrap[i] = 0.0;
       }
    }
    pmdecp = (double*)malloc(NTargets*sizeof(double));
    if (pmdecID) {
       SdsSize(pmdecID,&buffer,&status);
       SdsPointer(pmdecID,&pmdecptr,&buffer,&status);
       pmdec = (double*)pmdecptr;
       memcpy(pmdecp,pmdec,NTargets*sizeof(double));
    } else {
       for(i=0;i<NTargets;i++){
          pmdecp[i] = 0.0;
       }
    }
       
    
    SdsSize(pIdID,&buffer,&status);
    SdsPointer(pIdID,&pIdptr,&buffer,&status);
    pId = (short*)pIdptr;
    pi = (short*)malloc(NTargets*sizeof(short));
    memcpy(pi,pId,NTargets*sizeof(short));
    
    SdsSize(nameID,&buffer,&status);
    SdsPointer(nameID,&nameptr,&buffer,&status);

    {
        /* This block just allows us to put these declarations here using NTargets value determined above */
        char name[NTargets][NAMELEN];
        char comment[NTargets][NAMELEN];

        for(i=0;i<NTargets;i++){
            strncpy(name[i],nameptr+(i*NAMELEN),NAMELEN);
        }
        
        SdsSize(commentID,&buffer,&status);
        SdsPointer(commentID,&commentptr,&buffer,&status);
        
        for(i=0;i<NTargets;i++){
            strncpy(comment[i],commentptr+(i*NAMELEN),NAMELEN);
        }
        
        if(DeltaRaCBSArcsec != 0 && DeltaDecCBSArcsec == 0){
            /* The so-called "trivial" case with the offset only in RA */
            for(i=0;i<NTargets;i++){
                if(tt[i] != 'S'){
                    
                    /* This formula needs CHECKING! */
                    /* Should the dec_cen be substituted for dd[i] (dec of each object) */
                    rapair[i] = rr[i] + DeltaRaCBS/cos(dec_cen/15);
                    decpair[i] = dd[i];
                    strcat(name[i],"-cbspair");
                    /*strcat(comment[i],"-cbspair");*/
                    ConfDoAddCBSPair(topid,name[i],rapair[i],decpair[i],
                                     tt[i],pp[i],mm[i],pi[i],wlenpp[i],
                                     pmrap[i],pmdecp[i],comment[i],&status);
                }
            }
        }
        if(DeltaDecCBSArcsec != 0){
            
            double T;
            double dra;
            double ddec;
            double shiftedra;
            /* The more complicated case where we have an offset in DEC (and possibly RA) */
            for(i=0;i<NTargets;i++){
                if(tt[i] != 'S'){
                    /* These formulae need CHECKING! */
                    shiftedra = rr[i] + DeltaRaCBS/cos(dec_cen/15);               
                    T = atan(1/(-tan(fabs(shiftedra-ra_cen))*sin(dec_cen)));
                    dra = -DeltaDecCBS*cos(T)/cos(dd[i]/15);
                    ddec= -DeltaDecCBS*sin(T);
                    
                    rapair[i] = shiftedra + dra;
                    decpair[i] = dd[i] + ddec;
                    strcat(name[i],"-cbspair");
                    /*strcat(comment[i],"-cbspair");*/
                    ConfDoAddCBSPair(topid,name[i],rapair[i],decpair[i],
                                     tt[i],pp[i],mm[i],pi[i],wlenpp[i],
                                     pmrap[i],pmdecp[i],comment[i],&status);
                }
            }
        }
    
    }  /* Block for declaration of name and commentID */

    if(IncludeFid == 1)
    {
        /*Nasty...*/
        SdsFreeId(uid,&status);
        SdsFreeId(fid,&status);
        SdsFreeId(nameID,&status);
        SdsFreeId(raID,&status);
        SdsFreeId(decID,&status);
        SdsFreeId(typeID,&status);
        SdsFreeId(priorityID,&status);
        SdsFreeId(magnitudeID,&status);
        SdsFreeId(pIdID,&status);
        SdsFreeId(commentID,&status);
        if (rr != NULL) { free (rr); rr = NULL; } 
        if (rapair != NULL) { free (rapair); rapair = NULL; } 
        if (dd != NULL) { free (dd); dd = NULL; } 
        if (decpair != NULL) { free (decpair); decpair = NULL; } 
        if (tt != NULL) { free (tt); tt = NULL; } 
        if (pp != NULL) { free (pp); pp = NULL; } 
        if (mm != NULL) { free (mm); mm = NULL; } 
        if (pi != NULL) { free (pi); pi = NULL; } 
        if (wlenpp != NULL) { free (wlenpp) ; wlenpp = NULL; }
    
        SdsFind(topid, "unallocGuide", &uid, &status);
        SdsFind(uid, "name", &nameID, &status);
        SdsFind(uid, "ra", &raID, &status);
        SdsFind(uid, "dec", &decID, &status);
        SdsFind(uid, "type", &typeID, &status);
        SdsFind(uid, "priority", &priorityID, &status);
        SdsFind(uid, "magnitude", &magnitudeID, &status);
        SdsFind(uid, "pId", &pIdID, &status);
        SdsFind(uid, "comment", &commentID, &status);
        
        SdsSize(raID,&buffer,&status);
        SdsPointer(raID,&raptr,&buffer,&status);
        NTargets = buffer/sizeof(double);
        ra = (double*)raptr;
        rapair = (double*)malloc(NTargets*sizeof(double));
        rr = (double*)malloc(NTargets*sizeof(double));
        memcpy(rr,ra,NTargets*sizeof(double));
        
        SdsSize(decID,&buffer,&status);
        SdsPointer(decID,&decptr,&buffer,&status);
        dec = (double*)decptr;
        decpair = (double*)malloc(NTargets*sizeof(double));
        dd = (double*)malloc(NTargets*sizeof(double));
        memcpy(dd,dec,NTargets*sizeof(double));
        
        SdsSize(typeID,&buffer,&status);
        SdsPointer(typeID,&typeptr,&buffer,&status);
        type = (char*)typeptr;
        tt = (char*)malloc(NTargets*sizeof(char));
        memcpy(tt,type,NTargets*sizeof(char));
        
        SdsSize(priorityID,&buffer,&status);
        SdsPointer(priorityID,&priorityptr,&buffer,&status);
        priority = (short*)priorityptr;
        pp = (short*)malloc(NTargets*sizeof(short));
        memcpy(pp,priority,NTargets*sizeof(short));
       
        SdsSize(magnitudeID,&buffer,&status);
        SdsPointer(magnitudeID,&magnitudeptr,&buffer,&status);
        magnitude = (double*)magnitudeptr;
        mm = (double*)malloc(NTargets*sizeof(double));
        memcpy(mm,magnitude,NTargets*sizeof(double));

        SdsSize(pIdID,&buffer,&status);
        SdsPointer(pIdID,&pIdptr,&buffer,&status);
        pId = (short*)pIdptr;
        pi = (short*)malloc(NTargets*sizeof(short));
        memcpy(pi,pId,NTargets*sizeof(short));
        
        SdsSize(nameID,&buffer,&status);
        SdsPointer(nameID,&nameptr,&buffer,&status);
        
        {
            /* This block just allows us to put these declarations here using NTargets value determined above */
            char fidname[NTargets][NAMELEN];
            char fidcomment[NTargets][NAMELEN];

            for(i=0;i<NTargets;i++){
                strncpy(fidname[i],nameptr+(i*NAMELEN),NAMELEN);
            }
            
            SdsSize(commentID,&buffer,&status);
            SdsPointer(commentID,&commentptr,&buffer,&status);
            
            for(i=0;i<NTargets;i++){
                strncpy(fidcomment[i],commentptr+(i*NAMELEN),NAMELEN);
            }
            if(DeltaRaCBSArcsec != 0 && DeltaDecCBSArcsec == 0){
                /* The so-called "trivial" case with the offset only in RA */
                for(i=0;i<NTargets;i++){
                    if(tt[i] == 'F'){
                        
                        /* This formula needs CHECKING! */
                        /* Should the dec_cen be substituted for dd[i] (dec of each object) */
                        rapair[i] = rr[i] + DeltaRaCBS/cos(dec_cen/15);
                        decpair[i] = dd[i];
                        strcat(fidname[i],"-cbspair");
                        /*strcat(fidcomment[i],"-cbspair");*/
                        ConfDoAddCBSPair(topid,fidname[i],rapair[i],decpair[i],
                                     tt[i],pp[i],mm[i],pi[i],-1,
                                     pmrap[i],pmdecp[i],fidcomment[i],&status);
                    }
                }
            }
            if(DeltaDecCBSArcsec != 0){
                
                double T;
                double dra;
                double ddec;
                double shiftedra;
                /* The more complicated case where we have an offset in DEC (and possibly RA) */
                for(i=0;i<NTargets;i++){
                    if(tt[i] == 'F'){
                        /* These formulae need CHECKING! */
                        shiftedra = rr[i] + DeltaRaCBS/cos(dec_cen/15);               
                        T = atan(1/(-tan(fabs(shiftedra-ra_cen))*sin(dec_cen)));
                        dra = -DeltaDecCBS*cos(T)/cos(dd[i]/15);
                        ddec= -DeltaDecCBS*sin(T);
                        
                        rapair[i] = shiftedra + dra;
                        decpair[i] = dd[i] + ddec;
                        strcat(fidname[i],"-cbspair");
                        /*strcat(fidcomment[i],"-cbspair");*/
                        ConfDoAddCBSPair(topid,fidname[i],rapair[i],decpair[i],
                                     tt[i],pp[i],mm[i],pi[i],-1,
                                     pmrap[i],pmdecp[i],fidcomment[i],&status);
                    }
                }
            }
        } /* Block for declaration of fidname and fidcomment */
    
    }  /* Unallocated guide */
    /* There really should be some OTHER proper cleanup things here 
     * for all the malloc'd arrays and void* ptrs above
     */
    
    
    /* If this is uncommented then bad things happen...
     * beware the jabberwocky
     */
    /* 
       SdsFreeId(uid,&status);
       SdsFreeId(fid,&status);
    */
    SdsFreeId(nameID,&status);
    SdsFreeId(raID,&status);
    SdsFreeId(decID,&status);
    SdsFreeId(typeID,&status);
    SdsFreeId(priorityID,&status);
    SdsFreeId(magnitudeID,&status);
    SdsFreeId(wlenIndexID,&status);
    SdsFreeId(pIdID,&status);
    SdsFreeId(commentID,&status);
    if (rr != NULL)      { free (rr)      ; rr = NULL;      } 
    if (rapair != NULL)  { free (rapair)  ; rapair = NULL;  } 
    if (dd != NULL)      { free (dd)      ; dd = NULL;      } 
    if (decpair != NULL) { free (decpair) ; decpair = NULL; } 
    if (tt != NULL)      { free (tt)      ; tt = NULL;      } 
    if (pp != NULL)      { free (pp)      ; pp = NULL;      } 
    if (mm != NULL)      { free (mm)      ; mm = NULL;      } 
    if (pi != NULL)      { free (pi)      ; pi = NULL;      } 
    if (wlenpp != NULL)  { free (wlenpp)  ; wlenpp = NULL;  }

  
    if (status == STATUS__OK) {
        interp->result = "OK";
        return TCL_OK;
    } else {
        char buff[200];
        MessGetMsg(status, 0, sizeof(buff), buff);
        Tcl_AppendResult(interp, "\nError Generating CBS Pairs\n", (char *) NULL);
        Tcl_AppendResult(interp, buff, (char *) NULL);
        return TCL_ERROR;

    }
    
}


/*  ------------------------------------------------------------------------- */

static void AddOneSky(
    SdsIdType topid, double radius, int is_2df_diff_wlen, char targetType,
    double sine, double cosine, long  * objno, int  * sky, StatusType * status);

/*
 *                    S e t  S k y  G r i d  { }
 */
 
static int ConfSetSkyGrid(ClientData clientData DUNUSED, Tcl_Interp * interp,
    int argc DUNUSED, char *argv[])
    
/*+
 *  Command name:
 *    SetSkyGrid
 *
 *  Function:
 *    Sets up a grid of sky positions.
 *  
 *  Description:
 *    This routine creates a set of sky targets in the unallocated objects
 *    section of the configuration data, based on a three rings.
 *
 *  Call:
 *    SetSkyGrid id [ringNumOnly [program]]
 *
 *  Parameters:
 *
 *    (>) id  (Sds Id)  Id of top level of the configuration data structure.
 *    (>) ringNumOnly (int)  1, 2 or 3 - indications that only the specified
 *                          ring should be set up.  This is used for creating
 *                          test set ups.  If not supplied, all are set up.
 *    (>) program (string) IF supplied and equal to "program" then instead of
&
 *
 *   Support:
 *     Gavin Dalton, Oxford
 *-
 */
 {
    double th, thint, tp1, s, c;
    double xrad1, xrad2, xrad3;
    double fieldrad;
    SdsIdType topid, oid, rid;
    SdsCodeType code;
    long ndims, objno;
    int Npiv, Ipiv;
    unsigned long dims[7];
    StatusType status = STATUS__OK;
    char mbuf[100];
    char name[NAMELEN];
    short allocSkyCnt, unallocSkyCnt;
    SdsIdType field;
    int sky;
    int is_2df_diff_wlen = 0;
    char targetType, spectType = 0;
    int ring=0;  /* Which ring of targets to do.  IF 0, all thread. Otherwise - 1 to 3*/


    if ((argc < 2)||(argc > 4)) {
        interp->result = "Wrong # args - must have at least one, max three.";
        return TCL_ERROR;
    }
    FpilTargetSkyType(*Instrument, &targetType, &spectType);
    if (argc > 2) {
        ring=atol(argv[2]);
        if (argc >= 4)
        {
            if (strcmp(argv[3],"program") == 0)
            {
                targetType = 'P';
            }
            else
            {
                sprintf(interp->result,"3rd argument can only be \"program\", not %s",
                        argv[3]);
                return TCL_ERROR;
            }
        }
        
    }

    fieldrad = FpilGetFieldRadius(*Instrument);

    topid = (SdsIdType) (atol(argv[1]));
    SdsFind(topid, "objects", &oid, &status);
    SdsFind(oid, "ra", &rid, &status);
    SdsInfo(rid, name, &code, &ndims, dims, &status);
    Npiv = dims[0];
    SdsFreeId(rid, &status);
    SdsFreeId(oid, &status);
/*
 *  Determine the angle between each pivot. 
 */
    thint = 2 * PI / Npiv;
/*
 *  Get half that value.
 */
    th = thint / 2;
/*
 *  Determine the three radii we will work at.  We use the following
 *  proportions of the field size, which are chosen to get roughly
 *  the same dimensions as the original code which assumed a fixed
 *  field radius (for 2dF). 
 */
#   define SKYGRID_RADIUS_1 0.87
#   define SKYGRID_RADIUS_2 0.64
#   define SKYGRID_RADIUS_3 0.37

    xrad1 = fieldrad * SKYGRID_RADIUS_1;
    xrad2 = fieldrad * SKYGRID_RADIUS_2;
    xrad3 = fieldrad * SKYGRID_RADIUS_3;

    sprintf(mbuf, "SkyGrid: %d", Npiv);
    ConfMessage(interp, mbuf);


/*
 * Need a unique name for the sky.
 */
    SdsFind(topid, "fieldData", &field, &status);
    ArgGets(field, "unallocSky", &unallocSkyCnt, &status);
    if (status == STATUS__OK) {
        allocSkyCnt = 0;
        ArgGets(field, "allocSky", &allocSkyCnt, &status);
        if (status == SDS__NOITEM)
            ErsAnnul(&status);
    }
    SdsFreeId(field, &status);

/*
 *  Number for first sky's name.
 */
    sky = unallocSkyCnt + allocSkyCnt + 1;


    s = c = 0;

    is_2df_diff_wlen = (strncmp(FpilGetInstName(*Instrument), "2dF",3) == 0) &&
                            (gwave != gwave2); 


/*
 *  Create the grid.
 */
    for (Ipiv = 0; Ipiv < Npiv; Ipiv++) {

        tp1 = th + Ipiv * thint;
        s = sin(tp1);
        c = cos(tp1);
        if ((Ipiv % 5) == 0) {
            /*
             * We put a sky at every Npiv/5 in the outer ring.
             */
            if ((ring == 0)||(ring == 1))
                AddOneSky(topid, xrad1, is_2df_diff_wlen, targetType, s, c, 
                          &objno, &sky, &status);
        }
        if ((Ipiv % 8) == 0) {
            /*
             * We put a sky at every Npiv/8 in the middle ring.
             */
            if ((ring == 0)||(ring == 2))
                AddOneSky(topid, xrad2, is_2df_diff_wlen, targetType, s, c, 
                          &objno, &sky, &status);
        }
        if ((Ipiv % 13) == 0) {
            /*
             * We put a sky at every Npiv/13 in the small ring.
             */
            if ((ring == 0)||(ring == 3))
                AddOneSky(topid, xrad3, is_2df_diff_wlen, targetType, s, c, 
                          &objno, &sky, &status);
        }
    }
    if (status == STATUS__OK) {
        interp->result = "OK";
        return TCL_OK;
    } else {
        char buff[200];
        MessGetMsg(status, 0, sizeof(buff), buff);
        Tcl_AppendResult(interp, "\nError Setting Sky Grid\n", (char *) NULL);
        Tcl_AppendResult(interp, buff, (char *) NULL);
        return TCL_ERROR;

    }
}

static void AddOneSky(
    const SdsIdType topid,
    const double radius,
    const int is_2df_diff_wlen,
    const char targetType,
    const double sine,
    const double cosine,
    long  * const objno,
    int  * const sky,
    StatusType * const status)
{
    double xx;
    double yy;
    char label[NAMELEN];
    char *type = "Sky";

    if (*status != STATUS__OK) return;


    if (targetType == 'P')
        type = "Obj";
    else if (targetType == 'F')
        type = "Gui";

    xx = radius * cosine;
    yy = radius * sine;
    if (is_2df_diff_wlen) {
        /* We are using 2df with different wavelengths of the two
         * spectrograps. Create two sky positions rather than just one, one
       * for each spectrpgraph. The target types are hard wired, which is
       * truly awful, but it did quickly get the job done.
       */
        sprintf(label, "Grid_%s_%d_s1", type, *sky);
        ConfDoAddSkyPosition(topid, xx, yy, label, 'B', objno, status);
        sprintf(label, "Grid_%s_%d_s2", type, *sky);
        ConfDoAddSkyPosition(topid, xx, yy, label, 'C', objno, status);
    } else {
        sprintf(label, "Grid_%s_%d", type, *sky);
        ConfDoAddSkyPosition(topid, xx, yy, label, targetType, objno, status);
    }
    ++(*sky);
}



/*  ------------------------------------------------------------------------- */

/*
 *                    F i x  B r e a k a g e s  { }
 */
 
static int ConfFixBreakages(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc, char *argv[])
/*+
 *  Command name:
 *    FixBreakages
 *
 *  Function:
 *    Deallocates any broken fibres.
 *  
 *  Description:
 *    This routine deallocates any allocations that have been made to fibres
 *    flagged as broken.
 *
 *  Call:
 *    FixBreakages PivotId Cfid
 *
 *  Parameters:
 *
 *    (>) PivotId  (Sds Id)  Id of pivot data structure.
 *    (>) Cfid     (Sds Id)  Id of top level of configuration data structure.
 *
 *   Support:
 *     Gavin Dalton, Oxford
 *-
 */
{
    SdsIdType pivid, topid, objid = 0;
    unsigned fibre_no;
    StatusType status = STATUS__OK;
    short s;
    char c;
    int doneOne = 0;
    unsigned NumPivots = FpilGetNumPivots(*Instrument);


    if (argc != 3) {
        interp->result = "Wrong # args";
        return TCL_ERROR;
    }
/*
 *  Fetch arguments, pivot structure and configuration.
 */
    pivid = (SdsIdType) (atol(argv[1]));
    topid = (SdsIdType) (atol(argv[2]));
/* 
 *  Find the allocated objects.
 */
    SdsFind(topid, "objects", &objid, &status);
    ConfSdsCheck("Error finding objects structure", &status);
/*
 *  For each pivot, if we are allowed to change pivots associated with
 *  the spectrograph, get the inUse flag.
 */
    for (fibre_no = 0; fibre_no < NumPivots; fibre_no++) {
        int spec;
        spec = ConfWhichSpec((int) fibre_no, 0);
        if (UseFibreType[spec]) {
            ConfArrayGets(pivid, "inUse", fibre_no, &s, &status);
/*
 *          If the pivot is not in use, then we grab the type of the object
 *          associated with this fibre.
 */
            if (s != 1) {
                ConfArrayGetc(objid, "type", fibre_no, &c, &status);
                if (c != FpilUnallocatedType(*Instrument)) {
                    char command[80];
                    char message[80];
                    doneOne = 1;
/*
 *                  If an object is allocated, deallocate it using the 
 *                  DummyDeallocateFib Tcl Command.
 */
                    sprintf(message, "Deassigning Allocated Broken Fibre %d",
                        fibre_no + 1);
                    ConfMessage(interp, message);
                    sprintf(command, "DummyDeallocateFib %d", fibre_no + 1);
                    if (Tcl_Eval(interp, command) == TCL_ERROR) {
                        Tcl_AddErrorInfo(interp,
                            "\n(Error deallocating broken fibre)");
                        SdsFreeId(objid, &status);
                        return TCL_ERROR;
                    }
                }
            }
        }
    }
    SdsFreeId(objid, &status);
    if (!doneOne) {
        ConfMessage(interp, "No Allocated Broken Fibres were found");

    }
    return TCL_OK;
}

/*  ------------------------------------------------------------------------- */

/*
 *                       R e m o v e  S k y  { }
 */
 
static int ConfRemoveSky(ClientData clientData DUNUSED, Tcl_Interp * interp,
    int argc, char *argv[])
    
/*+
 *  Command name:
 *    RemoveSky
 *
 *  Function:
 *    Deallocates any fibres allocated to sky targets.
 *  
 *  Description:
 *    This routine deallocates any allocations that have been made to sky
 *    targets.
 *
 *  Call:
 *    RemoveSky PivotId Cfid
 *
 *  Parameters:
 *
 *    (>) PivotId  (Sds Id)  Id of pivot data structure.
 *    (>) Cfid     (Sds Id)  Id of top level of configuration data structure.
 *
 *   Support:
 *     Gavin Dalton, Oxford
 *-
 */
{
    SdsIdType pivid, topid, objid = 0;
    unsigned fibre_no;
    StatusType status = STATUS__OK;
    short s;
    char c;
    int doneOne = 0;
    unsigned NumPivots = FpilGetNumPivots(*Instrument);
    int spec;

    if (argc != 3) {
        interp->result = "Wrong # args";
        return TCL_ERROR;
    }
/*
 *  Fetch arguments, pivot structure and configuration.
 */
    pivid = (SdsIdType) (atol(argv[1]));
    topid = (SdsIdType) (atol(argv[2]));
/* 
 *  Find the allocated objects.
 */
    SdsFind(topid, "objects", &objid, &status);
    ConfSdsCheck("Error finding objects structure", &status);
/*
 *  For each pivot on a spectrograph we are allowed to change, 
 *  get the inUse flag.
 */
    for (fibre_no = 0; fibre_no < NumPivots; fibre_no++) {
        spec = ConfWhichSpec((int) fibre_no, 0);
        if (UseFibreType[spec]) {
            ConfArrayGets(pivid, "inUse", fibre_no, &s, &status);
            if (s == 1) {
/*
 *          Get any object allocation to this pivot. If the object
 *          is a Sky object, deallocate it by calling the
 *          Tcl command "DummyDeallocateFib".
 */
                ConfArrayGetc(objid, "type", fibre_no, &c, &status);
                if (FpilIsTargetSky(*Instrument, c)) {
                    char command[80];
                    char message[80];
                    doneOne = 1;
                    sprintf(message,
                        "Deassigning Allocated Sky Fibre %d", fibre_no + 1);
                    ConfMessage(interp, message);
                    sprintf(command, "DummyDeallocateFib %d", fibre_no + 1);

                    if (Tcl_Eval(interp, command) == TCL_ERROR) {
                        Tcl_AddErrorInfo(interp,
                            "\n(Error deallocating sky fibre)");
                        SdsFreeId(objid, &status);
                        return TCL_ERROR;
                    }
                }
            }
        }
    }
    SdsFreeId(objid, &status);

    if (!doneOne) {
        ConfMessage(interp, "No Allocated Sky Fibres were found");
    }

    return TCL_OK;
}

/*  ------------------------------------------------------------------------- */

#define ALLOC_TO_UNALLOC     0
#define ALLOC_ADD_TO_UNALLOC 1
/*
 *                    A l l o c T o U n a l l o c  { }
 *                    A l l o c A d d T o U n a l l o c  { }
 */
 
static int ConfAllocToUnalloc(ClientData clientdata,
    Tcl_Interp * interp, int argc, char *argv[])
    
/*+                      
 *  Command name:
 *    AllocToUnalloc
 *    AllocAddToUnalloc
 *
 *  Function:
 *    Put the allocated objects into the unallocated part of a structure.
 *
 *  Description:
 *      This command is used to prepare a currently configured field for
 *      reconfiguration.  
 *
 *      For the first form (AllocToUnalloc) the intention is that whilst we 
 *      may need to reconfigure, we should only be configuring the objects 
 *      that were configured on the other field, and we will
 *      delete all existing unallocated program and guide objects.
 *
 *      For the second from (AllocAddtoUnalloc) the intention is a quick
 *      deallocation of all allocated fibres, adding the deallocated objects
 *      to the unallocated and guide arrays.  THE STRUCTURE MUST HAVE 
 *      BEEN COMPRESSED FIRST for this to work. (CompressUnalloc command, or
 *      it should be a file read in).
 *
 *      We copy each allocated object to the appropiate part of the structure
 *       
 *      We delete the allocated array from the unallocGuide and unallocObject
 *      structures, if they exist.
 *
 *      In the "field" part of the structure, we 
 *
 *          set allocSky1 = 0
 *              allocSky2 = 0
 *               allocSky  = 0
 *               allocOby  = 0
 *               allocGui  = 0
 *
 *      For AllocToUnalloc
 *          set unallocSky to previous value of allocSky
 *          set unallocObj to previous value of allocObj
 *          set unallocGui to previous value of allocGui
 *      For AllocToUnalloc
 *          set unallocSky to previous value of allocSky + previous unallocSky
 *          set unallocObj to previous value of allocObj + previous unallocObj
 *          set unallocGui to previous value of allocGui + previous unallocGui
 *
 *
 *      The resulting structure is appropiate for reading into the program, but
 *      note that all the unallocGuide and unallocObject details in the 
 *      originial structure are lost.
 *
 *      The client data item, when cast to an int, will be either
 *         ALLOC_TO_UNALLOC or ALLOC_ADD_TO_UNALLOC to indicate the command.
 *
 *  Call:
 *     AllocToUnalloc id
 *     AllocAddToUnalloc id
 *
 *  Parameters:
 *     (>) id (int) SDS id of the structure to modify.
 *
 *  Language:
 *     Tcl
 *
 *  Support: Tony Farrell, AAO
 *-
 *  History:
 *       13-Apr-2000 - TJF - Original version (DtclTrnName)
 *  
 */

{
    SdsIdType topId;
    SdsIdType fieldId = 0;
    SdsIdType objectsId = 0;
    SdsIdType unallocObjId = 0;
    SdsIdType unallocGuiId = 0;
    SdsIdType tid;
    StatusType status = STATUS__OK;
    StatusType ignore = STATUS__OK;
    int result = TCL_OK;

    char name[SDS_C_NAMELEN];
    long ndims;
    unsigned long dims[7];
    SdsCodeType code;

    long allocObj = 0;
    long allocSky = 0;
    long allocGui = 0;

    short unallocObj = 0;
    short unallocSky = 0;
    short unallocGui = 0;

    long unallocObjArraySize = 0;
    long unallocGuiArraySize = 0;

    char *typePnt;
    unsigned long length;

    register unsigned i;
    long nextObj;
    long nextGui;

    int addCommand = 0;


    if (argc != 2) {
        sprintf(interp->result, "%s:Wrong # args", argv[0]);
        return TCL_ERROR;
    }
/*
 *  Work out the command type.
 */
    if ((long)clientdata == ALLOC_ADD_TO_UNALLOC)
        addCommand = 1;

    topId = atoi(argv[1]);

/*
 *  Find the various components we want.  Note, ArgFind is like SdsFind except
 *  that it uses ErsRep if it gets an error to report the name of the
 *  object concerned.
 */
    ArgFind(topId, "fieldData", &fieldId, &status);
    ArgFind(topId, "objects", &objectsId, &status);
    ArgFind(topId, "unallocObject", &unallocObjId, &status);
    ArgFind(topId, "unallocGuide", &unallocGuiId, &status);

    if (status != STATUS__OK)
        goto Exit;

/*
 *  Work out how many of each type of allocated object we have.
 */
    ArgGeti(fieldId, "allocObj", &allocObj, &status);
    if (status == SDS__NOITEM)
        ErsAnnul(&status);
    ArgGeti(fieldId, "allocGui", &allocGui, &status);
    if (status == SDS__NOITEM)
        ErsAnnul(&status);
    ArgGeti(fieldId, "allocSky", &allocSky, &status);
    if (status == SDS__NOITEM)
        ErsAnnul(&status);

    if (status != STATUS__OK)
        goto Exit;

/*
 *  If we don't have any allocated objects, this makes no sense.
 */
    if ((allocObj + allocGui + allocSky) == 0) {
        sprintf(interp->result,
            "%s:This makes no sense, there are no allocated objects", argv[0]);
        result = TCL_ERROR;
        goto Exit;
    }
/*
 *  If we are adding the allocated items to the existing unallocated
 *  structure, then we need the unallocated values.  The variables
 *  will have zero if we are deleing the existing unallocated objects.
 */
    if (addCommand)
    {
        /*
         * We must not be using an uncompressed structure.  Such structures
         * are indicated by having an "allocated" item in the unallocObjects
         * component.
         */
        SdsFind(unallocObjId, "allocated", &tid, &status);
        if (status == SDS__NOITEM)
            status = STATUS__OK;
        else if (status == STATUS__OK)
        {
            SdsFreeId(tid, &ignore);
            sprintf(interp->result, "%s:You have tried to run this command un an uncompressed structure, this does not work", argv[0]);
            result = TCL_ERROR;
            goto Exit;
        }

        ArgGets(fieldId, "unallocObj", &unallocObj, &status);
        ArgGets(fieldId, "unallocGui", &unallocGui, &status);
        ArgGets(fieldId, "unallocSky", &unallocSky, &status);
    }
/*
 *  All objects we are keeping are to be unallocated.
 */ 
    ArgPuts(fieldId, "unallocObj", allocObj + unallocObj, &status);
    ArgPuts(fieldId, "unallocGui", allocGui + unallocGui, &status);
    ArgPuts(fieldId, "unallocSky", allocSky + unallocSky, &status);


/*
 *  Set the field values appropriately for the transfer of the allocated
 *  details to the unallocated details.
 */
    ArgPuts(fieldId, "allocObj", 0, &status);
    ArgPuts(fieldId, "allocGui", 0, &status);
    ArgPuts(fieldId, "allocSky", 0, &status);
    for (i = 0; i < InstNumFibreTypes; ++i) {
        char buffer[20];
        sprintf(buffer, "allocSky%d", i + 1);
        ArgPuts(fieldId, buffer, 0, &status);
    }

/*
 *  Resize our structures.
 */
    unallocObjArraySize = unallocObj + unallocSky + allocObj + allocSky;
    unallocGuiArraySize = unallocGui + allocGui;
    ConfResizeAllItems(unallocObjId, unallocObjArraySize, &status);
    ConfResizeAllItems(unallocGuiId, unallocGuiArraySize, &status);

    SdsFind(objectsId, "x", &tid, &status);
    SdsInfo(tid, name, &code, &ndims, dims, &status);
    SdsFreeId(tid, &status);


    tid = 0;
    SdsFind(objectsId, "type", &tid, &status);
    SdsPointer(tid, (void **) &typePnt, &length, &status);

/*
 *  Indices to where the next allocated object of each time will go
 *  in the unallocated structures.
 */
    nextGui = unallocGui;
    nextObj = unallocObj + unallocSky;

    for (i = 0; (status == STATUS__OK) && (i < dims[0]); ++i) {
        SdsIdType uid;
        unsigned int obj_no;
        if (typePnt[i] == FpilUnallocatedType(*Instrument)) {
/*  
 *          Must be an unallocated pivot, goto the next object.
 */
            continue;
        } else if (FpilIsTargetGuide(*Instrument, typePnt[i])) {
/*
 *          Guide objects go in the unallocGuide array.
 */
            uid = unallocGuiId;
            obj_no = nextGui++;

            if (nextGui > unallocGuiArraySize) {
                fprintf(stderr, "Programming error, nextGui %ld, value  %ld\n",
                        nextGui, allocGui + unallocGui);
                exit(-1);
            }

            ConfArrayCopy(objectsId, uid, "dhdx", sizeof(double), i, obj_no,
                &status);
            ConfArrayCopy(objectsId, uid, "dhdy", sizeof(double), i, obj_no,
                &status);
            ConfArrayCopy(objectsId, uid, "dddx", sizeof(double), i, obj_no,
                &status);
            ConfArrayCopy(objectsId, uid, "dddy", sizeof(double), i, obj_no,
                &status);
        } else {
/*
 *          Program objects or sky objects go in the unallocObject array.
 */
            uid = unallocObjId;
            obj_no = nextObj++;

            if (obj_no >= unallocObjArraySize) {
                fprintf(stderr,
                        "Programming error, nextObj %ld, values  %ld + %ld\n",
                        nextObj, allocObj+unallocObj, allocSky+unallocSky);
                exit(-1);
            }

            ConfArrayCopy(objectsId, uid, "spectrograph", sizeof(char),
                i, obj_no, &status);
            ConfArrayCopy(objectsId, uid, "wlenIndex", sizeof(short),
                i, obj_no, &status);

        }

/*
 *      Details common to both arrays.
 */
        ConfArrayCopy(objectsId, uid, "name", NAMELEN, NAMELEN * i,
            NAMELEN * obj_no, &status);
        ConfArrayCopy(objectsId, uid, "ra", sizeof(double), i, obj_no, &status);
        ConfArrayCopy(objectsId, uid, "dec", sizeof(double), i, obj_no,
            &status);
        ConfArrayCopy(objectsId, uid, "x", sizeof(int), i, obj_no, &status);
        ConfArrayCopy(objectsId, uid, "y", sizeof(int), i, obj_no, &status);
        ConfArrayCopy(objectsId, uid, "type", sizeof(char), i, obj_no, &status);
        ConfArrayCopy(objectsId, uid, "priority", sizeof(short), i,
            obj_no, &status);
        ConfArrayCopy(objectsId, uid, "magnitude", sizeof(double), i,
            obj_no, &status);
        ConfArrayCopy(objectsId, uid, "pId", sizeof(short), i, obj_no, &status);
        ConfArrayCopy(objectsId, uid, "comment", CMTLEN, CMTLEN * i,
            CMTLEN * obj_no, &status);


 


    }

    if (tid != 0)
        SdsFreeId(tid, &ignore);


/*
 *  Remove the objects structure
 */
    SdsDelete(objectsId, &status);

/*
 *  Remove the allocated structure of the unallocObj and unallocGui 
 *  structures.
 */
    if ((status == STATUS__OK)&&(!addCommand))
    {
        SdsFind(unallocObjId, "allocated", &tid, &status);
        SdsDelete(tid, &status);
        SdsFreeId(tid, &status);

        SdsFind(unallocGuiId, "allocated", &tid, &status);
        SdsDelete(tid, &status);
        SdsFreeId(tid, &status);
    }

/*
 *  Tidy up.
 */
  Exit:
    if (fieldId)
        SdsFreeId(fieldId, &ignore);
    if (objectsId)
        SdsFreeId(objectsId, &ignore);
    if (unallocObjId)
        SdsFreeId(unallocObjId, &ignore);
    if (unallocGuiId)
        SdsFreeId(unallocGuiId, &ignore);

    if (status != STATUS__OK) {
        char buff[200];
        StatusType status2 = status;
        sprintf(buff, "%s:Bad status failure\n", argv[0]);

        ErsFlush(&status2);  /* Must do this before setting result */

        Tcl_AppendResult(interp, buff, (char *) NULL);
        MessGetMsg(status, 0, sizeof(buff), buff);
        Tcl_AppendResult(interp, buff, (char *) NULL);


        return TCL_ERROR;

    }
    return result;

}
/*  ------------------------------------------------------------------------- */

/*
 *                    C h e c k S d S  { }
 */
 
static int ConfCheckSds(ClientData clientdata DUNUSED,
    Tcl_Interp * interp, int argc, char *argv[])
    
/*+                      
 *  Command name:
 *    CheckSds
 *
 *  Function:
 *    Check an SDS file being opened.
 *
 *  Description:
 *    This checks an existing allocation file being read back in to see
 *    if it can be used.  If the file has an "instData" component, that
 *    is check to see if it is for the same instrument and plate.
 *
 *    The number of object (allocated) fibres is checked for validity.
 *
 *    If the instrument, plate and number of fibres are all ok, TCL_OK
 *    is return and an empty string is returned as the return value.  If 
 *    we examined the file ok, but something is wrong, then a TCL_OK is
 *    returned but a message is written as the return value.  In this case,
 *    the user might be able to proceded if the fibres are deallocated.  If
 *    the first character of the message is "+", then it is acceptiable to
 *    try to continue anyway. 
 *
 *    An error will be return if we can't read the structure correctly.
 *
 *  Call:
 *     CheckSds id
 *
 *  Parameters:
 *     (>) id (int) SDS id of the structure to check.
 *
 *  Language:
 *     Tcl
 *
 *  Support: Tony Farrell, AAO
 *-
 *  History:
 *       05-Jan-2005 - TJF - Original version 
 *  
 */

{
    SdsIdType topId;
    SdsIdType instId = 0;
    SdsIdType objectsId = 0;
    SdsIdType raId = 0;
    StatusType status = STATUS__OK;
    StatusType ignore = STATUS__OK;
    int result = TCL_OK;

    char name[SDS_C_NAMELEN];
    long ndims;
    unsigned long dims[7];
    SdsCodeType code;

    if (argc != 2) {
        sprintf(interp->result, "%s:Wrong # args", argv[0]);
        return TCL_ERROR;
    }

    topId = atoi(argv[1]);

/*
 *  Find the various components we want.  Note, ArgFind is like SdsFind except
 *  that it uses ErsRep if it gets an error to report the name of the
 *  object concerned.
 */
    ErsPush();
    ArgFind(topId, "instData", &instId, &status);
    if (status == STATUS__OK)
    {
        char instName[100]="";
        long plate;
        long curPlate = CurrentField;
        if (strcmp(FpilGetInstName(*Instrument),"6dF") == 0) /* Adjust 6dF current field plate */
            ++curPlate;
        
        /*
         * Check the instrument data.
         */
        ArgGetString(instId,"instrument", sizeof(instName), instName, &status);
        ArgGeti(instId, "plate", &plate, &status);
        if (status != STATUS__OK) goto Exit;
        if (strcmp(FpilGetInstName(*Instrument),instName) != 0)
        {
            ErsSPrintf(TCL_RESULT_SIZE, interp->result,
                       "This Configuration file is for \"%s\" rather then \"%s\",you must deallocate to proceed", 
                       instName, FpilGetInstName(*Instrument));
            goto Exit;
        }
        else if (plate != curPlate)
        {
            ErsSPrintf(TCL_RESULT_SIZE, interp->result,
                           "-%d:This Configuration file is for plate %d, rather then plate %d. You probably want to change plate or deallocate before proceeding (or you can do so later).", 
                           plate, plate, curPlate);
        }
       
        
    }
    
    else if (status == SDS__NOITEM)
    {
        /* Old files don't have instData */
        ErsAnnul(&status);
    }
    else
        goto Exit;

    ArgFind(topId, "objects", &objectsId, &status);
    if (status == SDS__NOITEM) {
    
       /*  Some files, for example, those created using AllocToUnalloc,
        *  don't have an 'objects' section. Configure can handle this,
        *  but we can't check it for consistency with the instrument
        *  data. (I'm not sure if there is a sensible check we can try
        *  in these circumstances. KS.)
        */
        
       ErsAnnul(&status);
       goto Exit;
    }
    ArgFind(objectsId, "ra", &raId, &status);
    SdsInfo(raId, name, &code, &ndims, dims, &status);
    if (status != STATUS__OK) goto Exit;

    if (ndims != 1)
    {
        status = SDS__INVDIMS;
        ErsRep(0, &status,
               "Objects structure exists but error getting  ra item details ");
        goto Exit;
    }
    if (dims[0] != FpilGetNumPivots(*Instrument))
    {
        /*
         * This problem should only occur if we don't have an Instrument
         * section - since the change in the number of pivots in 2dF
         * occured at the same time as we started writing the instrument
         * section.
         */
        if ((dims[0] == 404)&&
            (strcmp(FpilGetInstName(*Instrument),"2dF") == 0))
        {
            /* 2dF special case - we want a better message then the default
             * for the reduction in the number of fibres
             */
            Tcl_AppendResult(interp,
                       "This Configuration file was created for the 404 fibre 2dF.  2dF now has 400 fibres after a plate redesign in prepration for AAOMega. You can only proceed by removing the allocation. You can view 404 fibre allocations using the \"2dF-old-404\" instrument when starting configure", 0);
        }
        else
        {
            ErsSPrintf(TCL_RESULT_SIZE, interp->result,
                       "This Configuration file is not appropiate for %s,\nit contains an existing configuration with %d fibres. %s has %d fibres.  You may be able to proceed by removing the allocation. ",
                       FpilGetInstName(*Instrument),
                       dims[0],
                       FpilGetInstName(*Instrument),
                       FpilGetNumPivots(*Instrument));
        }
    }

   
    

/*
 *  Tidy up.
 */
  Exit:

    if (objectsId)
        SdsFreeId(objectsId, &ignore);
    if (instId)
        SdsFreeId(instId, &ignore);
    if (raId)
        SdsFreeId(raId, &ignore);
    

    if (status != STATUS__OK) {
        char buff[200];
        StatusType status2 = status;
        sprintf(buff, "%s:Bad status failure checking SDS structure\n", 
                argv[0]);

        ErsFlush(&status2); /* Must do this before setting result */

        Tcl_AppendResult(interp, buff, (char *) NULL);

        MessGetMsg(status, 0, sizeof(buff), buff);
        Tcl_AppendResult(interp, buff, (char *) NULL);


        return TCL_ERROR;

    }
    return result;

}

/*  ------------------------------------------------------------------------- */

/*
 *                    A d d I n s t r u m e n t D a t a  { }
 */
 
static int ConfAddInstrumentData(ClientData clientdata DUNUSED,
    Tcl_Interp * interp, int argc, char *argv[])
    
/*+                      
 *  Command name:
 *    AddInstrumentData
 *
 *  Function:
 *    Add details of the instrument to the configuration structure.
 *
 *  Description:
 *    Adds an instData component to the configuration structure.  If it
 *    already exists, the values are set correctly.
 *
 *  Call:
 *     AddInstrumentData id
 *
 *  Parameters:
 *     (>) id (int) SDS id of the structure to check.
 *
 *  Language:
 *     Tcl
 *
 *  Support: Tony Farrell, AAO
 *-
 *  History:
 *       06-Jan-2005 - TJF - Original version 
 *  
 */

{
    SdsIdType topId;
    SdsIdType instId = 0;
    StatusType status = STATUS__OK;
    StatusType ignore = STATUS__OK;

    if (argc != 2) {
        sprintf(interp->result, "%s:Wrong # args", argv[0]);
        return TCL_ERROR;
    }

    topId = atoi(argv[1]);

    /*
     * Look for the instData componet of the structure.
     */
    SdsFind(topId, "instData", &instId, &status);
    if (status == SDS__NOITEM)
    {
        /*
         * Structure does not exist - create it.
         *
         */
        status = STATUS__OK;
        SdsNew(topId, "instData", 0, 0, SDS_STRUCT, 0, 0, &instId, &status);
    }
    ArgFind(topId, "instData", &instId, &status);
    if (status == STATUS__OK)
    {
        /*
         * Add/Update the instrument data.  (For 6dF we want the displayed plate
         * rather the the CurrentField value.
         */
        ArgPutString(instId,"instrument", 
                     FpilGetInstName(*Instrument), 
                     &status);
        if (strcmp(FpilGetInstName(*Instrument),"6dF") == 0) /* Adjust 6dF current field plate */
            ArgPuti(instId, "plate", CurrentField+1, &status);
        else
            ArgPuti(instId, "plate", CurrentField, &status);
    }

/*
 *  Tidy up.
 */
    if (instId)
        SdsFreeId(instId, &ignore);

    if (status != STATUS__OK) {
        char buff[200];
        StatusType status2 = status;
        sprintf(buff, "%s:Bad status failure adding instrument data to SDS structure\n", 
                argv[0]);

        ErsFlush(&status2); /* Must do this before setting result */

        Tcl_AppendResult(interp, buff, (char *) NULL);

        MessGetMsg(status, 0, sizeof(buff), buff);
        Tcl_AppendResult(interp, buff, (char *) NULL);


        return TCL_ERROR;

    }
    return TCL_OK;

}

/*  ------------------------------------------------------------------------- */

/*
 *                 T r a n s l a t e N a m e  { } 
 */
 
static int TranslateName(ClientData clientdata DUNUSED,
    Tcl_Interp * interp, int argc, char *argv[])
 
/*+                      
 *  Command name:
 *    TranslateName
 *
 *  Function:
 *    Translate a logical name/environment variable.
 *
 *  Description:
 *    Translate the value of the specified environment variable.
 *
 *  Call:
 *    TranslateName name [options]
 *
 *  Parameters:
 *    (>) name (string) Name of the environment variable to translate.
 *
 *  Options:
 *     [-file (string)]   We intend to access the specified file with
 *                        name being the path to that file.  Returns an
 *                        appropriate file name for use in open calls
 *     [-default (string)] If there is no such name, then use the specified
 *                        string (file will not  be attached) as the result. If 
 *                        default  is not specified and the name does 
 *                        not exist, an error is returned.
 *
 *  Language:
 *     Tcl
 *
 *  Support: Tony Farrell, AAO
 *-
 *  History:
 *     10-May-1994 - TJF - Original version (DtclTrnName)
 *     24-Mar-1995 - TJF - Use DtclCmdError to report errors.
 *     21-Jul-1998 - TJF - Call DitsGetSymbol instead of getenv() to translate
 *                         symbols.
 *     04-Nov-1998 - TJF - Configure specific version.  Go back to getenv
 *                         to remove IMP dependency.  Also remove calls
 *                         to DctlCmdError and VMS and WIN32 code.
 *     15-Mar-2000 - TJF - Incorporate code from KS's OzPoz configure demo
 *                         to support macintosh - separator is a colon
 *                         instead of a slash.
 */

{
    char *defval = "";
    char *file = 0;
    int isfile = 0;
    int hasdefault = 0;
    register int i;
    char Separator[2];

#ifdef macintosh
    strcpy(Separator, ":");
#else
    strcpy(Separator, "/");
#endif

    if (argc < 2) {
        sprintf(interp->result, "%s:wrong # args", argv[0]);
        return TCL_ERROR;
    }

    for (i = 2; i < argc; ++i) {
        if (strcmp(argv[i], "-file") == 0) {
            ++i;
            file = argv[i];
            isfile = 1;
        } else if (strcmp(argv[i], "-default") == 0) {
            ++i;
            defval = argv[i];
            hasdefault = 1;
        } else {
            sprintf(interp->result,
                "%s:Unknown option - \"%s\"\n", argv[0], argv[i]);
            return TCL_ERROR;
        }
    }
/*
 * Get the symbol value.
 */
    if (getenv(argv[1]) == 0) {
/*
 *     No symbol value.  Use default if there is one.
 */
        if (hasdefault) {
            Tcl_SetResult(interp, defval, TCL_VOLATILE);
            return TCL_OK;
        } else {
            sprintf(interp->result, "%s:No translation for \"%s\"",
                argv[0], argv[1]);
            return TCL_ERROR;
        }
    }
    strcpy(interp->result, getenv(argv[1]));
    if (isfile) {
/*
 *     Append file.  Separator is system-dependent.   
 */
        Tcl_AppendResult(interp, Separator, file, NULL);
    }
    return TCL_OK;
}


/*  ------------------------------------------------------------------------- */

/*
 *                        E x e  D i r  { }
 */
 
static int ExeDir(ClientData clientdata DUNUSED, Tcl_Interp * interp,
    int argc DUNUSED, char *argv[]DUNUSED)
    
/*+                
 *  Command name:
 *     ExeDir
 *
 *  Function:
 *     Return the executable directory of this program.
 *
 *  Description:
 *     Returns the executable directory saved when main was invoked.
 *
 *  Call:
 *     ExeDir 
 *
 *  Language:
 *     Tcl
 *
 *  Support: Tony Farrell, AAO
 *-
 *
 *  History:
 *       04-Nov-1998 - TJF - Original version
 *  
 */

{
    Tcl_SetResult(interp, ExecutableDir, TCL_STATIC);
    return TCL_OK;
}

/*  ------------------------------------------------------------------------- */

/*
 *                        E r s F l u s h  { }
 */
 
static int ConfigErsFlush(ClientData clientdata DUNUSED,
    Tcl_Interp * interp DUNUSED, int argc DUNUSED, char *argv[]DUNUSED)

/*+                     
 *  Command name:
 *     ErsFlush
 *
 *  Function:
 *    Invoke ErsFlush routine.
 *
 *  Description:
 *    Invokes the Ers routine ErsFlush. This flushes out any queued error
 *    messages, ensuring that the user sees them.
 *
 *  Call:
 *     ErsFlush
 *
 *  Language:
 *     Tcl
 *
 *  Support: Tony Farrell, AAO
 *-
 *  History:
 *       18-Dec-1998 - TJF - Original version
 */

{
    StatusType status = STATUS__OK;
    ErsFlush(&status);
    return TCL_OK;
}

/*  ------------------------------------------------------------------------- */

/*
 *                          M i m i c  { }
 */

static int ConfMimic(ClientData clientdata DUNUSED,
    Tcl_Interp * interp, int argc, char *argv[])

/*+                     
 *  Command name:
 *     Mimic
 *
 *  Function:
 *      Return information about the mimic.
 *
 *  Description:
 *      Various values are used in both the C code and the mimic Tcl
 *      code to determine the size and scalling of the mimic.  This
 *      command returns these values.
 *
 *      Note that FieldRadius represents the area of the configable field,
 *      but the park positions of the fibres will be outside this, but
 *      within CircleRadius.
 *
 *  Call:
 *      Mimic value
 *
 *  Parameters:
 *      (>) value (string) The value to fetch, WindowSize, CircleRadius
 *                           CircleOffset or FieldRadius.
 *
 *  Language:
 *      Tcl
 *
 *  Support: Tony Farrell, AAO
 *-
 *  History:
 *       18-Dec-1998 - TJF - Original version
 *  
 */

{
    if (argc != 2) {
        sprintf(interp->result, "%s:wrong # args", argv[0]);
        return TCL_ERROR;
    }
    if (strcmp(argv[1], "WindowSize") == 0) {
        sprintf(interp->result, "%d", MimicWindowSize);
    } else if (strcmp(argv[1], "CircleRadius") == 0) {
        sprintf(interp->result, "%d", MimicCircleRadius);
    } else if (strcmp(argv[1], "CircleOffset") == 0) {
        sprintf(interp->result, "%d", MimicCircleOffset);
    } else if (strcmp(argv[1], "FieldRadius") == 0) {
        /* Grab the field radius and convert it to pixels */
        double fieldRadius = FpilGetFieldRadius(*Instrument);
        double scale = (double) MimicCircleRadius /
            (double) FpilGetFieldRadiusIncPark(*Instrument);
        sprintf(interp->result, "%d", (int) (fieldRadius * scale));
    } else {
        sprintf(interp->result,
            "%s:Invalid argument, must be one of \"WindowSize\"", argv[0]);
        strcat(interp->result, " \"CircleRadius\" or \"CircleOffset\"\n");
        return TCL_ERROR;
    }
    return TCL_OK;
}


/*  ------------------------------------------------------------------------- */

/*                      P l a t e  X  t o  D i s p  X  { }
 *
 *                      P l a t e  Y  t o  D i s p  Y  { }
 */
 
/*+
 *  Command name:
 *     PlateXtoDispX,PlateXtoDispX
 *
 *  Function:
 *     Converts between an plate position and a display position.
 *
 *  Description:
 *     These routines take a plate position in X or Y (in microns) and return
 *     the corresponding X or Y coordinate on the display (in pixels).
 *
 *  Call:
 *     PlateXtoDispX PlatePosn zoom
 *     PlateYtoDispY PlatePosn zoom
 *
 *  Language:
 *     Tcl
 *      
 *  Parameter:
 *     PlatePosn  (int) Plate X or Y coordinate in microns
 *     Zoom       (double) Current zoom factor.
 *
 *  Returns:
 *     Display pixel position.
 *
 *  Support: Keith Shortridge, AAO
 *-
 *  History:
 *       23-May-2001 - KS - Original version
 */
 
static int PlateXtoDispX(ClientData clientdata DUNUSED, Tcl_Interp * interp,
    int argc, char *argv[])
    
{
    char String[32];
    if (argc != 3) {
        sprintf(interp->result, "%s:wrong # args", argv[0]);
        return TCL_ERROR;
    }
    sprintf (String,"%d",FieldPlateToDispX(0,atoi(argv[1]),atof(argv[2])));
    Tcl_SetResult(interp, String, TCL_VOLATILE);
    return TCL_OK;
}

static int PlateYtoDispY(ClientData clientdata DUNUSED, Tcl_Interp * interp,
    int argc, char *argv[])
{
    char String[32];
    if (argc != 3) {
        sprintf(interp->result, "%s:wrong # args", argv[0]);
        return TCL_ERROR;
    }
    sprintf (String,"%d",FieldPlateToDispY(0,atoi(argv[1]),atof(argv[2])));
    Tcl_SetResult(interp, String, TCL_VOLATILE);
    return TCL_OK;
}


/*  ------------------------------------------------------------------------- */

/*+                     D i s p  X  t o  P l a t e  X  { }
 *
 *                      D i s p  Y  t o  P l a t e  Y  { }
 *
 *  Command name:
 *      DispXtoPlateX,DispYtoPlateY
 *
 *  Function:
 *      Converts between a display position and a plate position.
 *
 *  Description:
 *      These routines take a coordinate in X or Y (in pixels) and return
 *      the corresponding plate X or Y coordinate (in microns). Note that
 *      the pixel value is the pixel position in the scroll region of the
 *      canvas being displayed, not the pixel position in the window (which
 *      only displays a fraction of the canvas scroll region when zoomed).
 *
 *  Call:
 *      DispXtoPlateX DispPosn zoom
 *      DispYtoPlateY DispPosn zoom
 *
 *  Language:
 *      Tcl
 *      
 *  Parameter:
 *      DispPosn   (int) Display X or Y coordinate in pixels
 *      Zoom       (double) Current zoom factor.
 *
 *  Returns:
 *      Plate position in microns.
 *
 *  Support: Keith Shortridge, AAO
 *-
 *  History:
 *       07-Jun-2001 - KS - Original version
 */

static int DispXtoPlateX(ClientData clientdata DUNUSED, Tcl_Interp * interp,
    int argc, char *argv[])
{
    char String[32];
    if (argc != 3) {
        sprintf(interp->result, "%s:wrong # args", argv[0]);
        return TCL_ERROR;
    }
    
    /*  It is because the position is the position in the scroll region
     *  and not in the window that the second argument (the offset) here is
     *  zero. If it were the window position, we would need the offset.
     */
     
    sprintf (String,"%d",DispToFieldPlateX(atoi(argv[1]),0.0,atof(argv[2])));
    Tcl_SetResult(interp, String, TCL_VOLATILE);
    return TCL_OK;
}

static int DispYtoPlateY(ClientData clientdata DUNUSED, Tcl_Interp * interp,
    int argc, char *argv[])
{
    char String[32];
    if (argc != 3) {
        sprintf(interp->result, "%s:wrong # args", argv[0]);
        return TCL_ERROR;
    }
    sprintf (String,"%d",DispToFieldPlateY(atoi(argv[1]),0.0,atof(argv[2])));
    Tcl_SetResult(interp, String, TCL_VOLATILE);
    return TCL_OK;
}


/*  ------------------------------------------------------------------------- */

/*+                     P l a t e  X Y  t o  R a  D e c  { }
 *
 *  Command name:
 *      PlateXYtoRaDec
 *
 *  Function:
 *      Returns the mean Ra,Dec coordinates for a given plate position.
 *
 *  Description:
 *      This routine takes a plate position (the X and Y coordinates in
 *      microns) and returns the mean Ra,Dec coordinates that correspond to
 *      that plate position.
 *
 *  Call:
 *      PlateXYtoRaDec Cfid Xp Yp Ra Dec
 *
 *  Language:
 *      Tcl
 *      
 *  Parameters:
 *      Cfid       (Sds id) Id of the overall configuration structure
 *      Xp         (int) Plate X position in microns
 *      Yp         (int) Plate Y position in microns
 *      Ra         (double Tcl variable) Variable to receive Ra in radians
 *      Dec        (double Tcl variable) Variable to receive Dec in radians
 *
 *  Support: Keith Shortridge, AAO
 *-
 *  History:
 *       07-Jun-2001 - KS - Original version
 *       17-Oct-2002 - KS - Added use of FpilGetPointingWavelength()
 */

static int PlateXYtoRaDec(ClientData clientdata DUNUSED, Tcl_Interp * interp,
    int argc, char *argv[])
{
    SdsIdType id, idf;
    double cra, cdec, mjd;
    double xp,yp;
    double xf,yf;
    double ara,adec;
    double raMean,decMean;
    double gmap[21];                   /* Mean to apparent parameters  */
    char buff[200];
    double Wavelength;
    StatusType status = STATUS__OK;
    
    if (argc != 6) {
        interp->result = "wrong # args";
        return TCL_ERROR;
    }
    
    id = atol(argv[1]);
    
    /*
     *  Get field centre details
     */
     
    SdsFind(id, "fieldData", &idf, &status);
    ArgGetd(idf, "appRa", &cra, &status);
    ArgGetd(idf, "appDec", &cdec, &status);
    ArgGetd(idf, "configMjd", &mjd, &status);
    SdsFreeId(idf, &status);

    /*
     *  Set up x,y conversion parameters. We use the main pointing
     *  wavelength, since there isn't a specific target in question here,
     *  at least, not so far as we know.
     */

    slaMappa(2000.0, mjd, gmap);
    
    Wavelength = FpilGetPointingWavelength(*Instrument);
    FpilModelCvtInit(TelModel, mjd, 0.0, gtemp, gpress, ghumid, 
                       Wavelength, Wavelength, TelModelParams, &status);
        
    xp = atof(argv[2]);
    yp = atof(argv[3]);
    FpilModelPos2Xy(TelModel, xp, yp, &xf, &yf, &status);
    FpilModelXy2Rd(TelModel, cra, cdec, xf, yf, mjd, &ara, &adec,&status);
    slaAmpqk (ara,adec,gmap,&raMean,&decMean);
    sprintf(buff, "%.16g", raMean);
    Tcl_SetVar(interp, argv[4], buff, 0);
    sprintf(buff, "%.16g", decMean);
    Tcl_SetVar(interp, argv[5], buff, 0);
    
    if (status != STATUS__OK) {
        MessGetMsg(status, 0, sizeof(buff), buff);
        Tcl_AppendResult(interp, "\nError converting to RA,Dec\n",
            (char *) NULL);
        Tcl_AppendResult(interp, buff, (char *) NULL);
        return TCL_ERROR;
    }
    return TCL_OK;
}


/*  ------------------------------------------------------------------------- */

/*+                 R a  D e c  T o  P l a t e  X Y { }
 *
 *  Command name:
 *      RaDecToPlateXY
 *
 *  Function:
 *      Returns the plate position corresponding to mean Ra,Dec coordinates.
 *
 *  Description:
 *      This routine takes a pair of mean Ra,Dec coordinates and returns
 *      the corresponding plate position (the X and Y coordinates in
 *      microns).
 *
 *  Call:
 *      RaDecToPlateXY Cfid Ra Dec Xp Yp 
 *
 *  Language:
 *      Tcl
 *      
 *  Parameters:
 *      Cfid       (Sds id) Id of the overall configuration structure
 *      Ra         (double) Ra in radians
 *      Dec        (double) Dec in radians
 *      Xp         (int Tcl variable) Variable to receive Plate X position
 *                 in microns
 *      Yp         (int Tcl variable) Variable to receive Plate Y position
 *                 in microns
 *
 *  Support: Keith Shortridge, AAO
 *-
 *  History:
 *       07-Aug-2001 - KS - Original version
 *       17-Oct-2002 - KS - Added use of FpilGetPointingWavelength()
 */

static int RaDecToPlateXY(ClientData clientdata DUNUSED, Tcl_Interp * interp,
    int argc, char *argv[])
{
    SdsIdType id, idf;
    double cra, cdec, mjd;
    double xp,yp,xf,yf;
    double ara,adec;
    double raMean,decMean;
    double gmap[21];                   /* Mean to apparent parameters  */
    char buff[200];
    double Wavelength;
    StatusType status = STATUS__OK;
    
    if (argc != 6) {
        interp->result = "wrong # args";
        return TCL_ERROR;
    }
    
    id = atol(argv[1]);
    
    /*
     *  Get field centre details
     */
     
    SdsFind(id, "fieldData", &idf, &status);
    ArgGetd(idf, "appRa", &cra, &status);
    ArgGetd(idf, "appDec", &cdec, &status);
    ArgGetd(idf, "configMjd", &mjd, &status);
    SdsFreeId(idf, &status);

    /*
     *  Set up x,y conversion parameters. We use the main pointing
     *  wavelength, since there isn't a specific target in question here,
     *  at least, not so far as we know.

     */

    slaMappa(2000.0, mjd, gmap);
    
    Wavelength = FpilGetPointingWavelength(*Instrument);
    FpilModelCvtInit(TelModel, mjd, 0.0, gtemp, gpress, ghumid, 
                       Wavelength, Wavelength, TelModelParams, &status);
        
    raMean = atof(argv[2]);
    decMean = atof(argv[3]);
    
    slaMapqkz(raMean, decMean, gmap, &ara, &adec);
    FpilModelRd2Xy(TelModel, cra, cdec, ara, adec, mjd, &xf, &yf, &status);
    FpilModelXy2Pos(TelModel, xf, yf, &xp, &yp, &status);
            
    sprintf(buff, "%g", xp);
    Tcl_SetVar(interp, argv[4], buff, 0);
    sprintf(buff, "%g", yp);
    Tcl_SetVar(interp, argv[5], buff, 0);
    
    if (status != STATUS__OK) {
        MessGetMsg(status, 0, sizeof(buff), buff);
        Tcl_AppendResult(interp, "\nError converting from RA,Dec\n",
            (char *) NULL);
        Tcl_AppendResult(interp, buff, (char *) NULL);
        return TCL_ERROR;
    }
    return TCL_OK;
}

/*  ------------------------------------------------------------------------- */

/*+                     M e a n  T o  A p p  R a  D e c  { }
 *
 *  Command name:
 *      MeanToAppRaDec
 *
 *  Function:
 *      Converts mean Ra,Dec coordinates to apparent coordinates.
 *
 *  Description:
 *      This routine takes a pair of mean Ra,Dec coordinates and returns the
 *      corresponding apparent Ra,Dec coordinates, given the Mjd in the
 *      current configuration.
 *
 *  Call:
 *      MeanToAppRaDec Cfid MeanRa MeanDec AppRa AppDec
 *
 *  Language:
 *      Tcl
 *      
 *  Parameters:
 *      Cfid       (Sds id) Id of the overall configuration structure
 *      MeanRa     (double) Mean Ra in radians
 *      MeanDec    (double) Mean Dec in radians
 *      AppRa      (double Tcl variable) Variable to receive app Ra in radians
 *      AppDec     (double Tcl variable) Variable to receive app Dec in radians
 *
 *  Support: Keith Shortridge, AAO
 *-
 *  History:
 *       07-Jun-2001 - KS - Original version
 */

static int MeanToAppRaDec(ClientData clientdata DUNUSED, Tcl_Interp * interp,
    int argc, char *argv[])
{
    SdsIdType id, idf;
    double mjd;
    double raApp,decApp;
    double raMean,decMean;
    char buff[200];
    StatusType status = STATUS__OK;
    
    if (argc != 6) {
        interp->result = "wrong # args";
        return TCL_ERROR;
    }
    id = atol(argv[1]);
    
    /*
     *  Get Mjd from field centre details
     */
     
    SdsFind(id, "fieldData", &idf, &status);
    ArgGetd(idf, "configMjd", &mjd, &status);
    SdsFreeId(idf, &status);
    
    /*  Do the conversion */
    
    raMean = atof(argv[2]);
    decMean = atof(argv[3]);
    slaMap(raMean, decMean, 0 ,0,0,0,2000,  mjd, &raApp, &decApp);
    
    /*  Return the values */
    
    sprintf(buff, "%f", raApp);
    Tcl_SetVar(interp, argv[4], buff, 0);
    sprintf(buff, "%f", decApp);
    Tcl_SetVar(interp, argv[5], buff, 0);
    
    if (status != STATUS__OK) {
        MessGetMsg(status, 0, sizeof(buff), buff);
        Tcl_AppendResult(interp, "\nError converting mean to apparent Ra,Dec\n",
            (char *) NULL);
        Tcl_AppendResult(interp, buff, (char *) NULL);
        return TCL_ERROR;
    }
    return TCL_OK;
    
}

/*  ------------------------------------------------------------------------- */

/*
 *                    C o n f  D e l e t e  T a r g e t
 */

static void ConfDeleteTarget (SdsIdType StructId, int Index)

/*+
 *  Name:
 *    ConfDeleteTarget
 *
 *  Function:
 *    Deletes a specified target from a structure containing new targets. 
 *  
 *  Description:
 *    This is a utility routine used by ConfMergeCheckDuplicates(). It
 *    deletes a target in one of the unallocatedObject or unallocatedGuide
 *    structures by taking each array in that structure, obliterating the
 *    data for the target by shifting all the higher indexed data in
 *    that array down one index, and then resizing the array to reflect its
 *    new size.
 *
 *  Call:
 *    ConfDeleteTarget(StructId,Index)
 *
 *   Parameters:
 *
 *    (>) StructId  (Sds Id)   SDS identifier of the structure.
 *    (>) Index     (int) Index of target to be removed (from 0).
 *
 *   Support:
 *     Keith Shortridge, AAO
 *
 *   Version Date:
 *     21-Oct-2013  Original version. KS.    
 *-
 */

{
    StatusType Status = STATUS__OK;
    long CompIndex;
    SdsIdType CompId;
    SdsCodeType Code;
    char Name[NAMELEN];
    long Ndims;
    unsigned long Dims[7];
    unsigned int ItemSize;
    unsigned int Items;
    void* Pointer;
    unsigned char* Cptr;
    unsigned char* Cptr1;
    unsigned long Length;
    int BytesToMove;
    int I;
    
    /*  Work through each component in the structure, looking at the next
     *  until we get an error.
     */
     
    CompIndex = 1;
    for (;;) {
        SdsIndex (StructId,CompIndex,&CompId,&Status);
        if (Status != STATUS__OK) {
            ErsAnnul (&Status);
            break;
        } else {
        
            /*  Found a new component. Get its details. Work out the size
             *  of each element and the number of elements - we only expect 
             *  1D or 2D arrays.
             */
             
            SdsInfo(CompId, Name, &Code, &Ndims, Dims, &Status);
            if (Code != SDS_STRUCT) {
                ItemSize = 4;
                if (Code == SDS_CHAR || Code == SDS_BYTE || Code == SDS_UBYTE) {
                    ItemSize = 1;
                } else if (Code == SDS_SHORT || Code == SDS_USHORT) {
                    ItemSize = 2;
                } else if (Code == SDS_I64 || Code == SDS_UI64 ||
                                                         Code == SDS_DOUBLE) {
                    ItemSize = 8;
                }
                if (Ndims == 2) {
                    ItemSize*= Dims[0];
                    Items = Dims[1];
                } else {
                    Items = Dims[0];
                }
                
                /*  If there are more items in this component than our index
                 *  number, we remove the item at that index. If this is the
                 *  last item in the array, we just have to resize it. 
                 *  Otherwise we have to shuffle the data down first.
                 */
                 
                if (Index < Items) {
                    if (Index < (Items - 1)) {
                        SdsPointer (CompId,&Pointer,&Length,&Status);
                        if (Status == STATUS__OK) {
                            Cptr = (unsigned char*) Pointer + Index * ItemSize;
                            Cptr1 = Cptr + ItemSize;
                            BytesToMove = (Items - Index - 1) * ItemSize;
                            for (I = 0; I < BytesToMove; I++) {
                               *Cptr++ = *Cptr1++;
                            }
                        }
                    }
                    if (Ndims == 2) Dims[1]--;
                    else Dims[0]--;
                    SdsResize (CompId,Ndims,Dims,&Status);
                }
            }
            SdsFreeId(CompId,&Status);
        }
        CompIndex++;
    }
}
            
      
      

/*  ------------------------------------------------------------------------- */

/*
 *                    M e r g e  C h e c k  D u p l i c a t e s  { }
 */
 
static int ConfMergeCheckDuplicates(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc, char *argv[])

/*+
 *  Command name:
 *    MergeCheckDuplicates
 *
 *  Function:
 *    Removes already known targets from a structure containing new targets. 
 *  
 *  Description:
 *    The 'merge' function takes a structure containing new unallocated 
 *    program and guide objects and merges them with an existing allocated 
 *    field structure. It turned out (quite late in the day) that there are
 *    (unsurprisingly, really) problems introduced if we end up with duplicate
 *    targets as a result of the merge. That could be fixed by making the 
 *    merge process more intelligent, but it seemed safer to keep the changes
 *    in one place and introduce this routine. This runs through the 
 *    unallocated targets in the new structure and removes any that it finds
 *    match target names in the exisiting structure. Removing a target means
 *    modifying the SDS structure, which is a non-trivial operation, but this
 *    is handled by ConfDeleteTarget().
 *
 *  Call:
 *    MergeCheckDuplicates CurrentId NewId
 *
 *   Parameters:
 *
 *    (>) id  (Sds Id)   SDS identifier of existing structure.
 *    (>) pivid (Sds Id) SDS identifier of the new structure.
 *
 *   Returns:
 *     "OK" if succesful, "Error" otherwise
 *
 *   Support:
 *     Keith Shortridge, AAO
 *
 *   Version Date:
 *     21-Oct-2013  Original version. KS.    
 *-
 */
{
    char Buffer[300];
    SdsCodeType Code;
    long Ndims;
    unsigned long Actlen;
    unsigned long Dims[7];
    char Name[NAMELEN];
    char NewName[NAMELEN];
    char CurName[NAMELEN];
    SdsIdType CurTopId,CurUgId,CurUobjId,CurObjId;
    SdsIdType CurUgNameId,CurUobjNameId,CurObjNameId;
    SdsIdType NewTopId,NewUgId,NewUobjId,NewObjId;
    SdsIdType NewUgNameId,NewUobjNameId,NewObjNameId;
    int NCurUg,NCurUobj,NCurObj;
    int NNewUg,NNewUobj,NNewObj;
    short CurUgUsed,CurUobjUsed,CurObjUsed,NewUgUsed,NewUobjUsed,NewObjUsed;
    short CurUgNameUsed,CurUobjNameUsed,CurObjNameUsed;
    short NewUgNameUsed,NewUobjNameUsed,NewObjNameUsed;
    short Exists;
    int NDeleted;
    int I,J;
    StatusType Status = STATUS__OK;

    if (argc != 3) {
        interp->result = "Wrong # args";
        return TCL_ERROR;
    }

    /*  Get the passed SdsIds to the current and the new structures */
    
    CurTopId = (SdsIdType) (atol(argv[1]));
    NewTopId = (SdsIdType) (atol(argv[2]));
    
    /*  Keeping track of the SDS Id is a pain. You could probably get away
     *  with setting them to zero and freeing them if they're non-zero
     *  at the end, like allocated memory. But strictly you don't know that.
     */
    
    CurUgUsed = CONFIG_FALSE;
    CurUobjUsed = CONFIG_FALSE;
    CurObjUsed = CONFIG_FALSE;
    NewUgUsed = CONFIG_FALSE;
    NewUobjUsed = CONFIG_FALSE;
    NewObjUsed = CONFIG_FALSE;
    CurUgNameUsed = CONFIG_FALSE;
    CurUobjNameUsed = CONFIG_FALSE;
    CurObjNameUsed = CONFIG_FALSE;
    NewUgNameUsed = CONFIG_FALSE;
    NewUobjNameUsed = CONFIG_FALSE;
    NewObjNameUsed = CONFIG_FALSE;
    
    /*  Now we set ourselves up. We need the SdsIds of the name arrays in
     *  both input structures, and we need to know their sizes (which gives
     *  us the number of targets in each). Each structure has an unallocGuide,
     *  and unallocObject and a objects structure. (Actually, any of these may
     *  be missing, and we need to handle that as well.)
     *
     *  First, the unallocated guide targets in the current structure.
     */
     
    NCurUg = 0;
    SdsFind (CurTopId,"unallocGuide",&CurUgId,&Status);
    if (Status == 0) {
        CurUgUsed = CONFIG_TRUE;
        SdsFind (CurUgId,"name",&CurUgNameId,&Status);
        SdsInfo(CurUgNameId, Name, &Code, &Ndims, Dims, &Status);
        NCurUg = Dims[1];
        if (Status == 0) CurUgNameUsed = CONFIG_TRUE;
    } else if (Status == SDS__NOITEM) {
        ErsAnnul(&Status);
    } else {
        interp->result = "Error searching current unallocated guide structure";
        return TCL_ERROR;
    }
    
    /*  The unallocated object targets in the current structure. */
    
    NCurUobj = 0;
    SdsFind (CurTopId,"unallocObject",&CurUobjId,&Status);
    if (Status == 0) {
        CurUobjUsed = CONFIG_TRUE;
        SdsFind (CurUobjId,"name",&CurUobjNameId,&Status);
        SdsInfo(CurUobjNameId, Name, &Code, &Ndims, Dims, &Status);
        NCurUobj = Dims[1];
        if (Status == 0) CurUobjNameUsed = CONFIG_TRUE;
    } else if (Status == SDS__NOITEM) {
        ErsAnnul(&Status);
    } else {
        interp->result = "Error searching current unallocated obj structure";
        return TCL_ERROR;
    }
    
    /*  The allocated objects in the current structure. These can be target
     *  objects or guide objects.
     */
    
    NCurObj = 0;
    SdsFind (CurTopId,"objects",&CurObjId,&Status);
    if (Status == 0) {
        CurObjUsed = CONFIG_TRUE;
        SdsFind (CurObjId,"name",&CurObjNameId,&Status);
        SdsInfo(CurObjNameId, Name, &Code, &Ndims, Dims, &Status);
        NCurObj = Dims[1];
        if (Status == 0) CurObjNameUsed = CONFIG_TRUE;
    } else if (Status == SDS__NOITEM) {
        ErsAnnul(&Status);
    } else {
        interp->result = "Error searching current objects structure";
        return TCL_ERROR;
    }
    
    /*  The unallocated guide targets in the new structure. */
    
    NNewUg = 0;
    SdsFind (NewTopId,"unallocGuide",&NewUgId,&Status);
    if (Status == 0) {
        NewUgUsed = CONFIG_TRUE;
        SdsFind (NewUgId,"name",&NewUgNameId,&Status);
        SdsInfo(NewUgNameId, Name, &Code, &Ndims, Dims, &Status);
        NNewUg = Dims[1];
        if (Status == 0) NewUgNameUsed = CONFIG_TRUE;
    } else if (Status == SDS__NOITEM) {
        ErsAnnul(&Status);
    } else {
        interp->result = "Error searching new unallocated guide structure";
        return TCL_ERROR;
    }
    
    /*  The unallocated object targets in the new structure. */
    
    NNewUobj = 0;
    SdsFind (NewTopId,"unallocObject",&NewUobjId,&Status);
    if (Status == 0) {
        NewUobjUsed = CONFIG_TRUE;
        SdsFind (NewUobjId,"name",&NewUobjNameId,&Status);
        SdsInfo(NewUobjNameId, Name, &Code, &Ndims, Dims, &Status);
        NNewUobj = Dims[1];
        if (Status == 0) NewUobjNameUsed = CONFIG_TRUE;
    } else if (Status == SDS__NOITEM) {
        ErsAnnul(&Status);
    } else {
        interp->result = "Error searching new unallocated object structure";
        return TCL_ERROR;
    }
    
    /*  The allocated object targets in the new structure. Strictly, we
     *  don't need to bother with these, since the merge process ignores
     *  them - it only merges the unallocated targets in the new structure.
     *  But it seems worth it to check them anyway.
     */
     
    NNewObj = 0;
    SdsFind (NewTopId,"objects",&NewObjId,&Status);
    if (Status == 0) {
        NewObjUsed = CONFIG_TRUE;
        SdsFind (NewObjId,"name",&NewObjNameId,&Status);
        SdsInfo(NewObjNameId, Name, &Code, &Ndims, Dims, &Status);
        NNewObj = Dims[0];
        if (Status == 0) NewObjNameUsed = CONFIG_TRUE;
    } else if (Status == SDS__NOITEM) {
        ErsAnnul(&Status);
    } else {
        interp->result = "Error searching new objects structure";
        return TCL_ERROR;
    }
    
    /*  Now, we have up to three name arrays in the new structure, and
     *  up to three name arrays in the current structure. We start with
     *  the unallocated objects in the new structure. If any of these
     *  match an object in any of the name arrays in the current structure,
     *  then we have a match and need to delete that object from the 
     *  new structure. Note that when we delete the object, all subsequent
     *  objects move down in the name array and the array is truncated.
     *  So when we delete an object, we don't want to increment the index
     *  through the array - if we leave it unchanged, but modify the loop
     *  limit. Which is why the loop index isn't incrememted in the for
     *  statement. We probably shouldn't need to check a new unallocated
     *  object against the list of current unallocated guides, but it's
     *  playing safe.
     */
    
    NDeleted = 0;
    for (I = 0; I < NNewUobj; ) {
        Exists = CONFIG_FALSE;
        SdsGet(NewUobjNameId,NAMELEN,NAMELEN * I,NewName,&Actlen,&Status);
        for (J = 0; J < NCurUobj; J++) {
            SdsGet(CurUobjNameId,NAMELEN,NAMELEN * J,CurName,&Actlen,&Status);
            if (!strncmp(NewName,CurName,NAMELEN)) {
                ErsSPrintf(sizeof(Buffer), Buffer,
                 "Unallocated object '%s' matches current unallocated object",
                 NewName);
                ConfMessage(interp, Buffer);
                Exists = CONFIG_TRUE;
            }
        }
        for (J = 0; J < NCurObj; J++) {
            SdsGet(CurObjNameId,NAMELEN,NAMELEN * J,CurName,&Actlen,&Status);
            if (!strncmp(NewName,CurName,NAMELEN)) {
                ErsSPrintf(sizeof(Buffer), Buffer,
                 "Unallocated object '%s' matches current allocated object",
                 NewName);
                ConfMessage(interp, Buffer);
                Exists = CONFIG_TRUE;
            }
        }
        for (J = 0; J < NCurUg; J++) {
            SdsGet(CurUgNameId,NAMELEN,NAMELEN * J,CurName,&Actlen,&Status);
            if (!strncmp(NewName,CurName,NAMELEN)) {
                ErsSPrintf(sizeof(Buffer), Buffer,
                 "Unallocated object '%s' matches current unallocated guide",
                 NewName);
                ConfMessage(interp, Buffer);
                Exists = CONFIG_TRUE;
            }
        }
        if (Exists) {
            ConfDeleteTarget(NewUobjId,I);
            NNewUobj--;
            NDeleted++;
        } else {
            I++;
        }
    }
    
    /*  Now the same for the new unallocated guide targets */
    
    for (I = 0; I < NNewUg; ) {
        Exists = CONFIG_FALSE;
        SdsGet(NewUgNameId,NAMELEN,NAMELEN * I,NewName,&Actlen,&Status);
        for (J = 0; J < NCurUg; J++) {
            SdsGet(CurUgNameId,NAMELEN,NAMELEN * J,CurName,&Actlen,&Status);
            if (!strncmp(NewName,CurName,NAMELEN)) {
                ErsSPrintf(sizeof(Buffer), Buffer,
                 "Unallocated guide '%s' matches current unallocated guide",
                 NewName);
                ConfMessage(interp, Buffer);
                Exists = CONFIG_TRUE;
            }
        }
        for (J = 0; J < NCurObj; J++) {
            SdsGet(CurObjNameId,NAMELEN,NAMELEN * J,CurName,&Actlen,&Status);
            if (!strncmp(NewName,CurName,NAMELEN)) {
                ErsSPrintf(sizeof(Buffer), Buffer,
                 "Unallocated guide '%s' matches current allocated object",
                 NewName);
                ConfMessage(interp, Buffer);
                Exists = CONFIG_TRUE;
            }
        }
        for (J = 0; J < NCurUobj; J++) {
            SdsGet(CurUobjNameId,NAMELEN,NAMELEN * J,CurName,&Actlen,&Status);
            if (!strncmp(NewName,CurName,NAMELEN)) {
                ErsSPrintf(sizeof(Buffer), Buffer,
                 "Unallocated guide '%s' matches current unallocated object",
                 NewName);
                ConfMessage(interp, Buffer);
                Exists = CONFIG_TRUE;
            }
        }
        if (Exists) {
            ConfDeleteTarget(NewUgId,I);
            NNewUg--;
            NDeleted++;
        } else {
            I++;
        }
    }
    
    /*  And the allocated objects in the new structure. I know we don't
     *  need to do this, and it may even be confusing to do so, but it 
     *  seems worth being able to point out that are duplicates here, even
     *  if we don't remove them. (If we did remove them, ConfDeleteTargets{}
     *  is not the way to do it. We need to set the entry to unallocated
     *  and clear out the entry details.)
     */
     
    for (I = 0; I < NNewObj; I++) {
        Exists = CONFIG_FALSE;
        SdsGet(NewObjNameId,NAMELEN,NAMELEN * I,NewName,&Actlen,&Status);
        for (J = 0; J < NCurUg; J++) {
            SdsGet(CurUgNameId,NAMELEN,NAMELEN * J,CurName,&Actlen,&Status);
            if (!strncmp(NewName,CurName,NAMELEN)) {
                ErsSPrintf(sizeof(Buffer), Buffer,
                 "(Allocated object '%s' matches current unallocated guide)",
                 NewName);
                ConfMessage(interp, Buffer);
                Exists = CONFIG_TRUE;
            }
        }
        for (J = 0; J < NCurObj; J++) {
            SdsGet(CurObjNameId,NAMELEN,NAMELEN * J,CurName,&Actlen,&Status);
            if (!strncmp(NewName,CurName,NAMELEN)) {
                ErsSPrintf(sizeof(Buffer), Buffer,
                 "(Allocated object '%s' matches current allocated object)",
                 NewName);
                ConfMessage(interp, Buffer);
                Exists = CONFIG_TRUE;
            }
        }
        for (J = 0; J < NCurUobj; J++) {
            SdsGet(CurUobjNameId,NAMELEN,NAMELEN * J,CurName,&Actlen,&Status);
            if (!strncmp(NewName,CurName,NAMELEN)) {
                ErsSPrintf(sizeof(Buffer), Buffer,
                 "(Allocated object '%s' matches current unallocated object",
                 NewName);
                ConfMessage(interp, Buffer);
                Exists = CONFIG_TRUE;
            }
        }
    }
    
    if (NDeleted > 0) {
        ErsSPrintf(sizeof(Buffer),Buffer,
           "Removed %d duplicate unallocated targets prior to merge",NDeleted);
        ConfMessage(interp, Buffer);
    } else {
        ConfMessage(interp,"Checked for duplicate targets: none found");
    }


    /*  Release all the used SDD ids */
    
    if (CurUgUsed) SdsFreeId(CurUgId,&Status);
    if (CurUobjUsed)SdsFreeId(CurUobjId,&Status);
    if (CurObjUsed) SdsFreeId(CurObjId,&Status);
    if (NewUgUsed) SdsFreeId(NewUgId,&Status);
    if (NewUobjUsed) SdsFreeId(NewUobjId,&Status);
    if (NewObjUsed) SdsFreeId(NewObjId,&Status);
    if (CurUgNameUsed) SdsFreeId(CurUgNameId,&Status);
    if (CurUobjNameUsed) SdsFreeId(CurUobjNameId,&Status);
    if (CurObjNameUsed) SdsFreeId(CurObjNameId,&Status);
    if (NewUgNameUsed) SdsFreeId(NewUgNameId,&Status);
    if (NewUobjNameUsed) SdsFreeId(NewUobjNameId,&Status);
    if (NewObjNameUsed) SdsFreeId(NewObjNameId,&Status);
    
    return TCL_OK;
}

/*  ------------------------------------------------------------------------- */

/*+                  U n a l l o c a t e d  T y p e  { }
 *
 *  Command name:
 *      UnallocatedType
 *
 *  Function:
 *      Return the type code used for an unallocated fibre.
 *
 *  Description:
 *      The code used for an unallocated fibre in the "objects" section
 *      depends on the instrument. This routine returns the code (a single
 *      character) used for the currently selected instrument.
 *
 *  Call:
 *      UnallocatedType 
 *
 *  Language:
 *      Tcl
 *
 *  Support: Keith Shortridge, AAO
 *-
 *  History:
 *      31-Jul-2000 - KS - Original version
 */

static int UnallocatedType(ClientData clientdata DUNUSED, Tcl_Interp * interp,
    int argc DUNUSED, char *argv[]DUNUSED)
{
    char String[2];
    String[0] = FpilUnallocatedType(*Instrument);
    String[1] = '\0';
    Tcl_SetResult(interp, String, TCL_VOLATILE);
    return TCL_OK;
}

/*  ------------------------------------------------------------------------- */

/*+                   T y p e  I s  S k y  { }
 *
 *  Command name:
 *      TypeIsSky
 *
 *  Function:
 *      Tests if a specified target type indicates a sky position.
 *
 *  Description:
 *      It is possible to tell from just the 'type' character used if a given
 *      target represents a sky position, and this routine does so, returning 
 *      true if it is and false otherwise. 
 *
 *  Call:
 *      TypeIsSky targetType 
 *
 *   Parameters:
 * 
 *      (>) targetType  (char) The type code character for the target in
 *                             question.
 *     
 *  Support: Keith Shortridge, AAO
 *-
 *  History:
 *       16-Aug-2000 - KS - Original version
 */

static int TypeIsSky(ClientData clientdata DUNUSED, Tcl_Interp * interp,
    int argc, char *argv[])
{
    /*  Validate the arguments */

    if (argc != 2) {
        Tcl_SetResult (interp,"wrong # args",TCL_STATIC);
        return TCL_ERROR;
    }
    if (FpilIsTargetSky(*Instrument,*(argv[1]))) {
       Tcl_SetResult(interp, "1",TCL_STATIC);
    } else {
       Tcl_SetResult(interp, "0",TCL_STATIC);
    }
    return TCL_OK;
}

/*  ------------------------------------------------------------------------- */

/*+                   T y p e  I s  G u i d e  { }
 *
 *  Command name:
 *      TypeIsTelGuide
 *
 *  Function:
 *      Tests if a specified target type indicates a guide star.
 *
 *  Description:
 *      It is possible to tell from just the 'type' character used if a given
 *      target is a fiducial guide star, and this routine does so, returning 
 *      true if it is and false otherwise. Note that this also returns true
 *      if the target is a telescope guide object, so one may want to use 
 *      the TypeIsTelGuide command in conjunction with it to distinguish 
 *      between the two cases.
 *
 *  Call:
 *      TypeIsGuide targetType
 *
 *   Parameters:
 * 
 *      (>) targetType  (char) The type code character for the target in
 *                             question.
 *     
 *  Support: Keith Shortridge, AAO
 *-
 *  History:
 *      16-Aug-2000 - KS - Original version
 */

static int TypeIsGuide(ClientData clientdata DUNUSED, Tcl_Interp * interp,
    int argc, char *argv[])
{
    /*  Validate the arguments */

    if (argc != 2) {
        Tcl_SetResult (interp,"wrong # args",TCL_STATIC);
        return TCL_ERROR;
    }
    if (FpilIsTargetGuide(*Instrument,*(argv[1]))) {
       Tcl_SetResult(interp, "1",TCL_STATIC);
    } else {
       Tcl_SetResult(interp, "0",TCL_STATIC);
    }
    return TCL_OK;
}

/*  ------------------------------------------------------------------------- */

/*+                   T y p e  I s  T e l  G u i d e
 *
 *  Command name:
      TypeIsTelGuide

 *  Function:
        Tests if a specified target type indicates a telescope guide object.

 *  Description:
        Some instruments - such as the VLT - support the concept of a
        target type that is intended only for use by a telescope guider
        and not by a fibre. It is possible to tell from just the 'type'
        character used if a given target is such a telescope guider target,
        and this routine does so, returning true if it is and false
        otherwise.

 *  Call:
      TypeIsTelGuide  

 *   Parameters:
 
     (>) targetType  (char) The type code character for the target in question.
     
 *  Support: Keith Shortridge, AAO

 *-

 *  History:
       16-Aug-2000 - KS - Original version
  
*/

static int TypeIsTelGuide(ClientData clientdata DUNUSED, Tcl_Interp * interp,
    int argc, char *argv[])
{
    /*  Validate the arguments */

    if (argc != 2) {
        Tcl_SetResult (interp,"wrong # args",TCL_STATIC);
        return TCL_ERROR;
    }
    if (FpilIsTargetTelGuide(*Instrument,*(argv[1]))) {
       Tcl_SetResult(interp, "1",TCL_STATIC);
    } else {
       Tcl_SetResult(interp, "0",TCL_STATIC);
    }
    return TCL_OK;
}

/*  ------------------------------------------------------------------------- */

/*+                     U s e  G r e y
 *
 *  Command name:
      UseGrey

 *  Function:
        Tests if the grey scale display flag is set.

 *  Description:
        This routine returns the value of the global flag 'useGrey',
        which is used to indicate that a grey-scale colour scheme
        should be used.

 *  Call:
        UseGrey  

 *   Parameters: None
      
 *  Support: Keith Shortridge, AAO

 *-

 *  History:
       15-Jun-2001 - KS - Original version
  
*/

static int ConfUseGrey(ClientData clientdata DUNUSED, Tcl_Interp * interp,
    int argc DUNUSED, char *argv[] DUNUSED)
{
    if (useGrey) {
       Tcl_SetResult(interp, "1",TCL_STATIC);
    } else {
       Tcl_SetResult(interp, "0",TCL_STATIC);
    }
    return TCL_OK;
}

/*  ------------------------------------------------------------------------- */

/*+                         S e t  G r e y  { }
 *
 *  Command name:
 *      SetGrey
 *
 *  Function:
 *        Sets the grey scale display flag.
 *
 *  Description:
 *        This routine sets the value of the global flag 'useGrey',
 *        which is used to indicate that a grey-scale colour scheme
 *        should be used.
 *
 *  Call:
 *        SetGrey [flag]  
 *
 *  Parameters:
 * 
 *        (>) Flag  (int) Non-zero if a grey-scale scheme is to be used. If
 *                        omitted, the grey-scale flag is set true.
 *     
 *  Support: Keith Shortridge, AAO
 *-
 *  History:
 *       15-Jun-2001 - KS - Original version
 */

static int ConfSetGrey(ClientData clientdata DUNUSED, Tcl_Interp * interp,
    int argc, char *argv[])
{
    /*  Validate the arguments */

    if (argc == 1) {
        useGrey = 1;
    } else if (argc == 2) {
        useGrey = atoi(argv[1]);
    } else {
        Tcl_SetResult (interp,"wrong # args",TCL_STATIC);
        return TCL_ERROR;
    }
    return TCL_OK;
}

/*  ------------------------------------------------------------------------- */

/*+                        E x p e r t  A l l o w e d  { }
 *
 *  Command name:
      ExpertAllowed

 *  Function:
        Tests if the 'expert mode allowed' flag is set.

 *  Description:
        This routine returns the value of the global flag 'ExpertAllowed',
        which is used to indicate that the program is being run in a way
        that allows expert mode - in practice, this is if the command used
        to run the program included the word 'expert'. (The obvious way
        to do this is to have a link to the usual configure executable that
        has expert in the name - at least under UNIX.)

 *  Call:
        ExpertAllowed  

 *   Parameters: None
      
 *  Support: Keith Shortridge, AAO

 *-

 *  History:
        5-Nov-2001 - KS - Original version
  
*/

static int ConfExpertAllowed(ClientData clientdata DUNUSED,
        Tcl_Interp * interp, int argc DUNUSED, char *argv[] DUNUSED)
{
    if (ExpertAllowed) {
       Tcl_SetResult(interp, "1",TCL_STATIC);
    } else {
       Tcl_SetResult(interp, "0",TCL_STATIC);
    }
    return TCL_OK;
}

/*  ------------------------------------------------------------------------- */

/*
 *                        V L T  G u i d e r  { }
 */
 
static int VLTGuider(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc, char *argv[])
    
/*+
 *  Command Name:
 *    VLTGuider
 *
 *  Function:
 *    Determine the two guide probe positions for a specified target star.
 *
 *  Call:
 *    VLTGuider id index posn1 posn2
 *
 *  Description:
 *    This routine is passed the details of a possible VLT guide probe target
 *    star, in terms of the SDS Id for a sub-structure of the configuration
 *    data (in practice this should be the unallocated guide sub-structure)
 *    and an index. It calculates the two possible VLT guide probe positions
 *    and returns their details.
 *
 *  Parameters:
 *    (>)  id    (Sds Id)  Sds identifier of the substructure
 *    (>)  index (int)     index number - starting at one
 *    (<)  posn1 (char)    String giving "XP YP XS YS" for guide
 *                         probe in first position, where XP,YP is
 *                         location of guide pivot and XS,YS is
 *                         the location of the star.
 *    (<)  posn2 (char)    String giving "XP YP XS YS" for guide
 *                         probe in second position.
 *    (<)  pospos (int)    Either 1 or 2, depending on whether the
 *                         first position is classed as +ve or -ve.
 *                         If the first is +ve, this is returned as
 *                         1. If the second is +ve, it is returned
 *                         as 2.
 *
 *   Support:
 *     Keith Shortridge, AAO
 *
 *   Version Date:
 *     9-Aug-1999  Original version.
 *    22-Oct-2002  Code setting POS and NEG simplified and corrected.    
 *-
 */
{
    StatusType status;
    long index_no;
    SdsIdType id;
    SdsIdType topid;
    char type;
    unsigned long actlen;

    if (argc != 6) {
        interp->result = "wrong # args";
        return TCL_ERROR;
    }
    status = SDS__OK;
    topid = (SdsIdType) (atol(argv[1]));
    index_no = atol(argv[2]) - 1;

    SdsFind(topid, "type", &id, &status);
    SdsGet(id, sizeof(char), index_no, &type, &actlen, &status);
    SdsFreeId(id, &status);

    /*  This test for a 'Guide' target isn't strictly correct, as it will
     *  include ordinary guide stars as well as the more specialised VLT
     *  guide probe, which is what we should really be testing for here.
     */

    if (FpilIsTargetGuide(*Instrument, type)) {

        double GuideStarRadius;
        double GuideProbeLength = 560000;
        double Dx, Dy;
        int X1, Y1, X2, Y2;
        int X, Y;               /*  Guide star plate coordinates */
        double Theta, Phi, PhiOne, PhiTwo;
        double Pi = 3.14159;
        char Buffer[80];
        
        SdsFind(topid, "x", &id, &status);
        SdsGet(id, sizeof(int), index_no, &X, &actlen, &status);
        SdsFreeId(id, &status);
        SdsFind(topid, "y", &id, &status);
        SdsGet(id, sizeof(int), index_no, &Y, &actlen, &status);
        SdsFreeId(id, &status);

        /*  This calculation, which is messier than it needs to be, 
         *  works out two angles: Theta is the Angle between the
         *  line going straight up from the center of the field (0,0)
         *  and the line from the center of the field to the
         *  position of the guide star (X,Y), measured anti-clockwise.
         *  Now imagine a triangle. One of the sides is the 
         *  line from (0,0) to (X,Y), that is the line from the center
         *  of the field to the guide star. This line has a length 
         *  we calculate as GuideStarRadius. One of the sides is a line
         *  from the center of the field to the position of the rotator
         *  axis of the guide probe in one of the two possible positions
         *  it can use for this star. This line must be of length
         *  GuideProbeLength. The position of the rotator axis of the
         *  guide probe may be anywhere on the circle of radius 
         *  GuideProbeLength drawn around the center of the field. The
         *  third side of the triangle is the line from the center of
         *  the rotator axis to the guide star. This also has length
         *  GuideProbeLength - as far as I know the radius of the circle
         *  is exactly the length from the rotator axis to the center of
         *  the probe itself, although this isn't necessarily true.
         *  This is therefore an isosceles triangle, with two angles
         *  the same, both having the value we call Phi in this code.
         *  Trivial trig then gives Phi. This would be a lot
         *  easier with a diagram!
         */

        if ((abs(X) < 1) && (abs(Y) < 1)) {
            Theta = 0.0;
            GuideStarRadius = 0.0;
            Phi = 0.0;
        } else {
            Dx = abs(X);
            Dy = abs(Y);
            GuideStarRadius = sqrt((Dx * Dx) + (Dy * Dy));
            Phi = acos(GuideStarRadius / (GuideProbeLength * 2.0));
            Theta = atan(Dy / Dx);
            if ((X < 0) && (Y > 0))
                Theta = Pi - Theta;
            if ((X < 0) && (Y < 0))
                Theta += Pi;
            if ((X > 0) && (Y < 0))
                Theta = (2 * Pi) - Theta;
        }

        /*  You can now draw two lines from the center of the field to
         *  the two possible positions of the guide probe rotator axis.
         *  Both of these will be Phi radians away from the line
         *  joining the field center to the guide star, one on each
         *  side of that line. These angles (using the same convention
         *  as described above for Theta) we call PhiOne and PhiTwo.
         *  Having got those, we can then calculate the X,Y positions
         *  for the two possible rotator axis positions, (X1,Y1) and
         *  (X2,Y2). The code gets messy because of the different 
         *  quadrants, and I think that could be avoided, but it
         *  works.
         */

        PhiOne = Theta + Phi;
        if (PhiOne > (2.0 * Pi))
            PhiOne -= (2.0 * Pi);
        PhiTwo = Theta - Phi;
        if (PhiTwo < 0.0)
            PhiTwo = (2.0 * Pi) + PhiTwo;
        if (PhiOne > (3.0 * Pi / 2.0)) {
            PhiOne = (2.0 * Pi) - PhiOne;
            Y1 = -(int) (GuideProbeLength * sin(PhiOne));
            X1 = (int) (GuideProbeLength * cos(PhiOne));
        } else if (PhiOne > Pi) {
            PhiOne = PhiOne - Pi;
            Y1 = -(int) (GuideProbeLength * sin(PhiOne));
            X1 = -(int) (GuideProbeLength * cos(PhiOne));
        } else if (PhiOne > (Pi / 2.0)) {
            PhiOne = Pi - PhiOne;
            Y1 = (int) (GuideProbeLength * sin(PhiOne));
            X1 = -(int) (GuideProbeLength * cos(PhiOne));
        } else {
            Y1 = (int) (GuideProbeLength * sin(PhiOne));
            X1 = (int) (GuideProbeLength * cos(PhiOne));
        }
        if (PhiTwo > (3.0 * Pi / 2.0)) {
            PhiTwo = (2.0 * Pi) - PhiTwo;
            Y2 = -(int) (GuideProbeLength * sin(PhiTwo));
            X2 = (int) (GuideProbeLength * cos(PhiTwo));
        } else if (PhiTwo > Pi) {
            PhiTwo = PhiTwo - Pi;
            Y2 = -(int) (GuideProbeLength * sin(PhiTwo));
            X2 = -(int) (GuideProbeLength * cos(PhiTwo));
        } else if (PhiTwo > (Pi / 2.0)) {
            PhiTwo = Pi - PhiTwo;
            Y2 = (int) (GuideProbeLength * sin(PhiTwo));
            X2 = -(int) (GuideProbeLength * cos(PhiTwo));
        } else {
            Y2 = (int) (GuideProbeLength * sin(PhiTwo));
            X2 = (int) (GuideProbeLength * cos(PhiTwo));
        }
        sprintf(Buffer, "%d %d %d %d", X1, Y1, X, Y);
        Tcl_SetVar(interp, argv[3], Buffer, 0);
        sprintf(Buffer, "%d %d %d %d", X2, Y2, X, Y);
        Tcl_SetVar(interp, argv[4], Buffer, 0);
        
        /*  We need to calculate which of the two positions is 
         *  classified as +ve and which is -ve and return this
         *  as the final parameter. The convention is as follows:
         *
         *  POS is when the probe arm has to move CLOCKWISE around 
         *  its rotation point when moving from center position to the
         *  actual guide star, when you look at the adapter from outside.
         *
         *  (This is quoted directly from an e-mail from Luca Pasquini.
         *  We tried to clarify what 'outside' means, and got the reply
         *  that this means from the opposite side to that drawn by
         *  configure - so if you look down on the plate and see the guide
         *  probe between you and the plate, you're looking from 'inside'.)
         *
         *  There was an attempt made to code a test based on calculating
         *  the actual angles involved and comparing them, but it took
         *  Andreas Kaufer at ESO to point out that in fact, with Alpha1
         *  always being less than Phi in the above code, it follows that
         *  the point at X1,Y1 is in fact always the POS position and
         *  the point at X2,Y2 is always the NEG position. So nothing
         *  needs to be calculated. (AKA comment 2002-10-18)
         */
        
        Tcl_SetVar(interp, argv[5], "1", 0);
    } 

    if (status == SDS__OK) {
        return TCL_OK;
    } else {
        char mbuf[100];
        MessGetMsg(status, 0, 100, mbuf);
        sprintf(interp->result, "Error Getting Object Data\n%s", mbuf);
        return TCL_ERROR;
    }
}

/* End of VLTGuider */

/*  ------------------------------------------------------------------------- */

/*                V L T  G u i d e  P r o b e  C o n s t a n t s
 *
 *  The following values define the shape of the VLT Guide probe shaddow.
 *  The shaddow is assumed to consist of a circle centered on the pivot
 *  position of the guide probe, of radius GuideProbeOvalRadius (all
 *  values are in microns), and then a polygon with a number of vertices
 *  given by GuideProbeVertices. The positions of the vertices, relative
 *  to the pivot position, are given by the array GuideProbeVertexValues.
 *  This consists of a set of values, two for each vertex, giving the
 *  x,y coordinates of each vertex. The final vertex should be the same as
 *  the first, so this contains the complete set of boundary lines for
 *  the polygon. So it should have a number of entries
 *  given by 2 * GuideProbeVertices. Note that the code assumes that the
 *  polygon is symmetrical about the X axis. GuideProbeLength should
 *  be the length of the probe in microns - it should be the largest
 *  X-value found in GuideProbeVertexValues.
 *
 *  These vertex coordinates have been taken from the code for vfArmPolygon,
 *  as supplied by Rodrigo Amestica. The earlier versions of this code used
 *  a slightly simplified vertex set, taken from vltguiArmDisplay.tcl as
 *  supplied to AAO by Jason Spyromilio. After discussion with Rodrigo, it
 *  was decided to use the vfArmPolygon values. Luca Pasquini's report on the
 *  FLAMES commissioning included the suggestion 'Suggest enlarging the probe
 *  shadow by 2-3 arcseconds in CONFIGURE to allow deviations from 45 degrees
 *  position angle'. This code does not yet incorporate that additional 
 *  shaddow, as I wanted to experiment with the revised vertex values first.
 */

#define GuideProbeOvalRadius 137000
#define GuideProbeLength 594000
#define GuideProbeVertices 11

static int GuideProbeVertexValues[] =
    { 0, 137000, 212000, 101000, 314000, 64000, 496000, 50000, 594000, 37000,
    594000, -37000, 496000, -50000, 314000, -64000, 212000, -101000,
    0, -137000, 0, 137000
};

/*  ------------------------------------------------------------------------- */

/*
 *                    V L T  D r a w  G u i d e r  { }
 */
 
static int VLTDrawGuider(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc, char *argv[])
    
/*+
 *  Command Name:
 *    VLTDrawGuider
 *
 *  Function:
 *    Returns the Tk canvas drawing commands to draw the VLT guide probe.
 *
 *  Call:
 *    VLTDrawGuider posn zoom oval polygon
 *
 *  Description:
 *    Given a position string for the VLT guide probe (as generated by
 *    VLTGuider{}), this generates a series of parameters that can be used
 *    in Tk canvas drawing commands to draw the guide in the specified
 *    position.
 *
 *  Parameters:
 *    (>)  posn (char)    String giving "XP YP XS YS" for guide
 *                        probe in its position, where XP,YP is
 *                        location of guide pivot and XS,YS is
 *                        the location of the star.
 *    (>)  zoom (float)   The current zoom factor.
 *    (<)  pivot (char)   Numeric parameters for an oval canvas
 *                        command that will draw a small circle
 *                        to indicate the position of the guide
 *                        probe pivot.
 *    (<)  oval (char)    Numeric parameters for an oval canvas
 *                        command that will draw the circular
 *                        part of the guide probe.
 *    (<)  polygon (char) Numeric parameters for a polygon canvas
 *                        command that will draw the remaining
 *                        part of the guide probe.
 *
 *   Support:
 *     Keith Shortridge, AAO
 *
 *   Version Date:
 *     9-Aug-1999    
 *-
 */
{
    int XP, YP, XS, YS;
    int DeltaX, DeltaY;
    char Oval[64];
    char Polygon[512];
    int VIndex;
    float Zoom;
    double Theta;
    int Vertex;
    double Pi = 3.14159;
    if (argc != 6) {
        interp->result = "wrong # args";
        return TCL_ERROR;
    }
    sscanf(argv[1], "%d %d %d %d", &XP, &YP, &XS, &YS);
    Zoom = atof(argv[2]);
    sprintf(Oval, "%d %d %d %d",
        FieldPlateToDispX(0, XP, Zoom) - 10,
        FieldPlateToDispY(0, YP, Zoom) - 10,
        FieldPlateToDispX(0, XP, Zoom) + 10,
        FieldPlateToDispY(0, YP, Zoom) + 10);
    Tcl_SetVar(interp, argv[3], Oval, 0);
    sprintf(Oval, "%d %d %d %d",
        FieldPlateToDispX(0, XP - GuideProbeOvalRadius, Zoom),
        FieldPlateToDispY(0, YP - GuideProbeOvalRadius, Zoom),
        FieldPlateToDispX(0, XP + GuideProbeOvalRadius, Zoom),
        FieldPlateToDispY(0, YP + GuideProbeOvalRadius, Zoom));
    Tcl_SetVar(interp, argv[4], Oval, 0);
    Polygon[0] = '\0';
    DeltaX = XS - XP;
    DeltaY = YS - YP;
    Theta = atan((double) abs(DeltaY) / (double) abs(DeltaX));
    if ((DeltaX > 0) && (DeltaY < 0)) {
        Theta = Theta;
    } else if ((DeltaX > 0) && (DeltaY > 0)) {
        Theta = (Pi * 2.0) - Theta;
    } else if ((DeltaX < 0) && (DeltaY > 0)) {
        Theta = Pi + Theta;
    } else if ((DeltaX < 0) && (DeltaY < 0)) {
        Theta = Pi - Theta;
    }
    VIndex = 0;
    for (Vertex = 0; Vertex < GuideProbeVertices; Vertex++) {
        int XVertex;
        int YVertex;
        int XVertPos;
        int YVertPos;
        char Buffer[32];
        XVertex = GuideProbeVertexValues[VIndex++];
        YVertex = GuideProbeVertexValues[VIndex++];
        XVertPos = (cos(Theta) * XVertex) + (sin(Theta) * YVertex)
            + XP;
        YVertPos = (cos(Theta) * YVertex) - (sin(Theta) * XVertex)
            + YP;
        sprintf(Buffer, "%d %d ", FieldPlateToDispX(0, XVertPos, Zoom),
            FieldPlateToDispY(0, YVertPos, Zoom));
        strcat(Polygon, Buffer);
    }
    Tcl_SetVar(interp, argv[5], Polygon, 0);

    return TCL_OK;

}

/*  End of VLTDrawGuider */


/*  ------------------------------------------------------------------------- */

/*
 *                 V L T  D r a w  G u i d e r  O u t l i n e  { }
 */
 
static int VLTDrawGuiderOutline(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc, char *argv[])
    
/*+
 *  Command Name:
 *    VLTDrawGuiderOutline
 *
 *  Function:
 *    Returns Tk canvas drawing commands to draw the VLT guide probe outline.
 *
 *  Call:
 *    VLTDrawGuiderOutline posn zoom pivot lines arc
 *
 *  Description: 
 *    Given a position string for the VLT guide probe (as generated by
 *    VLTGuider{}, this generates a series of parameters that can be used
 *    in Tk canvas drawing commands to draw the guide in the specified
 *    position, in outline, as opposed to as a filled item.
 *
 *  Parameters:
 *    (>)  posn (char)    String giving "XP YP XS YS" for guide
 *                        probe in its position, where XP,YP is
 *                        location of guide pivot and XS,YS is
 *                        the location of the star.
 *    (>)  zoom (float)   The current zoom factor.
 *    (<)  pivot (char)   Numeric parameters for an oval canvas
 *                        command that will draw a small circle
 *                        to indicate the position of the guide
 *                        probe pivot.
 *    (<)  lines (char)   Numeric parameters for a line canvas
 *                        command that will draw the remaining
 *                        part of the guide probe outline.
 *    (<)  arc (char)     Numeric parameters for an arc canvas
 *                        command that will draw the required arc of
 *                        the circular part of the guide probe.
 *
 *   Support:
 *     Keith Shortridge, AAO
 *
 *   Version Date:
 *     21-Aug-2000    
 *-
 */
{
    int XP, YP, XS, YS;
    int DeltaX, DeltaY;
    char Arc[64];
    char Oval[64];
    char Lines[512];
    int VIndex;
    float Zoom;
    double Theta;
    double ThetaDegrees;
    int Vertex;
    double Pi = 3.14159;
    if (argc != 6) {
        interp->result = "wrong # args";
        return TCL_ERROR;
    }
    sscanf(argv[1], "%d %d %d %d", &XP, &YP, &XS, &YS);
    Zoom = atof(argv[2]);
    sprintf(Oval, "%d %d %d %d",
        FieldPlateToDispX(0, XP, Zoom) - 10,
        FieldPlateToDispY(0, YP, Zoom) - 10,
        FieldPlateToDispX(0, XP, Zoom) + 10,
        FieldPlateToDispY(0, YP, Zoom) + 10);
    Tcl_SetVar(interp, argv[3], Oval, 0);
    Lines[0] = '\0';
    
    /*  Theta is the angle between the positive X-axis and the line joining
     *  the center of the guide pivot to the guide star, measured in a 
     *  clockwise direction.
     */
     
    DeltaX = XS - XP;
    DeltaY = YS - YP;
    Theta = atan((double) abs(DeltaY) / (double) abs(DeltaX));
    if ((DeltaX > 0) && (DeltaY < 0)) {
        Theta = Theta;
    } else if ((DeltaX > 0) && (DeltaY > 0)) {
        Theta = (Pi * 2.0) - Theta;
    } else if ((DeltaX < 0) && (DeltaY > 0)) {
        Theta = Pi + Theta;
    } else if ((DeltaX < 0) && (DeltaY < 0)) {
        Theta = Pi - Theta;
    }
    VIndex = 0;
    
    /*  The set of parameters for the line command is generated in the
     *  same way as the polygon parameters generated by VLTDrawGuider(),
     *  but we omit the final closing line of the polygon, so we end
     *  up with an open line.
     */

    for (Vertex = 0; Vertex < GuideProbeVertices - 1; Vertex++) {
        int XVertex;
        int YVertex;
        int XVertPos;
        int YVertPos;
        char Buffer[32];
        XVertex = GuideProbeVertexValues[VIndex++];
        YVertex = GuideProbeVertexValues[VIndex++];
        XVertPos = (cos(Theta) * XVertex) + (sin(Theta) * YVertex)
            + XP;
        YVertPos = (cos(Theta) * YVertex) - (sin(Theta) * XVertex)
            + YP;
        sprintf(Buffer, "%d %d ", FieldPlateToDispX(0, XVertPos, Zoom),
            FieldPlateToDispY(0, YVertPos, Zoom));
        strcat(Lines, Buffer);
    }
    Tcl_SetVar(interp, argv[4], Lines, 0);
    
    /*  The parameters for the arc need to be based on the oval drawn
     *  by VLTDrawGuider(), but with a start and extent added. The start
     *  angle comes from the Theta calculated above, and the extent is 
     *  always going to be 180 degrees. We use -180 because the arc command
     *  convention is that the extent is measured in an anti-clockwise
     *  direction from the start angle. The start angle is the angle between
     *  the positive X-axis and the line joining the center of the arc to
     *  its start point and so is 270-Theta, in degrees.
     */
     
    ThetaDegrees = Theta * 180.0 / Pi;
    sprintf(Arc, "%d %d %d %d -start %f -extent %d",
        FieldPlateToDispX(0, XP - GuideProbeOvalRadius, Zoom),
        FieldPlateToDispY(0, YP - GuideProbeOvalRadius, Zoom),
        FieldPlateToDispX(0, XP + GuideProbeOvalRadius, Zoom),
        FieldPlateToDispY(0, YP + GuideProbeOvalRadius, Zoom),
        270.0 - ThetaDegrees, -180);
    Tcl_SetVar(interp, argv[5], Arc, 0);

    return TCL_OK;

}

/*  End of VLTDrawGuiderOutline */

/*  ------------------------------------------------------------------------  */

/*
 *                    V L T  G u i d e  S h a d d o w  { }
 */
 
static int VLTGuideShaddow(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc, char *argv[])
    
/*+
 *  Command Name:
 *    VLTGuideShaddow
 *
 *  Function:
 *    Determines if a target is shaddowed by the VLT guide probe.
 *
 *  Call:
 *    VLTGuideShaddow id index posn
 *
 *  Description: 
 *    This routine is passed the details of a target star, in terms of the 
 *    SDS Id for a sub-structure of the configuration data  and an index.
 *    Given a position string for the VLT guide probe (as generated by
 *    VLTGuider{}), this routine determines whether or not the target will
 *    be sahddowed by the guide probe. If it is, it flags the target star
 *    as disabled.
 *
 *  Parameters:
 *    (>)  id    (Sds Id)  Sds identifier of the substructure
 *    (>)  index (int)     index number - starting at one
 *    (>)  posn  (char)    String giving "XP YP XS YS" for guide
 *                         probe in its position, where XP,YP is
 *                         location of guide pivot and XS,YS is
 *                         the location of the star. If all of
 *                         XP,YP,XS and YS are zero, then no guide
 *                         probe position has been selected.
 *   Returns:
 *     "OK" if the target is not shaddowed by the guide probe, "shaddowed"
 *     if it is.
 *
 *   Side effects:
 *     The type field for the target object in the Sds structure is set to 
 *     indicate that the target is shaddowed or not.
 *
 *   Support:
 *     Keith Shortridge, AAO
 *
 *   Version Date:
 *     17-Aug-1999    
 *-
 */
{

    int XP, YP, XS, YS;
    int X, Y;
    int Shaddowed;
    double Dx, Dy;
    double DRadius;
    StatusType SdsStatus;
    SdsIdType TopId, Id;
    int IndexNo;
    double Pi = 3.14159;
    unsigned long Actlen;
    char Type;

    if (argc != 4) {
        interp->result = "wrong # args";
        return TCL_ERROR;
    }

    SdsStatus = SDS__OK;
    TopId = (SdsIdType) (atol(argv[1]));
    IndexNo = atol(argv[2]) - 1;
    SdsFind(TopId, "x", &Id, &SdsStatus);
    SdsGet(Id, sizeof(int), IndexNo, &X, &Actlen, &SdsStatus);
    SdsFreeId(Id, &SdsStatus);
    SdsFind(TopId, "y", &Id, &SdsStatus);
    SdsGet(Id, sizeof(int), IndexNo, &Y, &Actlen, &SdsStatus);
    SdsFreeId(Id, &SdsStatus);

    Shaddowed = FALSE;
    sscanf(argv[3], "%d %d %d %d", &XP, &YP, &XS, &YS);
    
    /*  Check for the special case where there is no guide pivot selected */
    
    if ((XP != 0) || (YP != 0) || (XS != 0) || (YS != 0)) {

        /*  Are we under the shaddow of the pivot circle? Are we closer
         *  to the pivot point than the circle radius, in other words?
         */
  
        Dx = (double) abs(X - XP);
        Dy = (double) abs(Y - YP);
        DRadius = (double) GuideProbeOvalRadius;
        if (((Dx * Dx) + (Dy * Dy)) <= (DRadius * DRadius)) {
            Shaddowed = TRUE;
        } else {

            /*  Now, we split the polygon describing the rest of the shaddow
             *  into a series of lines and see if the line joining our target
             *  to the point at the center of the shaddow crosses any of
             *  them. If it does, the target is outside the shaddow. (This
             *  isn't a valid general-purpose way of doing this sort of 
             *  'inside' analysis - we're making use of some knowledge about
             *  the symmetry and regularity of the shaddow.)
             */

            double Theta;
            double DeltaX, DeltaY;
            double XCent;
            double XCentPos, YCentPos;
            int VIndex;
            int Vertex;
            int Outside;

            DeltaX = XS - XP;
            DeltaY = YS - YP;
            Theta = atan((double) abs(DeltaY) / (double) abs(DeltaX));
            if ((DeltaX > 0) && (DeltaY < 0))
                Theta = Theta;
            if ((DeltaX > 0) && (DeltaY > 0))
                Theta = (Pi * 2.0) - Theta;
            if ((DeltaX < 0) && (DeltaY > 0))
                Theta = Pi + Theta;
            if ((DeltaX < 0) && (DeltaY < 0))
                Theta = Pi - Theta;

            XCent = GuideProbeLength / 2;
            XCentPos = (cos(Theta) * XCent) + XP;
            YCentPos = -(sin(Theta) * XCent) + YP;
            VIndex = 0;
            Outside = FALSE;
            for (Vertex = 0; Vertex < (GuideProbeVertices - 1); Vertex++) {
                int X1, Y1, X2, Y2;
                double X1Pos, Y1Pos, X2Pos, Y2Pos;
                X1 = GuideProbeVertexValues[VIndex++];
                Y1 = GuideProbeVertexValues[VIndex++];
                X2 = GuideProbeVertexValues[VIndex];
                Y2 = GuideProbeVertexValues[VIndex + 1];
                X1Pos = (cos(Theta) * X1) + (sin(Theta) * Y1) + XP;
                Y1Pos = (cos(Theta) * Y1) - (sin(Theta) * X1) + YP;
                X2Pos = (cos(Theta) * X2) + (sin(Theta) * Y2) + XP;
                Y2Pos = (cos(Theta) * Y2) - (sin(Theta) * X2) + YP;
                if (LinesIntersect((double) X, (double) Y, XCentPos, YCentPos,
                        X1Pos, Y1Pos, X2Pos, Y2Pos)) {
                    Outside = TRUE;
                    break;
                }
            }
            Shaddowed = !Outside;
        }
    }
    
    /*  We make sure that the target type flag reflects its current
     *  shaddowed state - shaddowed or not, as the case may be - and we
     *  set our return value accordingly.
     */
     
    SdsFind(TopId, "type", &Id, &SdsStatus);
    SdsGet(Id, sizeof(char), IndexNo, &Type, &Actlen, &SdsStatus);
    if (Shaddowed) {
        Tcl_SetResult(interp, "shaddowed", TCL_STATIC);
        Type = FpilDisabledTargetType(*Instrument, Type);
    } else {
        Tcl_SetResult(interp, "OK", TCL_STATIC);
        Type = FpilEnabledTargetType(*Instrument, Type);
    }
    SdsPut(Id, sizeof(char), IndexNo, &Type, &SdsStatus);
    SdsFreeId(Id, &SdsStatus);
    
    return (TCL_OK);
}

/*  ------------------------------------------------------------------------- */

/*
 *                   W a v e l e n g t h  C o m b o s  { }
 */

static int ConfWavelengthCombos(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc DUNUSED, char *argv[]DUNUSED)

/*+
 *  Command Name:
 *    WavelengthCombos
 *
 *  Function:
 *    Returns the set of possible wavelength combinations.
 *
 *  Call:
 *    WavelengthCombos
 *
 *  Description:
 *    Returns a list giving the names of the the possible wavelength 
 *    combinations supported by the instrument given the current fibre
 *    combination. The current fibre combination should have been set
 *    using SetFibreCombo{}.
 *
 *  Parameters: none
 *
 *   Support:
 *     Keith Shortridge, AAO
 *
 *   Version Date:
 *     25-Mar-2002
 *-
 */
{

    int NumCombos;
    char* ComboNames[16];
    register unsigned i;
    Tcl_DString Names;

    if (Instrument == 0) {
        sprintf(interp->result, "%s:Configure not yet initialised", argv[0]);
        return TCL_ERROR;
    }

    FpilWavelengthCombos(*Instrument, sizeof(ComboNames)/sizeof(char*),
                                                   &NumCombos, ComboNames);

    Tcl_DStringInit(&Names);
    for (i = 0; i < NumCombos; ++i) {
        Tcl_DStringAppendElement(&Names, ComboNames[i]);
    }
    Tcl_DStringResult(interp, &Names);
    Tcl_DStringFree(&Names);

    return TCL_OK;
}

/*  ------------------------------------------------------------------------- */

/*
 *                    S e t  O b s  W a v e l e n g t h  { }
 */
 
static int ConfSetObsWavelength(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc, char *argv[])

/*+
 *  Command Name:
 *    SetObsWavelength
 *
 *  Function:
 *    Sets the current observing wavelength, for astrometric purposes.
 *
 *  Call:
 *    SetObsWavelength Wavelength
 *
 *  Description:
 *    This routine sets the current observing wavelength. This will be
 *    used by the Fpil routines that return the wavelength to be used
 *    by different target types for different wavelength combinations.
 *    This routine provides a direct Tcl interface to FpilSetObsWavelength().
 *
 *  Parameters:
 *    (>)  Wavelength  (double)  The main observing wavelength.
 *
 *   Support:
 *     Keith Shortridge, AAO
 *
 *   Version Date:
 *     21st Mar 2002.  KS  Original version.
 *     11th Oct 2003.  CR  Handle optional 2nd Wavelength  
 *-
 */
{
    double Wavelength;
       
    if ((argc != 2) && (argc != 3)) {
        interp->result = "wrong # args";
        return TCL_ERROR;
    }
    Wavelength = atof(argv[1]);
    FpilSetObsWavelength(*Instrument,Wavelength);

    if (argc == 3) {
        Wavelength = atof(argv[2]);
        FpilSetObsWavelength2(*Instrument,Wavelength);
    }
    return TCL_OK;
}

/*  ------------------------------------------------------------------------- */

/*
 *                    S e t  F i b r e  C o m b o  { }
 */
 
static int ConfSetFibreCombo(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc, char *argv[])

/*+
 *  Command Name:
 *    SetFibreCombo
 *
 *  Function:
 *    Sets the current fibre combination, for astrometric purposes.
 *
 *  Call:
 *    SetFibreCombo FibreComboId 
 *
 *  Description:
 *    This routine sets the current fibre combination. This will be
 *    used by the Fpil routines that return the wavelength to be used
 *    by different target types for different wavelength combinations.
 *    (It may not seem obvious that the fibre combination is relevant
 *    here, but in fact which combination of fibres is in use affects
 *    the number of possible wavelength combinations - more wavelength
 *    combinations are possible if there are fibres in use that are fed
 *    to more than one spectrograph.) This routine provides a direct
 *    Tcl interface to FpilSetFibreCombo().
 *
 *  Parameters:
 *    (>)  FibreComboId  (int)  The fibre combination id.
 *
 *   Support:
 *     Keith Shortridge, AAO
 *
 *   Version Date:
 *     21st Mar 2002.  KS  Original version.  
 *-
 */
{
    int FibreComboId;
       
    if (argc != 2) {
        interp->result = "wrong # args";
        return TCL_ERROR;
    }
    FibreComboId = atoi(argv[1]);
    FpilSetFibreCombo(*Instrument,FibreComboId);
    return TCL_OK;
}

/*  ------------------------------------------------------------------------- */

/*
 *               S e t  W a v e l e n g t h  C o m b o  { }
 */
 
static int ConfSetWavelengthCombo(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc, char *argv[])

/*+
 *  Command Name:
 *    SetWavelengthCombo
 *
 *  Function:
 *    Sets the current wavelength combination, for astrometric purposes.
 *
 *  Call:
 *    SetWavelengthCombo WavelengthComboId 
 *
 *  Description:
 *    This routine sets the current fibre combination. This will be
 *    used by the Fpil routines that return the wavelength to be used
 *    by different target types for different wavelength combinations.
 *    This routine provides a direct Tcl interface to FpilSetWavelengthCombo().
 *
 *  Parameters:
 *    (>)  WavelengthComboId  (int)  The wavelength combination id.
 *
 *   Support:
 *     Keith Shortridge, AAO
 *
 *   Version Date:
 *     21st Mar 2002.  KS  Original version.  
 *-
 */
{
    int WavelengthComboId;
       
    if (argc != 2) {
        interp->result = "wrong # args";
        return TCL_ERROR;
    }
    WavelengthComboId = atoi(argv[1]);
    FpilSetWavelengthCombo(*Instrument,WavelengthComboId);
    return TCL_OK;
}

/*  ------------------------------------------------------------------------- */

/*                       A r g u s   R o u t i n e s
 *
 *   ARGUS is part of the FLAMES instrument, but can be treated as a general
 *   facility for the purposes of this code - for any instrument other than
 *   FLAMES, we simply set it as not in use.  It is specified, for our purposes
 *   here, by three quantities: an 'in use' flag; a position angle in degrees;
 *   and a scale which is one of '1:1' or '1:1.67' and which we maintain as
 *   a flag set for 1:1. The only real interraction with this code, other
 *   than the need to maintain the values and make them available to the Tcl
 *   layer, is that when the position angle is changed, we need to invoke
 *   the FpilSetPositionAngle() routine.
 *
 *   The ARGUS values are maintained by the C layer, which keeps them in the
 *   three global variables defined here. They may be set by the Tcl layer,
 *   as the result of an ARGUS settings dialogue, or by the C layer when
 *   a file is read in. So we provide both Tcl and C interfaces to these
 *   three global variables.
 *
 *   Note that the variable here called ArgusAngle is the value that is 
 *   used as the first Telescope model parameter for FLAMES (see JAB's
 *   document on the OzPoz Astrometric Model). This is the what the
 *   astrometric code calls the position angle for the observation. 
 *   This angle, which we call Aa ('a' for astrometric) is defined as
 *   follows:  (Sorry this is so long, but it took a while to get this
 *   straight during commissioning, and it's worth recording the details.)
 *
 *   The positioner is an R-theta system, with theta being zero for the line
 *   from the center of the plate to retractor 1, which defines the +ve X-axis
 *   of the plate, and 90 degrees for the line from the center of the plate
 *   to to (roughly) retractor 68, which is more or less the +ve Y-axis of the
 *   plate. 
 *   
 *   The job of the telescope software is to use the rotator so that a 
 *   specified angle on the plate aligns with North, and to keep it
 *   that way throughout the exposure. Loosely, we think of a point on the
 *   circumfrence of the plate being made to point North. A position 
 *   angle is passed to the telescope control software, which we call Ap ('p'
 *   for PAF, because this is the angle written into the PAF file as
 *   INS.ARGS.ANGLE. If this is zero, a specific point on the plate circumfrence
 *   will be made to point North. Call this point the 'zero position angle
 *   reference point'. In general, this point will be made to point Ap degrees
 *   from North, in the sense NESW. The original assumption when the astrometric
 *   code was written was that this point would correspond to the +ve Y-axis of
 *   the plate. So the angle that we call Aa here (sometimes called ypa in the
 *   astrometric model documentation, on the assumption that it would be the
 *   position angle of the y-axis) was originally defined so that when it was
 *   zero, the +ve Y axis (~fibre 64) would point North, and when it was 90
 *   degrees the +ve Y axis would point East.
 *   
 *   However, because of considerations to do with providing the maximum
 *   possible duration for an observation, the final choice of the zero
 *   position angle reference point was the line at theta = 45 degrees,
 *   roughly corresponding to fibre 34. This choice was implemented in the
 *   astrometric code not by changing the definition of Aa, which might
 *   have been more logical (but was a more complex software change) but
 *   by treating this as a field rotation of 45 degrees. The effect is
 *   that Aa becomes the position angle (the angle from North, in the sense
 *   NESW) of the line joining the field center to the home position of fibre
 *   34. (More strictly, the zero position angle reference point.)
 *   
 *   This is actually quite sensible. The result is that both Aa and Ap are
 *   the same, and correspond to the position angle of the line from the 
 *   center of the field to the zero position angle reference point (roughly
 *   fibre 34). The fact that this point does not correspond to the natural
 *   axes of the positioner, or to the coordinate system used by FPOSS, is
 *   handled by the introduction of a field rotation of 45 degrees when
 *   converting into plate coordinates.
 *   
 *   Now, when the Observer specifies that she wants ARGUS to be at a 
 *   position angle Ao (the value she enters in the GUI), she wants
 *   the long axis of ARGUS to point Ao degrees from North, in the sense NESW.
 *   Or, putting it the other way round, if the long axis of ARGUS points Ao 
 *   degrees from North in the sense NESW, the observer will say that the
 *   Argus position angle is Ao. ARGUS itself is mounted at an essentially
 *   arbitrary angle on the field plate, however, and we have to allow for
 *   that. 
 *   
 *   Define Am to be the angle from fibre 34 (the zero position reference
 *   point) to the long axis of ARGUS in the sense NESW. Now, given this, 
 *   it's easy. When the system applies a position angle Ap, fibre 34 will
 *   point Ap degrees from North, and the long axis of ARGUS will point a
 *   further Am degrees from North.  So the angle that Argus points
 *   from North in the sense NESW, which in the previous paragraph IS the
 *   ARGUS position angle, Ao, is given by:
 *   
 *      Ao = Ap + Am
 *   
 *   The Tcl code used by FPOSS calculates (this was written some time before
 *   all the details of the field rotation etc were known, just on the
 *   assumption that there would be some sort of offset involved)
 *    
 *      Ao = Aa - ArgusOffset
 *      
 *   From which it follows that the value defined here for ArgusOffset should
 *   be -Am, with Am defined as above. As it happens, Am turns out to be 90
 *   degrees, hence the value included here.
 *      
 */
 
static int ArgusInUse = 0;                 /* If non-zero, Argus is in use */
static double ArgusAngle = 0.0;            /* Argus angle in degrees */
static int ArgusScale1to1 = 1;             /* If true, 1:1, if false, 1:1.67 */
static double ArgusOffset = -90.0;         /* Argus offset in degrees */

/*  ------------------------------------------------------------------------- */

/*
 *                    G e t  A r g u s  D a t a  { }
 */
 
static int ConfGetArgusData(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc, char *argv[])

/*+
 *  Command Name:
 *    GetArgusData
 *
 *  Function:
 *    Get all the information about the current ARGUS state.
 *
 *  Call:
 *    GetArgusData InUse Angle Scale Offset
 *
 *  Description:
 *    This routine returns all the information known to the system about the
 *    ARGUS settings.
 *
 *  Parameters:
 *    (<)  InUse  (Tcl int variable)     Set to 1 if in use, 0 if not.
 *    (<)  Angle  (Tcl double variable)  Set to the argus angle, in degrees.
 *    (<)  Scale  (Tcl string variable)  Set to the current scale in use - a 
 *                                       string, either '1:1' or '1:1.67'
 *    (<)  Offset (Tcl double variable)  Set to the argus offset, in degrees.
 *
 *   Support:
 *     Keith Shortridge, AAO
 *
 *   Version Date:
 *     15-Aug-2001  KS  Original version. 
 *     23-Aug-2001  KS  Added Offset.
 *-
 */
{
    char buffer[80];
    int IntAngle;

    if (argc != 5) {
        interp->result = "wrong # args";
        return TCL_ERROR;
    }
    if (ArgusInUse) {
        Tcl_SetVar(interp, argv[1], "1", 0);
    } else {
        Tcl_SetVar(interp, argv[1], "0", 0);
    }
    IntAngle = (int)ArgusAngle;
    if ((float)IntAngle == ArgusAngle) {  
       sprintf(buffer, "%d", IntAngle);
    } else {
       sprintf(buffer, "%f", ArgusAngle);
    }
    Tcl_SetVar(interp, argv[2], buffer, 0);
    if (ArgusScale1to1) {
       Tcl_SetVar(interp, argv[3], "1:1", 0);
    } else {
       Tcl_SetVar(interp, argv[3], "1:1.67", 0);
    }
    IntAngle = (int)ArgusOffset;
    if ((float)IntAngle == ArgusOffset) {  
       sprintf(buffer, "%d", IntAngle);
    } else {
       sprintf(buffer, "%f", ArgusOffset);
    }
    Tcl_SetVar(interp, argv[4], buffer, 0);
    return TCL_OK;
}

/*  ------------------------------------------------------------------------- */

/*
 *                    S e t  A r g u s  D a t a  { }
 */
 
static int ConfSetArgusData(ClientData clientData DUNUSED,
    Tcl_Interp * interp, int argc, char *argv[])

/*+
 *  Command Name:
 *    SetArgusData
 *
 *  Function:
 *    Sets all the information about the current ARGUS state.
 *
 *  Call:
 *    SetArgusData InUse Angle Scale Offset
 *
 *  Description:
 *    This routine sets all the information needed by the system about the
 *    ARGUS settings.
 *
 *  Parameters:
 *    (>)  InUse  (int)     1 if in use, 0 if not.
 *    (>)  Angle  (double)  The argus angle, in degrees.
 *    (>)  Scale  (string)  The current scale in use - a  string, either '1:1'
 *                          or '1:1.67'
 *    (>)  Offset (double)  The argus offset, in degrees.
 *
 *   Support:
 *     Keith Shortridge, AAO
 *
 *   Version Date:
 *     15-Aug-2001  KS  Original version.  
 *-
 */
{
    int InUse;
    double Angle;
    char* Scale;
    double Offset;
    
    if (argc != 5) {
        interp->result = "wrong # args";
        return TCL_ERROR;
    }
    InUse = atoi(argv[1]);
    Angle = atof(argv[2]);
    Scale = argv[3];
    Offset = atof(argv[4]);
    ConfCSetArgusData(InUse,Angle,Scale,Offset);
    return TCL_OK;
}

/*  ------------------------------------------------------------------------- */
 
/*+
 *                    C o n f  C  S e t  A r g u s  D a t a
 *
 *  Routine:
 *    ConfCSetArgusData
 *
 *  Function:
 *    Sets all the information about the current ARGUS state.
 *
 *  Call:
 *    ConfCSetArgusData(InUse,Angle,Scale,Offset);
 *
 *  Description:
 *    This routine provides a C interface to set all the information needed by
 *    the system about the ARGUS settings.  If the angle specified has changed,
 *    it calls FpilSetPositionAngle() to make the new angle effective. If
 *    ARGUS is not in use, then it calls FpilSetPositionAngle() to set up
 *    the default position angle.
 *
 *  Parameters:
 *    (>)  InUse  (short)   1 if in use, 0 if not.
 *    (>)  Angle  (double)  The argus angle, in degrees.
 *    (>)  Scale  (char*)   The current scale in use - a  string, either '1:1'
 *                          or '1:1.67'
 *    (>)  Offset (double)  The argus offset, in degrees.
 *
 *   Support:
 *     Keith Shortridge, AAO
 *
 *   Version Date:
 *     15-Aug-2001  KS  Original version.  
 *     23-Aug-2001  KS  Added Offset.
 *-
 */
 
void ConfCSetArgusData(
    short InUse,
    double Angle,
    char* Scale,
    double Offset)
{
    double Radians;
    short UseDefault;
    
    /*  If the angle has changed, or if we have switched between having
     *  ARGUS in use and not in use, we need to set the position angle for
     *  the system.
     */
     
    ArgusScale1to1 = (!strcmp(Scale,"1:1"));
    if (InUse) {
        if ((ArgusAngle != Angle) || !ArgusInUse) {
            Radians = Angle * 3.14159265358979323846 / 180.0;
            UseDefault = 0;
            FpilSetPositionAngle(*Instrument,UseDefault,Radians,TelModelParams);
        }
    } else {
        if (ArgusInUse) {
            UseDefault = 1;
            FpilSetPositionAngle(*Instrument,UseDefault,0.0,TelModelParams);
         }
    }
    ArgusAngle = Angle;
    ArgusInUse = InUse;
    ArgusOffset = Offset;
}

/*  ------------------------------------------------------------------------- */

/*+
 *                    C o n f  C  G e t  A r g u s  D a t a
 *
 *  Routine:
 *    ConfCGetArgusData
 *
 *  Function:
 *    Gets all the information about the current ARGUS state.
 *
 *  Call:
 *    ConfGetCArgusData(InUse,Angle,Scale,Offset);
 *
 *  Description:
 *    This routine provides a C interface to get all the information needed by
 *    the system about the ARGUS settings.
 *
 *  Parameters:
 *    (<)  InUse  (short*)   1 if in use, 0 if not.
 *    (<)  Angle  (double*)  The argus angle, in degrees.
 *    (<)  Scale  (char*)    The current scale in use - a  string, either '1:1'
 *                           or '1:1.67'  Must be at least 7 characters long.
 *    (<)  Offset (double)   The argus offset, in degrees.
 *
 *   Support:
 *     Keith Shortridge, AAO
 *
 *   Version Date:
 *     15-Aug-2001  KS  Original version.  
 *     23-Aug-2001  KS  Added Offset.
 *-
 */
 
void ConfCGetArgusData(
    short* InUse,
    double* Angle,
    char* Scale,
    double* Offset)
{    
    *InUse = ArgusInUse;
    if (ArgusScale1to1) {
        strcpy(Scale,"1:1");
    } else {
        strcpy(Scale,"1:1.67");
    }
    *Angle = ArgusAngle;
    *Offset = ArgusOffset;
}

/*  ------------------------------------------------------------------------- */

/*
 *                    C o n f i g  V e r  N u m  { }
 */
 
static int ConfigVerNum(ClientData clientdata DUNUSED, Tcl_Interp * interp,
    int argc DUNUSED, char *argv[]DUNUSED)
    
/*+
 *  Command Name:
 *    ConfigVerNum
 *
 *  Function:
 *    Returns the current version number for the configuration program.
 *
 *  Call:
 *    ConfigVerNum
 *
 *  Description:
 *    This routine returns the current version number for the configuration
 *    program. Note that this is obtained from configversion.c.
 *
 *  Parameters:  None.
 *
 *  Returns:
 *    The version number, as a string.
 *-
 */
{
    interp->result = (char *) ConfigVersion;
    return TCL_OK;
}

/*
 *                    C o n f i g  S y s t e m V e r    { }
 */
 
static int ConfigSystemVer(ClientData clientdata DUNUSED, Tcl_Interp * interp,
    int argc DUNUSED, char *argv[]DUNUSED)
    
/*+
 *  Command Name:
 *    ConfigSystemVer
 *
 *  Function:
 *    Returns the current configure system version.
 *
 *  Call:
 *    ConfigSystemVer
 *
 *  Description:
 *    This routine returns the current system version for the configuration
 *    program. Note that this is obtained from configversion.c.
 *
 *  Parameters:  None.
 *
 *  Returns:
 *    The version number, as a string.
 *-
 */
{
    interp->result = (char *) ConfigSystemVersion;
    return TCL_OK;
}

/*  ------------------------------------------------------------------------- */

/*
 *                    C o n f i g  V e r  D a t e  { }
 */
 
static int ConfigVerDate(ClientData clientdata DUNUSED, Tcl_Interp * interp,
    int argc DUNUSED, char *argv[]DUNUSED)
    
/*+
 *  Command Name:
 *    ConfigVerDate
 *
 *  Function:
 *    Returns the current version date for the configuration program.
 *
 *  Call:
 *    ConfigVerDate
 *
 *  Description:
 *    This routine returns the current version date for the configuration
 *    program. Note that this is obtained from configversion.c.
 *
 *  Parameters:  None.
 *
 *  Returns:
 *    The version date, as a string.
 *-
 */
{
    char acmmVer[50];
    static char buf[200] = "";

    if (sscanf(ConfigSystemVersion, 
               "%*s %*s %*s %s", 
               acmmVer) != 1)
    {
        strcpy(acmmVer, "Unknown");
    }

    sprintf(buf, "Version %s  %s", acmmVer, ConfigDate);
    interp->result = (char *) buf;
    return TCL_OK;
}

/*  ------------------------------------------------------------------------- */

/*                           B i t m a p s
 *
 *  This section defines a couple of specialised bitmaps used by the 
 *  Tk interface to the configuration program. They are passed to the
 *  Tcl/Tk system during execution of ConfAddCommands().
 */
 
/*  Definition of the arrow bitmap used for highlighting */

#define arrow_width 16
#define arrow_height 16
static char arrow_bits[] = {
   0x00, 0xc0, 0x00, 0xe0, 0x00, 0xf0, 0x04, 0x78, 0x0c, 0x3c, 0x0c, 0x1e,
   0x0c, 0x0f, 0x8c, 0x07, 0xcc, 0x03, 0xec, 0x01, 0xfc, 0x00, 0x7c, 0x00,
   0xfc, 0x0f, 0xfc, 0x1f, 0x00, 0x00, 0x00, 0x00};

/*  And of the tick bitmap used in the handholder pane */

#define tick_width 16
#define tick_height 16
static char tick_bits[] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x60, 0x00, 0x30, 0x00, 0x18, 0x00, 0x0c,
   0x00, 0x06, 0x00, 0x03, 0x80, 0x01, 0xc0, 0x00, 0x62, 0x00, 0x36, 0x00,
   0x1c, 0x00, 0x08, 0x00, 0x00, 0x00, 0x00, 0x00};

/*  ------------------------------------------------------------------------- */

/*+
 *                    C o n f  A d d  C o m m a n d s
 *
 *  Initialise this application into Tcl, adding all the required commands.
 *  This routine is set up by main() so that it is called during Tcl/Tk
 *  initialisation. It defines all the Tcl/Tk extensions used by the
 *  configuration program - the C-implemented additional Tcl commands, the
 *  bitmaps used, and the X-error handler.
 */
 
static int ConfAddCommands(Tcl_Interp * interp)
 {
    Tk_Window win;

    if (Tcl_Init(interp) == TCL_ERROR) {
        return TCL_ERROR;
    }
    if (!batchMode) {
        if (Tk_Init(interp) == TCL_ERROR) {
            return TCL_ERROR;
        }
        Tcl_StaticPackage(interp, "Tk", Tk_Init, Tk_SafeInit);
    }

#ifdef macintosh
    SetupMainInterp(interp);
    Tcl_SetVar(interp, "tcl_rcFileName", CONF_TCLFILE, TCL_GLOBAL_ONLY);
#endif


    Sds_Init(interp);
    Tcl_CreateCommand(interp, "GetFibre", ConfGetFibre, (ClientData *) NULL,
        (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "GetButton", ConfGetButton, (ClientData *) NULL,
        (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "GetObject", ConfGetObject, (ClientData *) NULL,
        (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "FindObjectByName", 
         ConfFindObjectByName, (ClientData *) NULL,(Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "GetObjectData", ConfGetObjectData,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "SetFibre", ConfSetFibre, (ClientData *) NULL,
        (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "SetButtonAngle", ConfSetButtonAngle,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "ConvertAscii", ConfConvertAscii,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "CountAsciiObjects", ConfCountAsciiObjects,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "SetUTMeridian", ConfSetUTMeridian,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "FixHistoricalSDS", ConfFixHistoricalSDS,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "ConvertXy", ConfConvertXy,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "ParamsChanged", ConfParamsChanged,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "Ra2string", ConfRa2string,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "Dec2string", ConfDec2string,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "Mjd2date", ConfMjd2Date,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "Mjd2time", ConfMjd2Time,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "RaDec2Ha", ConfRaDec2Ha,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "RaDec2Zd", ConfRaDec2Zd,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "MakeAllocated", ConfMakeAllocated,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "UpdateFibreStats", ConfUpdateFibreStats,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "GetFibreStats", ConfGetFibreStats,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "SetSkyCounts", ConfSetSkyCounts,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "AddIndex", ConfAddIndex,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "AllocateFibre", ConfAllocateFibre,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "LockAllocations", ConfLockAllocations,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "ClearLocking", ConfClearLocking,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "LockedFibres", ConfLockedFibres,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "FibreLocked", ConfFibreLocked,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "PivotInUse", ConfPivotInUse,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "MarkFibreType", ConfMarkFibreType,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "DeallocateFibre", ConfDeallocateFibre,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "SdsFileName", ConfFileName,
        (ClientData *) ".sds", (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "NewSdsFileName", ConfFileName,
        (ClientData *) "_new.sds", (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "AsciiFileName", ConfFileName,
        (ClientData *) ".fld", (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "ImportFileName", ConfFileName,
        (ClientData *) ".imp", (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "DSSFileName", ConfFileName,
        (ClientData *) ".dss", (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "CompressUnalloc", ConfCompressUnalloc,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "StartUpdate", ConfStartUpdate,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "DoAllocation", ConfDoAllocation,
        (ClientData *) ALLOCATE_NORMAL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "DoAllocationBatch", ConfDoAllocation,
        (ClientData *) ALLOCATE_BATCH, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "DoCheck", ConfDoCheck,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "DoCheck1", ConfDoCheck1,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "ParameterDetails", ConfParameterDetails,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "NumberParameters", ConfNumberParameters,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "AlgorithmString", ConfAlgorithmString,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "SetParameter", ConfSetParameter,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "GetParameter", ConfGetParameter,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "SetUT", ConfSetUT,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "NextDay", ConfNextDay,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "MakeListing", ConfMakeListing,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "SkycatListing", ConfSkycatListing,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "ClassifyFile", ConfClassifyFile,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "WritePAF", ConfWritePAF,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "PAFFileName", ConfPAFFileName,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "ReadPAF", ConfReadPAF,
        (ClientData *) ".ins", (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "MakeSkyvals", ConfMakeSkyvals,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "MakeDSSListing", ConfMakeDSSListing,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "ListFileName", ConfFileName,
        (ClientData *) ".lis", (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "PivotsId", ConfPivotsId,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "Initialise", ConfInitialise,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "Instruments", ConfInstruments,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "Instrument", ConfInstrument,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "NumPivots", ConfNumPivots,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "NumFields", ConfNumFields,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "FibreTypes", ConfFibreTypes,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "TargetDescr", ConfTargetDescr,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "TargetCriteria", ConfTargetCriteria,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "FibreCombos", ConfFibreCombos,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "FibreComboMode", ConfFibreComboMode,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "FibreComboTypes", ConfFibreComboTypes,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "Telescope", ConfTelescope,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "GetWavelength", ConfGetWavelength,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "SetWavelength", ConfSetWavelength,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "GetRefPars", ConfGetRefPars,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "SetRefPars", ConfSetRefPars,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "GetTelPars", ConfGetTelPars,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "SetTelPars", ConfSetTelPars,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "GetPlate", ConfGetPlate,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "SetPlate", ConfSetPlate,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "GetType", ConfGetType,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "ReallyExists", ConfReallyExists,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "AddSkyPosition", ConfAddSkyPosition,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "SetCBSPairs", ConfSetCBSPairs,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "SetSkyGrid", ConfSetSkyGrid,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "SetImportFile", ConfSetImportFile,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "SetImportOnlyMode", ConfSetImportOnlyMode,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "TranslateName", TranslateName,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "AllocToUnalloc", ConfAllocToUnalloc,
        (ClientData *) ALLOC_TO_UNALLOC, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "AllocAddToUnalloc", ConfAllocToUnalloc,
        (ClientData *) ALLOC_ADD_TO_UNALLOC, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "CheckSds", ConfCheckSds,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "AddInstrumentData", ConfAddInstrumentData,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "ExeDir", ExeDir,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "ErsFlush", ConfigErsFlush,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "MergeCheckDuplicates", ConfMergeCheckDuplicates,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "UnallocatedType", UnallocatedType,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "TypeIsTelGuide", TypeIsTelGuide,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "TypeIsGuide", TypeIsGuide,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "TypeIsSky", TypeIsSky,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "PlateXtoDispX", PlateXtoDispX,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "PlateYtoDispY", PlateYtoDispY,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "DispXtoPlateX", DispXtoPlateX,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "DispYtoPlateY", DispYtoPlateY,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "PlateXYtoRaDec", PlateXYtoRaDec,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "RaDecToPlateXY", RaDecToPlateXY,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "MeanToAppRaDec", MeanToAppRaDec,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "GetAllocEntries", ConfGetAllocEntries,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);

    Tcl_CreateCommand(interp, "ConfigVerNum", ConfigVerNum,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "ConfigSystemVer", ConfigSystemVer,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);

    Tcl_CreateCommand(interp, "ConfigVerDate", ConfigVerDate,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "UseGrey", ConfUseGrey,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "SetGrey", ConfSetGrey,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "ExpertAllowed", ConfExpertAllowed,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);

    Tcl_CreateCommand(interp, "LeakPrep", TclLeakPrep,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "LeakCheck", TclLeakCheck,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);


    Tcl_CreateCommand(interp, "FixBreakages", ConfFixBreakages,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "RemoveSky", ConfRemoveSky,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);

    Tcl_CreateCommand(interp, "Mimic", ConfMimic,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);

    Tcl_CreateCommand(interp, "CheckTargetCriteria", ConfCheckTargetCriteria,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);

    Tcl_CreateCommand(interp, "InvokeDebugger", ConfInvokeDebugger,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);


    /* VLTGuider commands  */
    
    Tcl_CreateCommand(interp, "VLTGuider", VLTGuider,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "VLTDrawGuider", VLTDrawGuider,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "VLTDrawGuiderOutline", VLTDrawGuiderOutline,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "VLTGuideShaddow", VLTGuideShaddow,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
        
    /* ARGUS commands  */
    
    Tcl_CreateCommand(interp, "SetArgusData", ConfSetArgusData,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "GetArgusData", ConfGetArgusData,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
        
    /*  Commands used by the wavelength combination stuff */
    
    Tcl_CreateCommand(interp, "WavelengthCombos", ConfWavelengthCombos,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "SetObsWavelength", ConfSetObsWavelength,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "SetWavelengthCombo", ConfSetWavelengthCombo,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
    Tcl_CreateCommand(interp, "SetFibreCombo", ConfSetFibreCombo,
        (ClientData *) NULL, (Tcl_CmdDeleteProc *) NULL);
        
    /*
     *  Setup an X error handler, and define any bitmaps that are used.
     *  Note that batchMode is a global set by main().
     */
     
    if (!batchMode) {
        win = Tk_MainWindow(interp);
#       ifndef NO_X11  /* Don't need this if we don't have X11 */
        Tk_CreateErrorHandler(Tk_Display(win), -1, -1, -1, XError,
                                                       (ClientData) interp);
#       endif
        Tk_DefineBitmap(interp, Tk_GetUid("arrow"), arrow_bits,
            arrow_width, arrow_height);
        Tk_DefineBitmap(interp, Tk_GetUid("tick"), tick_bits,
            tick_width, tick_height);
    }

    return TCL_OK;

}


/* -------------------------------------------------------------------------- */

/*
 *                   L i n e s   I n t e r s e c t
 *  Name:
 *     LinesIntersect
 *
 *  Function:
 *     Determine if two lines intersect.
 *
 *  Description:
 *     This routine is passed the end coordinates of a pair of lines,
 *     and returns true or false depending on whether or not they
 *     intersect (true if they intersect, false if they don't). Note
 *     that just touching at the ends counts as intersecting. This is
 *     a rather heavy-handed routine - I believe there are far more
 *     elegant algorithms.
 *
 */

#ifndef TRUE
#define TRUE 1
#endif

#ifndef FALSE
#define FALSE 0
#endif

static int LinesIntersect(
    double VecX1,               /* X-coordinate of start of first line */
    double VecY1,               /* Y-coordinate of start of first line */
    double VecX2,               /* X-coordinate of end of first line */
    double VecY2,               /* Y-coordinate of end of first line */
    double X1,                  /* X-coordinate of start of second line */
    double Y1,                  /* Y-coordinate of start of second line */
    double X2,                  /* X-coordinate of end of second line */
    double Y2)
{                               /* Y-coordinate of end of second line */
    /*  Note: the use of 'Vec' to refer to the first line is historical,
     *  since the two lines are really quite interchangeable, but we have
     *  to distinguish between them somehow. (This code was lifted from
     *  an experimental C++ vector class which never got finished properly.)
     */

    double DeltaX;
    double DeltaXVec;
    double MaxX, MaxY, VecMaxX, VecMaxY;
    double MinX, MinY, VecMinX, VecMinY;

    int Intersect;              /* Function return value. */
    int MayIntersect;           /* True if the line extremes overlap */

    Intersect = FALSE;
    MayIntersect = TRUE;

    /*  First, a quick check to see if there is an overlap of the
     *  extreme positions of the lines. If there isn't, they can't
     *  overlap and that's that.
     */

    MaxX = (X1 < X2) ? X2 : X1;
    MaxY = (Y1 < Y2) ? Y2 : Y1;
    VecMaxX = (VecX1 < VecX2) ? VecX2 : VecX1;
    VecMaxY = (VecY1 < VecY2) ? VecY2 : VecY1;
    MinX = (X1 > X2) ? X2 : X1;
    MinY = (Y1 > Y2) ? Y2 : Y1;
    VecMinX = (VecX1 > VecX2) ? VecX2 : VecX1;
    VecMinY = (VecY1 > VecY2) ? VecY2 : VecY1;

    if (MaxX < VecMinX)
        MayIntersect = FALSE;
    if (MinX > VecMaxX)
        MayIntersect = FALSE;
    if (MaxY < VecMinY)
        MayIntersect = FALSE;
    if (MinY > VecMaxY)
        MayIntersect = FALSE;

    if (MayIntersect) {

        /*  The extremes overlap, so the lines might. We have to do
         *  this one the hard way, solving for the individual line
         *  equations and looking for any intersection. Along the
         *  way, we have to allow for the special cases where lines
         *  are vertical, since we can't handle infinite slopes.
         */

        DeltaX = X1 - X2;
        DeltaXVec = VecX1 - VecX2;

        if ((DeltaX == 0.0) || (DeltaXVec == 0.0)) {

            /*  At least one of the vectors is vertical. There will be an
             *  intersection of the projected lines so long as both are not
             *  vertical.
             */

            if ((DeltaX == 0.0) && (DeltaXVec == 0.0)) {

                /*  Both are vertical. This is really a special case.
                 *  Since they have an overlap of extremes, they must
                 *  actually be co-linear, ie must both have the same
                 *  X value. The overlap is therefore a vector, not a
                 *  point. We return a value of true.
                 */

                Intersect = TRUE;

            } else {

                /*  Only one is vertical. The projected lines will intersect
                 *  somewhere, so work out where and see if it is in the
                 *  overlap region of the lines.
                 */

                double X;       /* X-coordinate of intersection point */
                double Y;       /* Y-coordinate of intersection point */

                if (DeltaX == 0.0) {
                    double VecSlope = (VecY2 - VecY1) / (VecX2 - VecX1);
                    double VecIntercept = VecY1 - (VecSlope * VecX1);
                    X = X1;
                    Y = (VecSlope * X) + VecIntercept;
                } else {
                    double Slope = (Y2 - Y1) / (X2 - X1);
                    double Intercept = Y1 - (Slope * X1);
                    X = VecX1;
                    Y = (Slope * X) + Intercept;
                }
                if ((Y >= MinY) && (Y <= MaxY) &&
                    (Y >= VecMinY) && (Y <= VecMaxY)) {
                    Intersect = TRUE;
                }
            }
        } else {

            /*  Neither vector is vertical, so we don't have problems
             *  calculating slopes. This is really the general case.
             */

            double VecSlope = (VecY2 - VecY1) / (VecX2 - VecX1);
            double VecIntercept = VecY1 - (VecSlope * VecX1);
            double Slope = (Y2 - Y1) / (X2 - X1);
            double Intercept = Y1 - (Slope * X1);

            /*  If the slopes are the same, we probably don't have
             *  any intersection, but we should try to check for 
             *  co-linearity.
             */

            if ((Slope == VecSlope)) {
                if ((Intercept == VecIntercept)) {

                    /*  Slopes and intercepts are the same. We know there's
                     *  an overlap, so we return true.
                     */

                    Intersect = TRUE;
                }

            } else {

                /*  Slopes are not the same, so the projected lines intersect.
                 *  Calculate where, and see if it lies in the range of both
                 *  the lines.
                 */

                double X = (Intercept - VecIntercept) / (VecSlope - Slope);
                if ((X >= MinX) && (X <= MaxX) &&
                    (X >= VecMinX) && (X <= VecMaxX)) {
                    Intersect = TRUE;
                }
            }
        }
    }

    return (Intersect);

}

/*  ------------------------------------------------------------------------- */

/*+
 *                       R C S  V e r s i o n
 *
 *  Routine:
 *    RCSVersion
 *
 *  Function:
 *    Returns the current RCS version string, if available.
 *
 *  Call:
 *    RCSVersion(Version,Length);
 *
 *  Description:
 *    This routine returns a string in the form "n.n" giving the current
 *    RCS version number for the file in which it is included.
 *
 *  Parameters:
 *    (<)  Version  (char*)  The RCS version. Allow for at least nnn.nnn, so
 *                           make this at least 8 characters long. Will be
 *                           truncated if necessary, but will always be
 *                           properly nul-terminated.
 *    (>)  Length (double)   The number of characters available at Version.
 *
 *   Support:
 *     Keith Shortridge, AAO
 *
 *   Version Date:
 *     03-Jan-2002  KS  Original version.  
 *-
 */
 
static void RCSVersion(
   char* Version,
   int Length)
{
   /*  This string is the key to this routine. This string contains an RCS
    *  keyword, and so is modified each time the version is checked in and
    *  out. So this code doesn't actually know what is going to be in this
    *  string, except that it ought to have the form $Revision: 12.7 $". If
    *  RCS is not being used, then this string will at least remain unchanged,
    *  and the routine should at least return a valid (if uninformative) string.
    */
    
   char* VersionString = "$Revision: 12.7 $";
   
   /*  Local variables */
   
   int VersionLength;
   int OutIndex = 0;
   int StartIndex;
   int Index;
   
   /*  Work through the version string, looking for the ':' that indicates
    *  the end of the keyword string
    */
   
   StartIndex = 0;
   VersionLength = strlen(VersionString);
   for (Index = 0; Index < VersionLength; Index++) {
      if (VersionString[Index] == ':') {
         StartIndex = Index + 1;
         break;
      }
   }
   
   /*  Once we have the ':', copy from there to the terminating '$', ignoring
    *  any blank characters. The break out on a nul character shouldn't
    *  ever happen, but is there just in case.
    */
    
   if (StartIndex > 0) {
      for (Index = StartIndex; Index < VersionLength; Index++) {
         char TheChar = VersionString[Index];
         if (TheChar == '$') break;
         if (TheChar != ' ') {
            Version[OutIndex++] = TheChar;
            if (OutIndex >= Length) break;
         }
      }
   }
   
   /*  Try to terminate the output string nicely. In any case, make sure
    *  it's terminated.
    */
    
   if (OutIndex < Length) Version[OutIndex] = '\0';
   Version[Length - 1] = '\0';
            
}

/*  ------------------------------------------------------------------------- */

/*
 *                           X E r r o r 
 *
 * Routine name:
 *   XError
 *  
 * Description:
 *   Handle X errors.  This routine is called by Tk when X errors occur.  
 *   We attempt to continue, after outputing messages.  The effect of 
 *   this is unclear  but is probably more errors!   It is hoped that 
 *   experience will show which errors we can actually continue from.
 *   
 *   Note that we cannot do any output to the X display within this handler,
 *   all messages must go to stderr.
 *  
 * History:
 *   30-Oct-1995 - TJF - Original version (for tdfct)
 *   04-Apr-1997 - TJF - tdFPos version.  Exit if too many errors.
 *   27-Oct-1997 - TJF - Configure version.
 *   03-Nov-1998 - TJF - Revamp for non-DRAMA configure.
 *   15-Mar-2000 - TJF - Support macintosh. 
 *   11-Jan-2005 - TJF - Don't compile this if we don't have X11.
 */
#ifndef NO_X11  /* Don't need this if we don't have X11 */
static int XError(ClientData clientData, XErrorEvent * errorEventPtr)
{
#   if !defined(WIN32) && !defined(macintosh)
    char err_msg[80];
#   endif
    Tcl_Interp *interp = (Tcl_Interp *) clientData;


    Tcl_Channel errChannel;
    errChannel = Tcl_GetStdChannel(TCL_STDERR);
    if (errChannel) {
        char buffer[100];

/*
 *      Get the details of this error message.
 */
#       if !defined(WIN32) && !defined(macintosh)
        XGetErrorText(Tk_Display(Tk_MainWindow(interp)),
            errorEventPtr->error_code, err_msg, sizeof(err_msg));
#       endif


        XErrorCount++;          /* Increment error count */
/*
 *      Output the details
 */
        Tcl_Write(errChannel,
            "+++++++++++++++++++++++++++++++++++++++++++++++++++++++\n", -1);

        sprintf(buffer,
            "Configure:X error handler hit\07 (hit number %d)\n", XErrorCount);
        Tcl_Write(errChannel, buffer, -1);

#       if !defined(WIN32) && !defined(macintosh)
        sprintf(buffer, "Error detected:\n %s\n", err_msg);
        Tcl_Write(errChannel, buffer, -1);

        sprintf(buffer, "  Protocol request: %d\n",
            errorEventPtr->request_code);
        Tcl_Write(errChannel, buffer, -1);

        sprintf(buffer, "  Resource ID:      0x%lx\n",
            errorEventPtr->resourceid);
        Tcl_Write(errChannel, buffer, -1);

        sprintf(buffer, "  Minor Op code:    %d\n", errorEventPtr->minor_code);
        Tcl_Write(errChannel, buffer, -1);

        sprintf(buffer, "  Serial Number:    %ld\n", errorEventPtr->serial);
        Tcl_Write(errChannel, buffer, -1);
#       endif

/*
 *      If we are in a Tcl procedure, output the location.  This may not
 *      achieve much due to asychronous behaviour of X windows.
 */

        Tcl_Eval(interp, "if { [info level] > 0 } { \n\
                  puts stderr \"Error location is \" \n\
                       for {set i [info level]} {$i > 0} {incr i -1} { \n\
                                puts stderr \"Level $i: [info level $i]\" \n\
                       } \n\
                  } else { puts stderr \"Error occured at global level\" } \n");

/*
 *      On second error, enable X syncrhnoize mode to help catch errors.
 */

#       if !defined(WIN32) && !defined(macintosh)
        if (XErrorCount == 2) {
            Tcl_Write(errChannel, "Setting X protocol to Synchronous\n", -1);
            XSynchronize(Tk_Display(Tk_MainWindow(interp)), 1);

        }
#       endif


        if (XErrorCount < MAX_XERRORS) {
            Tcl_Write(errChannel, "Attempting to continue\n", -1);
            Tcl_Write(errChannel,
                "+++++++++++++++++++++++++++++++++++++++++++++++++++++++\n",
                -1);
            return 0;
        } else {
            Tcl_Write(errChannel,
                "CONFIGURE:too many errors, am exiting\n", -1);
            Tcl_Eval(interp, "exit 1");

        }
    }
    return 1;
}
#endif

/*  ------------------------------------------------------------------------- */

/*                         C o n f  M e s s a g e
 *
 *  This routine is used internally and passed to ConfigInitialise() for
 *  outputing informational messages to the user during allocation.
 *
 *  We invoke the Tcl command MsgOut to output the message.
 */
 
static void ConfMessage(void *argument, const char *string)
{
    Tcl_Interp *interp = (Tcl_Interp *) argument;
    Tcl_DString command;

    Tcl_DStringInit(&command);
    Tcl_DStringAppend(&command, "MsgOut", -1);
    Tcl_DStringAppendElement(&command, string);

    if (Tcl_GlobalEval(interp, command.string) != TCL_OK) {
        char s[200];
        sprintf(s,
            "\n(Configure InfoMessage routine for string \"%s\")", string);
        Tcl_AddErrorInfo(interp, s);
        Tcl_BackgroundError(interp);
    }

}

/*  ------------------------------------------------------------------------- */

/*                      C o n f  R p t  C o n s t  F i l e
 *
 *  This callback routine is invoked when the instrument support layer
 *  opens its constants files.  Tcl user interface will not yet be running
 *  so we report it using printf().
 */
 
static void ConfRptConstFile(void *argument DUNUSED,
    const char *string, StatusType * status)
{
    if (*status != STATUS__OK)
        return;
    fprintf(stdout, "Opened constants file \"%s\"\n", string);
}

/*  ------------------------------------------------------------------------- */

/*                    C o n f  R p t  A f f e c t  F i l e
 *
 *  This callback routine is invoked when the instrument support layer
 *  opens its affectation files.  Tcl user interface will not yet be running
 *  so we report it using printf(). Note that this can receive reports of
 *  failures to open files as well as reports of files opened. This returns
 *  zero in all cases, which will keep FpilReadAffectData() reading even if
 *  if fails to find all the affectation files. 
 */
 
static short ConfRptAffectFile(void *argument DUNUSED,
    short Failed, const char *string, StatusType * status)
{
    if (*status != STATUS__OK) return 0;
    if (Failed) {
        fprintf (stdout,"Failed to open affectation file (ignored):\n");
        fprintf (stdout,"%s\n",string);
    } else {
        fprintf(stdout, "Opened affectation file \"%s\"\n", string);
    }
    return 0;
}

/*  ------------------------------------------------------------------------- */

/*                 C o n f  R p t  T e l  M o d e l  F i l e
 *
 *  This routine is invoked when the instrument support layer opens
 *  its telescope model files.  
 *
 *  If the Tcl user interface is running at this stage, the argument
 *  is a pointer to the TCL interpreter and you should use ConfMessage
 *  to report the message.
 *
 *  Otherwise, report it using printf().
 */
 
static void ConfRptTelModelFile(void *argument, const char *string,
    StatusType * status)
{
    if (*status != STATUS__OK)
        return;
    if (argument) {
        char buffer[300];
        ErsSPrintf(sizeof(buffer), buffer,
            "Opened telescope model file \"%s\"", string);
        ConfMessage(argument, buffer);
    } else
        fprintf(stdout, "Opened telescope model file \"%s\"\n", string);
}

/*  ------------------------------------------------------------------------- */

/*                           E r s  H a n d l e r
 *
 *  This routine is invoked to handle the output of Ers messages.  It is
 *  invoked by the Ers C code, being passed one or more messages that are
 *  to be output. When ErsStart() is invoked, this routine is passed to it
 *  as the reporting routine to be used.
 *  
 */
 
static void ErsHandler(void *outArg, unsigned int count,
    const ErsMessageType messages[], StatusType * status)
 {
    register unsigned int i;
    Tcl_Interp *interp = (Tcl_Interp *) outArg;

    char buffer[2000];
    int n;
    int result;
    Tcl_DString command;
    int highlight = 0, bell = 0, alarm = 0;
    buffer[0] = 0;
    if (*status != STATUS__OK)
        return;
/* 
 *      Construct the error message.
 */
    for (i = 0; i < count; i++) {
        n = 2000 - strlen(buffer) - 2;
        strncat(buffer, messages[i].message, n);
        buffer[n] = '\0';
        if (i + 1 < count)
            strcat(buffer, "\n");
        if (messages[i].flags & ERS_M_ALARM)
            alarm = 1;
        if (messages[i].flags & ERS_M_HIGHLIGHT)
            highlight = 1;
        if (messages[i].flags & ERS_M_BELL)
            bell = 1;
    }
/*
 *  Convert to a C language string.
 */
    Tcl_DStringInit(&command);
    Tcl_DStringAppend(&command, "ErsOut", -1);
    Tcl_DStringAppendElement(&command, buffer);

/*
 *  Add Ers flags if required.
 */
    if (alarm)
        Tcl_DStringAppendElement(&command, "-alarm");
    if (highlight)
        Tcl_DStringAppendElement(&command, "-highlight");
    if (bell)
        Tcl_DStringAppendElement(&command, "-bell");

/*
 *  Output the message.
 */
    result = Tcl_GlobalEval(interp, command.string);
/*
 *  IF we have a failure whist reporting an error, then we can't
 *  use the normal background error reporter routine, as it uses
 *  ErsRep and we can't use ErsRep whlist in the middle of flushing
 *  Ers messages.  We must report it ourselves.  This code is a
 *  simplification of the code in DtclBackgroundError().
 */
    if (result != TCL_OK) {
        char *targv[2];
        char *scommand;
        char *error;
        char *errorInfo;
        char *tmp;

/*
 *      Add our own error context.
 */
        Tcl_AddErrorInfo(interp, "\n(Configure Ers Handler)");

/*
 *      Copy the error message.
 */
        error = malloc((unsigned) (strlen(interp->result) + 1));
        strcpy(error, interp->result);
/*
 *      Copy the contents of errorInfo.  (Note, we are ensured there is
 *      one since we just added to it)
 */
        tmp = Tcl_GetVar(interp, "errorInfo", TCL_GLOBAL_ONLY);
        errorInfo = malloc((unsigned) (strlen(tmp) + 1));
        strcpy(errorInfo, tmp);
/*
 *      Create the command to execute and do so.
 */
        targv[0] = "tkerror";
        targv[1] = error;
        scommand = Tcl_Merge(2, targv);
        result = Tcl_GlobalEval(interp, scommand);
/*
 *      Failure in tkerror.  This is the only place where we know
 *      if the command actually existed.  If it did not exist, we just
 *      report the original error to stderr.  
 */
        if (result != TCL_OK) {
            if ((strcmp(interp->result,
                        "invalid command name: \"tkerror\"") == 0) ||
                (strcmp(interp->result,
                        "invalid command name \"tkerror\"") == 0)) 
               fprintf(stderr, "%s\n", errorInfo);
            else {

/*
 *              This indicates an error in the actual call to
 *              tkerror.  tkerror does exist.  We want to report
 *              both errors.
 */
                fprintf(stderr, "tkerror failed to handle background error.\n");
                fprintf(stderr, "    Original error: %s\n\n", errorInfo);
                fprintf(stderr, "    Error in tkerror: %s\n", tmp);
            }
        }
        Tcl_ResetResult(interp);
        free(scommand);
        free(error);
        free(errorInfo);

    }
    Tcl_DStringFree(&command);
}

/* -------------------------------------------------------------------------- */

/*                 C h e c k s u m  R o u t i n e s
 *
 *  The following set of routines have been lifted directly from the
 *  file checksum.c obtained from ftp://iraf.noao.edu/misc/checksum.
 *  They implement the 1's complement checksum scheme described for FITS 
 *  by Seaman and Pence and which was adopted by ESO as the PAF file
 *  checksum.
 *
 *  I have made the following changes to the distributed code in order for
 *  it to compile cleanly under gcc -Wall -ansi:
 *  1) checksum() is now declared as a static routine returning void. 
 *  2) char_encode() is now declared as a static routine returning void. 
 */

/* Explicitly exclude those ASCII characters that fall between the
 * upper and lower case alphanumerics (<=>?@[\]^_`) from the encoding.
 * Which is to say that only the digits 0-9, letters A-Z, and letters
 * a-r should appear in the ASCII coding for the unsigned integers.
 */
#define  NX 13
unsigned exclude[NX] = { 0x3a, 0x3b, 0x3c, 0x3d, 0x3e, 0x3f, 0x40,
          0x5b, 0x5c, 0x5d, 0x5e, 0x5f, 0x60 };

int offset = 0x30;      /* ASCII 0 (zero) character */


/* CHECKSUM -- Increment the checksum of a character array.  The
 * calling routine must zero the checksum initially.  Shorts are
 * assumed to be 16 bits, ints 32 bits.
 */

/* Internet checksum algorithm, 16/32 bit unsigned integer version:
 */
static void checksum (buf, length, sum16, sum32)
char     *buf;
int      length;
unsigned short *sum16;
unsigned int   *sum32;
{
   unsigned short *sbuf;
   int      len, remain, i;
   unsigned int   hi, lo, hicarry, locarry, tmp16;

   sbuf = (unsigned short *) buf;

   len = 2*(length / 4);   /* make sure len is even */
   remain = length % 4; /* add remaining bytes below */

   /* Extract the hi and lo words - the 1's complement checksum
    * is associative and commutative, so it can be accumulated in
    * any order subject to integer and short integer alignment.
    * By separating the odd and even short words explicitly, both
    * the 32 bit and 16 bit checksums are calculated (although the
    * latter follows directly from the former in any case) and more
    * importantly, the carry bits can be accumulated efficiently
    * (subject to short integer overflow - the buffer length should
    * be restricted to less than 2**17 = 131072).
    */
   hi = (*sum32 >> 16);
   lo = (*sum32 << 16) >> 16;

   for (i=0; i < len; i+=2) {
       hi += sbuf[i];
       lo += sbuf[i+1];
   }

   /* any remaining bytes are zero filled on the right
    */
   if (remain) {
       if (remain >= 1)
      hi += buf[2*len] * 0x100;
       if (remain >= 2)
      hi += buf[2*len+1];
       if (remain == 3)
      lo += buf[2*len+2] * 0x100;
   }

   /* fold the carried bits back into the hi and lo words
    */
   hicarry = hi >> 16;
   locarry = lo >> 16;

   while (hicarry || locarry) {
       hi = (hi & 0xFFFF) + locarry;
       lo = (lo & 0xFFFF) + hicarry;
       hicarry = hi >> 16;
       locarry = lo >> 16;
   }

   /* simply add the odd and even checksums (with carry) to get the
    * 16 bit checksum, mask the two to reconstruct the 32 bit sum
    */
   tmp16 = hi + lo;
   while (tmp16 >> 16)
       tmp16 = (tmp16 & 0xFFFF) + (tmp16 >> 16);

   *sum16 = tmp16;
   *sum32 = (hi << 16) + lo;
}


/* CHAR_ENCODE -- Encode an unsigned integer into a printable ASCII
 * string.  The input bytes are each represented by four output bytes
 * whose sum is equal to the input integer, offset by 0x30 per byte.
 * The output is restricted to alphanumerics.
 *
 * This is intended to be used to embed the complement of a file checksum
 * within an (originally 0'ed) ASCII field in the file.  The resulting
 * file checksum will then be the 1's complement -0 value (all 1's).
 * This is an additive identity value among other nifty properties.  The
 * embedded ASCII field must be 16 or 32 bit aligned, or the characters
 * can be permuted to compensate.
 *
 * To invert the encoding, simply subtract the offset from each byte
 * and pass the resulting string to checksum.
 */

static void char_encode (value, ascii, nbytes, permute)
unsigned int   value;
char     *ascii;  /* at least 17 characters long */
int      nbytes;
int      permute;
{
   int   byte, quotient, remainder, ch[4], check, i, j, k;
   char  asc[32];

   for (i=0; i < nbytes; i++) {
       byte = (value << 8*(i+4-nbytes)) >> 24;

       /* Divide each byte into 4 that are constrained to be printable
        * ASCII characters.  The four bytes will have the same initial
        * value (except for the remainder from the division), but will be
        * shifted higher and lower by pairs to avoid special characters.
        */
       quotient = byte / 4 + offset;
       remainder = byte % 4;

       for (j=0; j < 4; j++)
      ch[j] = quotient;

       /* could divide this between the bytes, but the 3 character
        * slack happens to fit within the ascii alphanumeric range
        */
       ch[0] += remainder;

       /* Any run of adjoining ASCII characters to exclude must be
        * shorter (including the remainder) than the runs of regular
        * characters on either side.
        */
       check = 1;
       while (check)
      for (check=0, k=0; k < NX; k++)
          for (j=0; j < 4; j+=2)
         if (ch[j]==exclude[k] || ch[j+1]==exclude[k]) {
             ch[j]++;
             ch[j+1]--;
             check++;
         }

       /* ascii[j*nbytes+(i+permute)%nbytes] = ch[j]; */
       for (j=0; j < 4; j++)
      asc[j*nbytes+i] = ch[j];
   }

   for (i=0; i < 4*nbytes; i++)
       ascii[i] = asc[(i+4*nbytes-permute)%(4*nbytes)];

   ascii[4*nbytes] = 0;
}

/*
         1         2         3         4         5         6         7         8
12345678901234567890123456789012345678901234567890123456789012345678901234567890
*/
